//#  OTDBconnection.cc: Manages the connection with the OTDB database.
//#
//#  Copyright (C) 2002-2004
//#  ASTRON (Netherlands Foundation for Research in Astronomy)
//#  P.O.Box 2, 7990 AA Dwingeloo, The Netherlands, softwaresupport@astron.nl
//#
//#  This program is free software; you can redistribute it and/or modify
//#  it under the terms of the GNU General Public License as published by
//#  the Free Software Foundation; either version 2 of the License, or
//#  (at your option) any later version.
//#
//#  This program is distributed in the hope that it will be useful,
//#  but WITHOUT ANY WARRANTY; without even the implied warranty of
//#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//#  GNU General Public License for more details.
//#
//#  You should have received a copy of the GNU General Public License
//#  along with this program; if not, write to the Free Software
//#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//#
//#  $Id$

//# Always #include <lofar_config.h> first!
#include <lofar_config.h>

//# Includes
#include<Common/LofarLogger.h>
#include<Common/StringUtil.h>
#include<OTDB/OTDBconnection.h>
#include<OTDB/OTDBconstants.h>

#include <boost/date_time/posix_time/posix_time.hpp>
#include <pqxx/transaction>

using namespace pqxx;
using namespace boost::posix_time;

namespace LOFAR {
  namespace OTDB {

//
// OTDBconnection(user, passwd, database)
//
// Just creates an object and registers the connection parameters.
OTDBconnection::OTDBconnection (const string&	username,
								const string&	passwd,
								const string&	database,
								const string&	hostname,
								const string&	port):
	itsUser		  (username),
	itsPassword	  (passwd),
	itsDatabase	  (database),
	itsHost		  (hostname),
	itsPort       (port),
	itsIsConnected(false),
	itsConnection (0),
	itsAuthToken  (0),
	itsError	  ("")
{ }

//
// ~OTDBconnection()
//
OTDBconnection::~OTDBconnection ()
{
	if (itsConnection) {
		delete itsConnection;
		itsConnection = 0;
	}
}

//
// disconnect()
//
// Explicitly close the connection
void OTDBconnection::disconnect()
{
	if (itsConnection) {
		itsConnection->disconnect();
	}
	itsIsConnected = false;
	delete itsConnection;
	itsConnection = 0;
}

//
// connect()
//
// To reconnect in case the connection was lost
bool OTDBconnection::connect()
{
	if (itsIsConnected) {
		return (true);
	}

	LOG_TRACE_FLOW_STR ("Trying to connect to database " << itsDatabase);

	// Note: we connect to the database as user Lofar, the real DBaccess
	// is implemented in the SP's we will call.
	string	connectString("host=" + itsHost + " port=" + itsPort +
						  " dbname=" + itsDatabase +
						  " user=" + itsUser +
                          " password=" + itsPassword);

	// try to make the connection to the database
	itsConnection = new connection(connectString);
	if (!itsConnection) {
		itsError = "Unable to connect to the database";
		return (false);
	}

	uint32		authToken;
	try {
		work 	xAction(*itsConnection, "authenticate");
		result  res = xAction.exec("SELECT OTDBlogin('paulus','boskabouter')");
		res[0][0].to(authToken);

		if (authToken == 0) {
			itsError = "Authentication failed";
			delete itsConnection;
			itsConnection = 0;
			return (false);
		}
	}
	catch (std::exception& ex) {
		itsError = string("Exception during authentication:") + ex.what();
		delete itsConnection;
		itsConnection = 0;
		return (false);
	}	
	
	// everything is Ok.
	itsIsConnected = true;
	itsAuthToken   = authToken;
	return (true);
}

//
// getTreeInfo (treeID)
//
// Get info of one specific tree
//
OTDBtree	OTDBconnection::getTreeInfo (treeIDType		aTreeID,
										 bool			isMomID)
{
	OTDBtree 	empty;
	if (!itsIsConnected && !connect()) {
		return (empty); 
	}

	LOG_TRACE_FLOW_STR ("OTDB:getTreeInfo(" << aTreeID << "," << toString(isMomID) << ")");
	
	try {
		// construct a query that calls a stored procedure.
		work	xAction(*itsConnection, "getTreeInfo");
		string	momFlag = isMomID ? "true" : "false";
		string	query("SELECT * from getTreeInfo('" + toString(aTreeID) + "','" + momFlag + "')");

		// execute query
		result	res = xAction.exec(query);

		// any records found?
		if (res.empty()) {
			return (empty); 
		}

		return (OTDBtree (res[0]));
	}
	catch (std::exception&	ex) {
		itsError = string("Exception during retrieval of TreeInfo:") + ex.what();
	}

	return (empty); 
}

//
// getTreeList(treeType, classification, groupID, processType, processSubtype, strategy): vector<OTDBtree>
//
// To get a list of the OTDB trees available in the database.
//
vector<OTDBtree> OTDBconnection::getTreeList(
					treeType 		aTreeType,
					classifType		aClassification,
					uint32			aGroupID,
					const string&	aProcessType,
					const string&	aProcessSubtype,
					const string&	aStrategy)
{
	if (!itsIsConnected && !connect()) {
		vector<OTDBtree> 	empty;
		return (empty); 
	}

	LOG_TRACE_FLOW_STR ("OTDB:getTreeList(" << aTreeType << "," << aClassification << "," << aGroupID << ",'"
						<< aProcessType << "','" << aProcessSubtype << "','" << aStrategy << "')");
	try {
		// construct a query that calls a stored procedure.
		work	xAction(*itsConnection, "getTreeList");
		string	query("SELECT * from getTreeList('" +
						toString(aTreeType) + "','" +
						toString(aClassification) + "','" +
						toString(aGroupID) + "','" +
						aProcessType + "','" + 
						aProcessSubtype + "','" + 
						aStrategy + "')");

		// execute query
		result	res = xAction.exec(query);

		// show how many records found
		result::size_type	nrRecords = res.size();
		LOG_DEBUG_STR (nrRecords << " records in treeList(" << aTreeType << "," << aClassification << "," << aGroupID << ",'"
						<< aProcessType << "','" << aProcessSubtype << "','" << aStrategy << "')");
	
		// copy information to output vector
		vector<OTDBtree>	resultVec;
		for (result::size_type i = 0; i < nrRecords; ++i) {
			resultVec.push_back(OTDBtree(res[i]));
		}

		return (resultVec);
	}
	catch (std::exception&	ex) {
		itsError = string("Exception during retrieval of TreeInfoList:") + ex.what();
	}

	vector<OTDBtree> 	empty;
	return (empty);
}

//
// getStateList(treeIDType, beginDate): vector<TreeState>
//
// To get a list of the trees-state available in the database.
//
vector<TreeState> OTDBconnection::getStateList(
								treeIDType 		aTreeID,
								bool			isMomID,
								const ptime&	beginDate,
								const ptime&	endDate)
{
	vector<TreeState>		resultVec;

	if (!itsIsConnected && !connect()) {
		vector<TreeState> 	empty;
		return (empty); 
	}

	LOG_TRACE_FLOW_STR ("OTDB:getStateList(" << aTreeID << "," 
										<< toString(isMomID) << ","
										<< to_simple_string(beginDate) << ","
										<< to_simple_string(endDate) << ")");

	try {
		// construct a query that calls a stored procedure.
		work	xAction(*itsConnection, "getStateList");
		string	momFlag = isMomID ? "true" : "false";
		string	query("SELECT * from getStateList('" +
						toString(aTreeID) + "','" +
						momFlag + "','" +
						to_simple_string(beginDate) + "','" +
						to_simple_string(endDate) + "')");
		// execute query
		result	res = xAction.exec(query);

		// show how many records found
		result::size_type	nrRecords = res.size();
		LOG_DEBUG_STR (nrRecords << " records in stateList(" 
						<< aTreeID << ", " << beginDate << ")");
	
		// copy information to output vector
		vector<TreeState>	resultVec;
		for (result::size_type i = 0; i < nrRecords; ++i) {
			resultVec.push_back(TreeState(res[i]));
		}

		return (resultVec);
	}
	catch (std::exception&	ex) {
		itsError = string("Exception during retrieval of TreeStateList:")
					 + ex.what();
	}

	vector<TreeState> 	empty;
	return (empty);
}

//
// getDefaultTemplates
//
// Get a list of all default templates
//
vector<DefaultTemplate>	OTDBconnection::getDefaultTemplates()
{
	if (!itsIsConnected && !connect()) {
		vector<DefaultTemplate> 	empty;
		return (empty); 
	}

	LOG_TRACE_FLOW("OTDB:getDefaultTemplates()");
	
	try {
		// construct a query that calls a stored procedure.
		work	xAction(*itsConnection, "getDefaultTemplates");
		string	query("SELECT * from getDefaultTemplates()");

		// execute query
		result	res = xAction.exec(query);

		// show how many records found
		result::size_type	nrRecords = res.size();
		LOG_DEBUG_STR (nrRecords << " default templates found");

		vector<DefaultTemplate>	resultVec;
		for (result::size_type i = 0; i < nrRecords; ++i) {
			resultVec.push_back(DefaultTemplate(res[i]));
		}

		return (resultVec);
	}
	catch (std::exception&	ex) {
		itsError = string("Exception during retrieval of default templates:")
					 + ex.what();
	}

	vector<DefaultTemplate>	empty;
	return (empty); 
}

//
// getExecutableTrees(classif)
//
vector<OTDBtree> OTDBconnection::getExecutableTrees(classifType aClassification)
{
	if (!itsIsConnected && !connect()) {
		vector<OTDBtree> 	empty;
		return (empty); 
	}

	LOG_TRACE_FLOW_STR ("OTDB:getExecTrees(" << aClassification << ")");
	try {
		// construct a query that calls a stored procedure.
		work	xAction(*itsConnection, "getExecutableTrees");
		string	query("SELECT * from getExecutableTrees('" +
						toString(aClassification) + "')");

		// execute query
		result	res = xAction.exec(query);

		// show how many records found
		result::size_type	nrRecords = res.size();
		LOG_DEBUG_STR (nrRecords << " records in treeList(" 
									<< aClassification << ")");
	
		// copy information to output vector
		vector<OTDBtree>	resultVec;
		for (result::size_type i = 0; i < nrRecords; ++i) {
			resultVec.push_back(OTDBtree(res[i]));
		}

		return (resultVec);
	}
	catch (std::exception&	ex) {
		itsError = string("Exception during retrieval of getExecutableTrees:")
					 + ex.what();
	}

	vector<OTDBtree> 	empty;
	return (empty);
}

//
// getTreeGroup(groupType, periodInMinutes, cluster)
//
// 0 = to be scheduled, 1 = planned, 2 = active, 3 = finished, 4 = APPROVED trees
//
// Note: this function will probably make getExecutableTrees obsolete.
//
vector<OTDBtree> OTDBconnection::getTreeGroup(uint32	groupType, uint32 period, const string& cluster)
{
	if (!itsIsConnected && !connect()) {
		vector<OTDBtree> 	empty;
		return (empty); 
	}

	LOG_TRACE_FLOW_STR ("OTDB:getTreeGroup(" << groupType << "," << period << "," << cluster << ")");
	try {
		// construct a query that calls a stored procedure.
		work	xAction(*itsConnection, "getTreeGroup");
		string	query("SELECT * from getTreeGroup('" +
						toString(groupType) + "','" +
						toString(period) + "','" + 
                        cluster + "')");

		// execute query
		result	res = xAction.exec(query);

		// show how many records found
		result::size_type	nrRecords = res.size();
		LOG_DEBUG_STR (nrRecords << " records in treeGroup(" 
									<< groupType << "," << period << ")");
	
		// copy information to output vector
		vector<OTDBtree>	resultVec;
		for (result::size_type i = 0; i < nrRecords; ++i) {
			resultVec.push_back(OTDBtree(res[i]));
		}

		return (resultVec);
	}
	catch (std::exception&	ex) {
		itsError = string("Exception during retrieval of getTreeGroup:")
					 + ex.what();
	}

	vector<OTDBtree> 	empty;
	return (empty);
}

//
// getTreesInPeriod(treeType, beginDate, endDate): vector<OTDBtree>
//
// To get a list of the OTDB trees scheduled in that period
//
vector<OTDBtree> OTDBconnection::getTreesInPeriod(
					treeType 		aTreeType,
					const ptime&	beginDate,
					const ptime&	endDate)
{
	if (!itsIsConnected && !connect()) {
		vector<OTDBtree> 	empty;
		return (empty); 
	}

	LOG_TRACE_FLOW_STR ("OTDB:getTreesInPeriod(" << aTreeType << "," 
										<< to_simple_string(beginDate) << ","
										<< to_simple_string(endDate) << ")");
	try {
		// construct a query that calls a stored procedure.
		work	xAction(*itsConnection, "getTreesInPeriod");
		string	query("SELECT * from getTreesInPeriod('" +
						toString(aTreeType) + "','" +
						to_simple_string(beginDate) + "','" +
						to_simple_string(endDate) + "')");

		// execute query
		result	res = xAction.exec(query);

		// show how many records found
		result::size_type	nrRecords = res.size();
		LOG_DEBUG_STR (nrRecords << " records in Period(" 
										<< to_simple_string(beginDate) << ","
										<< to_simple_string(endDate) << ")");
	
		// copy information to output vector
		vector<OTDBtree>	resultVec;
		for (result::size_type i = 0; i < nrRecords; ++i) {
			resultVec.push_back(OTDBtree(res[i]));
		}

		return (resultVec);
	}
	catch (std::exception&	ex) {
		itsError = string("Exception during retrieval of getTreesInPeriod:")
					 + ex.what();
	}

	vector<OTDBtree> 	empty;
	return (empty);
}

//
// getModifiedTrees(after, [treetype])
//
vector<OTDBtree> OTDBconnection::getModifiedTrees(const ptime&	after, treeType		aTreeType)
{
	if (!itsIsConnected && !connect()) {
		vector<OTDBtree> 	empty;
		return (empty); 
	}

	LOG_TRACE_FLOW_STR ("OTDB:getModifiedTrees(" << to_simple_string(after) << "," << aTreeType << "')");
	try {
		// construct a query that calls a stored procedure.
		work	xAction(*itsConnection, "getModifiedTrees");
		string	query("SELECT * from getModifiedTrees('" +
						to_simple_string(after) + "','" +
						toString(aTreeType) + "')");

		// execute query
		result	res = xAction.exec(query);

		// show how many records found
		result::size_type	nrRecords = res.size();
		LOG_DEBUG_STR (nrRecords << " records in modifiedTrees(" << to_simple_string(after) << "," << aTreeType << "')");
	
		// copy information to output vector
		vector<OTDBtree>	resultVec;
		for (result::size_type i = 0; i < nrRecords; ++i) {
			resultVec.push_back(OTDBtree(res[i]));
		}

		return (resultVec);
	}
	catch (std::exception&	ex) {
		itsError = string("Exception during retrieval of getModifiedTrees:") + ex.what();
	}

	vector<OTDBtree> 	empty;
	return (empty);
}



//
// getMomID2treeIDMap(): map[momID]
//
// Get a map to convert a MomID to a treeID
//
map<uint, uint>	OTDBconnection::getMomID2treeIDMap()
{
	map<uint, uint>	theMap;
	if (!itsIsConnected && !connect()) {
		return (theMap); 
	}

	LOG_TRACE_FLOW_STR ("OTDB:getMomID2treeIDMap()");
	try {
		// construct a query that calls a stored procedure.
		work	xAction(*itsConnection, "getMomID2treeIDMap");
		string	query("SELECT * from getMomID2treeID()");

		// execute query
		result	res = xAction.exec(query);

		// any records found?
		if (res.empty()) {
			return (theMap); 
		}
	
		// construct map
		result::size_type	nrRecords = res.size();
		for (result::size_type i = 0; i < nrRecords; ++i) {
			uint treeID;
			res[i][0].to(treeID);
			uint MomID;
			res[i][1].to(MomID);
			theMap[MomID] = treeID;
		}
	}
	catch (std::exception&	ex) {
		itsError = string("Exception during getMoID2treeID:") + ex.what();
	}

	return (theMap);
}

//
// newGroupID(): groupID
//
// To get a uniq new groupID > 0
//
uint32 OTDBconnection::newGroupID()
{
	if (!itsIsConnected && !connect()) {
		return (0); 
	}

	LOG_TRACE_FLOW_STR ("OTDB:newGroupID()");
	try {
		// construct a query that calls a stored procedure.
		work	xAction(*itsConnection, "newGroupID");
		string	query("SELECT * from newGroupID()");

		// execute query
		result	res = xAction.exec(query);

		// any records found?
		if (res.empty()) {
			return (0); 
		}

		uint32	groupID;
		res[0][0].to(groupID);
		return (groupID);
	}
	catch (std::exception&	ex) {
		itsError = string("Exception during newGroupID:") + ex.what();
	}

	return (0);
}

//
// print(ostream&): os&
//
// Show connection characteristics.

ostream& OTDBconnection::print (ostream& os) const
{
	os << itsUser << "@" << itsDatabase << ":";
	if (!itsIsConnected) {
		os << "NOT ";
	}
	return (os << "connected.");
}

  } // namespace OTDB
} // namespace LOFAR
