#!/usr/bin/env python3
#
# Copyright (C) 2023
# ASTRON (Netherlands Institute for Radio Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.
#
# $Id$
import datetime

from lofar.messaging import DEFAULT_BUSNAME, DEFAULT_BROKER, RPCService, ServiceMessageHandler
from optparse import OptionParser, OptionGroup
import os
import logging
from lofar.sas.tmss.tmss.tmssapp.conversions import Pointing, compute_elevation

logger = logging.getLogger(__name__)

CALIBRATOR_LISTS = {
    'HBA': ['3C48', '3C147', '3C196', '3C295'],
    'LBA': ['3C196', '3C380']
}


class TMSSCalibratorSelectorServiceHandler(ServiceMessageHandler):
    def __init__(self):
        super().__init__()
        self.register_service_method("GetCalibratorForTargetCoordinates", self.get_calibrator_for_target_coordinates)

    def get_calibrator_for_target_coordinates(self, calibrator_type: str, timestamp: datetime.datetime, angle1: float, angle2: float, direction_type: str='J2000', min_elevation: float=.5):
        '''get calibrator with the highest elevation at given timestamp.
        Note: this for now is a very simple initial implementation that also ignores the provided pointing.
        '''

        if direction_type != 'J2000':
            raise NotImplementedError(f'Calibrator selection is only implemented for J2000 pointings (got: direction_type={direction_type}')

        if calibrator_type not in CALIBRATOR_LISTS:
            raise ValueError(f"Unknown calibrator_type={calibrator_type}, valid options are {list(CALIBRATOR_LISTS.keys())}")

        calibrators = CALIBRATOR_LISTS[calibrator_type]
        calibrator_elevations = {c: compute_elevation(Pointing.from_name(c), timestamp, 'CS002') for c in calibrators}
        highest_calibrator = max(calibrator_elevations, key=calibrator_elevations.get)

        if calibrator_elevations[highest_calibrator] >= min_elevation:
            return {'calibrator_name': highest_calibrator, 'elevation': calibrator_elevations[highest_calibrator]}
        else:
            raise ValueError(f'No calibrator found for angle1={angle1}, angle2={angle2} direction_type={direction_type} timestamp={timestamp}, min_elevation={min_elevation}.')


def create_calibrator_selector_service(exchange: str=DEFAULT_BUSNAME, broker: str=DEFAULT_BROKER):
    return RPCService(
        service_name="TMSSCalibratorSelector.Service",
        handler_type=TMSSCalibratorSelectorServiceHandler,
        handler_kwargs={},
        broker=broker,
        exchange=exchange,
        num_threads=1
    )


def main():
    # make sure we run in UTC timezone
    os.environ['TZ'] = 'UTC'
    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

    # Check the invocation arguments
    parser = OptionParser('%prog [options]',
                          description='run the calibrator selector service that can be called to select a calibrator for given target coordinates and time')

    group = OptionGroup(parser, 'Messaging options')
    group.add_option('-b', '--broker', dest='broker', type='string', default=DEFAULT_BROKER, help='Address of the message broker, default: %default')
    group.add_option('-e', "--exchange", dest="exchange", type="string", default=DEFAULT_BUSNAME, help="exchange where the rpc communication happens. [default: %default]")
    parser.add_option_group(group)

    (options, args) = parser.parse_args()

    from lofar.common.util import waitForInterrupt

    with create_calibrator_selector_service(options.exchange, options.broker):
        waitForInterrupt()

if __name__ == '__main__':
    main()
