#!/usr/bin/env python3
import datetime
import time
# Copyright (C) 2012-2015  ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

import unittest
import uuid

import logging
logger = logging.getLogger('lofar.'+__name__)

from lofar.common.test_utils import skip_integration_tests
if skip_integration_tests():
    exit(3)

from lofar.messaging.messagebus import TemporaryExchange, BusListenerJanitor

class TestCalibratorSelectorService(unittest.TestCase):
    '''
    Tests for the TestCalibratorSelectorService
    '''
    @classmethod
    def setUpClass(cls) -> None:
        cls.TEST_UUID = str(uuid.uuid1()).split('-')[0]

        cls.tmp_exchange = TemporaryExchange("%s_%s" % (cls.__name__, cls.TEST_UUID))
        cls.tmp_exchange.open()

        # override DEFAULT_BUSNAME
        import lofar
        lofar.messaging.config.DEFAULT_BUSNAME = cls.tmp_exchange.address
        lofar.messaging.DEFAULT_BUSNAME = cls.tmp_exchange.address

        # import here, and not at top of module, because DEFAULT_BUSNAME needs to be set before importing
        from lofar.sas.tmss.test.test_environment import TMSSTestEnvironment

        cls.tmss_test_env = TMSSTestEnvironment(exchange=cls.tmp_exchange.address,
                                                populate_schemas=False, populate_test_data=False,
                                                start_subtask_scheduler=False, start_postgres_listener=False,
                                                start_dynamic_scheduler=False, enable_viewflow=False)
        cls.tmss_test_env.start()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.tmss_test_env.stop()
        cls.tmp_exchange.close()

    def test_calibrator_selector_rpc(self):
        '''
        This test starts a calibrator selector service and queries it via the RPC client
        '''

        from lofar.sas.tmss.services.calibrator_selector import create_calibrator_selector_service
        from lofar.sas.tmss.services.calibrator_selector_rpc import TMSSCalibratorSelectorRPCClient

        # create and start the service (the object under test)
        service = create_calibrator_selector_service(exchange=self.tmp_exchange.address)
        with BusListenerJanitor(service):
            # now for the real test: use RPC client to query it
            with TMSSCalibratorSelectorRPCClient.create(exchange=self.tmp_exchange.address) as calibrator_selector_client:
                response = calibrator_selector_client.get_calibrator_for_target_coordinates(
                    calibrator_type='LBA',
                    timestamp=datetime.datetime(2024,1,1,15,0,0),
                    angle1=.2,
                    angle2=.1,
                    direction_type='J2000',
                    min_elevation=.5
                )

            self.assertEqual(response['calibrator_name'], '3C380')

    def test_get_calibrator_endpoint(self):
        '''
        This test starts a calibrator selector service and tmss, then queries the get calibrator endpoint on TMSS
        which then uses the service to determine a calibrator.
        '''

        from lofar.sas.tmss.services.calibrator_selector import create_calibrator_selector_service

        # create and start the service (the object under test)
        service = create_calibrator_selector_service(exchange=self.tmp_exchange.address)
        with BusListenerJanitor(service):
            # now for the real test: query TMSS endpoint that uses the service
            with self.tmss_test_env.create_tmss_client() as tmss_client:
                response = tmss_client.get_path_as_json_object('util/calibrator/?angle1=0.1&angle2=0.2&type=HBA&timestamp=2024-01-01T09')

            self.assertEqual(response['calibrator_name'], '3C295')


if __name__ == '__main__':
    #run the unit tests
    unittest.main()
