#!/usr/bin/env python3

# Copyright (C) 2012-2015  ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

import unittest
import uuid
import os
from unittest import mock

import logging
logger = logging.getLogger('lofar.'+__name__)

from lofar.common.test_utils import skip_integration_tests
if skip_integration_tests():
    exit(3)

from lofar.messaging.messagebus import TemporaryExchange, BusListenerJanitor

from time import sleep
from datetime import datetime, timedelta

class TestCopyService(unittest.TestCase):
    '''
    Tests for the TMSSService
    '''
    @classmethod
    def setUpClass(cls) -> None:
        cls.TEST_UUID = uuid.uuid1()

        cls.tmp_exchange = TemporaryExchange("%s_%s" % (cls.__name__, cls.TEST_UUID))
        cls.tmp_exchange.open()

        # override DEFAULT_BUSNAME
        import lofar
        lofar.messaging.config.DEFAULT_BUSNAME = cls.tmp_exchange.address
        lofar.messaging.DEFAULT_BUSNAME = cls.tmp_exchange.address

        # import here, and not at top of module, because DEFAULT_BUSNAME needs to be set before importing
        from lofar.sas.tmss.test.test_environment import TMSSTestEnvironment

        cls.tmss_test_env = TMSSTestEnvironment(exchange=cls.tmp_exchange.address, populate_schemas=True, populate_test_data=False,
                                                start_subtask_scheduler=True, start_postgres_listener=True,
                                                start_dynamic_scheduler=False, enable_viewflow=False)
        cls.tmss_test_env.start()


    @classmethod
    def tearDownClass(cls) -> None:
        cls.tmss_test_env.stop()
        cls.tmp_exchange.close()

    def setUp(self):
        def mocked_wrap_command_in_cep4_head_node_ssh_call(cmd, *args, **kwargs):
            logger.info('mocked_wrap_command_in_cep4_head_node_ssh_call returning original command (without ssh): %s', cmd)
            return cmd
        wrap_command_in_cep4_head_node_ssh_call_patcher = mock.patch('lofar.sas.tmss.services.copy_service.wrap_command_in_cep4_available_node_with_lowest_load_ssh_call')
        self.addCleanup(wrap_command_in_cep4_head_node_ssh_call_patcher.stop)
        self.wrap_command_in_cep4_head_node_ssh_call_mock = wrap_command_in_cep4_head_node_ssh_call_patcher.start()
        self.wrap_command_in_cep4_head_node_ssh_call_mock.side_effect = mocked_wrap_command_in_cep4_head_node_ssh_call

        def mocked_wrap_command_in_ssh_call(cmd, *args, **kwargs):
            logger.info('wrap_command_in_ssh_call returning original command (without ssh): %s', cmd)
            return cmd

        mocked_wrap_command_in_ssh_call_patcher = mock.patch('lofar.sas.tmss.services.copy_service.wrap_command_in_ssh_call')
        self.addCleanup(mocked_wrap_command_in_ssh_call_patcher.stop)
        self.mocked_wrap_command_in_ssh_call_patcher_mock = mocked_wrap_command_in_ssh_call_patcher.start()
        self.mocked_wrap_command_in_ssh_call_patcher_mock.side_effect = mocked_wrap_command_in_ssh_call

    def test_copy_managed_and_unmanaged(self):
        from lofar.sas.tmss.tmss.tmssapp.models import SchedulingUnitObservingStrategyTemplate, SchedulingSet, SubtaskType
        from lofar.sas.tmss.tmss.tmssapp.tasks import create_scheduling_unit_draft_from_observing_strategy_template, create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft, create_cleanuptask_for_scheduling_unit_blueprint
        from lofar.sas.tmss.tmss.tmssapp.subtasks import schedule_subtask, wait_for_subtask_status
        from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions
        from lofar.sas.tmss.test.tmss_test_data_django_models import SchedulingSet_test_data

        # create and start the service (the object under test)
        from lofar.sas.tmss.services.copy_service import create_copy_service
        service = create_copy_service(exchange=self.tmp_exchange.address, tmss_client_credentials_id=self.tmss_test_env.client_credentials.dbcreds_id)
        with BusListenerJanitor(service):
            for group_by_id in [True, False]:
                for managed_output in [True, False]:
                    self.tmss_test_env.delete_scheduling_unit_drafts_cascade()
                    self.tmss_test_env.delete_reservations_cascade()

                    logger.info('\n\n\n-----   group_by_id=%s   managed_output=%s   -----', group_by_id, managed_output)
                    DESTINATION_DIR = '/tmp/t_copy_service_%s' % uuid.uuid4().hex[:8]

                    strategy_template = SchedulingUnitObservingStrategyTemplate.get_version_or_latest(name="Simple Observation")
                    strategy_template.template['tasks']['Observation']['specifications_doc']['station_configuration']['SAPs'][0]['subbands'] = [0, 1, 2, 3]
                    strategy_template.template['tasks']['CopyPipeline'] = {
                        "specifications_doc": {
                            "destination": "localhost:" + DESTINATION_DIR,
                            "group_by_id": group_by_id,
                            "managed_output": managed_output
                        },
                        "specifications_template": {
                          "name": "copy pipeline",
                          "version": 1
                        }
                    }

                    # make the copytask copy the observation output
                    strategy_template.template['task_relations'] = [{
                        "consumer": "CopyPipeline",
                        "input": {
                          "dataformat": "MeasurementSet",
                          "datatype": "visibilities",
                          "role": "any"
                        },
                        "output": {
                          "dataformat": "MeasurementSet",
                          "datatype": "visibilities",
                          "role": "correlator"
                        },
                        "producer": "Observation",
                        "selection_doc": {},
                        "selection_template": {"name": "all" }
                    }]

                    # specs complete, create SU draft and blueprint
                    scheduling_unit_draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template, scheduling_set=SchedulingSet.objects.create(**SchedulingSet_test_data()), name="name", description="description")
                    scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

                    # "run" the observation (thus creating obs-dataproducts, and triggering the copy subtask to get scheduled
                    obs_subtask = scheduling_unit_blueprint.subtasks.filter(specifications_template__type__value=SubtaskType.Choices.OBSERVATION.value).first()
                    obs_subtask = schedule_subtask(obs_subtask, datetime.utcnow()+timedelta(minutes=1))
                    self.assertGreater(obs_subtask.output_dataproducts.count(), 0)

                    try:
                        # create some output files, as if the observation ran
                        for output_dp in obs_subtask.output_dataproducts.all():
                            os.makedirs(output_dp.directory, exist_ok=True)
                            logger.info('writing 1KB test dataproduct for subtask id=%s %s', obs_subtask.id, output_dp.filepath)
                            with open(output_dp.filepath, 'w') as file:
                                file.write(1024 * 'a')

                        # "finish" the observation.
                        obs_subtask = set_subtask_state_following_allowed_transitions(obs_subtask, 'finished')

                        # wait until the copy_subtask was picked up by the service
                        copy_subtask = scheduling_unit_blueprint.subtasks.filter(specifications_template__type__value=SubtaskType.Choices.COPY.value).first()
                        copy_subtask = wait_for_subtask_status(copy_subtask, 'finished')

                        # check the output, was it copied?
                        dest_dir = os.path.join(DESTINATION_DIR, ('L' + str(obs_subtask.id)) if group_by_id else '')
                        copied_files = os.listdir(dest_dir)
                        self.assertEqual(obs_subtask.output_dataproducts.count(), len(copied_files))

                        if managed_output:
                            self.assertEqual(obs_subtask.output_dataproducts.count(), copy_subtask.output_dataproducts.count())
                            for output_dp in copy_subtask.output_dataproducts.all():
                                self.assertTrue(os.path.exists(output_dp.filepath))

                    finally:
                        # cleanup in case of accidents
                        import shutil
                        shutil.rmtree(DESTINATION_DIR, ignore_errors=True)
                        for output_dp in obs_subtask.output_dataproducts.all():
                            shutil.rmtree(output_dp.directory, ignore_errors=True)

if __name__ == '__main__':
    #run the unit tests
    unittest.main()
