#!/usr/bin/env python3

# Copyright (C) 2015
# ASTRON (Netherlands Institute for Radio Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

"""
"""

import logging
from datetime import datetime, timedelta

logger = logging.getLogger(__name__)

# warning: we import the old qpid messagebus here. We have to, because cobalt/otdb still use it.
# TODO: introduce AMQP/RabbitMQ/Kombu messagebus in cobalt, get rid of otdb, then use the new messagebus here.
from lofar.messagebus import messagebus as old_qpid_messagebus

# name of the qpid queue where the qpid feedback for dataproducts/processing is received.
QPID_DATAPRODUCT_FEEDBACK_QUEUE = "otdb.task.feedback.dataproducts.for_tmss"
QPID_PROCESSING_FEEDBACK_QUEUE = "otdb.task.feedback.processing.for_tmss"

def read_feedback_message_and_process(old_qpid_frombuses, timeout: float=1):
    try:
        if old_qpid_frombuses is None:
            return

        start_timestamp = datetime.utcnow()

        while datetime.utcnow() - start_timestamp < timedelta(seconds=timeout):
            # read both the dataproducts- and processing-buses
            for old_qpid_frombus in old_qpid_frombuses:
                try:
                    while datetime.utcnow() - start_timestamp < timedelta(seconds=timeout):
                        # get message from messagebus
                        msg = old_qpid_frombus.get(0.1)

                        if msg is None:
                            break
                        else:
                            try:
                                content = msg.content()
                                logger.info("received message from qpid queue='%s'",  old_qpid_frombus.receiver.source.address)
                            except AttributeError as e:
                                # log errors from the parsing the qpid messages, see https://support.astron.nl/jira/projects/SDCH/queues/custom/81/SDCH-4144
                                logger.exception("%s\nignoring unprocessable message", str(e))
                            finally:
                                # always ack (even when the above AttributeError occurs), so we can continue with the next message
                                old_qpid_frombus.ack(msg)

                            # note: cobalt/rtcp creates feedback and assumes that the observationID has its origin in OTDB.
                            # hence, it stores this id in the 'sasid' property of the message.
                            # We know that TMSS sets its subtask_id in the parset in the Observation.ObsID field,
                            # so we can fetch the TMSS subtask_id from the msg's sasid.
                            tmss_subtask_id = int(content.sasid)
                            feedback = content.payload

                            logger.info("received feedback for sasid=%s", tmss_subtask_id)

                            process_feedback_and_set_to_finished_if_complete(tmss_subtask_id, feedback)
                except TimeoutError:
                    pass
    except Exception as e:
        logger.exception(str(e))

def process_feedback_and_set_to_finished_if_complete(subtask_id: int, feedback: str):
    if subtask_id < 1000000:
        logger.info('not processing feedback for non-TMSS observation id=%s', subtask_id)
        return

    from lofar.sas.tmss.tmss.tmssapp.models import Subtask
    from lofar.sas.tmss.tmss.tmssapp.adapters.feedback import process_feedback_for_subtask_and_set_to_finished_if_complete

    tmss_subtask = Subtask.objects.get(id=subtask_id)
    logger.info('submitting feedback for subtask id=%s to TMSS', subtask_id)
    updated_subtask = process_feedback_for_subtask_and_set_to_finished_if_complete(subtask=tmss_subtask, feedback_doc=feedback)
    logger.info('subtask id=%s with the processed feedback has state %s', subtask_id, updated_subtask.state.value)


def main():
    # make sure we run in UTC timezone
    import os
    os.environ['TZ'] = 'UTC'

    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

    from optparse import OptionParser, OptionGroup

    # Check the invocation arguments
    parser = OptionParser('%prog [options]',
                          description='run the tmss_feedback_handling_service which relays observation feedback from the old QPID messaging system into TMSS.')

    group = OptionGroup(parser, 'QPID Messaging options',
                        description="This feedback service connects to the (old and almost obsolete) QPID broker. This can and should be replaced by connecting to RabbitMQ when Cobalt and MAC have been adapted.")
    group.add_option('-q', '--qpid_broker', dest='qpid_broker', type='string', default=old_qpid_messagebus.broker_feedback, help='Address of the QPID broker, default: %default')
    parser.add_option_group(group)

    group = OptionGroup(parser, 'TMSS Django options')
    parser.add_option_group(group)
    group.add_option('-C', '--credentials', dest='dbcredentials', type='string', default=os.environ.get('TMSS_DBCREDENTIALS', 'TMSS'), help='django dbcredentials name, default: %default')
    (options, args) = parser.parse_args()

    # Django setup routine
    from lofar.sas.tmss.tmss import setup_and_check_tmss_django_database_connection_and_exit_on_error
    setup_and_check_tmss_django_database_connection_and_exit_on_error(options.dbcredentials)

    old_qpid_frombuses = [old_qpid_messagebus.FromBus(QPID_DATAPRODUCT_FEEDBACK_QUEUE, broker=options.qpid_broker),
                          old_qpid_messagebus.FromBus(QPID_PROCESSING_FEEDBACK_QUEUE, broker=options.qpid_broker)]

    while True:
        try:
            read_feedback_message_and_process(old_qpid_frombuses, timeout=5)
        except KeyboardInterrupt:
            break

    for bus in old_qpid_frombuses:
        bus.close()


if __name__ == '__main__':
    main()
