#!/usr/bin/env python3

# Copyright (C) 2015
# ASTRON (Netherlands Institute for Radio Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.
#

from lofar.common.dbcredentials import DBCredentials
from lofar.common.util import waitForInterrupt
from lofar.lta.ingest.client.ingestbuslistener import IngestEventMessageHandler, IngestEventMesssageBusListener
from lofar.lta.ingest.common.job import *
from lofar.lta.ingest.server.config import DEFAULT_INGEST_INCOMING_JOB_SUBJECT, INGEST_NOTIFICATION_PREFIX
from lofar.messaging.messagebus import ToBus, DEFAULT_BROKER, DEFAULT_BUSNAME, UsingToBusMixin
from lofar.messaging.messages import CommandMessage, EventMessage
from lofar.messaging.rpc import ServiceMessageHandler, RPCService
from lofar.sas.tmss.client.tmssbuslistener import TMSSBusListener, TMSSEventMessageHandler, TMSS_SUBTASK_STATUS_EVENT_PREFIX

import logging
logger = logging.getLogger(__name__)

class IngestEventMessageHandlerForIngestTMSSAdapter(UsingToBusMixin, IngestEventMessageHandler):
    '''The IngestBusListenerForIngestTMSSAdapter handles the Ingest EventMessages, adapts them handles them with updates on the TMSS REST API'''
    def __init__(self):
        IngestEventMessageHandler.__init__(self, ['JobStarted', 'JobFinished', 'JobTransferFailed', 'JobRemoved'])
        UsingToBusMixin.__init__(self)

    @staticmethod
    def is_tmss_job(job_dict) -> bool:
        return job_dict.get('type', job_dict.get('Type','')).lower() == 'tmss'

    def onJobStarted(self, job_dict):
        if self.is_tmss_job(job_dict):
            subtask_id = job_dict['export_id']
            from lofar.sas.tmss.tmss.tmssapp.models import Subtask, SubtaskState
            from lofar.sas.tmss.tmss.tmssapp.subtasks import wait_for_subtask_status
            subtask = Subtask.objects.get(id=subtask_id)

            if subtask.state.value in ('started', 'finished', 'cancelled', 'error', 'on hold'):
                pass # the ingest subtask was already started
            else:
                # wait until subtask was fully queued (or in error/cancelled/finished)
                subtask = wait_for_subtask_status(subtask, ('queued', 'cancelled', 'finished', 'error'), timeout=10, poll_interval=2)

                if subtask.state.value == 'queued':
                    # the ingest subtask was fully queued, and this is the first ingest transfer job that started
                    # so, set the ingest subtask to starting->started
                    for state in ('starting', 'started'):
                        subtask.state = SubtaskState.objects.get(value=state)
                        subtask.save()

    def onJobFailed(self, job_dict):
        if self.is_tmss_job(job_dict):
            from lofar.sas.tmss.tmss.tmssapp.models import Subtask, SubtaskState
            subtask_id = job_dict['export_id']
            subtask = Subtask.objects.get(id=subtask_id)
            subtask.state = SubtaskState.objects.get(value='on hold')
            subtask.error_reason = 'One or more dataproducts could not be ingested'
            subtask.save()

    def onJobFinished(self, job_dict):
        if self.is_tmss_job(job_dict):
            subtask_id = job_dict['export_id']
            from lofar.sas.tmss.tmss.tmssapp.models import Subtask, SubtaskState, Dataproduct, DataproductArchiveInfo
            ingest_subtask = Subtask.objects.get(id=subtask_id)

            if 'input_dataproduct_id' in job_dict:
                input_dataproduct = Dataproduct.objects.get(id=job_dict['input_dataproduct_id'])
                output_dataproduct = ingest_subtask.get_transformed_output_dataproduct(input_dataproduct.id)
                if not DataproductArchiveInfo.objects.filter(dataproduct=output_dataproduct).exists():
                    try:
                        previously_recorded_archive_info = DataproductArchiveInfo.objects.get(dataproduct__producer_transforms__input_id=input_dataproduct.id)
                        previously_ingested_dataproduct = previously_recorded_archive_info.dataproduct
                        logger.info("copying archive info from previously ingested dataproduct id=%s to output_dataproduct_id=%s for ingest_subtask_id=%s",
                                     previously_ingested_dataproduct.id, output_dataproduct.id, ingest_subtask.id)
                        output_dataproduct.copy_archive_information(previously_ingested_dataproduct)
                    except DataproductArchiveInfo.DoesNotExist:
                        logger.warning("Cannot copy archive info from any other previously ingested dataproduct for input_dataproduct_id=%s output_dataproduct=%s ingest_subtask_id=%s",
                                       input_dataproduct.id, output_dataproduct.id, ingest_subtask.id)

            # if progress is 100%, set status to finished (also updating the stoptime)
            progress = ingest_subtask.progress
            if abs(1.0-progress) < 1e-5 and ingest_subtask.state.value == SubtaskState.Choices.STARTED.value:
                try:
                    from django.db import transaction
                    with transaction.atomic():
                        for state in ('finishing', 'finished'):
                            ingest_subtask.state = SubtaskState.objects.get(value=state)
                            ingest_subtask.save()
                except Exception as e:
                    logger.exception(e)

            try:
                # send progress notification messages
                # reuse part of job_dict contents for new notification message
                job_dict2 = {}
                for key in ['job_id', 'export_id', 'project', 'type', 'ingest_server', 'tmss_id', 'lta_site']:
                    if key in job_dict:
                        job_dict2[key] = job_dict[key]

                message = 'Ingest progress %.1f%% for subtask %s' % (100.0*progress, subtask_id)
                logger.info(message)
                job_dict2['message'] = message
                job_dict2['percentage_done'] = 100.0*progress

                self._send_notification('TaskProgress', job_dict2)

                # special progress case: 100% -> finished
                if abs(1.0 - progress) < 1e-5:
                    job_dict2['message'] = 'All dataproducts of subtask id=%s were ingested' % (subtask_id,)
                    self._send_notification('TaskFinished', job_dict2)
            except Exception as e:
                logger.error(str(e))

    def onJobRemoved(self, job_dict):
        if self.is_tmss_job(job_dict):
            from lofar.sas.tmss.tmss.tmssapp.models import Subtask
            from lofar.sas.tmss.tmss.tmssapp.subtasks import cancel_subtask
            subtask_id = job_dict['export_id']
            subtask = Subtask.objects.get(id=subtask_id)
            cancel_subtask(subtask)

    def _send_notification(self, subject, content_dict):
        try:
            msg = EventMessage(subject="%s.%s" % (INGEST_NOTIFICATION_PREFIX, subject), content=content_dict)
            msg.ttl = 48*3600 #remove message from queue's when not picked up within 48 hours
            logger.info('Sending notification %s to %s: %s' % (msg.subject, self.exchange, str(content_dict).replace('\n', ' ')))
            self.send(msg)
        except Exception as e:
            logger.error(str(e))


class TMSSEventMessageHandlerForIngestTMSSAdapter(UsingToBusMixin, TMSSEventMessageHandler):
    def __init__(self):
        UsingToBusMixin.__init__(self)
        TMSSEventMessageHandler.__init__(self, log_event_messages=False)

    def start_handling(self):
        UsingToBusMixin.start_handling(self)
        TMSSEventMessageHandler.start_handling(self)

        # enqueue all runnable ingest subtasks upon startup. (should be none in normal circumstances)
        # this may result in dataproducts being submitted to ingestjobmanager multiple times.
        # That's ok. Ingest deals with that.
        self.queue_all_runnable_ingest_subtasks()


    def stop_handling(self):
        TMSSEventMessageHandler.stop_handling(self)
        UsingToBusMixin.stop_handling(self)

    def onSubTaskStatusChanged(self, id: int, status: str):
        super().onSubTaskStatusChanged(id, status)

        if status in ('scheduled', 'restarting'):
            # enqueue an all dataproduct for both a scheduled or restarting ingest subtask
            # ingest will only transfer the not-transferred-yet dataproducts.
            self.queue_ingest_subtask(id)

    def queue_ingest_subtask(self, subtask_id: int):
        from django.db import transaction
        from lofar.sas.tmss.tmss.tmssapp.models import Subtask, SubtaskState, SubtaskType, DataproductArchiveInfo
        subtask = Subtask.objects.get(id=subtask_id)

        if subtask.specifications_template.type.value == SubtaskType.Choices.INGEST.value and subtask.state.value in (SubtaskState.Choices.SCHEDULED.value, SubtaskState.Choices.RESTARTING.value):
            input_dataproducts = subtask.input_dataproducts.order_by('id').all()
            logger.info("TMSS Ingest subtask id=%s was %s. Trying to enqueue %s ingest jobs for all non-ingested dataproducts...", subtask_id, subtask.state.value, len(input_dataproducts))

            with transaction.atomic():
                subtask.state = SubtaskState.objects.get(value=SubtaskState.Choices.QUEUEING.value)
                subtask.save()

                for input_dp in input_dataproducts:
                    # check if the input_dp was already ingested before (its output_dp then has archive_info)
                    output_dp = subtask.get_transformed_output_dataproduct(input_dp.id)
                    if not DataproductArchiveInfo.objects.filter(dataproduct_id=output_dp.id).exists():
                        # not ingested before, so create and submit job
                        job = createJobXml(project_name=subtask.project.name,
                                           obs_id=input_dp.producer.subtask.id, # the name 'obs_id' is somewhat misleading, but that's a legacy name 'forced' by MoM/OTDB. TODO: refactor when removing MoM/OTDB.
                                           dataproduct_name=input_dp.filename,
                                           archive_id=input_dp.global_identifier_id,
                                           location=subtask.cluster.name + ':' + input_dp.filepath,
                                           tmss_ingest_subtask_id=subtask.id,
                                           tmss_input_dataproduct_id=input_dp.id)

                        msg = CommandMessage(content=job, subject=DEFAULT_INGEST_INCOMING_JOB_SUBJECT)
                        logger.info('submitting job %s to exchange %s with subject %s at broker %s',
                                    parseJobXml(job)['JobId'], self._tobus.exchange, msg.subject, self._tobus.broker)
                        self.send(msg)

                subtask.state = SubtaskState.objects.get(value=SubtaskState.Choices.QUEUED.value)
                subtask.save()
                logger.info("Created and enqueued ingest jobs for all non-ingested dataproducts in TMSS Ingest subtask id=%s", subtask_id)

    def queue_all_runnable_ingest_subtasks(self):
        from lofar.sas.tmss.tmss.tmssapp.models import Subtask, SubtaskState
        runnable_ingest_subtasks = Subtask.objects.filter(state__value=(SubtaskState.Choices.QUEUEING.value,
                                                                        SubtaskState.Choices.SCHEDULED.value,
                                                                        SubtaskState.Choices.RESTARTING.value),
                                                          specifications_template__type__value='ingest').all()

        for ingest_subtask in runnable_ingest_subtasks:
            self.queue_ingest_subtask(ingest_subtask.id)


class IngestTMSSServiceMessageHandler(ServiceMessageHandler):
    def init_service_handler(self, service_name: str='IngestTMSSService'):
        super().init_service_handler(service_name)
        self.register_service_method('get_SIP', self.get_SIP)
        self.register_service_method('get_output_dataproduct_id', self.get_output_dataproduct_id)
        self.register_service_method('store_archive_information', self.store_archive_information)
        self.register_service_method('delete_hashes_and_archive_information', self.delete_hashes_and_archive_information)

    def get_SIP(self, dataproduct_id: int) -> str:
        from lofar.sas.tmss.tmss.tmssapp.models import Dataproduct
        dataproduct = Dataproduct.objects.get(id=dataproduct_id)
        return dataproduct.sip_xml_doc

    def get_output_dataproduct_id(self, input_dataproduct_id: int, subtask_id: int) -> int:
        from lofar.sas.tmss.tmss.tmssapp.models import Subtask
        ingest_subtask = Subtask.objects.get(id=subtask_id)
        output_dataproduct = ingest_subtask.get_transformed_output_dataproduct(input_dataproduct_id)
        return output_dataproduct.id

    def store_archive_information(self, dataproduct_id: int, size: int, filepath: str, storage_ticket: str, md5_hash: str, adler32_hash: str):
        from lofar.sas.tmss.tmss.tmssapp.models import Dataproduct
        dataproduct = Dataproduct.objects.get(id=dataproduct_id)
        dataproduct.store_archive_information(storage_ticket=storage_ticket, size=size, filepath=filepath, md5_hash=md5_hash, adler32_hash=adler32_hash)

    def delete_hashes_and_archive_information(self, dataproduct_id: int):
        from lofar.sas.tmss.tmss.tmssapp.models import Dataproduct
        dataproduct = Dataproduct.objects.get(id=dataproduct_id)
        dataproduct.delete_hashes_and_archive_information()



class IngestTMSSAdapter:
    '''The IngestTMSSAdapter connects the TMSS and Ingest system via the messagebus and Django ORM rest api.
    It has three purpouses:
      1) create and enqueue ingest jobs upon receiving an ingest-subtask scheduled event
      2) track progress of the ingest-subtask (number of dataproducts transferred) and updating the (finished) state of the ingest-subtask
      3) Provice a messagebus RPC interface for administrating dataproduct information and getting SIPs'''
    def __init__(self, exchange=DEFAULT_BUSNAME, broker=DEFAULT_BROKER):
        from lofar.common import isDevelopmentEnvironment
        self.ingest2tmss_adapter = IngestEventMesssageBusListener(handler_type=IngestEventMessageHandlerForIngestTMSSAdapter,
                                                                  exchange=exchange,
                                                                  broker=broker)
        self.tmss2ingest_adapter = TMSSBusListener(handler_type=TMSSEventMessageHandlerForIngestTMSSAdapter,
                                                   routing_key=TMSS_SUBTASK_STATUS_EVENT_PREFIX+'.#',
                                                   exchange=exchange,
                                                   broker=broker)
        self.ingest_tmss_service = RPCService('IngestTMSSService',
                                              IngestTMSSServiceMessageHandler,
                                              exchange=exchange,
                                              broker=broker,
                                              num_threads=4)

    def open(self):
        self.ingest_tmss_service.start_listening()
        self.ingest2tmss_adapter.start_listening()
        self.tmss2ingest_adapter.start_listening()


    def close(self):
        self.ingest_tmss_service.stop_listening()
        self.ingest2tmss_adapter.stop_listening()
        self.tmss2ingest_adapter.stop_listening()

    def __enter__(self):
        try:
            self.open()
        except Exception as e:
            logger.exception(e)
            self.close()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


def main():
    # make sure we run in UTC timezone
    import os
    os.environ['TZ'] = 'UTC'

    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

    from optparse import OptionParser, OptionGroup

    # Check the invocation arguments
    parser = OptionParser('%prog [options]',
                          description='run the Ingest TMSS adapter, which receives jobs from TMSS, and updates ingest statuses to TMSS')

    group = OptionGroup(parser, 'Messaging options')
    group.add_option('-b', '--broker', dest='broker', type='string', default=DEFAULT_BROKER,
                     help='Address of the message broker, default: %default')
    group.add_option('-e', "--exchange", dest="exchange", type="string", default=DEFAULT_BUSNAME,
                     help="Bus or queue where the TMSS messages are published. [default: %default]")
    parser.add_option_group(group)

    group = parser.add_option_group("TMSS Django options")
    group.add_option('-C', '--credentials', dest='dbcredentials', type='string', default=os.environ.get('TMSS_DBCREDENTIALS', 'TMSS'), help='django dbcredentials name, default: %default')

    (options, args) = parser.parse_args()

    dbcreds = DBCredentials().get(options.dbcredentials)
    logger.info("Using TMSS db creds: %s" % dbcreds.stringWithHiddenPassword())

    from lofar.sas.tmss.tmss import setup_and_check_tmss_django_database_connection_and_exit_on_error
    setup_and_check_tmss_django_database_connection_and_exit_on_error(options.dbcredentials)

    logger.info('*****************************************')
    logger.info('Starting IngestTMSSAdapter...')
    logger.info('*****************************************')

    with IngestTMSSAdapter(options.exchange, options.broker):
        waitForInterrupt()

    logger.info('Stopped IngestTMSSAdapter')

if __name__ == '__main__':
    main()


