#!/usr/bin/env python3

import unittest
from datetime import datetime

import logging
logger = logging.getLogger(__name__)

from lofar.common.test_utils import exit_with_skipped_code_if_skip_integration_tests
exit_with_skipped_code_if_skip_integration_tests()

from lofar.messaging.messagebus import TemporaryExchange
tmp_exchange = TemporaryExchange("TestIngestTMSSAdapter")
tmp_exchange.open()

# override DEFAULT_BUSNAME
import lofar
lofar.messaging.DEFAULT_BUSNAME = tmp_exchange.address
lofar.messaging.config.DEFAULT_BUSNAME = tmp_exchange.address
lofar.messaging.messagebus.DEFAULT_BUSNAME = tmp_exchange.address

# tell unittest to stop (and automagically cleanup) the test exchange once all testing is done.
def tearDownModule():
    tmp_exchange.close()


try:
    # TODO: Can we create system-integration-tests which start both the LTA- and the TMSS-dockerimage, and have them work together?
    # For now, accept that we don't have such setup. And only run this test when a developer has both LTAIngest and TMSS installed on his system
    from lofar.sas.tmss.test.test_environment import TMSSTestEnvironment
except ImportError:
    print("Cannot run test because the TMSSTestEnvironment cannot be imported. Did you run cmake with BUILD_PACKAGES for both LTAIngest and TMSS?")
    exit(3)

from lofar.common.test_utils import integration_test
from lofar.common.util import single_line_with_single_spaces
from lofar.lta.ingest.server.ingest_tmss_adapter import *
from lofar.lta.ingest.common import config as ingest_config
from uuid import uuid4

@integration_test
class TestIngestTMSSAdapter(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.tmss_test_env = TMSSTestEnvironment(exchange=tmp_exchange.address,
                                                start_postgres_listener=True,
                                                populate_schemas=True)
        cls.tmss_test_env.start()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.tmss_test_env.stop()


    def test_ingest_subtask(self):
        from lofar.sas.tmss.test.tmss_test_data_django_models import SubtaskTemplate_test_data, Subtask_test_data, \
            TaskBlueprint_test_data, TaskTemplate_test_data, Dataproduct_test_data, SubtaskOutput_test_data, SubtaskInput_test_data
        from lofar.sas.tmss.tmss.tmssapp import models
        from lofar.sas.tmss.tmss.tmssapp.subtasks import wait_for_subtask_status, schedule_subtask
        from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions

        ####################################################
        # setup: create observation and link an ingest to it.
        ####################################################

        obs_task_template = models.TaskTemplate.get_latest(name='target observation')
        obs_task = models.TaskBlueprint.objects.create(**TaskBlueprint_test_data(specifications_template=obs_task_template))
        obs_subtask_template = models.SubtaskTemplate.get_latest(name='observation control')
        obs_subtask = models.Subtask.objects.create(**Subtask_test_data(subtask_template=obs_subtask_template, task_blueprint=obs_task))
        set_subtask_state_following_allowed_transitions(obs_subtask, models.SubtaskState.Choices.DEFINED.value)
        obs_subtask_output = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=obs_subtask))

        # Create SAP
        sap_template = models.SAPTemplate.objects.get(name="SAP")
        specifications_doc = sap_template.get_default_json_document_for_schema()
        sap = models.SAP.objects.create(specifications_doc=specifications_doc, specifications_template=sap_template)
        sap.save()

        feedback_template = models.DataproductFeedbackTemplate.objects.get(name='feedback')
        feedback_doc = {'percentage_written': 100, 'frequency': {'subbands': [156], 'station_subbands': [42], 'central_frequencies': [33593750.0], 'channel_width': 6103.515625, 'channels_per_subband': 32}, 'time': {'start_time': '2013-02-16T17:00:00', 'duration': 5.02732992172, 'sample_width': 2.00278016}, 'antennas': {'set': 'HBA_DUAL', 'fields': [{'type': 'HBA', 'field': 'HBA0', 'station': 'CS001'}, {'type': 'HBA', 'field': 'HBA1', 'station': 'CS001'}]}, 'target': {'pointing': {'angle1': 0, 'angle2': 0, 'direction_type': 'J2000', "target": "target1"}}, 'samples': {'polarisations': ['XX', 'XY', 'YX', 'YY'], 'type': 'float', 'bits': 32, 'writer': 'standard', 'writer_version': '2.2.0', 'complex': True}}
        feedback_doc = feedback_template.add_defaults_to_json_object_for_schema(feedback_doc)
        obs_dataproducts = [models.Dataproduct.objects.create(**Dataproduct_test_data(producer=obs_subtask_output, sap=sap, feedback_template=feedback_template, feedback_doc=feedback_doc)) for _ in range(4)]

        ingest_task_template = models.TaskTemplate.objects.create(**TaskTemplate_test_data(task_type_value='ingest'))
        ingest_task = models.TaskBlueprint.objects.create(**TaskBlueprint_test_data(scheduling_unit_blueprint=obs_subtask.task_blueprint.scheduling_unit_blueprint,
                                                                                    specifications_template=ingest_task_template))
        ingest_subtask_template = models.SubtaskTemplate.objects.create(**SubtaskTemplate_test_data(subtask_type_value='ingest'))
        ingest_subtask = models.Subtask.objects.create(**Subtask_test_data(subtask_template=ingest_subtask_template, task_blueprint=ingest_task))
        set_subtask_state_following_allowed_transitions(ingest_subtask, models.SubtaskState.Choices.DEFINED.value)
        ingest_subtask_input = models.SubtaskInput.objects.create(**SubtaskInput_test_data(subtask=ingest_subtask, producer=obs_subtask_output))
        ingest_input_dataproducts = obs_dataproducts
        ingest_subtask_input.dataproducts.set(models.Dataproduct.objects.filter(producer=obs_subtask_output).all())
        ingest_subtask_input.save()

        # mimic running the observation
        set_subtask_state_following_allowed_transitions(obs_subtask, models.SubtaskState.Choices.FINISHED.value)

        ####################################################
        # end of object setup
        ####################################################

        with tmp_exchange.create_tobus() as test_notifier_tobus:
            # helper method to mimic ingestjobmanagementserver's notification behaviour.
            def sendIngestNotification(event, job_id, export_id):
                content = {'job_id': job_id, 'Type': "TMSS", 'export_id': export_id, 'TMSSIngestSubtaskId': export_id}
                event_msg = EventMessage(subject="%s.%s" % (ingest_config.INGEST_NOTIFICATION_PREFIX, event), content=content)
                logger.info('sending test event message on %s subject=%s content=%s', test_notifier_tobus.exchange, event_msg.subject, event_msg.content)
                test_notifier_tobus.send(event_msg)

            # create a tmp job receiver queue
            with tmp_exchange.create_temporary_queue(routing_key=DEFAULT_INGEST_INCOMING_JOB_SUBJECT) as tmp_job_queue:
                with tmp_job_queue.create_frombus() as job_receiver:
                    # create a http rest client to check some results via the rest api
                    with self.tmss_test_env.create_tmss_client() as client:

                        # start the IngestTMSSAdapter, the object-under-test
                        with IngestTMSSAdapter(exchange=tmp_exchange.address):

                            # trigger the adapter to start creating job-xml messages by scheduling the ingest subtask
                            ingest_subtask.task_blueprint.scheduling_unit_blueprint.ingest_permission_granted_since = datetime.utcnow()
                            ingest_subtask.task_blueprint.scheduling_unit_blueprint.save()
                            ingest_subtask = schedule_subtask(ingest_subtask)

                            # wait until the object-under-test (the running IngestTMSSAdapter) has picked up and queued the subtask
                            self.assertEqual('queued', wait_for_subtask_status(ingest_subtask, 'queued').state.value)

                            # there should now be a job per dataproduct on the tmp_job_queue. receive and check it.
                            # and keep track of received jobs for later test usage
                            input_pd2jobs = {}
                            for dataproduct in ingest_input_dataproducts:
                                job_msg = job_receiver.receive(timeout=5)
                                self.assertIsNotNone(job_msg)
                                logger.info("received job xml: %s", single_line_with_single_spaces(job_msg.content))
                                job = parseJobXml(job_msg.content)
                                logger.info("extracted job: %s", job)
                                self.assertEqual("TMSS", job['Type'])
                                self.assertEqual(str(ingest_subtask.id), job['TMSSIngestSubtaskId'])
                                self.assertEqual(str(obs_subtask.id), job['ObservationId'])
                                self.assertTrue(job['Location'].endswith(dataproduct.filepath))
                                self.assertEqual(obs_task.scheduling_unit_blueprint.draft.scheduling_set.project.name, job['Project'])
                                input_pd2jobs[dataproduct] = job

                            # mimic the ingestjobmanagementserver and ingesttransferserver's behaviour
                            # by sending a job started and finished event, and by submitting the archive information for a 'transfered' output_dp
                            # "stall" at the final 2 dp's, and set status on hold
                            for i, input_dp in enumerate(ingest_input_dataproducts):
                                job = input_pd2jobs[input_dp]

                                # send the started event
                                sendIngestNotification('JobStarted', job['JobId'], export_id=job['job_group_id'])

                                # ingest subtask should now be 'started'
                                self.assertEqual('started', wait_for_subtask_status(ingest_subtask, 'started').state.value)
                                ingest_subtask.refresh_from_db()
                                self.assertAlmostEqual(float(i)/len(obs_dataproducts), ingest_subtask.progress, places=3)

                                output_dp = models.DataproductTransform.objects.get(input=input_dp).output

                                # mimic uploading the transfer info, the storageticket and hash (happens in the ingestpipeline after transfer)
                                storage_ticket = str(uuid4())
                                client.post_dataproduct_archive_information(output_dp.id, storage_ticket=storage_ticket, file_size=123,
                                                                            srm_url="srm://some.lta.site/project_x/%s_a1b2c3d4e5.tar" % (input_dp.filename,),
                                                                            md5_checksum="314159265359", adler32_checksum="27182818284")

                                # check SIP
                                SIP = client.get_dataproduct_SIP(output_dp.id, retry_count=10)
                                self.assertTrue(("<storageTicket>%s</storageTicket>"%(storage_ticket,)) in SIP)
                                logger.info(SIP)

                                # check progress of the ingest subtask
                                # the IngestTMSSAdapter uses the progress value to determine if the subtask should get the 'finished' state
                                self.assertAlmostEqual(float(i+1)/len(obs_dataproducts), client.get_subtask_progress(subtask_id=ingest_subtask.id)['progress'], places=3)

                                if i < len(ingest_input_dataproducts)-1:
                                    # send the finished event
                                    sendIngestNotification('JobFinished', job['JobId'], export_id=job['job_group_id'])
                                else:
                                    # send the failed event for the last job
                                    sendIngestNotification('JobFailed', job['JobId'], export_id=job['job_group_id'])

                            # ingest subtask should now be 'on hold' because the last job failed
                            self.assertEqual('on hold', wait_for_subtask_status(ingest_subtask, 'on hold', timeout=120).state.value)

                            from lofar.sas.tmss.tmss.tmssapp.subtasks import restart_on_hold_subtask
                            restart_on_hold_subtask(ingest_subtask)

                            # ingest subtask should now be 'queued' again
                            self.assertEqual('queued', wait_for_subtask_status(ingest_subtask, 'queued', timeout=120).state.value)

                            # resend the events for the last job, now started&finished not failed
                            sendIngestNotification('JobStarted', job['JobId'], export_id=job['job_group_id'])
                            # ingest subtask should now be 'started' again
                            self.assertEqual('started', wait_for_subtask_status(ingest_subtask, 'started', timeout=120).state.value)
                            sendIngestNotification('JobFinished', job['JobId'], export_id=job['job_group_id'])

                        # ingest subtask should now be 'finished'
                        self.assertEqual('finished', wait_for_subtask_status(ingest_subtask, 'finished', timeout=120).state.value)
                        self.assertAlmostEqual(1.0, client.get_subtask_progress(subtask_id=ingest_subtask.id)['progress'], places=3)


logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

if __name__ == '__main__':
    unittest.main()
