#!/usr/bin/env python3
import copy
import unittest
from unittest import mock

import logging
logger = logging.getLogger('lofar.'+__name__)

from lofar.messaging.messagebus import TemporaryExchange
from lofar.common.test_utils import integration_test
from lofar.common.util import single_line_with_single_spaces

from datetime import datetime, timedelta
from uuid import uuid4
import threading
import os

@integration_test
class TestIngestTMSSIntegration(unittest.TestCase):
    def setUp(self) -> None:
        self.TEST_DIR = '/tmp/ingest_tmss_integration_test/' + str(uuid4())

    def tearDown(self) -> None:
        import shutil
        shutil.rmtree(self.TEST_DIR, ignore_errors=True)

    def test(self):
        with TemporaryExchange("TestIngestTMSSIntegration") as tmp_exchange:
            # override DEFAULT_BUSNAME (which is used in a call from TMSS to RA to schedule)
            import lofar
            lofar.messaging.config.DEFAULT_BUSNAME = tmp_exchange.address

            from lofar.sas.tmss.test.test_environment import TMSSTestEnvironment
            with TMSSTestEnvironment(exchange=tmp_exchange.address,
                                     populate_schemas=True, start_postgres_listener=True,
                                     populate_test_data=False, enable_viewflow=False, start_dynamic_scheduler=False,
                                     start_subtask_scheduler=True, start_workflow_service=False) as tmss_test_env:
                tmss_test_env.delete_scheduling_unit_drafts_cascade()

                from lofar.sas.tmss.tmss.tmssapp import models
                from lofar.sas.tmss.test.tmss_test_data_django_models import SchedulingSet_test_data, SchedulingUnitDraft_test_data
                from lofar.sas.tmss.tmss.tmssapp.tasks import create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft, schedule_independent_subtasks_in_scheduling_unit_blueprint, create_scheduling_unit_draft_from_observing_strategy_template, update_task_graph_from_specifications_doc, update_task_blueprints_and_subtasks_graph_from_draft
                from lofar.sas.tmss.tmss.tmssapp.subtasks import wait_for_subtask_status
                from lofar.sas.tmss.test.test_environment import create_scheduling_unit_blueprint_simulator
                from lofar.common.json_utils import add_defaults_to_json_object_for_schema
                from lofar.messaging.messagebus import BusListener, BusListenerJanitor
                from lofar.common.dbcredentials import Credentials

                # create test dirs for dataproducts and jobs
                TEST_DATA_DIR = os.path.join(self.TEST_DIR, 'data')
                TEST_INGEST_JOBS_DIR = os.path.join(self.TEST_DIR, 'ingest_jobs')
                os.makedirs(TEST_DATA_DIR)
                os.makedirs(TEST_INGEST_JOBS_DIR)
                for fs in models.Filesystem.objects.all():
                    fs.directory = TEST_DATA_DIR
                    fs.save()

                # patch (mock) the LTAClient class during these tests.
                # when the ingestpipeline instantiates an LTAClient it will get the mocked class.
                # patch (mock) the convert_surl_to_turl method during these tests.
                with mock.patch('lofar.lta.ingest.common.srm.convert_surl_to_turl') as mock_convert_surl_to_turl, \
                     mock.patch('lofar.lta.ingest.server.ltaclient.LTAClient', autospec=True) as MockLTAClient:
                    ltamock = MockLTAClient.return_value

                    def stub_GetStorageTicket(project, filename, filesize, archive_id, job_id, obs_id,
                                              check_mom_id=True, id_source='MoM'):
                        result = {'primary_uri_rnd': 'srm://some.site.name:8443/some/path/data/lofar/ops/projects/%s/%s/%s' % (project, obs_id, filename),
                                  'result': 'ok', 'error': '', 'ticket': '3E0A47ED860D6339E053B316A9C3BEE2'}

                        # for the 2nd ingest job (see below), we return mock return that the dataproduct is already in the LTA,
                        # just like the normal LTA services do.
                        ingest_subtask_id = int(job_id.split('_')[1])
                        if models.Subtask.objects.get(id=ingest_subtask_id).task_blueprint.name == 'Ingest 2':
                            result['error'] = 'StorageTicket with mom ID "%s" and ID source "TMSS" already exists' % (archive_id,)
                            result['existing_ticket_id'] = '3E0A47ED860D6339E053B316A9C3BEE2'
                            result['existing_ticket_state'] = 40 # 40==successfully_ingested

                        return result

                    ltamock.GetStorageTicket.side_effect = stub_GetStorageTicket

                    mock_convert_surl_to_turl.side_effect = lambda surl: surl.replace('srm', 'gsiftp')

                    from lofar.lta.ingest.test import ltastubs
                    ltastubs.stub()

                    # end of mocking, now import ingest servers (which then use the mocked entities from above)
                    from lofar.lta.ingest.server.ingestjobmanagementserver import IngestJobManager
                    from lofar.lta.ingest.server.ingesttransferserver import IngestTransferServer
                    from lofar.lta.ingest.server.ingest_tmss_adapter import IngestTMSSAdapter

                    with IngestTMSSAdapter(exchange=tmp_exchange.address) as ingest_tmss_adapter, \
                         IngestJobManager(exchange=tmp_exchange.address, jobs_dir=TEST_INGEST_JOBS_DIR) as ingest_job_manager, \
                         IngestTransferServer(exchange=tmp_exchange.address, lta_credentials=Credentials(), mom_credentials=Credentials(), max_nr_of_parallel_jobs=1) as transfer_server:

                        # mock throttling method
                        transfer_server.enoughResourcesAvailable = lambda: True

                        # cleanup queues with janitor
                        with BusListenerJanitor(ingest_job_manager._incoming_jobs_listener), BusListenerJanitor(ingest_job_manager._ingest_event_listener), BusListenerJanitor(ingest_job_manager._ingest_service), \
                             BusListenerJanitor(ingest_tmss_adapter.ingest2tmss_adapter), BusListenerJanitor(ingest_tmss_adapter.tmss2ingest_adapter), BusListenerJanitor(transfer_server.incoming_jobs_listener):

                            scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data())
                            strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="IM HBA - 1 Beam")
                            specifications_doc_overrides = {
                                'tasks': {'Target Observation': { 'specifications_doc': { 'station_configuration': { 'SAPs': [{'subbands': [0]}]}}}}
                            }
                            scheduling_unit_draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template, scheduling_set=scheduling_set, specifications_doc_overrides=specifications_doc_overrides)
                            scheduling_unit_draft.task_drafts.filter(name__icontains='calibrator').delete()
                            scheduling_unit = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
                            scheduling_unit.output_pinned = True
                            scheduling_unit.save()

                            ingest_subtask = models.Subtask.objects.get(task_blueprint__scheduling_unit_blueprint__id=scheduling_unit.id,
                                                                        specifications_template__type__value=models.SubtaskType.Choices.INGEST.value)
                            schedule_independent_subtasks_in_scheduling_unit_blueprint(scheduling_unit, datetime.utcnow()+timedelta(minutes=5))

                            # start a simulator, forcing the scheduling_unit to "run" the observations and pipelines....
                            # and let the ingest server act on the eventmessages.
                            # as a result, the scheduling_unit should be finished at the end, and the dataproducts should be "archived" (not in the real LTA of course, because we faked the transfer)
                            stop_event = threading.Event()
                            with create_scheduling_unit_blueprint_simulator(scheduling_unit.id,
                                                                            handle_ingest=False, handle_observations=True, handle_QA=True, handle_pipelines=True, create_output_dataproducts=True,
                                                                            delay=0, duration=0, auto_grant_ingest_permission=True,
                                                                            exchange=tmp_exchange.address):

                                # wait until the observations/pipelines finished simulating, and the ingest_subtask finished
                                wait_for_subtask_status(ingest_subtask, 'finished', timeout=90)

                                # scheduling_unit (including ingest) should be ingested (not finished, cause cleanup did not run yet as we pinned the data)
                                scheduling_unit.refresh_from_db()
                                self.assertEqual("ingested", scheduling_unit.status.value)
                                ingest_subtask.refresh_from_db()
                                self.assertEqual("finished", ingest_subtask.state.value)

                                # check ingested dataproducts
                                self.assertGreater(ingest_subtask.output_dataproducts.count(), 0)
                                for output_dp in ingest_subtask.output_dataproducts.all():
                                    self.assertTrue(output_dp.filepath.startswith("srm://"))
                                    self.assertEqual(1, models.DataproductArchiveInfo.objects.filter(dataproduct__id=output_dp.id).count())

                            # SDCH-3705: suppose a user adds a 2nd ingest task, and retries to ingest the already ingested dataproducts....
                            original_spec = scheduling_unit.specifications_doc
                            updated_spec = copy.deepcopy(original_spec)
                            updated_spec['tasks']['Ingest 2'] = copy.deepcopy(updated_spec['tasks']['Ingest'])
                            for task_rel in list(updated_spec['task_relations']):
                                if task_rel['consumer'] == 'Ingest':
                                    new_task_rel = copy.deepcopy(task_rel)
                                    new_task_rel['consumer'] = 'Ingest 2'
                                    updated_spec['task_relations'].append(new_task_rel)

                            # update blueprint via draft
                            update_task_graph_from_specifications_doc(scheduling_unit_draft, updated_spec)
                            scheduling_unit = update_task_blueprints_and_subtasks_graph_from_draft(scheduling_unit)
                            ingest2_subtask = models.Subtask.objects.get(task_blueprint__scheduling_unit_blueprint__id=scheduling_unit.id,
                                                                         task_blueprint__name='Ingest 2',
                                                                         specifications_template__type__value=models.SubtaskType.Choices.INGEST.value)

                            # the new ingest subtask should still just be defined
                            self.assertEqual("defined", ingest2_subtask.state.value)
                            # and the scheduling_unit should now be "back" to 'processed' again, cause a new ingest task was introduced in the graph
                            self.assertEqual("processed", scheduling_unit.status.value)

                            # run the simulator again...
                            # should run only the 2nd ingest (sub)task
                            with create_scheduling_unit_blueprint_simulator(scheduling_unit.id,
                                                                            handle_ingest=False, handle_observations=True, handle_QA=True, handle_pipelines=True, create_output_dataproducts=True,
                                                                            delay=0, duration=0, auto_grant_ingest_permission=True,
                                                                            exchange=tmp_exchange.address):

                                # wait until the 2nd ingest finished
                                wait_for_subtask_status(ingest2_subtask, 'finished', timeout=90)

                                # scheduling_unit should be ingested (not finished, cause cleanup did not run yet as we pinned the data)
                                scheduling_unit.refresh_from_db()
                                self.assertEqual("ingested", scheduling_unit.status.value)
                                # and both ingests should be finished
                                ingest_subtask.refresh_from_db()
                                self.assertEqual("finished", ingest_subtask.state.value)
                                ingest2_subtask.refresh_from_db()
                                self.assertEqual("finished", ingest2_subtask.state.value)


logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

if __name__ == '__main__':
    unittest.main()
