#!/usr/bin/env python3
#
# Copyright (C) 2023
# ASTRON (Netherlands Institute for Radio Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.
#
# $Id$

import os
from optparse import OptionParser, OptionGroup

from lofar.common.util import waitForInterrupt
from lofar.messaging import DEFAULT_BUSNAME, DEFAULT_BROKER
from lofar.sas.tmss.client.tmss_http_rest_client import TMSSsession

from lofar.sas.tmss.services.lobster import service_handler
from lofar.sas.tmss.services.lobster import bus_listener
from lofar.sas.tmss.services.lobster.observation_pool import ObservationPool

import logging



logger = logging.getLogger(__name__)


create_service = service_handler.create_service
L2TMSSObservationControlBusListener = bus_listener.L2TMSSObservationControlBusListener

__all__ = [
    "Lobster"
    "main",
    "create_service",
    "L2TMSSObservationControlBusListener"
]

"""
Daemon controller that starts/stops TMSS-specified-and-scheduled observations on Lofar2 stations.
"""


class Lobster:
    def __init__(self,
                 exchange=DEFAULT_BUSNAME, broker=DEFAULT_BROKER,
                 tmss_client_credentials_id: str = None):
        self.observation_pool = ObservationPool()
        self.service = service_handler.create_service(exchange, broker,
                                                      observation_pool=self.observation_pool,
                                                      tmss_client_credentials_id=tmss_client_credentials_id)
        self.bus_listener = L2TMSSObservationControlBusListener(exchange, broker,
                                                                observation_pool=self.observation_pool,
                                                                tmss_client_credentials_id=tmss_client_credentials_id)

    def start(self):
        self.service.start_listening()
        self.bus_listener.start_listening()

    def stop(self):
        self.service.stop_listening()
        self.bus_listener.stop_listening()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()


def main():
    """Start the controller using the given command line options."""

    # make sure we run in UTC timezone
    os.environ['TZ'] = 'UTC'

    # Check the invocation arguments
    parser = OptionParser(
        '%prog [options]',
        description='run the Lofar2 TMSS observation controller daemon which starts observations on the Lofar2 stations given TMSS specified&scheduled observations.'
    )

    group = OptionGroup(parser, 'Messaging options')
    group.add_option('-b', '--broker', dest='broker', type='string', default=DEFAULT_BROKER, help='Address of the message broker, default: %default')
    group.add_option('-e', "--exchange", dest="exchange", type="string", default=DEFAULT_BUSNAME, help="exchange where the TMSS event messages are published. [default: %default]")
    parser.add_option_group(group)

    group = OptionGroup(parser, 'TMSS options')
    parser.add_option_group(group)
    group.add_option('-R', '--tmss_client_credentials_id', dest='tmss_client_credentials_id', type='string', default='TMSSClient', help='TMSS django REST API credentials name, default: %default')

    group = OptionGroup(parser, 'Logging options')
    parser.add_option_group(group)
    group.add_option('-d', '--debug', dest='debug', action='store_true', help='Enable debug logging.')

    (options, args) = parser.parse_args()

    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.DEBUG if options.debug else logging.INFO)

    # make sure TMSS is up and running
    TMSSsession.check_connection_and_exit_on_error(options.tmss_client_credentials_id)

    # start the service for rpc control, ServiceHandler and BusListener
    with Lobster(options.exchange, options.broker,
                 tmss_client_credentials_id=options.tmss_client_credentials_id):
        waitForInterrupt()


if __name__ == '__main__':
    main()
