#!/usr/bin/env python3
#
# Copyright (C) 2023
# ASTRON (Netherlands Institute for Radio Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.
#
# $Id$

from typing import Dict
from typing import List

from lofar_station_client.observation.multi_station_observation import (
    MultiStationObservation,
)

import logging
logger = logging.getLogger(__name__)


class ObservationPool:
    """ObservationPool for MultiStationObservation managed by id and subtask"""

    def __init__(self):
        self.pool = {}

    def create_multistationobservation(self, subtask_id: int, observation_id: int, stations: List[str],
                                       specifications: dict) -> MultiStationObservation:
        """Create a multistationobservation instance and add to pool if non-existent"""

        try:
            _ = self.get_multistationobservation(observation_id)
            raise ValueError("Observation: %s already in current observation pool!", observation_id)
        except KeyError:
            pass

        if not self.pool.get(subtask_id):
            self.pool[subtask_id] = {}

        self.pool[subtask_id][observation_id] = MultiStationObservation(specifications=specifications,
                                                                        hosts=stations)

        return self.pool[subtask_id][observation_id]

    def get_observations(self, subtask_id: int) -> Dict[int, MultiStationObservation]:
        """Get observations dict by subtask id

        :param subtask_id: The subtask id to retrieve observations for
        :raises KeyError: If subtask id not present in pool
        :return: Dictionary with observation id as key and
                 MultiStationObservation as values
        """

        return self.pool[subtask_id]

    def get_multistationobservation(self, observation_id: int) -> MultiStationObservation:
        """Get a MultiStationObservation instance by observation id

        :param observation_id: The unique id for the specific observation
        :raises KeyError: If observation id not present in pool
        :return: Instance of MultiStationObservation
        """

        for subtask in self.pool.values():
            if subtask.get(observation_id):
                return subtask[observation_id]

        raise KeyError("Observation: %s not present in pool", observation_id)

    def remove(self, observation_id: int):
        """Remove an observation

        :raises KeyError: If observation id not in current pool
        """

        subtask = None

        for _subtask, observations in self.pool.items():
            if observations.get(observation_id):
                subtask = _subtask
                break

        if subtask:
            del self.pool[subtask][observation_id]
            if len(self.pool[subtask]) is 0:
                del self.pool[subtask]
            return

        raise KeyError("Observation: %s not in current observation pool!", observation_id)
