#!/usr/bin/env python3
#
# Copyright (C) 2023
# ASTRON (Netherlands Institute for Radio Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.
#
# $Id$

from lofar.messaging import DEFAULT_BUSNAME, DEFAULT_BROKER, RPCService, ServiceMessageHandler
from lofar.sas.tmss.client.tmss_http_rest_client import TMSSsession
from lofar.sas.tmss.services.lobster.observation_pool import ObservationPool

import logging

logger = logging.getLogger(__name__)


class L2TMSSObservationControlServiceHandler(ServiceMessageHandler):
    def __init__(self, observation_pool: ObservationPool, tmss_client_credentials_id: str = None):
        super().__init__()
        self.register_service_method("AbortObservation", self.abort_observation)
        self.tmss_client = TMSSsession.create_from_dbcreds_for_ldap(tmss_client_credentials_id)
        self.observation_pool = observation_pool

    def start_handling(self):
        '''upon startup, connect to TMSS'''
        self.tmss_client.open()
        super().start_handling()

    def stop_handling(self):
        '''upon shutdown, close TMSS connection'''
        super().stop_handling()
        self.tmss_client.close()

    def abort_observation(self, observation_subtask_id: int):
        '''abort/kill the observation with observation_subtask_id if running.'''
        subtask = self.tmss_client.get_subtask(observation_subtask_id)

        if not subtask['state_value'] in ('queueing', 'queued', 'starting', 'started'):
            logger.warning("Cannot kill observation subtask id=%s because state='%s'",
                           observation_subtask_id, subtask['state_value'])
            return

        logger.info("killing observation subtask id=%s state='%s'...",
                    observation_subtask_id, subtask['state_value'])

        try:
            exceptions = []

            self.tmss_client.set_subtask_status(observation_subtask_id, "cancelling", retry_count=5)

            l2stationspecs = self.tmss_client.get_subtask_l2stationspecs(observation_subtask_id, retry_count=5)

            observation_id = l2stationspecs['stations'][0]['antenna_fields'][0][
                'observation_id'
            ]
            multi_station_observation = self.observation_pool.get_multistationobservation(observation_id)

            results = multi_station_observation.stop()
            for host, result in results.items():
                if isinstance(result, Exception):
                    exceptions.append(result)

            # TODO(Corne): Keep track of which hosts / observations did
            #              not stop correctly
            if len(exceptions) > 0:
                raise RuntimeError("Failed to abort subtask=%s with exceptions=%s", observation_subtask_id, exceptions)

            self.tmss_client.set_subtask_status(observation_subtask_id, "cancelled", retry_count=5)

        except Exception as e:
            logger.error(e)
            self.tmss_client.set_subtask_status(observation_subtask_id, "error", error_reason=str(e), retry_count=5)
            return False


def create_service(exchange=DEFAULT_BUSNAME, broker=DEFAULT_BROKER,
                   observation_pool: ObservationPool = ObservationPool(), tmss_client_credentials_id: str = None):
    return RPCService(
        service_name="L2TMSSObservationControl.Service",
        handler_type=L2TMSSObservationControlServiceHandler,
        handler_kwargs={
            "observation_pool": observation_pool,
            "tmss_client_credentials_id": tmss_client_credentials_id
        },
        broker=broker,
        exchange=exchange,
        num_threads=1
    )
