#!/usr/bin/env python3

# subtask_scheduling.py
#
# Copyright (C) 2015
# ASTRON (Netherlands Institute for Radio Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

import logging
import os
from optparse import OptionParser, OptionGroup

logger = logging.getLogger(__name__)

from lofar.sas.tmss.client.tmssbuslistener import *
from lofar.sas.tmss.client.tmss_http_rest_client import TMSSsession

class TMSSEventMessageHandlerForReportRefresh(TMSSEventMessageHandler):

    def __init__(self, tmss_client_credentials_id: str="TMSSClient"):
        super().__init__()
        self.tmss_client = TMSSsession.create_from_dbcreds_for_ldap(tmss_client_credentials_id)

    def start_handling(self):
        self.tmss_client.open()

    def stop_handling(self):
        self.tmss_client.close()

    def onSubTaskStatusChanged(self, id: int, status: str):

        # if end state has been reached, request refresh of cycle report
        if status in ['finished', 'cancelled', 'error']:
            logger.info('Subtask id=%s has reached end state=%s' % (id, status))
            from lofar.sas.tmss.tmss.tmssapp.models import Subtask
            subtask = Subtask.objects.get(id=id)
            for cycle in subtask.project.cycles.all():
                logger.info('Requesting recalculation of report for cycle=%s because related subtask=%s has reached its end state=%s' % (cycle.pk, id, status))
                self.tmss_client.get_path_as_json_object('/cycle/%s/report/?refresh=true&wipe=false&exclude_project=IDOLS,COM_LOFAR2' % cycle.pk)


def create_service(exchange: str=DEFAULT_BUSNAME, broker: str=DEFAULT_BROKER, tmss_client_credentials_id: str="TMSSClient"):
    return TMSSBusListener(handler_type=TMSSEventMessageHandlerForReportRefresh,
                           handler_kwargs={'tmss_client_credentials_id': tmss_client_credentials_id},
                           exchange=exchange, broker=broker)


def main():
    # make sure we run in UTC timezone
    os.environ['TZ'] = 'UTC'

    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

    # Check the invocation arguments
    parser = OptionParser('%prog [options]',
                          description='run the tmss_report_refresh_service which listens for TMSS event messages on the messagebus, and triggers a refresh of the cycle report on relevant changes.')

    group = OptionGroup(parser, 'Messaging options')
    group.add_option('-b', '--broker', dest='broker', type='string', default=DEFAULT_BROKER,
                     help='Address of the message broker, default: %default')
    group.add_option('-e', "--exchange", dest="exchange", type="string", default=DEFAULT_BUSNAME,
                     help="exchange where the TMSS event messages are published. [default: %default]")
    parser.add_option_group(group)

    group = OptionGroup(parser, 'Django options')
    parser.add_option_group(group)
    group.add_option('-R', '--tmss_client_credentials_id', dest='tmss_client_credentials_id', type='string', default='TMSSClient', help='TMSS django REST API credentials name, default: %default')
    group.add_option('-C', '--credentials', dest='dbcredentials', type='string', default=os.environ.get('TMSS_DBCREDENTIALS', 'TMSS'), help='django dbcredentials name, default: %default')

    (options, args) = parser.parse_args()

    # check TMSS is up and running via the client
    TMSSsession.check_connection_and_exit_on_error(options.tmss_client_credentials_id)

    # setup/check the django ORM as well
    from lofar.sas.tmss.tmss import setup_and_check_tmss_django_database_connection_and_exit_on_error
    setup_and_check_tmss_django_database_connection_and_exit_on_error(options.dbcredentials)

    with create_service(options.exchange, options.broker, options.tmss_client_credentials_id):
        waitForInterrupt()

if __name__ == '__main__':
    main()
