#!/usr/bin/env python3

# Copyright (C) 2012-2015  ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.
import time
import unittest
import uuid
from datetime import datetime

import logging
logger = logging.getLogger('lofar.' + __name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

from lofar.messaging.messagebus import TemporaryExchange, BusListenerJanitor

# create temporary queue
TEST_UUID = uuid.uuid1()
tmp_exchange = TemporaryExchange("%s_%s" % (__name__, TEST_UUID))
tmp_exchange.open()

# override DEFAULT_BUSNAME (before test env start)
import lofar
lofar.messaging.config.DEFAULT_BUSNAME = tmp_exchange.address

# start test env
from lofar.sas.tmss.test.test_environment import TMSSTestEnvironment
tmss_test_env = TMSSTestEnvironment(exchange=tmp_exchange.address, populate_test_data=False, populate_schemas=False, start_websocket=False, start_postgres_listener=True, enable_viewflow=False)
tmss_test_env.start()

# import what's needed for the test (after test env start)
from lofar.sas.tmss.services.report_refresh import create_service, TMSSEventMessageHandlerForReportRefresh
from lofar.common.test_utils import integration_test
from lofar.sas.tmss.test.tmss_test_data_django_models import *
from lofar.sas.tmss.tmss.tmssapp import models

def tearDownModule():
    tmss_test_env.stop()
    tmp_exchange.close()

@integration_test
class TestReportRefreshService(unittest.TestCase):
    '''
    Tests for the ReportRefreshService
    '''

    def test_refort_refresh_on_ending_subtask(self):
        '''
        This test starts a report refresh service. Creates and cancels a subtasl to check if this triggers the creation
        of a cycle report in the database.
        '''

        # test data setup
        project = models.Project.objects.create(**Project_test_data(name=str(uuid.uuid4())))
        cycle = models.Cycle.objects.create(**Cycle_test_data(name=str(uuid.uuid4())))
        project.cycles.set([cycle])
        project.save()
        scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project))
        scheduling_unit_template = models.SchedulingUnitTemplate.get_latest(name='scheduling unit')
        scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(**SchedulingUnitDraft_test_data(name=str(uuid.uuid4()), template=scheduling_unit_template, scheduling_set=scheduling_set))
        scheduling_unit_blueprint = models.SchedulingUnitBlueprint.objects.create(**SchedulingUnitBlueprint_test_data(name=str(uuid.uuid4()), draft=scheduling_unit_draft))
        task_blueprint = models.TaskBlueprint.objects.create(**TaskBlueprint_test_data(name=str(uuid.uuid4()), scheduling_unit_blueprint=scheduling_unit_blueprint))
        subtask_obs = models.Subtask.objects.create(**Subtask_test_data(state=models.SubtaskState.objects.get(value="defining"),
                                                                        subtask_template=models.SubtaskTemplate.get_latest(name='observation control'),
                                                                        task_blueprint=task_blueprint))
        subtask_obs.state = models.SubtaskState.objects.get(value="defined")
        subtask_obs.save()

        # create and start the service (the object under test)
        service = create_service(exchange=tmp_exchange.address, tmss_client_credentials_id=tmss_test_env.client_credentials.dbcreds_id)
        with BusListenerJanitor(service):
            # assert we did not have anything in the cache for the test cycle
            self.assertEqual(models.CycleReport.objects.filter(cycle=cycle).count(), 0)

            # trigger the service with a subtask status update
            subtask_obs.state = models.SubtaskState.objects.get(value="error")
            subtask_obs.save()

            # wait for the service to do its thing
            timeout = 30
            start = datetime.utcnow()
            while models.CycleReport.objects.filter(cycle=cycle).count() == 0:
                if datetime.utcnow() - start > timedelta(seconds=timeout):
                    raise TimeoutError("timeout while waiting for cycle report refresh to get triggered.")
                logger.info('waiting for cycle report refresh to get triggered.')
                time.sleep(0.5)
            while 'weekly_efficiency' not in models.CycleReport.objects.filter(cycle=cycle).first().data:
                if datetime.utcnow() - start > timedelta(seconds=timeout):
                    raise TimeoutError("timeout while waiting for refreshed cycle report to get created.")
                logger.info('waiting for refreshed cycle report to get created.')
                time.sleep(0.5)

            # assert we have a cache entry, so the refresh request worked
            self.assertEqual(models.CycleReport.objects.filter(cycle=cycle).count(), 1)

            # assert we have a valid report and not just a placeholder, so the actual cycle report generation worked
            self.assertIn('weekly_efficiency', models.CycleReport.objects.filter(cycle=cycle).first().data)


if __name__ == '__main__':
    #run the unit tests
    unittest.main()
