#!/usr/bin/env python3

# subtask_scheduling.py
#
# Copyright (C) 2015
# ASTRON (Netherlands Institute for Radio Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.
#
# $Id: subtask_scheduling.py 1580 2015-09-30 14:18:57Z loose $

"""
The subtask_scheduling service schedules TMSS subtasks.
It listens on the lofar notification message bus for state changes of TMSS subtasks; when a task finished,
it schedules (rest action) all successors that are in state 'defined'.
"""

from lofar.sas.tmss.client.tmssbuslistener import *
from lofar.sas.tmss.tmss.tmssapp.models import Subtask, SubtaskState, SubtaskType, SchedulingUnitBlueprint, TaskBlueprint, SchedulingUnitStatus
from lofar.sas.tmss.tmss.tmssapp.subtasks import schedule_subtask
from lofar.sas.tmss.tmss.exceptions import TMSSException
from lofar.common.datetimeutils import round_to_second_precision
from datetime import datetime, timedelta

import logging
logger = logging.getLogger(__name__)


class TMSSSubTaskSchedulingEventMessageHandler(TMSSEventMessageHandler):
    '''
    '''
    def try_schedule_subtasks(self, subtasks):
        '''try to schedule each subtask in the list'''
        for subtask in subtasks:
            try:
                # start with checking the prerequisites before we try schedule the subtask.
                # We could of course just try to schedule it and check for errors, 
                # but that pollutes the logs with ugly error messages indicating something bad happened.
                # It's much nice to present informative log messages here with reasonble explanations why a subtask cannot be scheduled.

                if subtask.is_obsolete:
                    logger.info("skipping scheduling of obsolete subtask %s", subtask.id)
                    continue

                if subtask.state.value != SubtaskState.Choices.DEFINED.value:
                    logger.info("skipping scheduling of subtask %s with state=%s", subtask.id, subtask.state)
                    continue

                predecessors = subtask.predecessors.all()
                if any([predecessor.state.value != SubtaskState.Choices.FINISHED.value and not predecessor.is_obsolete for predecessor in predecessors]):
                    logger.info("skipping scheduling of subtask %s because not all its (non-obsolete) predecessors are finished", subtask.id)
                    continue

                if subtask.specifications_template.type.value == SubtaskType.Choices.INGEST.value:
                    # check ingest permission
                    scheduling_unit_blueprint = subtask.task_blueprint.scheduling_unit_blueprint
                    if scheduling_unit_blueprint.ingest_permission_required:
                        if scheduling_unit_blueprint.ingest_permission_granted_since is None or scheduling_unit_blueprint.ingest_permission_granted_since > datetime.utcnow():
                            logger.info("skipping scheduling of ingest subtask %s project=%s which requires permission from scheduling_unit_blueprint id=%s first", subtask.id, subtask.project.name, scheduling_unit_blueprint.id)
                            continue

                if subtask.specifications_template.type.value == SubtaskType.Choices.CLEANUP.value:
                    # can we schedule this cleanup subtask already, or do we need to wait for ingest to finish and/or pinning
                    # check pinning
                    if subtask.predecessors.filter(task_blueprint__output_pinned=True).exists():
                        logger.info("skipping scheduling of cleanup subtask id=%s scheduling_unit_id=%s project=%s for which the data of to-be-cleaned-up predecessor task_blueprints is still been pinned",
                                    subtask.id, subtask.task_blueprint.scheduling_unit_blueprint.id, subtask.project.name)
                        continue

                    # check ingest status(es)
                    scheduling_unit_ingest_subtasks = Subtask.objects.filter(task_blueprint__scheduling_unit_blueprint__id=subtask.task_blueprint.scheduling_unit_blueprint.id)\
                                                                     .filter(specifications_template__type__value=SubtaskType.Choices.INGEST.value) \
                                                                     .exclude(state__value=SubtaskState.Choices.FINISHED.value) \
                                                                     .filter(obsolete_since__isnull=True).order_by('id').all()

                    if scheduling_unit_ingest_subtasks.exists():
                        logger.info("skipping scheduling of cleanup subtask id=%s because not all ingest subtasks in the same scheduling unit are finished yet", subtask.id)
                        continue

                # ok, all prerequisites have been checked.
                # try scheduling the subtask.
                # if it succeeds, then the state will be 'scheduled' afterwards
                # if there is a specification error, then the state will be 'error' afterwards
                # if there is another kind of error (like needing ingest-permission), then the state will be 'defined' afterwards, so you can retry.
                # for the ingest-permission we will retry automatically when that permission is granted
                scheduled_subtask = schedule_subtask(subtask, start_time=round_to_second_precision(datetime.utcnow()+timedelta(seconds=30)))
                logger.info("scheduled %s subtask id=%s status=%s start_time='%s'", scheduled_subtask.specifications_template.type.value, scheduled_subtask.id, scheduled_subtask.state.value, scheduled_subtask.scheduled_start_time)
            except TMSSException as e:
                # just log the error and continue with the next subtask.
                # the user should take action on "fixing" the unschedulable subtask
                logger.error(e)

    def start_handling(self):
        # upon startup of this handler, try to schedule all currently defined successor subtasks.
        # only those that meet all prerequisites will actually be scheduled.
        maybe_schedulable_subtasks = []
        for subtask in Subtask.objects.filter(state__value=SubtaskState.Choices.DEFINED.value,
                                              obsolete_since__isnull=True, inputs__isnull=False,
                                              task_blueprint__scheduling_unit_blueprint__status__value__in=SchedulingUnitStatus.OBSERVING_STATUS_OR_BEYOND_VALUES).distinct().order_by('id').all():
            if subtask.predecessors.filter(state__value='finished', obsolete_since__isnull=True).exists():
                maybe_schedulable_subtasks.append(subtask)

        self.try_schedule_subtasks(maybe_schedulable_subtasks)

        super().start_handling()

    def onSubTaskCreated(self, id: int):
        super().onSubTaskCreated(id)
        try:
            subtask = Subtask.objects.get(id=id)
            predecessors = subtask.predecessors.all()

            if predecessors.exists() and all([p.state.value==SubtaskState.Choices.FINISHED.value or p.is_obsolete for p in predecessors]):
                logger.info("subtask %s was just created and all its predecessors are finished or obsolete. trying to schedule it...", id)
                self.try_schedule_subtasks([subtask])
        except Subtask.DoesNotExist as e:
            logger.warning(e)

    def onSubTaskStatusChanged(self, id: int, status: str):
        super().onSubTaskStatusChanged(id, status)

        if status == SubtaskState.Choices.FINISHED.value:
            subtask = Subtask.objects.get(id=id)
            successors = subtask.successors.filter(state__value=SubtaskState.Choices.DEFINED.value)\
                                           .filter(obsolete_since__isnull=True).order_by('id').all()

            if successors:
                logger.info("subtask %s finished. trying to schedule %s defined non-obsolete successors: %s", id, successors.count(), ' '.join(str(id) for id in [s.id for s in successors]) or 'None')
                self.try_schedule_subtasks(successors)

            if subtask.specifications_template.type.value == SubtaskType.Choices.INGEST.value:
                # ingest subtasks have a special type of "successors": the cleanup subtask(s)
                # these cleanup subtasks do not depend on the ingest output data, so they are no direct data successor
                # these cleanup subtasks do depend on the ingest with the same scheduling unit to be finished.
                scheduling_unit_cleanup_subtasks = Subtask.objects.filter(task_blueprint__scheduling_unit_blueprint__id=subtask.task_blueprint.scheduling_unit_blueprint.id)\
                                                                  .filter(specifications_template__type__value=SubtaskType.Choices.CLEANUP.value) \
                                                                  .filter(state__value=SubtaskState.Choices.DEFINED.value) \
                                                                  .filter(obsolete_since__isnull=True).order_by('id').all()
                if scheduling_unit_cleanup_subtasks.exists():
                    logger.info("ingest subtask %s finished. trying to schedule %s defined non-obsolete cleanup successor(s): %s", id, scheduling_unit_cleanup_subtasks.count(), ' '.join(str(id) for id in [s.id for s in scheduling_unit_cleanup_subtasks]) or 'None')
                    self.try_schedule_subtasks(scheduling_unit_cleanup_subtasks)


    def onSchedulingUnitBlueprintIngestPermissionGranted(self, id: int, ingest_permission_granted_since: datetime):
        logger.info("ingest_permission_granted_since='%s' for scheduling_unit_blueprint id=%s, trying to schedule ingest subtasks if any, and prerequisites are met", ingest_permission_granted_since, id)
        scheduling_unit_ingest_subtasks = Subtask.objects.filter(task_blueprint__scheduling_unit_blueprint__id=id)\
                                                         .filter(specifications_template__type__value=SubtaskType.Choices.INGEST.value) \
                                                         .filter(state__value=SubtaskState.Choices.DEFINED.value) \
                                                         .filter(obsolete_since__isnull=True).order_by('id').all()
        self.try_schedule_subtasks(scheduling_unit_ingest_subtasks)

    def onTaskBlueprintOutputPinningUpdated(self, id: int, output_pinned: bool):
        logger.info("task_blueprint id=%s output_pinned changed to '%s'", id, output_pinned)
        try:
            task_blueprint = TaskBlueprint.objects.get(id=id)
            scheduling_unit_cleanup_subtasks = Subtask.objects.filter(task_blueprint__scheduling_unit_blueprint__id=task_blueprint.scheduling_unit_blueprint.id) \
                                                              .filter(specifications_template__type__value=SubtaskType.Choices.CLEANUP.value) \
                                                              .exclude(state__value=SubtaskState.Choices.FINISHED.value) \
                                                              .filter(obsolete_since__isnull=True).order_by('id').all()
            self.try_schedule_subtasks(scheduling_unit_cleanup_subtasks)
        except TaskBlueprint.DoesNotExist as e:
            logger.warning(e)

def create_subtask_scheduling_service(exchange: str=DEFAULT_BUSNAME, broker: str=DEFAULT_BROKER):
    return TMSSBusListener(handler_type=TMSSSubTaskSchedulingEventMessageHandler,
                           exchange=exchange, broker=broker)

