#!/usr/bin/env python3

# Copyright (C) 2012-2015  ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.
import time
import typing
import unittest
import uuid
from copy import deepcopy

import astropy.coordinates
from astropy.coordinates import Angle

import logging
logger = logging.getLogger(__name__)

from lofar.common.test_utils import skip_integration_tests
if skip_integration_tests():
    exit(3)

TEST_UUID = str(uuid.uuid1()).split('-')[0]

from datetime import datetime, timedelta
from lofar.common.datetimeutils import round_to_second_precision
from lofar.common.json_utils import get_default_json_object_for_schema, add_defaults_to_json_object_for_schema
from lofar.messaging.messagebus import TemporaryExchange, BusListenerJanitor

tmp_exchange = TemporaryExchange("t_dynamic_scheduling_%s" % (TEST_UUID,))
tmp_exchange.open()

# override DEFAULT_BUSNAME
import lofar
lofar.messaging.DEFAULT_BUSNAME = tmp_exchange.address
lofar.messaging.config.DEFAULT_BUSNAME = tmp_exchange.address

from lofar.sas.tmss.test.test_environment import TMSSTestEnvironment
tmss_test_env = TMSSTestEnvironment(exchange=tmp_exchange.address,
                                    populate_schemas=True, populate_test_data=False,
                                    start_postgres_listener=False, start_subtask_scheduler=False,
                                    enable_viewflow=False,
                                    start_dynamic_scheduler=False)  # do not start the dynamic scheduler in the testenv, because it is the object-under-test.
tmss_test_env.start()

def tearDownModule():
    tmss_test_env.stop()
    tmp_exchange.close()

from lofar.sas.tmss.test.tmss_test_data_django_models import *
from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.tmssapp.tasks import create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft, update_task_graph_from_specifications_doc, create_scheduling_unit_draft_from_observing_strategy_template
from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions, wait_for_scheduling_unit_blueprint_status
from lofar.sas.tmss.services.scheduling.constraints import get_boundary_stations_from_list

# the modules under test
from lofar.sas.tmss.services.scheduling.constraints import *
from lofar.sas.tmss.services.scheduling.dynamic_scheduling import *


class BaseSchedulingConstraintsTestCase(unittest.TestCase):
    '''
    Base class for all Scheduling constraints tests, includes common setup, in-between-test-cleanup, common scheduling unit creation methods, etc
    '''
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.project = models.Project.objects.create(**Project_test_data(name=str(uuid.uuid4()), project_state=models.ProjectState.objects.get(value=models.ProjectState.Choices.ACTIVE.value)))
        cls.cycle = models.Cycle.objects.create(**Cycle_test_data(name=str(uuid.uuid4()), start=datetime(2020, 1, 1), stop=datetime(2030, 1, 1)))
        cls.project.cycles.set([cls.cycle])
        cls.project.save()

        cls.scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=cls.project))

        cls.scheduler = Scheduler()

    def setUp(self) -> None:
        super().setUp()

        # wipe all scheduling_unit_drafts in between tests, so the tests don't influence each other
        tmss_test_env.delete_scheduling_unit_drafts_cascade()
        tmss_test_env.delete_reservations_cascade()
        wipe_evaluate_constraints_caches()

    @staticmethod
    def create_simple_observation_scheduling_unit(name:str=None, scheduling_set=None,
                                                  obs_duration:int=60,
                                                  constraints=None,
                                                  constraint_version: int=8) -> models.SchedulingUnitDraft:
        constraints_template = models.SchedulingConstraintsTemplate.get_version_or_latest(name="constraints", version=constraint_version)

        # we're testing at scheduling unit level here.
        # so, use a simple template with just one observation.
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_version_or_latest(name="Simple Observation")
        scheduling_unit_spec = add_defaults_to_json_object_for_schema(strategy_template.template,
                                                                      strategy_template.scheduling_unit_template.schema)
        scheduling_unit_spec['tasks']['Observation']['specifications_doc']['duration'] = obs_duration
        scheduling_unit_spec['tasks']['Observation']['specifications_doc']['station_configuration']['station_groups'][0]['stations'] = ['CS002', 'CS003']
        scheduling_unit_spec['tasks']['Observation']['specifications_doc']['station_configuration']['station_groups'][0]['max_nr_missing'] = 0

        # we're testing at scheduling unit level here.
        # so, keep the unit small for speed. Disable QA subtask creation.
        scheduling_unit_spec['tasks']['Observation']['specifications_doc']['QA']['file_conversion']['enabled'] = False
        scheduling_unit_spec['tasks']['Observation']['specifications_doc']['QA']['plots']['enabled'] = False

        if constraints is None:
            # by default, apply no constraints, and let the caller be specific on the constraints needed
            constraints = {'scheduler': 'dynamic'}

        if scheduling_unit_spec['scheduling_constraints_template']['version'] != constraints_template.version:
            scheduling_unit_spec['scheduling_constraints_template']['version'] = constraints_template.version
            scheduling_unit_spec['scheduling_constraints_doc'] = constraints

        if scheduling_set is None:
            scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=BaseSchedulingConstraintsTestCase.project))

        # add the scheduling_unit_doc to a new SchedulingUnitDraft instance, and were ready to use it!
        scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(name=name[:models.SchedulingUnitDraft.name.field.max_length],
                                                                          scheduling_set=scheduling_set,
                                                                          specifications_template=strategy_template.scheduling_unit_template,
                                                                          observation_strategy_template=strategy_template,
                                                                          scheduling_constraints_doc=constraints,
                                                                          scheduling_constraints_template=constraints_template)

        scheduling_unit_draft = update_task_graph_from_specifications_doc(scheduling_unit_draft, scheduling_unit_spec)
        return scheduling_unit_draft


class TestUtils(BaseSchedulingConstraintsTestCase):

    def test_get_boundary_stations_from_list(self):
        station_list = ['IE613', 'CS002', 'SE607', 'FR606', 'PL612', 'DE602', 'UK608']
        expected_stations = ('SE607', 'PL612', 'FR606', 'IE613')
        response = get_boundary_stations_from_list(tuple(station_list))
        self.assertEqual(response, expected_stations)

        station_list = ['RS409', 'RS508']
        expected_stations = ('RS508', 'RS508', 'RS409', 'RS409')
        response = get_boundary_stations_from_list(tuple(station_list))
        self.assertEqual(response, expected_stations)


class TestDailyConstraints(BaseSchedulingConstraintsTestCase):
    '''
    Tests for the constraint checkers used in dynamic scheduling
    '''

    def setUp(self) -> None:
        super().setUp()

        # scheduling unit
        self.obs_duration = 120 * 60
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit("SU %s" % self._testMethodName,
                                                                               scheduling_set=self.scheduling_set,
                                                                               obs_duration=self.obs_duration)

        specifications_doc = scheduling_unit_draft.specifications_doc
        specifications_doc['tasks']['Observation']['specifications_doc']['station_configuration']['station_groups'][0]['stations'] = ['CS002']
        update_task_graph_from_specifications_doc(scheduling_unit_draft, specifications_doc)
        self.scheduling_unit_blueprint_cs002 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        specifications_doc = scheduling_unit_draft.specifications_doc
        specifications_doc['tasks']['Observation']['specifications_doc']['station_configuration']['station_groups'][0]['stations'] = ['CS002', 'DE601']
        update_task_graph_from_specifications_doc(scheduling_unit_draft, specifications_doc)
        self.scheduling_unit_blueprint_cs002_de601 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

    # require_day

    def test_get_earliest_possible_start_time_with_daytime_constraint_returns_day_start(self):
        self.scheduling_unit_blueprint_cs002.scheduling_constraints_doc = {'daily': {'require_day': True}, 'scheduler': 'dynamic'}
        self.scheduling_unit_blueprint_cs002.save()
        timestamp = datetime(2020, 1, 1, 4, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint_cs002, timestamp, gridder=Gridder(1))
        # close enough? (depends on astropy precision)
        self.assertTrue(abs((returned_time-datetime(2020, 1, 1, 9, 40, 0)).total_seconds()) < 300)

    def test_get_earliest_possible_start_time_with_daytime_constraint_returns_day_start_of_latest_station(self):
        self.scheduling_unit_blueprint_cs002_de601.scheduling_constraints_doc = {'daily': {'require_day': True}, 'scheduler': 'dynamic'}
        self.scheduling_unit_blueprint_cs002_de601.save()
        timestamp = datetime(2020, 1, 1, 4, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint_cs002_de601, timestamp, gridder=Gridder(1))
        # close enough? (depends on astropy precision)
        self.assertTrue(abs((returned_time-datetime(2020, 1, 1, 9, 40, 0)).total_seconds()) < 300)

    def test_get_earliest_possible_start_time_with_daytime_constraint_returns_timestamp(self):
        self.scheduling_unit_blueprint_cs002.scheduling_constraints_doc = {'daily': {'require_day': True}, 'scheduler': 'dynamic'}
        self.scheduling_unit_blueprint_cs002.save()
        timestamp = datetime(2020, 1, 1, 10, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint_cs002, timestamp, gridder=Gridder(1))
        self.assertEqual(returned_time, timestamp)

    def test_get_earliest_possible_start_time_with_daytime_constraint_returns_next_day_start(self):
        self.scheduling_unit_blueprint_cs002.scheduling_constraints_doc = {'daily': {'require_day': True}, 'scheduler': 'dynamic'}
        self.scheduling_unit_blueprint_cs002.save()
        timestamp = datetime(2020, 1, 1, 20, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint_cs002, timestamp, gridder=Gridder(1))
        # close enough? (depends on astropy precision)
        self.assertTrue(abs((returned_time-datetime(2020, 1, 2, 9, 39, 0)).total_seconds()) < 300)

    def test_get_earliest_possible_start_time_with_daytime_constraint_returns_next_day_start_when_obs_does_not_fit(self):
        self.scheduling_unit_blueprint_cs002.scheduling_constraints_doc = {'daily': {'require_day': True}, 'scheduler': 'dynamic'}
        self.scheduling_unit_blueprint_cs002.save()
        timestamp = datetime(2020, 1, 1, 14, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint_cs002, timestamp, gridder=Gridder(1))
        # close enough? (depends on astropy precision)
        self.assertTrue(abs((returned_time-datetime(2020, 1, 2, 9, 39, 0)).total_seconds()) < 300)

    def test_can_run_within_timewindow_with_daily_constraints_with_daytime_constraint_returns_true(self):
        self.scheduling_unit_blueprint_cs002.scheduling_constraints_doc = {'daily': {'require_day': True}, 'scheduler': 'dynamic'}
        self.scheduling_unit_blueprint_cs002.save()

        lower_bound = datetime(2020, 1, 1, 10, 0, 0)
        upper_bound = datetime(2020, 1, 1, 15, 0, 0)
        self.assertTrue(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint_cs002, lower_bound, upper_bound))

    def test_can_run_within_timewindow_with_daily_constraints_with_daytime_constraint_returns_false_when_not_daytime(self):
        self.scheduling_unit_blueprint_cs002.scheduling_constraints_doc = {'daily': {'require_day': True}, 'scheduler': 'dynamic'}
        self.scheduling_unit_blueprint_cs002.save()

        lower_bound = datetime(2020, 1, 1, 20, 0, 0)
        upper_bound = datetime(2020, 1, 1, 23, 0, 0)
        self.assertFalse(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint_cs002, lower_bound, upper_bound))

    def test_can_run_within_timewindow_with_daily_constraints_with_daytime_constraint_returns_false_when_partially_not_daytime(self):
        self.scheduling_unit_blueprint_cs002.scheduling_constraints_doc = {'daily': {'require_day': True}, 'scheduler': 'dynamic'}
        self.scheduling_unit_blueprint_cs002.save()

        lower_bound = datetime(2020, 1, 1, 14, 0, 0)
        upper_bound = datetime(2020, 1, 1, 18, 0, 0)
        self.assertFalse(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint_cs002, lower_bound, upper_bound))

        lower_bound = datetime(2020, 1, 1, 8, 0, 0)
        upper_bound = datetime(2020, 1, 1, 11, 0, 0)
        self.assertFalse(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint_cs002, lower_bound, upper_bound))

    def test_can_run_within_timewindow_with_daytime_constraint_returns_correct_value(self):
        # remove other constraints:
        self.scheduling_unit_blueprint_cs002.scheduling_constraints_doc['sky'] = {}

        # set constraint to test
        self.scheduling_unit_blueprint_cs002.scheduling_constraints_doc = {'daily': {'require_day': True}, 'scheduler': 'dynamic'}
        self.scheduling_unit_blueprint_cs002.save()

        # can run in day
        lower_bound = datetime(2020, 1, 1, 8, 0, 0)
        upper_bound = datetime(2020, 1, 1, 15, 0, 0)
        self.assertTrue(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint_cs002, lower_bound, upper_bound))

        # cannot run at night
        lower_bound = datetime(2020, 1, 1, 15, 0, 0)
        upper_bound = datetime(2020, 1, 1, 23, 0, 0)
        self.assertFalse(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint_cs002, lower_bound, upper_bound))

    # require_night

    def test_get_earliest_possible_start_time_with_nighttime_constraint_returns_night_start(self):
        self.scheduling_unit_blueprint_cs002.scheduling_constraints_doc = {'daily': {'require_night': True}, 'scheduler': 'dynamic'}
        self.scheduling_unit_blueprint_cs002.save()
        timestamp = datetime(2020, 1, 1, 14, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint_cs002, timestamp, gridder=Gridder(1))
        # close enough? (depends on astropy precision)
        self.assertTrue(abs((returned_time-datetime(2020, 1, 1, 16, 38, 0)).total_seconds()) < 300)

    def test_get_earliest_possible_start_time_with_nighttime_constraint_returns_night_start_of_latest_station(self):
        self.scheduling_unit_blueprint_cs002_de601.scheduling_constraints_doc = {'daily': {'require_night': True}, 'scheduler': 'dynamic', 'location': 'both'}
        self.scheduling_unit_blueprint_cs002_de601.save()
        timestamp = datetime(2020, 1, 1, 14, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint_cs002_de601, timestamp, gridder=Gridder(1))
        # close enough? (depends on astropy precision)
        self.assertTrue(abs((returned_time-datetime(2020, 1, 1, 16, 45, 0)).total_seconds()) < 300)

    def test_get_earliest_possible_start_time_with_nighttime_constraint_returns_timestamp(self):
        self.scheduling_unit_blueprint_cs002.scheduling_constraints_doc = {'daily': {'require_night': True}, 'scheduler': 'dynamic'}
        self.scheduling_unit_blueprint_cs002.save()

        # late night
        timestamp = datetime(2020, 1, 1, 23, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint_cs002, timestamp, gridder=Gridder(1))
        self.assertEqual(returned_time, timestamp)

        # early night
        timestamp = datetime(2020, 1, 1, 3, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint_cs002, timestamp, gridder=Gridder(1))
        self.assertEqual(returned_time, timestamp)

    def test_get_earliest_possible_start_time_with_nighttime_constraint_returns_next_night_start_when_obs_does_not_fit(self):
        self.scheduling_unit_blueprint_cs002.scheduling_constraints_doc = {'daily': {'require_night': True}, 'scheduler': 'dynamic'}
        self.scheduling_unit_blueprint_cs002.save()

        # early night
        timestamp = datetime(2020, 1, 1, 6, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint_cs002, timestamp, gridder=Gridder(1))
        # close enough? (depends on astropy precision)
        self.assertTrue(abs((returned_time-datetime(2020, 1, 1, 16, 38, 0)).total_seconds()) < 300)

    def test_can_run_within_timewindow_with_daily_constraints_with_nighttime_constraint_returns_true(self):
        self.scheduling_unit_blueprint_cs002.scheduling_constraints_doc = {'daily': {'require_night': True}, 'scheduler': 'dynamic'}
        self.scheduling_unit_blueprint_cs002.save()

        # early night
        lower_bound = datetime(2020, 1, 1, 1, 0, 0)
        upper_bound = datetime(2020, 1, 1, 3, 0, 0)
        self.assertTrue(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint_cs002, lower_bound, upper_bound))

        # late night
        lower_bound = datetime(2020, 1, 1, 20, 0, 0)
        upper_bound = datetime(2020, 1, 1, 23, 0, 0)
        self.assertTrue(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint_cs002, lower_bound, upper_bound))

        # night-night next day
        lower_bound = datetime(2020, 1, 1, 23, 0, 0)
        upper_bound = datetime(2020, 1, 2, 3, 0, 0)
        self.assertTrue(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint_cs002, lower_bound, upper_bound))

    def test_can_run_within_timewindow_with_daily_constraints_with_nighttime_constraint_returns_false_when_not_nighttime(self):
        self.scheduling_unit_blueprint_cs002.scheduling_constraints_doc = {'daily': {'require_night': True}, 'scheduler': 'dynamic'}
        self.scheduling_unit_blueprint_cs002.save()

        lower_bound = datetime(2020, 1, 1, 10, 0, 0)
        upper_bound = datetime(2020, 1, 1, 14, 0, 0)
        self.assertFalse(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint_cs002, lower_bound, upper_bound))

    def test_can_run_within_timewindow_with_daily_constraints_with_nighttime_constraint_returns_false_when_partially_not_nighttime(self):
        self.scheduling_unit_blueprint_cs002.scheduling_constraints_doc = {'daily': {'require_night': True}, 'scheduler': 'dynamic'}
        self.scheduling_unit_blueprint_cs002.save()

        # day-night same day
        lower_bound = datetime(2020, 1, 1, 14, 0, 0)
        upper_bound = datetime(2020, 1, 1, 17, 0, 0)
        self.assertFalse(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint_cs002, lower_bound, upper_bound))

        # night-day same day
        lower_bound = datetime(2020, 1, 1, 6, 0, 0)
        upper_bound = datetime(2020, 1, 1, 10, 0, 0)
        self.assertFalse(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint_cs002, lower_bound, upper_bound))


    def test_can_run_within_timewindow_with_nighttime_constraint_returns_correct_value(self):
        # remove other constraints:
        self.scheduling_unit_blueprint_cs002.scheduling_constraints_doc['sky'] = {}

        # set constraint to test
        self.scheduling_unit_blueprint_cs002.scheduling_constraints_doc = {'daily': {'require_night': True}, 'scheduler': 'dynamic'}
        self.scheduling_unit_blueprint_cs002.save()

        # cannot run in day
        lower_bound = datetime(2020, 1, 1, 8, 0, 0)
        upper_bound = datetime(2020, 1, 1, 15, 0, 0)
        self.assertFalse(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint_cs002, lower_bound, upper_bound))

        # can run at night
        lower_bound = datetime(2020, 1, 1, 15, 0, 0)
        upper_bound = datetime(2020, 1, 1, 23, 0, 0)
        self.assertTrue(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint_cs002, lower_bound, upper_bound))


    # avoid_twilight

    def test_get_earliest_possible_start_time_with_twilight_constraint_returns_day_start(self):
        self.scheduling_unit_blueprint_cs002.scheduling_constraints_doc = {'daily': {'avoid_twilight': True}, 'scheduler': 'dynamic'}
        self.scheduling_unit_blueprint_cs002.save()

        timestamp = datetime(2020, 1, 1, 9, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint_cs002, timestamp, gridder=Gridder(1))
        # close enough? (depends on astropy precision)
        self.assertTrue(abs((returned_time-datetime(2020, 1, 1, 9, 40, 0)).total_seconds()) < 300)

    def test_get_earliest_possible_start_time_with_twilight_constraint_returns_day_start_of_latest_station(self):
        self.scheduling_unit_blueprint_cs002_de601.scheduling_constraints_doc = {'daily': {'avoid_twilight': True}, 'scheduler': 'dynamic'}
        self.scheduling_unit_blueprint_cs002_de601.save()

        timestamp = datetime(2020, 1, 1, 9, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint_cs002_de601, timestamp, gridder=Gridder(1))
        # close enough? (depends on astropy precision)
        self.assertTrue(abs((returned_time-datetime(2020, 1, 1, 9, 39, 0)).total_seconds()) < 300)

    def test_get_earliest_possible_start_time_with_twilight_constraint_returns_night_start(self):
        self.scheduling_unit_blueprint_cs002.scheduling_constraints_doc = {'daily': {'avoid_twilight': True}, 'scheduler': 'dynamic'}
        self.scheduling_unit_blueprint_cs002.save()

        timestamp = datetime(2020, 1, 1, 16, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint_cs002, timestamp, gridder=Gridder(1))
        # close enough? (depends on astropy precision)
        self.assertTrue(abs((returned_time-datetime(2020, 1, 1, 16, 38, 0)).total_seconds()) < 300)

    def test_get_earliest_possible_start_time_with_twilight_constraint_returns_night_start_of_latest_station(self):
        self.scheduling_unit_blueprint_cs002_de601.scheduling_constraints_doc = {'daily': {'avoid_twilight': True}, 'scheduler': 'dynamic', 'location': 'both'}
        self.scheduling_unit_blueprint_cs002_de601.save()

        timestamp = datetime(2020, 1, 1, 14, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint_cs002_de601, timestamp, gridder=Gridder(1))
        # close enough? (depends on astropy precision)
        self.assertTrue(abs((returned_time-datetime(2020, 1, 1, 16, 45, 0)).total_seconds()) < 300)

    def test_get_earliest_possible_start_time_with_twilight_constraint_returns_timestamp(self):
        self.scheduling_unit_blueprint_cs002.scheduling_constraints_doc = {'daily': {'avoid_twilight': True}, 'scheduler': 'dynamic'}
        self.scheduling_unit_blueprint_cs002.save()

        # # daytime
        timestamp = datetime(2020, 1, 1, 10, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint_cs002, timestamp, gridder=Gridder(1))
        self.assertEqual(returned_time, timestamp)

        # late time
        timestamp = datetime(2020, 1, 1, 20, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint_cs002, timestamp, gridder=Gridder(1))
        self.assertEqual(returned_time, timestamp)

        # early night
        timestamp = datetime(2020, 1, 1, 3, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint_cs002, timestamp, gridder=Gridder(1))
        self.assertEqual(returned_time, timestamp)

    def test_get_earliest_possible_start_time_with_twilight_constraint_returns_day_or_night_start_when_obs_does_not_fit(self):
        self.scheduling_unit_blueprint_cs002.scheduling_constraints_doc = {'daily': {'avoid_twilight': True}, 'scheduler': 'dynamic'}
        self.scheduling_unit_blueprint_cs002.save()

        timestamp = datetime(2020, 1, 1, 15, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint_cs002, timestamp, gridder=Gridder(1))
        # close enough? (depends on astropy precision)
        self.assertTrue(abs((returned_time-datetime(2020, 1, 1, 16, 38, 0)).total_seconds()) < 300)

        timestamp = datetime(2020, 1, 1, 7, 0, 0)
        returned_time = get_earliest_possible_start_time(self.scheduling_unit_blueprint_cs002, timestamp, gridder=Gridder(1))
        # close enough? (depends on astropy precision)
        self.assertTrue(abs((returned_time-datetime(2020, 1, 1, 9, 40, 0)).total_seconds()) < 300)

    def test_can_run_within_timewindow_with_daily_constraints_with_twilight_constraint_returns_true(self):
        self.scheduling_unit_blueprint_cs002.scheduling_constraints_doc = {'daily': {'avoid_twilight': True}, 'scheduler': 'dynamic'}
        self.scheduling_unit_blueprint_cs002.save()

        lower_bound = datetime(2020, 1, 1, 10, 0, 0)
        upper_bound = datetime(2020, 1, 1, 15, 0, 0)
        self.assertTrue(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint_cs002, lower_bound, upper_bound))

    def test_can_run_within_timewindow_with_daily_constraints_with_twilight_constraint_returns_false_when_in_twilight(self):
        self.scheduling_unit_blueprint_cs002.scheduling_constraints_doc = {'daily': {'avoid_twilight': True}, 'scheduler': 'dynamic'}
        self.scheduling_unit_blueprint_cs002.save()

        lower_bound = datetime(2020, 1, 1, 8, 0, 0)
        upper_bound = datetime(2020, 1, 1, 9, 0, 0)
        self.assertFalse(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint_cs002, lower_bound, upper_bound))

        lower_bound = datetime(2020, 1, 1, 16, 0, 0)
        upper_bound = datetime(2020, 1, 1, 17, 0, 0)
        self.assertFalse(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint_cs002, lower_bound, upper_bound))

    def test_can_run_within_timewindow_with_daily_constraints_with_twilight_constraint_returns_false_when_partially_in_twilight(self):
        self.scheduling_unit_blueprint_cs002.scheduling_constraints_doc = {'daily': {'avoid_twilight': True}, 'scheduler': 'dynamic'}
        self.scheduling_unit_blueprint_cs002.save()

        lower_bound = datetime(2020, 1, 1, 15, 0, 0)
        upper_bound = datetime(2020, 1, 1, 17, 0, 0)
        self.assertFalse(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint_cs002, lower_bound, upper_bound))

        lower_bound = datetime(2020, 1, 1, 6, 0, 0)
        upper_bound = datetime(2020, 1, 1, 10, 0, 0)
        self.assertFalse(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint_cs002, lower_bound, upper_bound))

    def test_can_run_within_timewindow_with_twilight_constraint_returns_correct_value(self):
        # remove other constraints:
        self.scheduling_unit_blueprint_cs002.scheduling_constraints_doc['sky'] = {}

        # set constraint to test
        self.scheduling_unit_blueprint_cs002.scheduling_constraints_doc = {'daily': {'avoid_twilight': True}, 'scheduler': 'dynamic'}
        self.scheduling_unit_blueprint_cs002.save()

        # can run in day
        lower_bound = datetime(2020, 1, 1, 8, 0, 0)
        upper_bound = datetime(2020, 1, 1, 15, 0, 0)
        self.assertTrue(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint_cs002, lower_bound, upper_bound))

        # can run at night
        lower_bound = datetime(2020, 1, 1, 15, 0, 0)
        upper_bound = datetime(2020, 1, 1, 23, 0, 0)
        self.assertTrue(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint_cs002, lower_bound, upper_bound))


class TestSkyConstraints(BaseSchedulingConstraintsTestCase):
    '''
    Tests for the constraint checkers used in dynamic scheduling
    '''

    def setUp(self) -> None:
        super().setUp()

        # scheduling unit
        self.obs_duration = 120 * 60
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit("SU %s" % self._testMethodName,
                                                                               scheduling_set=self.scheduling_set,
                                                                               obs_duration=self.obs_duration)

        self.scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
        # keep only sky constraints
        self.scheduling_unit_blueprint.scheduling_constraints_doc = {'sky': {}, 'scheduler': 'dynamic'}
        self.scheduling_unit_blueprint.save()

    # min_distance

    def test_can_run_within_timewindow_with_sky_constraints_with_min_distance_constraint_returns_true_when_met(self):
        self.scheduling_unit_blueprint.scheduling_constraints_doc['sky'] = {'min_distance': {'sun': 0.1, 'moon': 0.1, 'jupiter': 0.1}}
        self.scheduling_unit_blueprint.save()
        timestamp = datetime(2020, 1, 1, 10, 0, 0)
        returned_value = can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint, timestamp, timestamp + self.scheduling_unit_blueprint.specified_main_observation_duration)
        self.assertTrue(returned_value)

    def test_can_run_within_timewindow_with_sky_constraints_with_min_distance_constraint_returns_false_when_not_met(self):
        self.scheduling_unit_blueprint.scheduling_constraints_doc['sky'] = {'min_distance': {'sun': 3.14, 'moon': 3.14, 'jupiter': 3.14}}
        self.scheduling_unit_blueprint.save()
        timestamp = datetime(2020, 1, 1, 10, 0, 0)
        returned_value = can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint, timestamp, timestamp + self.scheduling_unit_blueprint.specified_main_observation_duration)
        self.assertFalse(returned_value)

    def test_sky_constraints_with_min_distance_TMSS_2243_bugfix(self):
        '''Check that the (min_)distance calculation is correct.
        See https://support.astron.nl/jira/browse/TMSS-2243.
        And see:  https://docs.astropy.org/en/stable/coordinates/common_errors.html#object-separation
        '''
        timestamp = datetime(2023, 1, 16, 12, 0, 0)

        self.scheduling_unit_blueprint.scheduling_constraints_doc['sky'] = {'min_distance': {'sun': astropy.coordinates.Angle(100, unit=astropy.units.degree).rad }}
        self.scheduling_unit_blueprint.save()
        self.assertTrue(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint, timestamp, timestamp + self.scheduling_unit_blueprint.specified_main_observation_duration))

        wipe_evaluate_constraints_caches()

        self.scheduling_unit_blueprint.scheduling_constraints_doc['sky'] = {'min_distance': {'sun': astropy.coordinates.Angle(120, unit=astropy.units.degree).rad }}
        self.scheduling_unit_blueprint.save()
        self.assertFalse(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint, timestamp, timestamp + self.scheduling_unit_blueprint.specified_main_observation_duration))

    def test_sky_min_distance_constraints_with_calibrator_and_target(self):
        '''Check the new constraints template version >= 9 with both calibrator and target min_distance constraints
        '''
        constraints_template = models.SchedulingConstraintsTemplate.get_version_or_latest(name="constraints", version=9)
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_version_or_latest(name="IM HBA - 1 Beam")
        scheduling_unit_spec = strategy_template.template_doc_complete_with_defaults
        scheduling_unit_spec['scheduling_constraints_doc'] = constraints_template.get_default_json_document_for_schema()
        scheduling_unit_spec['scheduling_constraints_template'] = { 'name': constraints_template.name, 'version': constraints_template.version }
        # scheduling_unit_spec['tasks']['Target Observation']['specifications_doc']['station_configuration']['SAPs'][0]['digital_pointing']

        # add the scheduling_unit_doc to a new SchedulingUnitDraft instance, and were ready to use it!
        scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(name="test_sky_min_distance_constraints_with_calibrator_and_target",
                                                                          scheduling_set=self.scheduling_set,
                                                                          specifications_template=strategy_template.scheduling_unit_template,
                                                                          observation_strategy_template=strategy_template,
                                                                          scheduling_constraints_template=constraints_template)
        scheduling_unit_draft = update_task_graph_from_specifications_doc(scheduling_unit_draft, scheduling_unit_spec)

        scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
        scheduling_unit_blueprint.scheduling_constraints_doc = {'sky': {}, 'scheduler': 'dynamic'}
        scheduling_unit_blueprint.save()

        timestamp = datetime(2020, 1, 1, 19, 0, 0)

        small_distances_to_bodies = {'sun': 0.01, 'moon': 0.01, 'jupiter': 0.01}
        too_large_distances_to_bodies = {'sun': 3.14, 'moon': 3.14, 'jupiter': 3.14}

        scheduling_unit_blueprint.scheduling_constraints_doc['sky'] = {'min_distance': {}}
        scheduling_unit_blueprint.scheduling_constraints_doc['sky']['min_distance'] = {'target': small_distances_to_bodies}
        scheduling_unit_blueprint.save()
        self.assertTrue(can_run_within_timewindow_with_constraints(scheduling_unit_blueprint, timestamp, timestamp + scheduling_unit_blueprint.specified_observation_duration))

        wipe_evaluate_constraints_caches()
        scheduling_unit_blueprint.scheduling_constraints_doc['sky']['min_distance'] = {'target': too_large_distances_to_bodies}
        scheduling_unit_blueprint.save()
        self.assertFalse(can_run_within_timewindow_with_constraints(scheduling_unit_blueprint, timestamp, timestamp + scheduling_unit_blueprint.specified_observation_duration))

        wipe_evaluate_constraints_caches()
        scheduling_unit_blueprint.scheduling_constraints_doc['sky']['min_distance'] = {'calibrator': small_distances_to_bodies, 'target': small_distances_to_bodies}
        scheduling_unit_blueprint.save()
        self.assertTrue(can_run_within_timewindow_with_constraints(scheduling_unit_blueprint, timestamp, timestamp + scheduling_unit_blueprint.specified_observation_duration))

        wipe_evaluate_constraints_caches()
        scheduling_unit_blueprint.scheduling_constraints_doc['sky']['min_distance'] = {'calibrator': too_large_distances_to_bodies, 'target': small_distances_to_bodies}
        scheduling_unit_blueprint.save()
        self.assertFalse(can_run_within_timewindow_with_constraints(scheduling_unit_blueprint, timestamp, timestamp + scheduling_unit_blueprint.specified_observation_duration))




    # min_target_elevation

    def test_can_run_within_timewindow_with_sky_constraints_with_min_target_elevation_constraint_returns_true_when_met(self):
        self.scheduling_unit_blueprint.scheduling_constraints_doc['sky'] = {'min_elevation': {'target': 0.1} }
        self.scheduling_unit_blueprint.save()
        timestamp = datetime(2020, 1, 1, 10, 0, 0)
        returned_value = can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint, timestamp, timestamp + self.scheduling_unit_blueprint.specified_main_observation_duration)
        self.assertTrue(returned_value)


    def test_can_run_within_timewindow_with_sky_constraints_with_min_target_elevation_constraint_returns_false_when_not_met(self):
        self.scheduling_unit_blueprint.scheduling_constraints_doc['sky'] = {'min_elevation': {'target': 1.57} }
        self.scheduling_unit_blueprint.save()
        timestamp = datetime(2020, 1, 1, 15, 0, 0)
        returned_value = can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint, timestamp, timestamp + self.scheduling_unit_blueprint.specified_main_observation_duration)
        self.assertFalse(returned_value)

    # transit_offset

    def test_can_run_within_timewindow_with_sky_constraints_with_transit_offset_constraint_returns_true_when_met(self):
        # transit at 19:40, obs middle is at 20:00, so we have an offset of roughly -20min or -1200 seconds

        # big window should fit
        self.scheduling_unit_blueprint.scheduling_constraints_doc['sky'] = {'transit_offset': {'from': -43200, 'to': 43200}}
        self.scheduling_unit_blueprint.save()
        timestamp = datetime(2020, 1, 1, 19, 0, 0)
        returned_value = can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint, timestamp, timestamp + self.scheduling_unit_blueprint.specified_main_observation_duration)
        self.assertTrue(returned_value)

        # narrow window around transit should still fit
        # (not too narrow, take astropy grid into account)
        self.scheduling_unit_blueprint.scheduling_constraints_doc['sky'] = {'transit_offset': {'from': -1500, 'to': -1000}}
        self.scheduling_unit_blueprint.save()
        timestamp = datetime(2020, 1, 1, 19, 0, 0)
        returned_value = can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint, timestamp, timestamp + self.scheduling_unit_blueprint.specified_main_observation_duration)
        self.assertTrue(returned_value)


    def test_can_run_within_timewindow_with_sky_constraints_with_transit_offset_constraint_returns_false_when_not_met(self):
        # transit at ~19:40 (depending on astropy grid, give of take few minutes)

        # window after
        self.scheduling_unit_blueprint.scheduling_constraints_doc['sky'] = {'transit_offset': {'from': -10*60, 'to': 0}}
        self.scheduling_unit_blueprint.save()
        timestamp = datetime(2020, 1, 1, 19, 0, 0) # obs center is at 20:00  ->  offset ~ -20min
        returned_value = can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint, timestamp, timestamp + self.scheduling_unit_blueprint.specified_main_observation_duration)
        self.assertFalse(returned_value)

        # window before
        self.scheduling_unit_blueprint.scheduling_constraints_doc['sky'] = {'transit_offset': {'from': -50*60, 'to': -40*60}}
        self.scheduling_unit_blueprint.save()
        timestamp = datetime(2020, 1, 1, 19, 0, 0) # obs center is at 20:00  ->  offset ~ -20min
        returned_value = can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint, timestamp, timestamp + self.scheduling_unit_blueprint.specified_main_observation_duration)
        self.assertFalse(returned_value)


    def test_can_run_within_timewindow_with_sky_constraints_with_transit_offset_considers_reference_pointing_for_HBA(self):
        # check the correct angle is used when calculating the transit time, depending on reference pointing being enabled or not

        # only use sky constraints
        self.scheduling_unit_blueprint.scheduling_constraints_doc = {'sky':{}, 'scheduler': 'dynamic'}

        # reference_pointing disabled, should use 'normal' HBA tile beam pointing
        # transit at 19:40, obs middle is at 20:00  ->  offset ~ -20min
        self.scheduling_unit_blueprint.scheduling_constraints_doc['sky'] = {'transit_offset': {'from': -60*60, 'to': 30*60}, 'reference_pointing': {'enabled': False, 'pointing': {'angle1': 0, 'angle2': 1, 'direction_type': 'J2000', 'target': 'TargetC'}}}
        self.scheduling_unit_blueprint.save()
        timestamp = datetime(2020, 1, 1, 19, 0, 0)
        returned_value = can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint, timestamp, timestamp + self.scheduling_unit_blueprint.specified_main_observation_duration)
        self.assertTrue(returned_value)

        # reference_pointing enabled, enforce a different angle1, so a different transit, which does not fit anymore
        wipe_evaluate_constraints_caches()
        self.scheduling_unit_blueprint.scheduling_constraints_doc['sky'] = {'transit_offset': {'from': -60*60, 'to': 30*60}, 'reference_pointing': {'enabled': True, 'pointing': {'angle1': 0, 'angle2': 1, 'direction_type': 'J2000', 'target': 'TargetC'}}}
        self.scheduling_unit_blueprint.save()
        timestamp = datetime(2020, 1, 1, 19, 0, 0)
        returned_value = can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint, timestamp, timestamp + self.scheduling_unit_blueprint.specified_main_observation_duration)
        self.assertFalse(returned_value)

        # reference_pointing enabled, but use a copy of original tile beam. should fit again
        wipe_evaluate_constraints_caches()
        self.scheduling_unit_blueprint.scheduling_constraints_doc['sky'] = {'transit_offset': {'from': -60*60, 'to': 30*60}, 'reference_pointing': {} }
        self.scheduling_unit_blueprint.scheduling_constraints_doc['sky']['reference_pointing']['pointing'] = deepcopy(self.scheduling_unit_blueprint.specifications_doc['tasks']['Observation']['specifications_doc']['station_configuration']['tile_beam'])
        self.scheduling_unit_blueprint.scheduling_constraints_doc['sky']['reference_pointing']['enabled'] = True
        self.scheduling_unit_blueprint.save()
        timestamp = datetime(2020, 1, 1, 19, 0, 0)
        returned_value = can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint, timestamp, timestamp + self.scheduling_unit_blueprint.specified_main_observation_duration)
        self.assertTrue(returned_value)


    def test_can_run_within_timewindow_with_sky_constraints_with_transit_offset_considers_reference_pointing_for_LBA(self):
        # check the correct angle is used when calculating the transit time, depending on reference pointing being enabled or not

        specifications_doc = self.scheduling_unit_blueprint.specifications_doc
        specifications_doc['tasks']['Observation']['specifications_doc']['station_configuration']['antenna_set'] = 'LBA_INNER'
        specifications_doc['tasks']['Observation']['specifications_doc']['station_configuration']['SAPs'] = [{'name': 'TargetA', 'subbands': [0, 1], 'digital_pointing': {'angle1': 0.6624317181687094, 'angle2': 1.5579526427549426, 'direction_type': 'J2000', 'target': 'TargetA'}}]
        scheduling_unit_draft = update_task_graph_from_specifications_doc(self.scheduling_unit_blueprint.draft, specifications_doc)
        self.scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
        # only use sky constraints
        self.scheduling_unit_blueprint.scheduling_constraints_doc = {'sky':{}, 'scheduler': 'dynamic'}

        # check the correct angle is used when calculating the transit time, depending on reference pointing being enabled or not

        # reference_pointing disabled, should use 'normal' LBA SAP pointing(s)
        # transit at ~19:40, obs middle is at 20:00  ->  offset ~ -20min
        self.scheduling_unit_blueprint.scheduling_constraints_doc['sky'] = {'transit_offset': {'from': -60*60, 'to': 30*60}, 'reference_pointing': {'enabled': False, 'pointing': {'angle1': 0, 'angle2': 1, 'direction_type': 'J2000', 'target': 'TargetC'}}}
        self.scheduling_unit_blueprint.save()
        timestamp = datetime(2020, 1, 1, 19, 0, 0)
        returned_value = can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint, timestamp, timestamp + self.scheduling_unit_blueprint.specified_main_observation_duration, gridder=Gridder(Gridder.FINE_TIME_GRID))
        self.assertTrue(returned_value)

        # reference_pointing enabled, enforce a different angle1, so a different transit, which does not fit anymore
        wipe_evaluate_constraints_caches()
        self.scheduling_unit_blueprint.scheduling_constraints_doc['sky'] = {'transit_offset': {'from': -60*60, 'to': 30*60}, 'reference_pointing': {'enabled': True, 'pointing': {'angle1': 0, 'angle2': 1, 'direction_type': 'J2000', 'target': 'TargetC'}}}
        self.scheduling_unit_blueprint.save()
        timestamp = datetime(2020, 1, 1, 19, 0, 0)
        returned_value = can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint, timestamp, timestamp + self.scheduling_unit_blueprint.specified_main_observation_duration, gridder=Gridder(Gridder.FINE_TIME_GRID))
        self.assertFalse(returned_value)

        # reference_pointing enabled, but use a copy of original LBA SAP pointing. should fit again
        wipe_evaluate_constraints_caches()
        self.scheduling_unit_blueprint.scheduling_constraints_doc['sky'] = {'transit_offset': {'from': -60*60, 'to': 30*60}, 'reference_pointing': {} }
        self.scheduling_unit_blueprint.scheduling_constraints_doc['sky']['reference_pointing']['pointing'] = deepcopy(self.scheduling_unit_blueprint.specifications_doc['tasks']['Observation']['specifications_doc']['station_configuration']['SAPs'][0]['digital_pointing'])
        self.scheduling_unit_blueprint.scheduling_constraints_doc['sky']['reference_pointing']['enabled'] = True
        self.scheduling_unit_blueprint.save()
        timestamp = datetime(2020, 1, 1, 19, 0, 0)
        returned_value = can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint, timestamp, timestamp + self.scheduling_unit_blueprint.specified_main_observation_duration)
        self.assertTrue(returned_value)


    def test_can_run_within_timewindow_with_sky_constraints_with_transit_offset_constraint_all_SAPs_for_LBA(self):
        # sap1 transits at 12:43,
        # sap2 transits at 16:39
        # obs middle is 15:00, so with an offset of +/- 3hours they should fit

        specifications_doc = self.scheduling_unit_blueprint.specifications_doc
        specifications_doc['tasks']['Observation']['specifications_doc']['station_configuration']['antenna_set'] = 'LBA_INNER'
        specifications_doc['tasks']['Observation']['specifications_doc']['station_configuration']['SAPs'] = [{'name': 'CygA', 'subbands': [0, 1], 'digital_pointing': {'angle1': 5.233660650313663, 'angle2': 0.7109404782526458, 'direction_type': 'J2000', 'target': 'CygA', }},
                                                                                                             {'name': 'CasA', 'subbands': [2, 3], 'digital_pointing': {'angle1': 6.233660650313663, 'angle2': 0.6109404782526458, 'direction_type': 'J2000', 'target': 'CasA',}}]
        scheduling_unit_draft = update_task_graph_from_specifications_doc(self.scheduling_unit_blueprint.draft, specifications_doc)
        self.scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
        self.scheduling_unit_blueprint.scheduling_constraints_doc = {'scheduler': 'dynamic'}
        self.scheduling_unit_blueprint.scheduling_constraints_doc['sky'] = {'transit_offset': {'from': -3*60*60, 'to': 3*60*60}}
        self.scheduling_unit_blueprint.save()
        timestamp = datetime(2020, 1, 1, 14, 0, 0)
        returned_value = can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint, timestamp, timestamp + self.scheduling_unit_blueprint.specified_main_observation_duration)
        self.assertTrue(returned_value)

    def test_can_run_within_timewindow_with_sky_constraints_with_MOON_and_SUN_pointings(self):
        # this checks that constraints can deal with SUN and MOON, which should be both up and transit in the late morning/noon on Jan 1st.
        specifications_doc = self.scheduling_unit_blueprint.specifications_doc
        specifications_doc['tasks']['Observation']['specifications_doc']['station_configuration']['antenna_set'] = 'LBA_INNER'
        specifications_doc['tasks']['Observation']['specifications_doc']['station_configuration']['SAPs'] = [{'name': 'Sun', 'subbands': [0, 1], 'digital_pointing': {'angle1': 0.0, 'angle2': 0.0, 'direction_type': 'SUN', 'target': 'Sun', }},
                                                                                                             {'name': 'Moon', 'subbands': [2, 3], 'digital_pointing': {'angle1': 0.0, 'angle2': 0.0, 'direction_type': 'MOON', 'target': 'Moon',}}]
        scheduling_unit_draft = update_task_graph_from_specifications_doc(self.scheduling_unit_blueprint.draft, specifications_doc)
        self.scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
        self.scheduling_unit_blueprint.scheduling_constraints_doc = {'scheduler': 'dynamic'}
        wipe_evaluate_constraints_caches()
        self.scheduling_unit_blueprint.scheduling_constraints_doc['sky'] = {'transit_offset': {'from': -3*60*60, 'to': 3*60*60}, 'min_elevation': {'target': 0.0}}
        self.scheduling_unit_blueprint.save()
        timestamp = datetime(2022, 1, 1, 11, 0, 0)
        returned_value = can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint, timestamp, timestamp + self.scheduling_unit_blueprint.specified_main_observation_duration)
        self.assertTrue(returned_value)

        # # unmet min elevation
        # todo: I'd expect this to fail, and coordinates_timestamp_and_station_to_target_rise_and_set expectedly return that MOON is always below horizon, but for some reason can_run_within_timewindow_with_constraints return True nonetheless
        wipe_evaluate_constraints_caches()
        self.scheduling_unit_blueprint.scheduling_constraints_doc['sky'] = {'transit_offset': {'from': -3*60*60, 'to': 3*60*60}, 'min_elevation': {'target': 1.5}}
        self.scheduling_unit_blueprint.save()
        timestamp = datetime(2022, 1, 1, 11, 0, 0)
        returned_value = can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint, timestamp, timestamp + self.scheduling_unit_blueprint.specified_main_observation_duration)
        self.assertFalse(returned_value)

        # transit not met
        # todo: I'd expect this to fail, and target transit time computation works on SUN and MOON, but for some reason can_run_within_timewindow_with_constraints return True nonetheless
        #  I assume this is elated to the "Applying temporary sky.transit_offset=[-12h,+12h] as workaround to prevent unschedulable bug" I see in the logs.
        # wipe_evaluate_constraints_caches()
        # self.scheduling_unit_blueprint.scheduling_constraints_doc['sky'] = {'transit_offset': {'from': -1*60*60, 'to': 1*60*60}, 'min_elevation': {'target': 0.0}}
        # self.scheduling_unit_blueprint.save()
        # timestamp = datetime(2022, 1, 1, 11, 2, 2)
        # returned_value = can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint, timestamp, timestamp + self.scheduling_unit_blueprint.specified_main_observation_duration)
        # self.assertFalse(returned_value)



class TestTimeConstraints(BaseSchedulingConstraintsTestCase):
    """
    Tests for the time constraint checkers used in dynamic scheduling with different boundaries
    Possible time constraints are
    - at
    - after
    - before
    - between (one or more 'from-to')
    - not between (one or more 'from-to')
    """

    def add_time_at_constraint(self, at_timestamp):
        lst_at_constraint = self.scheduling_unit_blueprint.scheduling_constraints_doc
        lst_at_constraint['time']['at'] = at_timestamp.isoformat()
        self.scheduling_unit_blueprint.save()

    def add_time_between_constraint(self, from_timestamp, to_timestamp):
        if "time" not in self.scheduling_unit_blueprint.scheduling_constraints_doc:
            self.set_minimal_time_constraints()
        if "between" not in self.scheduling_unit_blueprint.scheduling_constraints_doc['time']:
            self.scheduling_unit_blueprint.scheduling_constraints_doc['time']['between'] = []
        lst_between_constraints = self.scheduling_unit_blueprint.scheduling_constraints_doc['time']["between"]
        time_constraint_dict = {"from": from_timestamp.isoformat(), "to": to_timestamp.isoformat()}
        lst_between_constraints.append(time_constraint_dict)
        self.scheduling_unit_blueprint.save()

    def add_time_not_between_constraint(self, from_timestamp, to_timestamp):
        if "not_between" not in self.scheduling_unit_blueprint.scheduling_constraints_doc['time']:
            self.scheduling_unit_blueprint.scheduling_constraints_doc['time']['not_between'] = []
        lst_between_constraints = self.scheduling_unit_blueprint.scheduling_constraints_doc['time']["not_between"]
        time_constraint_dict = {"from": from_timestamp.isoformat(), "to": to_timestamp.isoformat()}
        lst_between_constraints.append(time_constraint_dict)
        self.scheduling_unit_blueprint.save()

    def set_minimal_time_constraints(self):
        self.scheduling_unit_blueprint.scheduling_constraints_doc = {'scheduler': 'dynamic', 'time':{}}

    def setUp(self) -> None:
        super().setUp()

        # scheduling unit
        self.obs_duration = 120 * 60
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit("SU %s" % self._testMethodName,
                                                                               scheduling_set=self.scheduling_set,
                                                                               obs_duration=self.obs_duration)
        self.scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
        self.set_minimal_time_constraints()

    # 'after' constraint

    def test_can_run_within_after_returns_false(self):

        # Set datetime constraints before lower bounds, but with too short window for obs duration
        self.set_minimal_time_constraints()
        self.scheduling_unit_blueprint.scheduling_constraints_doc['time']["after"] = datetime(2020, 1, 1, 11, 0, 0).isoformat()
        self.assertFalse(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint,
                                                                            datetime(2020, 1, 1, 12, 0, 0),
                                                                            datetime(2020, 1, 1, 13, 0, 0)))

        # Set datetime constraints after lower bounds, and with too little space left in window for obs duration
        self.set_minimal_time_constraints()
        self.scheduling_unit_blueprint.scheduling_constraints_doc['time']["after"] = datetime(2020, 1, 1, 14, 0, 0).isoformat()
        self.assertFalse(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint,
                                                                            datetime(2020, 1, 1, 12, 0, 0),
                                                                            datetime(2020, 1, 1, 15, 0, 0)))

    def test_can_run_within_after_returns_true(self):

        # Set datetime constraints before lower bounds, and with sufficient window for obs duration
        self.set_minimal_time_constraints()
        self.scheduling_unit_blueprint.scheduling_constraints_doc['time']["after"] = datetime(2020, 1, 1, 11, 0, 0).isoformat()
        self.assertTrue(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint,
                                                                            datetime(2020, 1, 1, 12, 0, 0),
                                                                            datetime(2020, 1, 1, 14, 0, 0)))

        # Set datetime constraints after lower bounds, but with sufficient space left in window for obs duration
        self.set_minimal_time_constraints()
        self.scheduling_unit_blueprint.scheduling_constraints_doc['time']["after"] = datetime(2020, 1, 1, 13, 0, 0).isoformat()
        self.assertTrue(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint,
                                                                            datetime(2020, 1, 1, 12, 0, 0),
                                                                            datetime(2020, 1, 1, 16, 0, 0)))

    # 'before' constraint

    def test_can_run_within_before_returns_false(self):

        # Set datetime constraints after upper bound, but with too short window for obs duration
        self.set_minimal_time_constraints()
        self.scheduling_unit_blueprint.scheduling_constraints_doc['time']["before"] = datetime(2020, 1, 2, 13, 0, 0).isoformat()
        self.assertFalse(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint,
                                                                            datetime(2020, 1, 2, 11, 0, 0),
                                                                            datetime(2020, 1, 2, 12, 0, 0)))

        # Set datetime constraints after lower bound, and with too little space left in window for obs duration
        self.set_minimal_time_constraints()
        self.scheduling_unit_blueprint.scheduling_constraints_doc['time']["before"] = datetime(2020, 1, 1, 13, 0, 0).isoformat()
        self.assertFalse(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint,
                                                                            datetime(2020, 1, 1, 12, 0, 0),
                                                                            datetime(2020, 1, 2, 12, 0, 0)))

    def test_can_run_before_returns_false(self):
        self.set_minimal_time_constraints()
        self.scheduling_unit_blueprint.scheduling_constraints_doc['time']["before"] = datetime(2020, 1, 1, 13, 0, 0).isoformat()
        self.assertFalse(can_run_before(self.scheduling_unit_blueprint, datetime(2020, 1, 1, 12, 0, 0), gridder=Gridder()))


    def test_can_run_before_returns_true(self):

        # Set datetime constraints after upper bounds, and with sufficient window for obs duration
        self.set_minimal_time_constraints()
        self.scheduling_unit_blueprint.scheduling_constraints_doc['time']["before"] = datetime(2020, 1, 2, 13, 0, 0).isoformat()
        self.assertTrue(can_run_before(self.scheduling_unit_blueprint, datetime(2020, 1, 2, 11, 0, 0), gridder=Gridder()))

        # Set datetime constraints after lower bounds, but with sufficient space left in window for obs duration
        self.set_minimal_time_constraints()
        self.scheduling_unit_blueprint.scheduling_constraints_doc['time']["before"] = datetime(2020, 1, 1, 15, 0, 0).isoformat()
        self.assertTrue(can_run_before(self.scheduling_unit_blueprint, datetime(2020, 1, 1, 12, 0, 0), gridder=Gridder()))


    def test_can_run_within_before_returns_true(self):

        # Set datetime constraints after upper bounds, and with sufficient window for obs duration
        self.set_minimal_time_constraints()
        self.scheduling_unit_blueprint.scheduling_constraints_doc['time']["before"] = datetime(2020, 1, 2, 13, 0, 0).isoformat()
        self.assertTrue(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint,
                                                                            datetime(2020, 1, 1, 12, 0, 0),
                                                                            datetime(2020, 1, 2, 12, 0, 0)))

        # Set datetime constraints after lower bounds, but with sufficient space left in window for obs duration
        wipe_evaluate_constraints_caches()
        self.set_minimal_time_constraints()
        self.scheduling_unit_blueprint.scheduling_constraints_doc['time']["before"] = datetime(2020, 1, 1, 15, 0, 0).isoformat()
        self.assertTrue(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint,
                                                                            datetime(2020, 1, 1, 12, 0, 0),
                                                                            datetime(2020, 1, 2, 12, 0, 0)))

    # 'between' constraint

    def test_can_run_within_between_returns_true(self):
        """
        Test 'between' constraint with start/stop datetime constraints (within, not anywhere within)
        """
        # Set datetime constraints start > lower_bound and stop > upper_bound, large window
        self.set_minimal_time_constraints()
        self.add_time_between_constraint(datetime(2020, 1, 1, 13, 0, 0), datetime(2020, 1, 2, 12, 0, 0))
        self.assertTrue(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint,
                                         datetime(2020, 1, 1, 12, 0, 0), datetime(2020, 1, 1, 20, 0, 0)))

        # Set datetime constraints start = lower_bound and stop = upper_bound, window just large enough for obs
        self.set_minimal_time_constraints()
        self.add_time_between_constraint(datetime(2020, 1, 1, 12, 0, 0), datetime(2020, 1, 1, 14, 0, 0))
        self.assertTrue(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint,
                                         datetime(2020, 1, 1, 12, 0, 0), datetime(2020, 1, 1, 14, 10, 0)))

    def test_can_run_within_between_returns_false(self):
        """
        Test 'between' constraint with start/stop datetime constraints (within, not anywhere within)
        """
        # Set datetime constraints start < lower_bound and stop < upper_bound, too little overlap for obs
        self.set_minimal_time_constraints()
        self.add_time_between_constraint(datetime(2020, 1, 1, 10, 0, 0), datetime(2020, 1, 1, 13, 0, 0))
        self.assertFalse(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint,
                                         datetime(2020, 1, 1, 12, 0, 0), datetime(2020, 1, 1, 20, 0, 0)))

    # 'not between' contraint

    # several simultaneous time ranges in 'at' / 'between' / 'not between' constraints

    def execute_can_run_within_timewindow_of_24hour_boundary(self):
        """
        Just a simple wrapper to call 'can_run_within_timewindow_with_constraints' function
        with a 24 hours boundary 2020-01-01 12:00 - 2020-01-02 12:00
        """
        return (can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint,
                                                           datetime(2020, 1, 1, 12, 0, 0),
                                                           datetime(2020, 1, 2, 12, 0, 0)))

    def test_can_run_within_at_constraint(self):
        """
        Test "at" constraint with both boundary and 'inside' upper_bound and lower_bound
        """
        # no constraints defined so should be OK
        self.set_minimal_time_constraints()
        self.assertTrue(self.execute_can_run_within_timewindow_of_24hour_boundary())

        with self.assertRaises(RuleValidationException) as context:
            self.set_minimal_time_constraints()
            self.scheduling_unit_blueprint.scheduling_constraints_doc['scheduler'] = 'fixed_time'
            self.scheduling_unit_blueprint.save()
            self.add_time_at_constraint(datetime(2020, 1, 1, 11, 0, 0))
        self.assertTrue('''missing required property 'time.at' for 'fixed time' scheduler''' in str(context.exception))

        # Set datetime constraint before lower_bound
        self.set_minimal_time_constraints()
        self.add_time_at_constraint(datetime(2020, 1, 1, 11, 0, 0))
        self.assertFalse(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint,
                                                                            datetime(2020, 1, 1, 12, 0, 0),
                                                                            datetime(2020, 1, 1, 14, 0, 0)))

        # Set datetime constraint at lower_bound, but duration exceeds upper_bound
        self.set_minimal_time_constraints()
        self.add_time_at_constraint(datetime(2020, 1, 1, 12, 0, 0))
        self.assertFalse(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint,
                                                                            datetime(2020, 1, 1, 12, 0, 0),
                                                                            datetime(2020, 1, 1, 13, 0, 0)))

        # Set datetime constraint at upper_bound
        self.set_minimal_time_constraints()
        self.add_time_at_constraint(datetime(2020, 1, 1, 14, 0, 0))
        self.assertFalse(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint,
                                                                             datetime(2020, 1, 1, 12, 0, 0),
                                                                             datetime(2020, 1, 1, 14, 0, 0)))

        # Set datetime constraint after upper_bound
        self.set_minimal_time_constraints()
        self.add_time_at_constraint(datetime(2020, 1, 1, 15, 0, 0))
        self.assertFalse(can_run_within_timewindow_with_constraints(self.scheduling_unit_blueprint,
                                                                            datetime(2020, 1, 1, 12, 0, 0),
                                                                            datetime(2020, 1, 1, 14, 0, 0)))

        # Set datetime constraint at lower_bound
        self.set_minimal_time_constraints()
        self.add_time_at_constraint(datetime(2020, 1, 1, 12, 0, 0))
        self.assertTrue(self.execute_can_run_within_timewindow_of_24hour_boundary())

        # Set datetime constraint that fits the time window
        self.set_minimal_time_constraints()
        self.add_time_at_constraint(datetime(2020, 1, 1, 18, 30, 0))
        self.assertTrue(self.execute_can_run_within_timewindow_of_24hour_boundary())

        # Set datetime constraint so that obs lasts till exactly upper_bound
        self.set_minimal_time_constraints()
        self.add_time_at_constraint(datetime(2020, 1, 2, 9, 50, 0))
        self.assertTrue(self.execute_can_run_within_timewindow_of_24hour_boundary())

    def test_can_run_within_between_constraints(self):
        """
        Test multiple 'between' constraints within 24 boundary and check overall result of
        'can_run_within_timewindow_with_constraints'
        This function will iterate between the boundary with boundary shift of 1hr and boundary length of
        the observation duration, which is in this testcase two hour
        i.e. 12-14, 13-15, 14-16,..etc.., 9-11
        """
        # no constraints defined so should be OK
        self.set_minimal_time_constraints()
        # self.assertTrue(self.execute_can_run_within_timewindow_of_24hour_boundary())

        # Try to add constraint of 1hr, this won't fit the observation.
        with self.assertRaises(RuleValidationException) as context:
            self.add_time_between_constraint(datetime(2020, 1, 1, 13, 0, 0), datetime(2020, 1, 1, 14, 0, 0))
        self.assertTrue('shorter' in str(context.exception))
        self.scheduling_unit_blueprint.refresh_from_db()

        # Add constraints of 2hr, but partially outside the bounds, we still cannot run
        self.add_time_between_constraint(datetime(2020, 1, 2, 11, 0, 0), datetime(2020, 1, 2, 13, 0, 0))
        self.assertFalse(self.execute_can_run_within_timewindow_of_24hour_boundary())

        # Add constraints of 2hr, we can run again
        self.add_time_between_constraint(datetime(2020, 1, 1, 17, 0, 0), datetime(2020, 1, 1, 19, 0, 0))
        self.assertTrue(self.execute_can_run_within_timewindow_of_24hour_boundary())

        # Add constraint of 24hr constraint, we still 'can_run'
        self.add_time_between_constraint(datetime(2020, 1, 1, 12, 0, 0), datetime(2020, 1, 2, 12, 0, 0))
        self.assertTrue(self.execute_can_run_within_timewindow_of_24hour_boundary())

        # Clear all between constraints
        self.set_minimal_time_constraints()

        # Add constraints after the 24hr, now we 'can not run'
        self.add_time_between_constraint(datetime(2020, 1, 2, 13, 0, 0), datetime(2020, 1, 2, 15, 0, 0))
        self.add_time_between_constraint(datetime(2020, 1, 2, 16, 0, 0), datetime(2020, 1, 2, 20, 0, 0))
        self.assertFalse(self.execute_can_run_within_timewindow_of_24hour_boundary())

        # Add constraint before the 24hr, we 'still can not run'
        self.add_time_between_constraint(datetime(2020, 1, 1, 9, 0, 0), datetime(2020, 1, 1, 12, 0, 0))
        self.assertFalse(self.execute_can_run_within_timewindow_of_24hour_boundary())

        # add one 'inside' constraint of 3 hours, so overall must be ok again.
        # Note that 2 hrs would only be sufficient if they match the moving window exactly (here: full hour)
        self.add_time_between_constraint(datetime(2020, 1, 1, 14, 30, 0), datetime(2020, 1, 1, 17, 30, 0))
        self.assertTrue(self.execute_can_run_within_timewindow_of_24hour_boundary())

    def test_can_run_within_not_between_constraints(self):
        """
        Test multiple 'not_between' constraints within 24 boundary and check overall result of
        'can_run_within_timewindow_with_constraints'
        This function will iterate between the boundary with boundary shift of 1hr and boundary length of
        the observation duration, which is in this testcase two hour
        i.e. 12-14, 13-15, 14-16,..etc.., 9-11
        """
        # no constraints defined so should be OK
        self.set_minimal_time_constraints()
        self.assertTrue(self.execute_can_run_within_timewindow_of_24hour_boundary())

        # Add constraints of 1hr, we still 'can_run'
        wipe_evaluate_constraints_caches()
        self.add_time_not_between_constraint(datetime(2020, 1, 1, 13, 0, 0), datetime(2020, 1, 1, 14, 0, 0))
        self.add_time_not_between_constraint(datetime(2020, 1, 1, 16, 0, 0), datetime(2020, 1, 1, 17, 0, 0))
        self.assertTrue(self.execute_can_run_within_timewindow_of_24hour_boundary())

        # Add constraints of 2hr, we still 'can_run'
        wipe_evaluate_constraints_caches()
        self.add_time_not_between_constraint(datetime(2020, 1, 1, 18, 0, 0), datetime(2020, 1, 1, 20, 0, 0))
        self.assertTrue(self.execute_can_run_within_timewindow_of_24hour_boundary())

        # Add constraint of 20hr constraint, we still 'can_run'
        wipe_evaluate_constraints_caches()
        self.add_time_not_between_constraint(datetime(2020, 1, 1, 12, 0, 0), datetime(2020, 1, 2, 8, 0, 0))
        # JS 20220913: I'm not sure this check is correct.
        # self.assertTrue(self.execute_can_run_within_timewindow_of_24hour_boundary())

        # Add constraint of 2hr, to fill the 'last gap', now we can not run
        wipe_evaluate_constraints_caches()
        self.add_time_not_between_constraint(datetime(2020, 1, 2, 10, 0, 0), datetime(2020, 1, 2, 12, 0, 0))
        self.assertFalse(self.execute_can_run_within_timewindow_of_24hour_boundary())

        self.set_minimal_time_constraints()


    # combined time contraints tests

    def test_can_run_combined_time_constraints(self):
        """
        Test multiple time constraints in combination and make sure that they block the time window as expected,
        even though each constraint individually would allow the observation to run.
        """

        # Set before and after constraint with sufficient gap to fit observation, and assert True
        self.set_minimal_time_constraints()
        self.scheduling_unit_blueprint.scheduling_constraints_doc['time']["after"] = datetime(2020, 1, 1, 12, 59, 59).isoformat()
        self.scheduling_unit_blueprint.scheduling_constraints_doc['time']["before"] = datetime(2020, 1, 1, 15, 0, 1).isoformat()
        self.assertTrue(self.execute_can_run_within_timewindow_of_24hour_boundary())

        # set before and after constraint with slightly smaller gap for observation, and assert False
        self.set_minimal_time_constraints()
        self.scheduling_unit_blueprint.scheduling_constraints_doc['time']["after"] = datetime(2020, 1, 1, 13, 0, 1).isoformat()
        self.scheduling_unit_blueprint.scheduling_constraints_doc['time']["before"] = datetime(2020, 1, 1, 15, 0, 0).isoformat()
        self.assertFalse(self.execute_can_run_within_timewindow_of_24hour_boundary())

        # set before and after constraint with large gap
        # then and add additional between and not between constraints until window is blocked
        # can run 13-8h
        self.set_minimal_time_constraints()
        self.scheduling_unit_blueprint.scheduling_constraints_doc['time']["after"] = datetime(2020, 1, 1, 13, 0, 0).isoformat()
        self.scheduling_unit_blueprint.scheduling_constraints_doc['time']["before"] = datetime(2020, 1, 2, 8, 0, 0).isoformat()
        self.assertTrue(self.execute_can_run_within_timewindow_of_24hour_boundary())

        # can run 13h-20h
        self.add_time_between_constraint(datetime(2020, 1, 1, 11, 0, 0), datetime(2020, 1, 1, 20, 0, 0))
        self.assertTrue(self.execute_can_run_within_timewindow_of_24hour_boundary())

        # can run 13h-17h
        self.add_time_not_between_constraint(datetime(2020, 1, 1, 17, 0, 0), datetime(2020, 1, 2, 4, 0, 0))
        self.assertTrue(self.execute_can_run_within_timewindow_of_24hour_boundary())

        # can not run anymore
        self.add_time_not_between_constraint(datetime(2020, 1, 1, 12, 0, 0), datetime(2020, 1, 1, 16, 0, 0))
        self.assertFalse(self.execute_can_run_within_timewindow_of_24hour_boundary())

        # JS 20220913: I'm not sure this check is correct.
        # # add another between window, can run 4h-8h
        # self.add_time_between_constraint(datetime(2020, 1, 1, 2, 0, 0), datetime(2020, 1, 2, 12, 0, 0))
        # self.assertTrue(self.execute_can_run_within_timewindow_of_24hour_boundary())

        # JS 20220913: I'm not sure this check is correct.
        # # move before constraint, can not run anymore
        # self.scheduling_unit_blueprint.scheduling_constraints_doc['time']["before"] = datetime(2020, 1, 2, 5, 0, 0).isoformat()
        # self.assertFalse(self.execute_can_run_within_timewindow_of_24hour_boundary())


class TestCombinedConstraints(BaseSchedulingConstraintsTestCase):
    '''
    Tests for the combinations of constraints
    '''

    def test_bugfix_TMSS_2338(self) -> None:
        '''
        See: https://support.astron.nl/jira/browse/TMSS-2338
        Investigation showed that the root cause of this bug is that get_earliest_possible_start_time yields different results when argument upper_bound is/isn't None.
        Specifications are used from the Pulsar acceptance test for target 'J1024-0719' which experienced this bug.
        '''
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name='BF Pulsar Timing')
        pointing = {'angle1': Angle('10h24m38.68846').radian, 'angle2': Angle('-07d19m19.170').radian, 'direction_type': 'J2000', 'target': 'J1024-0719'}
        scheduling_unit_draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template,
                                                                                              self.scheduling_set,
                                                                                              name='test_bugfix_TMSS_2338_'+str(uuid.uuid4()),
                                                                                              description='J1024-0719',
                                                                                              priority_queue=models.PriorityQueueType.objects.get(value=models.PriorityQueueType.Choices.B.value),
                                                                                              specifications_doc_overrides={
                                                                                                  'tasks': {
                                                                                                      'Observation': {
                                                                                                          'short_description': 'J1024-0719',
                                                                                                          'specifications_doc': {
                                                                                                              'duration': 25*60,
                                                                                                              'station_configuration': {
                                                                                                                  'tile_beam': pointing,
                                                                                                                  'SAPs': [{ 'digital_pointing': pointing}] } } }
                                                                                                  },
                                                                                                  'scheduling_constraints_doc': {
                                                                                                      'scheduler': 'dynamic',
                                                                                                      'time': {'between': [{'from': '2023-04-22T00:00:00Z', 'to': '2023-04-24T00:00:00Z'}]},
                                                                                                      'daily': { 'require_day': False, 'require_night': False},
                                                                                                      'sky': {'min_elevation': { 'target': Angle('25 degrees').radian},
                                                                                                              'transit_offset': { 'to': 2 * 60 * 60, 'from': -2 * 60 * 60}}
                                                                                                  }})

        scheduling_unit = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        lower_bound = datetime(2023, 4, 22, 0, 16)
        upper_bound = datetime(2023, 4, 23, 23, 59)
        gridder = Gridder(5)

        # get_earliest_possible_start_time should yield datetime(2023, 4, 23, 18, 1, 5) regardless of whether an upper_bound was given or not.
        self.assertEqual(get_earliest_possible_start_time(scheduling_unit, lower_bound, gridder=gridder),
                         get_earliest_possible_start_time(scheduling_unit, lower_bound, upper_bound, gridder=gridder))

        # (double)-check inner workings of filter_scheduling_units_which_can_only_run_in_this_window
        self.assertTrue(can_run_within_timewindow_with_constraints(scheduling_unit, lower_bound, upper_bound, gridder=gridder))
        self.assertIsNotNone(get_earliest_possible_start_time(scheduling_unit, lower_bound, gridder=gridder))
        self.assertIsNotNone(get_earliest_possible_start_time(scheduling_unit, lower_bound, upper_bound, gridder=gridder))

        # and finally check the workings of filter_scheduling_units_which_can_only_run_in_this_window itself
        self.assertEqual([], filter_scheduling_units_which_can_only_run_in_this_window([scheduling_unit], lower_bound, upper_bound))


    def test_bugfix_TMSS_2344(self) -> None:
        '''
        See: https://support.astron.nl/jira/browse/TMSS-2344
        Using specs from https://tmss.lofar.eu/api/scheduling_unit_blueprint/2013/specifications_doc_with_just_the_parameters
        '''
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name='BF Pulsar Timing')
        pointing = {'angle1': 4.006754690909052, 'angle2': 0.8565406441094557, 'direction_type': 'J2000', 'target': 'J1518+4904'}
        scheduling_unit_draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template,
                                                                                              self.scheduling_set,
                                                                                              name='test_bugfix_TMSS_2344_'+str(uuid.uuid4()),
                                                                                              description='J1518+4904',
                                                                                              priority_queue=models.PriorityQueueType.objects.get(value=models.PriorityQueueType.Choices.B.value),
                                                                                              specifications_doc_overrides={
                                                                                                  'tasks': {
                                                                                                      'Observation': {
                                                                                                          'short_description': 'J1518+4904',
                                                                                                          'specifications_doc': {
                                                                                                              'duration': 1800,
                                                                                                              'station_configuration': {
                                                                                                                  'tile_beam': pointing,
                                                                                                                  'SAPs': [{ 'digital_pointing': pointing}] } } }
                                                                                                  },
                                                                                                  'scheduling_constraints_doc': {
                                                                                                      'scheduler': 'dynamic',
                                                                                                      'time': {},
                                                                                                      'daily': { 'require_day': False, 'require_night': False},
                                                                                                      'sky': {'min_elevation': { 'target': Angle('30 degrees').radian},
                                                                                                              'min_distance': {"jupiter": 0, "moon": 0.008726646259971648, "sun": 0.008726646259971648 },
                                                                                                              'transit_offset': { 'to': 7200, 'from': -7200}
                                                                                                              }
                                                                                                  }})

        scheduling_unit = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        # with a fine grid, the evaluate_sky_min_elevation_constraint yields True
        result = evaluate_sky_min_elevation_constraint(scheduling_unit, datetime(2023, 4, 3, 19, 5, 0), Gridder(1))
        logger.info(result)
        self.assertTrue(result.is_constraint_met)

        # at the same timestamp, with a coarse grid, the evaluate_sky_min_elevation_constraint yields False
        result = evaluate_sky_min_elevation_constraint(scheduling_unit, datetime(2023, 4, 3, 19, 5, 0), Gridder(60))
        logger.info(result)
        self.assertFalse(result.is_constraint_met)

        # TMSS-2344: this current version of this test shows that the same evaluate_sky_min_elevation_constraint yields different results
        # and yields different log lines, depending on the used grid.



logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

# show debug messages for scheduling modules
logging.getLogger('lofar.sas.tmss.services.scheduling').level = logging.DEBUG
logging.getLogger('lofar.sas.tmss.services.scheduling.constraints').level = logging.DEBUG
logging.getLogger('lofar.sas.tmss.tmss.tmssapp.subtasks').level = logging.DEBUG
# hide spam of postgreslistener messages
logging.getLogger('lofar.sas.tmss.services.tmss_postgres_listener').disabled = True

if __name__ == '__main__':
    #run the unit tests
    unittest.main()
