#!/usr/bin/env python3

# Copyright (C) 2012-2015  ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

import unittest
import uuid

import logging
logger = logging.getLogger('lofar.'+__name__)

from lofar.common.test_utils import skip_integration_tests
if skip_integration_tests():
    exit(3)

from lofar.messaging.messagebus import TemporaryExchange, BusListenerJanitor
from lofar.common.json_utils import get_default_json_object_for_schema, add_defaults_to_json_object_for_schema

from time import sleep
from datetime import datetime, timedelta

class TestSubtaskSchedulingService(unittest.TestCase):
    '''
    Tests for the SubtaskSchedulingService
    '''
    @classmethod
    def setUpClass(cls) -> None:
        cls.TEST_UUID = uuid.uuid1()

        cls.tmp_exchange = TemporaryExchange("%s_%s" % (cls.__name__, cls.TEST_UUID))
        cls.tmp_exchange.open()

        # override DEFAULT_BUSNAME
        import lofar
        lofar.messaging.config.DEFAULT_BUSNAME = cls.tmp_exchange.address
        lofar.messaging.DEFAULT_BUSNAME = cls.tmp_exchange.address

        # import here, and not at top of module, because DEFAULT_BUSNAME needs to be set before importing
        from lofar.sas.tmss.test.test_environment import TMSSTestEnvironment

        cls.tmss_test_env = TMSSTestEnvironment(exchange=cls.tmp_exchange.address, populate_schemas=True, populate_test_data=False,
                                                start_subtask_scheduler=True, start_postgres_listener=True,
                                                start_dynamic_scheduler=False, enable_viewflow=False)
        cls.tmss_test_env.start()

        from lofar.sas.tmss.test.tmss_test_data_rest import TMSSRESTTestDataCreator
        cls.test_data_creator = TMSSRESTTestDataCreator(cls.tmss_test_env.django_server.url,
                                                        (cls.tmss_test_env.ldap_server.dbcreds.user, cls.tmss_test_env.ldap_server.dbcreds.password))

    @classmethod
    def tearDownClass(cls) -> None:
        cls.tmss_test_env.stop()
        cls.tmp_exchange.close()

    @staticmethod
    def wait_for_subtask_to_get_status(tmss_client, subtask_id, expected_status, timeout=30):
        '''helper method to poll for a subtask's status.
        raises TimeoutError if expected_status is not met withing timout seconds.
        returns subtask when expected_status is met.'''
        start = datetime.utcnow()
        subtask = tmss_client.get_subtask(subtask_id)
        while subtask['state_value'] != expected_status:
            sleep(0.5)
            logger.info("Waiting for subtask id=%s to get status '%s'. Current status='%s'. Polling...", subtask_id, expected_status, subtask['state_value'])
            subtask = tmss_client.get_subtask(subtask_id)
            if datetime.utcnow() - start > timedelta(seconds=timeout):
                raise TimeoutError("timeout while waiting for subtask id=%s to get status '%s'. It currently has status '%s'" % (
                    subtask_id, expected_status, subtask['state_value']))
        return subtask

    def test_01_for_expected_behaviour_of_two_connected_subtasks(self):
        '''
        This test starts a scheduling service and tmss, creates a chain of subtasks, finishes the first, and checks if the successors are then scheduled.
        '''
        return

        logger.info(' -- test_01_for_expected_behaviour -- ')

        # create and start the service (the object under test)
        service = create_service(exchange=self.tmp_exchange.address, tmss_client_credentials_id=self.tmss_test_env.client_credentials.dbcreds_id)
        with BusListenerJanitor(service):
            # -------------------------
            # long setup of objects....

            # setup proper template
            subtask_template_url = self.test_data_creator.post_data_and_get_url(self.test_data_creator.SubtaskTemplate(subtask_type_url=self.test_data_creator.django_api_url + '/subtask_type/qa_files/'), '/subtask_template/')

            # create two subtasks
            subtask1 = self.test_data_creator.post_data_and_get_response_as_json_object(self.test_data_creator.Subtask(state="defined", specifications_template_url=subtask_template_url), '/subtask/')
            subtask2 = self.test_data_creator.post_data_and_get_response_as_json_object(self.test_data_creator.Subtask(state="defined", specifications_template_url=subtask_template_url, task_blueprint_urls=subtask1['task_blueprints']), '/subtask/')

            # connect them
            output_url = self.test_data_creator.post_data_and_get_url(self.test_data_creator.SubtaskOutput(subtask1['url']), '/subtask_output/')
            input_url = self.test_data_creator.post_data_and_get_url(self.test_data_creator.SubtaskInput(subtask_url=subtask2['url'], subtask_output_url=output_url), '/subtask_input/')

            # ... end of long setup of objects
            # --------------------------------

            # now for the real test: set subtask1_id status to finished, and check that subtask2 is then properly scheduled
            with self.tmss_test_env.create_tmss_client() as tmss_client:
                subtask1 = tmss_client.get_subtask(subtask1['id'])
                subtask2 = tmss_client.get_subtask(subtask2['id'])

                self.assertEqual(subtask1['state_value'], 'defined')
                self.assertEqual(subtask2['state_value'], 'defined')

                # the first subtask ran, and is now finished... set it's status. This should trigger the scheduling service to schedule the second subtask.
                tmss_client.set_subtask_status(subtask1['id'], 'finished')

                subtask2 = self.wait_for_subtask_to_get_status(tmss_client, subtask2['id'], 'scheduled')

                # subtask2 should now be scheduled
                self.assertEqual(subtask2['state_value'], 'scheduled')

    def test_02_for_expected_behaviour_of_UC1_scheduling_unit(self):
        '''
        This test starts a scheduling service and tmss, creates a chain of subtasks, finishes the first, and checks if the successors are then scheduled.
        '''

        logger.info(' -- test_02_for_expected_behaviour_of_UC1_scheduling_unit -- ')

        # import here, and not at top of module, because the tmsstestenv needs to be running before importing
        from lofar.sas.tmss.tmss.tmssapp.tasks import create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft, update_task_graph_from_specifications_doc
        from lofar.sas.tmss.tmss.tmssapp.subtasks import update_subtasks_start_times_for_scheduling_unit
        from lofar.sas.tmss.tmss.tmssapp import models
        from lofar.sas.tmss.test.tmss_test_data_django_models import SchedulingSet_test_data
        from lofar.sas.tmss.services.scheduling.subtask_scheduling import create_subtask_scheduling_service

        # create and start the service (the object under test)
        service = create_subtask_scheduling_service(exchange=self.tmp_exchange.address)
        with BusListenerJanitor(service):
            # -------------------------
            # setup of objects: create the UC1 scheduling unit, and then select the first runnable subtasks
            strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_version_or_latest(name="IM HBA LoTSS - 2 Beams")

            # ensure QA subtasks are enabled for this test
            for obs_task_name in ('Calibrator Observation 1', 'Calibrator Observation 2', 'Target Observation'):
                for qa_subtask_type in ('plots', 'file_conversion'):
                    strategy_template.template['tasks'][obs_task_name]['specifications_doc']['QA'][qa_subtask_type]['enabled'] = True
            strategy_template.save()

            constraints_template = models.SchedulingConstraintsTemplate.get_version_or_latest(name=strategy_template.template.get('scheduling_constraints_template', {}).get('name', 'constraints'),
                                                                                              version=strategy_template.template.get('scheduling_constraints_template', {}).get('version'))
            constraints_doc = constraints_template.get_default_json_document_for_schema()

            spec = add_defaults_to_json_object_for_schema(strategy_template.template, strategy_template.scheduling_unit_template.schema)
            scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(name="IM HBA LoTSS - 2 Beams",
                                                                              scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()),
                                                                              specifications_template=strategy_template.scheduling_unit_template,
                                                                              observation_strategy_template=strategy_template,
                                                                              scheduling_constraints_doc=constraints_doc,
                                                                              scheduling_constraints_template=constraints_template)

            update_task_graph_from_specifications_doc(scheduling_unit_draft, spec)
            scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

            # assign some non-overlapping starttimes, so the tasks can be scheduled
            update_subtasks_start_times_for_scheduling_unit(scheduling_unit_blueprint, datetime.utcnow()+timedelta(minutes=1))

            # scheduling_unit_blueprint now has task_blueprints and subtasks
            # "unpack" the whole graph, so we can "walk" it and see if the correct subtasks are scheduled once its predecessors are finished
            obs_cal1 = scheduling_unit_blueprint.task_blueprints.get(name="Calibrator Observation 1")
            obs_cal1_st_obs = obs_cal1.subtasks.get(specifications_template_id__in=models.SubtaskTemplate.objects.filter(type=models.SubtaskType.objects.get(value=models.SubtaskType.Choices.OBSERVATION.value)))
            obs_cal1_st_qa1 = obs_cal1.subtasks.get(specifications_template_id__in=models.SubtaskTemplate.objects.filter(type=models.SubtaskType.objects.get(value=models.SubtaskType.Choices.QA_FILES.value)))
            obs_cal1_st_qa2 = obs_cal1.subtasks.get(specifications_template_id__in=models.SubtaskTemplate.objects.filter(type=models.SubtaskType.objects.get(value=models.SubtaskType.Choices.QA_PLOTS.value)))

            pl_cal1 = scheduling_unit_blueprint.task_blueprints.get(name="Calibrator Pipeline 1")
            pl_cal1_st = pl_cal1.subtasks.first()

            obs_tgt = scheduling_unit_blueprint.task_blueprints.get(name="Target Observation")
            obs_tgt_st_obs = obs_tgt.subtasks.get(specifications_template_id__in=models.SubtaskTemplate.objects.filter(type=models.SubtaskType.objects.get(value=models.SubtaskType.Choices.OBSERVATION.value)))
            obs_tgt_st_qa1 = obs_tgt.subtasks.get(specifications_template_id__in=models.SubtaskTemplate.objects.filter(type=models.SubtaskType.objects.get(value=models.SubtaskType.Choices.QA_FILES.value)))
            obs_tgt_st_qa2 = obs_tgt.subtasks.get(specifications_template_id__in=models.SubtaskTemplate.objects.filter(type=models.SubtaskType.objects.get(value=models.SubtaskType.Choices.QA_PLOTS.value)))

            pl_tgt1 = scheduling_unit_blueprint.task_blueprints.get(name="Pipeline target1")
            pl_tgt1_st = pl_tgt1.subtasks.first()

            pl_tgt2 = scheduling_unit_blueprint.task_blueprints.get(name="Pipeline target2")
            pl_tgt2_st = pl_tgt2.subtasks.first()

            obs_cal2 = scheduling_unit_blueprint.task_blueprints.get(name="Calibrator Observation 2")
            obs_cal2_st_obs = obs_cal2.subtasks.get(specifications_template_id__in=models.SubtaskTemplate.objects.filter(type=models.SubtaskType.objects.get(value=models.SubtaskType.Choices.OBSERVATION.value)))
            obs_cal2_st_qa1 = obs_cal2.subtasks.get(specifications_template_id__in=models.SubtaskTemplate.objects.filter(type=models.SubtaskType.objects.get(value=models.SubtaskType.Choices.QA_FILES.value)))
            obs_cal2_st_qa2 = obs_cal2.subtasks.get(specifications_template_id__in=models.SubtaskTemplate.objects.filter(type=models.SubtaskType.objects.get(value=models.SubtaskType.Choices.QA_PLOTS.value)))

            pl_cal2 = scheduling_unit_blueprint.task_blueprints.get(name="Calibrator Pipeline 2")
            pl_cal2_st = pl_cal2.subtasks.first()

            # define the graph in an iterable way: as tuples of a subtask-successors-pair
            # the graph is also ordered in a 'data-flow' direction
            graph = (# calibrator1 obs, qa and pipeline
                     (obs_cal1_st_obs, (obs_cal1_st_qa1, pl_cal1_st)),
                     (obs_cal1_st_qa1, (obs_cal1_st_qa2,)),
                     (obs_cal1_st_qa2, tuple()),
                     (pl_cal1_st, tuple()),
                     #target obs, qa and pipelines
                     (obs_tgt_st_obs, (obs_tgt_st_qa1, pl_tgt1_st, pl_tgt2_st)),
                     (obs_tgt_st_qa1, (obs_tgt_st_qa2,)),
                     (obs_tgt_st_qa2, tuple()),
                     (pl_tgt1_st, tuple()),
                     (pl_tgt2_st, tuple()),
                     # calibrator2 obs, qa and pipeline
                     (obs_cal2_st_obs, (obs_cal2_st_qa1, pl_cal2_st)),
                     (obs_cal2_st_qa1, (obs_cal2_st_qa2,)),
                     (obs_cal2_st_qa2, tuple()),
                     (pl_cal2_st, tuple()) )

            logger.info(" --- test_02_for_expected_behaviour_of_UC1_scheduling_unit setup finished. starting actual test ---")
            # ... end of long setup of objects
            # --------------------------------

            # now for the real test: use only the http rest api to check statuses and call schedule methods
            with self.tmss_test_env.create_tmss_client() as tmss_client:
                # walk the graph in a "data-flow" direction
                for subtask, successors in graph:
                    # get up-to-date subtask via the rest client
                    subtask1 = tmss_client.get_subtask(subtask.id)
                    logger.info("subtask id=%s status=%s successors: %s", subtask1['id'], subtask1['state_value'], ','.join(str(s.id) for s in successors))

                    if subtask1['state_value'] == 'defined':
                        for successor in successors:
                            # get up-to-date subtask via the rest client
                            subtask2 = tmss_client.get_subtask(successor.id)
                            self.assertEqual(subtask2['state_value'], 'defined')

                        # simulate that some scheduler schedules the first subtask (which does not depend on predecessors)...
                        if len(tmss_client.get_subtask_predecessors(subtask1['id'])) == 0:
                            subtask1 = tmss_client.schedule_subtask(subtask1['id'])
                            self.assertEqual(subtask1['state_value'], 'scheduled')

                    if subtask1['state_value'] == 'scheduled':
                        # simulate that the first subtask ran, and is now finished...
                        # cycle over the 'run time' statuses, concluding with status to finished.
                        # The finished status should trigger the scheduling service to schedule the successor subtask(s).
                        for status in ['queueing', 'queued', 'starting', 'started', 'finishing', 'finished']:
                            tmss_client.set_subtask_status(subtask1['id'], status)

                        for successor in successors:
                            # get up-to-date subtask via the rest client
                            subtask2 = self.wait_for_subtask_to_get_status(tmss_client, successor.id, 'scheduled')
                            self.assertEqual(subtask2['state_value'], 'scheduled')


if __name__ == '__main__':
    #run the unit tests
    unittest.main()
