#!/usr/bin/env python3

# Copyright (C) 2021
# ASTRON (Netherlands Institute for Radio Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.


import logging
import os
from optparse import OptionParser, OptionGroup
from requests import session

logger = logging.getLogger(__name__)

from lofar.common.dbcredentials import DBCredentials
from lofar.sas.tmss.client.tmssbuslistener import *
from lofar.sas.tmss.client.tmss_http_rest_client import TMSSsession

class TMSSEventMessageHandlerForSlackWebhooks(TMSSEventMessageHandler):
    '''
    '''
    def __init__(self, slack_url: str, rest_client_creds_id: str="TMSSClient"):
        super().__init__(log_event_messages=False)
        self.slack_url = slack_url
        self.slack_session = session()
        # we use the http rest client only to construct urls, not to fetch objects (for speed/load)
        self.tmss_client = TMSSsession.create_from_dbcreds_for_ldap(rest_client_creds_id)

    def stop_handling(self):
        super().stop_handling()
        self.slack_session.close()

    def post_to_slack_webhook(self, message: str):
        logger.info("post_to_slack_webhook: %s", message)
        # post to slack, see https://api.slack.com/messaging/webhooks
        self.slack_session.post(url=self.slack_url, json={"text": message})

    def onProjectStatusUpdated(self, name: str, status: str):
        msg = "Project %s status: %s" % (self._get_formatted_project_string(name), status)
        self.post_to_slack_webhook(msg)

    def onSchedulingUnitBlueprintStatusChanged(self, id: int, status: str):
        msg = self._get_formatted_project_scheduling_unit_string(id)
        if status in ('scheduled', 'unschedulable', 'error'):
            from lofar.sas.tmss.tmss.tmssapp.models.specification import SchedulingUnitBlueprint
            scheduling_unit = SchedulingUnitBlueprint.objects.get(id=id)

            if status == 'scheduled':
                msg += " start: \'%s UTC\'" % (scheduling_unit.scheduled_start_time.isoformat().replace('T', ' '))
            elif status == 'unschedulable':
                msg += " reason: %s" % (scheduling_unit.unschedulable_reason)
            elif status == 'error':
                msg += " reason: %s" % (scheduling_unit.error_reason)
        self.post_to_slack_webhook(msg)


    def onSchedulingUnitBlueprintCreated(self, id: int):
        scheduling_unit = self.tmss_client.get_path_as_json_object('scheduling_unit_blueprint_slim/%s' % (id,))
        self.post_to_slack_webhook("%s was created\ndescription: %s" % (self._get_formatted_project_scheduling_unit_string(id),
                                                                        scheduling_unit['description'] or "<no description>"))

    def onReservationCreated(self, id: int):
        self._onReservationCreatedUpdated(id, "created")

    def onReservationUpdated(self, id: int):
        self._onReservationCreatedUpdated(id, "updated")

    def _onReservationCreatedUpdated(self, id: int, event: str):
        from lofar.sas.tmss.tmss.tmssapp.models.specification import Reservation
        reservation = Reservation.objects.get(id=id)
        reservation_url = self.tmss_client.get_full_url_for_path('reservation/%s' % (id,))
        reservation_ui_url = reservation_url.replace('/api/reservation/', '/reservation/view/')
        msg = "reservation <%s|\'%s\' [%s]> was %s. start='%s' end='%s'" % (reservation_ui_url, reservation.name, id, event,
                                                                            reservation.start_time, reservation.stop_time)
        self.post_to_slack_webhook(msg)

    def onReservationDeleted(self, id: int, start_time, stop_time):
        self.post_to_slack_webhook("reservation %d from '%s' until '%s' was deleted" % (id, start_time, stop_time))

    def _get_formatted_project_scheduling_unit_string(self, scheduling_unit_blueprint_id: int) -> str:
        from lofar.sas.tmss.tmss.tmssapp.models.specification import SchedulingUnitBlueprint
        scheduling_unit = SchedulingUnitBlueprint.objects.get(id=scheduling_unit_blueprint_id)
        scheduling_unit_url = self.tmss_client.get_full_url_for_path('scheduling_unit_blueprint/%s' % (scheduling_unit_blueprint_id,))

        project_name = scheduling_unit.draft.scheduling_set.project.name
        formatted_project_string = self._get_formatted_project_string(project_name)

        su_ui_url = scheduling_unit_url.replace('/api/scheduling_unit_blueprint/', '/schedulingunit/view/blueprint/')
        # make 'status' a url to the scheduling constraints plot if the unit is scheduled/unschedulable/observing.
        status = "<%s/scheduling_constraints_plot|%s>"%(scheduling_unit_url,scheduling_unit.status.value) if scheduling_unit.status.value in ('scheduled', 'unschedulable', 'observing') else scheduling_unit.status.value
        result = "%s - <%s|\'%s\' [%s]> status: %s" % (formatted_project_string, su_ui_url, scheduling_unit.name, scheduling_unit.id, status)
        return result

    def _get_formatted_project_string(self, project_name: str) -> str:
        project_url = self.tmss_client.get_full_url_for_path('project/%s' % (project_name,))
        project_ui_url = project_url.replace('/api/project/', '/project/view/')
        result = "<%s|\'%s\'>" % (project_ui_url, project_name)
        return result


def create_service(slack_url: str, rest_client_creds_id:str="TMSSClient", exchange: str=DEFAULT_BUSNAME, broker: str=DEFAULT_BROKER):
    return TMSSBusListener(handler_type=TMSSEventMessageHandlerForSlackWebhooks,
                           handler_kwargs={'slack_url': slack_url, 'rest_client_creds_id': rest_client_creds_id},
                           exchange=exchange, broker=broker)


def main():
    # make sure we run in UTC timezone
    os.environ['TZ'] = 'UTC'

    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

    # Check the invocation arguments
    parser = OptionParser('%prog [options]',
                          description='run the tmss_slack_webhook_service which listens for TMSS event messages on the messagebus, and posts the updates on the slack webhook api.')

    group = OptionGroup(parser, 'Slack options')
    parser.add_option_group(group)
    group.add_option('-s', '--slack_credentials', dest='slack_credentials', type='string', default='TMSSSlack', help='credentials name (for the lofar credentials files) containing the TMSS Slack Webhook URL, default: %default')

    group = OptionGroup(parser, 'Django options')
    parser.add_option_group(group)
    group.add_option('-C', '--credentials', dest='dbcredentials', type='string', default=os.environ.get('TMSS_DBCREDENTIALS', 'TMSS'), help='django dbcredentials name, default: %default')
    group.add_option('-R', '--rest_credentials', dest='rest_credentials', type='string', default=os.environ.get('TMSS_CLIENT_DBCREDENTIALS', 'TMSSClient'), help='django REST API credentials name, default: %default')

    group = OptionGroup(parser, 'Messaging options')
    group.add_option('-b', '--broker', dest='broker', type='string', default=DEFAULT_BROKER,
                     help='Address of the message broker, default: %default')
    group.add_option('-e', "--exchange", dest="exchange", type="string", default=DEFAULT_BUSNAME,
                     help="exchange where the TMSS event messages are published. [default: %default]")
    parser.add_option_group(group)

    (options, args) = parser.parse_args()

    from lofar.sas.tmss.tmss import setup_and_check_tmss_django_database_connection_and_exit_on_error
    setup_and_check_tmss_django_database_connection_and_exit_on_error(options.dbcredentials)
    TMSSsession.check_connection_and_exit_on_error(options.rest_credentials)

    # The TMSS slack app maintenance page (requires astron user creds): https://radio-observatory.slack.com/apps/A01SKUJHNKF-tmss

    # read the secrect slack webhook url from a lofar dbcredentials file.
    slack_url = DBCredentials().get(options.slack_credentials).host

    with create_service(slack_url=slack_url, rest_client_creds_id=options.rest_credentials, exchange=options.exchange, broker=options.broker):
        waitForInterrupt()

if __name__ == '__main__':
    main()
