#!/usr/bin/env python3

# subtask_scheduling.py
#
# Copyright (C) 2015
# ASTRON (Netherlands Institute for Radio Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.
from datetime import timedelta
import logging
import os
from optparse import OptionParser, OptionGroup

logger = logging.getLogger(__name__)

from lofar.sas.tmss.client.tmssbuslistener import *
from lofar.common.util import find_free_port

from enum import Enum
from json import dumps as JSONdumps
from json import loads as JSONloads
from SimpleWebSocketServer import SimpleWebSocketServer, WebSocket
from threading import Thread, Event
from django.apps import apps

DEFAULT_WEBSOCKET_PORT = 5678

'''
NOTE: 
HTTP Headers are not supported inside the WebSocket protocol (see https://tools.ietf.org/html/rfc6455#page-11).
WS traffic is either UTC-8 text or binary data (see https://tools.ietf.org/html/rfc6455#page-39).
'''


class TMSSWebSocket(WebSocket):

    def __init__(self, server, sock, address):
        super().__init__(server, sock, address)
        self._set_flags_to_default()

    def _set_flags_to_default(self):
        self.authenticated = False
        self.user = None

    def handleMessage(self):
        if self.authenticated:
            logger.debug('Client already authenticated, ignoring incoming message. User: %s from IP: %s' % (self.user, self.address[0]))
            return
        try:
            token_key = JSONloads(self.data).get('token', '')
            from rest_framework.authtoken.models import Token
            token_obj = Token.objects.filter(key=token_key).first()
            if token_obj:
                self.user = token_obj.user
                self.authenticated = True
                logger.info('Client authenticated. User: %s from IP: %s' % (self.user, self.address[0]))
            else:
                logger.info('Client not authenticated. IP: %s' % (self.address[0]))
                self.close(1011, u'Please login, so you have a token, and please submit the token in the 1st message after the connection was made.')
        except Exception as e:
            logger.exception('Error when handling websocket message of User: %s from IP: %s' % (self.user, self.address[0]))
            raise

    def handleConnected(self):
        # Enforce to initial values be safe
        self._set_flags_to_default()

    def handleClose(self):
        # Enforce to initial values be safe
        self._set_flags_to_default()


class TMSSEventMessageHandlerForWebsocket(TMSSEventMessageHandler):
    '''
    '''
    class ObjActions(Enum):
        CREATE = 'create'
        UPDATE = 'update'
        DELETE = 'delete'

    class ObjTypes(Enum):   # These values refer to the DRF APIs
        SCHED_UNIT_BLUEPRINT = 'scheduling_unit_blueprint'
        SCHED_UNIT_DRAFT = 'scheduling_unit_draft'
        SUBTASK = 'subtask'
        TASK_BLUEPRINT = 'task_blueprint'
        TASK_DRAFT = 'task_draft'
        RESERVATION = 'reservation'
        SUBSYSTEM = 'subsystem'

    def __init__(self, websocket_port: int=DEFAULT_WEBSOCKET_PORT):
        super().__init__(log_event_messages=True)
        self.websocket_port = websocket_port
        self._run_ws = True

    def start_handling(self):
        socket_started_event = Event()

        # Create and run a simple ws server
        def start_ws_server():
            logger.debug("starting websocket server on port %s", self.websocket_port)
            self._ws_server = SimpleWebSocketServer('', self.websocket_port, TMSSWebSocket)
            socket_started_event.set()
            logger.info("started websocket server on port %s", self.websocket_port)
            while self._run_ws: # Run the server till the stop_handling
                self._ws_server.serveonce()

        self.t = Thread(target=start_ws_server)
        self.t.start()
        if not socket_started_event.wait(10):
            raise RuntimeError("Could not start websocket server on port %s"%self.websocket_port)
        super().start_handling()

    def stop_handling(self):
        super().stop_handling()
        self._run_ws = False    # Stop the ws server
        self.t.join()

    def _get_authorised_clients_for_object_in_websocket(self, obj):
        from django.contrib.auth import get_user_model
        User = get_user_model()

        from lofar.sas.tmss.tmss.tmssapp.viewsets.permissions import get_project_roles_for_user, get_project_roles_with_permission
        from lofar.sas.tmss.tmss.tmssapp.models import ProjectRole

        auth_clients = []
        logger.info('Checking which of these users should receive websocket update for obj=%s: %s' % (obj, [ws.user for ws in list(self._ws_server.connections.values())]))
        for ws in list(self._ws_server.connections.values()):
            if not ws.authenticated:
                logger.info("%s websocket is not authenticated and will not receive websocket update for obj=%s" % (ws.user, obj))
                continue
            user = User.objects.filter(username=ws.user).first()
            if user is None:
                logger.info('User=%s does not exist in TMSS and will not receive websocket update for obj=%s' % (ws.user, obj))
                continue
            if user.is_superuser:
                logger.info('User=%s is superuser and will receive websocket update for obj=%s' % (user, obj))
                auth_clients.append(ws)
                continue
            if user.has_perm("tmssapp.view_%s" % type(obj).__name__.lower()):
                logger.info('User=%s has permission=%s and will receive websocket update for obj=%s' % (user, "tmssapp.view_%s" % type(obj).__name__.lower(), obj))
                auth_clients.append(ws)
                continue
            logger.info('User=%s has no permission=%s, checking for project-based permission to receive websocket update for obj=%s' % (user, "tmssapp.view_%s" % type(obj).__name__.lower(), obj))
            # project-based permission
            permitted_project_roles = get_project_roles_with_permission(type(obj).__name__.lower(), 'GET')
            user_project_roles = get_project_roles_for_user(user)
            related_project = getattr(obj, 'project', None)
            if related_project:
                related_project_name = related_project.name.lower()
                for project_role in user_project_roles:
                    if project_role['project'].lower() == related_project_name and \
                            ProjectRole.objects.get(value=project_role['role']) in permitted_project_roles:
                        auth_clients.append(ws)
                        logger.info("User=%s has project-based permission for project=%s and will receive websocket update for obj=%s" % (user, project_role['project'].lower(), obj))
                        break
        return auth_clients

    def _broadcast_notify_to_clients_websocket(self, msg, clients):
        # Send a broadcast message to all ws clients passed as argument
        json_msg = JSONdumps(msg)    # Stringify msg, so clients can parse it as JSON
        logger.debug("Sending json on websocket to %d clients: %s", len(clients), json_msg)
        for ws in clients:
            ws.sendMessage(json_msg)

    def _post_update_on_websocket(self, pk, object_type, action):

        # determine model class
        model_class = apps.get_model("tmssapp", object_type.value.replace('_', ''))

        # Prepare the json_blob_template
        json_blob = {'object_details': {}, 'object_type': object_type.value, 'action': action.value}
        # insert primary key name and value
        pk_name = model_class._meta.pk.attname
        json_blob['object_details'][pk_name] = pk

        auth_ws_clients = []

        if action == self.ObjActions.CREATE or action == self.ObjActions.UPDATE:

            # fetch fresh instance
            model_instance = model_class.objects.get(pk=pk)
            model_instance.refresh_from_db()

            # Get WS clients authorised to get object changes information
            auth_ws_clients = self._get_authorised_clients_for_object_in_websocket(model_instance)
            if len(auth_ws_clients) <= 0:  # There are no authorised clients, just return
                return

            try:
                # insert additional properties
                if hasattr(model_instance, 'process_start_time'):
                    json_blob['object_details']['process_start_time'] = model_instance.process_start_time.isoformat() if model_instance.process_start_time else None
                if hasattr(model_instance, 'process_stop_time'):
                    json_blob['object_details']['process_stop_time'] = model_instance.process_stop_time.isoformat() if model_instance.process_stop_time else None
                if hasattr(model_instance, 'on_sky_start_time'):
                    json_blob['object_details']['on_sky_start_time'] = model_instance.on_sky_start_time.isoformat() if model_instance.on_sky_start_time else None
                if hasattr(model_instance, 'on_sky_stop_time'):
                    json_blob['object_details']['on_sky_stop_time'] = model_instance.on_sky_stop_time.isoformat() if model_instance.on_sky_stop_time else None
                if hasattr(model_instance, 'start_time'):
                    json_blob['object_details']['start_time'] = model_instance.start_time.isoformat() if model_instance.start_time else None
                if hasattr(model_instance, 'stop_time'):
                    json_blob['object_details']['stop_time'] = model_instance.stop_time.isoformat() if model_instance.stop_time else None
                if hasattr(model_instance, 'duration'):
                    if isinstance(model_instance.duration, timedelta):
                        json_blob['object_details']['duration'] = model_instance.duration.total_seconds()
                    else:
                        json_blob['object_details']['duration'] = model_instance.duration or 0
                if hasattr(model_instance, 'on_sky_duration'):
                    if isinstance(model_instance.on_sky_duration, timedelta):
                        json_blob['object_details']['on_sky_duration'] = model_instance.on_sky_duration.total_seconds()
                    else:
                        json_blob['object_details']['on_sky_duration'] = model_instance.on_sky_duration or 0
                if hasattr(model_instance, 'status'):
                    json_blob['object_details']['status_value'] = model_instance.status.value
                    # We have status and status_value, where the former should be a url.
                    # Some endpoints (e.g. scheduling_unit_blueprint_slim) currently seem to break that scheme, and the
                    # frontend relies on the status attribute for e.g. the week view timeline. For now, we add this key
                    # to the websocket update as well:
                    # todo: use status_value instead of status consistently and remove the status key from the update.
                    json_blob['object_details']['status'] = model_instance.status.value
                if hasattr(model_instance, 'state'):
                    json_blob['object_details']['state_value'] = model_instance.state.value
            except Exception as e:
                logger.error("Cannot get object details for %s: %s", json_blob, e)

        if action == self.ObjActions.DELETE:
            # The item has been deleted from the database already and we cannot determine who should get the info.
            # Inform everyone.
            auth_ws_clients = list(self._ws_server.connections.values())

        # Send the json_blob as a broadcast message to all connected and authorised WS clients
        self._broadcast_notify_to_clients_websocket(json_blob, auth_ws_clients)

    def onSubTaskCreated(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.SUBTASK, self.ObjActions.CREATE)

    def onSubTaskUpdated(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.SUBTASK, self.ObjActions.UPDATE)

    def onSubTaskDeleted(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.SUBTASK, self.ObjActions.DELETE)

    def onTaskDraftCreated(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.TASK_DRAFT, self.ObjActions.CREATE)

    def onTaskDraftUpdated(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.TASK_DRAFT, self.ObjActions.UPDATE)

    def onTaskDraftDeleted(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.TASK_DRAFT, self.ObjActions.DELETE)

    def onTaskBlueprintCreated(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.TASK_BLUEPRINT, self.ObjActions.CREATE)

    def onTaskBlueprintUpdated(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.TASK_BLUEPRINT, self.ObjActions.UPDATE)

    def onTaskBlueprintDeleted(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.TASK_BLUEPRINT, self.ObjActions.DELETE)

    def onSchedulingUnitDraftCreated(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.SCHED_UNIT_DRAFT, self.ObjActions.CREATE)

    def onSchedulingUnitDraftUpdated(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.SCHED_UNIT_DRAFT, self.ObjActions.UPDATE)

    def onSchedulingUnitDraftDeleted(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.SCHED_UNIT_DRAFT, self.ObjActions.DELETE)

    def onSchedulingUnitBlueprintCreated(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.SCHED_UNIT_BLUEPRINT, self.ObjActions.CREATE)

    def onSchedulingUnitBlueprintUpdated(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.SCHED_UNIT_BLUEPRINT, self.ObjActions.UPDATE)

    def onSchedulingUnitBlueprintDeleted(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.SCHED_UNIT_BLUEPRINT, self.ObjActions.DELETE)

    def onReservationCreated(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.RESERVATION, self.ObjActions.CREATE)

    def onReservationUpdated(self, id: int):
        self._post_update_on_websocket(id, self.ObjTypes.RESERVATION, self.ObjActions.UPDATE)

    def onReservationDeleted(self, id: int, start_time, stop_time):
        self._post_update_on_websocket(id, self.ObjTypes.RESERVATION, self.ObjActions.DELETE)

    def onSubsystemStatusUpdated(self, name: str, status: str):
        self._post_update_on_websocket(name, self.ObjTypes.SUBSYSTEM, self.ObjActions.UPDATE)


def create_service(websocket_port: int=DEFAULT_WEBSOCKET_PORT, exchange: str=DEFAULT_BUSNAME, broker: str=DEFAULT_BROKER):
    return TMSSBusListener(handler_type=TMSSEventMessageHandlerForWebsocket,
                           handler_kwargs={'websocket_port': websocket_port},
                           exchange=exchange, broker=broker)


def main():
    # make sure we run in UTC timezone
    os.environ['TZ'] = 'UTC'

    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

    # Check the invocation arguments
    parser = OptionParser('%prog [options]',
                          description='run the tmss_websocket_service which listens for TMSS event messages on the messagebus, and posts the updates on the websocket for htpp clients.')

    parser.add_option('-w', '--websocket_port', dest='websocket_port', type='int', default=find_free_port(DEFAULT_WEBSOCKET_PORT),
                      help='The port on which the websocket clients can connect, default: %default')

    group = OptionGroup(parser, 'Messaging options')
    group.add_option('-b', '--broker', dest='broker', type='string', default=DEFAULT_BROKER,
                     help='Address of the message broker, default: %default')
    group.add_option('-e', "--exchange", dest="exchange", type="string", default=DEFAULT_BUSNAME,
                     help="exchange where the TMSS event messages are published. [default: %default]")
    parser.add_option_group(group)

    group = OptionGroup(parser, 'Django options')
    parser.add_option_group(group)
    group.add_option('-C', '--credentials', dest='dbcredentials', type='string', default=os.environ.get('TMSS_DBCREDENTIALS', 'TMSS'), help='django dbcredentials name, default: %default')

    (options, args) = parser.parse_args()

    from lofar.sas.tmss.tmss import setup_and_check_tmss_django_database_connection_and_exit_on_error
    setup_and_check_tmss_django_database_connection_and_exit_on_error(options.dbcredentials)

    with create_service(options.websocket_port, options.exchange, options.broker):
        waitForInterrupt()

if __name__ == '__main__':
    main()
