#!/usr/bin/env python3

# workflow_service.py
#
# Copyright (C) 2015
# ASTRON (Netherlands Institute for Radio Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.
#

import os
import logging
logger = logging.getLogger(__name__)

from lofar.sas.tmss.client.tmssbuslistener import *

class SchedulingUnitEventMessageHandler(TMSSEventMessageHandler):

    def onSchedulingUnitBlueprintStatusChanged(self, id: int, status: str):
        try:
            # import here and not at top of module because we need the django.setup() to be run first, either from this module's main, or from the TMSSTestEnvironment
            from lofar.sas.tmss.tmss.workflowapp.signals import scheduling_unit_blueprint_schedulable_signal, scheduling_unit_blueprint_scheduled_signal, scheduling_unit_blueprint_observing_signal, scheduling_unit_blueprint_observed_signal, scheduling_unit_blueprint_ingested_signal, scheduling_unit_blueprint_finished_signal
            from lofar.sas.tmss.tmss.workflowapp.models import SchedulingUnitTask
            from lofar.sas.tmss.tmss.tmssapp.models import SchedulingUnitBlueprint, SchedulingUnitStatus, TaskType, SubtaskType

            logger.info("workflow onSchedulingUnitBlueprintStatusChanged: id=%s status changed to '%s'", id, status)
            scheduling_unit_blueprint = SchedulingUnitBlueprint.objects.get(pk=id)
            if status == SchedulingUnitStatus.Choices.SCHEDULABLE.value:
                logger.info("workflow scheduling_unit_blueprint id=%s is schedulable, signalling workflow...", id)
                scheduling_unit_blueprint_schedulable_signal.send(sender=self.__class__, scheduling_unit_blueprint=scheduling_unit_blueprint)
            elif status == SchedulingUnitStatus.Choices.SCHEDULED.value:
                logger.info("workflow scheduling_unit_blueprint id=%s is scheduled, signalling workflow...", id)
                scheduling_unit_blueprint_scheduled_signal.send(sender=self.__class__, scheduling_unit_blueprint=scheduling_unit_blueprint)
            elif status == SchedulingUnitStatus.Choices.OBSERVING.value:
                logger.info("workflow scheduling_unit_blueprint id=%s is observing, signalling workflow...", id)
                scheduling_unit_blueprint_observing_signal.send(sender=self.__class__, scheduling_unit_blueprint=scheduling_unit_blueprint)
            elif status == SchedulingUnitStatus.Choices.OBSERVED.value:
                logger.info("workflow scheduling_unit_blueprint id=%s is fully observed, signalling workflow...", id)
                scheduling_unit_blueprint_observed_signal.send(sender=self.__class__, scheduling_unit_blueprint=scheduling_unit_blueprint)
            elif status == SchedulingUnitStatus.Choices.INGESTED.value:
                logger.info("workflow scheduling_unit_blueprint id=%s is fully ingested, signalling workflow...", id)
                scheduling_unit_blueprint_ingested_signal.send(sender=self.__class__, scheduling_unit_blueprint=scheduling_unit_blueprint)
            elif status == SchedulingUnitStatus.Choices.FINISHED.value:
                logger.info("workflow scheduling_unit_blueprint id=%s is fully finished, signalling workflow...", id)
                scheduling_unit_blueprint_finished_signal.send(sender=self.__class__, scheduling_unit_blueprint=scheduling_unit_blueprint)
        except Exception as e:
            logger.exception("Exception in workflow onSchedulingUnitBlueprintStatusChanged: %s" % (str(e),))




def create_workflow_service(handler_type = SchedulingUnitEventMessageHandler, exchange: str=DEFAULT_BUSNAME, broker: str=DEFAULT_BROKER):
    return TMSSBusListener(handler_type=handler_type,
                           handler_kwargs={},
                           exchange=exchange, broker=broker)

def main():
    # make sure we run in UTC timezone
    os.environ['TZ'] = 'UTC'

    # enable viewflow in TMSS for this service
    os.environ['TMSS_ENABLE_VIEWFLOW'] = 'True'

    from optparse import OptionParser, OptionGroup
    from lofar.common import dbcredentials

    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

    # Check the invocation arguments
    parser = OptionParser('%prog [options]',
                          description='run the tmss_workflow_service which forwards TMSS events to the workflow engine.')

    group = OptionGroup(parser, 'Messaging options')
    group.add_option('-b', '--broker', dest='broker', type='string', default=DEFAULT_BROKER,
                     help='Address of the message broker, default: %default')
    group.add_option('-e', "--exchange", dest="exchange", type="string", default=DEFAULT_BUSNAME,
                     help="Bus or queue where the TMSS messages are published. [default: %default]")
    parser.add_option_group(group)

    group = OptionGroup(parser, 'Django options')
    parser.add_option_group(group)
    group.add_option('-C', '--credentials', dest='dbcredentials', type='string', default=os.environ.get('TMSS_DBCREDENTIALS', 'TMSS'), help='django dbcredentials name, default: %default')

    (options, args) = parser.parse_args()

    from lofar.sas.tmss.tmss import setup_and_check_tmss_django_database_connection_and_exit_on_error
    setup_and_check_tmss_django_database_connection_and_exit_on_error(options.dbcredentials)

    with create_workflow_service(exchange=options.exchange, broker=options.broker):
        waitForInterrupt()

if __name__ == '__main__':
    main()
