import unittest
import requests
import json
import uuid

import logging
logger = logging.getLogger('lofar.'+__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

from lofar.common.test_utils import skip_integration_tests
if skip_integration_tests():
    exit(3)

from lofar.messaging.messagebus import TemporaryExchange, BusListenerJanitor

TEST_UUID = uuid.uuid1()
tmp_exchange = TemporaryExchange("%s_%s" % (__name__, TEST_UUID))
tmp_exchange.open()
# override DEFAULT_BUSNAME
import lofar
lofar.messaging.config.DEFAULT_BUSNAME = tmp_exchange.address
# import here, and not at top of module, because DEFAULT_BUSNAME needs to be set before importing
from lofar.sas.tmss.test.test_environment import TMSSTestEnvironment
tmss_test_env = TMSSTestEnvironment(exchange=tmp_exchange.address, populate_schemas=True, populate_test_data=False,
                                        start_subtask_scheduler=False, start_postgres_listener=True,
                                        start_dynamic_scheduler=False, enable_viewflow=True, start_workflow_service=False,
                                        populate_permissions=True)
tmss_test_env.start()

def tearDownModule():
    tmss_test_env.stop()
    tmp_exchange.close()

from lofar.sas.tmss.services.workflow_service import create_workflow_service, SchedulingUnitEventMessageHandler

from time import sleep
from datetime import timedelta
from threading import Event
from lofar.sas.tmss.client.tmssbuslistener import *


class SchedulingUnitFlowTest(unittest.TestCase):
    sync_event_bp_scheduled = Event()
    sync_event_bp_observed = Event()

    class TestSchedulingUnitEventMessageHandler(SchedulingUnitEventMessageHandler):
        def onSchedulingUnitBlueprintStatusChanged(self, id: int, status: str):
            try:
                super().onSchedulingUnitBlueprintStatusChanged(id=id, status=status)
            except Exception as e:
                logger.exception(e)

            from lofar.sas.tmss.tmss.tmssapp.models import SchedulingUnitStatus
            if status == SchedulingUnitStatus.Choices.SCHEDULED.value:
                logging.info("Status is %s, sending sync event", status)
                SchedulingUnitFlowTest.sync_event_bp_scheduled.set()
            elif status == SchedulingUnitStatus.Choices.OBSERVED.value:
                logging.info("Scheduling Unit Blueprint with id %s was fully observed, sending sync event", id)
                SchedulingUnitFlowTest.sync_event_bp_observed.set()

    def setUp(self) -> None:
        logger.info('---------------------------------------------------------------------------------')
        self.sync_event_bp_scheduled.clear()
        self.sync_event_bp_observed.clear()

    @classmethod
    def setUpClass(cls) -> None:
        cls.TEST_UUID = TEST_UUID

        cls.BASE_URL_WF_API = tmss_test_env.django_server.url.rstrip('/').replace('/api','/workflow_api')
        cls.AUTH = requests.auth.HTTPBasicAuth(tmss_test_env.ldap_server.dbcreds.user, tmss_test_env.ldap_server.dbcreds.password)

    def test_qa_workflow_complete(self):
        from lofar.sas.tmss.tmss.tmssapp import models
        from lofar.sas.tmss.tmss.tmssapp.models import TaskType

        from lofar.sas.tmss.tmss.tmssapp.tasks import create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft, update_task_graph_from_specifications_doc
        from lofar.sas.tmss.test.tmss_test_data_django_models import SchedulingSet_test_data

        from lofar.sas.tmss.tmss.workflowapp.models.schedulingunitflow import SchedulingUnitProcess
        from lofar.sas.tmss.tmss.workflowapp.models.schedulingunitflow import SchedulingUnitTask as Task

        service = create_workflow_service(handler_type=SchedulingUnitFlowTest.TestSchedulingUnitEventMessageHandler, exchange=tmp_exchange.address)
        with BusListenerJanitor(service):
            strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_version_or_latest(name="IM HBA LoTSS - 2 Beams")
            # ensure QA subtasks are enabled for this test
            for obs_task_name in ('Calibrator Observation 1', 'Calibrator Observation 2', 'Target Observation'):
                for qa_subtask_type in ('plots', 'file_conversion'):
                    strategy_template.template['tasks'][obs_task_name]['specifications_doc']['QA'][qa_subtask_type]['enabled'] = True
            strategy_template.save()


            scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(
                                            name="Test Scheduling Unit UC1",
                                            specifications_template=strategy_template.scheduling_unit_template,
                                            observation_strategy_template=strategy_template,
                                            scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))

            update_task_graph_from_specifications_doc(scheduling_unit_draft, strategy_template.template)
            scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
            scheduling_unit_draft.refresh_from_db()

            # there is no signal that SchedulingUnitProcess instance was created,
            # so we have to wait an poll before we can proceed with the test
            poll_starttime = datetime.utcnow()
            while True:
                if SchedulingUnitProcess.objects.filter(su=scheduling_unit_blueprint).count() > 0:
                    break
                sleep(0.5)
                if datetime.utcnow()-poll_starttime > timedelta(seconds=30):
                    raise TimeoutError("SchedulingUnitProcess not created within expected time")

            # Yes! the SchedulingUnitProcess was created, let's get it.
            scheduling_unit_process_id = SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).id

            prev_ingest_permission_granted_since = models.SchedulingUnitBlueprint.objects.get(pk=scheduling_unit_process_id).ingest_permission_granted_since
            self.assertEqual(prev_ingest_permission_granted_since, None)

            #check the active task name
            self.assertEqual("wait_scheduled", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[0].flow_task.name, 'start')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[0].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].flow_task.name, 'wait_scheduled')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].status, 'NEW')

            # Change observation subtask status to scheduled
            from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions
            for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
                if task_blueprint.specifications_template.type.value == TaskType.Choices.OBSERVATION.value:
                    for subtask in task_blueprint.subtasks.all():
                        set_subtask_state_following_allowed_transitions(subtask, 'scheduled')

            # wait until scheduling unit is scheduled
            logging.info("waiting for sync_event_bp_scheduled...")
            if not self.sync_event_bp_scheduled.wait(timeout=10):
                logging.info("sync_event_bp_scheduled event not received, raising TimeoutError")
                raise TimeoutError()
            else:
                logging.info("Received self.sync_event_bp_scheduled event")
                poll_starttime = datetime.utcnow()
                while True:
                    if SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name == "wait_observing":
                        break
                    sleep(0.5)
                    if datetime.utcnow()-poll_starttime > timedelta(seconds=30):
                        raise TimeoutError("Task not activated within expected time")

            self.sync_event_bp_scheduled.clear()

            #check the active task name
            self.assertEqual("wait_observing", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].flow_task.name, 'wait_scheduled')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].flow_task.name, 'wait_observing')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].status, 'NEW')

            # finish the observations
            for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
                if task_blueprint.specifications_template.type.value == TaskType.Choices.OBSERVATION.value:
                    for subtask in task_blueprint.subtasks.all():
                        set_subtask_state_following_allowed_transitions(subtask, 'finished')

            logging.info("waiting for sync_event_bp_observed...")
            if not self.sync_event_bp_observed.wait(timeout=10):
                logging.info("sync_event_bp_observed event not received, raising TimeoutError")
                raise TimeoutError()
            else:
                logging.info("Received self.sync_event_bp_observed event")
                poll_starttime = datetime.utcnow()
                while True:
                    if SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name == "qa_reporting_to":
                        break
                    sleep(0.5)
                    if datetime.utcnow()-poll_starttime > timedelta(seconds=30):
                        raise TimeoutError("Task not activated within expected time")

            self.sync_event_bp_observed.clear()

            #check the active task name
            self.assertEqual("qa_reporting_to", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].flow_task.name, 'wait_observing')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'NEW')

            # schedule, run, finish the pipelines and ingest et al
            for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
                if task_blueprint.specifications_template.type.value != TaskType.Choices.OBSERVATION.value:
                    for subtask in task_blueprint.subtasks.all():
                        set_subtask_state_following_allowed_transitions(subtask, 'finished')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['id']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'ASSIGNED')

            #API: Perform qa_reporting_to step
            headers = {'content-type': 'application/json'}
            data = '{"operator_report": "Test report", "operator_accept": true}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[4].flow_task.name, 'check_operator_accept')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[4].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].flow_task.name, 'qa_reporting_sos')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].status, 'NEW')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['id']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            #API: Perform qa_reporting_sos step
            headers = {'content-type': 'application/json'}
            data = '{"sos_report": "Test report", "quality_within_policy": true, "sos_accept_show_pi": true}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].flow_task.name, 'qa_reporting_sos')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[6].flow_task.name, 'check_sos_accept_show_pi')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[6].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].flow_task.name, 'pi_verification')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].status, 'NEW')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['id']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            #API: Perform pi_verification step
            headers = {'content-type': 'application/json'}
            data = '{"pi_report": "Test report", "pi_accept": true}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].flow_task.name, 'pi_verification')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].flow_task.name, 'decide_acceptance')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].status, 'NEW')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['id']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            #API: Perform decide_acceptance step
            headers = {'content-type': 'application/json'}
            data = '{"sos_accept_after_pi": true}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].flow_task.name, 'decide_acceptance')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[9].flow_task.name, 'check_sos_accept_after_pi')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[9].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[10].flow_task.name, 'mark_sub_accepted')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[10].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[11].flow_task.name, 'allow_ingest')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[11].status, 'DONE')

            #verify that ingest_permission_granted_since is now a valid datetime
            ingest_permission_granted_since = models.SchedulingUnitBlueprint.objects.get(pk=scheduling_unit_process_id).ingest_permission_granted_since
            self.assertEqual(True,isinstance(ingest_permission_granted_since, datetime))

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[12].flow_task.name, 'wait_ingest_done')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[12].status, 'DONE')

            output_pinned = models.SchedulingUnitBlueprint.objects.get(pk=scheduling_unit_process_id).output_pinned
            # self.assertEqual(True,output_pinned) # todo: progress to check_cleanup_task_present when implemented

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[13].flow_task.name, 'check_data_pinned')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[13].status, 'DONE')

            # todo: progress to check_cleanup_task_present when implemented
            #       remove these two task asserts when implemented and adapt commented asserts properly.
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[14].flow_task.name, 'wait_finished')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[14].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[15].flow_task.name, 'end')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[15].status, 'DONE')
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[14].flow_task.name, 'decide_unpin_data')
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[14].status, 'NEW')
            #
            # #API: Get current task
            # data = ""
            # response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            # content = response.content.decode('utf-8')
            # r_dict = json.loads(content)
            # task_id=r_dict[0]['id']
            #
            # #API: Assign to an user
            # data = ""
            # response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            # content = response.content.decode('utf-8')
            # r_dict = json.loads(content)
            # self.assertEqual(200,response.status_code)
            #
            # #API: Perform decide_acceptance step
            # headers = {'content-type': 'application/json'}
            # data = '{"unpin_data": true}'
            # response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            # content = response.content.decode('utf-8')
            # r_dict = json.loads(content)
            # self.assertEqual(200,response.status_code)
            #
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[14].flow_task.name, 'unpin_data')
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[14].status, 'DONE')
            #
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[15].flow_task.name, 'delete_data')
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[15].status, 'DONE')
            #
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[16].flow_task.name, 'end')
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[16].status, 'DONE')

            # # assert we can filter flow_task
            # response = requests.get(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/?flow_task=Unpin%20Data,Decide%20Acceptance,Mark%20Sub', auth=self.__class__.AUTH)
            # content = response.content.decode('utf-8')
            # r_dict = json.loads(content)
            # self.assertEqual(r_dict['count'], 3)

    def test_qa_workflow_qa_reporting_to_no(self):
        from lofar.sas.tmss.tmss.tmssapp import models
        from lofar.sas.tmss.tmss.tmssapp.models import TaskType

        from lofar.sas.tmss.tmss.tmssapp.tasks import create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft, update_task_graph_from_specifications_doc
        from lofar.sas.tmss.test.tmss_test_data_django_models import SchedulingSet_test_data

        from lofar.sas.tmss.tmss.workflowapp.models.schedulingunitflow import SchedulingUnitProcess
        from lofar.sas.tmss.tmss.workflowapp.models.schedulingunitflow import SchedulingUnitTask as Task

        service = create_workflow_service(handler_type=SchedulingUnitFlowTest.TestSchedulingUnitEventMessageHandler, exchange=tmp_exchange.address)
        with BusListenerJanitor(service):

            strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_version_or_latest(name="IM HBA LoTSS - 2 Beams")

            scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(
                                            name="Test Scheduling Unit UC1",
                                            specifications_template=strategy_template.scheduling_unit_template,
                                            observation_strategy_template=strategy_template,
                                            scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))
            update_task_graph_from_specifications_doc(scheduling_unit_draft, strategy_template.template)

            scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
            scheduling_unit_draft.refresh_from_db()

            # there is no signal that SchedulingUnitProcess instance was created,
            # so we have to wait an poll before we can proceed with the test
            poll_starttime = datetime.utcnow()
            while True:
                if SchedulingUnitProcess.objects.filter(su=scheduling_unit_blueprint).count() > 0:
                    break
                sleep(0.5)
                if datetime.utcnow()-poll_starttime > timedelta(seconds=30):
                    raise TimeoutError("SchedulingUnitProcess not created within expected time")

            # Yes! the SchedulingUnitProcess was created, let's get it.
            scheduling_unit_process_id = SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).id

            prev_ingest_permission_granted_since = models.SchedulingUnitBlueprint.objects.get(pk=scheduling_unit_process_id).ingest_permission_granted_since
            self.assertEqual(prev_ingest_permission_granted_since, None)


            self.assertEqual("wait_scheduled", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[0].flow_task.name, 'start')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[0].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].flow_task.name, 'wait_scheduled')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].status, 'NEW')


            # Change observation subtask status to scheduled
            from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions
            for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
                if task_blueprint.specifications_template.type.value == TaskType.Choices.OBSERVATION.value:
                    for subtask in task_blueprint.subtasks.all():
                        set_subtask_state_following_allowed_transitions(subtask, 'scheduled')

            # wait until scheduling unit is scheduled
            logging.info("waiting for sync_event_bp_scheduled...")
            if not self.sync_event_bp_scheduled.wait(timeout=10):
                logging.info("sync_event_bp_scheduled event not received, raising TimeoutError")
                raise TimeoutError()
            else:
                logging.info("Received self.sync_event_bp_scheduled event")
                poll_starttime = datetime.utcnow()
                while True:
                    if SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name == "wait_observing":
                        break
                    sleep(0.5)
                    if datetime.utcnow()-poll_starttime > timedelta(seconds=30):
                        raise TimeoutError("Task not activated within expected time")

            self.sync_event_bp_scheduled.clear()


            #check the active task name
            self.assertEqual("wait_observing", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].flow_task.name, 'wait_scheduled')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].flow_task.name, 'wait_observing')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].status, 'NEW')

            # Change observation subtask status to scheduled
            from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions
            for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
                if task_blueprint.specifications_template.type.value == TaskType.Choices.OBSERVATION.value:
                    task_blueprint.output_pinned = True
                    task_blueprint.save()

                    for subtask in task_blueprint.subtasks.all():
                        set_subtask_state_following_allowed_transitions(subtask, 'finished')

            logging.info("waiting for sync_event_bp_observed...")
            if not self.sync_event_bp_observed.wait(timeout=10):
                logging.info("sync_event_bp_observed event not received, raising TimeoutError")
                raise TimeoutError()
            else:
                logging.info("Received self.sync_event_bp_observed event")
                poll_starttime = datetime.utcnow()
                while True:
                    if SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name == "qa_reporting_to":
                        break
                    sleep(0.5)
                    if datetime.utcnow()-poll_starttime > timedelta(seconds=30):
                        raise TimeoutError("Task not activated within expected time")

            self.sync_event_bp_observed.clear()


            #check the active task name
            self.assertEqual("qa_reporting_to", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].flow_task.name, 'wait_observing')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'NEW')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['id']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'ASSIGNED')

            #API: Perform qa_reporting_to step
            headers = {'content-type': 'application/json'}
            data = '{"operator_report": "Test report", "operator_accept": false}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[4].flow_task.name, 'check_operator_accept')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[4].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].flow_task.name, 'mark_sub_not_accepted')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[6].flow_task.name, 'check_data_pinned')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[6].status, 'DONE')

            # todo: progress to check_cleanup_task_present when implemented
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].flow_task.name, 'unpin_data')
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].status, 'NEW')
            #
            #
            # #API: Get current task
            # data = ""
            # response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            # content = response.content.decode('utf-8')
            # r_dict = json.loads(content)
            # task_id=r_dict[0]['id']
            #
            # #API: Assign to an user
            # data = ""
            # response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            # content = response.content.decode('utf-8')
            # r_dict = json.loads(content)
            # self.assertEqual(200,response.status_code)
            #
            # #API: Perform decide_acceptance step
            # headers = {'content-type': 'application/json'}
            # data = '{"unpin_data": true}'
            # response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            # content = response.content.decode('utf-8')
            # r_dict = json.loads(content)
            # self.assertEqual(200,response.status_code)
            #
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].flow_task.name, 'unpin_data')
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].status, 'DONE')
            #
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].flow_task.name, 'delete_data')
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].status, 'DONE')
            #
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[9].flow_task.name, 'end')
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[9].status, 'DONE')

    def test_qa_workflow_qa_reporting_sos_no(self):
        from lofar.sas.tmss.tmss.tmssapp import models
        from lofar.sas.tmss.tmss.tmssapp.models import TaskType

        from lofar.sas.tmss.tmss.tmssapp.tasks import create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft, update_task_graph_from_specifications_doc
        from lofar.sas.tmss.test.tmss_test_data_django_models import SchedulingSet_test_data

        from lofar.sas.tmss.tmss.workflowapp.models.schedulingunitflow import SchedulingUnitProcess
        from lofar.sas.tmss.tmss.workflowapp.models.schedulingunitflow import SchedulingUnitTask as Task

        service = create_workflow_service(handler_type=SchedulingUnitFlowTest.TestSchedulingUnitEventMessageHandler, exchange=tmp_exchange.address)
        with BusListenerJanitor(service):
            strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_version_or_latest(name="IM HBA LoTSS - 2 Beams")

            scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(
                                            name="Test Scheduling Unit UC1",
                                            specifications_template=strategy_template.scheduling_unit_template,
                                            observation_strategy_template=strategy_template,
                                            scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))

            update_task_graph_from_specifications_doc(scheduling_unit_draft, strategy_template.template)

            scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
            scheduling_unit_draft.refresh_from_db()

            # there is no signal that SchedulingUnitProcess instance was created,
            # so we have to wait an poll before we can proceed with the test
            poll_starttime = datetime.utcnow()
            while True:
                if SchedulingUnitProcess.objects.filter(su=scheduling_unit_blueprint).count() > 0:
                    break
                sleep(0.5)
                if datetime.utcnow()-poll_starttime > timedelta(seconds=30):
                    raise TimeoutError("SchedulingUnitProcess not created within expected time")

            # Yes! the SchedulingUnitProcess was created, let's get it.
            scheduling_unit_process_id = SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).id

            prev_ingest_permission_granted_since = models.SchedulingUnitBlueprint.objects.get(pk=scheduling_unit_process_id).ingest_permission_granted_since
            self.assertEqual(prev_ingest_permission_granted_since, None)

            #check the active task name
            self.assertEqual("wait_scheduled", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[0].flow_task.name, 'start')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[0].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].flow_task.name, 'wait_scheduled')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].status, 'NEW')


            # Change observation subtask status to scheduled
            from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions
            for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
                if task_blueprint.specifications_template.type.value == TaskType.Choices.OBSERVATION.value:
                    for subtask in task_blueprint.subtasks.all():
                        set_subtask_state_following_allowed_transitions(subtask, 'scheduled')

            # wait until scheduling unit is scheduled
            logging.info("waiting for sync_event_bp_scheduled...")
            if not self.sync_event_bp_scheduled.wait(timeout=10):
                logging.info("sync_event_bp_scheduled event not received, raising TimeoutError")
                raise TimeoutError()
            else:
                logging.info("Received self.sync_event_bp_scheduled event")
                poll_starttime = datetime.utcnow()
                while True:
                    if SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name == "wait_observing":
                        break
                    sleep(0.5)
                    if datetime.utcnow()-poll_starttime > timedelta(seconds=30):
                        raise TimeoutError("Task not activated within expected time")

            self.sync_event_bp_scheduled.clear()


            #check the active task name
            self.assertEqual("wait_observing", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].flow_task.name, 'wait_scheduled')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].flow_task.name, 'wait_observing')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].status, 'NEW')

            # Change observation subtask status to scheduled
            from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions
            for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
                if task_blueprint.specifications_template.type.value == TaskType.Choices.OBSERVATION.value:
                    task_blueprint.output_pinned = True
                    task_blueprint.save()

                    for subtask in task_blueprint.subtasks.all():
                        set_subtask_state_following_allowed_transitions(subtask, 'finished')

            logging.info("waiting for sync_event_bp_observed...")
            if not self.sync_event_bp_observed.wait(timeout=10):
                logging.info("sync_event_bp_observed event not received, raising TimeoutError")
                raise TimeoutError()
            else:
                logging.info("Received self.sync_event_bp_observed event")
                poll_starttime = datetime.utcnow()
                while True:
                    if SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name == "qa_reporting_to":
                        break
                    sleep(0.5)
                    if datetime.utcnow()-poll_starttime > timedelta(seconds=30):
                        raise TimeoutError("Task not activated within expected time")

            self.sync_event_bp_observed.clear()

            #check the active task name
            self.assertEqual("qa_reporting_to", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].flow_task.name, 'wait_observing')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'NEW')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['id']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'ASSIGNED')

            #API: Perform qa_reporting_to step
            headers = {'content-type': 'application/json'}
            data = '{"operator_report": "Test report", "operator_accept": true}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[4].flow_task.name, 'check_operator_accept')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[4].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].flow_task.name, 'qa_reporting_sos')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].status, 'NEW')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['id']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            #API: Perform qa_reporting_sos step
            headers = {'content-type': 'application/json'}
            data = '{"sos_report": "Test report", "quality_within_policy": true, "sos_accept_show_pi": false}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].flow_task.name, 'qa_reporting_sos')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[6].flow_task.name, 'check_sos_accept_show_pi')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[6].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].flow_task.name, 'mark_sub_not_accepted')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].status, 'DONE')

            output_pinned = models.SchedulingUnitBlueprint.objects.get(pk=scheduling_unit_process_id).output_pinned
            #self.assertEqual(True,output_pinned)  # todo: progress to check_cleanup_task_present when implemented

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].flow_task.name, 'check_data_pinned')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].status, 'DONE')

            # todo: progress to check_cleanup_task_present when implemented
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[9].flow_task.name, 'unpin_data')
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[9].status, 'NEW')
            #
            # #API: Get current task
            # data = ""
            # response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            # content = response.content.decode('utf-8')
            # r_dict = json.loads(content)
            # task_id=r_dict[0]['id']
            #
            # #API: Assign to an user
            # data = ""
            # response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            # content = response.content.decode('utf-8')
            # r_dict = json.loads(content)
            # self.assertEqual(200,response.status_code)
            #
            # #API: Perform decide_acceptance step
            # headers = {'content-type': 'application/json'}
            # data = '{"unpin_data": true}'
            # response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            # content = response.content.decode('utf-8')
            # r_dict = json.loads(content)
            # self.assertEqual(200,response.status_code)
            #
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[9].flow_task.name, 'unpin_data')
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[9].status, 'DONE')
            #
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[10].flow_task.name, 'delete_data')
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[10].status, 'DONE')
            #
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[11].flow_task.name, 'end')
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[11].status, 'DONE')

    def test_qa_workflow_qa_quality_acceptable_no(self):
        from lofar.sas.tmss.tmss.tmssapp import models
        from lofar.sas.tmss.tmss.tmssapp.models import TaskType

        from lofar.sas.tmss.tmss.tmssapp.tasks import create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft, update_task_graph_from_specifications_doc
        from lofar.sas.tmss.test.tmss_test_data_django_models import SchedulingSet_test_data

        from lofar.sas.tmss.tmss.workflowapp.models.schedulingunitflow import SchedulingUnitProcess
        from lofar.sas.tmss.tmss.workflowapp.models.schedulingunitflow import SchedulingUnitTask as Task

        service = create_workflow_service(handler_type=SchedulingUnitFlowTest.TestSchedulingUnitEventMessageHandler, exchange=tmp_exchange.address)
        with BusListenerJanitor(service):
            strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_version_or_latest(name="IM HBA LoTSS - 2 Beams")

            scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(
                                            name="Test Scheduling Unit UC1",
                                            specifications_template=strategy_template.scheduling_unit_template,
                                            observation_strategy_template=strategy_template,
                                            scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))
            update_task_graph_from_specifications_doc(scheduling_unit_draft, strategy_template.template)

            scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
            scheduling_unit_draft.refresh_from_db()

            # there is no signal that SchedulingUnitProcess instance was created,
            # so we have to wait an poll before we can proceed with the test
            poll_starttime = datetime.utcnow()
            while True:
                if SchedulingUnitProcess.objects.filter(su=scheduling_unit_blueprint).count() > 0:
                    break
                sleep(0.5)
                if datetime.utcnow()-poll_starttime > timedelta(seconds=60):
                    raise TimeoutError("SchedulingUnitProcess not created within expected time")

            # Yes! the SchedulingUnitProcess was created, let's get it.
            scheduling_unit_process_id = SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).id

            prev_ingest_permission_granted_since = models.SchedulingUnitBlueprint.objects.get(pk=scheduling_unit_process_id).ingest_permission_granted_since
            self.assertEqual(prev_ingest_permission_granted_since, None)

            #check the active task name
            self.assertEqual("wait_scheduled", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[0].flow_task.name, 'start')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[0].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].flow_task.name, 'wait_scheduled')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].status, 'NEW')


            # Change observation subtask status to scheduled
            from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions
            for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
                if task_blueprint.specifications_template.type.value == TaskType.Choices.OBSERVATION.value:
                    for subtask in task_blueprint.subtasks.all():
                        set_subtask_state_following_allowed_transitions(subtask, 'scheduled')

            # wait until scheduling unit is scheduled
            logging.info("waiting for sync_event_bp_scheduled...")
            if not self.sync_event_bp_scheduled.wait(timeout=10):
                logging.info("sync_event_bp_scheduled event not received, raising TimeoutError")
                raise TimeoutError()
            else:
                logging.info("Received self.sync_event_bp_scheduled event")
                poll_starttime = datetime.utcnow()
                while True:
                    if SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name == "wait_observing":
                        break
                    sleep(0.5)
                    if datetime.utcnow()-poll_starttime > timedelta(seconds=30):
                        raise TimeoutError("Task not activated within expected time")

            self.sync_event_bp_scheduled.clear()


            #check the active task name
            self.assertEqual("wait_observing", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].flow_task.name, 'wait_scheduled')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].flow_task.name, 'wait_observing')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].status, 'NEW')

            # Change observation subtask status to scheduled
            from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions
            for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
                if task_blueprint.specifications_template.type.value == TaskType.Choices.OBSERVATION.value:
                    task_blueprint.output_pinned = True
                    task_blueprint.save()

                    for subtask in task_blueprint.subtasks.all():
                        set_subtask_state_following_allowed_transitions(subtask, 'finished')

            logging.info("waiting for sync_event_bp_observed...")
            if not self.sync_event_bp_observed.wait(timeout=10):
                logging.info("sync_event_bp_observed event not received, raising TimeoutError")
                raise TimeoutError()
            else:
                logging.info("Received self.sync_event_bp_observed event")
                poll_starttime = datetime.utcnow()
                while True:
                    if SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name == "qa_reporting_to":
                        break
                    sleep(0.5)
                    if datetime.utcnow()-poll_starttime > timedelta(seconds=30):
                        raise TimeoutError("Task not activated within expected time")

            self.sync_event_bp_observed.clear()

            #check the active task name
            self.assertEqual("qa_reporting_to", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].flow_task.name, 'wait_observing')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'NEW')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['id']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'ASSIGNED')

            #API: Perform qa_reporting_to step
            headers = {'content-type': 'application/json'}
            data = '{"operator_report": "Test report", "operator_accept": true}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[4].flow_task.name, 'check_operator_accept')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[4].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].flow_task.name, 'qa_reporting_sos')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].status, 'NEW')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['id']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            #API: Perform qa_reporting_sos step
            headers = {'content-type': 'application/json'}
            data = '{"sos_report": "Test report", "quality_within_policy": true, "sos_accept_show_pi": true}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].flow_task.name, 'qa_reporting_sos')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[6].flow_task.name, 'check_sos_accept_show_pi')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[6].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].flow_task.name, 'pi_verification')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].status, 'NEW')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['id']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            #API: Perform pi_verification step
            headers = {'content-type': 'application/json'}
            data = '{"pi_report": "Test report", "pi_accept": true}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].flow_task.name, 'pi_verification')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].flow_task.name, 'decide_acceptance')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].status, 'NEW')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['id']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            #API: Perform decide_acceptance step
            headers = {'content-type': 'application/json'}
            data = '{"sos_accept_after_pi": false}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].flow_task.name, 'decide_acceptance')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[9].flow_task.name, 'check_sos_accept_after_pi')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[9].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[10].flow_task.name, 'mark_sub_not_accepted')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[10].status, 'DONE')

            output_pinned = models.SchedulingUnitBlueprint.objects.get(pk=scheduling_unit_process_id).output_pinned
            #self.assertEqual(True,output_pinned)   # todo: progress to check_cleanup_task_present when implemented

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[11].flow_task.name, 'check_data_pinned')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[11].status, 'DONE')

            # todo: progress to check_cleanup_task_present when implemented
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[12].flow_task.name, 'unpin_data')
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[12].status, 'NEW')
            #
            # #API: Get current task
            # data = ""
            # response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            # content = response.content.decode('utf-8')
            # r_dict = json.loads(content)
            # task_id=r_dict[0]['id']
            #
            # #API: Assign to an user
            # data = ""
            # response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            # content = response.content.decode('utf-8')
            # r_dict = json.loads(content)
            # self.assertEqual(200,response.status_code)
            #
            # #API: Perform decide_acceptance step
            # headers = {'content-type': 'application/json'}
            # data = '{"unpin_data": true}'
            # response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            # content = response.content.decode('utf-8')
            # r_dict = json.loads(content)
            # self.assertEqual(200,response.status_code)
            #
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[12].flow_task.name, 'unpin_data')
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[12].status, 'DONE')
            #
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[13].flow_task.name, 'delete_data')
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[13].status, 'DONE')
            #
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[14].flow_task.name, 'end')
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[14].status, 'DONE')

    def test_qa_workflow_qa_is_data_pinned_no(self):
        from lofar.sas.tmss.tmss.tmssapp import models
        from lofar.sas.tmss.tmss.tmssapp.models import TaskType

        from lofar.sas.tmss.tmss.tmssapp.tasks import create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft, update_task_graph_from_specifications_doc
        from lofar.sas.tmss.test.tmss_test_data_django_models import SchedulingSet_test_data

        from lofar.sas.tmss.tmss.workflowapp.models.schedulingunitflow import SchedulingUnitProcess
        from lofar.sas.tmss.tmss.workflowapp.models.schedulingunitflow import SchedulingUnitTask as Task

        service = create_workflow_service(handler_type=SchedulingUnitFlowTest.TestSchedulingUnitEventMessageHandler, exchange=tmp_exchange.address)
        with BusListenerJanitor(service):
            strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_version_or_latest(name="IM HBA LoTSS - 2 Beams")

            scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(
                                            name="Test Scheduling Unit UC1",
                                            specifications_template=strategy_template.scheduling_unit_template,
                                            observation_strategy_template=strategy_template,
                                            scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))

            update_task_graph_from_specifications_doc(scheduling_unit_draft, strategy_template.template)

            scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
            scheduling_unit_draft.refresh_from_db()

            # there is no signal that SchedulingUnitProcess instance was created,
            # so we have to wait an poll before we can proceed with the test
            poll_starttime = datetime.utcnow()
            while True:
                if SchedulingUnitProcess.objects.filter(su=scheduling_unit_blueprint).count() > 0:
                    break
                sleep(0.5)
                if datetime.utcnow()-poll_starttime > timedelta(seconds=60):
                    raise TimeoutError("SchedulingUnitProcess not created within expected time")

            # Yes! the SchedulingUnitProcess was created, let's get it.
            scheduling_unit_process_id = SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).id

            prev_ingest_permission_granted_since = models.SchedulingUnitBlueprint.objects.get(pk=scheduling_unit_process_id).ingest_permission_granted_since
            self.assertEqual(prev_ingest_permission_granted_since, None)

            #check the active task name
            self.assertEqual("wait_scheduled", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[0].flow_task.name, 'start')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[0].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].flow_task.name, 'wait_scheduled')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].status, 'NEW')


            # Change observation subtask status to scheduled
            from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions
            for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
                if task_blueprint.specifications_template.type.value == TaskType.Choices.OBSERVATION.value:
                    for subtask in task_blueprint.subtasks.all():
                        set_subtask_state_following_allowed_transitions(subtask, 'scheduled')

            # wait until scheduling unit is scheduled
            logging.info("waiting for sync_event_bp_scheduled...")
            if not self.sync_event_bp_scheduled.wait(timeout=10):
                logging.info("sync_event_bp_scheduled event not received, raising TimeoutError")
                raise TimeoutError()
            else:
                logging.info("Received self.sync_event_bp_scheduled event")
                poll_starttime = datetime.utcnow()
                while True:
                    if SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name == "wait_observing":
                        break
                    sleep(0.5)
                    if datetime.utcnow()-poll_starttime > timedelta(seconds=30):
                        raise TimeoutError("Task not activated within expected time")

            self.sync_event_bp_scheduled.clear()


            #check the active task name
            self.assertEqual("wait_observing", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].flow_task.name, 'wait_scheduled')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].flow_task.name, 'wait_observing')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].status, 'NEW')

            # Change observation subtask status to scheduled
            from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions
            for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
                if task_blueprint.specifications_template.type.value == TaskType.Choices.OBSERVATION.value:
                    task_blueprint.output_pinned = False
                    task_blueprint.save()

                    for subtask in task_blueprint.subtasks.all():
                        set_subtask_state_following_allowed_transitions(subtask, 'finished')

            logging.info("waiting for sync_event_bp_observed...")
            if not self.sync_event_bp_observed.wait(timeout=10):
                logging.info("sync_event_bp_observed event not received, raising TimeoutError")
                raise TimeoutError()
            else:
                logging.info("Received self.sync_event_bp_observed event")
                poll_starttime = datetime.utcnow()
                while True:
                    if SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name == "qa_reporting_to":
                        break
                    sleep(0.5)
                    if datetime.utcnow()-poll_starttime > timedelta(seconds=30):
                        raise TimeoutError("Task not activated within expected time")

            self.sync_event_bp_observed.clear()

            #check the active task name
            self.assertEqual("qa_reporting_to", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].flow_task.name, 'wait_observing')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'NEW')

            # schedule, run, finish the pipelines and ingest et al
            for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
                if task_blueprint.specifications_template.type.value != TaskType.Choices.OBSERVATION.value:
                    for subtask in task_blueprint.subtasks.all():
                        set_subtask_state_following_allowed_transitions(subtask, 'finished')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['id']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'ASSIGNED')

            #API: Perform qa_reporting_to step
            headers = {'content-type': 'application/json'}
            data = '{"operator_report": "Test report", "operator_accept": true}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name, 'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[4].flow_task.name, 'check_operator_accept')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[4].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].flow_task.name, 'qa_reporting_sos')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].status, 'NEW')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['id']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            #API: Perform qa_reporting_sos step
            headers = {'content-type': 'application/json'}
            data = '{"sos_report": "Test report", "quality_within_policy": true, "sos_accept_show_pi": true}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].flow_task.name, 'qa_reporting_sos')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[5].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[6].flow_task.name, 'check_sos_accept_show_pi')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[6].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].flow_task.name, 'pi_verification')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].status, 'NEW')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['id']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            #API: Perform pi_verification step
            headers = {'content-type': 'application/json'}
            data = '{"pi_report": "Test report", "pi_accept": true}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].flow_task.name, 'pi_verification')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[7].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].flow_task.name, 'decide_acceptance')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].status, 'NEW')

            #API: Get current task
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id=r_dict[0]['id']

            #API: Assign to an user
            data = ""
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/'.format(task_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            #API: Perform decide_acceptance step
            headers = {'content-type': 'application/json'}
            data = '{"sos_accept_after_pi": true}'
            response = requests.post(self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/perform/'.format(scheduling_unit_process_id), data=data, auth=self.__class__.AUTH, headers=headers)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            self.assertEqual(200,response.status_code)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].flow_task.name, 'decide_acceptance')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[8].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[9].flow_task.name, 'check_sos_accept_after_pi')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[9].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[10].flow_task.name, 'mark_sub_accepted')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[10].status, 'DONE')

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[11].flow_task.name, 'allow_ingest')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[11].status, 'DONE')

            #verify that ingest_permission_granted_since is now a valid datetime
            ingest_permission_granted_since = models.SchedulingUnitBlueprint.objects.get(pk=scheduling_unit_process_id).ingest_permission_granted_since
            self.assertEqual(True,isinstance(ingest_permission_granted_since, datetime))

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[12].flow_task.name, 'wait_ingest_done')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[12].status, 'DONE')

            output_pinned = models.SchedulingUnitBlueprint.objects.get(pk=scheduling_unit_process_id).output_pinned
            #self.assertEqual(True,output_pinned)   # todo: progress to check_cleanup_task_present when implemented

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[13].flow_task.name, 'check_data_pinned')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[13].status, 'DONE')

            # todo: progress to check_cleanup_task_present when implemented
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[14].flow_task.name, 'unpin_data')
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[14].status, 'NEW')

            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[15].flow_task.name, 'end')
            # self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[15].status, 'DONE')

    def test_assign_workflow_task_to_user_by_email_or_role(self):

        from lofar.sas.tmss.tmss.tmssapp import models
        from lofar.sas.tmss.tmss.tmssapp.models import TaskType

        from lofar.sas.tmss.tmss.tmssapp.tasks import create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft, update_task_graph_from_specifications_doc
        from lofar.sas.tmss.test.tmss_test_data_django_models import SchedulingSet_test_data

        from lofar.sas.tmss.tmss.workflowapp.models.schedulingunitflow import SchedulingUnitProcess
        from lofar.sas.tmss.tmss.workflowapp.models.schedulingunitflow import SchedulingUnitTask as Task

        from django.contrib.auth import get_user_model
        User = get_user_model()
        User.objects.create(email="someone@astron.nl", username='someone')  # create the user we need in this test

        service = create_workflow_service(handler_type=SchedulingUnitFlowTest.TestSchedulingUnitEventMessageHandler, exchange=tmp_exchange.address)
        with BusListenerJanitor(service):
            strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_version_or_latest(name="IM HBA LoTSS - 2 Beams")

            scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(
                name="Test Scheduling Unit UC1",
                specifications_template=strategy_template.scheduling_unit_template,
                observation_strategy_template=strategy_template,
                scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))

            update_task_graph_from_specifications_doc(scheduling_unit_draft, strategy_template.template)

            scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(
                scheduling_unit_draft)
            scheduling_unit_draft.refresh_from_db()

            # there is no signal that SchedulingUnitProcess instance was created,
            # so we have to wait an poll before we can proceed with the test
            poll_starttime = datetime.utcnow()
            while True:
                if SchedulingUnitProcess.objects.filter(su=scheduling_unit_blueprint).count() > 0:
                    break
                sleep(0.5)
                if datetime.utcnow() - poll_starttime > timedelta(seconds=30):
                    raise TimeoutError("SchedulingUnitProcess not created within expected time")

            # Yes! the SchedulingUnitProcess was created, let's get it.
            scheduling_unit_process_id = SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).id

            prev_ingest_permission_granted_since = models.SchedulingUnitBlueprint.objects.get(
                pk=scheduling_unit_process_id).ingest_permission_granted_since
            self.assertEqual(prev_ingest_permission_granted_since, None)

            # check the active task name
            self.assertEqual("wait_scheduled", SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[0].flow_task.name)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[0].flow_task.name,'start')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[0].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].flow_task.name, 'wait_scheduled')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].status, 'NEW')

            # Change observation subtask status to scheduled
            from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions
            for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
                if task_blueprint.specifications_template.type.value == TaskType.Choices.OBSERVATION.value:
                    for subtask in task_blueprint.subtasks.all():
                        set_subtask_state_following_allowed_transitions(subtask, 'scheduled')

            # wait until scheduling unit is scheduled
            logging.info("waiting for sync_event_bp_scheduled...")
            if not self.sync_event_bp_scheduled.wait(timeout=10):
                logging.info("sync_event_bp_scheduled event not received, raising TimeoutError")
                raise TimeoutError()
            else:
                logging.info("Received self.sync_event_bp_scheduled event")
                poll_starttime = datetime.utcnow()
                while True:
                    if SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[
                        0].flow_task.name == "wait_observing":
                        break
                    sleep(0.5)
                    if datetime.utcnow() - poll_starttime > timedelta(seconds=30):
                        raise TimeoutError("Task not activated within expected time")

            self.sync_event_bp_scheduled.clear()

            # check the active task name
            self.assertEqual("wait_observing",
                             SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[
                                 0].flow_task.name)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].flow_task.name,
                             'wait_scheduled')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[1].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].flow_task.name,
                             'wait_observing')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].status, 'NEW')

            # Change observation subtask status to scheduled
            from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions
            for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
                if task_blueprint.specifications_template.type.value == TaskType.Choices.OBSERVATION.value:
                    task_blueprint.output_pinned = False
                    task_blueprint.save()

                    for subtask in task_blueprint.subtasks.all():
                        set_subtask_state_following_allowed_transitions(subtask, 'finished')

            logging.info("waiting for sync_event_bp_observed...")
            if not self.sync_event_bp_observed.wait(timeout=10):
                logging.info("sync_event_bp_observed event not received, raising TimeoutError")
                raise TimeoutError()
            else:
                logging.info("Received self.sync_event_bp_observed event")
                poll_starttime = datetime.utcnow()
                while True:
                    if SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[
                        0].flow_task.name == "qa_reporting_to":
                        break
                    sleep(0.5)
                    if datetime.utcnow() - poll_starttime > timedelta(seconds=30):
                        raise TimeoutError("Task not activated within expected time")

            self.sync_event_bp_observed.clear()

            # check the active task name
            self.assertEqual("qa_reporting_to",
                             SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint.id).active_tasks()[
                                 0].flow_task.name)

            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].flow_task.name,
                             'wait_observing')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[2].status, 'DONE')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].flow_task.name,
                             'qa_reporting_to')
            self.assertEqual(Task.objects.filter(process=scheduling_unit_process_id).order_by('id')[3].status, 'NEW')

            # API: Get current task
            data = ""
            response = requests.post(
                self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_process/{}/current_task/'.format(
                    scheduling_unit_process_id), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')
            r_dict = json.loads(content)
            task_id = r_dict[0]['id']
            task = Task.objects.get(id=task_id)

            # assert task is still unassigned
            task.refresh_from_db()
            self.assertNotEqual(task.status, 'ASSIGNED')
            self.assertEqual(task.owner, None)

            # API: Assign to a user by email, as scientist, which should NOT work
            data = ""
            user_email = 'someone@astron.nl'
            from requests.auth import HTTPDigestAuth
            response = requests.post(
                self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/?user_email={}'.format(
                    task_id, user_email), json=data, auth=('scientist', 'scientist'))
            content = response.content.decode('utf-8')
            self.assertEqual(403, response.status_code)
            self.assertIn("You do not have permission to perform this action", content)

            # assert task is still unassigned
            task.refresh_from_db()
            self.assertNotEqual(task.status, 'ASSIGNED')
            self.assertEqual(task.owner, None)

            # API: Assign to a user by email, as operator, which should work
            data = ""
            user_email = 'someone@astron.nl'
            response = requests.post(
                self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/?user_email={}'.format(
                    task_id, user_email), json=data, auth=('operator', 'operator'))
            content = response.content.decode('utf-8')
            self.assertEqual(200, response.status_code)
            self.assertIn(user_email, content)

            # assert task is assigned to correct user
            user = User.objects.get(email=user_email)
            task.refresh_from_db()
            self.assertEqual(task.status, 'ASSIGNED')
            self.assertEqual(task.owner, user)

            # API: Assign to a different user by role
            data = ""
            project_role = 'contact'
            response = requests.post(
                self.__class__.BASE_URL_WF_API + '/scheduling_unit_flow/qa_scheduling_unit_task/{}/assign/?project_role={}'.format(
                    task_id, project_role), json=data, auth=self.__class__.AUTH)
            content = response.content.decode('utf-8')

            # assert we contact Keycloak for the info who has the role in the project.
            # (this fails since we don't have admin credentials for Keycloak, so the task is not actually reassigned)
            # todo: find a way to mock this out or have an integration test with working credentials in place
            self.assertIn("Keycloak admin API token could not be obtained", content)
            self.assertEqual(500, response.status_code)


if __name__ == '__main__':
    #run the unit tests
    unittest.main()
