#!/usr/bin/python3

# Copyright (C) 2020  ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

from dateutil import parser
from typing import List
from lofar.sas.tmss.tmss.tmssapp.models import *
from lofar.sas.tmss.tmss.tmssapp.conversions import antennafields_for_antennaset_and_station
from lofar.parameterset import parameterset
from lofar.common.util import single_line_with_single_spaces
from lofar.common.datetimeutils import formatDatetime
from lofar.sas.tmss.tmss.exceptions import SubtaskInvalidStateException, SubtaskException

import logging
logger = logging.getLogger(__name__)

def parse_parset_vector(parset_value: str, parameterset_func=parameterset.getIntVector) -> List[int]:
    """ Parse a parameterset vector, by passing it through the parameterset class.

        Provide the actual function the parameterset class uses to get the vector. """
    p = parameterset()
    p.add("key", parset_value)
    return parameterset_func(p, "key", expandable=True)

def parse_antennafield_name(name: str) -> dict:
    """ Parse a name like CS001HBA into a dict denoting the station, field, and type as used in the feedback_doc. """

    if name.endswith("HBA"):
        return { "station": name[:-3], "field": "HBA", "type": "HBA" }
    elif name.endswith("LBA"):
        return { "station": name[:-3], "field": "LBA", "type": "LBA" }
    elif name.endswith("HBA0"):
        return { "station": name[:-4], "field": "HBA0", "type": "HBA" }
    elif name.endswith("HBA1"):
        return { "station": name[:-4], "field": "HBA1", "type": "HBA" }
    else:
        raise Exception("Invalid station name supplied: %s" % name)


def translate_storagewriter_parset_to_doc(storagewriter: str) -> str:
    """ Translate a storagewriter as mentioned in the feedback parset to one we can use in the feedback_doc. """
    storagewriter = storagewriter.lower()

    if storagewriter == "casa":
        return "standard"
    elif storagewriter == "lofar":
        return "lofarstman"
    elif storagewriter == "dysco":
        return "dysco"
    elif storagewriter == "hdf5default":
        return "standard"
    else:
        return storagewriter

def observation_correlated_feedback_to_feedback_doc(dp_feedback: dict) -> dict:
    """ Transform feedback for Correlated dataproducts from observations into a feedback_doc dictionary."""

    # This feedback is generated in RTCP/Cobalt/CoInterface/src/LTAFeedback.cc

    feedback_doc = {
        "percentage_written": int(dp_feedback['percentageWritten']),
        "frequency": {
            # we explicitely choose to rename from inconsistent lofar1 station dataproduct feedback names to consistent tmss feedback names
            # this is mainly due to historical use and lack of a proper definition.
            # so, even though it might look like we've erroneously swapped the two here, this is actually correct.
            # also, see the SIP adapter which also swaps them back.

            # Cobalt/pipeline feedback (and LTA) definition of:
            # 'stationsubband':  the frequency index from the station PPF, thus an integer in the [0, 511] range
            # 'subband':         the measurementset number from the correlator, thus an integer in the [0, 487] range 

            # TMSS definition of:
            # 'subbands':        the frequency index from the station PPF, thus an integer in the [0, 511] range
            # 'beamlet_indices': the measurementset number from the correlator, thus an integer in the [0, 487] range 
            "subbands": [int(dp_feedback['stationSubband'])],
            "beamlet_indices": [int(dp_feedback['subband'])],
            "central_frequencies": [float(dp_feedback['centralFrequency'])],
            "channel_width": float(dp_feedback['channelWidth']),
            "channels_per_subband": int(dp_feedback['channelsPerSubband']),
            "spectral_leakage": False,
            "topocentric_correction": bool(dp_feedback.get('dopplerCorrection', False)),
        },
        "time": {
            "start_time": parser.parse(dp_feedback['startTime'], ignoretz=True).isoformat()+'Z',
            "duration": float(dp_feedback['duration']),
            "sample_width": float(dp_feedback['integrationInterval']),
        },
        "antennas": {
            "set": dp_feedback['antennaSet'],
            "fields": [parse_antennafield_name(field) for field in parse_parset_vector(dp_feedback['antennaFields'], parameterset.getStringVector)],
        },
        "target": {
            "pointing": {
                "direction_type": dp_feedback['Pointing.directionType'],
                "angle1": float(dp_feedback['Pointing.angle1']),
                "angle2": float(dp_feedback['Pointing.angle2']),
                "target": dp_feedback['Pointing.target']
            }
        },
        "samples": {
            "polarisations": ["XX","XY","YX","YY"],
            "type": "float",
            "bits": 32,
            "writer": translate_storagewriter_parset_to_doc(dp_feedback['storageWriter']),
            "writer_version": dp_feedback['storageWriterVersion'],
            "complex": True
        }
    }

    return feedback_doc

def observation_beamformed_feedback_to_feedback_doc(dp_feedback: dict) -> dict:
    """ Transform feedback for Beamformed dataproducts from observations into a feedback_doc dictionary."""

    # This feedback is generated in RTCP/Cobalt/CoInterface/src/LTAFeedback.cc

    # Determine which type of beam we're dealing with, to address the right array.
    # TODO: Remove this array layer from the feedback in COBALT
    beam_prefix = dp_feedback["beamPrefix"] # f.e. CoherentStokesBeam[0].

    feedback_doc = {
        "percentage_written": int(dp_feedback['percentageWritten']),
        "frequency": {
            "subbands": parse_parset_vector(dp_feedback[beam_prefix + 'stationSubbands'], parameterset.getIntVector),
            "central_frequencies": parse_parset_vector(dp_feedback[beam_prefix + 'centralFrequencies'], parameterset.getDoubleVector),
            "channel_width": float(dp_feedback[beam_prefix + 'channelWidth']),
            "channels_per_subband": int(dp_feedback[beam_prefix + 'channelsPerSubband']),
            "spectral_leakage": bool(dp_feedback.get('beamformingSpectralLeakage', False)),
            "topocentric_correction": False,
        },
        "time": {
            "start_time": parser.parse(dp_feedback['startTime'], ignoretz=True).isoformat()+'Z',
            "duration": float(dp_feedback['duration']),
            "sample_width": float(dp_feedback[beam_prefix + 'samplingTime']),
        },
        "antennas": {
            "set": dp_feedback[beam_prefix + 'antennaSet'],
            "fields": [parse_antennafield_name(field) for field in parse_parset_vector(dp_feedback[beam_prefix + 'antennaFields'], parameterset.getStringVector)],
        },
        "target": {
            "pointing": {
                "direction_type": dp_feedback[beam_prefix + 'Pointing.directionType'],
                "angle1": float(dp_feedback[beam_prefix + 'Pointing.angle1']),
                "angle2": float(dp_feedback[beam_prefix + 'Pointing.angle2']),
                "target": dp_feedback[beam_prefix + 'Pointing.target']
            },
            "coherent": not beam_prefix.startswith("IncoherentStokesBeam")
        },
        "samples": {
            "polarisations": parse_parset_vector(dp_feedback[beam_prefix + 'stokes'], parameterset.getStringVector),
            "type": dp_feedback[beam_prefix + 'sampleType'],
            "bits": int(dp_feedback[beam_prefix + 'sampleNrBits']),
            "writer": translate_storagewriter_parset_to_doc(dp_feedback['storageWriter']),
            "writer_version": dp_feedback['storageWriterVersion'],
            "complex": False # Complex Voltages write complex values, but split real and imaginary values into separate files
        }
    }

    return feedback_doc

def preprocessing_pipeline_feedback_to_feedback_doc(input_dp_feedback_doc: dict, dp_feedback: dict) -> dict:
    """ Transform feedback for Correlated dataproducts from preprocessing pipelines into a feedback_doc dictionary."""

    # This feedback is generated in CEP/Pipeline/recipes/sip/helpers/metadata.py

    feedback_doc = {
        "percentage_written": int(dp_feedback['percentageWritten']),
        "frequency": {
            # we explicitely choose to rename from inconsistent lofar1 station dataproduct feedback names to consistent tmss feedback names
            # this is mainly due to historical use and lack of a proper definition.
            # so, even though it might look like we've erroneously swapped the two here, this is actually correct.
            # also, see the SIP adapter which also swaps them back.
            # Thus, we swap at receiving the feedback -> then have the beamlet_indices and subbands internally in tmss consistent, and swap back here when creating the sip.
            # Furthermore... the preprocessing pipeline does not provide a proper value for 'stationSubband' (it is always 0)
            # so, we copy it from the corresponding upstream input dataproduct's feedback.
            "beamlet_indices": [int(dp_feedback['subband'])],
            "subbands": input_dp_feedback_doc['frequency']['subbands'],
            "central_frequencies": [float(dp_feedback['centralFrequency'])],
            "channel_width": float(dp_feedback['channelWidth']),
            "channels_per_subband": int(dp_feedback['channelsPerSubband']),
        },
        "time": {
            "start_time": parser.parse(dp_feedback['startTime'], ignoretz=True).isoformat()+'Z',
            "duration": float(dp_feedback['duration']),
            "sample_width": float(dp_feedback['integrationInterval']),
        },
        "antennas": {
            "set": input_dp_feedback_doc["antennas"]['set'],
            "fields": input_dp_feedback_doc["antennas"]['fields'],
        },
        "target": {
            "pointing": input_dp_feedback_doc["target"]["pointing"],
        },
        "samples": {
            "polarisations": ["XX","XY","YX","YY"],
            "type": "float",
            "bits": 32,
            "writer": translate_storagewriter_parset_to_doc(dp_feedback['storageWriter']),
            "writer_version": dp_feedback['storageWriterVersion'],
            "complex": True
        }
    }

    return feedback_doc


def pulsar_pipeline_summary_feedback_to_feedback_doc(dp_feedback: dict) -> dict:
    """ Transform feedback for pulsar pipeline "summary" dataproducts from observations into a feedback_doc dictionary."""

    # This feedback is generated in https://git.astron.nl/ro/pulp/-/blob/master/pulp_feedback.py

    feedback_doc = {
        "percentage_written": int(dp_feedback['percentageWritten']),
        # set of files must be unique, but PULP sends us duplicates 
        "files": list(set(parse_parset_vector(dp_feedback['fileContent'], parameterset.getStringVector))),
        "target": {
            "coherent": dp_feedback['datatype'] != "SummaryIncoherentStokes"
        }
    }

    return feedback_doc


def pulsar_pipeline_analysis_feedback_to_feedback_doc(input_dp_feedback_doc: dict, dp_feedback: dict) -> dict:
    """ Transform feedback for pulsar pipeline "analysis" dataproducts from observations into a feedback_doc dictionary.
        
        input_dp_feedback_doc can be any input dataproduct related to this output dataproduct. Typically,
        there is only 1, but in case of ComplexVoltages, there are 4 inputs (each of Xre/Xim/Yre/Yim). """

    # This feedback is generated in https://git.astron.nl/ro/pulp/-/blob/master/pulp_feedback.py

    for beam_prefix in ["FlysEyeBeam.", "CoherentStokesBeam.", "IncoherentStokesBeam."]:
        if any(k.startswith(beam_prefix) for k in dp_feedback.keys()):
            break
    else:
        raise ValueError("Could not determine beam prefix. Feedback: %s" % dp_feedback)

    feedback_doc = {
        "percentage_written": int(dp_feedback['percentageWritten']),
        # set of files must be unique, but PULP sends us duplicates 
        "files": list(set(parse_parset_vector(dp_feedback['fileContent'], parameterset.getStringVector))),
        "frequency": {
            "subbands": parse_parset_vector(dp_feedback[beam_prefix + 'stationSubbands'], parameterset.getIntVector),
            "central_frequencies": parse_parset_vector(dp_feedback[beam_prefix + 'centralFrequencies'], parameterset.getDoubleVector),
            "channel_width": float(dp_feedback[beam_prefix + 'channelWidth']),
            "channels_per_subband": int(dp_feedback[beam_prefix + 'channelsPerSubband']),
        },
        "time": {
            "start_time": input_dp_feedback_doc["time"]["start_time"],
            "duration":  input_dp_feedback_doc["time"]["duration"],
            "sample_width": float(dp_feedback[beam_prefix + 'samplingTime']),
        },
        "antennas": {
            "set": input_dp_feedback_doc["antennas"]['set'],
            "fields": [parse_antennafield_name(dp_feedback[beam_prefix + 'stationName'] + dp_feedback[beam_prefix + 'antennaFieldName'])] if beam_prefix == "FlysEyeBeam." else input_dp_feedback_doc["antennas"]['fields'],
        },
        "target": {
            "pointing": {
                "direction_type": "J2000",
                "angle1": float(dp_feedback[beam_prefix + 'Pointing.angle1']),
                "angle2": float(dp_feedback[beam_prefix + 'Pointing.angle2']),
                "target": input_dp_feedback_doc["target"]["pointing"]["target"]
            } if beam_prefix == "CoherentStokesBeam." else input_dp_feedback_doc["target"]["pointing"],
            "coherent": beam_prefix != "IncoherentStokesBeam."
        },
        "samples": {
            "polarisations": parse_parset_vector(dp_feedback[beam_prefix + 'stokes'], parameterset.getStringVector),
            "type": input_dp_feedback_doc["samples"]["type"],
            "bits": input_dp_feedback_doc["samples"]["bits"],
            "writer": "standard",
            "writer_version": "UNKNOWN",
            "complex": dp_feedback['datatype'] == "ComplexVoltages",
        }
    }

    return feedback_doc


def extract_dataproduct_feedback(feedback: parameterset) -> dict:
    """ Returns a dict of the feedback strings per dataproduct, extracted from a feedback parset. Example:

        extract_dataproduct_feedback(
            parameterset.fromString("LOFAR.ObsSW.Observation.Dataproducts.Output_Beamformed_[0].key1=value1")
        ) ==
        { "Output_Beamformed_[0]": { "key1": "value1" } }
    """

    # locate where the dataproducts are, to cut off any prefix like "LOFAR.ObsSW."
    dp_feedback = feedback.makeSubset(feedback.fullModuleName('Observation.DataProducts') + '.')

    # extract the unique dataproduct keys (f.e. "Output_Beamformed_[0]"), so we can loop over them
    dp_keys = sorted(list(set([key[:key.find('.')] for key in dp_feedback.keys() if key.startswith('Output_')])))

    # extract all info for each dataproduct
    return {dp_key: dp_feedback.makeSubset(dp_key+'.').dict() for dp_key in dp_keys}

def observation_process_feedback_to_feedback_doc_and_template(feedback: parameterset) -> (dict, SubtaskFeedbackTemplate):
    """ Process processing feedback from COBALT. """

    if not feedback.isDefined('Observation.startTime'):
        return None, None

    template = SubtaskFeedbackTemplate.get_version_or_latest(name='observation')
    doc = template.get_default_json_document_for_schema()
    doc["on_sky"] = {
        "start_time": parser.parse(feedback.getString('Observation.startTime'), ignoretz=True).isoformat()+"Z",
        "stop_time":  parser.parse(feedback.getString('Observation.stopTime'), ignoretz=True).isoformat()+"Z",
    }

    return (doc, template)


def process_feedback_into_subtask_properties(subtask:Subtask, feedback: parameterset) -> bool:
    """ Process feedback that is not dataproduct specific and update the subtask's processing_doc & processing_template accordingly.

        Processes which have no feedback will get processing_template.name == 'empty'.

        Returns the subtask. """

    if subtask.state.value != SubtaskState.objects.get(value='finishing').value:
        raise SubtaskInvalidStateException("Cannot process feedback for subtask id=%s because the state is '%s' and not '%s'" % (subtask.id, subtask.state.value, SubtaskState.Choices.FINISHING.value))

    try:
        if subtask.is_observation:
            # update actual start/stop times from the feedback
            doc, template = observation_process_feedback_to_feedback_doc_and_template(feedback)

            if doc is not None and template is not None:
                subtask.processing_feedback_doc = doc
                subtask.processing_feedback_template = template

                # the on_sky_start/stop_times are stored both in the feedback_doc and in the subtask model.
                # That is duplication of information, which is "bad"... but...
                # We do reason about on_sky_start/stop_times in the database. We compute derived/aggregated properties, which is more difficult on json_docs in sql.
                # Apart from that, there are more places where some properties are stored twice, in the feedback_doc and in some property. See for example dataproduct.size.
                # To guard against tinkering with these times and to prevent that they get 'out-of-sync', these properties are immutable in the database once set.
                start_time = subtask.processing_feedback_doc.get('on_sky', {}).get('start_time')
                stop_time = subtask.processing_feedback_doc.get('on_sky', {}).get('stop_time')
                if start_time is not None and stop_time is not None and (subtask.actual_on_sky_start_time is None or subtask.actual_on_sky_stop_time is None):
                    subtask.actual_on_sky_start_time = parser.parse(start_time, ignoretz=True)
                    subtask.actual_on_sky_stop_time = parser.parse(stop_time, ignoretz=True)

                subtask.save()
        else:
            # this subtask type has no processing feedback
            (subtask.processing_feedback_doc,
            subtask.processing_feedback_template) = ({}, SubtaskFeedbackTemplate.get_version_or_latest(name='empty'))
            subtask.save()
    except Exception as e:
        # just log any feedback handling/processing exceptions, do not go to error, stay in FINISHING state so the user can repair.
        logger.exception(f"Cannot process feedback for subtask id={subtask.id} of type {subtask.specifications_template.type.value}. \nFeedback:\m{feedback}")

    return subtask

def process_feedback_into_subtask_dataproducts(subtask:Subtask, feedback: parameterset) -> Subtask:
    if subtask.specifications_template.type.value not in [SubtaskType.Choices.OBSERVATION.value, SubtaskType.Choices.PIPELINE.value]:
            raise SubtaskException("Cannot process feedback for subtask id=%s since type=%s not in %s" %
                                   (subtask.id, subtask.specifications_template.type.value,
                                    [SubtaskType.Choices.OBSERVATION.value, SubtaskType.Choices.PIPELINE.value]))

    if subtask.state.value != SubtaskState.objects.get(value='finishing').value:
        raise SubtaskInvalidStateException("Cannot process feedback for subtask id=%s because the state is '%s' and not '%s'" % (subtask.id, subtask.state.value, SubtaskState.Choices.FINISHING.value))

    logger.info('processing feedback into the dataproducts of subtask id=%s type=%s feedback:\n%s', subtask.id, subtask.specifications_template.type.value, str(feedback))

    # process each dataproduct subset
    for dp_key, dp_feedback in extract_dataproduct_feedback(feedback).items():
        # determine corresponding TMSS dataproduct
        try:
            dataproduct = subtask.output_dataproducts.get(filename=dp_feedback.get('filename'))
        except Exception as e:
            # process reported about a file we didn't know it would write!

            logger.exception("cannot process feedback: %s. No output dataproduct known for subtask id=%s feedback: %s", e, subtask.id, dp_feedback)
            continue

        try:
            logger.info('processing feedback for dataproduct id=%s filename=%s of subtask id=%s feedback: %s', dataproduct.id, dataproduct.filename, subtask.id, single_line_with_single_spaces(str(dp_feedback)))

            # derive values or collect for different subtask types
            if subtask.is_observation:
                if dataproduct.datatype.value == "visibilities":
                    dataproduct.feedback_template = DataproductFeedbackTemplate.get_version_or_latest(name='feedback')
                    dataproduct.feedback_doc = observation_correlated_feedback_to_feedback_doc(dp_feedback)
                elif dataproduct.datatype.value == "time series":
                    dataproduct.feedback_template = DataproductFeedbackTemplate.get_version_or_latest(name='feedback')
                    dataproduct.feedback_doc = observation_beamformed_feedback_to_feedback_doc(dp_feedback)
                else:
                    logger.error("cannot process feedback for dataproduct with specification template %s", dataproduct.specifications_template.name)

            elif subtask.specifications_template.type.value == SubtaskType.Choices.PIPELINE.value:
                if subtask.specifications_template.name == "preprocessing pipeline":
                    # each file has exactly 1 input.
                    input_dataproduct = DataproductTransform.objects.get(output=dataproduct).input

                    dataproduct.feedback_template = DataproductFeedbackTemplate.get_version_or_latest(name='feedback')
                    dataproduct.feedback_doc = preprocessing_pipeline_feedback_to_feedback_doc(input_dataproduct.feedback_doc, dp_feedback)

                elif subtask.specifications_template.name == "pulsar pipeline":
                    if dataproduct.specifications_template.name == "pulp summary":
                        # pulp summary file
                        dataproduct.feedback_template = DataproductFeedbackTemplate.get_version_or_latest(name='pulp summary')
                        dataproduct.feedback_doc = pulsar_pipeline_summary_feedback_to_feedback_doc(dp_feedback)
                    else:
                        # pulp analysis file, which has 1-4 inputs. just use the first.
                        input_dataproduct = DataproductTransform.objects.filter(output=dataproduct).first().input

                        dataproduct.feedback_template = DataproductFeedbackTemplate.get_version_or_latest(name='feedback')
                        dataproduct.feedback_doc = pulsar_pipeline_analysis_feedback_to_feedback_doc(input_dataproduct.feedback_doc, dp_feedback)
                else:
                    logger.error("cannot process feedback for pipelines with schema named %s", subtask.specifications_template.name)

            # set value of dataproduct size field
            dataproduct.size = dp_feedback.get('size', dataproduct.size)

            # commit to database
            dataproduct.save()
            logger.info('saved processed feedback into dataproduct id=%s filename=%s feedback_doc=%s', dataproduct.id, dataproduct.filename, dataproduct.feedback_doc)
        except Exception as e:
            logger.exception('error while processing feedback for dataproduct id=%s filename=%s feedback=%s', dataproduct.id, dataproduct.filename, dp_feedback)

    subtask.refresh_from_db()
    return subtask


def process_feedback_for_subtask_and_set_to_finished_if_complete(subtask: Subtask, feedback_doc: str) -> Subtask:
    """
    Translates raw feedback from a subtask (which has been provided by Cobalt or pipelines) and translate it to
    json documents for the individual dataproducts.
    """
    if subtask.state.value == SubtaskState.objects.get(value='started').value:
        logger.info("received feedback for subtask id=%s while it is still running (state=%s). Setting state to 'finishing' and stop_time to now.", subtask.id, subtask.state.value)
        subtask.state = SubtaskState.objects.get(value='finishing')
        subtask.process_stop_time = datetime.utcnow()
        subtask.save()

    # the submitted feedback_doc is (should be) a plain text document in parset format
    # convert it as a parset
    new_feedback_parset = parameterset.fromString(feedback_doc)

    # and process it for the dataproducts
    subtask = process_feedback_into_subtask_dataproducts(subtask=subtask, feedback=new_feedback_parset)

    # and process the overall parameters. do this last, as if this fails, we at least parsed
    # the bulk of the feedback (the dataproducts)
    subtask = process_feedback_into_subtask_properties(subtask=subtask, feedback=new_feedback_parset)

    # if complete, set subtask state to finished
    if subtask.is_feedback_complete:
        logger.info("Feedback for subtask id=%s is complete. Setting state to 'finished'", subtask.id)
        subtask.state = SubtaskState.objects.get(value='finished')
        subtask.save()

    return subtask


def create_feedback_for_observation_subtask_from_specifications(subtask: Subtask) -> parameterset:
    '''creates a feedback document in parset format from the observation subtask's specification'''
    if not subtask.is_observation:
        raise SubtaskException("Cannot create observation feedback for subtask id=%s of type=%s" % (subtask.id, subtask.specifications_template.type.value))

    # must be scheduled or beyond
    if subtask.state.value not in SubtaskState.ACTIVE_OR_FINISHED_STATUS_VALUES:
        raise SubtaskException("Cannot create observation feedback for subtask id=%s with state=%s" % (subtask.id, subtask.state.value))

    # create the feedback parset
    # for reference, see: https://git.astron.nl/lofar2.0/cobalt/-/blob/main/CoInterface/src/Parset.cc
    parset = parameterset()
    parset.add('feedback_version', '03.02.00')

    # on_sky_start/stop_time reflects the schedule time when scheduled, or the actual times when beyond scheduled
    parset.add('Observation.startTime', formatDatetime(subtask.on_sky_start_time))
    parset.add('Observation.stopTime', formatDatetime(subtask.on_sky_stop_time))

    dataproducts_correlated = subtask.output_dataproducts.filter(dataformat__value=Dataformat.Choices.MEASUREMENTSET.value).filter(datatype__value=Datatype.Choices.VISIBILITIES.value).order_by('filename').all()
    dataproducts_beamformed = subtask.output_dataproducts.filter(dataformat__value=Dataformat.Choices.BEAMFORMED.value).filter(datatype__value=Datatype.Choices.TIME_SERIES.value).all()
    parset.add('Observation.DataProducts.nrOfOutput_Correlated_', str(dataproducts_correlated.count()))
    parset.add('Observation.DataProducts.nrOfOutput_Beamformed_', str(dataproducts_beamformed.count()))

    # use short name for the full specifications_doc
    spec = subtask.specifications_doc

    if spec['COBALT']['correlator']['enabled']:
        # generic correlator feedback
        subband_width = 200.0e6 / 1024.0  # fixed value. 200MHz is the only support clock
        nrChannelsPerSubband = spec['COBALT']['correlator']['channels_per_subband']
        blocksize = int(spec['COBALT'].get('blocksize', 196608))
        nrIntegrationsPerBlock = int(spec['COBALT']['correlator'].get('nrIntegrationsPerBlock', 1))
        nrBlocksPerIntegration = int(spec['COBALT']['correlator'].get('nrBlocksPerIntegration', 1))
        channel_width = subband_width / nrChannelsPerSubband
        nrSamplesPerBlock = blocksize / nrChannelsPerSubband
        nrSamplesPerIntegration = nrSamplesPerBlock / (nrIntegrationsPerBlock * nrBlocksPerIntegration)
        integrationTime = nrSamplesPerIntegration / channel_width;

        parset.add('Observation.Correlator.channelWidth', str(channel_width))
        parset.add('Observation.Correlator.channelsPerSubband', str(nrChannelsPerSubband))
        parset.add('Observation.Correlator.integrationInterval',  str(float(nrSamplesPerIntegration) / float(channel_width)))

        def nyquistZone():
            '''little helper routine'''
            if 'LBA' in spec['stations']['filter']:
                return 1
            if 'HBA_110_190' in spec['stations']['filter']:
                return 2
            elif 'HBA' in spec['stations']['filter']:
                return 3
            raise ValueError("unknown band filter: " + spec['stations']['filter'])

        # correlated dataproducts
        for file_nr, dp in enumerate(dataproducts_correlated.order_by('filename').all()):
            prefix = "Observation.DataProducts.Output_Correlated_[%d]." % (file_nr,)
            parset.add(prefix + "fileFormat", "AIPS++/CASA")
            parset.add(prefix + "filename", dp.filename)
            parset.add(prefix + "location", subtask.cluster.name + ':' + dp.directory)
            #parset.add(prefix + "size", "0") # skip size, remains NULL in db, which is ok
            parset.add(prefix + "storageWriter", "LOFAR")
            parset.add(prefix + "storageWriterVersion", spec.get('storagemanager', ''))
            parset.add(prefix + "percentageWritten", "-1") # unknown
            parset.add(prefix + "startTime", formatDatetime(subtask.on_sky_start_time))
            parset.add(prefix + "duration", str((subtask.on_sky_stop_time-subtask.on_sky_start_time).total_seconds()))
            parset.add(prefix + "subband", str(file_nr))

            sap_spec = dp.sap.specifications_doc
            antenna_fields = [f['station']+f['field'] for f in sap_spec['antennas']['fields']]
            antenna_fields_str = '[' + ','.join(antenna_fields) + ']'
            parset.add(prefix + "antennaFields", antenna_fields_str)
            parset.add(prefix + "antennaSet", sap_spec['antennas']['antenna_set'])
            parset.add(prefix + "integrationInterval", str(integrationTime))
            parset.add(prefix + "channelWidth", str(channel_width))
            parset.add(prefix + "channelsPerSubband", str(nrChannelsPerSubband))
            parset.add(prefix + "dopplerCorrection", "true" if spec['COBALT']['correlator'].get('doppler_correction') else "false")
            parset.add(prefix + "Pointing.directionType", sap_spec['pointing']['direction_type'])
            parset.add(prefix + "Pointing.angle1", str(sap_spec['pointing']['angle1']))
            parset.add(prefix + "Pointing.angle2", str(sap_spec['pointing']['angle2']))
            parset.add(prefix + "Pointing.target", sap_spec['pointing'].get('target', ''))
            parset.add(prefix + "SAP", "0") # obsolete, 0 is ok.
            stationSubband = dp.specifications_doc['subband']
            parset.add(prefix + "stationSubband", str(stationSubband))
            subbandOffset = 512 * (nyquistZone() - 1)
            centralFrequency = channel_width * nrChannelsPerSubband * (stationSubband + subbandOffset)
            parset.add(prefix + "centralFrequency", str(centralFrequency))

    if dataproducts_beamformed.exists():
        # TODO: implement for beamformed as well if needed/desired. It's quite complicated and needs lots of checking with subtasks.py tab_dataproducts.
        # see: https://support.astron.nl/jira/browse/TMSS-2668
        raise NotImplementedError("feedback from specifications for beamformed dataproducts is not implemented")

        # beamformed feedback
        # COBALT also outputs 'generic' info about (in)coherent/flyseye which is in fact just the info of the first tab/fe pipeline.
        # See: https://git.astron.nl/lofar2.0/cobalt/-/blob/main/CoInterface/src/LTAFeedback.cc#L284
        # TMSS does not process this 'generic' info, which was incomplete in the first place.
        # We do not need backwards compatibility with MoM, so we skip this 'generic' part, and just treat each dataproduct.
        # See for dataproduct feedback: https://git.astron.nl/lofar2.0/cobalt/-/blob/main/CoInterface/src/LTAFeedback.cc#L107

        spec_bf = spec['COBALT'].get('beamformer',{})
        spec_tab_pipelines = spec_bf.get('tab_pipelines',[])
        spec_fe_pipelines = spec_bf.get('flyseye_pipelines',[])

        for file_nr, dp in enumerate(dataproducts_beamformed.order_by('filename').all()):
            sap_spec = dp.sap.specifications_doc

            # generic/common dp feedback
            prefix = "Observation.DataProducts.Output_Beamformed_[%d]." % (file_nr,)
            parset.add(prefix + "fileFormat", "HDF5")
            parset.add(prefix + "filename", dp.filename)
            parset.add(prefix + "location", subtask.cluster.name + ':' + dp.directory)
            #parset.add(prefix + "size", "0") # skip size, remains NULL in db, which is ok
            parset.add(prefix + "storageWriter", "HDF5DEFAULT")
            parset.add(prefix + "storageWriterVersion", spec.get('storagemanager', 'UNKNOWN'))
            parset.add(prefix + "percentageWritten", "-1") # unknown
            parset.add(prefix + "beamformingSpectralLeakage", 'true' if spec_bf.get('input_ppf') else 'false')
            parset.add(prefix + "startTime", formatDatetime(subtask.on_sky_start_time)) # For now we do not support repointings, thus all starttimes for all dp's are the same
            parset.add(prefix + "duration", str((subtask.on_sky_stop_time-subtask.on_sky_start_time).total_seconds()))

            # beam specific dp feedback
            beam_type = "FlysEyeBeam" if len(spec_bf.get('flyseye_pipelines', [])) > 0 else "CoherentStokesBeam" if dp.specifications_doc.get('coherent', False) else "IncoherentStokesBeam"
            beam_prefix = "%s[0]." % (beam_type,)
            parset.add(prefix + "beamPrefix", beam_prefix)
            beam_prefix = prefix + beam_prefix
            parset.add(beam_prefix + "SAP", str(dp.specifications_doc['identifiers'].get('sap_index', 0)))
            parset.add(beam_prefix + "TAB", str(dp.specifications_doc['identifiers'].get('tab_index', 0)))
            parset.add(beam_prefix + "pipeline", str(dp.specifications_doc['identifiers'].get('pipeline_index', 0)))
            parset.add(beam_prefix + "pipelineTAB", str(dp.specifications_doc['identifiers'].get('tab_index', 0))) # same as tab??
            # parset.add(beam_prefix + "samplingTime",      str(format("%f") % (settings.sampleDuration() * stokesSet.nrChannels * stokesSet.timeIntegrationFactor)));
            # parset.add(beam_prefix + "sampleType", spec_bf['']       stokesSet.quantizerSettings.enabled ? "integer" : "float");
            # parset.add(beam_prefix + "sampleNrBits",      str(format("%u") % (stokesSet.quantizerSettings.enabled ? stokesSet.quantizerSettings.nrBits : 32)));
            #
            #       ostringstream centralFreqsStr;
            #       centralFreqsStr << "[";
            #       for (size_t i = f.firstSubbandIdx; i < f.lastSubbandIdx; ++i) {
            #         if (i > f.firstSubbandIdx)
            #           centralFreqsStr << ", ";
            #         centralFreqsStr << str(format("%.4lf") % settings.subbands[i].centralFrequency);
            #       }
            #       centralFreqsStr << "]";
            #
            # parset.add(beam_prefix + "centralFrequencies", centralFreqsStr.str());
            #
            #       ostringstream stationSubbandsStr;
            #       stationSubbandsStr << "[";
            #       for (size_t i = f.firstSubbandIdx; i < f.lastSubbandIdx; ++i) {
            #         if (i > f.firstSubbandIdx)
            #           stationSubbandsStr << ", ";
            #         stationSubbandsStr << str(format("%u") % settings.subbands[i].stationIdx);
            #       }
            #       stationSubbandsStr << "]";
            #
            # parset.add(beam_prefix + "stationSubbands",  stationSubbandsStr.str());

            antenna_fields = [f['station']+f['field'] for f in sap_spec['antennas']['fields']]
            antenna_fields_str = '[' + ','.join(antenna_fields) + ']'
            parset.add(prefix + "antennaFields", antenna_fields_str)
            parset.add(beam_prefix + "antennaSet", sap_spec['antennas']['antenna_set'])

            parset.add(prefix + "Pointing.directionType", sap_spec['pointing']['direction_type'])
            parset.add(prefix + "Pointing.angle1", str(sap_spec['pointing']['angle1']))
            parset.add(prefix + "Pointing.angle2", str(sap_spec['pointing']['angle2']))
            parset.add(prefix + "Pointing.target", sap_spec['pointing'].get('target', ''))



            #
            # parset.add(beam_prefix + "channelWidth",      str(format("%f") % (settings.subbandWidth() / stokesSet.nrChannels)));
            # parset.add(beam_prefix + "channelsPerSubband",str(format("%u") % stokesSet.nrChannels));
            # parset.add(beam_prefix + "stokes",            str(format("[%s]") % stokesVars[f.stokesNr]));
            #
            #       const struct ObservationSettings::Direction &tabDir =
            #         f.pointingNr >= 0 ? tab.pointings[f.pointingNr].direction : tab.direction;
            #       const string &tabTarget =
            #         f.pointingNr >= 0 ? tab.pointings[f.pointingNr].target : tab.target;
            #
            # parset.add(beam_prefix + "Pointing.equinox",   "J2000");
            # parset.add(beam_prefix + "Pointing.coordType", "RA-DEC");
            # parset.add(beam_prefix + "Pointing.directionType", tabDir.type);
            # parset.add(beam_prefix + "Pointing.angle1",    str(format("%f") % tabDir.angle1));
            # parset.add(beam_prefix + "Pointing.angle2",    str(format("%f") % tabDir.angle2));
            # parset.add(beam_prefix + "Pointing.target",    tabTarget);
            #
            #       if (type == "CoherentStokesBeam") {
            #         parset.add(beam_prefix + "Offset.equinox",     "J2000");
            #         parset.add(beam_prefix + "Offset.coordType",   "RA-DEC");
            #         parset.add(beam_prefix + "Offset.directionType", tabDir.type);
            #         parset.add(beam_prefix + "Offset.angle1",      str(format("%f") % (tabDir.angle1 - sap.direction.angle1)));
            #         parset.add(beam_prefix + "Offset.angle2",      str(format("%f") % (tabDir.angle2 - sap.direction.angle2)));
            #
            #         ostringstream antennaFieldNamesStr;
            #         antennaFieldNamesStr << settings.beamFormer.pipelines[tab.pipelineNr].antennaFieldNames;
            #         parset.add(beam_prefix + "antennaFields",      antennaFieldNamesStr.str());
            #       } else if (type == "IncoherentStokesBeam") {
            #         ostringstream antennaFieldNamesStr;
            #         antennaFieldNamesStr << settings.beamFormer.pipelines[tab.pipelineNr].antennaFieldNames;
            #         parset.add(beam_prefix + "antennaFields",      antennaFieldNamesStr.str());
            #       } else if (type == "FlysEyeBeam") {
            #         auto& antennaFieldName = settings.beamFormer.pipelines.at(tab.pipelineNr).antennaFieldNames.at(tab.localTabNr);
            #         parset.add(beam_prefix + "stationName", antennaFieldName.station);
            #         parset.add(beam_prefix + "antennaFieldName", antennaFieldName.antennaField);
            #
            #         ostringstream antennaFieldNamesStr;
            #         antennaFieldNamesStr << "[" << antennaFieldName << "]";
            #         parset.add(beam_prefix + "antennaFields", antennaFieldNamesStr.str());
            #       }



    return parset


def create_and_use_feedback_for_observation_subtask_from_specifications(subtask: Subtask) -> Subtask:
    '''creates and uses feedback from the observation subtask's specification, and sets the subtask's state to finished'''
    feedback_parset = create_feedback_for_observation_subtask_from_specifications(subtask)
    feedback_doc = str(feedback_parset)
    logger.info("create_and_use_feedback_for_observation_subtask_from_specifications for subtask id=%s state=%s feedback_doc=\n%s",
                subtask.id, subtask.state.value, feedback_doc)
    return process_feedback_for_subtask_and_set_to_finished_if_complete(subtask, feedback_doc)
