#!/usr/bin/python3
import copy

# Copyright (C) 2020  ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

from lofar.sas.tmss.tmss.tmssapp import models
from lofar.common.json_utils import validate_json_against_schema

import logging
logger = logging.getLogger(__name__)

def convert_to_l2stationspecs_dict(subtask: models.Subtask) -> dict:
    '''
    Convert the specifications in the subtask to a Lofar 2.0 station specifications dict.
    It is meant to be send to a LOFAR 2.0 to set it up for the observation that this subtask represents.
    '''

    antenna_set_mapping = {
        "HBA_DUAL":        { "CS": "ALL",         "RS": "ALL",         "INT": "ALL"         },
        "HBA_DUAL_INNER":  { "CS": "ALL",         "RS": "INNER",       "INT": "INNER"       },
        "HBA_ONE":         { "CS": "ALL",         "RS": "ALL",         "INT": "ALL"         },
        "HBA_ONE_INNER":   { "CS": "ALL",         "RS": "INNER",       "INT": "INNER"       },
        "HBA_ZERO":        { "CS": "ALL",         "RS": "ALL",         "INT": "ALL"         },
        "HBA_ZERO_INNER":  { "CS": "ALL",         "RS": "INNER",       "INT": "INNER"       },
        "LBA_INNER":       { "CS": "INNER",       "RS": "INNER",       "INT": "INNER"       },
        "LBA_OUTER":       { "CS": "OUTER",       "RS": "OUTER",       "INT": "OUTER"       },
        "LBA_SPARSE_EVEN": { "CS": "SPARSE_EVEN", "RS": "SPARSE_EVEN", "INT": "SPARSE_EVEN" },
        "LBA_SPARSE_ODD":  { "CS": "SPARSE_ODD",  "RS": "SPARSE_ODD",  "INT": "SPARSE_ODD"  },
        "LBA_ALL":         { "CS": "ALL",         "RS": "ALL",         "INT": "ALL"         },
    }

    antenna_field_mapping = {
        "HBA_DUAL":        { "CS": ["HBA0","HBA1"], "RS": ["HBA"], "INT": ["HBA"] },
        "HBA_DUAL_INNER":  { "CS": ["HBA0","HBA1"], "RS": ["HBA"], "INT": ["HBA"] },
        "HBA_ONE":         { "CS": ["HBA1"],        "RS": ["HBA"], "INT": ["HBA"] },
        "HBA_ONE_INNER":   { "CS": ["HBA1"],        "RS": ["HBA"], "INT": ["HBA"] },
        "HBA_ZERO":        { "CS": ["HBA0"],        "RS": ["HBA"], "INT": ["HBA"] },
        "HBA_ZERO_INNER":  { "CS": ["HBA0"],        "RS": ["HBA"], "INT": ["HBA"] },
        "LBA_INNER":       { "CS": ["LBA"],         "RS": ["LBA"], "INT": ["LBA"] },
        "LBA_OUTER":       { "CS": ["LBA"],         "RS": ["LBA"], "INT": ["LBA"] },
        "LBA_SPARSE_EVEN": { "CS": ["LBA"],         "RS": ["LBA"], "INT": ["LBA"] },
        "LBA_SPARSE_ODD":  { "CS": ["LBA"],         "RS": ["LBA"], "INT": ["LBA"] },
        "LBA_ALL":         { "CS": ["LBA"],         "RS": ["LBA"], "INT": ["LBA"] },
    }

    def station_to_type(station: str) -> str:
        """Whether a station is core/remote/international."""
        if station.startswith("CS"):
            return "CS"
        if station.startswith("RS"):
            return "RS"
        return "INT"

    # ToDo: make a composite schema for the list of station specs per station
    response = {'stations': []}

    # get the template with the json schema to check validity of the l2stationspecs
    l2station_template = models.CommonSchemaTemplate.get_latest("lofar2station")

    for station in subtask.specifications_doc["stations"]["station_list"]:  # todo: I guess we would simply filter for lofar2-stations here (L2SS-1189)?
        antenna_set = subtask.specifications_doc["stations"]["antenna_set"]
        station_type = station_to_type(station)

        l2stationspecs = l2station_template.get_default_json_document_for_schema()

        l2stationspecs['station'] = station
        l2fields = l2stationspecs['antenna_fields']

        for field_index, antenna_field in enumerate(antenna_field_mapping[antenna_set][station_type]):
            if len(l2fields) <= field_index:
                # when using more than one field for this station:
                # copy specs of the first field so that we get all default values
                l2fields.append(copy.deepcopy(l2fields[0]))
            l2fields[field_index]['observation_id'] = subtask.id
            l2fields[field_index]['start_time'] = subtask.scheduled_start_time.isoformat()
            l2fields[field_index]['stop_time'] = subtask.scheduled_stop_time.isoformat()
            l2fields[field_index]['antenna_field'] = antenna_field
            l2fields[field_index]['antenna_set'] = antenna_set_mapping[antenna_set][station_type]
            l2fields[field_index]['HBA']['tile_beam'] = subtask.specifications_doc["stations"]["analog_pointing"]
            l2fields[field_index]['filter'] = subtask.specifications_doc["stations"]["filter"]
            l2fields[field_index]['SAPs'] = [{'pointing' : dp['pointing'], 'subbands' : dp['subbands'] } for dp in subtask.specifications_doc["stations"]["digital_pointings"]]

        # raises when something is not valid, silently passes when valid.
        l2station_template.validate_document(l2stationspecs)

        response['stations'].append(l2stationspecs)

    return response
