#!/usr/bin/python3

# Copyright (C) 2020  ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
plt.switch_backend('agg')

import math
from datetime import datetime, timedelta, date
from typing import Union, Iterable
from django.db.models import Q

from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.tmssapp.conversions import compute_elevation, coordinates_and_timestamps_to_separation_from_bodies, coordinates_timestamp_and_station_to_target_transit, create_astroplan_observer_for_station, local_sidereal_time_for_utc_and_station, calculate_and_get_sunrise_and_sunset_of_observer_day, local_sidereal_time_for_utc_and_station_as_pytime
from lofar.sas.tmss.services.scheduling.constraints import get_target_sap_pointings, get_calibrator_sap_pointings, get_sap_pointings, get_transit_offset_reference_pointing, get_transit_offset_pointings, get_earliest_possible_start_time, get_optimal_start_time, Gridder, get_at_constraint_timestamp, get_stations_to_be_evaluated, get_minmax_elevation_over_stations_and_pointings_for_observation_task,         get_transit_offset_pointings

from lofar.common.datetimeutils import round_to_second_precision

import logging
logger = logging.getLogger(__name__)


class UTC_LST_DateFormatter(mpl.dates.DateFormatter):
    '''Formatter which renders timestamps in UTC<enter>LST'''
    def __init__(self, *args, **kwargs):
        fmt = kwargs.pop('fmt', '%H:%M')
        super().__init__(fmt=fmt, *args, **kwargs)

    def __call__(self, x, pos=0):
        timestamp = mpl.dates.num2date(x, self.tz)
        timestamp_lst = local_sidereal_time_for_utc_and_station_as_pytime(timestamp)
        if math.isclose(x, self.locs[0], rel_tol=1e-6):
            # add UTC and LST labels for the first tick
            return timestamp.strftime(self.fmt) + " UTC\n" + timestamp_lst.strftime(self.fmt) + " LST"
        return timestamp.strftime(self.fmt) + "\n" + timestamp_lst.strftime(self.fmt)


def scheduling_constraints_plot(scheduling_units: Union[models.SchedulingUnitBlueprint, models.SchedulingUnitDraft, Iterable[models.SchedulingUnitBlueprint]],
                                station: str=None, proposed_start_time: datetime=None, proposed_date: date=None, include_min_max_elevations: bool=False) -> bytes:
    if station is None:
        station = 'CS002'

    # helper var and function to get a consistent target->color mapping within this scheduling_constraints_plot call
    __name2color = {}
    __colors = ('#ffdd00', '#0095ff', '#00ff8c', '#7700ff',
                '#00a85d', '#4f00a8', '#a8005a', '#a8a800', '#0062a8',
                '#52ffb1', '#ffff52', '#52b7ff', '#ff0088', '#a352ff')
    def get_color(name: str) -> tuple:
        '''always get the same color for the given name, picked from a palette'''
        if name not in __name2color:
            __name2color[name] = __colors[len(__name2color) % len(__colors)]
        return __name2color[name]

    single_unit_plot = isinstance(scheduling_units, models.SchedulingUnitDraft) or isinstance(scheduling_units, models.SchedulingUnitBlueprint)

    # if the given scheduling_unit(s) is actually just one instance, wrap it in a list to loop over
    if single_unit_plot:
        scheduling_units = [scheduling_units]

    # only plot distance axes when there is only one unit, with a min_distance constraint.
    do_plot_min_distance = single_unit_plot and 'min_distance' in scheduling_units[0].scheduling_constraints_doc.get('sky', {})

    # create title
    if single_unit_plot:
        title = '%s %s SU[%s] %s' % (scheduling_units[0].project.name, station, scheduling_units[0].id, scheduling_units[0].name)
        if proposed_start_time is None and proposed_date is None:
            if isinstance(scheduling_units[0], models.SchedulingUnitBlueprint):
                # add status and (maybe) unschedulable reason to title
                title += " (%s)" % (scheduling_units[0].status.value,)
                if scheduling_units[0].status.value == models.SchedulingUnitStatus.Choices.UNSCHEDULABLE.value:
                    title += "\n" + scheduling_units[0].unschedulable_reason
    else:
        title = 'Schedule for %s' % (proposed_date)

    # create the figure
    fig, axes = plt.subplots(figsize=(16, 10),
                             nrows=2 if do_plot_min_distance else 1)
    fig.suptitle(title)

    elevation_axes = axes if isinstance(axes, mpl.axes.Axes) else axes[0]
    elevation_axes.set_ylim(0, 90)
    elevation_axes.set_ylabel('elevation [deg]')


    # determine x-axis time range and timestamp_to_x_value function
    if proposed_date is not None:
        lower_timestamp = datetime(proposed_date.year, proposed_date.month, proposed_date.day, 0, 0, 0)
        center_timestamp = lower_timestamp + timedelta(hours=12)
        upper_timestamp = lower_timestamp + timedelta(hours=24)
    elif proposed_start_time is not None:
        lower_timestamp = proposed_start_time
        upper_timestamp = lower_timestamp + max([s.specified_observation_duration for s in scheduling_units])
        center_timestamp = lower_timestamp + 0.5 * (upper_timestamp - lower_timestamp)
    else:
        if single_unit_plot and isinstance(scheduling_units[0], models.SchedulingUnitBlueprint):
            lower_timestamp = min([obs_task.on_sky_start_time for obs_task in scheduling_units[0].observation_tasks])
            upper_timestamp = max([obs_task.on_sky_stop_time for obs_task in scheduling_units[0].observation_tasks])
        else:
            # just pick 'now' for drafts if proposed_start_time/proposed_date is not given
            lower_timestamp = datetime.utcnow()
            upper_timestamp = lower_timestamp + max([s.specified_observation_duration for s in scheduling_units])
        # extend the time range to at least an hour
        extra = (timedelta(hours=3) - (upper_timestamp-lower_timestamp))/2
        if extra > timedelta(0):
            lower_timestamp -= extra
            upper_timestamp += extra
        center_timestamp = lower_timestamp + 0.5 * (upper_timestamp - lower_timestamp)

    elevation_axes.xaxis.set_major_locator(mpl.dates.HourLocator() if (upper_timestamp-lower_timestamp)>=timedelta(hours=4) else
                                           mpl.dates.MinuteLocator(interval=15) if (upper_timestamp-lower_timestamp)>=timedelta(hours=2) else
                                           mpl.dates.MinuteLocator(interval=10) if (upper_timestamp-lower_timestamp)>=timedelta(hours=1) else
                                           mpl.dates.MinuteLocator(interval=5) if (upper_timestamp-lower_timestamp)>=timedelta(minutes=30) else
                                           mpl.dates.MinuteLocator(interval=1) )
    elevation_axes.xaxis.set_major_formatter(UTC_LST_DateFormatter())
    elevation_axes.set_xlabel('Time in UTC/LST on %s' % (lower_timestamp.date().strftime('%Y-%m-%d'),))
    elevation_axes.set_xlim(lower_timestamp, upper_timestamp)
    elevation_axes.grid(True)

    if do_plot_min_distance:
        # setup separate min_distance_axes for bar charts
        min_distance_axes = axes[1]
        min_distance_axes.set_ylim(0, 180)
        min_distance_axes.set_ylabel('min angular separation [deg]')
        min_distance_axes.yaxis.grid(True)
        min_distance_axes.set_axisbelow(True)
    else:
        min_distance_axes = None

    observer = create_astroplan_observer_for_station(station)

    # plot day/night/twilight if daily_contraints or range is large enough
    if any([any(s.scheduling_constraints_doc.get('daily',{}).values()) for s in scheduling_units]) or \
       upper_timestamp-lower_timestamp > timedelta(hours=8):
        try:
            # get sun rise/set events
            sun_events_yesterday = calculate_and_get_sunrise_and_sunset_of_observer_day(observer, center_timestamp - timedelta(days=1))
            sun_events_today = calculate_and_get_sunrise_and_sunset_of_observer_day(observer, center_timestamp)
            sun_events_tomorrow = calculate_and_get_sunrise_and_sunset_of_observer_day(observer, center_timestamp + timedelta(days=1))

            # deduce day/night/twilight start/stop times
            start_prev_night = sun_events_yesterday[1]['end']
            end_prev_night = sun_events_today[0]['start']
            end_morning_twilight = sun_events_today[0]['end']
            start_evening_twilight = sun_events_today[1]['start']
            end_evening_twilight = sun_events_today[1]['end']
            end_night = sun_events_tomorrow[0]['start']
            start_morning_twilight = end_prev_night
            start_day = end_morning_twilight
            end_day = start_evening_twilight
            start_night = end_evening_twilight

            # shade night/twilight areas
            elevation_axes.fill_between((start_prev_night, end_prev_night), (90, 90),
                                        facecolor='dimgray', alpha=0.2)
            elevation_axes.fill_between((start_morning_twilight, end_morning_twilight), (90, 90),
                                        facecolor='gray', alpha=0.1)
            elevation_axes.fill_between((start_evening_twilight, end_evening_twilight), (90, 90),
                                        facecolor='gray', alpha=0.1)
            elevation_axes.fill_between((start_night, end_night), (90, 90),
                                        facecolor='dimgray', alpha=0.2)
        except TypeError as e:
            # Note: when the target is always above or below horizon, astroplan excepts with the not very
            #  meaningful error: 'numpy.float64' object does not support item assignment
            if "numpy.float64" in str(e):
                logger.error("sun is always up/down at station %s", station)
            else:
                raise

    for scheduling_unit in scheduling_units:
        # get the scheduling_unit's observation starttime either from the given proposed_start_time, or as the first observation's starttime
        # derive the su's stoptime and center time from that.
        su_obs_start_time = scheduling_unit.scheduled_start_time
        if single_unit_plot:
            if proposed_date is not None:
                # plot this unit at the transit
                su_obs_start_time = get_optimal_start_time(scheduling_unit, lower_timestamp, gridder=Gridder(Gridder.COARSE_TIME_GRID))
            elif proposed_start_time is not None:
                su_obs_start_time = proposed_start_time

        if su_obs_start_time is None:
            continue
        su_obs_stop_time = su_obs_start_time + scheduling_unit.specified_observation_duration

        # loop over the constraints, and plot a green 'required_*' bar, or a red 'avoid_*' bar
        for constraint, value in scheduling_unit.scheduling_constraints_doc.get('daily',{}).items():
            if value:
                start_end_pairs = ((start_day, end_day),) if constraint=='require_day' else \
                                  ((start_prev_night, end_prev_night), (start_night, end_night),) if constraint == 'require_night' else \
                                  ((start_morning_twilight, end_morning_twilight), (start_evening_twilight, end_evening_twilight))

                for start, end in start_end_pairs:
                    clipped_start = min(max(lower_timestamp, start), upper_timestamp)
                    clipped_end = max(min(upper_timestamp, end), lower_timestamp)
                    if clipped_start >= lower_timestamp and clipped_end <= upper_timestamp and clipped_start != clipped_end:
                        elevation_axes.fill_between((clipped_start, clipped_end), (3, 3),
                                                    alpha=0.5,
                                                    color='lime' if constraint.startswith('require_') else 'red')
                        elevation_axes.text(clipped_start if start < center_timestamp else clipped_end, 0,
                                            constraint.replace('_', ' '),
                                            horizontalalignment='left' if start < center_timestamp else 'right',
                                            verticalalignment='bottom')

        # prepare helper values for min_distance_constraints
        min_distance_constraints = scheduling_unit.scheduling_constraints_doc.get('sky', {}).get('min_distance', {})
        total_nr_of_pointing_body_pairs = sum([len(get_sap_pointings(x))*len(min_distance_constraints.get('target', min_distance_constraints) if x.is_target_observation else min_distance_constraints.get('calibrator', {}))
                                               for x in scheduling_unit.observation_tasks])
        pointing_body_pair_cntr = 0

        # loop over all observation tasks
        obs_tasks = sorted(scheduling_unit.observation_tasks, key=lambda t: t.relative_start_time)
        for obs_idx, obs_task in enumerate(obs_tasks):
            # create x-values as timestamps
            obs_start_timestamp = su_obs_start_time + obs_task.relative_start_time
            obs_stop_timestamp = su_obs_start_time + obs_task.relative_stop_time
            obs_timestamps = [round_to_second_precision(obs_start_timestamp + timedelta(seconds=sec)) for sec in range(0, int(round(obs_task.specified_duration.total_seconds())), 10*60)]
            if obs_timestamps[-1] < obs_stop_timestamp:
                obs_timestamps.append(obs_stop_timestamp)

            # gather all relevant pointings to plot
            cal_pointings = get_calibrator_sap_pointings(obs_task)
            target_pointings = get_target_sap_pointings(obs_task)
            pointings = cal_pointings + target_pointings
            ref_pointing = get_transit_offset_reference_pointing(obs_task)
            if ref_pointing and obs_task.is_target_observation:
                pointings.append(ref_pointing)

            # plot an elevation line chart per pointing
            for pointing in pointings:
                obs_elevations = [compute_elevation(pointing=pointing, timestamp=timestamp, station=station) * 180.0 / 3.1415 for timestamp in obs_timestamps]
                # simple line plot
                elevation_axes.plot(obs_timestamps, obs_elevations,
                                    label=pointing.str_astro() + (" Cal" if pointing in cal_pointings else ''),
                                    linestyle='dashdot' if pointing in cal_pointings else 'solid',
                                    linewidth=3 if pointing==ref_pointing else 1.5 if pointing in cal_pointings else 2.5,
                                    alpha=1 if ref_pointing is None or pointing==ref_pointing else 0.5,
                                    color=get_color(pointing.target))

                # fill between line and time-axis.
                elevation_axes.fill_between(obs_timestamps, obs_elevations,
                                            linewidth=0,
                                            alpha=0.2,
                                            color=get_color(pointing.target))

                if include_min_max_elevations:
                    boundary_stations = get_stations_to_be_evaluated(scheduling_unit.main_observation_task, obs_start_timestamp, 10e3)
                    if len(boundary_stations) > 1:
                        boundary_elevations = [[compute_elevation(pointing=pointing, timestamp=timestamp, station=station) * 180.0 / 3.1415 for station in boundary_stations] for timestamp in obs_timestamps]
                        min_boundary_elevations = [min(elv) for elv in boundary_elevations]
                        max_boundary_elevations = [max(elv) for elv in boundary_elevations]
                        elevation_axes.plot(obs_timestamps, min_boundary_elevations, linestyle='dotted', linewidth=1.5, alpha=1, color=get_color(pointing.target))
                        elevation_axes.plot(obs_timestamps, max_boundary_elevations, linestyle='dotted', linewidth=1.5, alpha=1, color=get_color(pointing.target))

            if obs_task.is_target_observation:
                # indicate the center of the observation (so you can compare it to transit/transit_offset)
                obs_center_timestamp = obs_start_timestamp + 0.5 * obs_task.specified_duration
                min_elevation_at_center, max_elevation_at_center = get_minmax_elevation_over_stations_and_pointings_for_observation_task(obs_task, obs_center_timestamp, True)
                min_elevation_at_center *= 180.0 / 3.1415
                max_elevation_at_center *= 180.0 / 3.1415
                center_lst = local_sidereal_time_for_utc_and_station(obs_center_timestamp, station=station)
                center_lst_time = timedelta(minutes=int(round(center_lst.hour * 60.0)))
                center_label = "%s [%s] %s\n%sUTC %sLST" % (scheduling_unit.project.name,
                                                            scheduling_unit.id,
                                                            scheduling_unit.name,
                                                            obs_center_timestamp.strftime('%H:%M'),
                                                            str(center_lst_time)[:-3])
                elevation_axes.annotate(center_label,
                                        xy=(obs_center_timestamp, 0), xycoords='data',
                                        xytext=(obs_center_timestamp, max_elevation_at_center-obs_idx*8), textcoords='data',
                                        arrowprops=dict(color="slateblue", arrowstyle="-", linestyle='dashed', alpha=0.75),
                                        horizontalalignment='center', verticalalignment='bottom', alpha=0.8, annotation_clip=None)

                if single_unit_plot:
                    transit_offset_constraints = scheduling_unit.scheduling_constraints_doc.get('sky', {}).get('transit_offset', {})

                    # for target observation(s) also annotate the transit info (transit timestamp in UTC & LST, and elevation)
                    for p_idx, pointing in enumerate(get_transit_offset_pointings(obs_task)):
                        transit_timestamp = coordinates_timestamp_and_station_to_target_transit(pointing, obs_center_timestamp, observer, n_grid_points=200)

                        # plot the limits within which this transit for this pointing should occur
                        center_limit_timestamps = {}
                        for offset in ('from', 'to'):
                            if offset in transit_offset_constraints:
                                center_limit_timestamp = transit_timestamp + timedelta(seconds=transit_offset_constraints[offset])
                                center_limit_timestamps[offset] = center_limit_timestamp
                                if single_unit_plot:
                                    center_limit_lst = local_sidereal_time_for_utc_and_station(center_limit_timestamp, station=station)
                                    center_limit_lst_time = timedelta(minutes=int(round(center_limit_lst.hour * 60.0)))
                                    center_limit_label = 'transit offset %s limit\n%sUTC %sLST' % ('lower' if offset == 'from' else 'upper',
                                                                                                   center_limit_timestamp.strftime('%H:%M'),
                                                                                                   str(center_limit_lst_time)[:-3])
                                else:
                                    center_limit_label = None

                                elevation_axes.annotate(center_limit_label,
                                                        xy=(center_limit_timestamp, 0),
                                                        xytext=(center_limit_timestamp, 5+p_idx*8 if center_limit_label else 5),
                                                        xycoords='data',
                                                        textcoords='data',
                                                        arrowprops=dict(color=get_color(pointing.target),
                                                                        arrowstyle="-",
                                                                        linestyle='dashed',
                                                                        linewidth=1),
                                                        horizontalalignment='center',
                                                        verticalalignment='bottom')

                        if 'from' in center_limit_timestamps and 'to' in center_limit_timestamps:
                            elevation_axes.fill_between((center_limit_timestamps['from'], center_limit_timestamps['to']),
                                                        (5, 5),
                                                        facecolor=get_color(pointing.target),
                                                        alpha=.5)

                        if transit_timestamp >= su_obs_start_time and transit_timestamp <= su_obs_stop_time:
                            # plot the transit for this pointing with
                            #  * an arrow indicating
                            #    - the transit timestamp in UTC and LST
                            #    - the elevation
                            #  * an arrow indicating
                            #    - the offset in minutes between center_of_obs and transit
                            elevation_at_transit = compute_elevation(pointing, transit_timestamp, station) * 180.0 / 3.1415
                            lst = local_sidereal_time_for_utc_and_station(transit_timestamp, station=station)
                            lst_time = timedelta(minutes=int(round(lst.hour*60.0)))
                            label = "%sUTC %sLST %.1fdeg" % (transit_timestamp.strftime('%H:%M'),
                                                             str(lst_time)[:-3],
                                                             elevation_at_transit)

                            elevation_axes.annotate(label,
                                                    xy=(transit_timestamp, elevation_at_transit), xycoords='data',
                                                    xytext=(transit_timestamp, elevation_at_transit-(p_idx+2)*8), textcoords='data',
                                                    arrowprops=dict(facecolor=get_color(pointing.target), alpha=0.5, edgecolor='dimgray', width=2, headwidth=5),
                                                    horizontalalignment='center',
                                                    verticalalignment='top' )

                            offset_from_transit = obs_center_timestamp-transit_timestamp
                            if abs(offset_from_transit.total_seconds()) >= 300:
                                # plot an arrow and offset-label if offset is big-ish
                                offset_label = "%dmin" % (offset_from_transit.total_seconds()/60.0,)
                                elevation_axes.annotate(offset_label,
                                                        xy=(transit_timestamp, elevation_at_transit+2-(p_idx+2)*8), xycoords='data',
                                                        xytext=(obs_center_timestamp, elevation_at_transit+2-(p_idx+2)*8), textcoords='data',
                                                        arrowprops=dict(color=get_color(pointing.target), arrowstyle="<|-|>", alpha=0.5),
                                                        horizontalalignment='left' if offset_from_transit > timedelta(seconds=0) else 'right',
                                                        verticalalignment='center')


            if 'min_elevation' in scheduling_unit.scheduling_constraints_doc.get('sky', {}):
                for obs_type in ('calibrator', 'target'):
                    if (obs_type == 'calibrator' and obs_task.is_calibrator_observation) or (obs_type == 'target' and obs_task.is_target_observation):
                        min_elevation = scheduling_unit.scheduling_constraints_doc['sky']['min_elevation'][obs_type]  * 180.0 / 3.1415
                        elevation_axes.plot((obs_timestamps[0], obs_timestamps[-1]), (min_elevation, min_elevation),
                                            label='_nolegend_',
                                            linestyle='dashed',
                                            linewidth=1,
                                            alpha=.5,
                                            color='#cc8585' if obs_type=='target' else '#eaad66')
                        elevation_axes.fill_between((obs_timestamps[0], obs_timestamps[-1]), (min_elevation, min_elevation),
                                                    facecolor='red' if obs_type=='target' else 'orange',
                                                    linewidth=0,
                                                    alpha=.15)
                        if single_unit_plot:
                            elevation_axes.text(obs_timestamps[0] if obs_start_timestamp < center_timestamp else obs_timestamps[-1],
                                                min_elevation + 1,
                                                'min %s elevation %.1f[deg]' % (obs_type, min_elevation),
                                                horizontalalignment='left' if obs_start_timestamp < center_timestamp else 'right',
                                                verticalalignment='bottom')


            # plot min_distances
            if do_plot_min_distance and 'min_distance' in scheduling_unit.scheduling_constraints_doc.get('sky', {}):
                # disect the constraints in a 'target' and 'calibrator' part which works for all versions
                # In TMSS-2877, constraints template version 9 was created with two sections for 'target' and 'calibrator'.
                # before version 9, we only had one section which was for the target only.
                min_distance_obs_constraints = min_distance_constraints.get('target', min_distance_constraints) if obs_task.is_target_observation else min_distance_constraints.get('calibrator', {})
                sap_pointings = target_pointings if obs_task.is_target_observation else cal_pointings

                # Angular separation between target and body hardly varies over the unit's observation window
                # So, plotting them as a line would produce lots of near-horizontal boring graphs.
                # So, plot them as bar, per target-body pair.
                pointing_body_pairs = []
                pointing_body_separations = []
                pointing_body_colors = []
                # loop over the alphabetically sorted bodies
                bodies = sorted(min_distance_obs_constraints.keys())
                for b_idx, body in enumerate(bodies):
                    min_distance_constraint = min_distance_obs_constraints[body]
                    for p_idx, pointing in enumerate(sap_pointings):
                        # Find minimum separation over observation window
                        min_distance = min([d.degree for d in coordinates_and_timestamps_to_separation_from_bodies(pointing=pointing,
                                                                                                                   timestamps=obs_timestamps,
                                                                                                                   bodies=[body])[body].values()])

                        min_distance_deg = min_distance_constraint * 180.0 / 3.1415
                        pointing_body_pairs.append("%s\n%s" % (pointing.target, body))
                        pointing_body_separations.append(min_distance)
                        pointing_body_colors.append(get_color(pointing.target))

                        x_low = pointing_body_pair_cntr-0.5
                        x_high = x_low + 1
                        pointing_body_pair_cntr += 1

                        min_distance_axes.plot([x_low, x_high],
                                               [min_distance_deg, min_distance_deg],
                                               linestyle='dashed',
                                               linewidth=1.5,
                                               color=get_color(body))
                        min_distance_axes.text(x_low,
                                               min_distance_deg + 1,
                                               'min %s %.1f[deg]'%(body, min_distance_deg),
                                               horizontalalignment='left',
                                               verticalalignment='bottom')

                        min_distance_axes.fill_between([x_low, x_high],
                                                       [min_distance_deg, min_distance_deg],
                                                       facecolor=get_color(body),
                                                       alpha=.15,
                                                       zorder=4)

                min_distance_axes.bar(pointing_body_pairs, pointing_body_separations, color=pointing_body_colors)

    # add legends
    if single_unit_plot:
        elevation_axes.legend(loc='upper right')

    # let matplotlib do its layouting
    fig.tight_layout()

    # save in-memory to a png "file", and return the png-byte-array
    from io import BytesIO
    with BytesIO() as mem_file:
        plt.savefig(mem_file, format='png')
        mem_file.seek(0)
        return mem_file.getvalue()


def scheduling_constraints_plot_for_date(proposed_date: date, station: str=None) -> bytes:
    '''create a scheduling_constraints plot for all scheduled/running/finished scheduling_units on the given date'''
    scheduling_units = models.SchedulingUnitBlueprint.objects.filter(status__value__in=(models.SchedulingUnitStatus.Choices.SCHEDULED.value,
                                                                                          models.SchedulingUnitStatus.Choices.OBSERVING.value,
                                                                                          models.SchedulingUnitStatus.Choices.OBSERVED.value,
                                                                                          models.SchedulingUnitStatus.Choices.PROCESSING.value,
                                                                                          models.SchedulingUnitStatus.Choices.PROCESSED.value,
                                                                                          models.SchedulingUnitStatus.Choices.INGESTING.value,
                                                                                          models.SchedulingUnitStatus.Choices.INGESTED.value,
                                                                                          models.SchedulingUnitStatus.Choices.FINISHED.value))
    lower_bound = datetime(proposed_date.year, proposed_date.month, proposed_date.day, 0, 0, 0)
    upper_bound = lower_bound + timedelta(hours=24)
    scheduling_units = scheduling_units.filter(scheduled_stop_time__gte=lower_bound)
    scheduling_units = scheduling_units.filter(scheduled_start_time__lt=upper_bound)
    scheduling_units = list(scheduling_units.all())

    return scheduling_constraints_plot(scheduling_units, station=station, proposed_date=proposed_date,
                                       proposed_start_time=None, include_min_max_elevations=False)


def lst_pressure_plot(lower_bound: datetime=None, upper_bound: datetime=None, include_calibrators: bool=True, projects: list=None) -> bytes:
    # create "bins" per minute for a full sidereal 23h56m day
    # start with 0 pressure for each minute-bin
    if lower_bound is None:
        lower_bound = datetime.utcnow().replace(hour=0, minute=0, second=0, microsecond=0)

    if upper_bound is not None and upper_bound <= lower_bound:
        raise ValueError("upper_bound should be larger than lower_bound")

    lower_bound_day_start = lower_bound.replace(hour=0, minute=0, second=0, microsecond=0)
    lst_values = [lower_bound_day_start + timedelta(minutes=m) for m in range(24*60-3)] # LST day is 23h56m4s, so we have 24*60 minus 3 full minutes bins
    num_bins = len(lst_values)
    pressure_series = {'scheduled': {'color': 'blue', 'data': [0 for _ in lst_values]},
                       'schedulable': {'color': 'lightblue', 'data': [0 for _ in lst_values]},
                       'unschedulable': {'color': 'darkred', 'data': [0 for _ in lst_values]},
                       'observing': {'color': 'yellow', 'data': [0 for _ in lst_values]},
                       'observed': {'color': 'green', 'data': [0 for _ in lst_values]},
                       'cancelled': {'color': 'orange', 'data': [0 for _ in lst_values]},
                       'error': {'color': 'red', 'data': [0 for _ in lst_values]}
                       }

    # prepare some filters to query for observation tasks with certain statuses
    filters = {
               'scheduled': Q(status__value=models.TaskStatus.Choices.SCHEDULED.value),
               'schedulable': Q(status__value=models.SchedulingUnitStatus.Choices.SCHEDULABLE.value),
               'unschedulable': Q(status__value=models.TaskStatus.Choices.UNSCHEDULABLE.value),
               'observing': Q(status__value__in=(models.TaskStatus.Choices.QUEUED.value, models.TaskStatus.Choices.STARTED.value)),
               'observed': Q(status__value=models.TaskStatus.Choices.FINISHED.value),
               'error': Q(status__value=models.TaskStatus.Choices.ERROR.value),
               'cancelled': Q(status__value=models.TaskStatus.Choices.CANCELLED.value)
               }

    for status, q_filter in filters.items():
        observation_tasks = models.TaskBlueprint.objects.filter(q_filter).filter(obsolete_since__isnull=True).filter(specifications_template__type__value=models.TaskType.Choices.OBSERVATION.value)

        if projects:
            observation_tasks = observation_tasks.filter(scheduling_unit_blueprint__draft__scheduling_set__project__name__in=projects)

        if status not in ('schedulable', 'unschedulable'):
            if lower_bound:
                observation_tasks = observation_tasks.filter(on_sky_stop_time__gte=lower_bound)
            if upper_bound:
                observation_tasks = observation_tasks.filter(on_sky_start_time__lt=upper_bound)

        for obs in observation_tasks.order_by('id').all():
            try:
                # skip calibrator-only observations if requested (keep combined target/calibrator)
                if not include_calibrators and obs.is_calibrator_observation and not obs.is_target_observation:
                    continue

                if status in ('schedulable', 'unschedulable'):
                    # (un)schedulable tasks do not have a set starttime.
                    # so, derive the expected starttime from the pointing(s) transit lst time.
                    pointings = get_transit_offset_pointings(obs)
                    if include_calibrators:
                        pointings += get_calibrator_sap_pointings(obs)

                    if pointings:
                        # how much does each pointing contribute to the overall pressure?
                        # for the largest part of the observation, observation-window for each target overlaps with the other windows.
                        # when all windows for all pointings overlap, then the combined pressure contribution/share is 1.0
                        pointing_pressure_share = 1.0/len(pointings)
                        duration_in_min = int(math.ceil(obs.specified_duration.total_seconds() / 60.0))
                        for pointing in pointings:
                            # transit is/shouldbe in the center of the obs, so start is transit minus half duration
                            start_lst = datetime.combine(lower_bound_day_start.date(), pointing.transit_lst) - 0.5*obs.specified_duration

                            # how many minutes since day start?
                            start_lst_in_minutes = int(round((start_lst - lower_bound_day_start).total_seconds() / 60.0))

                            # fill bins: loop over each full minute in the observation
                            pressures = pressure_series[status]['data']
                            for minute in range(start_lst_in_minutes, start_lst_in_minutes+duration_in_min):
                                bin_number = minute % num_bins # wrap around 23h56m
                                pressures[bin_number] += pointing_pressure_share
                else:
                    # get start/stop times, bounded by lower-/upper_bound
                    start_utc = max(lower_bound, obs.on_sky_start_time)
                    stop_utc = min(upper_bound, obs.on_sky_stop_time) if upper_bound else obs.on_sky_stop_time
                    duration_in_min = int(math.ceil((stop_utc-start_utc).total_seconds()/60.0))

                    # convert to UTC -> LST
                    start_lst = local_sidereal_time_for_utc_and_station_as_pytime(start_utc)
                    # how many minutes since day start?
                    start_lst_in_minutes = int(round((datetime.combine(lower_bound_day_start.date(), start_lst) - lower_bound_day_start).total_seconds() / 60.0))

                    # fill bins: loop over each full minute in the observation
                    pressures = pressure_series[status]['data']
                    for minute in range(start_lst_in_minutes, start_lst_in_minutes+duration_in_min):
                        bin_number = minute % num_bins # wrap around 23h56m
                        pressures[bin_number] += 1
            except Exception as e:
                logger.error(e)

    # apply some nice median filters to remove visually annoying spikes.
    try:
        from scipy import signal
        for status, series in pressure_series.items():
            series['data'] = signal.medfilt(series['data'], 5)
    except Exception as e:
        pass

    if upper_bound is None:
        # derive upper_bound from selected units
        from django.db.models import Max
        scheduling_units = models.SchedulingUnitBlueprint.objects.filter(status__value=models.SchedulingUnitStatus.Choices.SCHEDULED.value)
        scheduling_units = scheduling_units.filter(obsolete_since__isnull=True)
        upper_bound = scheduling_units.aggregate(Max('on_sky_stop_time'))['on_sky_stop_time__max'] or lower_bound+timedelta(days=365)

    # used for reporting in title
    upper_bound_day_start = upper_bound.replace(hour=0, minute=0, second=0, microsecond=0)
    if upper_bound_day_start == lower_bound_day_start:
        upper_bound_day_start += timedelta(days=1)

    # create the figure
    fig, axes = plt.subplots(figsize=(16, 10))
    fig.suptitle("LST pressure %s - %s %s%s" % (lower_bound.strftime('%Y-%m-%d'),
                                                upper_bound_day_start.strftime('%Y-%m-%d'),
                                                ','.join(projects)+' ' if projects else '',
                                                'targets and calibrators' if include_calibrators else 'targets only'))

    axes.set_xlim(lower_bound_day_start, lower_bound_day_start+timedelta(hours=23, minutes=56, seconds=4))
    # axes.set_ylim(0, max(pressures))
    # See todo above about normalization
    # axes.set_ylabel('Pressure [#obs/#opportunities as %]')
    axes.set_ylabel('Pressure [#obs]')

    axes.xaxis.set_major_formatter(mpl.dates.DateFormatter('%H:%M'))
    axes.xaxis.set_major_locator(mpl.dates.HourLocator())
    axes.set_xlabel('Local Sidereal Time')
    axes.grid(True)

    # plot the lst-pressure graph
    labels = ['observed', 'error', 'cancelled', 'observing', 'scheduled', 'schedulable', 'unschedulable']
    y_values = [pressure_series[label]['data'] for label in labels]
    colors = [pressure_series[label]['color'] for label in labels]
    axes.stackplot(lst_values, y_values, labels=labels, colors=colors, alpha=0.75)
    axes.legend(loc='upper right')

    # let matplotlib do its layouting
    fig.tight_layout()

    # save in-memory to a png "file", and return the png-byte-array
    from io import BytesIO
    with BytesIO() as mem_file:
        plt.savefig(mem_file, format='png')
        mem_file.seek(0)
        return mem_file.getvalue()
