from lofar.sas.tmss.tmss.exceptions import *
from lofar.sas.tmss.tmss.tmssapp.models.scheduling import Dataproduct, SubtaskType, Subtask, SubtaskOutput, SubtaskState, SIPidentifier, HashAlgorithm
from lofar.sas.tmss.tmss.tmssapp.models.specification import Datatype, Dataformat
from lofar.lta.sip import siplib, ltasip, validator, constants
from lofar.sas.tmss.tmss.tmssapp.adapters.parset import convert_to_parset
from lofar.sas.tmss.tmss.tmssapp.subtasks import used_pipeline_stations

import logging
import isodate
import datetime
from functools import lru_cache
logger = logging.getLogger(__name__)

mapping_antennaset_type_TMSS_2_SIP = {
    "HBA_DUAL": constants.ANTENNASETTYPE_HBA_DUAL,
    "HBA_DUAL_INNER": constants.ANTENNASETTYPE_HBA_DUAL_INNER,
    "HBA_JOINED": constants.ANTENNASETTYPE_HBA_JOINED,
    "HBA_JOINED_INNER": constants.ANTENNASETTYPE_HBA_JOINED_INNER,
    "HBA_ONE": constants.ANTENNASETTYPE_HBA_ONE,
    "HBA_ONE_INNER": constants.ANTENNASETTYPE_HBA_ONE_INNER,
    "HBA_ZERO": constants.ANTENNASETTYPE_HBA_ZERO,
    "HBA_ZERO_INNER": constants.ANTENNASETTYPE_HBA_ZERO_INNER,
    "LBA_INNER": constants.ANTENNASETTYPE_LBA_INNER,
    "LBA_OUTER": constants.ANTENNASETTYPE_LBA_OUTER,
    "LBA_SPARSE_EVEN": constants.ANTENNASETTYPE_LBA_SPARSE_EVEN,
    "LBA_SPARSE_ODD": constants.ANTENNASETTYPE_LBA_SPARSE_ODD,
    "LBA_ALL": "NOT SUPPORTED IN LTA YET"
}

mapping_filterset_type_TMSS_2_SIP = {
    "LBA_10_90": constants.FILTERSELECTIONTYPE_10_90_MHZ,
    "LBA_30_90": constants.FILTERSELECTIONTYPE_30_90_MHZ,
    "HBA_110_190": constants.FILTERSELECTIONTYPE_110_190_MHZ,
    "HBA_210_250": constants.FILTERSELECTIONTYPE_210_250_MHZ
}

mapping_beamformer_stokes_type_TMSS_2_SIP = {
    "I": ["I"],
    "IQUV": ["I", "Q", "U", "V"],
    "XXYY": ["Xre", "Xim", "Yre", "Yim"]
}

mapping_pulsar_selection_type_TMSS_2_SIP = {
    "manual": "Pulsars in observation specs",
    "meta": "Pulsar specified in dataproduct",
    "sapfind": "Brightest known pulsar in SAP",
    "sapfind3": "Three brightest known pulsars in SAP",
    "tabfind": "Brightest known pulsar in TAB",
    "tabfind+": "Pulsars in observation specs, file and brightest in SAP and TAB"
}


def get_number_of_dataproducts_of_format(subtask, dataproduct_dataformat):
    """
    Retrieve the number of dataproducts of given data format in subtask:
    :param subtask:
    :param dataproduct_dataformat: Dataformat like "MeasurementSet", "Beamformed", "QA_HDF5", "QA_Plots", "pulp summary", "pulp analysis"
    :return:
    """
    nbr_dataproduct = 0
    subtask_outputs = list(SubtaskOutput.objects.filter(subtask__id=subtask.id))
    for subtask_output in subtask_outputs:
        dataproducts = list(Dataproduct.objects.filter(producer__id=subtask_output.id,
                                                       dataformat=dataproduct_dataformat))
        nbr_dataproduct += len(dataproducts)
    return nbr_dataproduct


def get_siplib_stations_list(stationlist, antennaset):
    """
    Retrieve a list of siplib Stations objects which is extracted from the station_list and the antennaset
    :param stationlist:
    :param antennaset:
    :return:

    Conversion logic see here: https://support.astron.nl/confluence/display/TMSS/Dataproduct+Provenance
    """
    siplib_station_list = []
    for station in stationlist:
        if antennaset.startswith('LBA'):
            antennafieldtypes = ['LBA']
        elif antennaset.startswith('HBA') and not station.startswith('CS'):
            antennafieldtypes = ['HBA']
        elif antennaset.startswith('HBA_DUAL'):
            antennafieldtypes = ['HBA0', 'HBA1']
        elif antennaset.startswith('HBA_ZERO'):
            antennafieldtypes = ['HBA0']
        elif antennaset.startswith('HBA_ONE'):
            antennafieldtypes = ['HBA1']
        else:
            raise ValueError('Cannot determine antennafields for station=%s antennaset=%s' % (station, antennaset))
        siplib_station_list.append(siplib.Station.preconfigured(station, antennafieldtypes))
    return siplib_station_list


def _create_siplib_identifier_object(source: str, unique_id: int, name: str, suppress_warning=False):
    """
    Create a SIP identifier object without using a TMSS SIPidentifer object.
    This should ONLY be used directly for identifiers that are NOT actually referring to an LTA entity but are used for
    grouping and such things, and hence don't require a global identifier.
    """
    identifier = siplib.Identifier.__new__(siplib.Identifier)
    identifier._set_pyxb_identifier(
        ltasip.IdentifierType(
            source=source,
            identifier=unique_id,
            name=name,
            label=None),
        suppress_warning=suppress_warning)
    return identifier


def get_siplib_identifier(sipid_obj: SIPidentifier, context_str="", name:str=None) -> siplib.Identifier:
    """
    Retrieve an Identifier object. Get the unique_identifier and source of the given sip object and covert that to
    a siblib object
    :param sipid_obj: SIP Identifier object
    :param context_str: In case of failure add some context info
    :return: A lofar.lta.sip.siplib.Identifier object
    """
    unique_id = sipid_obj.unique_identifier
    if unique_id is None:
        raise TMSSException("Can not create SIP Identifier, the value of 'unique_id' is None for %s" % context_str)
    source = sipid_obj.source
    if source is None:
        raise TMSSException("Can not create SIP Identifier, the value 'source' is None for %s" % context_str)
    return _create_siplib_identifier_object(source, unique_id, name, suppress_warning=True)


# todo: TMSS defines relations on task level. Technically, I guess it is more correct to go to subtask level when
#  adding provenance info with all intermediate dataproducts. So in case a task consists of several chaines processes,
#  you get the individual applies steps described in the SIP. Clarify that this is really what we want.
def add_subbtask_and_input_dataproducts_to_sip(subtask: Subtask, sip: siplib.Sip):
    """
    add a subtask including its input dataproducts to the sip, according to the specific observation or pipeline type they represent
    :param subtask: the Subtask object to add to the sip
    :param sip: the Sip object to add to
    """
    if subtask.state.value != SubtaskState.Choices.FINISHED.value:
        logger.debug("Skipping addition of subtask id %s of type %s to sip because it is not finished but %s." % (subtask.id, subtask.specifications_template.type.value, subtask.state.value))
        return

    if subtask.is_obsolete:
        logger.debug("Skipping addition of subtask id %s of type %s to sip because it is obsolete." % (subtask.id, subtask.specifications_template.type.value))
        return

    logger.debug("Adding subtask id %s global_id %s of type %s to sip." % (subtask.id, subtask.global_identifier.unique_identifier, subtask.specifications_template.type.value))
    parset_sip_identifier = get_siplib_identifier(subtask.global_parset_identifier, "Parset for Subtask id=%s" % subtask.id)
    sip_subtask = create_sip_representation_for_subtask(subtask, parset_sip_identifier)
    if isinstance(sip_subtask, siplib.Observation):
        sip.add_observation(observation=sip_subtask, return_xml=False)
    else:
        sip.add_pipelinerun(pipeline=sip_subtask, return_xml=False)
        for input in subtask.inputs.all():
            if input.producer.subtask.is_obsolete:
                logger.debug("Skipping addition of dataproducts to sip which belong to subtask input id=%s, because their producing subtask %s is obsolete." % (input.id, input.producer.subtask.id))
                continue
            logger.debug("Adding %d related input dataproducts for subtask id=%s from input id=%s and producer subtask id=%s to sip." % (input.dataproducts.count(), subtask.id, input.id, input.producer.subtask.id))
            for dataproduct in input.dataproducts.all():
                logger.debug("Adding dataproduct id %s global_id %s to sip." % (dataproduct.id, dataproduct.global_identifier.unique_identifier))
                sip_dataproduct = create_sip_representation_for_dataproduct(dataproduct)
                sip.add_related_dataproduct(sip_dataproduct, return_xml=False)
    sip.add_parset(parset_sip_identifier, contents=str(convert_to_parset(subtask)))

@lru_cache(maxsize=10, typed=False)
def create_sip_representation_for_subtask(subtask: Subtask, parset_sip_identifier=None):
    """
    Extract info from the TMSS subtask and create a sip object representation corresponding to the type of the subtask
    :param subtask:
    :return: A siplib.Observation object or one of the various siplib pipeline object flavors
    """
    # determine common properties
    subtask_sip_identifier = get_siplib_identifier(subtask.global_identifier, "Subtask id=%s" % subtask.id, name=subtask.task_blueprint.short_description or None)
    name = str(subtask.id)

    # We have to create an identifier object that is used to disclose the former SAS ID in the LTA (which corresponds
    # to the related primary observation subtask in TMSS).
    # Note that while this is the same SIP structure as used for cross-referencing LTA entities, we do not reference
    # the primary subtask by its global identifier here, but for some reason the SIP expects an identifier that does not
    # point to anything. It merely carries semantics (aka 'control ID'). In TMSS that just happens to coincide with the
    # subtask id, but we do not actually want to refer to the observation itself here.
    observationId_identifier = _create_siplib_identifier_object('TMSS', subtask.id, name=None, suppress_warning=True)
    scheduling_unit_blueprint = subtask.task_blueprint.scheduling_unit_blueprint
    process_map_args = {'strategyname': subtask.specifications_template.name,
                        'strategydescription': subtask.specifications_template.description,
                        'starttime': subtask.process_start_time, # note below that we use on_sky_start_time for observations
                        'duration': isodate.duration_isoformat(datetime.timedelta(seconds=round(subtask.duration.total_seconds()))), # note below that we use on_duration for observations
                        'identifier': subtask_sip_identifier,
                        'observation_identifier': observationId_identifier,
                        'relations': [],  # (GroupID is deprecated)
                        'parset_identifier': parset_sip_identifier}

    # determine subtask specific properties and add subtask representation to Sip object
    if subtask.is_observation:
        # use on_sky_start_time, on_sky_duration for observations
        process_map_args['starttime'] = subtask.on_sky_start_time
        process_map_args['duration'] = isodate.duration_isoformat(datetime.timedelta(seconds=round(subtask.on_sky_duration.total_seconds())))

        subarraypointings = []
        beamnumber = 0
        for sap in subtask.SAPs.order_by('id').all():
            coherentstokesprocessing = None
            incoherentstokesprocessing = None
            flyseyeprocessing = None
            correlatorprocessing = None
            clock = 200000000
            if 'COBALT' in subtask.specifications_doc and 'beamformer' in subtask.specifications_doc['COBALT']: 
                if 'tab_pipelines' in subtask.specifications_doc['COBALT']['beamformer'] and len(subtask.specifications_doc['COBALT']['beamformer']['tab_pipelines']) > 0:
                    # only add the processing if it is applicable to the SAP
                    # unfortunately, we have to do the partial name matching here because the SAP objects get a name
                    # prefix assigned (encoding subtask ID and sequence number) and we cannot change that without
                    # difficult to resolve inconsistencies in the database. The partial matching is not ideal but pragmatic.
                    for beamformer_sap in subtask.specifications_doc['COBALT']['beamformer']['tab_pipelines'][0]['SAPs']:
                        if sap.specifications_doc['name'].endswith(beamformer_sap['name']):
                            # determine the correct set of stations by intersecting the stations actually used in the
                            # observation and the ones that the beamformer has been specified for.
                            stations = used_pipeline_stations(subtask.specifications_doc['COBALT']['beamformer']['tab_pipelines'][0], subtask.specifications_doc)
                            stations = get_siplib_stations_list(stations, subtask.specifications_doc['stations']['antenna_set'])
                            if 'coherent' in subtask.specifications_doc['COBALT']['beamformer']['tab_pipelines'][0] and any([tab['coherent'] for tab in beamformer_sap['tabs']]):
                                # todo: how to deal with more than one pipeline? For now, always use ['tab_pipelines'][0] and ['flyseye_pipelines'][0]
                                channelspersubband = subtask.specifications_doc['COBALT']['beamformer']['tab_pipelines'][0]['coherent']['channels_per_subband']
                                rawsamplingtime =  1/clock * 1024 * channelspersubband
                                timesamplingdownfactor = subtask.specifications_doc['COBALT']['beamformer']['tab_pipelines'][0]['coherent']['time_integration_factor']
                                coherentstokesprocessing = siplib.CoherentStokesProcessing(rawsamplingtime=rawsamplingtime,
                                                                                           rawsamplingtime_unit = 's',
                                                                                           timesamplingdownfactor=timesamplingdownfactor,
                                                                                           samplingtime= rawsamplingtime * timesamplingdownfactor,
                                                                                           samplingtime_unit = 's',
                                                                                           stokes = mapping_beamformer_stokes_type_TMSS_2_SIP[subtask.specifications_doc['COBALT']['beamformer']['tab_pipelines'][0]['coherent']['stokes']],
                                                                                           numberofstations = len(stations),
                                                                                           stations = stations,
                                                                                           frequencydownsamplingfactor=None,  # no longer used
                                                                                           numberofcollapsedchannels=None,   # no longer used
                                                                                           channelwidth_frequency = clock / 1024 / channelspersubband,
                                                                                           channelwidth_frequencyunit='Hz',
                                                                                           channelspersubband=channelspersubband,
                                                                                           processingtype="Coherent Stokes")

                            if 'incoherent' in subtask.specifications_doc['COBALT']['beamformer']['tab_pipelines'][0] and not all([tab['coherent'] for tab in beamformer_sap['tabs']]):
                                channelspersubband = subtask.specifications_doc['COBALT']['beamformer']['tab_pipelines'][0]['incoherent']['channels_per_subband']
                                rawsamplingtime =  1/clock * 1024 * channelspersubband
                                timesamplingdownfactor = subtask.specifications_doc['COBALT']['beamformer']['tab_pipelines'][0]['incoherent']['time_integration_factor']
                                incoherentstokesprocessing = siplib.IncoherentStokesProcessing(rawsamplingtime = rawsamplingtime,
                                                                                               rawsamplingtime_unit = 's',
                                                                                               timesamplingdownfactor = timesamplingdownfactor,
                                                                                               samplingtime = rawsamplingtime * timesamplingdownfactor,
                                                                                               samplingtime_unit = 's',
                                                                                               stokes = mapping_beamformer_stokes_type_TMSS_2_SIP[subtask.specifications_doc['COBALT']['beamformer']['tab_pipelines'][0]['incoherent']['stokes']],
                                                                                               numberofstations = len(stations),
                                                                                               stations = stations,
                                                                                               frequencydownsamplingfactor=None,   # no longer used
                                                                                               numberofcollapsedchannels=None,   # no longer used
                                                                                               channelwidth_frequency= clock / 1024 / channelspersubband,
                                                                                               channelwidth_frequencyunit='Hz',
                                                                                               channelspersubband=channelspersubband,
                                                                                               processingtype="Incoherent Stokes")
                            else:
                                incoherentstokesprocessing = None

                if 'flyseye_pipelines' in  subtask.specifications_doc['COBALT']['beamformer'] and len(subtask.specifications_doc['COBALT']['beamformer']['flyseye_pipelines']) > 0:
                    channelspersubband = subtask.specifications_doc['COBALT']['beamformer']['flyseye_pipelines'][0]['coherent']['channels_per_subband']
                    rawsamplingtime =  1/clock * 1024 * channelspersubband
                    timesamplingdownfactor = subtask.specifications_doc['COBALT']['beamformer']['flyseye_pipelines'][0]['coherent']['time_integration_factor']
                    flyseyeprocessing = siplib.FlysEyeProcessing(rawsamplingtime = rawsamplingtime,
                                                                 rawsamplingtime_unit = 's',
                                                                 timesamplingdownfactor = timesamplingdownfactor,
                                                                 samplingtime = rawsamplingtime * timesamplingdownfactor,
                                                                 samplingtime_unit = 's',
                                                                 stokes = mapping_beamformer_stokes_type_TMSS_2_SIP[subtask.specifications_doc['COBALT']['beamformer']['flyseye_pipelines'][0]['coherent']['stokes']],
                                                                 channelwidth_frequency= clock / 1024 / channelspersubband,
                                                                 channelwidth_frequencyunit='Hz',
                                                                 channelspersubband=channelspersubband,
                                                                 processingtype="Fly's Eye")
            if 'COBALT' in subtask.specifications_doc and 'correlator' in subtask.specifications_doc['COBALT'] and subtask.specifications_doc['COBALT']['correlator']['enabled']:
                integrationinterval = subtask.specifications_doc['COBALT']['blocksize'] * 1024 / clock / subtask.specifications_doc['COBALT']['correlator']['integrations_per_block'] * subtask.specifications_doc['COBALT']['correlator']['blocks_per_integration']
                channelwidth_frequency = clock / subtask.specifications_doc['COBALT']['correlator']['channels_per_subband']
                correlatorprocessing = siplib.CorrelatorProcessing(integrationinterval=integrationinterval,
                                                                   integrationinterval_unit='s',
                                                                   channelwidth_frequency=channelwidth_frequency,
                                                                   channelwidth_frequencyunit='Hz',
                                                                   channelspersubband=subtask.specifications_doc['COBALT']['correlator']['channels_per_subband'])

            measurement_type_map = {"target": "Target", "calibrator": "Calibration", "all sky": "All Sky"}

            sip_sap = siplib.SubArrayPointing(pointing=siplib.PointingRaDec(ra_angle=sap.specifications_doc['pointing']['angle1'],
                                                                            ra_angleunit="radians",
                                                                            dec_angle=sap.specifications_doc['pointing']['angle2'],
                                                                            dec_angleunit="radians",
                                                                            equinox=sap.specifications_doc['pointing']['direction_type']),
                                              beamnumber=beamnumber,
                                              identifier=get_siplib_identifier(sap.global_identifier, "SAP id=%s" % sap.id),
                                              measurementtype=measurement_type_map.get(sap.specifications_doc.get('purpose'), 'Miscellaneous'),
                                              targetname=sap.specifications_doc["pointing"]['target'],
                                              starttime=sap.specifications_doc['time']['start_time'],
                                              duration=isodate.duration_isoformat(datetime.timedelta(seconds=round(sap.specifications_doc['time']['duration']))),
                                              numberofprocessing=len([p for p in (correlatorprocessing, coherentstokesprocessing, incoherentstokesprocessing, flyseyeprocessing) if p is not None]),
                                              numberofcorrelateddataproducts=sap.dataproducts.filter(dataformat=Dataformat.objects.get(value=Dataformat.Choices.MEASUREMENTSET.value), producer__id__in=[subtask_output.id for subtask_output in SubtaskOutput.objects.filter(subtask__id=subtask.id)]).count(),
                                              numberofbeamformeddataproducts=sap.dataproducts.filter(dataformat=Dataformat.objects.get(value=Dataformat.Choices.BEAMFORMED.value), producer__id__in=[subtask_output.id for subtask_output in SubtaskOutput.objects.filter(subtask__id=subtask.id)]).count(),
                                              relations=[],  # (GroupID is deprecated)
                                              correlatorprocessing=correlatorprocessing,
                                              coherentstokesprocessing=coherentstokesprocessing,
                                              incoherentstokesprocessing=incoherentstokesprocessing,
                                              flyseyeprocessing=flyseyeprocessing,
                                              nonstandardprocessing=None, # only used for things that are not modeled. Don't think w ehave this in TMSS?
                                              measurementdescription=sap.specifications_doc['name']
                                              )
            subarraypointings.append(sip_sap)
            beamnumber += 1

        stations = get_siplib_stations_list(subtask.specifications_doc['stations']['station_list'],
                                            subtask.specifications_doc['stations']['antenna_set'])

        observation = siplib.Observation(observingmode=constants.OBSERVINGMODETYPE_BEAM_OBSERVATION,  # can be hardcoded for an observation
                                         instrumentfilter=mapping_filterset_type_TMSS_2_SIP[subtask.specifications_doc['stations']['filter']],
                                         clock_frequency="200",  # fixed,
                                         clock_frequencyunit=constants.FREQUENCYUNIT_MHZ,  # fixed
                                         stationselection=constants.STATIONSELECTIONTYPE_CUSTOM,  # fixed
                                         antennaset=mapping_antennaset_type_TMSS_2_SIP[subtask.specifications_doc['stations']['antenna_set']],
                                         timesystem="UTC",  # fixed
                                         numberofstations=len(subtask.specifications_doc['stations']['station_list']),
                                         stations=stations,
                                         numberofsubarraypointings=len(subtask.specifications_doc['stations']['digital_pointings']),
                                         numberoftbbevents=0,  # fixed
                                         numberofcorrelateddataproducts=get_number_of_dataproducts_of_format(subtask, Dataformat.Choices.MEASUREMENTSET.value),
                                         numberofbeamformeddataproducts=get_number_of_dataproducts_of_format(subtask, Dataformat.Choices.BEAMFORMED.value),
                                         numberofbitspersample=8,  # fixed
                                         process_map=siplib.ProcessMap(**process_map_args),
                                         channelwidth_frequency=None, # NA any more ('BlueGene compatibility' see comment in LTA-SIP.xsd)
                                         channelwidth_frequencyunit=constants.FREQUENCYUNIT_HZ,  # fixed
                                         observationdescription=subtask.task_blueprint.description,
                                         channelspersubband=None,  # NA any more ('BlueGene compatibility' see comment in LTA-SIP.xsd)
                                         subarraypointings=subarraypointings,
                                         transientbufferboardevents=None  # fixed
        )

        return observation
    elif subtask.specifications_template.type.value == SubtaskType.Choices.PIPELINE.value:
        sourcedata_identifiers = []
        for input in subtask.inputs.all():
            if input.producer.subtask.is_obsolete:
                logger.debug("Skipping addition of sourcedata identifiers to sip which belong to subtask input id=%s, because their producing subtask %s is obsolete." % (input.id, input.producer.subtask.id))
                continue
            sourcedata_identifiers += [get_siplib_identifier(dp.global_identifier, "Dataproduct id=%s" % dp.id) for dp in input.dataproducts.all()]
        if not sourcedata_identifiers:
            raise TMSSException("There seems to be no subtask input associated to your pipeline subtask id %s. Please define what data the pipeline processed." % subtask.id)

        # since we have a pipeline, change the observationId reference to point to the pipeline itself
        if subtask is None:
            raise TMSSException('Could not identify primary predecessor observation subtask')
        else:
            process_map_args.update({
                'duration': isodate.duration_isoformat(datetime.timedelta(seconds=round((subtask.process_stop_time - subtask.process_start_time).total_seconds()))),
                'observation_identifier': _create_siplib_identifier_object('TMSS', subtask.id, name=None, suppress_warning=True)
            })

        software_version = subtask.specifications_doc.get('software_version', 'unknown')  # todo: check that this is what we want or this should be taken from feedback --> TMSS-1263
        if ':latest' in software_version or ':tmss' in software_version:
            software_version = 'unknown'  # todo: can we determine somehow what version is in these containers?
        pipeline_map = siplib.PipelineMap(
                name=subtask.task_blueprint.short_description,
                version=software_version,
                sourcedata_identifiers=sourcedata_identifiers,
                process_map=siplib.ProcessMap(**process_map_args))

        if subtask.specifications_template.name == "preprocessing pipeline":  #  todo: re-evaluate this because schema name might change --> TMSS-1264
            spec = subtask.specifications_doc # extract settings from specifications_doc. This should be complete and immutable once run.
            pipeline = siplib.AveragingPipeline(  # <-- this is what we need for UC1
                pipeline_map,
                numberofcorrelateddataproducts=get_number_of_dataproducts_of_format(subtask, Dataformat.Choices.MEASUREMENTSET.value),
                frequencyintegrationstep=spec['demixer']['frequency_steps'] if spec['demixer']['enabled'] else 1,
                timeintegrationstep=spec['demixer']['time_steps'] if spec['demixer']['enabled'] else 1,
                flagautocorrelations=spec['preflagger1']['enabled'] and spec['preflagger1']['corrtype'] == 'auto',
                demixing=spec['demixer']['enabled'] and (len(spec['demixer']['demix_always']) > 0 or len(spec['demixer']['demix_if_needed']) > 0)
            )
        elif subtask.specifications_template.name == "pulsar pipeline":
            spec = subtask.specifications_doc # extract settings from specifications_doc. This should be complete and immutable once run.
            pipeline = siplib.PulsarPipeline(pipeline_map,
                pulsarselection=mapping_pulsar_selection_type_TMSS_2_SIP[subtask.task_blueprint.specifications_doc['pulsar']['strategy']],
                pulsars=[spec['pulsar']] if spec['pulsar'] not in mapping_pulsar_selection_type_TMSS_2_SIP.keys() else [],
                dosinglepulseanalysis=spec['single_pulse'],
                convertRawTo8bit=spec['output']['raw_to_8bit'],
                subintegrationlength=spec['dspsr']['tsubint'],
                subintegrationlength_unit='s',
                skiprfiexcision=spec['dspsr']['norfi'],
                skipdatafolding=spec['presto']['nofold'],
                skipoptimizepulsarprofile=spec['dspsr']['nopdmp'],
                skipconvertrawintofoldedpsrfits=spec['dspsr']['skip_dspsr'],
                runrotationalradiotransientsanalysis=spec['presto']['rrats'],
                skipdynamicspectrum=spec['output']['skip_dynamic_spectrum'],
                skipprefold=spec['presto']['skip_prepfold'])
        # todo: distinguish and create other pipeline types. Probably most of these can be filled in over time as needed,
        #  but they are not required for UC1. Here are stubs to start from for the other types the LTA supports:
        # elif subtask.specifications_template.name == "<???>": # todo
        #     pipeline = siplib.ImagingPipeline(pipeline_map,
        #                                imagerintegrationtime=10,  # todo
        #                                imagerintegrationtime_unit="ms",  #todo
        #                                numberofmajorcycles=5,  #todo
        #                                numberofinstrumentmodels=5,  #todo
        #                                numberofcorrelateddataproducts=1,  #todo
        #                                numberofskyimages=1, #todo
        #                                frequencyintegrationstep=None, #todo
        #                                timeintegrationstep=None, #todo
        #                                skymodeldatabase=None,  #todo
        #                                demixing=None)  #todo
        # elif subtask.specifications_template.name == "<???>":  # todo
        #     pipeline = siplib.CalibrationPipeline(
        #     pipeline_map,
        #     skymodeldatabase = "db", #todo
        #     numberofinstrumentmodels = 1,  #todo
        #     numberofcorrelateddataproducts = 1,  #todo
        #     frequencyintegrationstep = 1,  #todo
        #     timeintegrationstep = 1,  #todo
        #     flagautocorrelations = True,  #todo
        #     demixing = False)  #todo
        else:
            logger.warning('Could not identify LTA pipeline flavor for subtask=%s, template name=%s). Adding as SimplePipeline to SIP.'
                           % (subtask, subtask.specifications_template.name))
            pipeline = siplib.SimplePipeline(pipeline_map)
        return pipeline
    else:
        # todo: also cover, if needed: COPY, INSPECTION, QA_FILES, QA_PLOTS, DELETION, MANUAL, OTHER
        #  could be handled as UnspecifiedProcess, or require modeling in the SIP as well?
        raise TMSSException("The Subtask type %s cannot be represented in a SIP" % subtask.specifications_template.type.value)


@lru_cache(maxsize=10*488, typed=False)
def create_sip_representation_for_dataproduct(dataproduct: Dataproduct):
    """
    Extract info from the TMSS dataproduct and creates a SIP object representation
    :param dataproduct:
    :return: One of the siplib dataproduct object flavors.
    """

    # Make sure feedback is not empty
    if 'percentage_written' not in dataproduct.feedback_doc:
        raise ValueError("The feedback_doc of dataproduct id=%s is incomplete. Has feedback of the producing subtask been processed?" % dataproduct.id)

    # Note: this is for the type property present on all dataproduct flavors, dataproduct classes are
    #  differentiated in addition to that below
    type_map = {Datatype.Choices.VISIBILITIES.value: constants.DATAPRODUCTTYPE_CORRELATOR_DATA,
                Datatype.Choices.TIME_SERIES.value: constants.DATAPRODUCTTYPE_BEAM_FORMED_DATA,
                Datatype.Choices.INSTRUMENT_MODEL.value : constants.DATAPRODUCTTYPE_INSTRUMENT_MODEL,
                Datatype.Choices.IMAGE.value : constants.DATAPRODUCTTYPE_SKY_IMAGE,
                Datatype.Choices.QUALITY.value : constants.DATAPRODUCTTYPE_UNKNOWN,
                # Datatype.Choices.???.value : constants.DATAPRODUCTTYPE_PULSAR_PIPELINE_SUMMARY_OUTPUT,  # todo: we have an extra type for the summary in the XSD, but not in TMSS. Check if we need to add that (example SIPs apparently use the non-summary type for both flavors.)
                Datatype.Choices.PULSAR_PROFILE.value : constants.DATAPRODUCTTYPE_PULSAR_PIPELINE_OUTPUT}

    # Note: this is for the fileformat property present on all dataproduct flavors,
    #  dataproduct classes are differentiated in addition to that below
    fileformat_map = {Dataformat.Choices.MEASUREMENTSET.value: constants.FILEFORMATTYPE_AIPS___CASA,
                      Dataformat.Choices.BEAMFORMED.value: constants.FILEFORMATTYPE_HDF5,
                      Dataformat.Choices.QA_HDF5.value: constants.FILEFORMATTYPE_HDF5,
                      Dataformat.Choices.QA_PLOTS.value: constants.FILEFORMATTYPE_UNDOCUMENTED,
                      Dataformat.Choices.PULP_SUMMARY.value: constants.FILEFORMATTYPE_PULP,
                      Dataformat.Choices.PULP_ANALYSIS.value: constants.FILEFORMATTYPE_PULP}

    storage_writer_map = {"dysco": constants.STORAGEWRITERTYPE_DYSCOSTORAGEMANAGER,
                          "unknown": constants.STORAGEWRITERTYPE_UNKNOWN,
                          "standard": constants.STORAGEWRITERTYPE_LOFARSTORAGEMANAGER,
                          "lofarstman": constants.STORAGEWRITERTYPE_LOFARSTORAGEMANAGER}

    try:
        dataproduct_type = type_map[dataproduct.datatype.value]
    except Exception as err:
        dataproduct_type = constants.DATAPRODUCTTYPE_UNKNOWN
        logger.warning("Could not determine the dataproduct type (datatype) of dataproduct id %s (%s). Falling back to %s" % (dataproduct.id, err, dataproduct_type))

    try:
        dataproduct_fileformat = fileformat_map[dataproduct.dataformat.value]  # todo same as with type? Why is this not with the data? Why is this so different from the LTA datamodel?
    except Exception as err:
        dataproduct_fileformat = constants.FILEFORMATTYPE_UNDOCUMENTED
        logger.warning("Could not determine the fileformat type (dataformat) of dataproduct id %s (%s). Falling back to %s" % (dataproduct.id, err, dataproduct_fileformat))

    if "pulp" in dataproduct.dataformat.value:
        storage_writer = constants.STORAGEWRITERTYPE_UNKNOWN
    elif 'samples' in dataproduct.feedback_doc and dataproduct.feedback_doc["samples"]["writer"] == "standard" and dataproduct.dataformat.value == Dataformat.Choices.BEAMFORMED.value and dataproduct.producer.subtask.is_observation:
        storage_writer = constants.STORAGEWRITERTYPE_HDF5DEFAULT
    else:
        storage_writer = storage_writer_map[dataproduct.feedback_doc["samples"]["writer"] if 'samples' in dataproduct.feedback_doc else 'unknown'] # todo: It works for the preprocessing pipelines. Verify also for other types that we can use the feedback_doc here and remove the old method | storage_writer_map[dataproduct.producer.subtask.task_blueprint.specifications_doc.get("storagemanager", 'unknown')],

    dataproduct_map = siplib.DataProductMap(type=dataproduct_type,
                                            identifier=get_siplib_identifier(dataproduct.global_identifier, "Dataproduct %s" % dataproduct.id),
                                            size=dataproduct.size if dataproduct.size else 0,
                                            filename=dataproduct.filename,
                                            fileformat=dataproduct_fileformat,
                                            storage_writer=storage_writer,
                                            storage_writer_version=dataproduct.feedback_doc["samples"]["writer_version"] if 'samples' in dataproduct.feedback_doc else 'unknown',
                                            process_identifier=get_siplib_identifier(dataproduct.producer.subtask.global_identifier, "Producer Subtask %s" % dataproduct.producer.subtask.id))

    # CorrelatedDataProduct
    if dataproduct.dataformat.value == Dataformat.Choices.MEASUREMENTSET.value:
        sip_dataproduct = siplib.CorrelatedDataProduct(
            dataproduct_map,
            subarraypointing_identifier=get_siplib_identifier(dataproduct.sap.global_identifier, "SAP %s" % dataproduct.sap.id),
            starttime=dataproduct.feedback_doc['time']['start_time'],
            duration=isodate.duration_isoformat(datetime.timedelta(seconds=round(dataproduct.feedback_doc['time']['duration']))),
            integrationinterval=dataproduct.feedback_doc['time']['sample_width'],
            integrationintervalunit="s",
            central_frequency=dataproduct.feedback_doc['frequency']['central_frequencies'][0],
            central_frequencyunit="Hz",
            channelwidth_frequency=dataproduct.feedback_doc['frequency']['channel_width'],
            channelwidth_frequencyunit="Hz",
            channelspersubband=dataproduct.feedback_doc['frequency']['channels_per_subband'],
            # the SIP concepts stationsubband and subband are swapped compared to the TMSS concepts beamlet_indices and subband
            # this is mainly due to historical use and lack of a proper definition.

            # TMSS definition of :
            # 'subbands':        the frequency index from the station PPF, thus an integer in the [0, 511] range
            # 'beamlet_indices': the measurementset number from the correlator, thus an integer in the [0, 487] range 

            # LTA definition of :
            # 'stationsubband':  the frequency index from the station PPF, thus an integer in the [0, 511] range
            # 'subband':         the measurementset number from the correlator, thus an integer in the [0, 487] range 

            # Here we map the TMSS-'subbands' definition on the LTA-'stationsubband', 
            # and the TMSS-'beamlet_indices' onto the LTA-'subband'.
            # This may look "strange" like we swapped them, but it entirely depends on the definition 
            # which is indeed swapped between TMSS and the LTA. From their own respective viewpoints, both are correct.
            stationsubband=dataproduct.feedback_doc['frequency']['subbands'][0],
            subband=dataproduct.feedback_doc['frequency'].get('beamlet_indices', [None])[0]
        )

    # BeamFormedDataProduct
    elif dataproduct.dataformat.value == Dataformat.Choices.BEAMFORMED.value:
        beams = []
        beam_map = siplib.ArrayBeamMap(
            subarraypointing_identifier=get_siplib_identifier(dataproduct.sap.global_identifier, "SAP %s" % dataproduct.sap.id),
            beamnumber=dataproduct.specifications_doc.get('identifiers', {}).get('tab_index', 0), # todo: verify
            dispersionmeasure=0,  # fixed
            numberofsubbands=len(dataproduct.feedback_doc['frequency']['subbands']),
            stationsubbands=dataproduct.feedback_doc['frequency']['subbands'],
            samplingtime=dataproduct.feedback_doc['time']['sample_width'],
            samplingtimeunit="s",
            centralfrequencies=dataproduct.feedback_doc['frequency']['central_frequencies'],
            centralfrequencies_unit="Hz",
            channelwidth_frequency=dataproduct.feedback_doc['frequency']['channel_width'],
            channelwidth_frequencyunit="Hz",
            channelspersubband=dataproduct.feedback_doc['frequency']['channels_per_subband'],
            stokes=dataproduct.feedback_doc['samples']['polarisations'])

        # FlysEyeBeam
        if len(dataproduct.feedback_doc['antennas']['fields']) == 1:
            field = dataproduct.feedback_doc['antennas']['fields'][0]
            beams.append(
                siplib.FlysEyeBeam(
                    arraybeam_map=beam_map,
                    station=siplib.Station.preconfigured(field['station'], [field['field']]))
            )

        # IncoherentStokesBeam
        elif dataproduct.feedback_doc.get('target', {}).get('coherent', None) is False:
            beams.append(
                siplib.IncoherentStokesBeam(arraybeam_map=beam_map)
            )

        # CoherentStokesBeam
        else:
            pointing = dataproduct.feedback_doc['target']['pointing']
            sap_pointing = dataproduct.sap.specifications_doc['pointing']
            beams.append(
                siplib.CoherentStokesBeam(
                    arraybeam_map=beam_map,
                    pointing=siplib.PointingRaDec(ra_angle=pointing['angle1'], ra_angleunit=siplib.constants.ANGLEUNIT_RADIANS,
                                                  dec_angle=pointing['angle2'], dec_angleunit=siplib.constants.ANGLEUNIT_RADIANS,
                                                  equinox=pointing['direction_type'])._get_pyxb_pointing(suppress_warning=True),
                    offset=siplib.PointingRaDec(ra_angle=(pointing['angle1'] - sap_pointing['angle1']),
                                                ra_angleunit=siplib.constants.ANGLEUNIT_RADIANS,
                                                dec_angle=(pointing['angle2'] - sap_pointing['angle2']),
                                                dec_angleunit=siplib.constants.ANGLEUNIT_RADIANS,
                                                equinox=pointing['direction_type'])._get_pyxb_pointing(suppress_warning=True),
                )
            )

        sip_dataproduct = siplib.BeamFormedDataProduct(
            dataproduct_map,
            beams
        )

    elif dataproduct.dataformat.value == Dataformat.Choices.PULP_SUMMARY.value:
        sip_dataproduct = siplib.PulpSummaryDataProduct(
            dataproduct_map,
            filecontent=dataproduct.feedback_doc['files'],
            datatype="SummaryCoherentStokes" if dataproduct.feedback_doc['target']['coherent'] else "SummaryIncoherentStokes"
        )

    elif dataproduct.dataformat.value == Dataformat.Choices.PULP_ANALYSIS.value:
        datatype = "IncoherentStokes"
        if dataproduct.feedback_doc['target']['coherent']:
            if dataproduct.feedback_doc['samples']['complex']:
                datatype = "ComplexVoltages"
            else:
                datatype = "CoherentStokes"

        beam_map = siplib.ArrayBeamMap(
            subarraypointing_identifier=get_siplib_identifier(dataproduct.sap.global_identifier, "SAP %s" % dataproduct.sap.id),
            beamnumber=dataproduct.specifications_doc.get('identifiers', {}).get('tab_index', 0),  # todo: verify
            dispersionmeasure=0,  # fixed
            numberofsubbands=len(dataproduct.feedback_doc['frequency']['subbands']),
            stationsubbands=dataproduct.feedback_doc['frequency']['subbands'],
            samplingtime=dataproduct.feedback_doc['time']['sample_width'],
            samplingtimeunit="s",
            centralfrequencies=dataproduct.feedback_doc['frequency']['central_frequencies'],
            centralfrequencies_unit="Hz",
            channelwidth_frequency=dataproduct.feedback_doc['frequency']['channel_width'],
            channelwidth_frequencyunit="Hz",
            channelspersubband=dataproduct.feedback_doc['frequency']['channels_per_subband'],
            stokes=dataproduct.feedback_doc['samples']['polarisations']
        )

        # FlysEyeBeam
        if len(dataproduct.feedback_doc['antennas']['fields']) == 1:
            field = dataproduct.feedback_doc['antennas']['fields'][0]  # todo: can we have more than one here?
            beam = siplib.FlysEyeBeam(
                beam_map,
                station=siplib.Station.preconfigured(field['station'], [field['field']])
            )

        # IncoherentStokesBeam
        elif dataproduct.feedback_doc.get('target', {}).get('coherent', None) is False:
            beam = siplib.IncoherentStokesBeam(arraybeam_map=beam_map)

        # CoherentStokesBeam
        else:
            pointing = dataproduct.feedback_doc['target']['pointing']
            sap_pointing = dataproduct.sap.specifications_doc['pointing']
            beam = siplib.CoherentStokesBeam(
                    arraybeam_map=beam_map,
                    pointing=siplib.PointingRaDec(ra_angle=pointing['angle1'],
                                                  ra_angleunit=siplib.constants.ANGLEUNIT_RADIANS,
                                                  dec_angle=pointing['angle2'],
                                                  dec_angleunit=siplib.constants.ANGLEUNIT_RADIANS,
                                                  equinox=pointing['direction_type'])._get_pyxb_pointing(suppress_warning=True),
                    offset=siplib.PointingRaDec(ra_angle=(pointing['angle1'] - sap_pointing['angle1']),
                                                ra_angleunit=siplib.constants.ANGLEUNIT_RADIANS,
                                                dec_angle=(pointing['angle2'] - sap_pointing['angle2']),
                                                dec_angleunit=siplib.constants.ANGLEUNIT_RADIANS,
                                                equinox=pointing['direction_type'])._get_pyxb_pointing(suppress_warning=True),
                )

        sip_dataproduct = siplib.PulpDataProduct(
            dataproduct_map,
            filecontent=dataproduct.feedback_doc['files'],
            datatype=datatype,
            arraybeam=beam
        )

    # todo: distinguish and create other dataproduct types. Probably most of these can be filled in over time as needed,
    #  but they are not required for UC1. Here are stubs to start from for the other types the LTA supports:
    # elif dataproduct.dataformat.value == Dataformat.Choices.<???>.value: # todo
    #     sip_dataproduct = siplib.SkyModelDataProduct(dataproduct_map)
    # elif dataproduct.dataformat.value == Dataformat.Choices. <??? >.value:  # todo
    #     sip_dataproduct = siplip.GenericDataProduct(dataproduct_map)
    # elif dataproduct.dataformat.value == Dataformat.Choices. <??? >.value:  # todo
    #     sip_dataproduct = siplip.UnspecifiedDataProduct(dataproduct_map)
    # elif dataproduct.dataformat.value == Dataformat.Choices. <??? >.value:  # todo
    #     sip_dataproduct = siplip.PixelMapDataProduct(dataproduct_map,
    #                                                  numberofaxes=5, # todo
    #                                                  coordinates=[siplib.SpectralCoordinate(  # todo
    #                                                      quantity_type="Frequency",  # todo
    #                                                      quantity_value=20.0,  #todo
    #                                                      axis=siplib.LinearAxis(  #todo
    #                                                          number=5,  #todo
    #                                                          name="bla",  #todo
    #                                                          units="parsec",  #todo
    #                                                          length=5,  #todo
    #                                                          increment=5,  #todo
    #                                                          referencepixel=7.5,  #todo
    #                                                          referencevalue=7.4))])  #todo
    # elif dataproduct.dataformat.value == Dataformat.Choices. <??? >.value:  # todo
    #     sip_dataproduct = siplib.SkyImageDataProduct(
    #         dataproduct_map,  # todo
    #         numberofaxes=2,  # todo
    #         coordinates=[  # todo
    #             siplib.SpectralCoordinate(  # todo
    #                 quantity_type=constants.SPECTRALQUANTITYTYPE_VELOCITYAPPRADIAL,  # todo
    #                 quantity_value=20.0,  # todo
    #                 axis=siplib.LinearAxis(   # todo
    #                     number=5,  # todo
    #                     name="bla",  # todo
    #                     units="unit",  # todo
    #                     length=5,  # todo
    #                     increment=5,  # todo
    #                     referencepixel=7.5,  # todo
    #                     referencevalue=7.4)),  # todo
    #             siplib.DirectionCoordinate(  # todo
    #                 linearaxis_a=siplib.LinearAxis(  # todo
    #                     number=5,  # todo
    #                     name="bla",  # todo
    #                     units="unit",  # todo
    #                     length=5,  # todo
    #                     increment=5,  # todo
    #                     referencepixel=7.5,  # todo
    #                     referencevalue=7.4),  # todo
    #                 linearaxis_b=siplib.LinearAxis(  # todo
    #                     number=5,  # todo
    #                     name="blb",  # todo
    #                     units="unit",  # todo
    #                     length=5,  # todo
    #                     increment=5,  # todo
    #                     referencepixel=7.5,  # todo
    #                     referencevalue=7.4),  # todo
    #                 pc0_0=0.0,  # todo
    #                 pc0_1=0.1,  # todo
    #                 pc1_0=1.0,  # todo
    #                 pc1_1=1.1,  # todo
    #                 equinox=constants.EQUINOXTYPE_JUPITER,  # todo
    #                 radecsystem=constants.RADECSYSTEM_FK4_NO_E,  # todo
    #                 projection="rear",  # todo
    #                 projectionparameters=[1.0, 1.0, 1.0],  # todo
    #                 longitudepole_angle=1.0,  # todo
    #                 longitudepole_angleunit=constants.ANGLEUNIT_RADIANS,  # todo
    #                 latitudepole_angle=2.0,  # todo
    #                 latitudepole_angleunit=constants.ANGLEUNIT_ARCSEC,  # todo
    #             ),
    #             siplib.PolarizationCoordinate(  # todo
    #                 tabularaxis=siplib.TabularAxis(  # todo
    #                     number=5,  # todo
    #                     name="bla",  # todo
    #                     units="someunit",  # todo
    #                     length=5,  # todo
    #                 ),
    #                 polarizations=[constants.POLARIZATIONTYPE_LR, constants.POLARIZATIONTYPE_XRE]  # todo
    #             ),
    #         ],
    #         locationframe=constants.LOCATIONFRAME_LOCAL_GROUP,  # todo
    #         timeframe="timeframe",  # todo
    #         observationpointing=siplib.PointingRaDec(  # todo
    #             ra_angle=1.0,  # todo
    #             ra_angleunit=constants.ANGLEUNIT_DEGREES,  # todo
    #             dec_angle=42.0,  # todo
    #             dec_angleunit=constants.ANGLEUNIT_DEGREES,  # todo
    #             equinox=constants.EQUINOXTYPE_B1950  # todo
    #         ),
    #         restoringbeammajor_angle=1.0,  # todo
    #         restoringbeammajor_angleunit=constants.ANGLEUNIT_DEGREES,  # todo
    #         restoringbeamminor_angle=2.0,  # todo
    #         restoringbeamminor_angleunit=constants.ANGLEUNIT_DEGREES,  # todo
    #         rmsnoise=1.0  # todo
    #     )
    else:
        logger.warning('Could not identify corresponding SIP dataproduct flavor for dataproduct=%s, dataformat=%s. Adding as SimpleDataproduct to SIP.' % (dataproduct, dataproduct.dataformat))
        sip_dataproduct = siplib.SimpleDataProduct(dataproduct_map)
    return sip_dataproduct


def recursively_add_subbtask_and_input_dataproducts_to_sip(subtask: Subtask, sip: siplib.Sip):
    """
    add a subtask and the whole tree of its predecessor subtasks to the sip
    by traversing the subtask tree and repeatedly calling add_subbtask_to_sip().
    This will also add the input dataproducts of each subtask.
    :param subtask: the Subtask object to add to the sip
    :param sip: the Sip object to add to
    :return:
    """
    if subtask is not None and subtask.state.value==SubtaskState.Choices.FINISHED.value and not subtask.is_obsolete:
        add_subbtask_and_input_dataproducts_to_sip(subtask, sip)
        for predecessor in subtask.predecessors.all():
            recursively_add_subbtask_and_input_dataproducts_to_sip(predecessor, sip)


def generate_sip_for_dataproduct(dataproduct):
    """
    generate_sip_from_dataproduct
    :param dataproduct: The dataproduct to create the SIP for
    :return: The Sip object
    """

    # Create sip representation of the described/main dataproduct
    if dataproduct.producer.subtask.specifications_template.type.value == SubtaskType.Choices.INGEST.value:
        # the output dataproduct of an ingest subtask is a special case
        # generate the SIP for the corresponding input dataproduct
        ingest_input_dataproduct = dataproduct.producer.subtask.get_transformed_input_dataproduct(dataproduct)
        sip_dataproduct = create_sip_representation_for_dataproduct(ingest_input_dataproduct)

        # and enrich it SIP with the archive info of the output dataproduct
        if dataproduct.archive_info:
            sip_dataproduct.set_storageTicket(dataproduct.archive_info.storage_ticket)
            # enrich/overwrite the filename/size by the output dataproduct filename/size
            sip_dataproduct.set_fileName(dataproduct.filename)
            sip_dataproduct.set_size(dataproduct.size)

        if dataproduct.hashes:
            from django.core.exceptions import ObjectDoesNotExist
            try:
                sip_dataproduct.set_checksum_md5(dataproduct.hashes.get(hash_algorithm=HashAlgorithm.Choices.MD5.value).hash)
            except ObjectDoesNotExist:
                pass

            try:
                sip_dataproduct.set_checksum_adler32(dataproduct.hashes.get(hash_algorithm=HashAlgorithm.Choices.ADLER32.value).hash)
            except ObjectDoesNotExist:
                pass

        # for the rest of the sip generation, use the ingest_input_dataproduct
        dataproduct = ingest_input_dataproduct
    else:
        # normal case: a normal dataproduct generated by a normal observation/pipeline subtask
        sip_dataproduct = create_sip_representation_for_dataproduct(dataproduct)

    # Gather project details
    project_code = dataproduct.project.name
    project_primaryinvestigator = 'project_primaryinvestigator'  # left out for GDPR reason
    project_contactauthor = 'project_contactauthor'              # left out for GDPR reason
    project_description = dataproduct.project.description

    # Create the Sip object
    sip = siplib.Sip(project_code=project_code,
                     project_primaryinvestigator=project_primaryinvestigator,
                     project_contactauthor=project_contactauthor,
                     project_description=project_description,
                     dataproduct=sip_dataproduct)

    recursively_add_subbtask_and_input_dataproducts_to_sip(dataproduct.producer.subtask, sip)

    validator.check_consistency(sip)
    return sip


def get_sip_xml_for_dataproduct(dataproduct: Dataproduct) -> str:
    '''look up the sip document for the provided dataproduct (generate it first if it does not yet exist)'''
    sip = generate_sip_for_dataproduct(dataproduct)
    sip_xml = sip.get_prettyxml()

    # the sip-lib creates an xml document without a 'schemaLocation' where the schema can be found.
    # pragmatically insert the tmss production xsd url.
    # ugly string manipulation to inject the schemaLocation attribute (saves expensive XML parsing & DOM manipulation)
    sip_xml = sip_xml.replace('<sip:ltaSip ', '<sip:ltaSip xsi:schemaLocation="http://www.astron.nl/SIP-Lofar https://tmss.lofar.eu/api/xsd/LTA-SIP.xsd" ')

    return sip_xml
