import math

import numpy
from astropy.time import Time
import astropy.units
from datetime import datetime, timedelta, time as dtime
from astropy.coordinates.earth import EarthLocation
from astropy.coordinates import Angle, get_body
import astropy.time
from functools import lru_cache
from typing import Tuple
from lofar.common.datetimeutils import round_to_second_precision
from lofar.sas.tmss.tmss.tmssapp.models.calculations import StationTimeline
from lofar.sas.tmss.tmss.tmssapp.models.specification import CommonSchemaTemplate
from django.db.utils import IntegrityError

from django.core.exceptions import ObjectDoesNotExist, MultipleObjectsReturned

import logging
logger = logging.getLogger(__name__)


class Pointing():
    '''struct for pointing parameters
    '''
    def __init__(self, angle1: float, angle2: float, direction_type: str="J2000", target: str=None):
        self.angle1 = angle1
        self.angle2 = angle2
        self.direction_type = direction_type
        self.target = target
        if self.target is None:
            try:
                # try to set the target as hms,dms coord-string.
                self.target = self.as_SkyCoord().to_string('hmsdms')
            except ValueError:
                # just keep the self.target as None
                pass

    def __hash__(self):
        return hash((self.angle1, self.angle2,self.direction_type))

    def __eq__(self, other):
        # ToDo: compare floats as almost equal, or use angle gridder
        # in practice, we read angle1 and angle2 from strings,
        # and in python float('1.23')==float('1.23') always yields True
        return isinstance(other, Pointing) and \
               self.direction_type == other.direction_type and \
               self.angle1 == other.angle1 and \
               self.angle2 == other.angle2

    def __str__(self):
        '''represent the pointing in degrees, direction_type, targetname'''
        return "%s=%.3f[deg] %s=%.3f[deg]%s%s" % ('RA' if self.direction_type=='J2000' else 'angle1',
                                                  astropy.coordinates.Angle(self.angle1, unit=astropy.units.rad).deg,
                                                  'DEC' if self.direction_type == 'J2000' else 'angle2',
                                                  astropy.coordinates.Angle(self.angle2, unit=astropy.units.rad).deg,
                                                  '' if self.direction_type=='J2000' else self.direction_type,
                                                  (" '"+self.target+"'") if self.target else "")

    def str_astro(self):
        '''represent the pointing in astropy hmsdms, direction_type, targetname'''
        return "%s%s%s" % (self.as_SkyCoord().to_string('hmsdms'),
                           '' if self.direction_type=='J2000' else self.direction_type,
                           (" '"+self.target+"'") if self.target else "")

    @property
    def transit_lst(self) -> dtime:
        '''get the transit time for this pointing in Local Sidereal Time (expressed in python datetime.time object)'''
        if self.direction_type == "J2000":
            hms = astropy.coordinates.Angle(self.angle1, astropy.units.rad).hms
            py_time = dtime(hour=int(hms.h), minute=int(hms.m), second=int(hms.s))
            return py_time
        raise NotImplementedError()

    def as_SkyCoord(self, timestamp=None, location=None):
        '''
        return this pointing as astropy SkyCoord.
        The optional timestamp and observer location are relevant when pointing at celestial bodies, whose coordinates
        depend on these parameters. If no timestamp is provided, the current sky coordinates are returned.
        If no location is provided, the center of the Earth is assumed.
        '''
        if timestamp is None:
            timestamp = datetime.utcnow()
        if self.direction_type == "J2000":
            return astropy.coordinates.SkyCoord(ra=self.angle1, dec=self.angle2, unit=astropy.units.rad)
        elif self.direction_type in ['SUN', 'MOON', 'MERCURY', 'VENUS', 'MARS', 'JUPITER', 'SATURN', 'URANUS', 'NEPTUNE']:
            return astropy.coordinates.get_body(self.direction_type.lower(), Time(timestamp), location)
        raise ValueError("Do not know how to convert direction_type=%s to SkyCoord" % self.direction_type)

    @staticmethod
    def from_SkyCoord(sky_coord: astropy.coordinates.SkyCoord, target: str=None):
        return Pointing(sky_coord.ra.rad, sky_coord.dec.rad, 'J2000')

    @staticmethod
    def from_name(target_name: str):
        coord = astropy.coordinates.SkyCoord.from_name(target_name)
        return Pointing.from_SkyCoord(coord, target_name)

    def as_dict(self):
        return {'angle1': self.angle1,
                'angle2': self.angle2,
                'direction_type': self.direction_type,
                'target': self.target or ''}

@lru_cache(maxsize=60, typed=False)
def create_location_for_station(station: str) -> EarthLocation:
    '''
    returns an astropy EarthLocation for a given station, located in the LBA center of the given station
    :param station: a station name, e.g. "CS002"
    :return: astropy.coordinates.earth.EarthLocation
    '''
    from lofar.common import station_coordinates
    coords = station_coordinates.parse_station_coordinates()["%sLBA" % station.upper()]
    location = EarthLocation.from_geocentric(x=coords['x'], y=coords['y'], z=coords['z'],  unit=astropy.units.m)
    return location


@lru_cache(maxsize=60, typed=False)
def create_astroplan_observer_for_station(station: str) -> 'Observer':
    '''
    returns an astroplan observer for object for a given station, located in the LBA center of the given station
    :param station: a station name, e.g. "CS002"
    :return: astroplan.observer.Observer object
    '''
    from astroplan.observer import Observer # import here in the function to prevent this module from doing an astroplan setup-and-check-which-takes-to-long-at-startup
    location = create_location_for_station(station)
    observer = Observer(location, name=station, timezone="UTC")
    return observer


# default angle to the horizon at which the sunset/sunrise starts and ends, as per LOFAR definition.
SUN_SET_RISE_ANGLE_TO_HORIZON = Angle(10, unit=astropy.units.deg)
# default n_grid_points; higher is more precise but very costly; astropy defaults to 150, errors now can be in the minutes, increase if this is not good enough
# TODO: To be considered, now we store the sunset/sunrise data in advanced, we can increase the number of points!!
SUN_SET_RISE_PRECISION = 96

def timestamps_and_stations_to_sun_rise_and_set(timestamps: tuple, stations: tuple, angle_to_horizon: Angle=SUN_SET_RISE_ANGLE_TO_HORIZON,
                                                create_when_not_found=False, n_grid_points: int=SUN_SET_RISE_PRECISION) -> dict:
    """
    Retrieve for given stations and given timestamps the sunrise/sunset/day/night data as dictionary
    If station/timestamp is already calculated it will be retrieved from database otherwise it will be calculated
    and added to the database for possible future retrieval (optional parameter must be true).
    Storing the pre-calculated data into a database makes retrieval faster.

    The sunrise/sunset is always on the date of the timestamp.

    :param timestamps: tuple of datetimes, e.g. datetime(2020, 1, 1)
    :param stations: tuple of station names, e.g. ("CS002")
    :param angle_to_horizon: the angle between horizon and given coordinates for which rise and set times are returned
    :param: create_when_not_found: Add data to database if not found in database and so calculated for first time
    :return A dict that maps station names to a nested dict that contains lists of start and end times for sunrise,
            sunset, on each requested date.
        E.g.
        {"CS002":
            {   "sunrise": [{"start": datetime(2020, 1, 1, 6, 0, 0)), "end": datetime(2020, 1, 1, 6, 30, 0)},
                            {"start": datetime(2020, 1, 2, 6, 0, 0)), "end": datetime(2020, 1, 2, 6, 30, 0)}],
                "sunset": [{"start": datetime(2020, 1, 1, 18, 0, 0)), "end": datetime(2020, 1, 1, 18, 30, 0)},
                           {"start": datetime(2020, 1, 2, 18, 0, 0)), "end": datetime(2020, 1, 2, 18, 30, 0)}]
            }
        }
    """
    return_dict = {}
    for station in stations:
        # Create overall result
        return_dict.setdefault(station, {})
        return_dict[station].setdefault("sunrise", [])
        return_dict[station].setdefault("sunset", [])

        # TODO: cache these with the memcached server, see TMSS-1980. Remove StationTimeline model and table.

        observer = create_astroplan_observer_for_station(station)
        for timestamp in timestamps:
            try:
                obj = StationTimeline.objects.get(station_name=station, timestamp=datetime.date(timestamp))
                sunrise_dict = {"start": obj.sunrise_start, "end": obj.sunrise_end}
                sunset_dict = {"start": obj.sunset_start, "end": obj.sunset_end}
            except ObjectDoesNotExist:
                # Not found in database so calculate it
                sunrise_dict, sunset_dict = calculate_and_get_sunrise_and_sunset_of_observer_day(observer, timestamp, angle_to_horizon, n_grid_points=n_grid_points)

                # Add to database
                if create_when_not_found:
                    try:
                        station_timeline = StationTimeline.objects.create(
                                                    station_name=station,
                                                    timestamp=timestamp,
                                                    sunrise_start=sunrise_dict['start'],
                                                    sunrise_end=sunrise_dict['end'],
                                                    sunset_start=sunset_dict['start'],
                                                    sunset_end=sunset_dict['end'])
                        logger.debug("StationTimeline %s calculated and created for station=%s, timestamp=%s" %
                                    (station_timeline, station, timestamp))
                    except IntegrityError:
                        # already exists in db
                        pass
            except Exception as e:
                logger.error("timestamps_and_stations_to_sun_rise_and_set: %s", str(e))

            # Create overall result
            return_dict.setdefault(station, {})
            return_dict[station]["sunrise"].append(sunrise_dict)
            return_dict[station]["sunset"].append(sunset_dict)

    return return_dict

def timestamps_and_stations_to_sun_rise_set_day_night(timestamps: tuple, stations: tuple, angle_to_horizon: Angle=SUN_SET_RISE_ANGLE_TO_HORIZON,
                                                      create_when_not_found=False, n_grid_points: int=SUN_SET_RISE_PRECISION) -> dict:
    """
    Same as method timestamps_and_stations_to_sun_rise_and_set, but now with the 'day' and 'night' as well.

    The day/sunrise/sunset is always on the date of the timestamp.
    The night is usually the one _starting_ on the date of the time stamp, unless the given timestamp falls before
    sunrise, in which case it is the night _ending_ on the timestamp date.

    :param timestamps: tuple of datetimes, e.g. datetime(2020, 1, 1)
    :param stations: tuple of station names, e.g. ("CS002")
    :param angle_to_horizon: the angle between horizon and given coordinates for which rise and set times are returned
    :param: create_when_not_found: Add data to database if not found in database and so calculated for first time
    :return A dict that maps station names to a nested dict that contains lists of start and end times for sunrise,
            sunset, day and night, on each requested date.
        E.g.
        {"CS002":
            {   "sunrise": [{"start": datetime(2020, 1, 1, 6, 0, 0)), "end": datetime(2020, 1, 1, 6, 30, 0)},
                            {"start": datetime(2020, 1, 2, 6, 0, 0)), "end": datetime(2020, 1, 2, 6, 30, 0)}],
                "sunset": [{"start": datetime(2020, 1, 1, 18, 0, 0)), "end": datetime(2020, 1, 1, 18, 30, 0)},
                           {"start": datetime(2020, 1, 2, 18, 0, 0)), "end": datetime(2020, 1, 2, 18, 30, 0)}],
                "day": [{"start": datetime(2020, 1, 1, 6, 30, 0)), "end": datetime(2020, 1, 1, 18, 00, 0)},
                        {"start": datetime(2020, 1, 2, 6, 30, 0)), "end": datetime(2020, 1, 2, 18, 00, 0)}],
                "night": [{"start": datetime(2020, 1, 1, 18, 30, 0)), "end": datetime(2020, 1, 2, 6, 0, 0)},
                          {"start": datetime(2020, 1, 2, 18,3 0, 0)), "end": datetime(2020, 1, 3, 6, 0, 0)}],
            }
        }
    """
    return_dict = {}
    # Derive day/night from sunset/sunrise
    for station in stations:
        # Create overall result
        return_dict.setdefault(station, {})
        return_dict[station].setdefault("sunrise", [])
        return_dict[station].setdefault("sunset", [])
        return_dict[station].setdefault("day", [])
        return_dict[station].setdefault("night", [])

        for timestamp in timestamps:
            today = timestamps_and_stations_to_sun_rise_and_set([timestamp], [station], angle_to_horizon, create_when_not_found, n_grid_points)
            sunrise_dict = today[station]['sunrise'][0]
            sunset_dict = today[station]['sunset'][0]
            return_dict[station]['sunrise'].append(sunrise_dict)
            return_dict[station]['sunset'].append(sunset_dict)

            # the 'day' start/end is always for 'today' of this timestamp
            day_dict = {"start": sunrise_dict["end"], "end": sunset_dict["start"]}
            return_dict[station]['day'].append(day_dict)

            # 'night', can be either next- or previous night
            if timestamp >= sunrise_dict["start"]:
                # Determine next sunrise start
                tomorrow = timestamps_and_stations_to_sun_rise_and_set([timestamp + timedelta(days=1)], [station], angle_to_horizon, create_when_not_found, n_grid_points)
                next_sunrise_dict = tomorrow[station]['sunrise'][0]
                night_dict = {"start": sunset_dict["end"], "end": next_sunrise_dict['start']}
            else:
                # Determine previous sunset end
                yesterday = timestamps_and_stations_to_sun_rise_and_set([timestamp - timedelta(days=1)], [station], angle_to_horizon, create_when_not_found, n_grid_points)
                prev_sunset_dict = yesterday[station]['sunset'][0]
                night_dict = {"start": prev_sunset_dict["end"], "end": sunrise_dict["start"]}
            return_dict[station]['night'].append(night_dict)

    return return_dict


# a guesstimate for a reasonable cache size for astropy results
# roughly 50 stations, 365*24*12 timestamps (1year, 24 hours, 12/h), 1000 coordinates
# in reality the numbers are more likely to be: 4 stations, 150*24*12 timestamps (0.5year, 24 hours, 12/h), 250 coordinates
CACHE_SIZE = 50 * 365 * 24 * 12 * 1000

@lru_cache(maxsize=CACHE_SIZE, typed=False)
def calculate_and_get_sunrise_and_sunset_of_observer_day(observer, timestamp: datetime, angle_to_horizon: Angle=SUN_SET_RISE_ANGLE_TO_HORIZON, n_grid_points: int=SUN_SET_RISE_PRECISION) -> (dict, dict):
    """
    Compute sunrise, sunset of the given observer object (station) at the given timestamp.
    :param observer: observer object
    :param timestamp: Datetime of a day (datetime(2020, 1, 1)
    :param the angle between horizon and given coordinates for which rise and set times are returned
    :return: dictionaries (with 'start' and 'end' defined) of sunrise, sunset
    """
    # prepare the return dicts
    sunrise_dict = {"start": None, "end": None}
    sunset_dict = {"start": None, "end": None}

    noon = datetime.combine(timestamp.date(), dtime(12, 0, 0))
    prev_midnight = noon - timedelta(hours=12)
    next_midnight = noon + timedelta(hours=12)

    sunrise_start = observer.sun_rise_time(time=Time(noon),
                                           horizon=-angle_to_horizon, which='previous',
                                           n_grid_points=n_grid_points)
    try:
        sunrise_dict['start'] = round_to_second_precision(sunrise_start.to_datetime())
    except (ValueError, AttributeError):
        # sun never rises / astropy bug
        sun_body = astropy.coordinates.get_body("sun", Time(noon), observer.location)
        if observer.target_is_up(target=sun_body, time=Time(noon), horizon=-angle_to_horizon):
            sunrise_dict['start'] = prev_midnight
        else:
            sunrise_dict['start'] = noon

    sunrise_end = observer.sun_rise_time(time=Time(sunrise_dict['start']), horizon=angle_to_horizon, which='next',
                                         n_grid_points=n_grid_points)
    try:
        sunrise_dict['end'] = round_to_second_precision(sunrise_end.to_datetime())
    except (ValueError, AttributeError):
        # sun never rises / astropy bug
        sun_body = astropy.coordinates.get_body("sun", Time(sunrise_dict['start']), observer.location)
        if observer.target_is_up(target=sun_body, time=Time(sunrise_dict['start']), horizon=angle_to_horizon):
            sunrise_dict['end'] = prev_midnight
        else:
            sunrise_dict['end'] = noon

    sunset_start = observer.sun_set_time(time=Time(noon), horizon=angle_to_horizon, which='next',
                                         n_grid_points=n_grid_points)
    try:
        sunset_dict['start'] = round_to_second_precision(sunset_start.to_datetime())
    except (ValueError, AttributeError):
        # sun never rises / astropy bug
        sun_body = astropy.coordinates.get_body("sun", Time(noon), observer.location)
        if observer.target_is_up(target=sun_body, time=Time(noon), horizon=angle_to_horizon):
            sunset_dict['start'] = next_midnight
        else:
            sunset_dict['start'] = noon

    sunset_end = observer.sun_set_time(time=Time(sunset_dict['start']), horizon=-angle_to_horizon, which='next',
                                       n_grid_points=n_grid_points)
    try:
        sunset_dict['end'] = round_to_second_precision(sunset_end.to_datetime())
    except (ValueError, AttributeError):
        # sun never rises / astropy bug
        sun_body = astropy.coordinates.get_body("sun", Time(noon), observer.location)
        if observer.target_is_up(target=sun_body, time=Time(noon), horizon=-angle_to_horizon):
            sunset_dict['end'] = next_midnight
        else:
            sunset_dict['end'] = noon

    return sunrise_dict, sunset_dict


@lru_cache(maxsize=CACHE_SIZE, typed=False)
def coordinates_and_timestamp_to_separation_from_body(pointing: Pointing, timestamp: datetime, observer: 'Observer', body: str) -> dict:
    """
    compute angular distance of a given sky coordinate from the given solar system body at the given timestamp (seen from the observer location)
    :param pointing: A Pointing object containing the celestial coordinates
    :param timestamp: a datetime for which to calculate the angle
    :param body: a solar system body as string, e.g. 'sun', 'moon' or 'jupiter'
    :return: a separation angle between the given coordinate and the solar system body at the given time
    """
    coord = pointing.as_SkyCoord(timestamp, observer.location)
    body_coord = get_body(body=body, time=astropy.time.Time(timestamp), location=observer.location)

    # Order of arguments matters!
    # body_coord.separation(coord) is correct
    # coord.separation(body_coord) is incorrect
    # See https://docs.astropy.org/en/stable/coordinates/common_errors.html#object-separation
    angle = body_coord.separation(coord)

    return angle


def coordinates_and_timestamps_to_separation_from_bodies(pointing: Pointing, timestamps: tuple, bodies: tuple, station: str="CS002") -> dict:
    """
    compute angular distances of the given sky coordinates from the given solar system bodies at the given timestamps (seen from LOFAR core or alternatively provided station)
    :param pointing: A Pointing object containing the celestial coordinates
    :param timestamps: tuple of datetimes, e.g. (datetime(2020, 1, 1, 15, 0, 0), datetime(2020, 1, 1, 16, 0, 0))
    :param bodies: tuple of solar system bodies, e.g. ('sun', 'moon', 'jupiter')
    :param station: a station name
    :return A dict that maps each body to a dict that maps the given timestamp to a separation angle from the given coordinate.
        E.g.
        {
           "sun": {datetime(2020, 1, 1, 6, 0, 0): Angle("0.7rad"), datetime(2020, 1, 1, 7, 0, 0): Angle("0.7rad")},
           "moon": {datetime(2020, 1, 1, 6, 0, 0): Angle("0.4rad"), datetime(2020, 1, 1, 7, 0, 0): Angle("0.4rad")},
           "jupiter": {datetime(2020, 1, 1, 6, 0, 0): Angle("2.7rad"), datetime(2020, 1, 1, 7, 0, 0): Angle("2.7rad")}
        }
    """
    return_dict = {}
    observer = create_astroplan_observer_for_station(station)
    for body in bodies:
        for timestamp in timestamps:
            angle = coordinates_and_timestamp_to_separation_from_body(pointing, timestamp, observer, body)
            return_dict.setdefault(body, {})[timestamp] = angle
    return return_dict


# default angle above horizon, above which the target it reporte as 'up'
TARGET_SET_RISE_ANGLE_TO_HORIZON = Angle(0, unit=astropy.units.deg)  # if default should be non-zero, should we include it explicitly in response?
# default n_grid_points; higher is more precise but very costly; astropy defaults to 150, note that errors can be in the minutes with a lower values
TARGET_SET_RISE_PRECISION = 50

@lru_cache(maxsize=CACHE_SIZE, typed=False)
def coordinates_timestamp_and_station_to_target_rise_and_set(pointing: Pointing, timestamp: datetime, observer: 'Observer', angle_to_horizon: Angle=TARGET_SET_RISE_ANGLE_TO_HORIZON, n_grid_points: int=TARGET_SET_RISE_PRECISION) -> dict:
    """
    Compute rise and set times of the given coordinates above the provided horizon, for the given station and timestamp.
    The set time is always the one following the provided timestamp.
    This implies that if the target is up at a given timestamp, the surrounding rise and set times are returned.
    Otherwise both rise and set times follow the timestamp.
    :param pointing: A Pointing object containing the celestial coordinates
    :param timestamps: tuple of datetimes, e.g. (datetime(2020, 1, 1), datetime(2020, 1, 2))
    :param observer: station location as astroplan Observer object
    :param angle_to_horizon: the angle between horizon and given coordinates for which rise and set times are returned
    :return A dict with rise and set times for the requested date and station. If rise and set are None, the target is always above or below horizon, and the respective boolean is True.
        e.g. {"rise": datetime(2020, 1, 1, 4, 0, 0), "set": datetime(2020, 1, 1, 11, 0, 0), "always_above_horizon": False, "always_below_horizon": False}
    """
    try:
        if pointing.direction_type == 'J2000':
            coord = pointing.as_SkyCoord()
            target_set = observer.target_set_time(target=coord, time=Time(timestamp), horizon=angle_to_horizon, which='next', n_grid_points=n_grid_points)
            target_rise = observer.target_rise_time(target=coord, time=Time(target_set), horizon=angle_to_horizon, which='previous', n_grid_points=n_grid_points)
        elif pointing.direction_type == 'SUN':
            target_set = observer.sun_set_time(time=Time(timestamp), horizon=angle_to_horizon, which='next', n_grid_points=n_grid_points)
            target_rise = observer.sun_rise_time(time=Time(target_set), horizon=angle_to_horizon, which='previous', n_grid_points=n_grid_points)
        elif pointing.direction_type == 'MOON':
            target_set = observer.moon_set_time(time=Time(timestamp), horizon=angle_to_horizon, which='next', n_grid_points=n_grid_points)
            target_rise = observer.moon_rise_time(time=Time(target_set), horizon=angle_to_horizon, which='previous', n_grid_points=n_grid_points)
        else:
            raise ValueError("Do not know how to determine an appropriate target for rise and set time calculation with direction_type=%s" % pointing.direction_type)
            # todo: fixed sky coordinates won't work for a moving celestial body and NonFixedTarget does not seem to be
            #  implemented/supported by target_rise/set_time, so we'd have to sample ourselves, I guess.
        return {"rise": target_rise.to_datetime(),
                "set": target_set.to_datetime(),
                "always_above_horizon": False,
                "always_below_horizon": False}
    except (TypeError, ValueError) as e:
        if "numpy.float64" in str(e) or "could not broadcast input array" in str(e) or "Cannot cast array data from dtype" in str(e):
            # Note: when the target is always above or below horizon, astroplan excepts with the not very
            #  meaningful error: 'numpy.float64' object does not support item assignment
            # Determine whether the target is always above or below horizon so that we can return some useful
            # additional info, e.g. for scheduling purposes.
            if pointing.direction_type == 'J2000':
                is_up = observer.target_is_up(target=coord, time=Time(timestamp), horizon=angle_to_horizon)
            elif pointing.direction_type == 'SUN':
                is_up = observer.is_night(Time(timestamp), horizon=angle_to_horizon) is False
            elif pointing.direction_type == 'MOON':
                #is_up = observer.moon_altaz(Time(timestamp)).alt.radian > angle_to_horizon
                coord = get_body(body='moon', time=Time(timestamp), location=observer.location)
                is_up = observer.target_is_up(target=coord, time=Time(timestamp), horizon=angle_to_horizon)
            else:
                raise ValueError("Do not know how to determine if target is always up for direction_type=%s" % pointing.direction_type)
            return {"rise": None,
                    "set": None,
                    "always_above_horizon": is_up,
                    "always_below_horizon": not is_up}
        else:
            raise



def coordinates_timestamps_and_stations_to_target_rise_and_set(pointing: Pointing, timestamps: tuple, stations: tuple, angle_to_horizon: Angle=TARGET_SET_RISE_ANGLE_TO_HORIZON, n_grid_points: int=TARGET_SET_RISE_PRECISION) -> dict:
    """
    Compute rise and set times of the given coordinates above the provided horizon, for each given station and timestamp.
    The set time is always the one following the provided timestamp.
    This implies that if the target is up at a given timestamp, the surrounding rise and set times are returned.
    Otherwise both rise and set times follow the timestamp.
    :param pointing: A Pointing object containing the celestial coordinates    :param timestamps: tuple of datetimes, e.g. (datetime(2020, 1, 1), datetime(2020, 1, 2))
    :param stations: tuple of station names, e.g. ("CS002",)
    :param angle_to_horizon: the angle between horizon and given coordinates for which rise and set times are returned
    :return A dict that maps station names to a list of dicts with rise and set times for each requested date.
            If rise and set are None, the target is always above or below horizon, and the respective boolean is True.
        E.g.
        {"CS002": [{"rise": datetime(2020, 1, 1, 4, 0, 0), "set": datetime(2020, 1, 1, 11, 0, 0), "always_above_horizon": False, "always_below_horizon": False},
                   {"rise": datetime(2020, 1, 2, 4, 0, 0), "set": datetime(2020, 1, 2, 11, 0, 0), "always_above_horizon": False, "always_below_horizon": False}]
        }
    """
    return_dict = {}
    for station in stations:
        observer = create_astroplan_observer_for_station(station)
        for timestamp in timestamps:
            target_rise_and_set = coordinates_timestamp_and_station_to_target_rise_and_set(pointing, timestamp, observer, angle_to_horizon=angle_to_horizon, n_grid_points=n_grid_points)
            return_dict.setdefault(station, []).append(target_rise_and_set)
    return return_dict


# default n_grid_points; higher is more precise but very costly; astropy defaults to 150, note that errors can be in the minutes with a lower values
TARGET_TRANSIT_PRECISION = 100

@lru_cache(maxsize=CACHE_SIZE, typed=False)
def coordinates_timestamp_and_station_to_target_transit(pointing: Pointing, timestamp: datetime, observer: 'Observer', n_grid_points: int=TARGET_TRANSIT_PRECISION, which: str='nearest') -> datetime:
    """
    Compute nearest/next/previous meridian transit times of the given coordinates for a single station and timestamp.
    :param pointing: A Pointing object containing the celestial coordinates
    :param timestamp: the datetime you want the closest transit time for
    :param observer: station location as astroplan Observer object
    :param which: which transit, the 'nearest', the 'next', or the 'previous'
    :return The nearest/next/previous transit time to the input timestamp as datetime.
    """
    try:
        if pointing.direction_type == 'J2000':
            coord = pointing.as_SkyCoord()
            target_transit = observer.target_meridian_transit_time(target=coord, time=Time(timestamp), which=which, n_grid_points=n_grid_points).to_datetime()
        elif pointing.direction_type == 'SUN':
            target_transit = observer.noon(time=Time(timestamp), n_grid_points=n_grid_points, which=which).to_datetime()
        elif pointing.direction_type == 'MOON':
            if which != 'nearest':
                raise NotImplementedError()
            moon_set = observer.sun_set_time(time=Time(timestamp), which='next', n_grid_points=n_grid_points)
            moon_rise = observer.sun_rise_time(time=Time(moon_set), which='previous', n_grid_points=n_grid_points)
            target_transit = (moon_rise + 0.5 * (moon_set - moon_rise)).to_datetime()
        else:
            # todo: fixed sky coordinates won't work for a moving celestial body and NonFixedTarget does not seem to be
            #  implemented/supported by target_meridian_transit_time, so we'd have to sample ourselves, I guess.
            raise NotImplementedError("Do not know how to determine an appropriate target for transit time calculation with direction_type=%s" % pointing.direction_type)

        if isinstance(target_transit, numpy.ndarray):
            # a workaround for astropy bug.
            # sometimes (when target is always above or below the horizon) the astropy Time.to_datetime() returns an array
            # the actual value is the first item in the array. Use that.
            target_transit = target_transit[0]

            # a workaround for an even deeper astropy bug.
            # if the result is still an ndarray, recurse with slightly altered arguments
            # might result in a recursion error, but alas, at least we tried.
            if isinstance(target_transit, numpy.ndarray):
                return coordinates_timestamp_and_station_to_target_transit(pointing,
                                                                           timestamp + timedelta(minutes=1),
                                                                           observer,
                                                                           n_grid_points+1,
                                                                           which=which)

        if which == 'nearest' and abs(target_transit-timestamp) > timedelta(hours=12):
            # a workaround for astropy bug.
            # when the input timestamp is near to transit already then astropy has difficulty computing the 'nearest' transit (as it sits on the edge of the 'previous' and 'next' timestamp-grid
            # So, fix that by computing the previous and next transit, but then for a shifted input timestamp (to NOT be near to the timestamp-grid)
            prev_offset_target_transit = coordinates_timestamp_and_station_to_target_transit(pointing, timestamp + timedelta(hours=3), observer, n_grid_points, which='previous')
            next_offset_target_transit = coordinates_timestamp_and_station_to_target_transit(pointing, timestamp - timedelta(hours=3), observer, n_grid_points, which='next')
            # they should both be almost equal (depending on grid resolution)
            assert(abs(prev_offset_target_transit-next_offset_target_transit) < timedelta(hours=3))

            # also check the assumption about astropy's bug. Is the timestamp indeed near to transit?
            assert(abs(prev_offset_target_transit-timestamp) <= timedelta(hours=3))
            assert(abs(next_offset_target_transit-timestamp) <= timedelta(hours=3))

            # all fine. Return even more precise averaged result.
            target_transit = prev_offset_target_transit + 0.5*(next_offset_target_transit-prev_offset_target_transit)
        return target_transit

    except (ValueError, IndexError):
        # workaround for astropy bug
        # on rare occasions, astropy can't compute noon.
        # probably due to rounding/precision.
        # try again (recurse) with a slightly altered timestamp and n_grid_points
        # in the very worst case this leads to a recursion stack overflow
        return coordinates_timestamp_and_station_to_target_transit(pointing, timestamp + timedelta(minutes=1), observer, n_grid_points+1, which)


def coordinates_timestamps_and_stations_to_target_transit(pointing: Pointing, timestamps: tuple, stations: tuple, n_grid_points: int=TARGET_TRANSIT_PRECISION, which: str='nearest') -> dict:
    """
    Compute nearest/next/previous meridian transit times of the given coordinates for each given station and timestamp.
    :param pointing: A Pointing object containing the celestial coordinates
    :param timestamps: tuple of datetimes, e.g. (datetime(2020, 1, 1), datetime(2020, 1, 2))
    :param stations: tuple of station names, e.g. ("CS002",)
    :param which: which transit, the 'nearest', the 'next', or the 'previous'
    :return A dict that maps station names to a list of transit times (nearest/next/previous transit for each requested timestamp).
        E.g.
        {"CS002": [datetime(2020, 1, 1, 4, 0, 0), datetime(2020, 1, 2, 4, 0, 0)]}
    """
    return_dict = {}
    for station in stations:
        observer = create_astroplan_observer_for_station(station)
        for timestamp in timestamps:
            target_transit = coordinates_timestamp_and_station_to_target_transit(pointing, timestamp, observer, n_grid_points=n_grid_points, which=which)
            return_dict.setdefault(station, []).append(target_transit)
    return return_dict


@lru_cache(maxsize=CACHE_SIZE, typed=False)
def compute_elevation(pointing: Pointing, timestamp: datetime, station: str='CS002') -> float:
    """
    Compute the elevation of a pointing for the given pointing, timestamp and station.
    :param pointing: A Pointing object containing the celestial coordinates
    :param timestamp: the datetime you want the elevation at
    :param station: station name
    :return the elevation in radians
    """
    observer = create_astroplan_observer_for_station(station)
    coord = pointing.as_SkyCoord(timestamp, observer.location)
    altaz = observer.altaz(target=coord, time=Time(timestamp))  # alt(itude) and az(imuth)
    return altaz.alt.rad

def local_sidereal_time_for_utc_and_station_as_pytime(timestamp: datetime = None,
                                                      station: str = 'CS002',
                                                      field: str = 'LBA',
                                                      kind: str = "apparent") -> dtime:
    """
    calculate local sidereal time for given utc time and station
    Same as local_sidereal_time_for_utc_and_station but returning the lst time as native python time object
    """
    # convert the UTC timestamp to LST at CS002 (expressed as astropy Longitude)
    lst_longitude = local_sidereal_time_for_utc_and_station(timestamp, station, field, kind)

    # convert the astropy Longitude LST time back to a native python time (ignore microsecond)
    # we're "forced" to convert via datetime, evan though the date is irrelevant (and discarded)
    lst_longitude_as_timedelta = timedelta(hours=lst_longitude.hour)
    lst_longitude_as_py_time = (datetime(2022, 1, 1, 0, 0, 0) + lst_longitude_as_timedelta).time().replace(microsecond=0)
    return lst_longitude_as_py_time


def local_sidereal_time_for_utc_and_station(timestamp: datetime = None,
                                            station: str = 'CS002',
                                            field: str = 'LBA',
                                            kind: str = "apparent") -> astropy.coordinates.Longitude:
    """
    calculate local sidereal time for given utc time and station
    :param timestamp: timestamp as datetime object
    :param station: station name
    :param field: antennafield, 'LBA' or 'HBA'
    :param kind: 'mean' or 'apparent'
    :return:
    """
    if timestamp is None:
        timestamp = datetime.utcnow()
    location = create_location_for_station(station)
    return local_sidereal_time_for_utc_and_longitude(timestamp=timestamp, longitude=location.lon.to_string(decimal=True), kind=kind)


def local_sidereal_time_for_utc_and_longitude(timestamp: datetime = None,
                                              longitude: float = 6.8693028,
                                              kind: str = "apparent") -> astropy.coordinates.Longitude:
    """
    :param timestamp: timestamp as datetime object
    :param longitude: decimal longitude of observer location (defaults to CS002 LBA center)
    :param kind: 'mean' or 'apparent'
    :return:
    """
    if timestamp is None:
        timestamp = datetime.utcnow()
    t = Time(timestamp, format='datetime', scale='utc')
    return t.sidereal_time(kind=kind, longitude=longitude)


def antennafields_for_antennaset_and_station(antennaset:str, station:str) -> list:
    """
    convert an antennaset to a list of antennafields
    :param antennaset: A string identifier for an antennaset, like 'HBA_DUAL'
    :param station: A string identifier for a station, like 'CS001'
    :return: a list of antennafields that the station uses for the given antennaset ['HBA0', 'HBA1']
    """
    if antennaset.startswith('LBA'):
        fields = ['LBA']
    elif antennaset.startswith('HBA') and not station.startswith('CS'):
        fields = ['HBA']
    elif antennaset.startswith('HBA_DUAL'):
        fields = ['HBA0', 'HBA1']
    elif antennaset.startswith('HBA_ZERO'):
        fields = ['HBA0']
    elif antennaset.startswith('HBA_ONE'):
        fields = ['HBA1']
    else:
        raise ValueError('Cannot determine antennafields for station=%s antennaset=%s' % (station, antennaset))

    return fields


@lru_cache()
def get_all_stations() -> Tuple[str]:
    """
    returns all LOFAR (1&2) stations.
    """
    try:
        station_schema_template = CommonSchemaTemplate.get_version_or_latest(name="stations")
        groups = station_schema_template.schema['definitions']['station_group']['anyOf']
        selected_group = next(g for g in groups if g['title'].lower() == 'all')
        return tuple(sorted(set(selected_group['properties']['stations']['enum'][0])))
    except:
        return tuple()

@lru_cache()
def get_lofar2_stations() -> Tuple[str]:
    """
    returns all LOFAR2 stations.
    """
    try:
        station_schema_template = CommonSchemaTemplate.get_latest(name="stations")
        groups = station_schema_template.schema['definitions']['station_group']['anyOf']
        selected_group = next(g for g in groups if g['title'].upper() == 'LOFAR2')
        return tuple(sorted(set(selected_group['properties']['stations']['enum'][0])))
    except:
        return tuple()

@lru_cache()
def get_lofar1_stations() -> Tuple[str]:
    """
    returns all LOFAR1 stations.
    """
    return tuple(sorted(set(get_all_stations()) - set(get_lofar2_stations())))
