

from django.db import migrations

from lofar.sas.tmss.tmss.tmssapp.populate import populate_on_hold_status_and_transitions


class Migration(migrations.Migration):

    dependencies = [
        ('tmssapp', '0036_subtask_scheduled_central_lst'),
    ]

    operations = [migrations.RunPython(populate_on_hold_status_and_transitions),

         # the following migration step alters the tmssapp_subtask_compute_aggretates_for_taskblueprint trigger function from migration 0020_task_queued_status
         # the only difference is that there is a new SubtaskState and TaskStatus ON_HOLD.
         # yes, this is ugly/annoying that we have to replicate code, but that's how database migration steps/deltas/patches work...
         migrations.RunSQL('''CREATE OR REPLACE FUNCTION tmssapp_subtask_compute_aggretates_for_taskblueprint()
         RETURNS trigger AS
         $BODY$
         DECLARE
           agg_scheduled_start_time timestamp with time zone;
           agg_scheduled_stop_time timestamp with time zone;
           agg_actual_process_start_time timestamp with time zone;
           agg_actual_process_stop_time timestamp with time zone;
           agg_actual_on_sky_start_time timestamp with time zone;
           agg_actual_on_sky_stop_time timestamp with time zone;
           agg_process_start_time timestamp with time zone;
           agg_process_stop_time timestamp with time zone;
           agg_on_sky_start_time timestamp with time zone;
           agg_on_sky_stop_time timestamp with time zone;
           agg_status character varying(128);
           agg_obsolete_since timestamp with time zone;
           agg_on_sky_duration interval;
           agg_duration interval;
         BEGIN

               -- aggregate the various timestamps
               -- consider all the subtasks for this taskblueprint, excluding non-primary obsolete subtasks
               -- use a WITH Common Table Expression followed by a SELECT INTO variables for speed 
               WITH subtask_agg AS (SELECT MIN(scheduled_start_time) as scheduled_start_time,
                                           MAX(scheduled_stop_time) as scheduled_stop_time,
                                           MIN(actual_process_start_time) as actual_process_start_time,
                                           MAX(actual_process_stop_time) as actual_process_stop_time,
                                           MIN(actual_on_sky_start_time) as actual_on_sky_start_time,
                                           MAX(actual_on_sky_stop_time) as actual_on_sky_stop_time,
                                           MIN(process_start_time) as process_start_time,
                                           MAX(process_stop_time) as process_stop_time,
                                           MIN(on_sky_start_time) as on_sky_start_time,
                                           MAX(on_sky_stop_time) as on_sky_stop_time,
                                           MIN(obsolete_since) as obsolete_since
                                           FROM tmssapp_subtask
                                           WHERE tmssapp_subtask.task_blueprint_id=NEW.task_blueprint_id
                                           AND NOT (tmssapp_subtask.primary=false AND tmssapp_subtask.obsolete_since IS NOT NULL))
               SELECT     scheduled_start_time,      scheduled_stop_time,    actual_process_start_time,     actual_process_stop_time,     actual_on_sky_start_time,     actual_on_sky_stop_time,     on_sky_start_time,     on_sky_stop_time,     process_start_time,     process_stop_time,     on_sky_start_time,     on_sky_stop_time,     obsolete_since
               INTO   agg_scheduled_start_time, agg_scheduled_stop_time, agg_actual_process_start_time, agg_actual_process_stop_time, agg_actual_on_sky_start_time, agg_actual_on_sky_stop_time, agg_on_sky_start_time, agg_on_sky_stop_time, agg_process_start_time, agg_process_stop_time, agg_on_sky_start_time, agg_on_sky_stop_time, agg_obsolete_since
               FROM subtask_agg;

                -- compute the durations using the just computed on_sky/process start/stoptimes
                agg_on_sky_duration := AGE(agg_on_sky_stop_time, agg_on_sky_start_time); 
                agg_duration := AGE(agg_process_stop_time, agg_process_start_time); 


               -- aggregate the subtask statuses into one taskblueprint status 
               -- See design: https://support.astron.nl/confluence/display/TMSS/Specification+Flow/#SpecificationFlow-TaskBlueprints                       
               -- consider all the subtasks for this taskblueprint, excluding non-primary obsolete subtasks
               -- use a WITH Common Table Expression followed by a SELECT INTO variables for speed 
               WITH subtask_statuses AS (SELECT tmssapp_subtask.state_id, tmssapp_subtasktemplate.type_id
                                           FROM tmssapp_subtask
                                           INNER JOIN tmssapp_subtasktemplate on tmssapp_subtasktemplate.id=tmssapp_subtask.specifications_template_id
                                           WHERE tmssapp_subtask.task_blueprint_id=NEW.task_blueprint_id
                                           AND NOT (tmssapp_subtask.primary=false AND tmssapp_subtask.obsolete_since IS NOT NULL))
               SELECT CASE
                   WHEN EXISTS(SELECT true FROM subtask_statuses WHERE state_id='defining') THEN 'defined'
                   WHEN (SELECT COUNT(true) FROM subtask_statuses)=0 THEN 'defined'
                   WHEN (SELECT COUNT(true) FROM subtask_statuses WHERE state_id='finished')=(SELECT COUNT(true) FROM subtask_statuses) THEN 'finished'
                   WHEN EXISTS(SELECT true FROM subtask_statuses WHERE state_id IN ('cancelling', 'cancelled')) THEN 'cancelled'
                   WHEN EXISTS(SELECT true FROM subtask_statuses WHERE state_id='error') THEN 'error'
                   WHEN EXISTS(SELECT true FROM subtask_statuses WHERE state_id='on hold') THEN 'on hold'
                   WHEN EXISTS(SELECT true FROM subtask_statuses WHERE type_id='observation' AND state_id IN ('finishing', 'finished')) THEN 'observed'
                   WHEN EXISTS(SELECT true FROM subtask_statuses WHERE state_id IN ('starting','started','finishing','finished')) THEN 'started'
                   WHEN EXISTS(SELECT true FROM subtask_statuses WHERE state_id IN ('queueing','queued','restarting')) THEN 'queued'
                   WHEN EXISTS(SELECT true FROM subtask_statuses WHERE state_id='scheduled') THEN 'scheduled'
                   WHEN EXISTS(SELECT true FROM subtask_statuses WHERE state_id='unschedulable') THEN 'unschedulable'
                   ELSE 'schedulable'
                 END
               INTO agg_status
               FROM subtask_statuses;

               -- default when no subtasks available
               IF agg_status IS NULL THEN
                   agg_status := 'schedulable';
               END IF;

               -- all aggregated timestamps and statuses were computed
               -- now update the referred taskblueprint with these aggregated values  
               UPDATE tmssapp_taskblueprint
               SET scheduled_start_time=agg_scheduled_start_time, 
                   scheduled_stop_time=agg_scheduled_stop_time,
                   actual_process_start_time=agg_actual_process_start_time, 
                   actual_process_stop_time=agg_actual_process_stop_time, 
                   actual_on_sky_start_time=agg_actual_on_sky_start_time, 
                   actual_on_sky_stop_time=agg_actual_on_sky_stop_time, 
                   process_start_time=agg_process_start_time, 
                   process_stop_time=agg_process_stop_time, 
                   on_sky_start_time=agg_on_sky_start_time, 
                   on_sky_stop_time=agg_on_sky_stop_time, 
                   status_id=agg_status, 
                   obsolete_since=agg_obsolete_since,
                   on_sky_duration=agg_on_sky_duration,
                   duration=agg_duration
               WHERE id=NEW.task_blueprint_id;
           RETURN NEW;
           END;
           $BODY$
           LANGUAGE plpgsql VOLATILE;
           DROP TRIGGER IF EXISTS tmssapp_subtask_compute_aggretates_for_taskblueprint ON tmssapp_subtask ;
           CREATE TRIGGER tmssapp_subtask_compute_aggretates_for_taskblueprint
           AFTER INSERT OR UPDATE ON tmssapp_subtask
           FOR EACH ROW EXECUTE PROCEDURE tmssapp_subtask_compute_aggretates_for_taskblueprint();
         '''),

         # the following migration step alters the tmssapp_taskblueprint_compute_aggretates_for_schedulingunitblueprint trigger function from migration 0015_db_aggregates
         # the only difference is that there is a new TaskStatus QUEUED.
         # yes, this is ugly/annoying that we have to replicate code, but that's how database migration steps/deltas/patches work...
         migrations.RunSQL('''CREATE OR REPLACE FUNCTION tmssapp_taskblueprint_compute_aggretates_for_schedulingunitblueprint()
         RETURNS trigger AS
         $BODY$
         DECLARE
           agg_scheduled_start_time timestamp with time zone;
           agg_scheduled_stop_time timestamp with time zone;
           agg_actual_process_start_time timestamp with time zone;
           agg_actual_process_stop_time timestamp with time zone;
           agg_actual_on_sky_start_time timestamp with time zone;
           agg_actual_on_sky_stop_time timestamp with time zone;
           agg_process_start_time timestamp with time zone;
           agg_process_stop_time timestamp with time zone;
           agg_on_sky_start_time timestamp with time zone;
           agg_on_sky_stop_time timestamp with time zone;
           agg_observed_start_time timestamp with time zone;
           agg_observed_stop_time timestamp with time zone;
           agg_status character varying(128);
           agg_obsolete_since timestamp with time zone;
           agg_on_sky_duration interval;
           agg_observed_duration interval;
           agg_duration interval;
         BEGIN
    
               -- aggregate the various timestamps
               -- consider all the tasks for this schedulingunitblueprint
               -- use a WITH Common Table Expression followed by a SELECT INTO variables for speed 
               WITH task_agg AS (SELECT MIN(scheduled_start_time) as scheduled_start_time,
                                        MAX(scheduled_stop_time) as scheduled_stop_time,
                                        MIN(actual_process_start_time) as actual_process_start_time,
                                        MAX(actual_process_stop_time) as actual_process_stop_time,
                                        SUM(duration) as duration,
                                        MIN(actual_on_sky_start_time) as actual_on_sky_start_time,
                                        MAX(actual_on_sky_stop_time) as actual_on_sky_stop_time,
                                        MIN(process_start_time) as process_start_time,
                                        MAX(process_stop_time) as process_stop_time,
                                        MIN(on_sky_start_time) as on_sky_start_time,
                                        MAX(on_sky_stop_time) as on_sky_stop_time,
                                        SUM(on_sky_duration) as on_sky_duration,
                                        MIN(obsolete_since) as obsolete_since
                                        FROM tmssapp_taskblueprint
                                        WHERE tmssapp_taskblueprint.scheduling_unit_blueprint_id=NEW.scheduling_unit_blueprint_id
                                        AND tmssapp_taskblueprint.obsolete_since IS NULL)
               SELECT     scheduled_start_time,      scheduled_stop_time,    actual_process_start_time,     actual_process_stop_time,     actual_on_sky_start_time,     actual_on_sky_stop_time,     on_sky_start_time,     on_sky_stop_time,     process_start_time,     process_stop_time,     on_sky_start_time,     on_sky_stop_time,     obsolete_since,     duration,     on_sky_duration
               INTO   agg_scheduled_start_time, agg_scheduled_stop_time, agg_actual_process_start_time, agg_actual_process_stop_time, agg_actual_on_sky_start_time, agg_actual_on_sky_stop_time, agg_on_sky_start_time, agg_on_sky_stop_time, agg_process_start_time, agg_process_stop_time, agg_on_sky_start_time, agg_on_sky_stop_time, agg_obsolete_since, agg_duration, agg_on_sky_duration
               FROM task_agg;
    
               -- compute observed_start/stop_time observerd_duration. These are derived from finished/observed tasks only.
               -- use a WITH Common Table Expression followed by a SELECT INTO variables for speed 
               WITH observed_task_agg AS (SELECT MIN(on_sky_start_time) as observed_start_time,
                                                 MAX(on_sky_stop_time) as observed_stop_time,
                                                 SUM(on_sky_duration) as observed_duration
                                                 FROM tmssapp_taskblueprint
                                                 WHERE tmssapp_taskblueprint.scheduling_unit_blueprint_id=NEW.scheduling_unit_blueprint_id
                                                 AND tmssapp_taskblueprint.status_id in ('observed', 'finished')
                                                 AND tmssapp_taskblueprint.obsolete_since IS NULL)
               SELECT     observed_start_time,     observed_stop_time,     observed_duration
               INTO   agg_observed_start_time, agg_observed_stop_time, agg_observed_duration
               FROM observed_task_agg;
    
               -- aggregate the task statuses into one schedulingunit status 
               -- See design: https://support.astron.nl/confluence/display/TMSS/Specification+Flow/#SpecificationFlow-SchedulingUnitBlueprints
               -- consider all the taskblueprints for this schedulingunitblueprint
               -- use a WITH Common Table Expression followed by a SELECT INTO variables for speed 
               WITH task_statuses AS (SELECT tmssapp_taskblueprint.status_id, tmssapp_tasktemplate.type_id
                                      FROM tmssapp_taskblueprint
                                      INNER JOIN tmssapp_tasktemplate on tmssapp_tasktemplate.id=tmssapp_taskblueprint.specifications_template_id
                                      WHERE tmssapp_taskblueprint.scheduling_unit_blueprint_id=NEW.scheduling_unit_blueprint_id
                                      AND tmssapp_taskblueprint.obsolete_since IS NULL)
               SELECT CASE
                   WHEN (SELECT COUNT(true) FROM task_statuses)=0 THEN 'defined'
                   WHEN EXISTS(SELECT true FROM task_statuses WHERE status_id='defined') THEN 'defined'
                   WHEN (SELECT COUNT(true) FROM task_statuses WHERE status_id in ('finished', 'observed'))=(SELECT COUNT(true) FROM task_statuses) THEN 'finished'
                   WHEN EXISTS(SELECT true FROM task_statuses WHERE status_id='cancelled') THEN 'cancelled'
                   WHEN EXISTS(SELECT true FROM task_statuses WHERE status_id='error') THEN 'error'
                   WHEN EXISTS(SELECT true FROM task_statuses WHERE status_id='unschedulable') THEN 'unschedulable'
                   WHEN EXISTS(SELECT true FROM task_statuses WHERE status_id IN ('queued', 'started', 'observed', 'finished', 'on hold')) THEN 
                       CASE
                           WHEN (SELECT COUNT(true) FROM task_statuses WHERE type_id='observation' AND status_id IN ('observed', 'finished'))<(SELECT COUNT(true) FROM task_statuses WHERE type_id='observation') THEN 'observing'
                           WHEN NOT EXISTS(SELECT true FROM task_statuses WHERE type_id='pipeline' AND status_id IN ('queued', 'started', 'finished')) THEN 'observed'
                           WHEN EXISTS(SELECT true FROM task_statuses WHERE type_id='ingest' AND status_id IN ('queued', 'started')) THEN 'ingesting'
                           WHEN (SELECT COUNT(true) FROM task_statuses WHERE type_id='ingest' AND status_id='finished')=(SELECT COUNT(true) FROM task_statuses WHERE type_id='ingest') AND (SELECT COUNT(true) FROM task_statuses WHERE type_id='ingest')>0 THEN 'ingested'
                           WHEN (SELECT COUNT(true) FROM task_statuses WHERE type_id='pipeline' AND status_id='finished')=(SELECT COUNT(true) FROM task_statuses WHERE type_id='pipeline') AND (SELECT COUNT(true) FROM task_statuses WHERE type_id='pipeline')>0 THEN 'processed'
                           WHEN EXISTS(SELECT true FROM task_statuses WHERE type_id='pipeline' AND status_id IN ('queued', 'started', 'scheduled', 'finished')) THEN 'processing'
                       END                            
                   WHEN EXISTS(SELECT true FROM task_statuses WHERE status_id='scheduled') THEN 'scheduled'
                 END
               INTO agg_status
               FROM task_statuses;
    
               -- default when no tasks available
               IF agg_status IS NULL THEN
                   agg_status := 'schedulable';
               END IF;
    
               -- all aggregated timestamps and statuses were computed
               -- now update the referred taskblueprint with these aggregated values  
               UPDATE tmssapp_schedulingunitblueprint
               SET scheduled_start_time=agg_scheduled_start_time, 
                   scheduled_stop_time=agg_scheduled_stop_time,
                   actual_process_start_time=agg_actual_process_start_time, 
                   actual_process_stop_time=agg_actual_process_stop_time, 
                   actual_on_sky_start_time=agg_actual_on_sky_start_time, 
                   actual_on_sky_stop_time=agg_actual_on_sky_stop_time, 
                   process_start_time=agg_process_start_time, 
                   process_stop_time=agg_process_stop_time, 
                   on_sky_start_time=agg_on_sky_start_time, 
                   on_sky_stop_time=agg_on_sky_stop_time, 
                   observed_start_time=agg_observed_start_time, 
                   observed_stop_time=agg_observed_stop_time, 
                   status_id=agg_status, 
                   obsolete_since=agg_obsolete_since,
                   on_sky_duration=agg_on_sky_duration,
                   observed_duration=agg_observed_duration,
                   duration=agg_duration
               WHERE id=NEW.scheduling_unit_blueprint_id;
           RETURN NEW;
           END;
           $BODY$
           LANGUAGE plpgsql VOLATILE;
           DROP TRIGGER IF EXISTS tmssapp_taskblueprint_compute_aggretates_for_schedulingunitblueprint ON tmssapp_taskblueprint ;
           CREATE TRIGGER tmssapp_taskblueprint_compute_aggretates_for_schedulingunitblueprint
           AFTER INSERT OR UPDATE ON tmssapp_taskblueprint
           FOR EACH ROW EXECUTE PROCEDURE tmssapp_taskblueprint_compute_aggretates_for_schedulingunitblueprint();
         '''),

         # the following migration step alters the tmssapp_block_subtask_immutable_fields_update trigger function from migration 0036_subtask_scheduled_central_lst
         # yes, this is ugly/annoying that we have to replicate code, but that's how database migration steps/deltas/patches work...
         migrations.RunSQL('''CREATE OR REPLACE FUNCTION tmssapp_block_subtask_immutable_fields_update()
              RETURNS trigger AS
              $BODY$
              BEGIN
                IF TG_OP = 'UPDATE' THEN
                  IF OLD.specifications_template_id <> NEW.specifications_template_id OR
                     OLD.processing_feedback_doc <> NEW.processing_feedback_doc OR
                     OLD.processing_feedback_template_id <> NEW.processing_feedback_template_id OR
                     OLD.primary <> NEW.primary OR
                     OLD.task_blueprint_id <> NEW.task_blueprint_id OR
                     OLD.cluster_id <> NEW.cluster_id THEN
                    RAISE EXCEPTION 'ILLEGAL UPDATE OF IMMUTABLE SUBTASK FIELD';
                  END IF;

                  IF OLD.specifications_doc <> NEW.specifications_doc THEN
                    IF OLD.state_id NOT IN ('defined', 'scheduling', 'unscheduling') THEN
                      RAISE EXCEPTION 'ILLEGAL UPDATE OF IMMUTABLE SUBTASK FIELD specifications_doc. state=%', OLD.state_id;
                    END IF;
                  END IF;

                  IF NEW.actual_process_start_time IS NOT NULL AND OLD.actual_process_start_time <> NEW.actual_process_start_time THEN
                      IF OLD.actual_process_start_time IS NOT NULL THEN
                          RAISE EXCEPTION 'subtask.actual_process_start_time may only be set once';
                      END IF;

                      IF NEW.state_id NOT IN ('finishing', 'finished', 'cancelling', 'cancelled', 'error') THEN
                        RAISE EXCEPTION 'subtask.actual_process_start_time cannot be set when state=%', NEW.state_id;
                      END IF;
                  END IF;

                  IF NEW.actual_process_stop_time IS NOT NULL AND OLD.actual_process_stop_time <> NEW.actual_process_stop_time THEN
                      IF OLD.actual_process_stop_time IS NOT NULL THEN
                          RAISE EXCEPTION 'subtask.actual_process_stop_time may only be set once';
                      END IF;

                      IF NEW.state_id NOT IN ('starting', 'started', 'cancelling', 'cancelled', 'error') THEN
                        RAISE EXCEPTION 'subtask.actual_process_stop_time cannot be set when state=%', NEW.state_id;
                      END IF;
                  END IF;

                  IF NEW.actual_on_sky_start_time IS NOT NULL AND OLD.actual_on_sky_start_time <> NEW.actual_on_sky_start_time THEN
                      IF OLD.actual_on_sky_start_time IS NOT NULL THEN
                          RAISE EXCEPTION 'subtask.actual_on_sky_start_time may only be set once';
                      END IF;

                      IF NEW.state_id NOT IN ('starting', 'started', 'cancelling', 'cancelled', 'error') THEN
                        RAISE EXCEPTION 'subtask.actual_on_sky_start_time cannot be set when state=%', NEW.state_id;
                      END IF;
                  END IF;

                  IF NEW.actual_on_sky_stop_time IS NOT NULL AND OLD.actual_on_sky_stop_time <> NEW.actual_on_sky_stop_time THEN
                      IF OLD.actual_on_sky_stop_time IS NOT NULL THEN
                          RAISE EXCEPTION 'subtask.actual_on_sky_stop_time may only be set once';
                      END IF;

                      IF NEW.state_id NOT IN ('finishing', 'finished', 'cancelling', 'cancelled', 'error') THEN
                        RAISE EXCEPTION 'subtask.actual_on_sky_stop_time cannot be set when state=%', NEW.state_id;
                      END IF;
                  END IF;

                  IF OLD.error_reason <> NEW.error_reason OR (OLD.error_reason IS NULL AND NEW.error_reason IS NOT NULL) THEN
                    IF OLD.error_reason IS NOT NULL THEN
                      RAISE EXCEPTION 'subtask.error_reason may only be set once';
                    END IF;
                  END IF;

                  IF (OLD.scheduled_start_time <> NEW.scheduled_start_time) OR (OLD.scheduled_stop_time <> NEW.scheduled_stop_time) THEN
                      -- observations can only update scheduled_start_time/scheduled_stop_time while (un)scheduling
                      IF EXISTS(SELECT true FROM tmssapp_subtask INNER JOIN tmssapp_subtasktemplate ON tmssapp_subtasktemplate.id=tmssapp_subtask.specifications_template_id WHERE tmssapp_subtask.id=NEW.id AND tmssapp_subtasktemplate.type_id='observation') THEN   
                          IF OLD.state_id NOT IN ('defined', 'scheduling', 'unscheduling') THEN
                              RAISE EXCEPTION 'subtask.scheduled_start_time/scheduled_stop_time cannot be set when state=%', OLD.state_id;
                          END IF;
                      ELSE
                          -- pipelines, ingest, etc can also update scheduled_start_time/scheduled_stop_time while scheduled/queued
                          IF OLD.state_id NOT IN ('defined', 'scheduling', 'unscheduling', 'scheduled', 'queueing', 'queued') THEN
                              RAISE EXCEPTION 'subtask.scheduled_start_time/scheduled_stop_time cannot be set when state=%', OLD.state_id;
                          END IF;
                      END IF;
                  END IF;
                END IF;

                IF NEW.error_reason IS NOT NULL AND OLD.error_reason <> NEW.error_reason AND NEW.state_id <> 'error' THEN
                  RAISE EXCEPTION 'subtask.error_reason may only be set when state==error';
                END IF;

                IF OLD.scheduled_central_lst <> NEW.scheduled_central_lst THEN 
                  IF OLD.state_id NOT IN ('scheduling', 'unscheduling', 'defined', 'unschedulable') THEN
                    RAISE EXCEPTION 'subtask.scheduled_central_lst may only be set before it is scheduled';
                  END IF;

                  IF NOT EXISTS(SELECT true FROM tmssapp_subtask INNER JOIN tmssapp_subtasktemplate ON tmssapp_subtasktemplate.id=tmssapp_subtask.specifications_template_id WHERE tmssapp_subtask.id=NEW.id AND tmssapp_subtasktemplate.type_id='observation') THEN   
                    RAISE EXCEPTION 'subtask.scheduled_central_lst may only be set for observations';
                  END IF;
                END IF;

              RETURN NEW;
              END;
              $BODY$
              LANGUAGE plpgsql VOLATILE;
              DROP TRIGGER IF EXISTS tmssapp_block_subtask_immutable_fields_update ON tmssapp_Subtask ;
              CREATE TRIGGER tmssapp_block_subtask_immutable_fields_update
              BEFORE UPDATE OR INSERT ON tmssapp_Subtask
              FOR EACH ROW EXECUTE PROCEDURE tmssapp_block_subtask_immutable_fields_update();
              ''')
                  ]

