"""
This file contains common constructs used by database models in other modules
"""

import logging
import typing

logger = logging.getLogger(__name__)

from django.db.models import Model, CharField, DateTimeField, IntegerField, UniqueConstraint, ForeignKey, PROTECT, QuerySet
from django.contrib.postgres.fields import ArrayField, JSONField
from django.contrib.postgres.indexes import GinIndex
from django.db import transaction
from lofar.common.json_utils import validate_json_against_its_schema, validate_json_against_schema, add_defaults_to_json_object_for_schema, get_default_json_object_for_schema, raise_on_self_refs, resolved_remote_refs, ref_resolved_url, get_refs, remove_from_schema_cache
from lofar.sas.tmss.tmss.exceptions import SchemaValidationException, ObsoleteValidationException, ObsoleteTemplateException
from django.urls import reverse as reverse_url
from enum import Enum
import json
import jsonschema
from datetime import timedelta
from typing import Union, Tuple, Collection, Callable
from django.utils.functional import cached_property
from lofar.sas.tmss.tmss.exceptions import TMSSException


class RefreshFromDbInvalidatesCachedPropertiesMixin:
    """Helper Mixin class which invalidates all 'cached_property' attributes on a model upon refreshing from the db"""
    def refresh_from_db(self, *args, **kwargs):
        self.invalidate_cached_properties()
        return super().refresh_from_db(*args, **kwargs)

    def invalidate_cached_properties(self):
        for klass in self.__class__.mro():
            for key, value in klass.__dict__.items():
                if isinstance(value, cached_property):
                    self.__dict__.pop(key, None)


class ProjectPropertyMixin:
    @cached_property
    def project(self): # -> Project:
        '''return the related project of this task
        '''
        if not hasattr(self, 'path_to_project'):
            return TMSSException("Please define a 'path_to_project' attribute on the %s object for the ProjectPropertyMixin to function." % type(self))
        if type(self).__name__ == 'Project':
            return self
        obj = self
        for attr in self.path_to_project.split('__'):
            obj = getattr(obj, attr)
            if attr == 'project':
                return obj
            if obj and not isinstance(obj, Model):  # ManyToMany fields
                obj = obj.first()
            if obj is None:
                logger.warning("The element '%s' in the path_to_project of the %s object returned None for pk=%s" % (attr, type(self), self.pk))
                return None


class BasicCommon(Model):
    # todo: we cannot use foreign keys in the array here, so we have to keep the Tags table up to date by trigger or so.
    # todo: we could switch to a manytomany field instead?
    tags = ArrayField(CharField(max_length=128), size=8, blank=True, help_text='User-defined search keywords for object.', default=list)
    created_at = DateTimeField(auto_now_add=True, help_text='Moment of object creation.')
    updated_at = DateTimeField(auto_now=True, help_text='Moment of last object update.')

    @property
    def is_used(self) -> bool:
        '''Is this model instance used by any of its related objects?'''
        for rel_obj in self._meta.related_objects:
            if rel_obj.related_model.objects.filter(**{rel_obj.field.attname: self}).count() > 0:
                return True
        return False

    class Meta:
        abstract = True
        indexes = [GinIndex(fields=['tags'])]


class NamedCommon(BasicCommon):
    name = CharField(max_length=128, help_text='Human-readable name of this object.', null=False)
    description = CharField(max_length=255, help_text='A longer description of this object.', blank=True, default="")

    def __str__(self):
        return self.name

    class Meta:
        abstract = True


class NamedCommonPK(NamedCommon):
    name = CharField(max_length=128, help_text='Human-readable name of this object.', null=False, primary_key=True)

    class Meta:
        abstract = True


class AbstractChoice(Model):
    """
    Abstract class for all derived 'choices' models.
    We define a 'choice' as an item that you can pick from a predefined list.
    In the derived classes, we use an enum.Enum to define such a predefined list here in code.
    All values of the enums are then put automagically into the database in the populate module, which
    is/can_be/should_be called in the last migration step to populate the database with inital values
    for our 'static choices'.

    Design decision: Django also provides the 'choices' property on fields which sort of limits the number of choices
    one can make, and which sort of does some validation. In our opinion the validation is done in the wrong place, and
    no data consistency is enforced.
    So, we decided to follow Django's own hint, see https://docs.djangoproject.com/en/2.0/ref/models/fields/#choices
    "you’re probably better off using a proper database table with a ForeignKey"

    You can find the derived AbstractChoice classes being used as ForeignKey in other models, thus enforcing data
    consistency at database level.
    """
    value = CharField(max_length=128, editable=True, null=False, blank=False, unique=True, primary_key=True)

    class Meta:
        abstract = True

    def __str__(self):
        return self.value


class TemplatePurpose(AbstractChoice):
    """Defines the model and predefined list of possible TemplatePurpose's of a template.
    The items in the Choices class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        PRODUCTION = "production"
        SCIENTIFIC_COMMISSIONING = "scientific_commissioning"
        TECHNICAL_COMMISSIONING = "technical_commissioning"
        SYSTEM_HEALTH = "system_health"


class TemplateState(AbstractChoice):
    """Defines the model and predefined list of possible TemplateState's of a template.
    The items in the Choices class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        DEVELOPMENT = "development"
        ACTIVE = "active"
        LEGACY = "legacy"
        OBSOLETE = "obsolete"


class AbstractTemplate(NamedCommon):
    version = IntegerField(editable=False, null=False, help_text='Version of this template (with respect to other templates of the same name)')
    state = ForeignKey('TemplateState', null=False, default=TemplateState.Choices.DEVELOPMENT.value, on_delete=PROTECT, help_text='The state of this template: Development: Strategy under construction. - Active: Available for observing. - Legacy: not offered, but still working. Obsolete: broken or succeeded by newer definitions.')
    purpose = ForeignKey('TemplatePurpose', null=False, default=TemplatePurpose.Choices.TECHNICAL_COMMISSIONING.value, on_delete=PROTECT, help_text='The purpose of this template: Production: offered to the user for cycle and DDT observations. - Scientific Commissioning: scientific/astronomical commissioning. - Technical Commissioning: technical/system commissioning. - System Health: system health monitoring.')

    class Meta:
        abstract = True
        constraints = [UniqueConstraint(fields=['name', 'version'], name='%(class)s_unique_name_version')]

    @classmethod
    def get_latest(cls, name: str):
        '''get the latest non-obsolete version of this template'''
        return cls.objects.exclude(state__value=TemplateState.Choices.OBSOLETE.value).filter(name=name).order_by('-version').first()

    @classmethod
    def get_version_or_latest(cls, name: str, version: Union[int, str, None]=None):
        '''get the non-obsolete template with the given specific version, or the latest non-obsolete version if None or 'latest' '''
        try:
            if version is None or version=='latest':
                return cls.get_latest(name=name)

            if isinstance(version, str):
                version = int(version)

            qs = cls.objects.exclude(state__value=TemplateState.Choices.OBSOLETE.value)
            if isinstance(version, int):
                qs = qs.filter(version=version)
            return qs.get(name=name)
        except (cls.DoesNotExist, cls.MultipleObjectsReturned, ValueError):
            latest = cls.get_latest(name=name)
            if latest:
                logger.warning("Could not get %s name='%s' version=%s, returning latest version=%s", cls.__name__, name, version, latest.version)
                return latest
        # no template with given version nor any latest...
        return None


    @property
    def _need_to_auto_increment_version_number(self) -> bool:
        if self.pk is None:
            # this is a new instance, so we need a new version number
            return True

        # this is an existing instance which is being updated. Check if it is being used.
        # if used, then this template is used by others, so "editing"/updating is forbidden,
        # and we need a new version number
        return self.is_used

    def auto_set_version_number(self) -> Tuple[int, int]:
        '''An instance of a subclass of this class, for example a template, cannot/shouldnot be updated if it is already being used.
        So, update the version number if the template is already used, else keep it.
        Returns a tuple of the previous and new version number (which can be equal if there was no need to update it)'''
        if self._need_to_auto_increment_version_number:
            self.pk = None # this forces the creation of a new instance
            same_name_obj = self.__class__.objects.filter(name=self.name).order_by('-version').first()
            old_version = self.version
            new_version = same_name_obj.version + 1 if same_name_obj is not None else 1
            self.version = new_version
            return (old_version, new_version)

        return (self.version, self.version)



class AbstractSchemaTemplate(AbstractTemplate):
    schema = JSONField(help_text='Schema for the configurable parameters needed to use this template.')

    class Meta(AbstractTemplate.Meta):
        abstract = True

    @cached_property
    def ref_resolved_schema(self) -> dict:
        '''get the schema with all remote references fully resolved and replaced with references to local definitions, or None if not possible'''
        try:
            return resolved_remote_refs(self.schema)
        except Exception as e:
            logger.error("Could not resolve all references: %s", e)
            return None

    def get_default_json_document_for_schema(self) -> dict:
        '''get a json document object (dict) which complies with this template's schema and with all the defaults filled in.'''
        default_json_doc = get_default_json_object_for_schema(self.schema)
        # make sure we point to the ref_resolved version of the schema
        default_json_doc['$schema'] = ref_resolved_url(default_json_doc['$schema'])
        return default_json_doc

    def add_defaults_to_json_object_for_schema(self, json_object: dict) -> dict:
        '''add any missing default propery to the given json document object (dict).'''
        return add_defaults_to_json_object_for_schema(json_object, self.schema)

    def annotate_and_validate_json_schema(self):
        '''annotate this schema with the template's name, description and version,
           check for the required properties '$id', '$schema', 'title', 'description',
           and validate the schema itself against its own schema.'''
        try:
            if isinstance(self.schema, str):
                self.schema = json.loads(self.schema)
        except json.JSONDecodeError as e:
            raise SchemaValidationException(str(e))

        # sync up the template properties with the schema
        self.schema['title'] = self.name
        self.schema['description'] = self.description
        self.schema['version'] = self.version

        # enforce no additional properties at document root level, other than specified and $schema.
        self.schema['additionalProperties'] = False
        self.schema['patternProperties']= {"^[$]schema$": {}}

        # check for missing properties
        missing_properties = [property for property in ['$id', '$schema', 'title', 'description'] if property not in self.schema]
        if missing_properties:
            raise SchemaValidationException("Missing required properties '%s' for %s name='%s' version=%s in schema:\n%s" % (', '.join(missing_properties),
                                                                                                                             self.__class__.__name__, self.name, self.version,
                                                                                                                             json.dumps(self.schema, indent=2)))

        # check for valid url's
        invalid_url_properties = [property for property in ['$id', '$schema'] if not self.schema[property].startswith('http')]
        if invalid_url_properties:
            raise SchemaValidationException("Properties '%s' should contain a valid URL's for %s name='%s' version=%s in schema:\n%s" % (', '.join(invalid_url_properties),
                                                                                                                                         self.__class__.__name__, self.name, self.version,
                                                                                                                                         json.dumps(self.schema, indent=2)))

        try:
            # construct full url for $id of this schema
            path = reverse_url('get_template_json_schema', kwargs={'template': self._meta.model_name,
                                                                  'name': self.name,
                                                                  'version': self.version}).rstrip('/')
            parts = self.schema['$id'].split('/')
            scheme_host = '%s//%s' % (parts[0], parts[2])
            id_url = '%s%s#' % (scheme_host, path)
            self.schema['$id'] = id_url
        except Exception as e:
            logger.error("Could not override schema $id with auto-generated url: %s", e)

        # check on forbidden remote self references
        raise_on_self_refs(self.schema)

        # this template's schema has a schema of its own (usually the draft-06 meta schema). Validate it.
        validate_json_against_its_schema(self.schema)

    def validate_document(self, json_doc: typing.Union[str, dict], raise_on_obsolete=True):
        '''validate the given json_doc against this template's schema and consistency/sanity rules
           If no exception is thrown, then the given json_doc validates.
           Optionally raise when template is obsolete. This check should only happen when the document is basis for an
           action in the system, but not when a (legacy) document is merely read.
           :raises ValidationException if the json_doc does not validate against this template's schema, or if some rule is violated, or the template is obsolete'''
        if raise_on_obsolete and self.state.value == TemplateState.Choices.OBSOLETE.value:
            raise ObsoleteValidationException("Cannot validate document against obsolete %s id=%s name='%ss version=%s" % (self.__class__.__name__, self.id, self.name, self.version))
        self.validate_document_against_schema(json_doc)
        self.validate_document_against_rules(json_doc)

    def validate_document_against_schema(self, json_doc: typing.Union[str, dict]):
        '''validate the given json_doc against this template's schema
           If no exception is thrown, then the given json_doc validates.
           :raises SchemaValidationException if the json_doc does not validate against this template's schema'''
        validate_json_against_schema(json_doc, self.schema)

    def validate_document_against_rules(self, json_doc: typing.Union[str, dict]):
        '''validate the given json_doc against this template's schema and consistency/sanity rules
           If no exception is thrown, then the given json_doc validates.
           :raises RuleValidationException if the json_doc does not validate against this template's schema'''
        # there are no shared/common rules for this abstract template class.
        # override and implement for specific rules in each subclass if needed.
        pass

    @property
    def _need_to_auto_increment_version_number(self) -> bool:
        '''For any Template instance, we only update the version number if the schema document itself changed'''
        if self.pk is None:
            # this is a new instance, so we need a new version number, if and only if no version was given
            return self.version is None

        if self._schema_changed:
            # the schema value itself changed, so do auto increment version number if this template instance is used
            return self.is_used or self._schema_used_in_other_schema_templates

        # the schema(value) did not change, so do NOT auto_increment the version number
        return False

    @property
    def _schema_changed(self) -> bool:
        '''check if the schema document in this model instance changed compared to the currectly stored database version'''
        # check current contents of template document in the database...
        prev_schema_value = self.__class__.objects.values('schema').get(pk=self.pk)['schema']
        return prev_schema_value != self.schema

    @property
    def _schema_used_in_other_schema_templates(self) -> bool:
        '''check if the schema document in this model instance is used in any other non-obsolete template'''
        schema_id = self.schema.get('$id')
        if schema_id:
            schema_id = schema_id.rstrip('%23').rstrip('#').rstrip('/')
            for template_class in [sc for sc in AbstractSchemaTemplate.__subclasses__() if sc != self.__class__]:
                for template_instance in template_class.objects.exclude(state__value=TemplateState.Choices.OBSOLETE.value).all():
                    template_schema_str = json.dumps(template_instance.schema)
                    if schema_id in template_schema_str:
                        return True # early exit at first found usage
        return False

    def get_referencing_schema_templates(self) -> Collection['AbstractSchemaTemplate']:
        '''Get all other non_obsolete schematemplates which are referencing this one.'''
        referencing_schema_templates = []
        schema_id = self.schema.get('$id')
        if schema_id:
            schema_id = schema_id.rstrip('%23').rstrip('#').rstrip('/')
            for template_class in [sc for sc in AbstractSchemaTemplate.__subclasses__() if sc != self.__class__]:
                for template_instance in template_class.objects.exclude(state__value=TemplateState.Choices.OBSOLETE.value).all():
                    template_schema_str = json.dumps(template_instance.schema)
                    if schema_id in template_schema_str:
                        referencing_schema_templates.append(template_instance)
        return referencing_schema_templates

    def get_using_non_obsolete_strategy_templates(self) -> Collection['AbstractStrategyTemplate']:
        '''Get the non-obsolete observing strategy template which are using this schema template'''
        using_strategy_templates = []
        for template_class in [sc for sc in AbstractStrategyTemplate.__subclasses__() if sc != self.__class__]:
            for template_instance in template_class.objects.exclude(state__value=TemplateState.Choices.OBSOLETE.value).all():
                try:
                    if template_instance.is_using_schema_template(self):
                        using_strategy_templates.append(template_instance)
                except NotImplementedError as e:
                    logger.error(e)
        return using_strategy_templates

    def _aggregate_over_all_referencing_schema_templates(self, method_name: str) -> QuerySet:
        '''convenience method to apply the method with the given method_name to all referencing schema templates, and aggregate the results in a union QuerySet'''
        result = None
        referencing_schema_templates = self.get_referencing_schema_templates()
        for referencing_schema_template in referencing_schema_templates:
            method = getattr(referencing_schema_template, method_name)
            sub_result = method()
            if result is None:
                result = sub_result
            else:
                result = result | sub_result
        return result

    def get_unfinished_using_task_blueprints(self) -> QuerySet:
        '''Get the unfinished task blueprints using this template (state schedulable and beyond)'''
        return self._aggregate_over_all_referencing_schema_templates('get_unfinished_using_task_blueprints')

    @property
    def is_used_in_unfinished_tasks_blueprints(self) -> bool:
        '''Is this template used by any task which still needs to run (state schedulable and beyond)'''
        qs = self.get_unfinished_using_task_blueprints()
        return False if qs is None else qs.exists()

    def get_non_blueprinted_using_task_drafts(self) -> QuerySet:
        '''Get the non-blueprinted task drafts using this template'''
        return self._aggregate_over_all_referencing_schema_templates('get_non_blueprinted_using_task_drafts')

    @property
    def is_used_in_non_blueprinted_using_task_drafts(self) -> bool:
        '''Is this template used by any taskdraft which still needs to be blueprinted?'''
        qs = self.get_non_blueprinted_using_task_drafts()
        return False if qs is None else qs.exists()

    def get_unfinished_using_reservations(self) -> QuerySet:
        '''Get the unfinished reservations using this template'''
        return self._aggregate_over_all_referencing_schema_templates('get_unfinished_using_reservations')

    @property
    def is_used_in_unfinished_reservations(self) -> bool:
        '''Is this template used by any unfinished reservations'''
        qs = self.get_unfinished_using_reservations()
        return False if qs is None else qs.exists()

    @property
    def is_used_in_non_obsolete_strategy_templates(self) -> bool:
        '''Is this schema template used in any non-obsolete observing strategy template'''
        return len(self.get_using_non_obsolete_strategy_templates()) > 0


    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if self.state.value == TemplateState.Choices.OBSOLETE.value and not self._state.adding and self.__class__.objects.get(id=self.id).state.value != TemplateState.Choices.OBSOLETE.value:
            # only allow changing to OBSOLETE if not used.
            if self.is_used_in_unfinished_tasks_blueprints or self.is_used_in_non_blueprinted_using_task_drafts or \
               self.is_used_in_unfinished_reservations or self.is_used_in_non_obsolete_strategy_templates:
                raise ObsoleteValidationException("%s id=%s name='%s' version=%s state=%s is still being used, and may not be set to state=obsolete" % (self.id,
                                                                                                                                                        self.__class__.__name__,
                                                                                                                                                        self.name,
                                                                                                                                                        self.version,
                                                                                                                                                        self.state.value))

        old_version, new_version = self.auto_set_version_number()
        self.annotate_and_validate_json_schema()

        super().save(force_insert or self.pk is None, force_update, using, update_fields)
        remove_from_schema_cache(self.schema.get('$id', '').rstrip('#').rstrip('/'))
        self._propagate_new_version_to_using_templates(old_version, new_version)

        if old_version is not None and old_version != new_version:
            old_template = self.__class__.objects.get(name=self.name, version=old_version)
            if old_template.state.value not in (TemplateState.Choices.OBSOLETE.value, TemplateState.Choices.LEGACY.value):
                logger.info("setting state to 'legacy' for %s name='%s' version=%s because the schema and version number were auto-updated", self.__class__.__name__, self.name, old_version)
                old_template.state = TemplateState.objects.get(value=TemplateState.Choices.LEGACY.value)
                old_template.save()

    def _propagate_new_version_to_using_templates(self, old_version: int, new_version: int):
        '''check if the version of this schematemplate was (auto)incremented,
        and propagate the use (references) of this schema into all other using schema templates'''
        if old_version is not None and old_version != new_version:
            # yep, version was auto incremented, so update the using templates
            # do this in a transaction, such that if any one fails, all are reverted, keeping the references consistent.
            with transaction.atomic():
                logger.info("%s name='%s' was updated from version=%s to version=%s", self.__class__.__name__, self.name, old_version, new_version)

                old_template = self.__class__.objects.get(name=self.name, version=old_version)
                new_template = self.__class__.objects.get(name=self.name, version=new_version)

                # strip any non-relevant '/' or '#' characters at the tail of the ids
                old_schema_id = old_template.schema.get('$id', '').rstrip('%23').rstrip('#').rstrip('/')
                new_schema_id = new_template.schema.get('$id', '').rstrip('%23').rstrip('#').rstrip('/')

                for template_subclass in AbstractSchemaTemplate.__subclasses__():
                    for template_instance in template_subclass.objects.exclude(state__value__in=(TemplateState.Choices.OBSOLETE.value, TemplateState.Choices.LEGACY.value)).all():
                        ref_ids = get_refs(template_instance.schema, only_ids=True)
                        if any(ref_id == old_schema_id for ref_id in ref_ids):
                            logger.info("updating reference(s) in %s name='%s' version=%s from %s to %s",
                                        template_subclass.__name__, template_instance.name, template_instance.version, old_schema_id, new_schema_id)
                            template_instance.schema = json.loads(json.dumps(template_instance.schema).replace(old_schema_id, new_schema_id))
                            template_instance.save()

                # then apply it to the observing strategy templates as well
                from . import SchedulingUnitObservingStrategyTemplate # import here to prevent circular imports
                SchedulingUnitObservingStrategyTemplate.adapt_to_updated_schema_template(self, old_version, new_version)


class AbstractStrategyTemplate(AbstractTemplate):
    '''
    A (subclass of a) BaseStrategyTemplate is a template in the sense that contains a 'predefined' JSON document. It is not a JSON-schema, but a document!
    '''
    template = JSONField(null=False, help_text='JSON-data document compliant with the JSON-schema in the referenced template. '
                                               'This strategy template is like a predefined recipe with all the correct settings, and defines which parameters the user can alter.')

    class Meta(AbstractTemplate.Meta):
        abstract = True

    @property
    def _need_to_auto_increment_version_number(self) -> bool:
        '''For any StrategyTemplate instance, we only update the version number if the template document itself changed'''
        if self.pk is None:
            # this is a new instance, so we need a new version number, if and only if no version was given
            return self.version is None

        # check if it's changed
        if self._is_template_doc_different_from_db_version:
            # the template value itself changed, so do auto increment version number if this template instance is used
            return self.is_used

        # the template(value) did not change, so do NOT auto_increment the version number
        return False

    @property
    def _is_template_doc_different_from_db_version(self) -> bool:
        '''Check if the template document itself changed compared to the version in the database (if present)'''
        if self.pk is None:
            # this is a new instance, the document is different from null
            return True

        # check current contents of template document in the database...
        prev_template_value = self.__class__.objects.values('template').get(pk=self.pk)['template']
        return prev_template_value != self.template

    @property
    def is_using_schema_template(self, schema_template: AbstractSchemaTemplate) -> bool:
        raise NotImplementedError()

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if self.pk is not None and \
           self.state.value in (TemplateState.Choices.OBSOLETE.value, TemplateState.Choices.LEGACY.value) and \
           self._is_template_doc_different_from_db_version:
            # do not allow obsolete/legacy template documents to be updated
            raise ObsoleteTemplateException("It is not allowed to save an updated template(doc) for a strategy template with state=%s, id=%s, name='%s' version=%s" % (
                self.state.value, self.id, self.name, self.version))

        self.auto_set_version_number()
        super().save(force_insert, force_update, using, update_fields)


class Tags(Model):
    # todo: figure out how to keep this in sync with tags columns (->BasicCommon)
    # todo: Or remove this altogether without keeping track of tags?
    title = CharField(max_length=128)
    description = CharField(max_length=255)


class TemplateSchemaMixin():
    '''The TemplateSchemaMixin class can be mixed in to models which do validate and add defaults to json documents given a json-schema.
    It uses an internal cache with a max age to minimize the number of requests to schema's, subschema's or referenced (sub)schema's.'''
    def validate_json_against_this_templates_schema(self, document_attr:str, template_attr:str) -> None:
        '''
        annotate, validate and add defaults to the JSON document in the model instance using the schema of the given template.
        '''
        try:
            # fetch the actual JSON document and template-model-instance
            document = getattr(self, document_attr, None)
            template = getattr(self, template_attr, None)

            if document is not None and template is not None:
                if isinstance(document, str):
                    document = json.loads(document)

                validate_json_against_schema(document, template.schema)
        except json.JSONDecodeError as e:
            raise SchemaValidationException("Invalid JSON.\nError: %s \ndata: %s" % (str(e), document))
        except jsonschema.ValidationError as e:
            raise SchemaValidationException(str(e))

    def validate_doc_using_template(self, document_attr:str, template_attr:str) -> None:
        '''
        validate the document in the self.<document_attr> against the schema in the template and against the rules.
        If no exception is thrown, then the given json_doc validates.
         :raises ValidationException if the json_doc does not validate against this template's schema'''
        # fetch the actual JSON document and template-model-instance
        document = getattr(self, document_attr, None)
        template = getattr(self, template_attr, None)
        if document is not None and template is not None:
            template.validate_document(document)

    def annotate_doc_using_template(self, document_attr:str, template_attr:str, add_defaults: bool=True) -> None:
        '''
        annotate the JSON document with required properties ($schema)
        '''
        # fetch the actual JSON document and template-model-instance
        document = getattr(self, document_attr, None)
        template = getattr(self, template_attr, None)

        if document is not None and template is not None:
            try:
                if isinstance(document, str):
                    document = json.loads(document)

                # always annotate the json data document with a $schema URI to the schema that it is based on.
                # this enables all users using this document (inside or outside of TMSS) to do their own validation and usage of editors which use the schema as UI template
                # by adding the '/ref_resolved' to the url, the read-only ref_resolved version of the schema is loaded (which saves the client many additional lookups)
                document['$schema'] = ref_resolved_url(template.schema['$id'])
            except (KeyError, TypeError, AttributeError) as e:
                raise SchemaValidationException("Cannot set $schema in json_doc to the schema's $id.\nError: %s \njson_doc: %s\nschema: %s" % (str(e), document, template.schema))
            except json.JSONDecodeError as e:
                raise SchemaValidationException("Invalid JSON.\nError: %s \ndata: %s" % (str(e), document))

            # update the model instance with the updated and validated document
            setattr(self, document_attr, document)

    def add_defaults_to_doc_using_template(self, document_attr:str, template_attr:str) -> None:
        '''
        add defaults to the JSON document in the model instance using the schema of the given template.
        '''
        try:
            # fetch the actual JSON document and template-model-instance
            document = getattr(self, document_attr, None)
            template = getattr(self, template_attr, None)

            if document is not None and template is not None:
                if isinstance(document, str):
                    document = json.loads(document)

                # add defaults for missing properies, and validate on the fly if requested
                document = validate_json_against_schema(document, template.schema, add_defaults=True)

                # update the model instance with the updated and validated document
                setattr(self, document_attr, document)
        except json.JSONDecodeError as e:
            raise SchemaValidationException("Invalid JSON.\nError: %s \ndata: %s" % (str(e), document))
        except jsonschema.ValidationError as e:
            raise SchemaValidationException(str(e))