"""
This module 'populate' defines methods to populate the database with predefined ('static') data,
according to the proposed Django way: https://docs.djangoproject.com/en/2.0/topics/migrations/#data-migrations

import this module in your empty migration step file, and add the following migration:

from ..populate import *

class Migration(migrations.Migration):

    dependencies = [ <the dependency is automatically inserted here> ]

    operations = [ migrations.RunPython(populate_choices) ]

"""

import logging
logger = logging.getLogger(__name__)

import inspect
import re
from datetime import timezone, datetime, date
from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.tmssapp import viewsets
import os
from lofar.sas.tmss.tmss.settings import isViewflowEnabled
if isViewflowEnabled():
    from lofar.sas.tmss.tmss.workflowapp import viewsets as workflow_viewsets
from lofar.sas.tmss.tmss.tmssapp.models.common import *
from lofar.sas.tmss.tmss.tmssapp.models.specification import *
from lofar.sas.tmss.tmss.tmssapp.models.scheduling import *
from lofar.sas.tmss.tmss.tmssapp.models.permissions import *
from lofar.sas.tmss.tmss.tmssapp.conversions import timestamps_and_stations_to_sun_rise_and_set, get_all_stations
from lofar.common import isTestEnvironment, isDevelopmentEnvironment
from concurrent.futures import ThreadPoolExecutor
from django.contrib.auth.models import Group, Permission
from django.contrib.auth import get_user_model
User = get_user_model()
from django.contrib.contenttypes.models import ContentType
from django.db.utils import IntegrityError
from django.db import transaction
from lofar.common import isProductionEnvironment

working_dir = os.path.dirname(os.path.abspath(__file__))

permission_groups = {'system_setting': [viewsets.AntennaSetViewSet,
                                        viewsets.ClusterViewSet,
                                        viewsets.CommonSchemaTemplateViewSet,
                                        viewsets.DataformatViewSet,
                                        viewsets.DataproductFeedbackTemplateViewSet,
                                        viewsets.DataproductSpecificationsTemplateViewSet,
                                        viewsets.DatatypeViewSet,
                                        viewsets.FilesystemViewSet,
                                        viewsets.HashAlgorithmViewSet,
                                        viewsets.IOTypeViewSet,
                                        viewsets.PeriodCategoryViewSet,
                                        viewsets.PriorityQueueTypeViewSet,
                                        viewsets.ProjectCategoryViewSet,
                                        viewsets.ProjectPermissionViewSet,
                                        viewsets.ProjectRoleViewSet,
                                        viewsets.QuantityViewSet,
                                        viewsets.ReservationTemplateViewSet,
                                        viewsets.ResourceTypeViewSet,
                                        viewsets.RoleViewSet,
                                        viewsets.SAPViewSet,                          # assuming JD means 'system setting' by 'system table'
                                        viewsets.SAPTemplateViewSet,
                                        viewsets.SchedulingConstraintsTemplateViewSet,
                                        viewsets.SchedulingRelationPlacementViewSet,  # assuming JD means 'system setting' by 'system table'
                                        viewsets.SchedulingUnitTemplateViewSet,
                                        viewsets.StationTypeViewSet,
                                        viewsets.SubtaskAllowedStateTransitionsViewSet,
                                        viewsets.SubtaskFeedbackTemplateViewSet,
                                        viewsets.SubtaskTemplateViewSet,
                                        viewsets.SubtaskTypeViewSet,
                                        viewsets.SystemEventTemplateViewSet,
                                        viewsets.TaskConnectorTypeViewSet,
                                        viewsets.TaskTemplateViewSet,
                                        viewsets.TaskTypeViewSet,
                                        viewsets.UserViewSet,                         # assuming JD means 'system setting' by 'system table'
                                        viewsets.SystemEventStatusViewSet,
                                        viewsets.SystemEventTypeViewSet,
                                        viewsets.SystemEventSubtypeViewSet,
                                        viewsets.SystemEventSeverityViewSet,
                                        viewsets.ProjectStateViewSet,
                                        viewsets.TemplateStateViewSet,
                                        viewsets.TemplatePurposeViewSet],
                     'system_feedback': [viewsets.DataproductViewSet,
                                         viewsets.DataproductArchiveInfoViewSet,
                                         viewsets.DataproductHashViewSet,
                                         viewsets.DataproductTransformViewSet,
                                         viewsets.SIPidentifierViewSet,
                                         viewsets.SubtaskInputViewSet,
                                         viewsets.SubtaskOutputViewSet,
                                         viewsets.SubtaskStateViewSet,
                                         viewsets.SubtaskStateLogViewSet],
                     'readable_all_system_roles_writable_operator_support': [  # we need this, because system setting is read-only, and other groups are not readable by all system roles (-> scientist).
                                         viewsets.CycleViewSet,
                                         viewsets.ReservationViewSet,   # todo: moved from cycle_administration, JDM please verify
                                         viewsets.SystemEventViewSet,
                                         viewsets.CycleQuotaViewSet,
                                         viewsets.SchedulingUnitObservingStrategyTemplateViewSet],
                     'cycle_administration': [viewsets.ProjectViewSet,
                                              viewsets.ProjectQuotaViewSet,
                                              viewsets.ProjectQuotaArchiveLocationViewSet,
                                              viewsets.ReservationStrategyTemplateViewSet],
                     'project_administration': [viewsets.SchedulingUnitDraftViewSet,
                                                viewsets.SchedulingSetViewSet,
                                                viewsets.TagsViewSet,           # todo: I don't think this should be here. Tags are present on almost every model we have. They are also unrelated to particular projects.
                                                viewsets.TaskDraftViewSet,
                                                viewsets.TaskRelationBlueprintViewSet,
                                                viewsets.TaskRelationDraftViewSet,
                                                viewsets.TaskRelationSelectionTemplateViewSet,
                                                viewsets.TaskSchedulingRelationDraftViewSet,
                                                viewsets.SubtaskViewSet],       # todo: moved from 'system setting', JDM please verify
                     'blueprint_alteration': [viewsets.SchedulingUnitBlueprintViewSet,
                                              viewsets.TaskSchedulingRelationBlueprintViewSet],
                     'blueprint_fixed': [viewsets.TaskBlueprintViewSet],
                     'operational': [viewsets.StationTimelineViewSet,
                                     viewsets.SettingViewSet,
                                     viewsets.SystemSettingFlagViewSet,
                                     viewsets.SubsystemViewSet],
                     'workflow': [],
                     'workflow_piverification': []
                     }

if isViewflowEnabled():
    permission_groups['workflow'] = [workflow_viewsets.SchedulingUnitTaskViewSet,
                                     workflow_viewsets.SchedulingUnitProcessViewSet,
                                     workflow_viewsets.SchedulingUnitTaskAssignViewSet,
                                     workflow_viewsets.SchedulingUnitTaskUnassignViewSet,
                                     workflow_viewsets.QAReportingTOViewSet,
                                     workflow_viewsets.QAReportingSOSViewSet,
                                     workflow_viewsets.DecideAcceptanceViewSet]
    permission_groups['workflow_piverification'] = [workflow_viewsets.PIVerificationViewSet,
                                                    workflow_viewsets.SchedulingUnitGetActiveTasksViewSet,
                                                    workflow_viewsets.SchedulingUnitTaskExecuteViewSet]


def populate_choices(apps, schema_editor):
    '''
    populate each 'choice' table in the database with the 'static' list of 'choice'.Choices for
    each 'choice'type in Role, Datatype, Dataformat
    :return: None
    '''
    choice_classes = [Role, IOType, Datatype, Dataformat,
                      SubtaskState, SubtaskType, StationType, HashAlgorithm, SchedulingRelationPlacement,
                      SystemSettingFlag, ProjectCategory, PeriodCategory, Quantity, TaskType, ProjectRole,
                      PriorityQueueType, ProjectState]

    # upload choices in parallel
    with ThreadPoolExecutor() as executor:
        executor.map(lambda choice_class: choice_class.objects.bulk_create([choice_class(value=x.value) for x in choice_class.Choices]),
                     choice_classes)

def populate_system_event_choices(apps, schema_editor):
    '''
    populate each 'choice' table in the database with the 'static' list of 'choice'.Choices for
    each SystemEvent-related 'choice'type that has been added in a later migration than the initial populate.
    :return: None
    '''
    choice_classes = [SystemEventType, SystemEventSubtype, SystemEventSeverity, SystemEventStatus]

    # upload choices in parallel
    with ThreadPoolExecutor() as executor:
        executor.map(lambda choice_class: choice_class.objects.bulk_create([choice_class(value=x.value) for x in choice_class.Choices]),
                     choice_classes)

def populate_task_schedulingunit_choices(apps, schema_editor):
    '''
    populate each 'choice' table in the database with the 'static' list of 'choice'.Choices for
    each SystemEvent-related 'choice'type that has been added in a later migration than the initial populate.
    :return: None
    '''
    for choice in TaskStatus.Choices:
        TaskStatus.objects.get_or_create(value=choice.value) #get_or_create so existing rows in existing databases are not overwritten or duplicated

    for choice in SchedulingUnitStatus.Choices:
        SchedulingUnitStatus.objects.get_or_create(value=choice.value) #get_or_create so existing rows in existing databases are not overwritten or duplicated

def populate_choices_template_state_purpose(apps, schema_editor):
    '''
    populate additional 'choice' tables for  in the database with the 'static' list of 'choice'.
    '''
    for choice in TemplateState.Choices:
        TemplateState.objects.get_or_create(value=choice.value) #get_or_create so existing rows in existing databases are not overwritten or duplicated

    for choice in TemplatePurpose.Choices:
        TemplatePurpose.objects.get_or_create(value=choice.value) #get_or_create so existing rows in existing databases are not overwritten or duplicated

def populate_choices_subsystem_status(apps, schema_editor):
    '''
    populate additional 'choice' tables for SubsystemStatus in the database with the 'static' list of 'choice'.
    '''
    for choice in SubsystemStatus.Choices:
        SubsystemStatus.objects.get_or_create(value=choice.value) #get_or_create so existing rows in existing databases are not overwritten or duplicated

    # populate the 'scheduler' subsystem
    Subsystem.objects.get_or_create(name="scheduler",
                                    description="The scheduler subsystem takes care of scheduling all scheduling_units taking into account all their constraints and station availability",
                                    status=models.SubsystemStatus.objects.get(value=models.SubsystemStatus.Choices.IDLE.value))


def populate_subtask_allowed_state_transitions(apps, schema_editor):
    '''populate the SubtaskAllowedStateTransitions table with the allowed state transitions as defined by the design in https://support.astron.nl/confluence/display/TMSS/Subtask+State+Machine'''
    DEFINING = SubtaskState.objects.get(value=SubtaskState.Choices.DEFINING.value)
    DEFINED = SubtaskState.objects.get(value=SubtaskState.Choices.DEFINED.value)
    SCHEDULING = SubtaskState.objects.get(value=SubtaskState.Choices.SCHEDULING.value)
    SCHEDULED = SubtaskState.objects.get(value=SubtaskState.Choices.SCHEDULED.value)
    UNSCHEDULING = SubtaskState.objects.get(value=SubtaskState.Choices.UNSCHEDULING.value)
    QUEUEING = SubtaskState.objects.get(value=SubtaskState.Choices.QUEUEING.value)
    QUEUED = SubtaskState.objects.get(value=SubtaskState.Choices.QUEUED.value)
    STARTING = SubtaskState.objects.get(value=SubtaskState.Choices.STARTING.value)
    STARTED = SubtaskState.objects.get(value=SubtaskState.Choices.STARTED.value)
    FINISHING = SubtaskState.objects.get(value=SubtaskState.Choices.FINISHING.value)
    FINISHED = SubtaskState.objects.get(value=SubtaskState.Choices.FINISHED.value)
    CANCELLING = SubtaskState.objects.get(value=SubtaskState.Choices.CANCELLING.value)
    CANCELLED = SubtaskState.objects.get(value=SubtaskState.Choices.CANCELLED.value)
    ERROR = SubtaskState.objects.get(value=SubtaskState.Choices.ERROR.value)
    UNSCHEDULABLE = SubtaskState.objects.get(value=SubtaskState.Choices.UNSCHEDULABLE.value)

    SubtaskAllowedStateTransitions.objects.bulk_create([
        SubtaskAllowedStateTransitions(old_state=None, new_state=DEFINING),
        SubtaskAllowedStateTransitions(old_state=DEFINING, new_state=DEFINED),
        SubtaskAllowedStateTransitions(old_state=DEFINED, new_state=SCHEDULING),
        SubtaskAllowedStateTransitions(old_state=SCHEDULING, new_state=SCHEDULED),
        SubtaskAllowedStateTransitions(old_state=SCHEDULING, new_state=UNSCHEDULABLE),
        SubtaskAllowedStateTransitions(old_state=SCHEDULED, new_state=STARTING), # this is an odd one, as most (all?) subtasks are queued before execution...
        SubtaskAllowedStateTransitions(old_state=SCHEDULED, new_state=QUEUEING),
        SubtaskAllowedStateTransitions(old_state=SCHEDULED, new_state=UNSCHEDULING),
        SubtaskAllowedStateTransitions(old_state=UNSCHEDULING, new_state=DEFINED),
        SubtaskAllowedStateTransitions(old_state=UNSCHEDULING, new_state=CANCELLING), # directly after unscheduling we want to be able to go to cancelling and not trigger any schedulers on the defined state
        SubtaskAllowedStateTransitions(old_state=QUEUEING, new_state=QUEUED),
        SubtaskAllowedStateTransitions(old_state=QUEUED, new_state=STARTING),
        SubtaskAllowedStateTransitions(old_state=STARTING, new_state=STARTED),
        SubtaskAllowedStateTransitions(old_state=STARTED, new_state=FINISHING),
        SubtaskAllowedStateTransitions(old_state=FINISHING, new_state=FINISHED),
        SubtaskAllowedStateTransitions(old_state=CANCELLING, new_state=CANCELLED),

        SubtaskAllowedStateTransitions(old_state=DEFINING, new_state=ERROR),
        SubtaskAllowedStateTransitions(old_state=SCHEDULING, new_state=ERROR),
        SubtaskAllowedStateTransitions(old_state=UNSCHEDULING, new_state=ERROR),
        SubtaskAllowedStateTransitions(old_state=QUEUEING, new_state=ERROR),
        SubtaskAllowedStateTransitions(old_state=STARTING, new_state=ERROR),
        SubtaskAllowedStateTransitions(old_state=STARTED, new_state=ERROR),
        SubtaskAllowedStateTransitions(old_state=FINISHING, new_state=ERROR),
        SubtaskAllowedStateTransitions(old_state=CANCELLING, new_state=ERROR),

        SubtaskAllowedStateTransitions(old_state=DEFINED, new_state=CANCELLING),
        SubtaskAllowedStateTransitions(old_state=SCHEDULED, new_state=CANCELLING),
        SubtaskAllowedStateTransitions(old_state=QUEUED, new_state=CANCELLING),
        SubtaskAllowedStateTransitions(old_state=STARTED, new_state=CANCELLING),
        SubtaskAllowedStateTransitions(old_state=FINISHING, new_state=CANCELLING)  # in case feedback is not arriving, a user might choose to cancel a subtask even in this stage
        ])

def populate_subtask_allowed_state_transitions_extra(apps, schema_editor):
    '''populate the SubtaskAllowedStateTransitions table with additional allowed state transitions'''
    DEFINED = SubtaskState.objects.get(value=SubtaskState.Choices.DEFINED.value)
    CANCELLING = SubtaskState.objects.get(value=SubtaskState.Choices.CANCELLING.value)
    ERROR = SubtaskState.objects.get(value=SubtaskState.Choices.ERROR.value)
    UNSCHEDULABLE = SubtaskState.objects.get(value=SubtaskState.Choices.UNSCHEDULABLE.value)

    SubtaskAllowedStateTransitions.objects.bulk_create([
        SubtaskAllowedStateTransitions(old_state=DEFINED, new_state=UNSCHEDULABLE),
        SubtaskAllowedStateTransitions(old_state=DEFINED, new_state=ERROR),
        SubtaskAllowedStateTransitions(old_state=UNSCHEDULABLE, new_state=DEFINED),
        SubtaskAllowedStateTransitions(old_state=UNSCHEDULABLE, new_state=CANCELLING),
        SubtaskAllowedStateTransitions(old_state=UNSCHEDULABLE, new_state=ERROR)
        ])

def populate_subtask_allowed_state_transitions_extra2(apps, schema_editor):
    '''populate the SubtaskAllowedStateTransitions table with additional allowed state transitions'''
    SubtaskAllowedStateTransitions.objects.create(old_state=SubtaskState.objects.get(value=SubtaskState.Choices.UNSCHEDULING.value),
                                                  new_state=SubtaskState.objects.get(value=SubtaskState.Choices.UNSCHEDULABLE.value))

def populate_on_hold_status_and_transitions(apps, schema_editor):
    '''populate the new ON_HOLD status for subtask and task, add allowed state transitions'''
    Task_ON_HOLD, created = TaskStatus.objects.get_or_create(value=TaskStatus.Choices.ON_HOLD.value)
    ON_HOLD, created = SubtaskState.objects.get_or_create(value=SubtaskState.Choices.ON_HOLD.value)
    QUEUEING = SubtaskState.objects.get(value=SubtaskState.Choices.QUEUEING.value)
    RESTARTING, created = SubtaskState.objects.get_or_create(value=SubtaskState.Choices.RESTARTING.value)
    STARTING = SubtaskState.objects.get(value=SubtaskState.Choices.STARTING.value)
    STARTED = SubtaskState.objects.get(value=SubtaskState.Choices.STARTED.value)
    CANCELLING = SubtaskState.objects.get(value=SubtaskState.Choices.CANCELLING.value)
    ERROR = SubtaskState.objects.get(value=SubtaskState.Choices.ERROR.value)

    SubtaskAllowedStateTransitions.objects.bulk_create([
        SubtaskAllowedStateTransitions(old_state=STARTED, new_state=ON_HOLD),
        # one can set the status of an ON_HOLD subtask to RESTARTING...
        SubtaskAllowedStateTransitions(old_state=ON_HOLD, new_state=RESTARTING),
        # ... and then the subtask's controller picks up the RESTARTING event to queue/start the subtask again.
        SubtaskAllowedStateTransitions(old_state=RESTARTING, new_state=QUEUEING),
        SubtaskAllowedStateTransitions(old_state=RESTARTING, new_state=STARTING),
        SubtaskAllowedStateTransitions(old_state=ON_HOLD, new_state=CANCELLING),
        SubtaskAllowedStateTransitions(old_state=ON_HOLD, new_state=ERROR)
        ])

def populate_settings(apps, schema_editor):
    Setting.objects.create(name=SystemSettingFlag.objects.get(value=SystemSettingFlag.Choices.DYNAMIC_SCHEDULING_ENABLED.value), value=isDevelopmentEnvironment())
    Setting.objects.create(name=SystemSettingFlag.objects.get(value=SystemSettingFlag.Choices.FIXED_TIME_SCHEDULING_ENABLED.value), value=True)

def populate_test_data():
    """
    Create a Test Schedule Set to be able to refer to when Scheduling Unit Draft is created from a
    scheduling unit json
    :return:
    """
    try:
        # only add (with  expensive setup time) example data when developing/testing and we're not unittesting
        if isTestEnvironment() or isDevelopmentEnvironment():
            from lofar.sas.tmss.tmss.exceptions import TMSSException
            from lofar.sas.tmss.test.tmss_test_data_django_models import SchedulingUnitDraft_test_data
            from lofar.sas.tmss.tmss.tmssapp.tasks import create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft, update_task_graph_from_specifications_doc
            from lofar.sas.tmss.tmss.tmssapp.subtasks import schedule_subtask
            from lofar.common.json_utils import get_default_json_object_for_schema

            constraints_template = models.SchedulingConstraintsTemplate.get_latest(name="constraints")
            constraints_spec = get_default_json_object_for_schema(constraints_template.schema)

            next_fixed_time_start_time = datetime.utcnow() + timedelta(minutes=5)

            cycle = models.Cycle.objects.create(name="Cycle test",
                                                description="Cycle for TMSSTestEnvironment test data",
                                                start=datetime.utcnow()-timedelta(days=1),
                                                stop=datetime.utcnow()+timedelta(days=7))

            projects = models.Project.objects.order_by('-rank').all()
            for tmss_project in projects:
                tmss_project.project_state = models.ProjectState.objects.get(value=models.ProjectState.Choices.ACTIVE.value)
                tmss_project.cycles.set([cycle])
                tmss_project.save()

            for strategy_template in models.SchedulingUnitObservingStrategyTemplate.objects.filter(state__value=models.TemplateState.Choices.ACTIVE.value).order_by('name', 'version').all():
                try:
                    # the 'template' in the strategy_template is a predefined json-data blob which validates against the given scheduling_unit_template
                    # a user might 'upload' a partial json-data blob, so add all the known defaults
                    scheduling_unit_spec = add_defaults_to_json_object_for_schema(strategy_template.template, strategy_template.scheduling_unit_template.schema)
                    # if strategy_template.name not in ('IM HBA - 1 Beam', 'IM LBA Survey - 3 Beams', 'IM HBA LoTSS - 2 Beams'):
                    #     continue

                    if strategy_template.name == 'IM LBA Survey - 3 Beams':
                        scheduling_unit_spec['tasks']['Combined Observation']['specifications_doc']['duration'] = 4*60*60
                        scheduling_unit_spec['tasks']['Combined Observation']['specifications_doc']['station_configuration']['SAPs'][0]['digital_pointing'] = {"angle1": 5.348343326580123, "angle2": 0.7946493780603804, "direction_type": "J2000", "target": "P306+46"}
                        scheduling_unit_spec['tasks']['Combined Observation']['specifications_doc']['station_configuration']['SAPs'][1]['digital_pointing'] = {"angle1": 5.219407128089042, "angle2": 0.7936273908206015, "direction_type": "J2000", "target": "P299+45"}
                        scheduling_unit_spec['tasks']['Combined Observation']['specifications_doc']['station_configuration']['SAPs'][2]['digital_pointing'] = {"angle1": 5.090543651650128, "angle2": 0.7926335227743269, "direction_type": "J2000", "target": "P291+45"}
                        scheduling_unit_spec['tasks']['Combined Observation']['specifications_doc']['calibrator']['pointing'] = {"angle1": 4.8412365306941805, "angle2": 0.8507809072856333, "direction_type": "J2000", "target": "3c380"}
                        del scheduling_unit_spec['scheduling_constraints_doc']

                        # set some constraints, so the scheduler has something to chew on.
                        constraints_spec['scheduler'] = 'fixed_time'
                        constraints_spec['time'] = {'at': "2023-02-07T09:00:00.00000Z" }
                        constraints_spec['sky']['min_elevation'] = { "calibrator": 0.5235987755982988, "target": 0.8726646259971648 }
                        constraints_spec['sky']['min_distance'] = {"target": { "sun": 30*3.1415/180.0, "moon": 15*3.1415/180.0, "jupiter": 20*3.1415/180.0} }
                        constraints_spec['sky']['transit_offset'] = { "from": -5400, "to": 5400 }
                        constraints_spec['sky']['reference_pointing'] = {"enabled": True, "pointing": {"angle1": 5.2, "angle2": 0.793, "direction_type": "J2000", "target": "REF"}}
                    elif strategy_template.name == 'IM HBA LoTSS - 2 Beams':
                        scheduling_unit_spec['tasks']['Target Observation']['specifications_doc']['duration'] = 4*60*60
                        scheduling_unit_spec['tasks']['Target Observation']['specifications_doc']['station_configuration']['SAPs'][0]['digital_pointing'] = {"angle1": 5.348343326580123, "angle2": 0.7946493780603804, "direction_type": "J2000", "target": "P306+46"}
                        scheduling_unit_spec['tasks']['Target Observation']['specifications_doc']['station_configuration']['SAPs'][1]['digital_pointing'] = {"angle1": 5.219407128089042, "angle2": 0.7936273908206015, "direction_type": "J2000", "target": "P299+45"}
                        scheduling_unit_spec['tasks']['Calibrator Observation 1']['specifications_doc']['calibrator']['pointing'] = {"angle1": 4.8412365306941805, "angle2": 0.8507809072856333, "direction_type": "J2000", "target": "3c380"}
                        scheduling_unit_spec['tasks']['Calibrator Observation 2']['specifications_doc']['calibrator']['pointing'] = {"angle1": 4.8412365306941805, "angle2": 0.8507809072856333, "direction_type": "J2000", "target": "3c380"}
                        del scheduling_unit_spec['scheduling_constraints_doc']

                        # set some constraints, so the scheduler has something to chew on.
                        constraints_spec['scheduler'] = 'fixed_time'
                        constraints_spec['time'] = {'at': "2023-02-08T09:00:00.00000Z" }
                        constraints_spec['sky']['min_elevation'] = { "calibrator": 0.25235987755982988, "target": 0.8726646259971648 }
                        constraints_spec['sky']['min_distance'] = {"target": { "sun": 30*3.1415/180.0, "moon": 15*3.1415/180.0, "jupiter": 20*3.1415/180.0} }
                        constraints_spec['sky']['transit_offset'] = { "from": -5400, "to": 5400 }
                        constraints_spec['sky']['reference_pointing'] = {"enabled": False, "pointing": {"angle1": 5.2, "angle2": 0.793, "direction_type": "J2000", "target": "REF"}}
                    elif strategy_template.name == 'IM HBA - 1 Beam':
                        scheduling_unit_spec['tasks']['Target Observation']['specifications_doc']['duration'] = 8*60*60
                        scheduling_unit_spec['tasks']['Target Observation']['specifications_doc']['station_configuration']['SAPs'][0]['digital_pointing'] = {"angle1": 5.348343326580123, "angle2": 0.7946493780603804, "direction_type": "J2000", "target": "P306+46"}
                        scheduling_unit_spec['tasks']['Calibrator Observation 1']['specifications_doc']['calibrator']['pointing'] = {"angle1": 4.8412365306941805, "angle2": 0.8507809072856333, "direction_type": "J2000", "target": "3c380"}
                        scheduling_unit_spec['tasks']['Calibrator Observation 2']['specifications_doc']['calibrator']['pointing'] = {"angle1": 4.8412365306941805, "angle2": 0.8507809072856333, "direction_type": "J2000", "target": "3c380"}
                        del scheduling_unit_spec['scheduling_constraints_doc']

                        # set some constraints, so the scheduler has something to chew on.
                        # constraints_spec['scheduler'] = 'fixed_time'
                        # constraints_spec['time'] = {'at': "2023-02-09T06:00:00.00000Z" }
                        constraints_spec['sky']['min_elevation'] = { "calibrator": 0.2, "target": 0.3 }
                        constraints_spec['sky']['min_distance'] = {"target": { "sun": 30*3.1415/180.0, "moon": 15*3.1415/180.0, "jupiter": 20*3.1415/180.0} }
                        constraints_spec['sky']['transit_offset'] = { "from": -5400, "to": 5400 }
                        constraints_spec['sky']['reference_pointing'] = {"enabled": False, "pointing": {"angle1": 5.2, "angle2": 0.793, "direction_type": "J2000", "target": "REF"}}


                    # add the scheduling_unit_doc to a new SchedulingUnitDraft instance, and were ready to use it!
                    scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(name=strategy_template.name,
                                                                                      scheduling_set=models.SchedulingSet.objects.order_by('?').first(), # attach to a random schedset
                                                                                      description="Test scheduling unit",
                                                                                      specifications_template=strategy_template.scheduling_unit_template,
                                                                                      observation_strategy_template=strategy_template,
                                                                                      scheduling_constraints_doc=constraints_spec,
                                                                                      scheduling_constraints_template=constraints_template)
                    scheduling_unit_draft = update_task_graph_from_specifications_doc(scheduling_unit_draft, scheduling_unit_spec)

                    logger.info('created test scheduling_unit_draft: %s', scheduling_unit_draft.name)

                    scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
                    return
                    if scheduling_unit_blueprint.scheduled_stop_time:
                        next_fixed_time_start_time = scheduling_unit_blueprint.scheduled_stop_time + timedelta(minutes=5)
                except Exception as e:
                    logger.exception("Could not instantiate scheduling unit from strategy id=%d '%s'-v%d: %s" % (strategy_template.id,
                                                                                                                 strategy_template.name,
                                                                                                                 strategy_template.version, e))
    except ImportError:
        pass


def populate_cycles(apps, schema_editor):
    #  Cycle 0 deviates from any patterns
    cycle = models.Cycle.objects.create(name="Cycle 00",
                                        description="Lofar Cycle 0",
                                        start=datetime(2013, 2, 11, 0, 0, 0, 0, tzinfo=timezone.utc),
                                        stop=datetime(2013, 11, 14, 0, 0, 0, 0, tzinfo=timezone.utc))

    models.CycleQuota.objects.bulk_create([models.CycleQuota(cycle=cycle,
                                                             resource_type=ResourceType.objects.get(
                                                                 name="LOFAR Observing Time"),
                                                             value=0.8 * cycle.duration.total_seconds()),
                                           # rough guess. 80% of total time available for observing
                                           models.CycleQuota(cycle=cycle,
                                                             resource_type=ResourceType.objects.get(
                                                                 name="CEP Processing Time"),
                                                             value=0.8 * cycle.duration.total_seconds()),
                                           models.CycleQuota(cycle=cycle,
                                                             resource_type=ResourceType.objects.get(name="LTA Storage"),
                                                             value=0),  # needs to be filled in by user (SOS)
                                           models.CycleQuota(cycle=cycle,
                                                             resource_type=ResourceType.objects.get(
                                                                 name="LOFAR Support Time"),
                                                             value=0),  # needs to be filled in by user (SOS)

                                           models.CycleQuota(cycle=cycle,
                                                             resource_type=ResourceType.objects.get(
                                                                 name="LOFAR Observing Time Commissioning"),
                                                             value=0.05 * cycle.duration.total_seconds()),
                                           # rough guess. 5% of total time available for observing
                                           models.CycleQuota(cycle=cycle,
                                                             resource_type=ResourceType.objects.get(
                                                                 name="LOFAR Observing Time prio A"),
                                                             value=0),  # needs to be filled in by user (SOS)
                                           models.CycleQuota(cycle=cycle,
                                                             resource_type=ResourceType.objects.get(
                                                                 name="LOFAR Observing Time prio B"),
                                                             value=0)  # needs to be filled in by user (SOS)
                                           ])

    #  Cycles 1-10 follow the same pattern
    for nr in range(1, 11):
        cycle = models.Cycle.objects.create(name="Cycle %02d" % nr,
                                            description="Lofar Cycle %s" % nr,
                                            start=datetime(2013+nr//2, 5 if nr%2==0 else 11, 15, 0, 0, 0, 0, tzinfo=timezone.utc),
                                            stop=datetime(2013+(nr+1)//2, 5 if nr%2==1 else 11, 14, 0, 0, 0, 0, tzinfo=timezone.utc))

        models.CycleQuota.objects.bulk_create([models.CycleQuota(cycle=cycle,
                                                                 resource_type=ResourceType.objects.get(name="LOFAR Observing Time"),
                                                                 value=0.8*cycle.duration.total_seconds()), # rough guess. 80% of total time available for observing
                                               models.CycleQuota(cycle=cycle,
                                                                 resource_type=ResourceType.objects.get(name="CEP Processing Time"),
                                                                 value=0.8*cycle.duration.total_seconds()),
                                               models.CycleQuota(cycle=cycle,
                                                                 resource_type=ResourceType.objects.get(name="LTA Storage"),
                                                                 value=0), # needs to be filled in by user (SOS)
                                               models.CycleQuota(cycle=cycle,
                                                                 resource_type=ResourceType.objects.get(name="LOFAR Support Time"),
                                                                 value=0),  # needs to be filled in by user (SOS)

                                               models.CycleQuota(cycle=cycle,
                                                                 resource_type=ResourceType.objects.get(name="LOFAR Observing Time Commissioning"),
                                                                 value=0.05*cycle.duration.total_seconds()), # rough guess. 5% of total time available for observing
                                               models.CycleQuota(cycle=cycle,
                                                                 resource_type=ResourceType.objects.get(name="LOFAR Observing Time prio A"),
                                                                 value=0), # needs to be filled in by user (SOS)
                                               models.CycleQuota(cycle=cycle,
                                                                 resource_type=ResourceType.objects.get(name="LOFAR Observing Time prio B"),
                                                                 value=0) # needs to be filled in by user (SOS)
                                               ])

    #  Cycle 11 deviates from any patterns
    cycle = models.Cycle.objects.create(name="Cycle 11",
                                        description="Lofar Cycle 11",
                                        start=datetime(2018, 11, 15, 0, 0, 0, 0, tzinfo=timezone.utc),
                                        stop=datetime(2019, 5, 31, 0, 0, 0, 0, tzinfo=timezone.utc))

    models.CycleQuota.objects.bulk_create([models.CycleQuota(cycle=cycle,
                                                             resource_type=ResourceType.objects.get(
                                                                 name="LOFAR Observing Time"),
                                                             value=0.8 * cycle.duration.total_seconds()),
                                           # rough guess. 80% of total time available for observing
                                           models.CycleQuota(cycle=cycle,
                                                             resource_type=ResourceType.objects.get(
                                                                 name="CEP Processing Time"),
                                                             value=0.8 * cycle.duration.total_seconds()),
                                           models.CycleQuota(cycle=cycle,
                                                             resource_type=ResourceType.objects.get(
                                                                 name="LTA Storage"),
                                                             value=0),  # needs to be filled in by user (SOS)
                                           models.CycleQuota(cycle=cycle,
                                                             resource_type=ResourceType.objects.get(
                                                                 name="LOFAR Support Time"),
                                                             value=0),  # needs to be filled in by user (SOS)

                                           models.CycleQuota(cycle=cycle,
                                                             resource_type=ResourceType.objects.get(
                                                                 name="LOFAR Observing Time Commissioning"),
                                                             value=0.05 * cycle.duration.total_seconds()),
                                           # rough guess. 5% of total time available for observing
                                           models.CycleQuota(cycle=cycle,
                                                             resource_type=ResourceType.objects.get(
                                                                 name="LOFAR Observing Time prio A"),
                                                             value=0),  # needs to be filled in by user (SOS)
                                           models.CycleQuota(cycle=cycle,
                                                             resource_type=ResourceType.objects.get(
                                                                 name="LOFAR Observing Time prio B"),
                                                             value=0)  # needs to be filled in by user (SOS)
                                           ])

    #  Cycles 12-19 follow the same pattern
    for nr in range(12, 20):
        cycle = models.Cycle.objects.create(name="Cycle %02d" % nr,
                                            description="Lofar Cycle %s" % nr,
                                            start=datetime(2013 + nr // 2, 6 if nr % 2 == 0 else 12, 1, 0, 0, 0, 0,
                                                           tzinfo=timezone.utc),
                                            stop=datetime(2013 + (nr + 1) // 2, 5 if nr % 2 == 1 else 11,
                                                          30 if nr % 2 == 0 else 31, 0, 0,
                                                          0, 0, tzinfo=timezone.utc))

        models.CycleQuota.objects.bulk_create([models.CycleQuota(cycle=cycle,
                                                                 resource_type=ResourceType.objects.get(
                                                                     name="LOFAR Observing Time"),
                                                                 value=0.8 * cycle.duration.total_seconds()),
                                               # rough guess. 80% of total time available for observing
                                               models.CycleQuota(cycle=cycle,
                                                                 resource_type=ResourceType.objects.get(
                                                                     name="CEP Processing Time"),
                                                                 value=0.8 * cycle.duration.total_seconds()),
                                               models.CycleQuota(cycle=cycle,
                                                                 resource_type=ResourceType.objects.get(
                                                                     name="LTA Storage"),
                                                                 value=0),  # needs to be filled in by user (SOS)
                                               models.CycleQuota(cycle=cycle,
                                                                 resource_type=ResourceType.objects.get(
                                                                     name="LOFAR Support Time"),
                                                                 value=0),  # needs to be filled in by user (SOS)

                                               models.CycleQuota(cycle=cycle,
                                                                 resource_type=ResourceType.objects.get(
                                                                     name="LOFAR Observing Time Commissioning"),
                                                                 value=0.05 * cycle.duration.total_seconds()),
                                               # rough guess. 5% of total time available for observing
                                               models.CycleQuota(cycle=cycle,
                                                                 resource_type=ResourceType.objects.get(
                                                                     name="LOFAR Observing Time prio A"),
                                                                 value=0),  # needs to be filled in by user (SOS)
                                               models.CycleQuota(cycle=cycle,
                                                                 resource_type=ResourceType.objects.get(
                                                                     name="LOFAR Observing Time prio B"),
                                                                 value=0)  # needs to be filled in by user (SOS)
                                               ])


def populate_projects(apps, schema_editor):
    for name, rank in (("high", ProjectRank.HIGHEST.value), ("normal", ProjectRank.LOWEST.value), ("low", ProjectRank.LOWEST.value), ("COM_LOFAR2", ProjectRank.HIGHEST.value)):
        try:
            with transaction.atomic():
                tmss_project = models.Project.objects.create(name=name,
                                                         description="Project for all TMSS tests and commissioning (%s priority)" % (name,),
                                                         rank=rank,
                                                         can_trigger=name=='high',
                                                         private_data=True,
                                                         expert=True,
                                                         filler=False,
                                                         project_state=models.ProjectState.objects.get(value=models.ProjectState.Choices.ACTIVE.value),
                                                         project_category=models.ProjectCategory.objects.get(value=models.ProjectCategory.Choices.TEST.value))
                tmss_project.tags = ["Commissioning"]
                tmss_project.cycles.set(models.Cycle.objects.all())
                tmss_project.save()

                # add some trigger quota
                models.ProjectQuota.objects.create(project=tmss_project, value=10,
                                                   resource_type=models.ResourceType.objects.get(
                                                       name="Number of triggers"))

                # for convenience, create a schedulingset for each project
                models.SchedulingSet.objects.create(name="Test Scheduling Set", project=tmss_project)

                project_quota = ProjectQuota.objects.create(project=tmss_project, value=1e12, resource_type=ResourceType.objects.get(name="LTA Storage"))
                sara_cluster = Cluster.objects.get(name="SARA")
                sara_fs = sara_cluster.filesystem_set.first()
                models.ProjectQuotaArchiveLocation.objects.create(project_quota=project_quota, archive_location=sara_fs)
        except Exception as e:
            logger.error("populate_projects: %s", e)


def populate_resources(apps, schema_editor):
    bytes_q = Quantity.objects.get(value=Quantity.Choices.BYTES.value)
    time_q = Quantity.objects.get(value=Quantity.Choices.TIME.value)
    number_q = Quantity.objects.get(value=Quantity.Choices.NUMBER.value)

    ResourceType.objects.bulk_create([ResourceType(name="LTA Storage", description="Amount of storage in the LTA (in bytes)", quantity=bytes_q),
                                      ResourceType(name="CEP Storage", description="Amount of storage on the CEP processing cluster (in bytes)", quantity=bytes_q),
                                      ResourceType(name="CEP Processing Time", description="Processing time on the CEP processing cluster (in seconds)", quantity=time_q),
                                      ResourceType(name="LOFAR Observing Time", description="Observing time (in seconds)", quantity=time_q),
                                      ResourceType(name="LOFAR Observing Time prio A", description="Observing time with priority A (in seconds)", quantity=time_q),
                                      ResourceType(name="LOFAR Observing Time prio B", description="Observing time with priority B (in seconds)", quantity=time_q),
                                      ResourceType(name="LOFAR Observing Time Commissioning", description="Observing time for Commissioning/DDT (in seconds)", quantity=time_q),
                                      ResourceType(name="LOFAR Support Time", description="Support time by human (in seconds)", quantity=time_q),
                                      ResourceType(name="Number of triggers", description="Number of trigger events (as integer)", quantity=number_q) ])


def populate_misc(apps, schema_editor):
    cep4_cluster = Cluster.objects.create(name="CEP4", location="CIT", archive_site=False)
    from lofar.common import isProductionEnvironment, isTestEnvironment
    if isProductionEnvironment():
        Filesystem.objects.create(name="Production Storage (CIT)", cluster=cep4_cluster, capacity=3.6e15, directory="/data/projects")
    elif isTestEnvironment():
        Filesystem.objects.create(name="Test Storage (CIT)", cluster=cep4_cluster, capacity=3.6e15, directory="/data/test-projects")
    else:
        Filesystem.objects.create(name="Development Test FS", cluster=cep4_cluster, capacity=1e9, directory="/tmp")

    # LTA-site clusters
    sara_cluster = Cluster.objects.create(name="SARA", location="SARA", archive_site=True)
    juelich_cluster = Cluster.objects.create(name="Jülich", location="Jülich", archive_site=True)
    poznan_cluster = Cluster.objects.create(name="Poznan", location="Poznan", archive_site=True)

    if isProductionEnvironment():
        Filesystem.objects.create(name="Lofar Storage (SARA)", cluster=sara_cluster, capacity=3.6e15,
                                  directory="srm://srm.grid.sara.nl:8443/pnfs/grid.sara.nl/data/lofar/ops/projects/")
        Filesystem.objects.create(name="Lofar Storage (Jülich)", cluster=juelich_cluster, capacity=3.6e15,
                                  directory="srm://lofar-srm.fz-juelich.de:8443/pnfs/fz-juelich.de/data/lofar/ops/projects/")
        Filesystem.objects.create(name="Lofar Storage (Poznan)", cluster=poznan_cluster, capacity=3.6e15,
                                  directory="srm://lta-head.lofar.psnc.pl:8443/lofar/ops/projects/")
    elif isTestEnvironment():
        Filesystem.objects.create(name="Lofar Test Storage (SARA)", cluster=sara_cluster, capacity=3.6e15,
                                  directory="srm://srm.grid.sara.nl:8443/pnfs/grid.sara.nl/data/lofar/ops/test/projects/")
        Filesystem.objects.create(name="Lofar User Disk Storage (SARA)", cluster=sara_cluster, capacity=3.6e15,
                                  directory="srm://srm.grid.sara.nl/pnfs/grid.sara.nl/data/lofar/user/disk/projects/")
    else:
        Filesystem.objects.create(name="Development Storage (fake LTA)", cluster=sara_cluster, capacity=0,
                                  directory="srm://dev.lta.site/projects/lofar")




def populate_connectors():
    # the TaskConnectorType's define how the Task[Draft/Blueprint] *can* be connected.

    # NOTE: This is an explicit list of each possible link between tasks. This model suffices
    # until the number of connectors throw too large. By then, we could consider introducing
    # wild cards, like output_of=NULL meaning "any".
    logger.info("Populating TaskConnectorType's")
    from django.db.utils import IntegrityError

    def create_task_connector_skip_duplicate(task_template_name:str, **kwargs):
        # wrapper func to silently skip duplicates
        for task_template in TaskTemplate.objects.filter(name=task_template_name).all():
            try:
                connector = TaskConnectorType.objects.create(task_template=task_template, **kwargs)
                logger.info('created task_type_connector id: %s role: %s datatype: %s dataformat: %s  iotype: %s task_template=\'%s-v%s\'',
                            connector.id, connector.role.value, connector.datatype.value, connector.dataformat.value, connector.iotype.value, connector.task_template.name, connector.task_template.version)
            except IntegrityError:
                # skipping duplicate
                pass

    # calibrator, target and combined imaging observations
    for task_template_name in ['calibrator observation', 'target observation', 'parallel calibrator target observation', 'parallel target and beamforming observation', 'parallel calibrator target and beamforming observation']:
        create_task_connector_skip_duplicate(role=Role.objects.get(value=Role.Choices.CORRELATOR.value),  # TODO: shouldn't this be TARGET role for e.g. target observation? And either ANY or both a CALIBRATOR and a TARGET role type for the combined obs?
                                         datatype=Datatype.objects.get(value=Datatype.Choices.VISIBILITIES.value),
                                         dataformat=Dataformat.objects.get(value=Dataformat.Choices.MEASUREMENTSET.value),
                                         task_template_name=task_template_name,
                                         iotype=IOType.objects.get(value=IOType.Choices.OUTPUT.value))

    # beamforming observation
    for task_template_name in ['beamforming observation', 'parallel target and beamforming observation', 'parallel calibrator target and beamforming observation']:
        create_task_connector_skip_duplicate(role=Role.objects.get(value=Role.Choices.BEAMFORMER.value),
                                     datatype=Datatype.objects.get(value=Datatype.Choices.TIME_SERIES.value),
                                     dataformat=Dataformat.objects.get(value=Dataformat.Choices.BEAMFORMED.value),
                                     task_template_name=task_template_name,
                                     iotype=IOType.objects.get(value=IOType.Choices.OUTPUT.value))

    # pulsar pipeline
    create_task_connector_skip_duplicate(role=Role.objects.get(value=Role.Choices.BEAMFORMER.value),
                                 datatype=Datatype.objects.get(value=Datatype.Choices.TIME_SERIES.value),
                                 dataformat=Dataformat.objects.get(value=Dataformat.Choices.BEAMFORMED.value),
                                 task_template_name='pulsar pipeline',
                                 iotype=IOType.objects.get(value=IOType.Choices.INPUT.value))

    create_task_connector_skip_duplicate(role=Role.objects.get(value=Role.Choices.ANY.value),
                                 datatype=Datatype.objects.get(value=Datatype.Choices.QUALITY.value),
                                 dataformat=Dataformat.objects.get(value=Dataformat.Choices.PULP_SUMMARY.value),
                                 task_template_name='pulsar pipeline',
                                 iotype=IOType.objects.get(value=IOType.Choices.OUTPUT.value))

    create_task_connector_skip_duplicate(role=Role.objects.get(value=Role.Choices.ANY.value),
                                 datatype=Datatype.objects.get(value=Datatype.Choices.PULSAR_PROFILE.value),
                                 dataformat=Dataformat.objects.get(value=Dataformat.Choices.PULP_ANALYSIS.value),
                                 task_template_name='pulsar pipeline',
                                 iotype=IOType.objects.get(value=IOType.Choices.OUTPUT.value))

    # preprocessing pipeline
    for iotype_value in (IOType.Choices.INPUT.value, IOType.Choices.OUTPUT.value):
        create_task_connector_skip_duplicate(role=Role.objects.get(value=Role.Choices.ANY.value),
                                         datatype=Datatype.objects.get(value=Datatype.Choices.VISIBILITIES.value),
                                         dataformat=Dataformat.objects.get(value=Dataformat.Choices.MEASUREMENTSET.value),
                                         task_template_name='preprocessing pipeline',
                                         iotype=IOType.objects.get(value=iotype_value))

    # Ingest and Cleanup can/should accept all kinds of data.
    # So we could loop over all combinations of Datatype, Dataformat and Role and create an input connector for each.
    # This would result however in "unrealistic"/non-existing types like: TIME_SERIES-MEASUREMENTSET, or VISIBILITIES-BEAMFORMED, etc, which do not make any sense.
    # So, instead, lets loop over all exising output connectors, and accept those as input.
    for task_template_name in ('ingest', 'cleanup'):
        # loop over all existing output types
        any_role = Role.objects.get(value=Role.Choices.ANY.value)
        for output_connector_type in TaskConnectorType.objects.filter(iotype=IOType.objects.get(value=IOType.Choices.OUTPUT.value)).all():
            # always create two input connectors for the specific output_connector_type.role and the any_role
            for role in [output_connector_type.role, any_role]:
                create_task_connector_skip_duplicate(role=role,
                                                     datatype=output_connector_type.datatype,
                                                     dataformat=output_connector_type.dataformat,
                                                     task_template_name=task_template_name,
                                                     iotype=IOType.objects.get(value=IOType.Choices.INPUT.value))

    # The Copy (sub)task accepts any kinds of data, and passes it on.
    task_template_name = "copy pipeline"
    for output_connector_type in TaskConnectorType.objects.filter(iotype=IOType.objects.get(value=IOType.Choices.OUTPUT.value)).all():
        any_role = Role.objects.get(value=Role.Choices.ANY.value)
        for iotype_value in [IOType.Choices.INPUT.value, IOType.Choices.OUTPUT.value]:
            create_task_connector_skip_duplicate(role=any_role,
                                                 datatype=output_connector_type.datatype,
                                                 dataformat=output_connector_type.dataformat,
                                                 task_template_name=task_template_name,
                                                 iotype=IOType.objects.get(value=iotype_value))


def populate_permissions():
    logger.info('Populating permissions')

    populate_project_permissions()
    populate_system_permissions()
    populate_system_roles()
    populate_system_test_users()


def populate_project_permissions():

    # For each viewset and for each extra action create a project permission entry.
    members = inspect.getmembers(viewsets)
    if isViewflowEnabled():
        members += inspect.getmembers(workflow_viewsets)
    for name, obj in members:
        if inspect.isclass(obj):
            try:
                permission_name = obj.serializer_class.Meta.model.__name__.lower()
                logger.info('creating project permission %s' % permission_name)
                try:
                    ProjectPermission.objects.create(name=permission_name)
                except IntegrityError as e:
                    logger.debug('Skipping project permission creation for obj=%s: %s' % (obj, e))
                extra_actions = obj.get_extra_actions()
                if extra_actions:
                    for action in extra_actions:
                        action_permission_name = '%s-%s' % (permission_name, action.__name__)
                        logger.info('creating project permission %s' % action_permission_name)
                        try:
                            ProjectPermission.objects.create(name=action_permission_name)
                        except IntegrityError as e:
                            logger.debug('Skipping project permission creation for obj=%s: %s' % (obj, e))

            except Exception as e:
                logger.debug('Skipping project permission creation for obj=%s: %s' % (obj, e))

    # Then Assign initial permission matrix
    #
    def _get_all_project_permissions_for_viewset(view):
        name = view.serializer_class.Meta.model.__name__.lower()

        # permission for the model itself
        perms = [ProjectPermission.objects.get(name=name)]

        # permissions for all extra actions
        extra_actions = view.get_extra_actions()
        if extra_actions:
            for action in extra_actions:
                action_permission_name = '%s-%s' % (name, action.__name__)
                perms.append(ProjectPermission.objects.get(name=action_permission_name))
        return perms

    # system settings
    for view in permission_groups['system_setting']:
        perms = _get_all_project_permissions_for_viewset(view)
        for perm in perms:
            perm.GET.set([ProjectRole.objects.get(value=role) for role in ['pi', 'co_i', 'contact', 'shared_support', 'friend_of_project', 'friend_of_project_primary']])
            # Note: we only grant GET permissions, since other methods are only permitted to users who are superuser anyway.

    # system feedback
    for view in permission_groups['system_feedback']:
        perms = _get_all_project_permissions_for_viewset(view)
        for perm in perms:
            perm.GET.set([ProjectRole.objects.get(value=role) for role in ['pi', 'co_i', 'contact', 'shared_support', 'friend_of_project', 'friend_of_project_primary']])
            # Note: we only grant GET permissions, since other methods are only permitted to users who are superuser anyway.

    # cycle administration
    for view in permission_groups['cycle_administration']:
        perms = _get_all_project_permissions_for_viewset(view)
        for perm in perms:
            perm.GET.set([ProjectRole.objects.get(value=role) for role in ['pi', 'co_i', 'contact', 'shared_support', 'friend_of_project', 'friend_of_project_primary']])
            # Note: we only grant GET permissions, since other methods are only permitted to users with particular system roles.

    # project administration
    for view in permission_groups['project_administration']:
        perms = _get_all_project_permissions_for_viewset(view)
        for perm in perms:
            perm.GET.set([ProjectRole.objects.get(value=role) for role in ['pi', 'co_i', 'contact', 'shared_support', 'friend_of_project', 'friend_of_project_primary']])
            perm.POST.set([ProjectRole.objects.get(value=role) for role in ['shared_support']])
            perm.PUT.set([ProjectRole.objects.get(value=role) for role in ['shared_support']])
            perm.PATCH.set([ProjectRole.objects.get(value=role) for role in ['shared_support']])
            perm.DELETE.set([ProjectRole.objects.get(value=role) for role in ['shared_support']])

    # blueprint alteration
    for view in permission_groups['blueprint_alteration']:
        perms = _get_all_project_permissions_for_viewset(view)
        for perm in perms:
            perm.GET.set([ProjectRole.objects.get(value=role) for role in ['pi', 'co_i', 'contact', 'shared_support', 'friend_of_project', 'friend_of_project_primary']])
            perm.POST.set([ProjectRole.objects.get(value=role) for role in ['shared_support']])
            perm.PUT.set([ProjectRole.objects.get(value=role) for role in ['shared_support']])
            perm.PATCH.set([ProjectRole.objects.get(value=role) for role in ['shared_support']])
            perm.DELETE.set([ProjectRole.objects.get(value=role) for role in ['shared_support']])

    # blueprint fixed
    for view in permission_groups['blueprint_fixed']:
        perms = _get_all_project_permissions_for_viewset(view)
        for perm in perms:
            perm.GET.set([ProjectRole.objects.get(value=role) for role in ['pi', 'co_i', 'contact', 'shared_support', 'friend_of_project', 'friend_of_project_primary']])
            # Note: we only grant GET permissions, since other methods are only permitted to users who are superuser anyway.

    # operational
    for view in permission_groups['operational']:
        perms = _get_all_project_permissions_for_viewset(view)
        for perm in perms:
            perm.GET.set([ProjectRole.objects.get(value=role) for role in ['pi', 'co_i', 'contact', 'shared_support', 'friend_of_project', 'friend_of_project_primary']])
            # Note: we only grant GET permissions, since other methods are only permitted to users who are superuser anyway.

    # workflow
    for view in permission_groups['workflow']:
        perms = _get_all_project_permissions_for_viewset(view)
        for perm in perms:
            perm.GET.set([ProjectRole.objects.get(value=role) for role in ['co_i']])  # all project members, all other project roles have this one anyway
            perm.POST.set([ProjectRole.objects.get(value=role) for role in ['shared_support', 'friend_of_project', 'friend_of_project_primary']])
            perm.PUT.set([ProjectRole.objects.get(value=role) for role in ['shared_support', 'friend_of_project', 'friend_of_project_primary']])
            perm.PATCH.set([ProjectRole.objects.get(value=role) for role in ['shared_support', 'friend_of_project', 'friend_of_project_primary']])

    # workflow PI verification step
    for view in permission_groups['workflow_piverification']:
        perms = _get_all_project_permissions_for_viewset(view)
        for perm in perms:
            perm.GET.set([ProjectRole.objects.get(value=role) for role in ['co_i']])  # all project members, all other project roles have this one anyway
            perm.POST.set([ProjectRole.objects.get(value=role) for role in ['pi', 'contact', 'shared_support', 'friend_of_project', 'friend_of_project_primary']])
            perm.PUT.set([ProjectRole.objects.get(value=role) for role in ['pi', 'contact', 'shared_support', 'friend_of_project', 'friend_of_project_primary']])
            perm.PATCH.set([ProjectRole.objects.get(value=role) for role in ['pi', 'contact', 'shared_support', 'friend_of_project', 'friend_of_project_primary']])

    # readable_all_system_roles_writable_operator_support
    for view in permission_groups['readable_all_system_roles_writable_operator_support']:
        perms = _get_all_project_permissions_for_viewset(view)
        for perm in perms:
            perm.GET.set([ProjectRole.objects.get(value=role) for role in ['pi', 'co_i', 'contact', 'shared_support', 'friend_of_project', 'friend_of_project_primary']])
            # Note: we only grant GET permissions, since other methods are only permitted to users with particular system roles.

    # re-assign permissions for extra actions that need something out of the ordinary, e.g. because they operate
    # on a different model than their base model (all actions initially get the same permissions as standard REST
    # actions on the same model).
    perm = ProjectPermission.objects.get(name='schedulingunitobservingstrategytemplate-create_scheduling_unit')
    perm.GET.set([])
    perm.POST.set([ProjectRole.objects.get(value=role) for role in ['shared_support']])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='reservationstrategytemplate-create_reservation')
    perm.GET.set([])
    perm.POST.set([])  # only system roles allow this
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='schedulingunitblueprint-copy_specifications_doc_back_into_draft')
    perm.GET.set([])
    perm.POST.set([ProjectRole.objects.get(value=role) for role in ['shared_support']])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='schedulingunitblueprint-copy_specifications_doc_including_copies_for_failed_tasks_back_into_draft')
    perm.GET.set([])
    perm.POST.set([ProjectRole.objects.get(value=role) for role in ['shared_support']])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='schedulingunitblueprint-copy_to_draft')
    perm.GET.set([])
    perm.POST.set([ProjectRole.objects.get(value=role) for role in ['shared_support']])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='schedulingunitblueprint-create_cleanuptask_for_scheduling_unit_blueprint')
    perm.GET.set([])
    perm.POST.set([ProjectRole.objects.get(value=role) for role in ['shared_support']])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='schedulingunitblueprint-create_copies_of_failed_tasks_via_draft')
    perm.GET.set([])
    perm.POST.set([ProjectRole.objects.get(value=role) for role in ['shared_support']])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='schedulingunitblueprint-update_task_blueprints_and_subtasks_graph_from_draft')
    perm.GET.set([])
    perm.POST.set([ProjectRole.objects.get(value=role) for role in ['shared_support']])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='schedulingunitblueprint-get_all_subtasks_log_urls')
    perm.GET.set([ProjectRole.objects.get(value=role) for role in ['pi', 'co_i', 'contact', 'shared_support', 'friend_of_project', 'friend_of_project_primary']])
    perm.POST.set([])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='schedulingunitblueprint-subtasks')
    perm.GET.set([ProjectRole.objects.get(value=role) for role in ['pi', 'co_i', 'contact', 'shared_support', 'friend_of_project', 'friend_of_project_primary']])
    perm.POST.set([])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='schedulingunitblueprint-system_events')
    perm.GET.set([ProjectRole.objects.get(value=role) for role in ['pi', 'co_i', 'contact', 'shared_support', 'friend_of_project', 'friend_of_project_primary']])
    perm.POST.set([])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='taskblueprint-system_events')
    perm.GET.set([ProjectRole.objects.get(value=role) for role in ['pi', 'co_i', 'contact', 'shared_support', 'friend_of_project', 'friend_of_project_primary']])
    perm.POST.set([])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='schedulingunitdraft-create_cleanuptask_for_scheduling_unit_draft')
    perm.GET.set([])
    perm.POST.set([ProjectRole.objects.get(value=role) for role in ['shared_support']])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='schedulingunitdraft-update_task_graph_from_specifications_doc')
    perm.GET.set([])
    perm.POST.set([ProjectRole.objects.get(value=role) for role in ['shared_support']])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='taskblueprint-copy_to_new_draft')
    perm.GET.set([])
    perm.POST.set([ProjectRole.objects.get(value=role) for role in ['shared_support']])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='schedulingunitdraft-create_observing_strategy_template')
    perm.GET.set([])
    perm.POST.set([])  # only system roles allow this
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='schedulingunitdraft-create_scheduling_unit_blueprint_and_tasks_and_subtasks_tree')
    perm.GET.set([])
    perm.POST.set([ProjectRole.objects.get(value=role) for role in ['shared_support']])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='subtask-input_dataproducts')
    perm.GET.set([ProjectRole.objects.get(value=role) for role in ['pi', 'co_i', 'contact', 'shared_support', 'friend_of_project', 'friend_of_project_primary']])
    perm.POST.set([])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='subtask-output_dataproducts')
    perm.GET.set([ProjectRole.objects.get(value=role) for role in ['pi', 'co_i', 'contact', 'shared_support', 'friend_of_project', 'friend_of_project_primary']])
    perm.POST.set([])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='subtask-transformed_output_dataproduct')
    perm.GET.set([ProjectRole.objects.get(value=role) for role in ['pi', 'co_i', 'contact', 'shared_support', 'friend_of_project', 'friend_of_project_primary']])
    perm.POST.set([])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='taskblueprint-create_subtasks')
    perm.GET.set([])
    perm.POST.set([ProjectRole.objects.get(value=role) for role in ['shared_support']])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='taskblueprint-schedule_independent_subtasks')
    perm.GET.set([])
    perm.POST.set([ProjectRole.objects.get(value=role) for role in ['shared_support']])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='taskdraft-create_task_blueprint')
    perm.GET.set([])
    perm.POST.set([ProjectRole.objects.get(value=role) for role in ['shared_support']])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='taskdraft-create_task_blueprint_and_subtasks')
    perm.GET.set([])
    perm.POST.set([ProjectRole.objects.get(value=role) for role in ['shared_support']])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='tasktemplate-connectors')
    perm.GET.set([ProjectRole.objects.get(value=role) for role in ['pi', 'co_i', 'contact', 'shared_support', 'friend_of_project', 'friend_of_project_primary']])
    perm.POST.set([])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='taskblueprint-cancel')
    perm.GET.set([])
    perm.POST.set([ProjectRole.objects.get(value=role) for role in ['shared_support', 'friend_of_project']])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='taskblueprint-mark_as_obsolete')
    perm.GET.set([])
    perm.POST.set([ProjectRole.objects.get(value=role) for role in ['shared_support', 'friend_of_project']])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='taskblueprint-restart_when_on_hold')
    perm.GET.set([])
    perm.POST.set([ProjectRole.objects.get(value=role) for role in ['shared_support', 'friend_of_project']])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='schedulingunitblueprint-cancel')
    perm.GET.set([])
    perm.POST.set([ProjectRole.objects.get(value=role) for role in ['shared_support', 'friend_of_project']])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='schedulingunitblueprint-create_lofar2_sibling')
    perm.GET.set([])
    perm.POST.set([])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='schedulingunitblueprint-mark_as_obsolete')
    perm.GET.set([])
    perm.POST.set([ProjectRole.objects.get(value=role) for role in ['shared_support', 'friend_of_project']])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='subtask-cancel')
    perm.GET.set([])
    perm.POST.set([ProjectRole.objects.get(value=role) for role in ['shared_support', 'friend_of_project']])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='subtask-mark_as_obsolete')
    perm.GET.set([])
    perm.POST.set([ProjectRole.objects.get(value=role) for role in ['shared_support', 'friend_of_project']])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='subtask-restart_when_on_hold')
    perm.GET.set([])
    perm.POST.set([ProjectRole.objects.get(value=role) for role in ['shared_support', 'friend_of_project']])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='subtask-create_and_process_feedback_from_specification_and_set_to_finished')
    perm.GET.set([])
    perm.POST.set([ProjectRole.objects.get(value=role) for role in ['shared_support']])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])

    perm = ProjectPermission.objects.get(name='cycle-report')
    perm.GET.set([])
    perm.POST.set([])
    perm.PUT.set([])
    perm.PATCH.set([])
    perm.DELETE.set([])


def populate_system_permissions():
    # For each viewset create custom permissions for extra actions.
    for name, obj in inspect.getmembers(viewsets):
        if inspect.isclass(obj):
            try:
                ct = ContentType.objects.get_for_model(obj.serializer_class.Meta.model)
                extra_actions = obj.get_extra_actions()
                if extra_actions:
                    for action in extra_actions:
                        codename = ("%s_%s" % (action.__name__, obj.serializer_class.Meta.model.__name__)).lower()
                        name = f'Can {action.__name__} {obj.serializer_class.Meta.model.__name__.lower()}'
                        Permission.objects.get_or_create(codename=codename, name=name, content_type=ct)
            except:
                pass


def populate_system_roles():
    operator_group = Group.objects.get_or_create(name='operator')
    support_group = Group.objects.get_or_create(name='support')
    developer_group = Group.objects.get_or_create(name='developer')
    admin_permission_group = Group.objects.get_or_create(name='admin')
    maintenance_group = Group.objects.get_or_create(name='maintenance')
    scientist_group = Group.objects.get_or_create(name='scientist')
    spectator_group = Group.objects.get_or_create(name='spectator')

    assign_system_permissions()


def assign_system_permissions():
    '''
    Assign system permission to each system role, accordingly.
    '''

    # Get system roles
    operator_group = Group.objects.get(name='operator')
    support_group = Group.objects.get(name='support')
    developer_group = Group.objects.get(name='developer')
    admin_group = Group.objects.get(name='admin')
    maintenance_group = Group.objects.get(name='maintenance')
    scientist_group = Group.objects.get(name='scientist')
    spectator_group = Group.objects.get(name='spectator')

    # first, handle permissions for extra actions that need something out of the ordinary, e.g. because they operate
    # on a different model than their base model (other actions will get the same permissions as standard REST actions
    # on the same model).
    already_handled = []
    perm = Permission.objects.get(codename='create_scheduling_unit_schedulingunitobservingstrategytemplate')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='create_reservation_reservationstrategytemplate')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='copy_specifications_doc_back_into_draft_schedulingunitblueprint')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='copy_specifications_doc_including_copies_for_failed_tasks_back_into_draft_schedulingunitblueprint')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='copy_to_draft_schedulingunitblueprint')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='create_cleanuptask_for_scheduling_unit_blueprint_schedulingunitblueprint')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='create_copies_of_failed_tasks_via_draft_schedulingunitblueprint')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='update_task_blueprints_and_subtasks_graph_from_draft_schedulingunitblueprint')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='get_all_subtasks_log_urls_schedulingunitblueprint')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    developer_group.permissions.add(perm)
    admin_group.permissions.add(perm)
    maintenance_group.permissions.add(perm)
    spectator_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='subtasks_schedulingunitblueprint')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    developer_group.permissions.add(perm)
    admin_group.permissions.add(perm)
    maintenance_group.permissions.add(perm)
    spectator_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='system_events_schedulingunitblueprint')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    developer_group.permissions.add(perm)
    admin_group.permissions.add(perm)
    maintenance_group.permissions.add(perm)
    scientist_group.permissions.add(perm)
    spectator_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='system_events_taskblueprint')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    developer_group.permissions.add(perm)
    admin_group.permissions.add(perm)
    maintenance_group.permissions.add(perm)
    scientist_group.permissions.add(perm)
    spectator_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='create_cleanuptask_for_scheduling_unit_draft_schedulingunitdraft')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='update_task_graph_from_specifications_doc_schedulingunitdraft')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='copy_to_new_draft_taskblueprint')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='create_observing_strategy_template_schedulingunitdraft')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='create_scheduling_unit_blueprint_and_tasks_and_subtasks_tree_schedulingunitdraft')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='input_dataproducts_subtask')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    developer_group.permissions.add(perm)
    admin_group.permissions.add(perm)
    maintenance_group.permissions.add(perm)
    spectator_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='output_dataproducts_subtask')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    developer_group.permissions.add(perm)
    admin_group.permissions.add(perm)
    maintenance_group.permissions.add(perm)
    spectator_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='transformed_output_dataproduct_subtask')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    developer_group.permissions.add(perm)
    admin_group.permissions.add(perm)
    maintenance_group.permissions.add(perm)
    spectator_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='create_subtasks_taskblueprint')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='schedule_independent_subtasks_taskblueprint')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='restart_when_on_hold_taskblueprint')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='create_task_blueprint_and_subtasks_taskdraft')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='create_task_blueprint_taskdraft')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='connectors_tasktemplate')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    developer_group.permissions.add(perm)
    admin_group.permissions.add(perm)
    maintenance_group.permissions.add(perm)
    scientist_group.permissions.add(perm)
    spectator_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='cancel_taskblueprint')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='mark_as_obsolete_taskblueprint')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='mark_as_obsolete_subtask')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='restart_when_on_hold_subtask')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    already_handled.append(perm)

    perm = Permission.objects.get(codename='report_cycle')
    operator_group.permissions.add(perm)
    support_group.permissions.add(perm)
    already_handled.append(perm)

    # system settings
    for view in permission_groups['system_setting']:
        name = view.serializer_class.Meta.model.__name__.lower()

        perm = Permission.objects.get(codename='view_%s' % name)
        operator_group.permissions.add(perm)
        support_group.permissions.add(perm)
        developer_group.permissions.add(perm)
        admin_group.permissions.add(perm)
        maintenance_group.permissions.add(perm)
        scientist_group.permissions.add(perm)
        spectator_group.permissions.add(perm)

        extra_actions = view.get_extra_actions()
        if extra_actions:
            for action in extra_actions:
                perm = Permission.objects.get(codename='%s_%s' % (action.__name__, name))
                if 'get' in action.mapping and perm not in already_handled:
                    operator_group.permissions.add(perm)
                    support_group.permissions.add(perm)
                    developer_group.permissions.add(perm)
                    admin_group.permissions.add(perm)
                    maintenance_group.permissions.add(perm)
                    scientist_group.permissions.add(perm)
                    spectator_group.permissions.add(perm)

        # Note: we only grant view permissions, since alteration is only permitted to users who are superuser anyway.

    # system feedback
    for view in permission_groups['system_feedback']:
        name = view.serializer_class.Meta.model.__name__.lower()

        perm = Permission.objects.get(codename='view_%s' % name)
        operator_group.permissions.add(perm)
        support_group.permissions.add(perm)
        developer_group.permissions.add(perm)
        admin_group.permissions.add(perm)
        maintenance_group.permissions.add(perm)
        spectator_group.permissions.add(perm)

        extra_actions = view.get_extra_actions()
        if extra_actions:
            for action in extra_actions:
                perm = Permission.objects.get(codename='%s_%s' % (action.__name__, name))
                if 'get' in action.mapping and perm not in already_handled:
                    operator_group.permissions.add(perm)
                    support_group.permissions.add(perm)
                    developer_group.permissions.add(perm)
                    admin_group.permissions.add(perm)
                    maintenance_group.permissions.add(perm)
                    spectator_group.permissions.add(perm)


        # Note: we only grant view permissions, since alteration is only permitted to users who are superuser anyway.

    # schema validate_specifications_doc actions
    for schema_template_viewset_subclass in viewsets.AbstractSchemaTemplateViewSet.__subclasses__():
        name = schema_template_viewset_subclass.serializer_class.Meta.model.__name__.lower()

        # make PUT action 'validate_specifications_doc' available for all groups
        perm = Permission.objects.get(codename='validate_specifications_doc_%s' % name)

        for group in (operator_group, support_group, developer_group,
                      admin_group, maintenance_group, scientist_group):
            group.permissions.add(perm)


    # cycle administration
    for view in permission_groups['cycle_administration']:
        name = view.serializer_class.Meta.model.__name__.lower()

        perm = Permission.objects.get(codename='view_%s' % name)
        operator_group.permissions.add(perm)
        support_group.permissions.add(perm)
        developer_group.permissions.add(perm)
        admin_group.permissions.add(perm)
        maintenance_group.permissions.add(perm)
        spectator_group.permissions.add(perm)

        extra_actions = view.get_extra_actions()
        if extra_actions:
            for action in extra_actions:
                perm = Permission.objects.get(codename='%s_%s' % (action.__name__, name))
                if 'get' in action.mapping and perm not in already_handled:
                    operator_group.permissions.add(perm)
                    support_group.permissions.add(perm)
                    developer_group.permissions.add(perm)
                    admin_group.permissions.add(perm)
                    maintenance_group.permissions.add(perm)
                    spectator_group.permissions.add(perm)

        perm = Permission.objects.get(codename='change_%s' % name)
        operator_group.permissions.add(perm)
        support_group.permissions.add(perm)

        perm = Permission.objects.get(codename='add_%s' % name)
        operator_group.permissions.add(perm)
        support_group.permissions.add(perm)

        perm = Permission.objects.get(codename='delete_%s' % name)
        operator_group.permissions.add(perm)
        support_group.permissions.add(perm)  # todo: confirm

        if extra_actions:
            for action in extra_actions:
                perm = Permission.objects.get(codename='%s_%s' % (action.__name__, name))
                if 'post' in action.mapping and perm not in already_handled:
                    operator_group.permissions.add(perm)
                    support_group.permissions.add(perm)

    # project administration
    for view in permission_groups['project_administration']:
        name = view.serializer_class.Meta.model.__name__.lower()

        perm = Permission.objects.get(codename='view_%s' % name)
        operator_group.permissions.add(perm)
        support_group.permissions.add(perm)
        developer_group.permissions.add(perm)
        admin_group.permissions.add(perm)
        maintenance_group.permissions.add(perm)
        spectator_group.permissions.add(perm)

        extra_actions = view.get_extra_actions()
        if extra_actions:
            for action in extra_actions:
                perm = Permission.objects.get(codename='%s_%s' % (action.__name__, name))
                if 'get' in action.mapping and perm not in already_handled:
                    operator_group.permissions.add(perm)
                    support_group.permissions.add(perm)
                    developer_group.permissions.add(perm)
                    admin_group.permissions.add(perm)
                    maintenance_group.permissions.add(perm)
                    spectator_group.permissions.add(perm)

        perm = Permission.objects.get(codename='change_%s' % name)
        operator_group.permissions.add(perm)
        support_group.permissions.add(perm)

        perm = Permission.objects.get(codename='add_%s' % name)
        operator_group.permissions.add(perm)
        support_group.permissions.add(perm)

        perm = Permission.objects.get(codename='delete_%s' % name)
        operator_group.permissions.add(perm)
        support_group.permissions.add(perm)

        if extra_actions:
            for action in extra_actions:
                perm = Permission.objects.get(codename='%s_%s' % (action.__name__, name))
                if 'post' in action.mapping and perm not in already_handled:
                    operator_group.permissions.add(perm)
                    support_group.permissions.add(perm)

    # blueprint alteration
    for view in permission_groups['blueprint_alteration']:
        name = view.serializer_class.Meta.model.__name__.lower()

        perm = Permission.objects.get(codename='view_%s' % name)
        operator_group.permissions.add(perm)
        support_group.permissions.add(perm)
        developer_group.permissions.add(perm)
        admin_group.permissions.add(perm)
        maintenance_group.permissions.add(perm)
        spectator_group.permissions.add(perm)

        extra_actions = view.get_extra_actions()
        if extra_actions:
            for action in extra_actions:
                perm = Permission.objects.get(codename='%s_%s' % (action.__name__, name))
                if 'get' in action.mapping and perm not in already_handled:
                    operator_group.permissions.add(perm)
                    support_group.permissions.add(perm)
                    developer_group.permissions.add(perm)
                    admin_group.permissions.add(perm)
                    maintenance_group.permissions.add(perm)
                    spectator_group.permissions.add(perm)

        perm = Permission.objects.get(codename='change_%s' % name)   # todo: allow 'add' to same groups?
        operator_group.permissions.add(perm)
        support_group.permissions.add(perm)

        perm = Permission.objects.get(codename='add_%s' % name)
        operator_group.permissions.add(perm)
        support_group.permissions.add(perm)

        perm = Permission.objects.get(codename='delete_%s' % name)
        operator_group.permissions.add(perm)
        support_group.permissions.add(perm)

        if extra_actions:
            for action in extra_actions:
                perm = Permission.objects.get(codename='%s_%s' % (action.__name__, name))
                if 'post' in action.mapping and perm not in already_handled:
                    operator_group.permissions.add(perm)
                    support_group.permissions.add(perm)

    # blueprint fixed
    for view in permission_groups['blueprint_fixed']:
        name = view.serializer_class.Meta.model.__name__.lower()

        perm = Permission.objects.get(codename='view_%s' % name)
        operator_group.permissions.add(perm)
        support_group.permissions.add(perm)
        developer_group.permissions.add(perm)
        admin_group.permissions.add(perm)
        maintenance_group.permissions.add(perm)
        spectator_group.permissions.add(perm)

        extra_actions = view.get_extra_actions()
        if extra_actions:
            for action in extra_actions:
                perm = Permission.objects.get(codename='%s_%s' % (action.__name__, name))
                if 'get' in action.mapping and perm not in already_handled:
                    operator_group.permissions.add(perm)
                    support_group.permissions.add(perm)
                    developer_group.permissions.add(perm)
                    admin_group.permissions.add(perm)
                    maintenance_group.permissions.add(perm)
                    spectator_group.permissions.add(perm)

        # Note: we only grant view permissions, since alteration is only permitted to users who are superuser anyway.

    # operational
    for view in permission_groups['operational']:
        name = view.serializer_class.Meta.model.__name__.lower()

        perm = Permission.objects.get(codename='view_%s' % name)
        operator_group.permissions.add(perm)
        support_group.permissions.add(perm)
        developer_group.permissions.add(perm)
        admin_group.permissions.add(perm)
        maintenance_group.permissions.add(perm)
        scientist_group.permissions.add(perm)
        spectator_group.permissions.add(perm)

        extra_actions = view.get_extra_actions()
        if extra_actions:
            for action in extra_actions:
                perm = Permission.objects.get(codename='%s_%s' % (action.__name__, name))
                if 'get' in action.mapping and perm not in already_handled:
                    operator_group.permissions.add(perm)
                    support_group.permissions.add(perm)
                    developer_group.permissions.add(perm)
                    admin_group.permissions.add(perm)
                    maintenance_group.permissions.add(perm)
                    scientist_group.permissions.add(perm)
                    spectator_group.permissions.add(perm)

        perm = Permission.objects.get(codename='change_%s' % name)
        operator_group.permissions.add(perm)

        perm = Permission.objects.get(codename='add_%s' % name)
        operator_group.permissions.add(perm)

        perm = Permission.objects.get(codename='delete_%s' % name)
        operator_group.permissions.add(perm)

        if extra_actions:
            for action in extra_actions:
                perm = Permission.objects.get(codename='%s_%s' % (action.__name__, name))
                if 'post' in action.mapping and perm not in already_handled:
                    operator_group.permissions.add(perm)

    # workflow + workflow_piverification
    for view in permission_groups['workflow'] + permission_groups['workflow_piverification']:
        name = view.serializer_class.Meta.model.__name__.lower()

        perm = Permission.objects.get(codename='view_%s' % name)
        operator_group.permissions.add(perm)
        support_group.permissions.add(perm)
        spectator_group.permissions.add(perm)

        extra_actions = view.get_extra_actions()
        if extra_actions:
            for action in extra_actions:
                perm = Permission.objects.get(codename='%s_%s' % (action.__name__, name))
                if 'get' in action.mapping and perm not in already_handled:
                    operator_group.permissions.add(perm)
                    support_group.permissions.add(perm)
                    spectator_group.permissions.add(perm)

        perm = Permission.objects.get(codename='change_%s' % name)
        operator_group.permissions.add(perm)
        support_group.permissions.add(perm)

        perm = Permission.objects.get(codename='add_%s' % name)
        operator_group.permissions.add(perm)
        support_group.permissions.add(perm)

        if extra_actions:
            for action in extra_actions:
                perm = Permission.objects.get(codename='%s_%s' % (action.__name__, name))
                if 'post' in action.mapping and perm not in already_handled:
                    operator_group.permissions.add(perm)
                    support_group.permissions.add(perm)

    # readable_all_system_roles_writable_operator_support
    for view in permission_groups['readable_all_system_roles_writable_operator_support']:
        name = view.serializer_class.Meta.model.__name__.lower()

        perm = Permission.objects.get(codename='view_%s' % name)
        operator_group.permissions.add(perm)
        support_group.permissions.add(perm)
        developer_group.permissions.add(perm)
        admin_group.permissions.add(perm)
        maintenance_group.permissions.add(perm)
        scientist_group.permissions.add(perm)
        spectator_group.permissions.add(perm)

        extra_actions = view.get_extra_actions()
        if extra_actions:
            for action in extra_actions:
                perm = Permission.objects.get(codename='%s_%s' % (action.__name__, name))
                if 'get' in action.mapping and perm not in already_handled:
                    operator_group.permissions.add(perm)
                    support_group.permissions.add(perm)
                    developer_group.permissions.add(perm)
                    admin_group.permissions.add(perm)
                    maintenance_group.permissions.add(perm)
                    scientist_group.permissions.add(perm)
                    spectator_group.permissions.add(perm)

        perm = Permission.objects.get(codename='change_%s' % name)
        operator_group.permissions.add(perm)
        support_group.permissions.add(perm)

        perm = Permission.objects.get(codename='add_%s' % name)
        operator_group.permissions.add(perm)
        support_group.permissions.add(perm)

        perm = Permission.objects.get(codename='delete_%s' % name)
        operator_group.permissions.add(perm)
        support_group.permissions.add(perm)

        if extra_actions:
            for action in extra_actions:
                perm = Permission.objects.get(codename='%s_%s' % (action.__name__, name))
                if 'post' in action.mapping and perm not in already_handled:
                    operator_group.permissions.add(perm)
                    support_group.permissions.add(perm)


def reset_all_permissions():
    '''
    Removes all assigned system permissions from groups and freshly reassigns them.
    Deletes all project permissions and freshly create them with initial project role requirements.
    '''

    logger.info('Clearing system permissions on groups...')
    for group_name in ['operator', 'support', 'developer', 'admin', 'maintenance', 'scientist', 'spectator']:
        group = Group.objects.get(name=group_name)
        group.permissions.clear()

    logger.info('Running populate system permissions to create those that might be missing...')
    populate_system_permissions()

    logger.info('Re-assigning system permissions on groups...')
    assign_system_permissions()

    logger.info('Deleting all project permissions...')
    models.ProjectPermission.objects.all().delete()

    logger.info('Recreate all project permissions...')
    populate_project_permissions()


def populate_system_test_users():
    # TODO: Set proper credentials (passwords at least) or remove before we hit production
    operator_user, _ = User.objects.get_or_create(username='operator', password='operator')
    operator_user.groups.add(Group.objects.get(name='operator'))
    support_user, _ = User.objects.get_or_create(username='support', password='support')
    support_user.groups.add(Group.objects.get(name='support'))
    developer_user, _ = User.objects.get_or_create(username='developer', password='developer')
    developer_user.groups.add(Group.objects.get(name='developer'))
    admin_user, _ = User.objects.get_or_create(username='admin', password='admin')
    admin_user.groups.add(Group.objects.get(name='admin'))
    maintenance_user, _ = User.objects.get_or_create(username='maintenance', password='maintenance')
    maintenance_user.groups.add(Group.objects.get(name='maintenance'))
    scientist_user, _ = User.objects.get_or_create(username='scientist', password='scientist')
    scientist_user.groups.add(Group.objects.get(name='scientist'))
    spectator_user, _ = User.objects.get_or_create(username='spectator', password='spectator')
    spectator_user.groups.add(Group.objects.get(name='spectator'))

    # todo: remove these test users when not needed any more (required for local frontend testing where Keycloak is unavailable)
    if not isProductionEnvironment():
        friend_user, _ = User.objects.get_or_create(username='tmss_friend', password='tmss_friend')
        friend_user.groups.add(Group.objects.get(name='support'))
        friend2_user, _ = User.objects.get_or_create(username='tmss_friend2', password='tmss_friend2')
        friend2_user.groups.add(Group.objects.get(name='support'))
        contact_user, _ = User.objects.get_or_create(username='tmss_contact', password='tmss_contact')
        contact_user.groups.add(Group.objects.get(name='scientist'))
        pi_user, _ = User.objects.get_or_create(username='tmss_PI', password='tmss_PI')
        pi_user.groups.add(Group.objects.get(name='scientist'))
        shared_support_user, _ = User.objects.get_or_create(username='tmss_shared_support', password='tmss_shared_support')
        shared_support_user.groups.add(Group.objects.get(name='scientist'))
        shared_support2_user, _ = User.objects.get_or_create(username='tmss_shared_support2', password='tmss_shared_support2')
        shared_support2_user.groups.add(Group.objects.get(name='scientist'))
        scientist_user, _ = User.objects.get_or_create(username='tmss_scientist', password='tmss_scientist')
        scientist_user.groups.add(Group.objects.get(name='scientist'))


def populate_sunrise_and_sunset_for_all_stations(nbr_days=3, start_date=date.today()):
    """
    Populate station timeline data of all stations for given number of days the starting at given date
    Note: If data is not in database yet, it will take about 6 seconds to calculate it for all (51) stations
    """
    starttime_for_logging = datetime.utcnow()
    logger.info("Populate sunrise and sunset for ALL known stations from %s up to %d days" % (start_date, nbr_days))
    lst_timestamps = []
    for i in range(0, nbr_days):
        dt = datetime.combine(start_date, datetime.min.time()) + timedelta(days=i)
        lst_timestamps.append(dt)

    timestamps_and_stations_to_sun_rise_and_set(tuple(lst_timestamps), tuple(get_all_stations()), create_when_not_found=True)
    logger.info("Populate sunrise and sunset done in %.1fs", (datetime.utcnow()-starttime_for_logging).total_seconds())
