"""
This file contains the serializers (for the elsewhere defined data models)
"""

from rest_framework import serializers
from .. import models
from django.core.exceptions import ImproperlyConfigured
from .widgets import JSONEditorField
from rest_flex_fields.serializers import FlexFieldsSerializerMixin

class FloatDurationField(serializers.FloatField):

    # Turn datetime to float representation in seconds.
    # (Timedeltas are otherwise by default turned into a string representation)
    def to_representation(self, value):
        return value.total_seconds()

class RelationalHyperlinkedModelSerializer(serializers.HyperlinkedModelSerializer):
    _accepted_pk_names = ('id', 'name', 'unique_identifier')

    def get_field_names(self, declared_fields, info):
        field_names = super().get_field_names(declared_fields, info)
        try:
            field_names.remove(self.url_field_name) # is added later, see retun statement
        except ValueError:
            pass

        if getattr(self.Meta, 'extra_fields', None):
            field_names += self.Meta.extra_fields

        # add 'plain'-values of the fieldnames which relate to AbstractChoice-'lookup-tables'
        choice_field_names = [name+'_value' for name, rel in info.forward_relations.items()
                              if issubclass(rel.related_model, models.AbstractChoice)]

        # add 'plain-id(s)'-values of the fieldnammes which to forward_relations
        forward_related_field_names = [name+'_ids' if rel.to_many else name+'_id'
                                       for name, rel in info.forward_relations.items()
                                       if rel.related_model._meta.pk.name in self._accepted_pk_names
                                       and name in field_names]

        # always add 'plain-id'-values of the fieldnammes which to reverse_relations
        reverse_related_field_names = [name+'_ids' for name, rel in info.reverse_relations.items()
                                       if rel.related_model._meta.pk.name in self._accepted_pk_names
                                       and name in field_names]

        # return them sorted alphabetically, with id and url first so it's easy to identify and 'click' them
        return [info.pk.name, self.url_field_name] + sorted(field_names + choice_field_names + forward_related_field_names + reverse_related_field_names)

    def build_field(self, field_name, info, model_class, nested_depth):
        '''override of super.build_field to handle 'choice' fields'''
        try:
            return super().build_field(field_name, info, model_class, nested_depth)
        except ImproperlyConfigured:
            if field_name.endswith('_ids'):
                return self.build_reverse_relations_ids_field(field_name, info, model_class, nested_depth)
            if field_name.endswith('_value'):
                return self.build_choice_field(field_name, info)
            raise

    def build_reverse_relations_ids_field(self, field_name, info, model_class, nested_depth):
        '''builds a PrimaryKeyRelatedField serializer for the 'reverse_relations_ids' fields'''
        return serializers.PrimaryKeyRelatedField, {'label':field_name,
                                                    'source':field_name[:-4], # cut '_ids' from end
                                                    'many':True,
                                                    'read_only':True}

    def build_choice_field(self, field_name, info):
        '''builds a StringRelatedField serializer for the 'choice' fields'''
        original_field_name = field_name[:-6] # cut '_value' from end
        if original_field_name in info.forward_relations.keys():
            return serializers.StringRelatedField, {'label':field_name,
                                                    'source': original_field_name,
                                                    'read_only':True}


class DynamicRelationalHyperlinkedModelSerializer(FlexFieldsSerializerMixin, RelationalHyperlinkedModelSerializer):
    """
    A RelationalHyperlinkedModelSerializer that allows to control the serialization depth via a depth url parameter,
    and that allows to select which fields to include/exclude or to serialize as nested objects instead of references
    via fields, omit, and expand url parameters (provided by FlexFieldsSerializerMixin).
    """

    def to_representation(self, instance):
        request = self.context['request']

        # temporarily set requested depth and get representation
        depth = request.GET.get('depth')
        if depth:
            real_depth = getattr(self.Meta, 'depth', 0)
            self.Meta.depth = int(depth)
        represenation = super().to_representation(instance)

        # revert original depth and return
        if depth:
            self.Meta.depth = real_depth
        return represenation


class AbstractTemplateSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        abstract = True

    def create(self, validated_data):
        try:
            # 'version' is a non-editable field for normal users in forms, but we need to accept it when uploaded from a client when populating
            validated_data['version'] = self.context.get('request').data['version']
        finally:
            return super().create(validated_data)


class AbstractSchemaTemplateSerializer(AbstractTemplateSerializer):
    schema = JSONEditorField(schema_source=None)
    ref_resolved_schema = JSONEditorField(schema_source=None, read_only=True, help_text="This is a read-only view on the schema with all references resolved.")


class TaskConnectorTypeSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.TaskConnectorType
        fields = '__all__'


class TemplatePurposeSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.TemplatePurpose
        fields = '__all__'


class TemplateStateSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.TemplateState
        fields = '__all__'


