"""
This file contains the serializers (for the elsewhere defined data models)
"""

import logging
logger = logging.getLogger(__name__)

from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from .. import models
from .widgets import JSONEditorField
from .common import FloatDurationField, RelationalHyperlinkedModelSerializer, AbstractSchemaTemplateSerializer, DynamicRelationalHyperlinkedModelSerializer, TaskConnectorTypeSerializer

class SubtaskStateSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.SubtaskState
        fields = '__all__'


class SubtaskAllowedStateTransitionsSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.SubtaskAllowedStateTransitions
        fields = '__all__'


class SubtaskStateLogSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.SubtaskStateLog
        fields = '__all__'


class SubtaskTypeSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.SubtaskType
        fields = '__all__'


class StationTypeSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.StationType
        fields = '__all__'


class HashAlgorithmSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.HashAlgorithm
        fields = '__all__'


class SubtaskTemplateSerializer(AbstractSchemaTemplateSerializer):
    class Meta:
        model = models.SubtaskTemplate
        fields = '__all__'


class DataproductSpecificationsTemplateSerializer(AbstractSchemaTemplateSerializer):
    class Meta:
        model = models.DataproductSpecificationsTemplate
        fields = '__all__'


class DataproductFeedbackTemplateSerializer(AbstractSchemaTemplateSerializer):
    class Meta:
        model = models.DataproductFeedbackTemplate
        fields = '__all__'


class SubtaskFeedbackTemplateSerializer(AbstractSchemaTemplateSerializer):
    class Meta:
        model = models.SubtaskFeedbackTemplate
        fields = '__all__'


class SubtaskSerializer(DynamicRelationalHyperlinkedModelSerializer):
    cluster_name = serializers.StringRelatedField(source='cluster', label='cluster_name', read_only=True, help_text='The cluster name as defined in the specifications template, provided here to safe an addition lookup.')
    subtask_type = serializers.StringRelatedField(source='specifications_template.type', label='subtask_type', read_only=True, help_text='The subtask type as defined in the specifications template, provided here to safe an addition lookup.')
    specifications_doc = JSONEditorField(schema_source='specifications_template.schema')
    processing_feedback_doc = JSONEditorField(schema_source='processing_feedback_template.schema', required=False)
    duration = FloatDurationField(read_only=True)
    primary = serializers.BooleanField()
    project = serializers.StringRelatedField(source=models.Subtask.path_to_project.replace('__', '.'), label='project', read_only=True, help_text='The project this subtask belongs to.')

    class Meta:
        model = models.Subtask
        fields = '__all__'
        extra_fields = ['inputs', 'outputs', 'is_using_lofar2_stations']
        read_only_fields = ['inputs', 'outputs', 'is_using_lofar2_stations']
        expandable_fields = {
            'input_dataproducts': ('lofar.sas.tmss.tmss.tmssapp.serializers.DataproductSerializer', {'many': True}),
            'output_dataproducts': ('lofar.sas.tmss.tmss.tmssapp.serializers.DataproductSerializer', {'many': True})
        }

    def create(self, validated_data):
        # override normal create, refresh_from_db after creating, because some fields are computed in the database and not returned by the normal serializer
        subtask = super().create(validated_data)
        subtask.refresh_from_db()
        return subtask


class SubtaskInputSerializer(DynamicRelationalHyperlinkedModelSerializer):
    selection_doc = JSONEditorField(schema_source='selection_template.schema')

    class Meta:
        model = models.SubtaskInput
        fields = '__all__'


class SubtaskInputExtendedSerializer(SubtaskInputSerializer):
    input_role = TaskConnectorTypeSerializer(read_only=True, help_text='The connector type which defines what kind of data this subtask consumes.')


class SubtaskOutputSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.SubtaskOutput
        fields = '__all__'
        extra_fields = ['dataproducts', 'consumers']
        read_only_fields = ['dataproducts', 'consumers']


class SubtaskOutputExtendedSerializer(SubtaskOutputSerializer):
    output_role = TaskConnectorTypeSerializer(read_only=True, help_text='The connector type which defines what kind of data this subtask produces.')


class SubtaskExtendedSerializer(SubtaskSerializer):
    inputs = SubtaskInputExtendedSerializer(read_only=True, many=True)
    outputs = SubtaskOutputExtendedSerializer(read_only=True, many=True)


class DataproductSerializer(DynamicRelationalHyperlinkedModelSerializer):
    specifications_doc = JSONEditorField(schema_source='specifications_template.schema')
    feedback_doc = JSONEditorField(schema_source='feedback_template.schema')

    class Meta:
        model = models.Dataproduct
        fields = '__all__'
        extra_fields = ['filepath', 'consumers']
        read_only_fields = ['filepath', 'consumers']


class AntennaSetSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.AntennaSet
        fields = '__all__'


class DataproductTransformSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.DataproductTransform
        fields = '__all__'


class FilesystemSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.Filesystem
        fields = '__all__'


class ClusterSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.Cluster
        fields = '__all__'


class DataproductArchiveInfoSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.DataproductArchiveInfo
        fields = '__all__'


class DataproductHashSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.DataproductHash
        fields = '__all__'


class SAPSerializer(DynamicRelationalHyperlinkedModelSerializer):
    specifications_doc = JSONEditorField(schema_source='specifications_template.schema')

    class Meta:
        model = models.SAP
        fields = '__all__'


class SAPTemplateSerializer(AbstractSchemaTemplateSerializer):
    class Meta:
        model = models.SAPTemplate
        fields = '__all__'


class SAPSerializer(DynamicRelationalHyperlinkedModelSerializer):
    specifications_doc = JSONEditorField(schema_source='specifications_template.schema')

    class Meta:
        model = models.SAP
        fields = '__all__'


class SIPidentifierSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = models.SIPidentifier
        fields = ['unique_identifier', 'source', 'url']

