"""
This file contains the serializers (for the elsewhere defined data models)
"""

from rest_framework import serializers
from .. import models
from .scheduling import SubtaskSerializer, SubtaskExtendedSerializer
from .common import FloatDurationField, RelationalHyperlinkedModelSerializer, AbstractTemplateSerializer, AbstractSchemaTemplateSerializer, DynamicRelationalHyperlinkedModelSerializer, TaskConnectorTypeSerializer
from .widgets import JSONEditorField
from ..models import TMSSUser as User

# This is required for keeping a user reference as ForeignKey in other models
# (I think so that the HyperlinkedModelSerializer can generate a URI)
class UserSerializer(serializers.Serializer):
    class Meta:
        model = User
        fields = '__all__'


class TaskStatusSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.TaskStatus
        fields = '__all__'


class SchedulingUnitStatusSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.SchedulingUnitStatus
        fields = '__all__'


class TagsSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.Tags
        fields = '__all__'


class CommonSchemaTemplateSerializer(AbstractSchemaTemplateSerializer):
    class Meta:
        model = models.CommonSchemaTemplate
        fields = '__all__'


class SchedulingUnitObservingStrategyTemplateSerializer(AbstractTemplateSerializer):
    template = JSONEditorField(schema_source="scheduling_unit_template.schema")

    class Meta:
        model = models.SchedulingUnitObservingStrategyTemplate
        fields = '__all__'


class SchedulingSetStrategyTemplateSerializer(AbstractTemplateSerializer):
    template = JSONEditorField(schema_source="scheduling_set_template.schema")

    class Meta:
        model = models.SchedulingSetStrategyTemplate
        fields = '__all__'


class SchedulingSetTemplateSerializer(AbstractSchemaTemplateSerializer):
    class Meta:
        model = models.SchedulingSetTemplate
        fields = '__all__'


class SchedulingUnitTemplateSerializer(AbstractSchemaTemplateSerializer):
    class Meta:
        model = models.SchedulingUnitTemplate
        fields = '__all__'


class SchedulingConstraintsTemplateSerializer(AbstractSchemaTemplateSerializer):
    class Meta:
        model = models.SchedulingConstraintsTemplate
        fields = '__all__'


class SchedulingConstraintsWeightFactorSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.SchedulingConstraintsWeightFactor
        fields = '__all__'


class TaskTemplateSerializer(AbstractSchemaTemplateSerializer):
    class Meta:
        model = models.TaskTemplate
        fields = '__all__'


class TaskTemplateExtendedSerializer(TaskTemplateSerializer):
    connector_types = TaskConnectorTypeSerializer(many=True, read_only=True, help_text='The connector types which define what kind of data this task template consumes/produces.')


class TaskRelationSelectionTemplateSerializer(AbstractSchemaTemplateSerializer):
    class Meta:
        model = models.TaskRelationSelectionTemplate
        fields = '__all__'


class RoleSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Role
        fields = '__all__'

class IOTypeSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.IOType
        fields = '__all__'

class SchedulingRelationPlacementSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.SchedulingRelationPlacement
        fields = '__all__'


class DatatypeSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Datatype
        fields = '__all__'


class DataformatSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.Dataformat
        fields = '__all__'


class QuantitySerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.Quantity
        fields = '__all__'


class CycleSerializer(DynamicRelationalHyperlinkedModelSerializer):
    duration = FloatDurationField(read_only=True, help_text="Duration of the cycle [seconds]")

    class Meta:
        model = models.Cycle
        fields = '__all__'
        extra_fields = ['projects', 'name', 'duration', 'quota']
        expandable_fields = {
            'projects': ('lofar.sas.tmss.tmss.tmssapp.serializers.ProjectSerializer', {'many': True}),
            'quota': ('lofar.sas.tmss.tmss.tmssapp.serializers.CycleQuotaSerializer', {'many': True})
        }


class CycleQuotaSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.CycleQuota
        fields = '__all__'
        extra_fields = ['resource_type']

class ProjectSerializer(DynamicRelationalHyperlinkedModelSerializer):
#    scheduling_sets = serializers.PrimaryKeyRelatedField(source='scheduling_sets', read_only=True, many=True)

    class Meta:
        model = models.Project
        fields = '__all__'
        extra_fields = ['name','quota', 'nr_of_allowed_triggers', 'nr_of_used_triggers'] #, 'scheduling_sets']
        read_only_fields = ['nr_of_allowed_triggers', 'nr_of_used_triggers']

        expandable_fields = {
            'cycles': ('lofar.sas.tmss.tmss.tmssapp.serializers.CycleSerializer', {'many': True}),
            'quota': ('lofar.sas.tmss.tmss.tmssapp.serializers.ProjectQuotaSerializer', {'many': True})
        }


class ProjectQuotaArchiveLocationSerializer(DynamicRelationalHyperlinkedModelSerializer):
    queryset = models.ProjectQuotaArchiveLocation.objects.all()

    # performance boost: select the related models in a single db call.
    queryset = queryset.select_related('project_quota', 'archive_location')

    class Meta:
        model = models.ProjectQuotaArchiveLocation
        fields = '__all__'
        extra_fields = ['archive_subdirectory', 'full_archive_uri']


class ProjectQuotaSerializer(DynamicRelationalHyperlinkedModelSerializer):
    project_quota_archive_location = ProjectQuotaArchiveLocationSerializer(many=True, read_only=True)

    queryset = models.ProjectQuota.objects.all()

    # performance boost: select the related models in a single db call.
    queryset = queryset.select_related('project_quota_archive_location')

    class Meta:
        model = models.ProjectQuota
        fields = '__all__'
        extra_fields = ['resource_type']
        read_only_fields = ['project_quota_archive_location']


class ResourceTypeSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.ResourceType
        fields = '__all__'
        extra_fields = ['name']


class SystemSettingFlagSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.SystemSettingFlag
        fields = '__all__'


class SettingSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = models.Setting
        fields = '__all__'


class SubsystemStatusSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.SubsystemStatus
        fields = '__all__'


class SubsystemSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.Subsystem
        fields = '__all__'


class ProjectCategorySerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.ProjectCategory
        fields = '__all__'


class PeriodCategorySerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.PeriodCategory
        fields = '__all__'


class SchedulingSetSerializer(DynamicRelationalHyperlinkedModelSerializer):

    class Meta:
        model = models.SchedulingSet
        fields = '__all__'
        extra_fields = ['scheduling_unit_drafts']
        expandable_fields = {
            'project': 'lofar.sas.tmss.tmss.tmssapp.serializers.ProjectSerializer',
            'scheduling_unit_drafts': ('lofar.sas.tmss.tmss.tmssapp.serializers.SchedulingUnitDraftSerializer', {'many': True}),
        }


class SchedulingUnitDraftSerializer(DynamicRelationalHyperlinkedModelSerializer):
    scheduling_constraints_doc = JSONEditorField(schema_source="scheduling_constraints_template.schema")
    duration = FloatDurationField(read_only=True)
    specifications_doc = serializers.JSONField(read_only=True)

    class Meta:
        model = models.SchedulingUnitDraft
        fields = '__all__'
        read_only_fields = ['interrupts_telescope', 'specifications_doc']
        extra_fields = ['scheduling_unit_blueprints', 'task_drafts', 'duration', 'specifications_doc']
        expandable_fields = {
            'observation_strategy_template': 'lofar.sas.tmss.tmss.tmssapp.serializers.SchedulingUnitObservingStrategyTemplateSerializer',
            'specifications_template': 'lofar.sas.tmss.tmss.tmssapp.serializers.SchedulingUnitTemplateSerializer',
            'scheduling_constraints_template': 'lofar.sas.tmss.tmss.tmssapp.serializers.SchedulingConstraintsTemplateSerializer',
            'scheduling_set': 'lofar.sas.tmss.tmss.tmssapp.serializers.SchedulingSetSerializer',
            'scheduling_unit_blueprints': ('lofar.sas.tmss.tmss.tmssapp.serializers.SchedulingUnitBlueprintSerializer', {'many': True}),
            'task_drafts': ('lofar.sas.tmss.tmss.tmssapp.serializers.TaskDraftSerializer', {'many': True})
        }


class SchedulingUnitBlueprintSerializer(DynamicRelationalHyperlinkedModelSerializer):
    scheduling_constraints_doc = JSONEditorField(schema_source="scheduling_constraints_template.schema", required=False)
    duration = FloatDurationField(read_only=True)
    observed_duration = FloatDurationField(read_only=True)
    on_sky_duration = FloatDurationField(read_only=True)
    status = serializers.StringRelatedField(source='status.value', label='status')
    scheduled_central_lst = serializers.DateTimeField(read_only=True, source='main_observation_scheduled_central_lst', help_text='The scheduled central local sidereal time (Computed for CS002 from scheduled_start_time and scheduled_stop_time for the main observation task)')

    class Meta:
        model = models.SchedulingUnitBlueprint
        fields = '__all__'
        read_only_fields = ['interrupts_telescope', 'unschedulable_reason', 'error_reason']
        extra_fields = ['task_blueprints', 'output_pinned', 'unschedulable_reason', 'error_reason']
        expandable_fields = {
            'specifications_template': 'lofar.sas.tmss.tmss.tmssapp.serializers.SchedulingUnitTemplateSerializer',
            'draft': 'lofar.sas.tmss.tmss.tmssapp.serializers.SchedulingUnitDraftSerializer',
            'task_blueprints': ('lofar.sas.tmss.tmss.tmssapp.serializers.TaskBlueprintSerializer', {'many': True})
        }

    def create(self, validated_data):
        # override normal create, refresh_from_db after creating, because some fields are computed in the database and not returned by the normal serializer
        scheduling_unit_blueprint = super().create(validated_data)
        scheduling_unit_blueprint.refresh_from_db()
        return scheduling_unit_blueprint


class SchedulingUnitBlueprintSlimSerializer(serializers.ModelSerializer):
    '''A small 'slim' serializer for the most relevant properties.'''
    scheduler = serializers.CharField(source='scheduling_constraints_doc.scheduler', label='scheduler', read_only=True)
    project = serializers.StringRelatedField(source='draft.scheduling_set.project.name', label='project', read_only=True)
    antenna_set = serializers.CharField(source='main_observation_antenna_set', label='antenna_set', read_only=True)
    specified_stations = serializers.ListField(source='main_observation_specified_stations', label='specified_stations', read_only=True)
    used_stations = serializers.ListField(source='main_observation_used_stations', label='used_stations', read_only=True)
    missing_stations = serializers.ListField(source='main_observation_missing_stations', label='missing_stations', read_only=True)
    duration = FloatDurationField(read_only=True)
    on_sky_duration = FloatDurationField(read_only=True)

    class Meta:
        model = models.SchedulingUnitBlueprint
        read_only_fields = ['id',
                            'name',
                            'description',
                            'status',
                            'unschedulable_reason',
                            'error_reason',
                            'scheduler',
                            'priority_queue',
                            'scheduling_constraints_template_id',
                            'process_start_time',
                            'process_stop_time',
                            'on_sky_start_time',
                            'on_sky_stop_time',
                            'duration',
                            'on_sky_duration',
                            'project',
                            'antenna_set',
                            'specified_stations',
                            'used_stations',
                            'missing_stations']
        fields = read_only_fields

    def create(self, validated_data):
        raise NotImplementedError("This serializer should only be used for fast querying existing scheduling units")

    def update(self, instance, validated_data):
        raise NotImplementedError("This serializer should only be used for fast querying existing scheduling units")


class TaskBlueprintSlimSerializer(serializers.ModelSerializer):
    '''A small 'slim' serializer for the most relevant properties.'''
    duration = FloatDurationField(read_only=True)
    on_sky_duration = FloatDurationField(read_only=True)
    task_type = serializers.StringRelatedField(source='specifications_template.type', label='task_type', read_only=True, help_text='The task type as defined in the specifications template.')

    class Meta:
        model = models.TaskBlueprint
        read_only_fields = ['id',
                            'name',
                            'description',
                            'scheduling_unit_blueprint',
                            'status',
                            'task_type',
                            'unschedulable_reason',
                            'error_reason',
                            'process_start_time',
                            'process_stop_time',
                            'on_sky_start_time',
                            'on_sky_stop_time',
                            'duration',
                            'on_sky_duration',
                            ]
        fields = read_only_fields

    def create(self, validated_data):
        raise NotImplementedError("This serializer should only be used for fast querying existing tasks")

    def update(self, instance, validated_data):
        raise NotImplementedError("This serializer should only be used for fast querying existing tasks")


class SchedulingUnitBlueprintSlimWithTasksSerializer(SchedulingUnitBlueprintSlimSerializer):
    '''A small 'slim' serializer for the most relevant properties including the 'slim' versions of the nested tasks.'''
    task_blueprints = TaskBlueprintSlimSerializer(many=True, read_only=True)

    class Meta:
        model = SchedulingUnitBlueprintSlimSerializer.Meta.model
        read_only_fields = SchedulingUnitBlueprintSlimSerializer.Meta.read_only_fields + ['task_blueprints']
        fields = read_only_fields


class TaskDraftSerializer(DynamicRelationalHyperlinkedModelSerializer):

    duration = FloatDurationField(read_only=True)
    relative_start_time = FloatDurationField(read_only=True)
    relative_stop_time = FloatDurationField(read_only=True)
    specifications_doc = JSONEditorField(schema_source='specifications_template.schema')
    task_type = serializers.StringRelatedField(source='specifications_template.type', label='task_type', read_only=True, help_text='The task type as defined in the specifications template.')
    connector_types = TaskConnectorTypeSerializer(source='specifications_template.connector_types', label='connector_types', many=True, read_only=True, help_text='The connector types which define what kind of data this task consumes/produces.')

    class Meta:
        model = models.TaskDraft
        fields = '__all__'
        extra_fields = ['task_blueprints', 'produced_by', 'consumed_by', 'first_scheduling_relation', 'second_scheduling_relation', 'duration', 'relative_start_time', 'relative_stop_time', 'task_type']
        expandable_fields = {
            'task_blueprints': ('lofar.sas.tmss.tmss.tmssapp.serializers.TaskBlueprintSerializer', {'many': True}),
            'scheduling_unit_draft': 'lofar.sas.tmss.tmss.tmssapp.serializers.SchedulingUnitDraftSerializer',
            'specifications_template': 'lofar.sas.tmss.tmss.tmssapp.serializers.TaskTemplateSerializer'
        }


class TaskBlueprintSerializer(DynamicRelationalHyperlinkedModelSerializer):

    duration = FloatDurationField(read_only=True)
    on_sky_duration = FloatDurationField(read_only=True)
    relative_start_time = FloatDurationField(read_only=True)
    relative_stop_time = FloatDurationField(read_only=True)
    specifications_doc = JSONEditorField(schema_source='specifications_template.schema')
    task_type = serializers.StringRelatedField(source='specifications_template.type', label='task_type', read_only=True, help_text='The task type as defined in the specifications template.')
    status = serializers.StringRelatedField(source='status.value', label='status')
    scheduled_central_lst = serializers.DateTimeField(read_only=True, help_text='The scheduled central local sidereal time (Computed for CS002 from scheduled_start_time and scheduled_stop_time for observations)')

    class Meta:
        model = models.TaskBlueprint
        fields = '__all__'
        extra_fields = ['subtasks', 'produced_by', 'consumed_by', 'first_scheduling_relation', 'second_scheduling_relation',
                        'relative_start_time', 'relative_stop_time', 'task_type', 'unschedulable_reason', 'error_reason']
        read_only_fields = ['unschedulable_reason', 'error_reason']
        expandable_fields = {
            'draft': 'lofar.sas.tmss.tmss.tmssapp.serializers.TaskDraftSerializer',
            'scheduling_unit_blueprint': 'lofar.sas.tmss.tmss.tmssapp.serializers.SchedulingUnitBlueprintSerializer',
            'specifications_template': 'lofar.sas.tmss.tmss.tmssapp.serializers.TaskTemplateSerializer',
            'subtasks': ('lofar.sas.tmss.tmss.tmssapp.serializers.SubtaskSerializer', {'many': True})
        }

    def create(self, validated_data):
        # override normal create, refresh_from_db after creating, because some fields are computed in the database and not returned by the normal serializer
        taskblueprint = super().create(validated_data)
        taskblueprint.refresh_from_db()
        return taskblueprint


class TaskRelationDraftSerializer(DynamicRelationalHyperlinkedModelSerializer):
    selection_doc = JSONEditorField(schema_source='selection_template.schema')

    class Meta:
        model = models.TaskRelationDraft
        fields = '__all__'
        extra_fields = ['blueprints']


class TaskRelationDraftExtendedSerializer(TaskRelationDraftSerializer):
    input_role = TaskConnectorTypeSerializer(help_text='The connector type which defines what kind of data this subtask consumes.')
    output_role = TaskConnectorTypeSerializer(help_text='The connector type which defines what kind of data this subtask produces.')


class TaskRelationBlueprintSerializer(DynamicRelationalHyperlinkedModelSerializer):
    selection_doc = JSONEditorField(schema_source='selection_template.schema')

    class Meta:
        model = models.TaskRelationBlueprint
        fields = '__all__'


class TaskRelationBlueprintExtendedSerializer(TaskRelationBlueprintSerializer):
    input_role = TaskConnectorTypeSerializer(read_only=True, help_text='The connector type which defines what kind of data this subtask consumes.')
    output_role = TaskConnectorTypeSerializer(read_only=True, help_text='The connector type which defines what kind of data this subtask produces.')


class TaskSchedulingRelationDraftSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.TaskSchedulingRelationDraft
        fields = '__all__'
        extra_fields = ['blueprints']


class TaskSchedulingRelationBlueprintSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.TaskSchedulingRelationBlueprint
        fields = '__all__'


class TaskTypeSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.TaskType
        fields = '__all__'


class PriorityQueueTypeSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.PriorityQueueType
        fields = '__all__'


class ReservationStrategyTemplateSerializer(AbstractTemplateSerializer):
    template = JSONEditorField(schema_source="reservation_template.schema")

    class Meta:
        model = models.ReservationStrategyTemplate
        fields = '__all__'


class ReservationTemplateSerializer(AbstractSchemaTemplateSerializer):
    class Meta:
        model = models.ReservationTemplate
        fields = '__all__'


class ReservationSerializer(DynamicRelationalHyperlinkedModelSerializer):
    specifications_doc = JSONEditorField(schema_source='specifications_template.schema')

    class Meta:
        model = models.Reservation
        fields = '__all__'
        extra_fields = ['duration']


class TaskBlueprintExtendedSerializer(TaskBlueprintSerializer):
    """
    Serializes not only the task blueprints, but also a subset of the referenced objects therein so that they are
    expanded into the json response for a single API call (for convenience/optimization).

    """
    subtasks = SubtaskExtendedSerializer(many=True)  # we set many=True because this field represents a sth-to-many relationship, i.e. we serialize a list of objects here
    specifications_template = TaskTemplateExtendedSerializer()


class TaskDraftExtendedSerializer(TaskDraftSerializer):
    """
    Serializes not only the task drafts, but also a subset of the referenced objects therein so that they are
    expanded into the json response for a single API call (for convenience/optimization).
    """
    task_blueprints = TaskBlueprintExtendedSerializer(many=True)
    specifications_template = TaskTemplateExtendedSerializer()

class SchedulingUnitDraftExtendedSerializer(SchedulingUnitDraftSerializer):
    """
    Serializes not only the scheduling unit drafts, but also a subset of the referenced objects therein so that they are
    expanded into the json response for a single API call (for convenience/optimization).
    """
    task_drafts = TaskDraftExtendedSerializer(many=True)


class SchedulingUnitBlueprintExtendedSerializer(SchedulingUnitBlueprintSerializer):
    """
    Serializes not only the scheduling unit blueprints, but also a subset of the referenced objects therein so that they
    are expanded into the json response for a single API call (for convenience/optimization).
    """
    task_blueprints = TaskBlueprintExtendedSerializer(many=True)


class ProjectStateSerializer(DynamicRelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.ProjectState
        fields = '__all__'


class SystemEventTypeSerializer(DynamicRelationalHyperlinkedModelSerializer):

    class Meta:
        model = models.SystemEventType
        fields = '__all__'


class SystemEventSubtypeSerializer(DynamicRelationalHyperlinkedModelSerializer):

    class Meta:
        model = models.SystemEventSubtype
        fields = '__all__'


class SystemEventSeveritySerializer(DynamicRelationalHyperlinkedModelSerializer):

    class Meta:
        model = models.SystemEventSeverity
        fields = '__all__'


class SystemEventStatusSerializer(DynamicRelationalHyperlinkedModelSerializer):

    class Meta:
        model = models.SystemEventStatus
        fields = '__all__'


class SystemEventTemplateSerializer(AbstractSchemaTemplateSerializer):

    class Meta:
        model = models.SystemEventTemplate
        fields = '__all__'


class SystemEventSerializer(DynamicRelationalHyperlinkedModelSerializer):
    affected_hardware_doc = JSONEditorField(schema_source='affected_hardware_template.schema')
    created_by = serializers.CharField(source='created_by.email', read_only=True, default=None)

    class Meta:
        model = models.SystemEvent
        fields = '__all__'
