import os
import sys
import logging
logger = logging.getLogger(__name__)

from django.http import HttpResponse, JsonResponse, Http404, FileResponse, HttpResponseNotFound, HttpResponseServerError
from rest_framework.response import Response as RestResponse
from rest_framework import status
from django.shortcuts import get_object_or_404, render
from lofar.sas.tmss.tmss.tmssapp import models, serializers
from lofar.common.json_utils import get_default_json_object_for_schema
from lofar.common.util import single_line_with_single_spaces
from lofar.sas.tmss.tmss.tmssapp.adapters.parset import convert_to_parset
from lofar.sas.tmss.tmss.tmssapp.tasks import create_scheduling_unit_draft_from_observing_strategy_template, create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft, update_task_graph_from_specifications_doc
from lofar.common.datetimeutils import round_to_second_precision
from drf_yasg.utils import swagger_auto_schema
from drf_yasg.openapi import Parameter
from rest_framework.authtoken.models import Token
from rest_framework.permissions import AllowAny
from rest_framework.decorators import authentication_classes, permission_classes
from django.apps import apps
import re
from lofar.sas.tmss.tmss.tmssapp.serializers import SchedulingUnitDraftSerializer, SchedulingUnitBlueprintSerializer
from rest_framework.decorators import api_view, renderer_classes
from lofar.sas.tmss.tmss.tmssapp.renderers.ImagePNGRenderer import ImagePNGRenderer
from datetime import datetime, date, time, timedelta
import dateutil.parser
from astropy.coordinates import Angle
import astropy.units
from lofar.sas.tmss.tmss.tmssapp.conversions import local_sidereal_time_for_utc_and_station_as_pytime, timestamps_and_stations_to_sun_rise_set_day_night, coordinates_and_timestamps_to_separation_from_bodies, coordinates_timestamps_and_stations_to_target_rise_and_set, coordinates_timestamps_and_stations_to_target_transit, Pointing, create_location_for_station
from lofar.sas.tmss.tmss.tmssapp.adapters.keycloak import get_names_by_role_in_project
from lofar.sas.tmss.tmss.tmssapp.viewsets.permissions import get_project_roles_for_user
from lofar.sas.tmss.tmss.tmssapp.adapters.reports import create_failure_report
from distutils.util import strtobool
from lofar.sas.tmss.tmss.tmssapp.subtasks import get_calibrator_for_pointing
from lofar.sas.tmss.services.calibrator_selector import CALIBRATOR_LISTS

from django.views.decorators.cache import cache_page

# Note: Decorate with @api_view to get this picked up by Swagger

def subtask_template_default_specification(request, subtask_template_pk:int):
    subtask_template = get_object_or_404(models.SubtaskTemplate, pk=subtask_template_pk)
    spec = get_default_json_object_for_schema(subtask_template.schema)
    return JsonResponse(spec)


def task_template_default_specification(request, task_template_pk:int):
    task_template = get_object_or_404(models.TaskTemplate, pk=task_template_pk)
    spec = get_default_json_object_for_schema(task_template.schema)
    return JsonResponse(spec)


def subtask_parset(request, subtask_pk:int):
    subtask = get_object_or_404(models.Subtask, pk=subtask_pk)
    parset = convert_to_parset(subtask)
    return HttpResponse(str(parset), content_type='text/plain')


def index(request):
    return render(request, os.path.join(os.path.dirname(os.path.dirname(os.path.realpath(__file__))), '../frontend','tmss_webapp/build/index.html'))
    #return render(request, "../../../frontend/frontend_poc/build/index.html")


@api_view(['DELETE'])
def revoke_token_deauth(request, *args, **kwargs):
    token = request.META['HTTP_AUTHORIZATION'].split(" ")[1]
    invalidate_token = Token.objects.filter(key=token)
    invalidate_token.delete()
    return HttpResponse(status=204)


def authentication_state(request):
    if not request.user.is_authenticated:
        username = request.GET.get('username', None)
        if username:
            # Check websocket token against known token for given username, to hint to the frontend that the user has
            # an active session. Proposed solution from TMSS-2904.
            # Todo: I feel this is not a correct solution.
            #  The websocket token should function independently from the application login with session cookie.
            #  Review what is really going on here. I cannot replicate the described behavior on my machine. I log in
            #  to TMSS, close my browser, open it again, and my session still works without another login handshake.
            #  If this does not work on all operating systems and/or browsers then there is most likely something amiss
            #  with the request getting handled by middleware. I wonder if you are now somehow trying to fall back to
            #  the ws token to get into the application via the Token middleware, because somehow the OIDC middleware is
            #  not properly triggered.
            try:
                token = request.META['HTTP_AUTHORIZATION'].split(" ")[1]
                token_obj = Token.objects.filter(key=token).first()
                valid = (token_obj.user.username == username)
            except:
                valid = False
            return JsonResponse({'is_authenticated': False,
                                 'websocket_token_valid': valid})
        return JsonResponse({'is_authenticated': False})

    token_obj = Token.objects.filter(user=request.user).first()
    token = token_obj.key if token_obj else None

    return JsonResponse({'is_authenticated': True,
                         'username': request.user.username,
                         'email': request.user.email,
                         'id': request.user.id,
                         'websocket_token': token})


# Allow everybody to GET our publicly available template-json-schema's
@permission_classes([AllowAny])
@authentication_classes([AllowAny])
@swagger_auto_schema(#method='GET',
                     responses={200: 'Get the JSON schema from the template with the requested <template>, <name> and <version>',
                                404: 'the schema with requested <template>, <name> and <version> is not available'},
                     operation_description="Get the JSON schema for the given <template> with the given <name> and <version> as application/json content response.")
#@api_view(['GET'])   # todo: !! decorating this as api_view somehow breaks json ref resolution !! fix this and double url issue in urls.py, then use decorator here to include in Swagger
def get_template_json_schema(request, template:str, name:str, version:str=None):
    template_model = apps.get_model("tmssapp", template)
    template_instance = template_model.get_version_or_latest(name=name, version=version)
    if template_instance is None:
        return HttpResponseNotFound("Could not find template json schema with name=%s and version=%s" % (name, version))

    schema = template_instance.schema
    response = JsonResponse(schema, json_dumps_params={"indent": 2, "sort_keys": True})

    # override Cache-Control, we suggest the client to store the template in local browser cache.
    response['Cache-Control'] = "public max-age=300"

    # config Access-Control. Our schemas use $ref url's to other schemas, mainly pointing to our own common schemas with base definitions.
    # We instruct the client to allow fetching those.
    response["Access-Control-Allow-Origin"] = "*"
    response["Access-Control-Allow-Methods"] = ", ".join(set(response.get("Access-Control-Allow-Methods", '').split(', ') + ['GET', 'OPTIONS']))
    return response


# Allow everybody to GET our publicly available template-json-schema's
@permission_classes([AllowAny])
@authentication_classes([AllowAny])
@swagger_auto_schema(#method='GET',
                     responses={200: 'Get a JSON document filled with all defaults compliant with the schema from the template with the requested <template>, <name> and <version>',
                                404: 'the schema with requested <template>, <name> and <version> is not available'},
                     operation_description="Get a JSON document filled with all defaults compliant with the schema for the given <template> with the given <name> and <version> as application/json content response.")
#@api_view(['GET'])   # todo: !! decorating this as api_view somehow breaks json ref resolution !! fix this and double url issue in urls.py, then use decorator here to include in Swagger
def get_template_json_schema_default_doc(request, template:str, name:str, version:str=None):
    template_model = apps.get_model("tmssapp", template)
    template_instance = template_model.get_version_or_latest(name=name, version=version)
    default_json_doc = template_instance.get_default_json_document_for_schema()
    response = JsonResponse(default_json_doc, json_dumps_params={"indent":2, "sort_keys": True})

    # override Cache-Control, we suggest the client to store the template in local browser cache.
    response['Cache-Control'] = "public max-age=300"

    # config Access-Control. Our schemas use $ref url's to other schemas, mainly pointing to our own common schemas with base definitions.
    # We instruct the client to allow fetching those.
    response["Access-Control-Allow-Origin"] = "*"
    response["Access-Control-Allow-Methods"] = ", ".join(set(response.get("Access-Control-Allow-Methods", '').split(', ') + ['GET', 'OPTIONS']))
    return response


# Allow everybody to GET our publicly available template-json-schema's
@permission_classes([AllowAny])
@authentication_classes([AllowAny])
@swagger_auto_schema(#method='GET',
                     responses={200: 'Get a JSON document filled with all defaults compliant with the schema from the template with the requested <template>, <name> and <version>',
                                404: 'the schema with requested <template>, <name> and <version> is not available'},
                     operation_description="Get a JSON document filled with all defaults compliant with the schema for the given <template> with the given <name> and <version> as application/json content response.")
#@api_view(['GET'])   # todo: !! decorating this as api_view somehow breaks json ref resolution !! fix this and double url issue in urls.py, then use decorator here to include in Swagger
def get_template_json_ref_resolved_schema(request, template:str, name:str, version:str=None):
    template_model = apps.get_model("tmssapp", template)
    template_instance = template_model.get_version_or_latest(name=name, version=version)
    response = JsonResponse(template_instance.ref_resolved_schema, json_dumps_params={'indent': 2, "sort_keys": True})

    # override Cache-Control, we suggest the client to store the template in local browser cache.
    response['Cache-Control'] = "public max-age=300"

    # config Access-Control. Our schemas use $ref url's to other schemas, mainly pointing to our own common schemas with base definitions.
    # We instruct the client to allow fetching those.
    response["Access-Control-Allow-Origin"] = "*"
    response["Access-Control-Allow-Methods"] = ", ".join(set(response.get("Access-Control-Allow-Methods", '').split(', ') + ['GET', 'OPTIONS']))
    return response

# Allow everybody to GET our publicly available LTA SIP XSD (XML Schema Definition for the LTA SIP)
@permission_classes([AllowAny])
@authentication_classes([AllowAny])
@swagger_auto_schema(#method='GET',
                     responses={200: 'Get the LTA SIP XSD',
                                404: 'not available'},
                     operation_description="Get the LTA SIP XSD.")
#@api_view(['GET'])   # todo: !! decorating this as api_view somehow breaks json ref resolution !! fix this and double url issue in urls.py, then use decorator here to include in Swagger
@cache_page(timeout=300) # cache page here in the server, and suggest client to cache it as well
def get_lta_sip_xsd(request):
    lta_sip_xsd_path = os.path.join(os.environ["LOFARROOT"], "share", "lta", "LTA-SIP.xsd")
    with open(lta_sip_xsd_path, 'rt') as file:
        xsd = file.read()
        response = HttpResponse(content=xsd, content_type='application/xml')
        # override Cache-Control, we suggest the client to store the template in local browser cache.
        response['Cache-Control'] = "public max-age=300"
        return response


# Allow everybody to GET our publicly available station group lookups
@permission_classes([AllowAny])
@authentication_classes([AllowAny])
@swagger_auto_schema(#method='GET',
                     responses={200: 'A JSON object with two properties: group:<the_group_name>, stations:<the_list_of_stations>',
                                404: 'No such group or template available'},
                     operation_description="Get a JSON list of stations for the given <station_group> name the the group definitions in the common_schema_template given by <template_name> and <template_version>")
#@api_view(['GET'])  # todo: fix double url issue in urls.py, then use decorator here to include in Swagger
def get_stations_in_group(request, template_name:str, template_version:str, station_group:str):
    station_schema_template = get_object_or_404(models.CommonSchemaTemplate, name=template_name, version=template_version)
    station_schema = station_schema_template.schema

    if 'station_group' not in station_schema.get('definitions', {}):
        raise Http404('The JSON schema in template %s version %s has no station_group definitions' % (template_name, template_version))

    groups = station_schema['definitions']['station_group']['anyOf']
    try:
        selected_group = next(g for g in groups if g['title'].lower() == station_group.lower())
    except StopIteration:
        raise Http404('No station_group with name "%s" found in the JSON schema. template=%s version=%s' % (station_group, template_name, template_version))

    stations = selected_group['properties']['stations']['enum'][0]
    response = JsonResponse({'group': station_group, 'stations': stations})

    # override Cache-Control, we suggest the client to store the template in local browser cache.
    response['Cache-Control'] = "public max-age=300"
    return response


@permission_classes([AllowAny])
@authentication_classes([AllowAny])
@swagger_auto_schema(method='GET',
                     responses={200: 'An isoformat timestamp of the current UTC clock of the system'},
                     operation_description="Get the current system time in UTC")
@api_view(['GET'])
def utc(request):
    return HttpResponse(datetime.utcnow().isoformat(), content_type='text/plain')

@permission_classes([AllowAny])
@authentication_classes([AllowAny])
@swagger_auto_schema(method='GET',
                     responses={200: 'The LST time in hms format at the given UTC time and station or longitude'},
                     operation_description="Get LST time for UTC time and station or longitude",
                     manual_parameters=[Parameter(name='station', required=False, type='string', in_='query',
                                                  description="A station names (defaults to CS002)"),
                                        Parameter(name='timestamp', required=False, type='string', in_='query',
                                                  description="A timestamp in isoformat (defaults to utcnow)"),
                                        Parameter(name='longitude', required=False, type='string', in_='query',
                                                  description="A longitude as float")
                                        ])
@api_view(['GET'])
def lst(request):
    # Handling optional parameters via django paths in urls.py is a pain, we access them on the request directly instead.
    timestamp = request.GET.get('timestamp', None)
    stations = request.GET.get('stations', 'CS002')

    # conversions
    if timestamp:
        timestamp = dateutil.parser.parse(timestamp, ignoretz=True)  #  iso to datetime
    else:
        timestamp = datetime.utcnow()

    if isinstance(stations, str):
        stations = [s.strip() for s in stations.split(',')]

    result = { 'UTC': round_to_second_precision(timestamp).isoformat()+'Z', 'LST': {} }
    for station in stations:
        lst = local_sidereal_time_for_utc_and_station_as_pytime(timestamp, station)
        result['LST'][station] = str(lst)
    return JsonResponse(result)


@permission_classes([AllowAny])
@authentication_classes([AllowAny])
@swagger_auto_schema(method='GET',
                     responses={200: 'A JSON dict with all stations and their properties (like latitude and longitude)'},
                     operation_description="Get all LOFAR stations and their properties")
@api_view(['GET'])
def stations(request):
    from lofar.sas.tmss.tmss.tmssapp.models import CommonSchemaTemplate
    stations_template = CommonSchemaTemplate.get_latest(name="stations")
    stations = stations_template.schema['definitions']['station']['enum']
    lofar2_station_group = next(sg for sg in stations_template.schema['definitions']['station_group']['anyOf'] if sg['title'] == 'LOFAR2')
    lofar2_stations = lofar2_station_group['properties']['stations']['enum'][0]

    result = {}
    for station in stations:
        # for now, we just add the location as lat/lon in deg.
        loc = create_location_for_station(station)
        # and the station version.
        version = 2 if station in lofar2_stations else 1

        result[station] = {'latitude': loc.lat.deg, 'longitude': loc.lon.deg, 'version': version }
    return JsonResponse(result)


@permission_classes([AllowAny])
@authentication_classes([AllowAny])
@swagger_auto_schema(method='GET',
                     responses={200: 'A JSON object with sunrise, sunset, day and night of the given stations at the given timestamps'},
                     operation_description="Get sunrise, sunset, day and night for stations and timestamps.\n\n"
                                           "Example request: /api/util/sun_rise_and_set?stations=CS002,CS005&timestamps=2020-05-01,2020-09-09T11-11-00",
                     manual_parameters=[Parameter(name='stations', required=False, type='string', in_='query',
                                                  description="comma-separated list of station names"),
                                        Parameter(name='timestamps', required=False, type='string', in_='query',
                                                  description="comma-separated list of isoformat timestamps")])
@api_view(['GET'])
def get_sun_rise_and_set(request):
    """
    returns sunrise and sunset at the given stations and timestamps, or today at LOFAR core if none specified.
    example request: /api/util/sun_rise_and_set?stations=CS002,CS005&timestamps=2020-05-01,2020-09-09T11-11-00
    """
    timestamps = request.GET.get('timestamps', None)
    stations = request.GET.get('stations', None)
    if timestamps is None:
        timestamps = (datetime.utcnow(),)
    else:
        timestamps = timestamps.split(',')
        timestamps = tuple([dateutil.parser.parse(timestamp, ignoretz=True) for timestamp in timestamps])  #  isot to datetime
    if stations is None:
        stations = ("CS002",)
    else:
        stations = tuple(stations.split(','))

    return JsonResponse(timestamps_and_stations_to_sun_rise_set_day_night(timestamps, stations))


@permission_classes([AllowAny])
@authentication_classes([AllowAny])
@swagger_auto_schema(method='GET',
                     responses={200: 'A JSON object with angular distances of the given sky coordinates from the given solar system bodies at the given timestamps (seen from LOFAR core)'},
                     operation_description="Get angular distances of the given sky coordinates from the given solar system bodies at all given timestamps. \n\n"
                                           "Example request: /api/util/angular_separation?angle1=1&angle2=1&timestamps=2020-01-01T15,2020-01-01T16",
                     manual_parameters=[Parameter(name='angle1', required=True, type='string', in_='query',
                                                  description="first angle of celectial coordinates as float, e.g. RA"),
                                        Parameter(name='angle2', required=True, type='string', in_='query',
                                                  description="second angle of celectial coordinates as float, e.g. RA"),
                                        Parameter(name='direction_type', required=False, type='string', in_='query',
                                                  description="direction_type of celectial coordinates as string, e.g. J2000"),
                                        Parameter(name='timestamps', required=False, type='string', in_='query',
                                                  description="comma-separated list of isoformat timestamps"),
                                        Parameter(name='bodies', required=False, type='string', in_='query',
                                                  description="comma-separated list of solar system bodies")])
@api_view(['GET'])
def get_angular_separation(request):
    '''
    returns angular distances of the given sky coordinates from the given astronomical objects at the given timestamps and stations
    '''
    timestamps = request.GET.get('timestamps', None)
    pointing = Pointing(angle1 = request.GET.get('angle1'),
                        angle2 = request.GET.get('angle2'),
                        direction_type = request.GET.get("direction_type", "J2000"))
    bodies = tuple(request.GET.get('bodies', "sun,moon,jupiter").split(','))

    if pointing.angle1 is None or pointing.angle2 is None:
        raise ValueError("Please provide celestial coordinates via 'angle1', 'angle2' (and optionally 'direction_type') properties.")

    if timestamps is None:
        timestamps = (datetime.utcnow(),)
    else:
        timestamps = timestamps.split(',')
        timestamps = tuple([dateutil.parser.parse(timestamp, ignoretz=True) for timestamp in timestamps])  #  isot to datetime

    # calculate
    sep_dict = coordinates_and_timestamps_to_separation_from_bodies(pointing, bodies=bodies, timestamps=timestamps)
    new_sep_dict = {}

    # serialize angles and datetimes for json response
    for body, timestamps in sep_dict.items():
        for timestamp, angle in timestamps.items():
            new_sep_dict.setdefault(body, {})[timestamp.isoformat()] = angle.rad

    return JsonResponse(new_sep_dict)



@permission_classes([AllowAny])
@authentication_classes([AllowAny])
@swagger_auto_schema(method='GET',
                     responses={200: 'A JSON object with rise and set times of the given coordinates above the provided horizon, for each given station and timestamp.'},
                     operation_description="Get rise and set times of the given coordinates above the provided horizon, for each given station and timestamp. \n\n"
                                           "Example request: /api/util/target_rise_and_set?angle1=0.5&angle2=0.5&timestamps=2020-01-01T15&horizon=0.3",
                     manual_parameters=[Parameter(name='angle1', required=True, type='string', in_='query',
                                                  description="first angle of celectial coordinates as float, e.g. RA"),
                                        Parameter(name='angle2', required=True, type='string', in_='query',
                                                  description="second angle of celectial coordinates as float, e.g. RA"),
                                        Parameter(name='direction_type', required=False, type='string', in_='query',
                                                  description="direction_type of celectial coordinates as string, e.g. J2000"),
                                        Parameter(name='timestamps', required=False, type='string', in_='query',
                                                  description="comma-separated list of isoformat timestamps"),
                                        Parameter(name='stations', required=False, type='string', in_='query',
                                                  description="comma-separated list of station names"),
                                        Parameter(name='horizon', required=False, type='string', in_='query',
                                                  description="Elevation above horizon for which to return rise/set times as float")])
@api_view(['GET'])
def get_target_rise_and_set(request):
    '''
    returns rise and set times of the given coordinates above the provided horizon, for each given station and timestamp.
    '''
    timestamps = request.GET.get('timestamps', None)
    pointing = Pointing(request.GET.get('angle1'), request.GET.get('angle2'), request.GET.get("direction_type", "J2000"))
    stations = tuple(request.GET.get('stations', "CS002").split(','))
    horizon = request.GET.get('horizon', None)

    if pointing.angle1 is None or pointing.angle2 is None:
        raise ValueError("Please provide celestial coordinates via 'angle1', 'angle2' (and optionally 'direction_type') properties.")

    if timestamps is None:
        timestamps = (datetime.utcnow(),)
    else:
        timestamps = timestamps.split(',')
        timestamps = tuple([dateutil.parser.parse(timestamp, ignoretz=True) for timestamp in timestamps])  #  isot to datetime

    if horizon is None:
        horizon = Angle(0, unit=astropy.units.rad)
    else:
        horizon = Angle(horizon, unit=astropy.units.rad)

    # calculate
    rise_set_dict = coordinates_timestamps_and_stations_to_target_rise_and_set(pointing, angle_to_horizon=horizon, timestamps=timestamps, stations=stations)
    return JsonResponse(rise_set_dict)


@api_view(['GET'])
def get_target_transit(request):
    '''
    returns transit times of the given coordinates for each given station and timestamp.
    '''
    timestamps = request.GET.get('timestamps', None)
    pointing = Pointing(request.GET.get('angle1'), request.GET.get('angle2'), request.GET.get("direction_type", "J2000"))
    stations = tuple(request.GET.get('stations', "CS002").split(','))

    if pointing.angle1 is None or pointing.angle2 is None:
        raise ValueError("Please provide celestial coordinates via 'angle1', 'angle2' (and optionally 'direction_type') properties.")

    if timestamps is None:
        timestamps = (datetime.utcnow(),)
    else:
        timestamps = timestamps.split(',')
        timestamps = tuple([dateutil.parser.parse(timestamp, ignoretz=True) for timestamp in timestamps])  #  isot to datetime

    # calculate
    transit_dict = coordinates_timestamps_and_stations_to_target_transit(pointing, timestamps=timestamps, stations=stations)
    return JsonResponse(transit_dict)


@api_view(['GET'])
def get_calibrator(request):
    '''
    returns calibrator for given pointing and timestamp.
    '''
    timestamp = request.GET.get('timestamp', None)
    pointing = {'angle1': request.GET.get('angle1', None), 'angle2': request.GET.get('angle2', None), 'direction_type': request.GET.get("direction_type", "J2000")}
    min_elevation = request.GET.get('min_elevation', .5)
    calibrator_type = request.GET.get('type', None)

    if pointing['angle1'] is None or pointing['angle2'] is None:
        raise ValueError("Please provide celestial coordinates via 'angle1', 'angle2' (and optionally 'direction_type') properties.")

    if calibrator_type is None:
        raise ValueError(f"Please provide the type of calibrator you need via the 'type' property (Choices: {list(CALIBRATOR_LISTS.keys())})")

    if timestamp:
        timestamp = dateutil.parser.parse(timestamp, ignoretz=True)  #  isot to datetime
    else:
        timestamp = datetime.utcnow()

    # calculate
    calibrator = get_calibrator_for_pointing(calibrator_type, timestamp, pointing, min_elevation)
    return JsonResponse(calibrator)


@api_view(['POST'])
def submit_trigger(request):
    trigger_doc = request.data
    logger.info("Received trigger submission: %s", single_line_with_single_spaces(trigger_doc))

    # check if doc is valid
    trigger_template = models.CommonSchemaTemplate.get_latest(name="triggers")
    trigger_template.validate_document(trigger_doc)

    # gather relevant objects from db
    strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_version_or_latest(name=trigger_doc['scheduling_unit_observing_strategy_template']['name'],
                                                                                             version=trigger_doc['scheduling_unit_observing_strategy_template'].get('version'))
    scheduling_set = get_object_or_404(models.SchedulingSet, pk=trigger_doc['scheduling_set_id'])

    # check permissions
    if not scheduling_set.project.can_trigger:
        msg = 'Project \'%s\' does not allow triggers' % scheduling_set.project
        logger.error(msg)
        return RestResponse(msg, status=status.HTTP_403_FORBIDDEN)

    # check quota
    if scheduling_set.project.nr_of_allowed_triggers == 0:
        msg = 'Project \'%s\' does not have any trigger quota' % scheduling_set.project
        logger.error(msg)
        return RestResponse(msg, status=status.HTTP_403_FORBIDDEN)

    from django.db import transaction
    with transaction.atomic():
        # try to create the draft from the trigger_doc
        # we always create the draft (and a blueprint if mode=='run') and check the number of allowed triggers upon scheduling time
        scheduling_unit_draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template,
                                                                                              scheduling_set,
                                                                                              name=trigger_doc['name'],
                                                                                              description=trigger_doc.get('description'),
                                                                                              specifications_doc_overrides=trigger_doc['scheduling_unit_observing_strategy_template'].get('overrides', {}))

        # indicate that we are allowed to interrupt the telescope
        scheduling_unit_draft.interrupts_telescope = True
        scheduling_unit_draft.save()

        # if the trigger mode is 'run', then turn it into a blueprint which the dynamic scheduler will try to pick up, given the scheduling constraints
        if trigger_doc['mode'].lower() == 'run':
            scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
            return RestResponse(SchedulingUnitBlueprintSerializer(scheduling_unit_blueprint, context={'request': request}).data,
                                status=status.HTTP_201_CREATED)

        return RestResponse(SchedulingUnitDraftSerializer(scheduling_unit_draft, context={'request': request}).data,
                            status=status.HTTP_201_CREATED)

# Allow everybody to GET our publicly available template-json-schema's
@permission_classes([AllowAny])
@authentication_classes([AllowAny])
@swagger_auto_schema(#method='GET',
                     responses={200: 'a zip-file in the attachement containing all templates, grouped by type in subdirectories.'},
                     operation_description="get all templates in TMSS in a zip file, grouped by type in subdirectories.")
#@api_view(['GET'])   # todo: !! decorating this as api_view somehow breaks json ref resolution !! fix this and double url issue in urls.py, then use decorator here to include in Swagger
def get_templates_as_zipfile(request):
    '''get all templates in TMSS in a zip file, grouped by type in subdirectories.'''
    import zipfile, io, json
    from django.utils.text import re_camel_case

    # generate nice descriptive zip_filename
    hostname = request.get_host().split(':')[0]
    zip_filename = 'tmss_templates_%s_%s.zip' % (hostname, datetime.utcnow().strftime('%Y%m%dT%H%M'))

    archive = io.BytesIO()
    with zipfile.ZipFile(archive, 'w', compression=zipfile.ZIP_DEFLATED) as zip_archive:
        for template_class in models.AbstractSchemaTemplate.__subclasses__() + models.AbstractStrategyTemplate.__subclasses__():
            template_type_name = re_camel_case.sub(r'_\1', template_class.__name__).lower().lstrip('_')
            for template in template_class.objects.exclude(state__value=models.TemplateState.Choices.OBSOLETE.value).all():
                # it is ok when the name or path contains spaces, but then you get quotes around the name when extracting again.
                template_path = '%s%s/%s-%s.json' % (template_type_name,
                                                     ('/'+str(template.type) if hasattr(template, 'type') else ''),
                                                     template.name.replace(' ', '_'), # we don't like spaces in filenames. The template name including spaces is part of the document content, so we can retreive the true name from there when using it.
                                                     template.version)

                with zip_archive.open(template_path, 'w') as file:
                    # file content is the schema/template, plus the name version and description to make the file content self descriptive json.
                    content = {'name': template.name,
                               'version': template.version,
                               'description': template.description,
                               'state': template.state.value,
                               'purpose': template.purpose.value}

                    if isinstance(template, models.AbstractSchemaTemplate):
                        content['schema'] = template.schema
                    elif isinstance(template, models.AbstractStrategyTemplate):
                        # file content the json template document, plus the name version and description to make the file content self descriptive json.
                        content['template'] = template.template

                        # ugly but functional way of encoding referenced templates
                        for refereced_template_field in ('scheduling_unit_template', 'scheduling_set_template', 'reservation_template'):
                            if hasattr(template, refereced_template_field):
                                refereced_template = getattr(template, refereced_template_field)
                                content[refereced_template_field] = {'name': refereced_template.name,
                                                                     'version': refereced_template.version}

                    file.write((json.dumps(content, indent=2, sort_keys=True)+"\n").encode('utf-8'))

        # add a nice Readme
        with zip_archive.open('readme.txt', 'w') as file:
            file.writelines([('This zip archive contains all templates from TMSS running at %s.\n' % (hostname,)).encode('utf-8'),
                             b'Each directory is a template type containing json schemas or json template docs.\n',
                             b'Some directories contain an intermediate dir depicting a subtype for this specific template, for example "observation" or "pipeline" for task_template.\n',
                             b'\n',
                             b'You can unzip this archive in the TMSS source tree in the schemas dir, and then add new/updated files to git.\n',
                             b'Or you can upload these templates to any TMSS instance using tmss_populate --schemas which only uploads the new ones and/or patches the diffs.\n'])

    # rewind the BytesIO, so the django FileResponse object can read and stream it
    archive.seek(0)

    # return as a zipfile attachement
    # the django FileResponse object takes care of closing the archive BytesIO stream
    return FileResponse(archive,
                        filename=zip_filename,
                        as_attachment=True)



def response500(request):
    """
    A custom error handler for error code 500 that returns a json with the exception message as detail.
    """
    type_, value, traceback = sys.exc_info()
    data = {'detail': "%s: %s" % (value.__class__.__name__, str(value))}
    return JsonResponse(data=data, status=500)


@swagger_auto_schema(method='GET',
                     responses={200: 'For each project in TMSS, get list users with role "friend_of_project"',
                                403: 'forbidden'},
                     operation_description='For each project in TMSS, get list of users with role "friend_of_project"')
@api_view(['GET'])
def friends_of_projects(request):
    project_friends = {}
    for project in models.Project.objects.all():
        list_of_friends = get_names_by_role_in_project(models.ProjectRole.Choices.FRIEND_OF_PROJECT.value, project.name)
        project_friends[project.name] = list_of_friends
    return JsonResponse(project_friends)


@swagger_auto_schema(method='GET',
                     responses={200: 'Overview of all roles that the requesting user has in TMSS',
                                403: 'forbidden'},
                     operation_description="Overview of all roles that the requesting user has in TMSS")
@api_view(['GET'])
def my_roles(request):
    project_roles = {}
    for project_role in get_project_roles_for_user(request.user):
        project_roles.setdefault(project_role['project'], []).append(project_role['role'])
    system_roles = [group.name for group in request.user.groups.all()]
    if request.user.is_superuser:
        system_roles.append('superuser')
    return JsonResponse({'project_roles': project_roles, 'system_roles': system_roles})


@swagger_auto_schema(method='GET',
                     responses={200: 'A 24h scheduling_constraints plot (png) for all units on the given date (or today if no date given)',
                                403: 'forbidden'},
                     operation_description="Get a 24h scheduling_constraints plot (png) for all units on the given date (or today if no date given)")
@api_view(['GET'])
@renderer_classes([ImagePNGRenderer])
def daily_schedule_plot(request, datestamp: str=None):
    if datestamp is not None:
        datestamp = datetime.strptime(datestamp, '%Y-%m-%d').date()
    else:
        datestamp = datetime.utcnow().date()

    from lofar.sas.tmss.tmss.tmssapp.adapters.plots import scheduling_constraints_plot_for_date
    plot = scheduling_constraints_plot_for_date(station='CS002', # use the core's center as reference for all
                                                proposed_date=datestamp)
    return RestResponse(plot, content_type='image/png')


@swagger_auto_schema(method='GET',
                     responses={200: 'An html page with an overview of all 24h scheduling_constraints plots for all schedulable & unschedulable scheduling units on the given date (or today if no date given)',
                                403: 'forbidden'},
                     operation_description="Get an overview of 24h scheduling_constraints plots for all schedulable & unschedulable scheduling units on the given date (or today if no date given)")
@api_view(['GET'])
def daily_scheduling_constraints_overview(request, datestamp: str=None):
    from .models.specification import SchedulingUnitBlueprint, SchedulingUnitStatus
    from lofar.sas.tmss.services.scheduling.constraints import can_run_between_with_time_constraints

    if datestamp is not None:
        datestamp = datetime.strptime(datestamp, '%Y-%m-%d').date()
    else:
        datestamp = datetime.utcnow().date()

    date_start = datetime.combine(datestamp, time(0, 0, 0))
    date_end = date_start + timedelta(days=1) - timedelta(microseconds=1)
    datestamp_str = datestamp.strftime('%Y-%m-%d')

    description = '''<p>This page shows all scheduling constraints plots for the scheduling_units on %s<ul>''' % (datestamp_str,)
    table_rows = ''

    # add the daily_schedule_plot
    description += '''<li>the 24h view showing all elevations for all units in one plot</li>'''
    table_rows += '''<tr><td><img src="/api/daily_schedule_plot/%s"/></td></tr>\n''' % (datestamp_str,)

    # add all already scheduled/run scheduling_units for this date
    description += '''<li>all scheduled/observed scheduling_units on this date</li>'''
    scheduling_units = SchedulingUnitBlueprint.objects.filter(obsolete_since__isnull=True,
                                                              scheduled_start_time__lte=date_end,
                                                              scheduled_stop_time__gte=date_start)
    scheduling_units = scheduling_units.exclude(status__value__in=SchedulingUnitStatus.PRE_ACTIVE_STATUS_VALUES)
    scheduling_units = list(scheduling_units.all())

    if date_end > datetime.utcnow():
        # add all not-scheduled-yet scheduling_units
        description += '''<li>all not-scheduled-yet scheduling_units which could run on this date given their time constraints</li>'''
        scheduling_units.extend(SchedulingUnitBlueprint.objects.filter(obsolete_since__isnull=True,
                                                                       status__value__in=SchedulingUnitStatus.PRE_ACTIVE_STATUS_VALUES).all())
        scheduling_units = [su for su in scheduling_units if can_run_between_with_time_constraints(su, date_start, date_end)]


    scheduling_units = sorted(scheduling_units, key=lambda su: su.id)
    description += '''</ul></p>'''
    table_rows += '\n'.join(['''<tr><td>%s%s\n<img src="/api/scheduling_unit_blueprint/%s/scheduling_constraints_plot/%s"/></td></tr>''' % ('''%s id=<b>%s</b> name='%s' project_rank=%.4f unit_rank=%.4f CS/RS/IS=%s''' % (su.project, su.id, su.name, su.project.rank, su.rank, '/'.join(str(q) for q in su.main_observation_task.used_station_counts)),
                                                                                                                                            '' if su.status.value in SchedulingUnitStatus.PRE_ACTIVE_STATUS_VALUES else '''<br>start='%s'[UTC] center='%s'[LST] status=%s'''%(su.on_sky_start_time, su.main_observation_scheduled_central_lst, su.status.value),
                                                                                                                                            su.id,
                                                                                                                                            datestamp_str if su.status.value in SchedulingUnitStatus.PRE_ACTIVE_STATUS_VALUES else '')
                            for su in scheduling_units])
    html_doc = '''<!DOCTYPE html>
<html>
<head>
<style>
table {
  border-collapse: collapse;
  height: 100%%;
  width: 100%%;
}
tr:nth-child(even) {
  background-color: #EEEEEE;
}
tr:nth-child(odd) {
  background-color: #CCCCCC;
}
td img{
display: block;
margin-left: auto;
margin-right: auto;
}
</style>    
<title>Scheduling constraints plots for %s</title>
</head>
<body>
%s
<table>
%s
</table>
</body>
</html>
    ''' % (datestamp_str, description, table_rows)

    return HttpResponse(html_doc, content_type='text/html')


@swagger_auto_schema(method='GET',
                     responses={200: 'A 24h scheduling_constraints plot (png) for all units on the given date (or today if no date given)',
                                403: 'forbidden'},
                     operation_description="Get a 24h scheduling_constraints plot (png) for all units on the given date (or today if no date given)")
@api_view(['GET'])
@renderer_classes([ImagePNGRenderer])
def lst_pressure_plot(request, from_date: str=None, until_date: str=None):
    if from_date:
        from_date = datetime.strptime(from_date, '%Y-%m-%d')
    else:
        from_date = datetime.utcnow().replace(hour=0, minute=0, second=0, microsecond=0)

    if until_date:
        until_date = datetime.strptime(until_date, '%Y-%m-%d')

    from lofar.sas.tmss.tmss.tmssapp.adapters.plots import lst_pressure_plot
    plot = lst_pressure_plot(from_date, until_date, include_calibrators=True)
    return RestResponse(plot, content_type='image/png')


@swagger_auto_schema(method='GET',
                     responses={200: 'A 24h scheduling_constraints plot (png) for all units on the given date (or today if no date given)',
                                403: 'forbidden'},
                     operation_description="Get a 24h scheduling_constraints plot (png) for all units on the given date (or today if no date given)")
@api_view(['GET'])
@renderer_classes([ImagePNGRenderer])
def lst_target_pressure_plot(request, from_date: str=None, until_date: str=None):
    if from_date:
        from_date = datetime.strptime(from_date, '%Y-%m-%d')
    else:
        from_date = datetime.utcnow()

    if until_date:
        until_date = datetime.strptime(until_date, '%Y-%m-%d')

    from lofar.sas.tmss.tmss.tmssapp.adapters.plots import lst_pressure_plot
    plot = lst_pressure_plot(from_date, until_date, include_calibrators=False)
    return RestResponse(plot, content_type='image/png')



@swagger_auto_schema(method='GET',
                     responses={200: 'Failure Report Information as json',
                                403: 'forbidden'},
                     operation_description="Get Failure Report information for the period between the provided start and end timestamps.",
                     manual_parameters=[Parameter(name='start', required=True, type='string', in_='query',
                                                  description="Datetime in isoformat for start of reporting period"),
                                        Parameter(name='stop', required=True, type='string', in_='query',
                                                  description="Datetime in isoformat for stop of reporting period"),
                                        Parameter(name='ignore_time_range_for_histogram', required=False, type='string', in_='query',
                                                  description="for the lost time histogram, aggregate over all events known to TMSS"),
                                        ])
@api_view(['GET'])
def failure_report(request):

    # Check start and stop times
    start, stop = request.query_params.get('start', None), request.query_params.get('stop', None)
    ignore_time_range_for_histogram = strtobool(request.query_params.get('ignore_time_range_for_histogram', 'True'))
    if start:
        try:
            start = dateutil.parser.parse(start, ignoretz=True)
        except Exception:
            return RestResponse('Error: please specify an isoformat timestamp for start_time', status=400)
    if stop:
        try:
            stop = dateutil.parser.parse(stop, ignoretz=True)
        except Exception:
            return RestResponse('Error: please specify an isoformat timestamp for stop_time', status=400)

    result = create_failure_report(start, stop, ignore_time_range_for_histogram)
    return JsonResponse(result, status=status.HTTP_200_OK)


@permission_classes([AllowAny])
@authentication_classes([AllowAny])
@swagger_auto_schema(responses={200: 'A json list of the current public schedule.'},
                     operation_description="Get a json list of the current public schedule.")
def public_schedule(request):
    '''return the current and future(2 weeks) running and scheduled units in a simple html table'''
    return __create_html_schedule(request, past=False)


@permission_classes([AllowAny])
@authentication_classes([AllowAny])
@swagger_auto_schema(responses={200: 'A json list of the current public schedule.'},
                     operation_description="Get a json list of the current public schedule.")
def public_past_schedule(request):
    '''return the past schedule of observed units in the cycle in a simple html table'''
    return __create_html_schedule(request, past=True)


def __create_html_schedule(request, past: bool=False):
    # get all current and future running and scheduled units
    if 'json' in request.headers.get('Accept','').lower():
        return JsonResponse("Sorry, Json is not supported", status=status.HTTP_501_NOT_IMPLEMENTED, safe=False)

    base_url = request._current_scheme_host
    title = 'LOFAR observing schedule %sUTC' %(datetime.utcnow().strftime("%Y-%m-%d %H:%M"), )

    scheduling_units = models.SchedulingUnitBlueprint.objects.prefetch_related('task_blueprints__specifications_template__type') \
                                                             .select_related(models.SchedulingUnitBlueprint.path_to_project, 'status', 'specifications_template', 'specifications_template') \
                                                             .filter(status__value__in=models.SchedulingUnitStatus.ACTIVE_OR_FINISHED_STATUS_VALUES) \
                                                             .filter(obsolete_since__isnull=True)
    if past:
        current_cycle = models.Cycle.objects.filter(start__lte=datetime.utcnow()).order_by('-start').first()
        scheduling_units = scheduling_units.filter(on_sky_stop_time__gte=current_cycle.start) \
                                           .filter(on_sky_start_time__lte=datetime.utcnow())
        description = '''This page shows all %s's past observations. Please find the current observations <a href="%s">here</a>''' % (current_cycle.name, base_url.rstrip('/')+'/schedule')
    else:
        scheduling_units = scheduling_units.filter(on_sky_stop_time__gte=datetime.utcnow()) \
                                           .filter(on_sky_start_time__lte=datetime.utcnow()+timedelta(days=14))
        description = '''This page shows all observations for the upcoming 2 weeks. Please find all past observations <a href="%s">here</a>''' % (base_url.rstrip('/')+'/past_schedule', )
        description += '''<br>This schedule is subject to change based on changing telescope operational circumstances'''

    scheduling_units = scheduling_units.order_by('on_sky_start_time').all()

    table_rows = '''<tr><th>Project</th><th>ID</th><th>Name</th><th>Status</th><th>Start [UTC]</th><th>End [UTC]</th><th>Antenna</th><th>Target(s)</th></tr>\n'''
    row_date = date.min
    for su in scheduling_units:
        if su.on_sky_start_time.date() > row_date:
            row_date = su.on_sky_start_time.date()
            table_rows += '''<tr><td style="font-weight: bold; background-color: #999999;" colspan=9>%s</td></tr>\n''' % (row_date.strftime('%Y-%m-%d'))

        table_rows += '''<tr><td><a href="%s/project/view/%s" title="%s">%s</a></td><td><a href="%s/schedulingunit/view/blueprint/%s">%s</a></td><td>%s</td><td>%s</td><td>%s</td><td>%s</td><td>%s</td><td>%s</td></tr>\n''' % (
                                    base_url, su.project.name, 'description hidden' if su.project.private_data else su.project.description , su.project.name,
                                    base_url, su.id, su.id,
                                    'hidden' if su.project.private_data else su.name, su.status.value,
                                    su.on_sky_start_time.strftime('%H:%M') if su.on_sky_start_time else '',
                                    su.on_sky_stop_time.strftime('%H:%M') if su.on_sky_stop_time else '',
                                    ', '.join(sorted(list([q for q in set([x['specifications_doc__station_configuration__antenna_set'] for x in su.observation_tasks.values('specifications_doc__station_configuration__antenna_set')]) if q is not None]))),
                                    'hidden' if su.project.private_data else su.target_pointings_astropy)

    html_doc = '''<!DOCTYPE html>
    <html>
    <head>
    <style>
    table {
      border-collapse: collapse;
      height: 100%%;
      width: 100%%;
    }
    th {
      background-color: #999999;
      font-weight: bold; 
      text-align: left; 
      position: sticky;
      top: 0px;
    }
    tr:nth-child(even) {
      background-color: #EEEEEE;
    }
    tr:nth-child(odd) {
      background-color: #CCCCCC;
    }
    tr:hover {
      background-color: #DDDDDD;
    }     
    </style>    
    <title>LOFAR observing schedule %sUTC</title>
    <meta http-equiv="refresh" content="300">
    </head>
    <body>
    <h1>%s</h1>
    <p>%s</p>
    <table>
    %s
    </table>
    </body>
    </html>
        ''' % (title,
               title,
               description,
               table_rows,)

    return HttpResponse(html_doc, content_type='text/html')

