"""
A viewset extended with annotation and other common generic behaviours.

Adds the following functionality:
- Swagger API annotation
"""

import logging
logger = logging.getLogger(__name__)

from rest_framework import viewsets
from rest_framework.permissions import AllowAny
from rest_framework.decorators import authentication_classes, permission_classes
from drf_yasg.utils import swagger_auto_schema
from rest_framework import mixins
import json
from django.shortcuts import get_object_or_404
from django.utils.decorators import method_decorator
from django.views.decorators.cache import never_cache
from django.http import JsonResponse
from django.urls import reverse as revese_url
from rest_framework.decorators import action
from lofar.common import json_utils
from lofar.sas.tmss.tmss.tmssapp.viewsets.metadata import LOFARDefaultMetadata
from lofar.sas.tmss.tmss.tmssapp.viewsets.permissions import TMSSPermissions, IsProjectMemberFilterBackend, get_project_roles_for_user, get_project_roles_with_permission
from lofar.sas.tmss.tmss.tmssapp.models import permissions, CommonSchemaTemplate
from lofar.sas.tmss.tmss.tmssapp.models.common import TemplateState, TemplatePurpose
from django_filters.rest_framework import DjangoFilterBackend, FilterSet, CharFilter
from django_filters import filterset
from django_filters import rest_framework
from rest_framework.filters import OrderingFilter
import django_property_filter as property_filters
from django.contrib.postgres.fields import JSONField, ArrayField
from django.db.models.fields import CharField, TextField
from copy import deepcopy

FILTER_OVERRIDES = {
        JSONField: {
            'filter_class': CharFilter,
            'extra': lambda f: {'lookup_expr': 'icontains'}
        },
        ArrayField: {
            'filter_class': CharFilter,
            'extra': lambda f: {'lookup_expr': 'icontains'}
        },
        CharField: {
            'filter_class': CharFilter,
            'extra': lambda f: {'lookup_expr': 'icontains'}
        },
        TextField: {
            'filter_class': CharFilter,
            'extra': lambda f: {'lookup_expr': 'icontains'}
        }
    }

class LOFARDefaultFilterSet(FilterSet):
    FILTER_DEFAULTS = deepcopy(filterset.FILTER_FOR_DBFIELD_DEFAULTS)
    FILTER_DEFAULTS.update(FILTER_OVERRIDES)


class LOFARFilterBackend(DjangoFilterBackend):
    default_filter_set = LOFARDefaultFilterSet


class LOFARViewSet(viewsets.ModelViewSet):
    """
    If you're using format suffixes, make sure to also include
    the `format=None` keyword argument for each action.
    """
    permission_classes = (TMSSPermissions,)
    metadata_class = LOFARDefaultMetadata
    filter_backends = (LOFARFilterBackend, OrderingFilter, IsProjectMemberFilterBackend,)
    filter_fields = '__all__'
    ordering_fields = '__all__'

    # Override extra_action_permission_specs to change the way that project-based permission checks are performed on
    # extra actions on this viewset. This is required when an action operates on a different model and the permission
    # check needs to check the there-referenced project. Or when the reference is handed as a url parameter instead pf
    # the POST data. Format:
    # {<extra_action_name>: {'model': models.<_name_of_model_this_action_operates_on>,
    #                        'project_ref_override_pk_url_param': <url_parameter_name>,
    #                        'project_ref_override_attr': <path_to_project_attribute_name>,
    #                        'project_ref_override_model': models.<name_of_model_the_path_to_project_attribute_refers_to>}}
    extra_action_permission_specs = {}

    @staticmethod
    def _user_has_permission_for_any_project(user, method, permission_name):
                permitted_project_roles = get_project_roles_with_permission(permission_name, method)
                user_project_roles = set([project_role['role'] for project_role in get_project_roles_for_user(user)])
                for role in permissions.ProjectRole.objects.filter(value__in=list(user_project_roles)).all():
                    if role in permitted_project_roles:
                        return True
                return False

    def _get_permitted_extra_actions(self, request):
        # determined the extra actions a user is allowed to perform (these may have different permissions than the base
        # model, so we list which ones are allowed next to the allowed REST methods. This should be fine as actions only
        # support either POST or GET so there is no ambiguity.)

        permission_name = self.serializer_class.Meta.model.__name__.lower()
        allowed_extra_actions = []
        extra_actions = self.get_extra_actions()
        actual_method = request.method  # keep the actually requested method so that we can set it when we are done
        actual_action = self.action  # keep the actually requested action so that we can set it when we are done
        for extra_action in extra_actions:
            action = extra_action.__name__
            for method in extra_action.mapping:
                self.action = action  # pretend to do something else and check if we are allowed to do that
                request.method = method.upper()  # pretend to do something else and check if we are allowed to do that
                if TMSSPermissions().has_permission(request=request, view=self):
                    if action not in allowed_extra_actions:
                        allowed_extra_actions.append(action)
                # show extra action as permitted if the user is allowed to perform it in case the  correct project is
                # referenced (which then only works if the user posts data that links the object to the right project).
                if actual_action == 'list' and action not in allowed_extra_actions:
                    action_permission_name = '%s-%s' % (permission_name, action)
                    if self._user_has_permission_for_any_project(request.user, method.upper(), action_permission_name):
                            allowed_extra_actions.append(action)
        self.action = actual_action
        request.method = actual_method

        return allowed_extra_actions

    def _get_permitted_methods(self, request, pk=None):
        # Django returns an "Allow" header that reflects what methods the model supports in principle, but not what
        # the current user is actually has permission to perform. We use the "Access-Control-Allow-Methods" header
        # to disclose read/write permission to the frontend, so that it can render its views accordingly.
        allowed_methods = []

        # check allowed REST methods
        actual_method = request.method   # keep the actually requested method so that we can set it when we are done
        for method in ['GET', 'POST']:
            request.method = method  # pretend to do something else and check if we are allowed to do that
            if TMSSPermissions().has_permission(request=request, view=self):
                allowed_methods.append(method)
                if method == "POST":
                    # optimization: we grant POST/PUT/PATCH/DELETE permissions en block, so we only need to check one
                    # of them and skip the expensive actual permission check for the rest. Note: this only affects the
                    # header and the permission will still be properly checked when the user performs one of these.
                    allowed_methods += ['PUT', 'PATCH', 'DELETE']
            elif pk is not None:
                obj = get_object_or_404(self.queryset.model, pk=pk)
                if TMSSPermissions().has_object_permission(request=request, view=self, obj=obj):
                    allowed_methods.append(method)
        request.method = actual_method

        permission_name = self.serializer_class.Meta.model.__name__.lower()

        # add POST permission if project permission allows creation if correct project is referenced
        if self.action == 'list' and 'POST' not in allowed_methods:
            # has_permission on list actions only returns true if there is a system permission that allows POSTing
            # generally. Hence we need to check if there is the theoretical possibility of POSTing an object based
            # on project permissions (which then only works if the user posts data that links the object to the
            # right project).
            if self._user_has_permission_for_any_project(request.user, 'POST', permission_name):
                allowed_methods.append('POST')
                # optimization: we grant POST/PUT/PATCH/DELETE permissions en block, so we only need to check one
                # of them and skip the expensive actual permission check for the rest. Note: this only affects the
                # header and the permission will still be properly checked when the user performs one of these.
                allowed_methods += ['PUT', 'PATCH', 'DELETE']

        allowed_methods += self._get_permitted_extra_actions(request)
        return allowed_methods

    @swagger_auto_schema(responses={403: 'forbidden'})
    @method_decorator(never_cache) # by default, do not cache any lofar viewset, unless overridden in the subclass viewset
    def list(self, request, **kwargs):
        response = super(LOFARViewSet, self).list(request, **kwargs)
        if "Access-Control-Allow-Methods" not in response:
            response["Access-Control-Allow-Methods"] = ", ".join(self._get_permitted_methods(request))
        return response

    @swagger_auto_schema(responses={400: 'invalid specification', 403: 'forbidden'})
    def create(self, request, **kwargs):
        return super(LOFARViewSet, self).create(request, **kwargs)

    @swagger_auto_schema(responses={403: 'forbidden', 404: 'not found'})
    @method_decorator(never_cache) # by default, do not cache any lofar viewset, unless overridden in the subclass viewset
    def retrieve(self, request, pk=None, **kwargs):
        response = super(LOFARViewSet, self).retrieve(request, pk, **kwargs)
        if "Access-Control-Allow-Methods" not in response:
            response["Access-Control-Allow-Methods"] = ", ".join(self._get_permitted_methods(request, pk))
        return response

    @swagger_auto_schema(responses={400: 'invalid specification', 403: 'forbidden', 404: 'not found'})
    def update(self, request, pk=None, **kwargs):
        return super(LOFARViewSet, self).update(request, pk, **kwargs)

    @swagger_auto_schema(responses={400: 'invalid specification', 403: 'forbidden', 404: 'not found'})
    def partial_update(self, request, pk=None, **kwargs):
        return super(LOFARViewSet, self).partial_update(request, pk, **kwargs)

    @swagger_auto_schema(responses={403: 'forbidden', 404: 'not found'})
    def destroy(self, request, pk=None, **kwargs):
        return super(LOFARViewSet, self).destroy(request, pk, **kwargs)

class LOFARNestedViewSet(mixins.CreateModelMixin,
                                mixins.ListModelMixin,
                                #mixins.RetrieveModelMixin,
                                viewsets.GenericViewSet):

    @swagger_auto_schema(responses={403: 'forbidden'})
    def list(self, request, **kwargs):
        return super(LOFARNestedViewSet, self).list(request, **kwargs)

    @swagger_auto_schema(responses={400: 'invalid specification', 403: 'forbidden'})
    def create(self, request, **kwargs):
        return super(LOFARNestedViewSet, self).create(request, **kwargs)


class NumberInFilter(rest_framework.BaseInFilter, rest_framework.NumberFilter):
    """
    Custom filter that accepts a comma-separated lists of numbers.
    Returns objects where the (single) numeric value of the filtered field is in the provided list.
    """
    pass

class AbstractTemplatePropertyFilter(property_filters.PropertyFilterSet):
    state = rest_framework.ModelMultipleChoiceFilter(field_name='state', queryset=TemplateState.objects.all())
    purpose = rest_framework.ModelMultipleChoiceFilter(field_name='purpose', queryset=TemplatePurpose.objects.all())
    name = rest_framework.CharFilter(field_name='name', lookup_expr='iexact')
    version = NumberInFilter(field_name='version', lookup_expr='in')
    version_min = rest_framework.NumberFilter(field_name='version', lookup_expr='gte')
    version_max = rest_framework.NumberFilter(field_name='version', lookup_expr='lte')

    class Meta:
        pass


class AbstractTemplateViewSet(LOFARViewSet):
    filter_class = AbstractTemplatePropertyFilter    # note that this breaks other filter backends from LOFARViewSet

    def list(self, request, **kwargs):
        response = super().list(request, **kwargs)
        # override Cache-Control, we suggest the client to store the template in local browser cache.
        response['Cache-Control'] = "public max-age=300"
        return response

    def retrieve(self, request, pk=None, **kwargs):
        response = super().retrieve(request, pk, **kwargs)

        # override Cache-Control, we suggest the client to store the template in local browser cache.
        response['Cache-Control'] = "public max-age=300"

        # config Access-Control. Our schemas use $ref url's to other schemas, mainly pointing to our own common schemas with base definitions.
        # We instruct the client to allow fetching those.
        response["Access-Control-Allow-Origin"] = "*"
        response["Access-Control-Allow-Methods"] = ", ".join(set(response.get("Access-Control-Allow-Methods", '').split(', ') + ['GET', 'OPTIONS']))
        return response


class AbstractSchemaTemplateViewSet(AbstractTemplateViewSet):

    @swagger_auto_schema(responses={200: 'The schema as a JSON object',
                                    403: 'forbidden'},
                         operation_description="Get the schema as a JSON object.")
    @action(methods=['get'], detail=True)
    def schema(self, request, pk=None):
        template = get_object_or_404(self.queryset.model, pk=pk)
        response = JsonResponse(template.schema, json_dumps_params={'indent': 2, 'sort_keys': True})

        # override Cache-Control, we suggest the client to store the template in local browser cache.
        response['Cache-Control'] = "public max-age=300"
        return response


    @swagger_auto_schema(responses={200: 'The schema as a JSON object',
                                    403: 'forbidden'},
                         operation_description="Get the schema as a JSON object with all $ref URL's resolved, resulting in a complete and self describing schema.")
    @action(methods=['get'], detail=True)
    def ref_resolved_schema(self, request, pk=None):
        template = get_object_or_404(self.queryset.model, pk=pk)
        response = JsonResponse(template.ref_resolved_schema, json_dumps_params={'indent': 2, 'sort_keys': True})

        # override Cache-Control, we suggest the client to store the template in local browser cache.
        response['Cache-Control'] = "public max-age=300"
        return response

    @swagger_auto_schema(responses={200: 'JSON object with all the defaults from the schema filled in',
                                    403: 'forbidden'},
                         operation_description="Get a JSON object with all the defaults from the schema filled in.")
    @action(methods=['get'], detail=True)
    def default(self, request, pk=None):
        template = get_object_or_404(self.queryset.model, pk=pk)
        spec = template.get_default_json_document_for_schema()
        response = JsonResponse(spec, json_dumps_params={'indent': 2, 'sort_keys': True})

        # override Cache-Control, we suggest the client to store the template in local browser cache.
        response['Cache-Control'] = "public max-age=300"
        return response


    # Allow everybody to validate their docs against our publicly available templates
    @permission_classes([AllowAny])
    @authentication_classes([AllowAny])
    @swagger_auto_schema(responses={200: 'a validation json document'},
                         operation_description="do a full validation of the submitted specifications_doc, both checking the document against its json schema(s), and doing a full 'sanity check' if this document is observable.")
    @action(methods=['put'], detail=True, url_name='validate_specifications_doc')
    def validate_specifications_doc(self, request, pk=None):
        validation_template = CommonSchemaTemplate.get_latest('validation')
        validation_doc = validation_template.get_default_json_document_for_schema()
        try:
            template = get_object_or_404(self.queryset.model, pk=pk)

            if 'specifications_doc' not in request.data:
                validation_doc['valid'] = False
                validation_doc['message'] = "No specifications_doc submitted"
            else:
                specifications_doc = request.data['specifications_doc']
                validation_doc['specifications_doc'] = specifications_doc

                # try to validate the specifications_doc against its template's schema and rules.
                # if not valid, an exception is raised, which is handled below.
                template.validate_document(specifications_doc)

                # ok, initial simple json schema validation passed
                validation_doc['valid'] = True
                validation_doc['message'] = "specifications_doc is valid"
        except Exception as e:
            validation_doc['valid'] = False
            validation_doc['message'] = "Could not validate the specifications_doc. Error: " + str(e)

        logger.log(level=logging.DEBUG if validation_doc['valid'] else logging.ERROR, msg=validation_doc['message'])

        return JsonResponse(validation_doc, json_dumps_params={'indent': 2, 'sort_keys': True})

