"""
This file contains permissions and filters that are used in the viewsets
"""

from rest_framework import permissions as drf_permissions, filters as drf_filters
from .. import models
from lofar.sas.tmss.tmss.exceptions import *
from django.core.exceptions import ObjectDoesNotExist, FieldError
import logging
logger = logging.getLogger(__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)
from django.urls import resolve
import urllib.parse
from lofar.common import isProductionEnvironment, isDevelopmentEnvironment
from django.db.models import Q
from django.contrib.auth import get_user_model
User = get_user_model()
from django.shortcuts import get_object_or_404

#
# Permissions
#
# Note: These prevent accessing the object directly, and return a nice permission error.
#       However, object permissions are not considered on a listing by default, so make sure to either
#       A. apply a filter to prevent object without permission to be included in the list with full details, or
#       B. customize get_queryset on the view to call check_object_permissions.


def get_project_roles_for_user(user):

        if not isProductionEnvironment():
            # todo: stupid hack to make test pass, because we so far have failed mocking this function out successfully.
            try:
                if user == User.objects.get(username='paulus'):
                    return ({'project': 'test_user_is_shared_support', 'role': 'shared_support'},
                            {'project': 'test_user_is_contact', 'role': 'contact'},
                            {'project': 'test_user_is_friend', 'role': 'friend_of_project'})
            except Exception as e:
                pass

        if not isProductionEnvironment():
            # todo: remove these test users when not needed any more (required for local frontend testing where Keycloak is unavailable)
            if user == User.objects.get(username='tmss_friend'):
                return ({'project': 'high', 'role': 'friend_of_project'},
                        {'project': 'high', 'role': 'friend_of_project_primary'},
                        {'project': 'high', 'role': 'co_i'})
            if user == User.objects.get(username='tmss_friend2'):
                return ({'project': 'high', 'role': 'friend_of_project'},
                        {'project': 'high', 'role': 'co_i'})
            if user == User.objects.get(username='tmss_contact'):
                return ({'project': 'high', 'role': 'contact'},
                        {'project': 'high', 'role': 'co_i'})
            if user == User.objects.get(username='tmss_PI'):
                return ({'project': 'high', 'role': 'pi'},
                        {'project': 'high', 'role': 'co_i'})
            if user == User.objects.get(username='tmss_shared_support'):
                return ({'project': 'high', 'role': 'shared_support'},
                        {'project': 'high', 'role': 'shared_support_primary'},
                        {'project': 'high', 'role': 'co_i'})
            if user == User.objects.get(username='tmss_shared_support2'):
                return ({'project': 'high', 'role': 'shared_support'},
                        {'project': 'high', 'role': 'co_i'})
            if user == User.objects.get(username='tmss_scientist'):
                return ({'project': 'high', 'role': 'co_i'},)
        try:
            return tuple(user.project_roles)
        except (AttributeError, TypeError):
            # the user is a non-TMSSUser, for example anonyous/not-logged-in, or project_roles are None
            # return empty project roles
            return tuple()


def get_project_roles_with_permission(permission_name, method='GET'):
    try:
        logger.info('checking permission name=%s method=%s' % (permission_name, method))
        # ...retrieve ProjectPermission object
        project_permission = models.ProjectPermission.objects.get(name=permission_name)
        # ...determine what project roles are allowed to perform the requested action
        return getattr(project_permission, method).all()
    except ObjectDoesNotExist:
        logger.error("This action was configured to enforce project permissions, but no project permission with name '%s' has been defined." % permission_name)
        return []


class IsProjectMember(drf_permissions.DjangoObjectPermissions):
    """
    Object-level permission to only allow users of the related project to access it.
    Note: instance must have a project attribute.

    Define a filter_project_roles attribute on the view to further restrict access to a list of specific project roles.
    (Some user role background can be found here https://support.astron.nl/confluence/display/TMSS/User+roles).
    """
    def has_object_permission(self, request, view, obj):
        # GET detail, PATCH, and DELETE
        # we always have permission as superuser (e.g. in test environment, where a regular user is created to test permission specifically)
        if request.user.is_superuser:
            logger.debug("IsProjectMember: User=%s is superuser. Not enforcing project permissions!" % request.user)
            logger.debug('IsProjectMember.has_object_permission %s %s True' % (request._request, request.method))
            return True

        # todo: do we want to restrict access for that as well? Then we add it to the ProjectPermission model, but it seems cumbersome...?
        if request.method == 'OPTIONS':
            return True

        # determine which roles are allowed to access this object...
        permission_name = view.serializer_class.Meta.model.__name__.lower()
        if view.action not in ['retrieve', 'update', 'partial_update', 'destroy', 'list', 'create', None]:
            # Add suffix for extra (= non-REST) actions
            # Note: adding extra permission entries with this dash-notation allows to add both new actions as well as
            # new project roles without a DB change.
            # We could also add extra fields to the ProjectPermission model (which would be irrelevant in most cases),
            # or have project roles as fields and list all permitted actions as field values (REST and extra actions
            # in same ArrayField, which is probably the straight-forward choice, but has the downside that it requires
            # a DB change when adding a new project role.
            permission_name += '-%s' % view.action  # todo: not sure about the delimiter, dot breaks DRF unfortunately.

        permitted_project_roles = get_project_roles_with_permission(permission_name, request.method)

        # determine what project roles a user has
        user_project_roles = get_project_roles_for_user(request.user)

        # check whether the related project of this object is one that the user has permission to see
        related_project = getattr(obj, 'project', None)
        for project_role in user_project_roles:
            if related_project:
                if project_role['project'].lower() == related_project.name.lower() and \
                        models.ProjectRole.objects.get(value=project_role['role']) in permitted_project_roles:
                    logger.debug('user=%s is permitted to access object=%s' % (request.user, obj))
                    logger.debug('IsProjectMember.has_object_permission %s %s True' % (request._request, request.method))
                    return True
            else:
                logger.debug("No project property on object %s, so cannot check project permission." % obj)
                # todo: how to deal with objects that do not have a unique project associated to them?
                #  Do need users need the required role in all of them? Or just one?
                #  Also, we need to support lists in 'path_to_project' if we want to handle this.
                if issubclass(type(obj), models.AbstractTemplate) and view.action == 'default':
                    # todo: review this. Why is the default action called in a GET of an object that has a template reference?
                    logger.warning("'%s' is a Template and action is '%s' so granting object access nonetheless." % (obj, view.action))
                    return True

        logger.debug('User=%s is not permitted to access object=%s with related project=%s since it requires one of project_roles=%s' % (request.user, obj, related_project, permitted_project_roles))
        logger.debug('IsProjectMember.has_object_permission %s False' % (request._request))
        return False

    def has_permission(self, request, view):
        # GET list and POST

        # we have to generally allow OPTIONS in list view so users that only have a project
        # permission (and hence no general GET permission on the endpoint) can call OPTIONS.
        if request.method == 'OPTIONS':
            return True

        # We need to check the project name before the object is created, but project is an object property.
        # Turning the path to the project into a (static) model attribute, allows us to use it both in the object
        # property as well as here (where we did not create an object yet).
        # Note: the Viewflow actions are currently create actions but they do not actually create anything
        # (and do not contain the expected data to create an instance of the underlying model, but instead e.g. the
        # data required to perform a workflow step). So we exclude the offending views here.
        regular_create_action = view.action == 'create' and request.data \
                and not 'SchedulingUnitTaskExecuteViewSet' in str(view) \
                and not 'SchedulingUnitTaskAssignViewSet' in str(view)
        extra_action_specs = view.extra_action_permission_specs.get(view.action, None) if hasattr(view, 'extra_action_permission_specs') else None
        create_via_extra_action = extra_action_specs and (request.data or
                                                          extra_action_specs['project_ref_override_pk_url_param'] in request.query_params)

        if regular_create_action or create_via_extra_action:
            obj = None

            # Some extra actions create objects on other models. By default, we use the project reference of the base
            # model of the action to check whether the user has one of the required roles in this project. On base models
            # that are project-independent (e.g. SchedulingUnitObservingStrategyTemplate), we instead check the target
            # model for the path to project.
            model = extra_action_specs.get('model') if create_via_extra_action else view.serializer_class.Meta.model

            if model == models.Project:
                return False  # project creation solely depends on system role
            if hasattr(model, 'path_to_project'):
                attrs = model.path_to_project.split('__')
            elif hasattr(model, 'project'):
                attrs = ['project']
            else:
                raise AttributeError(f'The model {model} requires an attribute "project" or "path_to_project" to check project permissions.')
            for attr in attrs:
                obj_resolved_in_this_iteration = False
                if extra_action_specs \
                        and attr == extra_action_specs.get('project_ref_override_attr') and \
                                    extra_action_specs.get('project_ref_override_pk_url_param') in request.query_params:
                    # some extra actions do not carry all info in their POSTed data, but include a url parameter
                    # reference to the related project. Use that to resolve the project here.
                    obj = get_object_or_404(extra_action_specs.get('project_ref_override_model'),
                                            pk=request.query_params[extra_action_specs.get('project_ref_override_pk_url_param')])
                    continue
                if not obj:
                    # on first iteration, the referenced object needs to be resolved from POSTed FQDN
                    obj_ref = request.data[attr]
                    path = urllib.parse.urlparse(obj_ref).path
                    resolved_func, _, resolved_kwargs = resolve(path)
                    obj = resolved_func.cls().get_queryset().get(pk=resolved_kwargs['pk'])
                    obj_resolved_in_this_iteration = True
                if attr == 'project':
                    # has_object_permission checks the project from obj, so we can just check project permission on
                    # something that has the correct project attribute
                    p = self.has_object_permission(request, view, obj)
                    return p
                if not obj_resolved_in_this_iteration:
                    obj = getattr(obj, attr)

        pk = view.kwargs.get('pk', None)
        if pk:
            obj = view.serializer_class.Meta.model.objects.get(pk=pk)
            p = self.has_object_permission(request, view, obj)
        elif view.action == 'list':
            p = super().has_permission(request, view)
        else:
            p = False
        logger.debug('IsProjectMember has_permission view=%s action=%s -> %s' % (view, view.action, p))
        return p


class IsProjectMemberOrReadOnly(IsProjectMember):
    """
    Object-level permission to only allow users of the related project to modify it.
    Note: instance must have a project attribute.

    Define a filter_project_roles attribute on the view to further restrict access to a list of specific project roles.
    (Some user role background can be found here https://support.astron.nl/confluence/display/TMSS/User+roles).
    """
    def has_object_permission(self, request, view, obj):

        # generally allow reading, otherwise check permissions:
        if request.method in drf_permissions.SAFE_METHODS:
            return True
        else:
            return super().has_object_permission(request, view, obj)


class TMSSDjangoModelPermissions(drf_permissions.DjangoModelPermissions):
    """
    Modify the vanilla DjangoModelPermissions, which are apparently readonly by default
    """
    view_permissions = ['%(app_label)s.view_%(model_name)s']
    perms_map = {
        'GET': view_permissions,
        'OPTIONS': view_permissions,
        'HEAD': view_permissions,
        'POST': drf_permissions.DjangoModelPermissions.perms_map['POST'],
        'PUT': drf_permissions.DjangoModelPermissions.perms_map['PUT'],
        'PATCH': drf_permissions.DjangoModelPermissions.perms_map['PATCH'],
        'DELETE': drf_permissions.DjangoModelPermissions.perms_map['DELETE'],
    }

    def has_permission(self, request, view):
        extra_actions = [a.__name__ for a in view.get_extra_actions()]
        if view.action in extra_actions:
            permission_name = f'{view.action}_{view.serializer_class.Meta.model.__name__.lower()}'
            p = request.user.has_perm(f'tmssapp.{permission_name}')
        else:
            p = super().has_permission(request, view)
        return p


class TMSSPermissions(drf_permissions.DjangoObjectPermissions):
        """
        Create custom permission class
        Note: required because the composition using & and | in the permission_classes does not seem to work as it should.
        """
        project_permissions = IsProjectMember()
        model_permissions = TMSSDjangoModelPermissions()

        def has_permission(self, request, view):
            return (self.model_permissions.has_permission(request, view) or
                    self.project_permissions.has_permission(request, view)) and request.user.is_authenticated

        def has_object_permission(self, request, view, obj):
            return (self.model_permissions.has_permission(request, view) or
                    self.project_permissions.has_object_permission(request, view, obj)) and request.user.is_authenticated


#
# Custom Filters
#

class IsProjectMemberFilterBackend(drf_filters.BaseFilterBackend):
    """
    Filter that only allows users to see objects that are related to their projects.

    This only returns objects that belong to a project for which the requesting user has the required project role.
    Define a filter_project_roles attribute on the view to further restrict access to a list of specific project roles.
    (Some user role background can be found here https://support.astron.nl/confluence/display/TMSS/User+roles).
    """
    def filter_queryset(self, request, queryset, view):

        # Note that while filtering excludes certain items from the list view as expected, it also means that no
        # permission errors are raised when we do not restrict filtering to list view, and 'not found' is returned
        # instead, which might be confusing. So we want to explicitly only filter listings here:
        if view.action != 'list':
            return queryset

        # if a system role allows general access to the model, do not filter
        if TMSSDjangoModelPermissions().has_permission(request, view):
            return queryset

        # we don't filer for superuser (e.g. in test environment, where a regular user is created to test filtering specifically)
        if request.user.is_superuser:
            logger.debug("IsProjectMemberFilterBackend: User=%s is superuser. Not enforcing project permissions!" % request.user)
            return queryset

        # Early exit if the objects of this model are not related to particular projects. At this point, the user does
        # not have general model access, and since nothing project-related can be returned, return an empty list.
        if not (hasattr(view.serializer_class.Meta.model, 'path_to_project') or hasattr(view.serializer_class.Meta.model, 'project')):
            logger.debug(f'The model {view.serializer_class.Meta.model} requires an attribute "project" or "path_to_project" to check project permissions.')
            return queryset.none()

        # determine what project roles a user has
        user_project_roles = get_project_roles_for_user(request.user)

        # determine which project roles are allowed to access this object...
        permission_name = view.serializer_class.Meta.model.__name__.lower()
        filter_project_roles = get_project_roles_with_permission(permission_name, request.method)
        filter_project_role_names = [role.value for role in filter_project_roles]

        # determine all objects with related project in which the requesting user has one of the required project roles
        permitted_projects = None
        permitted_project_names = [project_role['project'] for project_role in user_project_roles if project_role['role'] in filter_project_role_names]
        if permitted_project_names:
            filter_query = Q()
            for permitted_project_name in permitted_project_names:
                filter_query |= Q(name__iexact=permitted_project_name)
            permitted_projects = models.Project.objects.filter(filter_query)
        else:
            permitted_projects = []

        if permitted_projects:
            if view.serializer_class.Meta.model == models.Project:
                return permitted_projects
            
            if hasattr(view.serializer_class.Meta.model, 'path_to_project'):
                path_to_project = view.serializer_class.Meta.model.path_to_project
                try:
                    return queryset.filter(**{f"{path_to_project}__in": permitted_projects})
                except FieldError as e:
                    # this is expected for models.schedulingunitflow.SchedulingUnitTask
                    logger.warning('Cannot perform in-db filtering for project permissions on model=%s with path_to_project=%s: %s  -- applying slow filtering based on project property.' % (view.serializer_class.Meta.model, path_to_project, e))
                    ids = []
                    for candidate in queryset.all():
                        if candidate.project in permitted_projects:
                            ids.append(candidate.id)
                    return queryset.filter(id__in=ids)

            if hasattr(view.serializer_class.Meta.model, 'project'):
                return queryset.filter(project__in=permitted_projects)

        return queryset.none()
