"""
This file contains the viewsets (based on the elsewhere defined data models and serializers)
"""


import logging
logger = logging.getLogger(__name__)

from django.shortcuts import get_object_or_404
from rest_framework import viewsets
from .lofar_viewset import LOFARViewSet, LOFARNestedViewSet, AbstractSchemaTemplateViewSet, FILTER_OVERRIDES
from .specification import NumberInFilter
from .. import models
from .. import serializers
from django_filters import rest_framework as filters
from lofar.sas.tmss.tmss.tmssapp.models.scheduling import Subtask
# Don't use Ordering class from the django_filters but use rest_framework instead !!
from rest_framework.filters import OrderingFilter
from drf_yasg import openapi
from drf_yasg.utils import swagger_auto_schema
from drf_yasg.inspectors import SwaggerAutoSchema
from drf_yasg.openapi import Parameter

from django.utils.decorators import method_decorator
from django.views.decorators.cache import cache_page, cache_control, never_cache

from rest_framework.decorators import action
from rest_framework.decorators import permission_classes
from django.http import HttpResponse, JsonResponse, HttpResponseRedirect, HttpResponseNotFound
from rest_framework.response import Response as RestResponse

from lofar.common import isProductionEnvironment, isTestEnvironment
from lofar.sas.tmss.tmss.tmssapp.viewsets.lofar_viewset import LOFARViewSet
from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.tmssapp import serializers
from lofar.sas.tmss.tmss.exceptions import TMSSException

from datetime import datetime
from lofar.common.json_utils import get_default_json_object_for_schema
from lofar.common.datetimeutils import formatDatetime
from lofar.sas.tmss.tmss.tmssapp.adapters.parset import convert_to_parset
from lofar.sas.tmss.tmss.tmssapp.adapters.l2stationspecs import convert_to_l2stationspecs_dict
from drf_yasg.renderers import _SpecRenderer
import json

from lofar.sas.tmss.tmss.tmssapp.renderers import PlainTextRenderer
from rest_framework.views import APIView
from rest_framework.decorators import api_view, renderer_classes
from django.core.exceptions import ObjectDoesNotExist
import django_property_filter as property_filters
from django.db import transaction


class TextPlainAutoSchema(SwaggerAutoSchema):
    def get_produces(self):
        return ["text/plain"]


class SubtaskStateViewSet(LOFARViewSet):
    queryset = models.SubtaskState.objects.all()
    serializer_class = serializers.SubtaskStateSerializer


class SubtaskAllowedStateTransitionsViewSet(LOFARViewSet):
    queryset = models.SubtaskAllowedStateTransitions.objects.all()
    serializer_class = serializers.SubtaskAllowedStateTransitionsSerializer


class SubtaskStateLogViewSet(LOFARViewSet):
    queryset = models.SubtaskStateLog.objects.all()
    serializer_class = serializers.SubtaskStateLogSerializer

    def get_queryset(self):
        if 'subtask_id' in self.kwargs:
            subtask = get_object_or_404(models.Subtask, pk=self.kwargs['subtask_id'])
            return subtask.subtaskstatelog_set.order_by('created_at').all()

        queryset = models.SubtaskStateLog.objects.all()

        # query by subtask
        subtask = self.request.query_params.get('subtask', None)
        if subtask is not None:
            return queryset.filter(subtask=subtask)

        return queryset


class SubtaskTypeViewSet(LOFARViewSet):
    queryset = models.SubtaskType.objects.all()
    serializer_class = serializers.SubtaskTypeSerializer

class StationTypeViewSet(LOFARViewSet):
    queryset = models.StationType.objects.all()
    serializer_class = serializers.StationTypeSerializer

class HashAlgorithmViewSet(LOFARViewSet):
    queryset = models.HashAlgorithm.objects.all()
    serializer_class = serializers.HashAlgorithmSerializer

class SubtaskTemplateViewSet(AbstractSchemaTemplateViewSet):
    queryset = models.SubtaskTemplate.objects.all()
    serializer_class = serializers.SubtaskTemplateSerializer

    def get_queryset(self):
        queryset = models.SubtaskTemplate.objects.all()

        # query by name
        name = self.request.query_params.get('name', None)
        if name is not None:
            return queryset.filter(name=name)

        return queryset


class DataproductSpecificationsTemplateViewSet(AbstractSchemaTemplateViewSet):
    queryset = models.DataproductSpecificationsTemplate.objects.all()
    serializer_class = serializers.DataproductSpecificationsTemplateSerializer


class DataproductFeedbackTemplateViewSet(AbstractSchemaTemplateViewSet):
    queryset = models.DataproductFeedbackTemplate.objects.all()
    serializer_class = serializers.DataproductFeedbackTemplateSerializer


class SubtaskFeedbackTemplateViewSet(AbstractSchemaTemplateViewSet):
    queryset = models.SubtaskFeedbackTemplate.objects.all()
    serializer_class = serializers.SubtaskFeedbackTemplateSerializer


class SubTaskFilter(property_filters.PropertyFilterSet):
    id = NumberInFilter(field_name='id', lookup_expr='in')
    id_min = filters.NumberFilter(field_name='id', lookup_expr='gte')
    id_max = filters.NumberFilter(field_name='id', lookup_expr='lte')
    state = filters.ModelMultipleChoiceFilter(field_name='state', queryset=models.SubtaskState.objects.all())
    name = filters.CharFilter(field_name='task_blueprint__scheduling_unit_blueprint__name', lookup_expr='icontains')   # todo: correct name?
    error_reason = filters.CharFilter(field_name='error_reason', lookup_expr='icontains')
    on_sky_start_time__lt = filters.IsoDateTimeFilter(field_name='on_sky_start_time', lookup_expr='lt')
    on_sky_start_time__gt = filters.IsoDateTimeFilter(field_name='on_sky_start_time', lookup_expr='gt')
    on_sky_stop_time__lt = filters.IsoDateTimeFilter(field_name='on_sky_stop_time', lookup_expr='lt')
    on_sky_stop_time__gt = filters.IsoDateTimeFilter(field_name='on_sky_stop_time', lookup_expr='gt')
    process_start_time__lt = filters.IsoDateTimeFilter(field_name='process_start_time', lookup_expr='lt')
    process_start_time__gt = filters.IsoDateTimeFilter(field_name='process_start_time', lookup_expr='gt')
    process_stop_time__lt = filters.IsoDateTimeFilter(field_name='process_stop_time', lookup_expr='lt')
    process_stop_time__gt = filters.IsoDateTimeFilter(field_name='process_stop_time', lookup_expr='gt')
    scheduled_start_time__lt = filters.IsoDateTimeFilter(field_name='scheduled_start_time', lookup_expr='lt')
    scheduled_start_time__gt = filters.IsoDateTimeFilter(field_name='scheduled_start_time', lookup_expr='gt')
    scheduled_stop_time__lt = filters.IsoDateTimeFilter(field_name='scheduled_stop_time', lookup_expr='lt')
    scheduled_stop_time__gt = filters.IsoDateTimeFilter(field_name='scheduled_stop_time', lookup_expr='gt')
    actual_on_sky_start_time__lt = filters.IsoDateTimeFilter(field_name='actual_on_sky_start_time', lookup_expr='lt')
    actual_on_sky_start_time__gt = filters.IsoDateTimeFilter(field_name='actual_on_sky_start_time', lookup_expr='gt')
    actual_on_sky_stop_time__lt = filters.IsoDateTimeFilter(field_name='actual_on_sky_stop_time', lookup_expr='lt')
    actual_on_sky_stop_time__gt = filters.IsoDateTimeFilter(field_name='actual_on_sky_stop_time', lookup_expr='gt')
    actual_process_start_time__lt = filters.IsoDateTimeFilter(field_name='actual_process_start_time', lookup_expr='lt')
    actual_process_start_time__gt = filters.IsoDateTimeFilter(field_name='actual_process_start_time', lookup_expr='gt')
    actual_process_stop_time__lt = filters.IsoDateTimeFilter(field_name='actual_process_stop_time', lookup_expr='lt')
    actual_process_stop_time__gt = filters.IsoDateTimeFilter(field_name='actual_process_stop_time', lookup_expr='gt')

    project = filters.ModelMultipleChoiceFilter(field_name=models.Subtask.path_to_project, label='project', queryset=models.Project.objects.all())
    subtask_type = filters.ModelMultipleChoiceFilter(field_name='specifications_template__type', queryset=models.SubtaskType.objects.all())
    is_using_lofar2_stations = property_filters.PropertyBooleanFilter(field_name='is_using_lofar2_stations')

    class Meta:
        model = Subtask
        fields = {
            'state__value': ['exact'],
            'cluster__name': ['exact', 'icontains'],
        }
        filter_overrides = FILTER_OVERRIDES


class SubtaskViewSet(LOFARViewSet):
    queryset = models.Subtask.objects.all()
    serializer_class = serializers.SubtaskSerializer
    filter_class = SubTaskFilter
    ordering = ('scheduled_start_time',)

    # performance boost: select the related models in a single db call.
    queryset = queryset.select_related('state', 'specifications_template', 'specifications_template__type', 'cluster', 'created_or_updated_by_user')

    @swagger_auto_schema(auto_schema=TextPlainAutoSchema,
                         responses={200: 'A LOFAR parset for this subtask (as plain text)',
                                    403: 'forbidden',
                                    404: 'Not found'},
                         produces='text/plain',
                         operation_description="Get a LOFAR parset for the specifications of this subtask")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, renderer_classes=[PlainTextRenderer])
    def parset(self, request, pk=None):
        subtask = get_object_or_404(models.Subtask, pk=pk)
        parset = convert_to_parset(subtask)

        header = "# THIS PARSET WAS GENERATED BY TMSS FROM THE SPECIFICATION OF SUBTASK ID=%d ON %s\n" % (subtask.pk, formatDatetime(datetime.utcnow()))
        parset_str = header + str(parset).replace('"','').replace("'","") # remove quotes
        return HttpResponse(parset_str, content_type='text/plain')


    @swagger_auto_schema(responses={200: 'The predecessor subtasks of this subtask',
                                    403: 'forbidden'},
                         operation_description="Get the predecessor subtasks of this subtask.")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, url_name="predecessors")
    def predecessors(self, request, pk=None):
        subtask = get_object_or_404(models.Subtask, pk=pk)
        predecessors = self.filter_queryset(subtask.predecessors)
        serializer = self.get_serializer(predecessors, many=True)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The successor subtasks of this subtask',
                                    403: 'forbidden'},
                         operation_description="Get the successor subtasks of this subtask.")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, url_name="successors")
    def successors(self, request, pk=None):
        subtask = get_object_or_404(models.Subtask, pk=pk)
        successors = self.filter_queryset(subtask.successors)
        serializer = self.get_serializer(successors, many=True)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The scheduled version of this subtask',
                                    403: 'forbidden',
                                    500: 'The subtask could not be scheduled'},
                         operation_description="Try to schedule this subtask.")
    @action(methods=['post'], detail=True, url_name="schedule")
    def schedule(self, request, pk=None):
        subtask = get_object_or_404(models.Subtask, pk=pk)
        from lofar.sas.tmss.tmss.tmssapp.subtasks import schedule_subtask_and_update_successor_start_times
        scheduled_subtask = schedule_subtask_and_update_successor_start_times(subtask)
        serializer = self.get_serializer(scheduled_subtask)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The unscheduled version of this subtask',
                                    403: 'forbidden',
                                    500: 'The subtask could not be unscheduled'},
                         operation_description="Try to unschedule this subtask, deleting all the output dataproducts and setting status back to 'defined'.")
    @action(methods=['post'], detail=True, url_name="unschedule")
    def unschedule(self, request, pk=None):
        subtask = get_object_or_404(models.Subtask, pk=pk)
        from lofar.sas.tmss.tmss.tmssapp.subtasks import unschedule_subtask
        unscheduled_subtask = unschedule_subtask(subtask)
        serializer = self.get_serializer(unscheduled_subtask)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The cancelled version of this subtask',
                                    403: 'forbidden',
                                    500: 'The subtask could not be cancelled'},
                         operation_description="Try to cancel this subtask.")
    @action(methods=['post'], detail=True, url_name="cancel")
    def cancel(self, request, pk=None):
        subtask = get_object_or_404(models.Subtask, pk=pk)
        from lofar.sas.tmss.tmss.tmssapp.subtasks import cancel_subtask
        cancelled_subtask = cancel_subtask(subtask)
        serializer = self.get_serializer(cancelled_subtask)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The obsolete version of this subtask',
                                    403: 'forbidden',
                                    500: 'The subtask could not be marked as obsolete'},
                         operation_description="Try to mark this subtask as obsolete.")
    @action(methods=['post'], detail=True, url_name="mark_as_obsolete")
    def mark_as_obsolete(self, request, pk=None):
        subtask = get_object_or_404(models.Subtask, pk=pk)
        subtask.obsolete_since = datetime.utcnow()
        subtask.save()
        serializer = self.get_serializer(subtask)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The restarted version of this subtask',
                                    403: 'forbidden',
                                    500: 'The subtask could not be restarted'},
                         operation_description="Try to restart this subtask when it is on hold.")
    @action(methods=['post'], detail=True, url_name="restart_when_on_hold")
    def restart_when_on_hold(self, request, pk=None):
        subtask = get_object_or_404(models.Subtask, pk=pk)
        from lofar.sas.tmss.tmss.tmssapp.subtasks import restart_on_hold_subtask
        restarted_subtask = restart_on_hold_subtask(subtask)
        serializer = self.get_serializer(restarted_subtask)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The state log for this Subtask.',
                                    403: 'forbidden'},
                         operation_description="Get the state log for this Subtask.")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True)
    def state_log(self, request, pk=None):
        subtask = get_object_or_404(models.Subtask, pk=pk)
        serializer = serializers.SubtaskStateLogSerializer(subtask.subtaskstatelog_set.order_by('created_at').all(),
                                                           many=True,
                                                           context={'request': request})
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The input dataproducts of this subtask.',
                                    403: 'forbidden'},
                         operation_description="Get the input dataproducts of this subtask.")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, url_name='input_dataproducts')
    def input_dataproducts(self, request, pk=None):
        subtask = get_object_or_404(models.Subtask, pk=pk)
        # fetch the subtask's input_dataproducts and all related properties in one queryset. Major speedup.
        input_dataproducts = subtask.input_dataproducts.select_related('dataformat', 'datatype', 'specifications_template', 'producer', 'feedback_template', 'sap', 'global_identifier', 'archive_info').prefetch_related('consumers')
        serializer = serializers.DataproductSerializer(input_dataproducts, many=True, context={'request': request})
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The output dataproducts of this subtask.',
                                    403: 'forbidden'},
                         operation_description="Get the output dataproducts of this subtask.")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, url_name='output_dataproducts')
    def output_dataproducts(self, request, pk=None):
        subtask = get_object_or_404(models.Subtask, pk=pk)
        # fetch the subtask's output_dataproducts and all related properties in one queryset. Major speedup.
        output_dataproducts = subtask.output_dataproducts.select_related('dataformat', 'datatype', 'specifications_template', 'producer', 'feedback_template', 'sap', 'global_identifier', 'archive_info').prefetch_related('consumers')
        serializer = serializers.DataproductSerializer(output_dataproducts, many=True, context={'request': request})
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'Returns the (updated) subtask itself.',
                                    403: 'forbidden'},
                         operation_description="Mark all output dataproducts of this subtask as deleted.")
    @action(methods=['post'], detail=True, url_name='mark_output_dataproducts_as_deleted')
    def mark_output_dataproducts_as_deleted(self, request, pk=None):
        with transaction.atomic():
            now = datetime.utcnow()
            subtask = get_object_or_404(models.Subtask, pk=pk)
            for dp in subtask.output_dataproducts.all():
                dp.deleted_since = now
                dp.save()

        subtask.refresh_from_db()
        serializer = self.get_serializer(subtask)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The transformed output dataproduct of this subtask for the given input_dataproduct_id.',
                                    403: 'forbidden'},
                         operation_description='Getthe transformed output dataproducts for the given input_dataproduct_id.',
                         manual_parameters=[Parameter(name='input_dataproduct_id', required=True, type='integer', in_='query',
                                                      description="the id of the input dataproduct for which you want to get the transformed output dataproduct")])
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, url_name='transformed_output_dataproduct')
    def transformed_output_dataproduct(self, request, pk=None):
        '''return the transformed output dataproduct for the given input_dataproduct_id.'''
        subtask = get_object_or_404(models.Subtask, pk=pk)
        input_dataproduct_id = request.query_params['input_dataproduct_id']
        try:
            output_dataproduct = subtask.get_transformed_output_dataproduct(input_dataproduct_id)
        except models.Dataproduct.DoesNotExist:
            return HttpResponseNotFound('Cannot find transformed output dataproduct for subtask id=%s and input_dataproduct_id=%s.' % (pk, input_dataproduct_id))

        serializer = serializers.DataproductSerializer(output_dataproduct, many=False, context={'request': request})
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The updated (maybe) finished version of this subtask.',
                                    403: 'forbidden',
                                    500: 'The feedback for this subtask could not be processed'},
                         operation_description="Process the feedback_doc (which can be for one or more or all dataproducts), and process it into json feedback per dataproduct. Sets the subtask to finished if all dataproducts are processed, which may require multiple postings of partial feedback docs.")
    @action(methods=['post'], detail=True, url_path='process_feedback_and_set_to_finished_if_complete')
    def process_feedback_and_set_to_finished_if_complete(self, request, pk=None):
        from lofar.sas.tmss.tmss.tmssapp.adapters.feedback import process_feedback_for_subtask_and_set_to_finished_if_complete
        subtask = get_object_or_404(models.Subtask, pk=pk)
        feedback_doc = request.body.decode('utf-8')
        updated_subtask = process_feedback_for_subtask_and_set_to_finished_if_complete(subtask, feedback_doc)
        serializer = self.get_serializer(updated_subtask)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'Get progress of this subtask ranging from 0.0 when just started up to 1.0 when finished.',
                                    403: 'forbidden',
                                    500: 'Could not compute the progress'},
                         operation_description="Get progress of this subtask ranging from 0.0 when just started up to 1.0 when finished.")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, url_name='get_progress')
    def get_progress(self, request, pk=None):
        subtask = get_object_or_404(models.Subtask, pk=pk)
        return JsonResponse({'id': subtask.id, 'progress': subtask.progress})


    @swagger_auto_schema(responses={200: 'Get the list of stations that were requested in the task specification, but which were not available to the scheduled observation.',
                                    403: 'forbidden',
                                    500: 'Server Error'},
                         operation_description="Get the list of stations that were requested in the task specification, but which were not available to the scheduled observation.")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, url_name='missing_stations')
    def missing_stations(self, request, pk=None):
        from lofar.sas.tmss.tmss.tmssapp.subtasks import get_missing_stations
        subtask = get_object_or_404(models.Subtask, pk=pk)
        missing_stations = get_missing_stations(subtask)
        return JsonResponse({'missing_stations': missing_stations})

    @swagger_auto_schema(auto_schema=TextPlainAutoSchema,
                         responses={200: 'A LOFAR 2.0 station spec for this subtask (as json)',
                                    403: 'forbidden',
                                    404: 'Not found'},
                         operation_description="Get a LOFAR 2.0 station spec for this subtask")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, url_name='l2stationspecs')
    def l2stationspecs(self, request, pk=None):
        subtask = get_object_or_404(models.Subtask, pk=pk)
        l2stationspecs = convert_to_l2stationspecs_dict(subtask)
        return JsonResponse(l2stationspecs, json_dumps_params={'indent': 2, 'sort_keys': True})


    @swagger_auto_schema(auto_schema=TextPlainAutoSchema,
                         responses={200: 'A LOFAR feedback document in parset format for this observation subtask (as plain text)',
                                    403: 'forbidden',
                                    404: 'Not found'},
                         produces='text/plain',
                         operation_description="Get a LOFAR feedback document in parset format (as plain text) for this observation subtask, generated from the specification.")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, renderer_classes=[PlainTextRenderer])
    def feedback_from_specification(self, request, pk=None):
        from lofar.sas.tmss.tmss.tmssapp.adapters.feedback import create_feedback_for_observation_subtask_from_specifications
        subtask = get_object_or_404(models.Subtask, pk=pk)
        feedback_parset = create_feedback_for_observation_subtask_from_specifications(subtask)
        feedback_parset = str(feedback_parset).replace('"', '').replace("'", "")
        feedback_parset = '\n'.join(sorted(feedback_parset.split('\n')))
        header = "# THIS FEEDBACK PARSET WAS GENERATED BY TMSS FROM THE SPECIFICATION OF SUBTASK ID=%d ON %s\n" % (subtask.pk, formatDatetime(datetime.utcnow().replace(microsecond=0)))
        feedback_parset_str = header + feedback_parset # remove quotes
        return HttpResponse(feedback_parset_str, content_type='text/plain')


    @swagger_auto_schema(responses={200: 'The finished version of this subtask with processed feedback.',
                                    403: 'forbidden',
                                    500: 'The feedback for this subtask could not be generated and/or processed'},
                         operation_description="Create a feedback document for this observation subtask generated from the specification, and process it resulting in a finished subtask. Use only for support/maintenance!")
    @action(methods=['post'], detail=True, url_path='create_and_process_feedback_from_specification_and_set_to_finished')
    def create_and_process_feedback_from_specification_and_set_to_finished(self, request, pk=None):
        from lofar.sas.tmss.tmss.tmssapp.adapters.feedback import create_and_use_feedback_for_observation_subtask_from_specifications
        subtask = get_object_or_404(models.Subtask, pk=pk)
        logger.info("create_and_process_feedback_from_specification_and_set_to_finished for subtask id=%s state=%s", subtask.id, subtask.state.value)
        with transaction.atomic():
            if subtask.state.value != models.SubtaskState.Choices.FINISHING.value:
                from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions
                subtask = set_subtask_state_following_allowed_transitions(subtask, models.SubtaskState.Choices.FINISHING.value)

            updated_subtask = create_and_use_feedback_for_observation_subtask_from_specifications(subtask)
            serializer = self.get_serializer(updated_subtask)
            return RestResponse(serializer.data)



class SubtaskExtendedViewSet(SubtaskViewSet):
    serializer_class = serializers.SubtaskExtendedSerializer


class SubtaskNestedViewSet(LOFARNestedViewSet):
    queryset = models.Subtask.objects.all()
    serializer_class = serializers.SubtaskSerializer
    filter_class = SubTaskFilter
    ordering = ('scheduled_start_time',)

    # performance boost: select the related models in a single db call.
    queryset = queryset.select_related('state', 'specifications_template', 'specifications_template__type', 'cluster', 'created_or_updated_by_user')

    def get_queryset(self):
        if 'task_blueprint_id' in self.kwargs:
            task_blueprint = get_object_or_404(models.TaskBlueprint, pk=self.kwargs['task_blueprint_id'])
            return task_blueprint.subtasks.all()


class SubtaskInputViewSet(LOFARViewSet):
    queryset = models.SubtaskInput.objects.all()
    serializer_class = serializers.SubtaskInputSerializer


class SubtaskInputExtendedViewSet(SubtaskInputViewSet):
    serializer_class = serializers.SubtaskInputExtendedSerializer


class SubtaskOutputViewSet(LOFARViewSet):
    queryset = models.SubtaskOutput.objects.all()
    serializer_class = serializers.SubtaskOutputSerializer


class SubtaskOutputExtendedViewSet(SubtaskOutputViewSet):
    serializer_class = serializers.SubtaskOutputExtendedSerializer


class DataproductViewSet(LOFARViewSet):
    queryset = models.Dataproduct.objects.all()
    serializer_class = serializers.DataproductSerializer

    # performance boost: select the related models in a single db call.
    queryset = queryset.select_related('dataformat', 'datatype', 'specifications_template', 'feedback_template', 'sap', 'global_identifier')

    @swagger_auto_schema(responses={200: 'The SIP for this dataproduct',
                                    403: 'forbidden'},
                         produces='application/xml',
                         operation_description="Get the Submission Information Package (SIP) for this dataproduct")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, url_name="sip")
    def sip(self, request, pk=None):
        dataproduct = get_object_or_404(models.Dataproduct, pk=pk)
        return HttpResponse(dataproduct.sip_xml_doc, content_type='application/xml')


    @swagger_auto_schema(responses={200: 'The SIP graph for this dataproduct',
                                    403: 'forbidden'},
                         operation_description="Get a visualization of the Submission Information Package (SIP) for this dataproduct")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, url_name="sip_graph")
    def sip_graph(self, request, pk=None):
        dataproduct = get_object_or_404(models.Dataproduct, pk=pk)
        from lofar.lta.sip import visualizer
        from lofar.sas.tmss.tmss.tmssapp.adapters.sip import generate_sip_for_dataproduct
        return HttpResponse(visualizer.visualize_sip(generate_sip_for_dataproduct(dataproduct)), content_type='image/svg+xml')


    @swagger_auto_schema(responses={200: 'The updated dataproduct.',
                                    403: 'forbidden',
                                    500: 'The feedback of this subtask could not be processed'},
                         operation_description="Store the archive information for this dataproduct.")
    @action(methods=['post'], detail=True, url_name='post_archive_information')
    def post_archive_information(self, request, pk=None):
        dataproduct = get_object_or_404(models.Dataproduct, pk=pk)

        json_doc = json.loads(request.body.decode('utf-8'))
        dataproduct.store_archive_information(storage_ticket=json_doc['storage_ticket'],
                                              size=int(json_doc['file_size']),
                                              filepath=json_doc['srm_url'],
                                              md5_hash=json_doc['md5_checksum'],
                                              adler32_hash=json_doc['adler32_checksum'])

        dataproduct.refresh_from_db()
        serializer = self.get_serializer(dataproduct)
        return RestResponse(serializer.data)

    @swagger_auto_schema(responses={200: 'The updated dataproduct.',
                                    403: 'forbidden',
                                    500: 'Could not delete archive information for dataproduct'},
                         operation_description="Delete the archive information for this dataproduct.")
    @action(methods=['post'], detail=True, url_name='delete_archive_information')
    def delete_archive_information(self, request, pk=None):
        dataproduct = get_object_or_404(models.Dataproduct, pk=pk)
        dataproduct.delete_hashes_and_archive_information()
        serializer = self.get_serializer(dataproduct)
        return RestResponse(serializer.data)


class AntennaSetViewSet(LOFARViewSet):
    queryset = models.AntennaSet.objects.all()
    serializer_class = serializers.AntennaSetSerializer


class DataproductTransformViewSet(LOFARViewSet):
    queryset = models.DataproductTransform.objects.all()
    serializer_class = serializers.DataproductTransformSerializer


class FilesystemViewSet(LOFARViewSet):
    queryset = models.Filesystem.objects.all()
    serializer_class = serializers.FilesystemSerializer


class ClusterViewSet(LOFARViewSet):
    queryset = models.Cluster.objects.all()
    serializer_class = serializers.ClusterSerializer


class DataproductArchiveInfoViewSet(LOFARViewSet):
    queryset = models.DataproductArchiveInfo.objects.all()
    serializer_class = serializers.DataproductArchiveInfoSerializer


class DataproductHashViewSet(LOFARViewSet):
    queryset = models.DataproductHash.objects.all()
    serializer_class = serializers.DataproductHashSerializer


class SAPViewSet(LOFARViewSet):
    queryset = models.SAP.objects.all()
    serializer_class = serializers.SAPSerializer

    @swagger_auto_schema(responses={200: 'The dataproducts in this SAP',
                                    403: 'forbidden'},
                         operation_description="The dataproducts in this SAP.")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, url_name='dataproducts')
    def dataproducts(self, request, pk=None):
        sap = get_object_or_404(models.SAP, pk=pk)
        serializer = serializers.DataproductSerializer(sap.dataproducts, many=True, context={'request': request})
        return RestResponse(serializer.data)


class SAPTemplateViewSet(AbstractSchemaTemplateViewSet):
    queryset = models.SAPTemplate.objects.all()
    serializer_class = serializers.SAPTemplateSerializer


class SIPidentifierViewSet(LOFARViewSet):
    queryset = models.SIPidentifier.objects.all()
    serializer_class = serializers.SIPidentifierSerializer

