"""
This file contains the viewsets (based on the elsewhere defined data models and serializers)
"""

from django.shortcuts import get_object_or_404, get_list_or_404, render

from django.contrib.auth import get_user_model
User = get_user_model()
from django_filters import rest_framework as filters
from django_filters import fields as filter_fields
from django_filters.constants import EMPTY_VALUES
import django_property_filter as property_filters
from rest_framework.viewsets import ReadOnlyModelViewSet
from rest_framework import status
from rest_framework.response import Response

from django.http import (
    Http404, HttpResponse, HttpResponsePermanentRedirect, HttpResponseRedirect,
)

from distutils.util import strtobool
from rest_framework.decorators import permission_classes
from rest_framework.permissions import IsAuthenticated
from rest_framework.decorators import action
from rest_framework.response import Response as RestResponse
from rest_framework.request import QueryDict, Request as RestRequest

from django.utils.decorators import method_decorator
from django.views.decorators.cache import cache_page, cache_control, never_cache

from drf_yasg.utils import swagger_auto_schema
from drf_yasg.openapi import Parameter
from drf_yasg import openapi

from lofar.sas.tmss.tmss.tmssapp.renderers.ImagePNGRenderer import ImagePNGRenderer
from lofar.sas.tmss.tmss.tmssapp.viewsets.lofar_viewset import LOFARViewSet, LOFARNestedViewSet, AbstractTemplateViewSet, AbstractSchemaTemplateViewSet, LOFARFilterBackend, FILTER_OVERRIDES, NumberInFilter
from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.tmssapp import serializers
from lofar.sas.tmss.tmss.tmssapp.adapters.reports import create_cycle_report, create_project_report
from lofar.sas.tmss.tmss.tmssapp.adapters.keycloak import get_names_by_role_in_project
from django.http import JsonResponse, HttpResponseBadRequest

from datetime import datetime
from lofar.common.json_utils import get_default_json_object_for_schema
from lofar.common.datetimeutils import formatDatetime
from lofar.sas.tmss.tmss.tmssapp.tasks import *
from lofar.sas.tmss.tmss.tmssapp.subtasks import *
from lofar.sas.tmss.tmss.tmssapp.viewsets.permissions import TMSSDjangoModelPermissions, get_project_roles_for_user

from django.urls  import resolve, get_script_prefix,Resolver404
from rest_framework.filters import OrderingFilter

import json
import logging
import dateutil

from django.core.exceptions import ObjectDoesNotExist
from django import forms


logger = logging.getLogger(__name__)


# This is required for keeping a user reference as ForeignKey in other models
# (I think so that the HyperlinkedModelSerializer can generate a URI)
class UserViewSet(ReadOnlyModelViewSet):
    queryset = User.objects.all()
    serializer_class = serializers.UserSerializer


class TagsViewSet(LOFARViewSet):
    queryset = models.Tags.objects.all()
    serializer_class = serializers.TagsSerializer


class CommonSchemaTemplateViewSet(AbstractSchemaTemplateViewSet):
    queryset = models.CommonSchemaTemplate.objects.all()
    serializer_class = serializers.CommonSchemaTemplateSerializer


class TaskStatusViewSet(LOFARViewSet):
    queryset = models.TaskStatus.objects.all()
    serializer_class = serializers.TaskStatusSerializer


class SchedulingUnitStatusViewSet(LOFARViewSet):
    queryset = models.SchedulingUnitStatus.objects.all()
    serializer_class = serializers.SchedulingUnitStatusSerializer


class SchedulingSetStrategyTemplateViewSet(AbstractTemplateViewSet):
    queryset = models.SchedulingSetStrategyTemplate.objects.all()
    serializer_class = serializers.SchedulingSetStrategyTemplateSerializer


class SchedulingUnitObservingStrategyTemplateViewSet(AbstractTemplateViewSet):
    queryset = models.SchedulingUnitObservingStrategyTemplate.objects.all()
    serializer_class = serializers.SchedulingUnitObservingStrategyTemplateSerializer

    # override default method, to make sure that we perform the correct project-based permission checks
    extra_action_permission_specs = {'create_scheduling_unit': {'model': models.SchedulingUnitDraft,
                                     'project_ref_override_pk_url_param': 'scheduling_set_id',
                                     'project_ref_override_attr': 'scheduling_set',
                                     'project_ref_override_model': models.SchedulingSet}}

    @swagger_auto_schema(responses={status.HTTP_201_CREATED: 'The newly created scheduling unit',
                                    status.HTTP_403_FORBIDDEN: 'forbidden'},
                         operation_description="Create a new SchedulingUnit based on this SchedulingUnitObservingStrategyTemplate, with the given <name> and <description> and make it a child of the given <scheduling_set_id>",
                         request_body=openapi.Schema(type=openapi.TYPE_OBJECT,
                                                     description="[Optional] specifications_doc containing the override values for the parameters in the template. See endpoint template_doc_with_just_the_parameters on this instance."),
                         manual_parameters=[Parameter(name='scheduling_set_id', required=True, type='integer', in_='query',
                                                      description="The id of the scheduling_set which will be the parent of the newly created scheduling_unit"),
                                            Parameter(name='name', required=False, type='string', in_='query',
                                                      description="The name for the newly created scheduling_unit"),
                                            Parameter(name='description', required=False, type='string', in_='query',
                                                      description="The description for the newly created scheduling_unit"),
                                            Parameter(name='rank', required=False, type='number', in_='query',
                                                      description="The rank of the scheduling unit."),
                                            Parameter(name='priority_queue', required=False, type='string', in_='query',
                                                      description="The priority queue of the scheduling unit.")
                                            ])
    @action(methods=['post'], detail=True)
    def create_scheduling_unit(self, request, pk=None):
        strategy_template = get_object_or_404(models.SchedulingUnitObservingStrategyTemplate, pk=pk)
        scheduling_set = get_object_or_404(models.SchedulingSet, pk=request.query_params['scheduling_set_id'])
        name = request.query_params.get('name', "scheduling unit")
        description = request.query_params.get('description', "")
        rank = request.query_params.get('rank', None)
        priority_queue = get_object_or_404(models.PriorityQueueType, value=request.query_params.get('priority_queue', "A"))

        specifications_doc_overrides = request.data if isinstance(request.data, dict) else None

        scheduling_unit_draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template, scheduling_set, name, description, rank, priority_queue, specifications_doc_overrides)

        return Response(serializers.SchedulingUnitDraftSerializer(scheduling_unit_draft, context={'request':request}).data,
                        status=status.HTTP_201_CREATED)


    @swagger_auto_schema(responses={200: 'The template document as JSON object with all defaults filled in.',
                                    403: 'forbidden'},
                         operation_description="Get the template document as JSON object with all defaults filled in. This template document is the foundation from which a scheduling unit draft is created. The intended use is to carefully inspect this JSON document for all specification details for this strategy template.")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True)
    def template_doc_complete_with_defaults(self, request, pk=None):
        strategy_template = get_object_or_404(models.SchedulingUnitObservingStrategyTemplate, pk=pk)
        return JsonResponse(strategy_template.template_doc_complete_with_defaults, json_dumps_params={'indent': 2, 'sort_keys': True})


    @swagger_auto_schema(responses={200: 'The template document as JSON object with just the items that are in the parameters list.',
                                    403: 'forbidden'},
                         operation_description="Get the slimmed down version of the template document as JSON object with all defaults filled in, but then only for the allowed-to-be-edited parameters. This template document is the foundation from which a scheduling unit draft is created. The intended use is to carefully inspect this JSON document for all specification details for this strategy template.")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True)
    def template_doc_with_just_the_parameters(self, request, pk=None):
        strategy_template = get_object_or_404(models.SchedulingUnitObservingStrategyTemplate, pk=pk)
        return JsonResponse(strategy_template.template_doc_with_just_the_parameters, json_dumps_params={'indent': 2, 'sort_keys': True})


    @swagger_auto_schema(responses={200: 'Get a trigger JSON document ready to be used for trigger submission',
                                    403: 'forbidden'},
                         operation_description="Get a trigger JSON document ready to be used for trigger submission.")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True)
    def trigger_doc(self, request, pk=None):
        '''Convenience endpoint to easily fetch a JSON doc for the requested strategy_template which can be used for trigger submission'''
        strategy_template = get_object_or_404(models.SchedulingUnitObservingStrategyTemplate, pk=pk)
        trigger_template = models.CommonSchemaTemplate.get_latest(name='triggers')
        triger_doc = trigger_template.get_default_json_document_for_schema()
        triger_doc['name'] = 'my_trigger_name'
        triger_doc['description'] = 'my_description_for_this_trigger'
        triger_doc['scheduling_unit_observing_strategy_template']['name'] = strategy_template.name
        triger_doc['scheduling_unit_observing_strategy_template']['version'] = int(strategy_template.version)
        triger_doc['scheduling_unit_observing_strategy_template']['overrides'] = strategy_template.template_doc_with_just_the_parameters

        return JsonResponse(triger_doc, json_dumps_params={'indent': 2, 'sort_keys': True})



class SchedulingUnitObservingStrategyTemplateNestedViewSet(LOFARNestedViewSet):
    queryset = models.SchedulingUnitObservingStrategyTemplate.objects.all()
    serializer_class = serializers.SchedulingUnitObservingStrategyTemplateSerializer

    def get_queryset(self):
            if 'scheduling_set_id' in self.kwargs:
                scheduling_unit_drafts = get_list_or_404(models.SchedulingUnitDraft, scheduling_set__id=self.kwargs['scheduling_set_id'])
                return [draft.observation_strategy_template for draft in scheduling_unit_drafts]
            else:
                return models.SchedulingUnitObservingStrategyTemplate.objects.all()


class SchedulingSetTemplateViewSet(AbstractSchemaTemplateViewSet):
    queryset = models.SchedulingSetTemplate.objects.all()
    serializer_class = serializers.SchedulingSetTemplateSerializer


class SchedulingUnitTemplateViewSet(AbstractSchemaTemplateViewSet):
    queryset = models.SchedulingUnitTemplate.objects.all()
    serializer_class = serializers.SchedulingUnitTemplateSerializer


class SchedulingConstraintsTemplateViewSet(AbstractSchemaTemplateViewSet):
    queryset = models.SchedulingConstraintsTemplate.objects.all()
    serializer_class = serializers.SchedulingConstraintsTemplateSerializer


class SchedulingConstraintsWeightFactorViewSet(LOFARViewSet):
    queryset = models.SchedulingConstraintsWeightFactor.objects.all()
    serializer_class = serializers.SchedulingConstraintsWeightFactorSerializer


class TaskTemplateViewSet(AbstractSchemaTemplateViewSet):
    queryset = models.TaskTemplate.objects.all()
    serializer_class = serializers.TaskTemplateSerializer

    @swagger_auto_schema(responses={200: 'All TaskConnectorTypes referencing this template.',
                                    403: 'forbidden'},
                         operation_description="Get the all TaskConnectorTypes referencing this template.")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, url_name='connectors')
    def connectors(self, request, pk=None):
        connectors = models.TaskConnectorType.objects.filter(task_template__id=pk)
        serializer = serializers.TaskConnectorTypeSerializer(connectors, many=True, context={'request': request})
        return RestResponse(serializer.data)


class TaskRelationSelectionTemplateViewSet(AbstractSchemaTemplateViewSet):
    queryset = models.TaskRelationSelectionTemplate.objects.all()
    serializer_class = serializers.TaskRelationSelectionTemplateSerializer


class ReservationStrategyTemplateViewSet(AbstractTemplateViewSet):
    queryset = models.ReservationStrategyTemplate.objects.all()
    serializer_class = serializers.ReservationStrategyTemplateSerializer

    @swagger_auto_schema(responses={status.HTTP_201_CREATED: 'The newly created reservation',
                                    status.HTTP_403_FORBIDDEN: 'forbidden'},
                         operation_description="Create a new Reservation based on this ReservationStrategyTemplate, "
                                               "with the given <name>, <description>, <start_time> and <stop_time>",
                         manual_parameters=[Parameter(name='start_time', required=True, type='string', in_='query',
                                                      description="The start time as a timestamp string in isoformat"),
                                            Parameter(name='stop_time', required=True, type='string', in_='query',
                                                      description="The stop time as a timestamp string in isoformat"),
                                            Parameter(name='name', required=False, type='string', in_='query',
                                                      description="The name for the newly created reservation"),
                                            Parameter(name='description', required=False, type='string', in_='query',
                                                      description="The description for the newly created reservation"),
                                            Parameter(name='project_id', required=False, type='integer', in_='query',
                                                      description="the id of the project which will be the parent of the newly created reservation"),
                                            ])
    @action(methods=['post'], detail=True)
    def create_reservation(self, request, pk=None):
        strategy_template = get_object_or_404(models.ReservationStrategyTemplate, pk=pk)
        reservation_template_spec = add_defaults_to_json_object_for_schema(strategy_template.template,
                                                      strategy_template.reservation_template.schema)

        start_time = request.query_params.get('start_time', None)
        stop_time = request.query_params.get('stop_time', None)
        if start_time:
            start_time = dateutil.parser.parse(start_time)  # string to datetime
        else:
            start_time = datetime.now()
        if stop_time:
            stop_time = dateutil.parser.parse(stop_time)  # string to datetime
        else:
            stop_time = None

        project_id = request.query_params.get('project_id', None)
        if project_id:
            project = get_object_or_404(models.Project, pk=request.query_params['project_id'])
        else:
            project = None

        stations = request.query_params.get('stations', None)
        if isinstance(stations, str):
            stations = [s.strip() for s in stations.split(',')]
        if stations:
            reservation_template_spec['resources']['stations'] = stations

        reservation = Reservation.objects.create(name=request.query_params.get('name', "reservation"),
                                                 description=request.query_params.get('description', ""),
                                                 project=project,
                                                 specifications_template=strategy_template.reservation_template,
                                                 specifications_doc=reservation_template_spec,
                                                 start_time=start_time,
                                                 stop_time=stop_time)

        reservation_strategy_template_path = request._request.path
        base_path = reservation_strategy_template_path[:reservation_strategy_template_path.find('/reservation_strategy_template')]
        reservation_path = '%s/reservation/%s/' % (base_path, reservation.id,)

        # return a response with the new serialized Reservation
        return Response(serializers.ReservationSerializer(reservation, context={'request':request}).data,
                        status=status.HTTP_201_CREATED)


class ReservationTemplateViewSet(AbstractSchemaTemplateViewSet):
    queryset = models.ReservationTemplate.objects.all()
    serializer_class = serializers.ReservationTemplateSerializer


all_stations_tuple = tuple((i, i) for i in ('CS001', 'CS002', 'CS003', 'CS004', 'CS005', 'CS006', 'CS007', 'CS011',
                                            'CS013', 'CS017', 'CS021', 'CS024', 'CS026', 'CS028', 'CS030', 'CS031',
                                            'CS032', 'CS101', 'CS103', 'CS201', 'CS301', 'CS302', 'CS401', 'CS501',
                                            'RS106', 'RS205', 'RS208', 'RS210', 'RS305', 'RS306', 'RS307', 'RS310',
                                            'RS406', 'RS407', 'RS409', 'RS503', 'RS508', 'RS509', 'DE601', 'DE602',
                                            'DE603', 'DE604', 'DE605', 'DE609', 'FR606', 'SE607', 'UK608', 'PL610',
                                            'PL611', 'PL612', 'IE613', 'LV614'))


class ReservationPropertyFilter(property_filters.PropertyFilterSet):
    id = NumberInFilter(field_name='id', lookup_expr='in')
    id_min = filters.NumberFilter(field_name='id', lookup_expr='gte')
    id_max = filters.NumberFilter(field_name='id', lookup_expr='lte')
    project = filters.CharFilter(field_name='project__name', lookup_expr='icontains')
    start_time = filters.IsoDateTimeFromToRangeFilter(field_name='start_time')
    stop_time = filters.IsoDateTimeFromToRangeFilter(field_name='stop_time')
    stop_time_isnull = filters.BooleanFilter(field_name='stop_time', lookup_expr='isnull')
    duration_min = property_filters.PropertyNumberFilter(field_name='duration', lookup_expr='gte')
    duration_max = property_filters.PropertyNumberFilter(field_name='duration', lookup_expr='lte')
    duration_isnull = property_filters.PropertyBooleanFilter(field_name='duration', lookup_expr='isnull')
    # Specification properties
    planned = filters.BooleanFilter(label='planned', field_name='specifications_doc__activity__planned')
    lba_rfi = filters.BooleanFilter(label='lba_rfi', field_name='specifications_doc__effects__lba_rfi')
    hba_rfi = filters.BooleanFilter(label='hba_rfi', field_name='specifications_doc__effects__lba_rfi')
    expert = filters.BooleanFilter(label='expert', field_name='specifications_doc__effects__expert')
    fixed_time = filters.BooleanFilter(label='fixed_time', field_name='specifications_doc__schedulability__fixed_time')
    dynamic = filters.BooleanFilter(label='dynamic', field_name='specifications_doc__schedulability__dynamic')
    project_exclusive = filters.BooleanFilter(label='project_exclusive', field_name='specifications_doc__schedulability__project_exclusive')
    # TODO: Remove hard coded values. See TMSS-984.
    reservation_type = filters.ChoiceFilter(label='reservation_type', field_name='specifications_doc__activity__type',
                                            choices=tuple((i, i) for i in ('maintenance', 'test', 'upgrade', 'outage', 'pr', 'stand-alone mode', 'test system', 'other', "mowing", "stopday", "software rollout", "windmill standstill")))
    subject = filters.ChoiceFilter(label='subject', field_name='specifications_doc__activity__subject',
                                   choices=tuple((i, i) for i in ('environment', 'hardware', 'firmware', 'software', 'system', 'network', 'nothing')))
    stations_any = filters.MultipleChoiceFilter(label='stations [any]', field_name='specifications_doc__resources__stations',
                                                choices=all_stations_tuple, lookup_expr='icontains')
    stations_all = filters.MultipleChoiceFilter(label='stations [all]', field_name='specifications_doc__resources__stations',
                                                choices=all_stations_tuple, lookup_expr='icontains', conjoined=True)

    class Meta:
        model = models.Reservation
        fields = '__all__'
        filter_overrides = FILTER_OVERRIDES


class ReservationViewSet(LOFARViewSet):
    queryset = models.Reservation.objects.all()
    serializer_class = serializers.ReservationSerializer
    filter_class = ReservationPropertyFilter     # note that this breaks other filter backends from LOFARViewSet


class RoleViewSet(LOFARViewSet):
    queryset = models.Role.objects.all()
    serializer_class = serializers.RoleSerializer


class IOTypeViewSet(LOFARViewSet):
    queryset = models.IOType.objects.all()
    serializer_class = serializers.IOTypeSerializer


class SchedulingRelationPlacementViewSet(LOFARViewSet):
    queryset = models.SchedulingRelationPlacement.objects.all()
    serializer_class = serializers.SchedulingRelationPlacementSerializer


class DatatypeViewSet(LOFARViewSet):
    queryset = models.Datatype.objects.all()
    serializer_class = serializers.DatatypeSerializer


class DataformatViewSet(LOFARViewSet):
    queryset = models.Dataformat.objects.all()
    serializer_class = serializers.DataformatSerializer


class TaskConnectorTypeViewSet(LOFARViewSet):
    queryset = models.TaskConnectorType.objects.all()
    serializer_class = serializers.TaskConnectorTypeSerializer


class CycleViewSet(LOFARViewSet):
    permission_classes = (TMSSDjangoModelPermissions,)      # override default project permission
    filter_backends = (LOFARFilterBackend, OrderingFilter)                                    # override default project permission
    queryset = models.Cycle.objects.all()
    serializer_class = serializers.CycleSerializer
    ordering = ['start']

    @swagger_auto_schema(responses={200: 'The Report information',
                                    403: 'forbidden'},
                         operation_description="Get Report information for the cycle.")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, url_name="report", name="Get Report")
    def report(self, request, pk=None):
        cycle = get_object_or_404(models.Cycle, pk=pk)
        # Check start and stop times
        start, stop = request.query_params.get('start', None), request.query_params.get('stop', None)
        refresh = request.query_params.get('refresh', 'false').lower() in ['1', 'true', 'yes']
        wipe = request.query_params.get('wipe', 'false').lower() in ['1', 'true', 'yes']
        exclude_projects = request.query_params.get('exclude_projects', None)
        filter_projects = request.query_params.get('filter_projects', None)
        if start:
            try:
                start = dateutil.parser.parse(start, ignoretz=True)
            except Exception:
                return Response('Error: please specify an isoformat timestamp for start_time', status=400)
        if stop:
            try:
                stop = dateutil.parser.parse(stop, ignoretz=True)
            except Exception:
                return Response('Error: please specify an isoformat timestamp for stop_time', status=400)

        if not start or start < cycle.start:
            start = cycle.start
        if not stop or stop > cycle.stop:
            stop = cycle.stop

        if wipe:
            # wipe all reports for this cycle (not just request period), so we have no outdated info in the cache.
            # After this, a new report will then be generated and contain a placeholder until the report is done.
            logger.info('Wipe requested, wiping all cached reports for cycle=%s' % cycle.pk)
            models.CycleReport.objects.filter(cycle=cycle).delete()
        try:
            report = models.CycleReport.objects.get(cycle=cycle, start=start, stop=stop, exclude_projects=exclude_projects, filter_projects=filter_projects)
        except models.CycleReport.DoesNotExist:
            report = models.CycleReport.objects.create(cycle=cycle, start=start, stop=stop,
                                                       exclude_projects=exclude_projects, filter_projects=filter_projects,
                                                       data="Your report is getting prepared. Please check again in a few minutes")
            models.CycleReport.submit_cycle_for_report_creation(report)
        if refresh:
            # start re-computation of the requested report in background, but we still return the existing report data
            # so that the user can get the standard report at all times and there are no waiting times when the refresh
            # service triggers a re-computation. (By also setting the wipe flag, we can choose to instead receive the
            # placeholder until the refresh is done.)
            logger.info('Refresh requested, submitting cycle=%s' % cycle.pk)
            models.CycleReport.submit_cycle_for_report_creation(report)

        return Response(report.data, status=status.HTTP_200_OK)

class CycleQuotaViewSet(LOFARViewSet):
    queryset = models.CycleQuota.objects.all()
    serializer_class = serializers.CycleQuotaSerializer

    def get_queryset(self):
        queryset = models.CycleQuota.objects.all()

        # query by project
        project = self.request.query_params.get('project', None)
        if project is not None:
            return queryset.filter(project=project)

        return queryset


class ProjectViewSet(LOFARViewSet):
    queryset = models.Project.objects.all()
    serializer_class = serializers.ProjectSerializer
    ordering = ['name']

    def get_queryset(self):
        queryset = models.Project.objects.all()

        # query by cycle
        try:
            cycle = self.request.query_params.get('cycle', None)
            if cycle is not None:
                return queryset.filter(cycles__name=cycle)
        except AttributeError:
            pass

        return queryset

    @swagger_auto_schema(responses={200: 'The Report information',
                                    403: 'forbidden'},
                         operation_description="Get Report information for the project.")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, url_name="report", name="Get Report")
    def report(self, request, pk=None):
        project = get_object_or_404(models.Project, pk=pk)

        # Check start and stop times
        start, stop = request.query_params.get('start', None), request.query_params.get('stop', None)
        if start:
            try:
                start = dateutil.parser.parse(start, ignoretz=True)
            except Exception:
                return Response('Error: please specify an isoformat timestamp for start_time', status=400)
        if stop:
            try:
                stop = dateutil.parser.parse(stop, ignoretz=True)
            except Exception:
                return Response('Error: please specify an isoformat timestamp for stop_time', status=400)

        result = create_project_report(project, start, stop)
        return Response(result, status=status.HTTP_200_OK)

    @swagger_auto_schema(responses={200: 'List of users that have the "friend of project" role for this project',
                                    403: 'forbidden'},
                         operation_description="Get the list of users that have the 'friend of project' role for this project")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, url_name="friend", name="Friend(s) of this project")
    def friend(self, request, pk=None):
        project = get_object_or_404(models.Project, pk=pk)
        result = get_names_by_role_in_project(models.ProjectRole.Choices.FRIEND_OF_PROJECT.value, project.name)
        return Response(result, status=status.HTTP_200_OK)

    @swagger_auto_schema(responses={200: 'List of roles that the requesting user has in this project',
                                    403: 'forbidden'},
                         operation_description="Get the list of roles that the requesting user has in this project")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, url_name="my_roles", name="My roles in this project")
    def my_roles(self, request, pk=None):
        project = get_object_or_404(models.Project, pk=pk)
        project_roles = get_project_roles_for_user(request.user)
        result = []
        for project_role in project_roles:
            if project.name.lower() == project_role['project'].lower():
                result.append(project_role['role'])
        return Response(result, status=status.HTTP_200_OK)


    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, renderer_classes=(ImagePNGRenderer,))
    def pressure_plot(self, request, pk=None):
        project = get_object_or_404(models.Project, pk=pk)

        from lofar.sas.tmss.tmss.tmssapp.adapters.plots import lst_pressure_plot
        plot = lst_pressure_plot(lower_bound=project.earliest_possible_cycle_start_time,
                                 upper_bound=project.latest_possible_cycle_stop_time,
                                 include_calibrators=False,
                                 projects=[project.name])
        return Response(plot, content_type='image/png', status=status.HTTP_201_CREATED)


class ProjectNestedViewSet(LOFARNestedViewSet):
    queryset = models.Project.objects.all()
    serializer_class = serializers.ProjectSerializer

    def get_queryset(self):
        if 'cycle_id' in self.kwargs:
            cycle = get_object_or_404(models.Cycle, pk=self.kwargs['cycle_id'])
            return cycle.projects.all()
        else:
            return models.Project.objects.all()


class ProjectQuotaViewSet(LOFARViewSet):
    queryset = models.ProjectQuota.objects.all()
    serializer_class = serializers.ProjectQuotaSerializer

    def get_queryset(self):
        queryset = models.ProjectQuota.objects.all()

        # query by project
        project = self.request.query_params.get('project', None)
        if project is not None:
            return queryset.filter(project=project)

        return queryset


class ProjectQuotaArchiveLocationViewSet(LOFARViewSet):
    queryset = models.ProjectQuotaArchiveLocation.objects.all()
    serializer_class = serializers.ProjectQuotaArchiveLocationSerializer

    def get_queryset(self):
        queryset = models.ProjectQuotaArchiveLocation.objects.all()

        # query by project
        project = self.request.query_params.get('project', None)
        if project is not None:
            return queryset.filter(project=project)

        return queryset


class ResourceTypeViewSet(LOFARViewSet):
    queryset = models.ResourceType.objects.all()
    serializer_class = serializers.ResourceTypeSerializer


class SchedulingSetViewSet(LOFARViewSet):
    queryset = models.SchedulingSet.objects.all()
    serializer_class = serializers.SchedulingSetSerializer

    @swagger_auto_schema(responses={201: 'The newly copied SchedulingSet',
                                    403: 'forbidden'},
                         operation_description="Copy this SchedulingSet including copies of all SchedulingUnitDraft's in this set.")
    @action(methods=['post'], detail=True, url_name="copy", name="Copy")
    def copy(self, request, pk=None):
        scheduling_set = get_object_or_404(models.SchedulingSet, pk=pk)
        scheduling_set_copy = models.SchedulingSet.objects.create(name=scheduling_set.name,
                                                                  project=scheduling_set.project)

        for scheduling_unit_draft in scheduling_set.scheduling_unit_drafts.all():
            copy_scheduling_unit_draft(scheduling_unit_draft, scheduling_set_copy)

        return Response(serializers.SchedulingSetSerializer(scheduling_set_copy, context={'request':request}).data,
                        status=status.HTTP_201_CREATED)


class SystemSettingFlagViewSet(LOFARViewSet):
    queryset = models.SystemSettingFlag.objects.all()
    serializer_class = serializers.SystemSettingFlagSerializer


class SettingViewSet(LOFARViewSet):
    queryset = models.Setting.objects.all()
    serializer_class = serializers.SettingSerializer


class SubsystemStatusViewSet(LOFARViewSet):
    queryset = models.SubsystemStatus.objects.all()
    serializer_class = serializers.SubsystemStatusSerializer


class SubsystemViewSet(LOFARViewSet):
    queryset = models.Subsystem.objects.all()
    serializer_class = serializers.SubsystemSerializer


class QuantityViewSet(LOFARViewSet):
    queryset = models.Quantity.objects.all()
    serializer_class = serializers.QuantitySerializer


class PeriodCategoryViewSet(LOFARViewSet):
    queryset = models.PeriodCategory.objects.all()
    serializer_class = serializers.PeriodCategorySerializer


class ProjectCategoryViewSet(LOFARViewSet):
    queryset = models.ProjectCategory.objects.all()
    serializer_class = serializers.ProjectCategorySerializer


class CharInFilter(filters.BaseInFilter, filters.CharFilter):
    """
    Custom filter that accepts a comma-separated lists of strings.
    Filters for objects where the (single) string value of the filtered field is in the provided list.
    """
    pass


class ListContainsAnyFilter(filters.BaseCSVFilter):
    """
    Custom filter for where both the filtered field and the filter input are comma-separated lists.
    Filters for objects where any of the values in the provided list is contained in the filtered field.
    """
    def filter(self, qs, values):
        from django.db.models import Q
        query = Q()
        if not values:
            return qs
        else:
            for value in values:
                query |= Q(**{self.field_name+'__contains': value.strip()})
        return qs.filter(query)


class FloatRangeField(filter_fields.RangeField):
    """
    Custom range field for float values. The normal django filter_fields.RangeField uses (integer) decimal fields.
    """
    def __init__(self, fields=None, *args, **kwargs):
        if fields is None:
            fields = (forms.FloatField(),forms.FloatField())
        super().__init__(fields, *args, **kwargs)


class FloatRangeFilter(filters.RangeFilter):
    """Custom RangeFilter using the FloatRangeField."""
    field_class = FloatRangeField


class IsoDateTimeNullOrFromToRangeFilter(filters.IsoDateTimeFromToRangeFilter):
    """
    Custom filter that works like the IsoDateTimeFromToRangeFilter but also returns objects where the filtered field
    is None.
    """
    def filter(self, qs, values):
        super_qs = super().filter(qs, values)
        from django.db.models import Q
        query = Q(**{self.field_name+'__isnull': True})
        return super_qs | qs.filter(query)


class TargetObservationSpecificationDocFilterMixin:
    """This filter mixin can be used to filter for values in the json specification_doc of target observation tasks in the scheduling units"""
    def filter(self, qs, value):
        """this filter method filters the scheduling_units in the queryset qs by applying the the filter in the self.field_name on the json specifications_doc for target observation tasks and then keeping the scheduling units with matching tasks."""
        if value in EMPTY_VALUES:
            return qs

        target_obs_tasks = TaskDraft.objects.filter(id__in=qs.values('task_drafts__id')).\
                                             filter(specifications_template__type__value=TaskType.Choices.OBSERVATION.value).\
                                             filter(specifications_template__name__icontains='target')

        filtered_target_obs_tasks = super().filter(target_obs_tasks, value)
        filtered_scheduling_unit_ids = filtered_target_obs_tasks.values('scheduling_unit_draft_id')
        return qs.filter(id__in=filtered_scheduling_unit_ids)


class TargetObservationSpecificationDocFloatRangeFilter(TargetObservationSpecificationDocFilterMixin, FloatRangeFilter):
    pass

class TargetObservationSpecificationFilterCharFilter(TargetObservationSpecificationDocFilterMixin, filters.CharFilter):
    pass


class SchedulingUnitDraftPropertyFilter(property_filters.PropertyFilterSet):
    project = filters.CharFilter(field_name='scheduling_set__project__name', lookup_expr='icontains')
    scheduling_set = filters.CharFilter(field_name='scheduling_set__name', lookup_expr='icontains')  # remove to filter based on list of options
    rank = filters.RangeFilter(field_name='rank')
    id = NumberInFilter(field_name='id', lookup_expr='in')
    id_min = filters.NumberFilter(field_name='id', lookup_expr='gte')
    id_max = filters.NumberFilter(field_name='id', lookup_expr='lte')
    scheduling_unit_blueprints = filters.ModelMultipleChoiceFilter(field_name='scheduling_unit_blueprints', queryset=models.SchedulingUnitBlueprint.objects.all())
    blueprinted = filters.BooleanFilter(label='blueprinted', field_name='scheduling_unit_blueprints', lookup_expr='isnull', exclude=True)
    duration_min = property_filters.PropertyDurationFilter(field_name='duration', lookup_expr='gte')
    duration_max = property_filters.PropertyDurationFilter(field_name='duration', lookup_expr='lte')
    created_at = filters.IsoDateTimeFromToRangeFilter(field_name='created_at')
    updated_at = filters.IsoDateTimeFromToRangeFilter(field_name='updated_at')
    angle1 = TargetObservationSpecificationDocFloatRangeFilter(field_name='specifications_doc__station_configuration__SAPs__0__digital_pointing__angle1')
    angle2 = TargetObservationSpecificationDocFloatRangeFilter(field_name='specifications_doc__station_configuration__SAPs__0__digital_pointing__angle2')
    direction_type = TargetObservationSpecificationFilterCharFilter(field_name='specifications_doc__station_configuration__SAPs__0__digital_pointing__direction_type', lookup_expr='iexact')

    observation_strategy_template_name = filters.CharFilter(field_name='observation_strategy_template__name', lookup_expr='icontains')
    observation_strategy_template_description = filters.CharFilter(field_name='observation_strategy_template__description', lookup_expr='icontains')

    class Meta:
        model = models.SchedulingUnitDraft
        fields = '__all__'
        filter_overrides = FILTER_OVERRIDES


class SchedulingUnitDraftViewSet(LOFARViewSet):
    queryset = models.SchedulingUnitDraft.objects.all()
    serializer_class = serializers.SchedulingUnitDraftSerializer
    filter_class = SchedulingUnitDraftPropertyFilter    # note that this breaks other filter backends from LOFARViewSet

    # prefetch all reverse related references from other models on their related_name to avoid a ton of duplicate queries
    queryset = queryset.prefetch_related('scheduling_unit_blueprints')\
                       .prefetch_related('task_drafts')\
                       .prefetch_related('task_drafts__specifications_template')\
                       .prefetch_related('task_drafts__specifications_template__type')\
                       .prefetch_related('task_drafts__specifications_template__connector_types')

    # use select_related for forward related references
    queryset = queryset.select_related('scheduling_set',
                                       'specifications_template',
                                       'specifications_template__state',
                                       'specifications_template__purpose',
                                       'observation_strategy_template',
                                       'observation_strategy_template__state',
                                       'observation_strategy_template__purpose',
                                       'scheduling_constraints_template',
                                       'scheduling_constraints_template__state',
                                       'scheduling_constraints_template__purpose',
                                       'scheduling_set__project',
                                       'priority_queue')

    @staticmethod
    def _extract_specifications_doc_from_request(request: RestRequest) -> dict:
        if 'specifications_doc' in request.data:
            # see: https://docs.djangoproject.com/en/3.2/ref/request-response/#django.http.QueryDict.copy
            # POST request.data objects are immutable QueryDicts.
            # by making a copy they become mutable.
            # We need to have mutable QueryDicts in order to pop the specifications_doc
            # because the specifications_doc is not a SUD property, but used in update_task_graph_from_specifications_doc
            if isinstance(request.data, QueryDict):
                request._full_data = request.data.copy()
            return request.data.pop('specifications_doc', {})
        return None

    def create(self, request, **kwargs):
        '''special convenience override of create. Perform a normal create of the scheduling_unit_draft,
        and if a specifications_doc was provided in the request then instantiate the task graph from that specifications_doc as well.'''
        with transaction.atomic():
            specifications_doc = self._extract_specifications_doc_from_request(request)

            # do normal create
            response = super().create(request, **kwargs)

            if specifications_doc:
                # also instantiate the task graph from the specifications_doc
                scheduling_unit_draft = response.data.serializer.instance
                scheduling_unit_draft = update_task_graph_from_specifications_doc(scheduling_unit_draft, specifications_doc)

                # and feed the updated scheduling_unit_draft with the graph into the serializer and response
                serializer = self.get_serializer(scheduling_unit_draft)
                response.data = serializer.data

            return response

    def update(self, request, pk=None, **kwargs):
        '''special convenience override of update. Perform an update of the scheduling_unit_draft,
        and if a specifications_doc was provided in the request then instantiate the task graph from that specifications_doc as well.'''
        with transaction.atomic():
            specifications_doc = self._extract_specifications_doc_from_request(request)

            # do normal create
            response = super().update(request, pk, **kwargs)

            if specifications_doc is not None:
                # also (re)instantiate the task graph from the specifications_doc
                scheduling_unit_draft = response.data.serializer.instance
                scheduling_unit_draft = update_task_graph_from_specifications_doc(scheduling_unit_draft, specifications_doc)

                # and feed the updated scheduling_unit_draft with the graph into the serializer and response
                serializer = self.get_serializer(scheduling_unit_draft)
                response.data = serializer.data

            return response


    @swagger_auto_schema(responses={201: 'The Created SchedulingUnitBlueprint',
                                    403: 'forbidden'},
                         operation_description="Carve this SchedulingUnitDraft and its TaskDraft(s) in stone, and make blueprint(s) out of it and create their subtask(s)")
    @action(methods=['post'], detail=True, url_name="create_scheduling_unit_blueprint_and_tasks_and_subtasks_tree", name="Create Blueprints-Tree")
    def create_scheduling_unit_blueprint_and_tasks_and_subtasks_tree(self, request, pk=None):
        scheduling_unit_draft = get_object_or_404(models.SchedulingUnitDraft, pk=pk)
        scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        # return a response with the new serialized SchedulingUnitBlueprintExtendedSerializer
        return Response(serializers.SchedulingUnitBlueprintExtendedSerializer(scheduling_unit_blueprint, context={'request':request}).data,
                        status=status.HTTP_201_CREATED)


    @swagger_auto_schema(responses={201: 'The new copied SchedulingUnitDraft',
                                    403: 'forbidden'},
                         operation_description="Copy this SchedulingUnitDraft",
                         request_body=openapi.Schema(type=openapi.TYPE_INTEGER,
                                                     description="the id of the scheduling_set which will be the parent of the newly created scheduling_unit"))
    @action(methods=['post'], detail=True, url_name="copy", name="Copy")
    def copy(self, request, pk=None):
        scheduling_unit_draft = get_object_or_404(models.SchedulingUnitDraft, pk=pk)

        # see if there is a given scheduling_set_id in the request
        if 'scheduling_set_id' in request.data:
            scheduling_set_id = request.data.pop('scheduling_set_id')
            scheduling_set = models.SchedulingSet.objects.get(id=scheduling_set_id)
        else:
            # otherwise, use this draft's scheduling_set as parent
            scheduling_set = scheduling_unit_draft.scheduling_set

        scheduling_unit_draft_copy = copy_scheduling_unit_draft(scheduling_unit_draft, scheduling_set)

        # return a response with the new serialized SchedulingUnitBlueprintExtendedSerializer
        return Response(serializers.SchedulingUnitDraftExtendedSerializer(scheduling_unit_draft_copy, context={'request':request}).data,
                        status=status.HTTP_201_CREATED)


    @swagger_auto_schema(responses={201: 'The updated scheduling_unit_draft with references to its new/updated task_drafts in the updated graph',
                                    403: 'forbidden'},
                         operation_description="Update the task graph from given the specifications_doc")
    @action(methods=['post'], detail=True, url_name="update_task_graph_from_specifications_doc", name="Update the task graph from given the specifications_doc")
    def update_task_graph_from_specifications_doc(self, request, pk=None):
        specifications_doc = request.data.get('specifications_doc', {})

        scheduling_unit_draft = get_object_or_404(models.SchedulingUnitDraft, pk=pk)
        update_task_graph_from_specifications_doc(scheduling_unit_draft, specifications_doc)

        # just return as a response the serialized scheduling_unit_draft (with references to the created task_draft(s)
        return Response(serializers.SchedulingUnitDraftSerializer(scheduling_unit_draft, context={'request':request}).data,
                        status=status.HTTP_201_CREATED)


    @swagger_auto_schema(responses={201: "A JSON document reflecting the current scheduling unit draft's task graph and settings",
                                    403: 'forbidden'},
                         operation_description="Get a json specifications_doc reflecting the current scheduling unit draft's task graph and settings.")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, url_name="specifications_doc", name="Get a json specifications_doc reflecting the current scheduling unit draft's task graph and settings.")
    def specifications_doc(self, request, pk=None):
        scheduling_unit_draft = get_object_or_404(models.SchedulingUnitDraft, pk=pk)
        return JsonResponse(scheduling_unit_draft.specifications_doc, json_dumps_params={'indent': 2, 'sort_keys': True})


    @swagger_auto_schema(responses={201: "Create a new Observation Strategy Template based on the specifications_doc of this scheduling_unit",
                                    403: 'forbidden'},
                         operation_description="Create a new Observation Strategy Template based on the specifications_doc of this scheduling_unit.")
    @action(methods=['post'], detail=True, url_name="create_observing_strategy_template", name="Create a new Observation Strategy Template based on the specifications_doc of this scheduling_unit")
    def create_observing_strategy_template(self, request, pk=None):
        scheduling_unit_draft = get_object_or_404(models.SchedulingUnitDraft, pk=pk)
        observing_strategy_template = scheduling_unit_draft.create_observing_strategy_template(name= request.data.pop('name', 'observing_strategy_template from scheduling_unit_draft id=%s'%(pk,)))

        # return a response with the new serialized observing_strategy_template
        return Response(serializers.SchedulingUnitObservingStrategyTemplateSerializer(observing_strategy_template, context={'request':request}).data,
                        status=status.HTTP_201_CREATED)

    @swagger_auto_schema(responses={201: "This SchedulingUnitDraft, with references to the created Cleanup TaskDrafts.",
                                    403: 'forbidden'},
                         operation_description="Create a cleanup task for this scheduling unit.")
    @action(methods=['post'], detail=True, url_name="create_cleanuptask", name="Create a cleanup task for this scheduling unit")
    def create_cleanuptask_for_scheduling_unit_draft(self, request, pk=None):
        scheduling_unit_draft = get_object_or_404(models.SchedulingUnitDraft, pk=pk)
        scheduling_unit_draft = create_cleanuptask_for_scheduling_unit_draft(scheduling_unit_draft)

        # return a response with the new serialized scheduling_unit_draft (with references to the created task_draft)
        return Response(serializers.SchedulingUnitDraftSerializer(scheduling_unit_draft, context={'request':request}).data,
                        status=status.HTTP_201_CREATED)

    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, renderer_classes=(ImagePNGRenderer,), url_path='scheduling_constraints_plot/?(?P<station>([a-zA-Z]{2}\d{3}))?/?(?P<proposed_start_time_or_date>(\d{4}-[01]\d-[0-3]\d(T[0-2]\d:[0-5]\d)?))?')
    def scheduling_constraints_plot(self, request, pk=None, station=None, proposed_start_time_or_date=None):
        scheduling_unit = get_object_or_404(models.SchedulingUnitDraft, pk=pk)

        proposed_start_time = None
        proposed_date = None
        if proposed_start_time_or_date is not None:
            parsed_start_time_or_date = dateutil.parser.parse(proposed_start_time_or_date)
            if len(proposed_start_time_or_date) > len("2023-01-01"):
                proposed_start_time = parsed_start_time_or_date
            else:
                proposed_date = parsed_start_time_or_date.date()

        from lofar.sas.tmss.tmss.tmssapp.adapters.plots import scheduling_constraints_plot
        plot = scheduling_constraints_plot(scheduling_unit, station=station, proposed_start_time=proposed_start_time,
                                           proposed_date=proposed_date, include_min_max_elevations=True)
        return Response(plot, content_type='image/png')



class SchedulingUnitDraftExtendedViewSet(SchedulingUnitDraftViewSet):
    serializer_class = serializers.SchedulingUnitDraftExtendedSerializer


class SchedulingUnitDraftNestedViewSet(LOFARNestedViewSet):
    queryset = models.SchedulingUnitDraft.objects.all()
    serializer_class = serializers.SchedulingUnitDraftSerializer

    def get_queryset(self):
        if 'scheduling_set_id' in self.kwargs:
            scheduling_set = get_object_or_404(models.SchedulingSet, pk=self.kwargs['scheduling_set_id'])

            # prefetch all reverse related references from other models on their related_name to avoid a ton of duplicate queries
            queryset = scheduling_set.scheduling_unit_drafts.prefetch_related('scheduling_unit_blueprints') \
                                                            .prefetch_related('task_drafts') \
                                                            .prefetch_related('task_drafts__specifications_template')

            # use select_related for forward related references
            queryset = queryset.select_related('scheduling_set',
                                               'specifications_template',
                                               'observation_strategy_template',
                                               'scheduling_constraints_template',
                                               'scheduling_set__project')
            return queryset.all()
        else:
            return models.SchedulingUnitDraft.objects.all()


class ModelChoiceInFilter(filters.BaseInFilter, filters.ModelChoiceFilter):
    """
    Custom filter for comma-separated lists of references
    """
    pass

class SchedulingUnitBlueprintSlimPropertyFilter(property_filters.PropertyFilterSet):
    '''A small 'slim' PropertyFilterSet for the most relevant properties.'''
    on_sky_start_time = filters.IsoDateTimeFromToRangeFilter(field_name='on_sky_start_time')
    on_sky_stop_time = filters.IsoDateTimeFromToRangeFilter(field_name='on_sky_stop_time')
    process_start_time = filters.IsoDateTimeFromToRangeFilter(field_name='process_start_time')
    process_stop_time = filters.IsoDateTimeFromToRangeFilter(field_name='process_stop_time')
    project = filters.CharFilter(field_name='draft__scheduling_set__project__name', lookup_expr='icontains')
    status = filters.MultipleChoiceFilter(field_name='status', choices=[(c.value, c.value) for c in models.SchedulingUnitStatus.Choices])
    id = NumberInFilter(field_name='id', lookup_expr='in')
    id_min = filters.NumberFilter(field_name='id', lookup_expr='gte')
    id_max = filters.NumberFilter(field_name='id', lookup_expr='lte')
    duration_min = filters.DurationFilter(field_name='duration', lookup_expr='gte')
    duration_max = filters.DurationFilter(field_name='duration', lookup_expr='lte')
    on_sky_duration_min = filters.DurationFilter(field_name='on_sky_duration', lookup_expr='gte')
    on_sky_duration_max = filters.DurationFilter(field_name='on_sky_duration', lookup_expr='lte')
    observed_duration_min = filters.DurationFilter(field_name='observed_duration', lookup_expr='gte')
    observed_duration_max = filters.DurationFilter(field_name='observed_duration', lookup_expr='lte')
    created_at = filters.IsoDateTimeFromToRangeFilter(field_name='created_at')
    updated_at = filters.IsoDateTimeFromToRangeFilter(field_name='updated_at')

    class Meta:
        model = models.SchedulingUnitBlueprint
        fields = ['on_sky_start_time', 'on_sky_stop_time', 'process_start_time', 'process_stop_time', 'project', 'status',
                  'id', 'duration', 'on_sky_duration', 'observed_duration', 'created_at', 'updated_at']
        filter_overrides = FILTER_OVERRIDES


class SchedulingUnitBlueprintSlimViewSet(LOFARViewSet):
    '''A small 'slim' viewset for the most relevant properties.'''
    queryset = models.SchedulingUnitBlueprint.objects.all()
    serializer_class = serializers.SchedulingUnitBlueprintSlimSerializer
    filter_class = SchedulingUnitBlueprintSlimPropertyFilter

    queryset = queryset.prefetch_related('draft__scheduling_set__project',
                                         'status',
                                         'priority_queue',
                                         'task_blueprints',
                                         'task_blueprints__status',
                                         'task_blueprints__specifications_template',
                                         'task_blueprints__specifications_template__type',
                                         )

class SchedulingUnitBlueprintSlimWithTasksViewSet(SchedulingUnitBlueprintSlimViewSet):
    '''A small 'slim' viewset for the most relevant properties including the 'slim' versions of the nested tasks.'''
    serializer_class = serializers.SchedulingUnitBlueprintSlimWithTasksSerializer

    def retrieve(self, request, pk=None, **kwargs):
        return super().retrieve(request, pk, **kwargs)


class SchedulingUnitBlueprintPropertyFilter(property_filters.PropertyFilterSet):
    on_sky_start_time = filters.IsoDateTimeFromToRangeFilter(field_name='on_sky_start_time')
    on_sky_stop_time = filters.IsoDateTimeFromToRangeFilter(field_name='on_sky_stop_time')
    process_start_time = filters.IsoDateTimeFromToRangeFilter(field_name='process_start_time')
    process_stop_time = filters.IsoDateTimeFromToRangeFilter(field_name='process_stop_time')
    scheduled_start_time = filters.IsoDateTimeFromToRangeFilter(field_name='scheduled_start_time')
    scheduled_stop_time = filters.IsoDateTimeFromToRangeFilter(field_name='scheduled_stop_time')
    actual_on_sky_start_time = filters.IsoDateTimeFromToRangeFilter(field_name='actual_on_sky_start_time')
    actual_on_sky_stop_time = filters.IsoDateTimeFromToRangeFilter(field_name='actual_on_sky_stop_time')
    actual_process_start_time = filters.IsoDateTimeFromToRangeFilter(field_name='actual_process_start_time')
    actual_process_stop_time = filters.IsoDateTimeFromToRangeFilter(field_name='actual_process_stop_time')
    project = filters.CharFilter(field_name='draft__scheduling_set__project__name', lookup_expr='icontains')
    status = filters.MultipleChoiceFilter(field_name='status', choices=[(c.value, c.value) for c in models.SchedulingUnitStatus.Choices])
    rank = filters.RangeFilter(field_name='rank')
    id = NumberInFilter(field_name='id', lookup_expr='in')
    id_min = filters.NumberFilter(field_name='id', lookup_expr='gte')
    id_max = filters.NumberFilter(field_name='id', lookup_expr='lte')
    duration_min = filters.DurationFilter(field_name='duration', lookup_expr='gte')
    duration_max = filters.DurationFilter(field_name='duration', lookup_expr='lte')
    on_sky_duration_min = filters.DurationFilter(field_name='on_sky_duration', lookup_expr='gte')
    on_sky_duration_max = filters.DurationFilter(field_name='on_sky_duration', lookup_expr='lte')
    observed_duration_min = filters.DurationFilter(field_name='observed_duration', lookup_expr='gte')
    observed_duration_max = filters.DurationFilter(field_name='observed_duration', lookup_expr='lte')
    created_at = filters.IsoDateTimeFromToRangeFilter(field_name='created_at')
    updated_at = filters.IsoDateTimeFromToRangeFilter(field_name='updated_at')
    draft = ModelChoiceInFilter(field_name='draft', lookup_expr='in', queryset=models.SchedulingUnitDraft.objects.all())
    draft_min = filters.NumberFilter(field_name='draft__id', lookup_expr='gte')
    draft_max = filters.NumberFilter(field_name='draft__id', lookup_expr='lte')
    angle1 = TargetObservationSpecificationDocFloatRangeFilter(field_name='specifications_doc__station_configuration__SAPs__0__digital_pointing__angle1')
    angle2 = TargetObservationSpecificationDocFloatRangeFilter(field_name='specifications_doc__station_configuration__SAPs__0__digital_pointing__angle2')
    direction_type = TargetObservationSpecificationFilterCharFilter(field_name='specifications_doc__station_configuration__SAPs__0__digital_pointing__direction_type', lookup_expr='iexact')
    scheduling_set = filters.CharFilter(field_name='draft__scheduling_set__name', lookup_expr='icontains')  # remove to filter based on list of options
    observation_strategy_template = filters.ModelChoiceFilter(field_name='draft__observation_strategy_template', queryset=models.SchedulingUnitObservingStrategyTemplate.objects.all())
    observation_strategy_template_name = filters.CharFilter(field_name='draft__observation_strategy_template__name', lookup_expr='icontains')
    observation_strategy_template_description = filters.CharFilter(field_name='draft__observation_strategy_template__description', lookup_expr='icontains')

    class Meta:
        model = models.SchedulingUnitBlueprint
        fields = '__all__'
        filter_overrides = FILTER_OVERRIDES


class SchedulingUnitBlueprintViewSet(LOFARViewSet):
    queryset = models.SchedulingUnitBlueprint.objects.all()
    serializer_class = serializers.SchedulingUnitBlueprintSerializer
    filter_class = SchedulingUnitBlueprintPropertyFilter  # note that this breaks other filter backends from LOFARViewSet

    # prefetch all reverse related references from other models on their related_name to avoid a ton of duplicate queries
    queryset = queryset.prefetch_related('task_blueprints',
                                         'task_blueprints__specifications_template',
                                         'task_blueprints__specifications_template__connector_types',
                                         'task_blueprints__subtasks',
                                         'task_blueprints__first_scheduling_relation',
                                         'task_blueprints__second_scheduling_relation',
                                         'task_blueprints__first_scheduling_relation__placement',
                                         'task_blueprints__second_scheduling_relation__placement',
                                         'task_blueprints__produced_by',
                                         'task_blueprints__consumed_by',
                                         'task_blueprints__status',
                                         'task_blueprints__draft',
                                         'task_blueprints__draft__first_scheduling_relation',
                                         'task_blueprints__draft__second_scheduling_relation',
                                         'task_blueprints__draft__first_scheduling_relation__placement',
                                         'task_blueprints__draft__second_scheduling_relation__placement',
                                         'task_blueprints__draft__produced_by',
                                         'task_blueprints__draft__consumed_by',
                                         'task_blueprints__draft__specifications_template__type',
                                         'task_blueprints__draft__specifications_template__connector_types'
                                         )

    # use select_related for forward related references
    queryset = queryset.select_related('specifications_template',
                                       'draft',
                                       'draft__observation_strategy_template',
                                       'draft__observation_strategy_template__state',
                                       'draft__observation_strategy_template__purpose',
                                       'draft__scheduling_set',
                                       'draft__priority_queue',
                                       'status',
                                       'priority_queue',
                                       'scheduling_constraints_template',
                                       'global_identifier')

    @swagger_auto_schema(responses={201: "This SchedulingUnitBlueprint, with references to its created TaskBlueprints and Subtasks.",
                                    403: 'forbidden'},
                         operation_description="update the task_blueprints and subtasks graph to reflect the draft graph.")
    @action(methods=['post'], detail=True, url_name="update_task_blueprints_and_subtasks_graph_from_draft")
    def update_task_blueprints_and_subtasks_graph_from_draft(self, request, pk=None):
        scheduling_unit_blueprint = get_object_or_404(models.SchedulingUnitBlueprint, pk=pk)
        scheduling_unit_blueprint = update_task_blueprints_and_subtasks_graph_from_draft(scheduling_unit_blueprint)

        # return a response with the new serialized scheduling_unit_blueprint (with references to the created task_blueprint(s) and (scheduled) subtasks)
        return Response(serializers.SchedulingUnitBlueprintSerializer(scheduling_unit_blueprint, context={'request':request}).data,
                        status=status.HTTP_201_CREATED)

    @swagger_auto_schema(responses={200: 'The available logging urls for all Subtasks of this SchedulingUnitBlueprint.',
                                    403: 'forbidden'},
                         operation_description="Get the subtask logging urls of this schedulingunit blueprint.")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, url_name='get_all_subtasks_log_urls')
    def get_all_subtasks_log_urls(self, request, pk=None):
        subtasks = models.Subtask.objects.filter(task_blueprint__scheduling_unit_blueprint_id=pk)
        result = []
        for subtask in subtasks:
            if subtask.log_url != "":
                result.append({"subtaskid": subtask.id, "type": subtask.specifications_template.type.value, "log_url": subtask.log_url})
        # TypeError: In order to allow non-dict objects to be serialized set the safe parameter to False.
        # result is list of dict so thats why
        return JsonResponse(result, safe=False)

    @swagger_auto_schema(responses={200: 'The cancelled version of this scheduling_unit',
                                    403: 'forbidden',
                                    500: 'The scheduling_unit could not be cancelled'},
                         operation_description="Try to cancel this scheduling_unit.")
    @action(methods=['post'], detail=True, url_name="cancel")
    def cancel(self, request, pk=None):
        scheduling_unit_blueprint = get_object_or_404(models.SchedulingUnitBlueprint, pk=pk)
        from lofar.sas.tmss.tmss.tmssapp.tasks import cancel_scheduling_unit_blueprint
        scheduling_unit_blueprint = cancel_scheduling_unit_blueprint(scheduling_unit_blueprint)
        serializer = self.get_serializer(scheduling_unit_blueprint)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The obsolete version of this scheduling_unit',
                                    403: 'forbidden',
                                    500: 'The scheduling_unit could not be marked as obsolete. Was it in a state from which it is allowed to mark it as obsolete?'},
                         operation_description="Try to mark this scheduling_unit as obsolete (when the status is cancelled/error)")
    @action(methods=['post'], detail=True, url_name="mark_as_obsolete")
    def mark_as_obsolete(self, request, pk=None):
        scheduling_unit_blueprint = get_object_or_404(models.SchedulingUnitBlueprint, pk=pk)
        from lofar.sas.tmss.tmss.tmssapp.tasks import mark_scheduling_unit_blueprint_as_obsolete
        scheduling_unit_blueprint = mark_scheduling_unit_blueprint_as_obsolete(scheduling_unit_blueprint)
        serializer = self.get_serializer(scheduling_unit_blueprint)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: "All Subtasks in this SchedulingUnitBlueprint",
                                    403: 'forbidden'},
                         operation_description="Get all subtasks for this scheduling_unit")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, url_name="subtasks", name="all subtasks in this scheduling_unit")
    def subtasks(self, request, pk=None):
        subtasks = models.Subtask.objects.all().filter(task_blueprint__scheduling_unit_blueprint_id=pk). \
                                                select_related('state', 'specifications_template', 'specifications_template__type', 'cluster', 'created_or_updated_by_user').order_by('id').all()

        # return a response with the new serialized scheduling_unit_blueprint (with references to the created task_blueprint(s) and (scheduled) subtasks)
        return Response(serializers.SubtaskSerializer(subtasks, many=True, context={'request':request}).data,
                        status=status.HTTP_200_OK)


    @swagger_auto_schema(responses={201: "This SchedulingUnitBlueprint, with references to the created Cleanup TaskBlueprints.",
                                    403: 'forbidden'},
                         operation_description="Create a cleanup task for this scheduling unit.")
    @action(methods=['post'], detail=True, url_name="create_cleanuptask", name="Create a cleanup task for this scheduling unit")
    def create_cleanuptask_for_scheduling_unit_blueprint(self, request, pk=None):
        scheduling_unit_blueprint = get_object_or_404(models.SchedulingUnitBlueprint, pk=pk)
        scheduling_unit_blueprint = create_cleanuptask_for_scheduling_unit_blueprint(scheduling_unit_blueprint)

        # return a response with the new serialized scheduling_unit_blueprint (with references to the created task_blueprint and subtask)
        return Response(serializers.SchedulingUnitBlueprintSerializer(scheduling_unit_blueprint, context={'request':request}).data,
                        status=status.HTTP_201_CREATED)


    @swagger_auto_schema(responses={201: "A JSON document reflecting the current scheduling unit blueprint's task graph and settings",
                                    403: 'forbidden'},
                         operation_description="Get a json specifications_doc reflecting the current scheduling unit draft's task graph and settings.")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, url_name="specifications_doc", name="Get a json specifications_doc reflecting the current scheduling unit blueprint's task graph and settings.")
    def specifications_doc(self, request, pk=None):
        scheduling_unit_blueprint = get_object_or_404(models.SchedulingUnitBlueprint, pk=pk)
        return JsonResponse(scheduling_unit_blueprint.specifications_doc, json_dumps_params={'indent': 2, 'sort_keys': True})


    @swagger_auto_schema(responses={201: "A JSON document reflecting the current scheduling unit blueprint's task graph and settings, with just the allowed-to-be-changed parameters from the template. Rely on defaults for the rest.",
                                    403: 'forbidden'},
                         operation_description="Get a json specifications_doc reflecting the current scheduling unit blueprint's task graph and settings, with just the allowed-to-be-changed parameters from the template. Rely on defaults for the rest")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, url_name="specifications_doc_with_just_the_parameters", name="Get a json specifications_doc (parameters only)")
    def specifications_doc_with_just_the_parameters(self, request, pk=None):
        scheduling_unit_blueprint = get_object_or_404(models.SchedulingUnitBlueprint, pk=pk)
        return JsonResponse(scheduling_unit_blueprint.specifications_doc_with_just_the_parameters, json_dumps_params={'indent': 2, 'sort_keys': True})


    @swagger_auto_schema(responses={201: "A JSON document reflecting the current scheduling unit blueprint's task graph and settings, including copies of tasks and relations for failed tasks, intended to be used to update the draft and (re)create the copies the failed tasks for a 'rerun'",
                                    403: 'forbidden'},
                         operation_description="Get a json specifications_doc reflecting the current scheduling unit draft's task graph and settings.")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, url_name="specifications_doc_including_copies_for_failed_tasks", name="Get a json specifications_doc reflecting the current scheduling unit blueprint's task graph and settings, including copies of tasks and relations for failed tasks, intended to be used to update the draft and (re)create the copies the failed tasks for a 'rerun'")
    def specifications_doc_including_copies_for_failed_tasks(self, request, pk=None):
        scheduling_unit_blueprint = get_object_or_404(models.SchedulingUnitBlueprint, pk=pk)
        return JsonResponse(scheduling_unit_blueprint.specifications_doc_including_copies_for_failed_tasks, json_dumps_params={'indent': 2, 'sort_keys': True})


    @swagger_auto_schema(responses={201: "This SchedulingUnitBlueprint",
                                    403: 'forbidden'},
                         operation_description="Copy this blueprint's specifications_doc back into the originating draft scheduling unit.")
    @action(methods=['post'], detail=True, url_name="copy_specifications_doc_back_into_draft", name="Copy this blueprint's specifications_doc back into the originating draft scheduling unit")
    def copy_specifications_doc_back_into_draft(self, request, pk=None):
        scheduling_unit_blueprint = get_object_or_404(models.SchedulingUnitBlueprint, pk=pk)
        update_task_graph_from_specifications_doc(scheduling_unit_blueprint.draft, scheduling_unit_blueprint.specifications_doc)

        # return this scheduling_unit_blueprint (which containts a link to the draft with the updated specifications_doc)
        scheduling_unit_blueprint.refresh_from_db()
        serializer = self.get_serializer(scheduling_unit_blueprint)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={201: "This SchedulingUnitBlueprint",
                                    403: 'forbidden'},
                         operation_description="Copy this blueprint's specifications_doc including copies of failed tasks back into the originating draft scheduling unit to prepare and edit reruns.")
    @action(methods=['post'], detail=True, url_name="copy_specifications_doc_including_copies_for_failed_tasks_back_into_draft", name="Copy this blueprint's specifications_doc including copies of failed tasks back into the originating draft scheduling unit to prepare and edit reruns")
    def copy_specifications_doc_including_copies_for_failed_tasks_back_into_draft(self, request, pk=None):
        scheduling_unit_blueprint = get_object_or_404(models.SchedulingUnitBlueprint, pk=pk)
        update_task_graph_from_specifications_doc(scheduling_unit_blueprint.draft, scheduling_unit_blueprint.specifications_doc_including_copies_for_failed_tasks)

        # return this scheduling_unit_blueprint (which containts a link to the draft with the updated specifications_doc)
        scheduling_unit_blueprint.refresh_from_db()
        serializer = self.get_serializer(scheduling_unit_blueprint)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={201: "This SchedulingUnitBlueprint including new TaskBlueprints for the failed tasks",
                                    403: 'forbidden'},
                         operation_description="Create runnable TaskBlueprint copies of failed tasks via the SchedulingUnitDraft and add them to this blueprint.")
    @action(methods=['post'], detail=True, url_name="create_copies_of_failed_tasks_via_draft", name="Create runnable TaskBlueprint copies of failed tasks via the SchedulingUnitDraft and add them to this blueprint")
    def create_copies_of_failed_tasks_via_draft(self, request, pk=None):
        scheduling_unit_blueprint = get_object_or_404(models.SchedulingUnitBlueprint, pk=pk)

        # update the task graph via the draft with copies of the failed tasks
        scheduling_unit_draft = scheduling_unit_blueprint.draft
        update_task_graph_from_specifications_doc(scheduling_unit_draft, scheduling_unit_blueprint.specifications_doc_including_copies_for_failed_tasks)

        # and then reflect the new draft graph in this blueprint
        scheduling_unit_blueprint = update_task_blueprints_and_subtasks_graph_from_draft(scheduling_unit_blueprint)

        serializer = self.get_serializer(scheduling_unit_blueprint)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={201: "This blueprint's SchedulingUnitDraft including new editable TaskDrafts copies for the failed TaskBlueprints",
                                    403: 'forbidden'},
                         operation_description="Create editable TaskDraft copies of failed tasks in the original SchedulingUnitDraft.")
    @action(methods=['post'], detail=True, url_name="create_copies_of_failed_tasks_in_draft", name="Create editable TaskDraft copies of failed tasks in the original SchedulingUnitDraft")
    def create_copies_of_failed_tasks_in_draft(self, request, pk=None):
        scheduling_unit_blueprint = get_object_or_404(models.SchedulingUnitBlueprint, pk=pk)

        # update the task graph in the draft with copies of the failed tasks
        scheduling_unit_draft = scheduling_unit_blueprint.draft
        scheduling_unit_draft = update_task_graph_from_specifications_doc(scheduling_unit_draft, scheduling_unit_blueprint.specifications_doc_including_copies_for_failed_tasks)

        # return the updated draft
        return Response(serializers.SchedulingUnitDraftSerializer(scheduling_unit_draft, context={'request':request}).data,
                        status=status.HTTP_200_OK)


    @swagger_auto_schema(responses={201: 'The newly copied SchedulingUnitDraft',
                                    403: 'forbidden'},
                         operation_description="Copy this blueprint into a new SchedulingUnitDraft instance")
    @action(methods=['post'], detail=True, url_name="copy_to_draft", name="Copy into new draft")
    def copy_to_draft(self, request, pk=None):
        scheduling_unit_blueprint = get_object_or_404(models.SchedulingUnitBlueprint, pk=pk)

        remove_lofar1_stations = strtobool(request.query_params.get('remove_lofar1_stations', 'False'))
        remove_lofar2_stations = strtobool(request.query_params.get('remove_lofar2_stations', 'False'))
        scheduling_unit_draft_copy = create_scheduling_unit_draft_from_scheduling_unit_blueprint(scheduling_unit_blueprint, remove_lofar1_stations, remove_lofar2_stations)

        # return a response with the new serialized SchedulingUnitBlueprintExtendedSerializer
        return Response(serializers.SchedulingUnitDraftExtendedSerializer(scheduling_unit_draft_copy, context={'request':request}).data,
                        status=status.HTTP_201_CREATED)


    @swagger_auto_schema(responses={201: 'The newly created SchedulingUnitBlueprint-copy with the given station(s) removed',
                                    403: 'forbidden'},
                         operation_description="Create a new SchedulingUnitBlueprint-copy with the given station(s) removed, cancel this unit and mark it obsolete")
    @action(methods=['post'], detail=True, url_name="create_copy_without_given_stations")
    def create_copy_without_given_stations(self, request, pk=None):
        scheduling_unit_blueprint = get_object_or_404(models.SchedulingUnitBlueprint, pk=pk)
        stations_to_be_removed = set([s.upper() for s in request.data['stations_to_be_removed']])
        with transaction.atomic():
            # start with creating a draft copy
            scheduling_unit_draft_copy = create_scheduling_unit_draft_from_scheduling_unit_blueprint(scheduling_unit_blueprint, add_copy_annotations=False)

            # and then remove the given stations from each task
            for task in scheduling_unit_draft_copy.tasks.all():
                if 'station_configuration' in task.specifications_doc:
                    if 'station_groups' in task.specifications_doc['station_configuration']:
                        for group in task.specifications_doc['station_configuration']['station_groups']:
                            group['stations'] = sorted(list(set(group['stations']) - stations_to_be_removed))
                            group['max_nr_missing'] = min(len(group['stations'])-1, group.get('max_nr_missing', 0))
                        task.save()

            # blueprint the copy
            scheduling_unit_blueprint_copy = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft_copy)

            # and cancel and obsolete the original
            mark_scheduling_unit_blueprint_as_obsolete(cancel_scheduling_unit_blueprint(scheduling_unit_blueprint))

            # return the new blueprint
            return Response(serializers.SchedulingUnitBlueprintSerializer(scheduling_unit_blueprint_copy, context={'request':request}).data,
                            status=status.HTTP_201_CREATED)


    @swagger_auto_schema(responses={200: 'The system events of this scheduling unit blueprint',
                                    403: 'forbidden'},
                         operation_description="Get the system events of this scheduling unit blueprint.")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, url_name="system_events")
    def system_events(self, request, pk=None):
        sub = get_object_or_404(models.SchedulingUnitBlueprint, pk=pk)
        system_events = self.filter_queryset(sub.system_events)
        serializer = self.get_serializer(system_events, many=True)
        return RestResponse(serializer.data)

    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, renderer_classes=(ImagePNGRenderer,), url_path='scheduling_constraints_plot/?(?P<station>([a-zA-Z]{2}\d{3}))?/?(?P<proposed_start_time_or_date>(\d{4}-[01]\d-[0-3]\d(T[0-2]\d:[0-5]\d)?))?')
    def scheduling_constraints_plot(self, request, pk=None, station=None, proposed_start_time_or_date=None):
        scheduling_unit = get_object_or_404(models.SchedulingUnitBlueprint, pk=pk)

        if proposed_start_time_or_date is None:
            from lofar.sas.tmss.services.scheduling.constraints import get_at_constraint_timestamp
            at_timestamp = get_at_constraint_timestamp(scheduling_unit)
            if at_timestamp is not None:
                proposed_start_time_or_date = at_timestamp.isoformat()
            else:
                return HttpResponseBadRequest("Cannot create scheduling constraints plot: scheduling_unit_blueprint id=%s is not scheduled yet." %(pk,))

        proposed_start_time = None
        proposed_date = None
        if proposed_start_time_or_date is not None:
            parsed_start_time_or_date = dateutil.parser.parse(proposed_start_time_or_date)
            if len(proposed_start_time_or_date) > len("2023-01-01"):
                proposed_start_time = parsed_start_time_or_date
            else:
                proposed_date = parsed_start_time_or_date.date()

        from lofar.sas.tmss.tmss.tmssapp.adapters.plots import scheduling_constraints_plot
        plot = scheduling_constraints_plot(scheduling_unit, station=station, proposed_start_time=proposed_start_time,
                                           proposed_date=proposed_date, include_min_max_elevations=True)
        return Response(plot, content_type='image/png')


    @swagger_auto_schema(responses={200: 'This scheduling_unit with updated constraints',
                                    403: 'forbidden'},
                         operation_description="Try to mark this scheduling_unit as fixed_time-scheduled, with the current scheduled_start_time as 'at' constraint")
    @action(methods=['post'], detail=True, url_name="mark_fixed_time_scheduled_at_scheduled_starttime")
    def mark_fixed_time_scheduled_at_scheduled_starttime(self, request, pk=None):
        scheduling_unit_blueprint = get_object_or_404(models.SchedulingUnitBlueprint, pk=pk)
        from lofar.sas.tmss.tmss.tmssapp.tasks import mark_scheduling_unit_fixed_time_scheduled_at_scheduled_starttime
        scheduling_unit_blueprint = mark_scheduling_unit_fixed_time_scheduled_at_scheduled_starttime(scheduling_unit_blueprint)
        serializer = self.get_serializer(scheduling_unit_blueprint)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'This scheduled scheduling_unit',
                                    403: 'forbidden'},
                         operation_description="Try to schedule this scheduling_unit at the given start_time")
    @action(methods=['post'], detail=True, url_name="schedule")
    def schedule(self, request, pk=None):
        scheduling_unit_blueprint = get_object_or_404(models.SchedulingUnitBlueprint, pk=pk)
        start_time = datetime.strptime(request.data.get('start_time', ''), '%Y-%m-%dT%H:%M:%S')
        from lofar.sas.tmss.tmss.tmssapp.tasks import schedule_independent_subtasks_in_scheduling_unit_blueprint
        with transaction.atomic():
            if scheduling_unit_blueprint.scheduling_constraints_doc is not None:
                if 'time' not in scheduling_unit_blueprint.scheduling_constraints_doc:
                    scheduling_unit_blueprint.scheduling_constraints_doc['time'] = {}
                scheduling_unit_blueprint.scheduling_constraints_doc['time']['at'] = start_time.isoformat()
                scheduling_unit_blueprint.scheduling_constraints_doc['scheduler'] = 'fixed_time'
                scheduling_unit_blueprint.save()

            if scheduling_unit_blueprint.status.value == SchedulingUnitStatus.Choices.SCHEDULED.value:
                scheduling_unit_blueprint = unschedule_subtasks_in_scheduling_unit_blueprint(scheduling_unit_blueprint)
            elif scheduling_unit_blueprint.status.value == SchedulingUnitStatus.Choices.UNSCHEDULABLE.value:
                scheduling_unit_blueprint = mark_subtasks_in_scheduling_unit_blueprint_as_schedulable(scheduling_unit_blueprint)
            scheduling_unit_blueprint = schedule_independent_subtasks_in_scheduling_unit_blueprint(scheduling_unit_blueprint, start_time)

        serializer = self.get_serializer(scheduling_unit_blueprint)
        return RestResponse(serializer.data)

    @swagger_auto_schema(responses={200: 'This unscheduled scheduling_unit=',
                                    403: 'forbidden'},
                         operation_description="Try to unschedule this scheduling_unit")
    @action(methods=['post'], detail=True, url_name="unschedule")
    def unschedule(self, request, pk=None):
        scheduling_unit_blueprint = get_object_or_404(models.SchedulingUnitBlueprint, pk=pk)
        from lofar.sas.tmss.tmss.tmssapp.tasks import unschedule_subtasks_in_scheduling_unit_blueprint
        scheduling_unit_blueprint = unschedule_subtasks_in_scheduling_unit_blueprint(scheduling_unit_blueprint)
        serializer = self.get_serializer(scheduling_unit_blueprint)
        return RestResponse(serializer.data)

    @swagger_auto_schema(responses={200: 'This unscheduled scheduling_unit=',
                                    403: 'forbidden'},
                         operation_description="Try to unschedule this scheduling_unit")
    @action(methods=['post'], detail=True, url_name="unschedule")
    def unschedule(self, request, pk=None):
        scheduling_unit_blueprint = get_object_or_404(models.SchedulingUnitBlueprint, pk=pk)
        from lofar.sas.tmss.tmss.tmssapp.tasks import unschedule_subtasks_in_scheduling_unit_blueprint
        scheduling_unit_blueprint = unschedule_subtasks_in_scheduling_unit_blueprint(scheduling_unit_blueprint)
        serializer = self.get_serializer(scheduling_unit_blueprint)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'This scheduling_unit with updated constraints',
                                    403: 'forbidden'},
                         operation_description="Try to mark this scheduling_unit as dynamically-scheduled, and let the scheduler find a start_time")
    @action(methods=['post'], detail=True, url_name="mark_dynamically_scheduled")
    def mark_dynamically_scheduled(self, request, pk=None):
        scheduling_unit_blueprint = get_object_or_404(models.SchedulingUnitBlueprint, pk=pk)
        from lofar.sas.tmss.tmss.tmssapp.tasks import mark_scheduling_unit_dynamically_scheduled
        scheduling_unit_blueprint = mark_scheduling_unit_dynamically_scheduled(scheduling_unit_blueprint)
        serializer = self.get_serializer(scheduling_unit_blueprint)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The schedule was reset',
                                    403: 'forbidden'},
                         operation_description="reset the schedule: marking the unschedulable units as schedulable, and then move all schedulable to a 'parking'-timestamp")
    @action(methods=['post'], detail=False, url_name="reset_schedule")
    def reset_schedule(self, request, pk=None):
        from lofar.sas.tmss.tmss.tmssapp.subtasks import update_subtasks_start_times_for_scheduling_unit, mark_subtasks_in_task_blueprint_as_schedulable
        for unit in SchedulingUnitBlueprint.objects.filter(status__value='unschedulable', obsolete_since__isnull=True).all():
            mark_subtasks_in_task_blueprint_as_schedulable(unit)

        if strtobool(request.query_params.get('scheduled', 'False')):
            for unit in SchedulingUnitBlueprint.objects.filter(status__value='scheduled', obsolete_since__isnull=True).all():
                unschedule_subtasks_in_scheduling_unit_blueprint(unit)

        # default 'parking'-timestamp is at noon seven days from now.
        timestamp = (datetime.utcnow() + timedelta(days=7)).replace(hour=12, minute=0, second=0, microsecond=0)
        timestamp_str = request.query_params.get('timestamp', None)
        if timestamp_str:
            timestamp = dateutil.parser.parse(timestamp_str)

        for unit in SchedulingUnitBlueprint.objects.filter(status__value='schedulable', obsolete_since__isnull=True).all():
            update_subtasks_start_times_for_scheduling_unit(unit, start_time=timestamp)

        # just return a success
        return HttpResponse(status=200)


    @swagger_auto_schema(responses={201: "The new lofar2 sibling SchedulingUnitBlueprint.",
                                    403: 'forbidden'},
                         operation_description="Create a copy of this unit for the lofar2 stations, and schedule it in parallel with this original.")
    @action(methods=['post'], detail=True, url_name="create_lofar2_sibling")
    def create_lofar2_sibling(self, request, pk=None):
        scheduling_unit_blueprint = get_object_or_404(models.SchedulingUnitBlueprint, pk=pk)
        sibling_blueprint = create_lofar2_sibling_scheduling_unit_draft_and_blueprint(scheduling_unit_blueprint)

        # return a response with the new serialized sibling_blueprint (with references to the created task_blueprint(s) and (scheduled) subtasks)
        return Response(serializers.SchedulingUnitBlueprintSerializer(sibling_blueprint, context={'request':request}).data,
                        status=status.HTTP_201_CREATED)




class SchedulingUnitBlueprintExtendedViewSet(SchedulingUnitBlueprintViewSet):
    serializer_class = serializers.SchedulingUnitBlueprintExtendedSerializer


class SchedulingUnitBlueprintNestedViewSet(LOFARNestedViewSet):
    queryset = models.SchedulingUnitBlueprint.objects.all()
    serializer_class = serializers.SchedulingUnitBlueprintSerializer

    def get_queryset(self):
        if 'scheduling_unit_draft_id' in self.kwargs:
            scheduling_unit_draft = get_object_or_404(models.SchedulingUnitDraft, pk=self.kwargs['scheduling_unit_draft_id'])
            return scheduling_unit_draft.scheduling_unit_blueprints.all()
        else:
            return models.SchedulingUnitBlueprint.objects.all()

class TaskDraftPropertyFilter(property_filters.PropertyFilterSet):
    id = NumberInFilter(field_name='id', lookup_expr='in')
    id_min = filters.NumberFilter(field_name='id', lookup_expr='gte')
    id_max = filters.NumberFilter(field_name='id', lookup_expr='lte')
    scheduling_unit_draft = ModelChoiceInFilter(field_name='scheduling_unit_draft', lookup_expr='in', queryset=models.SchedulingUnitDraft.objects.all())
    scheduling_unit_draft_min = filters.NumberFilter(field_name='scheduling_unit_draft__id', lookup_expr='gte')
    scheduling_unit_draft_max = filters.NumberFilter(field_name='scheduling_unit_draft__id', lookup_expr='lte')
    scheduling_unit_draft_name = filters.CharFilter(field_name='scheduling_unit_draft__name', lookup_expr='icontains')
    scheduling_set_name = filters.CharFilter(field_name='scheduling_unit_draft__scheduling_set__name', lookup_expr='icontains')
    project = filters.CharFilter(field_name='scheduling_unit_draft__scheduling_set__project__name', lookup_expr='icontains')
    task_blueprints = filters.ModelMultipleChoiceFilter(field_name='task_blueprints', queryset=models.TaskBlueprint.objects.all())
    duration_min = property_filters.PropertyDurationFilter(field_name='duration', lookup_expr='gte')
    duration_max = property_filters.PropertyDurationFilter(field_name='duration', lookup_expr='lte')
    created_at = filters.IsoDateTimeFromToRangeFilter(field_name='created_at')
    updated_at = filters.IsoDateTimeFromToRangeFilter(field_name='updated_at')
    relative_start_time_min = property_filters.PropertyDurationFilter(field_name='relative_start_time', lookup_expr='gte')
    relative_start_time_max = property_filters.PropertyDurationFilter(field_name='relative_start_time', lookup_expr='lte')
    relative_stop_time_min = property_filters.PropertyDurationFilter(field_name='relative_stop_time', lookup_expr='gte')
    relative_stop_time_max = property_filters.PropertyDurationFilter(field_name='relative_stop_time', lookup_expr='lte')

    class Meta:
        model = models.TaskDraft
        fields = '__all__'
        filter_overrides = FILTER_OVERRIDES


class TaskDraftViewSet(LOFARViewSet):
    queryset = models.TaskDraft.objects.all()
    serializer_class = serializers.TaskDraftSerializer
    filter_class = TaskDraftPropertyFilter

    # prefetch all reverse related references from other models on their related_name to avoid a ton of duplicate queries
    queryset = queryset.prefetch_related('first_scheduling_relation') \
                       .prefetch_related('second_scheduling_relation')\
                       .prefetch_related('produced_by')\
                       .prefetch_related('consumed_by')\
                       .prefetch_related('task_blueprints')\
                       .prefetch_related('scheduling_unit_draft')

    # prefetch nested references in reverse models to avoid duplicate lookup queries
    queryset = queryset.prefetch_related('first_scheduling_relation__placement') \
                       .prefetch_related('second_scheduling_relation__placement') \
                       .prefetch_related('task_blueprints__status') \
                       .prefetch_related('task_blueprints__specifications_template__type') \
                       .prefetch_related('task_blueprints__subtasks')\
                       .prefetch_related('task_blueprints__subtasks__state') \
                       .prefetch_related('task_blueprints__subtasks__cluster') \
                       .prefetch_related('task_blueprints__subtasks__specifications_template') \
                       .prefetch_related('task_blueprints__subtasks__specifications_template__type') \
                       .prefetch_related('specifications_template__type') \
                       .prefetch_related('specifications_template__connector_types') \
                       .prefetch_related('scheduling_unit_draft__scheduling_set') \

    @swagger_auto_schema(responses={201: 'The created task blueprint',
                                    403: 'forbidden'},
                         operation_description="Carve this draft task specification in stone, and make an (uneditable) blueprint out of it.")
    @action(methods=['post'], detail=True, url_name="create_task_blueprint", name="Create TaskBlueprint")
    def create_task_blueprint(self, request, pk=None):
        task_draft = get_object_or_404(models.TaskDraft, pk=pk)
        task_blueprint = create_task_blueprint_from_task_draft(task_draft)

        # return a response with the new serialized TaskBlueprint
        return Response(serializers.TaskBlueprintSerializer(task_blueprint, context={'request':request}).data,
                        status=status.HTTP_201_CREATED)

    @swagger_auto_schema(responses={201: "This TaskBlueprint, with its created subtask(s)",
                                    403: 'forbidden'},
                         operation_description="Create subtasks.")
    @action(methods=['post'], detail=True, url_name="create_task_blueprint_subtasks", name="Create TaskBlueprint and its Subtask(s)")
    def create_task_blueprint_and_subtasks(self, request, pk=None):
        task_draft = get_object_or_404(models.TaskDraft, pk=pk)
        task_blueprint = create_task_blueprint_and_subtasks_from_task_draft(task_draft)

        # return a response with the new serialized TaskBlueprint
        return Response(serializers.TaskBlueprintSerializer(task_blueprint, context={'request':request}).data,
                        status=status.HTTP_201_CREATED)



    @swagger_auto_schema(responses={201: 'The newly copied TaskDraft',
                                    403: 'forbidden'},
                         operation_description="Copy this TaskDraft to a new instance.")
    @action(methods=['post'], detail=True, url_name="copy", name="Create Copy")
    def copy(self, request, pk=None):
        task_draft = get_object_or_404(models.TaskDraft, pk=pk)

        remove_lofar1_stations = strtobool(request.query_params.get('remove_lofar1_stations', 'False'))
        remove_lofar2_stations = strtobool(request.query_params.get('remove_lofar2_stations', 'False'))

        task_draft_copy = copy_task_draft(task_draft, remove_lofar1_stations, remove_lofar2_stations)

        return Response(serializers.TaskDraftSerializer(task_draft_copy, context={'request':request}).data,
                        status=status.HTTP_201_CREATED)


    @swagger_auto_schema(responses={200: 'The predecessor task draft of this task draft',
                                    403: 'forbidden'},
                         operation_description="Get the predecessor task draft of this task draft.")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, url_name="predecessors")
    def predecessors(self, request, pk=None):
        task_draft = get_object_or_404(models.TaskDraft, pk=pk)
        predecessors = self.filter_queryset(task_draft.predecessors)
        serializer = self.get_serializer(predecessors, many=True)
        return Response(serializer.data)

    @swagger_auto_schema(responses={200: 'The successor subtasks of this subtask',
                                    403: 'forbidden'},
                         operation_description="Get the successor subtasks of this subtask.")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, url_name="successors")
    def successors(self, request, pk=None):
        task_draft = get_object_or_404(models.TaskDraft, pk=pk)
        successors = self.filter_queryset(task_draft.successors)
        serializer = self.get_serializer(successors, many=True)
        return Response(serializer.data)


class TaskDraftExtendedViewSet(TaskDraftViewSet):
    serializer_class = serializers.TaskDraftExtendedSerializer


class TaskDraftNestedViewSet(LOFARNestedViewSet):
     queryset = models.TaskDraft.objects.all()
     serializer_class = serializers.TaskDraftSerializer

     def get_queryset(self):
         if 'scheduling_unit_draft_id' in self.kwargs:
             scheduling_unit_draft = get_object_or_404(models.SchedulingUnitDraft, pk=self.kwargs['scheduling_unit_draft_id'])
             return scheduling_unit_draft.task_drafts.all()
         else:
             return models.TaskDraft.objects.all()


class TaskBlueprintPropertyFilter(property_filters.PropertyFilterSet):
    id = NumberInFilter(field_name='id', lookup_expr='in')
    id_min = filters.NumberFilter(field_name='id', lookup_expr='gte')
    id_max = filters.NumberFilter(field_name='id', lookup_expr='lte')
    scheduling_unit_blueprint = ModelChoiceInFilter(field_name='scheduling_unit_blueprint', lookup_expr='in', queryset=models.SchedulingUnitBlueprint.objects.all())
    scheduling_unit_blueprint_min = filters.NumberFilter(field_name='scheduling_unit_blueprint__id', lookup_expr='gte')
    scheduling_unit_blueprint_max = filters.NumberFilter(field_name='scheduling_unit_blueprint__id', lookup_expr='lte')
    scheduling_unit_blueprint_name = filters.CharFilter(field_name='scheduling_unit_blueprint__name', lookup_expr='icontains')
    scheduling_set_name = filters.CharFilter(field_name='draft__scheduling_unit_draft__scheduling_set__name', lookup_expr='icontains')
    project = filters.CharFilter(field_name='draft__scheduling_unit_draft__scheduling_set__project__name', lookup_expr='icontains')
    draft = ModelChoiceInFilter(field_name='draft',  lookup_expr='in', queryset=models.TaskDraft.objects.all())
    draft_min = filters.NumberFilter(field_name='draft__id', lookup_expr='gte')
    draft_max = filters.NumberFilter(field_name='draft__id', lookup_expr='lte')
    duration_min = filters.DurationFilter(field_name='duration', lookup_expr='gte')
    duration_max = filters.DurationFilter(field_name='duration', lookup_expr='lte')
    on_sky_duration_min = filters.DurationFilter(field_name='on_sky_duration', lookup_expr='gte')
    on_sky_duration_max = filters.DurationFilter(field_name='on_sky_duration', lookup_expr='lte')
    created_at = filters.IsoDateTimeFromToRangeFilter(field_name='created_at')
    updated_at = filters.IsoDateTimeFromToRangeFilter(field_name='updated_at')
    relative_start_time_min = property_filters.PropertyDurationFilter(field_name='relative_start_time', lookup_expr='gte')
    relative_start_time_max = property_filters.PropertyDurationFilter(field_name='relative_start_time', lookup_expr='lte')
    relative_stop_time_min = property_filters.PropertyDurationFilter(field_name='relative_stop_time', lookup_expr='gte')
    relative_stop_time_max = property_filters.PropertyDurationFilter(field_name='relative_stop_time', lookup_expr='lte')
    on_sky_start_time = filters.IsoDateTimeFromToRangeFilter(field_name='on_sky_start_time')
    on_sky_stop_time = filters.IsoDateTimeFromToRangeFilter(field_name='on_sky_stop_time')
    process_start_time = filters.IsoDateTimeFromToRangeFilter(field_name='process_start_time')
    process_stop_time = filters.IsoDateTimeFromToRangeFilter(field_name='process_stop_time')
    scheduled_start_time = filters.IsoDateTimeFromToRangeFilter(field_name='scheduled_start_time')
    scheduled_stop_time = filters.IsoDateTimeFromToRangeFilter(field_name='scheduled_stop_time')
    actual_on_sky_start_time = filters.IsoDateTimeFromToRangeFilter(field_name='actual_on_sky_start_time')
    actual_on_sky_stop_time = filters.IsoDateTimeFromToRangeFilter(field_name='actual_on_sky_stop_time')
    actual_process_start_time = filters.IsoDateTimeFromToRangeFilter(field_name='actual_process_start_time')
    actual_process_stop_time = filters.IsoDateTimeFromToRangeFilter(field_name='actual_process_stop_time')
    status = filters.MultipleChoiceFilter(field_name='status', choices=[(c.value, c.value) for c in models.TaskStatus.Choices])
    subtasks = filters.ModelMultipleChoiceFilter(field_name='subtasks', queryset=models.Subtask.objects.all())
    subtasks_min = filters.NumberFilter(field_name='subtasks__id', lookup_expr='gte')
    subtasks_max = filters.NumberFilter(field_name='subtasks__id', lookup_expr='lte')
    obsolete = filters.BooleanFilter(label='obsolete', field_name='obsolete_since', lookup_expr='isnull', exclude=True)
    task_type = filters.MultipleChoiceFilter(label='task type(s)', field_name='specifications_template__type', choices=[(c.value, c.value) for c in models.TaskType.Choices])

    class Meta:
        model = models.TaskBlueprint
        fields = '__all__'
        filter_overrides = FILTER_OVERRIDES


class TaskBlueprintViewSet(LOFARViewSet):
    queryset = models.TaskBlueprint.objects.all()
    serializer_class = serializers.TaskBlueprintSerializer
    filter_class = TaskBlueprintPropertyFilter

    # prefetch all reverse related references from other models on their related_name to avoid a ton of duplicate queries
    queryset = queryset.prefetch_related('first_scheduling_relation')\
                       .prefetch_related('second_scheduling_relation')\
                       .prefetch_related('produced_by')\
                       .prefetch_related('consumed_by')\
                       .prefetch_related('subtasks')

    # prefetch nested references in reverse models to avoid duplicate lookup queries
    queryset = queryset.prefetch_related('first_scheduling_relation__placement') \
                       .prefetch_related('second_scheduling_relation__placement') \
                       .prefetch_related('subtasks__specifications_template') \
                       .prefetch_related('specifications_template__connector_types')


    # use select_related for forward related references
    queryset = queryset.select_related('draft', 'specifications_template', 'specifications_template__type', 'scheduling_unit_blueprint', 'status')


    @swagger_auto_schema(responses={201: 'The newly copied TaskDraft',
                                    403: 'forbidden'},
                         operation_description="Copy this TaskDraft to a new instance.")
    @action(methods=['post'], detail=True, url_name="copy_to_new_draft", name="Copy to new TaskDraft")
    def copy_to_new_draft(self, request, pk=None):
        task_blueprint = get_object_or_404(models.TaskBlueprint, pk=pk)

        remove_lofar1_stations = strtobool(request.query_params.get('remove_lofar1_stations', 'False'))
        remove_lofar2_stations = strtobool(request.query_params.get('remove_lofar2_stations', 'False'))

        task_draft_copy = copy_task_blueprint_to_task_draft(task_blueprint, remove_lofar1_stations, remove_lofar2_stations)

        return Response(serializers.TaskDraftSerializer(task_draft_copy, context={'request':request}).data,
                        status=status.HTTP_201_CREATED)


    @swagger_auto_schema(responses={201: 'The newly copied TaskBlueprint',
                                    403: 'forbidden'},
                         operation_description="Copy this TaskBlueprint to a new instance (including subtasks) via a new TaskDraft.")
    @action(methods=['post'], detail=True, url_name="copy_to_new_blueprint", name="Copy to new TaskBlueprint")
    def copy_to_new_blueprint(self, request, pk=None):
        with transaction.atomic():
            task_blueprint = get_object_or_404(models.TaskBlueprint, pk=pk)

            remove_lofar1_stations = strtobool(request.query_params.get('remove_lofar1_stations', 'False'))
            remove_lofar2_stations = strtobool(request.query_params.get('remove_lofar2_stations', 'False'))

            task_blueprint_copy = copy_task_blueprint_via_task_draft_to_new_task_blueprint(task_blueprint, remove_lofar1_stations, remove_lofar2_stations)
            create_or_update_subtasks_from_task_blueprint(task_blueprint_copy)
            task_blueprint_copy.refresh_from_db()

            return Response(serializers.TaskBlueprintSerializer(task_blueprint_copy, context={'request':request}).data,
                            status=status.HTTP_201_CREATED)


    @swagger_auto_schema(responses={201: "This TaskBlueprint, with it is created subtasks",
                                    403: 'forbidden'},
                         operation_description="Create subtasks.")
    @action(methods=['post'], detail=True, url_name="create_subtasks", name="Create Subtasks")
    def create_subtasks(self, request, pk=None):
        task_blueprint = get_object_or_404(models.TaskBlueprint, pk=pk)
        create_or_update_subtasks_from_task_blueprint(task_blueprint)
        task_blueprint.refresh_from_db()

        # return a response with the new serialized task_blueprint (with references to the created subtasks)
        return Response(serializers.TaskBlueprintSerializer(task_blueprint, context={'request':request}).data,
                        status=status.HTTP_201_CREATED)

    @swagger_auto_schema(responses={201: "This TaskBlueprint, with the scheduled subtasks",
                                    403: 'forbidden'},
                         operation_description="Schedule the Subtasks that are not dependend on predecessors.")
    @action(methods=['post'], detail=True, url_name="schedule_independent_subtasks", name="Schedule independend Subtasks")
    def schedule_independent_subtasks(self, request, pk=None):
        task_blueprint = get_object_or_404(models.TaskBlueprint, pk=pk)
        schedule_independent_subtasks_in_task_blueprint(task_blueprint)
        task_blueprint.refresh_from_db()

        # return a response with the new serialized task_blueprint (with references to the created (and scheduled) subtasks)
        return Response(serializers.TaskBlueprintSerializer(task_blueprint, context={'request':request}).data,
                        status=status.HTTP_201_CREATED)

    @swagger_auto_schema(responses={200: 'The predecessor task draft of this task draft',
                                    403: 'forbidden'},
                         operation_description="Get the predecessor task draft of this task draft.")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, url_name="predecessors")
    def predecessors(self, request, pk=None):
        task_blueprint = get_object_or_404(models.TaskBlueprint, pk=pk)
        predecessors = self.filter_queryset(task_blueprint.predecessors)
        serializer = self.get_serializer(predecessors, many=True)
        return Response(serializer.data)

    @swagger_auto_schema(responses={200: 'The successor subtasks of this subtask',
                                    403: 'forbidden'},
                         operation_description="Get the successor subtasks of this subtask.")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, url_name="successors")
    def successors(self, request, pk=None):
        task_blueprint = get_object_or_404(models.TaskBlueprint, pk=pk)
        successors = self.filter_queryset(task_blueprint.successors)
        serializer = self.get_serializer(successors, many=True)
        return Response(serializer.data)

    @swagger_auto_schema(responses={200: 'The cancelled version of this task',
                                    403: 'forbidden',
                                    500: 'The task could not be cancelled'},
                         operation_description="Try to cancel this task.")
    @action(methods=['post'], detail=True, url_name="cancel")
    def cancel(self, request, pk=None):
        task_blueprint = get_object_or_404(models.TaskBlueprint, pk=pk)
        from lofar.sas.tmss.tmss.tmssapp.tasks import cancel_task_blueprint
        task_blueprint = cancel_task_blueprint(task_blueprint)
        serializer = self.get_serializer(task_blueprint)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The obsolete version of this task_blueprint',
                                    403: 'forbidden',
                                    500: 'The task_blueprint could not be marked as obsolete. Was it in a state from which it is allowed to mark it as obsolete?'},
                         operation_description="Try to mark this task_blueprint as obsolete (when the status is cancelled/error)")
    @action(methods=['post'], detail=True, url_name="mark_as_obsolete")
    def mark_as_obsolete(self, request, pk=None):
        task_blueprint = get_object_or_404(models.TaskBlueprint, pk=pk)
        from lofar.sas.tmss.tmss.tmssapp.tasks import mark_task_blueprint_as_obsolete
        task_blueprint = mark_task_blueprint_as_obsolete(task_blueprint)
        serializer = self.get_serializer(task_blueprint)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The restarted version of this task',
                                    403: 'forbidden',
                                    500: 'The task could not be restarted'},
                         operation_description="Try to restart all on hold subtasks in this task.")
    @action(methods=['post'], detail=True, url_name="restart_when_on_hold")
    def restart_when_on_hold(self, request, pk=None):
        task_blueprint = get_object_or_404(models.TaskBlueprint, pk=pk)
        from lofar.sas.tmss.tmss.tmssapp.tasks import restart_on_hold_task_blueprint
        task_blueprint = restart_on_hold_task_blueprint(task_blueprint)
        serializer = self.get_serializer(task_blueprint)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The system events of this task blueprint',
                                    403: 'forbidden'},
                         operation_description="Get the system events of this task blueprint.")
    @method_decorator(never_cache)
    @action(methods=['get'], detail=True, url_name="system_events")
    def system_events(self, request, pk=None):
        tb = get_object_or_404(models.TaskBlueprint, pk=pk)
        system_events = self.filter_queryset(tb.system_events)
        serializer = self.get_serializer(system_events, many=True)
        return RestResponse(serializer.data)


class TaskBlueprintExtendedViewSet(TaskBlueprintViewSet):
    serializer_class = serializers.TaskBlueprintExtendedSerializer


class TaskBlueprintNestedViewSet(LOFARNestedViewSet):
    queryset = models.TaskBlueprint.objects.all()
    serializer_class = serializers.TaskBlueprintSerializer

    def get_queryset(self):
        if 'task_draft_id' in self.kwargs:
            task_draft = get_object_or_404(models.TaskDraft, pk=self.kwargs['task_draft_id'])
            return task_draft.task_blueprints.all()
        else:
            return models.TaskBlueprint.objects.all()


class TaskBlueprintSlimPropertyFilter(property_filters.PropertyFilterSet):
    id = NumberInFilter(field_name='id', lookup_expr='in')
    id_min = filters.NumberFilter(field_name='id', lookup_expr='gte')
    id_max = filters.NumberFilter(field_name='id', lookup_expr='lte')
    scheduling_unit_blueprint = ModelChoiceInFilter(field_name='scheduling_unit_blueprint', lookup_expr='in', queryset=models.SchedulingUnitBlueprint.objects.all())
    duration_min = filters.DurationFilter(field_name='duration', lookup_expr='gte')
    duration_max = filters.DurationFilter(field_name='duration', lookup_expr='lte')
    on_sky_duration_min = filters.DurationFilter(field_name='on_sky_duration', lookup_expr='gte')
    on_sky_duration_max = filters.DurationFilter(field_name='on_sky_duration', lookup_expr='lte')
    created_at = filters.IsoDateTimeFromToRangeFilter(field_name='created_at')
    updated_at = filters.IsoDateTimeFromToRangeFilter(field_name='updated_at')
    on_sky_start_time = filters.IsoDateTimeFromToRangeFilter(field_name='on_sky_start_time')
    on_sky_stop_time = filters.IsoDateTimeFromToRangeFilter(field_name='on_sky_stop_time')
    process_start_time = filters.IsoDateTimeFromToRangeFilter(field_name='process_start_time')
    process_stop_time = filters.IsoDateTimeFromToRangeFilter(field_name='process_stop_time')
    status = filters.MultipleChoiceFilter(field_name='status', choices=[(c.value, c.value) for c in models.TaskStatus.Choices])
    task_type = filters.MultipleChoiceFilter(label='task type(s)', field_name='specifications_template__type', choices=[(c.value, c.value) for c in models.TaskType.Choices])

    class Meta:
        model = models.TaskBlueprint
        fields = ['id', 'scheduling_unit_blueprint', 'duration', 'on_sky_duration', 'created_at', 'updated_at',
                  'on_sky_start_time', 'on_sky_stop_time', 'process_start_time', 'process_stop_time', 'status', 'task_type']
        filter_overrides = FILTER_OVERRIDES


class TaskBlueprintSlimViewSet(LOFARViewSet):
    queryset = models.TaskBlueprint.objects.all()
    serializer_class = serializers.TaskBlueprintSlimSerializer
    filter_class = TaskBlueprintSlimPropertyFilter

    # use select_related for forward related references
    queryset = queryset.select_related('specifications_template__type', 'status', 'scheduling_unit_blueprint')


class TaskRelationDraftViewSet(LOFARViewSet):
    queryset = models.TaskRelationDraft.objects.all()
    serializer_class = serializers.TaskRelationDraftSerializer


class TaskRelationDraftExtendedViewSet(TaskRelationDraftViewSet):
    serializer_class = serializers.TaskRelationDraftExtendedSerializer


class TaskRelationDraftNestedViewSet(LOFARNestedViewSet):
    queryset = models.TaskRelationDraft.objects.all()
    serializer_class = serializers.TaskRelationDraftSerializer

    def get_queryset(self):
        if 'task_draft_id' in self.kwargs:
            task_draft = get_object_or_404(models.TaskDraft, pk=self.kwargs['task_draft_id'])
            return task_draft.produced_by.all() | task_draft.consumed_by.all()
        else:
            return models.TaskRelationDraft.objects.all()


class TaskRelationBlueprintViewSet(LOFARViewSet):
    queryset = models.TaskRelationBlueprint.objects.all()
    serializer_class = serializers.TaskRelationBlueprintSerializer


class TaskRelationBlueprintExtendedViewSet(TaskRelationBlueprintViewSet):
    serializer_class = serializers.TaskRelationBlueprintExtendedSerializer


class TaskSchedulingRelationBlueprintViewSet(LOFARViewSet):
    queryset = models.TaskSchedulingRelationBlueprint.objects.all()
    serializer_class = serializers.TaskSchedulingRelationBlueprintSerializer


class TaskSchedulingRelationDraftViewSet(LOFARViewSet):
    queryset = models.TaskSchedulingRelationDraft.objects.all()
    serializer_class = serializers.TaskSchedulingRelationDraftSerializer


class TaskRelationBlueprintNestedViewSet(LOFARNestedViewSet):
    queryset = models.TaskRelationBlueprint.objects.all()
    serializer_class = serializers.TaskRelationBlueprintSerializer


    def get_queryset(self):
        if 'task_blueprint_id' in self.kwargs:
            task_blueprint = get_object_or_404(models.TaskBlueprint, pk=self.kwargs['task_blueprint_id'])
            return task_blueprint.produced_by.all() | task_blueprint.consumed_by.all()
        elif 'task_relation_draft_id' in self.kwargs:
            task_relation_draft = get_object_or_404(models.TaskRelationDraft, pk=self.kwargs['task_relation_draft_id'])
            return task_relation_draft.blueprints.all()
        else:
            return models.TaskRelationBlueprint.objects.all()


class TaskTypeViewSet(LOFARViewSet):
    queryset = models.TaskType.objects.all()
    serializer_class = serializers.TaskTypeSerializer


class PriorityQueueTypeViewSet(LOFARViewSet):
    queryset = models.PriorityQueueType.objects.all()
    serializer_class = serializers.PriorityQueueTypeSerializer


class ProjectStateViewSet(LOFARViewSet):
    queryset = models.ProjectState.objects.all()
    serializer_class = serializers.ProjectStateSerializer


class SystemEventTypeViewSet(LOFARViewSet):
    queryset = models.SystemEventType.objects.all()
    serializer_class = serializers.SystemEventTypeSerializer


class SystemEventSubtypeViewSet(LOFARViewSet):
    queryset = models.SystemEventSubtype.objects.all()
    serializer_class = serializers.SystemEventSubtypeSerializer


class SystemEventSeverityViewSet(LOFARViewSet):
    queryset = models.SystemEventSeverity.objects.all()
    serializer_class = serializers.SystemEventSeveritySerializer


class SystemEventStatusViewSet(LOFARViewSet):
    queryset = models.SystemEventStatus.objects.all()
    serializer_class = serializers.SystemEventStatusSerializer


class SystemEventTemplateViewSet(AbstractSchemaTemplateViewSet):
    queryset = models.SystemEventTemplate.objects.all()
    serializer_class = serializers.SystemEventTemplateSerializer


class SystemEventFilter(filters.FilterSet):
    name = filters.CharFilter(field_name='name', lookup_expr='icontains')
    created_by = filters.CharFilter(field_name='created_by__email', lookup_expr='icontains')
    id = NumberInFilter(field_name='id', lookup_expr='in')
    id_min = filters.NumberFilter(field_name='id', lookup_expr='gte')
    id_max = filters.NumberFilter(field_name='id', lookup_expr='lte')
    created_at = filters.IsoDateTimeFromToRangeFilter(field_name='created_at')
    updated_at = filters.IsoDateTimeFromToRangeFilter(field_name='updated_at')
    description = filters.CharFilter(field_name='description', lookup_expr='icontains')
    start = filters.IsoDateTimeFromToRangeFilter(field_name='start')
    start_isnull = filters.BooleanFilter(field_name='start', lookup_expr='isnull')
    start_rangeornull = IsoDateTimeNullOrFromToRangeFilter(field_name='start')
    stop = filters.IsoDateTimeFromToRangeFilter(field_name='stop')
    stop_isnull = filters.BooleanFilter(field_name='stop', lookup_expr='isnull')
    stop_rangeornull = IsoDateTimeNullOrFromToRangeFilter(field_name='stop')
    issue_type = filters.ModelMultipleChoiceFilter(field_name='issue_type', queryset=models.SystemEventType.objects.all())
    issue_subtype = filters.ModelMultipleChoiceFilter(field_name='issue_subtype', queryset=models.SystemEventSubtype.objects.all())
    severity = filters.ModelMultipleChoiceFilter(field_name='severity', queryset=models.SystemEventSeverity.objects.all())
    notes = filters.CharFilter(field_name='notes', lookup_expr='icontains')
    status = filters.ModelMultipleChoiceFilter(field_name='status', queryset=models.SystemEventStatus.objects.all())
    affected_hardware_doc_stations_any = filters.MultipleChoiceFilter(label='stations [any]', field_name='affected_hardware_doc__stations',
                                                choices=all_stations_tuple, lookup_expr='icontains')
    affected_hardware_doc_stations_all = filters.MultipleChoiceFilter(label='stations [all]', field_name='affected_hardware_doc__stations',
                                                choices=all_stations_tuple, lookup_expr='icontains', conjoined=True)
    affected_hardware_template_name = filters.CharFilter(field_name='affected_hardware_template__name', lookup_expr='icontains')
    affected_tasks = filters.ModelMultipleChoiceFilter(field_name='affected_tasks', queryset=models.TaskBlueprint.objects.all())
    jira_url = filters.CharFilter(field_name='jira_url', lookup_expr='icontains')

    class Meta:
        model = models.SystemEvent
        fields = '__all__'
        filter_overrides = FILTER_OVERRIDES


class SystemEventViewSet(LOFARViewSet):
    queryset = models.SystemEvent.objects.all()
    serializer_class = serializers.SystemEventSerializer
    filter_class = SystemEventFilter

    def perform_create(self, serializer):
        serializer.save(created_by=self.request.user)
