"""tmss URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/2.0/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""
from django.shortcuts import render

from django.contrib import admin
from django.contrib.auth.views import LogoutView
from django.urls import path, re_path
from django.conf.urls import url, include
from django.views.generic.base import TemplateView, RedirectView

from collections import OrderedDict
from rest_framework import routers, permissions
from rest_framework.authtoken.views import obtain_auth_token
from .tmssapp import viewsets, models, serializers, views
from rest_framework.documentation import include_docs_urls
from drf_yasg.views import get_schema_view
from drf_yasg import openapi

from datetime import datetime
import os
from material.frontend import urls as frontend_urls
from lofar.sas.tmss.tmss.settings import isViewflowEnabled

#
# Django style patterns
#

# schema for Swagger API
swagger_schema_view = get_schema_view(
   openapi.Info(
      title="LOFAR Internal API",
      default_version='v1',
      description="ReST API for internal LOFAR services",
      terms_of_service="",
      contact=openapi.Contact(email="softwaresupport@astron.nl"),
      license=openapi.License(name="GPL License v3"),
   ),
   validators=['flex', 'ssv'],
   public=False,
  # public=True,
  # permission_classes=(permissions.AllowAny,),
)

# use re_path(r'<...>/?') to make trailing slash optional (double entries confuse Swagger)
urlpatterns = [
    path('admin/', admin.site.urls),
    path('logout/', LogoutView.as_view(), name='logout'),
    path('authentication_state/', views.authentication_state, name='authentication_state'),
    path('token-auth/', obtain_auth_token, name='obtain-token-auth'),
    path('token-deauth/', views.revoke_token_deauth, name='revoke-token-deauth'),
    path('docs/', include_docs_urls(title='TMSS API')),
    re_path(r'^swagger(?P<format>\.json|\.yaml)$', swagger_schema_view.without_ui(cache_timeout=0), name='schema-json'),
    path('swagger/', swagger_schema_view.with_ui('swagger', cache_timeout=0), name='schema-swagger-ui'),
    path('redoc/', swagger_schema_view.with_ui('redoc', cache_timeout=0), name='schema-redoc'),
    #re_path('schemas/<str:template>/<str:name>/<str:version>', views.get_template_json_schema, name='get_template_json_schema'),  # !! use of regex here breaks reverse url lookup
    path('schemas/<str:template>/<str:name>', views.get_template_json_schema, name='get_template_json_schema'),   # !! two urls for same view break Swagger, one url break json ref resolution !!
    path('schemas/<str:template>/<str:name>/', views.get_template_json_schema, name='get_template_json_schema'),   # !! two urls for same view break Swagger, one url break json ref resolution !!
    path('schemas/<str:template>/<str:name>/<str:version>', views.get_template_json_schema, name='get_template_json_schema'),   # !! two urls for same view break Swagger, one url break json ref resolution !!
    path('schemas/<str:template>/<str:name>/<str:version>/', views.get_template_json_schema, name='get_template_json_schema'),  # !! two urls for same view break Swagger, one url break json ref resolution !!
    path('schemas/<str:template>/<str:name>/<str:version>/default', views.get_template_json_schema_default_doc, name='get_template_json_schema_default_doc'),  # !! two urls for same view break Swagger, one url break json ref resolution !!
    path('schemas/<str:template>/<str:name>/<str:version>/ref_resolved', views.get_template_json_ref_resolved_schema, name='get_template_json_ref_resolved_schema'),  # !! two urls for same view break Swagger, one url break json ref resolution !!
    path('schemas/export_as_zip', views.get_templates_as_zipfile, name='get_templates_as_zipfile'),
    path('xsd/LTA-SIP.xsd', views.get_lta_sip_xsd, name='get_lta_sip_xsd'),
    #re_path('station_groups/<str:template_name>/<str:template_version>/<str:station_group>/?', views.get_stations_in_group, name='get_stations_in_group'), # !! use of regex here somehow breaks functionality (because parameters?) -> index page
    path('station_groups/<str:template_name>/<str:template_version>/<str:station_group>', views.get_stations_in_group, name='get_stations_in_group'),
    path('station_groups/<str:template_name>/<str:template_version>/<str:station_group>/', views.get_stations_in_group, name='get_stations_in_group'),
    re_path('util/sun_rise_and_set/?', views.get_sun_rise_and_set, name='get_sun_rise_and_set'),
    re_path('util/utc/?', views.utc, name="system-utc"),
    re_path('util/lst/?', views.lst, name="conversion-lst"),
    re_path('util/stations', views.stations, name="stations"),
    re_path('util/angular_separation/?', views.get_angular_separation, name='get_angular_separation'),
    re_path('util/target_rise_and_set/?', views.get_target_rise_and_set, name='get_target_rise_and_set'),
    re_path('util/target_transit/?', views.get_target_transit, name='get_target_transit'),
    re_path('util/calibrator/?', views.get_calibrator, name='get_calibrator'),
    re_path('submit_trigger', views.submit_trigger, name='submit_trigger'),
    re_path('friends_of_projects/', views.friends_of_projects, name='get_friends_of_projects'),
    re_path('my_roles_overview/', views.my_roles, name='get_my_roles'),
    re_path('plots/daily_schedule_plot/?(?P<datestamp>(\d{4}-[01]\d-[0-3]\d))?/?', views.daily_schedule_plot, name='daily_schedule_plot'),
    re_path('plots/daily_scheduling_constraints_overview/?(?P<datestamp>(\d{4}-[01]\d-[0-3]\d))?/?', views.daily_scheduling_constraints_overview, name='daily_scheduling_constraints_overview'),
    re_path('plots/lst_pressure_plot/?(?P<from_date>(\d{4}-[01]\d-[0-3]\d))?/?(?P<until_date>(\d{4}-[01]\d-[0-3]\d))?/?', views.lst_pressure_plot, name='lst_pressure_plot'),
    re_path('plots/lst_target_pressure_plot/?(?P<from_date>(\d{4}-[01]\d-[0-3]\d))?/?(?P<until_date>(\d{4}-[01]\d-[0-3]\d))?/?', views.lst_target_pressure_plot, name='lst_target_pressure_plot'),
    re_path('failure_report/', views.failure_report, name='failure_report'),
]

if os.environ.get('SHOW_DJANGO_DEBUG_TOOLBAR', False):
    import debug_toolbar
    urlpatterns.append(path('__debug__/', include(debug_toolbar.urls)))


#
# REST Router
#

class TMSSAPIRootView(routers.APIRootView):
    def get(self, request, *args, **kwargs):
        '''return the normal list of urls, but then sorted by url so they are easier to find in a list for us humble humans'''
        response = super().get(request, *args, **kwargs)
        sorted_by_key_urls = OrderedDict()

        for key in sorted(response.data.keys()):
            sorted_by_key_urls[key] = response.data[key]

        response.data = sorted_by_key_urls

        return response

class OptionalSlashRouter(routers.DefaultRouter):

    def __init__(self):
        super(routers.DefaultRouter, self).__init__()
        self.trailing_slash = '/?'


router = OptionalSlashRouter()
router.APIRootView = TMSSAPIRootView

# COMMON

router.register(r'tags', viewsets.TagsViewSet)

# SPECIFICATION

# choices
router.register(r'role', viewsets.RoleViewSet)
router.register(r'iotype', viewsets.IOTypeViewSet)
router.register(r'datatype', viewsets.DatatypeViewSet)
router.register(r'dataformat', viewsets.DataformatViewSet)
router.register(r'system_setting_flag', viewsets.SystemSettingFlagViewSet)
router.register(r'period_category', viewsets.PeriodCategoryViewSet)
router.register(r'project_category', viewsets.ProjectCategoryViewSet)
router.register(r'quantity', viewsets.QuantityViewSet)
router.register(r'task_type', viewsets.TaskTypeViewSet)
router.register(r'priority_queue_type', viewsets.PriorityQueueTypeViewSet)
router.register(r'project_state', viewsets.ProjectStateViewSet)
router.register(r'system_event_status', viewsets.SystemEventStatusViewSet)
router.register(r'system_event_type', viewsets.SystemEventTypeViewSet)
router.register(r'system_event_subtype', viewsets.SystemEventSubtypeViewSet)
router.register(r'system_event_severity', viewsets.SystemEventSeverityViewSet)
router.register(r'subsystem_status', viewsets.SubsystemStatusViewSet)

# templates
router.register(r'common_schema_template', viewsets.CommonSchemaTemplateViewSet)
router.register(r'scheduling_set_strategy_template', viewsets.SchedulingSetStrategyTemplateViewSet)
router.register(r'scheduling_set_template', viewsets.SchedulingSetTemplateViewSet)
router.register(r'scheduling_unit_observing_strategy_template', viewsets.SchedulingUnitObservingStrategyTemplateViewSet)
router.register(r'scheduling_unit_template', viewsets.SchedulingUnitTemplateViewSet)
router.register(r'scheduling_constraints_template', viewsets.SchedulingConstraintsTemplateViewSet)
router.register(r'task_template', viewsets.TaskTemplateViewSet)
router.register(r'task_relation_selection_template', viewsets.TaskRelationSelectionTemplateViewSet)
router.register(r'reservation_template', viewsets.ReservationTemplateViewSet)
router.register(r'task_connector_type', viewsets.TaskConnectorTypeViewSet)
router.register(r'reservation_strategy_template', viewsets.ReservationStrategyTemplateViewSet)
router.register(r'system_event_template', viewsets.SystemEventTemplateViewSet)

# instances
router.register(r'cycle', viewsets.CycleViewSet)
router.register(r'cycle_quota', viewsets.CycleQuotaViewSet)
router.register(r'project', viewsets.ProjectViewSet)
router.register(r'resource_type', viewsets.ResourceTypeViewSet)
router.register(r'project_quota', viewsets.ProjectQuotaViewSet)
router.register(r'project_quota_archive_location', viewsets.ProjectQuotaArchiveLocationViewSet)
router.register(r'setting', viewsets.SettingViewSet)
router.register(r'reservation', viewsets.ReservationViewSet)
router.register(r'system_event', viewsets.SystemEventViewSet)
router.register(r'subsystem', viewsets.SubsystemViewSet)

router.register(r'scheduling_set', viewsets.SchedulingSetViewSet)
router.register(r'scheduling_unit_draft_extended', viewsets.SchedulingUnitDraftExtendedViewSet)
router.register(r'scheduling_unit_draft', viewsets.SchedulingUnitDraftViewSet)  # ! The last registered view on a model is used for references to objects
router.register(r'scheduling_unit_blueprint_extended', viewsets.SchedulingUnitBlueprintExtendedViewSet)
router.register(r'scheduling_unit_blueprint_slim', viewsets.SchedulingUnitBlueprintSlimViewSet)
router.register(r'scheduling_unit_blueprint_slim_with_tasks', viewsets.SchedulingUnitBlueprintSlimWithTasksViewSet)
router.register(r'scheduling_unit_blueprint', viewsets.SchedulingUnitBlueprintViewSet)  # ! The last registered view on a model is used for references to objects
router.register(r'task_draft_extended', viewsets.TaskDraftExtendedViewSet)
router.register(r'task_draft', viewsets.TaskDraftViewSet)
router.register(r'task_blueprint_extended', viewsets.TaskBlueprintExtendedViewSet)
router.register(r'task_blueprint_slim', viewsets.TaskBlueprintSlimViewSet)
router.register(r'task_blueprint', viewsets.TaskBlueprintViewSet)
router.register(r'task_relation_draft', viewsets.TaskRelationDraftViewSet)
router.register(r'task_relation_blueprint', viewsets.TaskRelationBlueprintViewSet)
router.register(r'task_scheduling_relation_draft', viewsets.TaskSchedulingRelationDraftViewSet)
router.register(r'task_scheduling_relation_blueprint', viewsets.TaskSchedulingRelationBlueprintViewSet)
router.register(r'scheduling_constraints_weight_factor', viewsets.SchedulingConstraintsWeightFactorViewSet)

# nested
router.register(r'cycle/(?P<cycle_id>[\w\- ]+)/project', viewsets.ProjectNestedViewSet)
router.register(r'scheduling_set/(?P<scheduling_set_id>\d+)/scheduling_unit_draft', viewsets.SchedulingUnitDraftNestedViewSet)
router.register(r'scheduling_set/(?P<scheduling_set_id>\d+)/scheduling_unit_observing_strategy_template', viewsets.SchedulingUnitObservingStrategyTemplateNestedViewSet)
router.register(r'scheduling_unit_draft/(?P<scheduling_unit_draft_id>\d+)/scheduling_unit_blueprint', viewsets.SchedulingUnitBlueprintNestedViewSet)
router.register(r'scheduling_unit_draft/(?P<scheduling_unit_draft_id>\d+)/task_draft', viewsets.TaskDraftNestedViewSet)
router.register(r'task_draft/(?P<task_draft_id>\d+)/task_blueprint', viewsets.TaskBlueprintNestedViewSet)
router.register(r'task_draft/(?P<task_draft_id>\d+)/task_relation_draft', viewsets.TaskRelationDraftNestedViewSet)
router.register(r'task_relation_draft/(?P<task_relation_draft_id>\d+)/task_relation_blueprint', viewsets.TaskRelationBlueprintNestedViewSet)
router.register(r'task_blueprint/(?P<task_blueprint_id>\d+)/task_relation_blueprint', viewsets.TaskRelationBlueprintNestedViewSet)
router.register(r'task_blueprint/(?P<task_blueprint_id>\d+)/subtask', viewsets.SubtaskNestedViewSet)

# SCHEDULING

# choices
router.register(r'subtask_state', viewsets.SubtaskStateViewSet)
router.register(r'subtask_type', viewsets.SubtaskTypeViewSet)
router.register(r'station_type', viewsets.StationTypeViewSet)
router.register(r'hash_algorithm', viewsets.HashAlgorithmViewSet)
router.register(r'scheduling_relation_placement', viewsets.SchedulingRelationPlacementViewSet)
router.register(r'task_status', viewsets.TaskStatusViewSet)
router.register(r'scheduling_unit_status', viewsets.SchedulingUnitStatusViewSet)
router.register(r'template_state', viewsets.TemplateStateViewSet)
router.register(r'template_purpose', viewsets.TemplatePurposeViewSet)

# templates
router.register(r'subtask_template', viewsets.SubtaskTemplateViewSet)
router.register(r'subtask_feedback_template', viewsets.SubtaskFeedbackTemplateViewSet)
router.register(r'dataproduct_specifications_template', viewsets.DataproductSpecificationsTemplateViewSet)
router.register(r'dataproduct_feedback_template', viewsets.DataproductFeedbackTemplateViewSet)
router.register(r'sap_template', viewsets.SAPTemplateViewSet)

# instances
router.register(r'subtask_extended', viewsets.SubtaskExtendedViewSet)
router.register(r'subtask', viewsets.SubtaskViewSet)
router.register(r'dataproduct', viewsets.DataproductViewSet)
router.register(r'subtask_input_extended', viewsets.SubtaskInputExtendedViewSet)
router.register(r'subtask_input', viewsets.SubtaskInputViewSet)
router.register(r'subtask_output_extended', viewsets.SubtaskOutputExtendedViewSet)
router.register(r'subtask_output', viewsets.SubtaskOutputViewSet)
router.register(r'antenna_set', viewsets.AntennaSetViewSet)
router.register(r'dataproduct_transform', viewsets.DataproductTransformViewSet)
router.register(r'filesystem', viewsets.FilesystemViewSet)
router.register(r'cluster', viewsets.ClusterViewSet)
router.register(r'dataproduct_archive_info', viewsets.DataproductArchiveInfoViewSet)
router.register(r'dataproduct_hash', viewsets.DataproductHashViewSet)
router.register(r'subtask_allowed_state_transitions', viewsets.SubtaskAllowedStateTransitionsViewSet)
router.register(r'subtask_state_log', viewsets.SubtaskStateLogViewSet)
router.register(r'user', viewsets.UserViewSet)
router.register(r'sap', viewsets.SAPViewSet)
router.register(r'sip_identifier', viewsets.SIPidentifierViewSet)


# PERMISSIONS

router.register(r'project_role', viewsets.ProjectRoleViewSet)
router.register(r'project_permission', viewsets.ProjectPermissionViewSet)


# CONVERSIONS

router.register(r'station_timeline', viewsets.StationTimelineViewSet)


urlpatterns.extend(router.urls)

frontend_urlpatterns = [
    path("", views.index, name="index")
]


urlpatterns = [url(r'^api$', RedirectView.as_view(url='/api/')),
                url(r'^api/', include(urlpatterns)), url(r'^oidc$',
                RedirectView.as_view(url='/oidc/')),
                url(r'^oidc/', include('mozilla_django_oidc.urls')),
                url(r'^schedule', views.public_schedule),
                url(r'^past_schedule', views.public_past_schedule),
                url(r'', include(frontend_urls)),
                url(r'^.*', include(frontend_urlpatterns)),
]


# QA Workflow steps
if isViewflowEnabled():
    from .workflowapp import viewsets as workflow_viewsets
    viewflow_urlpatterns = []

    viewflow_router = OptionalSlashRouter()
    viewflow_router.APIRootView = TMSSAPIRootView

    from .workflowapp import viewsets as workflow_viewsets
    viewflow_router.register('scheduling_unit_flow/qa_reporting_to', workflow_viewsets.QAReportingTOViewSet, basename='qa_reporting_to')
    viewflow_router.register('scheduling_unit_flow/qa_reporting_sos', workflow_viewsets.QAReportingSOSViewSet, basename='qa_reporting_sos')
    viewflow_router.register('scheduling_unit_flow/qa_pi_verification', workflow_viewsets.PIVerificationViewSet, basename='qa_pi_verification')
    viewflow_router.register('scheduling_unit_flow/qa_decide_acceptance', workflow_viewsets.DecideAcceptanceViewSet, basename='qa_decide_acceptance')
    viewflow_router.register('scheduling_unit_flow/qa_scheduling_unit_process', workflow_viewsets.SchedulingUnitProcessViewSet, basename='qa_scheduling_unit_process')
    viewflow_router.register('scheduling_unit_flow/qa_scheduling_unit_task', workflow_viewsets.SchedulingUnitTaskViewSet, basename='qa_scheduling_unit_task')
    viewflow_router.register(r'scheduling_unit_flow/qa_scheduling_unit_task/(?P<pk>\d+)/assign', workflow_viewsets.SchedulingUnitTaskAssignViewSet)
    viewflow_router.register(r'scheduling_unit_flow/qa_scheduling_unit_task/(?P<pk>\d+)/unassign', workflow_viewsets.SchedulingUnitTaskUnassignViewSet)
    viewflow_router.register(r'scheduling_unit_flow/qa_scheduling_unit_process/(?P<pk>\d+)/current_task', workflow_viewsets.SchedulingUnitGetActiveTasksViewSet)
    viewflow_router.register(r'scheduling_unit_flow/qa_scheduling_unit_process/(?P<pk>\d+)/perform', workflow_viewsets.SchedulingUnitTaskExecuteViewSet)

    viewflow_urlpatterns.extend(viewflow_router.urls)

    urlpatterns.insert(0,url(r'^workflow$', RedirectView.as_view(url='/workflow/', permanent=False)))
    #Doesn't work if it is at the end of urlpatterns
    urlpatterns.insert(0,url(r'^workflow_api/',  include(viewflow_urlpatterns)))


# Custom error handlers
handler500 = 'lofar.sas.tmss.tmss.tmssapp.views.response500'