from django.utils.decorators import method_decorator
from viewflow import flow
from viewflow.base import this, Flow
from viewflow.activation import FuncActivation
from viewflow.flow.nodes import Signal

from .. import models
from .. import viewsets

from lofar.common.datetimeutils import round_to_second_precision

from lofar.sas.tmss.tmss.workflowapp.signals import *
from lofar.sas.tmss.tmss.tmssapp.models.specification import SchedulingUnitStatus

from viewflow import frontend, ThisObject
from viewflow.activation import STATUS
from viewflow.models import Process

from lofar.sas.tmss.client.tmssbuslistener import *

from datetime import datetime

import logging
logger = logging.getLogger(__name__)


class ConditionalActivation(FuncActivation):
    @classmethod
    def activate(cls, flow_task, prev_activation, token):
        """
        The condition_check is performed when the node becomes active, and if the condition holds it immediately marks
        the task as done so that the workflow progresses.
        :param flow_task:
        :param prev_activation:
        :param token:
        :return:
        """
        # The super activates the node so it starts waiting for a signal, which we want to omit in case the condition is already met.
        activation = super(ConditionalActivation, cls).activate(flow_task, prev_activation, token)

        if flow_task.condition_check(activation, None):
            # condition holds on activation
            activation.prepare()
            activation.done()

        return activation

class Conditional(Signal):
    """
    This is an augmented "If" and "Signal" node, which arguably implements an BPMN Conditional (https://camunda.com/bpmn/reference/#events-conditional).
    Initially this node performs a condition check and only if not met it waits for a signal.
    The purpose of this node is to avoid race conditions that can occur when a signal gets send before the Signal
    node is reached, but after a prior If node evaluated the corresponding TMSS property based on which it still expects
    a signal.
    See https://support.astron.nl/jira/browse/TMSS-656 for design notes.
    Also see https://support.astron.nl/jira/browse/TMSS-1562.
    """
    task_type = "HUMAN" # makes it show up in the unassigned task lists
    activation_class = ConditionalActivation

    def __init__(self, condition_check, signal, receiver=None, sender=None, task_loader=None, **kwargs):
        """
        Instantiate a Signal task.

        :param signal: A django signal to connect
        :param receiver: Callable[activation, **kwargs]
        :param sender: Optional signal sender
        :param task_loader: Callable[**kwargs] -> Task
        :param allow_skip: If True task_loader can return None if
                           signal could be skipped.

        You can skip a `task_loader` if the signal going to be
        sent with Task instance.
        """
        self.condition_check = condition_check
        super(Conditional, self).__init__(signal, receiver or self.signal_handler, sender, task_loader, **kwargs)

    @method_decorator(flow.flow_signal)
    def signal_handler(self, activation, sender, instance, **signal_kwargs):
        """Default signal handler used when a receiver is not provided."""

        if activation.get_status() == STATUS.DONE:
            # race condition -- condition was true on activation but we also receive the signal now
            return

        activation.prepare()
        if activation.flow_task.condition_check(activation, instance):
            activation.done()

    def ready(self):
        """Resolve internal `this`-references. and subscribe to the signal."""

        if isinstance(self.condition_check, ThisObject):
            self.condition_check = getattr(self.flow_class.instance, self.condition_check.name)

        super(Conditional, self).ready()


@frontend.register
class SchedulingUnitFlow(Flow):
    """
    This Flow describes the workflow for scheduling units. Each attribute of a Flow represents a flow task.
    Some tasks are done event-driven by the system. The signals are dispatched by the workflow service in response
    to SU state changes on the message bus.
    """
   
    process_class = models.SchedulingUnitProcess
    task_class = models.SchedulingUnitTask

    # 1. Start workflow on SUB becoming schedulable
    start = (
        flow.StartSignal(
          scheduling_unit_blueprint_schedulable_signal,
          this.on_scheduling_unit_blueprint_schedulable,
        ).Next(this.wait_scheduled)
    )

    # 2. Wait for SUB to get scheduled
    wait_scheduled = (
        Conditional(
          this.is_SUB_scheduled_state_or_beyond,
          scheduling_unit_blueprint_scheduled_signal,
          this.on_scheduling_unit_blueprint_scheduled,
          task_loader=this.get_scheduling_unit_task
        )
        .Next(this.wait_observing)
    )

    # 3. Wait for SUB to start observing
    wait_observing = (
        Conditional(
          this.is_SUB_observing_state_or_beyond,
          scheduling_unit_blueprint_observing_signal,
          this.on_scheduling_unit_blueprint_observing,
          task_loader=this.get_scheduling_unit_task
        )
        .Next(this.qa_reporting_to)
    )

    # 4. Human action in QA Reporting (TO) view
    qa_reporting_to = (
        flow.View(
            viewsets.QAReportingTOView,
            task_description='QA Reporting (TO)'
        ).Permission(
            auto_create=True
        ).Next(this.check_operator_accept)
    )
    
    # 5. if operator accepted -> 6., else -> 13F.
    check_operator_accept = (
        flow.If(lambda activation: activation.process.qa_reporting_to.operator_accept)
        .Then(this.qa_reporting_sos)
        .Else(this.mark_sub_not_accepted)
    )

    # 6. Human action in QA Reporting (SOS) view
    qa_reporting_sos = (
        flow.View(
            viewsets.QAReportingSOSView,
            task_description='QA Reporting (SOS)'
        ).Permission(
            auto_create=True
        ).Next(this.check_sos_accept_show_pi)
    )

    # 7. if SOS accepted -> 8., else -> 13F.
    check_sos_accept_show_pi = (
        flow.If(lambda activation: activation.process.qa_reporting_sos.sos_accept_show_pi)
        .Then(this.pi_verification)
        .Else(this.mark_sub_not_accepted)
    )

    # 8. Human action in PI Verification view
    pi_verification = (
        flow.View(
            viewsets.PIVerificationView,
            task_description='PI Verification'
        ).Permission(
            auto_create=True
        ).Next(this.decide_acceptance)
    )

    # 9. Human action in Decide Acceptance view
    decide_acceptance = (
        flow.View(
            viewsets.DecideAcceptanceView,
            task_description='Decide Acceptance'
        ).Permission(
            auto_create=True
        ).Next(this.check_sos_accept_after_pi)
    )

    # 10. If SOS accepted after PI -> 11., else -> 11-F.
    check_sos_accept_after_pi = (
        flow.If(lambda activation: activation.process.decide_acceptance.sos_accept_after_pi)
        .Then(this.mark_sub_accepted)
        .Else(this.mark_sub_not_accepted)
    )

    # 11. Mark SUB as accepted
    mark_sub_accepted = (
        flow.Handler(
            this.do_mark_sub_accepted
        ).Next(this.allow_ingest)
    )

    # 11-F. Mark SUB as not accepted
    mark_sub_not_accepted = (
        flow.Handler(
            this.do_mark_sub_not_accepted
        ).Next(this.check_data_pinned)   # for now we want to be safe, later we could simply remove the data by unpinning it and progress to wait_finished.
    )

    # 12. set ingest permission of SUB
    allow_ingest = (
        flow.Handler(
            this.signal_SUB_allow_ingest,
        )
        .Next(this.wait_ingest_done)
    )

    # 13. wait for ingest to be done (if needed)
    wait_ingest_done = (
        Conditional(
          this.is_SUB_ingested_state_or_beyond,
          scheduling_unit_blueprint_ingested_signal,
          task_loader=this.get_scheduling_unit_task
        )
        .Next(this.check_data_pinned)
    )

    # 14. if output pinned -> 15., else -> 17.
    check_data_pinned = (
        flow.If(lambda activation: activation.process.su.output_pinned)
        .Then(this.decide_unpin_data)
        .Else(this.wait_finished)
    )

    # 15. Human action in Unpin Data view.
    decide_unpin_data = (
        flow.View(
            viewsets.UnpinDataView,
            task_description='Unpin Data'
        ).Permission(
            auto_create=True
        ).Next(this.mark_sub_output_pinned)
    )

    # 16. Mark output_pinned on SUB
    mark_sub_output_pinned = (
        flow.Handler(
            this.do_mark_sub_output_pinned
        ).Next(this.wait_finished)
    )

    # TMSS-2130 goes here

    # 17. Delete the data
    # Note: We could do a 'check_unpin_decision' step prior to this, if we want to allow users to keep their data.
    # Currently, you are stuck in the workflow until the data has been unpinned in step 15.
    wait_finished = (
        Conditional(
          this.is_SUB_finished_state_or_beyond,
          scheduling_unit_blueprint_finished_signal,
          task_loader=this.get_scheduling_unit_task
        )
        .Next(this.end)
    )

    # 18. End
    end = flow.End()
    
    @method_decorator(flow.flow_start_signal)
    def on_scheduling_unit_blueprint_schedulable(self, activation, sender, scheduling_unit_blueprint, **signal_kwargs):
        try:
            process = models.SchedulingUnitProcess.objects.get(su=scheduling_unit_blueprint)
        except Process.DoesNotExist:
            # create the SchedulingUnitProcess for this SchedulingUnitBlueprint instance
            activation.prepare()
            activation.process.su = scheduling_unit_blueprint
            activation.done()
            logger.info("workflow process with id=%s started for scheduling unit id=%s name='%s'", activation.process.id, scheduling_unit_blueprint.id, scheduling_unit_blueprint.name)
        except Process.MultipleObjectsReturned:
            logger.info("Could not start QA workflow process for SUB id=%s because there are already multiple SchedulingUnitProcess objects for it.", scheduling_unit_blueprint.id)

        return activation

    @method_decorator(flow.flow_signal)
    def on_scheduling_unit_blueprint_scheduled(self, activation, sender, scheduling_unit_blueprint, **signal_kwargs):
        activation.prepare()
        activation.done()
        assert self.is_SUB_scheduled_state_or_beyond(activation, scheduling_unit_blueprint)
        logger.info('QA workflow on_scheduling_unit_blueprint_scheduled scheduling_unit id=%s flow_process=(\'%s\', %s) flow_task=(\'%s\', %s)',
                    scheduling_unit_blueprint.id, str(activation.process), activation.process.status, str(activation.flow_task), activation.status)

    @method_decorator(flow.flow_signal)
    def on_scheduling_unit_blueprint_observing(self, activation, sender, scheduling_unit_blueprint, **signal_kwargs):
        activation.prepare()
        activation.done()
        assert self.is_SUB_observing_state_or_beyond(activation, scheduling_unit_blueprint)
        logger.info('QA workflow on_scheduling_unit_blueprint_observing scheduling_unit id=%s flow_process=(\'%s\', %s) flow_task=(\'%s\', %s)',
                    scheduling_unit_blueprint.id, str(activation.process), activation.process.status, str(activation.flow_task), activation.status)

    @method_decorator(flow.flow_signal)
    def on_scheduling_unit_blueprint_observed(self, activation, sender, scheduling_unit_blueprint, **signal_kwargs):
        activation.prepare()
        activation.done()
        assert self.is_SUB_observed_state_or_beyond(activation, scheduling_unit_blueprint)
        logger.info('QA workflow on_scheduling_unit_blueprint_observed scheduling_unit id=%s flow_process=(\'%s\', %s) flow_task=(\'%s\', %s)',
                    scheduling_unit_blueprint.id, str(activation.process), activation.process.status, str(activation.flow_task), activation.status)

    def do_mark_sub(self, activation, accepted: bool):
        activation.process.su.results_accepted = accepted
        activation.process.su.save()
        activation.process.save()
        return activation

    def do_mark_sub_accepted(self, activation):
        return self.do_mark_sub(activation, True)

    def do_mark_sub_not_accepted(self, activation):
        return self.do_mark_sub(activation, False)

    def do_mark_sub_output_pinned(self, activation):
        activation.process.su.output_pinned = not activation.process.unpin_data.unpin_data
        activation.process.su.save()
        activation.process.save()
        return activation

    def signal_SUB_allow_ingest(self, activation):

        logger.info("granting ingest permission for scheduling unit blueprint id=%s", activation.process.su.id)
        if activation.process.su.ingest_permission_granted_since is None:
            activation.process.su.ingest_permission_granted_since = round_to_second_precision(datetime.utcnow())
        else:
            logger.warning("ingest permission on scheduling unit blueprint id=%s has already been granted on %s" % (activation.process.su.id, activation.process.su.ingest_permission_granted_since))
        activation.process.su.save()
        activation.process.save()

    def check_SchedulingUnitBlueprint_status_or_beyond(self, activation, scheduling_unit: models.SchedulingUnitBlueprint, expected_status):
        if scheduling_unit is None:
            scheduling_unit = activation.process.su
        scheduling_unit.refresh_from_db()

        logger.info("[check scheduling unit status] checking on %s, status=%s, expected_status=%s", scheduling_unit, scheduling_unit.status, expected_status)
        condition = scheduling_unit.status.value == expected_status

        if expected_status == SchedulingUnitStatus.Choices.INGESTED.value:
            condition = scheduling_unit.status.value in SchedulingUnitStatus.INGESTED_STATUS_OR_BEYOND_VALUES
        elif expected_status == SchedulingUnitStatus.Choices.OBSERVING.value:
            condition = scheduling_unit.status.value in SchedulingUnitStatus.OBSERVING_STATUS_OR_BEYOND_VALUES
        elif expected_status == SchedulingUnitStatus.Choices.OBSERVED.value:
            condition = scheduling_unit.status.value in SchedulingUnitStatus.OBSERVED_STATUS_OR_BEYOND_VALUES
        elif expected_status == SchedulingUnitStatus.Choices.SCHEDULED.value:
            condition = scheduling_unit.status.value in SchedulingUnitStatus.ACTIVE_OR_FINISHED_STATUS_VALUES
        elif expected_status == SchedulingUnitStatus.Choices.SCHEDULABLE.value:
            condition = scheduling_unit.status.value in SchedulingUnitStatus.SCHEDULABLE_ACTIVE_OR_FINISHED_STATUS_VALUES

        return condition

    def is_SUB_ingested_state_or_beyond(self, activation, scheduling_unit: models.SchedulingUnitBlueprint):
        return self.check_SchedulingUnitBlueprint_status_or_beyond(activation, scheduling_unit, SchedulingUnitStatus.Choices.INGESTED.value)

    def is_SUB_scheduled_state_or_beyond(self, activation, scheduling_unit: models.SchedulingUnitBlueprint):
        return self.check_SchedulingUnitBlueprint_status_or_beyond(activation, scheduling_unit, SchedulingUnitStatus.Choices.SCHEDULED.value)

    def is_SUB_observing_state_or_beyond(self, activation, scheduling_unit: models.SchedulingUnitBlueprint):
        return self.check_SchedulingUnitBlueprint_status_or_beyond(activation, scheduling_unit, SchedulingUnitStatus.Choices.OBSERVING.value)

    def is_SUB_observed_state_or_beyond(self, activation, scheduling_unit: models.SchedulingUnitBlueprint):
        return self.check_SchedulingUnitBlueprint_status_or_beyond(activation, scheduling_unit, SchedulingUnitStatus.Choices.OBSERVED.value)

    def is_SUB_finished_state_or_beyond(self, activation, scheduling_unit: models.SchedulingUnitBlueprint):
        return self.check_SchedulingUnitBlueprint_status_or_beyond(activation, scheduling_unit, SchedulingUnitStatus.Choices.FINISHED.value)

    def get_scheduling_unit_task(self, flow_task, sender, scheduling_unit_blueprint, **kwargs):
        try:
            return models.SchedulingUnitTask.objects.filter(process__schedulingunitprocess__su=scheduling_unit_blueprint).get(flow_task=flow_task)
        except models.SchedulingUnitTask.DoesNotExist as e:
            logger.warning(e)
            return None

