#!/usr/bin/env python3

# Copyright (C) 2018    ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

# $Id:  $

import os
os.environ['TZ'] = 'UTC'

from time import sleep
from datetime import datetime, timedelta
from astropy.coordinates import Angle
from astropy import units
import logging

if __name__ == "__main__":
    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

logger = logging.getLogger()

from lofar.sas.tmss.test.test_environment import TMSSTestEnvironment
tmss_test_env = None

# Check the invocation arguments
from optparse import OptionParser, OptionGroup
parser = OptionParser('%prog [options]',
                      description='run an acceptance test to see if we can import and/or schedule and/or simulate a number of scheduling units.')

group = OptionGroup(parser, 'target import options', description='Specify how many pulsar/lolls targets you want to import.')
parser.add_option_group(group)
group.add_option('-p', '--num_pulsars', dest='num_pulsars', type='int', default=-1, help='read at most this many pulsar targets (or all if -1), default: %default')
group.add_option('-l', '--num_lolls', dest='num_lolls', type='int', default=-1, help='read at most this many LoLLS targets (or all if -1), default: %default')
group.add_option('-t', '--lotss_angle_step', dest='lotss_angle_step', type='int', default=0, help='create a LoTSS survey across the sky with this angular step in degrees in RA and DEC (or no survey if <= 0), default: %default')
group.add_option('-d', '--dynamic_scheduling', dest='dynamic_scheduling', action='store_true', help='enable dynamic scheduling during import')

group = OptionGroup(parser, 'Simulation', description='Options for simulation of running and scheduling the units.')
parser.add_option_group(group)
group.add_option('-s', '--simulate', dest='simulate', action='store_true', help='simulate a run for each unit, allowing the scheduler to schedule the next unit when one finishes')
group.add_option('-e', '--event_delay', dest='event_delay', type='float', default=0, help='add a delay between each state transition (simulating the doing some work)')
group.add_option('-r', '--running_delay', dest='running_delay', type='float', default=1, help='add a delay (on top of event_delay) to simulate the observation/pipeline running duration')

group = OptionGroup(parser, 'TMSS Test Environment options', description='Run this acceptance test against a fresh TMSSTestEnvironment a running one (specfied by -C).')
parser.add_option_group(group)
group.add_option('-C', '--credentials', dest='dbcredentials', type='string',
                 default=os.environ.get('TMSS_DBCREDENTIALS', None),
                 help='django dbcredentials name, default: %default')
group.add_option('--wipe', dest='wipe', action='store_true', help='wipe all scheduling units in the fresh/existing TMSS instance. Only available when developing.')

(options, args) = parser.parse_args()

if options.dbcredentials is not None:
    # try to connect to existing TMSS
    from lofar.sas.tmss.tmss import setup_and_check_tmss_django_database_connection_and_exit_on_error
    setup_and_check_tmss_django_database_connection_and_exit_on_error(options.dbcredentials)


    # wipe scheduling units only in Development
    from lofar.common import isDevelopmentEnvironment
    if options.wipe and isDevelopmentEnvironment():
        from lofar.common.postgres import PostgresDatabaseConnection
        from lofar.common import dbcredentials
        dbcreds = dbcredentials.DBCredentials().get(options.dbcredentials)
        with PostgresDatabaseConnection(dbcreds) as db:
            db.executeQuery('TRUNCATE tmssapp_schedulingunitdraft CASCADE;')
            db.commit()

else:
    # create fresh TMSSTestEnvironment
    tmss_test_env = TMSSTestEnvironment(start_dynamic_scheduler=True, start_websocket=True,
                                        start_subtask_scheduler=False, start_workflow_service=True)
    tmss_test_env.start()
    tmss_test_env.populate_schemas_and_connectors()


from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.tmssapp.tasks import create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft
from lofar.sas.tmss.test.tmss_test_data_django_models import SchedulingSet_test_data, Project_test_data, Cycle_test_data
from lofar.sas.tmss.tmss.tmssapp.tasks import create_scheduling_unit_draft_from_observing_strategy_template
from lofar.sas.tmss.test.test_utils import assert_SUB_statuses, enable_scheduling, disable_scheduling, wait_for_units_to_get_scheduled_and_simulate_run, print_scheduling_results, TestStats
from dateutil import parser


def populate_pulsar_timing_test_campaign(project: models.Project, between=None, max_num_targets: int=-1):
    scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project,
                                                                                   name="pulsar timing run %s" % (min([parser.parse(x["from"], ignoretz=True) for x in between]).date() if between else datetime.today()),))
    with open(os.path.abspath(os.path.dirname(__file__)) + '/pulsar_targets.txt', 'r') as file:
        for line in file.readlines():
            if line.startswith('#'):
                continue

            if max_num_targets > -1 and scheduling_set.scheduling_unit_drafts.count() >= max_num_targets:
                break

            elems = line.split()
            target = elems[0]

            # if models.SchedulingUnitDraft.objects.filter(name=target).exists():
            #     logger.info("skipping creation of unit with name='%s' because it already exists", target)
            #     continue

            # read angle1 as h:m:s, and angle2 as d:m:s
            angle1 = Angle(elems[1].replace(':','h', 1).replace(':','m', 1)+'s')
            angle2 = Angle(elems[2].replace(':','d', 1).replace(':','m', 1)+'s')
            # create pointing for SAP and tile_beam
            pointing = {'angle1': angle1.radian, 'angle2': angle2.radian, 'direction_type': 'J2000', 'target': target}

            # derive a minimum elevation based on DEC (given by astronomer-commisionner)
            min_elevation = Angle('10 degrees') if angle2.deg < -15 else \
                            Angle('20 degrees') if angle2.deg < -10 else \
                            Angle('25 degrees') if angle2.deg < -5 else \
                            Angle('30 degrees')

            duration = int(elems[3]) * 60

            if "./templates/hba-cv-nopdmp-raw-freqres-template.xml" in elems:
                strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name='BF CV Pulsar Timing Scintillation')
            else:
                strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name='BF Pulsar Timing')

            logger.info('Creating SU for observation target=%s duration=%s strategy=%s' % (target, duration, strategy_template.name))
            scheduling_unit_draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template,
                                                                                                  scheduling_set,
                                                                                                  name=target,
                                                                                                  description="SU for test pulsar campaign target?%s" % target,
                                                                                                  priority_queue=models.PriorityQueueType.objects.get(value=models.PriorityQueueType.Choices.B.value),
                                                                                                  specifications_doc_overrides={
                                                                                                      'tasks': {
                                                                                                          'Observation': {
                                                                                                              'short_description': target,
                                                                                                              'specifications_doc': {
                                                                                                                  'duration': duration,
                                                                                                                  'station_configuration': {
                                                                                                                      'tile_beam': pointing,
                                                                                                                      'SAPs': [ {'digital_pointing': pointing} ]
                                                                                                                  }
                                                                                                              }
                                                                                                          }
                                                                                                      },
                                                                                                      'scheduling_constraints_doc': {
                                                                                                          'scheduler': 'dynamic',
                                                                                                          'time': {'between': between if between else []},
                                                                                                          'daily': {'require_day': False, 'require_night': False},
                                                                                                          'sky': { 'min_elevation': {'target': min_elevation.radian},
                                                                                                                  'transit_offset': {'to': 2*60*60, 'from': -2*60*60}}
                                                                                                      }
                                                                                                  })
            create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)


    if max_num_targets > 0:
        # add an additional "helper" fixed_time unit which is scheduled at the end of the between constraint,
        # such that the scheduler can schedule all pulsar B-units in the resulting gap between between.from and between.to
        misc_strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="Simple Observation")
        misc_scheduling_unit_draft = create_scheduling_unit_draft_from_observing_strategy_template(misc_strategy_template,
                                                                                                   scheduling_set,
                                                                                                   name="Helper",
                                                                                                   description="Helper SU for test pulsar campaign",
                                                                                                   priority_queue=models.PriorityQueueType.objects.get(value=models.PriorityQueueType.Choices.A.value),
                                                                                                   specifications_doc_overrides={
                                                                                                       'tasks': {
                                                                                                           'Observation': {
                                                                                                               'specifications_doc': {
                                                                                                                   'duration': 60,
                                                                                                                   'station_configuration': {
                                                                                                                       'station_groups': [{'stations': ['CS002']}]
                                                                                                                   }
                                                                                                               }
                                                                                                           }
                                                                                                       },
                                                                                                       'scheduling_constraints_doc': {
                                                                                                           'scheduler': 'fixed_time',
                                                                                                           'time': { 'at': between[-1]['to'] if between else (datetime.utcnow()+timedelta(days=2)).isoformat()},
                                                                                                           'sky': {'transit_offset': {'from': -12*60*60, 'to': 12*60*60},
                                                                                                                   'min_elevation': {'target': 0},
                                                                                                                   'min_distance': {} }
                                                                                                       }
                                                                                                   })
        create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(misc_scheduling_unit_draft)

    # assert_SUB_statuses(models.SchedulingUnitStatus.Choices.SCHEDULABLE.value)


def populate_LoLSS_test_campaign(project: models.Project, max_num_targets: int=-1):
    import csv
    scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project,
                                                                                   name="LoLLS run %s" % (datetime.today(),)))
    strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name='IM LBA Survey - 3 Beams')

    with open(os.path.abspath(os.path.dirname(__file__)) + '/LoLSS_targets.tsv', 'r') as file:
        rd = csv.reader(file, delimiter="\t", quotechar='"')
        for elems in rd:
            if elems[0].startswith('Name'):
                continue

            if max_num_targets > -1 and scheduling_set.scheduling_unit_drafts.count() >= max_num_targets:
                return

            su_name = elems[0]

            if models.SchedulingUnitDraft.objects.filter(name=su_name).exists():
                logger.info("skipping creation of unit with name='%s' because it already exists", su_name)
                continue

            su_description = elems[1]
            su_rank = float(elems[2])
            duration = int(elems[22])
            obs_target = elems[31]
            obs_desc = elems[24]
            pip1_desc = elems[25]
            pip2_desc = elems[26]
            pip3_desc = elems[27]
            pip_cal_desc = elems[28]
            pip1_angle1 = Angle(elems[33]).radian
            pip1_angle2 = Angle(elems[34]).radian
            pip1_target = elems[35]
            pip2_angle1 = Angle(elems[37]).radian
            pip2_angle2 = Angle(elems[38]).radian
            pip2_target = elems[39]
            pip3_angle1 = Angle(elems[41]).radian
            pip3_angle2 = Angle(elems[42]).radian
            pip3_target = elems[43]

            require_day = True if elems[10].startswith('require_day') else False
            min_target_elevation = Angle(elems[11]+' degrees').radian
            min_calibrator_elevation = Angle(elems[12]+' degrees').radian
            offset_window_from = int(elems[13])
            offset_window_to = int(elems[14])
            ref_pointing_angle1 = Angle(elems[15]).radian
            ref_pointing_angle2 = Angle(elems[16]).radian
            ref_pointing_target = elems[17]
            ref_pointing_direction_type = elems[18]
            min_distance_sun = Angle(elems[19]+' degrees').radian
            min_distance_moon = Angle(elems[20]+' degrees').radian
            min_distance_jupiter = Angle(elems[21]+' degrees').radian

            logger.info('Creating SU for observation target=%s duration=%s strategy=%s' % (obs_target, duration, strategy_template.name))
            scheduling_unit_draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template,
                                                                                                  scheduling_set,
                                                                                                  name=su_name,
                                                                                                  description=su_description,
                                                                                                  rank=su_rank,
                                                                                                  priority_queue=models.PriorityQueueType.objects.get(value=models.PriorityQueueType.Choices.B.value),
                                                                                                  specifications_doc_overrides={
                                                                                                      'tasks': {
                                                                                                          'Combined Observation': {
                                                                                                              'short_description': obs_desc,
                                                                                                              'specifications_doc': {
                                                                                                                  'duration': duration,
                                                                                                                  'station_configuration': {'SAPs': [
                                                                                                                      {'digital_pointing': {'angle1': pip1_angle1, 'angle2': pip1_angle2, 'direction_type': 'J2000', 'target': pip1_target}},
                                                                                                                      {'digital_pointing': {'angle1': pip2_angle1, 'angle2': pip2_angle2, 'direction_type': 'J2000', 'target': pip2_target}},
                                                                                                                      {'digital_pointing': {'angle1': pip3_angle1, 'angle2': pip3_angle2, 'direction_type': 'J2000', 'target': pip3_target}}]}
                                                                                                            }
                                                                                                          },
                                                                                                          # TODO: set pipeline settings from csv. Is this needed for acceptance testing?
                                                                                                          'Pipeline target1': {
                                                                                                              'short_description': pip1_desc
                                                                                                          },
                                                                                                          'Pipeline target2': {
                                                                                                              'short_description': pip2_desc
                                                                                                          },
                                                                                                          'Pipeline target3': {
                                                                                                              'short_description': pip3_desc
                                                                                                          },
                                                                                                          'Calibrator Pipeline': {
                                                                                                              'short_description': pip_cal_desc
                                                                                                          },
                                                                                                      },
                                                                                                      'scheduling_constraints_doc': {
                                                                                                          'scheduler': 'dynamic',
                                                                                                          'time': {},
                                                                                                          'daily': {'require_day': False, #require_day,
                                                                                                                    'require_night': False},
                                                                                                          'sky': {'min_distance': {'jupiter': min_distance_jupiter, 'moon': min_distance_moon, 'sun': min_distance_sun},
                                                                                                                  'min_elevation': {'target': min_target_elevation, 'calibrator': min_calibrator_elevation},
                                                                                                                  'transit_offset': {'to': offset_window_to, 'from': offset_window_from},
                                                                                                                  'reference_pointing': {'pointing': {'angle1': ref_pointing_angle1, 'angle2': ref_pointing_angle2, 'target': ref_pointing_target, 'direction_type': ref_pointing_direction_type}, 'enabled': False}
                                                                                                                  }
                                                                                                      }
                                                                                                  })

            create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

    # assert_SUB_statuses(models.SchedulingUnitStatus.Choices.SCHEDULABLE.value)


def populate_LoTSS_test_campaign(project: models.Project, ra_step_deg: int=5, dec_step_deg: int=5):
    scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project,
                                                                                   name="LoTSS run %s" % (datetime.today(),)))
    strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name='IM HBA LoTSS - 2 Beams')

    cal_pointing = {'angle1': 2.1537413875046543, 'angle2': 0.8415517080119604, 'direction_type': 'J2000', 'target': '3C196'}

    for ra in [Angle(angle, unit=units.degree) for angle in range(0, 360, ra_step_deg)]:
        ra1 = ra - Angle(2, unit=units.degree)
        ra2 = ra + Angle(2, unit=units.degree)
        for dec in [Angle(angle, unit=units.degree) for angle in range(30, 90, dec_step_deg)]:
            target1 = 'P%s%s' % (ra1.to_string(units.degree, fields=1, alwayssign=False), dec.to_string(units.degree, fields=1, alwayssign=True))
            target2 = 'P%s%s' % (ra2.to_string(units.degree, fields=1, alwayssign=False), dec.to_string(units.degree, fields=1, alwayssign=True))
            obs_target = target1+' '+target2
            pointing_ref = {'angle1': ra.rad, 'angle2': dec.rad, 'direction_type': 'J2000', 'target': obs_target+'_ref'}
            pointing1 = {'angle1': ra1.rad, 'angle2': dec.rad, 'direction_type': 'J2000', 'target': target1}
            pointing2 = {'angle1': ra2.rad, 'angle2': dec.rad, 'direction_type': 'J2000', 'target': target2}
            duration = timedelta(hours=8)

            logger.info('Creating SU for observation target=%s duration=%s strategy=%s' % (obs_target, duration, strategy_template.name))
            scheduling_unit_draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template,
                                                                                                  scheduling_set,
                                                                                                  name=obs_target,
                                                                                                  description=obs_target,
                                                                                                  rank=models.SchedulingUnitRank.DEFAULT.value,
                                                                                                  priority_queue=models.PriorityQueueType.objects.get(value=models.PriorityQueueType.Choices.A.value),
                                                                                                  specifications_doc_overrides={
                                                                                                      'tasks': {
                                                                                                          'Target Observation': {
                                                                                                              'short_description': obs_target,
                                                                                                              'specifications_doc': { 'duration': int(duration.total_seconds()),
                                                                                                                                      'station_configuration': {
                                                                                                                                          'SAPs': [ { 'digital_pointing': pointing1 },
                                                                                                                                                    { 'digital_pointing': pointing2 }
                                                                                                                                          ],
                                                                                                                                          'tile_beam': pointing_ref
                                                                                                                                      }}
                                                                                                          },
                                                                                                          'Calibrator Observation 1': {
                                                                                                              'short_description': cal_pointing['target'],
                                                                                                              'specifications_doc': { 'calibrator': {'pointing': cal_pointing}}
                                                                                                          },
                                                                                                          'Calibrator Observation 2': {
                                                                                                              'short_description': cal_pointing['target'],
                                                                                                              'specifications_doc': { 'calibrator': {'pointing': cal_pointing}}
                                                                                                          },
                                                                                                          'Calibrator Pipeline 1': {
                                                                                                              'short_description': cal_pointing['target']+'/PP'
                                                                                                          },
                                                                                                          'Calibrator Pipeline 2': {
                                                                                                              'short_description': cal_pointing['target']+'/PP'
                                                                                                          },
                                                                                                          'Pipeline target1': {
                                                                                                              'short_description': pointing1['target']+'/TP'
                                                                                                          },
                                                                                                          'Pipeline target2': {
                                                                                                              'short_description': pointing2['target']+'/TP'
                                                                                                          }
                                                                                                      },
                                                                                                      'scheduling_constraints_doc': {
                                                                                                          'scheduler': 'dynamic',
                                                                                                          'time': {},
                                                                                                          'sky': {
                                                                                                              'transit_offset': {'from': -3600, 'to': 3600 },
                                                                                                              'min_elevation': {'calibrator': Angle(5, unit=units.degree).rad},
                                                                                                              'min_distance': {'moon': Angle(10, unit=units.degree).rad,
                                                                                                                               'sun': Angle(10, unit=units.degree).rad,
                                                                                                                               'jupiter': Angle(10, unit=units.degree).rad},
                                                                                                              'reference_pointing': { 'enabled': False, 'pointing': pointing_ref}
                                                                                                          }
                                                                                                      }
                                                                                                  })

            create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)


def run_acceptance_test():
    # create a half-year cycle, and a project
    cycle, created = models.Cycle.objects.get_or_create(**Cycle_test_data(name="Acceptance Test Cycle", description="acceptance-test cycle for Pulsar Timing & LoLLS",
                                                                          start=datetime.today().replace(day=1, hour=0, minute=0, second=0, microsecond=0),
                                                                          stop=datetime.today().replace(day=1, hour=0, minute=0, second=0, microsecond=0)+timedelta(days=364/2)))
    project_pulsar, created = models.Project.objects.get_or_create(**Project_test_data(name="Pulsar Timing", description="acceptance-test project for pulsar timing"))
    project_pulsar.cycles.set([cycle])
    project_pulsar.save()

    project_lolls, created = models.Project.objects.get_or_create(**Project_test_data(name="LoLLS", description="acceptance-test project for LoLLS"))
    project_lolls.cycles.set([cycle])
    project_lolls.save()

    project_lotss, created = models.Project.objects.get_or_create(**Project_test_data(name="LoTSS", description="acceptance-test project for LoTSS"))
    project_lotss.cycles.set([cycle])
    project_lotss.save()

    stats = TestStats()
    if options.dynamic_scheduling:
        enable_scheduling()
    else:
        disable_scheduling()

    stats.start_test = datetime.utcnow()

    stats.start_populate = datetime.utcnow()
    # populate one campaign constraint to start pretty much right away, at next midnight.
    # They have to fit in a 48 hour window
    start_time = datetime.utcnow().replace(hour=0, minute=0, second=0, microsecond=0) + timedelta(hours=24)
    populate_LoLSS_test_campaign(project=project_lolls,
                                 max_num_targets=options.num_lolls)
    populate_pulsar_timing_test_campaign(project=project_pulsar,
                                         between=[{"from": start_time.isoformat()+'Z',
                                                   "to": (start_time+timedelta(hours=7*24)).isoformat()+'Z'}],
                                         max_num_targets=options.num_pulsars)
    if options.lotss_angle_step > 0:
        populate_LoTSS_test_campaign(project_lotss, options.lotss_angle_step, options.lotss_angle_step)

    stats.stop_populate = datetime.utcnow()
    # assert_SUB_statuses(models.SchedulingUnitStatus.Choices.SCHEDULABLE.value)

    stats.start_scheduling = datetime.utcnow()

    # start scheduling and wait for it to finish
    enable_scheduling()

    if options.simulate:
        from lofar.sas.tmss.test.test_environment import create_simulator_for_scheduled_units
        with create_simulator_for_scheduled_units(delay=options.event_delay, duration=options.running_delay):
            while True:
                # check progress
                num_schedulable = models.SchedulingUnitBlueprint.objects.filter(status__value=models.SchedulingUnitStatus.Choices.SCHEDULABLE.value).count()
                num_scheduled = models.SchedulingUnitBlueprint.objects.filter(status__value=models.SchedulingUnitStatus.Choices.SCHEDULED.value).count()
                num_active = models.SchedulingUnitBlueprint.objects.filter(status__value__in=models.SchedulingUnitStatus.ACTIVE_STATUS_VALUES).count()
                num_done = models.SchedulingUnitBlueprint.objects.filter(status__value__in=models.SchedulingUnitStatus.DONE_STATUS_VALUES).count()
                num_total = models.SchedulingUnitBlueprint.objects.count()

                # log progress
                logger.info('Scheduling unit statuses: #schedulable=%s #scheduled=%s #active=%s #done=%s progress=%.1f%%', num_schedulable, num_scheduled, num_active, num_done, 100.0*float(num_done)/float(num_total))

                # exit when everything has finished, or nothing schedulable anymore
                if num_done == num_total:
                    break

                sleep(1)
    else:
        while True:
            sleep(1)
            scheduler_subsytem = models.Subsystem.objects.get(name='scheduler')
            if scheduler_subsytem.status.value == models.SubsystemStatus.Choices.IDLE.value:
                break

    stats.stop_scheduling = datetime.utcnow()

    # assert success
    # assert_SUB_statuses(models.SchedulingUnitStatus.Choices.FINISHED.value)

    stats.stop_test = datetime.utcnow()
    return stats

stats = run_acceptance_test()

if tmss_test_env is not None:
    tmss_test_env.stop()

stats.print_test_stats()
