#!/usr/bin/env python3

# Copyright (C) 2018    ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

# $Id:  $

import os
import random
import unittest
import requests
import copy
import xml.etree.ElementTree as ET

import logging
logger = logging.getLogger('lofar.'+__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

from lofar.common.test_utils import exit_with_skipped_code_if_skip_integration_tests
exit_with_skipped_code_if_skip_integration_tests()

# Do Mandatory setup step:
# use setup/teardown magic for tmss test database, ldap server and django server
# (ignore pycharm unused import statement, python unittests does use at RunTime the tmss_test_environment_unittest_setup module)

from lofar.sas.tmss.test.test_environment import TMSSTestEnvironment
tmss_test_env = TMSSTestEnvironment(start_calibrator_selector_service=True)
try:
    tmss_test_env.start()
    tmss_test_env.populate_schemas_and_connectors()
except Exception as e:
    logger.exception(str(e))
    tmss_test_env.stop()
    exit(1)

# tell unittest to stop (and automagically cleanup) the test database once all testing is done.
def tearDownModule():
    tmss_test_env.stop()


AUTH = requests.auth.HTTPBasicAuth(tmss_test_env.ldap_server.dbcreds.user, tmss_test_env.ldap_server.dbcreds.password)
BASE_URL = tmss_test_env.django_server.url[:-1] if tmss_test_env.django_server.url.endswith('/')  else tmss_test_env.django_server.url

from lofar.sas.tmss.test.tmss_test_data_django_models import *

# import and setup rest test data creator
from lofar.sas.tmss.test.tmss_test_data_rest import TMSSRESTTestDataCreator
rest_data_creator = TMSSRESTTestDataCreator(BASE_URL, AUTH)

from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.tmssapp.adapters.parset import convert_to_parset, convert_to_parset_dict, _order_beamformer_dataproducts
from lofar.sas.tmss.tmss.tmssapp.adapters.l2stationspecs import convert_to_l2stationspecs_dict
from lofar.common.json_utils import get_default_json_object_for_schema, add_defaults_to_json_object_for_schema, resolved_remote_refs
from lofar.sas.tmss.tmss.tmssapp.adapters.sip import generate_sip_for_dataproduct, get_sip_xml_for_dataproduct, create_sip_representation_for_dataproduct
from lofar.lta.sip import constants
from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions
from lofar.sas.tmss.tmss.tmssapp.tasks import update_task_graph_from_specifications_doc, create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft
from lofar.sas.tmss.tmss.tmssapp.subtasks import schedule_subtask, cancel_subtask, create_observation_control_subtask_from_task_blueprint


class ObservationParsetAdapterTest(unittest.TestCase):

    @staticmethod
    def get_default_specifications():
        subtask_template = models.SubtaskTemplate.get_latest(name='observation control')
        return get_default_json_object_for_schema(subtask_template.schema)

    @staticmethod
    def create_subtask(specifications_doc):
        subtask_template = models.SubtaskTemplate.get_latest(name='observation control')
        subtask_data = Subtask_test_data(subtask_template=subtask_template, specifications_doc=specifications_doc, task_blueprint=models.TaskBlueprint.objects.create(**TaskBlueprint_test_data()))
        subtask:models.Subtask = models.Subtask.objects.create(**subtask_data)
        subtask_output = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=subtask))
        dataproduct:models.Dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=subtask_output))
        return subtask

    def test_correlator(self):
        specifications_doc = self.get_default_specifications()
        specifications_doc['COBALT']['version'] = 1
        specifications_doc['COBALT']['correlator']['enabled'] = True
        specifications_doc['stations']['digital_pointings'] = [
          { "name": "target1",
            "subbands": list(range(8))
          }
        ]

        nr_files = 8 # = nr of subbands

        subtask = self.create_subtask(specifications_doc)
        parset = convert_to_parset_dict(subtask)
        logger.info("test_correlator parset:",parset)

        self.assertEqual(True, parset["Observation.DataProducts.Output_Correlated.enabled"])
        self.assertEqual(False, parset["Observation.DataProducts.Output_CoherentStokes.enabled"])
        self.assertEqual(False, parset["Observation.DataProducts.Output_IncoherentStokes.enabled"])
        self.assertEqual(False, parset["Cobalt.BeamFormer.flysEye"])


    def test_piggyback_keys(self):
        specifications_doc = self.get_default_specifications()
        subtask = self.create_subtask(specifications_doc)
        parset = convert_to_parset_dict(subtask)
        sub = subtask.task_blueprint.scheduling_unit_blueprint

        # Assert the values are the same of the scheduling_unit_blueprint
        self.assertEqual(sub.piggyback_allowed_aartfaac, parset["ObservationControl.StationControl.aartfaacPiggybackAllowed"])
        self.assertEqual(sub.piggyback_allowed_tbb, parset["ObservationControl.StationControl.tbbPiggybackAllowed"])

    def test_flyseye(self):
        specifications_doc = self.get_default_specifications()
        specifications_doc['COBALT']['version'] = 1
        specifications_doc['COBALT']['correlator']['enabled'] = False
        specifications_doc['stations']['station_list'] = ['CS001', 'CS002', 'RS205']
        specifications_doc['stations']['antenna_set'] = 'HBA_DUAL'
        specifications_doc['stations']['digital_pointings'] = [
          { "name": "target1",
            "subbands": list(range(8))
          }
        ]

        specifications_doc['COBALT']['beamformer']['flyseye_pipelines'] = [
            { "coherent": {
                "stokes": "IQUV",
                "time_integration_factor": 4,
                "channels_per_subband": 16
              }
            }
        ]

        nr_files = 5 * 4 # 5 antenna fields (CS001HBA0, CS001HBA1, CS002HBA0, CS002HBA1, RS205HBA) * 4 stokes

        subtask = self.create_subtask(specifications_doc)
        parset = convert_to_parset_dict(subtask)
        logger.info("test_flyseye parset:",parset)

        self.assertEqual(True,     parset["Cobalt.BeamFormer.flysEye"])
        self.assertEqual(True,     parset["Observation.DataProducts.Output_CoherentStokes.enabled"])
        self.assertEqual(nr_files, len(parset["Observation.DataProducts.Output_CoherentStokes.filenames"]))


    def test_beamformer(self):
        specifications_doc = self.get_default_specifications()
        specifications_doc['COBALT']['version'] = 1
        specifications_doc['COBALT']['correlator']['enabled'] = False
        specifications_doc['stations']['digital_pointings'] = [
          { "name": "target1",
            "subbands": list(range(8))
          }
        ]

        specifications_doc['COBALT']['beamformer']['tab_pipelines'] = [
            { "coherent": {
                "stokes": "IQUV",
                "time_integration_factor": 4,
                "channels_per_subband": 16
              },
              "incoherent": {
                "stokes": "IQUV",
                "time_integration_factor": 4,
                "channels_per_subband": 16
              },

              "SAPs": [
                { "name": "target1",
                  "tabs": [
                    {
                      "coherent": True,
                      "pointings": [ { "pointing": { "angle1": 1.0, "angle2": 2.0, "target": "target1" } } ]
                    },
                    {
                      "coherent": False
                    },
                  ]
                }
              ]
            }
        ]

        nr_cs_files = 1 * 4 # 1 TAB * 4 stokes
        nr_is_files = 1 * 4 # 1 TAB * 4 stokes

        subtask = self.create_subtask(specifications_doc)
        parset = convert_to_parset_dict(subtask)
        logger.info("test_beamformer parset:",parset)

        self.assertEqual(subtask.task_blueprint.draft.scheduling_unit_draft.scheduling_set.project.description, parset['Observation.Campaign.title'])
        self.assertEqual(True,        parset["Observation.DataProducts.Output_CoherentStokes.enabled"])
        self.assertEqual(nr_cs_files, len(parset["Observation.DataProducts.Output_CoherentStokes.filenames"]))
        self.assertEqual(True,        parset["Observation.DataProducts.Output_IncoherentStokes.enabled"])
        self.assertEqual(nr_is_files, len(parset["Observation.DataProducts.Output_IncoherentStokes.filenames"]))


    def test_beamformer_multiple_pipelines(self):
        '''check the SIP's for multiple beamformer pipelines with a combination of CS and IS'''
        bf_task_template = models.TaskTemplate.get_latest('beamforming observation')
        bf_task_specifications_doc = bf_task_template.get_default_json_document_for_schema()
        bf_task_specifications_doc['station_configuration']['tile_beam']['target'] = "tile_beam_target"
        bf_task_specifications_doc['station_configuration']['SAPs'][0]['name'] = "SAP1"
        bf_task_specifications_doc['station_configuration']['SAPs'][0]['subbands'] = [0, 1, 2, 3]
        bf_task_specifications_doc['station_configuration']['SAPs'][0]['digital_pointing']['target'] = "SAP1Target"
        bf_task_specifications_doc['beamformer']['pipelines'] = [
            {
                "name": "bf_pipeline_1",
                "coherent": {
                    "SAPs": [
                        {
                            "name": "SAP1",
                            "tabs": [
                                {
                                    "pointing": {
                                        "angle1": 0,
                                        "angle2": 0,
                                        "target": "pl1_cs_sap1_tab1_target",
                                        "direction_type": "J2000"
                                    },
                                    "relative": True
                                },
                                {
                                    "pointing": {
                                        "angle1": 0,
                                        "angle2": 0,
                                        "target": "pl1_cs_sap1_tab2_target",
                                        "direction_type": "J2000"
                                    },
                                    "relative": True
                                }
                            ],
                            "subbands": {
                                "list": [0, 1, 2, 3],
                                "method": "copy"
                            }
                        }
                    ],
                    "settings": {
                        "stokes": "IQUV"
                    }
                },
                "flys eye": {
                    "enabled": False,
                },
                "incoherent": {
                    "SAPs": [
                        {
                            "name": "SAP1",
                            "subbands": {
                                "list": [0, 1, 2, 3],
                                "method": "copy"
                            }
                        }
                    ],
                    "settings": {
                        "stokes": "IQUV"
                    }
                }
            }
        ]

        bf_task_draft = models.TaskDraft.objects.create(**TaskDraft_test_data(name="test_beamformer_multiple_pipelines",
                                                                              specifications_template=bf_task_template,
                                                                              specifications_doc=bf_task_specifications_doc,
                                                                              scheduling_unit_draft=models.SchedulingUnitDraft.objects.create(**SchedulingUnitDraft_test_data(template=models.SchedulingUnitTemplate.get_latest('scheduling unit')))))

        from lofar.sas.tmss.tmss.tmssapp.tasks import create_task_blueprint_and_subtasks_from_task_draft
        bf_task_blueprint = create_task_blueprint_and_subtasks_from_task_draft(bf_task_draft)

        bf_subtask = bf_task_blueprint.subtasks.first()
        bf_subtask = schedule_subtask(bf_subtask, datetime.utcnow()+timedelta(minutes=1))
        parset = convert_to_parset_dict(bf_subtask)
        from pprint import pformat
        logger.info("test_beamformer_multiple_pipelines parset:\n%s", pformat(parset))

        exptected_nr_cs_files = 2 * 4 # 2 TAB * 4 stokes
        exptected_nr_is_files = 4 # 4 stokes
        self.assertEqual(True,        parset["Observation.DataProducts.Output_CoherentStokes.enabled"])
        self.assertEqual(exptected_nr_cs_files, len(parset["Observation.DataProducts.Output_CoherentStokes.filenames"]))
        self.assertEqual(True,        parset["Observation.DataProducts.Output_IncoherentStokes.enabled"])
        self.assertEqual(exptected_nr_is_files, len(parset["Observation.DataProducts.Output_IncoherentStokes.filenames"]))

        # submit feedback for dataproducts
        set_subtask_state_following_allowed_transitions(bf_subtask, 'finishing')
        dp_feedback_template = models.DataproductFeedbackTemplate.get_latest('feedback')
        for dp in bf_subtask.output_dataproducts.all():
            # give each dataproduct a feedback doc with the minimal working default parameters
            # overwrite/fill-in only those relevant for this test.
            dp_feedback_doc = dp_feedback_template.get_default_json_document_for_schema()
            dp_feedback_doc['target']['coherent'] = dp.specifications_doc.get('coherent', False)
            dp.feedback_doc = dp_feedback_doc
            dp.feedback_template = dp_feedback_template
            dp.save()

        # submit feedback for subtask
        subtask_feedback_template = models.SubtaskFeedbackTemplate.get_latest('observation')
        subtask_feedback_doc = subtask_feedback_template.get_default_json_document_for_schema()
        bf_subtask.feedback_template = subtask_feedback_template
        bf_subtask.feedback_doc = subtask_feedback_doc
        bf_subtask.save()

        # finish subtask
        set_subtask_state_following_allowed_transitions(bf_subtask, 'finished')

        # check SIP for all dataproducts
        # IS and CS pipeline should be there, and numberOfProcessing should be 2
        from lxml import etree
        for dp in bf_subtask.output_dataproducts.all():
            sip_xml_doc = dp.sip_xml_doc
            xml_root = etree.fromstring(sip_xml_doc)
            # check dataproduct type attribute via etree
            subArrayPointings = xml_root.xpath('observation/subArrayPointings')
            self.assertEqual(1, len(subArrayPointings))
            numberOfProcessing = int(xml_root.findall("observation/subArrayPointings/subArrayPointing/numberOfProcessing")[0].text)
            self.assertEqual(2, numberOfProcessing)
            self.assertEqual(1, len(xml_root.findall("observation/subArrayPointings/subArrayPointing/processing/coherentStokes")))
            self.assertEqual(1, len(xml_root.findall("observation/subArrayPointings/subArrayPointing/processing/incoherentStokes")))
            self.assertEqual(0, len(xml_root.findall("observation/subArrayPointings/subArrayPointing/processing/flysEye")))


    def test_beamformer_multiple_pipelines_different_SAPs(self):
        '''check the SIP's for multiple beamformer pipelines with a combination of CS and IS'''
        bf_task_template = models.TaskTemplate.get_latest('beamforming observation')
        bf_task_specifications_doc = bf_task_template.get_default_json_document_for_schema()
        bf_task_specifications_doc['station_configuration']['tile_beam']['target'] = "tile_beam_target"
        bf_task_specifications_doc['station_configuration']['SAPs'][0]['name'] = "SAP1"
        bf_task_specifications_doc['station_configuration']['SAPs'][0]['subbands'] = [0, 1, 2, 3]
        bf_task_specifications_doc['station_configuration']['SAPs'][0]['digital_pointing']['target'] = "SAP1Target"
        bf_task_specifications_doc['station_configuration']['SAPs'].append(copy.deepcopy(bf_task_specifications_doc['station_configuration']['SAPs'][0]))
        bf_task_specifications_doc['station_configuration']['SAPs'][1]['name'] = "SAP2"
        bf_task_specifications_doc['station_configuration']['SAPs'][1]['subbands'] = [0, 1, 2, 3]
        bf_task_specifications_doc['station_configuration']['SAPs'][1]['digital_pointing']['target'] = "SAP2Target"
        bf_task_specifications_doc['beamformer']['pipelines'] = [
            {
                "name": "bf_pipeline_1",
                "coherent": {
                    "SAPs": [
                        {
                            "name": "SAP1",
                            "tabs": [
                                {
                                    "pointing": {
                                        "angle1": 0,
                                        "angle2": 0,
                                        "target": "pl1_cs_sap1_tab1_target",
                                        "direction_type": "J2000"
                                    },
                                    "relative": True
                                },
                                {
                                    "pointing": {
                                        "angle1": 0,
                                        "angle2": 0,
                                        "target": "pl1_cs_sap1_tab2_target",
                                        "direction_type": "J2000"
                                    },
                                    "relative": True
                                }
                            ],
                            "subbands": {
                                "list": [0, 1, 2, 3],
                                "method": "copy"
                            }
                        }
                    ],
                    "settings": {
                        "stokes": "IQUV"
                    }
                },
                "flys eye": {
                    "enabled": False,
                },
                "incoherent": {
                    "SAPs": [
                        {
                            "name": "SAP2",
                            "subbands": {
                                "list": [0, 1, 2, 3],
                                "method": "copy"
                            }
                        }
                    ],
                    "settings": {
                        "stokes": "IQUV"
                    }
                },
                'station_groups': bf_task_specifications_doc['station_configuration']['station_groups']
            }
        ]

        bf_task_draft = models.TaskDraft.objects.create(**TaskDraft_test_data(name="test_beamformer_multiple_pipelines",
                                                                              specifications_template=bf_task_template,
                                                                              specifications_doc=bf_task_specifications_doc,
                                                                              scheduling_unit_draft=models.SchedulingUnitDraft.objects.create(**SchedulingUnitDraft_test_data(template=models.SchedulingUnitTemplate.get_latest('scheduling unit')))))

        from lofar.sas.tmss.tmss.tmssapp.tasks import create_task_blueprint_and_subtasks_from_task_draft
        bf_task_blueprint = create_task_blueprint_and_subtasks_from_task_draft(bf_task_draft)

        bf_subtask = bf_task_blueprint.subtasks.first()
        bf_subtask = schedule_subtask(bf_subtask, datetime.utcnow() + timedelta(minutes=100))

        # submit feedback for dataproducts
        set_subtask_state_following_allowed_transitions(bf_subtask, 'finishing')
        dp_feedback_template = models.DataproductFeedbackTemplate.get_latest('feedback')
        for dp in bf_subtask.output_dataproducts.all():
            # give each dataproduct a feedback doc with the minimal working default parameters
            # overwrite/fill-in only those relevant for this test.
            dp_feedback_doc = dp_feedback_template.get_default_json_document_for_schema()
            dp_feedback_doc['target']['coherent'] = dp.specifications_doc.get('coherent', False)
            dp.feedback_doc = dp_feedback_doc
            dp.feedback_template = dp_feedback_template
            dp.save()

        # submit feedback for subtask
        subtask_feedback_template = models.SubtaskFeedbackTemplate.get_latest('observation')
        subtask_feedback_doc = subtask_feedback_template.get_default_json_document_for_schema()
        bf_subtask.feedback_template = subtask_feedback_template
        bf_subtask.feedback_doc = subtask_feedback_doc
        bf_subtask.save()

        # finish subtask
        set_subtask_state_following_allowed_transitions(bf_subtask, 'finished')

        # check SIP for all dataproducts
        # IS and CS pipeline should be there, but numberOfProcessing should be 1 on each SAP
        from lxml import etree
        for dp in bf_subtask.output_dataproducts.all():
            sip_xml_doc = dp.sip_xml_doc
            xml_root = etree.fromstring(sip_xml_doc)
            # check dataproduct type attribute via etree
            subArrayPointings = xml_root.xpath('observation/subArrayPointings')
            self.assertEqual(1, len(subArrayPointings))
            numberOfProcessing = [int(item.text) for item in xml_root.findall("observation/subArrayPointings/subArrayPointing/numberOfProcessing")]
            self.assertEqual([1, 1], numberOfProcessing)
            self.assertEqual(1, len(xml_root.findall("observation/subArrayPointings/subArrayPointing/processing/coherentStokes")))
            self.assertEqual(1, len(xml_root.findall("observation/subArrayPointings/subArrayPointing/processing/incoherentStokes")))
            self.assertEqual(0, len(xml_root.findall("observation/subArrayPointings/subArrayPointing/processing/flysEye")))


    def test_order_beamformer_dataproducts(self):
        specifications_doc = self.get_default_specifications()
        specifications_doc['COBALT']['version'] = 1
        specifications_doc['COBALT']['correlator']['enabled'] = False
        specifications_doc['stations']['digital_pointings'] = [
          { "name": "target1",
            "subbands": list(range(8))
          },
          { "name": "target2",
            "subbands": list(range(8))
          }
        ]

        specifications_doc['COBALT']['beamformer']['tab_pipelines'] = [
            { "coherent": {
                "stokes": "IQUV",
                "time_integration_factor": 4,
                "channels_per_subband": 16
              },
              "incoherent": {
                "stokes": "IQUV",
                "time_integration_factor": 4,
                "channels_per_subband": 16
              },
              "SAPs": [
                { "name": "target2",
                  "tabs": [
                    {
                      "coherent": True,
                      "pointing": { "angle1": 1.0, "angle2": 2.0, "target": "target2.1" }
                    },
                    {
                      "coherent": True,
                      "pointing": { "angle1": 1.0, "angle2": 2.0, "target": "target2.2" }
                    },
                    {
                      "coherent": False
                    },
                  ]
                }
              ]
            },
            { "coherent": {
                "stokes": "I",
                "time_integration_factor": 4,
                "channels_per_subband": 16
              },
              "incoherent": {
                "stokes": "I",
                "time_integration_factor": 4,
                "channels_per_subband": 16
              },
              "SAPs": [
                { "name": "target2",
                  "tabs": [
                    {
                      "coherent": True,
                      "pointing": { "angle1": 1.0, "angle2": 2.0, "target": "target2.3" }
                    }
                  ]
                },
                { "name": "target1",
                  "tabs": [
                    {
                      "coherent": True,
                      "pointing": { "angle1": 1.0, "angle2": 2.0, "target": "target1.1" }
                    },
                    {
                      "coherent": False
                    },
                  ]
                }
              ]
            }
        ]

        # Add any missing defaults
        subtask_template = models.SubtaskTemplate.get_latest(name='observation control')
        specifications_doc = add_defaults_to_json_object_for_schema(specifications_doc, resolved_remote_refs(subtask_template.schema))

        # associated dataproducts
        dataproducts_in = [
          models.Dataproduct(specifications_doc={"sap": "target2", "identifiers": { "pipeline_index": 0, "tab_index": 0, "stokes_index": 0, "part_index": 0 }, "coherent": True}),
          models.Dataproduct(specifications_doc={"sap": "target2", "identifiers": { "pipeline_index": 0, "tab_index": 0, "stokes_index": 1, "part_index": 0 }, "coherent": True}),
          models.Dataproduct(specifications_doc={"sap": "target2", "identifiers": { "pipeline_index": 0, "tab_index": 0, "stokes_index": 2, "part_index": 0 }, "coherent": True}),
          models.Dataproduct(specifications_doc={"sap": "target2", "identifiers": { "pipeline_index": 0, "tab_index": 0, "stokes_index": 3, "part_index": 0 }, "coherent": True}),

          models.Dataproduct(specifications_doc={"sap": "target2", "identifiers": { "pipeline_index": 0, "tab_index": 1, "stokes_index": 0, "part_index": 0 }, "coherent": True}),
          models.Dataproduct(specifications_doc={"sap": "target2", "identifiers": { "pipeline_index": 0, "tab_index": 1, "stokes_index": 1, "part_index": 0 }, "coherent": True}),
          models.Dataproduct(specifications_doc={"sap": "target2", "identifiers": { "pipeline_index": 0, "tab_index": 1, "stokes_index": 2, "part_index": 0 }, "coherent": True}),
          models.Dataproduct(specifications_doc={"sap": "target2", "identifiers": { "pipeline_index": 0, "tab_index": 1, "stokes_index": 3, "part_index": 0 }, "coherent": True}),

          models.Dataproduct(specifications_doc={"sap": "target2", "identifiers": { "pipeline_index": 0, "tab_index": 2, "stokes_index": 0, "part_index": 0 }, "coherent": False}),
          models.Dataproduct(specifications_doc={"sap": "target2", "identifiers": { "pipeline_index": 0, "tab_index": 2, "stokes_index": 1, "part_index": 0 }, "coherent": False}),
          models.Dataproduct(specifications_doc={"sap": "target2", "identifiers": { "pipeline_index": 0, "tab_index": 2, "stokes_index": 2, "part_index": 0 }, "coherent": False}),
          models.Dataproduct(specifications_doc={"sap": "target2", "identifiers": { "pipeline_index": 0, "tab_index": 2, "stokes_index": 3, "part_index": 0 }, "coherent": False}),

          models.Dataproduct(specifications_doc={"sap": "target2", "identifiers": { "pipeline_index": 1, "tab_index": 0, "stokes_index": 0, "part_index": 0 }, "coherent": True}),

          models.Dataproduct(specifications_doc={"sap": "target1", "identifiers": { "pipeline_index": 1, "tab_index": 0, "stokes_index": 0, "part_index": 0 }, "coherent": True}),

          models.Dataproduct(specifications_doc={"sap": "target1", "identifiers": { "pipeline_index": 1, "tab_index": 1, "stokes_index": 0, "part_index": 0 }, "coherent": False}),
        ]

        # shuffle them
        random.shuffle(dataproducts_in)

        # order them
        coherent_dataproducts, incoherent_dataproducts = _order_beamformer_dataproducts(dataproducts_in, specifications_doc)

        # check length
        self.assertEqual(len(dataproducts_in), len(coherent_dataproducts) + len(incoherent_dataproducts), msg=f"{len(coherent_dataproducts)} coherent and {len(incoherent_dataproducts)} incoherent dataproducts were returned")

        # check order
        for is_coherent_list, dataproducts in ((True,coherent_dataproducts), (False,incoherent_dataproducts)):
            for prev, curr in zip([None] + dataproducts, dataproducts + [None]):
                if prev is None or curr is None:
                    continue

                # we provided all dataproducts, so there should be no stubs
                self.assertIsNotNone(prev)
                self.assertIsNotNone(curr)

                # returned dataproduct must have the 'coherent' flag as per the list it is in
                self.assertEqual(is_coherent_list, prev.specifications_doc["coherent"])
                self.assertEqual(is_coherent_list, curr.specifications_doc["coherent"])

                # dataproducts within each list must be in-order with respect
                # to what COBALT expects. python's comparison of tuples shines here.
                self.assertLessEqual(
                    (prev.specifications_doc["sap"],
                     prev.specifications_doc["identifiers"]["pipeline_index"],
                     prev.specifications_doc["identifiers"]["tab_index"],
                     prev.specifications_doc["identifiers"]["stokes_index"],
                     prev.specifications_doc["identifiers"]["part_index"]),
                    (curr.specifications_doc["sap"],
                     curr.specifications_doc["identifiers"]["pipeline_index"],
                     curr.specifications_doc["identifiers"]["tab_index"],
                     curr.specifications_doc["identifiers"]["stokes_index"],
                     curr.specifications_doc["identifiers"]["part_index"]),
                    msg=f"coherent list: {is_coherent_list}"
                )

class PulsarPipelineParsetAdapterTest(unittest.TestCase):
    def create_subtask(self, specifications_doc={}):
        subtask_template = models.SubtaskTemplate.get_latest(name='pulsar pipeline')
        specifications_doc = add_defaults_to_json_object_for_schema(specifications_doc, subtask_template.schema)

        subtask_data = Subtask_test_data(subtask_template=subtask_template, specifications_doc=specifications_doc, task_blueprint=models.TaskBlueprint.objects.create(**TaskBlueprint_test_data(short_description="Test pipeline task_blueprint")))
        subtask:models.Subtask = models.Subtask.objects.create(**subtask_data)

        subtask_output = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=subtask))
        for i in range(8):
            dataproduct:models.Dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=subtask_output, filename='myoutputtestfile_%s' % i))
        return subtask

    def test_pulp(self):
        # create obs subtask
        specifications_doc = ObservationParsetAdapterTest.get_default_specifications()
        specifications_doc['COBALT']['version'] = 1
        specifications_doc['COBALT']['correlator']['enabled'] = True
        specifications_doc['stations']['digital_pointings'] = [
            {"name": "target1",
             "subbands": list(range(8))
             }
        ]
        obs_subtask = ObservationParsetAdapterTest.create_subtask(specifications_doc)

        # create pulp subtask
        subtask = self.create_subtask()

        # Connect them, so there is a predecessor obs for parset generation
        subtask_output = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=obs_subtask))
        subtask_input = models.SubtaskInput.objects.create(**SubtaskInput_test_data(subtask=subtask, producer=subtask_output))
        subtask_preds_ids = [predecessor.id for predecessor in subtask.predecessors]

        # generate parset
        parset = convert_to_parset_dict(subtask)
        logger.info("test_pulp parset: %s" % parset)

        self.assertEqual(True, parset["Observation.DataProducts.Output_Pulsar.enabled"])
        self.assertEqual("Test pipeline task_blueprint", parset["Observation.Scheduler.taskName"])
        self.assertEqual(subtask_preds_ids, parset["Observation.Scheduler.predecessors"])
        self.assertEqual(parset['Observation.DataProducts.Output_Pulsar.filenames'], ['myoutputtestfile_%s' % i for i in range(8)])


class SIPadapterTest(unittest.TestCase):
    def setUp(self) -> None:
        tmss_test_env.delete_scheduling_unit_drafts_cascade()

    def test_simple_sip_generate_from_dataproduct(self):
        """
        Test if SIP is generated successfully when subtask, dataproduct and SAP objects are created
        Check some value in the SIP (xml) output
        Check that the SIP identifiers are in SIP (xml) output
        Check the number of SIP identifiers are increased with 3
        Check that all SIP identifiers are unique
        """
        subtask_template = models.SubtaskTemplate.get_latest(name='observation control')
        specifications_doc = get_default_json_object_for_schema(subtask_template.schema)
        specifications_doc['stations']['filter'] = "HBA_210_250"
        specifications_doc['COBALT']['correlator']['enabled'] = True
        feedback_template = models.DataproductFeedbackTemplate.get_latest(name='feedback')
        feedback_doc = {'percentage_written': 100, 'frequency': {'subbands': [156], 'station_subbands': [42], 'central_frequencies': [33593750.0], 'channel_width': 6103.515625, 'channels_per_subband': 32}, 'time': {'start_time': '2013-02-16T17:00:00', 'duration': 5.02732992172, 'sample_width': 2.00278016}, 'antennas': {'set': 'HBA_DUAL', 'fields': [{'type': 'HBA', 'field': 'HBA0', 'station': 'CS001'}, {'type': 'HBA', 'field': 'HBA1', 'station': 'CS001'}]}, 'target': {'pointing': {'angle1': 0, 'angle2': 0, 'direction_type': 'J2000', "target": "target1"}}, 'samples': {'polarisations': ['XX', 'XY', 'YX', 'YY'], 'type': 'float', 'bits': 32, 'writer': 'standard', 'writer_version': '2.2.0', 'complex': True}}
        feedback_doc = feedback_template.add_defaults_to_json_object_for_schema(feedback_doc)
        for dp in specifications_doc['stations']['digital_pointings']:
            dp['subbands'] = list(range(8))
        # Create SubTask(output)
        subtask_data = Subtask_test_data(subtask_template=subtask_template, specifications_doc=specifications_doc, task_blueprint=models.TaskBlueprint.objects.create(**TaskBlueprint_test_data()))
        subtask:models.Subtask = models.Subtask.objects.create(**subtask_data)
        subtask_output = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=subtask))
        # Create Dataproduct
        dataproduct: models.Dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(feedback_doc=feedback_doc,
                                                                                                    feedback_template=feedback_template,
                                                                                                    producer=subtask_output))

        # Create SAP
        sap_template = models.SAPTemplate.get_latest(name="SAP")
        specifications_doc = get_default_json_object_for_schema(sap_template.schema)
        sap = models.SAP.objects.create(specifications_doc=specifications_doc, specifications_template=sap_template)
        sap.save()

        dataproduct.sap = sap
        dataproduct.save()

        set_subtask_state_following_allowed_transitions(subtask, 'finished')

        sip = generate_sip_for_dataproduct(dataproduct)

        # double-check that SIP contains values from feedback and specifications docs
        self.assertIn(str(feedback_doc['frequency']['channel_width']), sip.get_prettyxml())
        self.assertIn(str(feedback_doc['time']['start_time']), sip.get_prettyxml())
        self.assertIn(constants.FILTERSELECTIONTYPE_210_250_MHZ, sip.get_prettyxml()) # specifications_doc: "HBA_210_250"

        self.assertIn(str(subtask.global_identifier.unique_identifier), sip.get_prettyxml())
        self.assertIn(str(dataproduct.global_identifier.unique_identifier), sip.get_prettyxml())
        self.assertIn(str(sap.global_identifier.unique_identifier), sip.get_prettyxml())

        # assert that a time MeasurementSet dataproduct in TMSS creates a CorrelatedDataproduct in the SIP
        self.assertIn(str('<dataProduct xsi:type="sip:CorrelatedDataProduct">'), sip.get_prettyxml())
        self.assertIn(str('<channelWidth'), sip.get_prettyxml().split('<correlator xsi:type="sip:Correlator">')[1].split('</correlator>')[0])

        # assert there is a correlator processing element because the correlator is enabled
        self.assertIn(str('<correlator xsi:type="sip:Correlator">'), sip.get_prettyxml())

    def test_simple_sip_generate_from_dataproduct_beamformed(self):
        """
        Test if SIP is generated successfully when subtask, dataproduct and SAP objects are created
        Check some value in the SIP (xml) output
        Check that the SIP identifiers are in SIP (xml) output
        Check the number of SIP identifiers are increased with 3
        Check that all SIP identifiers are unique
        """
        subtask_template = models.SubtaskTemplate.get_latest(name='observation control')
        specifications_doc = get_default_json_object_for_schema(subtask_template.schema)
        specifications_doc['stations']['filter'] = "HBA_210_250"
        specifications_doc['COBALT']['correlator']['enabled'] = False
        feedback_template = models.DataproductFeedbackTemplate.get_latest(name='feedback')
        feedback_doc = {'percentage_written': 100, 'frequency': {'subbands': [156], 'central_frequencies': [33593750.0], 'channel_width': 6103.515625, 'channels_per_subband': 32}, 'time': {'start_time': '2013-02-16T17:00:00', 'duration': 5.02732992172, 'sample_width': 2.00278016}, 'antennas': {'set': 'HBA_DUAL', 'fields': [{'type': 'HBA', 'field': 'HBA0', 'station': 'CS001'}, {'type': 'HBA', 'field': 'HBA1', 'station': 'CS001'}]}, 'target': {'pointing': {'angle1': 0, 'angle2': 0, 'direction_type': 'J2000', "target": "target1"}}, 'samples': {'polarisations': ['XX', 'XY', 'YX', 'YY'], 'type': 'float', 'bits': 32, 'writer': 'standard', 'writer_version': '2.2.0', 'complex': True}}
        feedback_doc = feedback_template.add_defaults_to_json_object_for_schema(feedback_doc)
        for dp in specifications_doc['stations']['digital_pointings']:
            dp['subbands'] = list(range(8))
        # Create SubTask(output)
        subtask_data = Subtask_test_data(subtask_template=subtask_template, specifications_doc=specifications_doc, task_blueprint=models.TaskBlueprint.objects.create(**TaskBlueprint_test_data()))
        subtask:models.Subtask = models.Subtask.objects.create(**subtask_data)
        subtask_output = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=subtask))

        # Create SAP
        sap_template = models.SAPTemplate.get_latest(name="SAP")
        specifications_doc = get_default_json_object_for_schema(sap_template.schema)
        sap = models.SAP.objects.create(specifications_doc=specifications_doc, specifications_template=sap_template)

        set_subtask_state_following_allowed_transitions(subtask, 'finished')

        # Create Dataproduct
        dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(feedback_doc=feedback_doc,
                                                                                feedback_template=feedback_template,
                                                                                producer=subtask_output,
                                                                                dataformat=models.Dataformat.objects.get(value="Beamformed"),
                                                                                datatype=models.Datatype.objects.get(value="time series"),
                                                                                sap=sap))



        # wipe sip cache and generate
        create_sip_representation_for_dataproduct.cache_clear()
        sip = generate_sip_for_dataproduct(dataproduct)

        # double-check that SIP contains values from feedback and specifications docs
        self.assertIn(str(feedback_doc['frequency']['channel_width']), sip.get_prettyxml())
        self.assertIn(constants.FILTERSELECTIONTYPE_210_250_MHZ, sip.get_prettyxml()) # specifications_doc: "HBA_210_250"
        for pol in feedback_doc['samples']['polarisations']:
            self.assertIn(str(pol), sip.get_prettyxml())

        self.assertIn(str(subtask.global_identifier.unique_identifier), sip.get_prettyxml())
        self.assertIn(str(dataproduct.global_identifier.unique_identifier), sip.get_prettyxml())
        self.assertIn(str(sap.global_identifier.unique_identifier), sip.get_prettyxml())

        # assert there are is no correlator processing element because the correlator is disabled
        self.assertNotIn(str('<correlator xsi:type="sip:Correlator">'), sip.get_prettyxml())

        # assert that a Beamformed dataproduct in TMSS creates a BeamformedDataproduct in the SIP
        self.assertIn(str('<dataProduct xsi:type="sip:BeamFormedDataProduct">'), sip.get_prettyxml())

        # assert we get a coherent stokes beam by default
        self.assertIn(str('CoherentStokesBeam'), sip.get_prettyxml())

        dp_spec_template = models.DataproductSpecificationsTemplate.get_latest(name='time series')

        # alternative dataproduct, create sip
        feedback_doc['target']['coherent'] = False
        dataproduct_data = Dataproduct_test_data(feedback_doc=feedback_doc,
                                                 feedback_template=feedback_template,
                                                 specifications_template=dp_spec_template,
                                                 producer=subtask_output,
                                                 dataformat=models.Dataformat.objects.get(value="Beamformed"),
                                                 datatype=models.Datatype.objects.get(value="time series"),
                                                 sap=sap)
        dataproduct = models.Dataproduct.objects.create(**dataproduct_data)


        # wipe sip cache and regenerate
        create_sip_representation_for_dataproduct.cache_clear()
        sip = generate_sip_for_dataproduct(dataproduct)

        # assert we get an incoherent stokes beam
        self.assertIn(str('<arrayBeam xsi:type="sip:IncoherentStokesBeam">'), sip.get_prettyxml())

        # alter dataproduct, recreate sip
        feedback_doc['antennas']['fields'] = [{'type': 'HBA', 'field': 'HBA0', 'station': 'CS001'}]
        dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(feedback_doc=feedback_doc,
                                                                                feedback_template=feedback_template,
                                                                                specifications_template=dp_spec_template,
                                                                                producer=subtask_output,
                                                                                dataformat=models.Dataformat.objects.get(value="Beamformed"),
                                                                                datatype=models.Datatype.objects.get(value="time series"),
                                                                                sap=sap))

        # wipe sip cache and regenerate
        create_sip_representation_for_dataproduct.cache_clear()
        sip = generate_sip_for_dataproduct(dataproduct)

        # assert we get a flyseye beam if we have a single antenna field
        self.assertIn(str('<arrayBeam xsi:type="sip:FlysEyeBeam">'), sip.get_prettyxml())


    def test_simple_sip_generate_from_dataproduct_pulp(self):
        """
        Test if SIP is generated successfully when subtask, dataproduct and SAP objects are created
        Check some value in the SIP (xml) output
        Check that the SIP identifiers are in SIP (xml) output
        Check the number of SIP identifiers are increased with 3
        Check that all SIP identifiers are unique
        """
        subtask_template = models.SubtaskTemplate.get_latest(name='observation control')
        specifications_doc = get_default_json_object_for_schema(subtask_template.schema)
        specifications_doc['stations']['filter'] = "HBA_110_190"
        feedback_template = models.DataproductFeedbackTemplate.get_latest(name='feedback')
        feedback_doc = {'percentage_written': 100, 'frequency': {'subbands': [152], 'central_frequencies': [33593750.0], 'channel_width': 3051.7578125, 'channels_per_subband': 64}, 'time': {'start_time': '2013-02-16T17:00:00', 'duration': 5.02732992172, 'sample_width': 2.00278016}, 'antennas': {'set': 'HBA_DUAL', 'fields': [{'type': 'HBA', 'field': 'HBA0', 'station': 'CS001'}, {'type': 'HBA', 'field': 'HBA1', 'station': 'CS001'}]}, 'target': {'pointing': {'angle1': 0, 'angle2': 0, 'direction_type': 'J2000', "target": "target1"}, 'coherent': True}, 'samples': {'polarisations': ['XX', 'XY', 'YX', 'YY'], 'type': 'float', 'bits': 32, 'writer': 'standard', 'writer_version': '2.2.0', 'complex': True}, 'files': ['stokes/SAP0/CS003HBA1/L773569_SAP000_B005_S0_P000_bf.h5', 'stokes/SAP0/RS106HBA/L773569_SAP000_B046_S0_P000_bf.h5']}
        feedback_doc = feedback_template.add_defaults_to_json_object_for_schema(feedback_doc)

        for dp in specifications_doc['stations']['digital_pointings']:
            dp['subbands'] = list(range(8))
        # Create SubTask(output)
        subtask_data = Subtask_test_data(subtask_template=subtask_template, specifications_doc=specifications_doc, task_blueprint=models.TaskBlueprint.objects.create(**TaskBlueprint_test_data()))
        subtask:models.Subtask = models.Subtask.objects.create(**subtask_data)
        subtask_output = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=subtask))

        # Create SAP
        sap_template = models.SAPTemplate.get_latest(name="SAP")
        specifications_doc = get_default_json_object_for_schema(sap_template.schema)
        sap = models.SAP.objects.create(specifications_doc=specifications_doc, specifications_template=sap_template)
        sap.save()

        set_subtask_state_following_allowed_transitions(subtask, 'finished')
        create_sip_representation_for_dataproduct.cache_clear()

        # PULP ANALYSIS
        # Create Dataproduct
        dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(feedback_doc=feedback_doc,
                                                                                feedback_template=feedback_template,
                                                                                producer=subtask_output,
                                                                                dataformat=models.Dataformat.objects.get(value="pulp analysis"),
                                                                                datatype=models.Datatype.objects.get(value="pulsar profile"),
                                                                                sap=sap))

        sip = generate_sip_for_dataproduct(dataproduct)

        # double-check that SIP contains values from feedback and specifications docs
        self.assertIn(str(feedback_doc['frequency']['channel_width']), sip.get_prettyxml())
        self.assertIn(constants.FILTERSELECTIONTYPE_110_190_MHZ, sip.get_prettyxml())
        for pol in feedback_doc['samples']['polarisations']:
            self.assertIn(str(pol), sip.get_prettyxml())

        self.assertIn(str(subtask.global_identifier.unique_identifier), sip.get_prettyxml())
        self.assertIn(str(dataproduct.global_identifier.unique_identifier), sip.get_prettyxml())
        self.assertIn(str(sap.global_identifier.unique_identifier), sip.get_prettyxml())

        # assert that a pulp analysis dataproduct in TMSS creates a PulpDataProduct in the SIP
        self.assertIn(str('<dataProduct xsi:type="sip:PulpDataProduct">'), sip.get_prettyxml())

        # assert beam type
        self.assertIn(str('CoherentStokesBeam'), sip.get_prettyxml())

        # assert datatype
        self.assertIn(str('<dataType>ComplexVoltages</dataType>'), sip.get_prettyxml())

        # assert fileformat
        self.assertIn(str('<fileFormat>PULP</fileFormat>'), sip.get_prettyxml())

        # create alternative dataproduct, recreate sip
        feedback_doc['target']['coherent'] = False
        dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(feedback_doc=feedback_doc,
                                                                                feedback_template=feedback_template,
                                                                                producer=subtask_output,
                                                                                dataformat=models.Dataformat.objects.get(value="pulp analysis"),
                                                                                datatype=models.Datatype.objects.get(value="pulsar profile"),
                                                                                sap=sap))

        # wipe sip cache and regenerate
        create_sip_representation_for_dataproduct.cache_clear()
        sip = generate_sip_for_dataproduct(dataproduct)

        # assert datatype reflects change of coherent flag
        self.assertIn(str('<dataType>IncoherentStokes</dataType>'), sip.get_prettyxml())

        # PULP SUMMARY

        # create alternative dataproduct, recreate sip
        feedback_doc['target']['coherent'] = False
        dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(feedback_doc=feedback_doc,
                                                                                feedback_template=feedback_template,
                                                                                producer=subtask_output,
                                                                                dataformat=models.Dataformat.objects.get(value="pulp summary"),
                                                                                datatype=models.Datatype.objects.get(value="pulsar profile"),
                                                                                sap=sap))

        # wipe sip cache and regenerate
        create_sip_representation_for_dataproduct.cache_clear()
        sip = generate_sip_for_dataproduct(dataproduct)

        # assert datatype reflects change of dataformat
        self.assertIn(str('<dataType>SummaryIncoherentStokes</dataType>'), sip.get_prettyxml())

        # assert that a pulp summary dataproduct in TMSS creates a PulpSummaryDataProduct in the SIP
        self.assertIn(str('<dataProduct xsi:type="sip:PulpSummaryDataProduct">'), sip.get_prettyxml())

        # assert fileformat
        self.assertIn(str('<fileFormat>PULP</fileFormat>'), sip.get_prettyxml())

    def test_pulp_dataproduct_from_pulsar_timing_strategy(self):
        '''Check the pulp dataproducts generated by a Pulsar Timing stategy.'''
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name='BF Pulsar Timing')
        scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(name="Pulsar timing",
                                                                          specifications_template=strategy_template.scheduling_unit_template,
                                                                          observation_strategy_template=strategy_template,
                                                                          scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))

        scheduling_unit_draft = update_task_graph_from_specifications_doc(scheduling_unit_draft, strategy_template.template)
        scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
        subtasks = models.Subtask.objects.filter(task_blueprint__scheduling_unit_blueprint__id=scheduling_unit_blueprint.id)
        obs_subtask = subtasks.get(specifications_template__type__value=models.SubtaskType.Choices.OBSERVATION.value)
        pulp_subtask = subtasks.get(specifications_template__type__value=models.SubtaskType.Choices.PIPELINE.value)

        for subtask in [obs_subtask, pulp_subtask]:
            subtask = schedule_subtask(subtask)
            set_subtask_state_following_allowed_transitions(subtask, 'finishing')

            feedback_template = models.DataproductFeedbackTemplate.get_latest(name='feedback')
            feedback_doc = {'percentage_written': 100, 'frequency': {'subbands': [152], 'central_frequencies': [33593750.0], 'channel_width': 3051.7578125, 'channels_per_subband': 64}, 'time': {'start_time': '2013-02-16T17:00:00', 'duration': 5.02732992172, 'sample_width': 2.00278016}, 'antennas': {'set': 'HBA_DUAL', 'fields': [{'type': 'HBA', 'field': 'HBA0', 'station': 'CS001'}, {'type': 'HBA', 'field': 'HBA1', 'station': 'CS001'}]}, 'target': {'pointing': {'angle1': 0, 'angle2': 0, 'direction_type': 'J2000', "target": "target1"}, 'coherent': True}, 'samples': {'polarisations': ['XX', 'XY', 'YX', 'YY'], 'type': 'float', 'bits': 32, 'writer': 'standard', 'writer_version': '2.2.0', 'complex': True}, 'files': ['stokes/SAP0/CS003HBA1/L773569_SAP000_B005_S0_P000_bf.h5', 'stokes/SAP0/RS106HBA/L773569_SAP000_B046_S0_P000_bf.h5']}
            feedback_doc = feedback_template.add_defaults_to_json_object_for_schema(feedback_doc)

            feedback_template_pulp_sum = models.DataproductFeedbackTemplate.get_latest(name='pulp summary')
            feedback_doc_pulp_sum = feedback_template_pulp_sum.get_default_json_document_for_schema()

            for dp in subtask.output_dataproducts.all():
                if dp.dataformat.value == models.Dataformat.Choices.PULP_SUMMARY.value:
                    dp.feedback_template = feedback_template_pulp_sum
                    dp.feedback_doc = feedback_doc_pulp_sum
                else:
                    dp.feedback_template = feedback_template
                    dp.feedback_doc = feedback_doc

                dp.save()
            set_subtask_state_following_allowed_transitions(subtask, 'finished')

        # wipe sip cache
        create_sip_representation_for_dataproduct.cache_clear()

        # generate sip for PULP_ANALYSIS dp
        sip = generate_sip_for_dataproduct(pulp_subtask.output_dataproducts.filter(dataformat__value=models.Dataformat.Choices.PULP_ANALYSIS.value).first())

        # prove that the sip contains a pipelineRun node of type PulsarPipeline, and check PulsarPipeline childnodes (settings)
        sipxml = sip.get_prettyxml()
        self.assertIn(str('<pipelineRun xsi:type="sip:PulsarPipeline">'), sipxml)
        self.assertIn(str('<storageWriter>Unknown</storageWriter>'), sipxml)
        self.assertNotIn(str('<name>Parset: </name>'), sipxml)
        self.assertNotIn(str('<name>Observation</name>'), sipxml)
        self.assertNotIn(str('<name>Pipeline</name>'), sipxml)
        self.assertNotIn(str('<pipelineName></pipelineName>'), sipxml)
        self.assertNotIn(str('<name></name>'), sipxml)
        #TODO: check PulsarPipeline childnodes (settings). This needs an implementation in create_sip_representation_for_subtask!

        # generate sip for PULP_SUMMARY dp
        sip = generate_sip_for_dataproduct(pulp_subtask.output_dataproducts.filter(dataformat__value=models.Dataformat.Choices.PULP_SUMMARY.value).first())

        # prove that the sip contains a pipelineRun node of type PulsarPipeline, and check PulsarPipeline childnodes (settings)
        sipxml = sip.get_prettyxml()
        self.assertIn(str('<pipelineRun xsi:type="sip:PulsarPipeline">'), sipxml)
        self.assertIn(str('<storageWriter>Unknown</storageWriter>'), sipxml)
        self.assertNotIn(str('<name>Parset: </name>'), sipxml)
        self.assertNotIn(str('<name>Observation</name>'), sipxml)
        self.assertNotIn(str('<name>Pipeline</name>'), sipxml)
        self.assertNotIn(str('<pipelineName></pipelineName>'), sipxml)
        self.assertNotIn(str('<name></name>'), sipxml)
        #TODO: more PULP_SUMMARY items to check in SIP ??

    def test_dataproduct_from_combined_calibrator_target(self):
        '''Check the SIP for a dataproduct from a parallel calibrator and target observation'''

        # create calibrator and target tasks with parallel placement
        task_template = models.TaskTemplate.get_latest(name="parallel calibrator target observation")
        task_spec = task_template.add_defaults_to_json_object_for_schema({})
        task_draft = models.TaskDraft.objects.create(**TaskDraft_test_data(specifications_template=task_template, specifications_doc=task_spec))
        task_blueprint = models.TaskBlueprint.objects.create(**TaskBlueprint_test_data(task_draft=task_draft))

        target_subtask = create_observation_control_subtask_from_task_blueprint(task_blueprint)

        import datetime
        subtask = schedule_subtask(target_subtask, datetime.datetime.utcnow() + datetime.timedelta(minutes=5))
        set_subtask_state_following_allowed_transitions(subtask, 'finishing')

        feedback_template = models.DataproductFeedbackTemplate.get_latest(name='feedback')
        feedback_doc = {'percentage_written': 100, 'frequency': {'subbands': [152], 'central_frequencies': [33593750.0], 'channel_width': 3051.7578125, 'channels_per_subband': 64}, 'time': {'start_time': '2013-02-16T17:00:00', 'duration': 5.02732992172, 'sample_width': 2.00278016}, 'antennas': {'set': 'HBA_DUAL', 'fields': [{'type': 'HBA', 'field': 'HBA0', 'station': 'CS001'}, {'type': 'HBA', 'field': 'HBA1', 'station': 'CS001'}]}, 'target': {'pointing': {'angle1': 0, 'angle2': 0, 'direction_type': 'J2000', "target": "target1"}, 'coherent': True}, 'samples': {'polarisations': ['XX', 'XY', 'YX', 'YY'], 'type': 'float', 'bits': 32, 'writer': 'standard', 'writer_version': '2.2.0', 'complex': True}, 'files': ['stokes/SAP0/CS003HBA1/L773569_SAP000_B005_S0_P000_bf.h5', 'stokes/SAP0/RS106HBA/L773569_SAP000_B046_S0_P000_bf.h5']}
        feedback_doc = feedback_template.add_defaults_to_json_object_for_schema(feedback_doc)

        for dp in subtask.output_dataproducts.all():
            dp.feedback_template = feedback_template
            dp.feedback_doc = feedback_doc
            dp.save()
        set_subtask_state_following_allowed_transitions(subtask, 'finished')

        # wipe sip cache
        create_sip_representation_for_dataproduct.cache_clear()

        # generate sip
        sip = generate_sip_for_dataproduct(subtask.output_dataproducts.first())

        # assert that there are subarrayPointings with measurementType Target and with Calibration on the SIP
        sipxml = sip.get_prettyxml()
        self.assertIn(str('<measurementType>Target</measurementType>'), sipxml)
        self.assertIn(str('<measurementType>Calibration</measurementType>'), sipxml)

    def test_sip_generate_with_related_dataproducts(self):
        """
        Test if a SIP is generated successfully when it is pipeline output.
        Assert that related intermediate dataproducts from the observation are referenced in the SIP.
        Produce a bunch of SIPs with a larger number of related dataproducts to make sure this performs well enough
        and finishes within test time.
        """

        # create obs subtask
        subtask_template = models.SubtaskTemplate.get_latest(name='observation control')
        specifications_doc = get_default_json_object_for_schema(subtask_template.schema)
        specifications_doc['stations']['filter'] = "HBA_210_250"
        feedback_template = models.DataproductFeedbackTemplate.get_latest(name='feedback')
        feedback_doc = {'percentage_written': 100, 'frequency': {'subbands': [156], 'central_frequencies': [33593750.0], 'channel_width': 6103.515625, 'channels_per_subband': 32}, 'time': {'start_time': '2013-02-16T17:00:00', 'duration': 5.02732992172, 'sample_width': 2.00278016}, 'antennas': {'set': 'HBA_DUAL', 'fields': [{'type': 'HBA', 'field': 'HBA0', 'station': 'CS001'}, {'type': 'HBA', 'field': 'HBA1', 'station': 'CS001'}]}, 'target': {'pointing': {'angle1': 0, 'angle2': 0, 'direction_type': 'J2000', "target": "target1"}}, 'samples': {'polarisations': ['XX', 'XY', 'YX', 'YY'], 'type': 'float', 'bits': 32, 'writer': 'standard', 'writer_version': '2.2.0', 'complex': True}, '$schema': 'http://127.0.0.1:8001/api/schemas/dataproductfeedbacktemplate/feedback/1#'}
        feedback_doc = feedback_template.add_defaults_to_json_object_for_schema(feedback_doc)
        for dp in specifications_doc['stations']['digital_pointings']:
            dp['subbands'] = list(range(8))

        # Create SubTask(output)
        obs_subtask_data = Subtask_test_data(subtask_template=subtask_template, specifications_doc=specifications_doc, task_blueprint=models.TaskBlueprint.objects.create(**TaskBlueprint_test_data()))
        obs_subtask:models.Subtask = models.Subtask.objects.create(**obs_subtask_data)
        obs_subtask_output = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=obs_subtask))

        # also create an obsolete obs subtask (to verify that that and its dataproducts get ignored)
        obs_subtask_data['primary'] = False
        obs_subtask_data['obsolete_since'] = datetime.utcnow()
        obsolete_subtask: models.Subtask = models.Subtask.objects.create(**obs_subtask_data)
        self.assertTrue(obsolete_subtask.is_obsolete)
        obsolete_subtask_output = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=obsolete_subtask))

        # Create SAP
        sap_template = models.SAPTemplate.get_latest(name="SAP")
        specifications_doc = get_default_json_object_for_schema(sap_template.schema)
        sap = models.SAP.objects.create(specifications_doc=specifications_doc, specifications_template=sap_template)
        sap.save()

        dp_spec_template = models.DataproductSpecificationsTemplate.get_latest(name='visibilities')

        # Create a bunch of related Dataproducts (observation output, pipeline input)
        self.assertEqual(obs_subtask.outputs.count(), 1)
        self.assertEqual(obsolete_subtask.outputs.count(), 1)
        for st in [obs_subtask, obsolete_subtask]:
            for i in range(15):
                dataproduct: models.Dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(filename='my_related_dataproduct_%s' % i + ('_obsolete' if st == obsolete_subtask else ''),
                                                                                                            feedback_doc=feedback_doc,
                                                                                                            feedback_template=feedback_template,
                                                                                                            producer=st.outputs.first(),
                                                                                                            dataformat=models.Dataformat.objects.get(value="Beamformed"),
                                                                                                            datatype=models.Datatype.objects.get(value="time series"),
                                                                                                            specifications_template=dp_spec_template))

                dataproduct.sap = sap
                dataproduct.specifications_doc.update({'sap': "foobar", 'subband': 1})
                dataproduct.save()

        set_subtask_state_following_allowed_transitions(obs_subtask, 'finished')
        set_subtask_state_following_allowed_transitions(obsolete_subtask, 'finished')

        # create pipeline subtask
        subtask_template = models.SubtaskTemplate.get_latest(name='preprocessing pipeline')
        specifications_doc = get_default_json_object_for_schema(subtask_template.schema)
        feedback_template = models.DataproductFeedbackTemplate.get_latest(name='feedback')
        feedback_doc = {'percentage_written': 100, 'frequency': {'subbands': [156], 'central_frequencies': [33593750.0], 'channel_width': 6103.515625, 'channels_per_subband': 32}, 'time': {'start_time': '2013-02-16T17:00:00', 'duration': 5.02732992172, 'sample_width': 2.00278016}, 'antennas': {'set': 'HBA_DUAL', 'fields': [{'type': 'HBA', 'field': 'HBA0', 'station': 'CS001'}, {'type': 'HBA', 'field': 'HBA1', 'station': 'CS001'}]}, 'target': {'pointing': {'angle1': 0, 'angle2': 0, 'direction_type': 'J2000', "target": "target1"}}, 'samples': {'polarisations': ['XX', 'XY', 'YX', 'YY'], 'type': 'float', 'bits': 32, 'writer': 'standard', 'writer_version': '2.2.0', 'complex': True} }
        feedback_doc = feedback_template.add_defaults_to_json_object_for_schema(feedback_doc)

        # Create SubTask(in/output)
        subtask_data = Subtask_test_data(subtask_template=subtask_template, specifications_doc=specifications_doc, task_blueprint=models.TaskBlueprint.objects.create(**TaskBlueprint_test_data()))
        subtask:models.Subtask = models.Subtask.objects.create(**subtask_data)
        pip_id = subtask.id
        for sto in [obs_subtask_output, obsolete_subtask_output]:
            subtask_input = models.SubtaskInput.objects.create(**SubtaskInput_test_data(subtask=subtask, producer=sto))
            subtask_input.dataproducts.set(sto.dataproducts.all())
        subtask_output = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=subtask))

        # create a bunch of main dataproducts (pipeline output)
        main_dataproducts = []
        for i in range(10):
            dataproduct: models.Dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(filename='my_main_dataproduct_%s' % i,
                                                                                                        feedback_doc=feedback_doc,
                                                                                                        feedback_template=feedback_template,
                                                                                                        producer=subtask_output,
                                                                                                        specifications_template=dp_spec_template,
                                                                                                        dataformat=models.Dataformat.objects.get(value="Beamformed"),
                                                                                                        datatype=models.Datatype.objects.get(value="time series")))

            dataproduct.sap = sap
            dataproduct.specifications_doc.update({'sap': "foobar", 'subband': 1})
            dataproduct.save()
            main_dataproducts.append(dataproduct)

        set_subtask_state_following_allowed_transitions(subtask, 'finished')

        # wipe sip cache
        create_sip_representation_for_dataproduct.cache_clear()

        # create their SIPs (separate loop since we needed to clear the cache in between):
        for i in range(10):
            dataproduct = main_dataproducts[i]
            prettyxml = get_sip_xml_for_dataproduct(dataproduct)
            self.assertIn(str('<fileName>my_related_dataproduct_13'), prettyxml)
            self.assertNotIn(str('<fileName>my_related_dataproduct_13_obsolete'), prettyxml)
            self.assertIn(str(f'<fileName>my_main_dataproduct_{i}'), prettyxml)
            self.assertNotIn(str(f'<fileName>my_main_dataproduct_{i+1}'), prettyxml)
            if i == 5:   # just do this once since it is expensive:
                etree = ET.fromstring(prettyxml)
                sip_obs_id = int(etree.find('pipelineRun').find('observationId').find('identifier').text)
                self.assertEqual(pip_id, sip_obs_id)

class L2stationspecAdapterTest(unittest.TestCase):

    @staticmethod
    def get_default_specifications():
        subtask_template = models.SubtaskTemplate.get_latest(name='observation control')
        return get_default_json_object_for_schema(subtask_template.schema)

    def create_subtask(self, specifications_doc=None):
        if specifications_doc == None:
            specifications_doc = {}
        subtask_template = models.SubtaskTemplate.get_latest(name='observation control')
        specifications_doc = add_defaults_to_json_object_for_schema(specifications_doc, subtask_template.schema)
        subtask_data = Subtask_test_data(subtask_template=subtask_template, specifications_doc=specifications_doc, task_blueprint=models.TaskBlueprint.objects.create(**TaskBlueprint_test_data()))
        subtask:models.Subtask = models.Subtask.objects.create(**subtask_data)

        subtask_output = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=subtask))
        dataproduct:models.Dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=subtask_output))
        return subtask

    def test_basic(self):
        # create obs subtask
        specifications_doc = L2stationspecAdapterTest.get_default_specifications()

        # explicitly specify which fields we want, since we verify the count later on
        specifications_doc['stations']['antenna_set'] = "HBA_DUAL"
        specifications_doc['stations']['station_list'] = ["CS001", "RS205", "DE601"]

        obs_subtask = ObservationParsetAdapterTest.create_subtask(specifications_doc)

        # generate l2stationspecs
        l2stationspecs = convert_to_l2stationspecs_dict(obs_subtask)

        self.assertEqual(3, len(l2stationspecs['stations']))  # one spec per station
        self.assertEqual(4, sum([len(station['antenna_fields']) for station in l2stationspecs['stations']]))  # we have 4 fields: CS001HBA0, CS001HBA1, RS205HBA, DE601HBA
        self.assertEqual(set(obs_subtask.specifications_doc['stations']['station_list']), set([station['station'] for station in l2stationspecs['stations']]))
        self.assertEqual(obs_subtask.id, l2stationspecs['stations'][0]['antenna_fields'][0]['observation_id'])
        # todo: add more checks


if __name__ == "__main__":
    os.environ['TZ'] = 'UTC'
    unittest.main()
