#!/usr/bin/env python3

# Copyright (C) 2018    ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

# $Id:  $

import os
import unittest
import requests

import logging
logger = logging.getLogger('lofar.'+__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

from lofar.common.test_utils import exit_with_skipped_code_if_skip_integration_tests
exit_with_skipped_code_if_skip_integration_tests()

from lofar.common.json_utils import validate_json_against_schema

# Do Mandatory setup step:
# use setup/teardown magic for tmss test database, ldap server and django server
# (ignore pycharm unused import statement, python unittests does use at RunTime the tmss_test_environment_unittest_setup module)
from lofar.sas.tmss.test.tmss_test_environment_unittest_setup import *
tmss_test_env.populate_schemas_and_connectors()

from lofar.sas.tmss.test.tmss_test_data_django_models import *

from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.exceptions import SubtaskInvalidStateException
from lofar.sas.tmss.tmss.tmssapp.adapters.feedback import *
from lofar.parameterset import parameterset
from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions


class FeedbackAdapterTest(unittest.TestCase):

    feedback_preprocessing_pipeline_complete = """LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].centralFrequency=33593750.0
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].channelWidth=6103.515625
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].channelsPerSubband=32
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].dopplerCorrection=false
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].duration=5.02732992172
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].fileFormat=AIPS++/CASA
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].filename=L99307_SB000_uv.dppp.MS
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].integrationInterval=2.00278016
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].location=locus001:/data/L99307
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].percentageWritten=100
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].returncode=0
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].size=15606123742
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].startTime=2013-02-16T17:00:00.000
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].stationSubband=0
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].storageWriter=CASA
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].storageWriterVersion=2.2.0
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[0].subband=0
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].centralFrequency=33789062.5
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].channelWidth=6103.515625
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].channelsPerSubband=32
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].dopplerCorrection=false
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].duration=5.02513194084
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].fileFormat=AIPS++/CASA
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].filename=L99307_SB001_uv.dppp.MS
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].integrationInterval=2.00278016
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].location=locus003:/data/L99307
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].percentageWritten=100
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].returncode=0
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].size=15606156518
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].startTime=2013-02-16T17:00:00.000
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].stationSubband=0
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].storageWriter=CASA
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].storageWriterVersion=2.2.0
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].subband=1
Observation.DataProducts.nrOfOutput_Beamformed_=0
Observation.DataProducts.nrOfOutput_Correlated_=2
_isCobalt=T
feedback_version=03.01.00
"""

    feedback_pulsar_pipeline_cv_analysis = """
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].CoherentStokesBeam.Offset.angle1=0.0
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].CoherentStokesBeam.Offset.angle2=0.0
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].CoherentStokesBeam.Offset.coordType=RA-DEC
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].CoherentStokesBeam.Offset.equinox=J2000
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].CoherentStokesBeam.Pointing.angle1=2.15980858832
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].CoherentStokesBeam.Pointing.angle2=1.30000703891
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].CoherentStokesBeam.Pointing.coordType=RA-DEC
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].CoherentStokesBeam.Pointing.equinox=J2000
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].CoherentStokesBeam.SAP=0
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].CoherentStokesBeam.TAB=0
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].CoherentStokesBeam.centralFrequencies=[110546875.000000,110742187.500000,110937500.000000,111132812.500000,111328125.000000,111523437.500000,111718750.000000,111914062.500000,112109375.000000,112304687.500000,112500000.000000,112695312.500000,112890625.000000,113085937.500000,113281250.000000,113476562.500000,113671875.000000,113867187.500000,114062500.000000,114257812.500000]
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].CoherentStokesBeam.channelWidth=195.3125
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].CoherentStokesBeam.channelsPerSubband=1
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].CoherentStokesBeam.dispersionMeasure=0.0
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].CoherentStokesBeam.nrSubbands=20
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].CoherentStokesBeam.samplingTime=5.12e-06
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].CoherentStokesBeam.stationSubbands=[54,55,56,57,58,59,60,61,62,63,64,65,66,67,68,69,70,71,72,73]
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].CoherentStokesBeam.stokes=[Xre,Xim,Yre,Yim]
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].datatype=ComplexVoltages
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].fileContent=[rawvoltages/SAP0/BEAM0/L808268_sap000_beam0000_part00.log, rawvoltages/SAP0/BEAM0/B0809+74_L808268_SAP0_BEAM0_PART0.ar, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S0_P000_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S1_P000_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S3_P000_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S2_P000_bf.h5, rawvoltages/SAP0/BEAM0/raw-8bit/L808268_SAP000_B000_S0_P000_bf.h5, rawvoltages/SAP0/BEAM0/raw-8bit/L808268_SAP000_B000_S1_P000_bf.h5, rawvoltages/SAP0/BEAM0/raw-8bit/L808268_SAP000_B000_S3_P000_bf.h5, rawvoltages/SAP0/BEAM0/raw-8bit/L808268_SAP000_B000_S2_P000_bf.h5, rawvoltages/SAP0/BEAM0/raw-8bit/L808268_SAP000_B000_S3_P000_bf.raw, rawvoltages/SAP0/BEAM0/raw-8bit/L808268_SAP000_B000_S0_P000_bf.raw, rawvoltages/SAP0/BEAM0/raw-8bit/L808268_SAP000_B000_S1_P000_bf.raw, rawvoltages/SAP0/BEAM0/raw-8bit/L808268_SAP000_B000_S2_P000_bf.raw, 0809+74.par, L808268_sap000_beam0000_part00.log]
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].fileFormat=PULP
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].filename=L808292_SAP000_B000_P000_bf.tar
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].location=cpu12:/data/projects/PipelineTests/L808292/pulp/cs
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].percentageWritten=100
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].size=4697600000
"""

    feedback_pulsar_pipeline_cv_summary = """
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[1].datatype=SummaryComplexVoltages
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[1].fileContent=[L808268_sap000_beam0000_part05.log, L808268_sap000_beam0000_part11.log, L808268_sap000_beam0000_part02.log, L808268_sap000_beam0000_part01.log, L808268_sap000_beam0000_part14.log, L808268_sap000_beam0000_part12.log, L808268_sap000_beam0000_part04.log, L808268_sap000_beam0000_part15.log, L808268_sap000_beam0000_part18.log, L808268_sap000_beam0000_part17.log, L808268_sap000_beam0000_part00.log, L808268_sap000_beam0000_part06.log, L808268_sap000_beam0000_part08.log, L808268_sap000_beam0000_part19.log, L808268_sap000_beam0000_part09.log, L808268_sap000_beam0000_part16.log, L808268_sap000_beam0000_part10.log, L808268_sap000_beam0000_part13.log, L808268_sap000_beam0000_part03.log, L808268_sap000_beam0000_part07.log, rawvoltages/SAP0/BEAM0/L808268_sap000_beam0000_part05.log, rawvoltages/SAP0/BEAM0/L808268_sap000_beam0000_part11.log, rawvoltages/SAP0/BEAM0/L808268_sap000_beam0000_part02.log, rawvoltages/SAP0/BEAM0/L808268_sap000_beam0000_part01.log, rawvoltages/SAP0/BEAM0/L808268_sap000_beam0000_part14.log, rawvoltages/SAP0/BEAM0/L808268_sap000_beam0000_part12.log, rawvoltages/SAP0/BEAM0/L808268_sap000_beam0000_part04.log, rawvoltages/SAP0/BEAM0/L808268_sap000_beam0000_part15.log, rawvoltages/SAP0/BEAM0/L808268_sap000_beam0000_part18.log, rawvoltages/SAP0/BEAM0/L808268_sap000_beam0000_part17.log, rawvoltages/SAP0/BEAM0/L808268_sap000_beam0000_part00.log, rawvoltages/SAP0/BEAM0/L808268_sap000_beam0000_part06.log, rawvoltages/SAP0/BEAM0/L808268_sap000_beam0000_part08.log, rawvoltages/SAP0/BEAM0/L808268_sap000_beam0000_part19.log, rawvoltages/SAP0/BEAM0/L808268_sap000_beam0000_part09.log, rawvoltages/SAP0/BEAM0/L808268_sap000_beam0000_part16.log, rawvoltages/SAP0/BEAM0/L808268_sap000_beam0000_part10.log, rawvoltages/SAP0/BEAM0/L808268_sap000_beam0000_part13.log, rawvoltages/SAP0/BEAM0/L808268_sap000_beam0000_part03.log, rawvoltages/SAP0/BEAM0/L808268_sap000_beam0000_part07.log, beam_process_node.txt, rawvoltages/SAP0/BEAM0/rm_spectrum.txt, rawvoltages/SAP0/BEAM0/B0809+74_L808268_SAP0_BEAM0_SFTd.ps, rawvoltages/SAP0/BEAM0/B0809+74_L808268_SAP0_BEAM0_j.ps, rawvoltages/SAP0/BEAM0/B0809+74_L808268_SAP0_BEAM0_pdmp.ps, rawvoltages/SAP0/BEAM0/B0809+74_L808268_SAP0_BEAM0.negRM.ps, rawvoltages/SAP0/BEAM0/B0809+74_L808268_SAP0_BEAM0.posRM.ps, rawvoltages/SAP0/BEAM0/B0809+74_L808268_SAP0_BEAM0_YFpd.ps, rawvoltages/SAP0/BEAM0/B0809+74_L808268_SAP0_BEAM0_GTpdf8.ps, status.png, status.th.png, rawvoltages/SAP0/BEAM0/B0809+74_L808268_SAP0_BEAM0_diag_pdmp.png, rawvoltages/SAP0/BEAM0/B0809+74_L808268_SAP0_BEAM0_diag.png, rawvoltages/SAP0/BEAM0/B0809+74_L808268_SAP0_BEAM0.ar, rawvoltages/SAP0/BEAM0/B0809+74_L808268_SAP0_BEAM0.fscr.AR, rawvoltages/SAP0/BEAM0/B0809+74_L808268_SAP0_BEAM0.paz.fscr.pdmp.AR, rawvoltages/SAP0/BEAM0/B0809+74_L808268_SAP0_BEAM0.paz.fscr.AR, rawvoltages/SAP0/BEAM0/pdmp.per, rawvoltages/SAP0/BEAM0/B0809+74_L808268_SAP0_BEAM0_pdmp.per, rawvoltages/SAP0/BEAM0/B0809+74_L808268_SAP0_BEAM0_pdmp.posn, rawvoltages/SAP0/BEAM0/pdmp.posn, rawvoltages/SAP0/BEAM0/B0809+74_L808268_SAP0_BEAM0_pdmp.ps, rawvoltages/SAP0/BEAM0/B0809+74_L808268_SAP0_BEAM0_diag_pdmp.png, rawvoltages/SAP0/BEAM0/B0809+74_L808268_SAP0_BEAM0.paz.fscr.pdmp.AR, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S3_P018_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S0_P000_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S0_P001_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S2_P014_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S2_P009_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S2_P005_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S3_P002_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S1_P000_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S3_P000_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S1_P019_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S1_P007_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S0_P014_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S3_P009_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S3_P012_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S0_P018_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S1_P008_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S0_P012_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S2_P007_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S2_P011_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S1_P003_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S1_P009_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S1_P010_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S1_P006_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S3_P006_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S2_P016_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S0_P013_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S0_P016_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S0_P011_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S0_P015_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S2_P001_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S1_P004_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S2_P010_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S0_P004_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S0_P017_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S2_P013_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S1_P013_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S1_P002_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S1_P005_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S2_P018_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S2_P017_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S1_P016_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S1_P012_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S3_P004_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S3_P015_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S3_P017_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S1_P018_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S0_P006_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S0_P005_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S0_P019_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S0_P007_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S1_P015_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S2_P006_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S2_P012_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S3_P008_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S2_P015_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S1_P014_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S3_P007_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S0_P010_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S2_P008_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S3_P014_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S1_P017_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S2_P002_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S0_P002_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S3_P011_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S2_P000_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S3_P003_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S2_P003_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S0_P009_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S3_P013_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S3_P019_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S1_P001_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S2_P004_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S3_P001_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S1_P011_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S3_P010_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S2_P019_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S0_P008_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S0_P003_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S3_P016_bf.h5, rawvoltages/SAP0/BEAM0/L808268_SAP000_B000_S3_P005_bf.h5, rawvoltages/SAP0/BEAM0/StokeStats.out, rawvoltages/SAP0/BEAM0/rmfit_results.out, rawvoltages/SAP0/BEAM0/all_bestRMs.out, 0809+74.par, rawvoltages/SAP0/BEAM0/pdmp.per, rawvoltages/SAP0/BEAM0/B0809+74_L808268_SAP0_BEAM0_pdmp.per, rawvoltages/SAP0/BEAM0/B0809+74_L808268_SAP0_BEAM0_pdmp.posn, rawvoltages/SAP0/BEAM0/pdmp.posn, rawvoltages/SAP0/BEAM0/StokeStats.out, rawvoltages/SAP0/BEAM0/rmfit_results.out, rawvoltages/SAP0/BEAM0/all_bestRMs.out, beam_process_node.txt, rawvoltages/SAP0/BEAM0/rm_spectrum.txt, L808268_summaryCV.log]
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[1].fileFormat=PULP
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[1].filename=L808292_summaryCV.tar
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[1].location=CEP4:/data//projects/PipelineTests/L808292/pulp/cs
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[1].percentageWritten=100
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[1].size=168908800
"""
    feedback_pulsar_pipeline_fe_analysis = """
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].FlysEyeBeam.SAP=0
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].FlysEyeBeam.TAB=0
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].FlysEyeBeam.antennaFieldName=HBA0
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].FlysEyeBeam.centralFrequencies=[115039062.500000,115234375.000000,115429687.500000,115625000.000000,115820312.500000,116015625.000000,116210937.500000,116406250.000000,116601562.500000,116796875.000000,116992187.500000,117187500.000000,117382812.500000,117578125.000000,117773437.500000,117968750.000000,118164062.500000,118359375.000000,118554687.500000,118750000.000000,118945312.500000,119140625.000000,119335937.500000,119531250.000000,119726562.500000,119921875.000000,120117187.500000,120312500.000000,120507812.500000,120703125.000000,120898437.500000,121093750.000000,121289062.500000,121484375.000000,121679687.500000,121875000.000000,122070312.500000,122265625.000000,122460937.500000,122656250.000000,122851562.500000,123046875.000000,123242187.500000,123437500.000000,123632812.500000,123828125.000000,124023437.500000,124218750.000000,124414062.500000,124609375.000000,124804687.500000,125000000.000000,125195312.500000,125390625.000000,125585937.500000,125781250.000000,125976562.500000,126171875.000000,126367187.500000,126562500.000000,126757812.500000,126953125.000000,127148437.500000,127343750.000000,127539062.500000,127734375.000000,127929687.500000,128125000.000000,128320312.500000,128515625.000000,128710937.500000,128906250.000000,129101562.500000,129296875.000000,129492187.500000,129687500.000000,129882812.500000,130078125.000000,130273437.500000,130468750.000000,130664062.500000,130859375.000000,131054687.500000,131250000.000000,131445312.500000,131640625.000000,131835937.500000,132031250.000000,132226562.500000,132421875.000000,132617187.500000,132812500.000000,133007812.500000,133203125.000000,133398437.500000,133593750.000000,133789062.500000,133984375.000000,134179687.500000,134375000.000000,134570312.500000,134765625.000000,134960937.500000,135156250.000000,135351562.500000,135546875.000000,135742187.500000,135937500.000000,136132812.500000,136328125.000000,136523437.500000,136718750.000000,136914062.500000,137109375.000000,137304687.500000,137500000.000000,137695312.500000,137890625.000000,138085937.500000,138281250.000000,138476562.500000,138671875.000000,138867187.500000,139062500.000000,139257812.500000,139453125.000000,139648437.500000,139843750.000000,140039062.500000,140234375.000000,140429687.500000,140625000.000000,140820312.500000,141015625.000000,141210937.500000,141406250.000000,141601562.500000,141796875.000000,141992187.500000,142187500.000000,142382812.500000,142578125.000000,142773437.500000,142968750.000000,143164062.500000,143359375.000000,143554687.500000,143750000.000000,143945312.500000,144140625.000000,144335937.500000,144531250.000000,144726562.500000,144921875.000000,145117187.500000,145312500.000000,145507812.500000,145703125.000000,145898437.500000,146093750.000000,146289062.500000,146484375.000000,146679687.500000,146875000.000000,147070312.500000,147265625.000000,147460937.500000,147656250.000000,147851562.500000,148046875.000000,148242187.500000,148437500.000000,148632812.500000,148828125.000000,149023437.500000,149218750.000000,149414062.500000,149609375.000000,149804687.500000,150000000.000000,150195312.500000,150390625.000000,150585937.500000,150781250.000000,150976562.500000,151171875.000000,151367187.500000,151562500.000000,151757812.500000,151953125.000000,152148437.500000,152343750.000000,152539062.500000,152734375.000000,152929687.500000,153125000.000000,153320312.500000,153515625.000000,153710937.500000,153906250.000000,154101562.500000,154296875.000000,154492187.500000,154687500.000000,154882812.500000,155078125.000000,155273437.500000,155468750.000000,155664062.500000,155859375.000000,156054687.500000,156250000.000000,156445312.500000,156640625.000000,156835937.500000,157031250.000000,157226562.500000,157421875.000000,157617187.500000,157812500.000000,158007812.500000,158203125.000000,158398437.500000,158593750.000000,158789062.500000,158984375.000000,159179687.500000,159375000.000000,159570312.500000,159765625.000000,159960937.500000,160156250.000000,160351562.500000,160546875.000000,160742187.500000,160937500.000000,161132812.500000,161328125.000000,161523437.500000,161718750.000000,161914062.500000,162109375.000000,162304687.500000,162500000.000000]
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].FlysEyeBeam.channelWidth=12.20703125
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].FlysEyeBeam.channelsPerSubband=16
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].FlysEyeBeam.dispersionMeasure=0.0
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].FlysEyeBeam.nrSubbands=244
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].FlysEyeBeam.samplingTime=0.01048576
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].FlysEyeBeam.stationName=CS001
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].FlysEyeBeam.stationSubbands=[77,78,79,80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95,96,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,112,113,114,115,116,117,118,119,120,121,122,123,124,125,126,127,128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143,144,145,146,147,148,149,150,151,152,153,154,155,156,157,158,159,160,161,162,163,164,165,166,167,168,169,170,171,172,173,174,175,176,177,178,179,180,181,182,183,184,185,186,187,188,189,190,191,192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223,224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239,240,241,242,243,244,245,246,247,248,249,250,251,252,253,254,255,256,257,258,259,260,261,262,263,264,265,266,267,268,269,270,271,272,273,274,275,276,277,278,279,280,281,282,283,284,285,286,287,288,289,290,291,292,293,294,295,296,297,298,299,300,301,302,303,304,305,306,307,308,309,310,311,312,313,314,315,316,317,318,319,320]
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].FlysEyeBeam.stokes=[I]
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].datatype=CoherentStokes
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].fileContent=[stokes/SAP0/CS001HBA0/B0809+74_L773569_SAP0_CS001HBA0_PSR_B0809+74.pfd.pdf, stokes/SAP0/CS001HBA0/B0809+74_L773569_SAP0_CS001HBA0_nomask_PSR_B0809+74.pfd.pdf, stokes/SAP0/CS001HBA0/B0809+74_L773569_SAP0_CS001HBA0_nomask_PSR_B0809+74.pfd.ps, stokes/SAP0/CS001HBA0/B0809+74_L773569_SAP0_CS001HBA0_GTpdf4.ps, stokes/SAP0/CS001HBA0/B0809+74_L773569_SAP0_CS001HBA0_DFTpd.ps, stokes/SAP0/CS001HBA0/L773569_SAP0_CS001HBA0_rfifind.ps, stokes/SAP0/CS001HBA0/B0809+74_L773569_SAP0_CS001HBA0_PSR_B0809+74.pfd.ps, stokes/SAP0/CS001HBA0/B0809+74_L773569_SAP0_CS001HBA0_j.ps, stokes/SAP0/CS001HBA0/B0809+74_L773569_SAP0_CS001HBA0_YFpd.ps, stokes/SAP0/CS001HBA0/B0809+74_L773569_SAP0_CS001HBA0_nomask_PSR_B0809+74.pfd, stokes/SAP0/CS001HBA0/B0809+74_L773569_SAP0_CS001HBA0_PSR_B0809+74.pfd, stokes/SAP0/CS001HBA0/B0809+74_L773569_SAP0_CS001HBA0_nomask_PSR_B0809+74.pfd.bestprof, stokes/SAP0/CS001HBA0/B0809+74_L773569_SAP0_CS001HBA0_PSR_B0809+74.pfd.bestprof, stokes/SAP0/CS001HBA0/B0809+74_L773569_SAP0_CS001HBA0_nomask_PSR_B0809+74.pfd.polycos, stokes/SAP0/CS001HBA0/B0809+74_L773569_SAP0_CS001HBA0_PSR_B0809+74.pfd.polycos, stokes/SAP0/CS001HBA0/L773569_SAP0_CS001HBA0_rfifind.inf, stokes/SAP0/CS001HBA0/L773569_SAP0_CS001HBA0.rfirep, stokes/SAP0/CS001HBA0/B0809+74_L773569_SAP0_CS001HBA0_nomask_PSR_B0809+74.pfd.th.png, stokes/SAP0/CS001HBA0/B0809+74_L773569_SAP0_CS001HBA0_PSR_B0809+74.pfd.th.png, stokes/SAP0/CS001HBA0/L773569_SAP0_CS001HBA0.sp.png, stokes/SAP0/CS001HBA0/B0809+74_L773569_SAP0_CS001HBA0_diag.png, stokes/SAP0/CS001HBA0/B0809+74_L773569_SAP0_CS001HBA0_nomask_PSR_B0809+74.pfd.png, stokes/SAP0/CS001HBA0/B0809+74_L773569_SAP0_CS001HBA0_PSR_B0809+74.pfd.png, stokes/SAP0/CS001HBA0/B0809+74_L773569_SAP0_CS001HBA0.ar, stokes/SAP0/CS001HBA0/B0809+74_L773569_SAP0_CS001HBA0.fscr.AR, stokes/SAP0/CS001HBA0/B0809+74_L773569_SAP0_CS001HBA0.paz.fscr.AR, stokes/SAP0/CS001HBA0/L773569_SAP0_CS001HBA0_rfifind.mask, stokes/SAP0/CS001HBA0/L773569_SAP0_CS001HBA0_rfifind.ps, stokes/SAP0/CS001HBA0/L773569_SAP0_CS001HBA0_rfifind.inf, stokes/SAP0/CS001HBA0/L773569_SAP0_CS001HBA0_rfifind.rfi, stokes/SAP0/CS001HBA0/L773569_SAP0_CS001HBA0_rfifind.bytemask, stokes/SAP0/CS001HBA0/L773569_SAP0_CS001HBA0_rfifind.stats, stokes/SAP0/CS001HBA0/L773569_SAP000_B000_S0_P000_bf.h5, stokes/SAP0/CS001HBA0/L773569_SAP0_CS001HBA0.fits, 0809+74.par, L773569_sap000_beam0000.log]
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].fileFormat=PULP
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].filename=L773575_SAP000_B000_P000_bf.tar
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].location=cpu21:/data/projects/FE_monitoring/L773575/pulp/cs
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].percentageWritten=100
LOFAR.ObsSW.Observation.DataProducts.Output_Pulsar_[0].size=308531200
"""

    feedback_preprocessing_pipeline_incomplete = """LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].centralFrequency=33789062.5
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].channelWidth=6103.515625
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].channelsPerSubband=32
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].dopplerCorrection=false
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].duration=5.02513194084
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].fileFormat=AIPS++/CASA
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].filename=L99307_SB001_uv.dppp.MS
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].integrationInterval=2.00278016
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].location=locus003:/data/L99307
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].percentageWritten=100
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].returncode=0
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].size=15606156518
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].startTime=2013-02-16T17:00:00.000
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].stationSubband=0
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].storageWriter=CASA
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].storageWriterVersion=2.2.0
LOFAR.ObsSW.Observation.DataProducts.Output_Correlated_[1].subband=1
Observation.DataProducts.nrOfOutput_Beamformed_=0
Observation.DataProducts.nrOfOutput_Correlated_=2
_isCobalt=T
feedback_version=03.01.00"""

    feedback_obs_correlator_complete = """
Observation.Correlator.channelWidth=3051.7578125
Observation.Correlator.channelsPerSubband=64
Observation.Correlator.integrationInterval=1.00663296
Observation.DataProducts.Output_Correlated_[0].Pointing.angle1=2.153736
Observation.DataProducts.Output_Correlated_[0].Pointing.angle2=0.841554
Observation.DataProducts.Output_Correlated_[0].Pointing.directionType=J2000
Observation.DataProducts.Output_Correlated_[0].Pointing.target=my_source
Observation.DataProducts.Output_Correlated_[0].SAP=0
Observation.DataProducts.Output_Correlated_[0].antennaFields=[CS001LBA,CS002LBA,CS003LBA,CS004LBA,CS005LBA,CS006LBA,CS007LBA,CS011LBA,CS013LBA,CS017LBA,CS021LBA,CS024LBA,CS026LBA,CS028LBA,CS030LBA,CS031LBA,CS032LBA,CS101LBA,CS103LBA,CS201LBA,CS301LBA,CS302LBA,CS401LBA,CS501LBA,RS106LBA,RS205LBA,RS208LBA,RS210LBA,RS305LBA,RS306LBA,RS307LBA,RS310LBA,RS406LBA,RS407LBA,RS409LBA,RS503LBA,RS508LBA,RS509LBA]
Observation.DataProducts.Output_Correlated_[0].antennaSet=LBA_INNER
Observation.DataProducts.Output_Correlated_[0].centralFrequency=30468750.000000
Observation.DataProducts.Output_Correlated_[0].channelWidth=3051.757812
Observation.DataProducts.Output_Correlated_[0].channelsPerSubband=64
Observation.DataProducts.Output_Correlated_[0].dopplerCorrection=false
Observation.DataProducts.Output_Correlated_[0].duration=0
Observation.DataProducts.Output_Correlated_[0].fileFormat=AIPS++/CASA
Observation.DataProducts.Output_Correlated_[0].filename=L220133_SAP000_SB000_uv.MS
Observation.DataProducts.Output_Correlated_[0].integrationInterval=1.006633
Observation.DataProducts.Output_Correlated_[0].location=locus001:/data/L220133/
Observation.DataProducts.Output_Correlated_[0].percentageWritten=0
Observation.DataProducts.Output_Correlated_[0].size=0
Observation.DataProducts.Output_Correlated_[0].startTime=2014-04-18 15:02:00
Observation.DataProducts.Output_Correlated_[0].stationSubband=156
Observation.DataProducts.Output_Correlated_[0].storageWriter=LOFAR
Observation.DataProducts.Output_Correlated_[0].storageWriterVersion=3
Observation.DataProducts.Output_Correlated_[0].subband=0
Observation.DataProducts.Output_Correlated_[1].Pointing.angle1=2.153736
Observation.DataProducts.Output_Correlated_[1].Pointing.angle2=0.841554
Observation.DataProducts.Output_Correlated_[1].Pointing.directionType=J2000
Observation.DataProducts.Output_Correlated_[1].Pointing.target=my_source
Observation.DataProducts.Output_Correlated_[1].SAP=0
Observation.DataProducts.Output_Correlated_[1].antennaFields=[CS001LBA,CS002LBA,CS003LBA,CS004LBA,CS005LBA,CS006LBA,CS007LBA,CS011LBA,CS013LBA,CS017LBA,CS021LBA,CS024LBA,CS026LBA,CS028LBA,CS030LBA,CS031LBA,CS032LBA,CS101LBA,CS103LBA,CS201LBA,CS301LBA,CS302LBA,CS401LBA,CS501LBA,RS106LBA,RS205LBA,RS208LBA,RS210LBA,RS305LBA,RS306LBA,RS307LBA,RS310LBA,RS406LBA,RS407LBA,RS409LBA,RS503LBA,RS508LBA,RS509LBA]
Observation.DataProducts.Output_Correlated_[1].antennaSet=LBA_INNER
Observation.DataProducts.Output_Correlated_[1].centralFrequency=30664062.500000
Observation.DataProducts.Output_Correlated_[1].channelWidth=3051.757812
Observation.DataProducts.Output_Correlated_[1].channelsPerSubband=64
Observation.DataProducts.Output_Correlated_[1].dopplerCorrection=false
Observation.DataProducts.Output_Correlated_[1].duration=0
Observation.DataProducts.Output_Correlated_[1].fileFormat=AIPS++/CASA
Observation.DataProducts.Output_Correlated_[1].filename=L220133_SAP000_SB001_uv.MS
Observation.DataProducts.Output_Correlated_[1].integrationInterval=1.006633
Observation.DataProducts.Output_Correlated_[1].location=locus003:/data/L220133/
Observation.DataProducts.Output_Correlated_[1].percentageWritten=0
Observation.DataProducts.Output_Correlated_[1].size=0
Observation.DataProducts.Output_Correlated_[1].startTime=2014-04-18 15:02:00
Observation.DataProducts.Output_Correlated_[1].stationSubband=157
Observation.DataProducts.Output_Correlated_[1].storageWriter=LOFAR
Observation.DataProducts.Output_Correlated_[1].storageWriterVersion=3
Observation.DataProducts.Output_Correlated_[1].subband=1
Observation.DataProducts.nrOfOutput_Beamformed_=0
Observation.DataProducts.nrOfOutput_Correlated_=2
Observation.startTime=2013-02-16T17:00:00.000
Observation.stopTime=2013-02-16T17:01:00.000
_isCobalt=T
feedback_version=03.01.00
"""

    feedback_obs_beamformer_complete = """
Observation.CoherentStokes.antennaSet=HBA_DUAL
Observation.CoherentStokes.channelWidth=12207.03125
Observation.CoherentStokes.channelsPerSubband=16
Observation.CoherentStokes.rawSamplingTime=8.192e-05
Observation.CoherentStokes.samplingTime=0.0004915200000000001
Observation.CoherentStokes.stationList=[CS007,CS005,CS002,CS004,CS006,CS003]
Observation.CoherentStokes.stokes=I
Observation.CoherentStokes.timeDownsamplingFactor=6
Observation.DataProducts.Output_Beamformed_[0].CoherentStokesBeam[0].Offset.angle1=-0.929342
Observation.DataProducts.Output_Beamformed_[0].CoherentStokesBeam[0].Offset.angle2=-0.917577
Observation.DataProducts.Output_Beamformed_[0].CoherentStokesBeam[0].Offset.coordType=RA-DEC
Observation.DataProducts.Output_Beamformed_[0].CoherentStokesBeam[0].Offset.equinox=J2000
Observation.DataProducts.Output_Beamformed_[0].CoherentStokesBeam[0].Pointing.angle1=0.000000
Observation.DataProducts.Output_Beamformed_[0].CoherentStokesBeam[0].Pointing.angle2=0.035002
Observation.DataProducts.Output_Beamformed_[0].CoherentStokesBeam[0].Pointing.coordType=RA-DEC
Observation.DataProducts.Output_Beamformed_[0].CoherentStokesBeam[0].Pointing.directionType=J2000
Observation.DataProducts.Output_Beamformed_[0].CoherentStokesBeam[0].Pointing.equinox=J2000
Observation.DataProducts.Output_Beamformed_[0].CoherentStokesBeam[0].Pointing.target=my_source
Observation.DataProducts.Output_Beamformed_[0].CoherentStokesBeam[0].SAP=0
Observation.DataProducts.Output_Beamformed_[0].CoherentStokesBeam[0].TAB=0
Observation.DataProducts.Output_Beamformed_[0].CoherentStokesBeam[0].antennaFields=[CS002HBA0,CS002HBA1,CS003HBA0,CS003HBA1,CS004HBA0,CS004HBA1,CS005HBA0,CS005HBA1,CS006HBA0,CS006HBA1,CS007HBA0,CS007HBA1]
Observation.DataProducts.Output_Beamformed_[0].CoherentStokesBeam[0].antennaSet=HBA_DUAL
Observation.DataProducts.Output_Beamformed_[0].CoherentStokesBeam[0].centralFrequencies=[119531250.0000, 119726562.5000, 119921875.0000, 120117187.5000, 120312500.0000, 120507812.5000, 120703125.0000, 120898437.5000, 121093750.0000, 121289062.5000, 121484375.0000, 121679687.5000, 121875000.0000, 122070312.5000, 122265625.0000, 122460937.5000, 122656250.0000, 122851562.5000, 123046875.0000, 123242187.5000, 123437500.0000, 123632812.5000, 123828125.0000, 124023437.5000, 124218750.0000, 124414062.5000, 124609375.0000, 124804687.5000, 125000000.0000, 125195312.5000, 125390625.0000, 125585937.5000, 125781250.0000, 125976562.5000, 126171875.0000, 126367187.5000, 126562500.0000, 126757812.5000, 126953125.0000, 127148437.5000, 127343750.0000, 127539062.5000, 127734375.0000, 127929687.5000, 128125000.0000, 128320312.5000, 128515625.0000, 128710937.5000, 128906250.0000, 129101562.5000, 129296875.0000, 129492187.5000, 129687500.0000, 129882812.5000, 130078125.0000, 130273437.5000, 130468750.0000, 130664062.5000, 130859375.0000, 131054687.5000, 131250000.0000, 131445312.5000, 131640625.0000, 131835937.5000, 132031250.0000, 132226562.5000, 132421875.0000, 132617187.5000, 132812500.0000, 133007812.5000, 133203125.0000, 133398437.5000, 133593750.0000, 133789062.5000, 133984375.0000, 134179687.5000, 134375000.0000, 134570312.5000, 134765625.0000, 134960937.5000, 135156250.0000, 135351562.5000, 135546875.0000, 135742187.5000, 135937500.0000, 136132812.5000, 136328125.0000, 136523437.5000, 136718750.0000, 136914062.5000, 137109375.0000, 137304687.5000, 137500000.0000, 137695312.5000, 137890625.0000, 138085937.5000, 138281250.0000, 138476562.5000, 138671875.0000, 138867187.5000, 139062500.0000, 139257812.5000, 139453125.0000, 139648437.5000, 139843750.0000, 140039062.5000, 140234375.0000, 140429687.5000, 140625000.0000, 140820312.5000, 141015625.0000, 141210937.5000, 141406250.0000, 141601562.5000, 141796875.0000, 141992187.5000, 142187500.0000, 142382812.5000, 142578125.0000, 142773437.5000, 142968750.0000, 143164062.5000, 143359375.0000, 143554687.5000, 143750000.0000, 143945312.5000, 144140625.0000, 144335937.5000, 144531250.0000, 144726562.5000, 144921875.0000, 145117187.5000, 145312500.0000, 145507812.5000, 145703125.0000, 145898437.5000, 146093750.0000, 146289062.5000, 146484375.0000, 146679687.5000, 146875000.0000, 147070312.5000, 147265625.0000, 147460937.5000, 147656250.0000, 147851562.5000, 148046875.0000, 148242187.5000, 148437500.0000, 148632812.5000, 148828125.0000, 149023437.5000, 149218750.0000, 149414062.5000, 149609375.0000, 149804687.5000, 150000000.0000, 150195312.5000, 150390625.0000, 150585937.5000, 150781250.0000, 150976562.5000]
Observation.DataProducts.Output_Beamformed_[0].CoherentStokesBeam[0].channelWidth=12207.031250
Observation.DataProducts.Output_Beamformed_[0].CoherentStokesBeam[0].channelsPerSubband=16
Observation.DataProducts.Output_Beamformed_[0].CoherentStokesBeam[0].dispersionMeasure=0.000000
Observation.DataProducts.Output_Beamformed_[0].CoherentStokesBeam[0].nrSubbands=162
Observation.DataProducts.Output_Beamformed_[0].CoherentStokesBeam[0].pipeline=0
Observation.DataProducts.Output_Beamformed_[0].CoherentStokesBeam[0].pipelineTAB=0
Observation.DataProducts.Output_Beamformed_[0].CoherentStokesBeam[0].sampleNrBits=32
Observation.DataProducts.Output_Beamformed_[0].CoherentStokesBeam[0].sampleType=float
Observation.DataProducts.Output_Beamformed_[0].CoherentStokesBeam[0].samplingTime=0.000492
Observation.DataProducts.Output_Beamformed_[0].CoherentStokesBeam[0].stationSubbands=[100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 256, 257, 258, 259, 260, 261]
Observation.DataProducts.Output_Beamformed_[0].CoherentStokesBeam[0].stokes=[I]
Observation.DataProducts.Output_Beamformed_[0].beamformingSpectralLeakage=true
Observation.DataProducts.Output_Beamformed_[0].beamPrefix=CoherentStokesBeam[0].
Observation.DataProducts.Output_Beamformed_[0].beamTypes=[]
Observation.DataProducts.Output_Beamformed_[0].duration=0
Observation.DataProducts.Output_Beamformed_[0].fileFormat=HDF5
Observation.DataProducts.Output_Beamformed_[0].filename=L221197_SAP000_B000_S0_P000_bf.h5
Observation.DataProducts.Output_Beamformed_[0].location=locus001:/data/L221197/
Observation.DataProducts.Output_Beamformed_[0].nrOfCoherentStokesBeams=1
Observation.DataProducts.Output_Beamformed_[0].nrOfFlysEyeBeams=0
Observation.DataProducts.Output_Beamformed_[0].nrOfIncoherentStokesBeams=0
Observation.DataProducts.Output_Beamformed_[0].percentageWritten=0
Observation.DataProducts.Output_Beamformed_[0].size=0
Observation.DataProducts.Output_Beamformed_[0].startTime=2014-04-22 15:37:00
Observation.DataProducts.Output_Beamformed_[0].storageWriter=HDF5DEFAULT
Observation.DataProducts.Output_Beamformed_[0].storageWriterVersion=UNKNOWN
Observation.DataProducts.Output_Beamformed_[1].CoherentStokesBeam[0].Offset.angle1=-0.876860
Observation.DataProducts.Output_Beamformed_[1].CoherentStokesBeam[0].Offset.angle2=-0.922267
Observation.DataProducts.Output_Beamformed_[1].CoherentStokesBeam[0].Offset.coordType=RA-DEC
Observation.DataProducts.Output_Beamformed_[1].CoherentStokesBeam[0].Offset.equinox=J2000
Observation.DataProducts.Output_Beamformed_[1].CoherentStokesBeam[0].Pointing.angle1=0.052482
Observation.DataProducts.Output_Beamformed_[1].CoherentStokesBeam[0].Pointing.angle2=0.030312
Observation.DataProducts.Output_Beamformed_[1].CoherentStokesBeam[0].Pointing.coordType=RA-DEC
Observation.DataProducts.Output_Beamformed_[1].CoherentStokesBeam[0].Pointing.directionType=J2000
Observation.DataProducts.Output_Beamformed_[1].CoherentStokesBeam[0].Pointing.equinox=J2000
Observation.DataProducts.Output_Beamformed_[1].CoherentStokesBeam[0].Pointing.target=my_source
Observation.DataProducts.Output_Beamformed_[1].CoherentStokesBeam[0].SAP=0
Observation.DataProducts.Output_Beamformed_[1].CoherentStokesBeam[0].TAB=1
Observation.DataProducts.Output_Beamformed_[1].CoherentStokesBeam[0].antennaFields=[CS002HBA0,CS002HBA1,CS003HBA0,CS003HBA1,CS004HBA0,CS004HBA1,CS005HBA0,CS005HBA1,CS006HBA0,CS006HBA1,CS007HBA0,CS007HBA1]
Observation.DataProducts.Output_Beamformed_[1].CoherentStokesBeam[0].antennaSet=HBA_DUAL
Observation.DataProducts.Output_Beamformed_[1].CoherentStokesBeam[0].centralFrequencies=[119531250.0000, 119726562.5000, 119921875.0000, 120117187.5000, 120312500.0000, 120507812.5000, 120703125.0000, 120898437.5000, 121093750.0000, 121289062.5000, 121484375.0000, 121679687.5000, 121875000.0000, 122070312.5000, 122265625.0000, 122460937.5000, 122656250.0000, 122851562.5000, 123046875.0000, 123242187.5000, 123437500.0000, 123632812.5000, 123828125.0000, 124023437.5000, 124218750.0000, 124414062.5000, 124609375.0000, 124804687.5000, 125000000.0000, 125195312.5000, 125390625.0000, 125585937.5000, 125781250.0000, 125976562.5000, 126171875.0000, 126367187.5000, 126562500.0000, 126757812.5000, 126953125.0000, 127148437.5000, 127343750.0000, 127539062.5000, 127734375.0000, 127929687.5000, 128125000.0000, 128320312.5000, 128515625.0000, 128710937.5000, 128906250.0000, 129101562.5000, 129296875.0000, 129492187.5000, 129687500.0000, 129882812.5000, 130078125.0000, 130273437.5000, 130468750.0000, 130664062.5000, 130859375.0000, 131054687.5000, 131250000.0000, 131445312.5000, 131640625.0000, 131835937.5000, 132031250.0000, 132226562.5000, 132421875.0000, 132617187.5000, 132812500.0000, 133007812.5000, 133203125.0000, 133398437.5000, 133593750.0000, 133789062.5000, 133984375.0000, 134179687.5000, 134375000.0000, 134570312.5000, 134765625.0000, 134960937.5000, 135156250.0000, 135351562.5000, 135546875.0000, 135742187.5000, 135937500.0000, 136132812.5000, 136328125.0000, 136523437.5000, 136718750.0000, 136914062.5000, 137109375.0000, 137304687.5000, 137500000.0000, 137695312.5000, 137890625.0000, 138085937.5000, 138281250.0000, 138476562.5000, 138671875.0000, 138867187.5000, 139062500.0000, 139257812.5000, 139453125.0000, 139648437.5000, 139843750.0000, 140039062.5000, 140234375.0000, 140429687.5000, 140625000.0000, 140820312.5000, 141015625.0000, 141210937.5000, 141406250.0000, 141601562.5000, 141796875.0000, 141992187.5000, 142187500.0000, 142382812.5000, 142578125.0000, 142773437.5000, 142968750.0000, 143164062.5000, 143359375.0000, 143554687.5000, 143750000.0000, 143945312.5000, 144140625.0000, 144335937.5000, 144531250.0000, 144726562.5000, 144921875.0000, 145117187.5000, 145312500.0000, 145507812.5000, 145703125.0000, 145898437.5000, 146093750.0000, 146289062.5000, 146484375.0000, 146679687.5000, 146875000.0000, 147070312.5000, 147265625.0000, 147460937.5000, 147656250.0000, 147851562.5000, 148046875.0000, 148242187.5000, 148437500.0000, 148632812.5000, 148828125.0000, 149023437.5000, 149218750.0000, 149414062.5000, 149609375.0000, 149804687.5000, 150000000.0000, 150195312.5000, 150390625.0000, 150585937.5000, 150781250.0000, 150976562.5000]
Observation.DataProducts.Output_Beamformed_[1].CoherentStokesBeam[0].channelWidth=12207.031250
Observation.DataProducts.Output_Beamformed_[1].CoherentStokesBeam[0].channelsPerSubband=16
Observation.DataProducts.Output_Beamformed_[1].CoherentStokesBeam[0].dispersionMeasure=0.000000
Observation.DataProducts.Output_Beamformed_[1].CoherentStokesBeam[0].nrSubbands=162
Observation.DataProducts.Output_Beamformed_[1].CoherentStokesBeam[0].pipeline=0
Observation.DataProducts.Output_Beamformed_[1].CoherentStokesBeam[0].pipelineTAB=1
Observation.DataProducts.Output_Beamformed_[1].CoherentStokesBeam[0].sampleNrBits=32
Observation.DataProducts.Output_Beamformed_[1].CoherentStokesBeam[0].sampleType=float
Observation.DataProducts.Output_Beamformed_[1].CoherentStokesBeam[0].samplingTime=0.000492
Observation.DataProducts.Output_Beamformed_[1].CoherentStokesBeam[0].stationSubbands=[100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255, 256, 257, 258, 259, 260, 261]
Observation.DataProducts.Output_Beamformed_[1].CoherentStokesBeam[0].stokes=[I]
Observation.DataProducts.Output_Beamformed_[1].beamformingSpectralLeakage=true
Observation.DataProducts.Output_Beamformed_[1].beamPrefix=CoherentStokesBeam[0].
Observation.DataProducts.Output_Beamformed_[1].beamTypes=[]
Observation.DataProducts.Output_Beamformed_[1].duration=0
Observation.DataProducts.Output_Beamformed_[1].fileFormat=HDF5
Observation.DataProducts.Output_Beamformed_[1].filename=L221197_SAP000_B001_S0_P000_bf.h5
Observation.DataProducts.Output_Beamformed_[1].location=locus003:/data/L221197/
Observation.DataProducts.Output_Beamformed_[1].nrOfCoherentStokesBeams=1
Observation.DataProducts.Output_Beamformed_[1].nrOfFlysEyeBeams=0
Observation.DataProducts.Output_Beamformed_[1].nrOfIncoherentStokesBeams=0
Observation.DataProducts.Output_Beamformed_[1].percentageWritten=0
Observation.DataProducts.Output_Beamformed_[1].size=0
Observation.DataProducts.Output_Beamformed_[1].startTime=2014-04-22 15:37:00
Observation.DataProducts.Output_Beamformed_[1].storageWriter=HDF5DEFAULT
Observation.DataProducts.Output_Beamformed_[1].storageWriterVersion=UNKNOWN
Observation.DataProducts.nrOfOutput_Beamformed_=2
Observation.DataProducts.nrOfOutput_Correlated_=0
Observation.IncoherentStokes.antennaSet=HBA_DUAL
Observation.IncoherentStokes.channelWidth=12207.03125
Observation.IncoherentStokes.channelsPerSubband=16
Observation.IncoherentStokes.rawSamplingTime=8.192e-05
Observation.IncoherentStokes.samplingTime=0.0004915200000000001
Observation.IncoherentStokes.stationList=[CS007,CS005,CS002,CS004,CS006,CS003]
Observation.IncoherentStokes.stokes=I
Observation.IncoherentStokes.timeDownsamplingFactor=6
Observation.startTime=2013-02-16T17:00:00.000
Observation.stopTime=2013-02-16T17:01:00.000
_isCobalt=T
feedback_version=03.01.00
    """

    def test_extract_dataproduct_feedback(self):
        # If we can extract the keys of the returned dict, we assume the values will be found as well
        feedback = parameterset.fromString(self.feedback_obs_correlator_complete)
        feedback_keys = sorted(list(extract_dataproduct_feedback(feedback).keys()))
        self.assertEqual(["Output_Correlated_[0]", "Output_Correlated_[1]"], feedback_keys)

        feedback = parameterset.fromString(self.feedback_obs_beamformer_complete)
        feedback_keys = sorted(list(extract_dataproduct_feedback(feedback).keys()))
        self.assertEqual(["Output_Beamformed_[0]", "Output_Beamformed_[1]"], feedback_keys)

        feedback = parameterset.fromString(self.feedback_preprocessing_pipeline_complete)
        feedback_keys = sorted(list(extract_dataproduct_feedback(feedback).keys()))
        self.assertEqual(["Output_Correlated_[0]", "Output_Correlated_[1]"], feedback_keys)


    def test_parsing_correlator_dataproduct_feedback(self):
        feedback = parameterset.fromString(self.feedback_obs_correlator_complete)
        for dp_key, dp_feedback in extract_dataproduct_feedback(feedback).items():
            # if these do not throw, they succeed in parsing
            feedback_doc = observation_correlated_feedback_to_feedback_doc(dp_feedback)

            self.assertNotEqual({}, feedback_doc)

    def test_parsing_correlator_processing_feedback(self):
        feedback = parameterset.fromString(self.feedback_obs_correlator_complete)

        (feedback_doc, feedback_template) = observation_process_feedback_to_feedback_doc_and_template(feedback)

        # make sure processing feedback got returned
        self.assertNotEqual({}, feedback_doc)
        self.assertNotEqual(None, feedback_template)

        # and that the feedback is valid
        validate_json_against_schema(feedback_doc, feedback_template.schema)

    def test_parsing_pulsar_pipeline_cv_analysis_feedback(self):
        """ Test whether we can convert a ComplexVoltages feedback parset subset for an analysis dataproduct
            into a JSON feedback doc for the dataproduct. """

        # we assume dataproduct 0 is an analysis doc, and dataproduct 1 a summary
        pipeline_feedback = parameterset.fromString(self.feedback_pulsar_pipeline_cv_analysis)
        pipeline_dp_feedback = extract_dataproduct_feedback(pipeline_feedback)

        # enough feedback for to augment the feedback of the pulsar pipeline
        beamformer_feedback_doc = {
            "time": {
                "start_time": "2021-06-25T09:54Z",
                "duration": 3600,
            },
            "antennas": {
                "set": "HBA_ZERO",
                "fields": [ { "station": "CS001", "field": "HBA0", "type": "HBA" } ],
            },
            "samples": {
                "type": "float",
                "bits": 32,
            },
            "target": {
                'pointing': {'angle1': 0, 'angle2': 0, 'direction_type': 'J2000', 'target': 'target1'}
            }
        }

        # if these do not throw, they succeed in parsing
        _, dp_feedback = pipeline_dp_feedback.popitem()
        analysis_feedback_doc = pulsar_pipeline_analysis_feedback_to_feedback_doc(beamformer_feedback_doc, dp_feedback)
        self.assertNotEqual({}, analysis_feedback_doc)


    def test_parsing_pulsar_pipeline_cv_summary_feedback(self):
        """ Test whether we can convert a ComplexVoltages feedback parset subset for a summary dataproduct
            into a JSON feedback doc for the dataproduct. """

        pipeline_feedback = parameterset.fromString(self.feedback_pulsar_pipeline_cv_summary)
        pipeline_dp_feedback = extract_dataproduct_feedback(pipeline_feedback)

        # if these do not throw, they succeed in parsing
        _, dp_feedback = pipeline_dp_feedback.popitem()
        summary_feedback_doc = pulsar_pipeline_summary_feedback_to_feedback_doc(dp_feedback)
        self.assertNotEqual({}, summary_feedback_doc)


    def test_parsing_pulsar_pipeline_fe_analysis_feedback(self):
        """ Test whether we can convert a Fly's Eye feedback parset subset for an analysis dataproduct
            into a JSON feedback doc for the dataproduct. """

        # we assume dataproduct 0 is an analysis doc, and dataproduct 1 a summary
        pipeline_feedback = parameterset.fromString(self.feedback_pulsar_pipeline_fe_analysis)
        pipeline_dp_feedback = extract_dataproduct_feedback(pipeline_feedback)

        # enough feedback for to augment the feedback of the pulsar pipeline
        beamformer_feedback_doc = {
            "time": {
                "start_time": "2021-06-25T09:54Z",
                "duration": 3600,
            },
            "antennas": {
                "set": "HBA_ZERO",
                # fields will be set from pulp feedback in case of Fly's Eye
            },
            "samples": {
                "type": "float",
                "bits": 32,
            },
            "target": {
                'pointing': {'angle1': 0, 'angle2': 0, 'direction_type': 'J2000', 'target': 'target1'}
            },
        }

        # if these do not throw, they succeed in parsing
        _, dp_feedback = pipeline_dp_feedback.popitem()
        analysis_feedback_doc = pulsar_pipeline_analysis_feedback_to_feedback_doc(beamformer_feedback_doc, dp_feedback)
        self.assertNotEqual({}, analysis_feedback_doc)

        # Field information should have been extracted from the feedback
        self.assertEqual("CS001", analysis_feedback_doc["antennas"]["fields"][0]["station"])
        self.assertEqual("HBA0",  analysis_feedback_doc["antennas"]["fields"][0]["field"])


    def test_parsing_beamformer_feedback(self):
        feedback = parameterset.fromString(self.feedback_obs_beamformer_complete)

        for dp_key, dp_feedback in extract_dataproduct_feedback(feedback).items():
            # if these do not throw, they succeed in parsing
            feedback_doc = observation_beamformed_feedback_to_feedback_doc(dp_feedback)

            self.assertNotEqual({}, feedback_doc)


    def test_parsing_preprocessing_pipeline_feedback(self):
        correlator_feedback = parameterset.fromString(self.feedback_obs_correlator_complete)
        correlator_dp_feedback = extract_dataproduct_feedback(correlator_feedback)

        pipeline_feedback = parameterset.fromString(self.feedback_preprocessing_pipeline_complete)
        pipeline_dp_feedback = extract_dataproduct_feedback(pipeline_feedback)

        for dp_key, dp_feedback in pipeline_dp_feedback.items():
            # we match keys 1:1 between correlator and pipeline, because that happens to work out
            correlator_feedback_doc = observation_correlated_feedback_to_feedback_doc(correlator_dp_feedback[dp_key])

            self.assertNotEqual({}, correlator_feedback_doc)

            # if these do not throw, they succeed in parsing
            feedback_doc = preprocessing_pipeline_feedback_to_feedback_doc(correlator_feedback_doc, dp_feedback)

            self.assertNotEqual({}, feedback_doc)


    def test_process_feedback_into_subtask_dataproducts_raises_on_wrong_subtask_state(self):
        subtask_data = Subtask_test_data()
        subtask:models.Subtask = models.Subtask.objects.create(**subtask_data)

        with self.assertRaises(SubtaskInvalidStateException) as cm:
            process_feedback_into_subtask_dataproducts(subtask, parameterset.fromString(self.feedback_obs_correlator_complete))

        self.assertIn("Cannot process feedback", str(cm.exception))


    def test_process_feedback_for_subtask_and_set_to_finished_if_complete_raises_on_wrong_subtask_state(self):
        subtask_data = Subtask_test_data()
        subtask:models.Subtask = models.Subtask.objects.create(**subtask_data)

        with self.assertRaises(SubtaskInvalidStateException) as cm:
            process_feedback_for_subtask_and_set_to_finished_if_complete(subtask, self.feedback_obs_correlator_complete)

        self.assertIn("Cannot process feedback", str(cm.exception))


    def test_generate_dataproduct_feedback_from_subtask_feedback_and_set_finished_fails_on_incomplete_feedback(self):
        subtask_data = Subtask_test_data(subtask_template=models.SubtaskTemplate.get_latest(name='observation control'))
        subtask_obs:models.Subtask = models.Subtask.objects.create(**subtask_data)
        set_subtask_state_following_allowed_transitions(subtask_obs, 'finishing')
        subtask_obs_output = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=subtask_obs))

        subtask_data = Subtask_test_data(subtask_template=models.SubtaskTemplate.get_latest(name='preprocessing pipeline'))
        subtask_pipe: models.Subtask = models.Subtask.objects.create(**subtask_data)
        set_subtask_state_following_allowed_transitions(subtask_pipe, 'finishing')
        subtask_pipe_output = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=subtask_pipe))

        test_dir = "/tmp/test/data/%s" % uuid.uuid4()
        empty_feedback_template = models.DataproductFeedbackTemplate.get_latest(name='empty')
        dataproduct_obs_out1:models.Dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(filename='L220133_SAP000_SB000_uv.MS', directory=test_dir, producer=subtask_obs_output, feedback_template=empty_feedback_template))
        dataproduct_obs_out2: models.Dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(filename='L220133_SAP000_SB001_uv.MS', directory=test_dir, producer=subtask_obs_output, feedback_template=empty_feedback_template))
        dataproduct_pipe_out1: models.Dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(filename='L99307_SB000_uv.dppp.MS', directory=test_dir, producer=subtask_pipe_output, feedback_template=empty_feedback_template))
        dataproduct_pipe_out2: models.Dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(filename='L99307_SB001_uv.dppp.MS', directory=test_dir, producer=subtask_pipe_output, feedback_template=empty_feedback_template))
        models.DataproductTransform.objects.create(input=dataproduct_obs_out1, output=dataproduct_pipe_out1, identity=True)
        models.DataproductTransform.objects.create(input=dataproduct_obs_out2, output=dataproduct_pipe_out2, identity=True)

        # assert dataproducts have no feedback docs before conversion
        self.assertNotIn('percentage_written', dataproduct_obs_out1.feedback_doc)
        self.assertNotIn('percentage_written', dataproduct_obs_out2.feedback_doc)
        self.assertNotIn('percentage_written', dataproduct_pipe_out1.feedback_doc)
        self.assertNotIn('percentage_written', dataproduct_pipe_out2.feedback_doc)

        process_feedback_for_subtask_and_set_to_finished_if_complete(subtask_obs, self.feedback_obs_correlator_complete)
        process_feedback_for_subtask_and_set_to_finished_if_complete(subtask_pipe, self.feedback_preprocessing_pipeline_incomplete) # <--- Note: test for incomplete feedback

        # assert not in FINISHED state
        self.assertEqual(models.SubtaskState.objects.get(value='finishing'), subtask_pipe.state)
        self.assertFalse(subtask_pipe.is_feedback_complete)

    def test_generate_dataproduct_feedback_from_subtask_feedback_and_set_finished(self):
        subtask_data = Subtask_test_data(subtask_template=models.SubtaskTemplate.get_latest(name='observation control'))
        subtask_obs:models.Subtask = models.Subtask.objects.create(**subtask_data)
        set_subtask_state_following_allowed_transitions(subtask_obs, 'finishing')
        subtask_obs_output = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=subtask_obs))

        subtask_data = Subtask_test_data(subtask_template=models.SubtaskTemplate.get_latest(name='preprocessing pipeline'))
        subtask_pipe: models.Subtask = models.Subtask.objects.create(**subtask_data)
        set_subtask_state_following_allowed_transitions(subtask_pipe, 'finishing')
        subtask_pipe_output = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=subtask_pipe))

        empty_feedback_template = models.DataproductFeedbackTemplate.get_latest(name='empty')
        dataproduct_obs_out1:models.Dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(filename='L220133_SAP000_SB000_uv.MS', producer=subtask_obs_output, feedback_template=empty_feedback_template))
        dataproduct_obs_out2: models.Dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(filename='L220133_SAP000_SB001_uv.MS', producer=subtask_obs_output, feedback_template=empty_feedback_template))
        dataproduct_pipe_out1: models.Dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(filename='L99307_SB000_uv.dppp.MS', producer=subtask_pipe_output, feedback_template=empty_feedback_template))
        dataproduct_pipe_out2: models.Dataproduct = models.Dataproduct.objects.create(**Dataproduct_test_data(filename='L99307_SB001_uv.dppp.MS', producer=subtask_pipe_output, feedback_template=empty_feedback_template))
        models.DataproductTransform.objects.create(input=dataproduct_obs_out1, output=dataproduct_pipe_out1, identity=True)
        models.DataproductTransform.objects.create(input=dataproduct_obs_out2, output=dataproduct_pipe_out2, identity=True)

        # assert dataproducts have no feedback docs before conversion
        for dataproduct in [dataproduct_obs_out1, dataproduct_obs_out2, dataproduct_pipe_out1, dataproduct_pipe_out2]:
            self.assertNotIn('percentage_written', dataproduct.feedback_doc)

        process_feedback_for_subtask_and_set_to_finished_if_complete(subtask_obs, self.feedback_obs_correlator_complete)
        process_feedback_for_subtask_and_set_to_finished_if_complete(subtask_pipe, self.feedback_preprocessing_pipeline_complete)

        # reload dataproducts and assert dataproduct feedback docs have feedback after conversion
        for dataproduct in [dataproduct_obs_out1, dataproduct_obs_out2, dataproduct_pipe_out1, dataproduct_pipe_out2]:
            dataproduct.refresh_from_db()

        self.assertIsNotNone(dataproduct_obs_out1.feedback_doc)
        self.assertIn('percentage_written', dataproduct_obs_out1.feedback_doc)
        self.assertIsNotNone(dataproduct_obs_out2.feedback_doc)
        self.assertIn('percentage_written', dataproduct_obs_out2.feedback_doc)
        self.assertIsNotNone(dataproduct_pipe_out1.feedback_doc)
        self.assertIn('percentage_written', dataproduct_pipe_out1.feedback_doc)
        self.assertIsNotNone(dataproduct_pipe_out2.feedback_doc)
        self.assertIn('percentage_written', dataproduct_pipe_out2.feedback_doc)

        # assert correct relations of feedback docs
        self.assertNotEqual(dataproduct_obs_out1.feedback_doc['frequency']['subbands'],
                            dataproduct_obs_out2.feedback_doc['frequency']['subbands'])
        self.assertNotEqual(dataproduct_obs_out1.feedback_doc['frequency']['subbands'],
                            dataproduct_pipe_out2.feedback_doc['frequency']['subbands'])
        self.assertEqual(dataproduct_obs_out1.feedback_doc['frequency']['subbands'],
                         dataproduct_pipe_out1.feedback_doc['frequency']['subbands'])
        self.assertEqual(dataproduct_obs_out1.feedback_doc['frequency']['subbands'],
                         dataproduct_pipe_out1.feedback_doc['frequency']['subbands'])

        # assert FINISHED states
        self.assertEqual(models.SubtaskState.objects.get(value='finished'), subtask_obs.state)
        self.assertEqual(models.SubtaskState.objects.get(value='finished'), subtask_pipe.state)

        self.assertTrue(subtask_obs.is_feedback_complete)
        self.assertTrue(subtask_pipe.is_feedback_complete)


if __name__ == "__main__":
    os.environ['TZ'] = 'UTC'
    unittest.main()
