#!/usr/bin/env python3

import copy
import unittest
from typing import Dict
from typing import List
from unittest import mock

import logging

logger = logging.getLogger('lofar.'+__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

from lofar_station_client.observation import (
    multi_station_observation
)

from lofar_station_client.observation.multi_station_observation import (
    MultiStationObservation
)

from lofar.sas.tmss.test.test_environment import TMSSTestEnvironment

from lofar.messaging.messagebus import TemporaryExchange, BusListenerJanitor
from lofar.common.test_utils import integration_test

try:
    from lofar.sas.tmss.services.lobster import Lobster
    from lofar.sas.tmss.services.lobster import ObservationPool
    from lofar.sas.tmss.services.lobster.rpc_client import  L2TMSSObservationControlRPCClient
except ModuleNotFoundError:
    print("Skipping test. could not import lobster 'TMSSLobster' service. Did you build it?")
    exit(3)

from datetime import datetime, timedelta
from uuid import uuid4


@integration_test
class TestL2StationTMSSIntegration(unittest.TestCase):

    DEFAULT_STATION_GROUPS = [{'stations': ['CS001'], 'max_nr_missing': 0}]

    @classmethod
    def setUpClass(cls) -> None:
        cls.TEST_UUID = uuid4()

        cls.tmp_exchange = TemporaryExchange()
        cls.tmp_exchange.open()

        cls.tmss_test_env = TMSSTestEnvironment(
            exchange=cls.tmp_exchange.address, populate_schemas=True, start_postgres_listener=True,
            populate_test_data=False, enable_viewflow=False, start_dynamic_scheduler=False,
            start_subtask_scheduler=True, start_workflow_service=False
        )
        cls.tmss_test_env.start()

    def setUp(self) -> None:
        # we should not depend on "previous" data
        self.tmss_test_env.delete_scheduling_unit_drafts_cascade()

        # Prevent unit test creating actual device proxies!
        proxy_patcher = mock.patch.object(
            multi_station_observation, "StationObservationFuture", autospec=True
        )
        self.m_station_obs_future = proxy_patcher.start()
        self.addCleanup(proxy_patcher.stop)

    @classmethod
    def tearDownClass(cls) -> None:
        cls.tmss_test_env.stop()
        cls.tmp_exchange.close()

    def prepare_observation_subtask(self,
        station_groups: List[Dict[str, any]] = None
    ):
        """Create an defined/schedulable observation subtask (and its parent task, schedulingunit) on TMSS
        and return the observation subtask with the given station_groups"""

        from lofar.sas.tmss.test.tmss_test_data_django_models import (
            Project_test_data,
            SchedulingSet_test_data
        )
        from lofar.sas.tmss.tmss.tmssapp import models
        from lofar.sas.tmss.tmss.tmssapp.tasks import (
            create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft,
            update_task_graph_from_specifications_doc
        )

        if not station_groups:
            station_groups = self.DEFAULT_STATION_GROUPS

        project = models.Project.objects.create(**Project_test_data(auto_ingest=True))
        scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project))

        # use a simple observation strategy with only one target imaging observation
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="Simple Observation")
        scheduling_unit_spec = copy.deepcopy(strategy_template.template)
        # overrule the normal stations for this template with just the one CS001 LOFAR2 station
        scheduling_unit_spec['tasks']['Observation']['specifications_doc']['station_configuration'][
            'station_groups'
        ] = station_groups

        # create a scheduling unit draft & blueprint
        su_draft = models.SchedulingUnitDraft.objects.create(
            name=str(uuid4()),
            scheduling_set=scheduling_set,
            specifications_template=strategy_template.scheduling_unit_template,
            observation_strategy_template=strategy_template
        )
        su_draft = update_task_graph_from_specifications_doc(su_draft, scheduling_unit_spec)
        su_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(su_draft)
        # get the observation subtask
        observation_subtask = su_blueprint.subtasks.filter(
            specifications_template__type__value=models.SubtaskType.Choices.OBSERVATION.value
        ).first()

        return observation_subtask

    def test_observation_pool_create(self):
        observation_subtask = self.prepare_observation_subtask()

        # get the lofar2 station specs via the REST API (just like lobster)
        # and check that the specs can be used in the ObservationPool
        with self.tmss_test_env.create_tmss_client() as tmss_client:
            specifications = tmss_client.get_subtask_l2stationspecs(observation_subtask.id)

        t_observation_pool = ObservationPool()

        t_observation_pool.create_multistationobservation(
            subtask_id=observation_subtask.id, observation_id=observation_subtask.id,
            specifications=specifications, stations=self.DEFAULT_STATION_GROUPS[0]["stations"]
        )

        self.assertIsInstance(
            t_observation_pool.get_multistationobservation(observation_subtask.id),
            multi_station_observation.MultiStationObservation
        )

        self.assertRaises(
            KeyError, t_observation_pool.get_multistationobservation, 12345
        )

    @mock.patch.object(MultiStationObservation, "_get_results")
    def test_simple_lofar2_observation(self, m_result):
        from lofar.sas.tmss.tmss.tmssapp import models
        from lofar.sas.tmss.tmss.tmssapp.subtasks import wait_for_subtask_status, schedule_subtask

        observation_subtask = self.prepare_observation_subtask()

        # Patch waiting and retrieving on future results
        m_result.return_value = [True] * len(
            self.DEFAULT_STATION_GROUPS[0]['stations']
        )

        # mock scp/ssh calls to cobalt
        with mock.patch('lofar.sas.tmss.services.lobster.message_handler._system'):
            # start the lobster buslistener, and the rpc service
            # this should pick up the scheduled observation and send instructions
            # to the (mocked) station. as a result, the observation subtask should
            # become STARTED
            with Lobster(self.tmp_exchange.address, self.tmp_exchange.broker,
                         tmss_client_credentials_id=self.tmss_test_env.client_credentials.dbcreds_id
            ):
                # the observation subtask should still be defined/schedulable
                self.assertEqual(models.SubtaskState.Choices.DEFINED.value, observation_subtask.state.value)

                # schedule the observation 15 seconds from now
                # That should generate a 'scheduled' event,
                # to which the L2TMSSObservationControlBusListener should respond by queuing and starting it.
                observation_subtask = schedule_subtask(observation_subtask,
                                                       datetime.utcnow() + timedelta(seconds=15)
                )
                self.assertEqual(models.SubtaskState.Choices.SCHEDULED.value, observation_subtask.state.value)

                # wait until L2TMSSObservationControlBusListener started the observation...
                observation_subtask = wait_for_subtask_status(
                    observation_subtask,
                    models.SubtaskState.Choices.STARTED.value,
                    timeout=60
                )

                self.assertEqual(models.SubtaskState.Choices.STARTED.value,
                                 observation_subtask.state.value)


                # test the rpc-client to abort the observation via the running service
                with L2TMSSObservationControlRPCClient.create(self.tmp_exchange.address,
                                                              self.tmp_exchange.broker) as rpc_client:

                    # abort the observation
                    rpc_client.abort_observation(observation_subtask.id)

                    # wait until the service aborted the observation...
                    observation_subtask = wait_for_subtask_status(
                        observation_subtask,
                        models.SubtaskState.Choices.CANCELLED.value,
                        timeout=60
                    )

                    self.assertEqual(models.SubtaskState.Choices.CANCELLED.value,
                                     observation_subtask.state.value)

if __name__ == '__main__':
    unittest.main()
