#!/usr/bin/env python3
import unittest
import logging
import json
from unittest import mock

from lofar_station_client.observation import multi_station_observation

try:
    from lofar.sas.tmss.services.lobster.observation_pool import ObservationPool
    from lofar.sas.tmss.services.lobster.specification_translation import (
        station_to_host
    )
    from lofar.sas.tmss.services.lobster.specification_translation import (
        extract_stations
    )
except ModuleNotFoundError:
    print("Skipping test. could not import lobster 'TMSSLobster' service. Did you build it?")
    exit(3)

logger = logging.getLogger('lofar.'+__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)


# 3 stations, 2 observations, station 1 and two share observation id
test_dummy_spec = json.loads("""
{
  "stations": [
    {
      "station": "CS001",
      "antenna_fields": {
        "SAPs": [
          {
            "name": "target",
            "pointing": {
              "angle1": 0.6624317181687094,
              "angle2": 1.5579526427549426,
              "direction_type": "J2000",
              "target": "target1"
            },
            "purpose": "target",
            "subbands": [
              104,
              105
            ]
          }
        ],
        "antenna_field": "HBA0",
        "antenna_set": "ALL",
        "filter": "HBA_110_190",
        "observation_id": 2021361,
        "stop_time": "2023-07-06T18:14:00"
      }
    },
    {
      "station": "CS002",
      "antenna_fields": {
        "SAPs": [
          {
            "name": "target",
            "pointing": {
              "angle1": 0.6624317181687094,
              "angle2": 1.5579526427549426,
              "direction_type": "J2000",
              "target": "target1"
            },
            "purpose": "target",
            "subbands": [
              104,
              105
            ]
          }
        ],
        "antenna_field": "HBA0",
        "antenna_set": "ALL",
        "filter": "HBA_110_190",
        "observation_id": 2021361,
        "stop_time": "2023-07-06T18:14:00"
      }
    },
    {
      "station": "CS003",
      "antenna_fields": {
        "SAPs": [
          {
            "name": "target",
            "pointing": {
              "angle1": 1.6624317181687094,
              "angle2": 0.5579526427549426,
              "direction_type": "J2000",
              "target": "target1"
            },
            "purpose": "target",
            "subbands": [
              104,
              105
            ]
          }
        ],
        "antenna_field": "HBA0",
        "antenna_set": "ALL",
        "filter": "HBA_110_190",
        "observation_id": 2021362,
        "stop_time": "2023-07-06T18:14:00"
      }
    }
  ]
}
""")


class TestL2StationTMSS(unittest.TestCase):

    def setUp(self):
        # Prevent unit test creating actual device proxies!
        proxy_patcher = mock.patch.object(multi_station_observation, "StationObservationFuture")
        self.m_station_obs_future = proxy_patcher.start()
        self.addCleanup(proxy_patcher.stop)

    def test_station_to_host(self):
        """Test station to host"""

        # TODO(Corne): https://support.astron.nl/jira/browse/TMSS-2861
        self.assertEqual("tango.service.consul:10000", station_to_host("irrelevant", 10000))
        # self.assertEqual("testc.control.lofar:10000", station_to_host("test", 10000))
        # self.assertEqual("cs001c.control.lofar:10000", station_to_host("cs001", 10000))

    def test_observation_pool_create_multistationobservation_get(self):
        """Create multistationobservation and retrieve observation from pool"""
        t_subtask_id = 1234
        t_observation_id = 2021361

        t_observation_pool = ObservationPool()

        t_observation_pool.create_multistationobservation(
            subtask_id=t_subtask_id, observation_id=t_observation_id,
            specifications={
                "stations": [{
                    "station": "cs001c",
                    "antenna_fields": [{
                        "antenna_field": "HBA",
                        "observation_id": 2021361
                    }]
                }]
            }, stations=["cs001c"]
        )

        self.assertIsInstance(
            t_observation_pool.get_multistationobservation(t_observation_id),
            multi_station_observation.MultiStationObservation
        )

        self.assertRaises(
            KeyError, t_observation_pool.get_multistationobservation, 12345
        )

    def test_observation_pool_create_multistationobservation_duplicate_error(
        self
    ):
        """Create multistationobs duplicate observations which should error"""
        t_subtask_id = 1234
        t_observation_id = 2021361

        t_observation_pool = ObservationPool()

        t_observation_pool.create_multistationobservation(
            subtask_id=t_subtask_id, observation_id=t_observation_id,
            specifications={
                "stations": [{
                    "station": "cs001c",
                    "antenna_fields": [{
                        "antenna_field": "HBA",
                        "observation_id": 2021361
                    }]
                }]
            }, stations=["cs001c"]
        )

        self.assertRaises(
            ValueError, t_observation_pool.create_multistationobservation,
            subtask_id=t_subtask_id,
            observation_id=t_observation_id,
            specifications={
                "stations": [{
                    "station": "cs001c",
                    "antenna_fields": [{
                        "antenna_field": "HBA",
                        "observation_id": 2021361
                    }]
                }]
            }, stations=["cs001c"]
        )

    def test_observation_pool_create_multistationobservation_remove(self):
        """create multistationobs and remove observations multiple times"""
        insertions = 10

        t_subtask_id = 1234
        t_observation_pool = ObservationPool()

        for i in range(insertions):
            t_observation_pool.create_multistationobservation(
                subtask_id=t_subtask_id, observation_id=i,
                specifications={
                    "stations": [{
                        "station": "cs001c",
                        "antenna_fields": [{
                            "antenna_field": "HBA",
                            "observation_id": i
                        }]
                    }]
                }, stations=["cs001c"]
            )

        self.assertEqual(1, len(t_observation_pool.pool))
        self.assertEqual(insertions, len(t_observation_pool.pool[t_subtask_id]))

        for i in range(insertions):
            self.assertIsInstance(
                t_observation_pool.get_multistationobservation(i),
                multi_station_observation.MultiStationObservation
            )
            t_observation_pool.remove(i)
            self.assertRaises(
                KeyError, t_observation_pool.get_multistationobservation, i
            )

        self.assertEqual(0, len(t_observation_pool.pool))

    def test_observation_pool_subtasks_get_observations(self):
        """Test associating observations to subtask using multi level dict"""
        num_subtasks = 5
        num_obs = 10

        t_observation_pool = ObservationPool()

        obs_id = 0
        for i in range(num_subtasks):
            for j in range(num_obs-i):
                t_observation_pool.create_multistationobservation(
                    subtask_id=i, observation_id=obs_id,
                    specifications={
                        "stations": [{
                            "station": "cs001c",
                            "antenna_fields": [{
                                "antenna_field": "HBA",
                                "observation_id": obs_id
                            }]
                        }]
                    }, stations=["cs001c"]
                )
                obs_id += 1

        for i in range(num_subtasks):
            self.assertEqual(
                num_obs-i, len(t_observation_pool.get_observations(i))
            )

    def test_extract_stations(self):
        "Test split_l2specification_per_observation"

        result = extract_stations(test_dummy_spec)

        # TODO(Corne): https://support.astron.nl/jira/browse/TMSS-2861
        self.assertIn("tango.service.consul:10000", result)
        # self.assertIn("CS001c.control.lofar:10000", result)
        # self.assertIn("CS002c.control.lofar:10000", result)
        #
        # self.assertIn("CS003c.control.lofar:10000", result)


if __name__ == '__main__':
    unittest.main()
