#!/usr/bin/env python3

import unittest

import logging
logger = logging.getLogger('lofar.'+__name__)

from lofar.common.test_utils import exit_with_skipped_code_if_skip_integration_tests
exit_with_skipped_code_if_skip_integration_tests()

from lofar.messaging.messagebus import TemporaryExchange
from lofar.common.test_utils import integration_test
from lofar.common.json_utils import validate_json_against_its_schema
from lofar.parameterset import parameterset

from datetime import datetime, timedelta
from dateutil import parser
from distutils.util import strtobool
from uuid import uuid4
import os
import shutil

@integration_test
class TestObservationStrategiesSpecificationAndScheduling(unittest.TestCase):
    '''The purpose of this test is to prove correctness of the specified and scheduled observations, pipelines and
    other (sub)tasks by checking the resulting statuses, the created subtask-specification_docs, parsets and dataproducts.
    For this test we regard TMSS and the services as a black box,
    and we can only use the http rest api (via the tmss_client) to specify, schedule and check the results.
    '''
    @classmethod
    def setUpClass(cls) -> None:
        cls.TEST_DIR = '/tmp/TestObservationStrategiesSpecificationAndScheduling/' + str(uuid4())
        os.makedirs(cls.TEST_DIR)

        cls.tmp_exchange = TemporaryExchange(cls.__class__.__name__)
        cls.tmp_exchange.open()

        # override DEFAULT_BUSNAME (so the RA services connect to this exchange)
        import lofar
        lofar.messaging.config.DEFAULT_BUSNAME = cls.tmp_exchange.address

        # create a blackbox TMSSTestEnvironment, and remember the purpose of this big test: we only care about the specifications and scheduling
        # so, there is no need to start all the fancy background services (for ingest, cleanup, viewflow, etc).
        from lofar.sas.tmss.test.test_environment import TMSSTestEnvironment
        cls.tmss_test_env = TMSSTestEnvironment(exchange=cls.tmp_exchange.address,
                                                populate_schemas=True, populate_legacy_schemas=True, start_postgres_listener=False,
                                                populate_test_data=False, enable_viewflow=False, start_dynamic_scheduler=False,
                                                start_subtask_scheduler=False, start_workflow_service=False)
        cls.tmss_test_env.start()

        cls.tmss_client = cls.tmss_test_env.create_tmss_client()
        cls.tmss_client.open()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.tmss_client.close()
        cls.tmss_test_env.stop()
        cls.tmp_exchange.close()
        shutil.rmtree(cls.TEST_DIR, ignore_errors=True)

    def setUp(self) -> None:
        self.tmss_test_env.delete_scheduling_unit_drafts_cascade()

        # prepare a new clean project and parent scheduling_set for each tested observation strategy template
        test_data_creator = self.tmss_test_env.create_test_data_creator()
        self.project = test_data_creator.post_data_and_get_response_as_json_object(test_data_creator.Project(auto_ingest=True), '/project/')
        self.scheduling_set = test_data_creator.post_data_and_get_response_as_json_object(test_data_creator.SchedulingSet(project_url=self.project['url']), '/scheduling_set/')

    def check_statuses(self, subtask_id, expected_subtask_status, expected_task_status, expected_schedunit_status):
        '''helper method to fetch the latest statuses of the subtask, its task, and its schedulingunit, and check for the expected statuses'''
        subtask = self.tmss_client.get_subtask(subtask_id)
        self.assertEqual(expected_subtask_status, subtask['state_value'])
        task = self.tmss_client.get_url_as_json_object(subtask['task_blueprint'])
        self.assertEqual(expected_task_status, task['status'])
        schedunit = self.tmss_client.get_url_as_json_object(task['scheduling_unit_blueprint'])
        self.assertEqual(expected_schedunit_status, schedunit['status'])

    def test_UC1(self):
        def check_parset(obs_subtask, is_target_obs:bool):
            '''helper function to check the parset for UC1 target/calibrator observations'''
            obs_parset = parameterset.fromString(self.tmss_client.get_subtask_parset(obs_subtask['id'])).dict()
            self.assertEqual(obs_subtask['id'], int(obs_parset['Observation.ObsID']))
            self.assertEqual('HBA', obs_parset['Observation.antennaArray'])
            self.assertEqual('HBA_DUAL_INNER', obs_parset['Observation.antennaSet'])
            self.assertEqual('HBA_110_190', obs_parset['Observation.bandFilter'])
            self.assertEqual(1, int(obs_parset['Observation.nrAnaBeams']))
            self.assertEqual(2 if is_target_obs else 1, int(obs_parset['Observation.nrBeams']))
            self.assertEqual('Observation', obs_parset['Observation.processType'])
            self.assertEqual('Beam Observation', obs_parset['Observation.processSubtype'])
            self.assertEqual(parser.parse(obs_subtask['scheduled_start_time']), parser.parse(obs_parset['Observation.startTime']))
            self.assertEqual(parser.parse(obs_subtask['scheduled_stop_time']), parser.parse(obs_parset['Observation.stopTime']))
            self.assertEqual(200, int(obs_parset['Observation.sampleClock']))
            self.assertEqual(243, len(obs_parset['Observation.Beam[0].subbandList'].split(',')))
            if is_target_obs:
                self.assertEqual(243, len(obs_parset['Observation.Beam[1].subbandList'].split(',')))
            self.assertEqual(True, strtobool(obs_parset['Observation.DataProducts.Output_Correlated.enabled']))
            self.assertEqual(486 if is_target_obs else 243, len(obs_parset['Observation.DataProducts.Output_Correlated.filenames'].split(',')))
            self.assertEqual(486 if is_target_obs else 243, len(obs_parset['Observation.DataProducts.Output_Correlated.locations'].split(',')))
            self.assertEqual(False, strtobool(obs_parset.get('Observation.DataProducts.Output_CoherentStokes.enabled','false')))
            self.assertEqual(False, strtobool(obs_parset.get('Observation.DataProducts.Output_IncoherentStokes.enabled','false')))
            self.assertEqual(False, strtobool(obs_parset.get('Observation.DataProducts.Output_Pulsar.enabled','false')))
            self.assertEqual(False, strtobool(obs_parset.get('Observation.DataProducts.Output_InstrumentModel.enabled','false')))
            self.assertEqual(False, strtobool(obs_parset.get('Observation.DataProducts.Output_SkyImage.enabled','false')))


        # import helper method to cycle through allowed state transitions
        from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions

        # setup: create a scheduling unit from the UC1 observation strategy template
        uc1_strategy_template = self.tmss_client.get_scheduling_unit_observing_strategy_template('IM HBA LoTSS - 2 Beams', 16)
        self.assertIsNotNone(uc1_strategy_template)


        # check before and after scheduling if the target observation task_template and the common datetime_tempalte is_used_in_unfinished_tasks_blueprints.
        from lofar.sas.tmss.tmss.tmssapp.models.specification import TaskTemplate, CommonSchemaTemplate, SchedulingUnitObservingStrategyTemplate
        tgtobs_task_template = TaskTemplate.get_version_or_latest("target observation", 9)
        datetime_template = CommonSchemaTemplate.get_version_or_latest('datetime', 9)
        strategy_template = SchedulingUnitObservingStrategyTemplate.get_version_or_latest(uc1_strategy_template['name'], uc1_strategy_template['version'])
        # should not be used by any task/unit yet.
        self.assertFalse(tgtobs_task_template.is_used_in_unfinished_tasks_blueprints)
        self.assertFalse(datetime_template.is_used_in_unfinished_tasks_blueprints)
        self.assertFalse(strategy_template.is_used_in_unfinised_schedulinging_unit_blueprints)

        scheduling_unit_draft = self.tmss_client.create_scheduling_unit_draft_from_strategy_template(uc1_strategy_template['id'], self.scheduling_set['id'])
        # check general object settings after creation
        self.assertEqual(uc1_strategy_template['url'], scheduling_unit_draft['observation_strategy_template'])
        self.assertFalse(scheduling_unit_draft['ingest_permission_required'])

        # TODO: check draft specification, constraints, etc according to UC1 requirements like antennaset, filters, subbands, etc.
        # for now, just check if the spec is ok according to schema.
        scheduling_unit_draft = self.tmss_client.get_schedulingunit_draft(scheduling_unit_draft['id'], include_specifications_doc=True)
        validate_json_against_its_schema(scheduling_unit_draft['specifications_doc'])

        scheduling_unit_blueprint = self.tmss_client.create_scheduling_unit_blueprint_and_tasks_and_subtasks_tree(scheduling_unit_draft['id'])

        # should now be used but not scheduled/running/finished yet.
        self.assertTrue(tgtobs_task_template.is_used_in_unfinished_tasks_blueprints)
        self.assertTrue(datetime_template.is_used_in_unfinished_tasks_blueprints)
        self.assertTrue(strategy_template.is_used_in_unfinised_schedulinging_unit_blueprints)

        scheduling_unit_blueprint_ext = self.tmss_client.get_schedulingunit_blueprint(scheduling_unit_blueprint['id'], extended=True, include_specifications_doc=True)
        self.assertFalse(scheduling_unit_blueprint_ext['ingest_permission_required'])

        # blueprint spec should be copied verbatim, so should be equal to (unchanged/unedited) draft
        self.assertEqual(scheduling_unit_draft['specifications_doc'], scheduling_unit_blueprint_ext['specifications_doc'])

        # observation(s) did not run yet, so observed_stop_time should be None
        self.assertIsNone(scheduling_unit_blueprint_ext['observed_stop_time'])
        self.assertEqual("schedulable", scheduling_unit_blueprint_ext['status'])

        # check the tasks
        tasks = scheduling_unit_blueprint_ext['task_blueprints']
        self.assertEqual(9, len(tasks))
        observation_tasks = [t for t in tasks if t['task_type'] == 'observation']
        self.assertEqual(3, len(observation_tasks))
        pipeline_tasks = [t for t in tasks if t['task_type'] == 'pipeline']
        self.assertEqual(4, len(pipeline_tasks))
        self.assertEqual(1, len([t for t in tasks if t['task_type'] == 'ingest']))
        ingest_task = next(t for t in tasks if t['task_type'] == 'ingest')
        self.assertEqual(1, len([t for t in tasks if t['task_type'] == 'cleanup']))
        cleanup_task = next(t for t in tasks if t['task_type'] == 'cleanup')

        cal_obs1_task = next(t for t in observation_tasks if t['name'] == 'Calibrator Observation 1')
        target_obs_task = next(t for t in observation_tasks if t['name'] == 'Target Observation')
        cal_obs2_task = next(t for t in observation_tasks if t['name'] == 'Calibrator Observation 2')

        # -------------------
        # schedule first calibrator obs
        self.assertEqual(1, len([st for st in cal_obs1_task['subtasks'] if st['subtask_type'] == 'observation']))
        cal_obs1_subtask = next(st for st in cal_obs1_task['subtasks'] if st['subtask_type'] == 'observation')
        cal_obs1_subtask = self.tmss_client.schedule_subtask(cal_obs1_subtask['id'])
        check_parset(cal_obs1_subtask, is_target_obs=False)
        self.check_statuses(cal_obs1_subtask['id'], "scheduled", "scheduled", "scheduled")

        # check output_dataproducts
        cal_obs1_output_dataproducts = self.tmss_client.get_subtask_output_dataproducts(cal_obs1_subtask['id'])
        self.assertEqual(243, len(cal_obs1_output_dataproducts))

        # "mimic" that the cal_obs1_subtask starts running
        set_subtask_state_following_allowed_transitions(cal_obs1_subtask['id'], 'started')
        self.check_statuses(cal_obs1_subtask['id'], "started", "started", "observing")

        # "mimic" that the cal_obs1_subtask finished (including qa subtasks)
        for subtask in cal_obs1_task['subtasks']:
            set_subtask_state_following_allowed_transitions(subtask['id'], 'finished')
        self.check_statuses(cal_obs1_subtask['id'], "finished", "finished", "observing")


        # -------------------
        # schedule target obs
        self.assertEqual(1, len([st for st in target_obs_task['subtasks'] if st['subtask_type'] == 'observation']))
        target_obs_subtask = next(st for st in target_obs_task['subtasks'] if st['subtask_type'] == 'observation')
        target_obs_subtask = self.tmss_client.schedule_subtask(target_obs_subtask['id'])
        check_parset(target_obs_subtask, is_target_obs=True)
        self.check_statuses(target_obs_subtask['id'], "scheduled", "scheduled", "observing")

        # check output_dataproducts
        target_obs_output_dataproducts = self.tmss_client.get_subtask_output_dataproducts(target_obs_subtask['id'])
        self.assertEqual(486, len(target_obs_output_dataproducts))

        # should now be used and scheduled, but not running/finished yet.
        self.assertTrue(tgtobs_task_template.is_used_in_unfinished_tasks_blueprints)
        self.assertTrue(datetime_template.is_used_in_unfinished_tasks_blueprints)
        self.assertTrue(strategy_template.is_used_in_unfinised_schedulinging_unit_blueprints)

        # "mimic" that the target_obs_subtask starts running
        set_subtask_state_following_allowed_transitions(target_obs_subtask['id'], 'started')
        self.check_statuses(target_obs_subtask['id'], "started", "started", "observing")

        # "mimic" that the target_obs_subtask finished (including qa subtasks)
        for subtask in target_obs_task['subtasks']:
            set_subtask_state_following_allowed_transitions(subtask['id'], 'finished')
        self.check_statuses(target_obs_subtask['id'], "finished", "finished", "observing")

        # should now be used and scheduled and finished.
        self.assertFalse(tgtobs_task_template.is_used_in_unfinished_tasks_blueprints)
        # the more common datetime_template should still be used in the other observations.
        self.assertTrue(datetime_template.is_used_in_unfinished_tasks_blueprints)
        self.assertTrue(strategy_template.is_used_in_unfinised_schedulinging_unit_blueprints)


        # -------------------
        # schedule second calibrator obs
        self.assertEqual(1, len([st for st in cal_obs2_task['subtasks'] if st['subtask_type'] == 'observation']))
        cal_obs2_subtask = next(st for st in cal_obs2_task['subtasks'] if st['subtask_type'] == 'observation')
        cal_obs2_subtask = self.tmss_client.schedule_subtask(cal_obs2_subtask['id'])
        check_parset(cal_obs2_subtask, is_target_obs=False)
        self.check_statuses(cal_obs2_subtask['id'], "scheduled", "scheduled", "observing")

        # check output_dataproducts
        cal_obs2_output_dataproducts = self.tmss_client.get_subtask_output_dataproducts(cal_obs2_subtask['id'])
        self.assertEqual(243, len(cal_obs2_output_dataproducts))

        # "mimic" that the cal_obs2_subtask starts running
        set_subtask_state_following_allowed_transitions(cal_obs2_subtask['id'], 'started')
        self.check_statuses(cal_obs2_subtask['id'], "started", "started", "observing")

        # "mimic" that the cal_obs2_subtask finished (including qa subtasks)
        for subtask in cal_obs2_task['subtasks']:
            set_subtask_state_following_allowed_transitions(subtask['id'], 'finished')
        self.check_statuses(cal_obs2_subtask['id'], "finished", "finished", "observed")

        # the more common datetime_template should now not be used anymore in unfinished tasks
        self.assertFalse(datetime_template.is_used_in_unfinished_tasks_blueprints)
        self.assertTrue(strategy_template.is_used_in_unfinised_schedulinging_unit_blueprints)


        # -------------------
        # check pipelines
        cal_pipe1_task = next(t for t in pipeline_tasks if t['name'] == 'Calibrator Pipeline 1')
        target_pipe1_task = next(t for t in pipeline_tasks if t['name'] == 'Pipeline target1')
        target_pipe2_task = next(t for t in pipeline_tasks if t['name'] == 'Pipeline target2')
        cal_pipe2_task = next(t for t in pipeline_tasks if t['name'] == 'Calibrator Pipeline 2')
        # TODO: check relations between tasks


        # -------------------
        # schedule first calibrator pipeline
        self.assertEqual(1, len([st for st in cal_pipe1_task['subtasks'] if st['subtask_type'] == 'pipeline']))
        cal_pipe1_subtask = next(st for st in cal_pipe1_task['subtasks'] if st['subtask_type'] == 'pipeline')
        cal_pipe1_subtask = self.tmss_client.schedule_subtask(cal_pipe1_subtask['id'])
        self.check_statuses(cal_pipe1_subtask['id'], "scheduled", "scheduled", "observed")

        # check dataproducts
        cal_pipe1_input_dataproducts = self.tmss_client.get_subtask_input_dataproducts(cal_pipe1_subtask['id'])
        cal_pipe1_output_dataproducts = self.tmss_client.get_subtask_output_dataproducts(cal_pipe1_subtask['id'])
        self.assertEqual(243, len(cal_pipe1_input_dataproducts))
        self.assertEqual(243, len(cal_pipe1_output_dataproducts))

        # "mimic" that the cal_pipe1_subtask starts running
        set_subtask_state_following_allowed_transitions(cal_pipe1_subtask['id'], 'started')
        self.check_statuses(cal_pipe1_subtask['id'], "started", "started", "processing")

        # "mimic" that the cal_pipe1_subtask finished
        set_subtask_state_following_allowed_transitions(cal_pipe1_subtask['id'], 'finished')
        self.check_statuses(cal_pipe1_subtask['id'], "finished", "finished", "processing")


        # -------------------
        # schedule first target pipeline
        self.assertEqual(1, len([st for st in target_pipe1_task['subtasks'] if st['subtask_type'] == 'pipeline']))
        target_pipe1_subtask = next(st for st in target_pipe1_task['subtasks'] if st['subtask_type'] == 'pipeline')
        target_pipe1_subtask = self.tmss_client.schedule_subtask(target_pipe1_subtask['id'])
        self.check_statuses(target_pipe1_subtask['id'], "scheduled", "scheduled", "processing")

        # check output_dataproducts
        target_pipe1_input_dataproducts = self.tmss_client.get_subtask_input_dataproducts(target_pipe1_subtask['id'])
        target_pipe1_output_dataproducts = self.tmss_client.get_subtask_output_dataproducts(target_pipe1_subtask['id'])
        self.assertEqual(243, len(target_pipe1_input_dataproducts))
        self.assertEqual(243, len(target_pipe1_output_dataproducts))

        # "mimic" that the target_pipe1_subtask starts running
        set_subtask_state_following_allowed_transitions(target_pipe1_subtask['id'], 'started')
        self.check_statuses(target_pipe1_subtask['id'], "started", "started", "processing")

        # "mimic" that the target_pipe1_subtask finished
        set_subtask_state_following_allowed_transitions(target_pipe1_subtask['id'], 'finished')
        self.check_statuses(target_pipe1_subtask['id'], "finished", "finished", "processing")


        # -------------------
        # schedule first target pipeline
        self.assertEqual(1, len([st for st in target_pipe2_task['subtasks'] if st['subtask_type'] == 'pipeline']))
        target_pipe2_subtask = next(st for st in target_pipe2_task['subtasks'] if st['subtask_type'] == 'pipeline')
        target_pipe2_subtask = self.tmss_client.schedule_subtask(target_pipe2_subtask['id'])
        self.check_statuses(target_pipe2_subtask['id'], "scheduled", "scheduled", "processing")

        # check output_dataproducts
        target_pipe2_input_dataproducts = self.tmss_client.get_subtask_input_dataproducts(target_pipe2_subtask['id'])
        target_pipe2_output_dataproducts = self.tmss_client.get_subtask_output_dataproducts(target_pipe2_subtask['id'])
        self.assertEqual(243, len(target_pipe2_input_dataproducts))
        self.assertEqual(243, len(target_pipe2_output_dataproducts))

        # "mimic" that the target_pipe2_subtask starts running
        set_subtask_state_following_allowed_transitions(target_pipe2_subtask['id'], 'started')
        self.check_statuses(target_pipe2_subtask['id'], "started", "started", "processing")

        # "mimic" that the target_pipe2_subtask finished
        set_subtask_state_following_allowed_transitions(target_pipe2_subtask['id'], 'finished')
        self.check_statuses(target_pipe2_subtask['id'], "finished", "finished", "processing")


        # -------------------
        # schedule second calibrator pipeline
        self.assertEqual(1, len([st for st in cal_pipe2_task['subtasks'] if st['subtask_type'] == 'pipeline']))
        cal_pipe2_subtask = next(st for st in cal_pipe2_task['subtasks'] if st['subtask_type'] == 'pipeline')
        cal_pipe2_subtask = self.tmss_client.schedule_subtask(cal_pipe2_subtask['id'])
        self.check_statuses(cal_pipe2_subtask['id'], "scheduled", "scheduled", "processing")

        # check dataproducts
        cal_pipe2_input_dataproducts = self.tmss_client.get_subtask_input_dataproducts(cal_pipe2_subtask['id'])
        cal_pipe2_output_dataproducts = self.tmss_client.get_subtask_output_dataproducts(cal_pipe2_subtask['id'])
        self.assertEqual(243, len(cal_pipe2_input_dataproducts))
        self.assertEqual(243, len(cal_pipe2_output_dataproducts))

        # "mimic" that the cal_pipe2_subtask starts running
        set_subtask_state_following_allowed_transitions(cal_pipe2_subtask['id'], 'started')
        self.check_statuses(cal_pipe2_subtask['id'], "started", "started", "processing")

        # "mimic" that the cal_pipe2_subtask finished
        set_subtask_state_following_allowed_transitions(cal_pipe2_subtask['id'], 'finished')
        self.check_statuses(cal_pipe2_subtask['id'], "finished", "finished", "processed")

        # observations and pipelines finished, but ingest&cleanup not yet,
        # so the template should indicate that it's still used
        self.assertTrue(strategy_template.is_used_in_unfinised_schedulinging_unit_blueprints)

        # cancel the rest of the tasks (ingst, cleanup)
        self.tmss_client.cancel_scheduling_unit_blueprint(scheduling_unit_blueprint['id'])

        # observations and pipelines finished, ingest&cleanup are cancelled
        # so the template should indicate that it's not used anymore
        self.assertFalse(strategy_template.is_used_in_unfinised_schedulinging_unit_blueprints)



    def test_lba_survey(self):
        def check_parset(obs_subtask):
            # todo: check that these values make sense!
            '''helper function to check the parset for UC1 target/calibrator observations'''
            obs_parset = parameterset.fromString(self.tmss_client.get_subtask_parset(obs_subtask['id'])).dict()
            self.assertEqual(obs_subtask['id'], int(obs_parset['Observation.ObsID']))
            self.assertEqual('LBA', obs_parset['Observation.antennaArray'])
            self.assertEqual('LBA_SPARSE_EVEN', obs_parset['Observation.antennaSet'])
            self.assertEqual('LBA_30_90', obs_parset['Observation.bandFilter'])
            self.assertEqual(1, int(obs_parset['Observation.nrAnaBeams']))
            self.assertEqual(4, int(obs_parset['Observation.nrBeams']))
            self.assertEqual('Observation', obs_parset['Observation.processType'])
            self.assertEqual('Beam Observation', obs_parset['Observation.processSubtype'])
            self.assertEqual(parser.parse(obs_subtask['scheduled_start_time']), parser.parse(obs_parset['Observation.startTime']))
            self.assertEqual(parser.parse(obs_subtask['scheduled_stop_time']), parser.parse(obs_parset['Observation.stopTime']))
            self.assertEqual(200, int(obs_parset['Observation.sampleClock']))
            self.assertEqual(122, len(obs_parset['Observation.Beam[0].subbandList'].split(',')))
            self.assertEqual(True, strtobool(obs_parset['Observation.DataProducts.Output_Correlated.enabled']))
            self.assertEqual(488, len(obs_parset['Observation.DataProducts.Output_Correlated.filenames'].split(',')))
            self.assertEqual(488, len(obs_parset['Observation.DataProducts.Output_Correlated.locations'].split(',')))
            self.assertEqual(False, strtobool(obs_parset.get('Observation.DataProducts.Output_CoherentStokes.enabled', 'false')))
            self.assertEqual(False, strtobool(obs_parset.get('Observation.DataProducts.Output_IncoherentStokes.enabled', 'false')))
            self.assertEqual(False, strtobool(obs_parset.get('Observation.DataProducts.Output_Pulsar.enabled', 'false')))
            self.assertEqual(False, strtobool(obs_parset.get('Observation.DataProducts.Output_InstrumentModel.enabled', 'false')))
            self.assertEqual(False, strtobool(obs_parset.get('Observation.DataProducts.Output_SkyImage.enabled', 'false')))

            # import helper method to cycle through allowed state transitions

        from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions

        # setup: create a scheduling unit from the UC1 observation strategy template
        uc1_strategy_template = self.tmss_client.get_scheduling_unit_observing_strategy_template('IM LBA Survey - 3 Beams', 38)

        scheduling_unit_draft = self.tmss_client.create_scheduling_unit_draft_from_strategy_template(
            uc1_strategy_template['id'], self.scheduling_set['id'])
        # check general object settings after creation
        self.assertEqual(uc1_strategy_template['url'], scheduling_unit_draft['observation_strategy_template'])
        self.assertFalse(scheduling_unit_draft['ingest_permission_required'])

        # TODO: check draft specification, constraints, etc according to LBA survey requirements like antennaset, filters, subbands, etc.
        # for now, just check if the spec is ok according to schema.
        specifications_doc = self.tmss_client.get_schedulingunit_draft_specifications_doc(scheduling_unit_draft['id'])
        validate_json_against_its_schema(specifications_doc)

        scheduling_unit_blueprint = self.tmss_client.create_scheduling_unit_blueprint_and_tasks_and_subtasks_tree(scheduling_unit_draft['id'])
        scheduling_unit_blueprint_ext = self.tmss_client.get_schedulingunit_blueprint(scheduling_unit_blueprint['id'], extended=True)
        self.assertFalse(scheduling_unit_blueprint_ext['ingest_permission_required'])

        # blueprint spec should be copied verbatim, so should be equal to (unchanged/unedited) draft
        self.assertEqual(specifications_doc, self.tmss_client.get_schedulingunit_blueprint_specifications_doc(scheduling_unit_blueprint['id']))

        # observation(s) did not run yet, so observed_stop_time should be None
        self.assertIsNone(scheduling_unit_blueprint_ext['observed_stop_time'])
        self.assertEqual("schedulable", scheduling_unit_blueprint_ext['status'])

        # check the tasks
        tasks = scheduling_unit_blueprint_ext['task_blueprints']
        self.assertEqual(7, len(tasks))
        observation_tasks = [t for t in tasks if t['task_type'] == 'observation']
        self.assertEqual(1, len(observation_tasks))
        pipeline_tasks = [t for t in tasks if t['task_type'] == 'pipeline']
        self.assertEqual(4, len(pipeline_tasks))
        self.assertEqual(1, len([t for t in tasks if t['task_type'] == 'ingest']))
        ingest_task = next(t for t in tasks if t['task_type'] == 'ingest')

        obs_task = next(t for t in observation_tasks if t['name'] == 'Combined Observation')

        # -------------------
        # schedule combined calibrator and target obs subtask
        self.assertEqual(1, len([st for st in obs_task['subtasks'] if st['subtask_type'] == 'observation']))
        obs_subtask = next(st for st in obs_task['subtasks'] if st['subtask_type'] == 'observation')
        obs_subtask = self.tmss_client.schedule_subtask(obs_subtask['id'])

        check_parset(obs_subtask)
        self.check_statuses(obs_subtask['id'], "scheduled", "scheduled", "scheduled")

        # check output_dataproducts
        obs_output_dataproducts = self.tmss_client.get_subtask_output_dataproducts(obs_subtask['id'])
        self.assertEqual(488, len(obs_output_dataproducts))

        # "mimic" that the cal_obs_subtask starts running
        set_subtask_state_following_allowed_transitions(obs_subtask['id'], 'started')
        self.check_statuses(obs_subtask['id'], "started", "started", "observing")

        # "mimic" that all obs_task subtask finished (including qa subtasks)
        for subtask in obs_task['subtasks']:
            set_subtask_state_following_allowed_transitions(subtask['id'], 'finished')

        # refresh all subtasks/tasks/schedunits for latest values
        obs_subtask = self.tmss_client.get_subtask(obs_subtask['id'])
        self.assertEqual("finished", obs_subtask['state_value'])
        obs_task = self.tmss_client.get_url_as_json_object(obs_task['url'])
        self.assertEqual("finished", obs_task['status'])
        scheduling_unit_blueprint = self.tmss_client.get_url_as_json_object(scheduling_unit_blueprint['url'])
        self.assertEqual("observed", scheduling_unit_blueprint['status'])

        # -------------------
        # check pipelines
        cal_pipe_task = next(t for t in pipeline_tasks if t['name'] == 'Calibrator Pipeline')
        target_pipe1_task = next(t for t in pipeline_tasks if t['name'] == 'Pipeline target1')
        target_pipe2_task = next(t for t in pipeline_tasks if t['name'] == 'Pipeline target2')
        target_pipe3_task = next(t for t in pipeline_tasks if t['name'] == 'Pipeline target3')
        # TODO: check relations between tasks

        # -------------------
        # schedule calibrator pipeline
        self.assertEqual(1, len([st for st in cal_pipe_task['subtasks'] if st['subtask_type'] == 'pipeline']))
        cal_pipe_subtask = next(st for st in cal_pipe_task['subtasks'] if st['subtask_type'] == 'pipeline')
        cal_pipe_subtask = self.tmss_client.schedule_subtask(cal_pipe_subtask['id'])
        self.check_statuses(cal_pipe_subtask['id'], "scheduled", "scheduled", "observed")

        # check dataproducts
        cal_pipe1_input_dataproducts = self.tmss_client.get_subtask_input_dataproducts(cal_pipe_subtask['id'])
        cal_pipe1_output_dataproducts = self.tmss_client.get_subtask_output_dataproducts(cal_pipe_subtask['id'])
        self.assertEqual(122, len(cal_pipe1_input_dataproducts))
        self.assertEqual(122, len(cal_pipe1_output_dataproducts))

        # "mimic" that the cal_pipe_subtask starts running
        set_subtask_state_following_allowed_transitions(cal_pipe_subtask['id'], 'started')
        self.check_statuses(cal_pipe_subtask['id'], "started", "started", "processing")

        # "mimic" that the cal_pipe_subtask finished
        set_subtask_state_following_allowed_transitions(cal_pipe_subtask['id'], 'finished')
        self.check_statuses(cal_pipe_subtask['id'], "finished", "finished", "processing")

        # -------------------
        # schedule first target pipeline
        self.assertEqual(1, len([st for st in target_pipe1_task['subtasks'] if st['subtask_type'] == 'pipeline']))
        target_pipe1_subtask = next(st for st in target_pipe1_task['subtasks'] if st['subtask_type'] == 'pipeline')
        target_pipe1_subtask = self.tmss_client.schedule_subtask(target_pipe1_subtask['id'])
        self.check_statuses(target_pipe1_subtask['id'], "scheduled", "scheduled", "processing")

        # check output_dataproducts
        target_pipe1_input_dataproducts = self.tmss_client.get_subtask_input_dataproducts(target_pipe1_subtask['id'])
        target_pipe1_output_dataproducts = self.tmss_client.get_subtask_output_dataproducts(target_pipe1_subtask['id'])
        self.assertEqual(122, len(target_pipe1_input_dataproducts))
        self.assertEqual(122, len(target_pipe1_output_dataproducts))

        # "mimic" that the target_pipe1_subtask starts running
        set_subtask_state_following_allowed_transitions(target_pipe1_subtask['id'], 'started')
        self.check_statuses(target_pipe1_subtask['id'], "started", "started", "processing")

        # "mimic" that the target_pipe1_subtask finished
        set_subtask_state_following_allowed_transitions(target_pipe1_subtask['id'], 'finished')
        self.check_statuses(target_pipe1_subtask['id'], "finished", "finished", "processing")

        # -------------------
        # schedule second target pipeline
        self.assertEqual(1, len([st for st in target_pipe2_task['subtasks'] if st['subtask_type'] == 'pipeline']))
        target_pipe2_subtask = next(st for st in target_pipe2_task['subtasks'] if st['subtask_type'] == 'pipeline')
        target_pipe2_subtask = self.tmss_client.schedule_subtask(target_pipe2_subtask['id'])
        self.check_statuses(target_pipe2_subtask['id'], "scheduled", "scheduled", "processing")

        # check output_dataproducts
        target_pipe2_input_dataproducts = self.tmss_client.get_subtask_input_dataproducts(target_pipe2_subtask['id'])
        target_pipe2_output_dataproducts = self.tmss_client.get_subtask_output_dataproducts(target_pipe2_subtask['id'])
        self.assertEqual(122, len(target_pipe2_input_dataproducts))
        self.assertEqual(122, len(target_pipe2_output_dataproducts))

        # "mimic" that the target_pipe2_subtask starts running
        set_subtask_state_following_allowed_transitions(target_pipe2_subtask['id'], 'started')
        self.check_statuses(target_pipe2_subtask['id'], "started", "started", "processing")

        # "mimic" that the target_pipe2_subtask finished
        set_subtask_state_following_allowed_transitions(target_pipe2_subtask['id'], 'finished')
        self.check_statuses(target_pipe2_subtask['id'], "finished", "finished", "processing")

        # -------------------
        # schedule third target pipeline
        self.assertEqual(1, len([st for st in target_pipe3_task['subtasks'] if st['subtask_type'] == 'pipeline']))
        target_pipe3_subtask = next(st for st in target_pipe3_task['subtasks'] if st['subtask_type'] == 'pipeline')
        target_pipe3_subtask = self.tmss_client.schedule_subtask(target_pipe3_subtask['id'])
        self.check_statuses(target_pipe3_subtask['id'], "scheduled", "scheduled", "processing")

        # check output_dataproducts
        target_pipe3_input_dataproducts = self.tmss_client.get_subtask_input_dataproducts(target_pipe3_subtask['id'])
        target_pipe3_output_dataproducts = self.tmss_client.get_subtask_output_dataproducts(target_pipe3_subtask['id'])
        self.assertEqual(122, len(target_pipe3_input_dataproducts))
        self.assertEqual(122, len(target_pipe3_output_dataproducts))

        # "mimic" that the target_pipe3_subtask starts running
        set_subtask_state_following_allowed_transitions(target_pipe3_subtask['id'], 'started')
        self.check_statuses(target_pipe3_subtask['id'], "started", "started", "processing")

        # "mimic" that the target_pipe3_subtask finished
        set_subtask_state_following_allowed_transitions(target_pipe3_subtask['id'], 'finished')
        self.check_statuses(target_pipe3_subtask['id'], "finished", "finished", "processed")

    def test_BF_CV_FRB_strategy(self):
        from lofar.sas.tmss.tmss.tmssapp.models import SchedulingUnitObservingStrategyTemplate

        # setup: create a scheduling unit from the 'BF CV FRB' observation strategy template
        template = SchedulingUnitObservingStrategyTemplate.get_latest('BF CV FRB')
        self.assertIsNotNone(template)

        scheduling_unit_draft = self.tmss_client.create_scheduling_unit_draft_from_strategy_template(template.id, self.scheduling_set['id'])
        self.assertIsNotNone(scheduling_unit_draft)
        # TODO: add additional checks for BF CV FRB once fully developed

    @unittest.skip("skipped because this is currently not an active template")
    def test_COM_COBALT2_BF_test_strategy_but_with_tab_rings(self):
        def check_parset(obs_subtask):
            '''helper function to check the parset for COM COBALT2 BF test observations'''
            obs_parset = parameterset.fromString(self.tmss_client.get_subtask_parset(obs_subtask['id'])).dict()

            # todo: add further checks for parset correctnes

        from lofar.sas.tmss.tmss.tmssapp.models import SchedulingUnitObservingStrategyTemplate, TaskDraft

        # setup: create a scheduling unit from the 'COM COBALT2 BF test' observation strategy template
        template = SchedulingUnitObservingStrategyTemplate.get_latest('COM COBALT2 BF test')
        self.assertIsNotNone(template)

        scheduling_unit_draft = self.tmss_client.create_scheduling_unit_draft_from_strategy_template(template.id, self.scheduling_set['id'])
        # change tab rings setting from strategy on the task draft
        task_draft = TaskDraft.objects.get(scheduling_unit_draft__id=scheduling_unit_draft['id'])
        task_draft.specifications_doc['beamformer']['pipelines'][0]['coherent']['SAPs'][0]['tab_rings']['count'] = 5
        task_draft.save()
        scheduling_unit_blueprint = self.tmss_client.create_scheduling_unit_blueprint_and_tasks_and_subtasks_tree(scheduling_unit_draft['id'])
        scheduling_unit_blueprint_ext = self.tmss_client.get_schedulingunit_blueprint(scheduling_unit_blueprint['id'], extended=True, include_specifications_doc=True)

        # check the tasks
        tasks = scheduling_unit_blueprint_ext['task_blueprints']
        observation_tasks = [t for t in tasks if t['task_type'] == 'observation']
        for obs_task in [t for t in observation_tasks if t['name'] == 'Observation']:
            for obs_subtask in [st for st in obs_task['subtasks'] if st['subtask_type'] == 'observation']:
                obs_subtask = self.tmss_client.schedule_subtask(obs_subtask['id'])
                check_parset(obs_subtask)

        # TODO: add additional checks for 'COM COBALT2 BF test' once fully developed

    @unittest.skip("no such template in git. still in development by sdco?")
    def test_IM_BF_2_CV_8_bit(self):
        def check_parset(obs_subtask):
            '''helper function to check the parset for the test observations'''
            obs_parset = parameterset.fromString(self.tmss_client.get_subtask_parset(obs_subtask['id'])).dict()

            # todo: add further checks for parset correctnes

        from lofar.sas.tmss.tmss.tmssapp.models import SchedulingUnitObservingStrategyTemplate, TaskDraft

        # setup: create a scheduling unit from the 'COM COBALT2 BF test' observation strategy template
        template = SchedulingUnitObservingStrategyTemplate.get_latest('IM + BF 2 CV - 8 bit')
        self.assertIsNotNone(template)

        scheduling_unit_draft = self.tmss_client.create_scheduling_unit_draft_from_strategy_template(template.id, self.scheduling_set['id'])
        scheduling_unit_blueprint = self.tmss_client.create_scheduling_unit_blueprint_and_tasks_and_subtasks_tree(scheduling_unit_draft['id'])
        scheduling_unit_blueprint_ext = self.tmss_client.get_schedulingunit_blueprint(scheduling_unit_blueprint['id'], extended=True, include_specifications_doc=True)

        # check the tasks
        tasks = scheduling_unit_blueprint_ext['task_blueprints']
        observation_tasks = [t for t in tasks if t['task_type'] == 'observation']
        for obs_task in [t for t in observation_tasks if t['name'] == 'Observation']:
            for obs_subtask in [st for st in obs_task['subtasks'] if st['subtask_type'] == 'observation']:
                obs_subtask = self.tmss_client.schedule_subtask(obs_subtask['id'])
                check_parset(obs_subtask)

        # TODO: add additional checks for 'COM COBALT2 BF test' once fully developed


logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

if __name__ == '__main__':
    unittest.main()
