#!/usr/bin/env python3
import unittest

import logging
logger = logging.getLogger('lofar.'+__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

from lofar.common.test_utils import exit_with_skipped_code_if_skip_regression_tests, assertEqualXML

exit_with_skipped_code_if_skip_regression_tests()

import json
import os
import re
import shutil
from datetime import datetime, timedelta
from uuid import uuid4
import xml.etree.ElementTree as ET

from lofar.sas.tmss.test.test_environment import TMSSTestEnvironment
tmss_test_env = TMSSTestEnvironment()
try:
    tmss_test_env.start()
    tmss_test_env.populate_schemas_and_connectors()
except Exception as e:
    logger.exception(str(e))
    tmss_test_env.stop()
    exit(1)

# tell unittest to stop (and automagically cleanup) the test database once all testing is done.
def tearDownModule():
    tmss_test_env.stop()

from lofar.sas.tmss.tmss.settings import BASE_DIR

from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.tmssapp.adapters.parset import convert_to_parset
from lofar.sas.tmss.tmss.tmssapp.adapters.sip import get_sip_xml_for_dataproduct
from lofar.sas.tmss.tmss.tmssapp.tasks import create_scheduling_unit_draft_from_observing_strategy_template, create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft
from lofar.sas.tmss.tmss.tmssapp.subtasks import schedule_subtask

from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions
from lofar.sas.tmss.test.tmss_test_data_django_models import Project_test_data, SchedulingSet_test_data


class TestObservingStrategiesRegression(unittest.TestCase):
    '''The purpose of this test is to prove correctness of the specified and scheduled observations, pipelines and
    other (sub)tasks by checking the resulting statuses, the created subtask-specification_docs, parsets and dataproducts.
    For this test we regard TMSS and the services as a black box,
    and we can only use the http rest api (via the tmss_client) to specify, schedule and check the results.
    '''
    @classmethod
    def setUpClass(cls) -> None:
        # enable large diffs in test framework to compare the large parsets and sips
        cls.maxDiff = None

        # make sure the base directories for the dataproducts are the same as in the gold-standard-parsets
        for fs in models.Filesystem.objects.all():
            fs.directory = "/data/test-projects/"
            fs.save()

        # we should not depend on "previous" data
        models.SubtaskInput.objects.all().delete()
        models.DataproductHash.objects.all().delete()
        models.DataproductArchiveInfo.objects.all().delete()
        models.DataproductTransform.objects.all().delete()
        models.Dataproduct.objects.all().delete()
        models.Subtask.objects.all().delete()

        # Create project and scheduling set for testing
        cls.project = models.Project.objects.create(**Project_test_data(name="test observing strategies regression", description="test observing strategies regression"))
        cls.scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=cls.project))

    def setUp(self) -> None:
        tmss_test_env.delete_scheduling_unit_drafts_cascade()

    def _create_SUB_and_tasks_and_subtasks_from_observing_strategy_template(self, strategy_template: models.SchedulingUnitObservingStrategyTemplate) -> models.SchedulingUnitBlueprint:
        """
        Help method to create a scheduling unit blueprint from an observing strategy template, including tasks and subtasks.
        :param strategy_template: the observing strategy template from which create the SUB.
        :return: the created SUB.
        """
        scheduling_unit_draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template=strategy_template, scheduling_set=self.scheduling_set)
        logger.info("Creating scheduling_unit_blueprint from draft for strategy_template '%s' version=%s", strategy_template.name, strategy_template.version)
        return create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

    def _set_fixed_scheduled_and_actual_process_times_and_schedule_to_finishing_state(self, subtask: models.Subtask, start_time=datetime(3000, 2, 16, 17, 0, 0)):
        duration = 600 if subtask.specifications_template.type.value == models.SubtaskType.Choices.OBSERVATION.value else 300
        stop_time = start_time + timedelta(seconds=duration)

        # Force scheduled_start/stop_time to be manually set.
        subtask.scheduled_start_time = start_time
        subtask.scheduled_stop_time = stop_time
        subtask.save()

        subtask = schedule_subtask(subtask)

        # Trick to force actual_process_start/stop_time to be manually set.
        set_subtask_state_following_allowed_transitions(subtask, models.SubtaskState.Choices.QUEUED.value)
        subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.STARTING.value)
        subtask.actual_process_start_time = start_time
        subtask.save()
        subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.STARTED.value)
        subtask.actual_process_stop_time = stop_time
        subtask.save()
        set_subtask_state_following_allowed_transitions(subtask, models.SubtaskState.Choices.FINISHING.value)

    def _sanitise_parset_dict(self, parset):
        """
        Help method to sanitise some fields in the parset that might dynamically change
        :param parset:
        :return: the sanitised parset
        """
        sanitised_parset = str(parset)
        sanitised_parset = re.sub('ID=[0-9]{7}', 'ID=2000000', sanitised_parset)
        sanitised_parset = re.sub('L[0-9]{7}', 'L2000000', sanitised_parset)
        sanitised_parset = re.sub('SAP[0-9]{3}', 'SAP000', sanitised_parset)
        sanitised_parset = re.sub('B[0-9]{3}', 'B000', sanitised_parset)
        sanitised_parset = re.sub('SB[0-9]{3}', 'SB000', sanitised_parset)
        sanitised_parset = re.sub('P[0-9]{3}', 'P000', sanitised_parset)
        sanitised_parset = re.sub('TMSS_subtask_[0-9]{7}\.SAP[0-9]{3}', 'TMSS_subtask_2000000.SAP000', sanitised_parset)

        header = "# THIS PARSET WAS GENERATED BY TMSS FROM THE SPECIFICATION OF SUBTASK ID=2000000 ON 2022-06-16 12:00:0\n"
        sanitised_parset = header + sanitised_parset.replace('"', '').replace("'", "")  # remove quotes

        return sanitised_parset

    def _sanitise_SIPs_xml(self, SIP_xml):
        """
        Help method to sanitise some fields in the SIPs that might dynamically change
        :param SIP_xml:
        :return: the sanitised SIP_xml
        """
        etree = ET.fromstring(SIP_xml)
        for e in etree.iter():
            if e.tag == 'identifier':
                children = list(e)
                if len(children) > 0:
                    e.text = '00000'
                else:
                    e.clear()
            elif e.tag == 'fileName':
                e.text = 'L2000000_SAP000_SB000_uv.MS'
            elif e.tag == 'measurementDescription':
                e.text = re.sub('L[0-9]{7}', 'L2000000', e.text)
            elif e.tag == 'parset':
                e.clear()
            elif e.tag == 'x':
                e.text = '380000.111'
            elif e.tag == 'y':
                e.text = '460000.222'
            elif e.tag == 'z':
                e.text = '5000000.333'

        return ET.tostring(etree).decode()

    def _get_SIPs_XML_and_parset_from_subtasks_in_SUB(self, scheduling_unit_blueprint: models.SchedulingUnitBlueprint):
        """
        Help method to retrieve SIPs and parsets from subtasks belonging to a scheduling unit blueprint.
        :param scheduling_unit_blueprint: the SUB to which the obs subtasks belong to.
        :return: a list of dicts with SIPs and parsets per subtask.
        """
        results = []

        obs_subtasks = models.Subtask.objects.filter(task_blueprint__scheduling_unit_blueprint=scheduling_unit_blueprint, specifications_template__type__value=models.SubtaskType.Choices.OBSERVATION.value)
        pip_subtasks = models.Subtask.objects.filter(task_blueprint__scheduling_unit_blueprint=scheduling_unit_blueprint, specifications_template__type__value=models.SubtaskType.Choices.PIPELINE.value)
        for subtasks in [obs_subtasks, pip_subtasks]:
            for subtask in subtasks:
                # Simulate the run of the subtask till the FINISHING state by scheduling the subtask.
                # Then we force the manual set of scheduled/actual_process times to avoid timestamp inconsistencies.
                self._set_fixed_scheduled_and_actual_process_times_and_schedule_to_finishing_state(subtask)

                # Add feedback_doc and feedback_template to dataproducts
                feedback_template = models.DataproductFeedbackTemplate.get_latest(name='feedback')
                if subtask.specifications_template.type.value == models.SubtaskType.Choices.PIPELINE.value:
                    feedback_doc = {'time':{'duration':120.16680908203125,'start_time':'3000-02-16T17:00:00','sample_width':4.00556032},'files':[],'target':{'coherent':True,'pointing':{'angle1':0.662432,'angle2':1.557953,'target':'Paaa+01','direction_type':'J2000'}},'$schema':'https://tmss.lofar.eu/api/schemas/dataproductfeedbacktemplate/feedback/1#','samples':{'bits':32,'type':'float','writer':'dysco','complex':True,'polarisations':['XX','XY','YX','YY'],'writer_version':'v1.2'},'antennas':{'set':'LBA_SPARSE_EVEN','fields':[{'type':'LBA','field':'LBA','station':'CS001'},{'type':'LBA','field':'LBA','station':'CS002'},{'type':'LBA','field':'LBA','station':'CS003'},{'type':'LBA','field':'LBA','station':'CS004'},{'type':'LBA','field':'LBA','station':'CS005'},{'type':'LBA','field':'LBA','station':'CS006'},{'type':'LBA','field':'LBA','station':'CS007'},{'type':'LBA','field':'LBA','station':'CS011'},{'type':'LBA','field':'LBA','station':'CS013'},{'type':'LBA','field':'LBA','station':'CS017'},{'type':'LBA','field':'LBA','station':'CS021'},{'type':'LBA','field':'LBA','station':'CS024'},{'type':'LBA','field':'LBA','station':'CS026'},{'type':'LBA','field':'LBA','station':'CS028'},{'type':'LBA','field':'LBA','station':'CS030'},{'type':'LBA','field':'LBA','station':'CS031'},{'type':'LBA','field':'LBA','station':'CS032'},{'type':'LBA','field':'LBA','station':'CS101'},{'type':'LBA','field':'LBA','station':'CS103'},{'type':'LBA','field':'LBA','station':'CS201'},{'type':'LBA','field':'LBA','station':'CS301'},{'type':'LBA','field':'LBA','station':'CS302'},{'type':'LBA','field':'LBA','station':'CS401'},{'type':'LBA','field':'LBA','station':'CS501'},{'type':'LBA','field':'LBA','station':'RS106'},{'type':'LBA','field':'LBA','station':'RS205'},{'type':'LBA','field':'LBA','station':'RS208'},{'type':'LBA','field':'LBA','station':'RS210'},{'type':'LBA','field':'LBA','station':'RS305'},{'type':'LBA','field':'LBA','station':'RS306'},{'type':'LBA','field':'LBA','station':'RS307'},{'type':'LBA','field':'LBA','station':'RS310'},{'type':'LBA','field':'LBA','station':'RS406'},{'type':'LBA','field':'LBA','station':'RS407'},{'type':'LBA','field':'LBA','station':'RS409'},{'type':'LBA','field':'LBA','station':'RS503'},{'type':'LBA','field':'LBA','station':'RS508'},{'type':'LBA','field':'LBA','station':'RS509'}]},'frequency':{'subbands':[329],'channel_width':24414.0625,'beamlet_indices':[0],'central_frequencies':[64257812.5],'channels_per_subband':8},'percentage_written':100}
                else:
                    feedback_doc = {'percentage_written':100,'frequency':{'subbands':[156],'central_frequencies':[33593750],'channel_width':6103.515625,'channels_per_subband':32},'time':{'start_time':'3000-02-16T17:00:00','duration':5.02732992172,'sample_width':2.00278016},'antennas':{'set':'HBA_DUAL','fields':[{'type':'HBA','field':'HBA0','station':'CS001'},{'type':'HBA','field':'HBA1','station':'CS001'}]},'target':{'pointing':{'angle1':0,'angle2':0,'direction_type':'J2000','target':'target1'}},'samples':{'polarisations':['XX','XY','YX','YY'],'type':'float','bits':32,'writer':'standard','writer_version':'2.2.0','complex':True},'$schema':'http://127.0.0.1:8001/api/schemas/dataproductfeedbacktemplate/feedback/1#'}
                feedback_doc = feedback_template.add_defaults_to_json_object_for_schema(feedback_doc)
                for dp in subtask.output_dataproducts.all():
                    dp.feedback_template = feedback_template
                    dp.feedback_doc = feedback_doc
                    dp.save()
                set_subtask_state_following_allowed_transitions(subtask, models.SubtaskState.Choices.FINISHED.value)

                # Aggregate SIPs and parsets
                result = {'task_blueprint': subtask.task_blueprint.name,
                          'SIP_xml': get_sip_xml_for_dataproduct(subtask.output_dataproducts.last()),
                          'parset': convert_to_parset(subtask)}
                results.append(result)

        return results

    def apply_regression_test_to_observing_strategy(self, test_strategy_name:str, sips_and_parsets:list):
        """
        Help method to apply regression test to a specific observing strategy template.
        :param test_strategy_name:
        :param sips_and_parset:
        """
        # Assert SIPs and parsets with expected values
        for sip_and_parset in sips_and_parsets:
            # Assert SIPs
            # Firstly sanitise some strings as they might dynamically change
            test_sip = self._sanitise_SIPs_xml(sip_and_parset.get('SIP_xml'))
            # Assert the sanitised SIPs
            with open(f'{os.path.join(BASE_DIR, f"../test/regressions_test/{test_strategy_name}/")}{test_strategy_name}_{sip_and_parset.get("task_blueprint")}.sip.xml', 'r') as f:
                expected_sip = f.read()
            assertEqualXML(test_sip, expected_sip)

            # Assert parsets
            test_parset = self._sanitise_parset_dict(sip_and_parset.get('parset'))
            with open(f'{os.path.join(BASE_DIR, f"../test/regressions_test/{test_strategy_name}/")}{test_strategy_name}_{sip_and_parset.get("task_blueprint")}.parset', 'r') as f:
                expected_parset = f.read()
            self.assertEqual(test_parset, expected_parset)

    def test_IM_LBA_Survey_3_Beams(self):
        """
        Regression test for the "IM LBA Survey - 3 Beams" observing strategy template.
        """

        # Create SUB and tasks and subtasks from strategy_template
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name='IM LBA Survey - 3 Beams')
        scheduling_unit_blueprint = self._create_SUB_and_tasks_and_subtasks_from_observing_strategy_template(strategy_template)

        # Get SIPs and parsets from the subtasks
        sips_and_parsets = self._get_SIPs_XML_and_parset_from_subtasks_in_SUB(scheduling_unit_blueprint)
        # Assert regression test
        self.apply_regression_test_to_observing_strategy('test_IM_LBA_Survey_3_Beams', sips_and_parsets)

    def test_BF_Pulsar_Timing(self):
        """
        Regression test for the "BF Pulsar Timing" observing strategy template.
        """

        # Create SUB and tasks and subtasks from strategy_template
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name='BF Pulsar Timing')
        scheduling_unit_blueprint = self._create_SUB_and_tasks_and_subtasks_from_observing_strategy_template(strategy_template)

        # Get SIPs and parsets from the subtasks
        sips_and_parsets = self._get_SIPs_XML_and_parset_from_subtasks_in_SUB(scheduling_unit_blueprint)
        # Assert regression test
        self.apply_regression_test_to_observing_strategy('test_BF_Pulsar_Timing', sips_and_parsets)

if __name__ == '__main__':
    unittest.main()
