#!/usr/bin/env python3
import datetime
# Copyright (C) 2018    ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

import os
import time
import unittest
import requests
from datetime import datetime

import logging
logger = logging.getLogger('lofar.'+__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)


from lofar.common.test_utils import exit_with_skipped_code_if_skip_integration_tests
exit_with_skipped_code_if_skip_integration_tests()

# Do Mandatory setup step:
# use setup/teardown magic for tmss test database, ldap server and django server
# (ignore pycharm unused import statement, python unittests does use at RunTime the tmss_test_environment_unittest_setup module)

from lofar.sas.tmss.test.test_environment import TMSSTestEnvironment
tmss_test_env = TMSSTestEnvironment()
try:
    tmss_test_env.start()
    tmss_test_env.populate_schemas_and_connectors()
except Exception as e:
    logger.exception(str(e))
    tmss_test_env.stop()
    exit(1)

# tell unittest to stop (and automagically cleanup) the test database once all testing is done.
def tearDownModule():
    tmss_test_env.stop()

from lofar.sas.tmss.test.tmss_test_data_django_models import *

from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.tmssapp.tasks import create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft, cancel_scheduling_unit_blueprint
from lofar.sas.tmss.tmss.tmssapp.subtasks import schedule_subtask
from lofar.common.json_utils import get_default_json_object_for_schema
from dateutil.relativedelta import relativedelta
from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions

from django.contrib.auth import get_user_model
User = get_user_model()


class ReportTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        tmss_test_env.delete_scheduling_unit_drafts_cascade()
        tmss_test_env.delete_reservations_cascade()

        # Create requirements
        # Project report setup
        cls.test_project_report = models.Project.objects.create(**Project_test_data(name='test_project_report', auto_ingest=True))
        cls.project_quota = models.ProjectQuota.objects.create(
            **ProjectQuota_test_data(project=cls.test_project_report, resource_type=models.ResourceType.objects.create(
                **ResourceType_test_data(quantity=models.Quantity.objects.get(value=models.Quantity.Choices.NUMBER.value)))))
        cls.setup_project_report(cls.test_project_report)

        # Cycle report setup
        cls.cycle = models.Cycle.objects.create(**Cycle_test_data(start=datetime.utcnow().isoformat(), stop=(datetime.utcnow() + timedelta(weeks=12)).isoformat()))

        project_unassigned = models.Project.objects.create(**Project_test_data(name='unassigned', auto_ingest=True))
        project_unassigned.cycles.set([cls.cycle.pk])

        project_filler = models.Project.objects.create(**Project_test_data(name='filler', auto_ingest=True))
        project_filler.cycles.set([cls.cycle.pk])
        project_filler.filler = True
        project_filler.save()

        cls.projects = {'unassigned': project_unassigned, 'filler': project_filler}
        for project_category in [c.value for c in models.ProjectCategory.Choices]:
            project = models.Project.objects.create(**Project_test_data(name=project_category, auto_ingest=True))
            project.cycles.set([cls.cycle.pk])
            project.project_category = models.ProjectCategory.objects.get(value=project_category)
            project.save()
            cls.projects[project_category] = project

        # Setup requirements for each project
        cycle_start = datetime.strptime(cls.cycle.start, '%Y-%m-%dT%H:%M:%S.%f')
        for p in cls.projects.values():
            if p.name in ('regular', 'commissioning'):
                cls.setup_project_report(p)
            elif p.name in ('unassigned'):
                cls._create_simple_finished_observation_SUB_for_project(p)
            elif p.name in ('test'):
                for i in range(3):
                    # 0: whole day but no efficiency (week 1); 1: whole day 100% efficiency (week 2); 2: between weeks (12 hours week 3, 12 hours week 4) 100% efficiency each week
                    step = timedelta(weeks=i+1, days=6, hours=0 if i != 2 else 12)
                    accepted = True if i != 0 else False
                    cls._create_simple_finished_observation_SUB_for_project(p, duration=86400, start_time=cycle_start+step, accepted=accepted)
                # Regular with no efficiency (week 5)
                cls._create_simple_finished_observation_SUB_for_project(p, start_time=cycle_start + timedelta(weeks=5), accepted=False)
            else:
                cls._create_simple_finished_observation_SUB_for_project(p, 60)
        # Add reservations
        cls._add_reservations_for_cycle_reporting()

    # Setups

    @classmethod
    def _add_reservations_for_cycle_reporting(cls) -> None:
        """
        Help method to add reservations for testing.
        """
        # Create generic and 'stand-alone' reservations
        reservation_template = models.ReservationTemplate.get_latest(name="reservation")
        reservation_template_spec = get_default_json_object_for_schema(reservation_template.schema)

        # reservation_template_spec['activity']['type'] is set to 'maintanance' by default
        reservation_template_spec['resources']['stations'] = ['DE601', 'DE602', 'DE603']
        reservation_no_project = models.Reservation.objects.create(start_time=datetime.now(),
                                                                   stop_time=datetime.now() + timedelta(seconds=300),
                                                                   name="No project", description="No project",
                                                                   specifications_template=reservation_template,
                                                                   specifications_doc=reservation_template_spec)
        reservation_template_spec['activity']['type'] = 'other'
        reservation_template_spec['resources']['stations'] = ['FR606', 'SE607', 'UK608', 'DE609', 'PL610']
        reservation_mixed = models.Reservation.objects.create(project=cls.projects['unassigned'],
                                                              start_time=datetime.now(),
                                                              stop_time=datetime.now() + timedelta(seconds=500),
                                                              name="Non-production project",
                                                              description="Non-production project",
                                                              specifications_template=reservation_template,
                                                              specifications_doc=reservation_template_spec)  # Non-production project
        reservation_template_spec['activity']['type'] = 'software rollout'
        reservation_template_spec['resources']['stations'] = ['CS030', 'CS101', 'CS301']
        reservation_project = models.Reservation.objects.create(project=cls.projects['regular'],
                                                                start_time=datetime.now(),
                                                                stop_time=datetime.now() + timedelta(seconds=600),
                                                                name="Production project",
                                                                description="Production project",
                                                                specifications_template=reservation_template,
                                                                specifications_doc=reservation_template_spec)  # Production project

        reservation_template_spec['activity']['type'] = 'stand-alone mode'
        reservation_template_spec['resources']['stations'] = ['CS001', 'CS002', 'CS003']
        reservation_no_project_sa_mode = models.Reservation.objects.create(start_time=datetime.now(),
                                                                           stop_time=datetime.now() + timedelta(seconds=1200),
                                                                           name="SA no project",
                                                                           description="SA no project",
                                                                           specifications_template=reservation_template,
                                                                           specifications_doc=reservation_template_spec)
        reservation_template_spec['resources']['stations'] = ['CS001', 'CS002', 'CS003', 'CS004', 'CS005']
        reservation_mixed_sa_mode = models.Reservation.objects.create(start_time=datetime.now(),
                                                                      stop_time=datetime.now() + timedelta(seconds=350),
                                                                      project=cls.projects['unassigned'],
                                                                      name="SA mixed no project",
                                                                      description="SA mixed no project",
                                                                      specifications_template=reservation_template,
                                                                      specifications_doc=reservation_template_spec)
        reservation_template_spec['resources']['stations'] = ['CS011', 'CS013', 'CS017', 'CS030']
        reservation_project_sa_mode = models.Reservation.objects.create(start_time=datetime.now(),
                                                                        stop_time=datetime.now() + timedelta(seconds=800),
                                                                        project=cls.projects['regular'],
                                                                        name="SA project", description="SA project",
                                                                        specifications_template=reservation_template,
                                                                        specifications_doc=reservation_template_spec)

    @staticmethod
    def _set_dataproducts_size_and_set_subtask_status(subtask:models.Subtask, status='finished', size=123) -> None:
        """
        Help method to give the dataproducts an explicit size and set subtask status to (default) 'finished'.
        """
        for dataproduct in subtask.output_dataproducts.all():
            dataproduct.size = size
            dataproduct.save()
        set_subtask_state_following_allowed_transitions(subtask, status)

    @classmethod
    def _create_simple_finished_observation_SUB_for_project(cls, project:models.Project, duration=600, start_time=None, accepted=None) -> models.SchedulingUnitBlueprint:
        """
        Help method to create a SUB with an observation with a default duration of 600s set to finished.
        """
        scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project))
        su_spec_template = models.SchedulingUnitTemplate.get_latest(name='scheduling unit')
        su_draft = models.SchedulingUnitDraft.objects.create(scheduling_set=scheduling_set,
                                                             specifications_template=su_spec_template,
                                                             ingest_permission_required=False)
        obs_task_template = models.TaskTemplate.get_latest(name='target observation')
        obs_specifications_doc = obs_task_template.get_default_json_document_for_schema()
        obs_specifications_doc['duration'] = duration
        obs_specifications_doc['station_configuration']['SAPs'][0]['subbands'] = [0, 1]
        obs_specifications_doc['QA']['file_conversion']['enabled'] = False
        obs_specifications_doc['QA']['plots']['enabled'] = False
        _ = models.TaskDraft.objects.create(name='observation', scheduling_unit_draft=su_draft,
                                            specifications_doc=obs_specifications_doc, specifications_template=obs_task_template)
        su_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(su_draft)
        if accepted is not None:
            su_blueprint.results_accepted = accepted
            su_blueprint.save()

        # Add some 'Beamformed Observation' dataproducts and set subtask status to "finished" (and so will be the SUB)
        obs_subtask = models.Subtask.objects.get(task_blueprint__scheduling_unit_blueprint=su_blueprint,
                                                 specifications_template__type__value='observation')
        if start_time:  # Simulate run to manually set start and stop times
            set_subtask_state_following_allowed_transitions(obs_subtask, 'queued')
            obs_subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.STARTING.value)
            obs_subtask.actual_process_start_time = start_time
            obs_subtask.save()
            obs_subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.STARTED.value)
            obs_subtask.actual_on_sky_start_time = start_time
            obs_subtask.actual_process_stop_time = start_time + timedelta(seconds=obs_specifications_doc['duration'])
            obs_subtask.actual_on_sky_stop_time = start_time + timedelta(seconds=obs_specifications_doc['duration'])
            obs_subtask.save()
            obs_subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.FINISHING.value)
            obs_subtask.save()

        subtask_output_obs = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=obs_subtask))
        dp_beamformed_obs = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=subtask_output_obs,
                                                                                      dataformat=models.Dataformat.objects.get(value="Beamformed"),
                                                                                      datatype=models.Datatype.objects.get(value="time series")))
        cls._set_dataproducts_size_and_set_subtask_status(obs_subtask)
        return su_blueprint

    @classmethod
    def setup_project_report(cls, project:models.Project) -> None:
        """
        Setup and populate requirements given a project to test Project reporting.
        """
        # Create requirements
        scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project))

        # Create four SUBs and respectively set their states to: 'finished' x2 (so we can create dataproducts, compare
        # their sizes, and have a successful and a failed SUBs), 'cancelled' and 'defined' (which means not cancelled).
        su_spec_template = models.SchedulingUnitTemplate.get_latest(name='scheduling unit')
        obs_task_template = models.TaskTemplate.get_latest(name='target observation')
        obs_specifications_doc = obs_task_template.get_default_json_document_for_schema()
        obs_specifications_doc['duration'] = 600
        obs_specifications_doc['station_configuration']['SAPs'][0]['subbands'] = [0, 1]
        obs_specifications_doc['QA']['file_conversion']['enabled'] = False
        obs_specifications_doc['QA']['plots']['enabled'] = False
        # Set custom stations to avoid stations conflicts between 'regular' and 'commissioning' projects
        if project.name == 'regular':
            obs_specifications_doc['station_configuration']['station_groups'][0]['stations'] = ['RS106', 'RS205', 'RS208']
            obs_specifications_doc['station_configuration']['station_groups'][0]['max_nr_missing'] = 2
        if project.name == 'commissioning':
            obs_specifications_doc['station_configuration']['station_groups'][0]['stations'] = ['RS305', 'RS306', 'RS307']
            obs_specifications_doc['station_configuration']['station_groups'][0]['max_nr_missing'] = 2

        ingest_task_template = models.TaskTemplate.get_latest(name='ingest')
        pipeline_task_template = models.TaskTemplate.get_latest(name='preprocessing pipeline')

        for i in range(5):  # Create five SUBs and set them to be respectively: successful, failed, cancelled, not cancelled and unschedulable.
            # Create an obs-ingest pair for each of the five SUBs
            su_draft = models.SchedulingUnitDraft.objects.create(scheduling_set=scheduling_set,
                                                                 specifications_template=su_spec_template,
                                                                 ingest_permission_required=False)
            obs_task_draft = models.TaskDraft.objects.create(name='observation for ingest',
                                                             scheduling_unit_draft=su_draft,
                                                             specifications_doc=obs_specifications_doc,
                                                             specifications_template=obs_task_template)

            ingest_task_draft = models.TaskDraft.objects.create(name='ingest',
                                                                scheduling_unit_draft=su_draft,
                                                                specifications_doc=ingest_task_template.get_default_json_document_for_schema(),
                                                                specifications_template=ingest_task_template)

            task_rel_sel_template = models.TaskRelationSelectionTemplate.get_latest(name='all')
            models.TaskRelationDraft.objects.create(producer=obs_task_draft,
                                                    consumer=ingest_task_draft,
                                                    output_role=obs_task_template.connector_types.first(),
                                                    input_role=ingest_task_template.connector_types.filter(
                                                        dataformat__value=models.Dataformat.Choices.MEASUREMENTSET.value).first(),
                                                    selection_template=task_rel_sel_template,
                                                    selection_doc=task_rel_sel_template.get_default_json_document_for_schema())

            if i == 3:  # Create an obs-pipeline pair for the 4th SUB (which is the not cancelled)
                obs_task_draft = models.TaskDraft.objects.create(name='observation for pipeline',
                                                                 scheduling_unit_draft=su_draft,
                                                                 specifications_doc=obs_specifications_doc,
                                                                 specifications_template=obs_task_template)

                pipeline_task_draft = models.TaskDraft.objects.create(name='pipeline',
                                                                      scheduling_unit_draft=su_draft,
                                                                      specifications_doc=pipeline_task_template.get_default_json_document_for_schema(),
                                                                      specifications_template=pipeline_task_template)
                task_rel_sel_template = models.TaskRelationSelectionTemplate.get_latest(name='all')
                models.TaskRelationDraft.objects.create(producer=obs_task_draft,
                                                        consumer=pipeline_task_draft,
                                                        output_role=obs_task_template.connector_types.first(),
                                                        input_role=pipeline_task_template.connector_types.filter(
                                                            dataformat__value=models.Dataformat.Choices.MEASUREMENTSET.value).first(),
                                                        selection_template=task_rel_sel_template,
                                                        selection_doc=task_rel_sel_template.get_default_json_document_for_schema())

            # Create SUB
            su_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(su_draft)
            if i in (0, 1):
                # Set results_accepted flag so we have a successful and a failed SUBs
                su_blueprint.results_accepted = i == 0
                su_blueprint.save()

            # Set subtasks statuses, Workflow flags, etc.
            obs_subtask = models.Subtask.objects.get(task_blueprint__scheduling_unit_blueprint=su_blueprint,
                                                   specifications_template__type__value='observation',
                                                   task_blueprint__draft__name='observation for ingest')
            ingest_subtask = models.Subtask.objects.get(task_blueprint__scheduling_unit_blueprint=su_blueprint,
                                                        specifications_template__type__value='ingest')

            # Create SubtaskOutput and Dataproducts
            subtask_output_obs = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=obs_subtask))
            dp_interferometric_obs = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=subtask_output_obs,
                                                                                               dataformat=models.Dataformat.objects.get(value="MeasurementSet"),
                                                                                               datatype=models.Datatype.objects.get(value="visibilities")))
            dp_beamformed_obs = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=subtask_output_obs,
                                                                                          dataformat=models.Dataformat.objects.get(value="Beamformed"),
                                                                                          datatype=models.Datatype.objects.get(value="time series")))

            if i == 3:  # 4th SUB (not cancelled) with pipeline
                pipeline_subtask = models.Subtask.objects.get(task_blueprint__scheduling_unit_blueprint=su_blueprint,
                                                              specifications_template__type__value='pipeline')
                subtask_output_pip = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=pipeline_subtask))
                dp_preprocessing_pip = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=subtask_output_pip,
                                                                                                 dataformat=models.Dataformat.objects.get(value="MeasurementSet"),
                                                                                                 datatype=models.Datatype.objects.get(value="visibilities")))
                dp_pulsar_pip1 = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=subtask_output_pip,
                                                                                           dataformat=models.Dataformat.objects.get(value="pulp summary"),
                                                                                           datatype=models.Datatype.objects.get(value="pulsar profile")))
                dp_pulsar_pip2 = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=subtask_output_pip,
                                                                                           dataformat=models.Dataformat.objects.get(value="pulp analysis"),
                                                                                           datatype=models.Datatype.objects.get(value="time series")))

                # Set explicit actual_process_duration to simulate the pipeline to be run within the specified time (600s)
                now = datetime.utcnow()
                set_subtask_state_following_allowed_transitions(pipeline_subtask, 'queued')
                pipeline_subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.STARTING.value)
                pipeline_subtask.actual_process_start_time = now
                pipeline_subtask.save()
                pipeline_subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.STARTED.value)
                pipeline_subtask.actual_process_stop_time = now + timedelta(seconds=obs_specifications_doc['duration'])
                pipeline_subtask.save()

                # Give dataproducts an explicit size and set pipeline subtask status to finished
                cls._set_dataproducts_size_and_set_subtask_status(pipeline_subtask)

            if i in (0, 1):  # Successful and failed SUBs
                schedule_subtask(obs_subtask, start_time=datetime.utcnow() + timedelta(seconds=2 * (i+1) * obs_specifications_doc['duration']))
                # Give dataproducts an explicit size and set obs subtask status to finished
                cls._set_dataproducts_size_and_set_subtask_status(obs_subtask)

                schedule_subtask(ingest_subtask, start_time=obs_subtask.process_stop_time)
                # Give dataproducts an explicit size and set ingest subtask status to finished
                cls._set_dataproducts_size_and_set_subtask_status(ingest_subtask)
            elif i == 2:    # Cancelled at half specified time duration (300s)
                # Set explicit on_sky_duration and actual_process_duration to simulate the obs to be run within half the specified time (300s)
                now = datetime.utcnow()
                set_subtask_state_following_allowed_transitions(obs_subtask, 'queued')
                obs_subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.STARTING.value)
                obs_subtask.actual_process_start_time = now
                obs_subtask.save()
                obs_subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.STARTED.value)
                obs_subtask.actual_on_sky_start_time = now
                obs_subtask.actual_process_stop_time = now + timedelta(seconds=obs_specifications_doc['duration'])
                obs_subtask.actual_on_sky_stop_time = now + timedelta(seconds=obs_specifications_doc['duration'])/2
                obs_subtask.save()
                obs_subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.FINISHING.value)
                obs_subtask.save()
                # Cancel the SUB
                cancel_scheduling_unit_blueprint(su_blueprint)
            elif i == 4:    # Unschedulable
                set_subtask_state_following_allowed_transitions(obs_subtask, 'unschedulable')
            else:
                # the rest is just defined/schedulable
                pass

    # Tests

    def test_create_cycle_report(self):
        """
        Test create_cycle_report extra action.
        """
        with tmss_test_env.create_tmss_client() as client:
            for trial in range(10):
                result = client.get_path_as_json_object('/cycle/%s/report' % self.cycle.pk)
                if "report is getting prepared" not in result:
                    break
                logger.info('waiting for report generation to complete...')
                time.sleep(2)

        self.assertNotIn("report is getting prepared", result)

        # Assertions

        # Assert telescope_time_distribution
        telescope_time_distribution = result['telescope_time_distribution']
        self.assertAlmostEqual(telescope_time_distribution['UNASSIGNED']['durations']['total'], 600, places=4)
        self.assertAlmostEqual(telescope_time_distribution['UNASSIGNED']['durations']['succeeded'], 600, places=4)
        self.assertAlmostEqual(telescope_time_distribution['UNASSIGNED']['durations']['failed'], 0, places=4)
        self.assertAlmostEqual(telescope_time_distribution['UNASSIGNED']['durations']['idle'], 0, places=4)
        self.assertAlmostEqual(telescope_time_distribution['FILLER']['durations']['total'], 60, places=4)
        self.assertAlmostEqual(telescope_time_distribution['REGULAR']['durations']['total'], 1200, places=4)
        self.assertAlmostEqual(telescope_time_distribution['REGULAR']['durations']['succeeded'], 600, places=4)
        self.assertAlmostEqual(telescope_time_distribution['REGULAR']['durations']['failed'], 600, places=4)
        self.assertAlmostEqual(telescope_time_distribution['REGULAR']['durations']['idle'], 0, places=4)
        self.assertAlmostEqual(telescope_time_distribution['USER_SHARED_SUPPORT']['durations']['total'], 60, places=4)
        self.assertAlmostEqual(telescope_time_distribution['COMMISSIONING']['durations']['total'], 1200, places=4)
        self.assertAlmostEqual(telescope_time_distribution['COMMISSIONING']['durations']['succeeded'], 600, places=4)
        self.assertAlmostEqual(telescope_time_distribution['COMMISSIONING']['durations']['failed'], 600, places=4)
        self.assertAlmostEqual(telescope_time_distribution['COMMISSIONING']['durations']['idle'], 0, places=4)
        self.assertAlmostEqual(telescope_time_distribution['DDT']['durations']['total'], 60, places=4)
        self.assertAlmostEqual(telescope_time_distribution['TEST']['durations']['total'], 600+86400*3, places=4)
        self.assertAlmostEqual(telescope_time_distribution['TEST']['durations']['succeeded'], 86400*2, places=4)
        self.assertAlmostEqual(telescope_time_distribution['TEST']['durations']['failed'], 600+86400, places=4)
        self.assertAlmostEqual(telescope_time_distribution['TEST']['durations']['idle'], 0, places=4)

        # Assert average_efficiency
        average_efficiency = result['average_efficiency']
        self.assertEqual(average_efficiency['target'], 0.65)        # TODO: Change it when implemented.
        self.assertEqual(average_efficiency['efficiency'], 0.04)  # 0.04 = (1200/3000 daily_eff + 86400/86400 daily_eff + 43200/43200 daily_eff + 43200/43200 daily_eff) / 85 days = 3.4 / 85

        # Assert completion_level
        completion_level = result['completion_level']
        self.assertEqual(completion_level['target'], 0.0)   # TODO: Change it when implemented.
        self.assertAlmostEqual(completion_level['total'], 262980, 4)  # 262980 = 600 'unassigned' + 1200 'regular' + 1200 'commissioning' + 86400*3+600 'test' + 3*60 'filler'
        self.assertAlmostEqual(completion_level['succeeded'], 174780, 4)  # 174780 = 600 'regular' + 600 'commissioning' + 86400*2 'test' + 780 'unassigned'
        self.assertEqual(completion_level['succeeded_perc'], 0.66)
        self.assertEqual(completion_level['prognosis'], 0.0)   # Two 'unschedulable' Subtasks (one in 'regular' and one in 'commissioning')

        # Assert observation_hours_per_category
        observation_durations_per_category = result['observation_durations_per_category']
        self.assertAlmostEqual(observation_durations_per_category['total_duration'], 262980, places=4)  # 262980 = 600 'unassigned' + 1200 'regular' + 1200 'commissioning' + 86400*3+600 'test' + 3*60 'filler'
        self.assertAlmostEqual(observation_durations_per_category['total_duration_successful'], 174780, places=4)  # 174000 = 600 'regular' + 600 'commissioning' + 86400*2 'test' + 780 unassigned
        self.assertAlmostEqual(observation_durations_per_category['total_duration_successful_A'], 174780, places=4)
        self.assertAlmostEqual(observation_durations_per_category['total_duration_successful_B'], 0, places=4)
        self.assertAlmostEqual(observation_durations_per_category['total_duration_failed'], 88200, places=4)  # 88200 = 600 'regular' + 600 'commissioning' + 86400+600 'test
        self.assertAlmostEqual(observation_durations_per_category['total_duration_idle'], 0, places=4)  # 'unassigned'
        self.assertAlmostEqual(observation_durations_per_category['DDT Com Rep'], 1260, places=4)   # = 0 + 1200 + 0 + 60 (DDT + Com + Rep + filler)
        self.assertAlmostEqual(observation_durations_per_category['TMSS Com'], 1200, places=4)
        self.assertAlmostEqual(observation_durations_per_category['System Unavailability'], 2100, places=4)  # = 300 + 600 + 1200 (maintanance + software rollout + TMSS Com)
        # Assert percentages
        self.assertEqual(observation_durations_per_category['successful_perc'], 0.66)
        self.assertEqual(observation_durations_per_category['idle_perc'], 0)
        self.assertEqual(observation_durations_per_category['failed_perc'], 0.34)
        self.assertEqual(observation_durations_per_category['ddt_com_rep_perc'], 0)
        self.assertEqual(observation_durations_per_category['tmss_com_perc'], 0)
        self.assertEqual(observation_durations_per_category['system_unavailability_perc'], 0.01)

        # Assert weekly_efficiency
        weekly_efficiency = result['weekly_efficiency']
        self.assertEqual(weekly_efficiency['weeks'][0]['efficiency'], 0.62)
        self.assertEqual(weekly_efficiency['weeks'][0]['total'], 3180)
        self.assertEqual(weekly_efficiency['weeks'][0]['total_succeeded'], 1980)
        self.assertEqual(weekly_efficiency['weeks'][1]['efficiency'], 0)
        self.assertEqual(weekly_efficiency['weeks'][1]['total'], 86400)
        self.assertEqual(weekly_efficiency['weeks'][1]['total_succeeded'], 0)
        self.assertEqual(weekly_efficiency['weeks'][2]['efficiency'], 1)
        self.assertEqual(weekly_efficiency['weeks'][2]['total'], 86400)
        self.assertEqual(weekly_efficiency['weeks'][2]['total_succeeded'], 86400)
        self.assertEqual(weekly_efficiency['weeks'][3]['efficiency'], 1)  # 24 hour-SUB between week 3 and week 4
        self.assertEqual(weekly_efficiency['weeks'][3]['total'], 43200)
        self.assertEqual(weekly_efficiency['weeks'][3]['total_succeeded'], 43200)

        self.assertEqual(weekly_efficiency['weeks'][4]['efficiency'], 1)  # 24 hour-SUB between week 3 and week 4
        self.assertEqual(weekly_efficiency['weeks'][4]['total'], 43200)
        self.assertEqual(weekly_efficiency['weeks'][4]['total_succeeded'], 43200)
        self.assertEqual(weekly_efficiency['weeks'][5]['total'], 600)
        self.assertEqual(weekly_efficiency['weeks'][5]['total_succeeded'], 0)
        self.assertIsNone(weekly_efficiency['weeks'][6]['efficiency'])

        # Assert data_ingested_per_site_and_category
        data_per_site_and_cat = result['data_ingested_per_site_and_category']
        # self.assertEqual(data_per_site_and_cat['sites']['CEP4']['size__sum'], 0)   # TODO: Update tests per LTA site after TMSS-1976.
        # self.assertEqual(data_per_site_and_cat['sites']['Sara']['size__sum'], 0)
        # self.assertEqual(data_per_site_and_cat['sites']['Jülich']['size__sum'], 0)
        # self.assertEqual(data_per_site_and_cat['sites']['Poznan']['size__sum'], 0)
        self.assertEqual(data_per_site_and_cat['categories']['Interferometric Observation']['size__sum'], 1476)  # 1476 = 492 'regular' + 492 'commissioning' + 246 'regular' (dp_interferometric_obs) + 246 'commissioning' (dp_interferometric_obs)
        self.assertEqual(data_per_site_and_cat['categories']['Beamformed Observation']['size__sum'], 1476)  # 1476 = 123x4 'other projects' + 246 'regular' (dp_beamformed_obs) + 246 'commissioning' (dp_beamformed_obs) + 492 'test'
        self.assertEqual(data_per_site_and_cat['categories']['Preprocessing Pipeline']['size__sum'], 246)  # 246 = 123 'regular' (dp_preprocessing_pip) + 123 'commissioning' (dp_preprocessing_pip)
        self.assertEqual(data_per_site_and_cat['categories']['Pulsar Pipeline']['size__sum'], 492)    # 492 = 246 'regular'  (dp_pulsar_pip1 + dp_pulsar_pip2) + 246 'commissioning' (dp_pulsar_pip1 + dp_pulsar_pip2)

        # Assert usage_mode
        usage_mode = result['usage_mode']
        self.assertAlmostEqual(usage_mode['all modes']['total'], 3750, places=4)
        self.assertAlmostEqual(usage_mode['all modes']['observing'], 1400, places=4)
        self.assertAlmostEqual(usage_mode['all modes']['idle/test'], 850, places=4)

        self.assertAlmostEqual(usage_mode['local-mode'], 1200, places=4)

        self.assertAlmostEqual(usage_mode['stand-alone mode']['total'], 2350, places=4)
        self.assertAlmostEqual(usage_mode['stand-alone mode']['no project'], 1200, places=4)
        self.assertAlmostEqual(usage_mode['stand-alone mode']['project'], 800, places=4)
        self.assertAlmostEqual(usage_mode['stand-alone mode']['mixed/no project'], 350, places=4)

        self.assertAlmostEqual(usage_mode['ILT mode']['total'], 1400, places=4)
        self.assertAlmostEqual(usage_mode['ILT mode']['observing'], 600, places=4)
        self.assertAlmostEqual(usage_mode['ILT mode']['idle/test'], 500, places=4)
        self.assertAlmostEqual(usage_mode['ILT mode']['ILT Dutch'], 600, places=4)    # = (600 + 1200 + 350 + 800) - (1200 + 350 + 800) ((Production project + SA no project + SA mixed no project + SA project) - (SA no project + SA mixed no project + SA project))

        # Assert failures
        failures = result['failures']
        # TODO: See TMSS-2276. Fix bug in weekly_efficiency calculation.
        # self.assertAlmostEqual(failures['months'][0]['total'], 262380, places=4)  # 262200 = 600 'unassigned' + 1200 'regular' + 1200 'commissioning' + 86400*2+43200+43200 'test' +3*60 'filler'
        # self.assertAlmostEqual(failures['months'][0]['total_failed'], 86400+1200, places=4)   # 87600 = 600 'regular' + 600 'commissioning' + 86400 'test'
        # self.assertEqual(failures['months'][0]['failed_perc'], 0.33)
        # self.assertAlmostEqual(failures['months'][1]['total'], 600, places=4)  # 600 'test'
        # self.assertAlmostEqual(failures['months'][1]['total_failed'], 600, places=4)
        # self.assertEqual(failures['months'][1]['failed_perc'], 1)
        # self.assertIsNone(failures['months'][2]['failed_perc'])  # Only two months

        # Assert contains_overlapping_observations is False
        self.assertEqual(result['contains_overlapping_observations'], False)

    def test_create_project_report(self):
        """
        Test create project extra action.
        """
        with tmss_test_env.create_tmss_client() as client:
            result = client.get_path_as_json_object('/project/%s/report' % self.test_project_report.pk)

        # Asserts

        # Assert Project and ProjectQuota ids
        self.assertEqual(result['project'], self.test_project_report.pk)
        self.assertEqual(result['quota'][0]['id'], self.project_quota.pk)

        # Assert durations are well calculated
        # We have five SUBs (successful, failed, cancelled, not cancelled and unschedulable)
        self.assertAlmostEqual(result['durations']['total_regular'], 600*5+600, places=0)  # 1 observation (600s) per SUB + 1 additional (600s) in the 4th not cancelled
        self.assertAlmostEqual(result['durations']['total_on_sky'], 3300, places=1)
        self.assertAlmostEqual(result['durations']['total_not_cancelled'], 3000, places=4)  # 3000 = 3300-300 (only the 3rd SUB (300s) is cancelled)
        self.assertAlmostEqual(result['durations']['total_observed'], 600*2, places=4)  #  successful + failed SUBs each with 600s
        self.assertAlmostEqual(result['durations']['total_observed_succeeded'], 600, places=4)
        self.assertAlmostEqual(result['durations']['total_observed_succeeded_A'], 600, places=4)
        self.assertAlmostEqual(result['durations']['total_observed_succeeded_B'], 0, places=4)
        self.assertAlmostEqual(result['durations']['total_observed_failed'], 600, places=4)
        # Assert percentages
        self.assertEqual(result['durations']['not_cancelled_perc'], 0.91)
        self.assertEqual(result['durations']['observed_perc'], 0.36)
        self.assertEqual(result['durations']['observed_succeeded_perc'], 0.18)
        self.assertEqual(result['durations']['observed_failed_perc'], 0.18)

        # There is only one successful SUB
        subs = models.SchedulingUnitBlueprint.objects.filter(draft__scheduling_set__project=self.test_project_report)
        self.assertEqual(result['SUBs']['successful'][0]['id'], subs.get(results_accepted=True).pk)
        self.assertEqual(result['SUBs']['successful'][0]['status'], 'finished')
        self.assertEqual(result['SUBs']['successful'][0]['ingested_data_size'], 246)
        # There is only one failed SUB
        self.assertEqual(result['SUBs']['failed'][0]['id'], subs.get(results_accepted=False).pk)
        self.assertEqual(result['SUBs']['failed'][0]['status'], 'finished')
        self.assertEqual(result['SUBs']['failed'][0]['ingested_data_size'], 246)

        # There are eight dataproducts, each has a size of 123
        self.assertEqual(result['LTA_dataproducts']['size'], 492)
        # Assert data_ingested_per_site_and_category
        data_per_site = result['data_ingested_per_site']
        # self.assertEqual(data_per_site['CEP4']['size__sum'], 0)   # TODO: Update tests per LTA site after TMSS-1976.
        # self.assertEqual(data_per_site['Sara']['size__sum'], 0)
        # self.assertEqual(data_per_site['Jülich']['size__sum'], 0)
        # self.assertEqual(data_per_site['Poznan']['size__sum'], 0)
        # Assert SAPs exposure
        self.assertAlmostEqual(result['SAPs_exposure']['_target_name_'], 600, places=1)   # only count exposure of the successful obs)
        # Assert processing resources
        self.assertEqual(result['processing_resources']['CPUtimeRAW'], None)  # TODO: Change it properly.
        self.assertAlmostEqual(result['processing_resources']['CPU_time_used'], 146.4, places=1)  # one pipeline -> 146.4 = 2(cores_per_task) x 122(parallel_tasks) x 600(duration) / 1000

        # Assert contains_overlapping_observations is False
        self.assertEqual(result['contains_overlapping_observations'], False)

class FailureReportTest(unittest.TestCase):

    def test_create_failure_report(self):
        # setup, create project with three SUs, each containing single observation
        project = models.Project.objects.create(**Project_test_data(name='failure_report_project'))
        su_1 = ReportTest._create_simple_finished_observation_SUB_for_project(project=project,
                                                                              duration=6666,
                                                                              start_time=datetime(2023, 1, 31, 23, 59))
        su_2 = ReportTest._create_simple_finished_observation_SUB_for_project(project=project,
                                                                              duration=3333,
                                                                              start_time=datetime(2023, 3, 31, 23, 59))
        su_3 = ReportTest._create_simple_finished_observation_SUB_for_project(project=project,
                                                                              duration=4444,
                                                                              start_time=datetime(2023, 5, 31, 23, 59))

        # setup, add system events, each affecting one of the SUs
        template = models.SystemEventTemplate.get_latest(name="affectedhardware")
        event_1 = models.SystemEvent.objects.create(name='event 1',
                                                    issue_type=models.SystemEventType.objects.get(value=models.SystemEventType.Choices.CEP.value),
                                                    issue_subtype=models.SystemEventSubtype.objects.get(value=models.SystemEventSubtype.Choices.CRASH.value),
                                                    severity=models.SystemEventSeverity.objects.get(value=models.SystemEventSeverity.Choices.FAILURE.value),
                                                    start=datetime(year=2023, month=1, day=31, hour=23),
                                                    stop=datetime(year=2023, month=2, day=1, hour=2),
                                                    affected_hardware_doc={},
                                                    affected_hardware_template=template,
                                                    created_by=User.objects.first(),
                                                    status=models.SystemEventStatus.objects.get(value=models.SystemEventStatus.Choices.OPEN.value),
                                                    )
        event_1.affected_tasks.set(su_1.task_blueprints.all())
        event_1.save()

        event_2 = models.SystemEvent.objects.create(name='event 2',
                                                    issue_type=models.SystemEventType.objects.get(value=models.SystemEventType.Choices.HUMAN.value),
                                                    issue_subtype=models.SystemEventSubtype.objects.get(value=models.SystemEventSubtype.Choices.DATALOSS.value),
                                                    severity=models.SystemEventSeverity.objects.get(value=models.SystemEventSeverity.Choices.FAILURE.value),
                                                    start=datetime(year=2023, month=3, day=31, hour=23),
                                                    stop=datetime(year=2023, month=4, day=1, hour=2),
                                                    affected_hardware_doc={},
                                                    affected_hardware_template=template,
                                                    created_by=User.objects.first(),
                                                    status=models.SystemEventStatus.objects.get(value=models.SystemEventStatus.Choices.ANALYSED.value),
                                                    )
        event_2.affected_tasks.set(su_2.task_blueprints.all())
        event_2.save()
        
        
        event_3 = models.SystemEvent.objects.create(name='event 3',
                                                    issue_type=models.SystemEventType.objects.get(value=models.SystemEventType.Choices.HUMAN.value),
                                                    issue_subtype=models.SystemEventSubtype.objects.get(value=models.SystemEventSubtype.Choices.NOISY.value),
                                                    severity=models.SystemEventSeverity.objects.get(value=models.SystemEventSeverity.Choices.MAJOR.value),
                                                    start=datetime(year=2023, month=5, day=31, hour=23),
                                                    stop=datetime(year=2023, month=6, day=1, hour=2),
                                                    affected_hardware_doc={},
                                                    affected_hardware_template=template,
                                                    created_by=User.objects.first(),
                                                    status=models.SystemEventStatus.objects.get(value=models.SystemEventStatus.Choices.CLOSED.value),
                                                    )
        event_3.affected_tasks.set(su_3.task_blueprints.all())
        event_3.save()


        # generate the report
        with tmss_test_env.create_tmss_client() as client:
            start = datetime(year=2023, month=1, day=1)
            stop = datetime(year=2024, month=1, day=1)
            report = client.get_path_as_json_object('/failure_report/?start=%s&stop=%s' % (start.isoformat(), stop.isoformat()))

        # assertions, SUs 1 and 2 should be in here, SU 3 should be ignored due to lack of severity
        # check grand total
        self.assertIn('total_duration_lost_event', report)
        self.assertEqual(report['total_duration_lost_event'], 21600)  # 2 failure events of 3 hours each
        self.assertAlmostEqual(report['total_percent_of_wall_time_lost_event'], 0.0684931506849315)
        self.assertEqual(report['total_duration_lost_on_sky'], 9999)  # 2 units: 6666 + 3333
        self.assertAlmostEqual(report['total_percent_of_wall_time_lost_on_sky'], 0.0003170662100456621)

        # check aggregates
        self.assertIn('human', str(report['system_event_summary']['by_issue_type']))   # event 1
        for d in report['system_event_summary']['by_issue_type']:
            self.assertEqual(d['count'], 1)   # only event 1 (but not 3) is accounted for
        self.assertNotIn('noisy', str(report['system_event_summary']['by_issue_subtype']))  # event 3 missing

        # check histogram (events get binned correctly?)
        self.assertEqual(report['lost_time_histogram_event']['2023-01-01'], 3600)  # event 1
        self.assertEqual(report['lost_time_histogram_event']['2023-02-01'], 7200)  # event 1
        self.assertEqual(report['lost_time_histogram_event']['2023-03-01'], 3600)  # event 2
        self.assertEqual(report['lost_time_histogram_event']['2023-04-01'], 7200)  # event 2
        self.assertEqual(report['lost_time_histogram_on_sky']['2023-01-01'], 6666)  # event 1
        self.assertEqual(report['lost_time_histogram_on_sky']['2023-03-01'], 3333)  # event 2

        self.assertNotIn('2023-05-01', report['lost_time_histogram_event'])   # event 3 missing
        self.assertNotIn('2023-06-01', report['lost_time_histogram_event'])   # event 3 missing
        self.assertNotIn('2023-05-01', report['lost_time_histogram_on_sky'])  # event 3 missing

        # check SUs
        self.assertEqual(len(report['failed_scheduling_units']), 2)
        for unit in report['failed_scheduling_units']:
            self.assertIn(unit['on_sky_duration'], [6666, 3333])
            self.assertEqual(unit['project'], 'failure_report_project')
            self.assertEqual(len(unit['subtasks']), 1)


if __name__ == "__main__":
    os.environ['TZ'] = 'UTC'
    unittest.main()
