#!/usr/bin/env python3

# Copyright (C) 2018    ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

# $Id:  $

import os
import unittest
from unittest import mock

import logging
logger = logging.getLogger('lofar.'+__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

from lofar.common.test_utils import exit_with_skipped_code_if_skip_integration_tests
exit_with_skipped_code_if_skip_integration_tests()

# create a module-wide TemporaryExchange, and use it in all communications between TMSSTestEnvironment, RA and ObservationControl
from lofar.messaging.messagebus import TemporaryExchange
tmp_exchange = TemporaryExchange('t_scheduling')
tmp_exchange.open()

# override DEFAULT_BUSNAME with tmp exchange, some modules import from lofar.messaging others from lofar.messaging.config...
import lofar
lofar.messaging.DEFAULT_BUSNAME = tmp_exchange.address
lofar.messaging.config.DEFAULT_BUSNAME = tmp_exchange.address

# before we import any django modules the DJANGO_SETTINGS_MODULE, TMSS_LDAPCREDENTIALS and TMSS_DBCREDENTIALS need to be known/set.
# import and start an isolated TMSSTestEnvironment (with fresh database and attached django and ldap server on free ports)
# this automagically sets the required  DJANGO_SETTINGS_MODULE, TMSS_LDAPCREDENTIALS and TMSS_DBCREDENTIALS envvars.
from lofar.sas.tmss.test.test_environment import TMSSTestEnvironment

tmss_test_env = TMSSTestEnvironment(populate_schemas=True,
                                    populate_legacy_schemas=True, # also check all non-active legacy/development schemas/strategies
                                    populate_test_data=False,
                                    start_postgres_listener=False, start_subtask_scheduler=False, start_dynamic_scheduler=False,
                                    enable_viewflow=False,
                                    exchange=tmp_exchange.address)

try:
    tmss_test_env.start()
except Exception as e:
    logger.exception(e)

    tmss_test_env.stop()
    tmp_exchange.close()
    exit(1)

# tell unittest to stop (and automagically cleanup) the test database once all testing is done.
def tearDownModule():
    tmss_test_env.stop()
    tmp_exchange.close()


from lofar.sas.tmss.test.tmss_test_data_django_models import *

# import and setup rest test data creator
from lofar.sas.tmss.test.tmss_test_data_rest import TMSSRESTTestDataCreator
test_data_creator = TMSSRESTTestDataCreator(tmss_test_env.django_server.url, (tmss_test_env.ldap_server.dbcreds.user, tmss_test_env.ldap_server.dbcreds.password))
from datetime import datetime, timedelta
from dateutil import parser
from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.tmssapp.subtasks import *
from lofar.sas.tmss.tmss.tmssapp.tasks import *
from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions
from lofar.messaging.rpc import RPCService, ServiceMessageHandler
import threading
import dateutil.parser
from lofar.common.datetimeutils import round_to_minute_precision


def create_subtask_object_for_testing(subtask_type_value, subtask_state_value):
    """
    Helper function to create a subtask object for testing with given subtask value and subtask state value
    as string (no object)
    For these testcases 'preprocessing pipeline' and 'observation control' is relevant
    """
    task_blueprint = models.TaskBlueprint.objects.create(**TaskBlueprint_test_data(specifications_template=models.TaskTemplate.get_latest(name='target observation' if subtask_type_value=='observation' else 'preprocessing pipeline')))
    subtask_template_obj = models.SubtaskTemplate.get_latest(name='observation control' if subtask_type_value=='observation' else 'preprocessing pipeline')
    subtask_data = Subtask_test_data(subtask_template=subtask_template_obj, task_blueprint=task_blueprint)
    subtask = models.Subtask.objects.create(**subtask_data)
    if subtask.state.value != subtask_state_value:
        set_subtask_state_following_allowed_transitions(subtask, subtask_state_value)
    return subtask

def create_reserved_stations_for_testing(stations: list, start_time: datetime=None, stop_time: datetime=None):
    """
    Helper function to create stations in reservation
    """
    if start_time is None:
        start_time = datetime.utcnow() - timedelta(hours=1)

    if stop_time is None:
        stop_time = datetime.utcnow() + timedelta(hours=2)

    reservation_template = models.ReservationTemplate.objects.first()
    reservation_doc = reservation_template.get_default_json_document_for_schema()
    reservation_doc['resources']['stations'] = stations
    reservation = models.Reservation.objects.create(description='', start_time=start_time, stop_time=stop_time,
                                                    specifications_doc=reservation_doc,
                                                    specifications_template=reservation_template)

def duplicates(l: list) -> list:
    # O(n^2), but that's good enough.
    uniques = []
    dupes = []

    for e in l:
      if e not in uniques:
        uniques.append(e)
      elif e not in dupes:
        dupes.append(e)

    return dupes

class SchedulingTest(unittest.TestCase):
    def setUp(self):
        tmss_test_env.delete_scheduling_unit_drafts_cascade()
        models.Reservation.objects.all().delete()
        test_data_creator.wipe_cache()

    @staticmethod
    def _create_target_observation_subtask(specification_doc: dict=None, task_station_groups: list=None) -> dict:
        '''create a target observation subtask in defined state and return the subtask as json dict.
        if the given specification_doc is None, then the defaults are used.'''
        if specification_doc is None:
            specification_doc = {'stations': {'digital_pointings': [{'name': 'target0', 'subbands': [0]}], 'station_list': ['CS001', 'CS002', 'CS003']}}

        with tmss_test_env.create_tmss_client() as client:
            subtask_template = client.get_subtask_template("observation control")
            specification_doc = add_defaults_to_json_object_for_schema(specification_doc, subtask_template['schema'])

            task_blueprint_data = test_data_creator.TaskBlueprint(template_url=client.get_task_template(name="target observation")['url'])
            task_blueprint_data['specifications_doc']['station_configuration']['SAPs'][0]['name'] = specification_doc['stations']['digital_pointings'][0]['name']
            if task_station_groups is None:
                task_station_groups = [{'stations': specification_doc.get('station_list', ['CS001']), 'max_nr_missing': 0}]
            task_blueprint_data['specifications_doc']['station_configuration']['station_groups'] = task_station_groups
            task_blueprint = test_data_creator.post_data_and_get_response_as_json_object(task_blueprint_data, '/task_blueprint/')

            filesystem = test_data_creator.post_data_and_get_response_as_json_object(test_data_creator.Filesystem(), '/filesystem/')
            cluster_url = filesystem['cluster']

            subtask_data = test_data_creator.Subtask(specifications_template_url=subtask_template['url'],
                                                     specifications_doc=specification_doc,
                                                     cluster_url=cluster_url,
                                                     scheduled_start_time=datetime.utcnow()+timedelta(minutes=5),
                                                     task_blueprint_url=task_blueprint['url'])
            subtask = test_data_creator.post_data_and_get_response_as_json_object(subtask_data, '/subtask/')
            outputrole = test_data_creator.post_data_and_get_response_as_json_object(test_data_creator.TaskConnectorType(dataformat='MeasurementSet', datatype='visibilities'), '/task_connector_type/')
            test_data_creator.post_data_and_get_url(test_data_creator.SubtaskOutput(subtask_url=subtask['url'], filesystem_url=filesystem['url'], outputrole_url=outputrole['url']), '/subtask_output/')

            client.set_subtask_status(subtask['id'], 'defined')
            return subtask

    def _test_schedule_observation_subtask_with_enough_resources_available(self, observation_specification_doc):
        with tmss_test_env.create_tmss_client() as client:
            # check before and after scheduling if the target observation task_template and the common datetime_tempalte is_used_in_unfinished_tasks_blueprints.
            task_template = models.TaskTemplate.get_latest("target observation")
            datetime_template = next(t for t in models.CommonSchemaTemplate.objects.filter(name='datetime').all() if task_template.id in [r.id for r in t.get_referencing_schema_templates()])

            # should not be used by any task yet.
            self.assertFalse(task_template.is_used_in_unfinished_tasks_blueprints)
            self.assertFalse(datetime_template.is_used_in_unfinished_tasks_blueprints)

            subtask = self._create_target_observation_subtask(observation_specification_doc)
            subtask_id = subtask['id']

            # should be used by task, which is not scheduled yet.
            self.assertTrue(task_template.is_used_in_unfinished_tasks_blueprints)
            self.assertTrue(datetime_template.is_used_in_unfinished_tasks_blueprints)

            subtask = client.schedule_subtask(subtask_id)
            self.assertEqual('scheduled', subtask['state_value'])

            # should be used by task, which is now scheduled, but not finished yet.
            self.assertTrue(task_template.is_used_in_unfinished_tasks_blueprints)
            self.assertTrue(datetime_template.is_used_in_unfinished_tasks_blueprints)

            # test whether all dataproduct specifications are unique
            outputs = Subtask.objects.get(pk=subtask_id).outputs.all()
            dataproduct_specifications_docs = [dp.specifications_doc for output in outputs for dp in output.dataproducts.all()]
            duplicate_dataproduct_specification_docs = duplicates(dataproduct_specifications_docs)

            self.assertEqual([], duplicate_dataproduct_specification_docs)

    def test_schedule_observation_subtask_with_enough_resources_available(self):
            spec = { "stations": { "digital_pointings": [ { "name": "target0", "subbands": [0] } ] },
                     "COBALT": { "correlator": { "enabled": True } } }
            self._test_schedule_observation_subtask_with_enough_resources_available(spec)


    @unittest.skip("TODO: add missing coherent stokes settings")
    def test_schedule_beamformer_observation_subtask_with_enough_resources_available(self):
            spec = {
              "stations": { "digital_pointings": [ { "name": "target0", "subbands": [0] } ] },
              "COBALT": {
                "version": 1,
                "correlator": { "enabled": False },
                "beamformer": {
                    "tab_pipelines": [
                      {
                        "SAPs": [ { "name": "target0", "tabs": [ { "coherent": False }, { "coherent": True } ] } ]
                      }
                    ]
                }
              }
            }
            self._test_schedule_observation_subtask_with_enough_resources_available(spec)


    def test_schedule_observation_subtask_to_close_to_other_failes(self):
        '''test if we can/cannot schedule one observation close to another (within/outside the DEFAULT_INTER_OBSERVATION_GAP)'''
        with tmss_test_env.create_tmss_client() as client:
            subtask1 = self._create_target_observation_subtask()
            test_data_creator.wipe_cache() # prevent that the subtasks share a common scheduling unit
            subtask2 = self._create_target_observation_subtask()

            # scheduling the first subtask should just work
            start_time1 = round_to_minute_precision(datetime.utcnow() + timedelta(hours=24))
            subtask1 = client.schedule_subtask(subtask1['id'], scheduled_start_time=start_time1)
            self.assertEqual('scheduled', subtask1['state_value'])
            self.assertEqual(start_time1, parser.parse(subtask1['scheduled_start_time'], ignoretz=True))

            # scheduling the second subtask should fail when positioning it too close AFTER the first
            start_time2 = parser.parse(subtask1['scheduled_stop_time'], ignoretz=True) + timedelta(seconds=10)
            with self.assertRaises(Exception):
                subtask2 = client.schedule_subtask(subtask2['id'], scheduled_start_time=start_time2)

            subtask2 = client.get_subtask(subtask2['id'])
            self.assertEqual('unschedulable', subtask2['state_value'])
            self.assertIsNotNone(subtask2['unschedulable_reason'])

            # reset
            client.set_subtask_status(subtask2['id'], 'defined')

            subtask2 = client.get_subtask(subtask2['id'])
            self.assertEqual('defined', subtask2['state_value'])
            self.assertIsNone(subtask2['unschedulable_reason'])

            # scheduling the second subtask should fail when positioning it too close BEFORE the first
            start_time2 = parser.parse(subtask1['scheduled_start_time'], ignoretz=True) - DEFAULT_INTER_OBSERVATION_GAP - timedelta(seconds=subtask2['duration']) + timedelta(seconds=10)
            with self.assertRaises(Exception):
                subtask2 = client.schedule_subtask(subtask2['id'], scheduled_start_time=start_time2)

            subtask2 = client.get_subtask(subtask2['id'])
            self.assertEqual('unschedulable', subtask2['state_value'])
            self.assertIsNotNone(subtask2['unschedulable_reason'])

            # reset
            client.set_subtask_status(subtask2['id'], 'defined')

            subtask2 = client.get_subtask(subtask2['id'])
            self.assertEqual('defined', subtask2['state_value'])
            self.assertIsNone(subtask2['unschedulable_reason'])

            # scheduling the second subtask should just work when positioning it far enough AFTER the first
            start_time2 = parser.parse(subtask1['scheduled_stop_time'], ignoretz=True) + DEFAULT_INTER_OBSERVATION_GAP
            subtask2 = client.schedule_subtask(subtask2['id'], scheduled_start_time=start_time2)
            self.assertEqual('scheduled', subtask2['state_value'])
            self.assertEqual(start_time2, parser.parse(subtask2['scheduled_start_time'], ignoretz=True))

            # unschedule
            subtask2 = client.unschedule_subtask(subtask2['id'])
            self.assertEqual('defined', subtask2['state_value'])

            # scheduling the second subtask should just work when positioning it far enough BEFORE the first
            start_time2 = parser.parse(subtask1['scheduled_start_time'], ignoretz=True) - DEFAULT_INTER_OBSERVATION_GAP - timedelta(seconds=subtask2['duration'])
            subtask2 = client.schedule_subtask(subtask2['id'], scheduled_start_time=start_time2)
            self.assertEqual('scheduled', subtask2['state_value'])
            self.assertEqual(start_time1, parser.parse(subtask1['scheduled_start_time'], ignoretz=True))

    def test_schedule_cancelled_observation_subtask_failes(self):
        with tmss_test_env.create_tmss_client() as client:
            subtask_template = client.get_subtask_template("observation control")
            spec = get_default_json_object_for_schema(subtask_template['schema'])
            spec['stations']['digital_pointings'][0]['subbands'] = [0]
            subtask = self._create_target_observation_subtask(spec)
            subtask_id = subtask['id']
            client.set_subtask_status(subtask_id, 'defined')

            # cancel it...
            subtask = client.cancel_subtask(subtask_id)
            self.assertEqual('cancelled', subtask['state_value'])

            # scheduling should fail
            with self.assertRaises(Exception):
                client.schedule_subtask(subtask_id)

            # and status should still be cancelled
            subtask = client.get_subtask(subtask_id)
            self.assertEqual('cancelled', subtask['state_value'])

            # mark it as obsolete... (the user thereby states that the cancelled subtask is not to be used again)
            self.assertIsNone(subtask['obsolete_since'])
            before = datetime.utcnow()
            subtask = client.mark_subtask_as_obsolete(subtask_id)
            obsolete_since = dateutil.parser.parse(subtask['obsolete_since'], ignoretz=True)
            self.assertIsNotNone(obsolete_since)
            self.assertGreaterEqual(obsolete_since, before)

            # scheduling should fail
            with self.assertRaises(Exception):
                client.schedule_subtask(subtask_id)

            # marking an obsolete subtask as obsolete again should be prevented
            with self.assertRaises(Exception) as context:
                subtask = client.get_subtask(subtask_id)
                logger.info("subtask id=%s obsolete_since='%s'", subtask['id'], subtask['obsolete_since'])
                subtask = client.mark_subtask_as_obsolete(subtask_id)
            self.assertIn("has been marked obsolete on %s" % obsolete_since, str(context.exception))

            # and obsolete_since timestamp should still be the same as before
            subtask = client.get_subtask(subtask_id)
            obsolete_since_new = dateutil.parser.parse(subtask['obsolete_since'], ignoretz=True)
            self.assertEqual(obsolete_since, obsolete_since_new)

    def test_cancel_scheduled_observation_subtask(self):
        with tmss_test_env.create_tmss_client() as client:
            subtask_template = client.get_subtask_template("observation control")
            spec = get_default_json_object_for_schema(subtask_template['schema'])
            spec['stations']['digital_pointings'][0]['subbands'] = [0]
            subtask = self._create_target_observation_subtask(spec)
            subtask_id = subtask['id']
            client.set_subtask_status(subtask_id, 'defined')
            # scheduling should succeed
            subtask = client.schedule_subtask(subtask_id)
            self.assertEqual('scheduled', subtask['state_value'])

            # cancel it...
            subtask = client.cancel_subtask(subtask_id)
            self.assertEqual('cancelled', subtask['state_value'])

    def test_cancel_started_observation_subtask(self):
        with tmss_test_env.create_tmss_client() as client:
            subtask_template = client.get_subtask_template("observation control")
            spec = get_default_json_object_for_schema(subtask_template['schema'])
            spec['stations']['digital_pointings'][0]['name'] = 'target0'
            spec['stations']['digital_pointings'][0]['subbands'] = [0]
            subtask = self._create_target_observation_subtask(spec)
            subtask_id = subtask['id']
            client.set_subtask_status(subtask_id, 'defined')
            # scheduling should succeed
            subtask = client.schedule_subtask(subtask_id)
            self.assertEqual('scheduled', subtask['state_value'])

            # mimic that the obs was started and is now running
            client.set_subtask_status(subtask_id, 'starting')
            client.set_subtask_status(subtask_id, 'started')

            observation_killed = threading.Event()
            class MockObsControlMessageHandler(ServiceMessageHandler):
                def __init__(self):
                    super(MockObsControlMessageHandler, self).__init__()
                    self.register_service_method("AbortObservation", self.abort_observation)

                def abort_observation(self, sas_id):
                    observation_killed.set()
                    return {'aborted': True}

            with RPCService(service_name=lofar.mac.config.DEFAULT_OBSERVATION_CONTROL_SERVICE_NAME,
                            handler_type=MockObsControlMessageHandler,
                            exchange=tmp_exchange.address):

                # cancel observation subtask... should kill the running observation
                # check that ObservationControlRPCClient.abort_observation was called
                subtask = client.cancel_subtask(subtask_id)
                self.assertEqual('cancelled', subtask['state_value'])

                observation_killed.wait(10)
                self.assertTrue(observation_killed.is_set())


    def test_schedule_observation_subtask_with_one_blocking_reservation_failed(self):
        """
        Set station CS001 to reserved
        Schedule subtask with station CS001
        Check if scheduling of the subtask fails
        """
        create_reserved_stations_for_testing(['CS001'])

        with tmss_test_env.create_tmss_client() as client:
            subtask_template = client.get_subtask_template("observation control")
            spec = get_default_json_object_for_schema(subtask_template['schema'])
            spec['COBALT']['correlator']['enabled'] = True
            spec['stations']['digital_pointings'][0]['name'] = 'target0'
            spec['stations']['digital_pointings'][0]['subbands'] = [0]
            spec['stations']['station_list'] = ['CS001']
            subtask = self._create_target_observation_subtask(spec)
            subtask_id = subtask['id']

            with self.assertRaises(Exception):
                client.schedule_subtask(subtask_id)

            subtask = client.get_subtask(subtask_id)
            self.assertEqual('unschedulable', subtask['state_value'])
            self.assertIsNotNone(subtask['unschedulable_reason'])

    def test_schedule_observation_subtask_with_blocking_reservations_failed(self):
        """
        Set station CS001, CS002, CS401, CS501 to reserved
        Schedule subtask with stations CS001, CS002, CS401
        Check if scheduling of the subtask fails
        """
        create_reserved_stations_for_testing(['CS001','CS002','CS501','CS401' ])

        with tmss_test_env.create_tmss_client() as client:
            subtask_template = client.get_subtask_template("observation control")
            spec = get_default_json_object_for_schema(subtask_template['schema'])
            spec['COBALT']['correlator']['enabled'] = True
            spec['stations']['digital_pointings'][0]['name'] = 'target0'
            spec['stations']['digital_pointings'][0]['subbands'] = [0]
            spec['stations']['station_list'] = ['CS001', 'CS002', 'CS401']
            subtask = self._create_target_observation_subtask(spec)
            subtask_id = subtask['id']

            with self.assertRaises(Exception):
                client.schedule_subtask(subtask_id)

            subtask = client.get_subtask(subtask_id)
            self.assertEqual('unschedulable', subtask['state_value'])
            self.assertTrue('Too many stations unavailable' in subtask['unschedulable_reason'])

            task = client.get_url_as_json_object(subtask['task_blueprint'])
            self.assertEqual('unschedulable', task['status_value'])
            self.assertTrue('Too many stations unavailable' in task['unschedulable_reason'])

            schedunit = client.get_url_as_json_object(task['scheduling_unit_blueprint'])
            self.assertEqual('unschedulable', schedunit['status_value'])
            self.assertTrue('Too many stations unavailable' in schedunit['unschedulable_reason'])

    def test_schedule_observation_subtask_with_blocking_reservation_ok(self):
        """
        Set station CS002, CS004 to reserved
        Schedule subtask with station CS002, CS003, CS004
        Check if scheduling of the subtask do not fail (it can schedule with station CS003)
        """
        create_reserved_stations_for_testing(['CS002','CS004'])

        with tmss_test_env.create_tmss_client() as client:
            subtask_template = client.get_subtask_template("observation control")
            spec = get_default_json_object_for_schema(subtask_template['schema'])
            spec['COBALT']['correlator']['enabled'] = True
            spec['stations']['digital_pointings'][0]['name'] = 'target0'
            spec['stations']['digital_pointings'][0]['subbands'] = [0]
            subtask = self._create_target_observation_subtask(spec, task_station_groups=[{'stations': ['CS002', 'CS003', 'CS004'], 'max_nr_missing': 2}])
            subtask_id = subtask['id']

            subtask = client.schedule_subtask(subtask_id)
            self.assertEqual('scheduled', subtask['state_value'])

    def test_schedule_observation_subtask_with_blocking_reservation_does_not_contain_reserved_stations(self):
        """
        Schedule subtask with station CS002, CS003, CS004
        Set station CS002 and subsequently CS004 to reserved
        Each time, check if re-scheduling of the subtask removes the stations
        """
        with tmss_test_env.create_tmss_client() as client:
            subtask_template = client.get_subtask_template("observation control")
            spec = get_default_json_object_for_schema(subtask_template['schema'])
            spec['COBALT']['correlator']['enabled'] = True
            spec['stations']['digital_pointings'][0]['name'] = 'target0'
            spec['stations']['digital_pointings'][0]['subbands'] = [0]
            subtask = self._create_target_observation_subtask(spec, task_station_groups=[{'stations': ['CS002', 'CS003', 'CS004'], 'max_nr_missing': 2}])
            subtask_id = subtask['id']
            self.assertNotIn('CS004', subtask['specifications_doc']['stations']['station_list'])

            subtask = client.schedule_subtask(subtask_id)
            self.assertEqual('scheduled', subtask['state_value'])
            self.assertIn('CS002', subtask['specifications_doc']['stations']['station_list'])
            self.assertIn('CS003', subtask['specifications_doc']['stations']['station_list'])
            self.assertIn('CS004', subtask['specifications_doc']['stations']['station_list'])

            create_reserved_stations_for_testing(['CS002'])
            subtask = client.unschedule_subtask(subtask_id)
            subtask = client.schedule_subtask(subtask_id)
            self.assertEqual('scheduled', subtask['state_value'])
            self.assertNotIn('CS002', subtask['specifications_doc']['stations']['station_list'])
            self.assertIn('CS003', subtask['specifications_doc']['stations']['station_list'])
            self.assertIn('CS004', subtask['specifications_doc']['stations']['station_list'])

            create_reserved_stations_for_testing(['CS004'])
            subtask = client.unschedule_subtask(subtask_id)
            subtask = client.schedule_subtask(subtask_id)
            self.assertEqual('scheduled', subtask['state_value'])
            self.assertNotIn('CS002', subtask['specifications_doc']['stations']['station_list'])
            self.assertIn('CS003', subtask['specifications_doc']['stations']['station_list'])
            self.assertNotIn('CS004', subtask['specifications_doc']['stations']['station_list'])

    def _setup_observation_and_pipeline(self, client, obs_spec, dataproduct_properties, pipeline_task_template_name, pipeline_subtask_template_name, pipeline_subtask_spec):
            filesystem = test_data_creator.post_data_and_get_response_as_json_object(test_data_creator.Filesystem(), '/filesystem/')
            cluster_url = filesystem['cluster']

            # setup: first create an observation, so the pipeline can have input.
            obs_task_blueprint_data = test_data_creator.TaskBlueprint(template_url=client.get_task_template(name="target observation")['url'])
            obs_task_blueprint = test_data_creator.post_data_and_get_response_as_json_object(obs_task_blueprint_data, '/task_blueprint/')
            obs_subtask_template = client.get_subtask_template("observation control")

            obs_subtask_data = test_data_creator.Subtask(specifications_template_url=obs_subtask_template['url'],
                                                         specifications_doc=obs_spec,
                                                         cluster_url=cluster_url,
                                                         task_blueprint_url=obs_task_blueprint['url'])
            obs_subtask = test_data_creator.post_data_and_get_response_as_json_object(obs_subtask_data, '/subtask/')
            obs_outputrole = test_data_creator.post_data_and_get_response_as_json_object(test_data_creator.TaskConnectorType(dataformat='MeasurementSet', datatype='visibilities'), '/task_connector_type/')
            obs_subtask_output_url = test_data_creator.post_data_and_get_url(test_data_creator.SubtaskOutput(subtask_url=obs_subtask['url'], filesystem_url=filesystem['url'], outputrole_url=obs_outputrole['url']), '/subtask_output/')
            
            dp_spec_template = client.get_dataproduct_specifications_template("visibilities")
            test_data_creator.post_data_and_get_url(test_data_creator.Dataproduct(**dataproduct_properties,
                                                                                  specifications_template_url=dp_spec_template['url'],
                                                                                  subtask_output_url=obs_subtask_output_url), '/dataproduct/')

            client.set_subtask_status(obs_subtask['id'], 'defined')

            # now create the pipeline...
            pipe_task_blueprint_data = test_data_creator.TaskBlueprint(template_url=client.get_task_template(name=pipeline_task_template_name)['url'])
            pipe_task_blueprint = test_data_creator.post_data_and_get_response_as_json_object(pipe_task_blueprint_data, '/task_blueprint/')

            pipe_subtask_template = client.get_subtask_template(pipeline_subtask_template_name)
            pipe_spec = add_defaults_to_json_object_for_schema(pipeline_subtask_spec, pipe_subtask_template['schema'])

            pipe_subtask_data = test_data_creator.Subtask(specifications_template_url=pipe_subtask_template['url'],
                                                          specifications_doc=pipe_spec,
                                                          task_blueprint_url=pipe_task_blueprint['url'],
                                                          cluster_url=cluster_url)
            pipe_subtask = test_data_creator.post_data_and_get_response_as_json_object(pipe_subtask_data, '/subtask/')
            
            # ...and connect it to the observation
            task_relation_selection_template_url = client.get_path_as_json_object('/task_relation_selection_template?name=all')[0]['url']
            test_data_creator.post_data_and_get_url(test_data_creator.SubtaskInput(subtask_url=pipe_subtask['url'], subtask_output_url=obs_subtask_output_url, task_relation_selection_template_url=task_relation_selection_template_url), '/subtask_input/')
            pipe_outputrole = test_data_creator.post_data_and_get_response_as_json_object(test_data_creator.TaskConnectorType(dataformat='MeasurementSet', datatype='visibilities'), '/task_connector_type/')
            test_data_creator.post_data_and_get_url(test_data_creator.SubtaskOutput(subtask_url=pipe_subtask['url'], filesystem_url=filesystem['url'], outputrole_url=pipe_outputrole['url']), '/subtask_output/')

            client.set_subtask_status(pipe_subtask['id'], 'defined')

            return obs_subtask, pipe_subtask

    def test_schedule_preprocessing_pipeline_subtask_with_enough_resources_available(self):
        with tmss_test_env.create_tmss_client() as client:
            obs_subtask_template = client.get_subtask_template("observation control")
            obs_spec = get_default_json_object_for_schema(obs_subtask_template['schema'])
            obs_spec['stations']['digital_pointings'][0]['name'] = 'target0'
            obs_spec['stations']['digital_pointings'][0]['subbands'] = [0]
            obs_spec['COBALT']['correlator']['enabled'] = True

            obs_subtask, pipe_subtask = self._setup_observation_and_pipeline(client,
                                                                obs_spec,
                                                                {"filename": "L123456_SB000.MS",
                                                                 "specifications_doc": {"sap": "target0", "subband": 0 } },
                                                                "preprocessing pipeline",
                                                                "preprocessing pipeline",
                                                                {})

            # make sure that pipeline's predecessor (the obs_subtask) is finished
            for state in ('defined', 'scheduling', 'scheduled', 'starting', 'started', 'finishing', 'finished'):
                client.set_subtask_status(obs_subtask['id'], state)

            subtask = client.schedule_subtask(pipe_subtask['id'])

            self.assertEqual('scheduled', subtask['state_value'])


    def test_schedule_pipeline_for_cancelled_observation(self):
        '''scheduling a pipeline as a successor of a cancelled observation should fail,
        except when the cancelled observation is also obsolete'''
        with tmss_test_env.create_tmss_client() as client:
            obs_subtask_template = client.get_subtask_template("observation control")
            obs_spec = get_default_json_object_for_schema(obs_subtask_template['schema'])
            obs_spec['stations']['digital_pointings'][0]['name'] = 'target0'
            obs_spec['stations']['digital_pointings'][0]['subbands'] = [0]

            obs_subtask, pipe_subtask = self._setup_observation_and_pipeline(client,
                                                                obs_spec,
                                                                {"filename": "L123456_SB000.MS",
                                                                 "specifications_doc": {"sap": "target0", "subband": 0 } },
                                                                "preprocessing pipeline",
                                                                "preprocessing pipeline",
                                                                {})

            # cancel the observation
            obs_subtask = client.cancel_subtask(obs_subtask['id'])

            # check, should be cancelled, but not obsolete
            self.assertEqual('cancelled', obs_subtask['state_value'])
            self.assertIsNone(obs_subtask['obsolete_since'])

            # scheduling pipeline should fail
            with self.assertRaises(Exception) as context:
                pipe_subtask = client.schedule_subtask(pipe_subtask['id'])
                self.assertTrue('Cannot schedule subtask' in str(context.exception))
                self.assertTrue('not FINISHED but state=cancelled' in str(context.exception))

            # now mark the cancelled observation as obsolete
            obs_subtask = client.mark_subtask_as_obsolete(obs_subtask['id'])

            # check, should (still) be cancelled, and now obsolete
            self.assertEqual('cancelled', obs_subtask['state_value'])
            self.assertIsNotNone(obs_subtask['obsolete_since'])

            # scheduling pipeline should now be a success
            pipe_subtask = client.schedule_subtask(pipe_subtask['id'])
            self.assertEqual('scheduled', pipe_subtask['state_value'])


    @unittest.skip("TODO: add missing coherent stokes settings")
    def test_schedule_pulsar_pipeline_subtask_with_enough_resources_available(self):
        with tmss_test_env.create_tmss_client() as client:
            obs_subtask_template = client.get_subtask_template("observation control")
            obs_spec = {
              "stations": { "digital_pointings": [ { "name": "target0", "subbands": [0] } ] },
              "COBALT": {
                "version": 1,
                "correlator": { "enabled": False },
                "beamformer": {
                    "tab_pipelines": [
                      {
                        "SAPs": [ { "name": "target0", "tabs": [ { "coherent": False }, { "coherent": True } ] } ]
                      }
                    ]
                }
              }
            }
            obs_spec = add_defaults_to_json_object_for_schema(obs_spec,obs_subtask_template['schema'])

            pulp_dataproduct_specifications_template = client.get_dataproduct_specifications_template("pulp summary")

            pipe_subtask = self._setup_observation_and_pipeline(client,
                                                                obs_spec,
                                                                {"filename": "L123456_SAP000_B000_S0_P000.h5",
                                                                 "specifications_template_url": pulp_dataproduct_specifications_template['url'],
                                                                 "specifications_doc": { "sap": "target0", "coherent": True, "stokes_set": "I", "identifiers": { "sap_index": 0, "tab_index": 0, "pipeline_index": 0, "part_index": 0, "stokes_index": 0 } } },
                                                                "pulsar pipeline",
                                                                "pulsar pipeline",
                                                                {})

            # make sure that pipeline's predecessors are finished
            for predecessor in client.get_subtask_predecessors(pipe_subtask['id']):
                for state in ('defined', 'scheduling', 'scheduled', 'starting', 'started', 'finishing', 'finished'):
                    client.set_subtask_status(predecessor['id'], state)

            subtask = client.schedule_subtask(pipe_subtask['id'])

            self.assertEqual('scheduled', subtask['state_value'])


    def test_schedule_ingest_subtask(self):
        with tmss_test_env.create_tmss_client() as client:
            cluster_url = client.get_path_as_json_object('/cluster/1')['url']

            # setup: first create an observation, so the ingest can have input.
            subtask_template = client.get_subtask_template("observation control")
            obs_spec = get_default_json_object_for_schema(subtask_template['schema'])
            obs_spec['stations']['digital_pointings'][0]['name'] = 'target0'
            obs_spec['stations']['digital_pointings'][0]['subbands'] = [0]
            obs_subtask = self._create_target_observation_subtask(obs_spec)
            obs_subtask_id = obs_subtask['id']
            obs_subtask_output_url = client.get_path_as_json_object('/subtask_output?subtask=%s'%obs_subtask_id)[0]['url']

            dp_spec_template_url = client.get_path_as_json_object('/dataproduct_specifications_template?name=visibilities')[0]['url']
            test_data_creator.post_data_and_get_url(test_data_creator.Dataproduct(filename="L%s_SB000.MS"%obs_subtask['id'],
                                                                                  specifications_template_url=dp_spec_template_url,
                                                                                  specifications_doc={"sap": "target0", "subband": 0},
                                                    subtask_output_url=obs_subtask_output_url), '/dataproduct/')

            # now create the ingest...
            ingest_subtask_template = client.get_subtask_template("ingest control")
            ingest_spec = get_default_json_object_for_schema(ingest_subtask_template['schema'])

            ingest_subtask_data = test_data_creator.Subtask(specifications_template_url=ingest_subtask_template['url'],
                                                          specifications_doc=ingest_spec,
                                                          task_blueprint_url=obs_subtask['task_blueprint'],
                                                          primary=False,
                                                          cluster_url=cluster_url)
            ingest_subtask = test_data_creator.post_data_and_get_response_as_json_object(ingest_subtask_data, '/subtask/')

            # ...and connect it to the observation
            task_relation_selection_template_url = client.get_path_as_json_object('/task_relation_selection_template?name=all')[0]['url']
            test_data_creator.post_data_and_get_url(test_data_creator.SubtaskInput(subtask_url=ingest_subtask['url'], subtask_output_url=obs_subtask_output_url, task_relation_selection_template_url=task_relation_selection_template_url), '/subtask_input/')
            test_data_creator.post_data_and_get_url(test_data_creator.SubtaskOutput(subtask_url=ingest_subtask['url']), '/subtask_output/')  # our subtask here has only one known related task

            for predecessor in client.get_subtask_predecessors(ingest_subtask['id']):
                for state in ('defined', 'scheduling', 'scheduled', 'starting', 'started', 'finishing', 'finished'):
                    client.set_subtask_status(predecessor['id'], state)

            client.set_subtask_status(ingest_subtask['id'], 'defined')

            task_blueprint = client.get_url_as_json_object(ingest_subtask['task_blueprint'])  # our subtask here has only one known related task
            schedulingunit_blueprint = client.get_url_as_json_object(task_blueprint['scheduling_unit_blueprint'])

            # first, make sure we need but do not have ingest persmission...
            client.session.patch(schedulingunit_blueprint['url'], json={'ingest_permission_required': True, 'ingest_permission_granted_since': None})

            with self.assertRaises(Exception) as context:
                subtask = client.schedule_subtask(ingest_subtask['id'])
                self.assertTrue('permission' in str(context.exception))

            subtask = client.get_subtask(ingest_subtask['id'])
            self.assertEqual('defined', subtask['state_value'])

            # now grant permission...
            client.session.patch(schedulingunit_blueprint['url'], json={'ingest_permission_required': True, 'ingest_permission_granted_since': datetime.utcnow().isoformat()})

            subtask = client.schedule_subtask(ingest_subtask['id'])

            self.assertEqual('scheduled', subtask['state_value'])
            self.assertEqual(models.Subtask.objects.get(id=ingest_subtask['id']).inputs.first().dataproducts.count(), 1)


    def test_schedule_schedulingunit_enough_resources_available(self):
        '''similar test as test_schedule_pipeline_subtask_with_enough_resources_available, but now created from a scheduling_unit'''
        with tmss_test_env.create_tmss_client() as client:
            scheduling_unit_template = client.get_schedulingunit_template("scheduling unit")
            scheduling_unit_doc = get_default_json_object_for_schema(scheduling_unit_template['schema'])

            # define an observation without QA
            obs_task = get_default_json_object_for_schema(client.get_task_template(name="target observation")['schema'])
            obs_task['QA']['plots']['enabled'] = False
            obs_task['QA']['file_conversion']['enabled'] = False
            obs_task['station_configuration']['SAPs'] = [{ 'subbands': [0,1] }]
            scheduling_unit_doc['tasks']["Observation"] = {"specifications_doc": obs_task,
                                                           "specifications_template": {"name": "target observation"}}

            # define a pipeline
            scheduling_unit_doc['tasks']["Pipeline"] = { "specifications_doc": get_default_json_object_for_schema(client.get_task_template(name="preprocessing pipeline")['schema']),
                                                         "specifications_template": {"name": "preprocessing pipeline"}}

            # connect obs to pipeline
            scheduling_unit_doc['task_relations'].append({"producer": "Observation",
                                                          "consumer": "Pipeline",
                                                          "input": { "role": "any", "datatype": "visibilities", "dataformat": "MeasurementSet"},
                                                          "output": { "role": "correlator", "datatype": "visibilities", "dataformat": "MeasurementSet"},
                                                          "selection_doc": {},
                                                          "selection_template": {"name": "all"} })

            # submit
            scheduling_unit_draft_data = test_data_creator.SchedulingUnitDraft(template_url=scheduling_unit_template['url'],
                                                                               specifications_doc=scheduling_unit_doc)
            scheduling_unit_draft = test_data_creator.post_data_and_get_response_as_json_object(scheduling_unit_draft_data, '/scheduling_unit_draft/')

            # create the whole blueprints tree...
            scheduling_unit_blueprint = client.create_scheduling_unit_blueprint_and_tasks_and_subtasks_tree(scheduling_unit_draft['id'])

            # fetch the created task_blueprints
            task_blueprints = [client.get_url_as_json_object(task_blueprint['url']) for task_blueprint in scheduling_unit_blueprint['task_blueprints']]
            self.assertEqual(2, len(task_blueprints))
            # and make sure they are ordered correctly
            if "Pipeline" in task_blueprints[0]['name']:
                task_blueprints.reverse()

            for task_blueprint in task_blueprints:
                self.assertEqual(1, len(task_blueprint['subtasks']))
                subtask = client.get_url_as_json_object(task_blueprint['subtasks'][0])

                client.session.patch(subtask['url'], {'scheduled_start_time': datetime.utcnow() + timedelta(minutes=5)})
                client.set_subtask_status(subtask['id'], 'defined')

                subtask = client.schedule_subtask(subtask['id'])

                self.assertEqual('scheduled', subtask['state_value'])

                for state in ('starting', 'started', 'finishing', 'finished'):
                    client.set_subtask_status(subtask['id'], state)


class SubtaskInputOutputTest(unittest.TestCase):
    """
    Subtask Input and Output test
    These testcases are located in the t_scheduling module, because during scheduling the output
    dataproducts are assigned
    """
    def setUp(self) -> None:
        tmss_test_env.delete_scheduling_unit_drafts_cascade()

    def test_specifications_doc_meets_selection_doc(self):
        # empty selection matches all
        self.assertTrue(specifications_doc_meets_selection_doc({'something else': 'target0'}, {}))

        # specification is a list? specification must be a subset of the selection
        self.assertTrue(specifications_doc_meets_selection_doc({'sap': ['target0']}, {'sap': ['target0']}))
        self.assertFalse(specifications_doc_meets_selection_doc({'sap': ['target0','target1','target2']}, {'sap': ['target0','target1']}))

        # specification is a value? it must appear in the selection
        self.assertTrue(specifications_doc_meets_selection_doc({'sap': 'target0'}, {'sap': ['target0']}))
        self.assertTrue(specifications_doc_meets_selection_doc({'sap': 'target0'}, {'sap': ['target0','target1']}))
        self.assertTrue(specifications_doc_meets_selection_doc({'sap': 'target0'}, {'sap': 'target0'}))

        # specification must contain the selection key
        self.assertFalse(specifications_doc_meets_selection_doc({'something else': 'target0'}, {'sap': 'target0'}))

    def test_schedule_pipeline_subtask_filters_predecessor_output_dataproducts_for_input(self):
        # setup:
        #   create observation subtask and outputs and dataproducts
        obs_st = create_subtask_object_for_testing('observation', 'finished')
        obs_out1 = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=obs_st))
        obs_out2 = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=obs_st))

        #   create connected pipeline subtask and inputs, specify input filtering
        trelsel_template = models.TaskRelationSelectionTemplate.get_latest(name='SAP')
        pipe_st = create_subtask_object_for_testing('pipeline', 'defined')
        filesystem = pipe_st.cluster.filesystem_set.first()
        output_role = models.TaskConnectorType.objects.create(**TaskConnectorType_test_data(datatype='visibilities', dataformat='MeasurementSet'))
        pipe_out = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=pipe_st, output_role=output_role, filesystem=filesystem)) # required by scheduling function
        pipe_in1 = models.SubtaskInput.objects.create(**SubtaskInput_test_data(subtask=pipe_st, producer=obs_out1, selection_template=trelsel_template,  selection_doc={'sap': ['target0']}))
        pipe_in2 = models.SubtaskInput.objects.create(**SubtaskInput_test_data(subtask=pipe_st, producer=obs_out2, selection_template=trelsel_template, selection_doc={'sap': ['target1']}))

        #   create obs output dataproducts with specs we can filter on
        dp_spec_template = models.DataproductSpecificationsTemplate.get_latest(name='visibilities')
        dp1_1 = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=obs_out1, specifications_template=dp_spec_template, specifications_doc={'sap': 'target0', 'subband': 0}))
        dp1_2 = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=obs_out1, specifications_template=dp_spec_template, specifications_doc={'sap': 'target1', 'subband': 0}))
        dp1_3 = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=obs_out1, specifications_template=dp_spec_template, specifications_doc={'sap': 'target0', 'subband': 1}))

        dp2_1 = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=obs_out2, specifications_template=dp_spec_template, specifications_doc={'sap': 'target0', 'subband': 0}))
        dp2_2 = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=obs_out2, specifications_template=dp_spec_template, specifications_doc={'sap': 'target1', 'subband': 0}))

        # trigger:
        #   schedule pipeline, which should attach the correct subset of dataproducts to the pipeline inputs
        schedule_pipeline_subtask(pipe_st)

        # assert:
        #   check correct input filtering
        self.assertEqual(set(pipe_in1.dataproducts.all()), {dp1_1, dp1_3})
        self.assertEqual(set(pipe_in2.dataproducts.all()), {dp2_2})

    def test_bug_TMSS_1124_ingest_cannot_be_scheduled_due_to_unfinished_subtask_input_output_implementation(self):
        '''
        See Jira: https://support.astron.nl/jira/browse/TMSS-1124
        This bug seemed to indicate that something is wrong in scheduling ingest tasks.
        The root-cause is the fact that currently conversion from task_relations into SubtaskInput/Subtaskoutputs
        does not filter for datatype/dataformat.
        In this initial version of this test we prove that the bug exists, and then we fix it in a later commit.
        '''
        # Setup: create a SchedulingUnitDraft/Blueprint with a BF Obs, Pulp, and Ingest
        parent_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data())
        su_draft = models.SchedulingUnitDraft.objects.create(scheduling_set=parent_set,
                                                             specifications_template=models.SchedulingUnitTemplate.get_latest(name='scheduling unit'))

        obs_task_template = models.TaskTemplate.get_latest(name='beamforming observation')
        obs_specifications_doc = obs_task_template.get_default_json_document_for_schema()
        obs_specifications_doc['station_configuration']['SAPs'][0]['subbands'] = [0,1,2,3]
        obs_specifications_doc['beamformer']['pipelines'][0]['coherent']['SAPs'] = []
        obs_specifications_doc['beamformer']['pipelines'][0]['incoherent']['SAPs'] = []
        obs_specifications_doc['beamformer']['pipelines'][0]['flys eye']['enabled'] = True
        obs_task_draft = models.TaskDraft.objects.create(name='obs',
                                                         scheduling_unit_draft=su_draft,
                                                         specifications_template=obs_task_template,
                                                         specifications_doc=obs_specifications_doc)

        pulp_task_template = models.TaskTemplate.get_latest(name='pulsar pipeline')
        pulp_task_draft = models.TaskDraft.objects.create(name='pulp',
                                                          scheduling_unit_draft=su_draft,
                                                          specifications_template=pulp_task_template,
                                                          specifications_doc=pulp_task_template.get_default_json_document_for_schema())

        # connect the obs and the pulp pipeline
        # it's a single relation of beamformed timeseries data
        trsel_template = models.TaskRelationSelectionTemplate.get_latest(name='all')
        trsel_obs_pulp_draft = models.TaskRelationDraft.objects.create(producer=obs_task_draft,
                                                                       consumer=pulp_task_draft,
                                                                       output_role=obs_task_template.connector_types.get(iotype__value=models.IOType.Choices.OUTPUT.value,
                                                                                                                         dataformat__value=models.Dataformat.Choices.BEAMFORMED.value,
                                                                                                                         datatype__value=models.Datatype.Choices.TIME_SERIES.value,
                                                                                                                         role__value=models.Role.Choices.BEAMFORMER.value),
                                                                       input_role=pulp_task_template.connector_types.get(iotype__value=models.IOType.Choices.INPUT.value,
                                                                                                                         dataformat__value=models.Dataformat.Choices.BEAMFORMED.value,
                                                                                                                         datatype__value=models.Datatype.Choices.TIME_SERIES.value,
                                                                                                                         role__value=models.Role.Choices.BEAMFORMER.value),
                                                                       selection_template=trsel_template,
                                                                       selection_doc=trsel_template.get_default_json_document_for_schema())

        ingest_task_template = models.TaskTemplate.get_latest(name='ingest')
        ingest_task_draft = models.TaskDraft.objects.create(name='ingest',
                                                            scheduling_unit_draft=su_draft,
                                                            specifications_template=ingest_task_template,
                                                            specifications_doc=ingest_task_template.get_default_json_document_for_schema())

        # connect the pulsar pipeline and ingest
        # there are two connections, 1) pulp summary, 2) pulsar profile
        # this should result in two connections at SubtaskInput/SubtaskOutput level, but it does not (which is the bug)
        trsel_pulp_ingest_draft1 = models.TaskRelationDraft.objects.create(producer=pulp_task_draft,
                                                                           consumer=ingest_task_draft,
                                                                           output_role=pulp_task_template.connector_types.get(iotype__value=models.IOType.Choices.OUTPUT.value,
                                                                                                                              dataformat__value=models.Dataformat.Choices.PULP_SUMMARY.value,
                                                                                                                              datatype__value=models.Datatype.Choices.QUALITY.value,
                                                                                                                              role__value=models.Role.Choices.ANY.value),
                                                                           input_role=ingest_task_template.connector_types.get(iotype__value=models.IOType.Choices.INPUT.value,
                                                                                                                               dataformat__value=models.Dataformat.Choices.PULP_SUMMARY.value,
                                                                                                                               datatype__value=models.Datatype.Choices.QUALITY.value,
                                                                                                                               role__value=models.Role.Choices.ANY.value),
                                                                           selection_template=trsel_template,
                                                                           selection_doc=trsel_template.get_default_json_document_for_schema())


        trsel_pulp_ingest_draft2 = models.TaskRelationDraft.objects.create(producer=pulp_task_draft,
                                                                           consumer=ingest_task_draft,
                                                                           output_role=pulp_task_template.connector_types.get(iotype__value=models.IOType.Choices.OUTPUT.value,
                                                                                                                              dataformat__value=models.Dataformat.Choices.PULP_ANALYSIS.value,
                                                                                                                              datatype__value=models.Datatype.Choices.PULSAR_PROFILE.value,
                                                                                                                              role__value=models.Role.Choices.ANY.value),
                                                                           input_role=ingest_task_template.connector_types.get(iotype__value=models.IOType.Choices.INPUT.value,
                                                                                                                               dataformat__value=models.Dataformat.Choices.PULP_ANALYSIS.value,
                                                                                                                               datatype__value=models.Datatype.Choices.PULSAR_PROFILE.value,
                                                                                                                               role__value=models.Role.Choices.ANY.value),
                                                                           selection_template=trsel_template,
                                                                           selection_doc=trsel_template.get_default_json_document_for_schema())


        su_draft.ingest_permission_required = False
        su_draft.save()

        # convert to blueprint
        su_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(su_draft)

        # fetch the created subtasks
        obs_subtak = models.Subtask.objects.get(task_blueprint__scheduling_unit_blueprint=su_blueprint, specifications_template__type__value='observation')
        pulp_subtak = models.Subtask.objects.get(task_blueprint__scheduling_unit_blueprint=su_blueprint, specifications_template__type__value='pipeline')
        ingest_subtak = models.Subtask.objects.get(task_blueprint__scheduling_unit_blueprint=su_blueprint, specifications_template__type__value='ingest')

        # scheduling the observation should succeed and creates the obs output dataproducts
        schedule_subtask(obs_subtak)
        set_subtask_state_following_allowed_transitions(obs_subtak, 'finished')

        # scheduling the pipeline should succeed and creates the pulp output dataproducts
        schedule_subtask(pulp_subtak)
        set_subtask_state_following_allowed_transitions(pulp_subtak, 'finished')

        # scheduling ingest should succeed and take all pulp's output_dataproducts as input via two SubtaskInput/Outputs
        schedule_subtask(ingest_subtak)

    def test_bug_TMSS_2291_do_not_create_flys_eye_dataproducts_for_unavailable_stations(self):
        '''
        See Jira: https://support.astron.nl/jira/browse/TMSS-2291
        When scheduling a flys eye subtask, do not create dataproducts for all stations that are in the pipeline specs,
        but only for those that are actually part of the observation.
        '''
        # Setup: create a SchedulingUnitDraft/Blueprint with a BF Obs, Pulp, and Ingest
        parent_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data())
        su_draft = models.SchedulingUnitDraft.objects.create(scheduling_set=parent_set,
                                                             specifications_template=models.SchedulingUnitTemplate.get_latest(name='scheduling unit'))

        obs_task_template = models.TaskTemplate.get_latest(name='beamforming observation')
        obs_specifications_doc = obs_task_template.get_default_json_document_for_schema()
        obs_specifications_doc['station_configuration']['SAPs'][0]['subbands'] = [0,1,2,3]
        station_groups = [{'stations': ['CS004', 'CS002', 'CS003'], 'max_nr_missing': 2}]
        obs_specifications_doc['station_configuration']['station_groups'] = station_groups
        obs_specifications_doc['beamformer']['pipelines'][0]['coherent']['SAPs'] = []
        obs_specifications_doc['beamformer']['pipelines'][0]['incoherent']['SAPs'] = []
        obs_specifications_doc['beamformer']['pipelines'][0]['flys eye']['enabled'] = True
        obs_specifications_doc['beamformer']['pipelines'][0]['station_groups'] = station_groups
        obs_task_draft = models.TaskDraft.objects.create(name='obs',
                                                         scheduling_unit_draft=su_draft,
                                                         specifications_template=obs_task_template,
                                                         specifications_doc=obs_specifications_doc)

        su_draft.ingest_permission_required = False
        su_draft.save()

        # convert to blueprint
        su_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(su_draft)

        # fetch the created subtasks
        obs_subtask = models.Subtask.objects.get(task_blueprint__scheduling_unit_blueprint=su_blueprint, specifications_template__type__value='observation')

        # scheduling the observation should succeed and creates the obs output dataproducts
        schedule_subtask(obs_subtask)
        self.assertEqual(['CS002', 'CS003', 'CS004'], obs_subtask.specifications_doc["stations"]["station_list"])
        self.assertEqual(obs_subtask.output_dataproducts.count(), 6)

        # create reservation to block CS004 and reschedule
        # check number of output dataproducts has changed accordingly
        unschedule_subtask(obs_subtask)
        create_reserved_stations_for_testing(['CS004'])
        schedule_subtask(obs_subtask)
        self.assertEqual(['CS002', 'CS003'], obs_subtask.specifications_doc["stations"]["station_list"])
        self.assertEqual(obs_subtask.output_dataproducts.count(), 4)

        # make CS004 available again, reserve the other stations and reschedule
        # check number of output dataproducts has changed accordingly
        models.Reservation.objects.all().delete()
        unschedule_subtask(obs_subtask)
        create_reserved_stations_for_testing(['CS002', 'CS003'])
        schedule_subtask(obs_subtask)
        self.assertEqual(['CS004'], obs_subtask.specifications_doc["stations"]["station_list"])
        self.assertEqual(obs_subtask.output_dataproducts.count(), 2)


class SAPTest(unittest.TestCase):
    """
    SAP test
    These testcases are located in the t_scheduling module, because the SAP entries are created/assigned during scheduling
    """

    def setUp(self) -> None:
        # make sure we're allowed to schedule
        setting = Setting.objects.get(name='dynamic_scheduling_enabled')
        setting.value = True
        setting.save()

        # make sure results from other tests are not interfering
        tmss_test_env.delete_scheduling_unit_drafts_cascade()

    def test_schedule_observation_subtask_creates_sap_with_correct_pointing(self):
        with tmss_test_env.create_tmss_client() as client:
            task_blueprint_data = test_data_creator.TaskBlueprint(template_url=client.get_task_template(name="target observation")['url'])
            task_blueprint_data['specifications_doc']['station_configuration']['SAPs'][0]['name'] = 'target0'
            task_blueprint = test_data_creator.post_data_and_get_response_as_json_object(task_blueprint_data, '/task_blueprint/')
            subtask_template = client.get_subtask_template("observation control")
            spec = get_default_json_object_for_schema(subtask_template['schema'])
            spec['stations']['digital_pointings'][0]['name'] = task_blueprint_data['specifications_doc']['station_configuration']['SAPs'][0]['name']
            spec['stations']['digital_pointings'][0]['subbands'] = [0]
            filesystem = test_data_creator.post_data_and_get_response_as_json_object(test_data_creator.Filesystem(), '/filesystem/')
            cluster_url = filesystem['cluster']
            pointing = {"angle1": 7.6, "angle2": 5.4, "direction_type": "J2000", "target": "target1"}
            spec['stations']['digital_pointings'][0]['pointing'] = pointing

            subtask_data = test_data_creator.Subtask(specifications_template_url=subtask_template['url'],
                                                     specifications_doc=spec,
                                                     cluster_url = cluster_url,
                                                     task_blueprint_url=task_blueprint['url'],
                                                     scheduled_start_time=datetime.utcnow() + timedelta(minutes=5),
                                                     scheduled_stop_time=datetime.utcnow() + timedelta(minutes=15))
            subtask = test_data_creator.post_data_and_get_response_as_json_object(subtask_data, '/subtask/')
            subtask_id = subtask['id']
            subtaskoutput_role = test_data_creator.post_data_and_get_response_as_json_object(test_data_creator.TaskConnectorType(dataformat="MeasurementSet", datatype='visibilities'), '/task_connector_type/')

            test_data_creator.post_data_and_get_url(test_data_creator.SubtaskOutput(subtask_url=subtask['url'], filesystem_url=filesystem['url'], outputrole_url=subtaskoutput_role['url']), '/subtask_output/')

            subtask_model = models.Subtask.objects.get(id=subtask_id)
            self.assertEqual(0, subtask_model.output_dataproducts.values('sap').count())

            client.set_subtask_status(subtask_id, 'defined')
            subtask = client.schedule_subtask(subtask_id)

            self.assertEqual(1, subtask_model.output_dataproducts.count())
            self.assertEqual(1, subtask_model.output_dataproducts.values('sap').count())
            self.assertEqual(subtask_model.output_dataproducts.first().sap.specifications_doc['pointing']['angle1'], pointing['angle1'])
            self.assertEqual(subtask_model.output_dataproducts.first().sap.specifications_doc['pointing']['angle2'], pointing['angle2'])

    def test_schedule_pipeline_subtask_copies_sap_from_input_to_output(self):
        # setup:
        #   create observation subtask and outputs and dataproducts
        obs_st = create_subtask_object_for_testing('observation', 'finished')
        obs_out = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=obs_st))

        #   create connected pipeline subtask and inputs, specify input filtering
        pipe_st = create_subtask_object_for_testing('pipeline', 'defined')
        pipe_filesystem = pipe_st.cluster.filesystem_set.first()
        pipe_output_role = models.TaskConnectorType.objects.create(**TaskConnectorType_test_data(datatype='visibilities', dataformat='MeasurementSet'))
        pipe_out = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=pipe_st, output_role=pipe_output_role, filesystem=pipe_filesystem)) # required by scheduling function
        pipe_in = models.SubtaskInput.objects.create(**SubtaskInput_test_data(subtask=pipe_st, producer=obs_out))

        #   create obs output dataproducts
        dp1_in = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=obs_out)) #, specifications_doc={"identifiers": { "sap_index": 0, "subband_index": 0 }}))
        dp2_in = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=obs_out)) #, specifications_doc={"identifiers": { "sap_index": 0, "subband_index": 1 }}))

        #   schedule pipeline, which should copy the SAP
        schedule_pipeline_subtask(pipe_st)

        # determine the newly created pipeline dataproducts
        dp1_out = DataproductTransform.objects.get(input=dp1_in).output
        dp2_out = DataproductTransform.objects.get(input=dp2_in).output

        # assert:
        self.assertEqual(dp1_in.sap, dp1_out.sap)
        self.assertEqual(dp2_in.sap, dp2_out.sap)


if __name__ == "__main__":
    os.environ['TZ'] = 'UTC'
    unittest.main()
