#!/usr/bin/env python3

# Copyright (C) 2018    ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

# $Id:  $
import copy
import os
import threading
import time
import unittest
import requests

import logging

from lofar.common.test_utils import exit_with_skipped_code_if_skip_integration_tests
from lofar.sas.tmss import client

exit_with_skipped_code_if_skip_integration_tests()

from lofar.common.json_utils import get_default_json_object_for_schema, add_defaults_to_json_object_for_schema, without_schema_property

# Do Mandatory setup step:
# use setup/teardown magic for tmss test database, ldap server and django server
from lofar.sas.tmss.test.test_environment import TMSSTestEnvironment
tmss_test_env = TMSSTestEnvironment(populate_schemas=True, start_postgres_listener=True)
try:
    tmss_test_env.start()
except Exception as e:
    print("Failed to start test environment")
    tmss_test_env.stop()
    exit(1)

# tell unittest to stop (and automagically cleanup) the test database once all testing is done.
def tearDownModule():
    tmss_test_env.stop()

from lofar.sas.tmss.test.tmss_test_data_django_models import *

# setup rest test data creator
rest_data_creator = tmss_test_env.create_test_data_creator()

from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.exceptions import SchemaValidationException, ObsoleteTemplateException, ValidationException, ObsoleteValidationException, RuleValidationException

import requests

from lofar.sas.tmss.tmss.tmssapp.tasks import create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft, create_scheduling_unit_blueprint_from_scheduling_unit_draft, update_task_blueprint_graph_from_draft, update_task_graph_from_specifications_doc, create_scheduling_unit_draft_from_observing_strategy_template, mark_task_blueprint_as_obsolete, cancel_task_blueprint, schedule_independent_subtasks_in_task_blueprint, create_scheduling_unit_draft_from_scheduling_unit_blueprint
from lofar.sas.tmss.tmss.tmssapp.subtasks import schedule_subtask, cancel_subtask, wait_for_subtask_status
from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions
from lofar.messaging.messagebus import BusListenerJanitor
from lofar.sas.tmss.tmss.tmssapp.conversions import get_lofar2_stations

logger = logging.getLogger(__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

class SchedulingUnitBlueprintStateTest(unittest.TestCase):
    """
    Test the Scheduling Blueprint State which is derived from the TaskBlueprint states.
    The result of each possible combination of these states will be checked
    See https://support.astron.nl/confluence/display/TMSS/Specification+Flow#SpecificationFlow-SchedulingBlueprints
    """

    def setUp(self) -> None:
        tmss_test_env.delete_scheduling_unit_drafts_cascade()
        tmss_test_env.delete_reservations_cascade()


    def create_tasks_and_subtasks(self, schedulingunit_blueprint):
        """
        Create three taskblueprint related to the schedulingunit_blueprint.
        These task are an observation, a pipeline and a ingest task.
        Also per task one subtask is instantiated (so makes three total) which is required to be able to set
        the task status which is a read-only property and is derived from the subtask states
        :param schedulingunit_blueprint:
        :return: dictionary with task and subtask objects
        """
        # Create observation task
        task_data = TaskBlueprint_test_data(name="Task Observation "+str(uuid.uuid4()), scheduling_unit_blueprint=schedulingunit_blueprint)
        task_obs = models.TaskBlueprint.objects.create(**task_data)
        subtask_data = Subtask_test_data(state=models.SubtaskState.objects.get(value="defining"),
                                         subtask_template=models.SubtaskTemplate.get_latest(name='observation control'),
                                         task_blueprint=task_obs)
        subtask_obs = models.Subtask.objects.create(**subtask_data)
        subtask_obs.state = models.SubtaskState.objects.get(value="defined")
        subtask_obs.save()

        # Create pipeline task
        task_data = TaskBlueprint_test_data(name="Task Pipeline", scheduling_unit_blueprint=schedulingunit_blueprint,
                                            specifications_template=models.TaskTemplate.objects.filter(type=models.TaskType.Choices.PIPELINE.value).first())
        task_pipe = models.TaskBlueprint.objects.create(**task_data)
        subtask_data = Subtask_test_data(state=models.SubtaskState.objects.get(value="defining"),
                                         subtask_template=models.SubtaskTemplate.get_latest(name='preprocessing pipeline'),
                                         task_blueprint=task_pipe)
        subtask_pipe = models.Subtask.objects.create(**subtask_data)
        subtask_pipe.state = models.SubtaskState.objects.get(value="defined")
        subtask_pipe.save()

        # Create ingest task
        task_data = TaskBlueprint_test_data(name="Task Ingest", scheduling_unit_blueprint=schedulingunit_blueprint,
                    specifications_template = models.TaskTemplate.objects.filter(type=models.TaskType.Choices.INGEST.value).first())
        task_ingest = models.TaskBlueprint.objects.create(**task_data)
        # There is no template defined for ingest yet ...but I can use preprocessing pipeline, only the template type matters
        # ....should become other thing in future but for this test does not matter
        subtask_data = Subtask_test_data(state=models.SubtaskState.objects.get(value="defining"),
                                         subtask_template=models.SubtaskTemplate.get_latest(name='preprocessing pipeline'),
                                         task_blueprint=task_ingest)

        subtask_ingest = models.Subtask.objects.create(**subtask_data)
        subtask_ingest.state = models.SubtaskState.objects.get(value="defined")
        subtask_ingest.save()

        return {"observation": {"task": task_obs, "subtask": subtask_obs},
                "pipeline": {"task": task_pipe, "subtask": subtask_pipe},
                "ingest": {"task": task_ingest, "subtask": subtask_ingest}}

    def set_task_state(self, task_state, task_type, task, subtask):
        """
        Set the taskblueprint state for given task_type
        State of task can only be set by setting the subtask state
        Do not set subtask state if subtask is None
        :param task_state: Task state to be set
        :param task_type: observation, pipeline or ingest
        :param task: TaskBlueprint object
        :param subtask: SubTask object
        """
        # Translate task state to subtask state, mostly one-o-one but two exceptions
        if task_state == "observed":
            subtask_state = "finishing"
        elif task_state == "schedulable":
            subtask_state = "scheduling"
        else:
            subtask_state = task_state

        if subtask is not None:
            from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions
            subtask = set_subtask_state_following_allowed_transitions(subtask, subtask_state)
            self.assertEqual(subtask_state, subtask.state.value, "Could not set subtask to expected state")

        task.refresh_from_db()
        # Check task.status as precondition
        self.assertEqual(task_state, task.status.value,
                         "INCORRECT PRECONDITION. Expected %s task to have status=%s, but actual status=%s)" % (
                         task_type, task_state, task.status.value))

    def test_state_with_no_tasks(self):
        """
        Test the schedulingunitblueprint state when tasks are not instantiated.
        the expected state should be 'defined'
        """
        schedulingunit_data = SchedulingUnitBlueprint_test_data(name="Scheduling Blueprint No Tasks")
        schedulingunit_blueprint = models.SchedulingUnitBlueprint.objects.create(**schedulingunit_data)
        self.assertEqual("defined", schedulingunit_blueprint.status.value)

    def test_states_with_observation_pipeline_ingest_tasks_subtasks(self):
        """
        Test the schedulingunitblueprint state when only one task is instantiated, an pipeline
        Subtask are also instantiated so minimal task state is schedulable !
        See next table where every row represents:
            Taskstate(obs),  Taskstate(pipeline), Taskstate(ingest), Expected SchedulingUnitBlueprint Status
        """
        test_table = [
            # normal behaviour
            ("error",       "schedulable", "schedulable",  "error"),
            ("cancelled",   "schedulable", "schedulable",  "cancelled"),
            ("schedulable", "schedulable", "schedulable",  "schedulable"),
            ("unschedulable","schedulable","schedulable",  "unschedulable"),
            ("scheduled",   "schedulable", "schedulable",  "scheduled"),
            ("queued",      "schedulable", "schedulable",  "observing"),
            ("started",     "schedulable", "schedulable",  "observing"),
            ("observed",    "schedulable", "schedulable",  "observed"),
            ("observed",    "scheduled",   "schedulable",  "observed"),
            ("observed",    "started",     "schedulable",  "processing"),
            ("observed",    "finished",    "schedulable",  "processed"),
            ("observed",    "finished",    "scheduled",    "processed"),
            ("observed",    "finished",    "started",      "ingesting"),
            ("observed",    "finished",    "finished",     "finished"),
            ("finished",    "schedulable", "schedulable",  "observed"),
            ("finished",    "scheduled",   "schedulable",  "observed"),
            ("finished",    "started",     "schedulable",  "processing"),
            ("finished",    "finished",    "schedulable",  "processed"),
            ("finished",    "finished",    "scheduled",    "processed"),
            ("finished",    "finished",    "started",      "ingesting"),
            ("finished",    "finished",    "finished",     "finished"),
            # any cancelled
            ("observed",    "cancelled",   "schedulable",  "cancelled"),
            ("observed",    "schedulable", "cancelled",    "cancelled"),
            ("observed",    "scheduled",   "cancelled",    "cancelled"),
            ("observed",    "started",     "cancelled",    "cancelled"),
            ("observed",    "cancelled",   "schedulable",  "cancelled"),
            ("observed",    "cancelled",   "unschedulable","cancelled"),
            ("observed",    "cancelled",   "scheduled",    "cancelled"),
            ("observed",    "cancelled",   "started",      "cancelled"),
            ("observed",    "cancelled",   "finished",     "cancelled"),
            ("finished",    "cancelled",   "schedulable",  "cancelled"),
            # any error
            ("observed",    "error",       "schedulable",  "error"),
            ("observed",    "schedulable", "error",        "error"),
            ("observed",    "unschedulable","error",       "error"),
            ("observed",    "scheduled",   "error",        "error"),
            ("observed",    "started",     "error",        "error"),
            ("observed",    "error",       "schedulable",  "error"),
            ("observed",    "error",       "scheduled",    "error"),
            ("observed",    "error",       "started",      "error"),
            ("observed",    "error",       "finished",     "error"),
            # cancelled over error
            ("error",       "error",       "cancelled",    "cancelled")
        ]
        # Do the actual test
        task_state_dict = {}
        for test_item in test_table:
            # Create schedulingblueprint
            schedulingunit_data = SchedulingUnitBlueprint_test_data(name="Task Blueprint With Three Tasks")
            schedulingunit_blueprint = models.SchedulingUnitBlueprint.objects.create(**schedulingunit_data)
            # Create related task and subtasks
            tasks_and_subtasks_dict = self.create_tasks_and_subtasks(schedulingunit_blueprint)

            task_state_dict["observation"], task_state_dict["pipeline"], task_state_dict["ingest"], expected_schedulingunit_status = test_item
            info_msg = "Test with with states observation='%s',pipeline='%s',ingest='%s' should result in schedulingunit_blueprint.status.value '%s'" \
                        % (task_state_dict["observation"], task_state_dict["pipeline"], task_state_dict["ingest"], expected_schedulingunit_status)
            logger.info(info_msg)
            for key in tasks_and_subtasks_dict:
                self.set_task_state(task_state_dict[key], key, tasks_and_subtasks_dict[key]["task"], tasks_and_subtasks_dict[key]["subtask"])

            schedulingunit_blueprint.refresh_from_db()
            # Check result
            self.assertEqual(expected_schedulingunit_status, schedulingunit_blueprint.status.value, info_msg)

    def test_TMSS_1635_bugfix(self):
        '''See https://support.astron.nl/jira/browse/TMSS-1635
        In this test I replicate the bug by creating a schedunit from the 'IM LBA Strategy - 3 Beams', make the pipeline go to error, and check for expected state'''
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="IM LBA Survey - 3 Beams")
        sudraft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template, scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))
        sublueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(sudraft)

        # run the observation
        obs_task = sublueprint.task_blueprints.filter(specifications_template__type__value=models.TaskType.Choices.OBSERVATION.value).first()
        for subtask in obs_task.subtasks.order_by('id').all():
            set_subtask_state_following_allowed_transitions(subtask,  'finished')
        obs_task.refresh_from_db()
        self.assertEqual(models.TaskStatus.Choices.FINISHED.value, obs_task.status.value)
        sublueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.OBSERVED.value, sublueprint.status.value)

        # run all pipelines except 'Pipeline target2' (which should fail, see below)
        pipelines = sublueprint.task_blueprints.filter(specifications_template__type__value=models.TaskType.Choices.PIPELINE.value).all()
        pipelines_wo_tgt2 = pipelines.exclude(name='Pipeline target2').all()
        for pipeline in pipelines_wo_tgt2:
            subtask = pipeline.subtasks.first()
            set_subtask_state_following_allowed_transitions(subtask,  'finished')
            pipeline.refresh_from_db()
            self.assertEqual(models.TaskStatus.Choices.FINISHED.value, pipeline.status.value)
        sublueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.PROCESSING.value, sublueprint.status.value)

        # make 'Pipeline target2' fail
        pipeline_tgt2 = pipelines.get(name='Pipeline target2')
        set_subtask_state_following_allowed_transitions(pipeline_tgt2.subtasks.first(), 'error')
        pipeline_tgt2.refresh_from_db()
        self.assertEqual(models.TaskStatus.Choices.ERROR.value, pipeline_tgt2.status.value)
        sublueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.ERROR.value, sublueprint.status.value)

        # now mark the ERROR 'Pipeline target2' as obsolete
        mark_task_blueprint_as_obsolete(pipeline_tgt2)
        pipeline_tgt2.refresh_from_db()
        self.assertEqual(models.TaskStatus.Choices.ERROR.value, pipeline_tgt2.status.value)
        sublueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.PROCESSED.value, sublueprint.status.value)

    def test_cannot_use_obsolete_template(self):
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.objects.create(name="test obsolete template",
                                                                                          state=models.TemplateState.objects.get(value=models.TemplateState.Choices.DEVELOPMENT.value),
                                                                                          scheduling_unit_template=models.SchedulingUnitTemplate.get_latest("scheduling unit"),
                                                                                          template={})

        strategy_template.state = models.TemplateState.objects.get(value=models.TemplateState.Choices.OBSOLETE.value)
        strategy_template.save()

        with self.assertRaises(ObsoleteTemplateException) as context:
            create_scheduling_unit_draft_from_observing_strategy_template(strategy_template, scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))
        self.assertTrue('is obsolete' in str(context.exception))


    def test_TMSS_1675_bugfix(self):
        '''See https://support.astron.nl/jira/browse/TMSS-1675
        In this test I replicate the bug by creating a schedunit from the 'IM LBA Survey - 3 Beams', 'run' the observation and pipelines, cancel the ingest task, and check for expected state'''
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="IM LBA Survey - 3 Beams")
        sudraft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template, scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))
        sublueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(sudraft)
        sublueprint.ingest_permission_granted_since = datetime.utcnow()
        sublueprint.save()

        # run the observation
        obs_task = sublueprint.task_blueprints.filter(specifications_template__type__value=models.TaskType.Choices.OBSERVATION.value).first()
        schedule_independent_subtasks_in_task_blueprint(obs_task, start_time=datetime.utcnow()+timedelta(minutes=5))
        for subtask in obs_task.subtasks.order_by('id').all():
            set_subtask_state_following_allowed_transitions(subtask,  'finished')
        obs_task.refresh_from_db()
        self.assertEqual(models.TaskStatus.Choices.FINISHED.value, obs_task.status.value)
        sublueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.OBSERVED.value, sublueprint.status.value)

        # run all pipelines
        pipelines = sublueprint.task_blueprints.filter(specifications_template__type__value=models.TaskType.Choices.PIPELINE.value).all()
        for pipeline in pipelines:
            schedule_subtask(pipeline.subtasks.first())
            set_subtask_state_following_allowed_transitions(pipeline.subtasks.first(),  'finished')
            pipeline.refresh_from_db()
            self.assertEqual(models.TaskStatus.Choices.FINISHED.value, pipeline.status.value)
        sublueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.PROCESSED.value, sublueprint.status.value)

        # from lofar.common.util import waitForInterrupt
        # waitForInterrupt()

        # schedule, run, on-hold, restart and then cancel 'Ingest' task
        ingest = sublueprint.task_blueprints.get(name='Ingest')
        ingest_subtask = ingest.subtasks.first()
        schedule_subtask(ingest_subtask)
        ingest.refresh_from_db()
        self.assertEqual(models.TaskStatus.Choices.SCHEDULED.value, ingest.status.value)
        sublueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.PROCESSED.value, sublueprint.status.value)

        set_subtask_state_following_allowed_transitions(ingest_subtask, 'started')
        sublueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.INGESTING.value, sublueprint.status.value)

        # See https://support.astron.nl/jira/browse/TMSS-2879
        set_subtask_state_following_allowed_transitions(ingest_subtask, 'on hold')
        sublueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.INGESTING.value, sublueprint.status.value)

        # restart
        set_subtask_state_following_allowed_transitions(ingest_subtask, 'started')
        sublueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.INGESTING.value, sublueprint.status.value)

        cancel_task_blueprint(ingest)
        ingest.refresh_from_db()
        self.assertEqual(models.TaskStatus.Choices.CANCELLED.value, ingest.status.value)
        sublueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.CANCELLED.value, sublueprint.status.value)

        # now mark the CANCELLED 'Ingest' as obsolete
        mark_task_blueprint_as_obsolete(ingest)
        ingest.refresh_from_db()
        self.assertEqual(models.TaskStatus.Choices.CANCELLED.value, ingest.status.value)
        sublueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.PROCESSED.value, sublueprint.status.value)


    def test_timestamps_when_SU_is_cancelled_after_obs_finished(self):
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="IM LBA Survey - 3 Beams")
        sudraft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template, scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))
        sublueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(sudraft)
        sublueprint.ingest_permission_granted_since = datetime.utcnow()
        sublueprint.save()

        # run the observation
        obs_task = sublueprint.task_blueprints.filter(specifications_template__type__value=models.TaskType.Choices.OBSERVATION.value).first()
        schedule_independent_subtasks_in_task_blueprint(obs_task, start_time=datetime.utcnow()+timedelta(minutes=5))
        for subtask in obs_task.subtasks.order_by('id').all():
            set_subtask_state_following_allowed_transitions(subtask,  'finished')
        obs_task.refresh_from_db()
        self.assertEqual(models.TaskStatus.Choices.FINISHED.value, obs_task.status.value)
        sublueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.OBSERVED.value, sublueprint.status.value)

        # cancel the SU
        time_before_cancel = datetime.utcnow()
        on_sky_start_time_before_cancel = sublueprint.on_sky_start_time
        on_sky_stop_time_before_cancel = sublueprint.on_sky_stop_time
        process_start_time_before_cancel = sublueprint.process_start_time
        for task_blueprint in sublueprint.task_blueprints.all():
            for subtask in task_blueprint.subtasks.all():
                try:
                    cancel_subtask(subtask)  # fails for some subtasks, so we cannot use cancel_scheduling_unit_blueprint # todo: check if cancel_scheduling_unit_blueprint/cancel_task_blueprint should try to kill the rest before raising
                except Exception:
                    continue
        sublueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.CANCELLED.value, sublueprint.status.value)
        time_after_cancel = datetime.utcnow()

        # assert that the timestamps are set correctly
        # process start time should be unaffected by cancel action (reflects when obs had started)
        # process stop time should reflect when the SU was cancelled
        # on sky start/stop time should be unaffected by cancel action (because the observation had already finished)
        self.assertEqual(sublueprint.process_start_time, process_start_time_before_cancel)
        self.assertGreaterEqual(sublueprint.process_stop_time, time_before_cancel)
        self.assertLessEqual(sublueprint.process_stop_time, time_after_cancel)
        self.assertEqual(sublueprint.on_sky_start_time, on_sky_start_time_before_cancel)
        self.assertEqual(sublueprint.on_sky_stop_time, on_sky_stop_time_before_cancel)


    def test_timestamps_when_SU_is_cancelled_while_obs_running(self):
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="IM LBA Survey - 3 Beams")
        sudraft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template, scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))
        sublueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(sudraft)
        sublueprint.ingest_permission_granted_since = datetime.utcnow()
        sublueprint.save()

        # run the observation
        obs_task = sublueprint.task_blueprints.filter(specifications_template__type__value=models.TaskType.Choices.OBSERVATION.value).first()
        schedule_independent_subtasks_in_task_blueprint(obs_task, start_time=datetime.utcnow()+timedelta(minutes=5))
        for subtask in obs_task.subtasks.order_by('id').all():
            set_subtask_state_following_allowed_transitions(subtask,  'started')
        obs_task.refresh_from_db()
        self.assertEqual(models.TaskStatus.Choices.STARTED.value, obs_task.status.value)
        sublueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.OBSERVING.value, sublueprint.status.value)

        # cancel the SU
        time_before_cancel = datetime.utcnow()
        on_sky_start_time_before_cancel = sublueprint.on_sky_start_time
        on_sky_stop_time_before_cancel = sublueprint.on_sky_stop_time
        process_start_time_before_cancel = sublueprint.process_start_time
        for task_blueprint in sublueprint.task_blueprints.all():
            for subtask in task_blueprint.subtasks.all():
                try:
                    cancel_subtask(subtask)  # fails for some subtasks, so we cannot use cancel_scheduling_unit_blueprint # todo: check if cancel_scheduling_unit_blueprint/cancel_task_blueprint should try to kill the rest before raising
                except Exception:
                    continue
        sublueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.CANCELLED.value, sublueprint.status.value)
        time_after_cancel = datetime.utcnow()

        # assert that the timestamps are set correctly
        # process start time should be unaffected by cancel action  (reflects when obs had started)
        # process stop time should reflect when the SU was cancelled
        # on sky start time should be unaffected by cancel action
        # on sky stop time should reflect when the SU was cancelled (because the observation was interrupted)
        self.assertEqual(sublueprint.process_start_time, process_start_time_before_cancel)
        self.assertGreaterEqual(sublueprint.process_stop_time, time_before_cancel)
        self.assertLessEqual(sublueprint.process_stop_time, time_after_cancel)
        self.assertEqual(sublueprint.on_sky_start_time, on_sky_start_time_before_cancel)
        self.assertNotEqual(sublueprint.on_sky_stop_time, on_sky_stop_time_before_cancel)
        self.assertGreaterEqual(sublueprint.on_sky_stop_time, time_before_cancel)
        self.assertLessEqual(sublueprint.on_sky_stop_time, time_after_cancel)


    def test_timestamps_when_SU_is_cancelled_before_obs_started(self):
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="IM LBA Survey - 3 Beams")
        sudraft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template, scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))
        sublueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(sudraft)
        sublueprint.ingest_permission_granted_since = datetime.utcnow()
        sublueprint.save()

        # cancel the SU
        time_before_cancel = datetime.utcnow()
        on_sky_start_time_before_cancel = sublueprint.on_sky_start_time
        on_sky_stop_time_before_cancel = sublueprint.on_sky_stop_time
        for task_blueprint in sublueprint.task_blueprints.all():
            for subtask in task_blueprint.subtasks.all():
                try:
                    cancel_subtask(subtask) # fails for some subtasks, so we cannot use cancel_scheduling_unit_blueprint # todo: check if cancel_scheduling_unit_blueprint/cancel_task_blueprint should try to kill the rest before raising
                except Exception:
                    continue
        sublueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.CANCELLED.value, sublueprint.status.value)
        time_after_cancel = datetime.utcnow()

        # assert that the timestamps are set correctly
        # process start/stop time should reflect when the SU was cancelled
        # on sky start/stop time should be unaffected by cancel action (observation had not started yet)
        self.assertGreaterEqual(sublueprint.process_start_time, time_before_cancel)
        self.assertLessEqual(sublueprint.process_start_time, time_after_cancel)
        self.assertGreaterEqual(sublueprint.process_stop_time, time_before_cancel)
        self.assertLessEqual(sublueprint.process_stop_time, time_after_cancel)
        self.assertEqual(sublueprint.on_sky_start_time, on_sky_start_time_before_cancel)
        self.assertEqual(sublueprint.on_sky_stop_time, on_sky_stop_time_before_cancel)


    def test_can_access_specs_of_scheduling_unit_draft_with_obsolete_template_but_not_create_blueprint(self):   # bugfix TMSS-1960
        # create draft and template
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.objects.create(name="test template",
                                                                                          state=models.TemplateState.objects.get(value=models.TemplateState.Choices.ACTIVE.value),
                                                                                          scheduling_unit_template=models.SchedulingUnitTemplate.get_latest("scheduling unit"),
                                                                                          template={})
        draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template, scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))

        # mark template obsolete, should not be possible
        with self.assertRaises(ObsoleteValidationException) as context:
            strategy_template.state = models.TemplateState.objects.get(value=models.TemplateState.Choices.OBSOLETE.value)
            strategy_template.save()



    def test_create_scheduling_unit_draft_from_strategy_template_only_contains_requested_station_groups(self):   # bugfix TMSS-2156
        # create draft from
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="IM HBA - 1 Beam")

        # override station_groups
        station_groups = [{"stations": ["CS002","CS003","CS004","CS005"], "max_nr_missing": 1}]
        override_doc = {"tasks": {"Target Observation": {"specifications_doc": {"station_configuration": {"station_groups": station_groups}}}}}
        draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template, scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()), specifications_doc_overrides=override_doc)

        # the template contains international stations and a custom group by default, both including DE605.
        self.assertNotIn("DE605", str(draft.specifications_doc))

        # assert the draft specs contain exactly the specified station groups:
        self.assertEqual(draft.specifications_doc["tasks"]["Target Observation"]["specifications_doc"]["station_configuration"]["station_groups"], station_groups)

    def test_remove_lofar2_station_while_create_scheduling_unit_draft_from_scheduling_unit_blueprint(self):   # TMSS-2637
        # create scheduling unit with several LOFAR1 stations
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="IM HBA - 1 Beam")
        station_groups = [{"stations": ["CS002","CS003","CS004","CS005","DE605"], "max_nr_missing": 1}]
        override_doc = {"tasks": {"Target Observation": {"specifications_doc": {"station_configuration": {"station_groups": station_groups}}}}}
        draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template, scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()), specifications_doc_overrides=override_doc)
        blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(draft)

        # check that the blueprint contains a particular station and all its tasks validate
        self.assertIn("DE605", str(blueprint.specifications_doc))
        for task in blueprint.task_blueprints.all():
            task.validate_specifications_doc()

        lofar2_stations = get_lofar2_stations()
        self.assertNotIn('DE605', lofar2_stations)

        # mark that station LOFAR2
        station_schema_template = models.CommonSchemaTemplate.get_latest(name="stations")
        groups = station_schema_template.schema['definitions']['station_group']['anyOf']
        for group in groups:
            if group['title'].upper() == 'LOFAR2':
                group['properties']['stations']['enum'][0].append('DE605')
        station_schema_template.save()
        get_lofar2_stations.cache_clear()
        lofar2_stations = get_lofar2_stations()
        self.assertIn('DE605', lofar2_stations)

        # make sure that the previously blueprinted tasks don't validate any more
        with self.assertRaises(RuleValidationException):
            for task in blueprint.task_blueprints.all():
                task.validate_specifications_doc()

        # start a subtask so that the unit is advance enough to skip rule-based validation on specs read access
        set_subtask_state_following_allowed_transitions(blueprint.subtasks.first(), 'started')
        blueprint.refresh_from_db()
        self.assertEqual(blueprint.status.value, models.SchedulingUnitStatus.Choices.OBSERVING.value)

        # try to create a draft copy, which should fail due to the invalid specs
        with self.assertRaises(ValidationException):
            create_scheduling_unit_draft_from_scheduling_unit_blueprint(blueprint)

        # try again, but this time remove lofar2 stations when creating the draft so that it can be blueprinted
        draft_copy = create_scheduling_unit_draft_from_scheduling_unit_blueprint(blueprint, remove_lofar2_stations=True)
        blueprint_from_copy = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(draft_copy)

        # check that the new blueprint does not contain the LOFAR2 station and all its tasks validate
        self.assertNotIn("DE605", str(blueprint_from_copy.specifications_doc))
        for task in blueprint_from_copy.task_blueprints.all():
            task.validate_specifications_doc()

        # revert the list of LOFAR2 stations, so other tests still work
        for group in groups:
            if group['title'].upper() == 'LOFAR2':
                group['properties']['stations']['enum'][0].remove('DE605')
        station_schema_template.save()
        get_lofar2_stations.cache_clear()


class TestFlatStations(unittest.TestCase):
    """
    Test the property of 'flat_stations', retrieve a list of all station as a flat list
    """
    def create_UC1_observation_scheduling_unit(self, name, scheduling_set):

        uc1_strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="IM HBA LoTSS - 2 Beams")
        scheduling_unit_spec = add_defaults_to_json_object_for_schema(uc1_strategy_template.template,
                                                                      uc1_strategy_template.scheduling_unit_template.schema)
        # limit target obs duration for demo data
        scheduling_unit_spec['tasks']['Calibrator Observation 1']['specifications_doc']['duration'] = 2 * 60
        scheduling_unit_spec['tasks']['Target Observation']['specifications_doc']['duration'] = 2 * 3600
        scheduling_unit_spec['tasks']['Calibrator Observation 2']['specifications_doc']['duration'] = 2 * 60

        constraints_template = models.SchedulingConstraintsTemplate.get_version_or_latest(name=scheduling_unit_spec.get('scheduling_constraints_template',{}).get('name', 'constraints'),
                                                                                          version=scheduling_unit_spec.get('scheduling_constraints_template', {}).get('version'))
        constraints = constraints_template.get_default_json_document_for_schema()


        # add the scheduling_unit_doc to a new SchedulingUnitDraft instance, and were ready to use it!
        draft = models.SchedulingUnitDraft.objects.create(name=name,
                                                         scheduling_set=scheduling_set,
                                                         specifications_template=uc1_strategy_template.scheduling_unit_template,
                                                         observation_strategy_template=uc1_strategy_template,
                                                         scheduling_constraints_doc=constraints,
                                                         scheduling_constraints_template=constraints_template)
        return update_task_graph_from_specifications_doc(draft, scheduling_unit_spec)


    def modify_stations_in_station_group(self, station_group_idx, lst_stations):
        """
        Modify for the scheduling_unit_blueprint created add setup, the list of stations for given group idx
        """
        target_obs_task = self.scheduling_unit_draft.task_drafts.get(name='Target Observation')
        station_groups = target_obs_task.specifications_doc['station_configuration']['station_groups']
        station_groups[station_group_idx]["stations"] = lst_stations
        station_groups[station_group_idx]["max_nr_missing"] = len(lst_stations)-1
        target_obs_task.save()

    def setUp(self) -> None:
        # scheduling unit
        my_scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data())
        self.scheduling_unit_draft = self.create_UC1_observation_scheduling_unit("UC1 scheduling unit for testing",  my_scheduling_set)
        self.scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(self.scheduling_unit_draft)

    def test_with_different_stations(self):
        """
        Test with different station list and station groups
        """
        list_expected_stations = ["CS002", "CS003", "CS004", "CS005", "CS006", "CS007", "CS011", "CS013", "CS017", "CS021", "CS024", "CS026", "CS028", "CS030", "CS031", "CS032", "CS101", "CS103", "CS201", "CS301", "CS302", "CS401", "CS501", "RS106", "RS205", "RS208", "RS210", "RS305", "RS306", "RS307", "RS310", "RS406", "RS407", "RS409", "RS503", "RS508", "RS509", "DE601", "DE602", "DE603", "DE604", "DE605", "DE609", "FR606", "SE607", "UK608", "PL610", "PL611", "PL612", "IE613", "LV614"]
        self.assertCountEqual(list_expected_stations, self.scheduling_unit_blueprint.main_observation_specified_stations)

        # Clear all stations should be forbidden, as we need at least one station
        nbr_station_groups = len(self.scheduling_unit_draft.specifications_doc['tasks']['Target Observation']['specifications_doc']['station_configuration']['station_groups'])
        for idx in range(nbr_station_groups):
            with self.assertRaises(ValidationException):
                self.modify_stations_in_station_group(idx, [])

        # Set two stations for all station_groups, check main_observation_specified_stations contains two stations
        for idx in range(nbr_station_groups):
            self.modify_stations_in_station_group(idx, ['CS002', 'CS003'])
        scheduling_unit_blueprint = create_scheduling_unit_blueprint_from_scheduling_unit_draft(self.scheduling_unit_draft)
        scheduling_unit_blueprint = update_task_blueprint_graph_from_draft(scheduling_unit_blueprint)
        self.assertCountEqual(['CS002', 'CS003'], scheduling_unit_blueprint.main_observation_specified_stations)

        # Set two stations for all station_groups, check main_observation_specified_stations contains all stations
        all_stations = ["CS002", "CS003", "CS004", "CS005", "CS006", "CS007", "CS011", "CS013", "CS017", "CS021", "CS024", "CS026", "CS028", "CS030", "CS031", "CS032", "CS101", "CS103", "CS201", "CS301", "CS302", "CS401", "CS501", "RS106", "RS205", "RS208", "RS210", "RS305", "RS306", "RS307", "RS310", "RS406", "RS407", "RS409", "RS503", "RS508", "RS509", "DE601", "DE602", "DE603", "DE604", "DE605", "DE609", "FR606", "SE607", "UK608", "PL610", "PL611", "PL612", "IE613", "LV614"]
        for idx in range(nbr_station_groups):
            self.modify_stations_in_station_group(idx, all_stations)
        scheduling_unit_blueprint = create_scheduling_unit_blueprint_from_scheduling_unit_draft(self.scheduling_unit_draft)
        scheduling_unit_blueprint = update_task_blueprint_graph_from_draft(scheduling_unit_blueprint)
        self.assertCountEqual(all_stations, scheduling_unit_blueprint.main_observation_specified_stations)


class SchedulingUnitBlueprintIndirectModificationsTestCase(unittest.TestCase):
    """
    Various tests for TMSS-850: SchedulingUnitBlueprints should be indirectly modifiable for certain cases, which are tested here.
    """
    @classmethod
    def setUpClass(cls) -> None:
        # enable large diff views in unittesting
        cls.maxDiff = None

    def setUp(self) -> None:
        tmss_test_env.delete_scheduling_unit_drafts_cascade()
        tmss_test_env.delete_reservations_cascade()

    def test_modify_draft_graph(self):
        '''core test to check if the task draft graph is correctly created/updated/removed'''
        with tmss_test_env.create_tmss_client() as client:
            # start with a default scheduling_unit_template specifications_doc with no tasks
            scheduling_unit_template = client.get_schedulingunit_template('scheduling unit', 8)
            scheduling_constraints_template = client.get_scheduling_constraints_template('constraints', 7)
            base_specifications_doc = client.get_schedulingunit_template_default_specification(scheduling_unit_template['name'], scheduling_unit_template['version'])
            if 'scheduling_constraints_template' in base_specifications_doc:
                base_specifications_doc['scheduling_constraints_doc'] = without_schema_property(client.get_url_as_json_object(scheduling_constraints_template['url']+'/default'))
            scheduling_unit_draft_test_data = rest_data_creator.SchedulingUnitDraft(specifications_doc=base_specifications_doc,
                                                                                    template_url=scheduling_unit_template['url'],
                                                                                    scheduling_constraints_doc=base_specifications_doc.get('scheduling_constraints_doc',{}),
                                                                                    scheduling_constraints_template_url=scheduling_constraints_template['url'])
            scheduling_unit_draft = client.post_to_path_and_get_result_as_json_object('scheduling_unit_draft',
                                                                                      json_data=scheduling_unit_draft_test_data)
            scheduling_unit_draft_id = scheduling_unit_draft['id']

            # check that there are no tasks
            self.assertEqual([], scheduling_unit_draft['task_drafts'])

            # is the exported/generated specifications_doc equal to the initial specification?
            self.assertEqual(base_specifications_doc, client.get_schedulingunit_draft_specifications_doc(scheduling_unit_draft_id))

            obs_task_template = client.get_task_template('target observation') # latest version
            ingest_task_template = client.get_task_template('ingest') # latest version
            cleanup_task_template = client.get_task_template('cleanup') # latest version
            selection_template = [template for template in client.get_path_as_json_object('task_relation_selection_template') if template['name']=='all'][0]

            # import a specifications_doc with a single task
            single_task_specifications_doc = copy.deepcopy(base_specifications_doc)
            single_task_specifications_doc['tasks'] = {'Task_A': {'specifications_doc': without_schema_property(client.get_task_template_default_specification(obs_task_template['name'], obs_task_template['version'])),
                                                                  'specifications_template': {"name": obs_task_template['name'], "version": obs_task_template['version']},
                                                                  'short_description': "",
                                                                  'description': "my task A"} }

            scheduling_unit_draft = client.update_task_graph_from_specifications_doc(scheduling_unit_draft['id'],
                                                                                        specifications_doc=single_task_specifications_doc)

            self.assertEqual(1, len(scheduling_unit_draft['task_drafts']))

            # is the exported/generated specifications_doc equal to the imported specification?
            self.assertEqual(single_task_specifications_doc, client.get_schedulingunit_draft_specifications_doc(scheduling_unit_draft_id))

            # modify a setting in the specification, check the task is updated
            org_duration = single_task_specifications_doc['tasks']['Task_A']['specifications_doc']['duration']
            new_duration = 2 * org_duration
            single_task_specifications_doc['tasks']['Task_A']['specifications_doc']['duration'] = new_duration

            scheduling_unit_draft = client.update_task_graph_from_specifications_doc(scheduling_unit_draft['id'],
                                                                                        specifications_doc=single_task_specifications_doc)
            updated_exported_single_task_specifications_doc = client.get_schedulingunit_draft_specifications_doc(scheduling_unit_draft_id)
            self.assertEqual(single_task_specifications_doc, updated_exported_single_task_specifications_doc)
            self.assertEqual(new_duration, updated_exported_single_task_specifications_doc['tasks']['Task_A']['specifications_doc']['duration'])


            # back to an empty graph with no tasks, and check results
            scheduling_unit_draft = client.update_task_graph_from_specifications_doc(scheduling_unit_draft['id'], specifications_doc=base_specifications_doc)
            self.assertEqual(0, len(scheduling_unit_draft['task_drafts']))
            self.assertEqual(base_specifications_doc, client.get_schedulingunit_draft_specifications_doc(scheduling_unit_draft_id))


            # import a specifications_doc with a dual task and relation
            dual_task_specifications_doc = copy.deepcopy(base_specifications_doc)
            dual_task_specifications_doc['tasks'] = {'Task_A': {'specifications_doc': without_schema_property(client.get_task_template_default_specification(obs_task_template['name'], obs_task_template['version'])),
                                                                'specifications_template': {"name": obs_task_template['name'], "version": obs_task_template['version']},
                                                                'short_description': "",
                                                                'description': "my task A"},
                                                     'Task_B': {'specifications_doc': without_schema_property(client.get_task_template_default_specification(ingest_task_template['name'], ingest_task_template['version'])),
                                                                'specifications_template': {"name": ingest_task_template['name'], "version": ingest_task_template['version']},
                                                                'short_description': "",
                                                                'description': "my task B"} }
            dual_task_specifications_doc['task_relations'] = [{ "input": { "role": "any",
                                                                           "datatype": "visibilities",
                                                                           "dataformat": "MeasurementSet" },
                                                                "output": { "role": "correlator",
                                                                            "datatype": "visibilities",
                                                                            "dataformat": "MeasurementSet" },
                                                                "producer": "Task_A",
                                                                "consumer": "Task_B",
                                                                "selection_doc": without_schema_property(client.get_url_as_json_object(selection_template['url']+"/default")),
                                                                "selection_template": {"name": selection_template['name'], "version": selection_template['version'] } }]
            scheduling_unit_draft = client.update_task_graph_from_specifications_doc(scheduling_unit_draft['id'],
                                                                                        specifications_doc=dual_task_specifications_doc)
            self.assertEqual(2, len(scheduling_unit_draft['task_drafts']))
            self.assertEqual(dual_task_specifications_doc, client.get_schedulingunit_draft_specifications_doc(scheduling_unit_draft_id))

            # check updates on settings
            updated_dual_task_specifications_doc = copy.deepcopy(dual_task_specifications_doc)
            updated_dual_task_specifications_doc['tasks']['Task_A']['specifications_doc']['duration'] *= 2

            scheduling_unit_draft = client.update_task_graph_from_specifications_doc(scheduling_unit_draft['id'],
                                                                                        specifications_doc=updated_dual_task_specifications_doc)
            self.assertEqual(2, len(scheduling_unit_draft['task_drafts']))
            self.assertEqual(updated_dual_task_specifications_doc, client.get_schedulingunit_draft_specifications_doc(scheduling_unit_draft_id))

            # check updates on task graph
            updated_dual_task_specifications_doc['tasks']['Task_C'] = {'specifications_doc': without_schema_property(client.get_task_template_default_specification(cleanup_task_template['name'], cleanup_task_template['version'])),
                                                                       'specifications_template': {"name": cleanup_task_template['name'], "version": cleanup_task_template['version']},
                                                                       'short_description': "",
                                                                       'description': "my task C"}
            updated_dual_task_specifications_doc['task_relations'].append({ "input": { "role": "any",
                                                                                       "datatype": "visibilities",
                                                                                       "dataformat": "MeasurementSet" },
                                                                            "output": { "role": "correlator",
                                                                                        "datatype": "visibilities",
                                                                                        "dataformat": "MeasurementSet" },
                                                                            "producer": "Task_A",
                                                                            "consumer": "Task_C",
                                                                            "selection_doc": without_schema_property(client.get_url_as_json_object(selection_template['url']+"/default")),
                                                                            "selection_template": { "name": selection_template['name'], "version": selection_template['version']} })

            scheduling_unit_draft = client.update_task_graph_from_specifications_doc(scheduling_unit_draft['id'],
                                                                                        specifications_doc=updated_dual_task_specifications_doc)
            self.assertEqual(3, len(scheduling_unit_draft['task_drafts']))
            self.assertEqual(updated_dual_task_specifications_doc, client.get_schedulingunit_draft_specifications_doc(scheduling_unit_draft_id))


            # back to an empty graph with no tasks, and check results
            scheduling_unit_draft = client.update_task_graph_from_specifications_doc(scheduling_unit_draft['id'], specifications_doc=base_specifications_doc)
            self.assertEqual(0, len(scheduling_unit_draft['task_drafts']))
            self.assertEqual(base_specifications_doc, client.get_schedulingunit_draft_specifications_doc(scheduling_unit_draft_id))


    def test_copy_blueprint_exported_specifications_doc_to_draft(self):
        '''core test to check if we can copy an exported specifications_doc of the task blueprint graph is correctly copied to the draft'''
        with tmss_test_env.create_tmss_client() as client:

            # setup: create a scheduling unit draft from a known observation strategy template
            project = rest_data_creator.post_data_and_get_response_as_json_object(rest_data_creator.Project(auto_ingest=True), '/project/')
            scheduling_set = rest_data_creator.post_data_and_get_response_as_json_object(rest_data_creator.SchedulingSet(project_url=project['url']), '/scheduling_set/')
            strategy_templates = client.get_path_as_json_object('scheduling_unit_observing_strategy_template')
            strategy_template = next(ost for ost in strategy_templates if ost['name'] == 'IM HBA - 1 Beam')
            self.assertIsNotNone(strategy_template)

            # create a scheduling_unit_draft from the template...
            scheduling_unit_draft = client.create_scheduling_unit_draft_from_strategy_template(strategy_template['id'], scheduling_set['id'])
            # ...and create a scheduling_unit_blueprint out of that
            scheduling_unit_blueprint = client.create_scheduling_unit_blueprint_and_tasks_and_subtasks_tree(scheduling_unit_draft['id'])

            # check that we have some tasks
            self.assertTrue(len(scheduling_unit_draft['task_drafts']) > 0)
            self.assertTrue(len(scheduling_unit_blueprint['task_blueprints']) > 0)
            self.assertEqual(len(scheduling_unit_draft['task_drafts']), len(scheduling_unit_blueprint['task_blueprints']))

            # and check that each task_blueprint has an existing reference to its draft
            for task_blueprint in scheduling_unit_blueprint['task_blueprints']:
                self.assertTrue(task_blueprint['draft'] in scheduling_unit_draft['task_drafts'])

            # the exported specifications_doc should be equal for both draft and blueprint
            original_draft_specifications_doc = client.get_schedulingunit_draft_specifications_doc(scheduling_unit_draft['id'])
            original_blueprint_specifications_doc = client.get_schedulingunit_blueprint_specifications_doc(scheduling_unit_blueprint['id'])
            self.assertEqual(original_draft_specifications_doc, original_blueprint_specifications_doc)


            # ---------------------------------------------------------------
            # try to use the specifications_doc from the draft for some 'user'-edits
            # this will fail, because we are not allowed to remove the drafts which are already referenced by blueprints
            # the returned doc should be equal to the original
            draft_specifications_doc = client.get_schedulingunit_draft_specifications_doc(scheduling_unit_draft['id'])
            # our user wants to wipe the draft
            edited_draft_specifications_doc = copy.deepcopy(draft_specifications_doc)
            edited_draft_specifications_doc['tasks'] = {}
            edited_draft_specifications_doc['task_relations'] = []
            edited_draft_specifications_doc['task_scheduling_relations'] = []
            scheduling_unit_draft = client.update_task_graph_from_specifications_doc(scheduling_unit_draft['id'], edited_draft_specifications_doc)
            after_edit_draft_specifications_doc = client.get_schedulingunit_draft_specifications_doc(scheduling_unit_draft['id'])
            self.assertEqual(draft_specifications_doc, after_edit_draft_specifications_doc)


            # but we can create a copy (which has no references anymore to blueprints), and edit this copy
            scheduling_unit_draft = client.copy_scheduling_unit_draft(scheduling_unit_draft['id'])
            edited_draft_specifications_doc = client.get_schedulingunit_draft_specifications_doc(scheduling_unit_draft['id'])
            edited_draft_specifications_doc['tasks'] = {}
            edited_draft_specifications_doc['task_relations'] = []
            edited_draft_specifications_doc['task_scheduling_relations'] = []
            scheduling_unit_draft = client.update_task_graph_from_specifications_doc(scheduling_unit_draft['id'],
                                                                                     edited_draft_specifications_doc)

            # check: the draft(copy) should have no tasks any more
            self.assertEqual(0, len(scheduling_unit_draft['task_drafts']))
            self.assertEqual(edited_draft_specifications_doc, client.get_schedulingunit_draft_specifications_doc(scheduling_unit_draft['id']))
            self.assertNotEqual(original_draft_specifications_doc, client.get_schedulingunit_draft_specifications_doc(scheduling_unit_draft['id']))


            # -----------------------------------------------------------------------------
            # use the specifications_doc from the draft to make later 'user'-edits possible.
            client.copy_scheduling_unit_blueprint_specifications_doc_back_into_draft(scheduling_unit_blueprint['id'])

            # note: we do not do any actual edits on the draft. We just check that we can copy it and reuse it.
            # Nothing should be changed afterwards, so apply the copied draft graph to the current blueprint...
            client.update_task_blueprints_and_subtasks_graph_from_draft(scheduling_unit_blueprint['id'])

            # ...fetch updates, and check that indeed nothing changed (compared to the original draft/blueprint)
            scheduling_unit_draft = client.get_schedulingunit_draft(scheduling_unit_blueprint['draft_id'], extended=False)
            scheduling_unit_blueprint = client.get_schedulingunit_blueprint(scheduling_unit_blueprint['id'], extended=True)

            # check that we have some tasks again
            self.assertTrue(len(scheduling_unit_draft['task_drafts']) > 0)
            self.assertTrue(len(scheduling_unit_blueprint['task_blueprints']) > 0)
            self.assertEqual(len(scheduling_unit_draft['task_drafts']), len(scheduling_unit_blueprint['task_blueprints']))

            # and check that each task_blueprint has an existing reference to its draft
            for task_blueprint in scheduling_unit_blueprint['task_blueprints']:
                self.assertTrue(task_blueprint['draft'] in scheduling_unit_draft['task_drafts'])

            # the exported specifications_doc should be equal for both draft and blueprint
            self.assertEqual(client.get_schedulingunit_draft_specifications_doc(scheduling_unit_draft['id']),
                             client.get_schedulingunit_blueprint_specifications_doc(scheduling_unit_blueprint['id']))
            # and it should be equal again to the original
            self.assertEqual(original_draft_specifications_doc, client.get_schedulingunit_draft_specifications_doc(scheduling_unit_draft['id']))
            self.assertEqual(original_blueprint_specifications_doc, client.get_schedulingunit_blueprint_specifications_doc(scheduling_unit_blueprint['id']))

    @unittest.skip("ToDo: re-enable this test and find out why it fails on the build server and not on dev-machines")
    def test_rerun_copy_of_failed_pipeline(self):
        '''Test concrete use case where we simulate that an observation ran fine, but a pipeline failed. Then create an exact copy of the pipeline and let that run.'''
        with tmss_test_env.create_tmss_client() as client:

            # setup: create a scheduling unit draft with a simple sequence observation->pipeline->ingest->cleanup
            project = rest_data_creator.post_data_and_get_response_as_json_object(rest_data_creator.Project(auto_ingest=True), '/project/')
            scheduling_set = rest_data_creator.post_data_and_get_response_as_json_object(rest_data_creator.SchedulingSet(project_url=project['url']), '/scheduling_set/')

            # create a target observation spec with many default and some special settings
            obs_specifications_doc = client.get_task_template_default_specification("target observation")
            obs_specifications_doc['QA']['file_conversion']['enabled'] = False
            obs_specifications_doc['QA']['plots']['enabled'] = False
            obs_specifications_doc['station_configuration']['SAPs'][0]['subbands'] = [0, 1, 2, 3]

            specifications_doc = {
                "tasks": { "observation": { "specifications_doc": obs_specifications_doc,
                                            "specifications_template": {"name": "target observation", "version": 2 }},
                           "pipeline": { "description": "my description",
                                         "specifications_doc": {},
                                         "specifications_template": {"name": "preprocessing pipeline", "version": 2 }},
                },
                "task_relations": [ { "input": { "role": "any",
                                                 "datatype": "visibilities",
                                                 "dataformat": "MeasurementSet" },
                                      "output": { "role": "correlator",
                                                  "datatype": "visibilities",
                                                  "dataformat": "MeasurementSet" },
                                      "consumer": "pipeline",
                                      "producer": "observation",
                                      "selection_doc": {},
                                      "selection_template": {"name": "all", "version": 2 } }
                ],
                "task_scheduling_relations": []
            }
            su_template = models.SchedulingUnitTemplate.get_latest("scheduling unit")
            specifications_doc['$schema'] = su_template.schema['$id']

            scheduling_unit_draft = client.create_scheduling_unit_draft_from_specifications_doc("test_rerun_copy_of_failed_pipeline", "no description", scheduling_set['id'], specifications_doc)

            # ...and create a scheduling_unit_blueprint out of that
            scheduling_unit_blueprint = client.create_scheduling_unit_blueprint_and_tasks_and_subtasks_tree(scheduling_unit_draft['id'])
            # fetch all subtasks
            subtasks = client.get_url_as_json_object(scheduling_unit_blueprint['url']+'/subtasks')
            self.assertEqual(2, len(subtasks))
            subtask_obs = next(s for s in subtasks if s['subtask_type']=='observation')
            subtask_pipe = next(s for s in subtasks if s['subtask_type']=='pipeline')

            # schedule the obs to create output dataproducts
            client.schedule_subtask(subtask_obs['id'], datetime.utcnow()+timedelta(minutes=1))
            obs_output_dataproducts = client.get_subtask_output_dataproducts(subtask_obs['id'])
            self.assertTrue(len(obs_output_dataproducts) > 0)

            # and simulate that it finished
            set_subtask_state_following_allowed_transitions(subtask_obs['id'], 'finished')

            # schedule the pipeline to link it to the observations output dataproducts
            client.schedule_subtask(subtask_pipe['id'], datetime.utcnow())
            pipe_input_dataproducts = client.get_subtask_input_dataproducts(subtask_pipe['id'])
            self.assertTrue(len(pipe_input_dataproducts) > 0)

            # fetch obs_output_dataproducts again, as they "changed" now that the pipeline was scheduled: the contain a reference to a consumer (the subtaskinput of the pipeline)
            obs_output_dataproducts = client.get_subtask_output_dataproducts(subtask_obs['id'])
            self.assertTrue(len(obs_output_dataproducts) > 0)
            self.assertEqual(obs_output_dataproducts, pipe_input_dataproducts)

            # and simulate that it runs and errors
            set_subtask_state_following_allowed_transitions(subtask_pipe['id'], 'started')
            set_subtask_state_following_allowed_transitions(subtask_pipe['id'], 'error')

            # the core method to test: create_copies_of_failed_tasks_via_draft
            scheduling_unit_blueprint = client.create_copies_of_failed_tasks_via_draft(scheduling_unit_blueprint['id'])

            # fetch all subtasks from the updated scheduling_unit_blueprint
            # now there should be an extra pipeline
            subtasks = client.get_url_as_json_object(scheduling_unit_blueprint['url']+'/subtasks')
            self.assertEqual(3, len(subtasks))
            subtask_obs = next(s for s in subtasks if s['subtask_type']=='observation')
            subtasks_pipe = [s for s in subtasks if s['subtask_type']=='pipeline']
            self.assertEqual(2, len(subtasks_pipe))
            failed_subtask_pipe = next(s for s in subtasks_pipe if s['state_value']=='error')
            copy_subtask_pipe = next(s for s in subtasks_pipe if s != failed_subtask_pipe)
            self.assertEqual('defined', copy_subtask_pipe['state_value'])

            # schedule the copy-pipeline to link it to the observations output dataproducts
            client.schedule_subtask(copy_subtask_pipe['id'], datetime.utcnow())

            # check the copy's dataproducts
            copy_subtask_pipe_input_dataproducts = client.get_subtask_input_dataproducts(copy_subtask_pipe['id'])
            self.assertTrue(len(copy_subtask_pipe_input_dataproducts) > 0)

            # fetch obs_output_dataproducts and pipe_input_dataproducts again, as they "changed" now that the copy-pipeline was scheduled: the contain a reference to a consumer (the subtaskinput of the pipeline)
            obs_output_dataproducts = client.get_subtask_output_dataproducts(subtask_obs['id'])
            pipe_input_dataproducts = client.get_subtask_input_dataproducts(subtask_pipe['id'])
            self.assertEqual(pipe_input_dataproducts, copy_subtask_pipe_input_dataproducts)
            self.assertEqual(obs_output_dataproducts, copy_subtask_pipe_input_dataproducts)

            # check the exported specifications_doc
            specifications_doc = client.get_schedulingunit_blueprint_specifications_doc(scheduling_unit_blueprint['id'])
            self.assertTrue('observation' in specifications_doc['tasks'])
            self.assertTrue('pipeline' in specifications_doc['tasks'])
            self.assertTrue('pipeline (copy for rerun 01)' in specifications_doc['tasks'])
            self.assertEqual(specifications_doc['tasks']['pipeline']['specifications_doc'],
                             specifications_doc['tasks']['pipeline (copy for rerun 01)']['specifications_doc'])

            # mark the original pipeline as obsolete...
            client.mark_subtask_as_obsolete(failed_subtask_pipe['id'])
            # then it should be gone from the spec, while the copy should still be there
            specifications_doc = client.get_schedulingunit_blueprint_specifications_doc(scheduling_unit_blueprint['id'])
            self.assertFalse('pipeline' in specifications_doc['tasks'])
            self.assertTrue('pipeline (copy for rerun 01)' in specifications_doc['tasks'])

            # Make the rerun-copy pipeline fail as well: simulate that it runs and errors
            set_subtask_state_following_allowed_transitions(copy_subtask_pipe['id'], 'started')
            set_subtask_state_following_allowed_transitions(copy_subtask_pipe['id'], 'error')

            # Test the the core method again to see if it can create a second rerun copy: create_copies_of_failed_tasks_via_draft
            scheduling_unit_blueprint = client.create_copies_of_failed_tasks_via_draft(scheduling_unit_blueprint['id'])

            # check the latest exported specifications_doc
            # should now contain two rerun copies
            specifications_doc = client.get_schedulingunit_blueprint_specifications_doc(scheduling_unit_blueprint['id'])
            self.assertTrue('observation' in specifications_doc['tasks'])
            self.assertFalse('pipeline' in specifications_doc['tasks'])
            self.assertTrue('pipeline (copy for rerun 01)' in specifications_doc['tasks'])
            self.assertTrue('pipeline (copy for rerun 02)' in specifications_doc['tasks'])
            self.assertEqual(specifications_doc['tasks']['pipeline (copy for rerun 01)']['specifications_doc'],
                             specifications_doc['tasks']['pipeline (copy for rerun 02)']['specifications_doc'])

            # mark the first error'ed rerun-copy pipeline as obsolete as well...
            client.mark_subtask_as_obsolete(copy_subtask_pipe['id'])
            # then it should be gone from the spec, while the 2nd copy should still be there
            specifications_doc = client.get_schedulingunit_blueprint_specifications_doc(scheduling_unit_blueprint['id'])
            self.assertFalse('pipeline' in specifications_doc['tasks'])
            self.assertFalse('pipeline (copy for rerun 01)' in specifications_doc['tasks'])
            self.assertTrue('pipeline (copy for rerun 02)' in specifications_doc['tasks'])


    def test_rerun_copy_of_failed_tasks_excludes_observations(self):
        '''Test concrete use case where we simulate that an observation failed. Then create a copy of failed tasks and make sure there is no copy of the observation.'''
        with tmss_test_env.create_tmss_client() as client:

            # setup: create a scheduling unit draft with a simple sequence observation->pipeline->ingest->cleanup
            project = rest_data_creator.post_data_and_get_response_as_json_object(rest_data_creator.Project(auto_ingest=True), '/project/')
            scheduling_set = rest_data_creator.post_data_and_get_response_as_json_object(rest_data_creator.SchedulingSet(project_url=project['url']), '/scheduling_set/')

            # create a target observation spec with many default and some special settings
            obs_specifications_doc = client.get_task_template_default_specification("target observation")
            obs_specifications_doc['QA']['file_conversion']['enabled'] = False
            obs_specifications_doc['QA']['plots']['enabled'] = False
            obs_specifications_doc['station_configuration']['SAPs'][0]['subbands'] = [0, 1, 2, 3]

            specifications_doc = {
                "tasks": { "observation": { "specifications_doc": obs_specifications_doc,
                                            "specifications_template": {"name": "target observation", "version": 2 }},
                           "pipeline": { "description": "my description",
                                         "specifications_doc": {},
                                         "specifications_template": {"name": "preprocessing pipeline", "version": 2 }},
                },
                "task_relations": [ { "input": { "role": "any",
                                                 "datatype": "visibilities",
                                                 "dataformat": "MeasurementSet" },
                                      "output": { "role": "correlator",
                                                  "datatype": "visibilities",
                                                  "dataformat": "MeasurementSet" },
                                      "consumer": "pipeline",
                                      "producer": "observation",
                                      "selection_doc": {},
                                      "selection_template": {"name": "all", "version": 2 } }
                ],
                "task_scheduling_relations": []
            }

            su_template = models.SchedulingUnitTemplate.get_latest("scheduling unit")
            specifications_doc['$schema'] = su_template.schema['$id']

            scheduling_unit_draft = client.create_scheduling_unit_draft_from_specifications_doc("test_rerun_copy_of_failed_tasks_excludes_observation", "no description", scheduling_set['id'], specifications_doc)

            # ...and create a scheduling_unit_blueprint out of that
            scheduling_unit_blueprint = client.create_scheduling_unit_blueprint_and_tasks_and_subtasks_tree(scheduling_unit_draft['id'])
            # fetch all subtasks
            subtasks = client.get_url_as_json_object(scheduling_unit_blueprint['url']+'/subtasks')
            self.assertEqual(2, len(subtasks))
            subtask_obs = next(s for s in subtasks if s['subtask_type']=='observation')
            subtask_pipe = next(s for s in subtasks if s['subtask_type']=='pipeline')

            # schedule the obs to create output dataproducts
            client.schedule_subtask(subtask_obs['id'], datetime.utcnow()+timedelta(minutes=60))
            obs_output_dataproducts = client.get_subtask_output_dataproducts(subtask_obs['id'])
            self.assertTrue(len(obs_output_dataproducts) > 0)

            # and simulate that it runs and errors
            set_subtask_state_following_allowed_transitions(subtask_obs['id'], 'started')
            set_subtask_state_following_allowed_transitions(subtask_obs['id'], 'error')

            # the core method to test: create_copies_of_failed_tasks_via_draft
            scheduling_unit_blueprint = client.create_copies_of_failed_tasks_via_draft(scheduling_unit_blueprint['id'])

            # fetch all subtasks from the updated scheduling_unit_blueprint
            # make sure this did nothing and there is no extra observation (even though it failed):
            subtasks = client.get_url_as_json_object(scheduling_unit_blueprint['url']+'/subtasks')
            self.assertEqual(2, len(subtasks))
            subtasks_obs = [s for s in subtasks if s['subtask_type']=='observation']
            subtasks_pipe = [s for s in subtasks if s['subtask_type']=='pipeline']
            self.assertEqual(1, len(subtasks_obs))
            self.assertEqual(1, len(subtasks_pipe))

    def test_rerun_copy_of_finished_pipeline_with_different_settings(self):
        '''Test concrete use case where we simulate that an observation ran fine, and the pipeline as well, but we want to run a copy of the pipeline with different settings.'''
        with tmss_test_env.create_tmss_client() as client:

            project = rest_data_creator.post_data_and_get_response_as_json_object(rest_data_creator.Project(auto_ingest=True), '/project/')
            scheduling_set = rest_data_creator.post_data_and_get_response_as_json_object(rest_data_creator.SchedulingSet(project_url=project['url']), '/scheduling_set/')

            # create a target observation spec with many default and some special settings
            obs_specifications_doc = client.get_task_template_default_specification("target observation")
            obs_specifications_doc['QA']['file_conversion']['enabled'] = False
            obs_specifications_doc['QA']['plots']['enabled'] = False
            obs_specifications_doc['station_configuration']['SAPs'][0]['subbands'] = [0, 1, 2, 3]

            # use default preprocessing pipeline spec for the first run
            pipe_specifications_doc = client.get_task_template_default_specification("preprocessing pipeline")
            pipe_specifications_doc['average']['time_steps'] = 1

            specifications_doc = {
                "tasks": { "observation": { "specifications_doc": obs_specifications_doc,
                                            "specifications_template": {"name": "target observation", "version": 2 }},
                           "pipeline": { "description": "my description",
                                         "specifications_doc": pipe_specifications_doc,
                                         "specifications_template": {"name": "preprocessing pipeline", "version": 2}},
                },
                "task_relations": [ { "input": { "role": "any",
                                                 "datatype": "visibilities",
                                                 "dataformat": "MeasurementSet" },
                                      "output": { "role": "correlator",
                                                  "datatype": "visibilities",
                                                  "dataformat": "MeasurementSet" },
                                      "consumer": "pipeline",
                                      "producer": "observation",
                                      "selection_doc": {},
                                      "selection_template": {"name": "all", "version": 2 } }
                ],
                "task_scheduling_relations": []
            }

            su_template = models.SchedulingUnitTemplate.get_latest("scheduling unit")
            specifications_doc['$schema'] = su_template.schema['$id']

            # create the scheduling_unit_draft
            scheduling_unit_draft = client.create_scheduling_unit_draft_from_specifications_doc("test_rerun_copy_of_finished_pipeline_with_different_settings", "no description", scheduling_set['id'], specifications_doc)
            self.assertEqual(2, len(scheduling_unit_draft['task_drafts']))

            # create a scheduling_unit_blueprint out of that
            scheduling_unit_blueprint = client.create_scheduling_unit_blueprint_and_tasks_and_subtasks_tree(scheduling_unit_draft['id'])
            self.assertEqual(2, len(scheduling_unit_blueprint['task_blueprints']))

            # get the extended json to see the task details
            scheduling_unit_draft = client.get_schedulingunit_draft(scheduling_unit_draft['id'], extended=True)
            task_draft_pipe = next(t for t in scheduling_unit_draft['task_drafts'] if t['name']=='pipeline')

            # create a copy of the task_draft
            task_draft_pipe_copy = client.copy_task_draft(task_draft_pipe['id'])

            # make some user setting edits, and update/patch the task_draft_pipe_copy
            edited_task_draft_pipe_specifications_doc = task_draft_pipe_copy['specifications_doc']
            edited_task_draft_pipe_specifications_doc['average']['time_steps'] = 2
            client.session.patch(task_draft_pipe_copy['url'], json={'specifications_doc': edited_task_draft_pipe_specifications_doc})

            # check if the edited settings are persistent
            scheduling_unit_draft = client.get_schedulingunit_draft(scheduling_unit_draft['id'], extended=True)
            task_draft_pipe_copy = next(t for t in scheduling_unit_draft['task_drafts'] if 'pipeline' in t['name'] and 'copy' in t['name'].lower())
            self.assertEqual(2, task_draft_pipe_copy['specifications_doc']['average']['time_steps'])

            # and the original should still have the original settings
            task_draft_pipe = next(t for t in scheduling_unit_draft['task_drafts'] if t['name']=='pipeline')
            self.assertEqual(1, task_draft_pipe['specifications_doc']['average']['time_steps'])

            # update the scheduling_unit_blueprint to reflect the new draft task graph
            client.update_task_blueprints_and_subtasks_graph_from_draft(scheduling_unit_blueprint['id'])

            # check the blueprint graph
            scheduling_unit_blueprint = client.get_schedulingunit_blueprint(scheduling_unit_blueprint['id'], extended=True)
            self.assertEqual(3, len(scheduling_unit_blueprint['task_blueprints']))
            task_blueprint_pipe = next(t for t in scheduling_unit_blueprint['task_blueprints'] if t['name']=='pipeline')
            task_blueprint_pipe_copy = next(t for t in scheduling_unit_blueprint['task_blueprints'] if 'pipeline' in t['name'] and 'copy' in t['name'].lower())
            self.assertEqual(1, task_blueprint_pipe['specifications_doc']['average']['time_steps'])
            self.assertEqual(2, task_blueprint_pipe_copy['specifications_doc']['average']['time_steps'])

    def test_rerun_copy_of_failed_pipeline_TMSS_1764_bugfix(self):
        '''Test concrete use case where we simulate a run for a UC1 observation, where a pipeline fails. Then create an exact copy of the pipeline and let that re-run.
        See https://support.astron.nl/jira/browse/TMSS-1764 for the expected behaviour.
        '''
        # Step 0: setup UC1 observation
        project = models.Project.objects.create(**Project_test_data(name='test', auto_ingest=True))
        scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project))

        uc1_strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="IM HBA LoTSS - 2 Beams")
        scheduling_unit_spec = add_defaults_to_json_object_for_schema(uc1_strategy_template.template, uc1_strategy_template.scheduling_unit_template.schema)
        # limit target obs duration for demo data
        scheduling_unit_spec['tasks']['Calibrator Observation 1']['specifications_doc']['duration'] = 2 * 60
        scheduling_unit_spec['tasks']['Target Observation']['specifications_doc']['duration'] = 2 * 3600
        scheduling_unit_spec['tasks']['Calibrator Observation 2']['specifications_doc']['duration'] = 2 * 60

        constraints_template = models.SchedulingConstraintsTemplate.get_version_or_latest(name=scheduling_unit_spec.get('scheduling_constraints_template',{}).get('name', 'constraints'),
                                                                                          version=scheduling_unit_spec.get('scheduling_constraints_template', {}).get('version'))
        constraints = constraints_template.get_default_json_document_for_schema()

        # add the scheduling_unit_doc to a new SchedulingUnitDraft instance, and were ready to use it!
        su_draft = models.SchedulingUnitDraft.objects.create(name="test",
                                                             scheduling_set=scheduling_set,
                                                             specifications_template=uc1_strategy_template.scheduling_unit_template,
                                                             observation_strategy_template=uc1_strategy_template,
                                                             scheduling_constraints_doc=constraints,
                                                             scheduling_constraints_template=constraints_template)
        su_draft = update_task_graph_from_specifications_doc(su_draft, scheduling_unit_spec)
        su_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(su_draft)

        with tmss_test_env.create_tmss_client() as client:
            # Step 1: simulate obs run
            scheduling_unit_blueprint_ext = client.get_schedulingunit_blueprint(su_blueprint.id, extended=True, include_specifications_doc=True)

            tasks = scheduling_unit_blueprint_ext['task_blueprints']
            observation_tasks = [t for t in tasks if t['task_type'] == 'observation']   # 3 obs
            pipeline_tasks = [t for t in tasks if t['task_type'] == 'pipeline']         # 4 pipes
            ingest_task = next(t for t in tasks if t['task_type'] == 'ingest')          # 1 ingest

            cal_obs1_task = next(t for t in observation_tasks if t['name'] == 'Calibrator Observation 1')
            target_obs_task = next(t for t in observation_tasks if t['name'] == 'Target Observation')
            cal_obs2_task = next(t for t in observation_tasks if t['name'] == 'Calibrator Observation 2')
            cal_pipe1_task = next(t for t in pipeline_tasks if t['name'] == 'Calibrator Pipeline 1')
            target_pipe1_task = next(t for t in pipeline_tasks if t['name'] == 'Pipeline target1')
            target_pipe2_task = next(t for t in pipeline_tasks if t['name'] == 'Pipeline target2')
            cal_pipe2_task = next(t for t in pipeline_tasks if t['name'] == 'Calibrator Pipeline 2')

            # -------------------
            # schedule first calibrator obs
            cal_obs1_subtask = next(st for st in cal_obs1_task['subtasks'] if st['subtask_type'] == 'observation')
            cal_obs1_subtask = client.schedule_subtask(cal_obs1_subtask['id'])
            # "mimic" that the cal_obs1_subtask starts running
            set_subtask_state_following_allowed_transitions(cal_obs1_subtask['id'], 'started')
            # "mimic" that the cal_obs1_subtask finished (including qa subtasks)
            for subtask in cal_obs1_task['subtasks']:
                set_subtask_state_following_allowed_transitions(subtask['id'], 'finished')

            # -------------------
            # schedule target obs
            target_obs_subtask = next(st for st in target_obs_task['subtasks'] if st['subtask_type'] == 'observation')
            target_obs_subtask = client.schedule_subtask(target_obs_subtask['id'])
            # "mimic" that the target_obs_subtask starts running
            set_subtask_state_following_allowed_transitions(target_obs_subtask['id'], 'started')
            # "mimic" that the target_obs_subtask finished (including qa subtasks)
            for subtask in target_obs_task['subtasks']:
                set_subtask_state_following_allowed_transitions(subtask['id'], 'finished')

            # -------------------
            # schedule second calibrator obs
            cal_obs2_subtask = next(st for st in cal_obs2_task['subtasks'] if st['subtask_type'] == 'observation')
            cal_obs2_subtask = client.schedule_subtask(cal_obs2_subtask['id'])
            # "mimic" that the cal_obs2_subtask starts running
            set_subtask_state_following_allowed_transitions(cal_obs2_subtask['id'], 'started')
            # "mimic" that the cal_obs2_subtask finished (including qa subtasks)
            for subtask in cal_obs2_task['subtasks']:
                set_subtask_state_following_allowed_transitions(subtask['id'], 'finished')

            # -------------------
            # schedule first calibrator pipeline
            cal_pipe1_subtask = next(st for st in cal_pipe1_task['subtasks'] if st['subtask_type'] == 'pipeline')
            cal_pipe1_subtask = client.schedule_subtask(cal_pipe1_subtask['id'])
            # "mimic" that the cal_pipe1_subtask starts running
            set_subtask_state_following_allowed_transitions(cal_pipe1_subtask['id'], 'started')
            # "mimic" that the cal_pipe1_subtask finished
            set_subtask_state_following_allowed_transitions(cal_pipe1_subtask['id'], 'finished')

            # -------------------
            # schedule first target pipeline
            target_pipe1_subtask = next(st for st in target_pipe1_task['subtasks'] if st['subtask_type'] == 'pipeline')
            target_pipe1_subtask = client.schedule_subtask(target_pipe1_subtask['id'])
            # "mimic" that the target_pipe1_subtask starts running
            set_subtask_state_following_allowed_transitions(target_pipe1_subtask['id'], 'started')
            # "mimic" that the target_pipe1_subtask finished
            set_subtask_state_following_allowed_transitions(target_pipe1_subtask['id'], 'finished')

            # -------------------
            # schedule second target pipeline
            target_pipe2_subtask = next(st for st in target_pipe2_task['subtasks'] if st['subtask_type'] == 'pipeline')
            target_pipe2_subtask = client.schedule_subtask(target_pipe2_subtask['id'])
            # "mimic" that the target_pipe2_subtask starts running
            set_subtask_state_following_allowed_transitions(target_pipe2_subtask['id'], 'started')
            # "mimic" that the target_pipe2_subtask finished
            set_subtask_state_following_allowed_transitions(target_pipe2_subtask['id'], 'finished')

            # -------------------
            # Step 2: cancel one of the pipelines
            # schedule second calibrator pipeline and cancel it to reproduce behaviour
            cal_pipe2_subtask = next(st for st in cal_pipe2_task['subtasks'] if st['subtask_type'] == 'pipeline')
            cal_pipe2_subtask = client.schedule_subtask(cal_pipe2_subtask['id'])
            # "mimic" that the cal_pipe2_subtask is cancelled
            client.cancel_subtask(cal_pipe2_subtask['id'])

            cancelled_pipe_task = models.TaskBlueprint.objects.get(status=models.TaskStatus.Choices.CANCELLED.value)
            self.assertEqual(models.TaskStatus.Choices.CANCELLED.value, cancelled_pipe_task.status.value)

            # Step 3: create a copy for failed tasks
            # NOTE: Crucial part as the new copied pipeline should be linked as predecessor for the ingest.
            # This did not happen before, causing the issue.
            client.create_copies_of_failed_tasks_via_draft(su_blueprint.id)

            copy_pipe_task = models.TaskBlueprint.objects.get(name='Calibrator Pipeline 2 (copy for rerun 01)')
            copy_pipe_subtask = models.Subtask.objects.get(task_blueprint=copy_pipe_task)
            self.assertEqual(models.SubtaskState.Choices.DEFINED.value, copy_pipe_subtask.state.value)

            # Step 4: mark the cancelled pipeline obsolete
            client.mark_task_blueprint_as_obsolete(cancelled_pipe_task.id)
            cancelled_pipe_task.refresh_from_db()
            self.assertTrue(cancelled_pipe_task.is_obsolete)

            # Step 5: check proper ingest behaviour by simulating the run of the copied pipeline and then schedule the ingest
            client.schedule_subtask(copy_pipe_subtask.id, datetime.utcnow())
            set_subtask_state_following_allowed_transitions(copy_pipe_subtask.id, 'started')
            set_subtask_state_following_allowed_transitions(copy_pipe_subtask.id, 'finished')
            copy_pipe_subtask.refresh_from_db()

            # NOTE: Here is another crucial part.
            # If we did not exclude obsolete producers when ingesting, this would trigger an exception.
            ingest_task = models.TaskBlueprint.objects.get(id=ingest_task['id'])
            ingest_subtask = models.Subtask.objects.get(task_blueprint=ingest_task)

            client.schedule_subtask(ingest_subtask.id, copy_pipe_subtask.process_stop_time)
            ingest_subtask.refresh_from_db()
            self.assertEqual(models.SubtaskState.Choices.SCHEDULED.value, ingest_subtask.state.value)

            # Assert the new relation exists
            copy_pipe_ingest_task_relations = models.TaskRelationBlueprint.objects.filter(producer__id=copy_pipe_task.id, consumer__id=ingest_task.id)
            self.assertEqual(1, len(copy_pipe_ingest_task_relations))
            self.assertIn(copy_pipe_task, ingest_task.predecessors)

            # Step 6: check if an input exists where the producer subtask is the re-run pipeline and the dataproducts from that are actually picked up
            ingest_subtask_input = ingest_subtask.inputs.get(producer__subtask=copy_pipe_subtask)
            self.assertIsNotNone(ingest_subtask_input)
            for copy_pipe_output_dp in copy_pipe_subtask.output_dataproducts.all():
                self.assertIn(copy_pipe_output_dp, ingest_subtask.input_dataproducts.all())


    def test_rerun_copy_of_failed_pipeline_before_observation_ran_TMSS_1839_bugfix(self):
        '''Test concrete use case where we simulate a run for a UC1 observation, where a pipeline is cancelled before the observation ran.
        See https://support.astron.nl/jira/browse/TMSS-1839 for the expected behaviour.
        '''
        project = models.Project.objects.create(**Project_test_data(name=str(uuid.uuid4()), auto_ingest=True))
        scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project))
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="IM HBA - 1 Beam")
        su_draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template, scheduling_set=scheduling_set)
        su_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(su_draft)

        with tmss_test_env.create_tmss_client() as client:
            # Step 1: cancel pipeline
            scheduling_unit_blueprint_ext = client.get_schedulingunit_blueprint(su_blueprint.id, extended=True, include_specifications_doc=True)

            tasks = scheduling_unit_blueprint_ext['task_blueprints']
            observation_tasks = [t for t in tasks if t['task_type'] == 'observation']   # 3 obs
            pipeline_tasks = [t for t in tasks if t['task_type'] == 'pipeline']         # 4 pipes

            target_obs_task = next(t for t in observation_tasks if t['name'] == 'Target Observation')
            target_pipe_task = next(t for t in pipeline_tasks if t['name'] == 'Target Pipeline')

            target_obs_subtask = next(st for st in target_obs_task['subtasks'] if st['subtask_type'] == 'observation')
            self.assertEqual('defined', target_obs_subtask['state_value'])

            # cancel the pipeline before the observation was scheduled and runs
            target_pipe_task = client.cancel_task_blueprint(target_pipe_task['id'])
            self.assertEqual('cancelled', target_pipe_task['status'])

            # the method under test:
            # in the ticket it is mentioned that this gives a 504 error.
            # however, it should 'just work'
            client.create_copies_of_failed_tasks_via_draft(su_blueprint.id)

            # ok, previous call worked, so continue that we can schedule the observation... and then the rerun pipeline.
            target_obs_subtask = next(st for st in target_obs_task['subtasks'] if st['subtask_type'] == 'observation')
            target_obs_subtask = client.schedule_subtask(target_obs_subtask['id'])
            self.assertEqual('scheduled', target_obs_subtask['state_value'])
            set_subtask_state_following_allowed_transitions(target_obs_subtask['id'], 'finished')


            # schedule and run the rerun pipeline
            scheduling_unit_blueprint_ext = client.get_schedulingunit_blueprint(su_blueprint.id, extended=True, include_specifications_doc=True)
            tasks = scheduling_unit_blueprint_ext['task_blueprints']
            pipeline_tasks = [t for t in tasks if t['task_type'] == 'pipeline']
            target_pipe_task = next(t for t in pipeline_tasks if t['name'] == 'Target Pipeline (copy for rerun 01)')
            target_pipe_subtask = next(st for st in target_pipe_task['subtasks'] if st['subtask_type'] == 'pipeline')
            target_pipe_subtask = client.schedule_subtask(target_pipe_subtask['id'])
            self.assertEqual('scheduled', target_pipe_subtask['state_value'])



    def test_create_cleanuptask_for_scheduling_unit_blueprint(self):
        '''Test concrete use case where we forgot to add a cleanup task and we add it after the blueprint was created'''
        with tmss_test_env.create_tmss_client() as client:

            # create project, and a scheduling set within it
            project = rest_data_creator.post_data_and_get_response_as_json_object(rest_data_creator.Project(auto_ingest=True), '/project/')
            scheduling_set = rest_data_creator.post_data_and_get_response_as_json_object(rest_data_creator.SchedulingSet(project_url=project['url']), '/scheduling_set/')

            # create 'Simple Observation' scheduling unit draft in the scheduling set
            observing_strategy_templates = client.get_path_as_json_object('scheduling_unit_observing_strategy_template')
            strategy_template = next(ost for ost in observing_strategy_templates if ost['name']=='Simple Observation')
            self.assertIsNotNone(strategy_template)
            scheduling_unit_draft = client.create_scheduling_unit_draft_from_strategy_template(strategy_template['id'], scheduling_set['id'])
            num_task_drafts_before = len(scheduling_unit_draft['task_drafts'])
            num_task_relation_drafts_before = len(client.get_path_as_json_object('task_relation_draft'))

            # turn draft into a blueprint
            scheduling_unit_blueprint = client.create_scheduling_unit_blueprint_and_tasks_and_subtasks_tree(scheduling_unit_draft['id'])
            num_task_blueprints_before = len(scheduling_unit_blueprint['task_blueprints'])
            num_task_relation_blueprints_before = len(client.get_path_as_json_object('task_relation_blueprint'))
            self.assertNotIn("Cleanup", [task['name'] for task in scheduling_unit_blueprint['task_blueprints']])

            # create a cleanup task for the blueprint
            client.create_cleanuptask_for_scheduling_unit_blueprint(scheduling_unit_blueprint['id'])

            # check the blueprint graph
            scheduling_unit_blueprint = client.get_schedulingunit_blueprint(scheduling_unit_blueprint['id'], extended=True)
            self.assertEqual(num_task_blueprints_before + 1, len(scheduling_unit_blueprint['task_blueprints']))
            self.assertIn("Cleanup", [task['name'] for task in scheduling_unit_blueprint['task_blueprints']])
            for task in scheduling_unit_blueprint['task_blueprints']:
                if task['name'] == "Cleanup":
                    self.assertEqual(len(task['produced_by']), 1)
            self.assertEqual(num_task_relation_blueprints_before + 1, len(client.get_path_as_json_object('task_relation_blueprint')))

            # check the draft graph
            scheduling_unit_draft = client.get_schedulingunit_draft(scheduling_unit_draft['id'])
            self.assertEqual(num_task_drafts_before + 1, len(scheduling_unit_draft['task_drafts']))
            self.assertIn("Cleanup (Copy)", [task['name'] for task in scheduling_unit_draft['task_drafts']])
            for task in scheduling_unit_draft['task_drafts']:
                if task['name'] == "Cleanup (Copy)":
                    self.assertEqual(len(task['produced_by']), 1)
            self.assertEqual(num_task_relation_drafts_before + 1, len(client.get_path_as_json_object('task_relation_draft')))

            # create another cleanup task for the scheduling unit and check name is as expected
            client.create_cleanuptask_for_scheduling_unit_blueprint(scheduling_unit_blueprint['id'])
            scheduling_unit_blueprint = client.get_schedulingunit_blueprint(scheduling_unit_blueprint['id'], extended=True)
            self.assertIn("Cleanup2", [task['name'] for task in scheduling_unit_blueprint['task_blueprints']])

    def test_TMSS_1585_bugfix_check_cleanup_starts(self):
        '''Test concrete use case where we forgot to add a cleanup task and we add it after the blueprint was created'''
        # we want to check if the cleanup (sub)task actually starts (gets scheduled) after creation.
        # So we need a  subtask_scheduling_service running in the background
        from lofar.sas.tmss.services.scheduling.subtask_scheduling import create_subtask_scheduling_service
        with tmss_test_env.create_tmss_client() as client:
            # create project, and a scheduling set within it
            project = rest_data_creator.post_data_and_get_response_as_json_object(rest_data_creator.Project(auto_ingest=True, auto_pin=True), '/project/')
            scheduling_set = rest_data_creator.post_data_and_get_response_as_json_object(rest_data_creator.SchedulingSet(project_url=project['url']), '/scheduling_set/')

            # create 'Simple Observation' scheduling unit draft in the scheduling set
            observing_strategy_templates = client.get_path_as_json_object('scheduling_unit_observing_strategy_template')
            strategy_template = next(ost for ost in observing_strategy_templates if ost['name'] == 'Simple Observation')
            self.assertIsNotNone(strategy_template)
            scheduling_unit_draft = client.create_scheduling_unit_draft_from_strategy_template(strategy_template['id'], scheduling_set['id'])

            # turn draft into a blueprint
            scheduling_unit_blueprint = client.create_scheduling_unit_blueprint_and_tasks_and_subtasks_tree(scheduling_unit_draft['id'])

            # check there is no cleanup yet
            self.assertNotIn("Cleanup", [task['name'] for task in scheduling_unit_blueprint['task_blueprints']])

            # check data is pinned
            self.assertTrue(scheduling_unit_blueprint['output_pinned'])

            # mimic that the observation ran
            obs_subtask = models.Subtask.objects.get(task_blueprint__scheduling_unit_blueprint__id=scheduling_unit_blueprint['id'], specifications_template__type__value='observation')
            obs_subtask = schedule_subtask(obs_subtask, datetime.utcnow() + timedelta(minutes=5))
            set_subtask_state_following_allowed_transitions(obs_subtask, 'finished')

            # is the output still pinned?
            obs_subtask.task_blueprint.refresh_from_db()
            self.assertTrue(obs_subtask.task_blueprint.output_pinned)

            # create a cleanup task for the blueprint
            client.create_cleanuptask_for_scheduling_unit_blueprint(scheduling_unit_blueprint['id'])

            # run the subtask scheduling service.
            # check/test that it reponds to the unpinning event by scheduling the cleanup subtask after unpinning
            with BusListenerJanitor(create_subtask_scheduling_service()):
                # is the observation output still pinned?
                obs_subtask.task_blueprint.refresh_from_db()
                self.assertTrue(obs_subtask.task_blueprint.output_pinned)

                cleanup_subtask = models.Subtask.objects.get(task_blueprint__scheduling_unit_blueprint__id=scheduling_unit_blueprint['id'],
                                                             specifications_template__type__value='cleanup')
                # the obs is finished but pinned, so the cleanup subtask should not be scheduled yet.
                self.assertEqual('defined', cleanup_subtask.state.value)

                # unpin, and wait until automatically scheduled
                logger.info("unpinning...")
                client.set_scheduling_unit_blueprint_pinning(scheduling_unit_blueprint['id'], False)
                logger.info("unpinned")

                cleanup_subtask = wait_for_subtask_status(cleanup_subtask, 'scheduled', timeout=10)
                self.assertEqual('scheduled', cleanup_subtask.state.value)


    def test_TMSS_2291_bugfix_FE_observation_finishes_when_some_stations_reserved(self):
        '''Test concrete use case where we some stations are reserved and thus not part of the observation.
        FE feedback should ignore these stations and not wait for feedback'''
        from lofar.sas.tmss.services.scheduling.subtask_scheduling import create_subtask_scheduling_service
        with tmss_test_env.create_tmss_client() as client:
            # create project, and a scheduling set within it
            project = rest_data_creator.post_data_and_get_response_as_json_object(rest_data_creator.Project(auto_ingest=True, auto_pin=True), '/project/')
            scheduling_set = rest_data_creator.post_data_and_get_response_as_json_object(rest_data_creator.SchedulingSet(project_url=project['url']), '/scheduling_set/')

            start_time = datetime.utcnow()+timedelta(hours=1)
            stop_time = start_time+timedelta(hours=1)

            # create an empty scheduling_unit_draft...
            scheduling_unit_template = next(template for template in client.get_path_as_json_object('scheduling_unit_template') if template['name']=='scheduling unit')
            scheduling_constraints_template = next(template for template in client.get_path_as_json_object('scheduling_constraints_template') if template['name']=='constraints')
            scheduling_constraints_doc = {"time": {"at": start_time.isoformat()}, "scheduler": "fixed_time"}
            scheduling_unit_draft_test_data = rest_data_creator.SchedulingUnitDraft(template_url=scheduling_unit_template['url'],
                                                                                    scheduling_constraints_template_url=scheduling_constraints_template['url'],
                                                                                    scheduling_constraints_doc=scheduling_constraints_doc)
            scheduling_unit_draft = client.post_to_path_and_get_result_as_json_object('scheduling_unit_draft',
                                                                                      json_data=scheduling_unit_draft_test_data)
            # ... and add a fly's eye beamforming observation to it
            fe_task_template = client.get_task_template('beamforming observation') # latest version
            fe_task_spec = client.get_task_template_default_specification(fe_task_template['name'], fe_task_template['version'])
            fe_task_spec['duration'] = (stop_time-start_time).total_seconds()
            # use two stations, allow 1 missing
            fe_task_spec['station_configuration']['station_groups'] = [{'max_nr_missing': 1, 'stations': ['CS011', 'CS013']}]
            # use LBA (because then we get one dataprodcut per station, in constrast to dual HBA fiels in core stations)
            fe_task_spec['station_configuration']['antenna_set'] = 'LBA_OUTER'
            # disable QA, coherent and incoherent
            fe_task_spec["QA"] = { "file_conversion": { "enabled": False }, "plots": { "enabled": False } }
            fe_task_spec["beamformer"]["pipelines"][0]["coherent"]["SAPs"] = []
            fe_task_spec["beamformer"]["pipelines"][0]["incoherent"]["SAPs"] = []
            # enable/setup fly's eye
            fe_task_spec["beamformer"]["pipelines"][0]["flys eye"]["enabled"] = True
            # use same stations as in station_configuration
            fe_task_spec["beamformer"]["pipelines"][0]['station_groups'] = fe_task_spec['station_configuration']['station_groups']

            # upload the FE obs to the schedulingunit
            su_specifications_doc = copy.deepcopy(scheduling_unit_draft['specifications_doc'])
            su_specifications_doc['tasks'] = {'FE Observation': {'specifications_doc': fe_task_spec,
                                                                 'specifications_template': {"name": fe_task_template['name'],
                                                                                             "version": fe_task_template['version']} } }
            scheduling_unit_draft = client.update_task_graph_from_specifications_doc(scheduling_unit_draft['id'],
                                                                                     specifications_doc=su_specifications_doc)

            # turn draft into a blueprint
            scheduling_unit_blueprint = client.create_scheduling_unit_blueprint_and_tasks_and_subtasks_tree(scheduling_unit_draft['id'])
            fe_obs_subtask = next(s for s in client.get_subtasks_in_scheduling_unit(scheduling_unit_blueprint['id']) if s['subtask_type']=='observation')

            # create a reservation for 1 station.
            # this is "acceptable", because "max_nr_missing" is >0
            # thus, the unit should be able the get scheduled and run.
            # The issue comes later, when we receive only feedback for the non-reserved stations.
            reservation_template = models.ReservationTemplate.get_latest(name="reservation")
            reservation_template_spec = get_default_json_object_for_schema(reservation_template.schema)
            reservation_template_spec["resources"] = {"stations": ["CS011"] }
            res = models.Reservation.objects.create(name="Station Reservation CS011",
                                                    description="Station reservation for testing",
                                                    specifications_template=reservation_template,
                                                    specifications_doc=reservation_template_spec,
                                                    start_time=start_time,
                                                    stop_time=stop_time)

            # schedule the FE subtask, and check it
            fe_obs_subtask = client.schedule_subtask(fe_obs_subtask['id'])
            self.assertEqual('scheduled', fe_obs_subtask['state_value'])

            # it is expected that the reserved station is missing
            missing_stations = client.get_url_as_json_object(fe_obs_subtask['url'].rstrip('/') + '/missing_stations')['missing_stations']
            self.assertEqual(['CS011'], missing_stations)
            used_stations = fe_obs_subtask['specifications_doc']['stations']['station_list']
            self.assertEqual(1, len(used_stations))
            self.assertEqual(['CS013'], used_stations)

        # setup via Rest Api done.
        # continue rest of test with django ORM api for simplicity.

        fe_obs_subtask = models.Subtask.objects.get(id=fe_obs_subtask['id'])

        # simulate running it, up until the finishing state, where we wait for feedback
        fe_obs_subtask = set_subtask_state_following_allowed_transitions(fe_obs_subtask, 'finishing')

        # no feedback was submitted yet, so is_feedback_complete should be False
        fe_obs_subtask.refresh_from_db()
        self.assertFalse(fe_obs_subtask.is_feedback_complete)

        # prepare/submit subtask feedback
        subtask_feedback_template = models.SubtaskFeedbackTemplate.get_latest('observation')
        subtask_feedback_doc = subtask_feedback_template.get_default_json_document_for_schema()
        fe_obs_subtask.processing_feedback_doc = subtask_feedback_doc
        fe_obs_subtask.processing_feedback_template = subtask_feedback_template
        fe_obs_subtask.save()

        # only subtask feedback was submitted, but no dataproduct feedback yet, so is_feedback_complete should be False
        fe_obs_subtask.refresh_from_db()
        self.assertFalse(fe_obs_subtask.is_feedback_complete)

        # an FE obs should create one dataproduct per used station
        output_dataproducts = list(fe_obs_subtask.output_dataproducts.all())
        self.assertEqual(len(used_stations), len(output_dataproducts))

        # now create feedback for the used stations, but NOT for the reserved/missing station (which causes the bug in TMSS-2291)
        dp_feedback_template = models.DataproductFeedbackTemplate.get_latest('feedback')
        for station, dp in zip(used_stations, output_dataproducts):
            dp_feedback_doc = dp_feedback_template.get_default_json_document_for_schema()
            dp_feedback_doc['antennas'] = { "set": "LBA_OUTER",
                                            "fields": [ {
                                                "type": "LBA",
                                                "field": "LBA",
                                                "station": station
                                            } ]
                                            }
            dp.feedback_doc = dp_feedback_doc
            dp.feedback_template = dp_feedback_template
            dp.save()

        # both subtask feedback and dataproduct feedback were submitted
        # so is_feedback_complete should be True
        fe_obs_subtask.refresh_from_db()
        self.assertTrue(fe_obs_subtask.is_feedback_complete)

    def test_TMSS_2275_bugfix_scheduling_unit_status_when_all_tasks_obsolete(self):
        '''See https://support.astron.nl/jira/browse/TMSS-2275'''
        # start with an empty unit, check status
        scheduling_unit_template = models.SchedulingUnitTemplate.get_latest('scheduling unit')
        unit = models.SchedulingUnitBlueprint.objects.create(**SchedulingUnitBlueprint_test_data(specifications_template=scheduling_unit_template))
        self.assertEqual(models.SchedulingUnitStatus.Choices.DEFINED.value, unit.status.value)

        # add one task, check status
        task_template = models.TaskTemplate.get_latest('observation')
        task = models.TaskBlueprint.objects.create(**TaskBlueprint_test_data(specifications_template=task_template, scheduling_unit_blueprint=unit))
        self.assertEqual(models.TaskStatus.Choices.DEFINED.value, task.status.value)
        unit.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.DEFINED.value, unit.status.value)

        # add one subtask, check status
        subtask_template = models.SubtaskTemplate.get_latest('observation control')
        subtask = models.Subtask.objects.create(**Subtask_test_data(subtask_template=subtask_template, task_blueprint=task))
        self.assertEqual(models.SubtaskState.Choices.DEFINING.value, subtask.state.value)
        task.refresh_from_db()
        unit.refresh_from_db()
        self.assertEqual(models.TaskStatus.Choices.DEFINED.value, task.status.value)
        self.assertEqual(models.SchedulingUnitStatus.Choices.DEFINED.value, unit.status.value)
        subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.DEFINED.value)
        subtask.save()
        subtask.refresh_from_db()
        task.refresh_from_db()
        unit.refresh_from_db()
        self.assertEqual(models.SubtaskState.Choices.DEFINED.value, subtask.state.value)
        self.assertEqual(models.TaskStatus.Choices.SCHEDULABLE.value, task.status.value)
        self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULABLE.value, unit.status.value)

        # cancel subtask, check status
        subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.CANCELLING.value)
        subtask.save()
        subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.CANCELLED.value)
        subtask.save()
        subtask.refresh_from_db()
        task.refresh_from_db()
        unit.refresh_from_db()
        self.assertEqual(models.SubtaskState.Choices.CANCELLED.value, subtask.state.value)
        self.assertEqual(models.TaskStatus.Choices.CANCELLED.value, task.status.value)
        self.assertEqual(models.SchedulingUnitStatus.Choices.CANCELLED.value, unit.status.value)

        # make subtask obsolete, check status
        subtask.obsolete_since = datetime.utcnow()
        subtask.save()
        subtask.refresh_from_db()
        task.refresh_from_db()
        unit.refresh_from_db()
        self.assertEqual(models.SubtaskState.Choices.CANCELLED.value, subtask.state.value)
        self.assertEqual(models.TaskStatus.Choices.CANCELLED.value, task.status.value)

        # the bug from TMSS-2275 used to be that the unit had status SCHEDULABLE. Now it has the correct status CANCELLED.
        self.assertEqual(models.SchedulingUnitStatus.Choices.CANCELLED.value, unit.status.value)

    def test_create_schedulingunitblueprint_copy_without_given_station(self):
        with tmss_test_env.create_tmss_client() as client:
            # create project, and a scheduling set within it
            project = rest_data_creator.post_data_and_get_response_as_json_object(rest_data_creator.Project(auto_ingest=True, auto_pin=True), '/project/')
            scheduling_set = rest_data_creator.post_data_and_get_response_as_json_object(rest_data_creator.SchedulingSet(project_url=project['url']), '/scheduling_set/')

            # create 'Simple Observation' scheduling unit draft in the scheduling set
            observing_strategy_templates = client.get_path_as_json_object('scheduling_unit_observing_strategy_template')
            strategy_template = next(ost for ost in observing_strategy_templates if ost['name'] == 'Simple Observation')
            self.assertIsNotNone(strategy_template)
            scheduling_unit_draft = client.create_scheduling_unit_draft_from_strategy_template(strategy_template['id'], scheduling_set['id'])

            # turn draft into a blueprint
            scheduling_unit_blueprint = client.create_scheduling_unit_blueprint_and_tasks_and_subtasks_tree(scheduling_unit_draft['id'])

            # schedule it
            scheduling_unit_blueprint = client.schedule_scheduling_unit(scheduling_unit_blueprint['id'], datetime.utcnow() + timedelta(minutes=10))
            self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint['status_value'])

            # get the originally specified group(s) of stations
            scheduling_unit_blueprint = client.get_schedulingunit_blueprint(scheduling_unit_blueprint['id'], include_specifications_doc=True)
            org_station_groups = scheduling_unit_blueprint['specifications_doc']['tasks']["Observation"]['specifications_doc']['station_configuration']['station_groups']

            stations_to_be_removed = ['CS002', 'CS005']

            # reassure these stations are in the original spec
            for station in stations_to_be_removed:
                self.assertIn(station, str(org_station_groups))

            # take out the stations (in a new copy)
            scheduling_unit_blueprint_copy = client.create_scheduling_unit_blueprint_copy_without_given_stations(scheduling_unit_blueprint['id'], stations_to_be_removed)

            # check it
            scheduling_unit_blueprint_copy = client.get_schedulingunit_blueprint(scheduling_unit_blueprint_copy['id'], include_specifications_doc=True)
            self.assertNotEqual(scheduling_unit_blueprint['id'], scheduling_unit_blueprint_copy['id'])

            # Check the core feature: are the stations taken out of the new copy?
            new_station_groups = scheduling_unit_blueprint_copy['specifications_doc']['tasks']["Observation"]['specifications_doc']['station_configuration']['station_groups']
            for station in stations_to_be_removed:
                self.assertNotIn(station, str(new_station_groups))

            # check the status of the original, is it cancelled and obsolete?
            scheduling_unit_blueprint = client.get_schedulingunit_blueprint(scheduling_unit_blueprint['id'], include_specifications_doc=True)
            self.assertEqual(models.SchedulingUnitStatus.Choices.CANCELLED.value, scheduling_unit_blueprint['status_value'])


if __name__ == "__main__":
    os.environ['TZ'] = 'UTC'
    unittest.main()
