#!/usr/bin/env python3

# Copyright (C) 2018    ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

# $Id:  $
import json
import os
import unittest

import logging
logger = logging.getLogger(__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

from lofar.common.test_utils import exit_with_skipped_code_if_skip_integration_tests
exit_with_skipped_code_if_skip_integration_tests()

# Do Mandatory setup step:
# use setup/teardown magic for tmss test database, ldap server and django server
# (ignore pycharm unused import statement, python unittests does use at RunTime the tmss_test_environment_unittest_setup module)
from lofar.sas.tmss.test.tmss_test_environment_unittest_setup import *
tmss_test_env.populate_schemas_and_connectors()

from lofar.sas.tmss.tmss.tmssapp import models
from lofar.common.json_utils import resolved_remote_refs, validate_json_against_schema, get_default_json_object_for_schema

class TestSchemas(unittest.TestCase):
    def check_schema(self, schema: dict):
        """ Check whether the given schema is valid. """

        # Can all $refs be actually resolved?
        try:
            ref_resolved_schema = resolved_remote_refs(schema)
        except Exception as e:
            raise Exception("Failed to resolve references in schema: %s" % (e,)) from e

        # Does this schema provide actually valid defaults?
        try:
            doc_with_defaults = get_default_json_object_for_schema(schema)
            validate_json_against_schema(doc_with_defaults, schema)
        except Exception as e:
            raise Exception("Failure in defaults in schema: %s" % (e,)) from e

    def check_template_table(self, model):
        """ Check all schemas present in the database for a given model. """

        for template in model.objects.all():
            try:
                self.check_schema(template.schema)
            except Exception as e:
                raise Exception("Error while checking schema for %s name='%s' version=%s: %s" % (template.__class__.__name__, template.name, template.version, e)) from e

    def test_commonschemas(self):
        self.check_template_table(models.CommonSchemaTemplate)

    def test_systemevent(self):
        self.check_template_table(models.SystemEventTemplate)

    def test_subtasks(self):
        self.check_template_table(models.SubtaskTemplate)
        self.check_template_table(models.SubtaskFeedbackTemplate)

    def test_dataproducts(self):
        self.check_template_table(models.DataproductSpecificationsTemplate)
        self.check_template_table(models.DataproductFeedbackTemplate)
        self.check_template_table(models.SAPTemplate)

    def test_tasks(self):
        self.check_template_table(models.TaskTemplate)
        self.check_template_table(models.TaskRelationSelectionTemplate)

    def test_scheduling_units(self):
        self.check_template_table(models.SchedulingUnitTemplate)
        self.check_template_table(models.SchedulingConstraintsTemplate)

    def test_reservations(self):
        self.check_template_table(models.ReservationTemplate)

if __name__ == "__main__":
    os.environ['TZ'] = 'UTC'
    unittest.main()
