#!/usr/bin/env python3

# Copyright (C) 2018    ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

# $Id:  $

import os
import unittest
from copy import deepcopy
import math

import logging
logger = logging.getLogger(__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

from lofar.common.test_utils import exit_with_skipped_code_if_skip_integration_tests
exit_with_skipped_code_if_skip_integration_tests()

# Do Mandatory setup step:
# use setup/teardown magic for tmss test database, ldap server and django server
# (ignore pycharm unused import statement, python unittests does use at RunTime the tmss_test_environment_unittest_setup module)
from lofar.sas.tmss.test.tmss_test_environment_unittest_setup import *
tmss_test_env.populate_schemas_and_connectors()


from lofar.sas.tmss.test.tmss_test_data_django_models import *
from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.tmssapp.subtasks import *
from lofar.sas.tmss.tmss.tmssapp.subtasks import _get_related_target_sap_by_name, _generate_tab_ring_pointings, _filter_subbands, project_pointing
from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions


def create_subtask_object_for_testing(subtask_type_value):
    """
    Helper function to create a subtask object for testing with given subtask value and subtask state value
    as string (no object)
    """
    template_type = models.SubtaskType.objects.get(value=subtask_type_value)
    subtask_template_obj = create_subtask_template_for_testing(template_type)
    subtask_data = Subtask_test_data(subtask_template=subtask_template_obj, task_blueprint=models.TaskBlueprint.objects.create(**TaskBlueprint_test_data()))
    subtask = models.Subtask.objects.create(**subtask_data)
    return subtask

def create_subtask_template_for_testing(template_type: object):
    """
    Helper function
    :param template_type:
    :return:
    """
    subtask_template_data = SubtaskTemplate_test_data()
    subtask_template_data['type'] = template_type
    return models.SubtaskTemplate.objects.create(**subtask_template_data)


def create_task_blueprint_object_for_testing(task_template_name="target observation", QA_enabled=False, task_spec: dict=None):
    """
    Helper function to create a task blueprint object for testing with given task template name value
    as string (no object)
    :param task_template_name: (Optional) name of schema target observation is target observation
    :param QA_enabled: (Optional) QA plots and file_conversion
    :return: task_blueprint_obj: Created Task Blueprint object
    """
    task_template = models.TaskTemplate.get_latest(name=task_template_name)
    task_spec = task_template.add_defaults_to_json_object_for_schema(task_spec or {})
    if 'QA' in task_spec:
        task_spec["QA"]['plots']['enabled'] = QA_enabled
        task_spec["QA"]['file_conversion']['enabled'] = QA_enabled

    task_draft_data = TaskDraft_test_data(specifications_template=task_template, specifications_doc=task_spec)
    task_draft_obj = models.TaskDraft.objects.create(**task_draft_data)

    task_name = "BlueprintTask with %s" % task_template_name
    task_blueprint_data = TaskBlueprint_test_data(name=task_name, task_draft=task_draft_obj)
    task_blueprint_obj = models.TaskBlueprint.objects.create(**task_blueprint_data)
    return task_blueprint_obj


def create_relation_task_blueprint_object_for_testing(blueprint_task_producer, blueprint_task_consumer):
    """
    Helper function to create a task relation blueprint object for testing for given task objects consumer and producer
    :param blueprint_task_producer: Blueprint task of producer, typical an observation
    :param blueprint_task_consumer: Blueprint task of consumer, typical a preprocessing pipeline
    :return: task_relation_obj: Created Task Relation Blueprint object
    """
    task_relation_data = TaskRelationBlueprint_test_data(blueprint_task_producer, blueprint_task_consumer)
    task_relation_obj = models.TaskRelationBlueprint.objects.create(**task_relation_data)
    return task_relation_obj


def create_scheduling_relation_task_blueprint_for_testing(first_task_blueprint, second_task_blueprint, placement='before'):
    """
    Helper function to create a task blueprint relation object between two task blueprint (calibrator and target observation)
    :param first_task_blueprint:
    :param second_task_blueprint:
    :return: task_relation_blueprint_obj: Created Task Relation Blueprint object
    """
    task_scheduling_rel_obj = models.TaskSchedulingRelationBlueprint.objects.create(
                                 tags=[],
                                 draft=models.TaskSchedulingRelationDraft.objects.create(first=first_task_blueprint.draft, second=second_task_blueprint.draft, placement=models.SchedulingRelationPlacement.objects.get(value=placement)),
                                 first=first_task_blueprint,
                                 second=second_task_blueprint,
                                 placement=models.SchedulingRelationPlacement.objects.get(value=placement),
                                 time_offset=60)
    return task_scheduling_rel_obj


class SubTasksCreationFromSubTask(unittest.TestCase):

    def test_create_qafile_subtask_from_pipeline_subtask_failed(self):
        """
        Test if creation of subtask qafile failed due to wrong type of the predecessor subtask
        Correct type should be 'observation' (for this test of course it is not)
        """
        pipeline_subtask = create_subtask_object_for_testing("pipeline")
        with self.assertRaises(ValueError):
            create_qafile_subtask_from_observation_subtask(pipeline_subtask)

    def test_create_qafile_subtask_from_observation_subtask_succeed(self):
        """
        Test if creation of subtask qafile succeed
        Subtask object is None because QA file conversion is  by default not enabled!!!!
        """
        predecessor_subtask = create_subtask_object_for_testing("observation")
        subtask = create_qafile_subtask_from_observation_subtask(predecessor_subtask)
        self.assertEqual(None, subtask)

    def test_create_qaplots_subtask_from_qafile_subtask_failed(self):
        """
        Test if creation of subtask qaplots failed due to wrong state or wrong type of the predecessor subtask
        Correct type should be 'qa_files' (for this test of course it is not)
        """
        subtasks = [create_subtask_object_for_testing("pipeline"),
                    create_subtask_object_for_testing("observation"),
                    create_subtask_object_for_testing("observation") ]
        for subtask in subtasks:
            with self.assertRaises(ValueError):
                create_qaplots_subtask_from_qafile_subtask(subtask)

    def test_create_qaplots_subtask_from_qafile_subtask_succeed(self):
        """
        Test if creation of subtask qaplots succeed
        Subtask object is None because QA plots is by default not enabled!!!!
        """
        predecessor_subtask = create_subtask_object_for_testing("qa_files")
        subtask = create_qaplots_subtask_from_qafile_subtask(predecessor_subtask)
        self.assertEqual(None, subtask)


class SubTasksCreationFromTaskBluePrint(unittest.TestCase):

    def setUp(self) -> None:
        tmss_test_env.delete_reservations_cascade()
        tmss_test_env.delete_scheduling_unit_drafts_cascade()

    def test_create_sequence_of_subtask_from_task_blueprint(self):
        """
        Create multiple subtasks from a task blueprint, executed in correct order.
        No exception should occur, check name, type and state of the subtask
        """
        task_blueprint = create_task_blueprint_object_for_testing()

        subtask = create_observation_control_subtask_from_task_blueprint(task_blueprint)
        self.assertEqual("defined", str(subtask.state))
        self.assertEqual("observation control", str(subtask.specifications_template.name))
        self.assertEqual("observation", str(subtask.specifications_template.type))

        # Next call requires an observation subtask already created
        subtask = create_qafile_subtask_from_task_blueprint(task_blueprint)
        # subtask object is None because QA file conversion is by default not enabled!!!!
        self.assertEqual(None, subtask)

        # Next call will fail due to no qa_files object
        subtask = create_qaplots_subtask_from_task_blueprint(task_blueprint)
        # subtask object is None because QA file conversion is by default not enabled!!!!
        self.assertEqual(None, subtask)


    def test_create_sequence_of_subtask_from_task_blueprint_with_QA_enabled(self):
        """
        Create multiple subtasks from a task blueprint, executed in correct order.
        QA plots and QA file conversion enabled
        No exception should occur, check name, type and state of the subtasks
        """
        # Create Observation Task Enable QA plot and QA conversion
        task_blueprint = create_task_blueprint_object_for_testing(QA_enabled=True)
        task_blueprint_preprocessing = create_task_blueprint_object_for_testing("preprocessing pipeline")

        subtask = create_observation_control_subtask_from_task_blueprint(task_blueprint)
        self.assertEqual("defined", str(subtask.state))
        self.assertEqual("observation control", str(subtask.specifications_template.name))
        self.assertEqual("observation", str(subtask.specifications_template.type))
        self.assertTrue(subtask.primary)
        # Next call requires an observation subtask already created
        subtask = create_qafile_subtask_from_task_blueprint(task_blueprint)
        self.assertEqual("defined", str(subtask.state))
        self.assertEqual("QA file conversion", str(subtask.specifications_template.name))
        self.assertEqual("qa_files", str(subtask.specifications_template.type))
        self.assertFalse(subtask.primary)
        # Next call requires an qaplots subtask already created
        subtask = create_qaplots_subtask_from_task_blueprint(task_blueprint)
        self.assertEqual("defined", str(subtask.state))
        self.assertEqual("QA plots", str(subtask.specifications_template.name))
        self.assertEqual("qa_plots", str(subtask.specifications_template.type))
        self.assertFalse(subtask.primary)
        # Next call will fail due to missing task relation
        with self.assertRaises(SubtaskCreationException):
            create_preprocessing_subtask_from_task_blueprint(task_blueprint_preprocessing)
        # Create that relation and check again
        create_relation_task_blueprint_object_for_testing(task_blueprint, task_blueprint_preprocessing)
        subtask = create_preprocessing_subtask_from_task_blueprint(task_blueprint_preprocessing)
        self.assertEqual("defined", str(subtask.state))
        self.assertEqual("preprocessing pipeline", str(subtask.specifications_template.name))
        self.assertEqual("pipeline", str(subtask.specifications_template.type))
        self.assertTrue(subtask.primary)

    def test_create_subtasks_from_task_blueprint_succeed(self):
        """
        """
        task_blueprint = create_task_blueprint_object_for_testing(QA_enabled=True)
        subtasks = create_or_update_subtasks_from_task_blueprint(task_blueprint)
        self.assertEqual(3, len(subtasks))

    def test_create_subtasks_from_task_blueprint_translates_SAP_names(self):
        task_blueprint = create_task_blueprint_object_for_testing('target observation')
        task_blueprint.specifications_doc['station_configuration']['SAPs'] = [{'name': 'target1', 'subbands': [1],
                                                      'digital_pointing': {'angle1': 0.1, 'angle2': 0.1,
                                                                           'direction_type': 'J2000',
                                                                           'target': 'target1'}},
                                                     {'name': 'target2', 'subbands': [1],
                                                      'digital_pointing': {'angle1': 0.2, 'angle2': 0.2,
                                                                           'direction_type': 'J2000',
                                                                           'target': 'target2'}}]
        subtask = create_observation_control_subtask_from_task_blueprint(task_blueprint)
        i = 0
        for sap in task_blueprint.specifications_doc['station_configuration']['SAPs']:
            subtask_pointing = subtask.specifications_doc['stations']['digital_pointings'][i]
            self.assertEqual(sap['name'], subtask_pointing['name'])
            self.assertEqual(sap['digital_pointing']['angle1'], subtask_pointing['pointing']['angle1'])
            i += 1

    def test_TMSS_2321_bug_get_transformed_output_dataproduct_returns_multiple(self):
        '''This test was written to reproduce bug reported in https://support.astron.nl/jira/browse/TMSS-2321
        After some interactive coding I found out that the problem is not per-se in the Ingest subtask,
        but that this faulty behaviour only manifested itself in this particular scheduling unit in the scheduling of the ingest subtask.

        So, this test only reproduces the relevant part(s), and not the entire scenario as described in the ticket.

        The root-cause is that it is possible to have multiple releations between tasks,
        and that results in the same dataproduct being reported multiple times in the Subtask.input_dataproducts and/or Subtask.output_dataproducts,
        and thus that the calling Subtask.get_transformed_input_dataproduct or Subtask.get_transformed_output_dataproduct raises on the non-uniqueness of the given id.
        '''
        # create a schedulingunit with two tasks
        # 1. BF&IM obs, with two types of output and ANY-type of output
        from lofar.sas.tmss.tmss.tmssapp.tasks import update_task_graph_from_specifications_doc, create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft, schedule_independent_subtasks_in_scheduling_unit_blueprint

        scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(**SchedulingUnitDraft_test_data(template=models.SchedulingUnitTemplate.get_latest('scheduling unit'),
                                                                                                          scheduling_constraints_template=models.SchedulingConstraintsTemplate.get_latest('constraints')))
        specifications_doc = scheduling_unit_draft.specifications_doc
        specifications_doc['tasks']['Observation'] = {'specifications_template': {'name': 'target observation'},
                                                      'specifications_doc': { 'QA': { 'file_conversion': {'enabled': False} , 'plots': {'enabled': False} }}}
        specifications_doc['tasks']['Ingest'] = {'specifications_template': {'name': 'ingest'}, 'specifications_doc': {}}
        specifications_doc['task_relations'].append(
            {
                "producer": "Observation", "output": { "dataformat": "MeasurementSet", "datatype": "visibilities", "role": "correlator" },
                "consumer": "Ingest", "input": { "dataformat": "MeasurementSet", "datatype": "visibilities", "role": "any" },
                "selection_doc": {}, "selection_template": { "name": "all" }
            })
        scheduling_unit_draft = update_task_graph_from_specifications_doc(scheduling_unit_draft, specifications_doc)

        # check standard relation between these tasks
        # there should only be one relation, observation outputting correlator-role data, into ingest accepting any-role data
        specifications_doc = scheduling_unit_draft.specifications_doc
        self.assertEqual(1, len(specifications_doc['task_relations']))
        task_rel = specifications_doc['task_relations'][0]
        self.assertEqual("Observation", task_rel['producer'])
        self.assertEqual("Ingest", task_rel['consumer'])
        self.assertEqual("correlator", task_rel['output']['role'])
        self.assertEqual("any", task_rel['input']['role'])

        # create a 2nd superflous/overspecified link
        # the user shouldn't do this, but hey... it's possible, and we know users, so they do it anyway.
        # It is not "forbidden" to create the 2nd link, and there may be a good use for it for certain cases.
        # That's why we just accept the superflous/overspecified link.
        extra_task_rel = deepcopy(specifications_doc['task_relations'][0])
        extra_task_rel['input']['role'] = "correlator"
        specifications_doc['task_relations'].append(extra_task_rel)
        scheduling_unit_draft = update_task_graph_from_specifications_doc(scheduling_unit_draft, specifications_doc)
        specifications_doc = scheduling_unit_draft.specifications_doc
        self.assertEqual(2, len(specifications_doc['task_relations']))

        # specification done.
        # convert to blueprint
        scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
        # are the same 2 relations also in the blueprint?
        self.assertEqual(2, len(scheduling_unit_blueprint.specifications_doc['task_relations']))
        for task_rel in scheduling_unit_blueprint.specifications_doc['task_relations']:
            self.assertEqual("Observation", task_rel['producer'])
            self.assertEqual("Ingest", task_rel['consumer'])

        obs_subtask = scheduling_unit_blueprint.subtasks.filter(specifications_template__type__value=SubtaskType.Choices.OBSERVATION.value).first()
        ingest_subtask = scheduling_unit_blueprint.subtasks.filter(specifications_template__type__value=SubtaskType.Choices.INGEST.value).first()

        # schedule/run the observation
        obs_subtask = schedule_subtask(obs_subtask, datetime.utcnow()+timedelta(minutes=1))
        set_subtask_state_following_allowed_transitions(obs_subtask, 'finished')

        # observation should generate only one dataproduct
        self.assertEqual(1, obs_subtask.output_dataproducts.count())

        scheduling_unit_blueprint.ingest_permission_granted_since = datetime.utcnow()
        scheduling_unit_blueprint.save()

        ingest_subtask = schedule_subtask(ingest_subtask, datetime.utcnow()+timedelta(minutes=10))

        # ingest should take only the one dataproduct from the observation as input
        # and it should be possible to find the input_dataproduct from a given ingest output_dataproduct
        # These 3 asserts used to fail, reproducing/confirming the bug. Now they (should) pass, proving the bug was solved.
        self.assertEqual(1, ingest_subtask.input_dataproducts.count())
        self.assertEqual(1, ingest_subtask.output_dataproducts.count())
        self.assertEqual(obs_subtask.output_dataproducts.first(), ingest_subtask.get_transformed_input_dataproduct(ingest_subtask.output_dataproducts.first()))


class SubTasksCreationFromTaskBluePrintCalibrator(unittest.TestCase):

    def test_create_sequence_of_subtask_from_task_blueprint_calibrator_failure(self):
        """
        Create multiple subtasks from a task blueprint when task is a calibrator
        Check that exception should occur due too missing related target observation
        """
        task_blueprint = create_task_blueprint_object_for_testing(task_template_name="calibrator observation")
        with self.assertRaises(SubtaskCreationException):
            create_observation_control_subtask_from_task_blueprint(task_blueprint)

    @unittest.skip("JS 2020-09-08: Cannot reproduce SubtaskCreationException. How is this test supposed to work??")
    def test_create_sequence_of_subtask_from_task_blueprint_calibrator(self):
        """
        Create multiple subtasks from a task blueprint when task is a calibrator and is related to task blueprint
        of a target observation
        Check that exception should occur due too missing pointing setting in target observation,
        the calibrator default is AutoSelect=True
        Check NO exception, when AutoSelect=False
        """
        cal_task_blueprint = create_task_blueprint_object_for_testing(task_template_name="calibrator observation")
        target_task_blueprint = create_task_blueprint_object_for_testing()
        create_scheduling_relation_task_blueprint_for_testing(cal_task_blueprint, target_task_blueprint)

        with self.assertRaises(SubtaskCreationException):
            create_observation_control_subtask_from_task_blueprint(cal_task_blueprint)

        cal_task_blueprint.specifications_doc['autoselect'] = False
        cal_task_blueprint.specifications_doc['pointing']['angle1'] = 1.111
        cal_task_blueprint.specifications_doc['pointing']['angle2'] = 2.222
        subtask = create_observation_control_subtask_from_task_blueprint(cal_task_blueprint)
        self.assertEqual("defined", str(subtask.state))
        self.assertEqual("observation control", str(subtask.specifications_template.name))
        self.assertEqual("observation", str(subtask.specifications_template.type))
        self.assertEqual('J2000', subtask.specifications_doc['stations']['analog_pointing']['direction_type'])
        self.assertEqual(1.111, subtask.specifications_doc['stations']['analog_pointing']['angle1'])
        self.assertEqual(2.222, subtask.specifications_doc['stations']['analog_pointing']['angle2'])


class SubTaskCreationFromTaskBlueprintPipelines(unittest.TestCase):

    def test_create_subtask_from_task_blueprint_preprocessing_pipeline(self):
        """
        Test that a preprocessing task blueprint can be turned into a preprocessing pipeline subtask
        """

        # setup
        software_version = '42.1234'
        observation_task_blueprint = create_task_blueprint_object_for_testing()
        pipeline_task_blueprint = create_task_blueprint_object_for_testing(task_template_name="preprocessing pipeline", task_spec={'software_version': software_version})
        create_relation_task_blueprint_object_for_testing(observation_task_blueprint, pipeline_task_blueprint)

        create_observation_control_subtask_from_task_blueprint(observation_task_blueprint)

        # trigger
        subtask = create_preprocessing_subtask_from_task_blueprint(pipeline_task_blueprint)

        # assert
        self.assertEqual("defined", str(subtask.state))
        self.assertEqual("preprocessing pipeline", str(subtask.specifications_template.name))
        self.assertEqual(models.SubtaskType.Choices.PIPELINE.value, str(subtask.specifications_template.type))
        self.assertEqual(subtask.specifications_doc['software_version'], software_version)

    def test_create_subtask_from_task_blueprint_pulsar_pipeline(self):
        """
        Test that ia pulsar task blueprint can be turned into a pulsar pipeline subtask
        """

        # setup
        observation_task_blueprint = create_task_blueprint_object_for_testing()
        pipeline_task_blueprint = create_task_blueprint_object_for_testing(task_template_name="pulsar pipeline")
        create_relation_task_blueprint_object_for_testing(observation_task_blueprint, pipeline_task_blueprint)

        create_observation_control_subtask_from_task_blueprint(observation_task_blueprint)

        # trigger
        subtask = create_pulsar_pipeline_subtask_from_task_blueprint(pipeline_task_blueprint)

        # assert
        self.assertEqual("defined", str(subtask.state))
        self.assertEqual("pulsar pipeline", str(subtask.specifications_template.name))
        self.assertEqual(models.SubtaskType.Choices.PIPELINE.value, str(subtask.specifications_template.type))
        self.assertTrue(subtask.primary)


class SubTaskCreationFromTaskBlueprintIngest(unittest.TestCase):

    def test_create_subtask_from_task_blueprint_ingest(self):
        """
        Test that ingest task blueprint can be turned into a ingest control subtask
        """

        # setup
        ingest_task_blueprint = create_task_blueprint_object_for_testing(task_template_name="ingest")

        # trigger
        subtask = create_ingest_subtask_from_task_blueprint(ingest_task_blueprint)

        # assert
        self.assertEqual("defined", str(subtask.state))
        self.assertEqual("ingest control", str(subtask.specifications_template.name))
        self.assertEqual(models.SubtaskType.Choices.INGEST.value, str(subtask.specifications_template.type))
        self.assertTrue(subtask.primary)


class SubtaskInputSelectionFilteringTest(unittest.TestCase):

    def setUp(self) -> None:
        # make sure we're allowed to schedule
        setting = Setting.objects.get(name='dynamic_scheduling_enabled')
        setting.value = True
        setting.save()

    def test_specifications_doc_meets_selection_doc_returns_true_on_empty_filter(self):
        specs = {}
        selection = {}
        self.assertTrue(specifications_doc_meets_selection_doc(specs, selection))

    def test_specifications_doc_meets_selection_doc_returns_true_when_filter_applies(self):
        # simple selection matches specs
        specs = {'sap': ['target0']}
        selection = {'sap': ['target0']}
        self.assertTrue(specifications_doc_meets_selection_doc(specs, selection))

        # extra specs are ignored
        specs = {'sap': ['target0'], 'not': 'relevant'}
        selection = {'sap': ['target0']}
        self.assertTrue(specifications_doc_meets_selection_doc(specs, selection))

        # complex selection matches specs; multiple keys and values
        specs = {'sap': ['target0'], 'is_relevant': True}
        selection = {'sap': ['target0', 'target1'], 'is_relevant': True}
        self.assertTrue(specifications_doc_meets_selection_doc(specs, selection))

    def test_specifications_doc_meets_selection_doc_returns_true_when_filter_does_not_apply(self):
        # selection mismatches specs
        specs = {'sap': ['target0']}
        selection = {'sap': ['target1']}
        self.assertFalse(specifications_doc_meets_selection_doc(specs, selection))

        # spec only partially selected
        specs = {'sap': ['target0', 'target1']}
        selection = {'sap': ['target1']}
        self.assertFalse(specifications_doc_meets_selection_doc(specs, selection))

        # selection not in specs
        specs = {'sap': ['target0']}
        selection = {'sap': ['target0'], 'is_relevant': True}
        self.assertFalse(specifications_doc_meets_selection_doc(specs, selection))


class SettingTest(unittest.TestCase):

    def test_schedule_observation_subtask_raises_when_flag_is_false(self):
        setting = Setting.objects.get(name='dynamic_scheduling_enabled')
        setting.value = False
        setting.save()
        obs_st = create_subtask_object_for_testing('observation')

        with self.assertRaises(SubtaskSchedulingException):
            schedule_observation_subtask(obs_st)


class SubTaskCreationFromTaskBlueprintBeamformer(unittest.TestCase):
    saps = [{"name": "target1", "subbands": [349, 372],
             "digital_pointing": {"angle1": 0.24, "angle2": 0.25, "direction_type": "J2000", "target": "target1"}},
            {"name": "target2", "subbands": [309, 302],
             "digital_pointing": {"angle1": 0.42, "angle2": 0.52, "direction_type": "J2000", "target": "target2"}}
    ]
    beamformers = [{"name": "beamformer1",
                    "coherent": {"settings": {"stokes": "I", "time_integration_factor": 8, "subbands_per_file": 244,
                                              "channels_per_subband": 8 },
                    "SAPs": [{"name": "target1",
                              "tabs": [{"relative": True,
                                        "pointing": {"direction_type": "J2000", "angle1": 0.1,  "angle2": 0.1, "target": "target1"}},
                                       {"relative": False, "pointing": {"direction_type": "J2000", "angle1": 0.2,  "angle2": 0.2, "target": "target1"}}],
                              "tab_rings": {"count": 8, "width": 0.02},
                              "subbands": {"method": "copy"}},
                             {"name": "target2",
                              "tabs": [{"relative": True,
                                        "pointing": {"direction_type": "J2000", "angle1": 0.1,  "angle2": 0.1, "target": "target2"}},
                                       {"relative": False, "pointing": {"direction_type": "J2000", "angle1": 0.2,  "angle2": 0.2, "target": "target2"}}],
                              "tab_rings": {"count": 7, "width": 0.03},
                              "subbands": {"list":[10,20,30], "method": "copy"}
                              }]},
                    "incoherent": {"settings": {"stokes": "I", "time_integration_factor": 4, "subbands_per_file": 244,
                                                "channels_per_subband": 8 },
                                   "SAPs": [{"name": "target1", "subbands": {"method": "copy"}}]},
                    "station_groups": [ {"stations":["CS001"], "max_nr_missing": 0} ]
                   },
                   {"name": "beamformer2",
                    "flys eye": {"enabled": True, "settings": {"stokes": "I", "time_integration_factor": 2, "subbands_per_file": 122,
                                              "channels_per_subband": 16 }},
                    "station_groups": [ {"stations":["DE609"], "max_nr_missing": 0} ] 
                   }]

    @unittest.skip('TODO: re-enable this test when we start supporting Cobalt2/LMM')
    def test_merge_beamformer_pipelines(self):
        task_blueprint = create_task_blueprint_object_for_testing(task_template_name="beamforming observation")
        task_blueprint.specifications_doc['station_configuration']['SAPs'] = self.saps
        task_blueprint.specifications_doc['beamformer']['pipelines'] = self.beamformers
        subtask = create_observation_control_subtask_from_task_blueprint(task_blueprint)

        # CS and IS pipelines should have been merged
        self.assertEqual(len(subtask.specifications_doc['COBALT']['beamformer']['tab_pipelines']), 1)

        # are the settings copied for both CS and IS? check a setting that differs
        self.assertEqual(subtask.specifications_doc['COBALT']['beamformer']['tab_pipelines'][0]["coherent"]["time_integration_factor"], 8)
        self.assertEqual(subtask.specifications_doc['COBALT']['beamformer']['tab_pipelines'][0]["incoherent"]["time_integration_factor"], 4)

        # are the TABs merged? assume target1 and target2 appear in that order in the result
        self.assertEqual(subtask.specifications_doc['COBALT']['beamformer']['tab_pipelines'][0]["SAPs"][0]["name"], "target1")
        self.assertEqual(subtask.specifications_doc['COBALT']['beamformer']['tab_pipelines'][0]["SAPs"][1]["name"], "target2")

        # is last TAB also incoherent (it should have been added after the coherent TABs)
        self.assertEqual(subtask.specifications_doc['COBALT']['beamformer']['tab_pipelines'][0]["SAPs"][0]["tabs"][0]["coherent"], True)
        self.assertEqual(subtask.specifications_doc['COBALT']['beamformer']['tab_pipelines'][0]["SAPs"][0]["tabs"][-1]["coherent"], False)

        # FE pipeline should still be there
        self.assertEqual(len(subtask.specifications_doc['COBALT']['beamformer']['flyseye_pipelines']), 1)

    @unittest.skip('TODO: re-enable this test when we start supporting Cobalt2/LMM')
    def test_unmergable_beamformer_pipelines(self):
        task_blueprint = create_task_blueprint_object_for_testing(task_template_name="beamforming observation")
        task_blueprint.specifications_doc['station_configuration']['SAPs'] = self.saps
        task_blueprint.specifications_doc['beamformer']['pipelines'] = deepcopy(self.beamformers)
        # make unmergable
        task_blueprint.specifications_doc['beamformer']['pipelines'][0]["incoherent"]["SAPs"][0]["subbands"] = {"list": [349], "method": "select subset" }
        subtask = create_observation_control_subtask_from_task_blueprint(task_blueprint)

        # CS and IS pipelines should NOT have been merged
        self.assertEqual(len(subtask.specifications_doc['COBALT']['beamformer']['tab_pipelines']), 2)

        # are the settings copied for both CS and IS? check a setting that differs
        self.assertEqual(subtask.specifications_doc['COBALT']['beamformer']['tab_pipelines'][0]["coherent"]["time_integration_factor"], 8)
        self.assertEqual(subtask.specifications_doc['COBALT']['beamformer']['tab_pipelines'][1]["incoherent"]["time_integration_factor"], 4)

        # is last TAB also incoherent (it should have been added after the coherent TABs)
        self.assertEqual(subtask.specifications_doc['COBALT']['beamformer']['tab_pipelines'][0]["SAPs"][0]["tabs"][0]["coherent"], True)
        self.assertEqual(subtask.specifications_doc['COBALT']['beamformer']['tab_pipelines'][1]["SAPs"][0]["tabs"][0]["coherent"], False)

        # FE pipeline should still be there
        self.assertEqual(len(subtask.specifications_doc['COBALT']['beamformer']['flyseye_pipelines']), 1)

    # todo: fix and enable test once we have a switch between standalone/add-on
    @unittest.skip('currently we hardcode standalone mode, where this test cannot be expected to fail')
    def test_create_sequence_of_subtask_from_task_blueprint_beamformer_failure(self):
        """
        Create a subtasks from a task blueprint when task is beamformer add-on.
        Assert that this fails without related target observation.
        # todo: mark as add-on when we have a switch
        """
        task_blueprint = create_task_blueprint_object_for_testing(task_template_name="beamforming observation")
        task_blueprint.specifications_doc['station_configuration']['SAPs'] = self.saps
        task_blueprint.specifications_doc['beamformer'] = self.beamformers
        with self.assertRaises(SubtaskCreationException):
            create_observation_control_subtask_from_task_blueprint(task_blueprint)

    def test_get_related_target_sap_by_name(self):
        beamformer_task_blueprint = create_task_blueprint_object_for_testing(task_template_name="beamforming observation")
        beamformer_task_blueprint.specifications_doc['station_configuration']['SAPs'] = self.saps
        beamformer_task_blueprint.specifications_doc['beamformer']['pipelines'] = self.beamformers

        target_task_spec= {'station_configuration':{'SAPs':self.saps}}
        target_task_blueprint = create_task_blueprint_object_for_testing(task_spec=target_task_spec)
        create_scheduling_relation_task_blueprint_for_testing(beamformer_task_blueprint, target_task_blueprint)

        # TODO: If we start using beamforming observations in parallel with target imaging observations, then we need to search for saps in the target imaging obs spec.
        # See git history for an initial implementation.
        # Also see git history for a test 'test_create_sequence_of_subtask_from_task_blueprint_beamformer' where we have a beamforming observation in parallel with target imaging observation
        sap = _get_related_target_sap_by_name(beamformer_task_blueprint, 'target2')
        self.assertEqual(sap, self.saps[1])

    def _assertInPointingAlmostEqual(self, expected_pointing, tab_pointings):
        expected_tab_found = False
        for p in tab_pointings:
            if p['target'] == expected_pointing['target']:
                self.assertAlmostEqual(p['angle1'], expected_pointing['angle1'])
                self.assertAlmostEqual(p['angle2'], expected_pointing['angle2'])
                expected_tab_found = True
        self.assertTrue(expected_tab_found, msg='No pointing with target="%s" found in collection=%s' % (expected_pointing['target'], tab_pointings))

    def test_generate_tab_ring_pointings_returns_correct_pointings(self):

        pointing = {"angle1": 0.11, "angle2": 0.22, "direction_type": "J2000", "target": "target1"}
        tab_rings = {"width": 1, "count": 1}

        # assert center pointing is returned
        tab_pointings = _generate_tab_ring_pointings(pointing, tab_rings)
        expected_pointing = pointing.copy()
        expected_pointing.update({'target': 'TAB 0 around target1'})
        self._assertInPointingAlmostEqual(expected_pointing, tab_pointings)

        # assert correct number of pointings is returned
        self.assertEqual(len(tab_pointings), 1+6)  # center + 1 ring
        tab_rings.update({'count': 3})
        tab_pointings = _generate_tab_ring_pointings(pointing, tab_rings)
        self.assertEqual(len(tab_pointings), 1+6+12+18)  # center + 3 rings

        # assert width is considered
        tab_rings.update({'width': 0.42})
        tab_pointings = _generate_tab_ring_pointings(pointing, tab_rings)
        self.assertEqual(len(tab_pointings), 1+6+12+18)  # center + 3 rings
        expected_pointing = pointing.copy()
        expected_pointing.update({'angle2': pointing['angle2']+tab_rings['width'], 'target': 'TAB 1 around target1'})
        self._assertInPointingAlmostEqual(expected_pointing, tab_pointings)

        # assert a know correct result is returned
        pointing = {"angle1": 0, "angle2": math.pi/2, "target": "target1", "direction_type": "J2000"}
        tab_rings = {"width": math.pi/2, "count": 1}
        tab_pointings = _generate_tab_ring_pointings(pointing, tab_rings)
        expected_pointing = {'angle1': 0, 'angle2': 0, 'target': 'TAB 4 around target1', "direction_type": "J2000"}
        self._assertInPointingAlmostEqual(expected_pointing, tab_pointings)

    def test_project_pointing(self):

        pointing_a = {"angle1": 0.11, "angle2": 0.22, "direction_type": "J2000", "target": "target_a"}
        pointing_b = {"angle1": 0.88, "angle2": 0.66, "direction_type": "J2000", "target": "target_b"}
        pointing_sum = project_pointing(pointing_a, pointing_b)
        self.assertEqual(pointing_sum, {"angle1": 1.0504442942831016, "angle2": 0.8737077041057875, "direction_type": "J2000", "target": "target_a"})

    def test_filter_subbands_filters_correctly(self):
        subbands = [1,3,4,5,10,11,12,13,19,20]

        # copy
        subband_selection = {'method': 'copy'}
        filtered_subbands = _filter_subbands(subbands, subband_selection)
        self.assertEqual(filtered_subbands, subbands)

        # subset
        subband_selection = {'method': 'select subset', 'list': [1,2,3,4,5,6,7,8,9,10]}
        filtered_subbands = _filter_subbands(subbands, subband_selection)
        self.assertEqual(filtered_subbands, [1,3,4,5,10])

        # largest continuous subset
        subband_selection = {'method': 'largest continuous subset'}
        filtered_subbands = _filter_subbands(subbands, subband_selection)
        self.assertEqual(filtered_subbands, [10,11,12,13])

class SubtaskAllowedStateTransitionsTest(unittest.TestCase):
    def test_successful_path(self):
        subtask = models.Subtask.objects.create(**Subtask_test_data())
        for state_value in (SubtaskState.Choices.DEFINING.value,
                            SubtaskState.Choices.DEFINED.value,
                            SubtaskState.Choices.SCHEDULING.value,
                            SubtaskState.Choices.SCHEDULED.value,
                            SubtaskState.Choices.QUEUEING.value,
                            SubtaskState.Choices.QUEUED.value,
                            SubtaskState.Choices.STARTING.value,
                            SubtaskState.Choices.STARTED.value,
                            SubtaskState.Choices.FINISHING.value,
                            SubtaskState.Choices.FINISHED.value):
            subtask.state = SubtaskState.objects.get(value=state_value)
            # no SubtaskIllegalStateTransitionException should be raised upon save. If it is raised, then test fails. No need for asserts.
            subtask.save()

    def test_helper_method_set_subtask_state_following_allowed_transitions_successful_path(self):
        for state_value in (SubtaskState.Choices.DEFINING.value,
                            SubtaskState.Choices.DEFINED.value,
                            SubtaskState.Choices.SCHEDULING.value,
                            SubtaskState.Choices.UNSCHEDULABLE.value,
                            SubtaskState.Choices.SCHEDULED.value,
                            SubtaskState.Choices.QUEUEING.value,
                            SubtaskState.Choices.QUEUED.value,
                            SubtaskState.Choices.STARTING.value,
                            SubtaskState.Choices.STARTED.value,
                            SubtaskState.Choices.FINISHING.value,
                            SubtaskState.Choices.FINISHED.value):
            # start with subtask in defining state each time
            subtask = models.Subtask.objects.create(**Subtask_test_data(state=SubtaskState.objects.get(value=SubtaskState.Choices.DEFINING.value)))
            self.assertEqual(SubtaskState.Choices.DEFINING.value, subtask.state.value)

            set_subtask_state_following_allowed_transitions(subtask, state_value)
            self.assertEqual(state_value, subtask.state.value)

    def test_helper_method_set_subtask_state_following_allowed_transitions_error_path(self):
        for intermediate_state_value in (SubtaskState.Choices.DEFINING.value,
                                         SubtaskState.Choices.SCHEDULING.value,
                                         SubtaskState.Choices.UNSCHEDULING.value,
                                         SubtaskState.Choices.UNSCHEDULABLE.value,
                                         SubtaskState.Choices.QUEUEING.value,
                                         SubtaskState.Choices.STARTING.value,
                                         SubtaskState.Choices.STARTED.value,
                                         SubtaskState.Choices.FINISHING.value):
            # start with subtask in defining state each time
            subtask = models.Subtask.objects.create(**Subtask_test_data(state=SubtaskState.objects.get(value=SubtaskState.Choices.DEFINING.value)))
            self.assertEqual(SubtaskState.Choices.DEFINING.value, subtask.state.value)

            # then go to the requested intermediate state
            set_subtask_state_following_allowed_transitions(subtask, intermediate_state_value)
            self.assertEqual(intermediate_state_value, subtask.state.value)

            # then go to the error state (should be allowed from any of these intermediate states)
            subtask.state = SubtaskState.objects.get(value=SubtaskState.Choices.ERROR.value)
            ERROR_MESSAGE = 'test_helper_method_set_subtask_state_following_allowed_transitions_error_path'
            subtask.error_reason = ERROR_MESSAGE
            subtask.save()
            self.assertEqual(SubtaskState.Choices.ERROR.value, subtask.state.value)
            self.assertEqual(ERROR_MESSAGE, subtask.error_reason)

            # overwriting error reason should not be allowed
            from django.db.utils import InternalError
            with self.assertRaises(InternalError) as context:
                subtask.error_reason = "overwriting error reason"
                subtask.save()

    def test_helper_method_set_subtask_state_following_allowed_transitions_cancel_path(self):
        for desired_end_state_value in (SubtaskState.Choices.CANCELLING.value,SubtaskState.Choices.CANCELLED.value):
            for state_value in (SubtaskState.Choices.DEFINED.value,
                                SubtaskState.Choices.UNSCHEDULABLE.value,
                                SubtaskState.Choices.SCHEDULED.value,
                                SubtaskState.Choices.QUEUED.value,
                                SubtaskState.Choices.STARTED.value):
                # start with subtask in defining state each time
                subtask = models.Subtask.objects.create(**Subtask_test_data(state=SubtaskState.objects.get(value=SubtaskState.Choices.DEFINING.value)))
                self.assertEqual(SubtaskState.Choices.DEFINING.value, subtask.state.value)

                # then go to the requested intermediate state
                set_subtask_state_following_allowed_transitions(subtask, state_value)
                self.assertEqual(state_value, subtask.state.value)

                # then go to the error state (should be allowed from any of these intermediate states)
                set_subtask_state_following_allowed_transitions(subtask, desired_end_state_value)
                self.assertEqual(desired_end_state_value, subtask.state.value)

    def test_helper_method_set_subtask_state_following_allowed_transitions_unscheduling_path(self):
        # start with subtask in defining state
        subtask = models.Subtask.objects.create(**Subtask_test_data(state=SubtaskState.objects.get(value=SubtaskState.Choices.DEFINING.value)))
        self.assertEqual(SubtaskState.Choices.DEFINING.value, subtask.state.value)

        # use helper method to follow the allowed path to 'unscheduling'
        set_subtask_state_following_allowed_transitions(subtask, SubtaskState.Choices.UNSCHEDULING.value)
        self.assertEqual(SubtaskState.Choices.UNSCHEDULING.value, subtask.state.value)

        # check transition path
        state_log = SubtaskStateLog.objects.filter(subtask=subtask).order_by('created_at').all()
        self.assertEqual(SubtaskState.Choices.DEFINING.value, state_log[0].new_state.value)
        self.assertEqual(SubtaskState.Choices.DEFINED.value, state_log[1].new_state.value)
        self.assertEqual(SubtaskState.Choices.SCHEDULING.value, state_log[2].new_state.value)
        self.assertEqual(SubtaskState.Choices.SCHEDULED.value, state_log[3].new_state.value)
        self.assertEqual(SubtaskState.Choices.UNSCHEDULING.value, state_log[4].new_state.value)

    def test_end_states(self):
        '''Check if the end states that we cannot get out of are according to the design'''
        # there should be no state to go to from ERROR
        allowed_states_from_ERROR = SubtaskAllowedStateTransitions.objects.filter(old_state__value=SubtaskState.Choices.ERROR.value)
        self.assertEqual(0, allowed_states_from_ERROR.count())

        # there should be no state to go to from FINISHED
        self.assertEqual(0, SubtaskAllowedStateTransitions.objects.filter(old_state__value=SubtaskState.Choices.FINISHED.value).count())

    def test_illegal_state_transitions(self):
        for state_value in [choice.value for choice in SubtaskState.Choices]:
            # assume helper method set_subtask_state_following_allowed_transitions is working (see other tests above)
            # use it to create subtask in desired initial state
            subtask = models.Subtask.objects.create(**Subtask_test_data(state=SubtaskState.objects.get(value=SubtaskState.Choices.DEFINING.value)))
            subtask = set_subtask_state_following_allowed_transitions(subtask, state_value)
            self.assertEqual(state_value, subtask.state.value)

            # derive the allowed and illegal state transitions states
            allowed_new_states = SubtaskAllowedStateTransitions.allowed_new_states(subtask.state)
            illegal_new_states = SubtaskAllowedStateTransitions.illegal_new_states(subtask.state)
            logger.info("test_illegal_state_transitions: old_state='%s' allowed_new_states=%s illegal_new_states=%s", state_value, [s.value for s in allowed_new_states], [s.value for s in illegal_new_states])

            for illegal_new_state in illegal_new_states:
                subtask.state = illegal_new_state
                try:
                    # check that the SubtaskIllegalStateTransitionException is raise for this illegal new state
                    with self.assertRaises(SubtaskIllegalStateTransitionException):
                        subtask.save()
                except Exception as e:
                    if 'subtask.actual_process_start_time/actual_process_stop_time cannot be set when state=' in str(e) or \
                       'subtask.actual_on_sky_start_time/actual_on_sky_stop_time cannot be set when state=' in str(e) or \
                        'subtask.error_reason may only be set when' in str(e):
                        # this exception is acceptable.
                        # It also prevents a subtask update, but for other good reasons.
                        # with sql triggers you don't know which is executed first.
                        logger.info(e) # yes, info level is ok here
                    else:
                        raise

                # state in database should still be the original
                subtask.refresh_from_db()
                self.assertEqual(state_value, subtask.state.value)

class SubTaskFilterOnRestAPITest(unittest.TestCase):

    def test_filter_for_subtask_type(self):
        # setup
        observation_task_blueprint = create_task_blueprint_object_for_testing(task_template_name="target observation")

        # trigger
        subtask = create_observation_control_subtask_from_task_blueprint(observation_task_blueprint)
        obs_subtask_id = subtask.id

        with tmss_test_env.create_tmss_client() as client:
            subtasks = client.get_subtasks(subtask_type='observation')
            ids = [s['id'] for s in subtasks]
            self.assertIn(obs_subtask_id, ids)

            subtasks = client.get_subtasks(subtask_type='pipeline')
            ids = [s['id'] for s in subtasks]
            self.assertNotIn(obs_subtask_id, ids)


class SubTaskImmutableFieldsTest(unittest.TestCase):
    '''The subtask model has db constraint that make certain fields immutable (sometimes depending on the state).'''
    # TODO: check more/all immutable fields.
    # TODO: how to handle updates from any client in start/stop-times, once they were set? Raise? Or silently ignore?

    def test_specification_doc_only_annotated_in_state_before_scheduled(self):
        '''See bug https://support.astron.nl/jira/browse/TMSS-1850
        Check that the subtask instance only annotates the specification_doc when it is mutable by state
        '''
        # setup, create task and subtask for observation
        task_template = models.TaskTemplate.get_latest('target observation')
        task = models.TaskBlueprint.objects.create(**TaskBlueprint_test_data(specifications_template=task_template))

        subtask_template = models.SubtaskTemplate.get_latest('observation control')
        specifications_doc = subtask_template.get_default_json_document_for_schema()

        subtask = models.Subtask.objects.create(**Subtask_test_data(specifications_doc=specifications_doc, subtask_template=subtask_template, task_blueprint=task, cluster=models.Cluster.objects.get(name="CEP4")))
        filesystem = subtask.cluster.filesystem_set.filter(name__icontains='test').first()
        output_role = models.TaskConnectorType.objects.create(**TaskConnectorType_test_data(datatype='visibilities', dataformat='MeasurementSet'))
        models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask, filesystem, output_role))
        subtask = set_subtask_state_following_allowed_transitions(subtask, 'defined')

        # the bug in TMSS-1850/SDCH-3388 was caused because the $schema was injected with
        # hostname 'scu001.control.lofar' and replaced upon save when starting with 'tmss'lofar.eu'
        # Both hostnames are valid, but they trigger this bug, because the $schema hostname is replaced,
        # which alters the specifications_doc, which is immutable by then.
        # reproduce that scenario here, and replace with some alternate url in the templates schema.
        other_schema_id = 'http://another.host.com/some-schema.json'
        from django.db import connection
        with connection.cursor() as cursor:
            cursor.execute('''UPDATE tmssapp_subtask set specifications_doc = jsonb_set("specifications_doc"::jsonb, '{$schema}', '"%s"')  WHERE id=%d;''' % (other_schema_id, subtask.id))
        subtask.refresh_from_db()
        self.assertEqual(other_schema_id, subtask.specifications_doc['$schema'])


        # now schedule and simulate running the subtask
        # this should cause several save actions on the subtask instance,
        # triggering the changed-immutable-field exception
        from lofar.sas.tmss.tmss.tmssapp.subtasks import schedule_subtask
        subtask = schedule_subtask(subtask, start_time=datetime.utcnow()+timedelta(hours=1))
        set_subtask_state_following_allowed_transitions(subtask, 'started')
        set_subtask_state_following_allowed_transitions(subtask, 'finished')

        # proof that the subtask specifications_doc still refers to the altered $schema id
        # and thus, that the subtask specifications_doc was not modified upon 'running' it
        subtask.refresh_from_db()
        self.assertEqual(other_schema_id, subtask.specifications_doc['$schema'])


if __name__ == "__main__":
    os.environ['TZ'] = 'UTC'
    unittest.main()
