#!/usr/bin/env python3

# Copyright (C) 2018    ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

# $Id:  $

import os
import unittest
import requests

import logging

logger = logging.getLogger(__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

from lofar.common.test_utils import exit_with_skipped_code_if_skip_integration_tests
exit_with_skipped_code_if_skip_integration_tests()

# Do Mandatory setup step:
# use setup/teardown magic for tmss test database, ldap server and django server
# (ignore pycharm unused import statement, python unittests does use at RunTime the tmss_test_environment_unittest_setup module)
from lofar.sas.tmss.test.tmss_test_environment_unittest_setup import *
tmss_test_env.populate_schemas_and_connectors()

from lofar.sas.tmss.test.tmss_test_data_django_models import *

# import and setup rest test data creator
from lofar.sas.tmss.test.tmss_test_data_rest import TMSSRESTTestDataCreator
rest_data_creator = TMSSRESTTestDataCreator(tmss_test_env.django_server.url, (tmss_test_env.ldap_server.dbcreds.user, tmss_test_env.ldap_server.dbcreds.password))

from lofar.sas.tmss.tmss.tmssapp.tasks import *
from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions

from lofar.sas.tmss.tmss.exceptions import SchemaValidationException
from django.db.models import Q


class CreationFromSchedulingUnitDraft(unittest.TestCase):
    """
    From scheduling_unit_draft should test:
    1. create_scheduling_unit_blueprint_from_scheduling_unit_draft(scheduling_unit_draft: models.SchedulingUnitDraft) -> models.SchedulingUnitBlueprint:
    6. create_task_drafts_from_scheduling_unit_draft(scheduling_unit_draft: models.SchedulingUnitDraft) -> [TaskDraft]:
    3. create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft: models.SchedulingUnitDraft) -> models.SchedulingUnitBlueprint:
    """
    def test_create_scheduling_unit_blueprint_from_scheduling_unit_draft(self):
        """
        Create Scheduling Unit Draft
        Check if the name draft (specified) is equal to name blueprint (created)
        Check with REST-call if NO tasks are created
        """
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="IM HBA LoTSS - 2 Beams")
        strategy_template.template['tasks'] = {}
        strategy_template.template['task_relations'] = []
        strategy_template.template['task_scheduling_relations'] = []

        scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(
                                   name="Test Scheduling Unit UC1",
                                   specifications_template=strategy_template.scheduling_unit_template,
                                   observation_strategy_template=strategy_template,
                                   scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()),
                                   ingest_permission_required=False,
                                   piggyback_allowed_tbb=True,
                                   piggyback_allowed_aartfaac=True,
                                   rank=0.25,
                                   interrupts_telescope=False)
        update_task_graph_from_specifications_doc(scheduling_unit_draft, strategy_template.template)

        scheduling_unit_blueprint = create_scheduling_unit_blueprint_from_scheduling_unit_draft(scheduling_unit_draft)
        self.assertEqual(scheduling_unit_draft.name, scheduling_unit_blueprint.draft.name)
        self.assertEqual(0, len(scheduling_unit_blueprint.task_blueprints.all()))
        self.assertEqual(scheduling_unit_draft.ingest_permission_required, scheduling_unit_blueprint.ingest_permission_required)
        self.assertEqual(scheduling_unit_draft.piggyback_allowed_tbb, scheduling_unit_blueprint.piggyback_allowed_tbb)
        self.assertEqual(scheduling_unit_draft.piggyback_allowed_aartfaac, scheduling_unit_blueprint.piggyback_allowed_aartfaac)
        self.assertEqual(scheduling_unit_draft.rank, scheduling_unit_blueprint.rank)
        self.assertEqual(scheduling_unit_draft.interrupts_telescope, scheduling_unit_blueprint.interrupts_telescope)

    def test_create_task_drafts_from_scheduling_unit_draft(self):
        """
        Create Scheduling Unit Draft (with empty specifications_doc)
        Check if NO tasks are created
        Check with REST-call if NO tasks are created
        """
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="IM HBA LoTSS - 2 Beams")
        strategy_template.template['tasks'] = {}
        strategy_template.template['task_relations'] = []
        strategy_template.template['task_scheduling_relations'] = []

        scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(
                                   name="Test Scheduling Unit UC1",
                                   specifications_template=strategy_template.scheduling_unit_template,
                                   observation_strategy_template=strategy_template,
                                   scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))

        update_task_graph_from_specifications_doc(scheduling_unit_draft, strategy_template.template)

        scheduling_unit_draft.refresh_from_db()
        task_drafts = scheduling_unit_draft.task_drafts.all()
        self.assertEqual(0, len(task_drafts))
        self.assertEqual(0, len(task_drafts))

    def test_create_task_drafts_from_scheduling_unit_draft_with_UC1_requirements(self):
        """
        Create Scheduling Unit Draft with specifications_doc (read from file)
        Create Task Blueprints (only)
        Check if tasks (7) are created
        """
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="IM HBA LoTSS - 2 Beams")

        scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(
                                   name="Test Scheduling Unit UC1",
                                   specifications_template=strategy_template.scheduling_unit_template,
                                   observation_strategy_template=strategy_template,
                                   scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))

        update_task_graph_from_specifications_doc(scheduling_unit_draft, strategy_template.template)

        scheduling_unit_draft.refresh_from_db()
        task_drafts = scheduling_unit_draft.task_drafts.all()
        self.assertEqual(9, len(task_drafts))

    def test_create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(self):
        """
        Create Scheduling Unit Draft with empty task specification
        Check if the name draft (specified) is equal to name blueprint (created)
        Check that a blueprint was created, even though the draft has no tasks
        """
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="IM HBA LoTSS - 2 Beams")
        strategy_template.template['tasks'] = {}
        strategy_template.template['task_relations'] = []
        strategy_template.template['task_scheduling_relations'] = []

        scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(
                                   name="Test Scheduling Unit UC1",
                                   specifications_template=strategy_template.scheduling_unit_template,
                                   observation_strategy_template=strategy_template,
                                   scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()),
                                   ingest_permission_required=True,
                                   piggyback_allowed_tbb=False,
                                   piggyback_allowed_aartfaac=False,
                                   rank=0.19,
                                   interrupts_telescope=False)

        update_task_graph_from_specifications_doc(scheduling_unit_draft, strategy_template.template)
        scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
        self.assertEqual(scheduling_unit_draft.name, scheduling_unit_blueprint.draft.name)
        self.assertEqual(1, len(scheduling_unit_draft.scheduling_unit_blueprints.all()))
        self.assertEqual(scheduling_unit_draft.ingest_permission_required, scheduling_unit_blueprint.ingest_permission_required)
        self.assertEqual(scheduling_unit_draft.piggyback_allowed_tbb, scheduling_unit_blueprint.piggyback_allowed_tbb)
        self.assertEqual(scheduling_unit_draft.piggyback_allowed_aartfaac, scheduling_unit_blueprint.piggyback_allowed_aartfaac)
        self.assertEqual(scheduling_unit_draft.rank, scheduling_unit_blueprint.rank)
        self.assertEqual(scheduling_unit_draft.interrupts_telescope, scheduling_unit_blueprint.interrupts_telescope)


    def test_create_multiple_scheduling_unit_blueprints_from_one_draft(self):
        """
        Test if we can create multiple scheduling_unit_blueprints from one draft.
        """
        # Create a SchedulingUnitDraft from a json specifications_doc with one linked observation and pipeline.
        specifications_template = models.SchedulingUnitTemplate.get_latest(name="scheduling unit")
        specifications_doc = specifications_template.get_default_json_document_for_schema()

        obs_task_template = models.TaskTemplate.get_latest(name="target observation")
        obs_task_doc = obs_task_template.get_default_json_document_for_schema()
        specifications_doc['tasks']['Observation'] = { 'specifications_doc': obs_task_doc, 'specifications_template': {'name': obs_task_template.name } }

        pipe_task_template = models.TaskTemplate.get_latest(name="preprocessing pipeline")
        pipe_task_doc = pipe_task_template.get_default_json_document_for_schema()
        specifications_doc['tasks']['Pipeline'] = { 'specifications_doc': pipe_task_doc, 'specifications_template': {'name': pipe_task_template.name } }

        specifications_doc['task_relations'].append({'producer': 'Observation',
                                                   'consumer': 'Pipeline',
                                                   'input': {'role': models.Role.Choices.ANY.value,
                                                             'datatype': models.Datatype.Choices.VISIBILITIES.value,
                                                             'dataformat': models.Dataformat.Choices.MEASUREMENTSET.value},
                                                   'output': {'role': models.Role.Choices.CORRELATOR.value,
                                                              'datatype': models.Datatype.Choices.VISIBILITIES.value,
                                                              'dataformat': models.Dataformat.Choices.MEASUREMENTSET.value},
                                                   'selection_template': {'name': 'all'} })

        # create the scheduling_unit_draft and its tasks
        scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(
                                   name="Test Scheduling Unit",
                                   specifications_template=specifications_template,
                                   scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))
        update_task_graph_from_specifications_doc(scheduling_unit_draft, specifications_doc)

        # check the results, the number of task_drafts, the relation between them...
        self.assertEqual(2, scheduling_unit_draft.task_drafts.count())
        draft_relations = models.TaskRelationDraft.objects.filter(Q(consumer__scheduling_unit_draft=scheduling_unit_draft) | Q(producer__scheduling_unit_draft=scheduling_unit_draft)).all()
        self.assertEqual(1, draft_relations.count())
        draft_relation = draft_relations[0]

        # create two blueprints from the same draft....
        scheduling_unit_blueprint1 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
        scheduling_unit_blueprint2 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        # ... and check the results.
        # first check that nothing changed in the drafts, the number of task_drafts, the relation between them...
        self.assertEqual(2, scheduling_unit_draft.task_drafts.count())
        draft_relations = models.TaskRelationDraft.objects.filter(Q(consumer__scheduling_unit_draft=scheduling_unit_draft) | Q(producer__scheduling_unit_draft=scheduling_unit_draft)).all()
        self.assertEqual(1, draft_relations.count())

        # then check both blueprints stem from the same draft
        self.assertEqual(scheduling_unit_draft, scheduling_unit_blueprint1.draft)

        # check both blueprints stem from the same draft, check their tasks and relations
        for scheduling_unit_blueprint in (scheduling_unit_blueprint1, scheduling_unit_blueprint2):
            # from the same draft?
            self.assertEqual(scheduling_unit_draft, scheduling_unit_blueprint.draft)

            # correct number of task_blueprints?
            self.assertEqual(2, scheduling_unit_blueprint.task_blueprints.count())

            # do they refer to the correct task_draft?
            for task_blueprint in scheduling_unit_blueprint.task_blueprints.all():
                self.assertEqual(scheduling_unit_draft, task_blueprint.draft.scheduling_unit_draft)
                self.assertEqual(task_blueprint.specifications_template, task_blueprint.draft.specifications_template)

            # one relation per scheduling_unit_blueprint?
            blueprint_relations = models.TaskRelationBlueprint.objects.filter(Q(consumer__scheduling_unit_blueprint=scheduling_unit_blueprint) | Q(producer__scheduling_unit_blueprint=scheduling_unit_blueprint)).all()
            self.assertEqual(1, blueprint_relations.count())


class CreationFromTaskDraft(unittest.TestCase):
    """
    From task draft should test:
     2. create_task_blueprint_from_task_draft(task_draft: models.TaskDraft) -> models.TaskBlueprint:
     5. create_task_blueprint_and_subtasks_from_task_draft(task_draft: models.TaskDraft) -> models.TaskBlueprint:
    """
    @staticmethod
    def create_task_object(task_draft_name, output_pinned):
        """
        Helper function to create a task object for testing
        """
        su_template = models.SchedulingUnitTemplate.get_latest(name='scheduling unit')
        su_draft_data = SchedulingUnitDraft_test_data(name="SU-"+task_draft_name, template=su_template)
        su_draft = models.SchedulingUnitDraft.objects.create(**su_draft_data)

        obs_task_template = models.TaskTemplate.get_latest(name='target observation')
        task_draft_data = TaskDraft_test_data(name=task_draft_name, specifications_template=obs_task_template, output_pinned=output_pinned, scheduling_unit_draft=su_draft)
        models.TaskDraft.objects.create(**task_draft_data)

    def test_create_task_blueprint_and_subtasks(self):
        """
        Create task draft
        Check if the name draft (specified) is equal to name blueprint (created)
        Check with REST-call if 3 subtasks are created and if these subtasks have state value 'defined'
        Check that the task blueprint gets created with same output_pinned flag as the task draft.
        """
        self.create_task_object(task_draft_name="Test Target Observation 1", output_pinned=False)

        task_draft = models.TaskDraft.objects.get(name="Test Target Observation 1")
        task_blueprint = create_task_blueprint_and_subtasks_from_task_draft(task_draft)
        self.assertEqual(task_draft.name, task_blueprint.draft.name)
        self.assertEqual(3, task_blueprint.subtasks.count())
        for subtask in task_blueprint.subtasks.all():
            subtask.state.value == 'defined'
        self.assertEqual(task_draft.output_pinned, task_blueprint.output_pinned)

    def test_create_task_blueprint(self):
        """
        Create task draft
        Check if the task draft name is equal to the task draft name specified in the created task blueprint
        Check with REST-call if NO subtasks are created
        Check that the task blueprint gets created with same output_pinned flag as the task draft.
        """
        self.create_task_object(task_draft_name="Test Target Observation 2", output_pinned=True)

        task_draft = models.TaskDraft.objects.get(name="Test Target Observation 2")
        task_blueprint = create_task_blueprint_from_task_draft(task_draft)
        self.assertEqual(task_draft.name, task_blueprint.draft.name)
        self.assertEqual(0, task_blueprint.subtasks.count())
        self.assertEqual(task_draft.output_pinned, task_blueprint.output_pinned)


class TaskBlueprintStateTest(unittest.TestCase):
    """
    Test the Task Blueprint State which is derived from the SubTask states.
    The result of each possible combination of these states will be checked
    See https://support.astron.nl/confluence/display/TMSS/Specification+Flow#SpecificationFlow-TaskBlueprints
    """

    def test_state_with_no_subtasks(self):
        """
        Test the taskblueprint state when subtasks are not instantiated.
        the expected state should be 'defined'
        """
        task_blueprint_data = TaskBlueprint_test_data(name="Task Blueprint No Subtasks")
        task_blueprint = models.TaskBlueprint.objects.create(**task_blueprint_data)

        self.assertEqual("defined", task_blueprint.status.value)

    def test_states_with_one_subtask(self):
        """
        Test the taskblueprint state when only one subtasks is instantiated, a pipeline
        See next tables where every row represents:
            Substate(Pipeline), Expected TaskBlueprint State
        """
        # Loop over multiple test_tables which follow the allowed state subtask state transitions up to the three allowed end states: finished, error and cancelled.
        test_tables = [[
            ("defining",    "defined"),
            ("defined",     "schedulable"),
            ("unschedulable","unschedulable"),
            ("defined",     "schedulable"),
            ("scheduling",  "schedulable"),
            ("unschedulable","unschedulable"),
            ("defined",     "schedulable"),
            ("scheduling",  "schedulable"),
            ("scheduled",   "scheduled"),
            ("queueing",    "queued"),
            ("queued",      "queued"),
            ("starting",    "started"),
            ("started",     "started"),
            ("finishing",   "started"),
            ("finished",    "finished")
        ], [
            ("defining",    "defined"),
            ("error",       "error")
        ], [
            ("defining",    "defined"),
            ("defined",     "schedulable"),
            ("unschedulable","unschedulable"),
            ("error",       "error")
        ], [
            ("defining",    "defined"),
            ("defined",     "schedulable"),
            ("cancelling",  "cancelled"),
            ("cancelled",   "cancelled")
        ], [
            ("defining",    "defined"),
            ("defined",     "schedulable"),
            ("unschedulable","unschedulable"),
            ("cancelling",  "cancelled"),
            ("cancelled",   "cancelled")
        ]]

        for test_table in test_tables:
            # Create taskblueprint
            task_blueprint_data = TaskBlueprint_test_data(name="Task Blueprint With One Subtask")
            task_blueprint = models.TaskBlueprint.objects.create(**task_blueprint_data)
            # Create pipeline subtask related to taskblueprint
            subtask_data = Subtask_test_data(subtask_template=models.SubtaskTemplate.get_latest(name='preprocessing pipeline'), task_blueprint=task_blueprint)
            subtask_pipe = models.Subtask.objects.create(**subtask_data)

            # Do the actual test
            for test_item in test_table:
                state_pipe, expected_task_state = test_item
                logger.info("Expected test result of substate pipeline='%s' should be '%s'" % (state_pipe, expected_task_state))
                subtask_pipe.state = models.SubtaskState.objects.get(value=state_pipe)
                subtask_pipe.save()

                task_blueprint.refresh_from_db()
                self.assertEqual(expected_task_state, task_blueprint.status.value)

    def test_states_with_observation_and_qa_subtask(self):
        """
        Test the taskblueprint state when two subtasks are instantiated, an observation and a QA.
        See next table where every row represents:
            Substate(Obs), Substate(QA), Expected TaskBlueprint State
        """
        test_tables = [[
            ("defining",    "defining",   "defined"),
            ("defining",    "defined",    "defined"),
            ("defined",     "defined",    "schedulable"),
            ("unschedulable", "defined",  "unschedulable"),
            ("defined",       "defined",  "schedulable"),
            ("scheduling",    "defined",  "schedulable"),
            ("unschedulable", "defined",  "unschedulable"),
            ("defined",       "defined",  "schedulable"),
            ("scheduling",  "defined",    "schedulable"),
            ("scheduled",   "defined",    "scheduled"),
            ("queueing",    "defined",    "queued"),
            ("queued",      "defined",    "queued"),
            ("starting",    "defined",    "started"),
            ("started",     "defined",    "started"),
            ("finishing",   "defined",    "observed"),
            ("finished",    "defined",    "observed"),
            ("finished",    "finished",   "finished")
        ], [
            ("cancelling",  "defined",    "cancelled"),
            ("cancelled",   "defined",    "cancelled")
        ], [
            ("unschedulable", "defined",  "unschedulable"),
            ("cancelling",    "defined",  "cancelled"),
            ("cancelled",     "defined",  "cancelled"),
        ], [
            ("error",       "defined",    "error")
        ], [
            ("unschedulable","defined",   "unschedulable"),
            ("error",       "defined",    "error")
        ], [
            # qa finishing/finished should be not observed
            ("defined",     "finishing",  "started"),
            ("defined",     "finished",   "started")
        ], [
            ("scheduled",   "finishing",  "started"),
            ("scheduled",   "finished",   "started")
        ], [
            # error and cancelled/ing
            ("scheduled",   "error",      "error")
        ], [
            ("scheduled",   "cancelling", "cancelled"),
            ("scheduled",   "cancelled",  "cancelled")
        ], [
            ("started",     "error",      "error")
        ], [
            ("started",     "cancelling", "cancelled"),
            ("started",     "cancelled",  "cancelled")
        ], [
            ("finished",    "error",      "error")
        ], [
            ("finished",    "cancelling", "cancelled"),
            ("finished",    "cancelled",  "cancelled")
        ], [
            # cancelled over error
            ("cancelling",  "error",      "cancelled"),
            ("cancelled",   "error",      "cancelled")
        ], [
            ("error",       "cancelling", "cancelled"),
            ("error",       "cancelling", "cancelled")
        ], [
            # qa scheduled
            ("queueing",    "scheduled",  "queued"),
            ("queued",      "scheduled",  "queued"),
            ("starting",    "scheduled",  "started"),
            ("started",     "scheduled",  "started"),
            ("finishing",   "scheduled",  "observed"),
            ("finished",    "scheduled",  "observed")
        ], [
            ("cancelling",  "scheduled",  "cancelled"),
            ("cancelled",   "scheduled",  "cancelled")
        ], [
            ("error",       "scheduled",  "error"),
        ] ]

        for test_table in test_tables:
            # Create taskblueprint
            task_blueprint_data = TaskBlueprint_test_data(name="Task Blueprint With Subtasks")
            task_blueprint = models.TaskBlueprint.objects.create(**task_blueprint_data)
            # Create observation and qa subtask related to taskblueprint
            subtask_data = Subtask_test_data(subtask_template=models.SubtaskTemplate.get_latest(name='observation control'), task_blueprint=task_blueprint, primary=True)
            subtask_obs = models.Subtask.objects.create(**subtask_data)
            subtask_data = Subtask_test_data(subtask_template=models.SubtaskTemplate.get_latest(name='QA file conversion'), task_blueprint=task_blueprint, primary=False)
            subtask_qa = models.Subtask.objects.create(**subtask_data)

            # Do the actual test
            for test_item in test_table:
                state_obs, state_qa, expected_task_state = test_item
                logger.info("Expected test result of substates observation='%s' and qa='%s' should be '%s'" % (state_obs, state_qa, expected_task_state))
                set_subtask_state_following_allowed_transitions(subtask_obs, state_obs)
                set_subtask_state_following_allowed_transitions(subtask_qa, state_qa)
                self.assertEqual(state_obs, subtask_obs.state.value)
                self.assertEqual(state_qa, subtask_qa.state.value)

                task_blueprint.refresh_from_db()
                self.assertEqual(expected_task_state, task_blueprint.status.value)

    def test_states_with_two_observation_and_two_qa_subtasks(self):
        """
        Test the taskblueprint state when four subtasks are instantiated, two observation and two QA.
        See next table where every row represents:
            Substate(Obs1), Substate(Obs2), Substate(QA1), Substate(QA2), Expected TaskBlueprint State
        """
        # Loop over multiple test_tables which follow the allowed state subtask state transitions up to the three allowed end states: finished, error and cancelled.
        test_tables = [[
            ("defined",     "defined",    "defined",    "defined",    "schedulable"),
            ("unschedulable","defined",   "defined",    "defined",    "unschedulable"),
            ("defined",     "defined",    "defined",    "defined",    "schedulable"),
            ("scheduling",  "defined",    "defined",    "defined",    "schedulable"),
            ("unschedulable","defined",   "defined",    "defined",    "unschedulable"),
            ("defined",     "defined",    "defined",    "defined",    "schedulable"),
            ("started",     "defined",    "defined",    "defined",    "started"),
            ("finishing",   "defined",    "defined",    "defined",    "observed"),
            ("finishing",   "started",    "defined",    "defined",    "observed"),
            ("finishing",   "finishing",  "defined",    "defined",    "observed"),
            ("finished",    "finished",   "defined",    "defined",    "observed"),
            ("finished",    "finished",   "scheduled",  "defined",    "observed"),
            ("finished",    "finished",   "finished",   "scheduled",  "observed"),
            ("finished",    "finished",   "finished",   "finished",   "finished")
        ], [
            ("finished",    "finished",   "finished",   "cancelled",  "cancelled"),
        ], [
            ("finished",    "finished",   "finished",   "error",      "error"),
        ], [
            ("error",       "finished",   "finished",   "cancelled",  "cancelled"),
        ]]

        for test_table in test_tables:
            # Create taskblueprint
            task_blueprint_data = TaskBlueprint_test_data(name="Task Blueprint With Subtasks")
            task_blueprint = models.TaskBlueprint.objects.create(**task_blueprint_data)
            # Create observation and qa subtasks related to taskblueprint
            subtask_data = Subtask_test_data(subtask_template=models.SubtaskTemplate.get_latest(name='observation control'), task_blueprint=task_blueprint, primary=True)
            subtask_obs1 = models.Subtask.objects.create(**subtask_data)
            subtask_data['primary'] = False
            subtask_obs2 = models.Subtask.objects.create(**subtask_data)
            subtask_data = Subtask_test_data(subtask_template=models.SubtaskTemplate.get_latest(name='QA file conversion'), task_blueprint=task_blueprint, primary=False)
            subtask_qa1 = models.Subtask.objects.create(**subtask_data)
            subtask_qa2 = models.Subtask.objects.create(**subtask_data)

            # Do the actual test
            for test_item in test_table:
                state_obs1, state_obs2, state_qa1, state_qa2, expected_task_state = test_item
                logger.info("Expected test result of substates observation='%s','%s' and qa='%s','%s' should be '%s'" %
                            (state_obs1, state_obs1, state_qa1, state_qa2, expected_task_state))

                # Set each subtask to its desired stated, always following allowed transitions only
                set_subtask_state_following_allowed_transitions(subtask_obs1, state_obs1)
                set_subtask_state_following_allowed_transitions(subtask_obs2, state_obs2)
                set_subtask_state_following_allowed_transitions(subtask_qa1, state_qa1)
                set_subtask_state_following_allowed_transitions(subtask_qa2, state_qa2)
                task_blueprint.refresh_from_db()

                self.assertEqual(expected_task_state, task_blueprint.status.value)


if __name__ == "__main__":
    os.environ['TZ'] = 'UTC'
    unittest.main()
