#!/usr/bin/env python3

# Copyright (C) 2018    ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

# $Id:  $
import copy
import os
import sys
from time import sleep
from datetime import datetime, timedelta

import logging
import typing

logger = logging.getLogger(__name__)
from lofar.sas.tmss.tmss.tmssapp import models


def assertDataWithUrls(self, data, expected):
    """
    object instances get returned as urls, check that the value is part of that url
    """

    from django.db import models as django_models

    for k, v in expected.items():
        if isinstance(v, django_models.Model):
            v = str(v.pk)
            v = v.replace(' ', '%20')
            err_msg = "The value '%s' (key is %s) is not in expected %s" % (str(v), str(data[k]), k)
            self.assertTrue(str(v) in data[k], err_msg)

        elif isinstance(v, datetime):
            # URL (data[k]) is string but the test_data object (v) is datetime format, convert latter to string format to compare
            self.assertEqual(v.isoformat(), data[k])
        elif isinstance(v, dict):
            # dict's are usually json docs in tmss. The '$schema' property is altered/filled in by the server, which makes it incomparable.
            # so, only compare the non-'$schema' properties.
            v_copy = copy.deepcopy(v)
            data_k_copy = copy.deepcopy(data[k])
            if '$schema' in v_copy:
                del v_copy['$schema']
            if '$schema' in data_k_copy:
                del data_k_copy['$schema']

            self.assertEqual(v_copy, data_k_copy)
        else:
            self.assertEqual(v, data[k])


def assertUrlList(self, url_list, expected_objects):
    """
    object instances get returned as urls, check that the expected projects are in that list
    """

    from django.db import models as django_models
    self.assertEqual(len(url_list), len(expected_objects))
    for v in expected_objects:
        if isinstance(v, django_models.Model):
            v = str(v.pk)
            v = v.replace(' ', '%20')
            self.assertTrue(any(str(v) in myurl for myurl in url_list))
        else:
            raise ValueError('Expected item is not a Django model instance: %s' % v)


def minimal_json_schema(title:str="my title", description:str="my description", id:str="http://localhost:8000/foo.json", properties:dict=None, required:list=None):
    if properties is None:
        properties = { 'foo': {'type': 'string', 'default': 'bar'}}

    if required is None:
        required = list(properties.keys())

    return {"$schema": "http://json-schema.org/draft-06/schema#",
            "$id": id,
            "title": title,
            "description": description,
            "type": "object",
            "properties": properties,
            "required": required,
            "default": {}
            }


def set_subtask_state_following_allowed_transitions(subtask: typing.Union[models.Subtask, int], desired_state_value:str) -> models.Subtask:
    '''helper function to set subtask state to the desired_state_value following allowed transitions
    Please note that this function is meant to be used in unit/intgration tests only to "simulate" subtask going
    from one state to the desired state, and thus preventing repetitive code like set_state A, set state B ... etc'''
    if isinstance(subtask, int):
        # the given subtask is an id. Fetch object.
        subtask = models.Subtask.objects.get(id=subtask)

    # end states that we cannot get out of according to the design
    END_STATE_VALUES = (models.SubtaskState.Choices.FINISHED.value)

    while subtask.state.value != desired_state_value and (subtask.state.value not in END_STATE_VALUES):
        # handle "unsuccessful path" to cancelled/canceling end state
        if desired_state_value in (models.SubtaskState.Choices.CANCELLED.value, models.SubtaskState.Choices.CANCELLING.value) and \
                subtask.state.value not in (models.SubtaskState.Choices.DEFINING.value,
                                            models.SubtaskState.Choices.QUEUEING.value,
                                            models.SubtaskState.Choices.STARTING.value,
                                            models.SubtaskState.Choices.FINISHING.value,
                                            models.SubtaskState.Choices.CANCELLING.value):
            subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.CANCELLING.value)

        # handle "unsuccessful path" to error end state
        elif desired_state_value == models.SubtaskState.Choices.ERROR.value and subtask.state.value in (models.SubtaskState.Choices.DEFINING.value,
                                                                                                 models.SubtaskState.Choices.QUEUEING.value,
                                                                                                 models.SubtaskState.Choices.STARTING.value,
                                                                                                 models.SubtaskState.Choices.FINISHING.value,
                                                                                                 models.SubtaskState.Choices.UNSCHEDULABLE.value,
                                                                                                 models.SubtaskState.Choices.CANCELLING.value):
            subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.ERROR.value)
            subtask.error_reason = 'set_subtask_state_following_allowed_transitions'

        # handle "unsuccessful path" to unschedulable state
        elif desired_state_value == models.SubtaskState.Choices.UNSCHEDULABLE.value and (subtask.state.value == models.SubtaskState.Choices.DEFINED.value or subtask.state.value == models.SubtaskState.Choices.SCHEDULING.value):
            subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.UNSCHEDULABLE.value)
        # handle "partially successfull path" to on hold state
        elif desired_state_value in (models.SubtaskState.Choices.ON_HOLD.value, models.SubtaskState.Choices.RESTARTING.value) and subtask.state.value == models.SubtaskState.Choices.STARTED.value:
            subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.ON_HOLD.value)
        # handle "partially successfull path" to restarting state
        elif desired_state_value in (models.SubtaskState.Choices.RESTARTING.value, models.SubtaskState.Choices.STARTED.value) and subtask.state.value == models.SubtaskState.Choices.ON_HOLD.value:
            subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.RESTARTING.value)
        elif desired_state_value == models.SubtaskState.Choices.STARTED.value and subtask.state.value == models.SubtaskState.Choices.RESTARTING.value:
            subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.STARTING.value)
        # handle reverse path to defined
        elif desired_state_value == models.SubtaskState.Choices.DEFINED.value and subtask.state.value == models.SubtaskState.Choices.UNSCHEDULABLE.value:
            subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.DEFINED.value)

        # handle reverse path to unscheduling
        elif desired_state_value == models.SubtaskState.Choices.UNSCHEDULING.value and subtask.state.value in (models.SubtaskState.Choices.SCHEDULED.value):
            subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.UNSCHEDULING.value)
        else:
            # handle "normal successful path"
            if subtask.state.value == models.SubtaskState.Choices.DEFINING.value:
                subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.DEFINED.value)
            elif subtask.state.value == models.SubtaskState.Choices.DEFINED.value:
                subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.SCHEDULING.value)
            elif subtask.state.value == models.SubtaskState.Choices.SCHEDULING.value:
                subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.SCHEDULED.value)
            elif subtask.state.value == models.SubtaskState.Choices.SCHEDULED.value:
                subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.QUEUEING.value)
            elif subtask.state.value == models.SubtaskState.Choices.QUEUEING.value:
                subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.QUEUED.value)
            elif subtask.state.value == models.SubtaskState.Choices.QUEUED.value:
                subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.STARTING.value)
            elif subtask.state.value == models.SubtaskState.Choices.STARTING.value:
                subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.STARTED.value)
            elif subtask.state.value == models.SubtaskState.Choices.STARTED.value:
                subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.FINISHING.value)
            elif subtask.state.value == models.SubtaskState.Choices.FINISHING.value:
                subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.FINISHED.value)
                # simulate that the actual_process_stop_time "makes sense" from specification view
                if subtask.actual_process_stop_time is None and subtask.actual_process_start_time is not None:
                    subtask.actual_process_stop_time = subtask.actual_process_start_time + (subtask.specified_duration if subtask.specified_duration else subtask.duration)
            elif subtask.state.value == models.SubtaskState.Choices.CANCELLING.value:
                subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.CANCELLED.value)
            elif subtask.state.value == models.SubtaskState.Choices.UNSCHEDULING.value:
                subtask.state = models.SubtaskState.objects.get(value=models.SubtaskState.Choices.DEFINED.value)

        logger.info("transitioning subtask id=%s to state='%s' desired_state='%s'", subtask.id, subtask.state.value, desired_state_value)
        subtask.save()
        subtask.refresh_from_db() # refresh with updated states/times/etc
        # loop, check in while statement at top if we reached the desired state already.

    return subtask


def wait_for_scheduling_unit_blueprint_status(scheduling_unit_blueprint_id: int, expected_status: typing.Union[str,list,tuple], timeout: int=20, poll_interval: float=0.25):
    """wait for the scheduling_unit_blueprint with the given id to have/get the expected status.
    If the expected_status is a string, then wait until the scheduling_unit_blueprint has exactly this expected status (or times out)
    If the expected_status is a list/tuple of strings, then wait until the scheduling_unit_blueprint has exactly one of the expected statuses (or times out)
    Returns the scheduling_unit_blueprint once it has the expected_status.
    """
    start_wait_timestamp = datetime.utcnow()
    expected_statuses = (expected_status,) if isinstance(expected_status, str) else expected_status
    scheduling_unit_blueprint = models.SchedulingUnitBlueprint.objects.get(id=scheduling_unit_blueprint_id)
    while (datetime.utcnow() - start_wait_timestamp).total_seconds() < timeout:
        start_fetch_timestamp = datetime.utcnow()

        if scheduling_unit_blueprint.status.value in expected_statuses:
            logger.info("wait_for_scheduling_unit_blueprint_status id=%s has expected status=%s waited for %.1fs", scheduling_unit_blueprint_id, scheduling_unit_blueprint.status.value, (datetime.utcnow()-start_wait_timestamp).total_seconds())
            return scheduling_unit_blueprint

        # wait a little...
        logger.info("wait_for_scheduling_unit_blueprint_status id=%s status=%s waiting for %s....", scheduling_unit_blueprint_id, scheduling_unit_blueprint.status.value, expected_statuses)
        fetch_elapsed = (datetime.utcnow() - start_fetch_timestamp).total_seconds()
        sleep(max(0, poll_interval-fetch_elapsed))
        # loop, try again
        scheduling_unit_blueprint.refresh_from_db()

    raise TimeoutError("Timeout while waiting for scheduling_unit_blueprint id=%s to get expected status '%s'. Current status: '%s'" % (scheduling_unit_blueprint_id, expected_status, scheduling_unit_blueprint.status.value))


def assert_SUB_statuses(expected_status_value):
    for scheduling_unit_blueprint in models.SchedulingUnitBlueprint.objects.all():
        if scheduling_unit_blueprint.status.value != expected_status_value:
            raise ValueError('SchedulingUnitBlueprint id=%s name=%s has unexpectedly status=%s instead of "%s"' % (scheduling_unit_blueprint.id, scheduling_unit_blueprint.name, scheduling_unit_blueprint.status, expected_status_value))


def disable_scheduling():
    for setting_name in ['dynamic_scheduling_enabled', 'fixed_time_scheduling_enabled']:
        setting = models.Setting.objects.get(name=setting_name)
        setting.value = False
        setting.save()


def enable_scheduling():
    for setting_name in ['dynamic_scheduling_enabled', 'fixed_time_scheduling_enabled']:
        setting = models.Setting.objects.get(name=setting_name)
        setting.value = True
        setting.save()


def wait_for_units_to_get_scheduled_and_simulate_run():
    success = False
    timeout_minutes = 60

    from lofar.sas.tmss.test.test_environment import create_simulator_for_scheduled_units
    with create_simulator_for_scheduled_units():
        loop_start = datetime.utcnow()
        while datetime.utcnow() - loop_start < timedelta(minutes=timeout_minutes):
            # check progress
            num_scheduled = models.SchedulingUnitBlueprint.objects.filter(status__value=models.SchedulingUnitStatus.Choices.SCHEDULED.value).count()
            num_schedulable = models.SchedulingUnitBlueprint.objects.filter(status__value=models.SchedulingUnitStatus.Choices.SCHEDULABLE.value).count()
            num_finished = models.SchedulingUnitBlueprint.objects.filter(status__value=models.SchedulingUnitStatus.Choices.FINISHED.value).count()
            num_done = models.SchedulingUnitBlueprint.objects.filter(status__value__in=models.SchedulingUnitStatus.DONE_STATUS_VALUES).count()
            units_in_unexpected_state = models.SchedulingUnitBlueprint.objects.exclude(status__value__in=(models.SchedulingUnitStatus.Choices.SCHEDULABLE.value,models.SchedulingUnitStatus.Choices.SCHEDULED.value,models.SchedulingUnitStatus.Choices.FINISHED.value)).all()
            if units_in_unexpected_state:
                logger.warning('SU ids=%s have unexpected statuses=%s' % ([su.id for su in units_in_unexpected_state], [su.status.value for su in units_in_unexpected_state]))

            # log progress
            logger.info('Scheduling unit statuses: finished=%s scheduled=%s schedulable=%s' % (num_finished, num_scheduled, num_schedulable))

            # exit when everything has finished, or nothing schedulable anymore
            if num_done == models.SchedulingUnitBlueprint.objects.count():
                success = True
                break

            # sleep a bit and loop again, reporting on (expected) status(es)
            sleep(1)

        if not success:
            raise TimeoutError('Could not schedule everything within %s minutes' % timeout_minutes)


def print_scheduling_results(scheduling_unit_blueprints):
    try:
        columns, rows = os.get_terminal_size(0)
    except:
        columns = 120

    title = 'Scheduling results'
    print('-' * (len(title) + 4))
    print('| %s |' % title)
    print('-' * columns)
    #print('SU_id,target,ra,dec,duration,min_elevation,transit_offset,scheduled_start_time,scheduled_stop_time')
    print('SU_id,target,scheduled_start_time,scheduled_stop_time')
    print('-' * columns)
    for scheduling_unit_blueprint in scheduling_unit_blueprints:
        #print('%s,%s,%s,%s,%s,%s,%s,%s,%s' % (scheduling_unit_blueprint.id, scheduling_unit_blueprint.specifications_doc,0,0,0,scheduling_unit_blueprint.scheduling_constraints_doc,0,scheduling_unit_blueprint.scheduled_start_time,scheduling_unit_blueprint.scheduled_stop_time))
        print('%s,%s,%s,%s' % (scheduling_unit_blueprint.id, scheduling_unit_blueprint.name,scheduling_unit_blueprint.scheduled_start_time.isoformat(),scheduling_unit_blueprint.scheduled_stop_time.isoformat()))
    print('-' * columns)
    sys.stdout.flush()


class TestStats:
    """
    Class for test statistics handling and prints.
    """
    def __init__(self):
        now = datetime.utcnow()
        self.start_test = now
        self.stop_test = now
        self.start_populate = now
        self.stop_populate = now
        self.start_scheduling = now
        self.stop_scheduling = now

    def print_test_stats(self):
        try:
            columns, rows = os.get_terminal_size(0)
        except:
            columns = 120

        title = 'Runtime summary'
        print('-' * (len(title) + 4))
        print('| %s |' % title)
        print('-' * columns)
        print('Overall test duration: %s' % (self.stop_test - self.start_test))
        print('Duration used on population: %s' % (self.stop_populate - self.start_populate))
        print('Duration used on scheduling: %s' % (self.stop_scheduling - self.start_scheduling))
        print('-' * columns)
