import argparse
import datetime
import time
import requests
from http.client import responses
import json
import html
from urllib.parse import quote
from threading import RLock
import os

import urllib3
urllib3.disable_warnings()

import logging
logger = logging.getLogger(__name__)


class TMSSsession(object):

    OPENID = "openid"
    BASICAUTH = "basicauth"

    POST_RETRY_COUNT = 0 # default number of retries (excluding the first normal attempt)
    POST_RETRY_INTERVAL = 10 # default number of seconds between POST retries

    def __init__(self, username, password, host: str='tmss.lofar.eu', port: int=443, authentication_method=OPENID):
        self.session = requests.session()
        self.username = username
        self.password = password
        self.host_url = "https://%s" % (host,) if port==443 else "http://%s:%s" % (host, port)
        self.api_url = "%s/api" % (self.host_url,)
        self.authentication_method = authentication_method
        self._lock = RLock()

    def __enter__(self):
        self.open()

        # return the request session for use within the context
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def open(self):
        '''open the request session and login'''
        with self._lock:
            self.session.__enter__()
            self.session.verify = False
            self.session.headers['Accept'] = 'application/json'

            if self.authentication_method == self.OPENID:
                # get authentication page of OIDC through TMSS redirect
                response = self.session.get(self.api_url.replace('/api', '/oidc/authenticate/'), allow_redirects=True)
                for resp in response.history:
                    logger.info("via %s: %s" % (resp.status_code, resp.url))
                logger.info("got %s: %s" % (response.status_code, response.url))
                # post user credentials to login page, also pass csrf token if present
                if 'csrftoken' in self.session.cookies:
                    # Mozilla OIDC provider
                    csrftoken = self.session.cookies['csrftoken']
                    data = {'username': self.username, 'password': self.password, 'csrfmiddlewaretoken': csrftoken}
                    response = self.session.post(url=response.url, data=data, allow_redirects=True)
                else:
                    # Keycloak
                    content = response.content.decode('utf-8')
                    if 'action' not in content:
                        raise Exception('Could not determine login form action from server response: %s' % content)
                    action = content.split('action="')[1].split('"')[0]
                    data = {'username': self.username, 'password': self.password, 'credentialId': ''}
                    response = self.session.post(url=html.unescape(action), data=data, allow_redirects=True)

                for resp in response.history:
                    logger.info("via %s: %s" % (resp.status_code, resp.url))
                logger.info("got %s: %s" % (response.status_code, response.url))

                # raise when sth went wrong
                if "The username and/or password you specified are not correct" in response.content.decode('utf8') \
                        or "Invalid username or password" in response.content.decode('utf8'):
                    raise ValueError("The username and/or password you specified are not correct")
                if response.status_code != 200:
                    raise ConnectionError(response.content.decode('utf8'))

            if self.authentication_method == self.BASICAUTH:
                self.session.auth = (self.username, self.password)

    def close(self):
        '''close the request session and logout'''
        with self._lock:
            try:
                # logout user
                self.session.get(self.api_url + '/logout/', allow_redirects=True)
                self.session.close()
            except:
                pass
            
    def _log_response(self, response: requests.Response):
        logger.log(level=logging.INFO if response.status_code >= 200 and response.status_code < 300 else logging.WARNING,
                   msg="%s %s %s in %.1fms%s on %s" % (response.request.method.upper(), response.status_code, responses.get(response.status_code),
                                                       response.elapsed.total_seconds()*1000, ' SLOW!' if response.elapsed > datetime.timedelta(seconds=1) else '',
                                                       response.request.url))
            
    def get_full_url_for_path(self, path: str) -> str:
        '''get the full URL for the given path'''
        return '%s/%s' % (self.api_url, quote(path.strip('/'), safe='/:=?-_&'))

    def get_path_as_json_object(self, path: str, params={}, json_data=None) -> object:
        '''get resource at the given path, interpret it as json, and return it as as native object (usually a dict or a list of dicts)'''
        return self.get_url_as_json_object(self.get_full_url_for_path(path=path), params=params, json_data=json_data)

    def get_path_as_string(self, path: str, params={}) -> str:
        '''get resource at the given path, and return it as as plain text'''
        return self.get_url_as_string(self.get_full_url_for_path(path=path), params=params)

    def get_url_as_string(self, full_url: str, params={}, json_data=None) -> str:
        '''get resource at the given full url (including http://<base_url>, interpret it as json, and return it as as plain text'''
        with self._lock:
            response = self.session.get(url=full_url, params=params, json=json_data, timeout=100000)
            self._log_response(response)

            if response.status_code >= 200 and response.status_code < 300:
                result = response.content.decode('utf-8')
                return result

            if response.status_code == 401:
                logger.warning("Unauthorized to get %s - did you login?", full_url)

            # ugly error message parsing
            content = response.text
            try:
                error_msg = content.split('\n')[1] # magic! error message is at 2nd line of response...
            except:
                error_msg= content

            raise Exception("Could not get %s - %s %s - %s" % (full_url, response.status_code, responses.get(response.status_code), error_msg))

    def get_url_as_json_object(self, full_url: str, params={}, json_data=None) -> object:
        '''get resource at the given full url (including http://<base_url>, interpret it as json, and return it as as native object (usually a dict or a list of dicts)'''
        result = self.get_url_as_string(full_url, params=params, json_data=json_data)
        result = json.loads(result)
        if isinstance(result, dict):
            result_object = result.get('results', result) # return the 'results' list if any, or else just the object itself

            if result.get('next'):
                # recurse, get the 'next' url, and return a concatenation of the results
                return result_object + self.get_url_as_json_object(result['next'])
            return result_object
        return result

    def post_to_url_and_get_result_as_as_string(self, full_url: str, json_data:dict=None, retry_count: int=POST_RETRY_COUNT, retry_interval: float=POST_RETRY_INTERVAL) -> str:
        '''post to the given full_url including http://<base_url>, and return the response as plain text
        Try to post, automatically retry 3 times with 10sec interval upon failure.
        '''
        with self._lock:
            attempt_count = retry_count+1
            for attempt_nr in range(attempt_count):
                response = self.session.post(url=full_url, timeout=100000, json=json_data,
                                             headers={'Referer': self.host_url, 'X-CSRFToken': self.session.cookies['csrftoken']})
                self._log_response(response)

                if response.status_code >= 200 and response.status_code < 300:
                    result = response.content.decode('utf-8')
                    return result

                if attempt_nr < retry_count:
                    time.sleep(retry_interval)

            # ugly error message parsing
            content = response.text
            try:
                error_msg = content.split('\n')[1] # magic! error message is at 2nd line of response...
            except:
                error_msg= content

            raise Exception("Could not post to %s - %s %s - %s" % (full_url, response.status_code, responses.get(response.status_code), error_msg))

    def post_to_url_and_get_result_as_json_object(self, full_url: str, json_data:dict=None, retry_count: int=POST_RETRY_COUNT, retry_interval: float=POST_RETRY_INTERVAL) -> object:
        '''post to the given full_url (including http://<base_url>), and return the response as native object (usually a dict or a list of dicts)'''
        result = self.post_to_url_and_get_result_as_as_string(full_url, json_data=json_data, retry_count=retry_count, retry_interval=retry_interval)
        return json.loads(result)

    def post_to_path_and_get_result_as_json_object(self, path: str, json_data:dict=None, retry_count: int=POST_RETRY_COUNT, retry_interval: float=POST_RETRY_INTERVAL) -> object:
        '''post to the given path, and return the response as native object (usually a dict or a list of dicts)'''
        return self.post_to_url_and_get_result_as_json_object(self.get_full_url_for_path(path=path), json_data=json_data, retry_count=retry_count, retry_interval=retry_interval)

    def get_scheduling_unit_observing_strategy_templates(self) -> list:
        '''get a list of all scheduling_unit_observing_strategy_template's as list of dicts'''
        return self.get_path_as_json_object('scheduling_unit_observing_strategy_template')

    def get_scheduling_unit_observing_strategy_template(self, name: str, version: int=None) -> dict:
        '''get the scheduling_unit_observing_strategy_template as dict for the given name (and version)'''
        clauses = {}
        if name is not None:
            clauses["name"] = name
        if version is not None:
            clauses["version"] = version
        else:
            # try to determine the latest version
            if name is not None:
                try:
                    templates = self.get_path_as_json_object('/scheduling_unit_observing_strategy_template', params={'name':name})
                    if templates:
                        templates = sorted(templates, key=lambda t: t['version'])
                        clauses["version"] = templates[-1]['version']
                except Exception as e:
                    # could not determine latest version
                    pass

        result = self.get_path_as_json_object('/scheduling_unit_observing_strategy_template', clauses)
        if isinstance(result, list):
            if len(result) > 1:
                raise ValueError("Found more then one scheduling_unit_observing_strategy_template for clauses: %s" % (clauses,))
            elif len(result) == 1:
                return result[0]
            raise ValueError("Could not find any template of type='scheduling_unit_observing_strategy_template' and clauses='%s'" % (clauses,))
        return result


    def get_trigger_specification_doc_for_scheduling_unit_observing_strategy_template(self, name: str, version: int = None) -> dict:
        '''get a trigger specification document for the scheduling_unit_observing_strategy_template with the given name (and version).
        You can edit this document to your needs, and submit it via method submit_trigger'''
        strategy_template = self.get_scheduling_unit_observing_strategy_template(name, version)
        return self.get_url_as_json_object(strategy_template['url'].rstrip('/') + '/trigger_doc')

    def submit_trigger(self, trigger_doc: dict, retry_count: int = 0, retry_interval: float = POST_RETRY_INTERVAL) -> dict:
        """create a scheduling_unit_draft with the interrupts_telescope flag set.
        When mode=='run' trigger_doc also create schedulable/runnable blueprints and subtasks, and trigger the dynamic scheduler to try to cancel to current observation and put this triggered observation in place.
        Returns the created scheduling_unit_draft/blueprint upon success, or raises.
        Use the method get_trigger_specification_doc_for_scheduling_unit_observing_strategy_template to get a valid trigger json doc, edit it, and sumbit using this method."""
        return self.post_to_path_and_get_result_as_json_object('/submit_trigger', json_data=trigger_doc, retry_count=retry_count, retry_interval=retry_interval)


def main_submit_trigger():
    '''
    '''
    parser = argparse.ArgumentParser(description='TMSS Client for Trigger Submission. Provides authentication handling and convenient of triggered observations of LOFAR from scripts or the command line.')
    parser.add_argument("observing_strategy_template", type=str, help="Which observing strategy template do you want to use to create the scheduling unit from? This template name should've been provided to you.")
    parser.add_argument("pointing_angle1", type=float, help="First angle of the pointing [rad]")
    parser.add_argument("pointing_angle2", type=float, help="Second angle of the pointing [rad]")
    parser.add_argument("--pointing_direction_type", type=str, default="J2000", help="The direction type of the pointing. Defaults to J2000.")
    parser.add_argument("--start_at", type=str, default=None, help="The start time of the observation as ISO string. If specified, a constraint will be added so that the observation must start exactly at this time. Defaults to None.")
    parser.add_argument("--end_before", type=str, default=None, help="The latest end time of the observation as ISO string. If specified, a constraint will be added so that the observation will run at any time before this time. Defaults to None.")
    parser.add_argument("--target_name", type=str, default=None, help="The name of the observation target. Defaults to None.")
    parser.add_argument("duration", type=int, help="The duration of the observation in seconds.")
    parser.add_argument("scheduling_set_id", type=int, help="The ID of the scheduling set the trigger should be placed in. This should've been provided to you.")
    parser.add_argument("--mode", type=str, required=True, help="Which trigger mode? 'test' or 'run'. Specify Whether this is a test submission or should be actually run.")
    parser.add_argument('--username', type=str, help='TMSS account username. Overrides username from run control file if set.')
    parser.add_argument('--password', type=str, help='TMSS account password. Overrides password from run control file if set.')
    parser.add_argument('--tmssrc', type=str, default='~/.tmssrc', help='Path to your run control file with credentials. On a shared system, this should be used to provide credentials instead of the command line for security reasons.')
    args = parser.parse_args()

    # parse credentials...
    try:
        # ...defaults
        credentials = {'host': 'tmss.lofar.eu',
                       'port': 443,
                       'authentication_method': TMSSsession.OPENID}
        try:
            # ...provided from file
            with open(os.path.expanduser(args.tmssrc), 'r') as f:
                for line in f.readlines():
                    key, value = line.split('=')
                    credentials[key.strip()] = int(value.strip()) if key.strip() == 'port' else value.strip()
        except FileNotFoundError as e:
            logger.info("Run control file %s not found. Using defaults and arguments only." % args.tmssrc)
            pass

        # ...provided as arguments
        credentials['username'] = args.username or credentials.get('username', None)
        credentials['password'] = args.password or credentials.get('password', None)

        if credentials['password'] is None or credentials['username'] is None:
            logger.error('Missing TMSS username or password. Please provide them via a run control file or as arguments.')
            exit(1)

        # start a session that handles the login via Keycloak to access the TMSS API
        with TMSSsession(username=credentials['username'], password=credentials['password'], host=credentials['host'], port=credentials['port'], authentication_method=credentials['authentication_method']) as session:
            # get a prepared valid trigger_doc for the requested template
            print("Preparing trigger specifications...")
            try:
                trigger_doc = session.get_trigger_specification_doc_for_scheduling_unit_observing_strategy_template(args.observing_strategy_template)
            except ValueError as e:
                # no such template available... get a list of all templates and print it, so the user can lookup the one that he wants
                print("ERROR:", e)
                print()
                print("Available templates:")
                for template in session.get_scheduling_unit_observing_strategy_templates():
                    print("id=%s name='%s' version=%s" % (template['id'], template['name'], template['version']))
                exit(1)

            # alter it with our parameters
            trigger_doc["name"] = "Trigger - %s - %s" % (args.observing_strategy_template, datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M"))
            trigger_doc["description"] = "Trigger submitted via CLI tools: %s" % args.observing_strategy_template
            trigger_doc["mode"] = args.mode
            trigger_doc["scheduling_set_id"] = args.scheduling_set_id
            overrides = trigger_doc['scheduling_unit_observing_strategy_template']['overrides']

            def dict_search_and_replace(d: dict, key, value):
                '''perform an in-place search-and-replace to replace all items for the given key by the given value'''
                if key in d:
                    d[key] = value

                # recurse over nested items.
                for k in d:
                    if isinstance(d[k], dict):
                        dict_search_and_replace(d[k], key, value)
                    elif isinstance(d[k], list):
                        for i in d[k]:
                            if isinstance(i, dict):
                                dict_search_and_replace(i, key, value)

            # this is a bit flaky, as it assumes that these parameters are available in the overrides dict
            dict_search_and_replace(overrides, 'angle1', args.pointing_angle1)
            dict_search_and_replace(overrides, 'angle2', args.pointing_angle2)
            dict_search_and_replace(overrides, 'duration', args.duration)

            if args.target_name:
                dict_search_and_replace(overrides, 'target', args.target_name)
                dict_search_and_replace(overrides, 'name', args.target_name)

            if args.end_before:
                overrides['scheduling_constraints_doc']['time']['before'] = args.end_before

            if args.start_at:
                overrides['scheduling_constraints_doc']['time']['at'] = args.start_at

            print("Submitting trigger...")
            scheduling_unit = session.submit_trigger(trigger_doc)
            print("Submitted trigger, and created Scheduling Unit:", scheduling_unit['url'].replace('api/scheduling_unit_', 'schedulingunit/view/'))

    except Exception as e:
        print(e)
        exit(1)

if __name__ == "__main__":
    main_submit_trigger()

