#!/usr/bin/env python3

# TMSSBusListener.py
#
# Copyright (C) 2015
# ASTRON (Netherlands Institute for Radio Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.
#
# $Id: TMSSBusListener.py 1580 2015-09-30 14:18:57Z loose $

"""
TMSSBusListener listens on the lofar notification message bus and calls (empty) on<SomeMessage> methods when such a message is received.
Typical usage is to derive your own subclass from TMSSBusListener and implement the specific on<SomeMessage> methods that you are interested in.
"""

from lofar.messaging.messagebus import BusListener, AbstractMessageHandler
from lofar.messaging import DEFAULT_BUSNAME, DEFAULT_BROKER, EventMessage
from lofar.messaging.exceptions import MessageHandlerUnknownSubjectError
from lofar.common.util import waitForInterrupt, single_line_with_single_spaces

from datetime import datetime
from dateutil import parser
import logging
logger = logging.getLogger(__name__)


_TMSS_EVENT_PREFIX_TEMPLATE                          = 'TMSS.Event.%s'
TMSS_SUBTASK_OBJECT_EVENT_PREFIX                     = _TMSS_EVENT_PREFIX_TEMPLATE % 'SubTask.Object'
TMSS_SUBTASK_STATUS_EVENT_PREFIX                     = _TMSS_EVENT_PREFIX_TEMPLATE % 'SubTask.Status'
TMSS_TASKBLUEPRINT_OBJECT_EVENT_PREFIX               = _TMSS_EVENT_PREFIX_TEMPLATE % 'TaskBlueprint.Object'
TMSS_TASKBLUEPRINT_STATUS_EVENT_PREFIX               = _TMSS_EVENT_PREFIX_TEMPLATE % 'TaskBlueprint.Status'
TMSS_TASKDRAFT_OBJECT_EVENT_PREFIX                   = _TMSS_EVENT_PREFIX_TEMPLATE % 'TaskDraft.Object'
TMSS_SCHEDULINGUNITBLUEPRINT_OBJECT_EVENT_PREFIX     = _TMSS_EVENT_PREFIX_TEMPLATE % 'SchedulingUnitBlueprint.Object'
TMSS_SCHEDULINGUNITBLUEPRINT_STATUS_EVENT_PREFIX     = _TMSS_EVENT_PREFIX_TEMPLATE % 'SchedulingUnitBlueprint.Status'
TMSS_SCHEDULINGUNITDRAFT_OBJECT_EVENT_PREFIX         = _TMSS_EVENT_PREFIX_TEMPLATE % 'SchedulingUnitDraft.Object'
TMSS_SCHEDULINGCONSTRAINTSWEIGHTFACTOR_OBJECT_EVENT_PREFIX = _TMSS_EVENT_PREFIX_TEMPLATE % 'SchedulingConstraintsWeightFactor.Object'
TMSS_PROJECT_OBJECT_EVENT_PREFIX                     = _TMSS_EVENT_PREFIX_TEMPLATE % 'Project.Object'
TMSS_PROJECT_STATUS_EVENT_PREFIX                     = _TMSS_EVENT_PREFIX_TEMPLATE % 'Project.Status'
TMSS_PROJECTQUOTAARCHIVELOCATION_OBJECT_EVENT_PREFIX = _TMSS_EVENT_PREFIX_TEMPLATE % 'ProjectQuotaArchiveLocation.Object'
TMSS_PROJECTCYCLES_OBJECT_EVENT_PREFIX               = _TMSS_EVENT_PREFIX_TEMPLATE % 'ProjectCycles.Object'
TMSS_SUBSYSTEM_STATUS_EVENT_PREFIX                   = _TMSS_EVENT_PREFIX_TEMPLATE % 'Subsystem.Status'
TMSS_SETTING_OBJECT_EVENT_PREFIX                     = _TMSS_EVENT_PREFIX_TEMPLATE % 'Setting.Object'
TMSS_RESERVATION_OBJECT_EVENT_PREFIX                 = _TMSS_EVENT_PREFIX_TEMPLATE % 'Reservation.Object'
TMSS_ALL_OBJECT_EVENTS_FILTER                        = _TMSS_EVENT_PREFIX_TEMPLATE % '.*.Object.#'
TMSS_ALL_STATUS_EVENTS_FILTER                        = _TMSS_EVENT_PREFIX_TEMPLATE % '.*.Status.#'
TMSS_ALL_EVENTS_FILTER                               = _TMSS_EVENT_PREFIX_TEMPLATE % '#'


class TMSSEventMessageHandler(AbstractMessageHandler):
    '''
    Base-type messagehandler for handling all TMSS event messages.
    Typical usage is to derive your own subclass from TMSSEventMessageHandler and implement the specific on<SomeMessage> methods that you are interested in.
    '''

    def __init__(self, log_event_messages: bool=False) -> None:
        self.log_event_messages = log_event_messages
        super().__init__()


    def handle_message(self, msg: EventMessage):
        if not isinstance(msg, EventMessage):
            raise ValueError("%s: Ignoring non-EventMessage: %s" % (self.__class__.__name__, msg))

        stripped_subject = msg.subject.replace(_TMSS_EVENT_PREFIX_TEMPLATE%('',), '')

        if self.log_event_messages:
            logger.info("%s %s: %s" % (self.__class__.__name__, stripped_subject, single_line_with_single_spaces(msg.content)))

        # sorry, very big if/elif/else tree.
        # it just maps all possible event subjects for all possible objects and statuses onto handler methods.
        if stripped_subject == 'SubTask.Object.Created':
            self.onSubTaskCreated(**msg.content)
        elif stripped_subject == 'SubTask.Object.Updated':
            self.onSubTaskUpdated(**msg.content)
        elif stripped_subject == 'SubTask.Object.Deleted':
            self.onSubTaskDeleted(**msg.content)
        elif stripped_subject == 'TaskBlueprint.Object.Created':
            self.onTaskBlueprintCreated(**msg.content)
        elif stripped_subject == 'TaskBlueprint.Object.Updated':
            self.onTaskBlueprintUpdated(**msg.content)
        elif stripped_subject == 'TaskBlueprint.Object.Deleted':
            self.onTaskBlueprintDeleted(**msg.content)
        elif stripped_subject == 'TaskDraft.Object.Created':
            self.onTaskDraftCreated(**msg.content)
        elif stripped_subject == 'TaskDraft.Object.Updated':
            self.onTaskDraftUpdated(**msg.content)
        elif stripped_subject == 'TaskDraft.Object.Deleted':
            self.onTaskDraftDeleted(**msg.content)
        elif stripped_subject == 'SchedulingUnitBlueprint.Object.Created':
            self.onSchedulingUnitBlueprintCreated(**msg.content)
        elif stripped_subject == 'SchedulingUnitBlueprint.Object.Updated':
            self.onSchedulingUnitBlueprintUpdated(**msg.content)
        elif stripped_subject == 'SchedulingUnitBlueprint.Object.Deleted':
            self.onSchedulingUnitBlueprintDeleted(**msg.content)
        elif stripped_subject == 'SchedulingUnitDraft.Object.Created':
            self.onSchedulingUnitDraftCreated(**msg.content)
        elif stripped_subject == 'SchedulingUnitDraft.Object.Updated':
            self.onSchedulingUnitDraftUpdated(**msg.content)
        elif stripped_subject == 'SchedulingUnitDraft.Object.Deleted':
            self.onSchedulingUnitDraftDeleted(**msg.content)
        elif stripped_subject.startswith('SubTask.Status.'):
            self.onSubTaskStatusChanged(**msg.content)
        elif stripped_subject.startswith('TaskBlueprint.Status.'):
            self.onTaskBlueprintStatusChanged(**msg.content)
        elif stripped_subject.startswith('SchedulingUnitBlueprint.Status.'):
            self.onSchedulingUnitBlueprintStatusChanged(**msg.content)
        elif stripped_subject == 'SchedulingConstraintsWeightFactor.Object.Updated':
            self.onSchedulingConstraintsWeightFactorUpdated(**msg.content)
        elif stripped_subject == 'Setting.Object.Updated':
            self.onSettingUpdated(**msg.content)
        elif stripped_subject == 'Subsystem.Status.Updated':
            self.onSubsystemStatusUpdated(**msg.content)
        elif stripped_subject == 'Project.Object.Created':
            self.onProjectCreated(**msg.content)
        elif stripped_subject == 'Project.Object.Updated':
            self.onProjectUpdated(**msg.content)
        elif stripped_subject.startswith('Project.Status.'):
            self.onProjectStatusUpdated(**msg.content)
        elif stripped_subject == 'Project.Object.Deleted':
            self.onProjectDeleted(**msg.content)
        elif stripped_subject == 'Project.Object.Rank.Updated':
            self.onProjectRankUpdated(**msg.content)
        elif stripped_subject == 'ProjectQuotaArchiveLocation.Object.Created':
            self.onProjectQuotaArchiveLocationCreated(**msg.content)
        elif stripped_subject == 'ProjectQuotaArchiveLocation.Object.Updated':
            self.onProjectQuotaArchiveLocationUpdated(**msg.content)
        elif stripped_subject == 'ProjectQuotaArchiveLocation.Object.Deleted':
            self.onProjectQuotaArchiveLocationDeleted(**msg.content)
        elif stripped_subject == 'ProjectCycles.Object.Created':
            self.onProjectCyclesCreated(**msg.content)
        elif stripped_subject == 'ProjectCycles.Object.Updated':
            self.onProjectCyclesUpdated(**msg.content)
        elif stripped_subject == 'ProjectCycles.Object.Deleted':
            self.onProjectCyclesDeleted(**msg.content)
        elif stripped_subject == 'SchedulingUnitBlueprint.Object.Constraints.Updated':
            self.onSchedulingUnitBlueprintConstraintsUpdated(**msg.content)
        elif stripped_subject == 'SchedulingUnitBlueprint.Object.OutputPinningUpdated':
            self.onSchedulingUnitBlueprintOutputPinningUpdated(**msg.content)
        elif stripped_subject == 'SchedulingUnitBlueprint.Object.Rank.Updated':
            self.onSchedulingUnitBlueprintRankUpdated(**msg.content)
        elif stripped_subject == 'SchedulingUnitBlueprint.Object.PriorityQueue.Updated':
            self.onSchedulingUnitBlueprintPriorityQueueUpdated(**msg.content)
        elif stripped_subject == 'SchedulingUnitBlueprint.Object.IngestPermissionGranted':
            self.onSchedulingUnitBlueprintIngestPermissionGranted(id=msg.content['id'],
                                                                  ingest_permission_granted_since=parser.parse(msg.content['ingest_permission_granted_since'], ignoretz=True))
        elif stripped_subject == 'TaskBlueprint.Object.OutputPinningUpdated':
            self.onTaskBlueprintOutputPinningUpdated(**msg.content)
        elif stripped_subject == 'Reservation.Object.Created':
            self.onReservationCreated(**msg.content)
        elif stripped_subject == 'Reservation.Object.Updated':
            self.onReservationUpdated(**msg.content)
        elif stripped_subject == 'Reservation.Object.Deleted':
            self.onReservationDeleted(**msg.content)
        else:
            raise MessageHandlerUnknownSubjectError("TMSSBusListener.handleMessage: unknown subject: %s" %  msg.subject)


    def onSubTaskStatusChanged(self, id: int, status:str):
        '''onSubTaskStatusChanged is called upon receiving a SubTask.Status.* message, which is sent when a SubTasks changes status.
        :param id: the TMSS id of the SubTask
        :param status: the new status of the SubTask
        '''
        pass

    def onTaskBlueprintStatusChanged(self, id: int, status:str):
        '''onTaskBlueprintStatusChanged is called upon receiving a TaskStatus.Choices.* message, which is sent when a TaskBlueprint changes status.
        :param id: the TMSS id of the TaskBlueprint
        :param status: the new status of the TaskBlueprint
        '''
        pass

    def onSchedulingUnitBlueprintStatusChanged(self, id: int, status:str):
        '''onSchedulingUnitBlueprintStatusChanged is called upon receiving a SchedulingUnitStatus.Choices.* message, which is sent when a SchedulingUnitBlueprints changes status.
        :param id: the TMSS id of the SchedulingUnitBlueprint
        :param status: the new status of the SchedulingUnitBlueprint
        '''
        pass

    def onSchedulingConstraintsWeightFactorUpdated(self, id: int):
        '''onSchedulingConstraintsWeightFactorUpdated is called upon receiving a SchedulingConstraintsWeightFactor.Object.Updated message, which is sent when a SchedulingConstraintsWeightFactor object updates.
        :param id: the TMSS id of the SchedulingConstraintsWeightFactor
        '''
        pass

    def onSubTaskCreated(self, id: int):
        '''onSubTaskCreated is called upon receiving a SubTask.Object.Created message, which is sent when a SubTasks was created.
        :param id: the TMSS id of the SubTask
        '''
        pass

    def onSubTaskUpdated(self, id: int):
        '''onSubTaskUpdated is called upon receiving a SubTask.Object.Updated message, which is sent when a SubTasks was created.
        :param id: the TMSS id of the SubTask
        '''
        pass

    def onSubTaskDeleted(self, id: int):
        '''onSubTaskDeleted is called upon receiving a SubTask.Object.Deleted message, which is sent when a SubTasks was created.
        :param id: the TMSS id of the SubTask
        '''
        pass

    def onTaskDraftCreated(self, id: int):
        '''onTaskDraftCreated is called upon receiving a TaskDraft.Object.Created message, which is sent when a TaskDrafts was created.
        :param id: the TMSS id of the TaskDraft
        '''
        pass

    def onTaskDraftUpdated(self, id: int):
        '''onTaskDraftUpdated is called upon receiving a TaskDraft.Object.Updated message, which is sent when a TaskDrafts was created.
        :param id: the TMSS id of the TaskDraft
        '''
        pass

    def onTaskDraftDeleted(self, id: int):
        '''onTaskDraftDeleted is called upon receiving a TaskDraft.Object.Deleted message, which is sent when a TaskDrafts was created.
        :param id: the TMSS id of the TaskDraft
        '''
        pass

    def onTaskBlueprintCreated(self, id: int):
        '''onTaskBlueprintCreated is called upon receiving a TaskBlueprint.Object.Created message, which is sent when a TaskBlueprints was created.
        :param id: the TMSS id of the TaskBlueprint
        '''
        pass

    def onTaskBlueprintUpdated(self, id: int):
        '''onTaskBlueprintUpdated is called upon receiving a TaskBlueprint.Object.Updated message, which is sent when a TaskBlueprints was created.
        :param id: the TMSS id of the TaskBlueprint
        '''
        pass

    def onTaskBlueprintDeleted(self, id: int):
        '''onTaskBlueprintDeleted is called upon receiving a TaskBlueprint.Object.Deleted message, which is sent when a TaskBlueprints was created.
        :param id: the TMSS id of the TaskBlueprint
        '''
        pass

    def onSchedulingUnitDraftCreated(self, id: int):
        '''onSchedulingUnitDraftCreated is called upon receiving a SchedulingUnitDraft.Object.Created message, which is sent when a SchedulingUnitDrafts was created.
        :param id: the TMSS id of the SchedulingUnitDraft
        '''
        pass

    def onSchedulingUnitDraftUpdated(self, id: int):
        '''onSchedulingUnitDraftUpdated is called upon receiving a SchedulingUnitDraft.Object.Updated message, which is sent when a SchedulingUnitDrafts was created.
        :param id: the TMSS id of the SchedulingUnitDraft
        '''
        pass

    def onSchedulingUnitBlueprintConstraintsUpdated(self, id: int, scheduling_constraints_doc: dict):
        '''onSchedulingUnitBlueprintConstraintsUpdated is called upon receiving a SchedulingUnitDraft.Object.Constraints.Updated message, which is sent when a the constraints on a SchedulingUnitDrafts were updated.
        :param id: the TMSS id of the SchedulingUnitDraft
        '''
        pass

    def onSchedulingUnitBlueprintRankUpdated(self, id: int, rank: float):
        '''onSchedulingUnitBlueprintRankUpdated is called upon receiving a SchedulingUnitBlueprint.Object.Rank.Updated message, which is sent when a the rank on a SchedulingUnitDrafts was updated.
        :param id: the TMSS id of the SchedulingUnitBlueprint
        '''
        pass

    def onSchedulingUnitBlueprintPriorityQueueUpdated(self, id: int, priority_queue: str):
        '''onSchedulingUnitBlueprintPriorityQueueUpdated is called upon receiving a SchedulingUnitBlueprint.Object.PriorityQueue.Updated message, which is sent when a the priority_queue on a SchedulingUnitDrafts was updated.
        :param id: the TMSS id of the SchedulingUnitBlueprint
        '''
        pass

    def onSchedulingUnitDraftDeleted(self, id: int):
        '''onSchedulingUnitDraftDeleted is called upon receiving a SchedulingUnitDraft.Object.Deleted message, which is sent when a SchedulingUnitDrafts was created.
        :param id: the TMSS id of the SchedulingUnitDraft
        '''
        pass

    def onSchedulingUnitBlueprintCreated(self, id: int):
        '''onSchedulingUnitBlueprintCreated is called upon receiving a SchedulingUnitBlueprint.Object.Created message, which is sent when a SchedulingUnitBlueprints was created.
        :param id: the TMSS id of the SchedulingUnitBlueprint
        '''
        pass

    def onSchedulingUnitBlueprintUpdated(self, id: int, **kwargs):
        '''onSchedulingUnitBlueprintUpdated is called upon receiving a SchedulingUnitBlueprint.Object.Updated message, which is sent when a SchedulingUnitBlueprints was created.
        :param id: the TMSS id of the SchedulingUnitBlueprint
        '''
        pass

    def onSchedulingUnitBlueprintOutputPinningUpdated(self, id: int, output_pinned: bool):
        pass

    def onSchedulingUnitBlueprintDeleted(self, id: int):
        '''onSchedulingUnitBlueprintDeleted is called upon receiving a SchedulingUnitBlueprint.Object.Deleted message, which is sent when a SchedulingUnitBlueprints was created.
        :param id: the TMSS id of the SchedulingUnitBlueprint
        '''
        pass

    def onProjectCreated(self, name: str):
        '''onProjectCreated is called upon receiving a Project.Object.Created message, which is sent when a Project was created.
        '''
        pass

    def onProjectUpdated(self, name: str):
        '''onProjectUpdated is called upon receiving a Project.Object.Updated message, which is sent when a Project was created.
        '''
        pass

    def onProjectStatusUpdated(self, name: str, status: str):
        '''onProjectStatusUpdated is called upon receiving a Project.Status.* message, which is sent when a Project was assigned a new status.
        '''
        pass

    def onProjectRankUpdated(self, name: str, rank: float):
        '''onProjectRankUpdated is called upon receiving a Project.Rank.Updated message, which is sent when a Project was assigned a new rank.
        '''
        pass

    def onProjectDeleted(self, name: str):
        '''onProjectDeleted is called upon receiving a Project.Object.Deleted message, which is sent when a Project was created.
        '''
        pass

    def onProjectQuotaArchiveLocationCreated(self, id: int):
        '''onProjectQuotaArchiveLocationCreated is called upon receiving a ProjectQuotaArchiveLocation.Object.Created message, which is sent when a ProjectQuotaArchiveLocation was created.
        :param id: the TMSS id of the ProjectQuotaArchiveLocation
        '''
        pass

    def onProjectQuotaArchiveLocationUpdated(self, id: int):
        '''onProjectQuotaArchiveLocationUpdated is called upon receiving a ProjectQuotaArchiveLocation.Object.Updated message, which is sent when a ProjectQuotaArchiveLocation was created.
        :param id: the TMSS id of the ProjectQuotaArchiveLocation
        '''
        pass

    def onProjectQuotaArchiveLocationDeleted(self, id: int):
        '''onProjectQuotaArchiveLocationDeleted is called upon receiving a ProjectQuotaArchiveLocation.Object.Deleted message, which is sent when a ProjectQuotaArchiveLocation was created.
        :param id: the TMSS id of the ProjectQuotaArchiveLocation
        '''
        pass

    def onProjectCyclesCreated(self, id: int):
        '''onProjectCyclesCreated is called upon receiving a ProjectCycles.Object.Created message, which is sent when a ProjectCycles was created.
        :param id: the TMSS id of the ProjectCycles
        '''
        pass

    def onProjectCyclesUpdated(self, id: int):
        '''onProjectCyclesUpdated is called upon receiving a ProjectCycles.Object.Updated message, which is sent when a ProjectCycles was created.
        :param id: the TMSS id of the ProjectCycles
        '''
        pass

    def onProjectCyclesDeleted(self, id: int):
        '''onProjectCyclesDeleted is called upon receiving a ProjectCycles.Object.Deleted message, which is sent when a ProjectCycles was created.
        :param id: the TMSS id of the ProjectCycles
        '''
        pass

    def onSettingUpdated(self, name: str, value):
        '''onSettingUpdated is called upon receiving a Setting.Object.Updated message, which is sent when a Setting was updated.
        :param name: the name of the Setting
        '''
        pass


    def onSubsystemStatusUpdated(self, name: str, status: str):
        '''onSubsystemStatusUpdated is called upon receiving a Subsystem.Status.Updated message, which is sent when a Subsystem's status was updated.
        :param name: the name of the subsystem
        :param status: the new status of the subsystem
        '''
        pass


    def onSchedulingUnitBlueprintIngestPermissionGranted(self, id: int, ingest_permission_granted_since: datetime):
        '''onSchedulingUnitBlueprintIngestPermissionGranted is called upon receiving a SchedulingUnitBlueprint.Object.IngestPermissionGranted message, usually as a result of setting the permissing in the database via the QA Workflow.
        :param id: the TMSS id of the SchedulingUnitBlueprint
        :param ingest_permission_granted_since: the timestamp when the permission was granted
        '''
        pass
    
    def onTaskBlueprintOutputPinningUpdated(self, id: int, output_pinned: bool):
        '''onTaskBlueprintOutputPinningUpdated is called upon receiving a TaskBlueprint.Object.OutputPinningUpdated message, usually as a result of a change on the TaskBlueprint output_pinned field.
        :param id: the TMSS id of the TaskBlueprint
        :param output_pinned: True if the output of this task is pinned to disk, that is, forbidden to be removed.
        '''
        pass

    def onReservationCreated(self, id: int):
        '''onReservationCreated is called upon receiving a Reservation.Object.Created message, which is sent when a Reservation was created.
        '''
        pass

    def onReservationUpdated(self, id: int):
        '''onReservationUpdated is called upon receiving a Reservation.Object.Updated message, which is sent when a Reservation was created.
        '''
        pass

    def onReservationDeleted(self, id: int, start_time, stop_time):
        '''onReservationDeleted is called upon receiving a Reservation.Object.Deleted message, which is sent when a Reservation was created.
        '''
        pass


class TMSSBusListener(BusListener):
    def __init__(self,
                 handler_type: TMSSEventMessageHandler.__class__ = TMSSEventMessageHandler,
                 handler_kwargs: dict = None,
                 exchange: str = DEFAULT_BUSNAME,
                 routing_key: str = TMSS_ALL_EVENTS_FILTER,
                 num_threads: int = 1,
                 broker: str = DEFAULT_BROKER,
                 queue_name_postfix: str = None):
        """
        TMSSSubTaskBusListener listens on the lofar notification message bus and calls on<SomeMessage> methods in the TMSSEventMessageHandler when such a message is received.
        Typical usage is to derive your own subclass from TMSSEventMessageHandler and implement the specific on<SomeMessage> methods that you are interested in.
        """
        if not issubclass(handler_type, TMSSEventMessageHandler):
            raise TypeError("handler_type should be a TMSSEventMessageHandler subclass")

        super().__init__(handler_type, handler_kwargs, exchange, routing_key, num_threads, broker, queue_name_postfix=queue_name_postfix)


if __name__ == '__main__':
    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

    class ExampleTMSSEventMessageHandler(TMSSEventMessageHandler):
        def onSubTaskStatusChanged(self, id: int, status:str):
            logger.debug("MyTMSSEventMessageHandler.onSubTaskStatusChanged(id=%s, status=%s)", id, status)

    from lofar.messaging.messagebus import BusListenerJanitor
    with BusListenerJanitor(TMSSBusListener(handler_type=ExampleTMSSEventMessageHandler)):
        waitForInterrupt()

