import AuthService from "../services/auth.service";
import AuthStore from "./auth.store";
import PermissionStackUtil from './permission.stack.handler';
import axios from "axios"
import {faro} from  '@grafana/faro-react';
/**
 * Global functions to authenticate user and get user details from browser local storage.
 */
const Auth = {
    /** To check if user already logged in 
     * - First it will check the API token is available in the browser local storage
     * - If not then check the Keycloak authentication
     * - Return false if no authentication found
     */
    isDebugLoggin: false,
    isTokenCheckInProgress: false,
    prevalidatedToken: "abcd-abcd",
    callCounter: 0,
    validateExistingToken: async (user) => {

        const loginType = localStorage.getItem("loginType");

        if (Auth.isDebugLoggin) console.log("  \\ validateExistingToken:", user, loginType)

        // Check if token validation is already in progress
        if (Auth.isTokenCheckInProgress) {
            // Another part of the application is already checking the token,
            // so wait for it to complete before continuing.
            await new Promise(resolve => {
                const intervalId = setInterval(() => {
                    if (!Auth.isTokenCheckInProgress) {
                        clearInterval(intervalId);
                        resolve();
                    }
                }, 100);
            });
        }
        Auth.isTokenCheckInProgress = true

        // another thread might have fixed this :)
        if (Auth.prevalidatedToken === user.token) {
            if (Auth.isDebugLoggin) {
                console.log("//  isAuthenticated : No Need anymore  to Check Token !")
                Auth.isTokenCheckInProgress = false
                return true
            }
        }
        if (loginType === "Keycloak") {
            if (Auth.isDebugLoggin) console.log("  - validateExistingToken: Checking for KeyCloak  State", user?.token)
            const currentAuthenticationState = await AuthService.getKeycloakAuthState();
            if (currentAuthenticationState?.is_authenticated) {
                if (Auth.isDebugLoggin) console.log("  // User was a valid KeyCloak user ", currentAuthenticationState)
                Auth.prevalidatedToken = user.token
                Auth.isTokenCheckInProgress = false
                 SetUserFromKeyCloak(currentAuthenticationState);
                
                return true; // we are still having a valid session.
            }
        } else if (loginType === "Application") {
            if (Auth.isDebugLoggin) console.log("  - validateExistingToken: Checking for Token State for Application ", user?.token)
            let currentAuthenticationState = await AuthService.isValidAuthenticationState(user?.name)
            if (currentAuthenticationState?.websocket_token_valid) {
                Auth.prevalidatedToken = user.token
                if (Auth.isDebugLoggin) console.log("  - validateExistingToken: Token Validated ", Auth.prevalidatedToken, currentAuthenticationState)
                Auth.isTokenCheckInProgress = false
                return true
            }
        }
        if (Auth.isDebugLoggin) console.log("  / validateExistingToken: Failed to Validate Token ")
        Auth.isTokenCheckInProgress = false
        Auth.prevalidatedToken = "abcd-abcd"
        return false
    },
    isAuthenticated: async () => {
        let user = localStorage.getItem("user");
        if (Auth.isDebugLoggin) console.log("\\ isAuthenticated User From localstorage:", user, Auth.prevalidatedToken, user?.token)

        // by adding the  user.authenticator, we reinforce all users to revalidate once, since it was missing in the previous version of the localstorage object
        if (user) { // this is the cached user from the system, could be either LDAP or Keycloak.
            user = JSON.parse(user);
            let authPermissions = await Auth.getUserPermission(user);
            if (Auth.prevalidatedToken !== user?.token) { // let's check if this is still a valid token.
                let isTokenValid = await Auth.validateExistingToken(user)
                if (!isTokenValid) {
                    if (Auth.isDebugLoggin) console.log("// isAuthenticated : No Valid Token")
                    return false
                } else {

                    faro?.api?.setUser({
                        id: user.name,
                        username: user.name
                      });
                    if (Auth.isDebugLoggin) console.log("-  isAuthenticated : Valid Existing Token")
                }
            } else {
                if (Auth.isDebugLoggin) console.log(" - isAuthenticated : Cache token match", Auth.prevalidatedToken)
            }
            if (Auth.isDebugLoggin) console.log("/ isAuthenticated : Retrieved authPermissions:", authPermissions)
            return authPermissions;
        } else {
            //Keycloak authentication, we  relying on the cookie. LDAP User will get a False, we cannot authenticate them.
            // but keycloak users  might have a session
            if (Auth.isDebugLoggin) console.log("- Check getKeycloakAuthState. ")
            const res = await AuthService.getKeycloakAuthState();
            if (res?.is_authenticated) {
                localStorage.setItem("loginType", 'Keycloak');
                const userDetails = SetUserFromKeyCloak(res);
                return Auth.getUserPermission(userDetails);
            }
        }
        await PermissionStackUtil.getPermissions(false);
        return false;
    },

    getUserPermission: async (user) => {
        if (user.token) {
            axios.defaults.headers.common['Authorization'] = `Token ${user.token}`;
            axios.defaults.headers.common['X-CSRFTOKEN'] = `${user.token}`;
            const permissions = await AuthStore.getState();
            if (!permissions.userRolePermission.project) {
                await PermissionStackUtil.getPermissions(true);
            }
            return true;
        }
    },
    /** Gets user details from browser local storage */
    getUser: () => {
        return JSON.parse(localStorage.getItem("user"));
    },
    /** Authenticate user from the backend and store user details in local storage */
    login: async (user, pass) => {
        const authData = await AuthService.authenticate(user, pass);
        if (authData) {
            localStorage.setItem("user", JSON.stringify({ name: user, token: authData.token, websocket_token: authData.token },));
            localStorage.setItem("loginType", 'Application');
            await PermissionStackUtil.getPermissions(true);
            return true;
        } else {
            return false;
        }
    },
    /** Remove user details from localstorage on logout and clear entire local storage*/
    logout: () => {
        AuthService.deAuthenticate();
        PermissionStackUtil.deleteStore();
        Auth.prevaliDationUser = "";
        localStorage.removeItem("user");
        localStorage.clear();
    }
}

export default Auth;

function SetUserFromKeyCloak(res) {
    const cookies = document.cookie.split(';').reduce((prev, current) => {
        const [name, value] = current.split(/\s?(.*?)=(.*)/).splice(1, 2);
        prev[name] = value;
        return prev;
    }, {});
    const userDetails = { 'name': res.username, 'token': cookies.csrftoken || res.csrftoken, 'websocket_token': res.websocket_token };
    localStorage.setItem("user", JSON.stringify(userDetails));
    return userDetails;
}
