import { BrowserRouter as Router } from 'react-router-dom';
import { act } from "react-dom/test-utils";
import { render, cleanup, fireEvent, screen, waitFor } from '@testing-library/react';
import '@testing-library/jest-dom';
import mockConsole from "jest-mock-console";
import _ from 'lodash';

import { ProjectEdit } from './edit';
import ProjectService from '../../services/project.service';
import CycleService from '../../services/cycle.service';

import ProjectServiceMock from '../../__mocks__/project.service.data';
import AuthServiceMock from '../../__mocks__/auth.service.data';
import AuthService from '../../services/auth.service';


let projectCategoriesSpy, allCycleSpy, periodCategoriesSpy, projectDetailsSpy, resourcesSpy, projectQuotaSpy, archiveLocationSpy,
    updateProjectSpy, savePQSpy, updatePQSpy, deletePQSpy, projectPermissionSpy, fileSystemSpy, clusterSpy;
let restoreConsole;

beforeEach(() => {
    restoreConsole = mockConsole();
    setMockSpy();
});

afterEach(() => {
    restoreConsole();
    // cleanup on exiting
    clearMockSpy();
    cleanup();
});

/**
 * To set mock spy for Services that have API calls to the back end to fetch data
 */
const setMockSpy = (() => {
    projectCategoriesSpy = jest.spyOn(ProjectService, 'getProjectCategories');
    projectCategoriesSpy.mockImplementation(() => { return Promise.resolve([{ url: "Regular", value: 'Regular' }, { url: "User Shared Support", value: 'User Shared Support' }]) });
    periodCategoriesSpy = jest.spyOn(ProjectService, 'getPeriodCategories');
    periodCategoriesSpy.mockImplementation(() => { return Promise.resolve([{ url: "Single Cycle", value: 'Single Cycle' }, { url: "Long Term", value: 'Long Term' }]) });
    allCycleSpy = jest.spyOn(CycleService, 'getAllCycles');
    allCycleSpy.mockImplementation(() => {
        return Promise.resolve([{ url: "https://localhost:3000/api/cycle/Cycle-0/", name: 'Cycle-0' },
        { url: "https://localhost:3000/api/cycle/Cycle-1/", name: 'Cycle-1' },
        { url: "https://192.168.99.100:8008/api/cycle/Cycle%200/", name: 'Cycle 0' }]);
    });
    projectDetailsSpy = jest.spyOn(ProjectService, 'getProjectDetails');
    projectDetailsSpy.mockImplementation((id) => {
        return Promise.resolve(_.find(ProjectServiceMock.project, { name: id }))
    });
    resourcesSpy = jest.spyOn(ProjectService, 'getResources');
    resourcesSpy.mockImplementation(() => {
        let resourceList = [];
        Object.assign(resourceList, ProjectServiceMock.resources);
        return Promise.resolve(resourceList);
    });
    projectQuotaSpy = jest.spyOn(ProjectService, 'getProjectQuota');
    projectQuotaSpy.mockImplementation((id) => {
        let quota = {};
        Object.assign(quota, _.find(ProjectServiceMock.projectQuota, { id: id }));
        return Promise.resolve(quota);
    });
    updateProjectSpy = jest.spyOn(ProjectService, 'updateProject');
    updateProjectSpy.mockImplementation((id, project) => {
        let updatedProject = {};
        Object.assign(updatedProject, _.find(ProjectServiceMock.project, { name: id }));
        updatedProject.name = project.name;
        updatedProject.updated_at = new Date();
        updatedProject.isUpdated = true
        return Promise.resolve(updatedProject);
    });
    savePQSpy = jest.spyOn(ProjectService, 'saveProjectQuota');
    savePQSpy.mockImplementation(() => {
        return Promise.resolve({ data: ProjectServiceMock.projectQuota[0] });
    });
    updatePQSpy = jest.spyOn(ProjectService, 'updateProjectQuota');
    updatePQSpy.mockImplementation((quota) => {
        return Promise.resolve({ data: _.find(ProjectServiceMock.projectQuota, { id: quota.id }) });
    });
    deletePQSpy = jest.spyOn(ProjectService, 'deleteProjectQuota');
    deletePQSpy.mockImplementation(() => {
        return Promise.resolve({ message: 'deleted' });
    });
    projectPermissionSpy = jest.spyOn(AuthService, 'getAccessControlMethod');
    projectPermissionSpy.mockImplementation(() => {
        return Promise.resolve(AuthServiceMock.projectId)
    })
    fileSystemSpy = jest.spyOn(ProjectService, 'getFileSystem');
    fileSystemSpy.mockImplementation(() => {
        return Promise.resolve(ProjectServiceMock.fileSystem)
    })
    clusterSpy = jest.spyOn(ProjectService, 'getCluster');
    clusterSpy.mockImplementation(() => {
        return Promise.resolve(ProjectServiceMock.cluster)
    })
    archiveLocationSpy = jest.spyOn(ProjectService, 'getProjectQuotaArchiveLocation');
    archiveLocationSpy.mockImplementation(() => {
        return Promise.resolve(ProjectServiceMock.archiveLocation)
    })

});

const clearMockSpy = (() => {
    projectCategoriesSpy.mockRestore();
    periodCategoriesSpy.mockRestore();
    projectDetailsSpy.mockRestore();
    resourcesSpy.mockRestore();
    projectQuotaSpy.mockRestore();
    updateProjectSpy.mockRestore();
    savePQSpy.mockRestore();
    updatePQSpy.mockRestore();
    deletePQSpy.mockRestore();
    projectPermissionSpy.mockRestore();
    fileSystemSpy.mockRestore();
    clusterSpy.mockRestore();
    archiveLocationSpy.mockRestore();
});

it("renders nothing if no project details found", async () => {
    console.log("renders nothing if no project details found..........");
    let content;
    await act(async () => {
        content = render(<Router><ProjectEdit match={{ params: { id: "OSR-12" } }} location={{}} /></Router>);
    });

    expect(content.queryByText("Project - Edit")).toBe(null);
});

it("renders input fields with Project details if found", async () => {
    console.log("renders input fields with Project details if found..........");
    let content;
    await act(async () => {
        content = render(<Router><ProjectEdit match={{ params: { id: "OSR-11" } }} location={{}} /></Router>);
    });

    expect(content.queryByText("Project - Edit")).not.toBe(null);
    expect(content.queryByText("OSR-11")).not.toBe(null);

    const spinButtons = content.queryAllByRole("spinbutton");
    const trigPrioInput = _.filter(spinButtons, { "id": "trig_prio" })[0];
    expect(trigPrioInput.value).toBe("1000");

    const rankInput = _.filter(spinButtons, { "id": "proj-rank" })[0];
    expect(rankInput.value).toBe("2");

    const trigger = content.getAllByLabelText(/trigger/i).filter((element) => { return element.id === "trigger" })[0];
    expect(trigger.hasAttribute("checked")).toBeFalsy();


    expect(content.queryAllByText('Select Project Category').length).toBe(2);
    expect(content.queryAllByText('Regular').length).toBe(0);


    expect(content.queryAllByText('Select Period Category').length).toBe(2);
    expect(content.queryAllByText('Single Cycle').length).toBe(0);

    expect(content.queryAllByText('Cycle 0').length).toBe(0);

    expect(_.filter(spinButtons, { "id": "CEP Processing Time" })[0].value).toBe("1");
    expect(_.filter(spinButtons, { "id": "LOFAR Observing Time" })[0].value).toBe("1");
    expect(_.filter(spinButtons, { "id": "LOFAR Observing Time prio A" })[0].value).toBe("1");
    expect(_.filter(spinButtons, { "id": "LOFAR Observing Time prio B" })[0].value).toBe("1");
    expect(_.filter(spinButtons, { "id": "LOFAR Support Time" })[0].value).toBe("1");
    expect(_.filter(spinButtons, { "id": "LTA Storage" })[0].value).toBe("1.099511627776");
    expect(_.filter(spinButtons, { "id": "Number of triggers" })[0].value).toBe("1");

    expect(content.queryByTestId('save-btn').hasAttribute("disabled")).toBeFalsy();

});

it("Save button disabled if mandatory field value is removed", async () => {
    console.log("renders input fields with Project details if found..........");
    let content;
    await act(async () => {
        content = render(<Router><ProjectEdit match={{ params: { id: "OSR-11" } }} location={{}} /></Router>);
    });

    expect(content.queryByText("Project - Edit")).not.toBe(null);
    expect(content.queryByText("OSR-11")).not.toBe(null);
    const descInput = content.queryByTestId('description');
    await act(async () => {
        fireEvent.change(descInput, { target: { value: '' } });
    });
    expect(content.queryByTestId('save-btn').hasAttribute("disabled"));

});

it("save Project after editing fields", async () => {
    console.log("save Project after editing fields ..........");
    let content;
    await act(async () => {
        content = render(<Router><ProjectEdit match={{ params: { id: "OSR-11" } }} location={{}} history={{ length: 1, go: () => { } }} /></Router>);
    });

    expect(content.queryByText("Project - Edit")).not.toBe(null);
    expect(content.queryByText("OSR-11")).not.toBe(null);

    const spinButtons = content.queryAllByRole("spinbutton");
    const trigPrioInput = _.filter(spinButtons, { "id": "trig_prio" })[0];
    fireEvent.blur(trigPrioInput, { target: { value: 900 } });
    expect(trigPrioInput.value).toBe("900");

    const rankInput = _.filter(spinButtons, { "id": "proj-rank" })[0];
    fireEvent.blur(rankInput, { target: { value: 2 } });
    expect(rankInput.value).toBe("2");

    const trigger = content.getAllByLabelText(/trigger/i).filter((element) => { return element.id === "trigger" })[0];
    await act(async () => {
    fireEvent.click(trigger);
    });
    expect(trigger.hasAttribute("checked")).toBeTruthy();
    await act(async () => {
        fireEvent.click(screen.getAllByText("Select Project Category")[0].parentElement.parentElement.parentElement.children[3]);
    });
    const projCatInput = screen.getAllByText("User Shared Support")[0];
    await act(async () => { fireEvent.click(projCatInput); });

    expect(content.queryByTestId("project-cat")).toBeInTheDocument();
    await waitFor(() => expect(content.queryAllByText('Regular').length).toBe(1));
    expect(content.queryAllByText('User Shared Support').length).toBe(3);
    await act(async () => {
        fireEvent.click(screen.getAllByText("Select Period Category")[0].parentElement.parentElement.parentElement.children[3]);
    });

    const projPeriodInput = screen.getAllByText("Long Term")[0];
    await act(async () => {
        fireEvent.click(projPeriodInput);
    });
    expect(content.queryByTestId("project-cat")).toBeInTheDocument();
    expect(content.queryAllByText('Single Cycle').length).toBe(1);
    expect(content.queryAllByText('Long Term').length).toBe(3);
    await act(async () => { fireEvent.click(screen.getAllByRole("listbox")[0].parentElement.parentElement); });
    const oldCycleInput = screen.getAllByText("Cycle-0")[0];
    const newCycleInput = screen.getAllByText("Cycle 0")[0];
    await act(async () => { fireEvent.click(oldCycleInput); });
    await act(async () => { fireEvent.click(newCycleInput); });
    
    // After selecting Cycle
    expect(content.queryAllByText('Cycle-0').length).toBe(1);
    expect(content.queryAllByText('Cycle 0').length).toBe(1);

    const lofarObsTimeInput = _.filter(spinButtons, { "id": "LOFAR Observing Time" })[0];
    await act(async () => { fireEvent.change(lofarObsTimeInput, { target: { value: 10 } }) });
    expect(lofarObsTimeInput.value).toBe('10');

    const cepProcTimeInput = _.filter(spinButtons, { "id": "CEP Processing Time" })[0];
    await act(async () => {
        fireEvent.change(cepProcTimeInput, { target: { value: 5 } });
    });
    expect(cepProcTimeInput.value).toBe('5');

    const ltaStorageInput = _.filter(spinButtons, { "id": "LTA Storage" })[0];
    await act(async () => {
        fireEvent.change(ltaStorageInput, { target: { value: 2 } });
    });
    expect(ltaStorageInput.value).toBe('2');

    expect(content.queryByTestId('save-btn').hasAttribute("disabled")).toBeFalsy();

    await act(async () => {
        fireEvent.click(content.queryByTestId('save-btn'));
    });

    // After saving project, Success dialog should be displayed
    expect(updateProjectSpy).toHaveBeenCalledTimes(1);
    expect(updatePQSpy).toHaveBeenCalledTimes(3);
    expect(content.queryByText("Success")).not.toBe(null);
});

it("save Project after adding, modifying and deleting resources", async () => {
    console.log("save Project after adding, modifying and deleting resource..........");
    let content;
    await act(async () => {
        content = render(<Router><ProjectEdit match={{ params: { id: "OSR-11" } }} location={{}} history={{ length: 1, go: () => { } }} /></Router>);
    });

    expect(content.queryByText("Project - Edit")).not.toBe(null);
    expect(content.queryByText("OSR-11")).not.toBe(null);
    await waitFor(() => screen.getAllByText("Add Resources"));


    let addedresurceschild = screen.getAllByText("Add Resources")[0].parentElement.parentElement.parentElement.children[3];
    // Adding new resource
    await act(async () => { fireEvent.click(addedresurceschild); });
    const addResourceInput = screen.getAllByText("CEP Storage")[0];
    await act(async () => { fireEvent.click(addResourceInput); });
    // After selecting New Resource
    expect(content.queryAllByText('CEP Storage').length).toBe(3);
    await waitFor(() => expect(content.queryAllByText('Add Resources').length == 1));


    const addResourceBtn = content.queryByTestId('add_res_btn');
    await act(async () => { fireEvent.click(addResourceBtn); });
    expect(content.queryAllByText('Add Resources').length).toBe(2);

    const spinButtons = content.queryAllByRole("spinbutton");
    const lofarSupHrsInput = _.filter(spinButtons, { "id": "CEP Storage" })[0];
    expect(lofarSupHrsInput).not.toBe(null);
    fireEvent.change(lofarSupHrsInput, { target: { value: 100 } });
    expect(lofarSupHrsInput.value).toBe('100');

    // Editing existing resource
    const lofarObsTimeInput = _.filter(spinButtons, { "id": "LOFAR Observing Time" })[0];
    fireEvent.change(lofarObsTimeInput, { target: { value: 15 } });
    expect(lofarObsTimeInput.value).toBe('15');

    // Deleting existing resource
    const removeResourceBtn = content.queryByTestId('LTA Storage-btn');
    fireEvent.click(removeResourceBtn);

    expect(content.queryByTestId('save-btn').hasAttribute("disabled")).toBeFalsy();

    await act(async () => {
        fireEvent.click(content.queryByTestId('save-btn'));
    });
    await new Promise((r) => setTimeout(r, 1000));
    // After saving project, Success dialog should be displayed
    expect(updateProjectSpy).toHaveBeenCalledTimes(1);
    expect(savePQSpy).toHaveBeenCalledTimes(1);
    expect(updatePQSpy).toHaveBeenCalledTimes(1);
    expect(deletePQSpy).toHaveBeenCalledTimes(1);
});

