import _ from 'lodash';
import ScheduleService from '../../services/schedule.service';
import WorkflowService from '../../services/workflow.service';

/**
 * Class with common functions used by both Timeline View and Week Overview components.
 */
class TimelineCommonUtils {

    // Constants used in timeline views
    ZOOM_LEVELS = [{ name: '30 Minutes', value: 30 * 60 },
    { name: '1 Hour', value: 1 * 60 * 60 },
    { name: '3 Hours', value: 3 * 60 * 60 },
    { name: '6 Hours', value: 6 * 60 * 60 },
    { name: '12 Hour', value: 12 * 60 * 60 },
    { name: '1 Day', value: 24 * 60 * 60 },
    { name: '2 Days', value: 2 * 24 * 60 * 60 },
    { name: '3 Days', value: 3 * 24 * 60 * 60 },
    { name: '5 Days', value: 5 * 24 * 60 * 60 },
    { name: '1 Week', value: 7 * 24 * 60 * 60 },
    { name: '2 Weeks', value: 14 * 24 * 60 * 60 },
    { name: '4 Weeks', value: 28 * 24 * 60 * 60 },
    { name: 'Custom', value: 0 * 60 * 60 }];
    DEFAULT_ZOOM_LEVEL = "2 Days";
    WORKFLOW_STATUS_MAP = {
        'wait scheduled': 'Waiting To Be Scheduled',
        'wait observing': 'Scheduled',
        'qa reporting to': 'QA Reporting (TO)',
        'qa reporting sos': 'QA Reporting (SDCO)',
        'pi verification': 'PI Verification',
        'decide acceptance': 'Decide Acceptance',
        'wait ingest done': 'Ingesting',
        'decide unpin data': 'Unpin Data',
        'end': 'Done'
    };

    constructor() {
        this.STATUS_BEFORE_SCHEDULED = ['defining', 'defined', 'schedulable'];  // Statuses before scheduled to get station_group
        this.loadMainStationGroups();
    }

    /**
     * To group the stations under CS,RS,IS to show the count in Popover
     */
    async loadMainStationGroups() {
        async function fetchMainStationGroups() {
            return await ScheduleService.getMainGroupStations()
        }

        this.mainStationGroups = await fetchMainStationGroups().catch(e => console.error("Couldn't get retrieve main station groups", e))
    }



    /**
     * Get all stations of the SU bleprint from the observation task or subtask based on the SU status.
     * @param {Object} suBlueprint
     */
    getSUStations(suBlueprint) {
        let stations = [];
        /* Get all observation tasks */
        let observationTasks = _.filter(suBlueprint.tasks, (task) => { return task.task_type.toLowerCase() === "observation" });
        if (observationTasks.length === 0) {
            observationTasks = _.filter(suBlueprint.task_blueprints, (task) => { return task.task_type.toLowerCase() === "observation" });
        }
        for (const observationTask of observationTasks) {
            /** If the status of SU is before scheduled, get all stations from the station_groups from the task specification_docs */
            if (this.STATUS_BEFORE_SCHEDULED.indexOf(suBlueprint.status.toLowerCase()) >= 0 && observationTask.specifications_doc.station_configuration?.station_groups) {
                for (const grpStations of _.map(observationTask.specifications_doc.station_configuration?.station_groups, "stations")) {
                    stations = _.concat(stations, grpStations);
                }
            } else if (this.STATUS_BEFORE_SCHEDULED.indexOf(suBlueprint.status.toLowerCase()) < 0 && observationTask.subtasks) {
                /** If the status of SU is scheduled or after get the stations from the subtask specification tasks */
                for (const subtask of observationTask.subtasks) {
                    if (subtask.specifications_doc.stations) {
                        stations = _.concat(stations, subtask.specifications_doc.stations.station_list);
                    }
                }
            }
        }
        return _.uniq(stations);
    }

    /**
     * Group the SU stations to main groups Core, Remote, International
     * @param {Object} stationList 
     */
    groupSUStations(stationList) {
        let suStationGroups = {};
        for (const group in this.mainStationGroups) {
            suStationGroups[group] = _.intersection(this.mainStationGroups[group], stationList);
        }
        return suStationGroups;
    }


    /**
     * To get all workflow processes and tasks for a specific set of SUB Ids passed as array.
     * @param {Array} suIds - Array of scheduling unit blueprint id
     * @returns - Object with all workflow processes and tasks as properties. {processes: [], tasks:[]}
     */
    async getWorkflowsAndTasks(suIds) {
        let workflow = {};
        let processes = [];
        let tasks = [];
        if (suIds.length > 0) {
            processes = await WorkflowService.getSUWorkflowProcesses(suIds.join(","));
            const processIds = _.map(processes, 'id');
            if (processIds.length > 0) {
                tasks = await WorkflowService.getWorkflowProcessesTasks(processIds);
            }
        }
        workflow.processes = processes;
        workflow.tasks = tasks;
        return workflow;
    }


    /**
     * Function to get the current workflow details for the scheduling unit
     * @param {Array} workflowProcesses - List of workflow processes filtered for the SUBs in the table view
     * @param {Array} workflowTasks - List of workflow tasks filtered for the workflow processes
     * @param {Number} scheduleunitId - Scheduling Unit blueprint id for which the status to be found
     * @returns workflow status as Object with status, results_accepted.
     */
    getWorkflowDetails(workflows, scheduleunitId) {
        let workflowDetails = {};
        const suWorkflow = _.find(workflows.processes, ['su', scheduleunitId]);
        const suWorkflowTasks = suWorkflow ? _.filter(workflows.tasks, ['process', suWorkflow.id]) : [];
        const lastWorkflowTask = suWorkflowTasks.length > 0 ? (_.orderBy(suWorkflowTasks, ['id'], ['desc']))[0] : null;
        if (lastWorkflowTask?.flow_task) {
            workflowDetails["dataAccepted"] = suWorkflow.results_accepted;  // TODO: replaced flag, now it's in the SUB model so update this ref.
            workflowDetails["status"] = this.WORKFLOW_STATUS_MAP[lastWorkflowTask.flow_task.toLowerCase()];
        }
        return workflowDetails;
    }


}

export default TimelineCommonUtils;