import moment from 'moment';
import _ from 'lodash';

import ScheduleService from "../../services/schedule.service";
import UIConstants from '../../utils/ui.constants';
import TaskService from "../../services/task.service";

const QaReportingTemplate = {
    /**
     * Function that return the QA Reporting template with values pre-populated from the Scheduling Unit
     * and tasks.
     * @param {number} id - Scheduling Unit blueprint id
     * @returns - QA Reporting template component
     */
    getReportTemplate : async function(id) {
        const SUB = await ScheduleService.getSchedulingUnitExtended('blueprint', id);
        let dataproducts = await ScheduleService.getSUDataproducts(id);
        let cycleName = SUB.scheduling_set_object.project.cycles_ids.join(", ");
        let tasks = SUB.task_blueprints;
        let project_id =  SUB.scheduling_set_object.project_id;

        let missingStationsContent = null;
        let missingStations = [];
        let observations = []
        for (let task of tasks) {
            let observation = {
                pipeline: []
            };
            if (task.task_type === "observation") {
                const taskSuccessors = await TaskService.getTaskSuccessors('blueprint', task.id)
                let taskDataproducts = _.find(dataproducts.task_blueprints, {'id': task.id})
                if(taskSuccessors) {
                    for(const successor of taskSuccessors) {
                        if(successor.task_type === "pipeline") {
                            successor.subtasks.map(successorSubtask => {
                                if(successorSubtask.primary === true){
                                    observation.pipeline.push(successorSubtask.id);
                                }
                                return successorSubtask;
                            })
                        }
                    }
                }
                let taskDP = [];
                let completeness = [];
                // From the dataproducts of all subtasks of the task filter the ones with feedback to get number of files with feedback.
                // Calculate the completeness from the percentage written in the feedback
                for (const subtask of taskDataproducts.subtasks) {
                    const dataproducts = _.filter(subtask.output_dataproducts, dataproduct => {
                        if (dataproduct.feedback_doc && dataproduct.feedback_doc.percentage_written) {
                            completeness.push(dataproduct.feedback_doc.percentage_written);
                        }
                        return dataproduct.feedback_doc?true:false;
                    });
                    taskDP.push(...dataproducts);
                }
                observation.dataproducts = taskDP.length;
                observation.completeness = _.mean(completeness) || 0;
                let subtasks = _.filter(task.subtasks, ['subtask_type', 'observation']);
                for (const subtask of subtasks) {
                    // Get missing stations of the target observation from its observation subtask
                    if (["target observation", "beamforming observation", "combined observation"].includes(task.specifications_template.name.toLowerCase())) {
                        missingStations = missingStations.concat(await TaskService.getSubtaskMissingStations(subtask.id));
                    }
                    observation.controlId = subtask.id;
                    observation.scheduleUnitId =  task.scheduling_unit_blueprint_id;
                    observation.taskId = task.id;
                    observation.project =  project_id;
                    observation.antenna = subtask.specifications_doc.stations.antenna_set;
                    observation.filter = subtask.specifications_doc.stations.filter;
                    observation.startTime = moment(subtask.on_sky_start_time, moment.ISO_8601).format(UIConstants.CALENDAR_DATETIME_FORMAT);
                    observation.stopTime = moment(subtask.on_sky_stop_time, moment.ISO_8601).format(UIConstants.CALENDAR_DATETIME_FORMAT);
                    observation.targetname = task.specifications_doc.station_configuration?.SAPs ?
                                                task.specifications_doc.station_configuration?.SAPs.map(sap => sap.digital_pointing.target): [];
                    observation.targetname = observation.targetname.join(", ");
                    observation.subbands = subtask.specifications_doc.stations?.digital_pointings ?
                                                subtask.specifications_doc.stations?.digital_pointings.map(pointing => pointing.subbands.length): [];
                    observation.subbands = observation.subbands.join(", ");
                    observations.push(observation);
                }
            }
        }
        // If there are missing stations, order them, group by station group, concat them and create content to display
        if (missingStations.length > 0) {
            const coreStations = (await ScheduleService.getStations("Core")).stations;
            const remoteStations = (await ScheduleService.getStations("Remote")).stations;
            const interStations = (await ScheduleService.getStations("International")).stations;
            missingStations = _.sortedUniq(missingStations);
            let coreMissingStations = _.filter(missingStations, station => { return _.includes(coreStations, station)});
            let remoteMissingStations = _.filter(missingStations, station => { return _.includes(remoteStations, station)});
            let interMissingStations = _.filter(missingStations, station => { return _.includes(interStations, station)});
            missingStations = coreMissingStations.concat(remoteMissingStations).concat(interMissingStations);
            missingStationsContent = `<p><strong>Stations not participating because of reservations: </strong> ${missingStations.join(", ")}</p>`;
        }
        let tableBody = ''
        const sortedObservations = _.orderBy(observations,['startTime'], ['desc'])
        for (const observation of sortedObservations) {
            tableBody = tableBody + `
            <tr>
                <td><a href="https://proxy.lofar.eu/inspect/HTML/${observation.controlId}/index.html">${observation.controlId}</a></td>
                <td>${observation.scheduleUnitId}</td>
                <td>${observation.taskId}</td>
                <td>${observation.pipeline.join(", ")}</td>
                <td>${observation.project}</td>
                <td>${observation.targetname}</td>
                <td><a href="https://proxy.lofar.eu/inspect/HTML/${observation.controlId}/Stations/station_beamlets.html">BST</a></td>
                <td><a href="https://proxy.lofar.eu/inspect/HTML/${observation.controlId}/index.html">${observation.completeness}%</a></td>
                <td>${observation.antenna}</td>
                <td>${observation.filter}</td>
                <td>${observation.startTime}</td>
                <td>${observation.stopTime}</td>
                <td>${observation.subbands}</td>
                <td>${observation.dataproducts}</td>
                <td><a href="https://proxy.lofar.eu/inspect/${observation.controlId}/rtcp-${observation.controlId}.parset">parset</a></td>
            </tr>`
        }
        
        return (
            `<p>Subject- Observation ${SUB.status.toLowerCase() === 'error'?'Failed': 'Success'}</p>
            <p>Dear Colleague,</p>
            <p> The following message contains information regarding a LOFAR ${cycleName} project for which you are listed as the contact author.
            Please forward this information to your collaborators.</p>
            <p> We would like to inform you that observations related to your LOFAR ${cycleName} project have been performed. Please find detailed information below.</p>
            <strong>General notes:</strong>
            <p> The validation plots linked below remains online for 3 weeks from the date of observation. After that, they are compressed and transferred to offline storage. users can request access to them by submitting a ticket through the <a href="https://support.astron.nl/sdchelpdesk">helpdesk</a>.</p>
            <strong>Observations:</strong>
            <table>
                <thead>
                    <th>Observation ID</th>
                    <th> Scheduling Unit ID</th>
                    <th> Task ID</th>
                    <th> Pipeline ID</th>
                    <th> Project</th>
                    <th> Target Name</th>
                    <th> Station plots</th>
                    <th> Completeness</th>
                    <th> Antenna Set</th>
                    <th> Filter</th>
                    <th> Start Time</th>
                    <th> End time</th>
                    <th> Number of Sub-bands</th>
                    <th> Number of files</th>
                    <th> Parset</th>
                </thead>
                <tbody>
                    ${tableBody}
                </tbody>
            </table>
            <br/>
            <p><strong>Performance of the system:</strong><br><br></p>
            <p><strong>Data recording:</strong><br><br></p>
            ${missingStationsContent?missingStationsContent:""}
            <p><strong>Data processing:</strong><br><br></p>
            <p><strong>Archiving:</strong><br><br></p>
            <strong style='display:inline'>Remarks:</strong><p style= "display:inline; position: relative"> Please analyse the validation plots at 
            <a href="https://proxy.lofar.eu/inspect/HTML/">https://proxy.lofar.eu/inspect/HTML/</a> within 24 hours after this notification and submit your 
            findings using the form at <a href="/schedulingunit/${id}/workflow">${window.location.origin}/schedulingunit/${id}/workflow </a>.
            After this time window has passed, we will assume that your judgement is that the observation was successful and we will complete the actions described above ourselves to support your run.</p>
            <br/>
            <p>From the moment the data are made available to you at the LTA you have four weeks to check their quality and to report any problems to the 
            Observatory through the <a href="https://support.astron.nl/sdchelpdesk">helpdesk</a>. After this time window has passed, no requests for re-observations will be considered.</p>

            <strong>Actions:</strong>
            <p>If you need any further clarification, please do not hesitate to contact us through the SDC helpdesk at <a href="https://support.astron.nl/sdchelpdesk">https://support.astron.nl/sdchelpdesk</a>, specifying your project code in the subject.

            <p>Best regards,</p>
            <p>TO name (Telescope Operations)</p>
            <p>SDCO name (Science Data center Operations)</p>`
        );
    }
    
};

export default QaReportingTemplate;