import axios from "axios"

const AuthService = {
    authenticate: async (user, pass) => {
        try {
            delete axios.defaults.headers.common['Authorization'];
            delete axios.defaults.headers.common['X-Csrftoken'];
            const response = await axios.post("/api/token-auth/", { username: user, password: pass }, { headers: { 'Content-Type': 'application/json' } });
            axios.defaults.headers.common['Authorization'] = `Token ${response.data.token}`;
            return response.data;
        } catch (error) {
            console.error(error);
            return null;
        }
    },
    /**
     * Get keycloak auth state
     * @returns auth state in JSON
     */
    getKeycloakAuthState: async () => {
        try {
            delete axios.defaults.headers.common['Authorization'];
            const response = await axios.get('/api/authentication_state/');
            return response.data;
        } catch (error) {
            console.error('Error [getKeycloakAuthState]', error);
            return null;
        }
    },
    /**
     * Deauthenticate the user login
     *  - Based on login type it will access the API/Keycloak url
     * @param {*} token 
     */
    deAuthenticate: async () => {
        let contentType = '';
        try {
            const loginType = localStorage.getItem("loginType");
            if (loginType && loginType === 'Application') {
                await axios.delete("/api/token-deauth/");
            } else {
                //If logintype is keycloak set Content-Type to 'application/x-www-form-urlencoded; charset=UTF-8'
                let token = '';
                let user = localStorage.getItem("user");
                if (user) {
                    user = JSON.parse(user);
                    token = user.token;
                }
                contentType = axios.defaults.headers.common['Content-Type'];
                const formData = 'csrfmiddlewaretoken=' + token;
                await axios.post('/oidc/logout/', formData, {
                    headers: {
                        'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8',
                    }
                });
            }
            localStorage.removeItem("loginType");
        } catch (error) {
            console.error(error);
        }
        axios.defaults.headers.common['Content-Type'] = contentType;
    },
    isValidAuthenticationState: async (username) => {
        let response
        try {
            // the server will first attempt to see if the user is already authenticated (based upon cookies)
            // and then if it fails, it will validate upon the token (present in the header) and the username
            // This gives us a change to both validate keycloak and application tokens.
            // when using keycloak, do not  supply the username.
            response = await axios.get("/api/authentication_state/?username=" + username)
            if (response?.status == 200) {
                return response.data
            }

        } catch (error) {
            console.error("// isValidAuthenticationState Requested. Got error", error, response);
        }
        return {}
    },
    getAccessControlMethod: async (subdomain, source) => {
        let res = null;
        try {
            const response = await axios.get(`/api/${subdomain}/?fields=name&limit=1`, {
                cancelToken: source
            });
            res = response.headers['access-control-allow-methods'];
        } catch (error) {
            console.error('[AuthService.getAccessControlMethod]', error);
            return error.response;
        }
        return res;
    }
}
export default AuthService;