import axios from "axios"

const ReservationService = {
    getReservationTemplates: async function () {
        try {
            const url = `/api/reservation_template`;
            const response = await axios.get(url);
            return response.data.results;
        } catch (error) {
            console.error(error);
        }
    },
    saveReservation: async function (reservation) {
        try {
            const response = await axios.post(('/api/reservation/'), reservation);
            return response.data;
          } catch (error) {
            console.error(error);
            return null;
          }
    },
    updateReservation: async function (reservation) {
        try {
            const response = await axios.put((`/api/reservation/${reservation.id}/`), reservation);
            return response.data;
          } catch (error) {
            console.error(error);
            return null;
          }
    },
    /** Gets all reservations  */
    getReservations: async() => {
        let reservations = [];
        try {
            let initialResponse = await axios.get(`/api/reservation/?ordering=id`);
            const totalCount = initialResponse.data.count;
            const initialCount = initialResponse.data.results.length
            reservations = reservations.concat(initialResponse.data.results);
            if (totalCount > initialCount) {
                let secondResponse = await axios.get(`/api/reservation/?ordering=id&limit=${totalCount-initialCount}&offset=${initialCount}`);
                reservations = reservations.concat(secondResponse.data.results);
            }
        }   catch(error) {
            console.error('[schedule.services.getReservations]',error);
        }
        return reservations;
    },
    /** Get the reservations that are present during the period specified by startTime and stopTime */
    getTimelineReservations: async(startTime, stopTime, id) => {
        let reservations = [];
        try {
            let url = `/api/reservation/?ordering=id`;
            if (stopTime) {     // Gets the reservations started and stopped between the period
                url = `${url}&start_time_before=${stopTime || ''}&stop_time_after=${startTime || ''}`;
            }   else if (startTime) {   // Gets the reservations started before the startTime and still exists
                url = `${url}&start_time_before=${startTime || ''}&stop_time_isnull=true`;
            }
            if (id) {
                url = `${url}&id=${id || ''}`;
            }
            let initialResponse = await axios.get(url);
            const totalCount = initialResponse.data.count;
            const initialCount = initialResponse.data.results.length;
            reservations = reservations.concat(initialResponse.data.results);
            if (totalCount > initialCount) {
                let secondResponse = await axios.get(`${url}&limit=${totalCount-initialCount}&offset=${initialCount}`);
                reservations = reservations.concat(secondResponse.data.results);
            }
            if (stopTime) {     // Get the reservations started before the stopTime and exists.
                let indefiniteReservations = await ReservationService.getTimelineReservations(stopTime);
                reservations = reservations.map(reservation => ( //filter on unique reservations (id)
                    {...reservation, ...indefiniteReservations.find(indefiniteReservation => indefiniteReservation.id === reservation.id)}
                ))
            }
        }   catch(error) {
            console.error('[schedule.services.getTimelineReservations]',error);
        }
        return reservations;
    },
    getReservation: async function (id) {
        try {
            const response = await axios.get(`/api/reservation/${id}`);
            return response.data;
        }   catch(error) {
            console.error(error);
            return null;
        }
    },
    getReservationTemplate: async function(templateId) {
        try {
          const response = await axios.get('/api/reservation_template/' + templateId);
          return response.data;
        } catch (error) {
          console.log(error);
        }
      },
     
    deleteReservation: async function(id) {
        try {
            const url = `/api/reservation/${id}`;
            await axios.delete(url);
            return true;
        } catch(error) {
            console.error(error);
            return false;
        }
    },
    getReservationStrategyTemplates: async function () {
        try {
            const url = `/api/reservation_strategy_template/?ordering=id`;
            const response = await axios.get(url);
            return response.data.results;
        } catch (error) {
            console.error(error);
        }
    },
    getReservationFilterDefinition: async function (){
        let res = [];
        try {
            res = await axios.options(`/api/reservation/`);
        }   catch(error) {
            console.error('[reservation.services.getReservationFilterDefinition]',error);
        }
        return res;
    },
    getReservationsWithFilter: async function (filters, orderBy, limit, offset){
        let response = null;
        let api = `/api/reservation/?`;
        try {
        
            api += (filters === '')? '' : filters+'&';
            api += (orderBy === '')? '' : orderBy+'&';
            api += 'limit='+limit+'&offset='
            api += (offset === '') ? 0 : offset;
            response = await axios.get(api);
        }   catch(error) {
            console.error('[reservation.services.getReservationsWithFilter]',api,error.message);
        }
        return response;
    },
}

export default ReservationService;
