import axios from "axios"

const WorkflowService = { 
    getWorkflowProcesses: async function (){
        let data = [];
        try {
            let initResponse = await axios.get('/workflow_api/scheduling_unit_flow/qa_scheduling_unit_process/?ordering=id');
            data = initResponse.data.results;
            const totalCount = initResponse.data.count;
            const initialCount = initResponse.data.results.length;
            if (initialCount < totalCount) {
                let nextResponse = await axios.get(`/workflow_api/scheduling_unit_flow/qa_scheduling_unit_process/?ordering=id&limit=${totalCount-initialCount}&offset=${initialCount}`);
                data = data.concat(nextResponse.data.results);
            }
        }   catch(error) {
            console.log(error);
        }
        return data;
    },
    /**
     * To get the workflow process object of the Scheduling Unit Blueprint
     * @param {Number} suId - Scheduling Unit Blueprint id
     * @returns - Workflow process object
     */
    getSUWorkflowProcess: async function (suId){
        let data = null;
        try {
            let response = await axios.get('/workflow_api/scheduling_unit_flow/qa_scheduling_unit_process/?su=' + suId);
            data = response.data.results.length > 0? response.data.results[0]: null;
        }   catch(error) {
            console.log(error);
        }
        return data;
    },
    /**
     * To get the workflow process objects of all the required Scheduling Unit Blueprints
     * @param {String} suId - Scheduling Unit Blueprint ids separated by comma
     * @returns - Array of Workflow process object
     */
     getSUWorkflowProcesses: async function (suId){
        let data = [];
        try {
            let url = '/workflow_api/scheduling_unit_flow/qa_scheduling_unit_process/?su=' + suId;
            let response = await axios.get(url);
            data = response.data.results;
            const totalCount = response.data.count;
            const initialCount = data.length;
            if (totalCount > initialCount) {
                url += `&offset=${initialCount}&limit=${totalCount-initialCount}`;
                let secondResponse = await axios.get(url);
                data = data.concat(secondResponse.data.results);
            }
        }   catch(error) {
            console.log(error);
        }
        return data;
    },
    getWorkflowTasks: async function (){
        let data = [];
        try {
            let initResponse = await axios.get('/workflow_api/scheduling_unit_flow/qa_scheduling_unit_task/?ordering=id');
            data = initResponse.data.results;
            const totalCount = initResponse.data.count;
            const initialCount = initResponse.data.results.length;
            if (initialCount < totalCount) {
                let nextResponse = await axios.get(`/workflow_api/scheduling_unit_flow/qa_scheduling_unit_task/?ordering=id&limit=${totalCount-initialCount}&offset=${initialCount}`);
                data = data.concat(nextResponse.data.results);
            }
        }   catch(error) {
            console.log(error);
        }
        return data;
    },
    /**
     * To get the individual workflow process details
     * @param {Number} id - Workflow process id
     * @returns Workflow process details
     */
    getWorkflowProcess: async function (id){
        let data = null;
        try {
            let initResponse = await axios.get(`/workflow_api/scheduling_unit_flow/qa_scheduling_unit_process/?id=${id}`);
            data = initResponse.data.results;
        }   catch(error) {
            console.log(error);
        }
        return data;
    },
    /** To get all related tasks of the workflow
     * @param {Number} id - Workflow process id
     * @returns Array of workflow task objects
     */
    getWorkflowProcessTasks: async function (id) {
        let data = [];
        try {
            let response = await axios.get('/workflow_api/scheduling_unit_flow/qa_scheduling_unit_task/?ordering=-id&process=' + id);
            data = response.data.results;
        }   catch(error) {
            console.log(error);
        }
        return data;
    },
    /** To get all related tasks of the workflow
     * @param {Number} id - Workflow process id
     * @returns Array of workflow task objects
     */
     getWorkflowProcessesTasks: async function (processIds) {
        let data = [];
        try {
            let url = '/workflow_api/scheduling_unit_flow/qa_scheduling_unit_task/?ordering=-id';
            for (const processId of processIds) {
                url = `${url}&process=${processId}`;
            }
            let response = await axios.get(url);
            data = response.data.results;
            const totalCount = response.data.count;
            const initialCount = data.length;
            if (totalCount > initialCount) {
                url += `&offset=${initialCount}&limit=${totalCount-initialCount}`;
                let secondResponse = await axios.get(url);
                data = data.concat(secondResponse.data.results);
            }
        }   catch(error) {
            console.log(error);
        }
        return data;
    },
    /**
     * To get the active workflow tasks with filters, sorting, limit and offset.
     * @param {String} filters - filter query
     * @param {String} orderBy - ordering field query params
     * @param {Number} limit - Number of records to fetch
     * @param {Number} offset - Number of records to offset and fetch
     * @returns Array of workflow tasks
     */
     getActiveWorkflowTasksWithFilter: async function (filters, orderBy, limit, offset){
        let response = {}
        try {
            let api = `/workflow_api/scheduling_unit_flow/qa_scheduling_unit_task/?`;
            api += (filters === '')? '' : filters+'&';
            api += (orderBy === '')? '' : orderBy+'&';
            api += 'limit='+limit+'&offset='
            api += (offset == '') ? 0 : offset;
            response = await axios.get(api);
        }    catch(error) {
            console.log(error);
        }
        return response;
    },
    getWorkflowTaskPermission: async function(id) {
        try {
            const response = await axios.get(`/workflow_api/scheduling_unit_flow/qa_scheduling_unit_task/${id}`);
            let res = {}
            res.headers = response.headers['access-control-allow-methods'];
            res.user_name = response.data.owner_username;
            return res;

        }   catch(error) {
            console.error('[workflow.services.getWorkflowTaskPermission]',error);
            return null;
        }
    },
    getWorkflowProcessPermission: async function(id) {
        try {
            const response = await axios.get(`/workflow_api/scheduling_unit_flow/qa_scheduling_unit_process/${id}`);
            const res = response.headers['access-control-allow-methods'];
            return res;

        }   catch(error) {
            console.error('[workflow.services.getWorkflowProcessPermission]',error);
            return null;
        }
    },
    updateAssignTo: async (id, toUser, data) => {
        try {
            const response = await axios.post(`/workflow_api/scheduling_unit_flow/qa_scheduling_unit_task/${id}/assign/?${toUser}`, data);
            return {status: "success", data: response.data} ;
        }   catch(error) {
            console.error('[workflow.services.updateAssignTo]',error);
            if (error.response.status === 422) {
                return { status:"error", error: error.response.statusText };
            }   else {
                return null;
            }
        }
    },
    unAssign: async function(id, data) {
        try {
            const response = await axios.post(`/workflow_api/scheduling_unit_flow/qa_scheduling_unit_task/${id}/unassign/`, data);
            return response.data;
        }   catch(error) {
            console.error('[workflow.services.updateAssignTo]',error);
            return null;
        }
    },
    reAssignTo: async function(id, toUser, data) {
        try {
            const unAssignStatus = await this.unAssign(id, data);
            if (unAssignStatus) {
                const response = this.updateAssignTo(id, toUser, data);
                return {status: "success", data: response.data} ;
            }   else {
                return null;
            }
        }   catch(error) {
            console.error('[workflow.services.updateAssignTo]',error);
            if (error.response.status === 422) {
                return { status:"error", error: error.response.statusText };
            }   else {
                return null;
            }
        }
    },
    updateQA_Perform: async (id, data) => {
        try {
            const response = await axios.post(`/workflow_api/scheduling_unit_flow/qa_scheduling_unit_process/${id}/perform/`, data);
            return response.data;
        }   catch(error) {
            console.error('[workflow.services.updateQA_Perform]',error);
            return null;
        }
    },
    getSchedulingUnitTask: async () => {
        try {
            const response = await axios.get(`/workflow_api/scheduling_unit_flow/qa_scheduling_unit_task/`);
            return response.data.results;
        }   catch(error) {
            console.error('[workflow.services.getSchedulingUnitTask]',error);
        }
    },
    getCurrentTask: async (id) => {
        let currentTask = null;
        try {
            const response = await axios.post(`/workflow_api/scheduling_unit_flow/qa_scheduling_unit_process/${id}/current_task/`);
            currentTask = response.data[0];
        }   catch(error) {
            console.error('[workflow.services.current_task]',error);
        }
        return currentTask;
    },
    getQAReportingTo: async (id) => {
        try {
            const response = await axios.get(`/workflow_api/scheduling_unit_flow/qa_reporting_to/${id}`);
            return response.data;
        }   catch(error) {
            console.error('[workflow.services.qa_reporting_to]',error);
        }
    },
    getQAReportingSOS: async (id) => {
        try {
            const response = await axios.get(`/workflow_api/scheduling_unit_flow/qa_reporting_sos/${id}`);
            return response.data;
        }   catch(error) {
            console.error('[workflow.services.qa_reporting_sos]',error);
        }
    },
    getQAPIverification: async (id) => {
        try {
            const response = await axios.get(`/workflow_api/scheduling_unit_flow/qa_pi_verification/${id}`);
            return response.data;
        }   catch(error) {
            console.error('[workflow.services.qa_pi_verification]',error);
        }
    },
    getQADecideAcceptance: async (id) => {
        try {
            const response = await axios.get(`/workflow_api/scheduling_unit_flow/qa_decide_acceptance/${id}`);
            return response.data;
        }   catch(error) {
            console.error('[workflow.services.qa_decide_acceptance]',error);
        }
    }
}

export default WorkflowService;