#!/usr/bin/env python3
import datetime
# Copyright (C) 2012-2015  ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

import time
import typing
import unittest
import uuid
from unittest import mock

import astropy.coordinates
from astropy.coordinates import Angle

from lofar.common.test_utils import skip_integration_tests
if skip_integration_tests():
    exit(3)

TEST_UUID = str(uuid.uuid1()).split('-')[0]

from lofar.common.datetimeutils import round_to_second_precision, round_to_minute_precision
from lofar.common.json_utils import get_default_json_object_for_schema, add_defaults_to_json_object_for_schema
from lofar.messaging.messagebus import TemporaryExchange, BusListenerJanitor

tmp_exchange = TemporaryExchange("t_dynamic_scheduling_%s" % (TEST_UUID,))
tmp_exchange.open()

# override DEFAULT_BUSNAME
import lofar
lofar.messaging.DEFAULT_BUSNAME = tmp_exchange.address
lofar.messaging.config.DEFAULT_BUSNAME = tmp_exchange.address

from lofar.sas.tmss.test.test_environment import TMSSTestEnvironment
tmss_test_env = TMSSTestEnvironment(exchange=tmp_exchange.address,
                                    populate_schemas=True, populate_legacy_schemas=True, populate_test_data=False,
                                    start_postgres_listener=True, start_subtask_scheduler=False,
                                    enable_viewflow=False,
                                    start_dynamic_scheduler=False)  # do not start the dynamic scheduler in the testenv, because it is the object-under-test.
tmss_test_env.start()

def tearDownModule():
    tmss_test_env.stop()
    tmp_exchange.close()

from lofar.sas.tmss.test.tmss_test_data_django_models import *
from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.tmssapp.tasks import create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft, update_task_graph_from_specifications_doc, create_scheduling_unit_draft_from_observing_strategy_template
from lofar.sas.tmss.test.test_utils import set_subtask_state_following_allowed_transitions, wait_for_scheduling_unit_blueprint_status
from lofar.sas.tmss.tmss.tmssapp.conversions import Pointing

# the module under test
from lofar.sas.tmss.services.scheduling.dynamic_scheduling import *


import logging
logger = logging.getLogger(__name__)

class BaseDynamicSchedulingTestCase(unittest.TestCase):
    '''
    Base class for all Scheduling tests, includes common setup, in-between-test-cleanup, common scheduling unit creation methods, etc
    '''
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.cycle = models.Cycle.objects.create(**Cycle_test_data(start=datetime.utcnow()-timedelta(days=1),
                                                                  stop=datetime.utcnow() + timedelta(days=7)))

        # make some re-usable projects with high/low priority
        cls.project_low = models.Project.objects.create(**Project_test_data("proj_%s_r=%s" % (str(uuid.uuid4()).split('-')[0], models.ProjectRank.LOWEST.value,), rank=models.ProjectRank.LOWEST.value))
        cls.project_medium = models.Project.objects.create(**Project_test_data("proj_%s_=%s" % (str(uuid.uuid4()).split('-')[0], models.ProjectRank.DEFAULT.value,), rank=models.ProjectRank.DEFAULT.value))
        cls.project_high = models.Project.objects.create(**Project_test_data("proj_%s_r=%s"% (str(uuid.uuid4()).split('-')[0], models.ProjectRank.HIGHEST.value,), rank=models.ProjectRank.HIGHEST.value))
        cls.cycle.projects.set([cls.project_low, cls.project_medium, cls.project_high])
        cls.cycle.save()

        cls.scheduling_set_low = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=cls.project_low))
        cls.scheduling_set_medium = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=cls.project_medium))
        cls.scheduling_set_high = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=cls.project_high))

        cls.scheduler = Scheduler()
        cls.message_handler = TMSSDynamicSchedulingMessageHandler()

    def setUp(self) -> None:
        self.clean_environment()
        self.reset_scheduling_constraints_weight_factors()

    def reset_scheduling_constraints_weight_factors(self, default_value: float=1, default_age_value: float=0):
        '''ensure all used weight_factors are set to the default_value to begin with (and allow individual tests to override them)'''
        for weight_factor in models.SchedulingConstraintsWeightFactor.objects.all():
            weight_factor.weight = default_value
            weight_factor.save()

        for key in ('age', 'unit_rank', 'project_rank', 'time.at', 'density', 'sky.transit_offset', 'sky.min_elevation', 'sky.min_distance', 'daily'):
            weight_factor, created = models.SchedulingConstraintsWeightFactor.objects.get_or_create(constraint_name=key,
                                                                                                    defaults={'weight': default_age_value if key=='age' else default_value})
            if not created:
                weight_factor.weight = default_age_value if key=='age' else default_value
                weight_factor.save()

    def _set_density_vs_optimal_weight_factor(self, value):
        # Note that placing this in setupClass will for some reason create the weight factor, and that will then
        # be picked up in the constraints checker, but the saved value will be thrown out and the default value for the
        # db field will be present.
        weight_factor, created = models.SchedulingConstraintsWeightFactor.objects.get_or_create(constraint_name="density_vs_optimal")
        weight_factor.weight = value
        weight_factor.save()

    @staticmethod
    def clean_environment():
        # wipe all scheduling_unit_drafts in between tests, so the tests don't influence each other
        tmss_test_env.delete_scheduling_unit_drafts_cascade()
        models.Reservation.objects.all().delete()
        wipe_evaluate_constraints_caches()

    @staticmethod
    def create_simple_observation_scheduling_unit(name:str=None, scheduling_set=None,
                                                  obs_duration:int=60,
                                                  constraints=None,
                                                  interrupts_telescope=False,
                                                  priority_queue=None,
                                                  unit_rank: float=0.5,
                                                  pointing: Pointing=None) -> models.SchedulingUnitDraft:
        constraints_template = models.SchedulingConstraintsTemplate.get_version_or_latest(name="constraints")
        constraints_given = constraints is not None
        constraints = add_defaults_to_json_object_for_schema(constraints or {}, constraints_template.schema)

        if not constraints_given:
            constraints['scheduler'] = 'dynamic'

        # by default, allow very loose sky constraints which can basically run anywhere
        constraints['sky']['transit_offset'] = {'from': -12 * 60 * 60, 'to': 12 * 60 * 60}
        constraints['sky']['min_elevation'] = {'target': 0, 'calibrator': 0}

        # we're testing at scheduling unit level here.
        # so, use a simple template with just one observation.
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_version_or_latest(name="Simple Observation")
        scheduling_unit_spec = add_defaults_to_json_object_for_schema(strategy_template.template,
                                                                      strategy_template.scheduling_unit_template.schema)
        scheduling_unit_spec['tasks']['Observation']['specifications_doc']['duration'] = obs_duration
        scheduling_unit_spec['tasks']['Observation']['specifications_doc']['station_configuration']['station_groups'][0]['stations'] = ['CS002', 'CS003']
        scheduling_unit_spec['tasks']['Observation']['specifications_doc']['station_configuration']['station_groups'][0]['max_nr_missing'] = 1

        if pointing is None:
            # use same name for pointing.target as for unit (convenient for debugging with plots)
            pointing = Pointing(angle1=0.6624317181687094,
                                angle2=1.5579526427549426,
                                direction_type='J2000',
                                target=name)

        scheduling_unit_spec['tasks']['Observation']['specifications_doc']['station_configuration']['SAPs'][0]['digital_pointing'] = pointing.as_dict()

        # we're testing at scheduling unit level here.
        # so, keep the unit small for speed. Disable QA subtask creation.
        scheduling_unit_spec['tasks']['Observation']['specifications_doc']['QA']['file_conversion']['enabled'] = False
        scheduling_unit_spec['tasks']['Observation']['specifications_doc']['QA']['plots']['enabled'] = False

        if priority_queue is None:
            priority_queue = models.PriorityQueueType.objects.get(value=models.PriorityQueueType.Choices.A.value)

        if scheduling_set is None:
            scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data())

        # ensure the (reused) project is active
        scheduling_set.project.project_state = models.ProjectState.objects.get(value=models.ProjectState.Choices.ACTIVE.value)
        scheduling_set.project.save()

        # ensure the (reused) project has a cycle
        if not scheduling_set.project.cycles.exists():
            scheduling_set.project.cycles.add(models.Cycle.objects.create(**Cycle_test_data(start=datetime.utcnow()-timedelta(days=7),
                                                                                            stop=datetime.utcnow()+timedelta(days=7))))

        # add the scheduling_unit_doc to a new SchedulingUnitDraft instance, and were ready to use it!
        scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(name=name,
                                                                          scheduling_set=scheduling_set,
                                                                          specifications_template=strategy_template.scheduling_unit_template,
                                                                          observation_strategy_template=strategy_template,
                                                                          scheduling_constraints_doc=constraints,
                                                                          scheduling_constraints_template=constraints_template,
                                                                          interrupts_telescope=interrupts_telescope,
                                                                          priority_queue=priority_queue,
                                                                          rank=unit_rank)

        scheduling_unit_spec['scheduling_constraints_doc'] = constraints
        scheduling_unit_draft = update_task_graph_from_specifications_doc(scheduling_unit_draft, scheduling_unit_spec)
        return scheduling_unit_draft

    @staticmethod
    def create_station_reservation(name, stations: [], start_time: datetime=datetime(2100, 1, 1, 0, 0, 0),
                                                       stop_time: datetime=datetime(2100, 1, 2, 0, 0, 0)):
        """
        Create a station reservation with given list of stations, start_time and stop_time (optional)
        Default duration is 24 hours
        """
        # create an overlapping reservation
        reservation_template = models.ReservationTemplate.objects.first() # there is only one
        reservation_doc = reservation_template.get_default_json_document_for_schema()
        reservation_doc['resources']['stations'] = stations
        return models.Reservation.objects.create(name=name,
                                                 start_time=start_time,
                                                 stop_time=stop_time,
                                                 specifications_doc=reservation_doc,
                                                 specifications_template=reservation_template,
                                                 description="Station reservation for testing")

    @staticmethod
    def _clear_constraints(scheduling_unit_draft: models.SchedulingUnitDraft):
        '''helper function to quickly clear the scheduling_constraints_doc of the given scheduling_unit_draft,
        so we can fill it later with our explicit constraints per test.'''
        scheduling_unit_draft.scheduling_constraints_doc['sky'] = {}
        scheduling_unit_draft.scheduling_constraints_doc['time'] = {"between": [], "not_between": [] }


class TestFixedTimeScheduling(BaseDynamicSchedulingTestCase):
    '''Test the behaviour of the dynamic scheduler for scheduling_units with a 'fixed_time' 'at' constraint'''
    @staticmethod
    def create_simple_observation_scheduling_unit_fixed_time(name:str=None, scheduling_set=None,
                                                             at: datetime=None,
                                                             obs_duration: typing.Union[int, timedelta]=None,
                                                             unit_rank: float=0.5) -> models.SchedulingUnitDraft:
        if name is None:
            name = 'scheduling_unit for fixed_time at constraint'

        if at is None:
            at = round_to_second_precision(datetime.utcnow())

        if obs_duration is None:
            obs_duration = timedelta(minutes=10)

        if isinstance(obs_duration, timedelta):
            obs_duration = obs_duration.total_seconds()

        scheduling_unit_draft = BaseDynamicSchedulingTestCase.create_simple_observation_scheduling_unit(name, scheduling_set=scheduling_set, obs_duration=obs_duration, unit_rank=unit_rank)
        BaseDynamicSchedulingTestCase._clear_constraints(scheduling_unit_draft)
        scheduling_unit_draft.scheduling_constraints_doc['scheduler'] = 'fixed_time'
        scheduling_unit_draft.scheduling_constraints_doc['time']['at'] = at.isoformat()
        scheduling_unit_draft.save()
        return scheduling_unit_draft

    def test_fixed_time_in_the_future_can_be_scheduled(self):
        """
        Test a simple observation with the 'at' constraint and 'fixed_time' scheduler.
        The 'at' timestamp in the future should make the scheduling_unit schedulable, and then scheduled.
        """
        # Use at constraint in the future (noon, tomorrow)
        at = round_to_second_precision(datetime.utcnow().replace(hour=12, minute=0, second=0, microsecond=0) + timedelta(days=1))
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit_fixed_time(at=at, unit_rank=0.5)
        scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        # assert blueprint has correct constraints, and is schedulable
        self.assertEqual('fixed_time', scheduling_unit_blueprint.scheduling_constraints_doc['scheduler'])
        self.assertEqual(at.isoformat(), scheduling_unit_blueprint.scheduling_constraints_doc['time']['at'])
        self.assertEqual(scheduling_unit_blueprint.status.value, models.SchedulingUnitStatus.Choices.SCHEDULABLE.value)

        # try the method-under-test: schedule_fixed_time_scheduling_units
        self.scheduler.schedule_fixed_time_scheduling_units()

        # Assert the scheduling_unit has been scheduled
        scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint.status.value)
        self.assertEqual(at, scheduling_unit_blueprint.scheduled_start_time)

        # extra check (which is not part of dynamic/fixed scheduling)
        # we want the on_sky_start_time to have the same value as the scheduled_start_time  when scheduled.
        self.assertEqual(scheduling_unit_blueprint.scheduled_start_time, scheduling_unit_blueprint.on_sky_start_time)


    def test_fixed_time_unit_can_be_scheduled_between_now_and_future_scheduled_unit(self):
        """
        Test a simple observation with the 'at' constraint and 'fixed_time' scheduler.
        It should be scheduled in between 'now' and an already scheduled unit.
        """
        # re-use scheduled unit from test_fixed_time_in_the_future_can_be_scheduled
        # assume that test_fixed_time_in_the_future_can_be_scheduled passes.
        self.test_fixed_time_in_the_future_can_be_scheduled()
        # as a result, there should be one scheduled scheduling_unit
        self.assertEqual(1, models.SchedulingUnitBlueprint.objects.count())
        scheduling_unit_blueprint1 = models.SchedulingUnitBlueprint.objects.all()[0]
        self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint1.status.value)

        # squeeze in a 2nd unit, in between now and the 1st one.
        at = round_to_second_precision(datetime.utcnow() + (scheduling_unit_blueprint1.scheduled_start_time-datetime.utcnow())/2)
        scheduling_unit_draft2 = self.create_simple_observation_scheduling_unit_fixed_time(at=at)
        scheduling_unit_blueprint2 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft2)

        # assert blueprint has correct constraints, and is schedulable
        self.assertEqual('fixed_time', scheduling_unit_blueprint2.scheduling_constraints_doc['scheduler'])
        self.assertEqual(at.isoformat(), scheduling_unit_blueprint2.scheduling_constraints_doc['time']['at'])
        self.assertEqual(scheduling_unit_blueprint2.status.value, models.SchedulingUnitStatus.Choices.SCHEDULABLE.value)

        # try the method-under-test: schedule_fixed_time_scheduling_units
        self.scheduler.schedule_fixed_time_scheduling_units()

        # Assert the scheduling_unit has been scheduled
        scheduling_unit_blueprint2.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint2.status.value)
        self.assertEqual(at, scheduling_unit_blueprint2.scheduled_start_time)
        self.assertGreater(scheduling_unit_blueprint2.scheduled_start_time, datetime.utcnow())
        self.assertLess(scheduling_unit_blueprint2.scheduled_start_time, scheduling_unit_blueprint1.scheduled_start_time)


    def test_fixed_time_unit_can_be_scheduled_after_future_scheduled_unit(self):
        """
        Test a simple observation with the 'at' constraint and 'fixed_time' scheduler.
        It should be scheduled after an already scheduled unit.
        """
        # re-use scheduled unit from test_fixed_time_in_the_future_can_be_scheduled
        # assume that test_fixed_time_in_the_future_can_be_scheduled passes.
        self.test_fixed_time_in_the_future_can_be_scheduled()
        # as a result, there should be one scheduled scheduling_unit
        self.assertEqual(1, models.SchedulingUnitBlueprint.objects.count())
        scheduling_unit_blueprint1 = models.SchedulingUnitBlueprint.objects.all()[0]
        self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint1.status.value)

        # add a 2nd unit, a day after the 1st already scheduled unit
        at = round_to_second_precision(scheduling_unit_blueprint1.scheduled_observation_stop_time + timedelta(days=1))
        scheduling_unit_draft2 = self.create_simple_observation_scheduling_unit_fixed_time(at=at)
        scheduling_unit_blueprint2 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft2)

        # assert blueprint has correct constraints, and is schedulable
        self.assertEqual('fixed_time', scheduling_unit_blueprint2.scheduling_constraints_doc['scheduler'])
        self.assertEqual(at.isoformat(), scheduling_unit_blueprint2.scheduling_constraints_doc['time']['at'])
        self.assertEqual(scheduling_unit_blueprint2.status.value, models.SchedulingUnitStatus.Choices.SCHEDULABLE.value)

        # try the method-under-test: schedule_fixed_time_scheduling_units
        self.scheduler.schedule_fixed_time_scheduling_units()

        # Assert the scheduling_unit has been scheduled
        scheduling_unit_blueprint2.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint2.status.value)
        self.assertEqual(at, scheduling_unit_blueprint2.scheduled_start_time)
        self.assertGreater(scheduling_unit_blueprint2.scheduled_start_time, scheduling_unit_blueprint1.scheduled_observation_stop_time)


    def test_fixed_time_unit_cannot_be_scheduled_overlapping_a_future_scheduled_unit(self):
        """
        Test a simple observation with the 'at' constraint and 'fixed_time' scheduler.
        It should not be schedulable when overlapping an already scheduled unit.
        """
        # set all weight factors to 0, except for scheduling_rank so we can control which unit wins
        self.reset_scheduling_constraints_weight_factors(default_value=0, default_age_value=0)
        weight_factor, created = models.SchedulingConstraintsWeightFactor.objects.get_or_create(constraint_name="unit_rank")
        weight_factor.weight = 1
        weight_factor.save()

        # re-use scheduled unit from test_fixed_time_in_the_future_can_be_scheduled
        # assume that test_fixed_time_in_the_future_can_be_scheduled passes.
        self.test_fixed_time_in_the_future_can_be_scheduled()
        # as a result, there should be one scheduled scheduling_unit
        self.assertEqual(1, models.SchedulingUnitBlueprint.objects.count())
        scheduling_unit_blueprint1 = models.SchedulingUnitBlueprint.objects.all()[0]
        self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint1.status.value)

        # test three different start_times for a 2nd unit, each overlapping the 1st scheduled unit
        duration = scheduling_unit_blueprint1.specified_main_observation_duration
        overlapping_start_times = (scheduling_unit_blueprint1.scheduled_start_time - duration + timedelta(seconds=1), # tail just touches head
                                   scheduling_unit_blueprint1.scheduled_start_time, # full overlap
                                   scheduling_unit_blueprint1.scheduled_start_time + duration - timedelta(seconds=1)) # head just touches tail

        for at in overlapping_start_times:
            scheduling_unit_draft2 = self.create_simple_observation_scheduling_unit_fixed_time(at=at, obs_duration=duration, unit_rank=1)
            scheduling_unit_blueprint2 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft2)

            # assert blueprint has correct constraints, and is schedulable
            self.assertEqual('fixed_time', scheduling_unit_blueprint2.scheduling_constraints_doc['scheduler'])
            self.assertEqual(at.isoformat(), scheduling_unit_blueprint2.scheduling_constraints_doc['time']['at'])
            self.assertEqual(scheduling_unit_blueprint2.status.value, models.SchedulingUnitStatus.Choices.SCHEDULABLE.value)

            # try the method-under-test: schedule_fixed_time_scheduling_units
            self.scheduler.schedule_fixed_time_scheduling_units()

            # Assert the scheduling_unit has been marked as unschedulable
            scheduling_unit_blueprint2.refresh_from_db()
            self.assertEqual(models.SchedulingUnitStatus.Choices.UNSCHEDULABLE.value, scheduling_unit_blueprint2.status.value)
            # even though it's unschedulable, it should still be positioned at the requested 'at' time (in the past)
            self.assertEqual(at, scheduling_unit_blueprint2.scheduled_start_time)


    def test_fixed_time_with_blocking_other_constraints_is_unschedulable(self):
        """
        Test a simple observation with the 'at' constraint and 'fixed_time' scheduler.
        Prove that it can be scheduled with no other constraints, and becomes unschedulable for blocking other contraints
        """
        # re-use scheduled unit from test_fixed_time_in_the_future_can_be_scheduled
        # assume that test_fixed_time_in_the_future_can_be_scheduled passes (check it!)
        self.test_fixed_time_in_the_future_can_be_scheduled()
        # as a result, there should be one scheduled scheduling_unit
        self.assertEqual(1, models.SchedulingUnitBlueprint.objects.count())
        scheduling_unit_blueprint = models.SchedulingUnitBlueprint.objects.all()[0]
        self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint.status.value)
        at = round_to_second_precision(scheduling_unit_blueprint.scheduled_start_time)

        # unschedule it, so we can schedule it again with new daily constraints
        unschedule_subtasks_in_scheduling_unit_blueprint(scheduling_unit_blueprint)
        # assert blueprint has correct constraints, and is schedulable again
        self.assertEqual('fixed_time', scheduling_unit_blueprint.scheduling_constraints_doc['scheduler'])
        self.assertEqual(at.isoformat(), scheduling_unit_blueprint.scheduling_constraints_doc['time']['at'])
        self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULABLE.value, scheduling_unit_blueprint.status.value)

        # set impossible constraints, require night while 'at' is during day
        scheduling_unit_blueprint.scheduling_constraints_doc['daily'] = { 'require_night': True}
        scheduling_unit_blueprint.save()

        wipe_evaluate_constraints_caches()

        # try the method-under-test: schedule_fixed_time_scheduling_units
        self.scheduler.schedule_fixed_time_scheduling_units()

        # Assert the scheduling_unit has been marked as unschedulable
        scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.UNSCHEDULABLE.value, scheduling_unit_blueprint.status.value)
        # even though it's unschedulable, it should still be positioned at the requested 'at' time
        self.assertEqual(at, scheduling_unit_blueprint.scheduled_start_time)


    def test_fixed_time_in_the_past_makes_unit_unschedulable(self):
        """
        Test a simple observation with the 'at' constraint and 'fixed_time' scheduler.
        The 'at' timestamp in the past should make the scheduling_unit unschedulable.
        """
        # Use at constraint in the past
        at = round_to_second_precision(datetime.utcnow() - timedelta(days=1))
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit_fixed_time(at=at)
        scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        # assert blueprint has correct constraints, and is schedulable
        self.assertEqual('fixed_time', scheduling_unit_blueprint.scheduling_constraints_doc['scheduler'])
        self.assertEqual(at.isoformat(), scheduling_unit_blueprint.scheduling_constraints_doc['time']['at'])
        self.assertEqual(scheduling_unit_blueprint.status.value, models.SchedulingUnitStatus.Choices.SCHEDULABLE.value)

        # try the method-under-test: schedule_fixed_time_scheduling_units
        self.scheduler.schedule_fixed_time_scheduling_units()

        # Assert the scheduling_unit has been marked as unschedulable
        scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.UNSCHEDULABLE.value, scheduling_unit_blueprint.status.value)
        # even though it's unschedulable, it should still be positioned at the requested 'at' time (in the past)
        self.assertEqual(at, scheduling_unit_blueprint.scheduled_start_time)

    def test_bug_fix_TMSS_2296_a_fixed_time_unit_blocked_by_another_should_become_unschedulable(self):
        """
        Test to reproduce the reported bug that a fixed_time unit is not made unschedulable when blocked by another.
        See: https://support.astron.nl/jira/browse/TMSS-2296
        """
        at = round_to_second_precision(datetime.utcnow() + timedelta(days=1))
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit_fixed_time(at=at)
        scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
        scheduling_unit_blueprint.rank = models.SchedulingUnitRank.HIGHEST.value
        scheduling_unit_blueprint.save()

        # assert blueprint has correct constraints, and is schedulable
        self.assertEqual('fixed_time', scheduling_unit_blueprint.scheduling_constraints_doc['scheduler'])
        self.assertEqual(at.isoformat(), scheduling_unit_blueprint.scheduling_constraints_doc['time']['at'])
        self.assertEqual(scheduling_unit_blueprint.status.value, models.SchedulingUnitStatus.Choices.SCHEDULABLE.value)

        self.scheduler.schedule_fixed_time_scheduling_units()

        # Assert the scheduling_unit has been scheduled
        scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint.status.value)
        self.assertEqual(at, scheduling_unit_blueprint.scheduled_start_time)

        # create a second unit, at the same time and try to schedule it.
        # that should not be possible, and the unit should become unschedulable.
        scheduling_unit_draft2 = self.create_simple_observation_scheduling_unit_fixed_time(at=at)
        scheduling_unit_blueprint2 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft2)
        scheduling_unit_blueprint2.rank = models.SchedulingUnitRank.LOWEST.value
        scheduling_unit_blueprint2.save()

        # assert blueprint has correct constraints, and is schedulable initially...
        self.assertEqual('fixed_time', scheduling_unit_blueprint2.scheduling_constraints_doc['scheduler'])
        self.assertEqual(at.isoformat(), scheduling_unit_blueprint2.scheduling_constraints_doc['time']['at'])
        self.assertEqual(scheduling_unit_blueprint2.status.value, models.SchedulingUnitStatus.Choices.SCHEDULABLE.value)

        # try to schedule the second unit...
        self.scheduler.schedule_fixed_time_scheduling_units()

        # Assert the scheduling_unit has been NOT scheduled (but is unschedulable)
        scheduling_unit_blueprint2.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.UNSCHEDULABLE.value, scheduling_unit_blueprint2.status.value)
        expected_msg = "Scheduling unit id=%s is blocking this unit from being scheduled" % (scheduling_unit_blueprint.id,)
        self.assertEqual(expected_msg, scheduling_unit_blueprint2.unschedulable_reason)
        self.assertEqual(at, scheduling_unit_blueprint2.scheduled_start_time)

        # Assert the original scheduling_unit is still scheduled
        scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint.status.value)
        self.assertEqual(at, scheduling_unit_blueprint.scheduled_start_time)

    def test_bug_fix_TMSS_2302_unschedulable_reason_too_generic(self):
        """
        Test to reproduce the reported bug that a fixed_time unit does not state a specific unschedulable reason.
        See: https://support.astron.nl/jira/browse/TMSS-2302
        We are using the specs from https://tmss.lofar.eu/schedulingunit/view/blueprint/2004 which had troubles being scheduled.
        """
        project = models.Project.objects.create(**Project_test_data(name=str(uuid.uuid4()), auto_ingest=True,
                                                                    project_state=models.ProjectState.objects.get(value=models.ProjectState.Choices.ACTIVE.value)))

        at = datetime(2023, 3, 30, 18, 40)
        cycle = models.Cycle.objects.create(name=str(uuid.uuid4()),
                                            start=at - timedelta(days=1),
                                            stop=datetime.utcnow() + timedelta(days=7))
        cycle.projects.set([project])
        cycle.save()

        scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project))

        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="IM HBA LoTSS - 2 Beams")
        su_draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template, scheduling_set=scheduling_set,
                                                                                 specifications_doc_overrides={
                                                                                     "scheduling_constraints_doc": {
                                                                                         "daily": {
                                                                                             "avoid_twilight": False,
                                                                                              "require_day": False,
                                                                                              "require_night": False
                                                                                            },
                                                                                            "scheduler": "fixed_time",
                                                                                            "location": "both",
                                                                                            "sky": {
                                                                                              "min_distance": {
                                                                                                "jupiter": 0.2617993877991494,
                                                                                                "moon": 0.3490658503988659,
                                                                                                "sun": 0.5235987755982988
                                                                                              },
                                                                                              "min_elevation": {
                                                                                                "calibrator": 0.5235987755982988,
                                                                                                "target": 0.5235987755982988
                                                                                              },
                                                                                              "transit_offset": {
                                                                                                "from": -3600,
                                                                                                "to": 3600
                                                                                              }
                                                                                            },
                                                                                            "time": {
                                                                                              "at": at.isoformat()
                                                                                            }
                                                                                          },
                                                                                          "tasks": {
                                                                                            "Calibrator Observation 1": {
                                                                                              "short_description": "Cal 3C196",
                                                                                              "specifications_doc": {
                                                                                                "calibrator": {
                                                                                                  "pointing": {
                                                                                                    "angle1": 2.1537413875046543,
                                                                                                    "angle2": 0.8415517080119602,
                                                                                                    "direction_type": "J2000",
                                                                                                    "target": "3C196"
                                                                                                  }
                                                                                                }
                                                                                              }
                                                                                            },
                                                                                            "Calibrator Observation 2": {
                                                                                              "short_description": "Cal 3C295",
                                                                                              "specifications_doc": {
                                                                                                "calibrator": {
                                                                                                  "pointing": {
                                                                                                    "angle1": 3.714680167406339,
                                                                                                    "angle2": 0.9111102054650467,
                                                                                                    "direction_type": "J2000",
                                                                                                    "target": "3C295"
                                                                                                  }
                                                                                                }
                                                                                              }
                                                                                            },
                                                                                            "Calibrator Pipeline 1": {
                                                                                              "short_description": "Cal1 3C196/PP"
                                                                                            },
                                                                                            "Calibrator Pipeline 2": {
                                                                                              "short_description": "Cal2 3C295/PP"
                                                                                            },
                                                                                            "Pipeline target1": {
                                                                                              "short_description": "COSMOS/TP"
                                                                                            },
                                                                                            "Pipeline target2": {
                                                                                              "short_description": "P145+04/TP"
                                                                                            },
                                                                                            "Target Observation": {
                                                                                              "short_description": "COSMOS_P145+04",
                                                                                              "specifications_doc": {
                                                                                                "duration": 14400,
                                                                                                "station_configuration": {
                                                                                                  "SAPs": [
                                                                                                    {
                                                                                                      "digital_pointing": {
                                                                                                        "angle1": 2.620073728683454,
                                                                                                        "angle2": 0.038499054416908256,
                                                                                                        "direction_type": "J2000",
                                                                                                        "target": "COSMOS"
                                                                                                      }
                                                                                                    },
                                                                                                    {
                                                                                                      "digital_pointing": {
                                                                                                        "angle1": 2.5359372232089807,
                                                                                                        "angle2": 0.07651038627062032,
                                                                                                        "direction_type": "J2000",
                                                                                                        "target": "P145+04"
                                                                                                      }
                                                                                                    }
                                                                                                  ],
                                                                                                    "station_groups": [
                                                                                                        {
                                                                                                            "max_nr_missing": 4,
                                                                                                            "stations": [ "CS002", "CS003", "CS004", "CS005", "CS006", "CS007", "CS011", "CS013", "CS017", "CS021", "CS024", "CS026", "CS028", "CS030", "CS031", "CS032", "CS101", "CS103", "CS201", "CS301", "CS302", "CS401", "CS501", "RS106", "RS205", "RS208", "RS210", "RS305", "RS306", "RS307", "RS310", "RS406", "RS407", "RS409", "RS503", "RS508", "RS509"]
                                                                                                        },
                                                                                                        {
                                                                                                            "max_nr_missing": 2,
                                                                                                            "stations": ["DE601", "DE602", "DE603", "DE604", "DE605", "DE609", "FR606", "SE607", "UK608", "PL610", "PL611", "PL612", "IE613", "LV614"]
                                                                                                        },
                                                                                                        {
                                                                                                            "max_nr_missing": 1,
                                                                                                            "stations": ["DE601", "DE605"]
                                                                                                        }
                                                                                                    ],
                                                                                                  "tile_beam": {
                                                                                                    "angle1": 2.578006639499052,
                                                                                                    "angle2": 0.05750510819470917,
                                                                                                    "direction_type": "J2000",
                                                                                                    "target": "COSMOSP145+04REF"
                                                                                                  }
                                                                                                }
                                                                                              }
                                                                                            }
                                                                                          }
                                                                                        })
        scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(su_draft)

        # assert blueprint has correct constraints, and is schedulable
        self.assertEqual('fixed_time', scheduling_unit_blueprint.scheduling_constraints_doc['scheduler'])
        self.assertEqual(at.isoformat(), scheduling_unit_blueprint.scheduling_constraints_doc['time']['at'])
        self.assertEqual(scheduling_unit_blueprint.status.value, models.SchedulingUnitStatus.Choices.SCHEDULABLE.value)

        self.scheduler.schedule_fixed_time_scheduling_units()

        # Assert the scheduling_unit is unschedulable, and has a specific unschedulable_reason
        scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.UNSCHEDULABLE.value, scheduling_unit_blueprint.status.value)

        # the bug used to yield a generic reason: constraint time.at='2023-03-30 18:40:00' falls outside of window ['2023-03-30 18:40:00', '2023-03-30 22:40:00']
        self.assertNotIn('falls outside of window', scheduling_unit_blueprint.unschedulable_reason)
        # assert that the bug is fixed and yields a specific unschedulable_reason
        self.assertIn('sky min_elevation, sky transit_offset constraints are not met', scheduling_unit_blueprint.unschedulable_reason)


    def test_bug_fix_TMSS_2496_reserved_stations_are_not_used_for_constraints(self):
        """
        See: https://support.astron.nl/jira/browse/TMSS-2496
        """
        project = models.Project.objects.create(**Project_test_data(name=str(uuid.uuid4()), auto_ingest=True,
                                                                    project_state=models.ProjectState.objects.get(value=models.ProjectState.Choices.ACTIVE.value)))

        at = datetime(2030, 5, 1, 20, 0)
        cycle = models.Cycle.objects.create(name=str(uuid.uuid4()),
                                            start=at - timedelta(days=1),
                                            stop=at + timedelta(days=7))
        cycle.projects.set([project])
        cycle.save()

        scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project))

        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="Simple Observation")
        su_draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template, scheduling_set=scheduling_set,
                                                                                 specifications_doc_overrides={
                                                                                     "scheduling_constraints_doc": {
                                                                                         "daily": {
                                                                                             "avoid_twilight": False,
                                                                                              "require_day": False,
                                                                                              "require_night": False
                                                                                            },
                                                                                            "scheduler": "fixed_time",
                                                                                            "location": "both",
                                                                                            "sky": {
                                                                                              "min_elevation": {
                                                                                                "target": 0.5235987755982988
                                                                                              },
                                                                                              "transit_offset": {
                                                                                                "from": -14400,
                                                                                                "to": 14400
                                                                                              }
                                                                                            },
                                                                                            "time": {
                                                                                              "at": at.isoformat()
                                                                                            }
                                                                                          },
                                                                                          "tasks": {
                                                                                            "Observation": {
                                                                                              "specifications_doc": {
                                                                                                "duration": 3600,
                                                                                                "station_configuration": {
                                                                                                  "SAPs": [
                                                                                                    {
                                                                                                      "digital_pointing": {
                                                                                                        "angle1": 2.5359372232089807,
                                                                                                        "angle2": 0.07651038627062032,
                                                                                                        "direction_type": "J2000",
                                                                                                        "target": "P145+04"
                                                                                                      }
                                                                                                    }
                                                                                                  ],
                                                                                                  "station_groups": [
                                                                                                  {
                                                                                                      "max_nr_missing": 1,
                                                                                                      "stations": [ "FR606", "LV614"]
                                                                                                  }
                                                                                                ]
                                                                                                }
                                                                                              }
                                                                                            }
                                                                                          }
                                                                                        })
        scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(su_draft)

        # assert blueprint has correct constraints, and is schedulable
        self.assertEqual('fixed_time', scheduling_unit_blueprint.scheduling_constraints_doc['scheduler'])
        self.assertEqual(at.isoformat(), scheduling_unit_blueprint.scheduling_constraints_doc['time']['at'])
        self.assertEqual(scheduling_unit_blueprint.status.value, models.SchedulingUnitStatus.Choices.SCHEDULABLE.value)

        self.scheduler.schedule_fixed_time_scheduling_units()

        # Assert the scheduling_unit is unschedulable, and unschedulable_reason says constraint for LV614
        scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.UNSCHEDULABLE.value, scheduling_unit_blueprint.status.value)
        self.assertIn('sky min_elevation constraint is not met', scheduling_unit_blueprint.unschedulable_reason)
        self.assertIn('LV614', scheduling_unit_blueprint.unschedulable_reason)

        # make it schedulable by reserving the LV614 station, which is then taken out, thus the constraints should be met.
        reservation_template = models.ReservationTemplate.get_latest('reservation')
        reservation_doc = reservation_template.get_default_json_document_for_schema()
        reservation_doc['resources']['stations'] = ['LV614']
        reservation_doc['schedulability']['fixed_time'] = False
        models.Reservation.objects.create(start_time=at,
                                          stop_time=at+timedelta(days=1),
                                          specifications_doc=reservation_doc,
                                          specifications_template=reservation_template,
                                          description='')

        wipe_evaluate_constraints_caches()
        mark_subtasks_in_scheduling_unit_blueprint_as_schedulable(scheduling_unit_blueprint)
        self.scheduler.schedule_fixed_time_scheduling_units()
        scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint.status.value)


    def test_bug_fix_TMSS_2520_inter_observation_gap(self):
        """
        See: https://support.astron.nl/jira/browse/TMSS-2520
        When scheduling an observation you can schedule behind an observation (1 minute)
        But you can not schedule 1 minute before an already scheduled observation, when increased to 1 min and 1 sec and it does work.
        """
        project = models.Project.objects.create(**Project_test_data(name=str(uuid.uuid4()), auto_ingest=True,
                                                                    project_state=models.ProjectState.objects.get(value=models.ProjectState.Choices.ACTIVE.value)))

        at = datetime(2030, 5, 2, 12, 0)
        cycle = models.Cycle.objects.create(name=str(uuid.uuid4()),
                                            start=at - timedelta(days=1),
                                            stop=at + timedelta(days=7))
        cycle.projects.set([project])
        cycle.save()

        scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project))

        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="Simple Observation")
        su_draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template,
                                                                                 scheduling_set=scheduling_set,
                                                                                 specifications_doc_overrides={
                                                                                     "scheduling_constraints_doc": {
                                                                                         "scheduler": "fixed_time",
                                                                                         "time" : {"at": at.isoformat()},
                                                                                         "sky": {
                                                                                             "min_elevation": {"target": 0},
                                                                                             "transit_offset": {
                                                                                                 "from": -28800,
                                                                                                 "to": 28800
                                                                                             }
                                                                                         }
                                                                                     }
                                                                                 })

        sub1 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(su_draft)
        sub1.scheduling_constraints_doc['time']['at'] = "2030-05-02T12:00:00Z"
        sub1.save()

        scheduled_units = self.scheduler.schedule_fixed_time_scheduling_units()
        self.assertEqual(1, len(scheduled_units))
        sub1.refresh_from_db()
        self.assertEqual(sub1.status.value, models.SchedulingUnitStatus.Choices.SCHEDULED.value)

        # create a 2nd unit
        sub2 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(su_draft)

        # confirm the 2nd unit can be scheduled 1 minute after the 1st
        sub2.scheduling_constraints_doc['time']['at'] = (sub1.scheduled_observation_stop_time + timedelta(minutes=1)).isoformat()
        sub2.save()
        scheduled_units = self.scheduler.schedule_fixed_time_scheduling_units()
        self.assertEqual(1, len(scheduled_units))
        sub2.refresh_from_db()
        self.assertEqual(sub2.status.value, models.SchedulingUnitStatus.Choices.SCHEDULED.value)

        # test the bug fix: unschedule and confirm the 2nd unit can be scheduled ending 1 minute before the 1st
        sub2 = unschedule_subtasks_in_scheduling_unit_blueprint(sub2)
        sub2.scheduling_constraints_doc['time']['at'] = (sub1.scheduled_observation_start_time - timedelta(minutes=1) - sub2.specified_observation_duration).isoformat()
        sub2.save()
        scheduled_units = self.scheduler.schedule_fixed_time_scheduling_units()
        self.assertEqual(1, len(scheduled_units))
        sub2.refresh_from_db()
        self.assertEqual(sub2.status.value, models.SchedulingUnitStatus.Choices.SCHEDULED.value)


    def test_exception_during_scheduling_sets_state_to_unschedulable(self):
        """
        Test a simple observation with the 'at' constraint and 'fixed_time' scheduler.
        When an exception is raised during scheduling, the state should go to unschedulable so the user can retry/fix.
        """
        # Use at constraint
        at = round_to_second_precision(datetime.utcnow() + timedelta(days=1))
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit_fixed_time(at=at)
        scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        # assert blueprint has correct constraints, and is schedulable
        self.assertEqual('fixed_time', scheduling_unit_blueprint.scheduling_constraints_doc['scheduler'])
        self.assertEqual(at.isoformat(), scheduling_unit_blueprint.scheduling_constraints_doc['time']['at'])
        self.assertEqual(scheduling_unit_blueprint.status.value, models.SchedulingUnitStatus.Choices.SCHEDULABLE.value)

        # mock the schedule_subtask method, make it raise an Exception
        with mock.patch('lofar.sas.tmss.tmss.tmssapp.subtasks.schedule_subtask') as mocked:
            def mocked_schedule_subtask(*args, **kwargs):
                raise Exception("test-exception with test-unschedulable-reason")
            mocked.side_effect = mocked_schedule_subtask

            # try the method-under-test: schedule_fixed_time_scheduling_units
            # via-via this calls the schedule_subtask, which is now raising an exception in the mocked_schedule_subtask
            self.scheduler.schedule_fixed_time_scheduling_units()

        # Assert the scheduling_unit has been marked as unschedulable
        scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.UNSCHEDULABLE.value, scheduling_unit_blueprint.status.value)
        # even though it's unschedulable, it should still be positioned at the requested 'at' time
        self.assertTrue("test-exception with test-unschedulable-reason" in scheduling_unit_blueprint.unschedulable_reason)
        # even though it's unschedulable, it should still be positioned at the requested 'at' time
        self.assertEqual(at, scheduling_unit_blueprint.scheduled_start_time)


    def test_fixed_time_is_only_schedulable_for_active_projects(self):
        """
        Test a simple observation with the 'at' constraint and 'fixed_time' scheduler.
        """
        # Use at constraint in the near future
        at = round_to_second_precision(datetime.utcnow() + timedelta(days=1))
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit_fixed_time(at=at)
        scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        # assert blueprint has correct constraints, and is schedulable
        self.assertEqual('fixed_time', scheduling_unit_blueprint.scheduling_constraints_doc['scheduler'])
        self.assertEqual(at.isoformat(), scheduling_unit_blueprint.scheduling_constraints_doc['time']['at'])
        self.assertEqual(scheduling_unit_blueprint.status.value, models.SchedulingUnitStatus.Choices.SCHEDULABLE.value)

        active_project_state = models.ProjectState.objects.get(value=models.ProjectState.Choices.ACTIVE.value)
        non_active_project_states = models.ProjectState.objects.exclude(value=models.ProjectState.Choices.ACTIVE.value).all()

        self.assertIsNotNone(active_project_state)
        self.assertTrue(len(non_active_project_states) > 0)

        for state in non_active_project_states:
            # make sure the unit is schedulable to start with
            mark_subtasks_in_scheduling_unit_blueprint_as_schedulable(scheduling_unit_blueprint)
            self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULABLE.value, scheduling_unit_blueprint.status.value)

            # set the project's state
            scheduling_unit_blueprint.project.project_state = state
            scheduling_unit_blueprint.project.save()

            # try the method-under-test: schedule_fixed_time_scheduling_units
            self.scheduler.schedule_fixed_time_scheduling_units()

            # Assert the scheduling_unit has been marked as unschedulable
            scheduling_unit_blueprint.refresh_from_db()
            self.assertEqual(models.SchedulingUnitStatus.Choices.UNSCHEDULABLE.value, scheduling_unit_blueprint.status.value)
            # even though it's unschedulable, it should still be positioned at the requested 'at' time (in the past)
            self.assertEqual(at, scheduling_unit_blueprint.scheduled_start_time)

        # now prove that it can be scheduled for project in active state
        scheduling_unit_blueprint.project.project_state = active_project_state
        scheduling_unit_blueprint.project.save()

        # make sure the unit is schedulable to start with
        mark_subtasks_in_scheduling_unit_blueprint_as_schedulable(scheduling_unit_blueprint)
        self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULABLE.value, scheduling_unit_blueprint.status.value)

        # try the method-under-test: schedule_fixed_time_scheduling_units
        self.scheduler.schedule_fixed_time_scheduling_units()

        # Assert the scheduling_unit has been scheduled at the requested time
        scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint.status.value)
        self.assertEqual(at, scheduling_unit_blueprint.scheduled_start_time)


    def test_project_activation_deactivation_event_handling(self):
        """
        Test a simple observation with the 'at' constraint and 'fixed_time' scheduler.
        First ensure it is unschedulable by making the project suspended.
        Then active the project, and see if the scheduler picks it up and schedules it.
        """
        # Use at constraint in the near future
        at = round_to_second_precision(datetime.utcnow() + timedelta(days=1))
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit_fixed_time(at=at)
        scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        # assert blueprint has correct constraints, and is schedulable
        self.assertEqual('fixed_time', scheduling_unit_blueprint.scheduling_constraints_doc['scheduler'])
        self.assertEqual(at.isoformat(), scheduling_unit_blueprint.scheduling_constraints_doc['time']['at'])
        self.assertEqual(scheduling_unit_blueprint.status.value, models.SchedulingUnitStatus.Choices.SCHEDULABLE.value)

        active_project_state = models.ProjectState.objects.get(value=models.ProjectState.Choices.ACTIVE.value)
        suspended_project_state = models.ProjectState.objects.get(value=models.ProjectState.Choices.SUSPENDED.value)
        self.assertIsNotNone(active_project_state)
        self.assertIsNotNone(suspended_project_state)

        # set the scheduling_unit_blueprint's project state to suspended
        project = scheduling_unit_blueprint.project
        project.project_state = suspended_project_state
        project.save()

        # start the dynamic_scheduling_service, which includes eventmessage handling, and a running scheduler
        with BusListenerJanitor(create_dynamic_scheduling_service()):
            # at scheduler startup, all units are evaluated for schedulability.
            # assert the scheduling_unit has been marked as unschedulable
            scheduling_unit_blueprint.refresh_from_db()
            self.assertEqual(models.SchedulingUnitStatus.Choices.UNSCHEDULABLE.value, scheduling_unit_blueprint.status.value)

            # set the project's state
            project.project_state = active_project_state
            project.save()

            # now wait and poll until unit is scheduled, or timeout
            scheduling_unit_blueprint = wait_for_scheduling_unit_blueprint_status(scheduling_unit_blueprint.id, models.SchedulingUnitStatus.Choices.SCHEDULED.value)

            # is it scheduled at the right time?
            self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint.status.value)
            self.assertEqual(at, scheduling_unit_blueprint.scheduled_start_time)

            # set the project's state to inactive again
            project.project_state = suspended_project_state
            project.save()

            # wait and poll until unit is unschedulable, or timeout
            scheduling_unit_blueprint = wait_for_scheduling_unit_blueprint_status(scheduling_unit_blueprint.id, models.SchedulingUnitStatus.Choices.UNSCHEDULABLE.value)

            # is it unschedulable at the right time?
            self.assertEqual(models.SchedulingUnitStatus.Choices.UNSCHEDULABLE.value, scheduling_unit_blueprint.status.value)
            self.assertEqual(at, scheduling_unit_blueprint.scheduled_start_time)


    def test_project_cycle_event_handling(self):
        """
        Test a simple observation with the 'at' constraint and 'fixed_time' scheduler.
        First ensure it is unschedulable by making the project not part of any cycle, so there is no cycle-window to schedule it in.
        Then add the project to a cycle, and see if the scheduler picks it up and schedules it.
        """
        project = models.Project.objects.create(**Project_test_data(name=str(uuid.uuid4()), project_state=models.ProjectState.objects.get(value=models.ProjectState.Choices.ACTIVE.value)))
        scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project))

        # Use at constraint in the near future
        at = round_to_second_precision(datetime.utcnow() + timedelta(days=1))
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit_fixed_time(at=at, scheduling_set=scheduling_set)
        scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        # make sure we start with no cycles
        models.CycleReport.objects.all().delete()
        project.cycles.all().delete()

        # assert blueprint has correct constraints, and is schedulable
        self.assertEqual('fixed_time', scheduling_unit_blueprint.scheduling_constraints_doc['scheduler'])
        self.assertEqual(at.isoformat(), scheduling_unit_blueprint.scheduling_constraints_doc['time']['at'])
        self.assertEqual(scheduling_unit_blueprint.status.value, models.SchedulingUnitStatus.Choices.SCHEDULABLE.value)

        # start the dynamic_scheduling_service, which includes eventmessage handling, and a running scheduler
        with BusListenerJanitor(create_dynamic_scheduling_service()):
            # at scheduler startup, a full scheduling run is done.
            # wait and poll until unit is marked as unschedulable, or timeout
            scheduling_unit_blueprint = wait_for_scheduling_unit_blueprint_status(scheduling_unit_blueprint.id, models.SchedulingUnitStatus.Choices.UNSCHEDULABLE.value)

            # add the project to a cycle
            cycle = models.Cycle.objects.create(**Cycle_test_data(start=at - timedelta(days=7),
                                                                  stop=at + timedelta(days=7)))
            project.cycles.add(cycle)

            # now wait and poll until unit is scheduled, or timeout
            scheduling_unit_blueprint = wait_for_scheduling_unit_blueprint_status(scheduling_unit_blueprint.id, models.SchedulingUnitStatus.Choices.SCHEDULED.value)

            # is it scheduled at the right time?
            self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint.status.value)
            self.assertEqual(at, scheduling_unit_blueprint.scheduled_start_time)

            # remove the project from the cycle again
            project.cycles.remove(cycle)

            # wait and poll until unit is unschedulable, or timeout
            scheduling_unit_blueprint = wait_for_scheduling_unit_blueprint_status(scheduling_unit_blueprint.id, models.SchedulingUnitStatus.Choices.UNSCHEDULABLE.value)

            # is it unschedulable at the right time?
            self.assertEqual(models.SchedulingUnitStatus.Choices.UNSCHEDULABLE.value, scheduling_unit_blueprint.status.value)
            self.assertEqual(at, scheduling_unit_blueprint.scheduled_start_time)


    def test_constraints_update_event_handling(self):
        """
        Test a simple observation with the 'at' constraint and 'fixed_time' scheduler.
        Check if the event handling in the dynamic scheduler picks up events from contraint updates, and reschedules the unit.
        """
        # Use at constraint in the near future
        at = round_to_second_precision(datetime.utcnow() + timedelta(days=1))
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit_fixed_time(at=at)
        scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        # assert blueprint has correct constraints, and is schedulable
        self.assertEqual('fixed_time', scheduling_unit_blueprint.scheduling_constraints_doc['scheduler'])
        self.assertEqual(at.isoformat(), scheduling_unit_blueprint.scheduling_constraints_doc['time']['at'])
        self.assertEqual(scheduling_unit_blueprint.status.value, models.SchedulingUnitStatus.Choices.SCHEDULABLE.value)

        # start the dynamic_scheduling_service, which includes eventmessage handling, and a running scheduler
        with BusListenerJanitor(create_dynamic_scheduling_service()):
            # upon starting the dynamic_scheduling_service, it should run at least one scheduling round
            # so, as a result the schedulable unit should now get scheduled
            # wait and poll until unit is scheduled, or timeout
            scheduling_unit_blueprint = wait_for_scheduling_unit_blueprint_status(scheduling_unit_blueprint.id, models.SchedulingUnitStatus.Choices.SCHEDULED.value)

            # is it scheduled at the right time?
            self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint.status.value)
            self.assertEqual(at, scheduling_unit_blueprint.scheduled_start_time)

            # move the 'at' constraint, upon saving the dynamic scheduler should reschedule it
            # move it into the past, so the unit should become unschedulable, which is a status(change) we can wait for.
            at = round_to_second_precision(datetime.utcnow() - timedelta(hours=1))
            scheduling_unit_blueprint.scheduling_constraints_doc['time']['at'] = at.isoformat()
            scheduling_unit_blueprint.save()

            # wait and poll until unit was unscheduled and became unschedulable, or timeout
            scheduling_unit_blueprint = wait_for_scheduling_unit_blueprint_status(scheduling_unit_blueprint.id, models.SchedulingUnitStatus.Choices.UNSCHEDULABLE.value)

            # is it unschedulable, and at the new timestap?
            self.assertEqual(models.SchedulingUnitStatus.Choices.UNSCHEDULABLE.value, scheduling_unit_blueprint.status.value)
            self.assertEqual(at, scheduling_unit_blueprint.scheduled_start_time)

            # move it again, now into the future. Thus, the unit should be schedulable again and get scheduled
            at = round_to_second_precision(datetime.utcnow() + timedelta(hours=1))
            scheduling_unit_blueprint.scheduling_constraints_doc['time']['at'] = at.isoformat()
            scheduling_unit_blueprint.save()

            # wait and poll until unit was scheduled, or timeout
            scheduling_unit_blueprint = wait_for_scheduling_unit_blueprint_status(scheduling_unit_blueprint.id, models.SchedulingUnitStatus.Choices.SCHEDULED.value)

            # is it scheduled at the new time?
            self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint.status.value)
            self.assertEqual(at, scheduling_unit_blueprint.scheduled_start_time)


    def test_fixed_time_blocked_by_reservation(self):
        """
        Test a simple observation with the 'at' constraint and 'fixed_time' scheduler.
        The 'at' timestamp in the future should make the scheduling_unit schedulable.
        However, we also make a blocking reservation. Thus the scheduling_unit should get the 'unschedulable' status.
        """
        # Use at constraint in the future
        at = round_to_second_precision(datetime.utcnow() + timedelta(days=1))
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit_fixed_time(at=at)
        scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        # assert blueprint has correct constraints, and is schedulable
        self.assertEqual('fixed_time', scheduling_unit_blueprint.scheduling_constraints_doc['scheduler'])
        self.assertEqual(at.isoformat(), scheduling_unit_blueprint.scheduling_constraints_doc['time']['at'])
        self.assertEqual(scheduling_unit_blueprint.status.value, models.SchedulingUnitStatus.Choices.SCHEDULABLE.value)

        # create an overlapping reservation
        reservation_template = models.ReservationTemplate.objects.first() # there is only one
        reservation_doc = reservation_template.get_default_json_document_for_schema()
        reservation_doc['resources']['stations'] = ['CS002', 'CS003']
        reservation_doc['schedulability']['fixed_time'] = False
        models.Reservation.objects.create(start_time=scheduling_unit_blueprint.scheduled_start_time,
                                          stop_time=scheduling_unit_blueprint.scheduled_observation_stop_time,
                                          specifications_doc=reservation_doc,
                                          specifications_template=reservation_template,
                                          description='')

        # try the method-under-test: schedule_fixed_time_scheduling_units
        self.scheduler.schedule_fixed_time_scheduling_units()

        # Assert the scheduling_unit is unschedulable (because of the blocking reservation)
        scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.UNSCHEDULABLE.value, scheduling_unit_blueprint.status.value)
        self.assertEqual(at, scheduling_unit_blueprint.scheduled_start_time)

    def test_reservation_removes_station_from_fixed_time_observation(self):
        """
        Schedule a simple observation with the 'at' constraint and 'fixed_time' scheduler.
        Adding a reservation that uses one of its stations should remove the station from the observation.
        The scheduling_unit should get the 'unschedulable' status when too many stations get reserved.
        """

        # Use at constraint in the future
        at = round_to_second_precision(datetime.utcnow() + timedelta(days=1))
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit_fixed_time(at=at)
        scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        # assert blueprint has correct constraints, and is schedulable
        self.assertEqual('fixed_time', scheduling_unit_blueprint.scheduling_constraints_doc['scheduler'])
        self.assertEqual(at.isoformat(), scheduling_unit_blueprint.scheduling_constraints_doc['time']['at'])
        self.assertEqual(scheduling_unit_blueprint.status.value, models.SchedulingUnitStatus.Choices.SCHEDULABLE.value)

        # schedule the scheduling unit with the method-under-test: schedule_fixed_time_scheduling_units
        self.scheduler.schedule_fixed_time_scheduling_units()
        scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(scheduling_unit_blueprint.status.value, models.SchedulingUnitStatus.Choices.SCHEDULED.value)
        self.assertIn('CS002', scheduling_unit_blueprint.subtasks.first().specifications_doc['stations']['station_list'])

        # create an overlapping reservation
        reservation_template = models.ReservationTemplate.objects.first() # there is only one
        reservation_doc = reservation_template.get_default_json_document_for_schema()
        reservation_doc['resources']['stations'] = ['CS002']
        reservation_doc['schedulability']['fixed_time'] = False
        reservation = models.Reservation.objects.create(start_time=scheduling_unit_blueprint.scheduled_start_time,
                                                        stop_time=scheduling_unit_blueprint.scheduled_observation_stop_time,
                                                        specifications_doc=reservation_doc,
                                                        specifications_template=reservation_template,
                                                        description='',
                                                        name='###1')

        # now the scheduling unit should be re-scheduled without the reserved station
        self.scheduler.schedule_fixed_time_scheduling_units()
        self.message_handler.onReservationCreated(reservation.id)

        # Assert the scheduling_unit is (again) scheduled but minus the reserved station
        scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(scheduling_unit_blueprint.status.value, models.SchedulingUnitStatus.Choices.SCHEDULED.value)
        self.assertNotIn('CS002', scheduling_unit_blueprint.subtasks.first().specifications_doc['stations']['station_list'])

        # create a second overlapping reservation for another station
        reservation_template = models.ReservationTemplate.objects.first() # there is only one
        reservation_doc = reservation_template.get_default_json_document_for_schema()
        reservation_doc['resources']['stations'] = ['CS003']
        reservation_doc['schedulability']['fixed_time'] = False
        reservation = models.Reservation.objects.create(start_time=scheduling_unit_blueprint.scheduled_start_time,
                                          stop_time=scheduling_unit_blueprint.scheduled_observation_stop_time,
                                          specifications_doc=reservation_doc,
                                          specifications_template=reservation_template,
                                          description='###2')

        # now the scheduling unit should be re-scheduled without the reserved station
        self.scheduler.schedule_fixed_time_scheduling_units()
        self.message_handler.onReservationCreated(reservation.id)

        # Assert the scheduling_unit is unschedulable (because of the blocking reservation)
        scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.UNSCHEDULABLE.value, scheduling_unit_blueprint.status.value)


    def test_reservation_event_handling(self):
        """
        Test a simple observation with the 'at' constraint and 'fixed_time' scheduler.
        First ensure it is unschedulable by reusing the unschedulable scheduling_unit from test_fixed_time_blocked_by_reservation
        Then move the reservation, and see if the scheduler picks it up and schedules it.
        """
        # reuse the unschedulable scheduling_unit from test_fixed_time_blocked_by_reservation
        self.test_fixed_time_blocked_by_reservation()
        self.assertEqual(1, models.SchedulingUnitBlueprint.objects.count())
        scheduling_unit_blueprint = models.SchedulingUnitBlueprint.objects.first()

        # assert blueprint has correct constraints, and is unschedulable
        at = parser.parse(scheduling_unit_blueprint.scheduling_constraints_doc['time']['at'], ignoretz=True)
        self.assertEqual('fixed_time', scheduling_unit_blueprint.scheduling_constraints_doc['scheduler'])
        self.assertEqual(at, scheduling_unit_blueprint.scheduled_start_time)
        self.assertEqual(scheduling_unit_blueprint.status.value, models.SchedulingUnitStatus.Choices.UNSCHEDULABLE.value)

        # reuse the reservation as well
        self.assertEqual(1, models.Reservation.objects.count())
        reservation = models.Reservation.objects.first()

        # start the dynamic_scheduling_service, which includes eventmessage handling, and a running scheduler
        with BusListenerJanitor(create_dynamic_scheduling_service()):
            # upon starting the dynamic_scheduling_service, it checks/sets all statuses according to project statuses and reservations
            # this unit should still be unschedulable due to the blocking reservation
            # wait and poll until unit is unschedulable, or timeout
            scheduling_unit_blueprint = wait_for_scheduling_unit_blueprint_status(scheduling_unit_blueprint.id, models.SchedulingUnitStatus.Choices.UNSCHEDULABLE.value)
            # is it unschedulable at the right time?
            self.assertEqual(models.SchedulingUnitStatus.Choices.UNSCHEDULABLE.value, scheduling_unit_blueprint.status.value)
            self.assertEqual(at, scheduling_unit_blueprint.scheduled_start_time)

            # move the reservation open-ended into the future
            reservation.start_time = scheduling_unit_blueprint.scheduled_observation_stop_time + timedelta(days=1)
            reservation.stop_time = None
            reservation.save() # should trigger the scheduler event handler....

            # now wait and poll until unit is scheduled, or timeout
            scheduling_unit_blueprint = wait_for_scheduling_unit_blueprint_status(scheduling_unit_blueprint.id, models.SchedulingUnitStatus.Choices.SCHEDULED.value)

            # is it scheduled at the right time?
            self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint.status.value)
            self.assertEqual(at, scheduling_unit_blueprint.scheduled_start_time)

            # move the reservation back into the past, but keep it open-ended, thus blocking the unit
            reservation.start_time = at - timedelta(days=1)
            reservation.stop_time = None
            reservation.save() # should trigger the scheduler event handler....

            # wait and poll until unit is unschedulable, or timeout
            scheduling_unit_blueprint = wait_for_scheduling_unit_blueprint_status(scheduling_unit_blueprint.id, models.SchedulingUnitStatus.Choices.UNSCHEDULABLE.value)

            # is it unschedulable at the right time?
            self.assertEqual(models.SchedulingUnitStatus.Choices.UNSCHEDULABLE.value, scheduling_unit_blueprint.status.value)
            self.assertEqual(at, scheduling_unit_blueprint.scheduled_start_time)

            # remove the reservation as a whole
            reservation.delete() # should trigger the scheduler event handler....

            # now wait and poll until unit is scheduled, or timeout
            scheduling_unit_blueprint = wait_for_scheduling_unit_blueprint_status(scheduling_unit_blueprint.id, models.SchedulingUnitStatus.Choices.SCHEDULED.value)

            # is it scheduled at the right time?
            self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint.status.value)
            self.assertEqual(at, scheduling_unit_blueprint.scheduled_start_time)

            # keep track of state changes for synchronization (waiting in this test, see below)
            obs_subtask = scheduling_unit_blueprint.subtasks.filter(specifications_template__type__value=models.SubtaskType.Choices.OBSERVATION.value, state__value=models.SubtaskState.Choices.SCHEDULED.value).first()
            scheduled_state_change = obs_subtask.subtaskstatelog_set.filter(new_state__value=models.SubtaskState.Choices.SCHEDULED.value).order_by('-updated_at').first()

            # create an new reservation overlapping with only one station
            reservation_template = models.ReservationTemplate.objects.first() # there is only one
            reservation_doc = reservation_template.get_default_json_document_for_schema()
            reservation_doc['resources']['stations'] = ['CS002']
            reservation_doc['schedulability']['fixed_time'] = False
            reservation = models.Reservation.objects.create(start_time=scheduling_unit_blueprint.scheduled_start_time,
                                                            stop_time=scheduling_unit_blueprint.scheduled_observation_stop_time,
                                                            specifications_doc=reservation_doc,
                                                            specifications_template=reservation_template,
                                                            description='')

            # now wait and poll until subtask/unit is unscheduled, and scheduled again, or timeout
            from time import sleep
            while True:
                latest_scheduled_state_change = obs_subtask.subtaskstatelog_set.filter(new_state__value=models.SubtaskState.Choices.SCHEDULED.value).order_by('-updated_at').first()
                if latest_scheduled_state_change.updated_at > scheduled_state_change.updated_at:
                    break
                sleep(0.1)
            # wait/check if the unit is also scheduled again
            scheduling_unit_blueprint = wait_for_scheduling_unit_blueprint_status(scheduling_unit_blueprint.id, models.SchedulingUnitStatus.Choices.SCHEDULED.value)

            # is it scheduled at the right time? and is reserved station CS002 taken out?
            self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint.status.value)
            self.assertEqual(at, scheduling_unit_blueprint.scheduled_start_time)
            self.assertEqual(['CS003'], scheduling_unit_blueprint.main_observation_used_stations)
            self.assertEqual(['CS002', 'CS003'], scheduling_unit_blueprint.main_observation_specified_stations)

            # keep track of state changes for synchronization (waiting in this test, see below)
            scheduled_state_change = obs_subtask.subtaskstatelog_set.filter(new_state__value=models.SubtaskState.Choices.SCHEDULED.value).order_by('-updated_at').first()

            # remove the reservation as a whole
            reservation.delete() # should trigger the scheduler event handler....

            # now wait and poll until subtask/unit is unscheduled, and scheduled again, or timeout
            while True:
                latest_scheduled_state_change = obs_subtask.subtaskstatelog_set.filter(new_state__value=models.SubtaskState.Choices.SCHEDULED.value).order_by('-updated_at').first()
                if latest_scheduled_state_change.updated_at > scheduled_state_change.updated_at:
                    break
                sleep(0.1)
            # wait/check if the unit is also scheduled again
            scheduling_unit_blueprint = wait_for_scheduling_unit_blueprint_status(scheduling_unit_blueprint.id, models.SchedulingUnitStatus.Choices.SCHEDULED.value)

            # is it scheduled at the right time? and is free station CS002 taken in again?
            self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint.status.value)
            self.assertEqual(at, scheduling_unit_blueprint.scheduled_start_time)
            self.assertEqual(['CS002', 'CS003'], scheduling_unit_blueprint.main_observation_used_stations)
            self.assertEqual(['CS002', 'CS003'], scheduling_unit_blueprint.main_observation_specified_stations)



    def test_can_schedule_all_observing_strategy_templates_with_default_constraints(self):
        ''''''
        from lofar.sas.tmss.tmss.tmssapp.tasks import create_scheduling_unit_draft_from_observing_strategy_template

        failed_templates = []
        for strategy_template in models.SchedulingUnitObservingStrategyTemplate.objects.filter(state__value='active').order_by('name', 'version').all():
            try:
                self.clean_environment()

                at = round_to_second_precision(datetime.utcnow()+timedelta(minutes=5))

                if 'IDOLS' in strategy_template.name:
                    # idols has a daytime and nighttime specific template.
                    # schedule it at noon or midnight, tomorrow so the timestamp is not in the past.
                    at = datetime.combine(datetime.today(), time(12, 0, 0) if 'daytime' in strategy_template.name.lower() else time(23, 59, 0)) + timedelta(days=1)
                elif 'solar' in strategy_template.name.lower():
                    # solar campaign needs to be observed at daytime.
                    # schedule it at noon, tomorrow so the timestamp is not in the past.
                    at = datetime.combine(datetime.today(), time(12, 0, 0)) + timedelta(days=1)

                scheduling_constraints_doc = {'scheduler': 'fixed_time',
                                              'time': {'at': at.isoformat()},
                                              'daily': {'require_day': False, 'require_night': False},
                                              'sky': {'transit_offset': {'from': -24 * 60 * 60, 'to': 24 * 60 * 60},
                                                      'min_elevation': {'target': 0, 'calibrator': 0}}}
                specifications_doc_overrides = {'scheduling_constraints_doc': scheduling_constraints_doc}

                logger.info("Creating scheduling_unit_draft from strategy_template '%s' version=%s", strategy_template.name, strategy_template.version)
                draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template=strategy_template,
                                                                                      scheduling_set=self.scheduling_set_high,
                                                                                      specifications_doc_overrides=specifications_doc_overrides)

                logger.info("Creating scheduling_unit_blueprint from draft for strategy_template '%s' version=%s", strategy_template.name, strategy_template.version)
                blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(draft)
                self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULABLE.value, blueprint.status.value)
                self.assertEqual('fixed_time', blueprint.scheduling_constraints_doc['scheduler'])

                # call the method-under-test: self.scheduler.schedule_fixed_time_scheduling_units
                # this test only checks if each strategy_template with plain default spec and default constraints *can* be scheduled by the dynamic scheduler without any exceptions/errors. The defaults should just work.
                logger.info("Scheduling scheduling_unit_blueprint for strategy_template '%s' version=%s", strategy_template.name, strategy_template.version)
                self.scheduler.schedule_fixed_time_scheduling_units()

                blueprint.refresh_from_db()
                self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, blueprint.status.value)
                self.assertEqual(at, blueprint.scheduled_start_time)

                logger.info("Created and scheduled scheduling_unit_blueprint for strategy_template '%s' version=%s successfully using fixed time scheduler with default constraints", strategy_template.name, strategy_template.version)
            except Exception as e:
                logger.exception("Could not create and fixed_time schedule strategy template '%s' version=%s: %s", strategy_template.name, strategy_template.version, e)
                failed_templates.append(strategy_template)

        # this test passes when there are no failed templates
        self.assertTrue(len(failed_templates)==0, msg='failed_templates: %s' % ([(t.name, t.version) for t in failed_templates]))


    def test_bug_fix_TMSS_2538_check_get_gaps_between_scheduled_units_in_window(self):
        """
        Test to check the core routine 'get_gaps_between_scheduled_units_in_window' for determining where to schedule B-prio units in gaps in between A-prio units.
        See: https://support.astron.nl/jira/browse/TMSS-2538
        """
        # create two units, 1 hour apart, and check the gaps in between
        at1 = round_to_minute_precision(datetime.utcnow() + timedelta(hours=2))
        at1 += timedelta(minutes=-at1.minute)
        scheduling_unit_draft1 = self.create_simple_observation_scheduling_unit_fixed_time(name='A1',
                                                                                           at=at1,
                                                                                           obs_duration=timedelta(hours=1))
        scheduling_unit_blueprint1 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft1)

        at2 = round_to_minute_precision(at1 + timedelta(hours=2))
        scheduling_unit_draft2 = self.create_simple_observation_scheduling_unit_fixed_time(name='A2',
                                                                                           at=at2,
                                                                                           obs_duration=timedelta(hours=1))
        scheduling_unit_blueprint2 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft2)

        self.scheduler.schedule_fixed_time_scheduling_units()

        # Assert the both units have been scheduled
        scheduling_unit_blueprint1.refresh_from_db()
        scheduling_unit_blueprint2.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint1.status.value)
        self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint2.status.value)
        self.assertEqual(at1, scheduling_unit_blueprint1.scheduled_start_time)
        self.assertEqual(at2, scheduling_unit_blueprint2.scheduled_start_time)

        # Get the observation subtasks, and assert the both have been scheduled
        subtask1 = scheduling_unit_blueprint1.subtasks.filter(specifications_template__type__value=models.SubtaskType.Choices.OBSERVATION.value).first()
        subtask2 = scheduling_unit_blueprint2.subtasks.filter(specifications_template__type__value=models.SubtaskType.Choices.OBSERVATION.value).first()
        self.assertEqual(at1, subtask1.scheduled_start_time)
        self.assertEqual(at2, subtask2.scheduled_start_time)

        # window surrounds both subtasks
        lower_bound = subtask1.scheduled_start_time - 0.5 * subtask1.duration
        upper_bound = subtask2.scheduled_stop_time + 0.5 * subtask2.duration
        gaps = get_gaps_between_scheduled_units_in_window(lower_bound=lower_bound, upper_bound=upper_bound)
        self.assertEqual(((lower_bound, subtask1.scheduled_start_time), (subtask1.scheduled_stop_time, subtask2.scheduled_start_time), (subtask2.scheduled_stop_time, upper_bound)),
                         gaps)

        # window starts mid first subtask, ends mid 2nd subtask
        lower_bound = subtask1.scheduled_start_time + 0.5 * subtask1.duration
        upper_bound = subtask2.scheduled_stop_time - 0.5 * subtask2.duration
        gaps = get_gaps_between_scheduled_units_in_window(lower_bound=lower_bound, upper_bound=upper_bound)
        self.assertEqual(( (subtask1.scheduled_stop_time, subtask2.scheduled_start_time),), gaps)

        # window starts mid first subtask, ends after 2nd subtask
        lower_bound = subtask1.scheduled_start_time + 0.5 * subtask1.duration
        upper_bound = subtask2.scheduled_stop_time + 0.5 * subtask2.duration
        gaps = get_gaps_between_scheduled_units_in_window(lower_bound=lower_bound, upper_bound=upper_bound)
        self.assertEqual(( (subtask1.scheduled_stop_time, subtask2.scheduled_start_time), (subtask2.scheduled_stop_time, upper_bound)), gaps)

        # window starts before first subtask, ends mid 2nd subtask
        lower_bound = subtask1.scheduled_start_time - 0.5 * subtask1.duration
        upper_bound = subtask2.scheduled_stop_time - 0.5 * subtask2.duration
        gaps = get_gaps_between_scheduled_units_in_window(lower_bound=lower_bound, upper_bound=upper_bound)
        self.assertEqual(((lower_bound, subtask1.scheduled_start_time), (subtask1.scheduled_stop_time, subtask2.scheduled_start_time)),
                         gaps)



    def test_bug_fix_TMSS_2538_B_prio_unit_should_be_scheduled_in_gaps_only(self):
        """
        Test to reproduce and fix the reported bug that a B-prio unit is scheduled overlapping an A-prio unit.
        See: https://support.astron.nl/jira/browse/TMSS-2538
        """
        # schedule a normal 4hour-long unit at 1hour into the future
        at = round_to_minute_precision(datetime.utcnow() + timedelta(hours=1))
        scheduling_unit_draft_A = self.create_simple_observation_scheduling_unit_fixed_time(name='fixed_A',
                                                                                            at=at,
                                                                                            obs_duration=timedelta(hours=4))
        scheduling_unit_blueprint_A = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft_A)
        scheduling_unit_blueprint_A.rank = models.SchedulingUnitRank.HIGHEST.value
        scheduling_unit_blueprint_A.save()

        # assert blueprint has correct constraints, and is schedulable with prio A
        self.assertEqual('fixed_time', scheduling_unit_blueprint_A.scheduling_constraints_doc['scheduler'])
        self.assertEqual(at.isoformat(), scheduling_unit_blueprint_A.scheduling_constraints_doc['time']['at'])
        self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULABLE.value, scheduling_unit_blueprint_A.status.value)
        self.assertEqual(models.PriorityQueueType.Choices.A.value, scheduling_unit_blueprint_A.priority_queue.value)

        self.scheduler.schedule_fixed_time_scheduling_units()

        # Assert the scheduling_unit has been scheduled
        scheduling_unit_blueprint_A.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint_A.status.value)
        self.assertEqual(at, scheduling_unit_blueprint_A.scheduled_start_time)

        # convert center timestamp of observation to lst, so we can use it as an angle in the 2nd unit's point,
        # thus forcing the 2nd unit to overlap with the first.
        center_time = scheduling_unit_blueprint_A.scheduled_observation_start_time + 0.5*scheduling_unit_blueprint_A.specified_observation_duration
        center_time_lst_longitude = local_sidereal_time_for_utc_and_station(center_time, 'CS002')
        scheduling_unit_draft_B = self.create_simple_observation_scheduling_unit(name='dynamic_B',
                                                                                 obs_duration=600,
                                                                                 priority_queue=models.PriorityQueueType.objects.get(value=models.PriorityQueueType.Choices.B.value))
        scheduling_unit_spec_B = scheduling_unit_draft_B.specifications_doc
        scheduling_unit_spec_B['tasks']['Observation']['specifications_doc']['station_configuration']['SAPs'][0]['digital_pointing']['angle1'] = center_time_lst_longitude.rad
        scheduling_unit_spec_B['tasks']['Observation']['specifications_doc']['station_configuration']['SAPs'][0]['digital_pointing']['angle2'] = Angle('60 degrees').rad
        scheduling_unit_spec_B['tasks']['Observation']['specifications_doc']['station_configuration']['SAPs'][0]['digital_pointing']['target'] = 'dynamic_B'
        scheduling_unit_spec_B['scheduling_constraints_doc']['scheduler'] = 'dynamic'
        scheduling_unit_spec_B['scheduling_constraints_doc']['sky'] = {'transit_offset': {'from': -9000, 'to': 9000}}
        scheduling_unit_draft_B = update_task_graph_from_specifications_doc(scheduling_unit_draft_B, scheduling_unit_spec_B)
        scheduling_unit_blueprint_B = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft_B)

        # assert blueprint has correct constraints, and is schedulable initially...
        self.assertEqual('dynamic', scheduling_unit_blueprint_B.scheduling_constraints_doc['scheduler'])
        self.assertEqual(scheduling_unit_blueprint_B.status.value, models.SchedulingUnitStatus.Choices.SCHEDULABLE.value)

        # try to schedule the B unit...
        self.scheduler.schedule_next_scheduling_unit(lower_bound=round_to_second_precision(datetime.utcnow()))

        # Assert the scheduling_unit has been scheduled.
        scheduling_unit_blueprint_B.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint_B.status.value)

        # check that the bug has been solved: unitB should NOT be overlapping with unitA.
        self.assertFalse(scheduling_unit_blueprint_B.scheduled_start_time > scheduling_unit_blueprint_A.scheduled_start_time and
                         scheduling_unit_blueprint_B.scheduled_start_time < scheduling_unit_blueprint_A.scheduled_stop_time)

        # try to schedule the B unit...
        self.scheduler.do_dynamic_schedule()

        # Assert the original scheduling_unit is still scheduled
        scheduling_unit_blueprint_A.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint_A.status.value)
        self.assertEqual(at, scheduling_unit_blueprint_A.scheduled_start_time)

        # check that the bug has been solved: unitB should NOT be overlapping with unitA.
        self.assertFalse(scheduling_unit_blueprint_B.scheduled_start_time > scheduling_unit_blueprint_A.scheduled_start_time and
                         scheduling_unit_blueprint_B.scheduled_start_time < scheduling_unit_blueprint_A.scheduled_stop_time)


    def test_bug_fix_TMSS_2563_schedule_B_priority_units_incorrect_start_time(self):
        """
        Test to reproduce and fix the reported bug that placing B-prio unit(s) in a gap between A-prio units sometimes yields to the B-unit being scheduled at the incorrect timestamp and blocking other B-units.
        See: https://support.astron.nl/jira/browse/TMSS-2563
        """
        # schedule two fixed-time 1hour-long units, starting in just 4min and in 2h
        at1 = round_to_minute_precision(datetime.utcnow() + timedelta(minutes=4))
        scheduling_unit_draft_A1 = self.create_simple_observation_scheduling_unit_fixed_time(name='fixed_A1',
                                                                                             at=at1,
                                                                                             obs_duration=timedelta(hours=1))
        scheduling_unit_blueprint_A1 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft_A1)

        # already schedule unit A1
        self.scheduler.schedule_fixed_time_scheduling_units()

        # prepare unit A2 and B1 and B2
        at2 = round_to_minute_precision(datetime.utcnow() + timedelta(minutes=4, hours=2.1))
        scheduling_unit_draft_A2 = self.create_simple_observation_scheduling_unit_fixed_time(name='fixed_A2',
                                                                                             at=at2,
                                                                                             obs_duration=timedelta(hours=1))
        scheduling_unit_blueprint_A2 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft_A2)

        # convert timestamp in between the two observations to lst, so we can use it as an angle in the B-prio units pointings,
        # with this pointing the B-prio units should fit right in between A1 and A2
        in_between_time = scheduling_unit_blueprint_A1.scheduled_observation_stop_time + 0.5*(scheduling_unit_blueprint_A2.scheduled_observation_start_time - scheduling_unit_blueprint_A1.scheduled_observation_stop_time)
        in_between_lst_longitude = local_sidereal_time_for_utc_and_station(in_between_time, 'CS002')
        scheduling_unit_draft_B = self.create_simple_observation_scheduling_unit(name='dynamic_B1',
                                                                                 obs_duration=3600,
                                                                                 priority_queue=models.PriorityQueueType.objects.get(value=models.PriorityQueueType.Choices.B.value))
        scheduling_unit_spec_B = scheduling_unit_draft_B.specifications_doc
        scheduling_unit_spec_B['tasks']['Observation']['specifications_doc']['station_configuration']['SAPs'][0]['digital_pointing']['angle1'] = in_between_lst_longitude.rad
        scheduling_unit_spec_B['tasks']['Observation']['specifications_doc']['station_configuration']['SAPs'][0]['digital_pointing']['angle2'] = Angle('80 degrees').rad
        scheduling_unit_spec_B['scheduling_constraints_doc']['scheduler'] = 'dynamic'
        scheduling_unit_spec_B['scheduling_constraints_doc']['sky']['min_elevation'] = {'target': Angle('10 degrees').rad}
        del scheduling_unit_spec_B['scheduling_constraints_doc']['sky']['min_distance']
        del scheduling_unit_spec_B['scheduling_constraints_doc']['sky']['transit_offset']
        del scheduling_unit_spec_B['scheduling_constraints_doc']['daily']
        # the 'before' constraint causes the bug
        scheduling_unit_spec_B['scheduling_constraints_doc']['time']['before'] = (datetime.utcnow()+timedelta(days=2)).isoformat()

        scheduling_unit_draft_B1 = update_task_graph_from_specifications_doc(scheduling_unit_draft_B, scheduling_unit_spec_B)
        scheduling_unit_blueprint_B1 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft_B1)

        # B2 has same specs and same RA, but lower DEC, so it should "win" from B1 in fighting over a scheduling spot.
        scheduling_unit_draft_B.name = 'dynamic_B2'
        scheduling_unit_draft_B.save()
        scheduling_unit_spec_B['tasks']['Observation']['specifications_doc']['station_configuration']['SAPs'][0]['digital_pointing']['target'] = scheduling_unit_draft_B.name
        scheduling_unit_spec_B['tasks']['Observation']['specifications_doc']['station_configuration']['SAPs'][0]['digital_pointing']['angle2'] = Angle('40 degrees').rad
        scheduling_unit_draft_B2 = update_task_graph_from_specifications_doc(scheduling_unit_draft_B, scheduling_unit_spec_B)
        scheduling_unit_blueprint_B2 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft_B2)

        # run scheduler, A2 and B2 should be scheduled first, followed by B1
        self.scheduler.do_full_schedule_computation()

        # Assert the A-prio scheduling_units have been scheduled
        scheduling_unit_blueprint_A1.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint_A1.status.value)
        self.assertEqual(at1, scheduling_unit_blueprint_A1.scheduled_start_time)
        scheduling_unit_blueprint_A2.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint_A2.status.value)
        self.assertEqual(at2, scheduling_unit_blueprint_A2.scheduled_start_time)

        scheduling_unit_blueprint_B2.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint_B2.status.value)
        self.assertTrue(scheduling_unit_blueprint_B2.scheduled_start_time > scheduling_unit_blueprint_A1.scheduled_stop_time)
        self.assertTrue(scheduling_unit_blueprint_B2.scheduled_stop_time < scheduling_unit_blueprint_A2.scheduled_start_time)

        scheduling_unit_blueprint_B1.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduling_unit_blueprint_B1.status.value)
        self.assertTrue(scheduling_unit_blueprint_B1.scheduled_start_time > scheduling_unit_blueprint_A2.scheduled_stop_time)


    def test_units_can_run_in_parallel_for_distinct_station_sets(self):
        """ Generic test to check if any two units can run in parallel if they use distinct sets of stations.
        (Assume Cobalt can handle this)
        """
        # 1st observation runs with CS032
        at = round_to_minute_precision(datetime.utcnow() + timedelta(hours=1))
        scheduling_unit_draft1 = self.create_simple_observation_scheduling_unit_fixed_time(name='A1',
                                                                                           at=at,
                                                                                           obs_duration=timedelta(hours=4))
        scheduling_unit_spec1 = scheduling_unit_draft1.specifications_doc
        scheduling_unit_spec1['tasks']['Observation']['specifications_doc']['station_configuration']['station_groups'][0]['stations'] = ['CS032']
        scheduling_unit_spec1['tasks']['Observation']['specifications_doc']['station_configuration']['station_groups'][0]['max_nr_missing'] = 0
        scheduling_unit_draft1 = update_task_graph_from_specifications_doc(scheduling_unit_draft1, scheduling_unit_spec1)

        # 2nd observation runs at same timestamp, but with CS002
        scheduling_unit_draft2 = self.create_simple_observation_scheduling_unit_fixed_time(name='A2',
                                                                                           at=at,
                                                                                           obs_duration=timedelta(hours=4))
        scheduling_unit_spec2 = scheduling_unit_draft2.specifications_doc
        scheduling_unit_spec2['tasks']['Observation']['specifications_doc']['station_configuration']['station_groups'][0]['stations'] = ['CS002']
        scheduling_unit_spec2['tasks']['Observation']['specifications_doc']['station_configuration']['station_groups'][0]['max_nr_missing'] = 0
        scheduling_unit_draft2 = update_task_graph_from_specifications_doc(scheduling_unit_draft2, scheduling_unit_spec2)

        # blueprint both
        scheduling_unit_blueprint1 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft1)
        scheduling_unit_blueprint2 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft2)

        scheduled_units = self.scheduler.schedule_fixed_time_scheduling_units()
        self.assertEqual(2, len(scheduled_units))
        self.assertIn(scheduling_unit_blueprint1, scheduled_units)
        self.assertIn(scheduling_unit_blueprint2, scheduled_units)


class TestDynamicScheduling(BaseDynamicSchedulingTestCase):
    '''Test the behaviour of the dynamic scheduler for scheduling_units with a 'dynamic' scheduler constraint (and sky/daily/time constraints)'''

    def test_simple_observation_with_at_constraint(self):
        """
        Test a simple observation with the 'at' constraint
        """
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit('scheduling_unit for at constraint')

        # Clear/set constraints
        at = datetime.utcnow().replace(minute=0, second=0, microsecond=0) + timedelta(hours=2)
        scheduling_unit_draft.scheduling_constraints_doc = {'scheduler': 'dynamic',
                                                            'time': {'at': at.isoformat()} }
        scheduling_unit_draft.save()

        # ensure the unit->set->project->cycle's start/end times are large enough
        if scheduling_unit_draft.project.cycles.count() == 0:
            cycle = models.Cycle.objects.create(**Cycle_test_data())
            scheduling_unit_draft.project.cycles.add(cycle)
            scheduling_unit_draft.project.save()

        cycle = scheduling_unit_draft.project.cycles.first()
        cycle.start = at - timedelta(days=1)
        cycle.stop = at + timedelta(days=7)
        cycle.save()

        scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
        logger.info("test_simple_observation_with_at_constraint: %s", scheduling_unit_blueprint.scheduling_constraints_doc)

        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()
        self.assertEqual(1, len(scheduled_scheduling_units))
        scheduled_scheduling_unit = scheduled_scheduling_units[0]

        # Assert the scheduling_unit has been scheduled and assert is has been scheduled at "at" timestamp
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduled_scheduling_unit.id, scheduling_unit_blueprint.id)
        self.assertEqual(scheduled_scheduling_unit.status.value, 'scheduled')
        self.assertEqual(scheduled_scheduling_unit.scheduled_start_time, at)


    def test_two_simple_observations_with_at_constraint_one_after_the_other(self):
        """
        Test two simple observation each with the 'at' constraint, one after the other with enough space, so they should both be scheduled
        """
        at1 = round_to_second_precision(datetime.utcnow() + timedelta(minutes=10))
        scheduling_unit_draft1 = self.create_simple_observation_scheduling_unit('scheduling_unit for at constraint',
                                                                                constraints={'scheduler': 'dynamic',
                                                                                             'time': {'at': at1.isoformat()},
                                                                                             'sky': { 'transit_offset': {'from': -12*60*60, 'to': 12*60*60} } })

        scheduling_unit_blueprint1 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft1)

        at2 = round_to_second_precision(at1 + scheduling_unit_blueprint1.specified_main_observation_duration + timedelta(minutes=10) )
        scheduling_unit_draft2 = self.create_simple_observation_scheduling_unit('scheduling_unit for at constraint',
                                                                                constraints={'scheduler': 'dynamic',
                                                                                             'time': {'at': at2.isoformat()},
                                                                                             'sky': { 'transit_offset': {'from': -12*60*60, 'to': 12*60*60} } })

        scheduling_unit_blueprint2 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft2)


        # ensure the transit score does not play a role in ordering
        transit_weight,_ = models.SchedulingConstraintsWeightFactor.objects.get_or_create(constraint_name='sky.transit_offset')
        transit_weight.weight = 0
        transit_weight.save()

        # call the method-under-test.
        # we expect the scheduling unit with the closest 'time.at' constraint to be scheduled
        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()
        scheduled_scheduling_unit = scheduled_scheduling_units[0]

        # Assert the scheduling_unit has been scheduled and assert it has been scheduled at "at" timestamp
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduled_scheduling_unit.id, scheduling_unit_blueprint1.id)
        self.assertEqual(scheduled_scheduling_unit.status.value, 'scheduled')
        self.assertEqual(scheduled_scheduling_unit.scheduled_start_time, at1)

        # mimic the subtask/scheduling_unit runs and gets finished, thereby making room for the next iteration of scheduling
        set_subtask_state_following_allowed_transitions(scheduled_scheduling_unit.subtasks.first(), 'finished')
        scheduled_scheduling_unit.refresh_from_db()
        self.assertEqual(scheduled_scheduling_unit.status.value, 'finished')

        # make sure the 2nd unit is schedulable
        unschedule_subtasks_in_scheduling_unit_blueprint(scheduling_unit_blueprint2)


        # call the method-under-test.
        # we expect the 2nd scheduling unit to be scheduled
        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()
        self.assertEqual(1, len(scheduled_scheduling_units))
        scheduled_scheduling_unit = scheduled_scheduling_units[0]

        # Assert the scheduling_unit has been scheduled and assert is has been scheduled at "at" timestamp
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduled_scheduling_unit.id, scheduling_unit_blueprint2.id)
        self.assertEqual(scheduled_scheduling_unit.status.value, 'scheduled')
        self.assertEqual(scheduled_scheduling_unit.scheduled_start_time, at2)

        # reset transit weight
        transit_weight,_ = models.SchedulingConstraintsWeightFactor.objects.get_or_create(constraint_name='sky.transit_offset')
        transit_weight.weight = models.SchedulingConstraintsWeightFactor.weight.field.default
        transit_weight.save()


    def test_n_simple_observations_one_at_constraint(self):
        """
        Test n simple observations where only one of them has an 'at' constraint
        """
        n = 5   # No of SU to be created
        target = 4  # SU id to be within the 'at' constraint
        target_scheduling_unit_blueprint = None # SU which will be our target

        # Create constraints to be assigned to all of the scheduling_units
        from_timestamp = round_to_second_precision(datetime.utcnow())
        to_timestamp = round_to_second_precision(datetime.utcnow() + timedelta(hours=12))
        between_constraints = [{"from": from_timestamp.isoformat(), "to": to_timestamp.isoformat()},]
        # Create at constraint to be assigned only to one of the scheduling_units
        at = round_to_second_precision((datetime.utcnow() + timedelta(minutes=30)))

        # Create n scheduling_units and set the proper constraints
        for su in range(1, n+1):
            scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data())
            scheduling_unit_draft = self.create_simple_observation_scheduling_unit('scheduling_unit %s' % su,
                                                                                   scheduling_set=scheduling_set)
            # Clear constraints
            scheduling_unit_draft.scheduling_constraints_doc = {'sky': {}, 'daily': {}, 'scheduler': 'dynamic',
                                                                'time': {"between": [], "not_between": []}}
            scheduling_unit_draft.save()
            if su == target:    # Only scheduling_unit with id 'target' is set within an 'at' constraint
                scheduling_unit_draft.scheduling_constraints_doc['time']['at'] = at.isoformat()
                scheduling_unit_draft.save()
                target_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
            else:
                create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()
        self.assertEqual(1, len(scheduled_scheduling_units))
        scheduled_scheduling_unit = scheduled_scheduling_units[0]

        # Assert the 'target' scheduling_unit has been scheduled with priority and assert it is has been scheduled at "at" timestamp
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduled_scheduling_unit.id, target_scheduling_unit_blueprint.id)
        self.assertEqual(scheduled_scheduling_unit.status.value, 'scheduled')
        self.assertEqual(scheduled_scheduling_unit.scheduled_start_time, at)

    def test_n_simple_observations_one_at_constraint(self):
        """
        Test n simple observations where only one of them has an 'at' constraint
        """
        n = 5   # No of SU to be created
        target = 4  # SU id to be within the 'at' constraint
        target_scheduling_unit_blueprint = None # SU which will be our target

        # Create constraints to be assigned to all of the scheduling_units
        from_timestamp = round_to_second_precision(datetime.utcnow())
        to_timestamp = round_to_second_precision(datetime.utcnow() + timedelta(hours=12))
        between_constraints = [{"from": from_timestamp.isoformat(), "to": to_timestamp.isoformat()},]
        # Create at constraint to be assigned only to one of the scheduling_units
        at = round_to_second_precision((datetime.utcnow() + timedelta(minutes=10)))

        # Create n scheduling_units and set the proper constraints
        for su in range(1, n+1):
            scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data())
            scheduling_unit_draft = self.create_simple_observation_scheduling_unit('scheduling_unit %s' % su,
                                                                                   scheduling_set=scheduling_set,
                                                                                   obs_duration=3600)
            # Clear constraints
            scheduling_unit_draft.scheduling_constraints_doc = {'sky': {}, 'daily': {}, 'scheduler': 'dynamic',
                                                                'time': {"between": [], "not_between": []}}
            scheduling_unit_draft.save()
            if su == target:    # Only scheduling_unit with id 'target' is set within an 'at' constraint
                scheduling_unit_draft.scheduling_constraints_doc['time']['at'] = at.isoformat()
                scheduling_unit_draft.save()
                target_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
            else:
                create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()
        # all should be scheduled
        self.assertEqual(5, len(scheduled_scheduling_units))

        # the first should be the 'at'-constrained one
        # Assert the 'target' scheduling_unit has been scheduled with priority and assert it is has been scheduled at "at" timestamp
        scheduled_scheduling_unit = scheduled_scheduling_units[0]
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduled_scheduling_unit.id, target_scheduling_unit_blueprint.id)
        self.assertEqual(scheduled_scheduling_unit.status.value, 'scheduled')
        self.assertEqual(scheduled_scheduling_unit.scheduled_start_time, at)


    def test_before_constraint(self):
        """
        Test 2 simple observations with the same projec/unit rank, target, etc, so all scores are equal, except for the before constraint.
        The unit with the earlier before time wins.
        """
        now = round_to_second_precision(datetime.utcnow())
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit('su test_before_constraint',
                                                                               scheduling_set=self.scheduling_set_high)
        # Clear constraints
        scheduling_unit_draft.scheduling_constraints_doc = {'sky': {
                                                                'min_elevation': {'target': 0},
                                                                'transit_offset': {'from': -12*60*60, 'to': 12*60*60}
                                                            },
                                                            'daily': {},
                                                            'scheduler': 'dynamic',
                                                            'time': {"between": [], "not_between": []}}
        scheduling_unit_draft.save()

        # create two equal units
        sub1 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
        sub2 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        # give one a before constraint. Now it should 'win'.
        sub1.scheduling_constraints_doc['time'] = {"before": (now+timedelta(days=1)).isoformat()}
        sub1.save()
        best_sub = self.scheduler.find_best_next_schedulable_unit([sub1, sub2], now, now+timedelta(hours=24)).scheduling_unit
        self.assertIn(best_sub, [sub1, sub2])

        # give the other the before constraint. Now that one should 'win'.
        sub1.scheduling_constraints_doc['time'] = {}
        sub1.save()
        sub2.scheduling_constraints_doc['time'] =  {"before": (now+timedelta(days=1)).isoformat()}
        sub2.save()
        best_sub = self.scheduler.find_best_next_schedulable_unit([sub1, sub2], now, now+timedelta(hours=24)).scheduling_unit
        self.assertEqual(sub2, best_sub)

        # give both a before constraint. Now the 'earliest' should 'win'.
        sub1.scheduling_constraints_doc['time'] = {"before": (now+timedelta(days=1)).isoformat()}
        sub1.save()
        sub2.scheduling_constraints_doc['time'] = {"before": (now+timedelta(days=2)).isoformat()}
        sub2.save()
        best_sub = self.scheduler.find_best_next_schedulable_unit([sub1, sub2], now, now+timedelta(hours=24)).scheduling_unit
        self.assertEqual(sub1, best_sub)

        # give both a before constraint. Now the other 'earliest' should 'win'.
        sub1.scheduling_constraints_doc['time'] = {"before": (now+timedelta(days=1)).isoformat()}
        sub1.save()
        sub2.scheduling_constraints_doc['time'] = {"before": (now+timedelta(days=2)).isoformat()}
        sub2.save()
        best_sub = self.scheduler.find_best_next_schedulable_unit([sub1, sub2], now, now+timedelta(hours=24)).scheduling_unit
        self.assertEqual(sub1, best_sub)


    def test_project_specific_reservation(self):

        # create two projects and two units.
        project_A = models.Project.objects.create(**Project_test_data("dynsched_project_%s_A"% (uuid.uuid4(),)))
        scheduling_set_A = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project_A))
        scheduling_unit_draft_A = self.create_simple_observation_scheduling_unit("scheduling unit project A", scheduling_set=scheduling_set_A)
        scheduling_unit_blueprint_A = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft_A)

        project_B = models.Project.objects.create(**Project_test_data("dynsched_project_%s_B"% (uuid.uuid4(),)))
        scheduling_set_B = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project_B))
        scheduling_unit_draft_B = self.create_simple_observation_scheduling_unit("scheduling unit project B", scheduling_set=scheduling_set_B)
        scheduling_unit_blueprint_B = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft_B)

        cycle = models.Cycle.objects.create(name=str(uuid.uuid4()),
                                            start=datetime.utcnow(),
                                            stop=datetime.utcnow() + timedelta(hours=24))
        cycle.projects.set([project_A, project_B])
        cycle.save()

        # make a reservation for the used stations
        reservation_template = models.ReservationTemplate.get_latest('reservation')
        reservation_doc = reservation_template.get_default_json_document_for_schema()
        used_stations = list(set(scheduling_unit_blueprint_A.main_observation_specified_stations) | set(scheduling_unit_blueprint_B.main_observation_specified_stations))
        reservation_doc['resources'] = {'stations': used_stations}
        reservation_doc['schedulability'] = {'project_exclusive': True, 'dynamic': True, 'fixed_time': True}
        reservation = models.Reservation.objects.create(start_time=datetime.utcnow(),
                                                        specifications_template=reservation_template,
                                                        specifications_doc=reservation_doc)

        # test both projects as exclusive
        for exclusive_project in (project_A, project_B):
            reservation.project = exclusive_project
            reservation.save()

            # # call the method-under-test.
            scheduled_unit = self.scheduler.schedule_next_scheduling_unit()

            # if the reservation is exclusive for project_A, then unit A should be scheduled, and unit B unchedulable
            # and vice versa
            if exclusive_project == project_A:
                self.assertEqual(scheduling_unit_blueprint_A, scheduled_unit)
                unschedule_subtasks_in_scheduling_unit_blueprint(scheduling_unit_blueprint_A) # prepare for next project in this test
            else:
                self.assertEqual(scheduling_unit_blueprint_B, scheduled_unit)
                unschedule_subtasks_in_scheduling_unit_blueprint(scheduling_unit_blueprint_B) # prepare for next project in this test


    def test_time_bound_unit_wins_even_at_lower_priority(self):
        # create two schedule units, one with high one with low prio.
        # first create them without any further constraints, and check if high prio wins.
        scheduling_unit_draft_low = self.create_simple_observation_scheduling_unit("scheduling unit low",
                                                                                   scheduling_set=self.scheduling_set_low,
                                                                                   obs_duration=3600)
        scheduling_unit_blueprint_low = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft_low)

        scheduling_unit_draft_high = self.create_simple_observation_scheduling_unit("scheduling unit high",
                                                                                    scheduling_set=self.scheduling_set_high,
                                                                                    obs_duration=3600)
        scheduling_unit_blueprint_high = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft_high)

        now = round_to_minute_precision(datetime.utcnow()+timedelta(seconds=30)) # for nice debugging friendly timestamps
        tomorrow = now+timedelta(days=1)

        # call the method-under-test.
        best_scored_scheduling_unit = self.scheduler.find_best_next_schedulable_unit([scheduling_unit_blueprint_low, scheduling_unit_blueprint_high], now, tomorrow)

        # we expect the scheduling_unit with the highest project rank to be scheduled first
        self.assertEqual(scheduling_unit_blueprint_high.id, best_scored_scheduling_unit.scheduling_unit.id)

        # now update the low prio unit with a time constraint, "forcing" it to be run 'now'.
        scheduling_unit_blueprint_low.scheduling_constraints_doc['time'] = { 'at': (now + DEFAULT_INTER_OBSERVATION_GAP).isoformat()+'Z' }
        scheduling_unit_blueprint_low.save()

        # call the method-under-test.
        best_scored_scheduling_unit = self.scheduler.find_best_next_schedulable_unit([scheduling_unit_blueprint_low, scheduling_unit_blueprint_high], now, tomorrow)

        # now we expect the scheduling_unit with the lowest project rank to be scheduled first because it can only run within this limited timewindow
        self.assertEqual(scheduling_unit_blueprint_low.id, best_scored_scheduling_unit.scheduling_unit.id)


        # update the low prio unit. enlarge the time window constraint a bit, so both low and high prio units can fit
        # this should result that the high prio goes first, and the low prio (which now fits as well) goes second
        latest_allowed_stop_time_of_low_prio_su = now +\
                                                  scheduling_unit_blueprint_low.specified_main_observation_duration +\
                                                  scheduling_unit_blueprint_high.specified_main_observation_duration +\
                                                  DEFAULT_INTER_OBSERVATION_GAP
        scheduling_unit_blueprint_low.scheduling_constraints_doc['time'] = { 'before': latest_allowed_stop_time_of_low_prio_su.isoformat()+'Z' }
        scheduling_unit_blueprint_low.save()

        # call the method-under-test.
        best_scored_scheduling_unit = self.scheduler.find_best_next_schedulable_unit([scheduling_unit_blueprint_low, scheduling_unit_blueprint_high], now, tomorrow)

        # now we expect the scheduling_unit with the higher project rank to be scheduled first (because there is enough room to schedule the low prio later)
        self.assertEqual(scheduling_unit_blueprint_high.id, best_scored_scheduling_unit.scheduling_unit.id)

        # call the method-under-test again but search after latest possible stop_time of the low_prio unit (should return high prio unit)
        best_scored_scheduling_unit = self.scheduler.find_best_next_schedulable_unit([scheduling_unit_blueprint_low, scheduling_unit_blueprint_high],
                                                                                     latest_allowed_stop_time_of_low_prio_su, tomorrow)
        self.assertEqual(scheduling_unit_blueprint_high.id, best_scored_scheduling_unit.scheduling_unit.id)
        # proof that this is due to the 'before'constraint, and not the higher prio
        scheduling_units_in_future_window = filter_scheduling_units_using_constraints([scheduling_unit_blueprint_low, scheduling_unit_blueprint_high],
                                                                                      latest_allowed_stop_time_of_low_prio_su, tomorrow)
        self.assertEqual([scheduling_unit_blueprint_high], scheduling_units_in_future_window)


    @unittest.mock.patch('lofar.sas.tmss.services.scheduling.dynamic_scheduling.datetime')
    def test_priority_queues_and_project_priority_and_unit_rank(self, mocked_scheduler_datetime):
        """
        All scheduling_units in queue "A" should be scheduled before any unit from queue "B" are scheduled
        Within the same queue, units should be scheduled in project ranking order
        And within the same queue and project rank, units should be scheduled in unit ranking order
        """
        # the scheduler start 'searching' for candidate scheduling_units at utcnow
        # normally, time progresses along while the observation is running.
        # in this test however, we simulate the observation runs, and time hardly advances.
        # so, we mock the datetime.utcnow() function in the scheduler, and make it return the time we want/simulate.
        # initially, use the normal current utcnow in the scheduler
        mocked_scheduler_datetime.utcnow.side_effect = datetime.utcnow

        # use fixed weights for deterministic test behaviour
        constraints_template = models.SchedulingConstraintsTemplate.get_version_or_latest(name="constraints", version=1)
        wf_unit_rank, created = models.SchedulingConstraintsWeightFactor.objects.get_or_create(constraint_name="unit_rank")
        wf_project_rank, created = models.SchedulingConstraintsWeightFactor.objects.get_or_create(constraint_name="project_rank")

        # make project rank just a tad more important than the unit_rank
        wf_unit_rank.weight = 0.9
        wf_project_rank.weight = 1.0
        wf_unit_rank.save()
        wf_project_rank.save()

        # age should not play a role
        wf_age, created = models.SchedulingConstraintsWeightFactor.objects.get_or_create(constraint_name="age")
        wf_age.weight = 0
        wf_age.save()

        queue_A = models.PriorityQueueType.objects.get(value=models.PriorityQueueType.Choices.A.value)

        # prepare 4 schedunits: only A queue, 2 per project (1 high unit rank, 1 low unit rank)
        # as "proof" that we don't just schedule in created order..., create them in reverse expected order :)
        for scheduling_set in (self.scheduling_set_high, self.scheduling_set_low):
            for unit_rank in (models.SchedulingUnitRank.HIGHEST.value, models.SchedulingUnitRank.LOWEST.value):
                scheduling_unit_draft = self.create_simple_observation_scheduling_unit('SU queue=%s project_rank=%s unit_rank=%s' % (queue_A.value, scheduling_set.project.rank, unit_rank),
                                                                                       scheduling_set=scheduling_set,
                                                                                       priority_queue=queue_A,
                                                                                       unit_rank=unit_rank,
                                                                                       obs_duration=3600)
                scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

                # ignore all other constraints
                scheduling_unit_blueprint.scheduling_constraints_doc = {'scheduler': 'dynamic'}
                scheduling_unit_blueprint.save()

                # check we're only using one subtask in this unit (for simplicity)
                self.assertEqual(1, scheduling_unit_blueprint.subtasks.count())

        # preperation complete, we now have 8 schedulable units
        self.assertEqual(4, models.SchedulingUnitBlueprint.objects.count())
        self.assertEqual(4, models.SchedulingUnitBlueprint.objects.filter(status__value=models.SchedulingUnitStatus.Choices.SCHEDULABLE.value).count())

        prev_scheduled_scheduling_unit = None

        # do 4 scheduling runs.
        # test if the expected unit is scheduled.
        for expected_queue in (queue_A,):
            # we first expect the project with the highest rank then the lowest rank
            for expected_project in (self.project_high, self.project_low):
                # we first expect the unit with the highest rank then the lowest rank
                for expected_unit_rank in (models.SchedulingUnitRank.HIGHEST.value, models.SchedulingUnitRank.LOWEST.value):
                    logger.info('----------------------------------------------------------------------')
                    scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()
                    # the first scheduled_scheduling_unit is the first upcoming one
                    scheduled_scheduling_unit = scheduled_scheduling_units[0]

                    # Assert the scheduling_unit has been scheduled
                    self.assertIsNotNone(scheduled_scheduling_unit)
                    self.assertEqual(scheduled_scheduling_unit.status.value, 'scheduled')

                    # check that queue_A takes precedence over queue_B
                    self.assertEqual(expected_queue, scheduled_scheduling_unit.priority_queue)

                    # check that within a queue, project rank takes precedence
                    self.assertEqual(expected_project, scheduled_scheduling_unit.project)

                    # check that within a queue, and equal project rank, the unit rank takes precedence
                    self.assertAlmostEqual(expected_unit_rank, scheduled_scheduling_unit.rank, 2)

                    # check that we are scheduled after the previous one
                    if prev_scheduled_scheduling_unit is not None:
                        self.assertGreater(scheduled_scheduling_unit.scheduled_start_time, prev_scheduled_scheduling_unit.scheduled_observation_stop_time)
                        self.assertGreater(scheduled_scheduling_unit.scheduled_start_time, prev_scheduled_scheduling_unit.actual_process_stop_time)

                    # mimic the subtask/scheduling_unit runs and gets finished, thereby making room for the next iteration of scheduling
                    set_subtask_state_following_allowed_transitions(scheduled_scheduling_unit.subtasks.first(), 'finished')
                    scheduled_scheduling_unit.refresh_from_db()
                    self.assertEqual(scheduled_scheduling_unit.status.value, 'finished')

                    # mock the progress of time in the scheduler.
                    def mocked_utcnow():
                        logger.info("scheduler mocked datetime.utcnow='%s' (the scheduled_observation_stop_time of scheduling_unit id=%s", scheduled_scheduling_unit.scheduled_observation_stop_time, scheduled_scheduling_unit.id)
                        return scheduled_scheduling_unit.scheduled_observation_stop_time
                    mocked_scheduler_datetime.utcnow.side_effect = mocked_utcnow

                    # keep track of the previous one, for comparison of start/stop times
                    prev_scheduled_scheduling_unit = scheduled_scheduling_unit

                    # unschedule the rest for the next scheduling round
                    for su in scheduled_scheduling_units[1:]:
                        unschedule_subtasks_in_scheduling_unit_blueprint(su)


    def test_A_B_priority_queues(self):
        """
        Check if B-prio units can be scheduled in the gaps between A-prio units.
        """
        queue_A = models.PriorityQueueType.objects.get(value=models.PriorityQueueType.Choices.A.value)
        queue_B = models.PriorityQueueType.objects.get(value=models.PriorityQueueType.Choices.B.value)
        project_A = models.Project.objects.create(**Project_test_data("Project A %s" % (str(uuid.uuid4()).split('-')[0], ), rank=models.ProjectRank.HIGHEST.value))
        project_B = models.Project.objects.create(**Project_test_data("Project B %s" % (str(uuid.uuid4()).split('-')[0], ), rank=models.ProjectRank.HIGHEST.value))
        cycle = models.Cycle.objects.create(**Cycle_test_data(start=datetime.utcnow() - timedelta(days=1), stop=datetime.utcnow() + timedelta(days=7)))
        project_A.cycles.add(cycle)
        project_B.cycles.add(cycle)
        scheduling_set_A = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project_A))
        scheduling_set_B = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project_B))
        DURATION = 3600

        # set density_vs_optimal to 0, meaning "as tight as possible", so we can squeeze in the B-units in the tight gap between the A-units
        weight_factor, created = models.SchedulingConstraintsWeightFactor.objects.get_or_create(constraint_name="density_vs_optimal")
        weight_factor.weight = 0
        weight_factor.save()

        self.clean_environment()

        # prepare 4 schedunits: 2 per queue (A/B)
        scheduling_units = {}
        for queue in (queue_A, queue_B):
            for i in range(2):
                name = "SU %s-%s" % (queue.value, i + 1)
                scheduling_unit_draft = self.create_simple_observation_scheduling_unit(name,
                                                                                       scheduling_set=scheduling_set_A if queue == queue_A else scheduling_set_B,
                                                                                       priority_queue=queue,
                                                                                       unit_rank=i,
                                                                                       obs_duration=DURATION)
                scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
                scheduling_unit_blueprint.scheduling_constraints_doc = {'scheduler': 'dynamic'}
                scheduling_unit_blueprint.save()
                scheduling_units[name] = scheduling_unit_blueprint

        # -----------------------------------------------------
        # preparation complete, we now have 4 schedulable units
        self.assertEqual(4, len(scheduling_units))
        self.assertEqual(4, models.SchedulingUnitBlueprint.objects.count())
        self.assertEqual(4, models.SchedulingUnitBlueprint.objects.filter(status__value=models.SchedulingUnitStatus.Choices.SCHEDULABLE.value).count())

        # schedule A-1 first, at just a tad in the future
        at_A1 = round_to_minute_precision(datetime.utcnow() + timedelta(minutes=5))
        scheduling_units['SU A-1'].scheduling_constraints_doc['time'] = {'at': at_A1.isoformat()}
        scheduling_units['SU A-1'].save()

        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()
        self.assertGreater(len(scheduled_scheduling_units), 0)
        scheduled_scheduling_unit = scheduled_scheduling_units[0]
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual('SU A-1', scheduled_scheduling_unit.name)
        self.assertEqual('scheduled', scheduled_scheduling_unit.status.value)
        self.assertEqual(queue_A, scheduled_scheduling_unit.priority_queue)
        self.assertEqual(at_A1, scheduled_scheduling_unit.scheduled_start_time)

        # mimic the subtask/scheduling_unit runs and gets finished, thereby making room for the next iteration of scheduling
        set_subtask_state_following_allowed_transitions(scheduled_scheduling_unit.subtasks.first(), 'finished')
        scheduled_scheduling_unit.refresh_from_db()
        self.assertEqual(scheduled_scheduling_unit.status.value, 'finished')

        # try to schedule A-2 just after A-1.
        # this should succeed, but it leaves no large enough gap for the B units to be squeezed in between
        at_A2 = round_to_minute_precision(scheduled_scheduling_unit.scheduled_observation_stop_time + DEFAULT_NEXT_STARTTIME_GAP + timedelta(minutes=1))
        scheduling_units['SU A-2'] = unschedule_subtasks_in_scheduling_unit_blueprint(scheduling_units['SU A-2'])
        scheduling_units['SU A-2'].scheduling_constraints_doc['time'] = {'at': at_A2.isoformat()}
        scheduling_units['SU A-2'].save()

        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()
        # Assert unit A-2 has been scheduled
        self.assertGreater(len(scheduled_scheduling_units), 0)
        scheduled_scheduling_unit = scheduled_scheduling_units[0]
        # Assert the scheduling_unit has been scheduled
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual('SU A-2', scheduled_scheduling_unit.name)
        self.assertEqual('scheduled', scheduled_scheduling_unit.status.value)
        self.assertEqual(queue_A, scheduled_scheduling_unit.priority_queue)
        self.assertEqual(at_A2, scheduled_scheduling_unit.scheduled_start_time)

        # try again to schedule A-2, but now leave a large enough gap for a single B unit in between
        at_A2 = round_to_minute_precision(scheduling_units['SU A-1'].scheduled_observation_stop_time + 2*DEFAULT_INTER_OBSERVATION_GAP + timedelta(seconds=DURATION) + timedelta(minutes=1))
        scheduling_units['SU A-2'] = unschedule_subtasks_in_scheduling_unit_blueprint(scheduling_units['SU A-2'])
        scheduling_units['SU A-2'].scheduling_constraints_doc['time'] = {'at': at_A2.isoformat()}
        scheduling_units['SU A-2'].save()

        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()
        # Assert unit A-2 has been scheduled, and unit B-x in between
        self.assertEqual(3, len(scheduled_scheduling_units))
        # check that the first scheduled unit is from the A-queue
        scheduled_scheduling_unit1 = scheduled_scheduling_units[0]
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual('SU A-2', scheduled_scheduling_unit1.name)
        self.assertEqual('scheduled', scheduled_scheduling_unit1.status.value)
        self.assertEqual(queue_A, scheduled_scheduling_unit1.priority_queue)
        self.assertEqual(at_A2, scheduled_scheduling_unit1.scheduled_start_time)
        # and check that the 2nd scheduled unit is from the B-queue
        scheduled_scheduling_unit2 = scheduled_scheduling_units[1]
        self.assertIsNotNone(scheduled_scheduling_unit2)
        self.assertEqual('scheduled', scheduled_scheduling_unit2.status.value)
        self.assertEqual(queue_B, scheduled_scheduling_unit2.priority_queue)
        scheduling_units['SU A-1'].refresh_from_db()
        scheduling_units['SU A-2'].refresh_from_db()
        self.assertTrue(scheduled_scheduling_unit2.scheduled_start_time > scheduling_units['SU A-1'].observation_tasks.first().scheduled_stop_time)
        self.assertTrue(scheduled_scheduling_unit2.observation_tasks.first().scheduled_stop_time < scheduling_units['SU A-2'].scheduled_start_time)

        # and try yet again to schedule A-2, but now leave an even bigger gap to fit both B units in between
        at_A2 = round_to_minute_precision(scheduling_units['SU A-1'].observation_tasks.first().scheduled_stop_time + 2*DEFAULT_INTER_OBSERVATION_GAP + timedelta(seconds=2*DURATION) + timedelta(minutes=2))
        scheduling_units['SU A-2'] = unschedule_subtasks_in_scheduling_unit_blueprint(scheduling_units['SU A-2'])
        scheduling_units['SU A-2'].scheduling_constraints_doc['time'] = {'at': at_A2.isoformat()}
        scheduling_units['SU A-2'].save()
        scheduling_units['SU B-1'] = unschedule_subtasks_in_scheduling_unit_blueprint(scheduling_units['SU B-1'])
        scheduling_units['SU B-2'] = unschedule_subtasks_in_scheduling_unit_blueprint(scheduling_units['SU B-2'])

        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()
        # Assert unit A-2 has been scheduled, and unit B-1 and B-2 in between
        self.assertEqual(3, len(scheduled_scheduling_units))
        # check that the first scheduled unit is from the A-queue
        scheduled_scheduling_unit1 = scheduled_scheduling_units[0]
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual('SU A-2', scheduled_scheduling_unit1.name)
        self.assertEqual('scheduled', scheduled_scheduling_unit1.status.value)
        self.assertEqual(queue_A, scheduled_scheduling_unit1.priority_queue)
        self.assertEqual(at_A2, scheduled_scheduling_unit1.scheduled_start_time)
        # and check that the 2nd and 3rd scheduled unit is from the B-queue
        for i in (1, 2):
            scheduled_scheduling_unit2 = scheduled_scheduling_units[i]
            self.assertIsNotNone(scheduled_scheduling_unit2)
            self.assertEqual('scheduled', scheduled_scheduling_unit2.status.value)
            self.assertEqual(queue_B, scheduled_scheduling_unit2.priority_queue)
            scheduling_units['SU A-1'].refresh_from_db()
            scheduling_units['SU A-2'].refresh_from_db()
            self.assertTrue(scheduled_scheduling_unit2.scheduled_start_time > scheduling_units['SU A-1'].observation_tasks.first().scheduled_stop_time)
            self.assertTrue(scheduled_scheduling_unit2.observation_tasks.first().scheduled_stop_time < scheduling_units['SU A-2'].scheduled_start_time)


    def test_weightfactor_update_event_handling(self):
        """
        Check if the event handling in the dynamic scheduler picks up events from weightfactor updates, and reschedules the units.
        """
        # start with weight su_rank=0 and project_rank=1
        wf_unit_rank, created = models.SchedulingConstraintsWeightFactor.objects.get_or_create(constraint_name="unit_rank")
        wf_unit_rank.weight = 0
        wf_unit_rank.save()

        wf_project_rank, created = models.SchedulingConstraintsWeightFactor.objects.get_or_create(constraint_name="project_rank")
        wf_project_rank.weight = 1
        wf_project_rank.save()

        # assert project rank is "higher" for scheduling_set_high than scheduling_set_medium
        self.assertLess(self.scheduling_set_high.project.rank, self.scheduling_set_low.project.rank)

        # create two scheduling units, 1 with high project rank and low su rank, the other vice versa
        for scheduling_set in (self.scheduling_set_high, self.scheduling_set_low):
            project = scheduling_set.project
            # 'invert' the unit_rank in comparison to project
            unit_rank = models.SchedulingUnitRank.LOWEST.value * (models.ProjectRank.LOWEST.value - project.rank) // (models.ProjectRank.LOWEST.value - models.ProjectRank.HIGHEST.value)

            scheduling_unit_draft = self.create_simple_observation_scheduling_unit('SU project_rank=%s unit_rank=%s' % (project.rank, unit_rank),
                                                                                   scheduling_set=scheduling_set,
                                                                                   unit_rank=unit_rank)
            scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

            # assert the unit_rank was properly copied into draft and then into the blueprint
            self.assertEqual(unit_rank, scheduling_unit_blueprint.rank)

            # ignore all other constraints
            scheduling_unit_blueprint.scheduling_constraints_doc = {'scheduler': 'dynamic'}
            scheduling_unit_blueprint.save()


        # preperation complete, we now have 2 schedulable units
        self.assertEqual(2, models.SchedulingUnitBlueprint.objects.count())
        self.assertEqual(2, models.SchedulingUnitBlueprint.objects.filter(status__value=models.SchedulingUnitStatus.Choices.SCHEDULABLE.value).count())

        # start the dynamic_scheduling_service, which includes eventmessage handling, and a running scheduler
        with BusListenerJanitor(create_dynamic_scheduling_service()):
            # upon starting the dynamic_scheduling_service, it should run at least one scheduling round
            # so, as a result the schedulable unit should now get scheduled
            # wait and poll until unit is scheduled, or timeout
            # we expect the scheduling_unit_blueprint for project_high to be scheduled first,
            # because it has a higher project prio, and the project_prio_weight=1 and su_prio_weight=0
            expected_scheduled_unit = models.SchedulingUnitBlueprint.objects.get(draft__scheduling_set__project=self.project_high)
            scheduled_unit_blueprint = wait_for_scheduling_unit_blueprint_status(expected_scheduled_unit.id, models.SchedulingUnitStatus.Choices.SCHEDULED.value)
            self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduled_unit_blueprint.status.value)

            # change the weightfactors (swap them compared to above)
            # as a result weightfactor-update-events are handled by the scheduler, triggering a new scheduling round
            with transaction.atomic():
                wf_unit_rank.weight = 1
                wf_unit_rank.save()
                wf_project_rank.weight = 0
                wf_project_rank.save()

            # now we expect the scheduling_unit_blueprint for project_low to be scheduled
            # because it has a higher su prio, and su_prio_weight=1 and the project_prio_weight=0
            # wait and poll until unit was scheduled, or timeout
            expected_scheduled_unit = models.SchedulingUnitBlueprint.objects.get(draft__scheduling_set__project=self.project_low)
            scheduled_unit_blueprint = wait_for_scheduling_unit_blueprint_status(expected_scheduled_unit.id, models.SchedulingUnitStatus.Choices.SCHEDULED.value)
            self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, scheduled_unit_blueprint.status.value)

            # wait and poll until the other unit was unscheduled, or timeout
            expected_unscheduled_unit = models.SchedulingUnitBlueprint.objects.get(draft__scheduling_set__project=self.project_high)
            unscheduled_unit_blueprint = wait_for_scheduling_unit_blueprint_status(expected_unscheduled_unit.id, models.SchedulingUnitStatus.Choices.SCHEDULABLE.value)
            self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULABLE.value, expected_unscheduled_unit.status.value)


    def test_unit_rank(self):
        """
        When all other constraints/priorities/queues are equal, the scheduling_unit's individual rank should determine the order.
        """
        for rank in (models.SchedulingUnitRank.LOWEST.value, models.SchedulingUnitRank.HIGHEST.value):
            scheduling_unit_draft = self.create_simple_observation_scheduling_unit('scheduling_unit rank=%s' % (rank,), scheduling_set=self.scheduling_set_low)
            scheduling_unit_draft.rank = rank
            scheduling_unit_draft.save()

            scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
            # ignore all other constraints
            scheduling_unit_blueprint.scheduling_constraints_doc = {'scheduler': 'dynamic'}
            scheduling_unit_blueprint.save()

            self.assertEqual(scheduling_unit_draft.rank, scheduling_unit_blueprint.rank)

        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()
        scheduled_scheduling_unit = scheduled_scheduling_units[0]

        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduled_scheduling_unit.status.value, 'scheduled')
        self.assertEqual(models.SchedulingUnitRank.HIGHEST.value, scheduled_scheduling_unit.rank)
        cancel_scheduling_unit_blueprint(scheduled_scheduling_unit)
        unschedule_subtasks_in_scheduling_unit_blueprint(scheduled_scheduling_units[1])

        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()
        scheduled_scheduling_unit = scheduled_scheduling_units[0]
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduled_scheduling_unit.status.value, 'scheduled')
        self.assertEqual(models.SchedulingUnitRank.LOWEST.value, scheduled_scheduling_unit.rank)


    def test_fixed_time_constraint_is_preventing_scheduling_unit_from_being_scheduled_dynamically(self):
        scheduling_unit_draft_fixed_time = self.create_simple_observation_scheduling_unit("scheduling unit fixed_time low", scheduling_set=self.scheduling_set_low,
                                                                                      constraints={'scheduler': 'fixed_time',
                                                                                                   'time':{'at': (datetime.utcnow()+timedelta(minutes=1)).isoformat()}})
        scheduling_unit_blueprint_fixed_time = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft_fixed_time)
        self.assertEqual(scheduling_unit_blueprint_fixed_time.status.value, "schedulable")

        # call the method-under-test.
        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()

        # we expect no scheduling_unit to be scheduled, because the only one is set to 'fixed_time' constraint
        self.assertEqual(0, len(scheduled_scheduling_units))

        # check the results
        scheduling_unit_blueprint_fixed_time.refresh_from_db()
        self.assertEqual(scheduling_unit_blueprint_fixed_time.status.value, 'schedulable')


    def test_fixed_time_scheduled_blocking_dynamically_scheduled(self):
        # override: ensure these constraints play no role in scoring to enforce deterministic behaviour
        for constraint_name in ('density', 'time.at', 'sky.transit_offset'):
            weight_factor, created = models.SchedulingConstraintsWeightFactor.objects.get_or_create(constraint_name=constraint_name)
            weight_factor.weight = 0
            weight_factor.save()


        scheduling_unit_draft_fixed_time = self.create_simple_observation_scheduling_unit("scheduling unit fixed_time", scheduling_set=self.scheduling_set_low)
        scheduling_unit_blueprint_fixed_time = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft_fixed_time)
        scheduling_unit_blueprint_fixed_time.scheduling_constraints_doc = {'scheduler': 'fixed_time',
                                                                           'time': {'at': (datetime.utcnow() + timedelta(minutes=5)).isoformat()}}
        scheduling_unit_blueprint_fixed_time.save()
        self.assertEqual(scheduling_unit_blueprint_fixed_time.status.value, "schedulable")

        scheduling_unit_blueprint_fixed_time = schedule_independent_subtasks_in_scheduling_unit_blueprint(scheduling_unit_blueprint_fixed_time, datetime.utcnow() + timedelta(minutes=5))
        self.assertEqual(scheduling_unit_blueprint_fixed_time.status.value, "scheduled")

        # create a dynamically scheduled unit.
        # force it to be scheduled before the scheduling_unit_blueprint_fixed_time stops (using time.before constraint)
        # this cannot be scheduled, which is what we want to check
        scheduling_unit_draft_dynamic = self.create_simple_observation_scheduling_unit("scheduling unit dynamic", scheduling_set=self.scheduling_set_high, obs_duration=3600)
        scheduling_unit_blueprint_dynamic = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft_dynamic)
        scheduling_unit_blueprint_dynamic.scheduling_constraints_doc = {'scheduler': 'dynamic',
                                                                        'time': { 'before': scheduling_unit_blueprint_fixed_time.scheduled_observation_stop_time.isoformat()}}
        scheduling_unit_blueprint_dynamic.save()

        # before the scheduler had a go at it, this scheduling_unit_blueprint_dynamic is/shouldbe schedulable
        self.assertEqual(scheduling_unit_blueprint_dynamic.status.value, 'schedulable')

        # call the method-under-test.
        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()

        # we expect the no scheduling_unit was scheduled, because the fixed_time is in the way
        self.assertEqual(0, len(scheduled_scheduling_units))

        # now that the scheduler had a go at it, this scheduling_unit_blueprint_dynamic is/shouldbe unschedulable
        scheduling_unit_blueprint_dynamic.refresh_from_db()
        self.assertEqual(scheduling_unit_blueprint_dynamic.status.value, 'unschedulable')

        # relax the constraint (remove the 'time.before') and try again
        # in the event driven scheduler this would result in a schedulable unit again. (for now, set it manually)
        scheduling_unit_blueprint_dynamic.scheduling_constraints_doc = {'scheduler': 'dynamic'}
        scheduling_unit_blueprint_dynamic.save()
        mark_subtasks_in_scheduling_unit_blueprint_as_schedulable(scheduling_unit_blueprint_dynamic)
        wipe_evaluate_constraints_caches()

        # assure that the scheduling_unit_blueprint_fixed_time is still scheduled (and thus, still in the way)
        scheduling_unit_blueprint_fixed_time.refresh_from_db()
        self.assertEqual(scheduling_unit_blueprint_fixed_time.status.value, "scheduled")

        # call the method-under-test.
        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()
        self.assertEqual(1, len(scheduled_scheduling_units))
        scheduled_scheduling_unit = scheduled_scheduling_units[0]
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduling_unit_blueprint_dynamic, scheduled_scheduling_unit)
        scheduling_unit_blueprint_dynamic.refresh_from_db()
        self.assertEqual(scheduling_unit_blueprint_dynamic.status.value, 'scheduled')

        # check scheduling_unit_blueprint_dynamic starts after the scheduling_unit_blueprint_fixed_time
        self.assertGreater(scheduling_unit_blueprint_dynamic.scheduled_start_time, scheduling_unit_blueprint_fixed_time.scheduled_start_time)

        # ensure DEFAULT_INTER_OBSERVATION_GAP between them
        self.assertGreaterEqual(scheduling_unit_blueprint_dynamic.scheduled_start_time - scheduling_unit_blueprint_fixed_time.scheduled_observation_stop_time, DEFAULT_INTER_OBSERVATION_GAP)


    def test_can_schedule_all_observing_strategy_templates_with_default_constraints(self):
        ''''''
        from lofar.sas.tmss.tmss.tmssapp.tasks import create_scheduling_unit_draft_from_observing_strategy_template

        failed_templates = []
        for strategy_template in models.SchedulingUnitObservingStrategyTemplate.objects.filter(state__value='active').order_by('name', 'version').all():
            try:
                self.clean_environment()

                logger.info("Creating scheduling_unit_draft from strategy_template '%s' version=%s", strategy_template.name, strategy_template.version)
                draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template=strategy_template, scheduling_set=self.scheduling_set_high)
                draft.scheduling_constraints_doc['scheduler'] = 'dynamic'
                if 'time' in draft.scheduling_constraints_doc and 'at' in draft.scheduling_constraints_doc['time']:
                    del draft.scheduling_constraints_doc['time']['at']
                # by default, allow very loose sky constraints which can basically run anywhere
                draft.scheduling_constraints_doc['sky']['transit_offset'] = {'from': -12 * 60 * 60, 'to': 12 * 60 * 60}
                draft.scheduling_constraints_doc['sky']['min_elevation'] = {'target': 0, 'calibrator': 0}
                draft.save()

                logger.info("Creating scheduling_unit_blueprint from draft for strategy_template '%s' version=%s", strategy_template.name, strategy_template.version)
                blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(draft)

                # call the method-under-test: self.scheduler.do_dynamic_schedule
                # this test only checks if each strategy_template with plain default spec and default constraints *can* be scheduled by the dynamic scheduler without any exceptions/errors. The defaults should just work.
                logger.info("Scheduling scheduling_unit_blueprint for strategy_template '%s' version=%s", strategy_template.name, strategy_template.version)
                scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()
                self.assertEqual(1, len(scheduled_scheduling_units))
                scheduled_scheduling_unit = scheduled_scheduling_units[0]
                self.assertIsNotNone(scheduled_scheduling_unit)
                self.assertEqual(blueprint.id, scheduled_scheduling_unit.id)
                self.assertEqual("scheduled", scheduled_scheduling_unit.status.value)
                logger.info("Created and scheduled scheduling_unit_blueprint for strategy_template '%s' version=%s successfully for dynamic scheduler with default constraints", strategy_template.name, strategy_template.version)
            except Exception as e:
                logger.exception("Could not create and dynamically schedule strategy template '%s' version=%s: %s", strategy_template.name, strategy_template.version, e)
                failed_templates.append(strategy_template)

        # this test passes when there are no failed templates
        self.assertEqual(0, len(failed_templates), msg='failed_templates: %s' % ([(t.name, t.version) for t in failed_templates]))

    def test_unschedulable_reasons_due_to_unmet_constraints(self):
        """
        Test if the correct unschedulable_reason is set when a constraint is not met.
        """
        # use a short cycle, far in the future, for which we known how the target behaves in this period (transit, elevation, etc)
        cycle = models.Cycle.objects.create(**Cycle_test_data(start=datetime(2030, 1, 1), stop=datetime(2030, 1, 7)))
        project = models.Project.objects.create(**Project_test_data(name=str(uuid.uuid4()), project_state=models.ProjectState.objects.get(value=models.ProjectState.Choices.ACTIVE.value)))
        project.cycles.add(cycle)
        project.save()
        scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project))
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(str(uuid.uuid4()), scheduling_set=scheduling_set, obs_duration=3600)

        # keep matters simple, use one station
        obs_task_draft = scheduling_unit_draft.task_drafts.first()
        obs_task_draft.specifications_doc['station_configuration']['station_groups'][0]['stations'] = ['CS002']
        obs_task_draft.save()

        scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        # set density_vs_optimal to 1, meaning "as close to transit as possible"
        weight_factor, created = models.SchedulingConstraintsWeightFactor.objects.get_or_create(constraint_name="density_vs_optimal")
        weight_factor.weight = 0
        weight_factor.save()

        for expected_reason_start, expected_specific_reason_part, constraints in (
                # ("constraint time.at='2029-12-31 00:00:00' falls outside of cycle bounds ['2030-01-01 00:00:00', '2030-01-07 00:00:00']", '',  {'scheduler': 'fixed_time', 'time': {'at': (cycle.start - timedelta(hours=24)).isoformat()}}),
                # ("constraint time.at='2029-12-31 00:00:00' falls outside of cycle bounds ['2030-01-01 00:00:00', '2030-01-07 00:00:00']", '',  {'scheduler': 'dynamic', 'time': {'at': (cycle.start - timedelta(hours=24)).isoformat()}}),
                # ("time constraint is not met anywhere between '2030-01-01 00:00:00' and '2030-01-01 00:00:00'", '',  {'scheduler': 'dynamic', 'time': {'before': cycle.start.isoformat()}}),
                # ("time constraint is not met anywhere between '2030-01-07 00:00:00' and '2030-01-07 00:00:00'", '',  {'scheduler': 'dynamic', 'time': {'after': cycle.stop.isoformat()}}),
                # ("sky min_elevation constraint is not met anywhere between '2030-01-01 00:00:00' and '2030-01-07 00:00:00'", "task_name='Observation' station=CS002 target='%s' elevation=" % (scheduling_unit_blueprint.name,),  {'scheduler': 'dynamic', 'sky': {'min_elevation': {'target': 1.57}}}),
                # ("sky min_elevation constraint is not met at '2030-01-01 00:00:00'", "task_name='Observation' station=CS002 target='%s' elevation=" %(scheduling_unit_blueprint.name,),  {'scheduler': 'fixed_time', 'sky': {'min_elevation': {'target': 1.57}}, 'time': {'at': '2030-01-01T00:00:00Z'}}),
                # ("sky min_distance constraint is not met anywhere between '2030-01-01 00:00:00' and '2030-01-07 00:00:00'", 'to body=sun < min_distance=149.9',  {'scheduler': 'dynamic', 'sky': {'min_distance': {'sun': 150 * 3.1415/180.0}}}),
                # ("sky min_distance constraint is not met at '2030-01-01 00:00:00'", 'to body=sun < min_distance=149.9',  {'scheduler': 'fixed_time', 'sky': {'min_distance': {'sun': 150 * 3.1415/180.0}}, 'time': {'at': '2030-01-01T00:00:00Z'}}),
                ("sky transit_offset constraint is not met anywhere between '2030-01-01 00:00:00' and '2030-01-01 07:00:00'", '',  {'scheduler': 'dynamic', 'sky': {'transit_offset': {'from': -900, 'to': 900}}, 'time': {'between': [{'from': '2030-01-01T00:00:00Z', 'to': '2030-01-01T07:00:00Z'}]}}),
                # ("sky transit_offset constraint is not met at 2030-01-01 00:00:00", '',  {'scheduler': 'fixed_time', 'sky': {'transit_offset': {'from': -900, 'to': 900}}, 'time': {'at': '2030-01-01T00:00:00Z'}}),
                # ("daily constraint is not met anywhere between '2030-01-01 00:00:00' and '2030-01-01 07:00:00'", '',  {'scheduler': 'dynamic', 'daily': {'require_day': True }, 'time': {'between': [{'from': '2030-01-01T00:00:00Z', 'to': '2030-01-01T07:00:00Z'}]}}),
                # ("daily constraint is not met at 2030-01-01 00:00:00", '',  {'scheduler': 'fixed_time', 'daily': {'require_day': True }, 'time': {'at': '2030-01-01T00:00:00Z'}}),
                # ("daily constraint is not met anywhere between '2030-01-01 11:00:00' and 2030-01-01 '15:00:00'", '',  {'scheduler': 'dynamic', 'daily': {'require_night': True }, 'time': {'between': [{'from': '2030-01-01T11:00:00Z', 'to': '2030-01-01T15:00:00Z'}]}}),
                # ("daily constraint is not met at 2030-01-01 11:00:00", '',  {'scheduler': 'fixed_time', 'daily': {'require_night': True }, 'time': {'at': '2030-01-01T11:00:00Z'}}),
                # ("daily constraint is not met anywhere between 2030-01-01 07:00:00 and 2030-01-01 09:00:00", '',  {'scheduler': 'dynamic', 'daily': {'avoid_twilight': True }, 'time': {'between': [{'from': '2030-01-01T07:00:00Z', 'to': '2030-01-01T09:00:00Z'}]}}),
                # ("daily constraint is not met at 2030-01-01 08:00:00", '', {'scheduler': 'fixed_time', 'daily': {'avoid_twilight': True}, 'time': {'at': '2030-01-01T08:00:00Z'}})
                ):

            logger.info('---------------------------------------------------------------------\nchecking for expected reason: %s', expected_reason_start)

            # reset unit...
            mark_subtasks_in_scheduling_unit_blueprint_as_schedulable(scheduling_unit_blueprint)
            self.assertEqual('', scheduling_unit_blueprint.unschedulable_reason)
            wipe_evaluate_constraints_caches()

            # set the constraints
            scheduling_unit_blueprint.scheduling_constraints_doc = constraints
            scheduling_unit_blueprint.save()

            # try to schedule, should fail.
            scheduled_scheduling_units = self.scheduler.do_full_schedule_computation()
            self.assertEqual(0, len(scheduled_scheduling_units))

            # Assert the scheduling_unit has not been scheduled and that it has the correct expected unschedulable_reason
            scheduling_unit_blueprint.refresh_from_db()
            self.assertEqual('unschedulable', scheduling_unit_blueprint.status.value)
            self.assertTrue(scheduling_unit_blueprint.unschedulable_reason.startswith(expected_reason_start), msg="expected: %s\nactual:%s"%(expected_reason_start, scheduling_unit_blueprint.unschedulable_reason))
            self.assertTrue(expected_specific_reason_part in scheduling_unit_blueprint.unschedulable_reason, msg="expected: %s\nactual:%s"%(expected_specific_reason_part, scheduling_unit_blueprint.unschedulable_reason))

    def test_clear_unschedulable_reason_TMSS_1881_bugfix(self):
        '''Test if the unschedulable reason is cleared for cleanup task.
        See https://support.astron.nl/jira/browse/TMSS-1881 for the expected behaviour.
        '''
        project = models.Project.objects.create(**Project_test_data(name=str(uuid.uuid4()), auto_ingest=True,
                                                                    project_state=models.ProjectState.objects.get(value=models.ProjectState.Choices.OPENED.value)))
        project.cycles.add(models.Cycle.objects.create(**Cycle_test_data(start=datetime.utcnow() - timedelta(days=365),
                                                                         stop=datetime.utcnow() + timedelta(days=365))))
        scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project))
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="IM HBA - 1 Beam")
        su_draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template, scheduling_set=scheduling_set)
        su_draft.scheduling_constraints_doc['scheduler'] = 'dynamic'
        # by default, allow very loose sky constraints which can basically run anywhere
        su_draft.scheduling_constraints_doc['sky']['transit_offset'] = {'from': -12 * 60 * 60, 'to': 12 * 60 * 60}
        su_draft.scheduling_constraints_doc['sky']['min_elevation'] = {'target': 0, 'calibrator': 0}
        su_draft.scheduling_constraints_doc['sky']['min_distance'] = {'sun': 0, 'moon': 0, 'jupiter': 0}
        su_draft.save()

        su_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(su_draft)

        # start the dynamic_scheduling_service, which includes eventmessage handling, and a running scheduler
        with BusListenerJanitor(create_dynamic_scheduling_service()):
            # at scheduler startup, all units are evaluated for schedulability.
            # assert the scheduling_unit has been marked as unschedulable
            # wait and poll until unit is unschedulable, or timeout
            su_blueprint = wait_for_scheduling_unit_blueprint_status(su_blueprint.id, models.SchedulingUnitStatus.Choices.UNSCHEDULABLE.value, timeout=60)
            self.assertEqual(models.SchedulingUnitStatus.Choices.UNSCHEDULABLE.value, su_blueprint.status.value)

            # all subtasks should have a unschedulable_reason
            self.assertTrue(all([s.unschedulable_reason is not None for s in su_blueprint.subtasks.all()]))
            self.assertTrue(all([s.unschedulable_reason == ('project %s is not active'%(project.name,)) for s in su_blueprint.subtasks.filter(specifications_template__type__value='observation').all()]))
            self.assertTrue(all([s.unschedulable_reason == 'predecessor is unschedulable' for s in su_blueprint.subtasks.exclude(specifications_template__type__value='observation').all()]))

            # check tasks and unit as well
            for t in su_blueprint.task_blueprints.all():
                t.refresh_from_db() # need to refresh, django keeps these cached
                self.assertIsNotNone(t.unschedulable_reason)
                self.assertTrue(len(t.unschedulable_reason) > 0)
            self.assertIsNotNone(su_blueprint.unschedulable_reason)
            self.assertTrue(len(su_blueprint.unschedulable_reason) > 0)

            # set the project's state
            project.project_state = models.ProjectState.objects.get(value=models.ProjectState.Choices.ACTIVE.value)
            project.save()

            # now wait and poll until unit is scheduled, or timeout
            su_blueprint = wait_for_scheduling_unit_blueprint_status(su_blueprint.id, models.SchedulingUnitStatus.Choices.SCHEDULED.value, timeout=60)

            # all subtasks should NOT have a unschedulable_reason anymore
            self.assertTrue(all([s.unschedulable_reason is None for s in su_blueprint.subtasks.all()]))

            # all subtasks should NOT have a unschedulable_reason anymore
            self.assertTrue(all([s.unschedulable_reason is None for s in su_blueprint.subtasks.all()]))

            # check tasks and unit as well
            for t in su_blueprint.task_blueprints.all():
                t.refresh_from_db() # need to refresh, django keeps these cached
                self.assertTrue(t.unschedulable_reason is None or t.unschedulable_reason == '')
            self.assertTrue(su_blueprint.unschedulable_reason is None or su_blueprint.unschedulable_reason == '')

    def test_compare_elevation_and_distance_constraints_with_luci_TMSS_2243_bugfix(self):
        '''Test dynamic scheduling and target-constraint-results against LUCI (LOFAR Unified Calculator for Imaging)
        See https://support.astron.nl/jira/browse/TMSS-2243.
        There are two issues at the same time:
         1) a bug in `coordinates_and_timestamp_to_separation_from_body`, resulting in incorrect distance (angular separation). This has been fixed now.
         2) when the min_distance constraint is not met (as was the case for the reported scheduling units), then the unschedulable reason was not usefull.

        This test uses a specification like https://tmss.lofar.eu/api/scheduling_unit_blueprint/1645/specifications_doc
        But with a min_distance constraint that is tuned to fail, such that we can check the usefulness of the unschedulable_reason.

        For all other constraints and reasons see test_unschedulable_reasons_due_to_unmet_constraints.
        '''
        project = models.Project.objects.create(**Project_test_data(name=str(uuid.uuid4()), project_state=models.ProjectState.objects.get(value=models.ProjectState.Choices.ACTIVE.value)))
        cycle = models.Cycle.objects.create(**Cycle_test_data(name=str(uuid.uuid4()), start=datetime(2020, 1, 1), stop=datetime(2030, 1, 1)))
        project.cycles.set([cycle])
        project.save()

        scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project))
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_version_or_latest(name="IM HBA LoTSS - 2 Beams", version=13)

        # At '2023-01-16 02:10:00' the distance between 'Pointing(angle1=182.963[deg] angle2=19.645[deg] dir_type=J2000 target='P182+20')' and the moon is 46.904[deg]
        # Tune the sky.min_distance.moon to "demand" a larger distance than that.
        spec = {
            "scheduling_constraints_doc": {
                "daily": {
                    "avoid_twilight": False,
                    "require_day": False,
                    "require_night": False
                },
                "scheduler": "fixed_time",
                "sky": {
                    "min_distance": {
                        "jupiter": 0.2617993877991494,
                        "moon": astropy.coordinates.Angle(60, unit=astropy.units.deg).rad,
                        "sun": 0.2617993877991494
                    },
                    "min_elevation": {
                        "calibrator": 0.5235987755982988,
                        "target": 0.5235987755982988
                    },
                    "transit_offset": {
                        "from": -5040,
                        "to": 5040
                    }
                },
                "time": {
                    "at": "2023-01-16T02:00:00.00000Z",
                    "between": [],
                    "not_between": []
                }
            },
            "tasks": {
                "Calibrator Observation 1": {
                    "specifications_doc": {
                        "calibrator": {
                            "pointing": {
                                "angle1": 2.1537413875046543,
                                "angle2": 0.8415517080119604,
                                "direction_type": "J2000",
                                "target": "3C196"
                            }
                        }
                    }
                },
                "Calibrator Observation 2": {
                    "specifications_doc": {
                        "calibrator": {
                            "pointing": {
                                "angle1": 3.714680167406339,
                                "angle2": 0.9111102054650465,
                                "direction_type": "J2000",
                                "target": "3C295"
                            }
                        }
                    }
                },
                "Target Observation": {
                    "specifications_doc": {
                        "duration": 14400,
                        "station_configuration": {
                            "SAPs": [
                                {
                                    "digital_pointing": {
                                        "angle1": 3.1933133043110806,
                                        "angle2": 0.3428675074858804,
                                        "direction_type": "J2000",
                                        "target": "P182+20"
                                    }
                                },
                                {
                                    "digital_pointing": {
                                        "angle1": 3.179489569414763,
                                        "angle2": 0.2551773177288743,
                                        "direction_type": "J2000",
                                        "target": "P182+15"
                                    }
                                }
                            ],
                            "station_groups": [
                                {
                                    "max_nr_missing": 4,
                                    "stations": [
                                        "CS002",
                                        "CS003",
                                        "CS004",
                                        "CS005",
                                        "CS006",
                                        "CS007",
                                        "CS011",
                                        "CS013",
                                        "CS017",
                                        "CS021",
                                        "CS024",
                                        "CS026",
                                        "CS028",
                                        "CS030",
                                        "CS031",
                                        "CS032",
                                        "CS101",
                                        "CS103",
                                        "CS201",
                                        "CS301",
                                        "CS302",
                                        "CS401",
                                        "CS501",
                                        "RS106",
                                        "RS205",
                                        "RS208",
                                        "RS210",
                                        "RS305",
                                        "RS306",
                                        "RS307",
                                        "RS310",
                                        "RS406",
                                        "RS407",
                                        "RS409",
                                        "RS503",
                                        "RS508",
                                        "RS509"
                                    ]
                                },
                                {
                                    "max_nr_missing": 2,
                                    "stations": [
                                        "DE601",
                                        "DE602",
                                        "DE603",
                                        "DE604",
                                        "DE605",
                                        "DE609",
                                        "FR606",
                                        "SE607",
                                        "UK608",
                                        "PL610",
                                        "PL611",
                                        "PL612",
                                        "IE613",
                                        "LV614"
                                    ]
                                },
                                {
                                    "max_nr_missing": 1,
                                    "stations": [
                                        "DE601",
                                        "DE605"
                                    ]
                                }
                            ],
                            "tile_beam": {
                                "angle1": 3.186402236805496,
                                "angle2": 0.2990237700856845,
                                "direction_type": "J2000",
                                "target": "P182+20_P182+15_ref"
                            }
                        }
                    }
                }
            }
        }

        su_draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template, scheduling_set=scheduling_set, specifications_doc_overrides=spec)
        su_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(su_draft)

        # call the 1st method-under-test.
        # The result should be None, as this unit is unschedulable due to unmet min_distance constraints
        scored_scheduling_unit = self.scheduler.find_best_next_schedulable_unit([su_blueprint],
                                                                                datetime(2023, 1, 16, 0, 0, 0),
                                                                                datetime(2023, 1, 17, 0, 0, 0))
        self.assertIsNone(scored_scheduling_unit)
        # Ok, if we made it here, then we've reproduced that the unit cannot be scheduled. (That is correct/expected because the required moon distance is not met)

        # call the 2nd method-under-test.
        su_blueprint = determine_unschedulable_reason_and_mark_unschedulable_if_needed(su_blueprint, datetime(2023, 1, 16, 0, 0, 0), datetime(2023, 1, 17, 0, 0, 0))
        # the unit should be unschedulable...
        self.assertEqual('unschedulable', su_blueprint.status.value)
        logger.info(su_blueprint.unschedulable_reason)
        # generic message about which constraint should be in the reason
        self.assertTrue("sky min_distance constraint is not met at '2023-01-16 02:00:00'" in su_blueprint.unschedulable_reason)
        # specific message(s) about which constraint detail should be in the reason
        self.assertTrue('to body=moon < min_distance=60.000[deg]' in su_blueprint.unschedulable_reason)


    def test_scheduling_near_transit_TMSS_2259_bugfix(self):
        '''Test scheduling works when the 'at' constraint  it at/near transit.
        See https://support.astron.nl/jira/browse/TMSS-2259.
        This test uses a specification from https://tmss.lofar.eu/api/scheduling_unit_blueprint/1662/specifications_doc
        '''
        project = models.Project.objects.create(**Project_test_data(name=str(uuid.uuid4()), project_state=models.ProjectState.objects.get(value=models.ProjectState.Choices.ACTIVE.value)))
        cycle = models.Cycle.objects.create(**Cycle_test_data(name=str(uuid.uuid4()), start=datetime(2020, 1, 1), stop=datetime(2030, 1, 1)))
        project.cycles.set([cycle])
        project.save()

        scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project))
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_version_or_latest(name="IM HBA LoTSS - 2 Beams", version=13)

        spec = {
            "scheduling_constraints_doc": {
                "daily": {
                   "avoid_twilight": False,
                   "require_day": False,
                   "require_night": False
                 },
                 "scheduler": "fixed_time",
                 "sky": {
                   "min_distance": {
                     "jupiter": 0,
                     "moon": 0,
                     "sun": 0
                   },
                   "min_elevation": {
                     "calibrator": 0.25,
                     "target": 0.25
                   },
                   "transit_offset": {
                     "from": -15840,
                     "to": 15840
                   }
                 },
                 "time": {
                   "at": "2023-01-24T02:30:00Z"
                 }
               },
               "tasks": {
                 "Calibrator Observation 1": {
                   "specifications_doc": {
                     "calibrator": {
                       "pointing": {
                         "angle1": 2.1537413875046543,
                         "angle2": 0.8415517080119604,
                         "direction_type": "J2000",
                         "target": "3C196"
                       }
                     }
                   }
                 },
                 "Calibrator Observation 2": {
                   "specifications_doc": {
                     "calibrator": {
                       "pointing": {
                         "angle1": 3.714680167406339,
                         "angle2": 0.9111102054650465,
                         "direction_type": "J2000",
                         "target": "3C295"
                       }
                     }
                   }
                 },
                 "Target Observation": {
                   "specifications_doc": {
                     "duration": 14400,
                     "station_configuration": {
                       "SAPs": [
                         {
                           "digital_pointing": {
                             "angle1": 3.48122499938163,
                             "angle2": 0.34438400468039104,
                             "direction_type": "J2000",
                             "target": "P199+20"
                           }
                         },
                         {
                           "digital_pointing": {
                             "angle1": 3.506401373841648,
                             "angle2": 0.25690034555153757,
                             "direction_type": "J2000",
                             "target": "P200+15"
                           }
                         }
                       ]
                     }
                   }
                 }
               }
             }

        su_draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template, scheduling_set=scheduling_set, specifications_doc_overrides=spec)
        su_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(su_draft)

        # the bug reported in TMSS-2259 states that this unit cannot be scheduled at exactly 2023-01-24T02:30:00Z.
        best_scored_scheduling_unit = self.scheduler.find_best_next_schedulable_unit([su_blueprint], datetime(2023, 1, 24, 0, 0),  datetime(2023, 1, 25, 0, 0))

        # in the first version of this test, it was reproduced that the unit could not be scheduled at/near transit.
        # now, after the fix, we check that it CAN be scheduled, thus that best_scored_scheduling_unit should NOT be None
        self.assertIsNotNone(best_scored_scheduling_unit)

        # Proof that it can actually be scheduled in a wide range of timestamps.
        for timestamp in (datetime(2023, 1, 24, 2, 0),
                          datetime(2023, 1, 24, 2, 25),
                          datetime(2023, 1, 24, 2, 29),
                          datetime(2023, 1, 24, 2, 30),
                          datetime(2023, 1, 24, 2, 35),
                          datetime(2023, 1, 24, 3, 30),
                          datetime(2023, 1, 24, 5, 30)):
            su_blueprint.refresh_from_db()
            su_blueprint.scheduling_constraints_doc['time']['at'] = timestamp.isoformat()
            su_blueprint.save()
            best_scored_scheduling_unit = self.scheduler.find_best_next_schedulable_unit([su_blueprint], datetime(2023, 1, 24, 0, 0),  datetime(2023, 1, 25, 0, 0))
            self.assertIsNotNone(best_scored_scheduling_unit)

    def test_TMSS_2582_bugfix_compute_optimal_start_time(self):
        '''Test if the dynamic scheduled can find an optimal starttime close enough (within 24h) to earliest_starttime, such that the two times can be properly interpolated yielding a runnable weighted_starttime.
        See https://support.astron.nl/jira/browse/TMSS-2582.
        This test uses a specification from https://tmss.lofar.eu/api/scheduling_unit_blueprint/2608/specifications_doc
        '''
        project = models.Project.objects.create(**Project_test_data(name=str(uuid.uuid4()), project_state=models.ProjectState.objects.get(value=models.ProjectState.Choices.ACTIVE.value)))
        cycle = models.Cycle.objects.create(**Cycle_test_data(name=str(uuid.uuid4()), start=datetime(2020, 1, 1), stop=datetime(2030, 1, 1)))
        project.cycles.set([cycle])
        project.save()

        scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project))
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_version_or_latest(name="IM HBA LoTSS - 2 Beams", version=13)

        spec = {
          "scheduling_constraints_doc": {
            "daily": {
              "avoid_twilight": False,
              "require_day": False,
              "require_night": False
            },
            "scheduler": "dynamic",
            "sky": {
              "min_distance": {
                "jupiter": 0.2617993877991494,
                "moon": 0.5235987755982988,
                "sun": 0.5235987755982988
              },
              "min_elevation": {
                "calibrator": 0.5235987755982988,
                "target": 0.5235987755982988
              },
              "reference_pointing": {
                "enabled": True,
                "pointing": {
                  "angle1": 3.6249423913078505,
                  "angle2": 0.7057902055554829,
                  "direction_type": "J2000",
                  "target": "A1763_P211+40_ref"
                }
              },
              "transit_offset": {
                "from": -3600,
                "to": 3600
              }
            },
            "time": {
              # "between": [{"from": "2023-06-10T15:00:00Z", "to": "2023-06-11T03:00:00Z"}]
            }
          },
          "tasks": {
            "Calibrator Observation 1": {
              "short_description": "3c196",
              "specifications_doc": {
                "QA": {"file_conversion": {"enabled": False}, "plots": {"enabled": False}},
                "calibrator": {
                  "pointing": {
                    "angle1": 2.1537387040609293,
                    "angle2": 0.8415523862663001,
                    "direction_type": "J2000",
                    "target": "3c196"
                  }
                }
              }
            },
            "Calibrator Observation 2": {
              "short_description": "3c295",
              "specifications_doc": {
                "QA": {"file_conversion": {"enabled": False}, "plots": {"enabled": False}},
                "calibrator": {
                  "pointing": {
                    "angle1": 3.7146801528619284,
                    "angle2": 0.9111107387600959,
                    "direction_type": "J2000",
                    "target": "3c295"
                  }
                }
              }
            },
            "Target Observation": {
              "short_description": "A1763_P211+40_refP211+40",
              "specifications_doc": {
                "QA": {"file_conversion": {"enabled": False}, "plots": {"enabled": False}},
                "duration": 28800,
                "station_configuration": {
                  "SAPs": [
                    {
                      "digital_pointing": {
                        "angle1": 3.557570056914787,
                        "angle2": 0.715604191939447,
                        "direction_type": "J2000",
                        "target": "A1763_P211+40_ref"
                      }
                    },
                    {
                      "digital_pointing": {
                        "angle1": 3.6923145729846056,
                        "angle2": 0.6959744733574531,
                        "direction_type": "J2000",
                        "target": "P211+40"
                      }
                    }
                  ]
                }
              }
            }
          }
        }

        su_draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template, scheduling_set=scheduling_set, specifications_doc_overrides=spec)
        su_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(su_draft)

        # try to let the scheduler find the optimal starttime in the given search window.
        # this should succeed, and should yield a starttime such that the unit's center is at transit.
        # for this target (ref_pointing), date and station we know the transit is at 20:08UTC (13:50LST), which means a start at 15:56UTC
        # try several search windows where it should fit. It should each time find the same start and lst.
        for lower_bound in (datetime(2023, 6, 10, 14, 0), datetime(2023, 6, 10, 15, 0), datetime(2023, 6, 10, 16, 0)):
            for window_width in (timedelta(hours=12), timedelta(hours=18), timedelta(hours=24)):
                upper_bound = lower_bound + window_width

                # try a set of time_constraints, to see if that has an influence
                for time_constraint in ({},
                                        {"before": "2023-06-11T03:00:00Z"},
                                        {"between": [{"from": "2023-06-10T15:00:00Z", "to": "2023-06-11T03:00:00Z"}]}):
                    # apply the time_constraint
                    su_blueprint.scheduling_constraints_doc['time'] = time_constraint
                    su_blueprint.save()

                    # check what the scheduler finds in the given window, with the given constraints
                    best_scored_scheduling_unit = self.scheduler.find_best_next_schedulable_unit([su_blueprint],
                                                                                                 lower_bound,
                                                                                                 upper_bound)
                    self.assertIsNotNone(best_scored_scheduling_unit)
                    self.assertEqual(su_blueprint.id, best_scored_scheduling_unit.scheduling_unit.id)

                    # check the expected_start_time and expected_central_lst
                    expected_start_time = datetime(2023, 6, 10, 15, 56)
                    self.assertTrue(abs((best_scored_scheduling_unit.start_time - expected_start_time).total_seconds()) < 300)

                    placed_scheduling_unit = update_subtasks_start_times_for_scheduling_unit(best_scored_scheduling_unit.scheduling_unit,
                                                                                             best_scored_scheduling_unit.start_time)
                    self.assertTrue(abs((placed_scheduling_unit.scheduled_start_time - expected_start_time).total_seconds()) < 300)

                    from datetime import date, time, datetime as dt
                    expected_central_lst = dt.combine(date(2023, 6, 10), time(13, 50))
                    actual_central_lst = dt.combine(date(2023, 6, 10), placed_scheduling_unit.main_observation_scheduled_central_lst)
                    self.assertTrue(abs((actual_central_lst - expected_central_lst).total_seconds()) < 300)

                    # for reference, print the schedule (with some nice elvations, transit, etc)
                    self.scheduler.log_schedule()


    def test_bugfix_TMSS_2295_min_elevation(self):
        '''Test the correct evaluation and setting/clearing of the min_elevation constraint for far-away stations.
        See https://support.astron.nl/jira/browse/TMSS-2295.
        This test uses a specification from https://tmss.lofar.eu/api/scheduling_unit_blueprint/1802/specifications_doc_with_just_the_parameters
        '''
        project = models.Project.objects.create(**Project_test_data(name=str(uuid.uuid4()), project_state=models.ProjectState.objects.get(value=models.ProjectState.Choices.ACTIVE.value)))
        cycle = models.Cycle.objects.create(**Cycle_test_data(name=str(uuid.uuid4()), start=datetime(2020, 1, 1), stop=datetime(2030, 1, 1)))
        project.cycles.set([cycle])
        project.save()

        scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project))
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="IM HBA LoTSS - 2 Beams")

        spec = {
            "scheduling_constraints_doc": {
                "daily": {
                    "avoid_twilight": False,
                    "require_day": False,
                    "require_night": False
            },
            "scheduler": "fixed_time",
            "location": "both",
            "sky": {
              "min_distance": {
                "jupiter": 0.2617993877991494,
                "moon": 0.5235987755982988,
                "sun": 0.5235987755982988
              },
              "min_elevation": {
                "calibrator": 0.3490659,
                "target": 0.5235988
              },
              "transit_offset": {
                "from": -5040,
                "to": 5040
              }
            },
            "time": {
              "at": "2023-02-16T23:22:00.00000Z",
              "between": [],
              "not_between": []
            }
          },
            "tasks": {
            "Calibrator Observation 1": {
              "short_description": "3C196",
              "specifications_doc": {
                "calibrator": {
                  "pointing": {
                    "angle1": 2.1537413875046543,
                    "angle2": 0.8415517080119604,
                    "direction_type": "J2000",
                    "target": "3C196"
                  }
                }
              }
            },
            "Calibrator Observation 2": {
              "short_description": "3C295",
              "specifications_doc": {
                "calibrator": {
                  "pointing": {
                    "angle1": 3.714680167406339,
                    "angle2": 0.9111102054650465,
                    "direction_type": "J2000",
                    "target": "3C295"
                  }
                }
              }
            },
            "Target Observation": {
              "short_description": "BootesAILT_P216+31",
              "specifications_doc": {
                "duration": 28800,
                "station_configuration": {
                  "SAPs": [
                    {
                      "digital_pointing": {
                        "angle1": 3.7824891904504576,
                        "angle2": 0.5956222112502504,
                        "direction_type": "J2000",
                        "target": "BootesA"
                      }
                    },
                    {
                      "digital_pointing": {
                        "angle1": 3.7746890231350867,
                        "angle2": 0.5431120414492766,
                        "direction_type": "J2000",
                        "target": "ILT_P216+31"
                      }
                    }
                  ],
                  "station_groups": [
                    {
                      "max_nr_missing": 4,
                      "stations": [
                        "CS002",
                        "CS003",
                        "CS004",
                        "CS005",
                        "CS006",
                        "CS007",
                        "CS011",
                        "CS013",
                        "CS017",
                        "CS021",
                        "CS024",
                        "CS026",
                        "CS028",
                        "CS030",
                        "CS031",
                        "CS101",
                        "CS103",
                        "CS201",
                        "CS301",
                        "CS302",
                        "CS401",
                        "CS501",
                        "RS106",
                        "RS205",
                        "RS208",
                        "RS210",
                        "RS305",
                        "RS306",
                        "RS307",
                        "RS310",
                        "RS406",
                        "RS407",
                        "RS409",
                        "RS503",
                        "RS508",
                        "RS509"
                      ]
                    },
                    {
                      "max_nr_missing": 2,
                      "stations": [
                        "DE601",
                        "DE602",
                        "DE603",
                        "DE604",
                        "DE605",
                        "DE609",
                        "FR606",
                        "SE607",
                        "UK608",
                        "PL610",
                        "PL611",
                        "PL612",
                        "IE613",
                        "LV614"
                      ]
                    },
                    {
                      "max_nr_missing": 1,
                      "stations": [
                        "DE601",
                        "DE605"
                      ]
                    }
                  ],
                  "tile_beam": {
                    "angle1": 3.778568017397644,
                    "angle2": 0.5693682899025981,
                    "direction_type": "J2000",
                    "target": "BootesAILT_P216+31_ref"
                  }
                }
              }
            }
            }
        }

        su_draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template, scheduling_set=scheduling_set, specifications_doc_overrides=spec)
        su_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(su_draft)

        # let the scheduler try to find this unit as schedulable.
        # it won't find anything, because the target elevation dips below 30deg for IE613
        best_scored_scheduling_unit = self.scheduler.find_best_next_schedulable_unit([su_blueprint], datetime(2023, 2, 16, 23, 0),  datetime(2023, 2, 17, 11, 0))
        self.assertIsNone(best_scored_scheduling_unit)

        # now determine the unschedulable reason..., and check that it now correct after the bugfix.
        su_blueprint = determine_unschedulable_reason_and_mark_unschedulable_if_needed(su_blueprint, datetime(2023, 2, 16, 23, 0),  datetime(2023, 2, 17, 11, 0), gridder=self.scheduler.fine_gridder)
        self.assertEqual(models.SchedulingUnitStatus.Choices.UNSCHEDULABLE.value, su_blueprint.status.value)
        logger.info(su_blueprint.unschedulable_reason)
        # bug fixed?: FR606 should not be in reason anymore, and IE613 should.
        self.assertFalse("FR606" in su_blueprint.unschedulable_reason)
        self.assertTrue("IE613" in su_blueprint.unschedulable_reason)

    def test_bugfix_TMSS_2823_coarse_fine_grid(self):
        '''Test the correct evaluation at both the coarse and fine grid for this tightly-constrained unit.
        See https://support.astron.nl/jira/browse/TMSS-2823.
        This test uses a specification from https://tmss.lofar.eu/api/scheduling_unit_blueprint/3703/specifications_doc_with_just_the_parameters
        '''
        project = models.Project.objects.create(**Project_test_data(name=str(uuid.uuid4()), project_state=models.ProjectState.objects.get(value=models.ProjectState.Choices.ACTIVE.value)))
        cycle = models.Cycle.objects.create(**Cycle_test_data(name=str(uuid.uuid4()), start=datetime(2020, 1, 1), stop=datetime(2030, 1, 1)))
        project.cycles.set([cycle])
        project.save()

        scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project))
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="IM LBA Survey - 3 Beams")

        spec = {
            "scheduling_constraints_doc": {
            "daily": {
              "avoid_twilight": True,
              "require_day": False,
              "require_night": True
            },
            "location": "center",
            "scheduler": "dynamic",
            "sky": {
              "min_distance": {
                "jupiter": 0.2617993877991494,
                "moon": 0.5235987755982988,
                "sun": 0.5235987755982988
              },
              "min_elevation": {
                "calibrator": 0.5235987755982988,
                "target": 0.6108652381980153
              },
              "reference_pointing": {
                "enabled": False,
                "pointing": {
                  "angle1": 0,
                  "angle2": 0,
                  "direction_type": "J2000",
                  "target": "_target_name_"
                }
              },
              "transit_offset": {
                "from": -8100,
                "to": -7500
              }
            },
            "time": {
              "between": [],
              "not_between": [],
                # "at": "2024-02-02T02:27:12Z"
            }
          },
          "tasks": {
            "Combined Observation": {
              "short_description": "O4.1_field-1_field-3_field-7",
              "specifications_doc": {
                "QA": {
                  "file_conversion": {
                    "enabled": False
                  },
                  "plots": {
                    "enabled": False
                  }
                },
                "calibrator": {
                  "pointing": {
                    "angle1": 3.714676749469887,
                    "angle2": 0.9111089934308438,
                    "direction_type": "J2000",
                    "target": "3c295"
                  }
                },
                "duration": 7200,
                "station_configuration": {
                  "SAPs": [
                    {
                      "digital_pointing": {
                        "angle1": 3.319577840049671,
                        "angle2": 0.25591239223217255,
                        "direction_type": "J2000",
                        "target": "field-1"
                      }
                    },
                    {
                      "digital_pointing": {
                        "angle1": 3.2392385892511197,
                        "angle2": 0.21168225832813228,
                        "direction_type": "J2000",
                        "target": "field-3"
                      }
                    },
                    {
                      "digital_pointing": {
                        "angle1": 3.291280816887087,
                        "angle2": 0.12433376525357202,
                        "direction_type": "J2000",
                        "target": "field-7"
                      }
                    }
                  ],
                  "station_groups": [
                    {
                      "max_nr_missing": 0,
                      "stations": [
                        "CS002"
                      ]
                    }
                  ]
                }
              }
            }
          }
        }

        # the original spec from https://tmss.lofar.eu/api/scheduling_unit_blueprint/3703/specifications_doc_with_just_the_parameters is not accepted with the new validation
        # so, proof that the validation works, and then adapt the spec for the rest of the test.
        with self.assertRaises(RuleValidationException) as context:
            su_draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template, scheduling_set=scheduling_set, specifications_doc_overrides=spec)
        self.assertTrue(str(context.exception).startswith("time between latest and earliest transit for targets in task"))

        # adapt the spec for the rest of the test.
        spec['scheduling_constraints_doc']['sky']['transit_offset'] = {"from": -1800, "to": 1800}
        su_draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template, scheduling_set=scheduling_set, specifications_doc_overrides=spec)
        su_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(su_draft)

        # let the scheduler search for this unit in a known schedulable window.
        # it should find this unit (and thus, in scheduler.do_dynamic_schedule it should get scheduled)
        best_next_schedulable_unit = self.scheduler.find_best_next_schedulable_unit([su_blueprint], datetime(2024, 2, 16), datetime(2024, 2, 17))
        self.assertIsNotNone(best_next_schedulable_unit)
        self.assertEqual(su_blueprint.id, best_next_schedulable_unit.scheduling_unit.id)


    def test_scheduling_blocks_outside_observation_window_TMSS_2298_bugfix(self):
        '''When scheduling two units, one a bit after the other, the first units erroneously blocks the second outside its observation window.
        See https://support.astron.nl/jira/browse/TMSS-2298.
        This test uses a specification from https://tmss.lofar.eu/api/scheduling_unit_blueprint/1977/specifications_doc
        and https://tmss.lofar.eu/api/scheduling_unit_blueprint/1978/specifications_doc
        '''
        project = models.Project.objects.create(**Project_test_data(name=str(uuid.uuid4()), project_state=models.ProjectState.objects.get(value=models.ProjectState.Choices.ACTIVE.value)))
        cycle = models.Cycle.objects.create(**Cycle_test_data(name=str(uuid.uuid4()), start=datetime(2020, 1, 1), stop=datetime(2030, 1, 1)))
        project.cycles.set([cycle])
        project.save()

        scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project))
        strategy_template_1beam = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="IM HBA - 1 Beam")
        strategy_template_2beams = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="IM HBA LoTSS - 2 Beams")

        spec_1977 = {
            "scheduling_constraints_doc": {
                "daily": {
                    "avoid_twilight": False,
                    "require_day": False,
                    "require_night": False
                },
                "scheduler": "dynamic",
                "sky": {
                    "min_distance": {
                        "jupiter": 0.5000368306963754,
                        "moon": 0.5000368306963754,
                        "sun": 0.5000368306963754
                    },
                    "min_elevation": {
                        "calibrator": 0,
                        "target": 0
                    },
                    "transit_offset": {
                        "from": -43200,
                        "to": 43200
                    }
                },
                "time": {
                }
            },
            "tasks": {
                "Calibrator Observation 1": {
                    "short_description": "Cal1 3Cabc",
                    "specifications_doc": {
                        "QA": {"file_conversion": {"enabled": False}, "plots": {"enabled": False}},
                        "calibrator": {
                            "pointing": {
                                "angle1": 0.6624317181687094,
                                "angle2": 1.5579526427549426,
                                "direction_type": "J2000",
                                "target": "3Cabc"
                            }
                        }
                    }
                },
                "Calibrator Observation 2": {
                    "short_description": "Cal2 3Cdef",
                    "specifications_doc": {
                        "QA": {"file_conversion": {"enabled": False}, "plots": {"enabled": False}},
                        "calibrator": {
                            "pointing": {
                                "angle1": 0.6624317181687094,
                                "angle2": 1.5579526427549426,
                                "direction_type": "J2000",
                                "target": "3Cdef"
                            }
                        }
                    }
                },
                "Target Observation": {
                    "short_description": "target obs",
                    "specifications_doc": {
                        "QA": {"file_conversion": {"enabled": False}, "plots": {"enabled": False}},
                        "duration": 120,
                        "station_configuration": {
                            "SAPs": [ {
                                "digital_pointing": {
                                    "angle1": 0.6624317181687094,
                                    "angle2": 1.5579526427549426,
                                    "direction_type": "J2000",
                                    "target": "target1"
                                }
                            } ]
                        }
                    }
                }
            }
        }

        su_draft_1977 = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template_1beam,
                                                                                      scheduling_set=scheduling_set,
                                                                                      specifications_doc_overrides=spec_1977)
        su_blueprint_1977 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(su_draft_1977)

        # schedule it dynamically to find a nice timestamp where it will be able to run
        scheduled_unit = self.scheduler.schedule_next_scheduling_unit()
        self.assertIsNotNone(scheduled_unit)
        self.assertEqual(su_blueprint_1977.id, scheduled_unit.id)

        su_blueprint_1977.refresh_from_db()
        su_blueprint_1977.scheduling_constraints_doc['time']['at'] = su_blueprint_1977.scheduled_start_time.isoformat()
        su_blueprint_1977.scheduling_constraints_doc['scheduler'] = 'fixed_time'
        su_blueprint_1977.save()

        # unschedule and schedule it again, but this time at the given fixed_time
        unschedule_subtasks_in_scheduling_unit_blueprint(su_blueprint_1977)
        scheduled_units = self.scheduler.schedule_fixed_time_scheduling_units()
        self.assertGreater(len(scheduled_units), 0)
        scheduled_unit = scheduled_units[0]
        self.assertIsNotNone(scheduled_unit)
        self.assertEqual(su_blueprint_1977.id, scheduled_unit.id)


        spec_1978 = {
            "scheduling_constraints_doc": {
                "daily": {
                    "avoid_twilight": False,
                    "require_day": False,
                    "require_night": False
                },
                "scheduler": "fixed_time",
                "sky": {
                    "min_distance": {
                        "jupiter": 0.2617993877991494,
                        "moon": 0.5235987755982988,
                        "sun": 0.5235987755982988
                    },
                    "min_elevation": {
                        "calibrator": 0.5235987755982988,
                        "target": 0.5235987755982988
                    },
                    "transit_offset": {
                        "from": -37440,
                        "to": 37440
                    }
                },
                "time": {
                    "at": "2023-03-14T17:35:00.00000Z"
                }
            },
            "tasks": {
                "Calibrator Observation 1": {
                    "specifications_doc": {
                        "QA": {"file_conversion": {"enabled": False}, "plots": {"enabled": False}},
                        "calibrator": {
                            "pointing": {
                                "angle1": 0.6624317181687094,
                                "angle2": 1.5579526427549426,
                                "direction_type": "J2000",
                                "target": "3Cabc"
                            }
                        }
                    }
                },
                "Calibrator Observation 2": {
                    "specifications_doc": {
                        "QA": {"file_conversion": {"enabled": False}, "plots": {"enabled": False}},
                        "calibrator": {
                            "pointing": {
                                "angle1": 0.6624317181687094,
                                "angle2": 1.5579526427549426,
                                "direction_type": "J2000",
                                "target": "3Cdef"
                            }
                        }
                    }
                },
                "Target Observation": {
                    "specifications_doc": {
                        "QA": {"file_conversion": {"enabled": False}, "plots": {"enabled": False}},
                        "duration": 120,
                        "station_configuration": {
                            "SAPs": [
                                {
                                    "digital_pointing": {
                                        "angle1": 0.6624317181687094,
                                        "angle2": 1.5579526427549426,
                                        "direction_type": "J2000",
                                        "target": "Paaa+01"
                                    }
                                },
                                {
                                    "digital_pointing": {
                                        "angle1": 0.6624317181687094,
                                        "angle2": 1.5579526427549426,
                                        "direction_type": "J2000",
                                        "target": "Paaa+02"
                                    }
                                }
                            ]
                        }
                    }
                }
            }
        }

        su_draft_1978 = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template_2beams,
                                                                                      scheduling_set=scheduling_set,
                                                                                      specifications_doc_overrides=spec_1978)
        su_blueprint_1978 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(su_draft_1978)

        # try to schedule 1978 5 min after. This should work, but it doesn't, confirming the bug.
        # we know that 2 calibrators are 10min long, the target is 2min long.
        su_blueprint_1977.refresh_from_db()
        su_blueprint_1978.scheduling_constraints_doc['time']['at'] = (su_blueprint_1977.scheduled_start_time +
                                                                      timedelta(minutes=10) + timedelta(minutes=1) +
                                                                      timedelta(minutes=2) + timedelta(minutes=1) +
                                                                      timedelta(minutes=10) +
                                                                      timedelta(minutes=5)).isoformat()
        su_blueprint_1978.scheduling_constraints_doc['scheduler'] = 'fixed_time'
        su_blueprint_1978.save()
        scheduled_units = self.scheduler.schedule_fixed_time_scheduling_units()
        self.assertGreater(len(scheduled_units), 0)
        scheduled_unit = scheduled_units[0]
        self.assertIsNotNone(scheduled_unit)
        self.assertEqual(su_blueprint_1978.id, scheduled_unit.id)


    def test_bugfix_TMSS_2549_avoid_twilight_constraint_fails_for_LV614_because_the_sun_never_sets(self):
        '''See https://support.astron.nl/jira/browse/TMSS-2549.
        This test uses a specification from https://tmss.lofar.eu/api/scheduling_unit_blueprint/2492/specifications_doc
        '''
        project = models.Project.objects.create(**Project_test_data(name=str(uuid.uuid4()), project_state=models.ProjectState.objects.get(value=models.ProjectState.Choices.ACTIVE.value)))
        cycle = models.Cycle.objects.create(**Cycle_test_data(name=str(uuid.uuid4()), start=datetime(2020, 1, 1), stop=datetime(2030, 1, 1)))
        project.cycles.set([cycle])
        project.save()

        scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project))
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="IM LBA Survey - 3 Beams")

        specifications_doc = {
          "scheduling_constraints_doc": {
            "daily": {
              "avoid_twilight": True,
              "require_day": True,
              "require_night": False
            },
            "scheduler": "fixed_time",
            "sky": {
              "min_distance": {
                "jupiter": 0.2617993877991494,
                "moon": 0.5235987755982988,
                "sun": 0.0872664625997
              },
              "min_elevation": {
                "calibrator": 0.5235987755982988,
                "target": 0.5235987755982988
              },
              "transit_offset": {
                "from": -7200,
                "to": 7200
              }
            },
            "time": {
              "at": "2023-06-05T09:45:00Z",
              "between": [],
              "not_between": []
            }
          },
          "tasks": {
            "Combined Observation": {
              "short_description": "O2.4_3c123_3c130_3c129",
              "specifications_doc": {
                "QA": {
                  "file_conversion": {
                    "enabled": False
                  },
                  "plots": {
                    "enabled": False
                  }
                },
                "calibrator": {
                  "pointing": {
                    "angle1": 2.153739925791406,
                    "angle2": 0.8415523862663001,
                    "direction_type": "J2000",
                    "target": "3c196"
                  }
                },
                "duration": 14400,
                "station_configuration": {
                  "SAPs": [
                    {
                      "digital_pointing": {
                        "angle1": 1.2089582951018352,
                        "angle2": 0.5178479157129776,
                        "direction_type": "J2000",
                        "target": "3c123"
                      }
                    },
                    {
                      "digital_pointing": {
                        "angle1": 1.2779322526996078,
                        "angle2": 0.9089639832953292,
                        "direction_type": "J2000",
                        "target": "3c130"
                      }
                    },
                    {
                      "digital_pointing": {
                        "angle1": 1.2616592390863257,
                        "angle2": 0.7855901496151677,
                        "direction_type": "J2000",
                        "target": "3c129"
                      }
                    }
                  ],
                  "station_groups": [
                    {
                      "max_nr_missing": 5,
                      "stations": [
                        "CS002",
                        "CS003",
                        "CS004",
                        "CS005",
                        "CS006",
                        "CS007",
                        "CS011",
                        "CS013",
                        "CS017",
                        "CS021",
                        "CS024",
                        "CS026",
                        "CS028",
                        "CS030",
                        "CS031",
                        "CS032",
                        "CS101",
                        "CS103",
                        "CS201",
                        "CS301",
                        "CS302",
                        "CS401",
                        "CS501",
                        "RS106",
                        "RS205",
                        "RS208",
                        "RS210",
                        "RS305",
                        "RS306",
                        "RS307",
                        "RS310",
                        "RS406",
                        "RS407",
                        "RS409",
                        "RS503",
                        "RS508",
                        "RS509"
                      ]
                    },
                    {
                      "max_nr_missing": 2,
                      "stations": [
                        "DE601",
                        "DE602",
                        "DE603",
                        "DE604",
                        "DE605",
                        "DE609",
                        "FR606",
                        "SE607",
                        "UK608",
                        "PL610",
                        "PL611",
                        "PL612",
                        "IE613",
                        "LV614"
                      ]
                    },
                    {
                      "max_nr_missing": 0,
                      "stations": [
                        "RS210",
                        "RS310"
                      ]
                    },
                    {
                      "max_nr_missing": 1,
                      "stations": [
                        "RS508",
                        "RS509"
                      ]
                    }
                  ]
                }
              }
            }
          }
        }

        su_draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template,
                                                                                 scheduling_set=scheduling_set,
                                                                                 specifications_doc_overrides=specifications_doc)
        su_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(su_draft)

        # schedule it dynamically to find a nice timestamp where it will be able to run
        scored_scheduling_unit = self.scheduler.find_best_next_schedulable_unit([su_blueprint], datetime(2023, 6, 5, 9, 0), datetime(2023, 6, 5, 21, 0))
        self.assertIsNotNone(scored_scheduling_unit)
        self.assertEqual(su_blueprint.id, scored_scheduling_unit.scheduling_unit.id)


    def test_bugfix_TMSS_2597_incorrect_unschedulable_reason_for_reserved_stations(self):
        '''See https://support.astron.nl/jira/browse/TMSS-2597.
        '''
        project = models.Project.objects.create(**Project_test_data(name=str(uuid.uuid4()), project_state=models.ProjectState.objects.get(value=models.ProjectState.Choices.ACTIVE.value)))
        # create a short cycle, so the midterm scheduler is done quickly and the bug is triggered
        cycle = models.Cycle.objects.create(**Cycle_test_data(name=str(uuid.uuid4()),
                                                              start=datetime.utcnow(),
                                                              stop=datetime.utcnow()+timedelta(days=1)))
        project.cycles.set([cycle])
        project.save()

        scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project))
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit("scheduling unit for %s" % self._testMethodName,
                                                                               scheduling_set=scheduling_set)
        scheduling_unit_draft.scheduling_constraints_doc = {'scheduler': 'dynamic',
                                                            'sky': {'min_elevation': {'target': 1.57}}} # unmeetable constraint
        scheduling_unit_draft.save()
        scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        # create a very short reservation, which is in the cycle window, but leaving plenty of room for the unit to be placed.
        self.create_station_reservation(name="reservation for %s" % self._testMethodName,
                                        stations=scheduling_unit_blueprint.main_observation_specified_stations,
                                        start_time=datetime.utcnow()+timedelta(hours=1),
                                        stop_time =datetime.utcnow()+timedelta(hours=1, minutes=1))

        # let the scheduler try to schedule the unit
        # it should not find a spot (because of the high min_elevation)
        self.scheduler.do_dynamic_schedule()

        # check that it's unschedulable
        scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual('unschedulable', scheduling_unit_blueprint.status.value)

        # check that the reported bug is gone.
        # the unschedulable_reason used to state that the unit is unschedulable due to station reservations.
        self.assertNotEqual("Stations  are reserved", scheduling_unit_blueprint.unschedulable_reason)

        # we want the unschedulable_reason to report on the min_elevation, if this passes the bug is resolved.
        self.assertTrue("min_elevation" in scheduling_unit_blueprint.unschedulable_reason)


    def test_bugfix_TMSS_2618_order_of_scheduling_parallel_observations_and_reservations_should_not_matter(self):
        '''See https://support.astron.nl/jira/browse/TMSS-2618.
        '''
        # setup two projects in one cycle
        project_idols = models.Project.objects.create(**Project_test_data(name="IDOLS_"+str(uuid.uuid4()), project_state=models.ProjectState.objects.get(value=models.ProjectState.Choices.ACTIVE.value)))
        project_other = models.Project.objects.create(**Project_test_data(name=str(uuid.uuid4()), project_state=models.ProjectState.objects.get(value=models.ProjectState.Choices.ACTIVE.value)))
        # create a short cycle, so the midterm scheduler is done quickly and the bug is triggered
        cycle = models.Cycle.objects.create(**Cycle_test_data(name=str(uuid.uuid4()),
                                                              start=datetime.utcnow(),
                                                              stop=datetime.utcnow()+timedelta(days=7)))
        cycle.projects.set([project_idols, project_other])
        cycle.save()

        # prepare scheduling sets
        scheduling_set_idols = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project_idols))
        scheduling_set_other = models.SchedulingSet.objects.create(**SchedulingSet_test_data(project=project_other))

        # use short night where the constraints match on any calendar day
        night_start = datetime.utcnow().replace(hour=23, minute=0, second=0, microsecond=0)
        night_end = night_start + timedelta(hours=2)

        # prepare scheduling units
        # prepare scheduling unit 1, idols, using CS0032 only
        strategy_idols = models.SchedulingUnitObservingStrategyTemplate.get_latest('IDOLS nighttime')
        spec_doc_idols = strategy_idols.template_doc_with_just_the_parameters
        spec_doc_idols['scheduling_constraints_doc']['scheduler'] = 'fixed_time'
        spec_doc_idols['scheduling_constraints_doc']['time']['at'] = night_start.strftime('%Y-%m-%dT%H:%M:%SZ')
        spec_doc_idols['tasks']['BF 1 station LBA']['specifications_doc']['duration'] = (night_end - night_start).total_seconds()
        spec_doc_idols['tasks']['BF 1 station LBA']['specifications_doc']['station_configuration']['station_groups'] = [{'max_nr_missing': 0, 'stations': ['CS032']}]
        su_draft_idols = create_scheduling_unit_draft_from_observing_strategy_template(strategy_idols, scheduling_set_idols, specifications_doc_overrides=spec_doc_idols)
        su_idols = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(su_draft_idols)
        self.assertIn('CS032', su_idols.main_observation_specified_stations)

        # prepare scheduling unit 2, the other. A short BF obs in the middle of the night
        # the template uses all dutch station, including CS032, but losing one is ok.
        strategy_other = models.SchedulingUnitObservingStrategyTemplate.get_latest('BF Pulsar Timing')
        spec_doc_other = strategy_other.template_doc_with_just_the_parameters
        spec_doc_other['scheduling_constraints_doc']['scheduler'] = 'fixed_time'
        spec_doc_other['scheduling_constraints_doc']['time']['at'] = (night_start + 0.5*(night_end-night_start)).strftime('%Y-%m-%dT%H:%M:%SZ')
        spec_doc_other['scheduling_constraints_doc']['sky'] = {'transit_offset': {'from': -43200,
                                                                                  'to': 43200 },
                                                               'min_distance': {'jupiter': 0,
                                                                                'sun': 0,
                                                                                'moon': 0},
                                                               'min_elevation': { 'target': 0,
                                                                                  'calibrator': 0}}
        spec_doc_other['tasks']['Observation']['specifications_doc']['duration'] = 60
        su_draft_other = create_scheduling_unit_draft_from_observing_strategy_template(strategy_other, scheduling_set_other, specifications_doc_overrides=spec_doc_other)
        su_other = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(su_draft_other)
        self.assertIn('CS032', su_other.main_observation_specified_stations)

        # set all weight factors to 0, except for scheduling_rank so we can control which unit wins
        self.reset_scheduling_constraints_weight_factors(default_value=0, default_age_value=0)
        weight_factor, created = models.SchedulingConstraintsWeightFactor.objects.get_or_create(constraint_name="unit_rank")
        weight_factor.weight = 1
        weight_factor.save()

        # preparation complete
        # now start testing scheduling these
        # test all sequences of units, with and without reservations
        for idols_rank, other_rank in ((models.SchedulingUnitRank.HIGHEST.value, models.SchedulingUnitRank.LOWEST.value),
                                       (models.SchedulingUnitRank.LOWEST.value, models.SchedulingUnitRank.HIGHEST.value)):
            su_idols.rank = idols_rank
            su_idols.save()
            su_other.rank = other_rank
            su_other.save()

            # wipe any lingering reservation(s), and/or scheduled units
            models.Reservation.objects.all().delete()
            unschedule_subtasks_in_scheduling_unit_blueprint(su_idols)
            unschedule_subtasks_in_scheduling_unit_blueprint(su_other)

            # 1st attempt/variant, idols first, then the other, no reservation
            # both should be scheduled, idols should have cs032, and the other should have cs032 removed because it was not available
            su_idols_scheduled = self.scheduler.try_schedule_unit(su_idols, get_at_constraint_timestamp(su_idols))
            su_other_scheduled = self.scheduler.try_schedule_unit(su_other, get_at_constraint_timestamp(su_other))
            self.assertIsNotNone(su_idols_scheduled)
            self.assertIsNotNone(su_other_scheduled)
            self.assertEqual('scheduled', su_idols_scheduled.status.value)
            self.assertEqual('scheduled', su_other_scheduled.status.value)
            self.assertIn('CS032', su_idols_scheduled.main_observation_used_stations)
            self.assertNotIn('CS032', su_other_scheduled.main_observation_used_stations)
            self.scheduler.log_schedule() # for reference

            # revert/reset
            unschedule_subtasks_in_scheduling_unit_blueprint(su_idols)
            unschedule_subtasks_in_scheduling_unit_blueprint(su_other)

            # 2nd attempt/variant, the other first, then idols, no reservation
            # both should be scheduled, idols should have cs032,
            # at first the other should use cs032, but after scheduling idols, cs032 should have been removed by implicit rescheduling
            su_other_scheduled = self.scheduler.try_schedule_unit(su_other, get_at_constraint_timestamp(su_other))
            self.assertIsNotNone(su_other_scheduled)
            self.assertEqual('scheduled', su_other_scheduled.status.value)
            self.assertIn('CS032', su_other_scheduled.main_observation_used_stations)
            # by scheduling the idols one, the other may be unscheduled or rescheduled depending on their scores.
            su_idols_scheduled = self.scheduler.try_schedule_unit(su_idols, get_at_constraint_timestamp(su_idols))
            self.assertIsNotNone(su_idols_scheduled)
            self.assertEqual('scheduled', su_idols_scheduled.status.value)
            self.assertIn('CS032', su_idols_scheduled.main_observation_used_stations)
            su_other.refresh_from_db()
            if su_idols.rank < su_other.rank: # lower rank means higher prio
                # the other was unscheduled, so no need to check if cs032 was taken out
                self.assertEqual('schedulable', su_other.status.value)
            else:
                # the other was (re)scheduled, and cs032 was taken out
                self.assertEqual('scheduled', su_other.status.value)
                self.assertNotIn('CS032', su_other_scheduled.main_observation_used_stations)
            self.scheduler.log_schedule() # for reference

            # revert/reset
            unschedule_subtasks_in_scheduling_unit_blueprint(su_idols)
            unschedule_subtasks_in_scheduling_unit_blueprint(su_other)

            # now test the cases with reservation, which is the best way, because then the user/operator is explicit in what they want.
            reservation_strategy_idols = models.ReservationStrategyTemplate.get_latest('IDOLS')
            reservation_idols = models.Reservation.objects.create(name="IDOLS", description="IDOLS", project=project_idols,
                                                                  specifications_template=reservation_strategy_idols.reservation_template,
                                                                  specifications_doc=reservation_strategy_idols.template,
                                                                  start_time=night_start, stop_time=night_end)

            # 3rd attempt/variant, idols first, then the other, with reservation
            su_idols_scheduled = self.scheduler.try_schedule_unit(su_idols, get_at_constraint_timestamp(su_idols))
            su_other_scheduled = self.scheduler.try_schedule_unit(su_other, get_at_constraint_timestamp(su_other))
            self.assertIsNotNone(su_idols_scheduled)
            self.assertIsNotNone(su_other_scheduled)
            self.assertEqual('scheduled', su_idols_scheduled.status.value)
            self.assertEqual('scheduled', su_other_scheduled.status.value)
            self.assertIn('CS032', su_idols.main_observation_used_stations)
            self.assertNotIn('CS032', su_other_scheduled.main_observation_used_stations)
            self.scheduler.log_schedule() # for reference

            # revert/reset
            unschedule_subtasks_in_scheduling_unit_blueprint(su_idols)
            unschedule_subtasks_in_scheduling_unit_blueprint(su_other)

            # 4th attempt/variant, the other first, then idols, with reservation
            su_other_scheduled = self.scheduler.try_schedule_unit(su_other, get_at_constraint_timestamp(su_other))
            su_idols_scheduled = self.scheduler.try_schedule_unit(su_idols, get_at_constraint_timestamp(su_idols))
            self.assertIsNotNone(su_other_scheduled)
            self.assertIsNotNone(su_idols_scheduled)
            self.assertEqual('scheduled', su_other_scheduled.status.value)
            self.assertEqual('scheduled', su_idols_scheduled.status.value)
            self.assertIn('CS032', su_idols.main_observation_used_stations)
            self.assertNotIn('CS032', su_other_scheduled.main_observation_used_stations)
            self.scheduler.log_schedule() # for reference

    def test_bugfix_TMSS_2638_TMSS_2639_schedule_B_prio_units_in_gaps_with_and_without_overlapping_IDOLS_observation(self):
        '''See https://support.astron.nl/jira/browse/TMSS-2638
           and https://support.astron.nl/jira/browse/TMSS-2639
        '''
        # use short day from 11:00 until 13:00 where the constraints match on any calendar day
        day_start = (datetime.utcnow()+timedelta(days=1)).replace(hour=11, minute=0, second=0, microsecond=0)
        day_end = day_start + timedelta(hours=2)

        # prepare helper scheduling units which should define the gap into which the B-prio unit should be placed later
        sud_fixed_one = self.create_simple_observation_scheduling_unit('su_fixed_one', obs_duration=3600, scheduling_set=self.scheduling_set_high)
        sud_fixed_one.scheduling_constraints_doc = {'scheduler': 'fixed_time', 'time': {'at': (day_start-timedelta(minutes=75)).isoformat()} }
        sud_fixed_one.save()
        sub_fixed_one = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(sud_fixed_one)
        self.assertEqual('A', sub_fixed_one.priority_queue.value)

        sud_fixed_two = self.create_simple_observation_scheduling_unit('su_fixed_two', obs_duration=3600, scheduling_set=self.scheduling_set_high)
        sud_fixed_two.scheduling_constraints_doc = {'scheduler': 'fixed_time', 'time': {'at': (day_end+timedelta(minutes=15)).isoformat()} }
        sud_fixed_two.save()
        sub_fixed_two = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(sud_fixed_two)
        self.assertEqual('A', sub_fixed_two.priority_queue.value)

        scheduled_units = self.scheduler.schedule_fixed_time_scheduling_units()
        self.assertEqual(2, len(scheduled_units))
        self.assertTrue(sub_fixed_one in scheduled_units)
        self.assertTrue(sub_fixed_two in scheduled_units)
        sub_fixed_one.refresh_from_db()
        sub_fixed_two.refresh_from_db()

        # check if the two units were scheduled at the expected times,
        # and that they thus leave an expected gap open in between them
        expected_gap = (day_start-timedelta(minutes=75)+timedelta(hours=1), day_end+timedelta(minutes=15))
        self.assertEqual(expected_gap[0], sub_fixed_one.scheduled_observation_stop_time)
        self.assertEqual(expected_gap[1], sub_fixed_two.scheduled_observation_start_time)

        # check that the method get_gaps_to_previous_and_next_observations_in_scheduling_unit yields the same gaps
        # for unit one
        first_gap, second_gap = get_gaps_to_previous_and_next_observations_in_scheduling_unit(sub_fixed_one)
        self.assertEqual(datetime.min, first_gap[0])
        self.assertEqual(sub_fixed_one.scheduled_start_time, first_gap[1])
        self.assertEqual(expected_gap[0], second_gap[0])
        self.assertEqual(expected_gap[1], second_gap[1])
        # for unit two
        first_gap, second_gap = get_gaps_to_previous_and_next_observations_in_scheduling_unit(sub_fixed_two)
        self.assertEqual(expected_gap[0], first_gap[0])
        self.assertEqual(expected_gap[1], first_gap[1])
        self.assertEqual(sub_fixed_two.scheduled_stop_time, second_gap[0])
        self.assertEqual(datetime.max, second_gap[1])

        # everything ok so far
        # we have two scheduled fixed units with an expected gap in between.
        # now let's add a short B-prio unit, and scheduled it in the gap
        # do give it a between constraint to prevent it being scheduled before sub_fixed_one, or after sub_fixed_two
        sud_B = self.create_simple_observation_scheduling_unit('su_B',
                                                               obs_duration=600,
                                                               priority_queue=models.PriorityQueueType.objects.get(value=models.PriorityQueueType.Choices.B.value),
                                                               scheduling_set=self.scheduling_set_high)
        sud_B.scheduling_constraints_doc = {'scheduler': 'dynamic', 'time': {'between': [{'from': sub_fixed_one.scheduled_start_time.isoformat(), 'to': sub_fixed_two.scheduled_stop_time.isoformat()}]} }
        sud_B.save()
        sub_B = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(sud_B)
        self.assertEqual('B', sub_B.priority_queue.value)

        # and make sure we enable dynamic (B-prio) scheduling for this moment ...
        setting = models.Setting.objects.get(name=models.SystemSettingFlag.Choices.DYNAMIC_SCHEDULING_ENABLED.value)
        setting.value = True
        setting.save()

        # try to schedule it via the schedule_B_priority_units_in_gaps_around_scheduling_unit for both sub_fixed_one and sub_fixed_two
        for sub_fixed in (sub_fixed_one, sub_fixed_two):
            scheduled_units = self.scheduler.schedule_B_priority_units_in_gaps_around_scheduling_unit(sub_fixed)
            self.assertEqual(1, len(scheduled_units))
            self.assertTrue(sub_B in scheduled_units)
            sub_B.refresh_from_db()
            self.assertEqual(sub_fixed_one.scheduled_observation_stop_time+DEFAULT_INTER_OBSERVATION_GAP, sub_B.scheduled_start_time)

            # reset/unschedule it
            sub_B = unschedule_subtasks_in_scheduling_unit_blueprint(sub_B)

        # again, everything ok so far
        # now we have two scheduled fixed units with an expected gap in between.
        # and a scheduled B unit in the gap.

        # add a "long" overlapping/parallel idols scheduling unit, using CS0032 only
        strategy_idols = models.SchedulingUnitObservingStrategyTemplate.get_latest('IDOLS daytime')
        spec_doc_idols = strategy_idols.template_doc_with_just_the_parameters
        spec_doc_idols['scheduling_constraints_doc']['scheduler'] = 'fixed_time'
        spec_doc_idols['scheduling_constraints_doc']['time']['at'] = day_start.strftime('%Y-%m-%dT%H:%M:%SZ')
        spec_doc_idols['tasks']['BF 1 station LBA']['specifications_doc']['duration'] = (day_end - day_start).total_seconds()
        spec_doc_idols['tasks']['BF 1 station LBA']['specifications_doc']['station_configuration']['station_groups'] = [{'max_nr_missing': 0, 'stations': ['CS032']}]
        su_draft_idols = create_scheduling_unit_draft_from_observing_strategy_template(strategy_idols, self.scheduling_set_high, name='IDOLS', specifications_doc_overrides=spec_doc_idols)
        sub_idols = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(su_draft_idols)
        self.assertIn('CS032', sub_idols.main_observation_specified_stations)

        # unschedule all...
        for su in models.SchedulingUnitBlueprint.objects.filter(status__value='scheduled').all():
            unschedule_subtasks_in_scheduling_unit_blueprint(su)

        # ... disable dynamic (B-prio) scheduling for this moment ...
        setting = models.Setting.objects.get(name=models.SystemSettingFlag.Choices.DYNAMIC_SCHEDULING_ENABLED.value)
        setting.value = False
        setting.save()

        # ...and schedule all fixed including the new idols unit.
        scheduled_units = self.scheduler.schedule_fixed_time_scheduling_units()
        self.assertEqual(3, len(scheduled_units))
        self.assertTrue(sub_fixed_one in scheduled_units)
        self.assertTrue(sub_fixed_two in scheduled_units)
        self.assertTrue(sub_idols in scheduled_units)
        sub_fixed_one.refresh_from_db()
        sub_fixed_two.refresh_from_db()
        sub_idols.refresh_from_db()

        # check the start/stop times against the expected_gap again
        self.assertEqual(expected_gap[0], sub_fixed_one.scheduled_observation_stop_time)
        self.assertEqual(expected_gap[1], sub_fixed_two.scheduled_observation_start_time)
        # check the start/stop times of idols
        self.assertEqual(day_start, sub_idols.scheduled_observation_start_time)
        self.assertEqual(day_end, sub_idols.scheduled_observation_stop_time)

        # check the gaps method again with idols present
        # check that the method get_gaps_to_previous_and_next_observations_in_scheduling_unit yields the same gaps
        # for unit one
        first_gap, second_gap = get_gaps_to_previous_and_next_observations_in_scheduling_unit(sub_fixed_one)
        self.assertEqual(datetime.min, first_gap[0])
        self.assertEqual(sub_fixed_one.scheduled_start_time, first_gap[1])
        self.assertEqual(expected_gap[0], second_gap[0])
        self.assertEqual(expected_gap[1], second_gap[1])
        # for unit two
        first_gap, second_gap = get_gaps_to_previous_and_next_observations_in_scheduling_unit(sub_fixed_two)
        self.assertEqual(expected_gap[0], first_gap[0])
        self.assertEqual(expected_gap[1], first_gap[1])
        self.assertEqual(sub_fixed_two.scheduled_stop_time, second_gap[0])
        self.assertEqual(datetime.max, second_gap[1])

        # enable dynamic (B-prio) scheduling for this moment again ...
        setting = models.Setting.objects.get(name=models.SystemSettingFlag.Choices.DYNAMIC_SCHEDULING_ENABLED.value)
        setting.value = True
        setting.save()

        # ... and try to schedule the B-prio unit again via the schedule_B_priority_units_in_gaps_around_scheduling_unit for both sub_fixed_one and sub_fixed_two
        for sub_fixed in (sub_fixed_one, sub_fixed_two):
            scheduled_units = self.scheduler.schedule_B_priority_units_in_gaps_around_scheduling_unit(sub_fixed)
            self.assertEqual(1, len(scheduled_units))
            self.assertTrue(sub_B in scheduled_units)
            sub_B.refresh_from_db()
            self.assertEqual(sub_fixed_one.scheduled_observation_stop_time+DEFAULT_INTER_OBSERVATION_GAP, sub_B.scheduled_start_time)

            # reset/unschedule it
            sub_B = unschedule_subtasks_in_scheduling_unit_blueprint(sub_B)

    def test_competing_units_are_scheduled_non_overlapping(self):
        """ Have a set of similar survey-like units, and make sure they are scheduled one after the other in a non-overlapping way"""

        # prepare the test with a fixed unit, schedule it, and make it run.
        next_full_hour = datetime.utcnow().replace(minute=0, second=0, microsecond=0) + timedelta(hours=1)
        pointing = Pointing(local_sidereal_time_for_utc_and_station(next_full_hour + timedelta(hours=0.5)).rad,
                            Angle(80, 'deg').rad, 'J2000')
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit("fixed and running",
                                                                               scheduling_set=self.scheduling_set_high,
                                                                               obs_duration=3600,
                                                                               pointing=pointing)
        # have loose constraints, except for a tight transit offset of +- 15min
        scheduling_unit_draft.scheduling_constraints_doc = {
            'sky': {
                'min_elevation': {'target': 0},
                'transit_offset': {'from': -900, 'to': 900}
            },
            'daily': {},
            'scheduler': 'fixed_time',
            'time': {"at": next_full_hour.isoformat() }}
        scheduling_unit_draft.save()
        su_fixed = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        # preparation, schedule it
        scheduled_units = self.scheduler.schedule_fixed_time_scheduling_units()
        self.assertEqual(1, len(scheduled_units))
        self.assertEqual(su_fixed.id, scheduled_units[0].id)

        # preparation, run it
        set_subtask_state_following_allowed_transitions(su_fixed.observation_tasks.first().subtasks.first(), 'started')
        su_fixed.refresh_from_db()

        self._set_density_vs_optimal_weight_factor(0)

        # for the actual test, create a total of 9 units
        # each unit is 8 hours, so we spread 3 of them over the 24h period using RA.
        # and we spread the 3 blocks of 3 units over 3 days.
        # the first unit partially overlaps with the running su_fixed unit from above.
        # It should be shifted/scheduled such that it starts right after the su_fixed unit.
        NUM_DAYS = 3
        NUM_UNITS_PER_DAY = 3
        for day in range(NUM_DAYS):
            for block in range(NUM_UNITS_PER_DAY):
                pointing = Pointing(local_sidereal_time_for_utc_and_station(su_fixed.on_sky_stop_time
                                                                            - 0.25*su_fixed.specified_observation_duration
                                                                            + timedelta(hours=4)
                                                                            + timedelta(hours=block*8)).rad,
                                    Angle(80, 'deg').rad, 'J2000')
                scheduling_unit_draft = self.create_simple_observation_scheduling_unit(pointing.as_SkyCoord().to_string('hmsdms'),
                                                                                       scheduling_set=self.scheduling_set_high,
                                                                                       obs_duration=6*3600,
                                                                                       unit_rank=float(day)/float(NUM_DAYS),
                                                                                       pointing=pointing)
                # have loose constraints, except for a relatively tight transit offset of +- 60min
                scheduling_unit_draft.scheduling_constraints_doc = {'sky': {
                                                                        'min_elevation': {'target': 0},
                                                                        'transit_offset': {'from': -3600, 'to': 3600}
                                                                    },
                                                                    'daily': {},
                                                                    'scheduler': 'dynamic',
                                                                    'time': {"between": [], "not_between": []}}
                scheduling_unit_draft.save()

                create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        self.scheduler.log_schedule()

        self.scheduler.do_full_schedule_computation()

        # now check the order of the scheduled units.
        scheduled_units = list(models.SchedulingUnitBlueprint.objects.exclude(id=su_fixed.id).order_by('scheduled_start_time').all())
        first_unit = scheduled_units[0]
        self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, first_unit.status.value)

        # the first unit partially overlaps with the running su_fixed unit from above.
        # It should be shifted/scheduled such that it start right after the su_fixed unit.
        self.assertTrue(first_unit.on_sky_start_time > su_fixed.on_sky_stop_time)

        # 20231107 JS: disabled this check which seems to be dependent on time-of-day/day-of-year.
        # the important non-overlapping check here is already done, namely that first_unit.on_sky_start_time > su_fixed.on_sky_stop_time
        # self.assertTrue((first_unit.on_sky_start_time - su_fixed.on_sky_stop_time) < timedelta(minutes=5))

        # and check that they cover the expected number of days
        last_unit = scheduled_units[-1]
        total_span = last_unit.on_sky_stop_time - first_unit.on_sky_start_time
        self.assertGreaterEqual(total_span, timedelta(days=NUM_DAYS-1))
        self.assertLessEqual(total_span, timedelta(days=NUM_DAYS+1))

        # make sure they don't overlap
        prev_unit = first_unit
        for unit in scheduled_units[1:]:
            self.assertEqual(models.SchedulingUnitStatus.Choices.SCHEDULED.value, unit.status.value)
            self.assertGreater(unit.on_sky_start_time, prev_unit.on_sky_stop_time)
            prev_unit = unit

    # def test_bugfix_TMSS_2847_coarse_vs_fine_gridder(self):
    #
    #     next_full_hour = datetime.utcnow().replace(minute=0, second=0, microsecond=0) + timedelta(hours=1)
    #     pointing = Pointing(local_sidereal_time_for_utc_and_station(next_full_hour + timedelta(hours=4)).rad,
    #                         Angle(80, 'deg').rad, 'J2000')
    #     scheduling_unit_draft = self.create_simple_observation_scheduling_unit(pointing.as_SkyCoord().to_string('hmsdms'),
    #                                                                            scheduling_set=self.scheduling_set_high,
    #                                                                            obs_duration=8 * 3600,
    #                                                                            pointing=pointing)
    #
    #     # have loose constraints, except for a relatively tight transit offset of +- 60min
    #     scheduling_unit_draft.scheduling_constraints_doc = {'sky': {
    #         'min_elevation': {'target': 0},
    #         'transit_offset': {'from': -3600, 'to': 3600}
    #     },
    #         'daily': {},
    #         'scheduler': 'dynamic',
    #         'time': {"between": [], "not_between": []}}
    #     scheduling_unit_draft.save()
    #
    #     scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
    #     self.scheduler.do_full_schedule_computation()


class TestReservedStationsTimeWindows(BaseDynamicSchedulingTestCase):
    """
    Tests for the reserved stations used in dynamic scheduling
    Test with different boundaries of scheduling unit start and stop times
    Reservation 'visualized'
          @ = station reservation start_time, * = station reservation stop_time
               SUB start_time     SUB stop_time                 Expected Result
     1.                |     @ ......|...*                      can NOT run
     2.             @..|..*          |                          can NOT run
     3.                |  @.....*    |                          can NOT run
     4.             @..|.............|......*                   can NOT run
     5.       @......* |             |                          can run
     6.                |             |   @.....*                can run
    """
    @classmethod
    def setUpClass(cls) -> None:
        cls.clean_environment()

        super().setUpClass()

        # create a three re-usable variants scheduling_unit_blueprint, based on the "IM HBA - 1 Beam" strategy
        # one with the default stations groups, the other with just CS002, the last with two core stations
        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="IM HBA LoTSS - 2 Beams")

        scheduling_unit_draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template, cls.scheduling_set_medium, "Test Scheduling Unit many stations")
        cls.scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
        # confirm that the station_groups have
        # - dutch station with max_nr_missing=4
        # - international with max_nr_missing=2
        # - international required with max_nr_missing=1
        # the reservation tests below depend on that assumption
        expected_station_groups = [{'stations': ['CS002', 'CS003', 'CS004', 'CS005', 'CS006', 'CS007', 'CS011', 'CS013', 'CS017', 'CS021', 'CS024', 'CS026', 'CS028', 'CS030', 'CS031', 'CS032', 'CS101', 'CS103', 'CS201', 'CS301', 'CS302', 'CS401', 'CS501', 'RS106', 'RS205', 'RS208', 'RS210', 'RS305', 'RS306', 'RS307', 'RS310', 'RS406', 'RS407', 'RS409', 'RS503', 'RS508', 'RS509'],
                                    'max_nr_missing': 4},
                                   {'stations': ['DE601', 'DE602', 'DE603', 'DE604', 'DE605', 'DE609', 'FR606', 'SE607', 'UK608', 'PL610', 'PL611', 'PL612', 'IE613', 'LV614'],
                                    'max_nr_missing': 2},
                                   {'stations': ['DE601', 'DE605'],
                                    'max_nr_missing': 1}]

        station_groups = cls.scheduling_unit_blueprint.specifications_doc['tasks']['Target Observation']['specifications_doc']['station_configuration']['station_groups']
        assert(expected_station_groups == station_groups)

        # create the other with just CS002
        scheduling_unit_draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template, cls.scheduling_set_medium, "Test Scheduling Unit just CS002")
        obs_task_draft = scheduling_unit_draft.task_drafts.get(name='Target Observation')
        obs_task_draft.specifications_doc['station_configuration']['station_groups'] = [{'stations': ['CS002'], 'max_nr_missing': 0}]
        obs_task_draft.save()
        cls.scheduling_unit_blueprint_cs002 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        # confirm that we have 1 station_groups with only CS002, which may not be missed
        expected_station_groups = [{'stations': ['CS002'], 'max_nr_missing': 0}]
        station_groups = cls.scheduling_unit_blueprint_cs002.specifications_doc['tasks']['Target Observation']['specifications_doc']['station_configuration']['station_groups']
        assert(expected_station_groups == station_groups)

        # create the last with just CS002, CS003 of which 1 may be missied
        scheduling_unit_draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template, cls.scheduling_set_medium, "Test Scheduling Unit just CS002&CS003")
        obs_task_draft = scheduling_unit_draft.task_drafts.get(name='Target Observation')
        obs_task_draft.specifications_doc['station_configuration']['station_groups'] = [{'stations': ['CS002', 'CS003'], 'max_nr_missing': 1}]
        obs_task_draft.save()
        cls.scheduling_unit_blueprint_cs002_cs003 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        # confirm that we have 1 station_groups with only CS002&CS003, of which one may be missed
        expected_station_groups = [{'stations': ['CS002', 'CS003'], 'max_nr_missing': 1}]
        station_groups = cls.scheduling_unit_blueprint_cs002_cs003.specifications_doc['tasks']['Target Observation']['specifications_doc']['station_configuration']['station_groups']
        assert(expected_station_groups == station_groups)

    def setUp(self) -> None:
        # keep the reusable self.scheduling_unit_blueprint,
        # wipe all reservations in between tests, so the tests don't influence each other
        models.Reservation.objects.all().delete()

    @staticmethod
    def set_1_reservation_start_time_gt_sub_start_time_and_stop_time_gt_sub_stop_time(station_reservation, scheduling_unit_blueprint):
        """
        Set (1) reservation start_time > SUB start_time and reservation stop_time > SUB stop_time
        """
        station_reservation.start_time = scheduling_unit_blueprint.scheduled_start_time + timedelta(seconds=1)
        station_reservation.stop_time = scheduling_unit_blueprint.scheduled_observation_stop_time + timedelta(seconds=1)
        station_reservation.save()

    @staticmethod
    def set_2_reservation_start_time_lt_sub_start_time_and_stop_time_lt_sub_stop_time(station_reservation, scheduling_unit_blueprint):
        """
        Set (2) reservation start_time < SUB start_time and reservation stop_time < SUB stop_time
        """
        station_reservation.start_time = scheduling_unit_blueprint.scheduled_start_time - timedelta(seconds=1)
        station_reservation.stop_time = scheduling_unit_blueprint.scheduled_observation_stop_time - timedelta(seconds=1)
        station_reservation.save()

    @staticmethod
    def set_3_reservation_start_time_gt_sub_start_time_and_stop_time_lt_sub_stop_time(station_reservation, scheduling_unit_blueprint):
        """
        Set (3) reservation start_time > SUB start_time and reservation stop_time < SUB stop_time
        """
        station_reservation.start_time = scheduling_unit_blueprint.scheduled_start_time + timedelta(seconds=1)
        station_reservation.stop_time = scheduling_unit_blueprint.scheduled_observation_stop_time - timedelta(seconds=1)
        station_reservation.save()

    @staticmethod
    def set_4_reservation_start_time_lt_sub_start_time_and_stop_time_gt_sub_stop_time(station_reservation, scheduling_unit_blueprint):
        """
        Set (4) reservation start_time < SUB start_time and reservation stop_time > SUB stop_time
        """
        station_reservation.start_time = scheduling_unit_blueprint.scheduled_start_time - timedelta(seconds=1)
        station_reservation.stop_time = scheduling_unit_blueprint.scheduled_observation_stop_time + timedelta(seconds=1)
        station_reservation.save()

    @staticmethod
    def set_5_reservation_start_time_and_stop_time_lt_sub_start_time(station_reservation, scheduling_unit_blueprint):
        """
        Set (5) reservation start_time and reservation stop_time < SUB start_time
        """
        station_reservation.start_time = scheduling_unit_blueprint.scheduled_start_time - timedelta(minutes=60)
        station_reservation.stop_time = scheduling_unit_blueprint.scheduled_start_time - timedelta(seconds=1)
        station_reservation.save()

    @staticmethod
    def set_6_reservation_start_time_and_stop_time_gt_sub_stop_time(station_reservation, scheduling_unit_blueprint):
        """
        Set (6) reservation start_time and reservation stop_time > SUB stop_time
        """
        station_reservation.start_time = scheduling_unit_blueprint.scheduled_observation_stop_time + timedelta(seconds=1)
        station_reservation.stop_time = scheduling_unit_blueprint.scheduled_observation_stop_time + timedelta(minutes=65)
        station_reservation.save()

    def test_one_station_reserved(self):
        """
        Test station reservation when 1 station (CS002) is reserved and station CS002 is used in scheduling_unit
        with different reservation start and stop times.
        """
        reservation_one = self.create_station_reservation("One", ["CS002"])
        # reservation start_time > SUB start_time and reservation stop_time > SUB stop_time
        self.set_1_reservation_start_time_gt_sub_start_time_and_stop_time_gt_sub_stop_time(reservation_one, self.scheduling_unit_blueprint_cs002)
        self.assertFalse(can_run_within_station_reservations(self.scheduling_unit_blueprint_cs002))

        # reservation start_time < SUB start_time and stop_time < SUB stop_time
        self.set_2_reservation_start_time_lt_sub_start_time_and_stop_time_lt_sub_stop_time(reservation_one, self.scheduling_unit_blueprint_cs002)
        self.assertFalse(can_run_within_station_reservations(self.scheduling_unit_blueprint_cs002))

        # reservation start_time > SUB start_time and stop_time < SUB stop_time
        self.set_3_reservation_start_time_gt_sub_start_time_and_stop_time_lt_sub_stop_time(reservation_one, self.scheduling_unit_blueprint_cs002)
        self.assertFalse(can_run_within_station_reservations(self.scheduling_unit_blueprint_cs002))

        # reservation start_time < SUB start_time and stop_time > SUB stop_time
        self.set_4_reservation_start_time_lt_sub_start_time_and_stop_time_gt_sub_stop_time(reservation_one, self.scheduling_unit_blueprint_cs002)
        self.assertFalse(can_run_within_station_reservations(self.scheduling_unit_blueprint_cs002))

        # Reservations outside boundary
        # start_time and stop_time < SUB start_time
        self.set_5_reservation_start_time_and_stop_time_lt_sub_start_time(reservation_one, self.scheduling_unit_blueprint_cs002)
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint_cs002))

        # start_time and stop_time > SUB stop_time
        self.set_6_reservation_start_time_and_stop_time_gt_sub_stop_time(reservation_one, self.scheduling_unit_blueprint_cs002)
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint_cs002))

    def test_two_stations_reserved(self):
        """
        Test station reservation when 2 station (CS002,CS003) are reserved and station CS002 is used in scheduling_unit
        with different reservation start and stop times
        """
        reservation_two = self.create_station_reservation("Two", ["CS002", "CS003"])
        # reservation start_time > SUB start_time and reservation stop_time > SUB stop_time
        self.set_1_reservation_start_time_gt_sub_start_time_and_stop_time_gt_sub_stop_time(reservation_two, self.scheduling_unit_blueprint_cs002_cs003)
        self.assertFalse(can_run_within_station_reservations(self.scheduling_unit_blueprint_cs002_cs003))

        # reservation start_time < SUB start_time and stop_time < SUB stop_time
        self.set_2_reservation_start_time_lt_sub_start_time_and_stop_time_lt_sub_stop_time(reservation_two, self.scheduling_unit_blueprint_cs002_cs003)
        self.assertFalse(can_run_within_station_reservations(self.scheduling_unit_blueprint_cs002_cs003))

        # reservation start_time > SUB start_time and stop_time < SUB stop_time
        self.set_3_reservation_start_time_gt_sub_start_time_and_stop_time_lt_sub_stop_time(reservation_two, self.scheduling_unit_blueprint_cs002_cs003)
        self.assertFalse(can_run_within_station_reservations(self.scheduling_unit_blueprint_cs002_cs003))

        # reservation start_time < SUB start_time and stop_time > SUB stop_time
        self.set_4_reservation_start_time_lt_sub_start_time_and_stop_time_gt_sub_stop_time(reservation_two, self.scheduling_unit_blueprint_cs002_cs003)
        self.assertFalse(can_run_within_station_reservations(self.scheduling_unit_blueprint_cs002_cs003))

        # Reservations outside boundary
        # start_time and stop_time < SUB start_time
        self.set_5_reservation_start_time_and_stop_time_lt_sub_start_time(reservation_two, self.scheduling_unit_blueprint_cs002_cs003)
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint_cs002_cs003))

        # start_time and stop_time > SUB stop_time
        self.set_6_reservation_start_time_and_stop_time_gt_sub_stop_time(reservation_two, self.scheduling_unit_blueprint_cs002_cs003)
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint_cs002_cs003))

    def test_two_stations_reserved_but_not_used(self):
        """
        Test station reservation when 2 stations (CS003, CS004) are reserved and station CS002 is used in scheduling_unit
        with different reservation start and stop times
        All possibilities should result in 'can run'
        """
        reservation_two_no_overlap = self.create_station_reservation("Two-NoOverlap", ["CS003", "CS004"])
        # reservation start_time > SUB start_time and reservation stop_time > SUB stop_time
        self.set_1_reservation_start_time_gt_sub_start_time_and_stop_time_gt_sub_stop_time(reservation_two_no_overlap, self.scheduling_unit_blueprint_cs002)
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint_cs002))

        # reservation start_time < SUB start_time and stop_time < SUB stop_time
        self.set_2_reservation_start_time_lt_sub_start_time_and_stop_time_lt_sub_stop_time(reservation_two_no_overlap, self.scheduling_unit_blueprint_cs002)
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint_cs002))

        # reservation start_time > SUB start_time and stop_time < SUB stop_time
        self.set_3_reservation_start_time_gt_sub_start_time_and_stop_time_lt_sub_stop_time(reservation_two_no_overlap, self.scheduling_unit_blueprint_cs002)
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint_cs002))

        # reservation start_time < SUB start_time and stop_time > SUB stop_time
        self.set_4_reservation_start_time_lt_sub_start_time_and_stop_time_gt_sub_stop_time(reservation_two_no_overlap, self.scheduling_unit_blueprint_cs002)
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint_cs002))

        # Reservations outside boundary
        # start_time and stop_time < SUB start_time
        self.set_5_reservation_start_time_and_stop_time_lt_sub_start_time(reservation_two_no_overlap, self.scheduling_unit_blueprint_cs002)
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint_cs002))

        # start_time and stop_time > SUB stop_time
        self.set_6_reservation_start_time_and_stop_time_gt_sub_stop_time(reservation_two_no_overlap, self.scheduling_unit_blueprint_cs002)
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint_cs002))

    def test_two_stations_reserved_with_duration_null(self):
        """
        Test station reservation when two stations (CS002,CS003) are reserved with duration null and so reserved indefinitely
        and station CS002 is used in scheduling_unit
        Test with different reservation start time and NO stop_time
        start_time after SUB stop_time 'can run' all others 'can NOT run'
        """
        reservation_two_no_duration = self.create_station_reservation("Two-NoDuration", ["CS002", "CS003"], stop_time=None)
        # reservation start_time > SUB start_time and < SUB stop_time
        reservation_two_no_duration.start_time = self.scheduling_unit_blueprint_cs002_cs003.scheduled_start_time + timedelta(seconds=1)
        reservation_two_no_duration.save()
        self.assertFalse(can_run_within_station_reservations(self.scheduling_unit_blueprint_cs002_cs003))

        # reservation start_time < SUB start_time (and < SUB stop_time of course)
        reservation_two_no_duration.start_time = self.scheduling_unit_blueprint_cs002_cs003.scheduled_start_time - timedelta(seconds=1)
        reservation_two_no_duration.save()
        self.assertFalse(can_run_within_station_reservations(self.scheduling_unit_blueprint_cs002_cs003))

        # reservation start_time > SUB stop time
        reservation_two_no_duration.start_time = self.scheduling_unit_blueprint_cs002_cs003.scheduled_observation_stop_time + timedelta(seconds=1)
        reservation_two_no_duration.save()
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint_cs002_cs003))

    def test_dutch_stations_conflicts_result_false(self):
        """
        Test conflict of 'Dutch' station which have a default of max_nr_missing=4,
        Create stations reservation equal to max_nr_missing+1 and check that it can not run
        """
        # Create a reservation within scheduling_unit
        self.create_station_reservation("Dutch", ['CS002', 'CS003', 'CS004', 'CS401', 'CS501'],
                                        start_time=self.scheduling_unit_blueprint.scheduled_start_time)
        self.assertFalse(can_run_within_station_reservations(self.scheduling_unit_blueprint))

    def test_dutch_stations_conflicts_result_true(self):
        """
        Test conflict of 'Dutch' station which have a default of max_nr_missing=4,
        Create stations reservation equal to max_nr_missing and check that it can run
        """
        # Create a reservation within scheduling_unit
        self.create_station_reservation("Dutch", ['CS002', 'CS003', 'CS401'],
                                        start_time=self.scheduling_unit_blueprint.scheduled_start_time)
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint))

    def test_international_stations_conflicts_result_false(self):
        """
        Test conflict of 'International' stations which have a default of max_nr_missing=2,
        Create stations reservation equal to max_nr_missing+1 and check that it can not run
        """
        # Create a reservation within scheduling_unit
        self.create_station_reservation("International", ['SE607', 'PL610', 'PL612'],
                                        start_time=self.scheduling_unit_blueprint.scheduled_start_time)
        self.assertFalse(can_run_within_station_reservations(self.scheduling_unit_blueprint))

    def test_international_stations_conflicts_result_true(self):
        """
        Test conflict of 'International' stations which are have a default of max_nr_missing=2,
        Create stations reservation equal to max_nr_missing and check that it can run
        """
        # Create a reservation within scheduling_unit
        self.create_station_reservation("International", ['SE607', 'PL610'],
                                        start_time=self.scheduling_unit_blueprint.scheduled_start_time)
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint))

    def test_international_required_stations_conflicts_result_false(self):
        """
        Test conflict of 'International Required' stations which have a default of max_nr_missing=1,
        Create stations reservation equal to max_nr_missing+1 and check that it can not run
        """
        # Create a reservation within scheduling_unit
        self.create_station_reservation("International Required", ['DE601', 'DE605'],
                                        start_time=self.scheduling_unit_blueprint.scheduled_start_time,
                                        stop_time=self.scheduling_unit_blueprint.scheduled_observation_stop_time)
        self.assertFalse(can_run_within_station_reservations(self.scheduling_unit_blueprint))

    def test_international_required_stations_conflicts_result_true(self):
        """
        Test conflict of 'International Required' stations which are have a default of max_nr_missing=1,
        Create stations reservation equal to max_nr_missing and check that it can run
        """
        # Create a reservation within scheduling_unit
        self.create_station_reservation("International Required", ['DE605'],
                                        start_time=self.scheduling_unit_blueprint.scheduled_start_time)
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint))

    def test_mixed_required_stations_conflicts_result_false(self):
        """
        Test conflict of 'mixed' stations which are have a default of max_nr_missing,
        Create stations reservation equal to max_nr_missing and one station group max_nr_missing+1
        and check that it can not run
        """
        # Create a reservation within scheduling_unit
        self.create_station_reservation("Mixed", ['DE605', 'SE607', 'PL610', 'CS002', 'CS003', 'CS401'],
                                        start_time=self.scheduling_unit_blueprint.scheduled_start_time)
        self.assertFalse(can_run_within_station_reservations(self.scheduling_unit_blueprint))

    def test_mixed_required_stations_conflicts_result_true(self):
        """
        Test conflict of 'mixed' stations which are have a default of max_nr_missing,
        Create stations reservation equal to max_nr_missing and check that it can run
        """
        # Create a reservation within scheduling_unit
        self.create_station_reservation("Mixed", ['DE605', 'PL610', 'CS002', 'CS003', 'CS401'],
                                        start_time=self.scheduling_unit_blueprint.scheduled_start_time)
        self.assertTrue(can_run_within_station_reservations(self.scheduling_unit_blueprint))

    def test_unschedulable_reason_when_reserved_stations_block_unit(self):
        """
        Test station reservation when 2 station (CS002,CS003) are reserved and stations (CS002, CS003) are used in scheduling_unit
        """
        reservation_two = self.create_station_reservation("Two", ["CS002", "CS003"])
        # reservation start_time > SUB start_time and reservation stop_time > SUB stop_time
        self.set_1_reservation_start_time_gt_sub_start_time_and_stop_time_gt_sub_stop_time(reservation_two, self.scheduling_unit_blueprint_cs002_cs003)

        # set the constraints for scheduling_unit_blueprint_cs002_cs003
        self.scheduling_unit_blueprint_cs002_cs003.scheduling_constraints_doc['scheduler'] = 'fixed_time'
        self.scheduling_unit_blueprint_cs002_cs003.scheduling_constraints_doc['time'] = { "at": reservation_two.start_time.isoformat() }
        # have very loose sky constraints, so this test is not time-dependent, and can run at any reservation_two.start_time
        self.scheduling_unit_blueprint_cs002_cs003.scheduling_constraints_doc['sky'] = {'transit_offset': {'from': -43200,
                                                                                                           'to': 43200 },
                                                                                        'min_distance': {'jupiter': 0,
                                                                                                         'sun': 0,
                                                                                                         'moon': 0},
                                                                                        'min_elevation': { 'target': 0,
                                                                                                           'calibrator': 0}}
        self.scheduling_unit_blueprint_cs002_cs003.save()

        # mark the other units as fixed time, so they won't interfere
        self.scheduling_unit_blueprint.scheduling_constraints_doc = {'scheduler': 'fixed_time', 'time': {'at': (reservation_two.stop_time + timedelta(days=1)).isoformat()}}
        self.scheduling_unit_blueprint.save()
        self.scheduling_unit_blueprint_cs002.scheduling_constraints_doc = {'scheduler': 'fixed_time', 'time': {'at': (reservation_two.stop_time + timedelta(days=2)).isoformat()}}
        self.scheduling_unit_blueprint_cs002.save()

        # try to schedule, should fail for scheduling_unit_blueprint_cs002_cs003.
        wipe_evaluate_constraints_caches()
        scheduled_scheduling_units = self.scheduler.schedule_fixed_time_scheduling_units()

        # Assert the scheduling_unit has not been scheduled and that it has the correct expected unschedulable_reason
        self.scheduling_unit_blueprint_cs002_cs003.refresh_from_db()
        self.assertEqual('unschedulable', self.scheduling_unit_blueprint_cs002_cs003.status.value)
        self.assertEqual("Stations CS002,CS003 are reserved at start_time='%s'"%(reservation_two.start_time,),
                         self.scheduling_unit_blueprint_cs002_cs003.unschedulable_reason)


class TestTriggers(BaseDynamicSchedulingTestCase):
    """
    Tests for scheduling behavior of triggered observations
    """
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()

        # create a scheduling set in a project that allows triggers
        cls.scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data())
        cls.scheduling_set.project.can_trigger = True
        cls.scheduling_set.project.trigger_priority = 1
        cls.scheduling_set.project.save()

        # create some trigger quota for this project
        models.ProjectQuota.objects.create(project=cls.scheduling_set.project, value=1000, resource_type=models.ResourceType.objects.get(quantity__value=models.Quantity.Choices.NUMBER.value, name__icontains='trigger'))

        # create a second scheduling set in a project that allows triggers and has higher trigger_priority.
        # Also increase rank, to rule out that things get priority based on that.
        cls.scheduling_set_high_trigger_priority = models.SchedulingSet.objects.create(**SchedulingSet_test_data())
        cls.scheduling_set_high_trigger_priority.project.can_trigger = True
        cls.scheduling_set_high_trigger_priority.project.trigger_priority = cls.scheduling_set.project.trigger_priority + 1
        cls.scheduling_set_high_trigger_priority.project.save()

        # create some trigger quota for this project as well
        models.ProjectQuota.objects.create(project=cls.scheduling_set_high_trigger_priority.project, value=1000, resource_type=models.ResourceType.objects.get(quantity__value=models.Quantity.Choices.NUMBER.value, name__icontains='trigger'))

        # create a second scheduling set in a project that allows triggers and has higher trigger_priority.
        # Also increase rank, to rule out that things get priority based on that.
        cls.scheduling_set_high_trigger_priority_worse_rank = models.SchedulingSet.objects.create(**SchedulingSet_test_data())
        cls.scheduling_set_high_trigger_priority_worse_rank.project.can_trigger = True
        cls.scheduling_set_high_trigger_priority_worse_rank.project.trigger_priority = cls.scheduling_set.project.trigger_priority + 1
        cls.scheduling_set_high_trigger_priority_worse_rank.project.rank = cls.scheduling_set.project.rank + 1
        cls.scheduling_set_high_trigger_priority_worse_rank.project.save()

        # create some trigger quota for this project as well
        models.ProjectQuota.objects.create(project=cls.scheduling_set_high_trigger_priority_worse_rank.project, value=1000, resource_type=models.ResourceType.objects.get(quantity__value=models.Quantity.Choices.NUMBER.value, name__icontains='trigger'))


    def test_simple_triggered_scheduling_unit_gets_scheduled(self):

        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "triggered scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set,
                                    interrupts_telescope=True)
        triggered_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()
        self.assertEqual(1, len(scheduled_scheduling_units))
        scheduled_scheduling_unit = scheduled_scheduling_units[0]

        # Assert the scheduling_unit has been scheduled
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduled_scheduling_unit.id, triggered_scheduling_unit_blueprint.id)
        self.assertEqual(scheduled_scheduling_unit.status.value, 'scheduled')

    def test_triggered_scheduling_unit_with_at_constraint_gets_scheduled_at_correct_time(self):

        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    'scheduling_unit for at %s' % self._testMethodName,
                                    interrupts_telescope=True,
                                    scheduling_set=self.scheduling_set)
        # Clear constraints
        scheduling_unit_draft.scheduling_constraints_doc = { 'sky': {}, 'daily': {}, 'scheduler': 'dynamic',
                                                             'time': {"between": [], "not_between": [] }}

        # Set at constraint
        at = round_to_second_precision(datetime.utcnow() + timedelta(minutes=17))
        scheduling_unit_draft.scheduling_constraints_doc['time']['at'] = at.isoformat()
        scheduling_unit_draft.save()
        triggered_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()
        self.assertEqual(1, len(scheduled_scheduling_units))
        scheduled_scheduling_unit = scheduled_scheduling_units[0]

        # Assert the scheduling_unit has been scheduled and assert is has been scheduled at "at" timestamp
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduled_scheduling_unit.id, triggered_scheduling_unit_blueprint.id)
        self.assertEqual(scheduled_scheduling_unit.status.value, 'scheduled')
        self.assertEqual(scheduled_scheduling_unit.scheduled_start_time, at)

    def test_triggered_scheduling_unit_has_priority_over_regular_observation(self):

        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set,
                                    interrupts_telescope=False)
        regular_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "triggered scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set,
                                    interrupts_telescope=True)
        triggered_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()
        scheduled_scheduling_unit = scheduled_scheduling_units[0]

        # Assert the triggered scheduling_unit has been scheduled
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduled_scheduling_unit.id, triggered_scheduling_unit_blueprint.id)

        regular_scheduling_unit_blueprint.refresh_from_db()
        triggered_scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(regular_scheduling_unit_blueprint.status.value, 'scheduled')
        self.assertEqual(triggered_scheduling_unit_blueprint.status.value, 'scheduled')

    def test_trigger_get_scheduled_close_to_submission_time(self):
        # we expect triggers to run as soon as possible

        # Set weight factor to prevent that dynamically scheduled units get scheduled in the future.
        # Note: If TMSS is configured with a non-zero weight factor, triggers require strict time constraints to make
        # them get scheduled close to submission time!
        self._set_density_vs_optimal_weight_factor(0)

        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "triggered scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set,
                                    interrupts_telescope=True)
        triggered_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()
        scheduled_scheduling_unit = scheduled_scheduling_units[0]

        # Assert the triggered scheduling_unit has been scheduled
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduled_scheduling_unit.id, triggered_scheduling_unit_blueprint.id)

        # Assert the trigger gets scheduled close to submission time
        self.assertLess(scheduled_scheduling_unit.scheduled_start_time, datetime.utcnow() + timedelta(minutes=5))

    def test_trigger_get_scheduled_according_to_before_constraint(self):
        # we expect triggers to start within the allowed time window.

        # Set weight factor to prevent that dynamically scheduled unit gets scheduled to start right away by default.
        # Triggers with strict time constraints should not start within the permitted time window nonetheless.
        self._set_density_vs_optimal_weight_factor(0.5)

        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "triggered scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set,
                                    interrupts_telescope=True)
        scheduling_unit_draft.scheduling_constraints_doc = {'scheduler': 'dynamic',
                                                            'time': {'before': (datetime.utcnow() + timedelta(minutes=5)).isoformat()}}
        scheduling_unit_draft.save()
        triggered_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()
        scheduled_scheduling_unit = scheduled_scheduling_units[0]

        # Assert the triggered scheduling_unit has been scheduled
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduled_scheduling_unit.id, triggered_scheduling_unit_blueprint.id)

        # Assert the trigger gets scheduled before the requested time
        self.assertLess(scheduled_scheduling_unit.scheduled_start_time, datetime.utcnow() + timedelta(minutes=5))

    def test_triggered_scheduling_unit_has_priority_over_regular_observation_fixed_time(self):
        # RT-PH-T1 / RT-PH-T2 (test code, see: https://support.astron.nl/confluence/display/TMSS/DOC-01-a+Test+Plan+-+Verification+tests)
        # a fixed time trigger that overlaps with fixed time regular observations should have priority and block them

        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set,
                                    obs_duration=600,
                                    interrupts_telescope=False)
        scheduling_unit_draft.scheduling_constraints_doc = {'scheduler': 'fixed_time',
                                                            'time': {'at': (datetime.utcnow() + timedelta(minutes=5)).isoformat()}}
        scheduling_unit_draft.save()
        regular_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)


        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set,
                                    obs_duration=600,
                                    interrupts_telescope=False)
        scheduling_unit_draft.scheduling_constraints_doc = {'scheduler': 'fixed_time',
                                                            'time': {'at': (datetime.utcnow() + timedelta(minutes=16)).isoformat()}}
        scheduling_unit_draft.save()
        regular_scheduling_unit_blueprint2 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        # triggered obs that overlaps with both regular obs
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "triggered scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set,
                                    obs_duration=600,
                                    interrupts_telescope=True)
        scheduling_unit_draft.scheduling_constraints_doc = {'scheduler': 'fixed_time',
                                                            'time': {'at': (datetime.utcnow() + timedelta(minutes=10)).isoformat()}}
        scheduling_unit_draft.save()
        triggered_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduled_scheduling_units = self.scheduler.schedule_fixed_time_scheduling_units()
        self.assertEqual(1, len(scheduled_scheduling_units))
        scheduled_scheduling_unit = scheduled_scheduling_units[0]

        # Assert the triggered scheduling_unit has been scheduled, overlapping obs are unschedulable
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduled_scheduling_unit.id, triggered_scheduling_unit_blueprint.id)

        regular_scheduling_unit_blueprint.refresh_from_db()
        regular_scheduling_unit_blueprint2.refresh_from_db()
        triggered_scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(regular_scheduling_unit_blueprint.status.value, 'unschedulable')
        self.assertEqual(regular_scheduling_unit_blueprint2.status.value, 'unschedulable')
        self.assertEqual(triggered_scheduling_unit_blueprint.status.value, 'scheduled')

    def test_triggered_scheduling_unit_lower_priority_than_regular_observation_fixed_time(self):
        # RT-PL-T1 / RT-PL-T2(test code, see: https://support.astron.nl/confluence/display/TMSS/DOC-01-a+Test+Plan+-+Verification+tests)
        # a fixed time trigger should have no precedence over overlapping regular fixed time observations with higher
        # trigger priority and hence be blocked by them.

        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set_high_trigger_priority_worse_rank,
                                    obs_duration=600,
                                    interrupts_telescope=False)
        scheduling_unit_draft.scheduling_constraints_doc = {'scheduler': 'fixed_time',
                                                            'time': {'at': (datetime.utcnow() + timedelta(minutes=5)).isoformat()}}
        scheduling_unit_draft.save()
        regular_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)


        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set_high_trigger_priority_worse_rank,
                                    obs_duration=600,
                                    interrupts_telescope=False)
        scheduling_unit_draft.scheduling_constraints_doc = {'scheduler': 'fixed_time',
                                                            'time': {'at': (datetime.utcnow() + timedelta(minutes=16)).isoformat()}}
        scheduling_unit_draft.save()
        regular_scheduling_unit_blueprint2 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        # triggered obs that overlaps with both regular obs
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "triggered scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set,
                                    obs_duration=600,
                                    interrupts_telescope=True)
        scheduling_unit_draft.scheduling_constraints_doc = {'scheduler': 'fixed_time',
                                                            'time': {'at': (datetime.utcnow() + timedelta(minutes=10)).isoformat()}}
        scheduling_unit_draft.save()
        triggered_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduled_scheduling_units = self.scheduler.schedule_fixed_time_scheduling_units()
        self.assertEqual(2, len(scheduled_scheduling_units))

        # Assert the triggered scheduling_unit is unschedulable scheduled and the regular obs remain scheduled
        self.assertNotIn(triggered_scheduling_unit_blueprint, scheduled_scheduling_units)

        regular_scheduling_unit_blueprint.refresh_from_db()
        regular_scheduling_unit_blueprint2.refresh_from_db()
        triggered_scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(regular_scheduling_unit_blueprint.status.value, 'scheduled')
        self.assertEqual(regular_scheduling_unit_blueprint2.status.value, 'scheduled')
        self.assertEqual(triggered_scheduling_unit_blueprint.status.value, 'unschedulable')

    def test_triggered_dynamically_scheduled_observation_has_priority_over_regular_fixed_time_observation(self):
        # RT-PH-T1(test code, see: https://support.astron.nl/confluence/display/TMSS/DOC-01-a+Test+Plan+-+Verification+tests)
        # a trigger that gets dynamically scheduled should have priority over a regular fixed time observation

        # Set weight factor to prevent that the dynamically scheduled trigger get scheduled in the future and the units
        # conflict with each other.
        self._set_density_vs_optimal_weight_factor(0)

        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set,
                                    obs_duration=600,
                                    interrupts_telescope=False)
        scheduling_unit_draft.scheduling_constraints_doc = {'scheduler': 'fixed_time',
                                                            'time': {'at': (datetime.utcnow() + timedelta(minutes=5)).isoformat()}}
        scheduling_unit_draft.save()
        regular_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "triggered scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set_high_trigger_priority,
                                    obs_duration=600,
                                    interrupts_telescope=True)
        triggered_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        logger.warning(triggered_scheduling_unit_blueprint.scheduling_constraints_doc)
        logger.warning('###>1')
        scheduled_scheduling_units = self.scheduler.do_full_schedule_computation()
        logger.warning('###>2')
        # self.assertEqual(1, len(scheduled_scheduling_units))  # todo: currently returns also the schedulable unit
        for unit in scheduled_scheduling_units:
            unit.refresh_from_db()
            if unit.status.value == models.SchedulingUnitStatus.Choices.SCHEDULED.value:
                scheduled_scheduling_unit = unit

        # Assert the triggered scheduling_unit has been scheduled
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduled_scheduling_unit.id, triggered_scheduling_unit_blueprint.id)

        regular_scheduling_unit_blueprint.refresh_from_db()
        triggered_scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(regular_scheduling_unit_blueprint.status.value, 'unschedulable')
        self.assertEqual(triggered_scheduling_unit_blueprint.status.value, 'scheduled')

    def test_triggered_fixed_time_observation_has_priority_over_regular_dynamically_scheduled_observation(self):
        # a trigger with fixed time constraint should have precedence over a dynamically scheduled unit.
        # the dynamically scheduled unit should be re-scheduled at a later point in time.
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set,
                                    obs_duration=600,
                                    interrupts_telescope=False)
        scheduling_unit_draft.scheduling_constraints_doc['time'] = {'before': (datetime.utcnow() + timedelta(minutes=10)).isoformat()}
        scheduling_unit_draft.save()
        regular_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "triggered scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set,
                                    obs_duration=600,
                                    interrupts_telescope=True)
        scheduling_unit_draft.scheduling_constraints_doc = {'scheduler': 'fixed_time',
                                                            'time': {'at': (datetime.utcnow() + timedelta(minutes=5)).isoformat()}}
        scheduling_unit_draft.save()
        triggered_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduled_scheduling_units = self.scheduler.do_full_schedule_computation()
        self.assertEqual(1, len(scheduled_scheduling_units))
        scheduled_scheduling_unit = scheduled_scheduling_units[0]

        # Assert the triggered scheduling_unit has been scheduled
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduled_scheduling_unit.id, triggered_scheduling_unit_blueprint.id)

        regular_scheduling_unit_blueprint.refresh_from_db()
        triggered_scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(regular_scheduling_unit_blueprint.status.value, 'unschedulable')
        self.assertEqual(triggered_scheduling_unit_blueprint.status.value, 'scheduled')

    def test_triggered_scheduling_unit_unschedules_regular_observation(self):
        # an already scheduled observation should go to unscheduled state when a trigger has precedence over it.

        # create a regular scheduling_unit
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "regular scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set,
                                    interrupts_telescope=False)
        regular_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()
        self.assertEqual(1, len(scheduled_scheduling_units))
        scheduled_scheduling_unit = scheduled_scheduling_units[0]

        # Assert the scheduling_unit has been scheduled
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduled_scheduling_unit.id, regular_scheduling_unit_blueprint.id)

        regular_scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(regular_scheduling_unit_blueprint.status.value, 'scheduled')

        # add a triggered scheduling_unit
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "triggered scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set,
                                    interrupts_telescope=True)
        triggered_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()
        self.assertEqual(2, len(scheduled_scheduling_units))
        # Assert now the new triggered scheduling_unit has been scheduled, and the regular one has been rescheduled
        regular_scheduling_unit_blueprint.refresh_from_db()
        triggered_scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(regular_scheduling_unit_blueprint.status.value, 'scheduled')
        self.assertEqual(triggered_scheduling_unit_blueprint.status.value, 'scheduled')
        self.assertTrue(triggered_scheduling_unit_blueprint.scheduled_start_time < regular_scheduling_unit_blueprint.scheduled_start_time)

    @unittest.skip('the behavior under test is debatable')
    def test_triggered_fixed_time_scheduling_unit_unschedules_dynamically_scheduled_regular_observations(self):
        # a fixed-time trigger unschedules a dynamically scheduled regular obs even if it has lower trigger priority
        # todo: Note that this behavior was requested in TMSS-2501, but feels inconsistent. To be discussed further.
        #   Currently, the trigger does not unschedule the regular obs, but is marked unschedulable, which is exactly
        #   what I would expect to happen.

        # create a regular scheduling_unit
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set_high_trigger_priority,
                                    obs_duration=600,
                                    interrupts_telescope=False)
        regular_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduled_scheduling_units = self.scheduler.do_full_schedule_computation()

        self.assertEqual(1, len(scheduled_scheduling_units))
        scheduled_scheduling_unit = scheduled_scheduling_units[0]

        # Assert the scheduling_unit has been scheduled
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduled_scheduling_unit.id, regular_scheduling_unit_blueprint.id)

        regular_scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(regular_scheduling_unit_blueprint.status.value, 'scheduled')
        original_start_time = regular_scheduling_unit_blueprint.scheduled_start_time

        # add an overlapping triggered scheduling_unit
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "triggered scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set,
                                    obs_duration=600,
                                    interrupts_telescope=True)
        scheduling_unit_draft.scheduling_constraints_doc = {'scheduler': 'fixed_time', 'time': {'at': (original_start_time + timedelta(minutes=5)).isoformat()}}
        scheduling_unit_draft.save()
        triggered_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduled_scheduling_units = self.scheduler.do_full_schedule_computation()

        # assert the trigger is scheduled and the regular unit is still schedulable
        # self.assertEqual(1, len(scheduled_scheduling_units))  # todo: currently returns also the schedulable unit

        regular_scheduling_unit_blueprint.refresh_from_db()
        triggered_scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(regular_scheduling_unit_blueprint.status.value, 'schedulable')
        self.assertEqual(triggered_scheduling_unit_blueprint.status.value, 'scheduled')

    @mock.patch("lofar.sas.tmss.tmss.tmssapp.subtasks.kill_observation_subtask")
    def test_triggered_scheduling_unit_cancels_regular_observation(self, kill_observation_subtask_mock):
        # an observation that is already happening should be interrupted when a trigger comes in so that the trigger
        # can run right away.

        # create a regular scheduling_unit
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set,
                                    interrupts_telescope=False)
        regular_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
        regular_scheduling_unit_blueprint.save()

        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()
        self.assertEqual(1, len(scheduled_scheduling_units))
        scheduled_scheduling_unit = scheduled_scheduling_units[0]

        # Assert the scheduling_unit has been scheduled
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduled_scheduling_unit.id, regular_scheduling_unit_blueprint.id)

        regular_scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(regular_scheduling_unit_blueprint.status.value, 'scheduled')

        # put obs to started state
        subtask = scheduled_scheduling_unit.task_blueprints.first().subtasks.first()
        subtask.state = models.SubtaskState.objects.get(value='starting')
        subtask.save()
        subtask.state = models.SubtaskState.objects.get(value='started')
        subtask.save()

        # assert obs it detected as running
        running_subtasks = get_running_observation_subtasks()
        self.assertIn(subtask, running_subtasks)

        # also assert cut-off date is considered
        running_subtasks = get_running_observation_subtasks(subtask.scheduled_stop_time + timedelta(minutes=5))
        self.assertNotIn(subtask, running_subtasks)

        # add a triggered scheduling_unit with higher priority
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "triggered scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set_high_trigger_priority_worse_rank,
                                    interrupts_telescope=True)
        triggered_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()
        self.assertEqual(1, len(scheduled_scheduling_units))
        scheduled_scheduling_unit = scheduled_scheduling_units[0]

        # assert the subtask has been cancelled and killed
        kill_observation_subtask_mock.assert_called_with(subtask)

        # Assert now the new triggered scheduling_unit has been scheduled
        # todo: TMSS-704: We should only cancel if the trigger cannot run afterwards due to constraints.
        #  Add such constraints once the scheduler considers that, since that will break this test.
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduled_scheduling_unit.id, triggered_scheduling_unit_blueprint.id)

        triggered_scheduling_unit_blueprint.refresh_from_db()
        regular_scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(triggered_scheduling_unit_blueprint.status.value, 'scheduled')
        self.assertEqual(regular_scheduling_unit_blueprint.status.value, 'cancelled')

    @mock.patch("lofar.sas.tmss.tmss.tmssapp.subtasks.kill_observation_subtask")
    def test_triggered_scheduling_unit_cancels_regular_observation_fixed_time(self, kill_observation_subtask_mock):
        # RT-PH-T3(test code, see: https://support.astron.nl/confluence/display/TMSS/DOC-01-a+Test+Plan+-+Verification+tests)
        # an observation that is already happening should be interrupted when a trigger comes in so that the trigger
        # can run right away, also with fixed time scheduling.

        # create a regular scheduling_unit
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set,
                                    interrupts_telescope=False)
        regular_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
        regular_scheduling_unit_blueprint.scheduling_constraints_doc = {'scheduler': 'fixed_time',
                                                                        'time': {'at': (datetime.utcnow() + timedelta(minutes=10)).isoformat()}}
        regular_scheduling_unit_blueprint.save()

        scheduled_scheduling_units = self.scheduler.schedule_fixed_time_scheduling_units()
        self.assertEqual(1, len(scheduled_scheduling_units))
        scheduled_scheduling_unit = scheduled_scheduling_units[0]

        # Assert the scheduling_unit has been scheduled
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduled_scheduling_unit.id, regular_scheduling_unit_blueprint.id)

        regular_scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(regular_scheduling_unit_blueprint.status.value, 'scheduled')

        # put obs to started state
        subtask = scheduled_scheduling_unit.task_blueprints.first().subtasks.first()
        subtask.state = models.SubtaskState.objects.get(value='starting')
        subtask.save()
        subtask.state = models.SubtaskState.objects.get(value='started')
        subtask.save()

        # assert obs it detected as running
        running_subtasks = get_running_observation_subtasks()
        self.assertIn(subtask, running_subtasks)

        # add a triggered scheduling_unit with higher priority
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "triggered scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set_high_trigger_priority_worse_rank,
                                    interrupts_telescope=True,
                                    obs_duration=600)
        triggered_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
        triggered_scheduling_unit_blueprint.scheduling_constraints_doc = {'scheduler': 'fixed_time',
                                                                          'time': {'at': (datetime.utcnow() + timedelta(minutes=5)).isoformat()}}
        triggered_scheduling_unit_blueprint.save()

        scheduled_scheduling_units = self.scheduler.schedule_fixed_time_scheduling_units()
        self.assertEqual(1, len(scheduled_scheduling_units))
        scheduled_scheduling_unit = scheduled_scheduling_units[0]

        # assert the subtask has been cancelled and killed
        kill_observation_subtask_mock.assert_called_with(subtask)

        # Assert now the new triggered scheduling_unit has been scheduled, the regular observation becomes unschedulable because the at-constraint is now blocked
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduled_scheduling_unit.id, triggered_scheduling_unit_blueprint.id)

        triggered_scheduling_unit_blueprint.refresh_from_db()
        regular_scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(triggered_scheduling_unit_blueprint.status.value, 'scheduled')
        self.assertEqual(regular_scheduling_unit_blueprint.status.value, 'cancelled')

    @mock.patch("lofar.sas.tmss.tmss.tmssapp.subtasks.kill_observation_subtask")
    def test_triggered_dynamic_scheduling_unit_cancels_regular_fixed_time_observation(self, kill_observation_subtask_mock):
        # RT-PH-T3(test code, see: https://support.astron.nl/confluence/display/TMSS/DOC-01-a+Test+Plan+-+Verification+tests)
        # an observation that is already happening should be interrupted when a trigger comes in so that the trigger
        # can run right away, also when the trigger is dynamically scheduled and the regular obs is fixed time.

        # Set weight factor to prevent that the dynamically scheduled trigger get scheduled in the future and the units
        # conflict with each other.
        self._set_density_vs_optimal_weight_factor(0)

        # create a regular scheduling_unit
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set,
                                    interrupts_telescope=False)
        regular_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
        regular_scheduling_unit_blueprint.scheduling_constraints_doc = {'scheduler': 'fixed_time',
                                                                        'time': {'at': (datetime.utcnow() + timedelta(minutes=5)).isoformat()}}
        regular_scheduling_unit_blueprint.save()

        scheduled_scheduling_units = self.scheduler.schedule_fixed_time_scheduling_units()
        self.assertEqual(1, len(scheduled_scheduling_units))
        scheduled_scheduling_unit = scheduled_scheduling_units[0]

        # Assert the scheduling_unit has been scheduled
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduled_scheduling_unit.id, regular_scheduling_unit_blueprint.id)

        regular_scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(regular_scheduling_unit_blueprint.status.value, 'scheduled')

        # put obs to started state
        subtask = scheduled_scheduling_unit.task_blueprints.first().subtasks.first()
        subtask.state = models.SubtaskState.objects.get(value='starting')
        subtask.save()
        subtask.state = models.SubtaskState.objects.get(value='started')
        subtask.save()

        # assert obs it detected as running
        running_subtasks = get_running_observation_subtasks()
        self.assertIn(subtask, running_subtasks)

        # add a triggered scheduling_unit with higher priority
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "triggered scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set_high_trigger_priority_worse_rank,
                                    interrupts_telescope=True,
                                    obs_duration=600)
        triggered_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduled_scheduling_units = self.scheduler.do_full_schedule_computation()
        self.assertEqual(1, len(scheduled_scheduling_units))
        scheduled_scheduling_unit = scheduled_scheduling_units[0]

        # assert the subtask has been cancelled and killed
        kill_observation_subtask_mock.assert_called_with(subtask)

        # Assert now the new triggered scheduling_unit has been scheduled
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduled_scheduling_unit.id, triggered_scheduling_unit_blueprint.id)

        triggered_scheduling_unit_blueprint.refresh_from_db()
        regular_scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(triggered_scheduling_unit_blueprint.status.value, 'scheduled')
        self.assertEqual(regular_scheduling_unit_blueprint.status.value, 'cancelled')

    @mock.patch("lofar.sas.tmss.tmss.tmssapp.subtasks.kill_observation_subtask")
    def test_triggered_fixed_time_scheduling_unit_cancels_regular_dynamically_scheduled_observation(self, kill_observation_subtask_mock):

        # create a regular scheduling_unit
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set,
                                    interrupts_telescope=False,
                                    obs_duration=600)
        regular_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()
        self.assertEqual(1, len(scheduled_scheduling_units))
        scheduled_scheduling_unit = scheduled_scheduling_units[0]

        # Assert the scheduling_unit has been scheduled
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduled_scheduling_unit.id, regular_scheduling_unit_blueprint.id)

        regular_scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(regular_scheduling_unit_blueprint.status.value, 'scheduled')

        # put obs to started state
        subtask = scheduled_scheduling_unit.task_blueprints.first().subtasks.first()
        subtask.state = models.SubtaskState.objects.get(value='starting')
        subtask.save()
        subtask.state = models.SubtaskState.objects.get(value='started')
        subtask.save()

        # assert obs it detected as running
        running_subtasks = get_running_observation_subtasks()
        self.assertIn(subtask, running_subtasks)

        # add a triggered scheduling_unit with higher priority
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "triggered scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set_high_trigger_priority_worse_rank,
                                    interrupts_telescope=True)
        triggered_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
        triggered_scheduling_unit_blueprint.scheduling_constraints_doc = {'scheduler': 'fixed_time',
                                                                          'time': {'at': (datetime.utcnow() + timedelta(minutes=5)).isoformat()}}
        triggered_scheduling_unit_blueprint.save()

        scheduled_scheduling_units = self.scheduler.schedule_fixed_time_scheduling_units()
        self.assertEqual(1, len(scheduled_scheduling_units))
        scheduled_scheduling_unit = scheduled_scheduling_units[0]

        # assert the subtask has been cancelled and killed
        kill_observation_subtask_mock.assert_called_with(subtask)

        # Assert now the new triggered scheduling_unit has been scheduled
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduled_scheduling_unit.id, triggered_scheduling_unit_blueprint.id)

        triggered_scheduling_unit_blueprint.refresh_from_db()
        regular_scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(triggered_scheduling_unit_blueprint.status.value, 'scheduled')
        self.assertEqual(regular_scheduling_unit_blueprint.status.value, 'cancelled')

    @mock.patch("lofar.sas.tmss.tmss.tmssapp.subtasks.cancel_subtask")
    def test_triggered_scheduling_unit_does_not_cancel_regular_observation_with_same_trigger_priority(self, cancel_mock):
        # RT-PL-T3(test code, see: https://support.astron.nl/confluence/display/TMSS/DOC-01-a+Test+Plan+-+Verification+tests)
        # a trigger should only cancel an observation when it has higher trigger prio, so let's make sure it does not when it doesn't have that.

        # create a regular scheduling_unit
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set,
                                    interrupts_telescope=False)
        regular_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()
        self.assertEqual(1, len(scheduled_scheduling_units))
        scheduled_scheduling_unit = scheduled_scheduling_units[0]

        # Assert the scheduling_unit has been scheduled
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduled_scheduling_unit.id, regular_scheduling_unit_blueprint.id)
        regular_scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(regular_scheduling_unit_blueprint.status.value, 'scheduled')

        # put obs to started state
        obs_subtask = scheduled_scheduling_unit.subtasks.get(state__value='scheduled', specifications_template__type__value='observation')
        set_subtask_state_following_allowed_transitions(obs_subtask, 'started')

        # assert obs is detected as running
        running_subtasks = get_running_observation_subtasks()
        self.assertIn(obs_subtask, running_subtasks)

        # also assert cut-off date is considered
        running_subtasks = get_running_observation_subtasks(obs_subtask.scheduled_stop_time + timedelta(minutes=5))
        self.assertNotIn(obs_subtask, running_subtasks)

        # add a triggered scheduling_unit with same trigger priority
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "triggered scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set,
                                    interrupts_telescope=True)
        triggered_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()

        # assert the triggered was scheduled
        self.assertEqual(1, len(scheduled_scheduling_units))

        # assert that the subtask has NOT been cancelled and is still in state 'started'
        cancel_mock.assert_not_called()
        obs_subtask.refresh_from_db()
        self.assertEqual(obs_subtask.state.value, 'started')

        # assert obs is  still detected as running
        running_subtasks = get_running_observation_subtasks()
        self.assertIn(obs_subtask, running_subtasks)

        # Assert that the new triggered scheduling_unit is scheduled later
        # and that the regular one is still observing
        triggered_scheduling_unit_blueprint.refresh_from_db()
        regular_scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(triggered_scheduling_unit_blueprint.status.value, 'scheduled')
        self.assertGreater(triggered_scheduling_unit_blueprint.scheduled_start_time, regular_scheduling_unit_blueprint.on_sky_stop_time)
        self.assertEqual(regular_scheduling_unit_blueprint.status.value, 'observing')


    @mock.patch("lofar.sas.tmss.tmss.tmssapp.subtasks.cancel_subtask")
    def test_triggered_scheduling_unit_does_not_cancel_regular_observation_if_it_cannot_run_anyway(self, cancel_mock):

        # set density_vs_optimal to 0, meaning "as tight as possible"
        weight_factor, created = models.SchedulingConstraintsWeightFactor.objects.get_or_create(constraint_name="density_vs_optimal")
        weight_factor.weight = 0
        weight_factor.save()

        # create a regular scheduling_unit
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set,
                                    interrupts_telescope=False)
        regular_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()
        self.assertEqual(1, len(scheduled_scheduling_units))
        scheduled_scheduling_unit = scheduled_scheduling_units[0]

        # Assert the scheduling_unit has been scheduled
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduled_scheduling_unit.id, regular_scheduling_unit_blueprint.id)
        regular_scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(regular_scheduling_unit_blueprint.status.value, 'scheduled')

        # put obs to started state
        subtask = scheduled_scheduling_unit.task_blueprints.first().subtasks.first()
        set_subtask_state_following_allowed_transitions(subtask, 'started')
        regular_scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(regular_scheduling_unit_blueprint.status.value, 'observing')

        # assert obs it detected as running
        running_subtasks = get_running_observation_subtasks()
        self.assertIn(subtask, running_subtasks)

        # also assert cut-off date is considered
        running_subtasks = get_running_observation_subtasks(subtask.scheduled_stop_time + timedelta(minutes=5))
        self.assertNotIn(subtask, running_subtasks)

        # add a triggered scheduling_unit with higher priority, but a between constraint that can never be met
        scheduling_unit_draft = self.create_simple_observation_scheduling_unit(
                                    "triggered scheduling unit for %s" % self._testMethodName,
                                    scheduling_set=self.scheduling_set_high_trigger_priority_worse_rank,
                                    interrupts_telescope=True)
        triggered_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)
        triggered_scheduling_unit_blueprint.scheduling_constraints_doc = {'scheduler': 'dynamic', 'time': {'between': [{"from": datetime.utcnow().isoformat(), "to": (datetime.utcnow()+timedelta(minutes=1)).isoformat()}] }}
        triggered_scheduling_unit_blueprint.save()

        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()

        # assert that the subtask has NOT been cancelled and is still in state 'started'
        cancel_mock.assert_not_called()
        self.assertEqual(subtask.state.value, 'started')

        # Assert that the new triggered scheduling_unit has NOT been scheduled, and the regular one is still observing
        self.assertEqual(0, len(scheduled_scheduling_units))
        regular_scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(regular_scheduling_unit_blueprint.status.value, 'observing')

        # triggered_scheduling_unit should be unschedulable, because we've given it impossible constraints.
        triggered_scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(triggered_scheduling_unit_blueprint.status.value, 'unschedulable')


    @mock.patch("lofar.sas.tmss.tmss.tmssapp.subtasks.cancel_subtask")
    def test_triggered_scheduling_unit_goes_to_unschedulable_if_it_cannot_cancel_and_does_not_fit(self, cancel_mock):

        # create three regular scheduling_units, two with high trigger priority, one with lower
        scheduling_unit_draft_high1 = self.create_simple_observation_scheduling_unit(
                                    "scheduling unit high1",
                                    scheduling_set=self.scheduling_set_high_trigger_priority,
                                    interrupts_telescope=False,
                                    unit_rank=0.0)
        regular_scheduling_unit_blueprint_high1 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft_high1)
        regular_scheduling_unit_blueprint_high1.scheduling_constraints_doc = {'scheduler': 'dynamic', 'time': {'between': [{"from": datetime.utcnow().isoformat(), "to": (datetime.utcnow()+timedelta(hours=24)).isoformat()},]}}
        regular_scheduling_unit_blueprint_high1.save()

        scheduling_unit_draft_high2 = self.create_simple_observation_scheduling_unit(
                                    "scheduling unit high2",
                                    scheduling_set=self.scheduling_set_high_trigger_priority,
                                    interrupts_telescope=False,
                                    unit_rank=0.1)
        regular_scheduling_unit_blueprint_high2 = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft_high2)
        regular_scheduling_unit_blueprint_high2.scheduling_constraints_doc = {'scheduler': 'dynamic', 'time': {'between': [{"from": datetime.utcnow().isoformat(), "to": (datetime.utcnow()+timedelta(hours=24)).isoformat()},]}}
        regular_scheduling_unit_blueprint_high2.save()

        scheduling_unit_draft_low = self.create_simple_observation_scheduling_unit(
                                    "scheduling unit low",
                                    scheduling_set=self.scheduling_set,
                                    interrupts_telescope=False,
                                    unit_rank=0.1)
        regular_scheduling_unit_blueprint_low = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft_low)
        regular_scheduling_unit_blueprint_low.scheduling_constraints_doc = {'scheduler': 'dynamic', 'time': {'between': [{"from": datetime.utcnow().isoformat(), "to": (datetime.utcnow()+timedelta(hours=24)).isoformat()},]}}
        regular_scheduling_unit_blueprint_low.save()

        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()
        scheduled_scheduling_unit = scheduled_scheduling_units[0]

        # Assert the regular_scheduling_unit_blueprint_high1 has been scheduled
        self.assertIsNotNone(scheduled_scheduling_unit)
        self.assertEqual(scheduled_scheduling_unit.id, regular_scheduling_unit_blueprint_high1.id)
        regular_scheduling_unit_blueprint_high1.refresh_from_db()
        self.assertEqual(regular_scheduling_unit_blueprint_high1.status.value, 'scheduled')

        # put first obs to started state
        subtask = scheduled_scheduling_unit.task_blueprints.first().subtasks.first()
        set_subtask_state_following_allowed_transitions(subtask, 'started')

        for su in scheduled_scheduling_units[1:]:
            unschedule_subtasks_in_scheduling_unit_blueprint(su)

        # add a triggered scheduling_unit with same trigger priority, and a between constraint that can only be met
        # when the regular obs would be cancelled (which is not allowed because it requires higher trigger priority).
        scheduling_unit_draft_trigger = self.create_simple_observation_scheduling_unit(
                                    "scheduling unit triggered",
                                    scheduling_set=self.scheduling_set_high_trigger_priority,
                                    interrupts_telescope=True)
        triggered_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft_trigger)
        triggered_scheduling_unit_blueprint.scheduling_constraints_doc = {'scheduler': 'dynamic', 'time': {'between': [{"from": datetime.utcnow().isoformat(), "to": (datetime.utcnow()+timedelta(hours=24)).isoformat()},]}}
        triggered_scheduling_unit_blueprint.save()

        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()

        regular_scheduling_unit_blueprint_low.refresh_from_db()
        regular_scheduling_unit_blueprint_high1.refresh_from_db()
        regular_scheduling_unit_blueprint_high2.refresh_from_db()
        triggered_scheduling_unit_blueprint.refresh_from_db()
        cancel_mock.assert_not_called()
        subtask.refresh_from_db()

        # assert that the subtask has NOT been cancelled and is still in state 'started', and its SU is observing
        self.assertEqual(subtask.state.value, 'started')
        self.assertEqual(regular_scheduling_unit_blueprint_high1.status.value, 'observing')

        # Assert that the new triggered scheduling_unit was scheduled, (but a bit later, after the observing one)
        self.assertEqual(triggered_scheduling_unit_blueprint.status.value, 'scheduled')
        self.assertGreater(triggered_scheduling_unit_blueprint.scheduled_start_time, subtask.on_sky_stop_time)
        # .. and assert regular observations remain scheduled
        self.assertEqual(regular_scheduling_unit_blueprint_high2.status.value, 'scheduled')
        self.assertEqual(regular_scheduling_unit_blueprint_low.status.value, 'scheduled')


    @mock.patch("lofar.sas.tmss.tmss.tmssapp.subtasks.kill_observation_subtask")
    def test_TMSS_2733_bugfix_do_cancel_calibrators_in_regular_unit(self, kill_observation_subtask_mock):
        scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data())
        scheduling_set.project.trigger_priority = 0
        scheduling_set.project.project_state = models.ProjectState.objects.get(value=models.ProjectState.Choices.ACTIVE.value)
        scheduling_set.project.cycles.add(models.Cycle.objects.create(**Cycle_test_data(start=datetime.utcnow()-timedelta(days=1),
                                                                                        stop=datetime.utcnow()+timedelta(days=1))))

        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="IM HBA - 1 Beam")
        regular_su_draft = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template, scheduling_set=scheduling_set)
        regular_su_draft.scheduling_constraints_doc['scheduler'] = 'fixed_time'
        regular_su_draft.scheduling_constraints_doc['time']['at'] = round_to_second_precision(datetime.utcnow() + timedelta(minutes=3)).isoformat()
        regular_su_draft.scheduling_constraints_doc['sky']['transit_offset'] = {'from': -12 * 60 * 60, 'to': 12 * 60 * 60}
        regular_su_draft.scheduling_constraints_doc['sky']['min_elevation'] = {'target': 0, 'calibrator': 0}
        regular_su_draft.scheduling_constraints_doc['sky']['min_distance'] = {'sun': 0, 'moon': 0, 'jupiter': 0}
        regular_su_draft.save()
        regular_su_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(regular_su_draft)

        scheduled_scheduling_units = self.scheduler.schedule_fixed_time_scheduling_units()
        self.assertEqual(1, len(scheduled_scheduling_units))
        self.assertEqual(regular_su_blueprint.id, scheduled_scheduling_units[0].id)

        # put first obs to started state
        first_obs_subtask = regular_su_blueprint.observation_tasks.order_by('scheduled_start_time').first().subtasks.first()
        set_subtask_state_following_allowed_transitions(first_obs_subtask, 'started')

        # create a triggered observation, overlapping with the first regular observation
        scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data())
        scheduling_set.project.can_trigger = True
        scheduling_set.project.trigger_priority = 10000
        scheduling_set.project.project_state = models.ProjectState.objects.get(value=models.ProjectState.Choices.ACTIVE.value)
        models.ProjectQuota.objects.create(project=scheduling_set.project.project, value=1000, resource_type=models.ResourceType.objects.get(quantity__value=models.Quantity.Choices.NUMBER.value, name__icontains='trigger'))
        scheduling_set.project.cycles.add(models.Cycle.objects.create(**Cycle_test_data(start=datetime.utcnow()-timedelta(days=1),
                                                                                        stop=datetime.utcnow()+timedelta(days=1))))
        scheduling_unit_draft_trigger = self.create_simple_observation_scheduling_unit("triggered scheduling unit for %s" % self._testMethodName,
                                                                                       obs_duration=60,
                                                                                       scheduling_set=scheduling_set,
                                                                                       interrupts_telescope=True)
        triggered_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft_trigger)
        triggered_scheduling_unit_blueprint.scheduling_constraints_doc = {'scheduler': 'dynamic', 'time': {'between': [{"from": first_obs_subtask.scheduled_start_time.isoformat(), "to": first_obs_subtask.scheduled_stop_time.isoformat()},]}}
        triggered_scheduling_unit_blueprint.save()

        # schedule the triggered observations
        scheduled_scheduling_units = self.scheduler.do_dynamic_schedule()
        self.assertEqual(1, len(scheduled_scheduling_units))
        self.assertEqual(triggered_scheduling_unit_blueprint.id, scheduled_scheduling_units[0].id)

        # the regular observations should have been cancelled.
        kill_observation_subtask_mock.assert_called()
        first_obs_subtask.refresh_from_db()
        self.assertEqual(models.SubtaskState.Choices.CANCELLED.value, first_obs_subtask.state.value)

        # and the rest of the tasks (observations & pipelines & etc) in the regular unit should be cancelled as well
        for task in regular_su_blueprint.tasks.all():
            task.refresh_from_db()
            self.assertEqual(models.TaskStatus.Choices.CANCELLED.value, task.status.value)

    @mock.patch("lofar.sas.tmss.tmss.tmssapp.subtasks.kill_observation_subtask")
    def test_TMSS_2881_bugfix_unschedulable_trigger_should_not_cancel_running_observations(self, kill_observation_subtask_mock):
        scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data())
        scheduling_set.project.trigger_priority = 0
        scheduling_set.project.project_state = models.ProjectState.objects.get(value=models.ProjectState.Choices.ACTIVE.value)
        scheduling_set.project.cycles.add(models.Cycle.objects.create(**Cycle_test_data(start=datetime.utcnow()-timedelta(days=1),
                                                                                        stop=datetime.utcnow()+timedelta(days=1))))

        strategy_template = models.SchedulingUnitObservingStrategyTemplate.get_latest(name="Simple Observation")
        regular_su_draft_past = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template, scheduling_set=scheduling_set)
        regular_su_blueprint_past = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(regular_su_draft_past)
        regular_su_blueprint_past = set_scheduling_unit_blueprint_start_times(regular_su_blueprint_past, round_to_second_precision(datetime.utcnow() - timedelta(days=1)))
        regular_obs_subtask_past = regular_su_blueprint_past.observation_tasks.first().subtasks.first()
        set_subtask_state_following_allowed_transitions(regular_obs_subtask_past, 'finished')

        regular_su_draft_running = create_scheduling_unit_draft_from_observing_strategy_template(strategy_template, scheduling_set=scheduling_set)
        regular_su_draft_running.scheduling_constraints_doc['scheduler'] = 'fixed_time'
        regular_su_draft_running.scheduling_constraints_doc['time']['at'] = round_to_second_precision(datetime.utcnow() + timedelta(minutes=3)).isoformat()
        regular_su_draft_running.scheduling_constraints_doc['sky']['transit_offset'] = {'from': -12 * 60 * 60, 'to': 12 * 60 * 60}
        regular_su_draft_running.scheduling_constraints_doc['sky']['min_elevation'] = {'target': 0, 'calibrator': 0}
        regular_su_draft_running.scheduling_constraints_doc['sky']['min_distance'] = {'target': {'sun': 0, 'moon': 0, 'jupiter': 0}}
        regular_su_draft_running.save()
        regular_su_blueprint_running = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(regular_su_draft_running)

        scheduled_scheduling_units = self.scheduler.schedule_fixed_time_scheduling_units()
        self.assertEqual(1, len(scheduled_scheduling_units))
        self.assertEqual(regular_su_blueprint_running.id, scheduled_scheduling_units[0].id)

        # put first obs to started state
        first_obs_subtask = regular_su_blueprint_running.observation_tasks.first().subtasks.first()
        set_subtask_state_following_allowed_transitions(first_obs_subtask, 'started')

        # create a triggered observation, erroneously in the past (which was the root cause of the bug, making it unschedulable)
        scheduling_set = models.SchedulingSet.objects.create(**SchedulingSet_test_data())
        scheduling_set.project.can_trigger = True
        scheduling_set.project.trigger_priority = 10000
        scheduling_set.project.project_state = models.ProjectState.objects.get(value=models.ProjectState.Choices.ACTIVE.value)
        models.ProjectQuota.objects.create(project=scheduling_set.project.project, value=1000, resource_type=models.ResourceType.objects.get(quantity__value=models.Quantity.Choices.NUMBER.value, name__icontains='trigger'))
        scheduling_set.project.cycles.add(models.Cycle.objects.create(**Cycle_test_data(start=datetime.utcnow()-timedelta(days=3),
                                                                                        stop=datetime.utcnow()+timedelta(days=1))))
        scheduling_unit_draft_trigger = self.create_simple_observation_scheduling_unit("triggered scheduling unit for %s" % self._testMethodName,
                                                                                       obs_duration=60,
                                                                                       scheduling_set=scheduling_set,
                                                                                       interrupts_telescope=True)
        triggered_scheduling_unit_blueprint = create_scheduling_unit_blueprint_and_tasks_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft_trigger)
        triggered_scheduling_unit_blueprint.scheduling_constraints_doc = {'scheduler': 'fixed_time',
                                                                          'time': {"at": regular_su_blueprint_past.scheduled_start_time.isoformat()} }
        triggered_scheduling_unit_blueprint.save()

        # schedule the triggered observations
        scheduled_scheduling_units = self.scheduler.schedule_fixed_time_scheduling_units()
        self.assertEqual(0, len(scheduled_scheduling_units))

        triggered_scheduling_unit_blueprint.refresh_from_db()
        self.assertEqual(models.SchedulingUnitStatus.Choices.UNSCHEDULABLE.value, triggered_scheduling_unit_blueprint.status.value)

        # the regular observations should NOT have been cancelled. (in the bug report, the running obs was killed)
        kill_observation_subtask_mock.assert_not_called()





logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

# # uncomment to show debug messages for scheduling modules
# logging.getLogger('lofar.sas.tmss.services.scheduling').level = logging.DEBUG
# logging.getLogger('lofar.sas.tmss.tmss.tmssapp.subtasks').level = logging.DEBUG

# hide spam of postgreslistener messages
logging.getLogger('lofar.sas.tmss.services.postgres_listener').disabled = True

if __name__ == '__main__':
    #run the unit tests
    unittest.main()

